#include "all.h"
#include <process.h>
#include <conio.h>
#include <io.h>
#include <direct.h>
#include <advanced.h>
#include <dta_read.h>
#include <ctype.h>
#include <windows.h>
#include <list>

//----------------------------

enum E_KEYWORD{
   KW_TEMP_DIR,
   KW_VSS_GET,
   KW_SND_COMPRESS,
   KW_DTA_MAKE,
   KW_SYNCHRONIZE,
   KW_VERBOSE,
   KW_WRITEABLE,
   KW_LAST
};

static const char *keywords[] = {
   "tmp",
   "vss",
   "snd",
   "dta",
   "syn",
   "v",
   "w",
   NULL
};

static E_KEYWORD FindKeyword(const char *cp){

   for(int i=0; keywords[i]; i++){
      if(!strcmp(keywords[i], cp))
         return (E_KEYWORD)i;
   }
   return KW_LAST;
}

//----------------------------

static bool WantEscape(){

   if(_kbhit() && _getch()==0x1b){
      cout<<"Are you sure to interrupt this operation? ";
      while(true){
         switch(_getch()){
         case 'y': case 'Y':
            return true;
         case 'n': case 'N':
            return false;
         }
      }
   }
   return false;
}

//----------------------------
static const char endl_1[] = "\\\n", endl_2[] = "..\n";

static bool MakeCRCOfDir(const char *base_dir, const char *base_mask, vector<char> &crc_buf){

   crc_buf.insert(crc_buf.end(), base_dir, base_dir + strlen(base_dir));
   crc_buf.insert(crc_buf.end(), endl_1, endl_1+2);

   WIN32_FIND_DATA fd;
   HANDLE h = FindFirstFile(C_fstring("%s\\%s", base_dir, base_mask), &fd);
   if(h!=INVALID_HANDLE_VALUE){
      do{
         if(fd.dwFileAttributes&_A_HIDDEN)
            continue;
         if(fd.dwFileAttributes&_A_SUBDIR){
                              //recurse sub-directories
            if(fd.cFileName[0]!='.'){
               if(!MakeCRCOfDir(C_fstring("%s\\%s", base_dir, fd.cFileName), base_mask, crc_buf)){
                  FindClose(h);
                  return false;
               }
            }
         }else{
                              //ignore *.scc files
            char *name = fd.cFileName;
            int name_size = strlen(name);
            for(int i=name_size; i--; )
               name[i] = tolower(name[i]);

            if(name_size>=4 && !strcmp(&name[name_size-4], ".scc"))
               continue;
                              //make CRC of the file
            crc_buf.insert(crc_buf.end(), name, name + name_size);
            crc_buf.push_back(' ');
            C_fstring file_time("%x,%x", fd.ftLastWriteTime.dwLowDateTime, fd.ftLastWriteTime.dwHighDateTime);
            crc_buf.insert(crc_buf.end(), file_time, file_time+file_time.Size());
            crc_buf.push_back('\n');

            if(WantEscape()){
               FindClose(h);
               return false;
            }
         }
      }while(FindNextFile(h, &fd));
      FindClose(h);
   }
   crc_buf.insert(crc_buf.end(), endl_2, endl_2+3);
   return true;
}

//----------------------------

static bool MakeCRC(const char *work_dir, const char *files, vector<char> &crc_buf){

   while(*files){
      char fname[256];
      int count;
      if(sscanf(files, "%s%n", fname, &count)!=1)
         return true;

      C_string base_mask;
                              //make crc of specified file
      for(int i=strlen(fname); i--; ){
         if(fname[i]=='\\'){
            base_mask = &fname[i+1];
            fname[i] = 0;
            break;
         }
      }
      files += count;
      if(!MakeCRCOfDir(C_fstring("%s\\%s", work_dir, fname), base_mask, crc_buf))
         return false;
   }
   return true;
};

//----------------------------

static bool CreateDirectoryTree(const char *dst){

   SECURITY_ATTRIBUTES sa;
   memset(&sa, 0, sizeof(sa));
   sa.nLength = sizeof(sa);
   int cpos, npos = 0;
   C_string str = dst;
   str.ToLower();
   while(true){
      cpos = npos;
      while(str[++npos]) if(str[npos]=='\\') break;
      if(cpos){
         str[cpos] = 0;
         bool ok =
            CreateDirectory(str, &sa);
         if(!ok){
            dword ec = GetLastError();
            if(ec!=ERROR_ALREADY_EXISTS)
               return false;
         }
         str[cpos] = '\\';
      }
      if(!str[npos])
         break;
   }
   return true;
}

//----------------------------

struct S_source_file{
   C_string source_dir;
   C_string file;
   FILETIME ftime[1];
   dword size;
   dword attr;
   S_source_file(){}
   S_source_file(const S_source_file &sf){ operator =(sf); }
   S_source_file &operator =(const S_source_file &sf){
      source_dir = sf.source_dir;
      file = sf.file;
      attr = sf.attr;
      size = sf.size;
      memcpy(ftime, sf.ftime, sizeof(ftime));
      return *this;
   }
};

//----------------------------
typedef list<S_source_file> t_source_files;

static bool AddMatchingFiles(const C_string &source_dir, const char *fmask, t_source_files &files){

   C_string path_base = fmask;
   for(int i=path_base.Size(); i--; ){
      if(path_base[i]=='\\'){
         path_base[i+1] = 0;
         path_base = (const char*)path_base;
         break;
      }
   }
   if(i==-1){
      path_base = NULL;
   }
   WIN32_FIND_DATA fd;
   HANDLE h = FindFirstFile(C_fstring("%s\\%s", (const char*)source_dir, fmask), &fd);
   if(h==INVALID_HANDLE_VALUE)
      return false;
   do{
      if(fd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)
         continue;
      files.push_back(S_source_file());
      S_source_file &f = files.back();
      f.file = path_base + fd.cFileName;
      f.file.ToLower();
      f.source_dir = source_dir;
      f.size = fd.nFileSizeLow;
      //memcpy(&f.ftime[0], &fd.ftCreationTime, sizeof(FILETIME));
      //memcpy(&f.ftime[1], &fd.ftLastAccessTime, sizeof(FILETIME));
      memcpy(&f.ftime[0], &fd.ftLastWriteTime, sizeof(FILETIME));
      f.attr = fd.dwFileAttributes & (FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_SYSTEM|FILE_ATTRIBUTE_NORMAL);
   }while(FindNextFile(h, &fd));
   FindClose(h);
   return true;
}

//----------------------------

static void CleanDirectory(const C_string &work_dir, const C_string &sub_dir, t_source_files &files,
   vector<C_string> &erase_files){

   C_string path = work_dir;
   if(sub_dir.Size()){
      path += "\\";
      path += sub_dir;
   }
   path += "\\*.*";
   WIN32_FIND_DATA fd;
   HANDLE h = FindFirstFile(path, &fd);
   if(h==INVALID_HANDLE_VALUE)
      return;
   do{
      if(fd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY){
         if(fd.cFileName[0] != '.'){
                              //recursive
            C_string dir;
            if(sub_dir.Size())
               dir = sub_dir + "\\";
            dir += fd.cFileName;
            CleanDirectory(work_dir, dir, files, erase_files);
         }
         continue;
      }
      C_string final_name = sub_dir;
      if(final_name.Size())
         final_name += "\\";
      final_name += fd.cFileName;
      final_name.ToLower();
                              //file found, check if exists and attributes match
      bool found = false;
      t_source_files::iterator it;
      for(it=files.begin(); it!=files.end(); it++){
         const S_source_file &f = (*it);
         if(f.file==final_name){
            found = true;
            if(!memcmp(&f.ftime[0], &fd.ftLastWriteTime, sizeof(FILETIME)) &&
               //!memcmp(&f.ftime[1], &fd.ftLastAccessTime, sizeof(FILETIME)) &&
               //!memcmp(&f.ftime[2], &fd.ftLastWriteTime, sizeof(FILETIME)) &&
               f.size==fd.nFileSizeLow &&
               f.attr==(fd.dwFileAttributes&(FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_SYSTEM))){

                              //files match, remove from copy list
               files.erase(it);
            }
            break;
         }
      }
      if(!found){
                              //redundant file, save
         erase_files.push_back(final_name);
      }
   }while(FindNextFile(h, &fd));
   FindClose(h);
}

//----------------------------
// This function cleans empty directories in specified tree.
// It returns number of files in direcotry.
static int CleanEmptyDirectories(const C_string &work_dir, int &num_rem){

   WIN32_FIND_DATA fd;
   HANDLE h = FindFirstFile(work_dir + "\\*.*", &fd);
   if(h==INVALID_HANDLE_VALUE)
      return 0;
   int num_files = 0;
   do{
      if(fd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY){
         if(fd.cFileName[0] != '.'){
            C_string dir = work_dir + "\\" + fd.cFileName;
            num_files = CleanEmptyDirectories(dir, num_rem);
            if(!num_files){
               bool ok = RemoveDirectory(dir);
               if(ok)
                  ++num_rem;
            }
         }
      }else{
         ++num_files;
      }
   }while(FindNextFile(h, &fd));
   FindClose(h);
   return num_files;
}

//----------------------------

static bool SyncDirs(ifstream &is_sync){

   C_string target_dir;
   C_string err;
   t_source_files source_files;
   
   cout<<"Synchronizing...\n";
   cout<<"Searching files... ";
   while(!is_sync.eof()){
      char line[1024], *cp = line;
      is_sync.getline(line, sizeof(line));
      if(!line[0] || line[0]==';') continue;

      char kw[256];
      int nc;
      if(sscanf(cp, "%s%n", kw, &nc) != 1){
         err = C_fstring("Invalid line: '%s'.", cp);
         goto fail;
      }
      cp += nc;
      C_string str;
      str = kw;
      if(str.Match("TargetDir")){
         if(sscanf(cp, "%s%n", kw, &nc) != 1){
            err = "TargetDir: expecting parameter.";
            goto fail;
         }
         target_dir = kw;
      }else
      if(str.Match("SourceDir")){
         if(sscanf(cp, "%s%n", kw, &nc) != 1){
            err = "SourceDir: expecting parameter.";
            goto fail;
         }
         C_string source_dir = kw;
         cp += nc;
         if(sscanf(cp, "%s", kw) != 1){
            err = "SourceDir: expecting {.";
            goto fail;
         }
         while(!is_sync.eof()){
            is_sync.getline(line, sizeof(line));
            cp = line;
            while(isspace(*cp)) ++cp;
            if(!*cp || *cp==';') continue;
            if(cp[0]=='}' && cp[1]==0)
               break;
            for(int i = strlen(cp); i && isspace(cp[i-1]); i--);
            if(!AddMatchingFiles(source_dir, cp, source_files)){
               err = C_fstring("Invalid file: '%s'", cp);
               goto fail;
            }
         }
      }else{
         err = C_fstring("Invalid keyword: '%s'.", kw);
         goto fail;
      }
   }
   cout<<(const char*)C_fstring("%i found.\n", source_files.size());

   {
      cout<<"Searching working directory... ";
      vector<C_string> erase_files;
      CleanDirectory(target_dir, "", source_files, erase_files);
      cout<<"Done\n";
      if(erase_files.size()){
         cout<<(const char*)C_fstring("Sync found %i files for deleting in target directory. Continue?\n", erase_files.size());
         while(true){
            char c = tolower(_getch());
            if(c=='n')
               return false;
            if(c=='y')
               break;
         }
         for(int i=erase_files.size(); i--; ){
            C_string del_name = target_dir + "\\" + erase_files[i];
            SetFileAttributes(del_name, FILE_ATTRIBUTE_NORMAL);
            bool ok = DeleteFile(del_name);
            if(!ok){
               cout<<(const char*)C_fstring("Failed to delete file: %s\n", (const char*)del_name);
            }
         }
      }
   }
   {
      int num_rem = 0;
      CleanEmptyDirectories(target_dir, num_rem);
      if(num_rem)
         cout<<(const char*)C_fstring("%i empty directories erased.\n", num_rem);
   }
   cout<<(const char*)C_fstring("Copying %i files.\n", source_files.size());
   {
      t_source_files::const_iterator it;
      for(it=source_files.begin(); it!=source_files.end(); it++){
         const S_source_file &sf = (*it);
         cout<<(const char*)C_fstring("%s ... ", (const char*)sf.file);
         C_string src_dir = sf.source_dir + "\\" + sf.file;
         C_string dst_dir = target_dir + "\\" + sf.file;
         bool ok = CreateDirectoryTree(dst_dir);
         if(ok){
            SetFileAttributes(dst_dir, FILE_ATTRIBUTE_NORMAL);
            ok = CopyFile(src_dir, dst_dir, false);
         }
         if(!ok){
            cout<<"Failed!\n";
            return false;
         }
         cout<<"OK\n";
         if(WantEscape())
            return false;
      }
   }
   return true;
fail:
   cout<<endl <<(const char*)err <<endl;
   return false;
}

//----------------------------

#define CHECK_NEXT_PARAM if(++i==argc){ cout<<"Expecting parameter after command " <<cp <<endl; return 1; }

//----------------------------

int __cdecl main(int argc, char *argv[]){

   cout<<"BuildMaker version 1.1 (c) 2002 Lonely Cat Games\n\n";

   if(argc==1){
      cout<<
         "Usage:\n"
         "  bm.exe [switches]\n"
         "  -tmp <dir>     ... set temporary folder for files\n"
         "  -vss <project> <script> ... retrieve files from VSS database\n"
         "                    (script specify file with folder names)\n"
         "  -snd <script>  ... compress sounds\n"
         "                    (script specify file with folder names)\n"
         "  -dta <script>  ... create DTA files\n"
         "                    (script specify file with folder and file names)\n"
         "  -syn <script>  ... synchronize directory\n"
         "                    (script specify file with source files)\n"
         "  -v             ... verbose move (write progress\n"
         "  -w             ... make files write-able during getting from VSS\n"
         ;
      return 0;
   }

   C_string vss_project;
   C_string work_dir;

   ifstream is_vss_script;     //file with VSS script
   bool vss_get = false;

   ifstream is_snd_script;     //file with sound script
   bool snd_compr = false;

   ifstream is_dta_script;     //file with dta script
   bool dta_make = false;

   ifstream is_sync;           //file with synchronize script
   bool do_sync = false;

   bool verbose = false;

   bool writeable = false;

   int i;

                              //parse the command line
   for(i=1; i<argc; i++){
      const char *cp = argv[i];
      if(*cp=='-' || *cp=='/'){
         E_KEYWORD kw = FindKeyword(cp+1);
         switch(kw){
         case KW_TEMP_DIR:
            CHECK_NEXT_PARAM;
            work_dir = argv[i];
            break;

         case KW_VSS_GET:
            CHECK_NEXT_PARAM;
            vss_project = argv[i];
            CHECK_NEXT_PARAM;
            is_vss_script.open(argv[i]);
            if(!is_vss_script.good()){
               cout<<"Failed to open VSS script: " <<argv[i] <<endl;
               goto fail;
            }
            vss_get = true;
            break;

         case KW_SND_COMPRESS:
            CHECK_NEXT_PARAM;
            is_snd_script.open(argv[i]);
            if(!is_snd_script.good()){
               cout<<"Failed to open sound script: " <<argv[i] <<endl;
               goto fail;
            }
            snd_compr = true;
            break;

         case KW_DTA_MAKE:
            CHECK_NEXT_PARAM;
            is_dta_script.open(argv[i]);
            if(!is_dta_script.good()){
               cout<<"Failed to open dta script: " <<argv[i] <<endl;
               goto fail;
            }
            dta_make = true;
            break;

         case KW_SYNCHRONIZE:
            CHECK_NEXT_PARAM;
            is_sync.open(argv[i]);
            if(!is_sync.good()){
               cout<<"Failed to open sync script: " <<argv[i] <<endl;
               goto fail;
            }
            do_sync = true;
            break;

         case KW_VERBOSE:
            verbose = true;
            break;

         case KW_WRITEABLE:
            writeable = true;
            break;

         default:
            cout<<"Invalid switch: " <<cp <<endl;
            goto fail;
         }
      }else{
         cout<<"Invalid switch: " <<cp <<endl;
         goto fail;
      }
   }

                              //validate parameters
   if(!work_dir.Size()){
      cout<<"Temporary working folder not set!" <<endl;
      goto fail;
   }

                              //retrieve files from VSS
   if(vss_get){
      while(!is_vss_script.eof()){
         char line[1024];
         is_vss_script.getline(line, sizeof(line));
                              //skip comments and blank lines
         if(!line[0] || line[0]==';')
            continue;
         char src[256], dst[256];
         int num = sscanf(line, "%s %s", src, dst);
         if(!num)
            continue;
         if(num<2)
            strcpy(dst, src);
         C_fstring dest_dir("%s\\%s", (const char*)work_dir, dst);
                              //make sure directory exists
         if(!CreateDirectoryTree(dest_dir+"\\")){
            cout<<"Failed to create directory: \"" <<(const char*)dest_dir <<"\"" <<endl;
            goto fail;
         }
         cout<<(const char*)C_fstring("VSS Get directory \"%s\"... ", (const char*)dest_dir);
         C_fstring cmd("Get $\\%s\\%s -GL\"%s\" -O&- -I- -R -GCK %s",
            (const char*)vss_project,
            src,
            (const char*)dest_dir,
            writeable ? "-W" : "");
         E_VSS_COMMAND_RESULT res = VSSCommand(cmd);
         if(res!=VSS_OK){
            cout<<endl;
            switch(res){
            case VSS_NOT_INSTALLED:  cout<<"VSS error: not installed on the system\n"; break;
            case VSS_FAILED_TO_RUN:  cout<<"VSS error: failed to run\n"; break;
            default: cout<<"VSS unknown error\n";
            }
            goto fail;
         }
         cout<<"OK\n";
         if(WantEscape())
            goto fail;
      }
      is_vss_script.close();
   }

                              //compress sounds
   if(snd_compr){
      while(!is_snd_script.eof()){
         char line[1024];
         is_snd_script.getline(line, sizeof(line));
                              //skip comments and blank lines
         if(!line[0] || line[0]==';')
            continue;
         char src[256], dst[256];
         float quality = .5f;
         int num = sscanf(line, "%s %s %f", src, dst, &quality);
         if(num < 2){
            cout<<"Converting sounds - invalid line: " <<line <<endl;
            goto fail;
         }
         C_fstring src_dir("%s\\%s", (const char*)work_dir, src);
         C_fstring dst_dir("%s\\%s", (const char*)work_dir, dst);

         cout<<(const char*)C_fstring("Converting sound directory \"%s\" to \"%s\"...\n",
            (const char*)src_dir, (const char*)dst_dir);
         C_fstring cmd_line("wav_conv.exe %s\\*.wav %s -co%f -d -f22050 %s", (const char*)src_dir, (const char*)dst_dir,
            quality, verbose ? "" : "-q");
         int rtn = spawnl(P_WAIT, "Tools\\wav_conv.exe", cmd_line, NULL);
         if(rtn){
            switch(rtn){
            case -1:
               cout<<"Failed to run 'Tools\\wav_conv.exe'.\n";
               break;
            default:
               cout<<(const char*)C_fstring("Failed to convert sounds '%s'\n", (const char*)src_dir);
            }
            goto fail;
         }
         cout<<"OK\n";
         if(WantEscape())
            return 2;
      }
      is_snd_script.close();
   }

                              //make DTA files
   if(dta_make){
      while(!is_dta_script.eof()){
         char line[1024], *cp = line;
         is_dta_script.getline(line, sizeof(line));
                              //skip comments and blank lines
         if(!line[0] || line[0]==';')
            continue;
         char file[256];
         int nc;
         int num = sscanf(cp, "%s%n", file, &nc);
         if(num!=1){
            cout<<"Creating DTA files - invalid line: " <<line <<endl;
            goto fail;
         }
         cp += nc;
         cout<<(const char*)C_fstring("Creating DTA file \"%s\"... ", (const char*)file);
                              //check if file need to be created
         bool need_create = true;
         
                              //create CRC of the file
         vector<char> crc_buf;
         bool crc_valid = MakeCRC(work_dir, cp, crc_buf);
         C_fstring crc_file("%s\\%s", (const char*)work_dir, file);
         if(crc_valid){
                              //compare crc with what's on the disk
            if(crc_file.Size()>4 && crc_file[(int)crc_file.Size()-4]=='.')
               strcpy(&crc_file[(int)crc_file.Size()-4], ".crc");
            else
               crc_file += ".crc";
            int sz_on_disk = GetFileSize(crc_file);
            if(crc_buf.size()==sz_on_disk){
               vector<char> crc_buf_on_disk(sz_on_disk);
               Load(crc_file, crc_buf_on_disk.begin(), sz_on_disk);

               if(crc_buf==crc_buf_on_disk){
                              //check if dta file exists
                  ifstream is;
                  is.open(C_fstring("%s\\%s", (const char*)work_dir, file), ios::binary);
                  if(is.good()){
                     is.close();
                     need_create = false;
                  }
               }
            }
         }
         if(need_create){
            char curr_dir[256];
            _getcwd(curr_dir, sizeof(curr_dir));
            if(_chdir(work_dir)==-1){
               cout<<(const char*)C_fstring("Failed to change to directory '%s'.\n", (const char*)work_dir);
               goto fail;
            }
                              //make the dta file
            _unlink(file);

            CreateDirectoryTree(file);

                              //disable compression, recurse directories,
                              // convert to lower case, exclude VSS *.scc files
            C_fstring cmd_line("rar.exe -m0 -r -cl u -x*.scc %s %s %s", verbose ? "" : "-inul", file, cp);
            int rtn = spawnl(P_WAIT, C_fstring("%s\\Tools\\rar.exe", curr_dir), cmd_line, NULL);
            if(rtn){
               switch(rtn){
               case -1:
                  cout<<"Failed to run 'Tools\\rar.exe'.\n";
                  break;
               default:
                  cout<<(const char*)C_fstring("Failed to create DTA filr '%s'\n", file);
               }
               goto fail;
            }
            _chdir(curr_dir);
                              //save crc now
            if(crc_valid)
               Save(crc_file, crc_buf.begin(), crc_buf.size());

            cout<<"OK\n";
         }else{
            cout<<"SKIPPING (CRC matches)\n";
         }
                           //create CNT file
         LPC_dta_read dta = dtaCreate(C_fstring("%s\\%s", (const char*)work_dir, file));
         if(!dta){
            cout<<(const char*)C_fstring("Failed to read created DTA file '%s'.\n", file);
            goto fail;
         }
         dta->Release();

         if(WantEscape())
            return 2;
      }
   }

                              //synchronize dirs
   if(do_sync){
      bool ok = SyncDirs(is_sync);
      if(!ok)
         goto fail;
   }
   cout<<"Finished!\n";
   getch();

   return 0;
fail:
   getch();
   return 1;
}

//----------------------------
