#include "all.h"
#include <win_reg.h>
#include <windows.h>

//----------------------------

bool CreateDirectoryTree(const char *dst){

   SECURITY_ATTRIBUTES sa;
   memset(&sa, 0, sizeof(sa));
   sa.nLength = sizeof(sa);
   int cpos, npos = 0;
   C_string str = dst;
   while(true){
      cpos = npos;
      while(str[++npos]) if(str[npos]=='\\') break;
      if(cpos){
         str[cpos] = 0;
         //bool ok =
            CreateDirectory(str, &sa);
         //if(!ok) return false;
         str[cpos] = '\\';
      }
      if(!str[npos])
         break;
   }
   return true;
}

//----------------------------
// Get path to VSS executable and current database.
bool VSSGetPath(C_string &vss_exe, C_string &vss_database){

   char name[256];
                           //get path to source safe
   int hk = RegkeyOpen("Software\\Microsoft\\SourceSafe");
   if(hk==-1)
      return false;
   int len = RegkeyRdata(hk, "SCCServerPath", name, sizeof(name));
   if(len==-1){
      RegkeyClose(hk);
      return false;
   }
   while(len && name[len-1]!='\\') --len;
   strcpy(&name[len], "ss.exe");
   vss_exe = name;

   len = RegkeyRdata(hk, "Current Database", name, sizeof(name));
   RegkeyClose(hk);
   if(len==-1)
      return false;
   vss_database = name;

   return true;
}

//----------------------------

E_VSS_COMMAND_RESULT VSSCommand(const char *in_cmdline){

   C_string vss_exe, vss_data;
   if(!VSSGetPath(vss_exe, vss_data))
      return VSS_NOT_INSTALLED;

   C_fstring cmdline("\"%s\" %s", (const char*)vss_exe, (const char*)in_cmdline);

   PROCESS_INFORMATION pi;
   memset(&pi, 0, sizeof(pi));

   STARTUPINFO si;
   memset(&si, 0, sizeof(si));
   si.cb = sizeof(si);
   si.dwFlags = STARTF_USESHOWWINDOW;// | STARTF_USESTDHANDLES;
   //si.hStdOutput = h_out;
   //si.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
   //si.hStdError = h_out;
   si.wShowWindow = SW_HIDE;

   SECURITY_ATTRIBUTES lsa;
   lsa.nLength = sizeof(lsa);
   lsa.lpSecurityDescriptor = NULL;
   lsa.bInheritHandle = true;

   char env[1024];
   strcpy(env, "SSUSER=Guest\0");
   int len = strlen(env);
   //strcat(env, vss_data);
   strcpy(env+len+1, C_fstr("SSDIR=%s", (const char*)vss_data));
   len += strlen(env+len+1) + 1;
   env[len+1] = 0;
   bool ok = CreateProcess(vss_exe,
      (char*)(const char*)cmdline,
      &lsa,                //process security attributes
      &lsa,                //thread security attributes
      true,                //handle inheritance flag
      0,                   //creation flags
      env,
      NULL,                //pointer to current directory name
      &si,
      &pi);
   if(!ok)
      return VSS_FAILED_TO_RUN;

   dword wr = WaitForSingleObject(pi.hProcess, INFINITE);
   switch(wr){
   default:
      return VSS_FAILED_TO_RUN;
   case WAIT_OBJECT_0: break;
   }
   dword cr;
   bool b = GetExitCodeProcess(pi.hProcess, &cr);
   assert(b && cr!=STILL_ACTIVE);
   b = CloseHandle(pi.hProcess);
   assert(b);
   b = CloseHandle(pi.hThread);
   assert(b);
   return (cr==0) ? VSS_OK : VSS_RETURN_ERROR;
}

//----------------------------
