#include "all.h"
#include "tables.h"
#include "net.h"

//----------------------------
extern const char *actor_names[];
const char *actor_names[ACTOR_LAST] = {
   "NULL",
   "Player",
   "Enemy",
   "GunShoot",
   "Cannon",
   "Particle",
   "GunHole",
   "Version",
   "AutoCannon",
   "Grenade",
   "Mine",
   "Resource",
   "Door",
   "Flare",
   "Explosion",
   "Rocket",
   "Automobil",
   "RandomGen",
   "TxtShift",
   "TimeBomb",
   "Detector",
   "Ladder",
   "KillAnim",
   "SnowRainElem",
};

//----------------------------

bool I3DAPI FlyColRespGeo(I3D_collision_data &cd){

   int mat_id = -1;
   PI3D_frame frm = cd.GetHitFrm();
   assert(frm);
   switch(frm->GetType()){
   case FRAME_VOLUME:
      mat_id = I3DCAST_VOLUME(frm)->GetCollisionMaterial();
      break;
   case FRAME_VISUAL:
      mat_id = I3DCAST_VISUAL(frm)->GetCollisionMaterial();
      break;
   case FRAME_JOINT:
      break;
   default: 
      assert(0);
   }
   if(mat_id != -1){
      if(mat_table->ItemB(TAB_B_MAT_BORDER, mat_id) ||
         mat_table->ItemB(TAB_B_MAT_WATER, mat_id))
         return false;
                           //fly through humans volumes(react only on joints)
      if(mat_id == HUMAN_SPHERE_MAT)
         return false;

      if(mat_table->ItemB(TAB_B_MAT_AI_SEE_THROUGH, mat_id)){
                              //for volumes its some chance out shots don't get through
                              // (second number means percentage by which fence absorbs shots)
         if((frm->GetType() == FRAME_VOLUME) && (S_int_random(100) > 20)){
            //ErrReport("Fence NOT hit", editor);
            return false;
         }
         //ErrReport("Fence HIT", editor);
      }
   }

                              //try find actor which we hit
                              //18.12.Jirka:comment out. Think that this code was here only for 
                              //hit player in first person mode, which is done now by I3D_ON_CONDITIONAL flag.
   /*do{
      LPC_actor a = (LPC_actor)frm->GetData();
      if(a){
                              //don't hit invisible enemy
         if(a->GetType()==ACTOR_ENEMY && !a->GetFrame()->IsOn()){
            return false;
         }
         break;
      }
   }while(frm = frm->GetParent());*/
   return true;
}

/*
//----------------------------
//!SM_COLLISIONS - could be removed when sweap to sm
bool I3DAPI FlyColRespGeo(const S_vector &collision_normal, float collision_dist,
   PI3D_frame frm1, void *context, void *special){

                              //try to find out which volume we've hit
   PI3D_volume vol = NULL;
   const I3D_frame *frm = frm1;
   do{
      vol = I3DCAST_VOLUME(frm->FindChildFrame(NULL, ENUMF_VOLUME));
      if(vol){
         int mat_id = vol->GetCollisionMaterial();
         if(mat_id){
            if(mat_table->ItemB(TAB_B_MAT_BORDER, mat_id) ||
               mat_table->ItemB(TAB_B_MAT_WATER, mat_id))
               return false;
         }
         LPC_actor a = (LPC_actor)vol->GetData();
         if(a){
                              //don't hit invisible enemy
            if(a->GetType()==ACTOR_ENEMY && !a->GetFrame()->IsOn())
               return false;
         }
         break;
      }
   }while(frm = frm->GetParent());
   return true;
}
*/

//----------------------------

bool I3DAPI FlyColResp(I3D_collision_data &cd){

   assert(cd.GetHitFrm());
   int mat_id = 0;
   switch(cd.GetHitFrm()->GetType()){
   case FRAME_VOLUME:
      mat_id = I3DCAST_VOLUME(cd.GetHitFrm())->GetCollisionMaterial();
      break;
   case FRAME_VISUAL:
      mat_id = I3DCAST_VISUAL(cd.GetHitFrm())->GetCollisionMaterial();
      break;
   default: 
      break;
   }
   
   if(mat_table->ItemB(TAB_B_MAT_BORDER, mat_id) ||
      mat_table->ItemB(TAB_B_MAT_WATER, mat_id))
      return false;
   if(mat_table->ItemB(TAB_B_MAT_AI_SEE_THROUGH, mat_id)){
                           //its some chance out shots don't get through
                           // (second number means percentage by which fence absorbs shots)
      if(S_int_random(100) > 20){
         //ErrReport("Fence NOT hit", editor);
         return false;
      }
      //ErrReport("Fence HIT", editor);
   }
   return true;
}

//----------------------------
//----------------------------

static void ShiftTextureCoords(PI3D_mesh_object mo, int time, int angle1, float speed, float *whole_count){

   if(!mo)
      return;

   float tsec = ((float)time)*.001f;
   float add[2];
   const float &angle = ((float)angle1) * PI / 180.0f;
   float spds = speed*tsec;
   add[0] = cos(angle) * spds;
   add[1] = -sin(angle) * spds;

   int i;
   for(i=0; i<2; i++){
      whole_count[i] += add[i];
      if(fabs(whole_count[i]) >= 1.0f){
         float f = (int)whole_count[i];
         whole_count[i] -= f;
         add[i] -= f;
      }
   }

   I3D_vertex_mesh *vp;
   vp = mo->LockVertices(0, I3DLOCK_UV);
   for(i=mo->NumVertices(0); i--; ){
      vp[i].tex[0].u += add[0];
      vp[i].tex[0].v += add[1];
   }
   mo->UnlockVertices(0);
}

//----------------------------

#ifdef EDITOR

//----------------------------

enum{
   TAB_I_TXS_ANGLE,
   TAB_F_TXS_SPEED,
   TAB_F_TXS_AMPL,
};

C_table_template templ_txt_shift = {
   // {9427B800-8E29-11d2-9B9E-004F4905FF1A}
   { 0x9427b800, 0x8e29, 0x11d2, { 0x9b, 0x9e, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Texture shift",
   NULL,
   {
      {TE_INT, TAB_I_TXS_ANGLE, "Angle", 0, 359},
      {TE_FLOAT, TAB_F_TXS_SPEED, "Speed", 0, 0, 1},
      {TE_FLOAT, TAB_F_TXS_AMPL, "Amplitude"},
      {TE_NULL}
   }
};

//----------------------------

class C_texture_shift: public C_actor{

   LPC_table tab;
   C_smart_ptr<I3D_object> obj;

   float whole_count[2];
public:
   C_texture_shift(C_game_mission &m, dword):
      C_actor(m, ACTOR_TEXTURE_SHIFT)
   {
      tab = CreateTable();
      tab->Open((dword)&templ_txt_shift, TABOPEN_TEMPLATE);
      whole_count[0] = whole_count[1] = 0.0f;
   }
   ~C_texture_shift(){
      if(tab) tab->Release();
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      C_actor::SetFrame(frm);

      if(frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_OBJECT){
         obj = I3DCAST_OBJECT(frm);
      }else{
         obj = NULL;
      }
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      if(obj){
         ShiftTextureCoords(obj->GetMeshObject(), tc.time, tab->ItemI(TAB_I_TXS_ANGLE), tab->ItemF(TAB_F_TXS_SPEED), whole_count);
         obj->SetUpdate();
      }
      return 0;
   }

//----------------------------

   virtual LPC_table GetTable(int index) const{
      if(index==0)
         return tab;
      return C_actor::GetTable(index);
   }

   virtual class C_table_template *GetTemplate(int) const{ return &templ_txt_shift; }
   virtual void MissionSave(C_chunk *ck) const{ SaveTable(ck, tab); }
   virtual bool MissionLoad(C_chunk *ck, int table_id){ return LoadTable(ck, tab); }
};

#endif//EDITOR

//----------------------------
//----------------------------

class C_snow_rain_elem: public C_actor{
   float time_scale;
public:
   C_snow_rain_elem(C_game_mission &m):
      C_actor(m, ACTOR_NULL),
      time_scale(.7f + S_float_random(.6f))
   {
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      C_actor::SetFrame(frm);
      if(frm && frm->GetType()==FRAME_MODEL){
         PI3D_model m = I3DCAST_MODEL(frm);
         PI3D_animation_set as = m->GetAnimationSet();
         if(as)
            m->SetAnimation(0, as);
      }
   }

//----------------------------

   virtual bool MaySaveGame() const{ return false; }

//----------------------------

   dword Tick(const S_tick_context &tc){

      int time = (float)tc.time * time_scale;

      bool destroy = false;
      if(model){
         model->Tick(time);
         /*
         if(!model->NumInterpolators()){
            PI3D_interpolator intp = model->GetInterpolators()[0];
            if(intp->GetCurrTime()==intp->GetEndTime())
               destroy = true;
         }
         */ 
         destroy = (!model->NumInterpolators());
      }
      if(destroy){
         mission.DestroyActor(this);
      }
      return 0;
   }
};

//----------------------------
//----------------------------

LPC_actor CreateDetector(C_game_mission&);
LPC_actor CreateKillAnim(C_game_mission&);
LPC_actor CreateTimeBomb(C_game_mission&, void*);
LPC_actor CreateGunShoot(C_game_mission&, void*);
LPC_actor CreateRocket(C_game_mission&, void*);
LPC_actor CreateGrenade(C_game_mission&, void*);
LPC_actor CreateFlare(C_game_mission&, void*);
LPC_actor CreateMine(C_game_mission&, void*);
LPC_actor CreateResource(C_game_mission&, void*);
LPC_actor CreateExplosion(C_game_mission&, void*);
LPC_actor CreateRandomGen(C_game_mission&);
LPC_actor CreateAutomobil(C_game_mission&);
LPC_actor CreateAutoCannon(C_game_mission&);
LPC_actor CreateParticle(C_game_mission&);
LPC_actor CreateGunGole(C_game_mission&);
LPC_actor CreateDoor(C_game_mission&);
LPC_actor CreateVersion(C_game_mission&);
LPC_actor CreateCannon(C_game_mission&);
LPC_actor CreateLadder(C_game_mission&);

//----------------------------

LPC_actor C_game_mission::CreateActor1(int type, dword data){

   LPC_actor ap=NULL;
   switch(type){
   case ACTOR_GUN_SHOOT: ap = CreateGunShoot(*this, (void*)data); break;
   case ACTOR_TIME_BOMB: ap = CreateTimeBomb(*this, (void*)data); break;
#ifdef EDITOR
   case ACTOR_TEXTURE_SHIFT: ap = new C_texture_shift(*this, data); break;
#endif
   case ACTOR_RANDOM_GEN: ap = CreateRandomGen(*this); break;
   case ACTOR_ROCKET: ap = CreateRocket(*this, (void*)data); break;
   case ACTOR_GRENADE: ap = CreateGrenade(*this, (void*)data); break;
   case ACTOR_MINE: ap = CreateMine(*this, (void*)data); break;
   case ACTOR_RESOURCE: ap = CreateResource(*this, (void*)data); break;
   case ACTOR_DOOR: ap = CreateDoor(*this); break;
   case ACTOR_FLARE: ap = CreateFlare(*this, (void*)data); break;
   case ACTOR_EXPLOSION: ap = CreateExplosion(*this, (void*)data); break;
   case ACTOR_PARTICLE: ap = CreateParticle(*this); break;
   case ACTOR_GUNHOLE: ap = CreateGunGole(*this); break;
   case ACTOR_CANNON: ap = CreateCannon(*this); break;
   case ACTOR_AUTOMOBIL: ap = CreateAutomobil(*this); break;
   case ACTOR_AUTO_CANNON: ap = CreateAutoCannon(*this); break;
   case ACTOR_VERSION: ap = CreateVersion(*this); break;
   case ACTOR_DETECTOR: ap = CreateDetector(*this); break;
   case ACTOR_LADDER: ap = CreateLadder(*this); break;
   case ACTOR_KILL_ANIM: ap = CreateKillAnim(*this); break;
   case ACTOR_SNOWRAIN_ELEM: ap = new C_snow_rain_elem(*this); break;
   }
   return ap;
}

//----------------------------
