#include "all.h"
#include "net.h"
#include <insanity\controller.h>
#include <isound2.h>
#include "loader.h"
#include "mainmenu.h"

//----------------------------
                              //helper functions
static int GetKeyWord(char* &line, const char *kwords){
   int i=0;
   while(kwords[0]){
      int linelen=strlen(kwords);
      if(!strnicmp(line, kwords, linelen)){
         line += linelen;
         return i;
      }
      ++i;
      kwords += strlen(kwords)+1;
   }
   return -1;
}

//----------------------------

static bool ReadInt(char* &line, int &i1){
   int i;
   if(sscanf(line, "%i%n", &i1, &i) == 1){
      line += i;
      return true;
   }
   return false;
}

//----------------------------

int ReadString(char* &line, char *wrd){
   int i=0;
   while(line[0] && isspace(line[0])) line++;
   if(*line=='"'){
      ++line;
      while(*line && *line!='"') wrd[i++]=*line++;
      ++line;
   }else
      while(line[0] && !isspace(line[0])) wrd[i++]=*line++;
   wrd[i++]=0;
   return i;
}

//----------------------------

int ReadFloat(char* &line, float &f){
   int i;
   sscanf(line, "%f%n", &f, &i);
   line += i;
   return i;
}

//----------------------------
                              //briefing implementation


#define BRIEF_TEXT_SIZE    .0375f
#define BRIEF_TEXT_TIME    1000
#define BRIEF_TEXT_COME    50 //time displaying single letter of text
#define BRIEF_TEXT_COLOR   0xffffff40
#define BRIEF_TEXT_FADE    400

class C_briefing: public C_mission{

   struct S_key{
      int id, key_time;
      C_string wav, txt;
      bool wide;
   };
   vector<S_key> keys;
   PISND_source snd_src;
   PISND_sound snd;
   C_poly_text texts;
   
   C_smart_ptr<C_text> text;
   int text_len;
   int countdown, text_come_cnt;
   C_fade<S_fade_color> fade_text;

   C_smart_ptr<C_class_to_be_ticked_hd> save_tc;

   int end_count;

   void MakeText(const C_string &str, bool wide){

      int i;

      float fmt_rect[4] = {.2f, 0, .8f, 0};
      text = texts.CreateText(str, BRIEF_TEXT_COLOR, .5f, .075f, BRIEF_TEXT_SIZE, TEXT_JUSTIFY_CENTER,
         wide, fmt_rect, NULL, NULL, NULL, 0,
         PI*.1f);
      text->Release();
      text_len = str.Size();
      if(wide)
         text_len = text_len*3/2;

      if(!wide){
         C_sprite_group *pg = text->GetAnsiText();
         const vector<C_smart_ptr<C_poly_sprite> > &sprs = pg->GetSPrites();
         for(i=0; i<sprs.size(); i++)
            sprs[i]->SetOn(false);
      }
      text_come_cnt = 1;
      fade_text.Stop();
   }

//----------------------------

   void FadeText(){
      if(text)
         fade_text.Setup(BRIEF_TEXT_FADE, BRIEF_TEXT_COLOR, BRIEF_TEXT_COLOR&0xffffff);
   }

//----------------------------

   static dword I3DAPI cbCBox(PI3D_frame frm, FRM_CALLBACK_MESSAGE msg, dword prm1, dword prm2){

      int i;
      C_briefing *brf = (C_briefing*)frm->GetData();
      switch(msg){
      case CBM_ANIMNOTE:
         {
                              //convert string to number
            const I3D_note_callback *nc = (I3D_note_callback*)prm1;
            prm2 = nc->note_time;
            prm1 = atoi(nc->note);
                              //decode type of message
            switch(prm1){
            case 99:          //suspend
               brf->FadeText();
               break;
            default:          //play sound/show message
               for(i=brf->keys.size(); i--; ) if(brf->keys[i].id==prm1) break;
               if(i!=-1){
                  S_key &key = brf->keys[i];
                  if(key.wav.Size() && brf->snd_src){
                     brf->snd->Stop();
                     C_string snd_name;
                     I3D_RESULT ir;
                     for(int i=sound_cache.NumDirs(); i--; ){
                        const char *snd_dir = sound_cache.GetDir(i);
                        snd_name = C_fstring("%sbriefing\\%s", snd_dir, (const char*)key.wav);

                              //strip extension
                        if(snd_name.Size()>=4 && snd_name[(int)snd_name.Size()-4]=='.'){
                           snd_name[(int)snd_name.Size()-4] = 0;
                        }
                        dword open_flags =
                           ISND_NOVOLUME | ISND_NOPANNING;
                        if(allow_stream_snd)
                           open_flags |= ISND_STREAMING;
                        ir = brf->snd_src->Open(C_fstring("%s.ogg", (const char*)snd_name), open_flags);
                        if(ISND_FAIL(ir))
                           ir = brf->snd_src->Open(C_fstring("%s.wav", (const char*)snd_name), open_flags);
                        if(ISND_SUCCESS(ir))
                           break;
                     }
                     if(ISND_FAIL(ir)){
                        ErrReport(C_fstring("can't load sound: %s", (const char*)snd_name), editor);
                     }else{
                        brf->snd->Open(brf->snd_src, ISND_SYSMEM);
                        brf->snd->Play();
                     }
                  }
                  if(key.txt.Size()){
                     brf->text = NULL;
                     brf->MakeText(key.txt, key.wide);
                  }
                  int &cnt = brf->countdown;
                  cnt = key.key_time;
                  if(cnt){
                     cnt = 100;
                              //don't skip sample
                     if(brf->snd_src && brf->snd->IsPlaying()){
                        int sam_left = brf->snd_src->GetPlayTime() - brf->snd->GetCurrTime();
                        cnt = Max(cnt, sam_left+500);
                     }
                              //don't skip text
                     if(brf->text && brf->text_come_cnt){
                        int text_left_time = brf->text_len * BRIEF_TEXT_COME - brf->text_come_cnt;
                        cnt = Max(cnt, text_left_time+500);
                     }
                  }
               }else{
                  ErrReport(C_fstring("undefined briefing key: %i", prm1), editor);
               }
            }
         }
         break;

      case CBM_ANIM_FINISH:
         brf->end_count = Max(2000, brf->countdown+500);
#if 0
         OUTTEXT(C_fstring("briefing - animation finshed, setting countdown to %i ms", brf->end_count));
#endif
         break;
      }
      return 0;
   }

   enum{
      NS_NULL,
      NS_ENTER,
      NS_LEAVE,
   };

   C_smart_ptr<C_net_sync> net_sync;
   C_smart_ptr<C_sync_menu> sync_menu;

//----------------------------

   void Finish(){

      if(snd)
         snd->Stop();
      if(!net){
         if(save_tc)
            tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(save_tc);
         else{
            game_music.FadeVolume(1.0f);
            RenderAllBackbuffers();
            tick_class = CreateMenu(GAME_MENU_4_OF_8);
            if(tick_class)
               tick_class->Release();
         }
         return;
      }
      net_sync->BeginEvent(NS_LEAVE);
      NetSend();
   }

   C_smart_ptr<I3D_frame> cbox;
public:
   C_briefing():
      countdown(0),
      snd_src(NULL),
      snd(NULL),
      text(NULL),
      end_count(0)
   {
      net_sync = CreateNetSync(NM_BRIEFING);
      net_sync->Release();
   }

   bool Init(const char *dir, C_class_to_be_ticked_hd *stc){

      save_tc = stc;

      E_MISSION_IO mio = Open(C_fstring("%s\\briefing", dir), OPEN_USE_SCENE_CAMERA | OPEN_NO_OPEN_FAIL_MSG);
      if(mio!=MIO_OK)
         return false;

      PI3D_camera cam = scene->GetActiveCamera();
      if(!cam)
         return false;
                              //read ini file
      C_cache is;
      if(!is.open(C_fstring("missions\\%s\\briefing\\brief.ini", dir), CACHE_READ))
         return false;
      char line[256], *cp;
      int line_num=0;
      if(is.good()){
         while(!is.eof()){
            is.getline(line, sizeof(line));
            ++line_num;
            if(line[0]==';')
               continue;
            cp = line;

            int time;
            int id;
            if(ReadInt(cp, id)){
               if(id == 999){
                              //near clipping plane
                  float ncp;
                  if(ReadFloat(cp, ncp)){
                     float nr, fr;
                     cam->GetRange(nr, fr);
                     cam->SetRange(ncp, fr);
                  }
                  continue;
               }

               if(!ReadInt(cp, time)) continue;
               char wav[128];
               wav[0]=0;//, txt[0]=0;
               const C_string *txt = NULL;
               bool wide;
               while(*cp){
                  while(*cp) if(!isspace(*cp)) break; else cp++;
                  switch(*cp++){
                  case 's':      //sound
                     ReadString(cp, wav);
                     break;
                  case 't':      //text
                     {
                        int tid;
                        bool b = ReadInt(cp, tid);
                        if(!b){
                           ErrReport(C_fstring("Invalid text in brief.ini, line %i", line_num), editor);
                        }else{
                           if(game_profile.IsDatadisk())
                              tid += 1e6;
                           txt = &all_txt[tid];
                           wide = all_txt.IsWide(tid);
                        }
                     }
                     break;
                  }
               }
               keys.push_back(S_key());
               S_key &key = keys.back();
               key.key_time = time;
               key.id = id;
               if(wav[0]) key.wav = wav;
               if(txt){
                  key.txt = *txt;
                  key.wide = wide;
               }
            }
         }
         is.close();
      }
      texts.Init();
      scene->Tick(0);
                              //create sample for voice playback
      if(isound){
         isound->CreateSource(&snd_src);
         isound->CreateSound(&snd);
      }
                              //modify interpolators to simple
                              // also find camerabox interpolator
      const PI3D_interpolator *intps = scene->GetInterpolators();
      for(int i=scene->NumInterpolators(); i--; ){
         PI3D_interpolator intp = intps[i];
         intp->SetAnimation(0, intp->GetAnimation(0));
         if(!cbox && C_string(intp->GetFrame()->GetName())=="camerabox"){
            cbox = intp->GetFrame();
            cbox->SetCallback(cbCBox, FRMCALLBACK_ANIMNOTE | FRMCALLBACK_ANIMFINISH);
            cbox->SetData((dword)this);
         }
      }

      if(net){
         net_sync->BeginEvent(NS_ENTER);
         NetSend();
      }
      return true;
   }

//----------------------------

   ~C_briefing(){

      if(cbox)
         cbox->SetData(0);
      if(snd_src){
         snd_src->Release();
         snd->Release();
      }
   }

//----------------------------

   void Tick(const S_tick_context &tc){


      if(net){
         dword event;
         switch(net_sync->GetState(NULL, &event)){
         case C_net_sync::SS_RESOLVED:
            {
               if(sync_menu){
                  sync_menu->AddAllCheckMarks();
                  RenderAllBackbuffers();
                  sync_menu = NULL;
               }
               switch(event){
               case NS_ENTER:
                  net_sync->Reset();
                  break;
               case NS_LEAVE:
                  {
                     NetSend();
                     if(save_tc)
                        tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(save_tc);
                     else{
                        game_music.FadeVolume(1.0f);
                        RenderAllBackbuffers();
                        tick_class = CreateMenu(GAME_MENU_4_OF_8);
                        if(tick_class)
                           tick_class->Release();
                     }
                  }
                  return;
               default:
                  assert(0);
               }
            }
            break;
         case C_net_sync::SS_PROGRESS:
            {
                              //waiting for sync event
               if(!sync_menu){
                  sync_menu = CreateSyncMenu(0);
                  sync_menu->Release();
               }
               vector<dword> pids;
               net_sync->GetAnsweredPIDs(pids);
               sync_menu->AddCheckMarks(pids.begin(), pids.size());
            }
            return;
         }
      }

      if(countdown && (countdown -= tc.time) <= 0){
         countdown = 0;
      }

      if(text){
         if(text_come_cnt){
            int beg = text_come_cnt/BRIEF_TEXT_COME;
            text_come_cnt += tc.time;
            int end = text_come_cnt/BRIEF_TEXT_COME;
            C_sprite_group *pg = text->GetAnsiText();
            if(pg){
               const vector<C_smart_ptr<C_poly_sprite> > &sprs = pg->GetSPrites();
               if(end > sprs.size()){
                  end = sprs.size();
                  text_come_cnt = 0;
               }
               for(; beg<end; beg++)
                  sprs[beg]->SetOn(true);
            }else
            if(end>text_len)  //japan
               text_come_cnt = 0;
         }
         if(!fade_text.IsEnd()){
            text->SetColor(fade_text.Tick(tc.time));
            if(fade_text.IsEnd())
               text = NULL;
         }
      }

      if(end_count){
         if((end_count -= tc.time) <= 0)
            Finish();
         return;
      }

      if(!countdown && !end_count)
         C_mission::Tick(tc);

      if(tc.p_ctrl->Get(GKEY_ESCAPE)){
         tc.p_ctrl->IgnoreValue(GKEY_ESCAPE);
         Finish();
         return;
      }
   }

//----------------------------

   void Render(){

      driver->BeginScene();
      C_mission::Render();
      texts.Render();

      if(sync_menu)
         sync_menu->Render();

      driver->EndScene();
   }

//----------------------------

   virtual void NetIn(E_NETWORK_MESSAGE msg_id, PC_net_in_msg in_msg){

      switch(msg_id){
      case NM_BRIEFING:
         {
            net_sync->NetIn(in_msg, true);
            dword event;
            switch(net_sync->GetState(NULL, &event)){
            case C_net_sync::SS_RESOLVED:
               {
                  if(sync_menu){
                     sync_menu->AddAllCheckMarks();
                     RenderScreen();
                     sync_menu = NULL;
                  }
                  switch(event){
                  case NS_ENTER:
                     net_sync->Reset();
                     break;
                  case NS_LEAVE:
                     {
                        NetSend();
                        if(save_tc)
                           tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(save_tc);
                        else{
                           game_music.FadeVolume(1.0f);
                           RenderAllBackbuffers();
                           tick_class = CreateMenu(GAME_MENU_4_OF_8);
                           if(tick_class)
                              tick_class->Release();
                        }
                     }
                     return;
                  default:
                     assert(0);
                  }
               }
               break;
            case C_net_sync::SS_PROGRESS:
               {
                                 //waiting for sync event
                  if(!sync_menu){
                     sync_menu = CreateSyncMenu(0);
                     sync_menu->Release();
                  }
                  vector<dword> pids;
                  net_sync->GetAnsweredPIDs(pids);
                  sync_menu->AddCheckMarks(pids.begin(), pids.size());
               }
               return;
            }
         }
         break;

      default:
         if(save_tc){
            save_tc->NetIn(msg_id, in_msg);
         }else{
            NET_LOG(C_fstring("Briefing: skipping chunk 0x%x", msg_id));
            in_msg->SkipChunk();
         }
      }
   }

//----------------------------

   virtual void OnPlayerDisconnect(dword pid, const char *name){

      net_sync->OnPlayerDisconnect(pid);
      C_mission::OnPlayerDisconnect(pid, name);
      if(save_tc)
         save_tc->OnPlayerDisconnect(pid, name);
   }

//----------------------------

   virtual dword GetID() const{ return 'BRFG'; }
};

//----------------------------

C_class_to_be_ticked_hd *CreateBriefing(const char *dir, C_class_to_be_ticked_hd *save_tc){

   C_briefing *brf = new C_briefing;
   if(!brf->Init(dir, save_tc)){
      brf->Release();
      brf = NULL;
   }
   return brf;
}

//----------------------------
