#include "all.h"

#if defined BETAVERSION || defined EDITOR

#include "winstuff.h"
#include <insanity\controller.h>
#include "net.h"
#include <advanced.h>

//----------------------------

#define CURSOR_BLINK_TIME 150

class C_bugreport: public C_class_to_be_ticked_hd{
   C_smart_ptr<C_class_to_be_ticked_hd> save_tc;

   bool local_database;
   int bug_id;
   bool ed_active;

   C_smart_ptr<C_text> text;
   C_smart_ptr<C_poly_sprite> spr_arrow;
   C_string database_dir;

   class C_description{
      enum{ BUFFER_SIZE = 300 };
      int cursor_count;
      bool cursor_phase;
      C_smart_ptr<C_text> text;

      void ShowText(const char *cp, bool rc){

         float fmt_rect[4] = {.02f, 0, .96f, 0};
         text = console_main.texts.CreateText(cp, -1, .02f, .12f, .04f, 0, false, fmt_rect);
         text->Release();
                                    //automatically reset cursor
         if(rc){
            cursor_phase = false;
            cursor_count = 0;
         }
      }
   public:
      void HideText(){ text = NULL; }

      char buffer[BUFFER_SIZE];
      int buf_pos;          //-1 means not in chat mode

      C_description()
      {
         buffer[buf_pos = 0] = 0;
         cursor_count = 0;
         cursor_phase = true;
      }

      void Tick(const S_tick_context &tc){

         char ascii_key = igraph->GetAscii(tc.key, igraph->GetShiftKeys()&SKEY_SHIFT);

         switch(tc.key){

         case K_BACKSPACE:
         case K_DEL:
            if(buf_pos){
               --buf_pos;
               buffer[buf_pos] = 0;
               ShowText(C_fstring("%s_", (const char*)buffer), true);
            }
            break;
         }
                                    //add char
         if(isprint(ascii_key)){
            if(buf_pos<BUFFER_SIZE-1){
               buffer[buf_pos++] = ascii_key;
               buffer[buf_pos] = 0;
               ShowText(C_fstring("%s_", (const char*)buffer), true);
            }
         }else{
                                 //blink chat cursor
            if((cursor_count -= tc.time) <= 0){
               cursor_count = CURSOR_BLINK_TIME;
               cursor_phase = !cursor_phase;
               ShowText(C_fstring("%s%s", buffer, !cursor_phase ? "" : "_"), false);
            }
         }
      }
   } desc;

public:
   C_bugreport(C_class_to_be_ticked_hd *stc):
      save_tc(stc),
      bug_id(0)
   {
                              //determine bug number
      char redirect[256];
      redirect[0] = 0;
      int i = Load("Bugs\\redirect.txt", &redirect, sizeof(redirect)-1);
      if(i!=-1){
         redirect[i] = 0;
         while(i-- && (isspace(redirect[i]) || iscntrl(redirect[i])))
            redirect[i] = 0;
         database_dir = redirect;
      }else
         database_dir = "Bugs";
      Load(C_fstring("%s\\bug_id.bin", (const char*)database_dir), &bug_id, sizeof(bug_id));

#ifdef EDITOR
      ed_active = editor->IsActive();
      if(ed_active)
         editor->SetActive(false);
#endif

      console_main.texts.Init();
      text = console_main.texts.CreateText(C_fstring("Bug number: %i\nText:", bug_id), -1, .02f, .05f, .04f);
      text->Release();

      spr_arrow = new C_poly_sprite(0, 0, .05f, &icons_img, 0);
      spr_arrow->SetColor(0xff00ff00);
      spr_arrow->Release();
      spr_arrow->SetScreenPos(igraph->Mouse_x(), igraph->Mouse_y());
   }

   virtual dword GetID() const{ return 'BUGR'; }

   virtual void Tick(const S_tick_context &tc){

      spr_arrow->SetScreenPos(igraph->Mouse_x(), igraph->Mouse_y());

      switch(tc.key){
      case K_ENTER:
         if(desc.buf_pos){
            Load(C_fstring("%s\\bug_id.bin", (const char*)database_dir), &bug_id, sizeof(bug_id));

            desc.buffer[desc.buf_pos] = 0;
                              //get file title
            C_string file_title = desc.buffer;
            {
               int last_space = 0;
               const int max_title_size = Min(60, (int)file_title.Size());
               for(int i=0; i<max_title_size; i++){
                  char c = file_title[i];
                  if(!isalnum(c))
                     c = ' ';
                  if(c==' ')
                     last_space = i;
                  file_title[i] = c;
               }
               if(file_title[i])
                  file_title[last_space] = 0;
            }
            C_fstring fname("%s\\%.3i %s.txt", (const char*)database_dir, bug_id, (const char*)file_title);
            OsCreateDirectoryTree(fname);

            C_string comp;
            OsGetComputerName(comp);

            S_date_time dt;
            OsGetCurrentTimeDate(dt);
            C_fstring title("Entered by: %s\r\nDate: %i.%i.%i %i:%.2i\r\n", (const char*)comp, dt.day, dt.month, dt.year, dt.hour, dt.minute);
            if(save_tc->GetID()=='GMIS'){
               C_game_mission *gm = (C_game_mission*)(C_class_to_be_ticked*)save_tc;
               const C_string &mname = gm->GetMissionName();
               const S_vector &pos = gm->GetScene()->GetActiveCamera()->GetWorldPos();

               title += C_fstring("Mission: '%s', camera: [%.1f, %.1f, %.1f]\r\n", (const char*)mname, pos.x, pos.y, pos.z);
            }
            title += C_fstring("-------------------\r\n\r\nBug description:\r\n----------------\r\n");

            vector<char> buf;
            buf.insert(buf.end(), title, title+title.Size());
            buf.insert(buf.end(), desc.buffer, desc.buffer+strlen(desc.buffer));

            C_fstring app("\r\n\r\n\r\nReponsible person:\r\n------------------\r\n\r\n\r\nStatus, date, sign:\r\n-------------------\r\n\r\n");
            buf.insert(buf.end(), app, app+app.Size());

            int i = Save(fname, buf.begin(), buf.size());
            if(i==-1){
               console_main.OutText("Failed to save bug report!", 0xffff0000);
            }else{

                              //increment bug ID
               ++bug_id;
               Save(C_fstring("%s\\bug_id.bin", (const char*)database_dir), &bug_id, sizeof(bug_id));

                              //save screenshot
               C_fstring fname("%s\\Shots\\%.3i.png", (const char*)database_dir, bug_id-1);
               OsCreateDirectoryTree(fname);
               C_cache ck;
               ck.open(fname, CACHE_WRITE);
               if(ck.good()){
                              //re-render without desc texts
                  desc.HideText();
                  text = NULL;
                  RenderScreen();

                  LPIImage img = igraph->CreateBackbufferCopy(true);
                  img->SaveShot(&ck, "png");
                  img->Release();
                  ck.close();
               }
            }
         }
                              //flow...
      case K_ESC:
                              //ignore pressed key
         {
            int slot, ii;
            if(tc.p_ctrl->FindSlot(C_controller::ID_KEYBOARD, tc.key, slot, ii)){
               tc.p_ctrl->IgnoreValue(slot, ii);
            }
         }
#ifdef EDITOR
         if(ed_active)
            editor->SetActive(true);
#endif         
         tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(save_tc);
         return;
      default:
         desc.Tick(tc);
      }

                              //must tick in net game
      if(net && save_tc->GetID()=='GMIS'){
         S_tick_context tc1 = tc;
         tc1.ClearInput();
         save_tc->Tick(tc1);
      }
   }

//----------------------------

   virtual void Render(){

      driver->BeginScene();
      save_tc->Render();

      bool use_zb = driver->GetState(RS_USEZB);
      if(use_zb) driver->SetState(RS_USEZB, false);
      spr_arrow->Render();
      if(use_zb) driver->SetState(RS_USEZB, true);

      driver->EndScene();
   }

//----------------------------

   virtual void NetIn(enum E_NETWORK_MESSAGE msg_id, class C_net_in_msg *in_msg){
      save_tc->NetIn(msg_id, in_msg);
   }
};

//----------------------------

C_class_to_be_ticked_hd *CreateBugReport(C_class_to_be_ticked_hd *save_tc){

   assert(save_tc);
   return new C_bugreport(save_tc);
}

#endif//BETAVERSION

//----------------------------
