#include "all.h"
#include "net.h"
#include "mainmenu.h"

//----------------------------
                              //campaign animation implementation
class C_camp_anim: public C_mission{

   C_smart_ptr<C_class_to_be_ticked_hd> save_tc;

   int end_count;
   PI3D_frame cam_target;

   static dword I3DAPI cbCBox(PI3D_frame frm, FRM_CALLBACK_MESSAGE msg, dword prm1, dword prm2){

      C_camp_anim *ca = (C_camp_anim*)frm->GetData();
      switch(msg){

      case CBM_ANIMNOTE:
         {
                              //convert string to number
            const I3D_note_callback *nc = (I3D_note_callback*)prm1;
            prm2 = nc->note_time;
            prm1 = atoi(nc->note);
            ca->Notify(frm, prm1, prm2);
         }
         break;

      case CBM_ANIM_FINISH:
         ca->end_count = 1000;
         break;
      }
      return 0;
   }

   enum{
      NS_NULL,
      NS_ENTER,
      NS_LEAVE,
   };

   C_smart_ptr<C_net_sync> net_sync;
   C_smart_ptr<C_sync_menu> sync_menu;

//----------------------------

   void Finish(){

      if(!net){
         if(save_tc && save_tc->GetID()=='MENU')
            game_music.Play("logo");
         tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(save_tc);
         return;
      }
      net_sync->BeginEvent(NS_LEAVE);
   }

   C_smart_ptr<I3D_frame> cbox;
public:
   C_camp_anim():
      end_count(0)
   {
      net_sync = CreateNetSync(NM_CAMP_ANIM);
      net_sync->Release();
   }

   bool Init(const char *anim_name, C_class_to_be_ticked_hd *stc, bool load_progress){

      E_MISSION_IO mio = Open(C_fstring("Cinematic\\%s", (const char*)anim_name),
         (load_progress ? 0 : OPEN_NO_SHOW_PROGRESS) |
         OPEN_USE_SCENE_CAMERA | OPEN_NO_OPEN_FAIL_MSG);
      if(mio!=MIO_OK)
         return false;

      save_tc = stc;
                              //modify interpolators to simple
                              // also find camerabox interpolator
      const PI3D_interpolator *intps = scene->GetInterpolators();
      for(int i=scene->NumInterpolators(); i--; ){
         PI3D_interpolator intp = intps[i];
         intp->SetAnimation(0, intp->GetAnimation(0));
         if(!cbox && C_string(intp->GetFrame()->GetName())=="camerabox"){
            cbox = intp->GetFrame();
            cbox->SetCallback(cbCBox, FRMCALLBACK_ANIMNOTE | FRMCALLBACK_ANIMFINISH);
            cbox->SetData((dword)this);
         }
      }
      if(net)
         net_sync->BeginEvent(NS_ENTER);
      return true;
   }

//----------------------------

   ~C_camp_anim(){
      if(cbox)
         cbox->SetData(0);
   }

//----------------------------

   void Tick(const S_tick_context &tc){

      if(net){
         dword event;
         switch(net_sync->GetState(NULL, &event)){
         case C_net_sync::SS_RESOLVED:
            {
               if(sync_menu){
                  sync_menu->AddAllCheckMarks();
                  RenderScreen();
                  sync_menu = NULL;
               }
               switch(event){
               case NS_ENTER:
                  net_sync->Reset();
                  break;
               case NS_LEAVE:
                  {
                     if(save_tc && save_tc->GetID()=='MENU')
                        game_music.Play("logo");
                     tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(save_tc);
                  }
                  return;
               default:
                  assert(0);
               }
            }
            break;
         case C_net_sync::SS_PROGRESS:
            {
                              //waiting for sync event
               if(!sync_menu){
                  sync_menu = CreateSyncMenu(0);
                  sync_menu->Release();
               }
               vector<dword> pids;
               net_sync->GetAnsweredPIDs(pids);
               sync_menu->AddCheckMarks(pids.begin(), pids.size());
            }
            return;
         }
      }

      switch(tc.key){
      case K_ESC:
         Finish();
         return;
      }

      if(end_count){
         if((end_count -= tc.time) <= 0)
            Finish();
         return;
      }

      C_mission::Tick(tc);
   }

//----------------------------

   virtual void NetIn(E_NETWORK_MESSAGE msg_id, PC_net_in_msg in_msg){

      switch(msg_id){
      case NM_CAMP_ANIM:
         net_sync->NetIn(in_msg, true);
         break;

      default:
         if(save_tc){
            save_tc->NetIn(msg_id, in_msg);
         }else{
            assert(0);
            in_msg->SkipChunk();
         }
      }
   }

//----------------------------

   virtual void OnPlayerDisconnect(dword pid, const char *name){

      net_sync->OnPlayerDisconnect(pid);
      C_mission::OnPlayerDisconnect(pid, name);
      if(save_tc)
         save_tc->OnPlayerDisconnect(pid, name);
   }

//----------------------------

   virtual void Render(){

      driver->BeginScene();
      C_mission::Render();

      if(sync_menu)
         sync_menu->Render();
      driver->EndScene();
   }

//----------------------------

   virtual dword GetID() const{ return 'CAMP'; }
};

//----------------------------

C_class_to_be_ticked_hd *CreateCampAnim(const char *anim_name, C_class_to_be_ticked_hd *save_tc, bool load_progress){

   C_camp_anim *ca = new C_camp_anim;
   if(!ca->Init(anim_name, save_tc, load_progress)){
      ca->Release();
      ca = NULL;
   }
   return ca;
}

//----------------------------
//----------------------------
