#include "all.h"
#include "commandline.h"
#include "winstuff.h"
#include "net.h"
#include <process.h>

//----------------------------

extern const char *lang_names[];

//----------------------------

const C_command_line_base::S_scan_data *C_command_line::GetScanData() const{
   static const S_scan_data scan_data[] = {
      {"help", CMD_NOP, 0, (t_Callback)ShowHelp},
      {"resolution", CMD_INT_X2, offsetof(C_command_line, res_x)},
      {"position", CMD_INT_X2, offsetof(C_command_line, pos_x)},
      {"releasedll", CMD_BOOL, offsetof(C_command_line, debug_use_release_dlls)},
      {"direct_trans", CMD_BOOL, offsetof(C_command_line, debug_direct_trans)},
      {"launch_app", CMD_NOP, 0},
      {"mission", CMD_STRING | CMD_GLOBAL_PTR, (dword)&game_configuration.net_cmdline_mission},
      {"stop", CMD_NOP, 0, (t_Callback)Stop},
      {"sleep", CMD_NOP, 0, (t_Callback)Sleep},
      {"net_host", CMD_BOOL | CMD_GLOBAL_PTR, (dword)&game_configuration.net_host},
      {"net_join", CMD_BOOL | CMD_GLOBAL_PTR, (dword)&game_configuration.net_join},
      {"net_log", CMD_BOOL | CMD_GLOBAL_PTR, (dword)&game_configuration.net_debug_log},
      {"net_num", CMD_INT | CMD_GLOBAL_PTR, (dword)&game_configuration.net_debug_numplayers},
      {"net_player", CMD_STRING | CMD_GLOBAL_PTR, (dword)&game_configuration.net_cmdline_plr_name},
      {"net_cpu_schedule", CMD_BOOL | CMD_GLOBAL_PTR, (dword)&game_configuration.net_debug_schedule_cpu},
      {"graph_caps", CMD_BOOL, offsetof(C_command_line, dump_graph_caps)},
      {"sound_caps", CMD_BOOL, offsetof(C_command_line, dump_sound_caps)},
      {"net_session_name", CMD_STRING | CMD_GLOBAL_PTR, (dword)&game_configuration.net_cmdline_sess_name},
      {"net_address", CMD_STRING | CMD_GLOBAL_PTR, (dword)&game_configuration.net_cmdline_tcpip_addr},
      {"net_port", CMD_STRING | CMD_GLOBAL_PTR, (dword)&game_configuration.net_cmdline_tcpip_port},
      {"datadisk", CMD_INT | CMD_GLOBAL_PTR, (dword)&game_configuration.datadisk_ver},
      {"profile", CMD_STRING, offsetof(C_command_line, profile_name)},
      {"language", CMD_NOP, 0, (t_Callback)ReadLang},
      {"safe", CMD_BOOL | CMD_GLOBAL_PTR, (dword)&game_configuration.safe_mode},
      NULL
   };
   return scan_data;
}

//----------------------------

bool C_command_line::ReadLang(const char **cl){

   C_string name;
   if(!ReadString(cl, name)){
      DisplayError("Expecting parameter after 'language'", APP_NAME);
      return false;
   }

   for(int i=0; lang_names[i]; i++){
      if(C_string(lang_names[i]).Matchi(name)){
         game_configuration.language = i;
         break;
      }
   }
   if(!lang_names[i]){
      DisplayError(C_fstring("Invalid language: %s", (const char*)name), APP_NAME);
      return false;
   }
   return true;
}

//----------------------------

bool C_command_line::Sleep(const char **cl){

   int time;
   if(!ReadInt(cl, time)){
      DisplayError("Expecting parameter after sleep", APP_NAME);
      return false;
   }
   OsSleep(time);
   return true;
}

//----------------------------
                              //macros causing to include specified help message
                              // only if EDITOR is #defined
#ifdef EDITOR
#define CLHELP_EDIT_TEXT(t) t
#else
#define CLHELP_EDIT_TEXT(t)
#endif


#ifdef _DEBUG
#define CLHELP_DEBUG_TEXT(t) t
#else
#define CLHELP_DEBUG_TEXT(t)
#endif

//----------------------------

C_command_line::C_command_line():
   C_command_line_base(3),
   res_x(0), res_y(0),
   pos_x(0x80000000), pos_y(0x80000000),
#ifndef NDEBUG
   debug_use_release_dlls(false),
#endif
   dump_graph_caps(false), dump_sound_caps(false),
   debug_direct_trans(false)
{
}

//----------------------------

bool C_command_line::ShowHelp(const char **cl){

   OsMessageBox(
                  "General:\n"
                  "  datadisk<n>\tset version to run\n"
                  "  language<name>\tset language to run\n"
                  "  graph_caps\tdump videocard caps\n"
                  "  sound_caps\tdump soundcard caps\n"
                  "  resolution <x> <y>\tset window resolution\n"
                  "  position <x> <y>\tset window position\n"
                  "  profile <name>\tset default profile\n"
CLHELP_EDIT_TEXT( "  releasedll\tuse release engine dlls\n")
CLHELP_EDIT_TEXT( "  game_mode\trun game mode\n")
CLHELP_EDIT_TEXT( "  mode <SNG|BM|DM|TDM|CTF|GL|F|MC>\tset specified mission mode\n")
CLHELP_EDIT_TEXT( "  no_editor\tdon't initialize editor\n")
CLHELP_EDIT_TEXT( "  mission \"name\"\tset specified mission\n")
CLHELP_EDIT_TEXT( "  model \"name\"\tview specified model\n")
CLHELP_EDIT_TEXT( "  sleep <n>\tsleep for specified period\n")
CLHELP_EDIT_TEXT( "  countdown <n>\trun only for period\n")
CLHELP_DEBUG_TEXT("  debug_break <n>\tset breakpoint at n-th allocation\n")
                  "Network:\n"
                  "  net_host\thost game\n"
                  "  net_join\tjoin game\n"
                  "  net_connect\tautomatically connect to network game\n"
                  "  net_player \"name\"\tset net player name\n"
CLHELP_DEBUG_TEXT("  net_cpu_schedule <n>\tschedule CPU time in net game\n")
CLHELP_DEBUG_TEXT("  net_num <n>\tnumber net players (auto-hosting)\n")
                  "  net_session_name \"name\"\tset net session name\n"
                  "  net_connection_provider <type>\tset net connection (tcpip|ipx|sim)\n"
                  "  net_address \"string\"\tset network address to connect to (tcpip)\n"
                  "  net_port \"string\"\tset network port to connect to (tcpip)\n"
CLHELP_DEBUG_TEXT("  net_log \"string\"\tlog network activity\n")
                  "Others:\n"
                  "  launch_app \"app_name\" [params]\trun another application\n"
                  "  stop\tstop processing cmdline\n"
      ,
      C_fstring("%s - command-line help",
      //app_name),
      APP_NAME),
      MBOX_OK);
   return false;
}

//----------------------------
//----------------------------
