#include "all.h"


//----------------------------
//----------------------------

struct S_out_text{
   LPC_text text;
   int num_lines;

   S_out_text():
      text(NULL)
   {}
   S_out_text(const char *cp, C_poly_text &texts, float x, float y,
      float line_top, float line_size, dword color, bool wide, dword justify):
      num_lines(1)
   {
      static float fmt_rect[4] = {.01f, 0, .99f, 0};
      text = texts.CreateText(cp, color, x, y+line_top, line_size,
         justify | //TEXT_JUSTIFY_CLIP_TO_SCREEN |
         TEXT_JUSTIFY_CONVERT_EOL,
         wide, fmt_rect, NULL, NULL, &num_lines);
   }
   ~S_out_text(){
      if(text)
         text->Release();
   }
};

//----------------------------

C_console::C_console(float x, float y, float ls, float ld, dword st, float b, dword j):
   justify(j),
   scroll_time(st),
   line_bottom(b),
   text_x(x),
   text_y(0.0f),
   line_top(y),
   line_size(ls),
   line_dist(ld),
   out_text_time(0)
{}

//----------------------------

C_console::~C_console(){
   Clear();
}

//----------------------------

void C_console::Clear(){

   for(int i=0; i<out_texts.size(); i++)
      delete out_texts[i];
   out_texts.clear();
   text_y = 0.0f;
}

//----------------------------

void C_console::OutText(const char *cp, dword color, bool wide){

   if(!out_texts.size())
      out_text_time = scroll_time * 2;

   text_y += line_dist;
   if(out_texts.size())
      text_y += line_dist * (out_texts.back()->num_lines - 1);

   if(!texts.IsInit())
      texts.Init();
   out_texts.push_back(new S_out_text(cp, texts, text_x, text_y, line_top, line_size, color, wide, justify));

   if(out_texts.size()>((line_bottom-line_top)/line_dist)){
      delete out_texts.front();
      out_texts.erase(&out_texts.front());
      OrderTexts();
   }
}

//----------------------------

void C_console::OrderTexts(){

   text_y = 0.0f;
   for(int i=0; i<out_texts.size(); i++){
      text_y += line_dist;
      out_texts[i]->text->SetPos(text_x, text_y+line_top);
      text_y += line_dist * (out_texts[i]->num_lines - 1);
   }
}

//----------------------------

void C_console::Tick(int t){

   if(out_texts.size()){   //scroll console messages
      while(out_texts.size()){
         int i = t * Min(18, (int)exp(out_texts.size()*.18f));
         if((out_text_time -= i)>0)
            break;
         out_text_time = Max(0, (int)(out_text_time + scroll_time));
         delete out_texts.front();
         out_texts.erase(&out_texts.front());
         OrderTexts();
      }
   }
}

//----------------------------
//----------------------------

void OUTTEXT(int i){
   OUTTEXT(C_fstring("%i", i));
}

//----------------------------

void OUTTEXT(float f){
   OUTTEXT(C_fstring("%.2f", f));
}

//----------------------------

void OUTTEXT(const char *cp, dword color){
   console_main.OutText(cp, color);
}

//----------------------------

void OUTTEXT(const S_vector &v){
   OUTTEXT(C_fstring("[%.2f %.2f %.2f]", v.x, v.y, v.z));
}

//----------------------------

void DEBUG(const char *cp){
   console_debug.OutText(cp, 0xc0ffffff);
}

void DEBUG(int i){
   DEBUG(C_fstring("%i", i));

}
void DEBUG(float f){
   DEBUG(C_fstring("%.2f", f));
}

void DEBUG(const S_vector &v){
   DEBUG(C_fstring("[%.2f %.2f %.2f]", v.x, v.y, v.z));
}

//----------------------------
