#include "all.h"
#include "actors.h"
#include "net.h"
#include "tables.h"


//----------------------------

#define DETECTOR_COUNTER 400

//----------------------------

                              //C_detector

                              //detectors over network:
                              // - primary detector:
                              //    - run detection
                              //    - send 'unlock' message when success
                              // - network detector:
                              //    - wait for unlock
                              //    - peform only some actions:
                              //       - play sound
                              //       - pause
                              // - network messages:
                              //    - bits:2 = code:
                              //       - 0 = unlock
                              //       - 1 = explode
                              //          - S_vector expl_pos
                              //       - 2 = listening (bits:2 E_listen_type, S_vector pos, float power)
#if defined _DEBUG && 1
 #define DET_DEBUG(n) if(tab_game_cfg->ItemB(TAB_B_GAME_DEBUG_DETECTOR)) OUTTEXT(n)
#else
 #define DET_DEBUG(n)
#endif

//----------------------------
                              //template for detector
static C_table_template templ_detector = {
   // {821E8220-CF4B-11d2-B347-0000B49B8E8B}
   { 0x821e8220, 0xcf4b, 0x11d2, { 0xb3, 0x47, 0x0, 0x0, 0xb4, 0x9b, 0x8e, 0x8b } },
   "Detector properties",
   NULL,
   {
      {TE_ARRAY, TAB_E_COND, "Detect what", 16},
         {TE_BRANCH, TAB_E_COND, NULL, 5, (dword)"%[0] \"%[1]\" (%[2]) \"%[3]\""},
            {TE_ENUM, TAB_E_COND, "Condition", (dword)
               "no\0"
               "Player in range (m)\0"
               "Any human in range (m)\0"
               "CountDown (msec)\0"

               "Resource in range (range, ID)\0"
               "All alive p. in range (m)\0"
               "Signal received (sig_id)\0"
               "Actor destroyed\0"

               "Explosion in range (m)\0"
               "Sounds (sensitivity %)\0"
               "All enemies dead\0"
               "All players dead\0"

               "Actor in range (range, , name)\0"
               "Res. in inv. in range (range, ID)\0"
               },
            {TE_INT, TAB_I_COND_DATA, "Condition - data"},
            {TE_INT, TAB_I_COND_DATA2, "Condition - data 2"},
            {TE_STRING, TAB_S32_COND_DATA_S, "Condition - string", 32},
            {TE_BOOL, TAB_B_DET_COND_INVERSE, "Inverse condition"},
      {TE_ARRAY, TAB_E_DET_ACTION, "Actions", 16},
         {TE_BRANCH, TAB_E_DET_ACTION, NULL, 3, (dword)"%[0] \"%[1]\" (%[2])"},
            {TE_ENUM, TAB_E_DET_ACTION, "Action", (dword)
               "no\0"
               "Play sound (name)\0"
               "Send signal (whom, sig_id)\0"
               "Mission successful\0"

               "Mission failed\0"
               "Pause detector\0"
               "Explosion (CP, expl_id)\0"
               "Set on/off (0=off, 1=on)\0"

               "Deactivate detector (name)\0"
               "Destroy actor (name, power)\0"
               "Objective completed (text id)\0"
               "Set Focus (actor name)\0"
               },
            {TE_STRING, TAB_S32_DET_ACTION_DATA, "Action - data string", 32},
            {TE_INT, TAB_I_DET_ACTION_DATA, "Action - data number"},
      {TE_INT, TAB_I_DET_RESET_TIME, "Reset time (-1 = never)", 0, 0, -1},
      {TE_NULL}
   }
};

//----------------------------

class C_detector: public C_actor{

   LPC_table tab;

   int reset_counter;         //0 = detecting, -1 = never reset

   int cond_index;
   int cond_count;

   int action_pause;
   int action_index;
   int count;

   void PerformActions(){
      bool loop = true;
      int &i = action_index;
      for(i=action_index; loop && i<tab->ArrayLen(TAB_E_DET_ACTION); i++){
                        //detection succeeded
                        //perform action
         byte action = tab->ItemE(TAB_E_DET_ACTION, i);
         switch(action){
         case 1:        //play sound
            {
               C_string snd_name = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
               if(snd_name.Size()>=4 && snd_name[(int)snd_name.Size()-4]=='.'){
                  snd_name[(int)snd_name.Size()-4] = 0;
               }
               mission.PlaySoundGlobal(snd_name);
            }
            break;

         case 3:        //mission successful
            if(network_actor) break;
            DET_DEBUG(C_fstring("Detector %s: mission successful", (const char*)GetName()));
            mission.SetGameState(MS_DONE_SUCCESS);
            break;

         case 4:        //mission fail
            if(network_actor) break;
            DET_DEBUG(C_fstring("Detector %s: mission failed", (const char*)GetName()));
            mission.SetGameState(MS_DONE_FAIL);
            break;

         case 2:        //send signal
            //if(network_actor) break;
            {
               LPC_actor dest = NULL;
               const char *cp = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
               if(*cp){
                  dest = mission.FindActor(cp);
                  if(!dest)
                     continue;
               }
               S_signal_data sd;
               sd.actor = this;
               sd.type = SIGNAL_DETECTOR;
               sd.sub_type = tab->ItemI(TAB_I_DET_ACTION_DATA, i);
               if(dest) dest->cbProc(CB_SIGNAL, (dword)&sd, 0);
               else mission.BroadcastMessage(CB_SIGNAL, (dword)&sd, 0);
            }
            break;

         case 5:              //pause detector
            action_pause = Max(1, tab->ItemI(TAB_I_DET_ACTION_DATA, i));
            loop = false;
            break;

         case 6:              //create explosion
            if(network_actor)
               break;
            {
               const char *cp1 = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
               LPC_checkpoint cp = mission.FindCP(cp1);
               if(cp){
                  S_explosion_init ei;
                  ei.pos = cp->GetPos();
                  ei.id = tab->ItemI(TAB_I_DET_ACTION_DATA, i);
                  mission.CreateActor(ACTOR_EXPLOSION, (dword)&ei);
               }
            }
            break;

         case 7:              //on/off frame
            //if(network_actor) break;
            {
               C_string str = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
               if(str.Size() >= 3 && str[(int)str.Size()-1]=='@' && str[(int)str.Size()-1]=='@'){
                  str[(int)str.Size()-3] = 0;
               }
               bool on = tab->ItemI(TAB_I_DET_ACTION_DATA, i);
               struct S_hlp{
                  static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                     frm->SetOn(c);
                     struct S_hlp{
                        static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                           if(frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()!=I3D_VISUAL_PARTICLE)
                              return I3DENUMRET_OK;
                           frm->SetOn(c);
                           return I3DENUMRET_OK;
                        }
                     };
                              //also all its particles, sounds and volumes
                     frm->EnumFrames(S_hlp::cbEnum, c, ENUMF_VISUAL | ENUMF_VOLUME | ENUMF_SOUND | ENUMF_LIGHT);
                     return I3DENUMRET_OK;
                  }
               };
               mission.GetScene()->EnumFrames(S_hlp::cbEnum, on, ENUMF_ALL | ENUMF_USEMASK, str);
            }
            break;

         case 8:        //deactivate detector
            if(network_actor)
               break;
            {
               LPC_actor dest = NULL;
               const char *cp = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
               if(*cp){
                  dest = mission.FindActor(cp);
                  if(!dest)
                     continue;
                  dest->cbProc(CB_DEACTIVATE, (dword)this);
               }
            }
            break;

         case 9:        //destroy actor
            if(network_actor)
               break;
            {
               const char *cp = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
               if(*cp){
                  const vector<C_smart_ptr<C_actor> > &actors = mission.GetActors();
                  for(int ai=actors.size(); ai--; ){
                     LPC_actor a = actors[ai];
                     if(C_string(a->GetName()).Match(cp)){
                        a->Explode(this, tab->ItemI(TAB_I_DET_ACTION_DATA, i));
                     }
                  }
               }
            }
            break;

         case 10:             //objective completed
            mission.DisplayObjectives();
            break;

         case 11:             //set focus
            {
               const char *cp = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
               if(*cp){
                  if(mission.game_cam.GetDistanceMode()!=2)
                     mission.game_cam.SetDistanceMode(2);
                              //find specified actor
                  LPC_actor a = mission.FindActor(cp);
                  if(a){
                              //set visible active player
                     LPC_actor aa = FindActivePlayer();
                     if(aa)
                        aa->cbProc(CB_SETFOCUS, false);
                              //set focus on wanted actor
                     if(!a->cbProc(CB_SETFOCUS, 1))
                        mission.game_cam.SetFocus(a->GetFrame(), a->GetFrame());
                  }
               }else{
                              //find active player
                  /*struct S_hlp{
                     static bool cbE(LPC_actor a, dword c){
                        if(a->IsActive()){
                           *(LPC_actor*)c = a;
                           return false;
                        }
                        return true;
                     }
                  };
                  LPC_actor aa = NULL;
                  mission.EnumActors(S_hlp::cbE, (dword)&aa);*/
                  LPC_actor aa = FindActivePlayer();
                  if(aa) aa->cbProc(CB_SETFOCUS, 1);
               }
            }
            break;
         }
      }
   }

//----------------------------

   LPC_actor FindActivePlayer(){

      struct S_hlp{
         static bool cbE(LPC_actor a, dword c){
            if(a->IsActive()){
               *(LPC_actor*)c = a;
               return false;
            }
            return true;
         }
      };
      LPC_actor aa = NULL;
      mission.EnumActors(S_hlp::cbE, (dword)&aa);
      return aa;
   }

//----------------------------

   void Unlock(){

      action_index = 0;
      if(net && !network_actor){
                              //send beginning over network
         assert(!network_actor);
         bool g = true;
         net_msg->Ascend(net_id);
         net_msg->AddBits(0, 2, g);
         net_msg->Descend();
      }
                              //setup reset counter
      reset_counter = tab->ItemI(TAB_I_DET_RESET_TIME);
      PerformActions();
   }

//----------------------------

   bool NextCond(bool deb = false){

      if(deb){
         DET_DEBUG(C_fstring("Detector %s - condition #%i passed", frame ? frame->GetName() : "<unknown>", cond_index));
      }
      if(++cond_index==tab->ArrayLen(TAB_E_COND)){
         Unlock();
         return true;
      }
      return false;
   }

public:
   C_detector(C_game_mission &m):
      C_actor(m, ACTOR_DETECTOR),
      count(0),
      action_pause(0),
      reset_counter(0),      
      cond_index(0),
      cond_count(0),
      action_index(-1),
      tab(CreateTable())
   {
      tab->Open((dword)&templ_detector, TABOPEN_TEMPLATE);
                              //only host is master for detector class
      if(net)
         network_actor = net->IsHost() ? 0 : net->GetHostPID();
   }

   ~C_detector(){
      if(tab) tab->Release();
   }

//----------------------------

   virtual bool MayListen() const{

      if(!reset_counter && !cond_count)
         return (tab->ItemE(TAB_E_COND, cond_index)==9);
      return false;
   }

//----------------------------

   virtual LPC_table GetTable(int index) const{
      if(index==0)
         return tab;
      return C_actor::GetTable(index);
   }

//----------------------------

   virtual class C_table_template *GetTemplate(int index) const{ return &templ_detector; }

//----------------------------

   virtual void TableUpdate(int index, bool not_load_prg){

      reset_counter = 0;
      count = 0;
   }

   virtual void MissionSave(C_chunk *ck) const{ SaveTable(ck, tab); }
   virtual bool MissionLoad(C_chunk *ck, int table_id){ return LoadTable(ck, tab); }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

                           //write data
      ck->Write(&reset_counter, sizeof(reset_counter));
      ck->Write(&cond_index, sizeof(cond_index));
      ck->Write(&cond_count, sizeof(cond_count));
      ck->Write(&action_pause, sizeof(action_pause));
      ck->Write(&action_index, sizeof(action_index));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

                              //read data
      ck->Read(&reset_counter, sizeof(reset_counter));
      ck->Read(&cond_index, sizeof(cond_index));
      ck->Read(&cond_count, sizeof(cond_count));
      ck->Read(&action_pause, sizeof(action_pause));
      ck->Read(&action_index, sizeof(action_index));
      return true;
   }

//----------------------------

   dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_ADD_OBJECTIVES:
         {
            vector<S_objective> &obj_list = *(vector<S_objective>*)prm1;
                              //scan all actions to see if there're some objectives
            for(int i=0; i<tab->ArrayLen(TAB_E_DET_ACTION); i++){
               if(tab->ItemE(TAB_E_DET_ACTION, i)==10){
                  obj_list.push_back(S_objective());
                  obj_list.back().text_id = tab->ItemI(TAB_I_DET_ACTION_DATA, i);
                  obj_list.back().completed = (action_index >= i);
               }
            }
         }
         break;

      case CB_SIGNAL:
         {
            if(network_actor)
               return false;
                              //check if we're awaiting signal
            if(!reset_counter && !cond_count)
            if(tab->ItemE(TAB_E_COND, cond_index)==6){
                              //check if it's signal from another detector
               LPS_signal_data sd = (LPS_signal_data)prm1;
               if(sd->type==SIGNAL_DETECTOR)
                              //check if sig_id matches
               if(sd->sub_type == tab->ItemI(TAB_I_COND_DATA, cond_index)){
                              //signal passed, move on!
                  DET_DEBUG(C_fstring("Detector %s: signal from %s received", (const char*)GetName(), sd->actor->GetName()));
                  NextCond(true);
               }
            }
         }
         break;

      case CB_DEACTIVATE:
         {
            reset_counter = -1;
            action_pause = 0;
            DET_DEBUG(C_fstring("Detector %s deactivated by %s", (const char*)GetName(), ((LPC_actor)prm1)->GetName()));
         }
         break;
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual bool MayExplode() const{

      if(network_actor){
         return (!reset_counter && !cond_count);
      }
      if(!reset_counter && !cond_count)
      if(tab->ItemE(TAB_E_COND, cond_index)==8)
         return true;
      return false;
   }

//----------------------------

   virtual bool Explode(C_actor *caller, int power){

      if(network_actor){
         if(!reset_counter && !cond_count){
            const bool g = true;
            net_msg->Ascend(net_id);
            net_msg->AddBits(1, 2, g);
            net_msg->AddVector(caller->GetFrame()->GetWorldPos(), g);
            net_msg->Descend();
         }
         return false;
      }
      if(!reset_counter && !cond_count){
         if(tab->ItemE(TAB_E_COND, cond_index)==8){
            float dist = (caller->GetFrame()->GetWorldPos()-GetFrame()->GetWorldPos()).Magnitude();
            if(dist<tab->ItemE(TAB_I_COND_DATA, cond_index))
               NextCond(true);
         }
      }
      return false;
   }

//----------------------------

   virtual void ReceiveListen(S_listen_source &ls, bool net_send){

      switch(ls.type){
      case LISTEN_SHOOT:
      case LISTEN_EXPLOSION:
         if(cond_index==tab->ArrayLen(TAB_E_COND))
            return;
         ls.power *= (float)tab->ItemI(TAB_I_COND_DATA, cond_index)/100.0f;
         if(net){
            if(net_send && network_actor){
                                 //don't process it directly, rather send it to network owner
               const bool g = false;
               net_msg->Ascend(net_id);
               net_msg->AddBits(2, 2, g);
               net_msg->AddBits(ls.type, 2, g);
               net_msg->AddVector(ls.pos, g);
               net_msg->AddFloat(ls.power, g);
               net_msg->Descend();
               return;
            }
            net_send = false;
         }
         C_actor::ReceiveListen(ls, net_send);
         return;
      }
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      if(action_pause && (action_pause-=tc.time)<=0){
         action_pause = 0;
         PerformActions();
         if(action_pause) return 0;
      }
                              //count for reset
      if(reset_counter==-1)
         return 0;
      if(reset_counter && (reset_counter-=tc.time)<=0){
         reset_counter = 0;
         cond_index = 0;
         DET_DEBUG(C_fstring("Detector %s: counter reset", (const char*)frame->GetName()));
      }
      if(reset_counter) return 0;
                              //network actor doesn't detect
      if(network_actor) return 0;

      if(cond_count){
         if((cond_count-=tc.time)>0) return 0;
         cond_count = 0;
         if(NextCond(true))
            return 0;
      }
                              //count for running detection
      if((count-=tc.time) <= 0){
         count = DETECTOR_COUNTER;

         do{
            const char *data_s = tab->ItemS(TAB_S32_COND_DATA_S, cond_index, 32);
            int data_i = tab->ItemI(TAB_I_COND_DATA, cond_index);
            int data_i_2 = tab->ItemI(TAB_I_COND_DATA2, cond_index);
            byte det_type = tab->ItemE(TAB_E_COND, cond_index);
            bool passed = false;
                              //run detection
            switch(det_type){
            case 0:           //not specified - always passed
               passed=true;
               break;
            case 1:           //any plr in range
            case 2:           //any human in range
            case 4:           //resource in range
            case 5:           //all alive players in range
            case 10:          //all enemies dead
            case 11:          //all players dead
            case 13:          //resource in human's inventory in range
               {
                  struct S_hlp{
                     LPC_actor this_a;
                     byte det_type;
                     int data_i;
                     int data_i_2;
                     int num_alive;
                     int num_passed;
                     S_hlp():
                        num_alive(0),
                        num_passed(0)
                     {}

                     static bool cbEnum(LPC_actor act, dword c){
                        S_hlp *hp = (S_hlp*)c;
                        int det_type = hp->det_type;
                                 //do only certain types
                        switch(act->GetType()){
                        case ACTOR_RESOURCE:
                           if(det_type==4){
                              if(act->GetType()==ACTOR_RESOURCE)
                              if(act->cbProc(CB_USE_RESOURCE, 1)==hp->data_i_2)
                                 break;
                           }
                           return true;
                        case ACTOR_PLAYER:
                           switch(det_type){
                           case 11:
                              if(act->IsAlive())
                                 ++hp->num_alive;
                              return true;
                           case 5:
                              if(act->IsAlive())
                                 ++hp->num_alive;
                              break;
                           case 1:
                           case 2:
                              break;
                           case 13:
                              {
                                 const S_vector *pos = act->GetPos();
                                 float dist = (*pos - hp->this_a->GetFrame()->GetWorldPos()).Magnitude();
                                 if(dist < hp->data_i){
                                                //search human's inventory
                                    const vector<C_smart_ptr<C_inventory::S_item> > *items = (const vector<C_smart_ptr<C_inventory::S_item> >*)act->cbProc(CB_GETINVENTORY);
                                    if(items){
                                       for(int i=items->size(); i--; ){
                                          if((*items)[i]->itm == hp->data_i_2)
                                             break;
                                       }
                                       if(i!=-1){
                                          ++hp->num_passed;
                                          return false;
                                       }
                                    }
                                 }
                                 return true;
                              }
                              break;
                           default:
                              return true;
                           }
                           break;
                        case ACTOR_ENEMY:
                           {
                              //don't count with non-germans
                              LPC_table tab = act->GetTable(0);
                              if(tab->ItemI(TAB_I_ENM_GROUP)!=0)
                                 return true;

                              switch(det_type){
                              case 2:
                                 break;
                              case 10:
                                 if(act->IsAlive()){
                                    ++hp->num_alive;
                                    return false;
                                 }
                                 return true;
                              default:
                                 return true;
                              }
                           }
                              //flow...
                        default: return true;
                        }
                        const S_vector *pos = act->GetPos();
                        float dist = (*pos - hp->this_a->GetFrame()->GetWorldPos()).Magnitude();
                        if(dist < hp->data_i){
                           ++hp->num_passed;
                           if(det_type!=5) return false;
                        }
                        return true;
                     }
                  } hlp;
                  hlp.this_a = this;
                  hlp.det_type = det_type;
                  hlp.data_i = data_i;
                  hlp.data_i_2 = data_i_2;

                  mission.EnumActors(S_hlp::cbEnum, (dword)&hlp);
                  switch(det_type){
                  case 5:     //all alive plrs in range
                     if(hlp.num_passed==hlp.num_alive && hlp.num_passed)
                        passed = true;
                     break;
                  case 10:    //all enemies dead
                  case 11:    //all players dead
                     if(!hlp.num_alive)
                        passed = true;
                     break;
                  default:
                     if(hlp.num_passed)
                        passed = true;
                  }
               }
               break;

            case 3:              //countdown
               cond_count = data_i;
               break;

            case 7:           //object destroyed
               {
                  LPC_actor act = mission.FindActor(data_s);
                  if(act){
                     switch(act->GetType()){
                     case ACTOR_PLAYER:
                     case ACTOR_ENEMY:
                        passed = !act->IsAlive();
                        break;
                     default:
                        passed = act->IsDestroyed();
                        break;
                     }
                  }
               }
               break;
            case 9:           //listen sounds
               {
                  const float v = 2.0f;
                  const S_listen_source *heard_sound = NULL;   //no sound heard

                  for(int j=listen_list.size(); j--; ){
                     const S_listen_source &ls = listen_list[j];
                     //if(ls.ignore_count) continue;
                                       //accept only sources over thresh
                     if(ls.power<v) continue;
                     if(heard_sound){
                        if(heard_sound->type<ls.type) continue;
                        if(heard_sound->type == ls.type && heard_sound->power>ls.power) continue;
                                       //more important sound found, keep it
                     }
                     heard_sound = &ls;
                  }
                  if(heard_sound){
                     //cbProc(CB_LISTEN_RESET);
                     ListenReset();
                     passed = true;
                     break;
                  }else
                     ListenTick(count);
               }
               break;

            case 12:          //actor in range
               {
                  LPC_actor act = mission.FindActor(data_s);
                  if(act){
                     S_vector apos;
                     act->GetCenterPos(apos);
                     float dist = (frame->GetWorldPos() - apos).Magnitude();
                     if(dist<data_i)
                        passed = true;
                  }
               }
               break;
            }
            if(tab->ItemB(TAB_B_DET_COND_INVERSE, cond_index)) passed = !passed;
            if(!passed) break;
            if(NextCond(det_type))
               break;
         }while(true);
      }
      return 0;
   }

//----------------------------

   virtual bool MaySaveGame() const{
      //return (bool)model;
                              //don't save end detector
      return (frame->GetName()[0]!=0);
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      in_msg->Ascend();
      while(in_msg->GetChunkSize()){
         switch(in_msg->GetBits(2)){
         case 0:                 //unlock
            NET_LOG("C_detector: unlocked by network message");
            Unlock();
            break;

         case 1:                 //explosion
            {
               S_vector pos = in_msg->GetVector();
               if(!network_actor)
               if(!reset_counter && !cond_count)
               if(tab->ItemE(TAB_E_COND, cond_index)==8){
                  float dist = (pos-GetFrame()->GetWorldPos()).Magnitude();
                  if(dist<tab->ItemE(TAB_I_COND_DATA, cond_index))
                     NextCond(true);
               }
            }
            break;

         case 2:              //listen source
            {
               S_listen_source ls;
               ls.type = (E_listen_type)in_msg->GetBits(2);
               ls.pos = in_msg->GetVector();
               ls.power = in_msg->GetFloat();
               if(!network_actor)   //for network_actor this information is useless
                  ReceiveListen(ls, false);
            }
            break;

         default:
            assert(0);
         }
      }
      in_msg->Descend();
   }
};

//----------------------------

LPC_actor CreateDetector(C_game_mission &gm){
   return new C_detector(gm);
}

//----------------------------
