#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"


//----------------------------

C_table_template templ_door = {
   // {807E29A0-AB9E-11d2-9B9E-004F4905FF1A}
   { 0x807e29a0, 0xab9e, 0x11d2, { 0x9b, 0x9e, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Door properties",
   NULL,
   {
      {TE_INT, TAB_I_DOR_ANGLE_CLOSE, "Close Angle", -360, 360},
      {TE_INT, TAB_I_DOR_ANGLE_OPEN, "Open Angle", -360, 360, 90},
      {TE_INT, TAB_I_DOR_AUTO_CLOSE_TIME, "Auto-close time (0-off)"},
      {TE_BOOL, TAB_B_DOR_LOCK, "Lock"},
      {TE_STRING, TAB_S12_DOR_SOUND, "Open sound", 12},
      {TE_NULL}
   }
};

//----------------------------

                              //door over network:
                              // - when open, send open message
#define DOOR_OPEN_TIME  1000

class C_door: public C_actor{

   enum{
      MODE_CLOSED, MODE_OPEN, MODE_OPENING, MODE_CLOSING
   } mode;
   int count;                 //for open/close
   C_ease_interpolator open_ease;
   LPC_table tab;
   PI3D_frame frm_door;

   PI3D_portal portal;
   PI3D_object object_dup;     //used only with portal

   S_vector lock_pos;
   int long_axis;             //X(0) or Z(2)

   void ApplyAngle(int indx){
      S_quat q;
      float a = tab->ItemI(indx)*PI/180.0f;
      q.Make(S_vector(0, 1, 0), a);
      frame->SetRot(q);
      frame->Update();
      if(portal){
         bool on(indx==TAB_I_DOR_ANGLE_OPEN);
         portal->SetOn(on);
         if(object_dup)
            object_dup->SetOn(!on);
      }
   }
   void Interpolate(float t1){

      float t = open_ease[t1];

      S_quat q;
      float a = tab->ItemI(TAB_I_DOR_ANGLE_CLOSE)*PI/180.0f;
      float b = tab->ItemI(TAB_I_DOR_ANGLE_OPEN)*PI/180.0f;
      q.Make(S_vector(0, 1, 0), a+(b-a)*t);
      frame->SetRot(q);
      frame->Update();
      if(portal){
         bool on = (t!=0.0f);
         portal->SetOn(on);
         if(object_dup)
            object_dup->SetOn(!on);
      }
   }

//----------------------------

   void PlaySound(){

      C_string name = tab->ItemS(TAB_S12_DOR_SOUND);
      if(!name[0])
         name = "door";
      else{
                              //remove extension
         for(int i=name.Size(); i--; ){
            if(name[i]=='.'){
               name[i] = 0;
               break;
            }
         }
      }
      mission.PlaySound(name, frame->GetWorldPos(), 2.0f, 6.0f);
   }

//----------------------------
                              //network_actor: yes

   void Open(LPC_actor opener, bool net_send){

      if(mode==MODE_CLOSED){
         mode = MODE_OPENING;
         count = DOOR_OPEN_TIME;
         PlaySound();
         S_listen_source ls(LISTEN_HUMAN, frame->GetWorldPos(), 15.0f);
         mission.EmitListen(ls,
            9.0f,             //max_dist
            false,            //net_send
            opener);

         if(net && net_send){
            bool g = true;
            net_msg->Ascend(net_id);
            net_msg->AddBits(0, 1, g);
            net_msg->Descend();
         }
      }
   }
public:
   C_door(C_game_mission &m):
      C_actor(m, ACTOR_DOOR),
      tab(CreateTable()),
      portal(NULL),
      object_dup(NULL),
      mode(MODE_CLOSED)
   {
      tab->Open((dword)&templ_door, TABOPEN_TEMPLATE);
   }
   ~C_door(){
      if(tab) tab->Release();
      if(portal){
         portal->Release();
      }
      if(object_dup) object_dup->Release();
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      C_actor::SetFrame(frm);

      lock_pos.Zero();
      if(frm->GetType()!=FRAME_VISUAL){
         frm = frm->FindChildFrame(NULL, ENUMF_VISUAL);
      }
      frm_door = frm;
      if(frm){
         I3D_bound_volume bvol;
         frame->ComputeHRBoundVolume(&bvol);

         //PI3D_object obj = I3DCAST_OBJECT(frm);
                              //determine position of door-lock
         //const I3D_bbox &bbox = obj->GetBoundVolume().bbox;
         const I3D_bbox &bbox = bvol.bbox;
                        //determine which side's greater
         int axis = (bbox.max.x-bbox.min.x) > (bbox.max.z-bbox.min.z) ? 0 : 2;
         long_axis = axis;

         lock_pos.y = bbox.min.y + 1.0f;
         lock_pos[axis] = bbox.min[axis] + (bbox.max[axis] - bbox.min[axis]) * .5f;

         //DebugLine(lock_pos*frame->GetMatrix(), lock_pos*frame->GetMatrix()+frame->GetMatrix()(2), 0);
         /*
                        //install portal
         PI3D_frame sct_door = obj->GetParent();
         while(sct_door && sct_door->GetType()!=FRAME_SECTOR)
            sct_door = sct_door->GetParent();
         if(sct_door){
            I3D_bbox bbox1;
            bbox1.min = bbox.min*frm->GetMatrix();
            bbox1.max = bbox.max*frm->GetMatrix();
            PI3D_sector sct = I3DCAST_SECTOR(sct_door);

            for(int i=sct->NumPortals(); i--; ){
               PI3D_portal prt = sct->GetPortal(i);
               if(prt->NumVertices()!=4) break;
               const S_vector *v = prt->GetVertices();
               for(int j=4; j--; ){
                  float d = (v[j]-bbox1.min).Magnitude();
                  if(d<.08f){
                     d = (v[j^2]-bbox1.max).Magnitude();
                     if(d<.08f){
                        portal = prt;
                        break;
                     }
                  }
               }
               if(portal)
                  break;
            }
            if(portal){
                        //find sector of door
               PI3D_sector sct=NULL;
               if(sct_door==portal->GetPrimSector()) sct=portal->GetConnectSector();
               else
               if(sct_door==portal->GetConnectSector()) sct = portal->GetPrimSector();
               if(sct){
                  portal->AddRef();
                  portal->SetOn(false);
                  object_dup = driver->CreateObject();
                  object_dup->SetMesh(obj->GetMesh());
                  object_dup->SetPos(obj->GetWorldPos());
                  object_dup->SetPivot(obj->GetPivot());
                  object_dup->SetRot(obj->GetWorldRot());
                  object_dup->SetScale(obj->GetWorldScale());
                  object_dup->Update();
                  object_dup->LinkTo(sct);
               }
            }
         }//else
            //OUTTEXT(C_fstring("door %s - cannot find portal", frm->GetName()));
         */
      }
   }

//----------------------------

   virtual int NumUsePositions(C_actor *asker = NULL) const{
      return (mode==MODE_CLOSED) ? 2 : 0;
   }

//----------------------------

   virtual const S_vector GetUsePos(int indx, bool abs_index_mode = false) const{

      S_vector v = lock_pos;
      v[2-long_axis] += (!indx ? -.75f : .75f);
      v *= frm_door->GetMatrix();
      return v;
   }

//----------------------------

   virtual float GetUsePosThresh() const{ return 1.5f; }

//----------------------------

   virtual bool GetUseDir(S_vector &ret, int indx, bool abs_index_mode = false) const{

      ret.Zero();
      ret[2-long_axis] = (!indx ? 1.0f : -1.0f);
      ret %= frm_door->GetMatrix();
      return true;
   }
   virtual float GetUseDirThresh() const{ return PI*.3f; }
   virtual dword CheckUseStay(dword suggest_stay_mode, int indx) const{ return SM_STAY; }

//----------------------------

   virtual int Use(LPC_actor user, dword prm, bool force_use_off){
      
      if(tab->ItemB(TAB_B_DOR_LOCK)){
         if(!user->cbProc(CB_USE_DOOR, 3)){
            mission.PlaySound("DoorLocked", lock_pos*frame->GetMatrix(), 1.0f, 6.0f);
            return 0;
         }
         mission.PlaySound("DoorUnlock", lock_pos*frame->GetMatrix(), 1.0f, 6.0f);
      }
      return user->cbProc(CB_USE_DOOR, 0, (dword)this);
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      ck->Write(&mode, sizeof(byte));
      ck->Write(&count, sizeof(count));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      ck->Read(&mode, sizeof(byte));
      ck->Read(&count, sizeof(count));
      if(mode==MODE_OPEN)
         ApplyAngle(1);

      return true;
   }

//----------------------------

   dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_TABLE_MESSAGE:
         switch(prm1){
         case TCM_FOCUS:
            switch(prm2){
            case TAB_I_DOR_ANGLE_CLOSE:
            case TAB_I_DOR_ANGLE_OPEN:
               ApplyAngle(prm2);
               break;
            }
            break;
         case TCM_CLOSE:
            ApplyAngle(TAB_I_DOR_ANGLE_CLOSE);
            break;
         }
         break;

      case CB_USE_DOOR:
         switch(prm1){
         case 1:              //open door
            Open((LPC_actor)prm2, true);
            break;
         case 2:              //check if closed
            return (mode==MODE_CLOSED);
         }
         break;
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      switch(mode){
      case MODE_OPENING:
      case MODE_CLOSING:
         {
            count -= tc.time;
            if(count <= 0) count=0;
            float f=(float)count/DOOR_OPEN_TIME;
            if(mode==MODE_OPENING) f=1.0f-f;
            Interpolate(f);
            if(!count){
               if(mode==MODE_OPENING){
                  mode = MODE_OPEN;
                  count = tab->ItemI(TAB_I_DOR_AUTO_CLOSE_TIME);
               }else mode=MODE_CLOSED;
            }
         }
         break;
      case MODE_OPEN:
         if(count && (count -= tc.time) <= 0){
            mode=MODE_CLOSING;
            count=DOOR_OPEN_TIME;
            PlaySound();
         }
         break;
      }
      return 0;
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      in_msg->Ascend();
      while(in_msg->GetChunkSize()){
         byte code = in_msg->GetBits(1);
         switch(code){
         case 0:              //open
            Open(NULL, false);
            
            S_tick_context tc;
            tc.Zero();
            tc.time = in_msg->GetLatency();
            Tick(tc);
            break;
         }
      }
      in_msg->Descend();
   }

//----------------------------

   virtual LPC_table GetTable(int index) const{
      if(index==0)
         return tab;
      return C_actor::GetTable(index);
   }

//----------------------------

   virtual class C_table_template *GetTemplate(int) const{ return &templ_door; }

   virtual void MissionSave(C_chunk *ck) const{ SaveTable(ck, tab); }
   virtual bool MissionLoad(C_chunk *ck, int table_id){ return LoadTable(ck, tab); }

//----------------------------

   virtual void TableUpdate(int index, bool not_load_prg){

      switch(index){
      case TAB_I_DOR_ANGLE_CLOSE:
      case TAB_I_DOR_ANGLE_OPEN:
         ApplyAngle(index);
         break;
      case -1:
         ApplyAngle(TAB_I_DOR_ANGLE_CLOSE);
         break;
      }
   }
};

//----------------------------

LPC_actor CreateDoor(C_game_mission &gm){

   return new C_door(gm);
}

//----------------------------
