#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"

//----------------------------

void TableToParticle(LPC_table tp, PI3D_particle sp);

static const S_vector little_y(.0f, .001f, .0f);
//----------------------------

                              //explosions over network:
                              // - primary explosion:
                              //    - evaluation of explosion
                              // - network explosion:
                              //    - display particles
                              // - sent over network:
                              //    - byte id
                              //    - S_vector position

class C_explosion: public C_actor{
   int count;
   bool exploded;
   float range;
   int power;
   LPC_actor creator;
   LPC_actor actor_reason;    //who has caused the explosion
   byte expl_id;
   bool vols_killed;

   int anim_time;

//----------------------------

   void InitModel(){

      //S_model_load s = {model, &mission};
      //model->EnumFrames(cbModelLoad, (dword)&s, ENUMF_DUMMY | ENUMF_USEMASK, "~0*");
      model->EnumFrames(cbAddVolumeToScene, (dword)mission.GetScene(), ENUMF_VOLUME);

      model->SetAnimation(0, NULL);
      mission.GetScene()->SetFrameSector(model);
                              //apply tables to model
      struct S_hlp{
         static I3DENUMRET I3DAPI cbPrt(PI3D_frame frm, dword dw){

            if(I3DCAST_VISUAL(frm)->GetVisualType() != I3D_VISUAL_PARTICLE)
               return I3DENUMRET_OK;

            PI3D_particle prt = (PI3D_particle)frm;

            LPC_table table = CreateTable();
            C_fstring str("tables\\expl\\%.1i\\%s.tab", dw, &frm->GetOrigName()[2]);
            bool b = table->Open((dword)(const char*)str, TABOPEN_FILENAME);
            if(b){
               TableToParticle(table, prt);
               prt->SetOn(true);
            }else{
               ErrReport(C_fstring("table open failed: %s", (const char*)str), editor);
            }
            table->Release();
                              //addiction - convert some to XZ mode
            if(frm->GetOrigName()[2]=='%')
               prt->SetProperty(PRT_MODE, 2);

            return I3DENUMRET_OK;
         }
      };
      model->EnumFrames(S_hlp::cbPrt, expl_id, ENUMF_VISUAL, NULL);

                              //setup lights and volumes
      struct S_hlp1{
         static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword sct){
            switch(frm->GetType()){
            case FRAME_LIGHT:
               {
                  PI3D_light lp = I3DCAST_LIGHT(frm);
                  ((PI3D_sector)sct)->AddLight(lp);
                  lp->SetMode(lp->GetMode() | I3DLIGHTMODE_DYNAMIC_LM);
               }
               break;
            case FRAME_VOLUME:
               {
                           //change to killing volumes
                  assert(mat_table->ItemE(TAB_E_MAT_KILL, 41)==1);
                  PI3D_volume vol = I3DCAST_VOLUME(frm);
                  vol->SetCollisionMaterial(41);
               }
               break;
            }
            return I3DENUMRET_OK;
         }
      };
      model->EnumFrames(S_hlp1::cbEnum, (dword)model->GetParent(), ENUMF_LIGHT | ENUMF_VOLUME, NULL);
   }
public:
   C_explosion(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_EXPLOSION),
      exploded(false),
      vols_killed(false),
      creator(NULL),
      actor_reason(NULL),
      anim_time(0),
      count(0)
   {
                              //allow init data to be NULL, used during LoadGame
      if(!init_data)
         return;

      const S_explosion_init &ei = *(S_explosion_init*)init_data;
      expl_id = ei.id;

      range = mat_table->ItemI(TAB_I_EXPL_RANGE, ei.id);
      power = mat_table->ItemI(TAB_I_EXPL_POWER, ei.id);
                              //load proper model
      frame = model = driver->CreateModel();

                              //play explosion sound
      const char *name = mat_table->ItemS(TAB_S12_EXPL_SOUND, ei.id, 12);
      if(!*name){
         name = "explose";
#ifdef EDITOR
         ReportActorError(C_fstring("no sound set for explosion #%i - using default", ei.id));
#endif
      }
      mission.PlaySound(name, ei.pos,
         tab_game_cfg->ItemF(TAB_F_GAME_SNDR_EXPL_MIN),
         tab_game_cfg->ItemF(TAB_F_GAME_SNDR_EXPL_MAX));

      if(I3D_FAIL(model_cache.Open(model, C_fstring("expl%.2i", ei.id), mission.GetScene(), 0, ErrReport, editor)))
         return;

      creator = ei.creator;
      actor_reason = ei.actor_reason;

      model->SetPos(ei.pos + little_y); //put little over ground:(for case that sector is exacly seme box as floor like in Greek4 before Kosta)
      model->Update();

      InitModel();

      {
                              //emit sound to listeners
         S_listen_source ls(LISTEN_EXPLOSION, ei.pos, 50.0f);
                              //don't send over net - shadows will emit it too
         mission.EmitListen(ls, tab_game_cfg->ItemF(TAB_F_GAME_SNDR_EXPL_MAX), false, creator);
      }


      network_actor = ei.netw_actor;
                              //send over network
      if(net && !network_actor){
         bool g = false;
         net_msg->AddWord(C_game_mission::NM_GAME_CREATE_EXPLOSION, g);
         net_msg->AddByte(ei.id, g);
         net_msg->AddVector(ei.pos, g);
      }
   }

//----------------------------

   virtual bool IsOnMap() const{ return true; }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      SG_SaveModelInfo(ck);
                           //write data
      ck->Write(&count, sizeof(count));
      ck->Write(&exploded, sizeof(exploded));
      ck->Write(&range, sizeof(range));
      ck->Write(&power, sizeof(power));
      ck->Write(&expl_id, sizeof(expl_id));
                        //write animation time
      ck->Write(&anim_time, sizeof(anim_time));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      SG_LoadModelInfo(ck);
                           //read data
      ck->Read(&count, sizeof(count));
      ck->Read(&exploded, sizeof(exploded));
      ck->Read(&range, sizeof(range));
      ck->Read(&power, sizeof(power));
      ck->Read(&expl_id, sizeof(expl_id));
      InitModel();
      ck->Read(&anim_time, sizeof(anim_time));
      for(int i=0; i<anim_time; i+=1000){
         model->Tick(1000);
      }
      model->Tick(anim_time%1000);

      return true;
   }

//----------------------------

   dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){
      case CB_SETPARAM:
         switch(prm1){
         case 2:              //get reason
            *(LPC_actor*)prm2 = actor_reason;
            break;
         }
         break;
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      anim_time += tc.time;
      model->Tick(tc.time);
                              //make check only time-to-time
      if((count += tc.time) > tab_game_cfg->ItemI(TAB_I_GAME_EXPL_DELAY)){
         count = 0;

         if(!exploded){
            exploded = true;

            if(!network_actor){
                              //send explosion message to all actors nearby
               struct S_hlp1{
                  S_vector pos;
                  float range;
                  int power;
                  LPC_actor a;
                  static bool cbAct(LPC_actor a, dword t1){
                     const S_hlp1 &t = *(S_hlp1*)t1;
                     S_vector apos;
                     if(a!=t.a){  //don't test itself
                        if(a->MayExplode()){
                           if(a->GetCenterPos(apos)){
                              float dist = (apos-t.pos).Magnitude();
                              if(dist < t.range){
                                       //let him to determine explision power
                                 float f = IntAsFloat(a->cbProc(CB_GET_EXPLODE_RATIO, (dword)t.a, (dword)&t.pos));
                                 int p = (float)t.power * (1.0f - dist/t.range) * f;
                                 if(p)
                                    a->Explode(t.a, p);
                              }
                           }
                        }
                     }
                     return true;
                  }
               } hlp;
               hlp.pos = model->GetWorldPos();
               //hlp.pos.y += 1.0f;
               hlp.range = range;
               hlp.power = power;
               hlp.a = this;
                                 //our volumes off during explosion
               SetVisible(false, true, false);
               if(creator) creator->SetVisible(false, true, false);
               mission.EnumActors(S_hlp1::cbAct, (dword)&hlp);
               SetVisible(true, true, false);
               if(creator) creator->SetVisible(true, true, false);
            }
                              //shake camera a bit
            if(mission.game_cam.GetDistanceMode() != 0){
               const S_vector &pos = mission.game_cam.GetCameraPos();
               float dist = (pos-frame->GetWorldPos()).Magnitude();
               float max_range = range*2;
               if(dist < max_range){
                  float power = dist/max_range;
                  int type = (power<.5f) ? 0 : (power<.75f) ? 1 : 2;
                  mission.game_cam.SetAnim(C_fstring("cam_shk%i.i3d", type));
               }
            }
         }
                              //check if all prts off
         struct S_hlp{
            static I3DENUMRET I3DAPI cbPrt(PI3D_frame frm, dword dw){

               if(I3DCAST_VISUAL(frm)->GetVisualType() != I3D_VISUAL_PARTICLE)
                  return I3DENUMRET_OK;
               PI3D_particle prt = (PI3D_particle)frm;
               if(prt->GetProperty(PRT_NUMELEMENTS)){
                  *(int*)dw |= 1;
                  //return I3DENUMRET_CANCEL;
                  if(prt->GetProperty(PRT_DRAW_MODE)==2)
                     *(int*)dw |= 2;
               }
               return I3DENUMRET_OK;
            }
            static I3DENUMRET I3DAPI cbLight(PI3D_frame frm, dword sct){
               PI3D_light lp = I3DCAST_LIGHT(frm);
               ((PI3D_sector)sct)->DeleteLight(lp);
               return I3DENUMRET_OK;
            }
            static I3DENUMRET I3DAPI cbVol(PI3D_frame frm, dword sct){
               frm->SetOn(false);
               return I3DENUMRET_OK;
            }
         };
         int on = 0;
         model->EnumFrames(S_hlp::cbPrt, (dword)&on, ENUMF_VISUAL, NULL);

         if(!on){
                              //un-setup lights
            model->EnumFrames(S_hlp::cbLight, (dword)model->GetParent(), ENUMF_LIGHT, NULL);
            mission.DestroyActor(this);
         }else
         if(!(on&2)){
            if(!vols_killed){
               model->EnumFrames(S_hlp::cbVol, 0, ENUMF_VOLUME, NULL);
               vols_killed = true;
            }
         }
      }
      return 0;
   }
};

//----------------------------

LPC_actor CreateExplosion(C_game_mission &gm, void *data){

   return new C_explosion(gm, data);
}

//----------------------------
