#include "all.h"
#include "net.h"
#include "actors.h"
#include "smokeray.h"
#include "tables.h"


#define FLARE_EMIT_SOUND_TIME 1000

void TableToParticle(LPC_table tp, PI3D_particle sp);

//----------------------------
                              //flare over network:
                              // - sent over network:
                              //    - S_vector pos
                              //    - S_vector dir

class C_flare: public C_actor{
   enum{
      MODE_FLY, MODE_LIE
   } mode;
   S_vector pos, dir;
   PI3D_sector curr_sct;
   C_smart_ptr<I3D_light> light[2]; //[vertex|lmap]
   int counter;
   int anim_time;
   int emit_snd_cnt;
   S_vector emit_snd_pos;
   float scene_lightness;
   C_smoke_ray *smoke_ray;

//----------------------------

   void SetupModel(){

      for(int i=0; i<2; i++){
         static const char *lname[] = {"vertex", "lightmap" };
         light[i] = I3DCAST_LIGHT(model->FindChildFrame(lname[i], ENUMF_LIGHT));
         assert(light[i]);
         if(light[i])
            light[i]->SetMode(!i ? I3DLIGHTMODE_VERTEX : I3DLIGHTMODE_DYNAMIC_LM);
      }
      model->SetAnimation(0, NULL);

      PI3D_particle prt = (PI3D_particle)model->FindChildFrame(NULL, ENUMF_VISUAL);
      if(prt){
         LPC_table table = CreateTable();
         int i = table->Open((dword)"tables\\flare.tab", TABOPEN_FILENAME);
         if(i){
            TableToParticle(table, prt);
         }
         table->Release();
      }

      delete smoke_ray;
      S_smoke_ray_init si;
      si.scene = mission.GetScene();
      si.start_pos = pos;
      si.start_dir = dir;
      si.tail_fade_time = 4000;
      si.tile_tile = 3.0f;
      si.dest_scale = 10.0f;
      si.tail_name = "tail03";
      smoke_ray = GreateSmokeRay(si);
   }
public:

   C_flare(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_FLARE),
      curr_sct(NULL),
      smoke_ray(NULL),
      mode(MODE_FLY),
      anim_time(0),
      emit_snd_cnt(0),
      scene_lightness(1.0f),
      counter(2000)
   {
                              //allow init data to be NULL, used during LoadGame
      if(!init_data)
         return;

      const S_flare_init &fi = *(S_flare_init*)init_data;

      pos = fi.pos;
      dir = fi.dir;
      emit_snd_pos = pos;
#if 1                         //get aproximate lightness at creation pos,
                              //it is used for scale emit sound.
      GetSceneLightness();
      /*S_vector color = mission.GetScene()->GetLightness(pos);
      scene_lightness = color[0]*.3f + color[1]*.6f + color[2]*.1f;
      scene_lightness -= .2f; //minimal light
      scene_lightness = Max(.0f, Min(1.0f, scene_lightness * .5f));
      //OUTTEXT(C_fstring("scene lightness = %.2f", scene_lightness));*/
#endif

      model = driver->CreateModel();
      frame = model;
      model_cache.Open(model, "flare", mission.GetScene(), 0, ErrReport, editor);
      mission.PlaySound("flare", model, 10, 100);

      frame->SetPos(pos);
      frame->Update();
      mission.GetScene()->SetFrameSector(frame);

      SetupModel();

      network_actor = fi.netw_actor;
                                 //send over network
      if(net && !network_actor){
         bool g = true;
         net_msg->AddWord(C_game_mission::NM_GAME_CREATE_FLARE, g);
         net_msg->AddVector(fi.pos, g);
         net_msg->AddVector(fi.dir, g);
         net_msg->AddWord(fi.shooter ? fi.shooter->GetNetID() : 0, g);
      }
                              //emit sound of flare fire
      S_listen_source ls = S_listen_source(LISTEN_SHOOT, pos, 30);
      mission.EmitListen(ls, tab_game_cfg->ItemF(TAB_F_GAME_SNDR_GUN_MAX), false, fi.shooter);
   }

//----------------------------

   inline void GetSceneLightness(){

      assert(mission.GetScene());
      S_vector color = mission.GetScene()->GetLightness(pos);
      scene_lightness = color[0]*.3f + color[1]*.6f + color[2]*.1f;
      scene_lightness -= .2f; //minimal light
      scene_lightness = Max(.0f, Min(1.0f, scene_lightness * .5f));
      //OUTTEXT(C_fstring("scene lightness = %.2f", scene_lightness));*/
   }


//----------------------------

   ~C_flare(){
      delete smoke_ray;
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      anim_time += tc.time;
      model->Tick(tc.time);

      bool animated = true;
      if(!model->NumInterpolators()){
         animated = false;
      }else{
         const PI3D_interpolator *intps = model->GetInterpolators();
                              //check only non-particle interpolators
         for(int i=model->NumInterpolators(); i--; ){
            PI3D_frame frm = intps[i]->GetFrame();
            if(frm->GetType()!=FRAME_VISUAL)
               break;
         }
         animated = (i!=-1);
      }
      if(!animated){
         bool dtr = true;
         PI3D_particle prt = (PI3D_particle)model->FindChildFrame(NULL, ENUMF_VISUAL);
         if(prt){
            prt->SetProperty(PRT_SETCOUNT, -1);
            dtr = (!prt->GetProperty(PRT_NUMELEMENTS));
         }
         if(dtr && smoke_ray->IsOver()){
            mission.DestroyActor(this);
            return -1;
         }
      }
      if(mode==MODE_LIE){
         smoke_ray->UpdateFade(tc.time);
         return -2;
      }

      emit_snd_cnt = Max(0, emit_snd_cnt - tc.time);
      if(!emit_snd_cnt){
         emit_snd_cnt = FLARE_EMIT_SOUND_TIME;
         S_listen_source ls;
         ls.type = LISTEN_FLARE;
         ls.pos = emit_snd_pos; //use position of creation for emit sound
         ls.power = .5f + 5.0f * (1.0f-scene_lightness);
                           //don't send over network - shadow shots will emit sound
         mission.EmitListen(ls, 200.0f, false, NULL);
      }

      float tsec = (float)tc.time*.001f;
      S_vector delta = dir * tsec;
      if(counter && (counter -= tc.time)<0){
         counter = 0;
                              //start falling slowly down
         dir.y = Min(0.0f, dir.y);
         dir.x *= .02f;
         dir.z *= .02f;
         smoke_ray->Finish();
      }

      if(counter){
                              //slow down
         dir *= 1.0f - (.2f * tsec);
      }else{
                              //fly down slowly
         dir.x *= 1.0f - .2f * tsec;
         dir.z *= 1.0f - .2f * tsec;
         dir.y = Max(-.75f, dir.y - 1.0f*tsec);
      }

                              //check collisions
      I3D_collision_data cd(pos, delta, 0, NULL, FlyColResp);
      if(mission.GetScene()->TestCollision(cd)){
         S_vector ret_norm = cd.GetHitNormal();
         ret_norm.Normalize();

         if(ret_norm.y < .5f){
            pos = cd.GetHitPosition() + ret_norm * .05f;
            float f = dir.Magnitude();
            if(f<1.5f){
               mode = MODE_LIE;
            }else{
               dir.Zero();
               if(dir.y > 0.0f)
                  dir.y *= .5f;
            }
            smoke_ray->Finish();
         }
      }else{
#if 0
         DebugLine(pos, pos+delta, 0);
#endif
         pos += delta;
      }

      frame->SetPos(pos);
      frame->Update();
      mission.GetScene()->SetFrameSector(frame);

      if(animated){
         smoke_ray->UpdateTrajectory(pos, dir);
      }
      smoke_ray->UpdateFade(tc.time);
      
      PI3D_frame frm = model->GetParent();
      PI3D_sector sct = NULL;
      if(frm->GetType()==FRAME_SECTOR)
         sct = I3DCAST_SECTOR(frm);
      if(!sct)
         sct = mission.GetScene()->GetPrimarySector();
      if(curr_sct != sct){
         for(int i=2; i--; ){
            if(light[i]){
               if(curr_sct)
                  curr_sct->DeleteLight(light[i]);
               sct->AddLight(light[i]);
            }
         }
         curr_sct = sct;
      }
      dir.y -= tsec * 5.0f;
      return 0;
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      SG_SaveModelInfo(ck);
                        //write data
      ck->Write(&mode, sizeof(byte));
      ck->Write(&pos, sizeof(pos));
      ck->Write(&dir, sizeof(dir));
      ck->Write(&counter, sizeof(counter));
      ck->Write(&anim_time, sizeof(anim_time));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      if(ck->Size()){
         SG_LoadModelInfo(ck);
                           //read data
         ck->Read(&mode, sizeof(byte));
         ck->Read(&pos, sizeof(pos));
         ck->Read(&dir, sizeof(dir));
         ck->Read(&counter, sizeof(counter));
         ck->Read(&anim_time, sizeof(anim_time));
         SetupModel();
         model->Tick(anim_time);
         emit_snd_pos = pos;
         GetSceneLightness();
      }

      return true;
   }
};

//----------------------------

LPC_actor CreateFlare(C_game_mission &gm, void *data){

   return new C_flare(gm, data);
}

//----------------------------
