#include "all.h"
#include <math\ease.hpp>


//----------------------------
#define PLR_MENU_COMPACTED(i) S_vector2(.06f+i*.15f, .72f)
#define PLR_MENU_EXPANDED(i)  S_vector2(.03f, .7125f-i*.045f)

                              //time axis length (on screen)
#define TIME_RANGE 60000
#define TIME_AXIS_LENGTH 90000
#define AXIS_NUM_TIME 10000

#define COMPACT_TIME 500

//----------------------------
// Check if a point is in 2D rectangle.
inline bool IsPtInRect(float x, float y, const float rc[4]){
   return (x>=rc[0] && y>=rc[1] && x<rc[2] && y<rc[3]);
}

//----------------------------
//----------------------------
                              //players menu

#define HEALTH_BAR_DELTA_X -.027f
#define HEALTH_BAR_DELTA_Y -.02025f

class C_plr_menu{
   C_game_menu &game_menu;

                              //main pannel
   C_smart_ptr<C_poly_sprite> spr_face;
   C_smart_ptr<C_poly_sprite> health_bar;
   dword health_bar_color;
   float curr_health;
   bool zero_health;
   C_fade<S_vector2> fade_pos_pannel;
   C_fade<S_fade_color> fade_col_pannel;
                              //time track
   C_smart_ptr<C_poly_sprite> spr_time_axis;
   C_fade<S_fade_color> fade_track;
   bool visible;
                              //list of icons
   struct S_icon: public C_unknown{
      C_smart_ptr<C_poly_sprite> spr;
      int time;

      S_icon(C_poly_sprite *s, int t=0):
         spr(s), time(t)
      {}
   };
   vector<C_smart_ptr<S_icon> > icon_list;
   C_fade<S_vector2> fade_pos_icon0;

   dword pos;
   bool active;
   C_game_menu::E_MENU_STATE state;

//----------------------------

   S_vector2 GetPannelPos(C_game_menu::E_MENU_STATE state){

      switch(state){
      case C_game_menu::STATE_COMPACT:
         return PLR_MENU_COMPACTED(pos);
      default:                //expand
         return PLR_MENU_EXPANDED(pos);
      }
   }

//----------------------------

   void AddTrackBar(){

      int time_offset = game_menu.GetTimeOffset();
      float f;
      game_menu.IconX(game_menu.GetTimeOffset(), f);
      spr_time_axis = new C_poly_sprite(f, GetPannelPos(C_game_menu::STATE_EXPAND).y, 0.0f, &icons_img, 4);
      spr_time_axis->SetColor(0);
      spr_time_axis->Release();

      float f1;
      game_menu.IconX(time_offset+TIME_RANGE, f1);
      spr_time_axis->SetScreenSize(f1-f, .0075f);

      f = (float)(time_offset%5000) / 5000.0f;
      spr_time_axis->SetTileX(f, f+TIME_RANGE/5000);
   }

//----------------------------

   dword GetPannelColor(){
      switch(state){
      case C_game_menu::STATE_COMPACT:
         //return active ? 0xe0ffffff : 0xa0c0c0ff;
         return active ? 0xffffffff : 0x60c0c0ff;
      default:                //expand
         return active ? 0xffffffff : 0x60a0a0a0;
      }
   }

//----------------------------

   dword GetTextColor(){
      switch(state){
      case C_game_menu::STATE_COMPACT:
         return active ? 0xc0ffffff : 0x80c0c0ff;
      default:                //expand
         return active ? 0xffffffff : 0xc0ffffff;
      }
   }
public:
   C_sprite_image face_img;
public:
   C_plr_menu(C_game_menu &gm):
      game_menu(gm),
      zero_health(false),
      curr_health(0.0f),
      health_bar_color(0xffffff),
      active(false),
      visible(true)
   {}

//----------------------------

   void Init(dword p1, C_game_menu::E_MENU_STATE st){

      state = st;
      pos = p1;
      int i;

      AddTrackBar();
      S_vector2 pannel_pos = GetPannelPos(state);

      switch(state){
      case C_game_menu::STATE_COMPACT:
         spr_time_axis->SetOn(false);
         spr_time_axis->SetColor(0x00ffffff);
         for(i=0; i<icon_list.size(); i++){
            icon_list[i]->spr->SetOn(false);
            icon_list[i]->spr->SetColor(0x00ffffff);
         }
         break;
      default:                //expand
         spr_time_axis->SetColor(0x00ffffff);
         fade_track.Setup(300, spr_time_axis->GetColor(), 0x80ffffff);
         for(i=0; i<icon_list.size(); i++){
            icon_list[i]->spr->SetColor(0x00ffffff);
         }
      }
      SetFace();
      spr_face->SetColor(0x00ffffff);
      fade_col_pannel.Setup(200, spr_face->GetColor(), GetPannelColor());

      health_bar = new C_poly_sprite(pannel_pos.x + HEALTH_BAR_DELTA_X, pannel_pos.y + HEALTH_BAR_DELTA_Y, .0f, &icons_img, 64);
      health_bar->Release();
      health_bar->SetColor(health_bar_color);
      SetHealth(1.0f);
   }

//----------------------------

   void ResetScreen(){

                              //face
      SetFace();
                              //health
      S_vector2 pannel_pos = GetPannelPos(state);
      health_bar = new C_poly_sprite(pannel_pos.x + HEALTH_BAR_DELTA_X, pannel_pos.y + HEALTH_BAR_DELTA_Y, .0f, &icons_img, 64);
      health_bar->Release();
      SetHealth(curr_health);
                              //icons
      for(int i=0; i<icon_list.size(); i++){
         S_icon *ic = icon_list[i];
         const S_vector2 &pos = ic->spr->GetPos();
         dword indx = ic->spr->GetSpriteIndex();

         C_poly_sprite *spr = new C_poly_sprite(pos.x, pos.y, .0375f, &icons_img, indx, 10);
         spr->SetOn(ic->spr->IsOn());
         ic->spr = spr;
         spr->Release();
      }
      AddTrackBar();
   }

//----------------------------

   void SetFace(){

      S_vector2 pannel_pos = GetPannelPos(state);
      spr_face = new C_poly_sprite(pannel_pos.x, pannel_pos.y, .045f, &face_img, 0);
      spr_face->Release();
   }

//----------------------------

   void SetState(C_game_menu::E_MENU_STATE st){

      int i;
      switch(state=st){
      case C_game_menu::STATE_EXPAND:
         spr_time_axis->SetOn(true);
         fade_track.Setup(COMPACT_TIME, spr_time_axis->GetColor(), 0xffffffff);

         for(i=0; i<icon_list.size(); i++)
            icon_list[i]->spr->SetOn(true);
         
         if(icon_list.size()){
            float f;
            game_menu.IconX(icon_list[0]->time, f);
            fade_pos_icon0.Setup(COMPACT_TIME, icon_list[0]->spr->GetPos(),
               S_vector2(f, GetPannelPos(C_game_menu::STATE_EXPAND).y));
         }
         break;
      case C_game_menu::STATE_COMPACT:
         fade_track.Setup(COMPACT_TIME, spr_time_axis->GetColor(), 0x00ffffff);

         if(icon_list.size())
            fade_pos_icon0.Setup(COMPACT_TIME, icon_list[0]->spr->GetPos(),
               GetPannelPos(state)+S_vector2(+.06f, 0));
         break;
      }
      fade_pos_pannel.Setup(COMPACT_TIME, spr_face->GetPos(), GetPannelPos(state));
      fade_col_pannel.Setup(COMPACT_TIME, spr_face->GetColor(), GetPannelColor());

   }

//----------------------------

   void SetHealth(float f){

      curr_health = f;
      if(f<MRG_ZERO){
         zero_health = true;
         health_bar->SetOn(false);
         return;
      }
      zero_health = false;
      f = Max(0.0f, Min(1.0f, f));

      assert(health_bar);
      health_bar->SetScreenSize(.0075f, .0405f*f);
                              //determine color of the bar
      static const struct S_bc{
         float val;
         float color[3];      //RGB values
      } bar_colors[]={
         //{1.0f, 1.0f, 1.0f, 0.0f},
         {1.0f, 0.0f, 1.0f, 0.0f},
         {0.5f, 1.0f, 1.0f, 0.0f},
         {0.0f, 1.0f, 0.0f, 0.0f}
      };
      for(int i=0; ; i++)
      if(f>=bar_colors[i].val){
         const S_bc &lo = bar_colors[i];
         S_vector c = (S_vector&)lo.color[0];
         if(i){
            const S_bc &hi = bar_colors[i-1];
            float f1 = f - lo.val;
            float delta = hi.val-lo.val;
            f1 = f / delta;
            S_vector c_delta(hi.color[0]-lo.color[0], hi.color[1]-lo.color[1], hi.color[2]-lo.color[2]);
            c += c_delta*f1;
         }
         health_bar_color =
            byte(c[0]*255.0f)<<16 |
            byte(c[1]*255.0f)<<8 |
            byte(c[2]*255.0f);
         break;
      }
      dword old_color = health_bar->GetColor();
      health_bar->SetColor((old_color&0xff000000) | health_bar_color);
   }

//----------------------------

   void SetActive(bool b){

      active = b;
      fade_col_pannel.Setup(200, spr_face->GetColor(), GetPannelColor());
   }

//----------------------------

   bool IsActive() const{ return active; }

//----------------------------

   void Tick(int time){
      int i;
                              //fade track bar
      if(!fade_track.IsEnd()){
         dword color = fade_track.Tick(time);
         spr_time_axis->SetColor(color);
         if(fade_track.IsEnd() && state==C_game_menu::STATE_COMPACT)
            spr_time_axis->SetOn(false);
         for(i=1; i<icon_list.size(); i++){
            icon_list[i]->spr->SetColor(color);
            if(fade_track.IsEnd() && state==C_game_menu::STATE_COMPACT)
               icon_list[i]->spr->SetOn(false);
         }
      }
                              //move pannel
      if(!fade_pos_pannel.IsEnd()){
         S_vector2 v=fade_pos_pannel.Tick(time);
         spr_face->SetPos(v.x, v.y);
         health_bar->SetPos(v.x+HEALTH_BAR_DELTA_X, v.y+HEALTH_BAR_DELTA_Y);
      }
                              //fade pannel
      if(!fade_col_pannel.IsEnd()){
         dword color=fade_col_pannel.Tick(time);
         spr_face->SetColor(color);
         health_bar->SetColor((color&0xff000000) | health_bar_color);
         if(icon_list.size())
            icon_list[0]->spr->SetColor(color);
      }
      if(!fade_pos_icon0.IsEnd() && icon_list.size()){
         S_vector2 v=fade_pos_icon0.Tick(time);
         icon_list[0]->spr->SetPos(v.x, v.y);
      }
   }

//----------------------------

   void Render(){

      spr_time_axis->Render();
      spr_face->Render();
      health_bar->Render();
      for(int i=icon_list.size(); i--; )
         icon_list[i]->spr->Render();
   }

//----------------------------

   void SetPrgList(const vector<S_prg_item> &plist, bool pos_only){

      int i;
      if(!pos_only){
                              //clear current icon list
         icon_list.clear();
      }

                              ////back-to-front, to get nice overlapping
      //for(i=plist.size(); i--; )
      for(i=0; i<plist.size(); i++){
         S_vector2 v;
         bool on=true;
         switch(state){
         case C_game_menu::STATE_EXPAND:
            on = game_menu.IconX(plist[i].show_time, v.x);
            v.y = GetPannelPos(C_game_menu::STATE_EXPAND).y;
            break;
         case C_game_menu::STATE_COMPACT:
            if(!i){
               v=GetPannelPos(state);
               v.x += .06f;
            }else{
               game_menu.IconX(plist[i].show_time, v.x);
               v.y=GetPannelPos(C_game_menu::STATE_EXPAND).y;
               on=false;
            }
            break;
         }
         if(pos_only){
            C_poly_sprite *spr = icon_list[i]->spr;
            assert(spr);
            spr->SetPos(v.x, v.y);
            spr->SetOn(on);
         }else{
            C_poly_sprite *spr = new C_poly_sprite(v.x, v.y, .0375f, &icons_img, plist[i].prg_item + 16, 10);
            if(!on)
               spr->SetOn(false);
            C_plr_menu::S_icon *icon = new C_plr_menu::S_icon(spr);
            icon_list.push_back(icon);
            spr->Release();
            icon->Release();
            icon_list.back()->time = plist[i].show_time;
         }
      }
   }

//----------------------------

   bool HitTest(float sx, float sy, int &sub_id){
                              //check face icon
      float rc[4];
      spr_face->GetExtent(rc[0], rc[1], rc[2], rc[3]);
      if(IsPtInRect(sx, sy, rc)){
         sub_id=-1;
         return true;
      }
                              //check all program icons
      for(int i=0; i<icon_list.size(); i++){
         icon_list[sub_id=i]->spr->GetExtent(rc[0], rc[1], rc[2], rc[3]);
         if(IsPtInRect(sx, sy, rc))
            return true;
      }
                              //check time axis
      spr_time_axis->GetExtent(rc[0], rc[1], rc[2], rc[3]);
      if(IsPtInRect(sx, sy, rc)){
         sub_id = -2;
         return true;
      }
      return false;
   }

//----------------------------

   void SetPrgKeyColor(dword key_id, dword color){

      assert(key_id<icon_list.size());
      icon_list[key_id]->spr->SetColor(color);
   }

//----------------------------

   dword GetPrgKeyColor(dword key_id){

      assert(key_id<icon_list.size());
      return icon_list[key_id]->spr->GetColor();
   }

//----------------------------

   void TimeAxisScrolled(){

                              //reposition all prg icons
      for(int i=0; i<icon_list.size(); i++){
         S_vector2 v;
         bool on=true;
         switch(state){
         case C_game_menu::STATE_EXPAND:
            on = game_menu.IconX(icon_list[i]->time, v.x);
            v.y = GetPannelPos(C_game_menu::STATE_EXPAND).y;
            break;
         case C_game_menu::STATE_COMPACT:
            if(!i)
               continue;
            game_menu.IconX(icon_list[i]->time, v.x);
            v.y = GetPannelPos(C_game_menu::STATE_EXPAND).y;
            on = false;
            break;
         }
         icon_list[i]->spr->SetPos(v.x, v.y);
         icon_list[i]->spr->SetOn(on);
      }
   }
};

//----------------------------
//----------------------------
#define COMPASS_SIZE .075f
#define INV_CAM_POS_Z -40
                              //game menu
C_game_menu::C_game_menu(C_poly_text &t):
   texts(t),
   num_players(0),
   state(STATE_COMPACT),
   curr_inv(NULL),
   time_offset(0),
   curr_comp_angle(0.0f)
{
   spr_compass = new C_poly_sprite(1.0f-COMPASS_SIZE*.6f, COMPASS_SIZE*.6f, COMPASS_SIZE, &icons_img, 1);
   spr_compass->Release();
   memset(pmenu, 0, sizeof(pmenu));
}

//----------------------------

C_game_menu::~C_game_menu(){

   SetInventory(NULL);
   inv_scene = NULL;
   for(int i=0; i<MAX_PLAYERS; i++)
      delete pmenu[i];
}

//----------------------------

void C_game_menu::ResetScreen(){

                              //reset menu items for new resolution

   spr_compass = new C_poly_sprite(1.0f-COMPASS_SIZE*.6f, COMPASS_SIZE*.6f, COMPASS_SIZE, &icons_img, 1);
   spr_compass->Release();

   for(int i=MAX_PLAYERS; i--; ){
      if(pmenu[i]){
         pmenu[i]->ResetScreen();
      }
   }
   if(inv_scene){
      inv_scene->SetViewport(igraph->Scrn_sx()*.68f, igraph->Scrn_sy()*.05f,
         igraph->Scrn_sx()*.98f, igraph->Scrn_sy()*(.05f + .5f), true);
   }
}

//----------------------------

bool C_game_menu::IconX(int time, float &f){

   if(time<time_offset || time>(time_offset+TIME_RANGE))
      return false;
   f = .075f + ((float)(time-time_offset)/TIME_RANGE) * .9f;
   return true;
}

//----------------------------

bool C_game_menu::CreateAxisNums(){

   if(axis_nums.size())
      return false;
   for(int i=0; i<=TIME_AXIS_LENGTH; i += AXIS_NUM_TIME){
      float f;
      if(!IconX(i, f))
         continue;
      char buf[32];
      sprintf(buf, "%u", i/1000);
      C_text *txt = texts.CreateText(buf, -1, f, .7462f, .019f, TEXT_JUSTIFY_CENTER);
      axis_nums.push_back(txt);
      txt->Release();
   }
   return true;
}

//----------------------------

void C_game_menu::SetInventory(C_inventory *inv){

   if(curr_inv==inv) return;
   if(!inv_scene && inv){
                              //init inventory scene, camera, ...
      inv_scene = driver->CreateScene();
      inv_scene->Release();

      inv_scene->SetViewport(igraph->Scrn_sx()*.68f, igraph->Scrn_sy()*.05f,
         igraph->Scrn_sx()*.98f, igraph->Scrn_sy()*(.05f + .5f), true);

      PI3D_camera cam = driver->CreateCamera();
      cam->SetName("inventory camera");
      inv_scene->AddFrame(cam);
      cam->SetPos(S_vector(0, 0, INV_CAM_POS_Z));
      cam->SetDir(S_vector(0, 0, 1), 0.0);
      cam->SetFOV(3.3f*GetPI()/180.0f);
      cam->SetRange(-INV_CAM_POS_Z - 2.0f, -INV_CAM_POS_Z + 2.0f);
      cam->Update();
      inv_scene->SetActiveCamera(cam);
      cam->Release();
      {
         PI3D_light lp = driver->CreateLight();
         lp->SetLightType(I3DLIGHT_DIRECTIONAL);
         lp->SetPos(S_vector(0, 0, 0));
         lp->SetDir(S_vector(1, -1, 3), 0.0f);
         lp->SetColor(S_vector(1, 1, 1));
         lp->SetOn(true);
         lp->LinkTo(inv_scene->GetPrimarySector());
         inv_scene->GetPrimarySector()->AddLight(lp);
         lp->Update();
         inv_scene->AddFrame(lp);
         lp->Release();
      }
      {
         PI3D_light lp = driver->CreateLight();
         lp->SetLightType(I3DLIGHT_AMBIENT);
         lp->SetPower(.3f);
         inv_scene->AddFrame(lp);
         inv_scene->GetPrimarySector()->AddLight(lp);
         lp->Release();
      }
      PI3D_model mod = driver->CreateModel();
      model_cache.Open(mod, "inv_frm", inv_scene, 0, ErrReport, editor);
      mod->SetPos(S_vector(0, 0, 1.9f));
      mod->LinkTo(inv_scene->GetPrimarySector());
      mod->Update();
      //mod->StopAnimation(0);
      mod->SetAnimation(0, NULL);
      inv_scene->AddFrame(mod);
      mod->Release();
   }

   if(curr_inv)
      curr_inv->ReleaseModels();
   curr_inv = inv;
   if(curr_inv)
      curr_inv->LoadModels(inv_scene, texts);
}

//----------------------------

void C_game_menu::Render(){

   bool was_zb = driver->GetState(RS_USEZB);
   if(was_zb) driver->SetState(RS_USEZB, false);

   for(int i=0; i<MAX_PLAYERS; i++){
      if(pmenu[i])
         pmenu[i]->Render();
   }
   spr_compass->Render();

   if(was_zb) driver->SetState(RS_USEZB, true);
   
   if(curr_inv){
      bool clr = driver->GetState(RS_CLEAR);
      if(clr) driver->SetState(RS_CLEAR, false);
      inv_scene->Render(0);
      driver->SetViewport(0, 0, igraph->Scrn_sx(), igraph->Scrn_sy());
      if(clr) driver->SetState(RS_CLEAR, true);
   }
}

//----------------------------

void C_game_menu::SetState(E_MENU_STATE st){

   if(state==st) return;
   state=st;
   for(int i=0; i<MAX_PLAYERS; i++)
      if(pmenu[i])
         pmenu[i]->SetState(state);

   switch(st){
   case STATE_COMPACT:
      time_num_fade.Setup(COMPACT_TIME, axis_nums[0]->GetColor(), 0x00ffffff);
      break;
   case STATE_EXPAND:
      i=CreateAxisNums();
      time_num_fade.Setup(COMPACT_TIME, i ? 0x00ffffff :
         axis_nums[0]->GetColor(), 0xffffffff);
      break;
   }
}

//----------------------------

void C_game_menu::SetCompassAngle(float a){

   if(curr_comp_angle!=a){
      curr_comp_angle = a;
      spr_compass->Rotate(a);
   }
}

//----------------------------

void C_game_menu::Tick(const S_tick_context &tc){

   int i;
   for(i=0; i<MAX_PLAYERS; i++)
      if(pmenu[i])
         pmenu[i]->Tick(tc.time);

   if(curr_inv){
      inv_scene->Tick(tc.time);
      curr_inv->Tick(tc.time);
   }
   if(!time_num_fade.IsEnd()){
      dword color=time_num_fade.Tick(tc.time);
      if(time_num_fade.IsEnd() && state==STATE_COMPACT)
         axis_nums.clear();
      else
      for(i=0; i<axis_nums.size(); i++)
         axis_nums[i]->SetColor(color);
   }
}

//----------------------------

dword C_game_menu::AddPlayerMenu(dword face_id){

   if(num_players==MAX_PLAYERS)
      return -1;
                              //find free menu
   for(int i=0; ; i++){
      if(!pmenu[i])
         break;
   }
   pmenu[i] = new C_plr_menu(*this);
                              //load proper face for human
   bool b = pmenu[i]->face_img.Init(driver, "tables\\face.prj", C_fstring("faces\\face_%.2i.png", face_id));
   if(!b){
      b = pmenu[i]->face_img.Init(driver, "tables\\face.prj", "faces\\face_00.png");
      assert(b);
   }
   pmenu[i]->Init(num_players, state);
   num_players++;
   return i;
}

//----------------------------

void C_game_menu::DestroyPlayerMenu(dword plr_id){

   if(plr_id >= MAX_PLAYERS)
      return;

   delete pmenu[plr_id];
   pmenu[plr_id] = NULL;
   --num_players;
}

//----------------------------

void C_game_menu::SetPlayerFace(dword plr_id, dword face_id){

   pmenu[plr_id]->face_img.Close();
                              //load proper face for human
   bool b = pmenu[plr_id]->face_img.Init(driver, "tables\\face.prj", C_fstring("faces\\face_%.2i.png", face_id));
   if(!b){
      b = pmenu[plr_id]->face_img.Init(driver, "tables\\face.prj", "faces\\face_00.png");
      assert(b);
   }
   if(pmenu[plr_id])
      pmenu[plr_id]->SetFace();
}

//----------------------------

void C_game_menu::SetPrgList(dword plr_id, const vector<S_prg_item> &pl, bool pos_only){

   if(plr_id>=MAX_PLAYERS || !pmenu[plr_id])
      return;
   pmenu[plr_id]->SetPrgList(pl, pos_only);
}

//----------------------------

void C_game_menu::SetActive(dword plr_id){

   for(int i=0; i<MAX_PLAYERS; i++){
      if(pmenu[i]){
         if(pmenu[i]->IsActive())
            pmenu[i]->SetActive(false);
         if(i==plr_id)
            pmenu[i]->SetActive(true);
      }
   }
}

//----------------------------

void C_game_menu::SetDeathFace(dword plr_id){

   assert(pmenu[plr_id]);
   if(!pmenu[plr_id]){
      assert(0);
      return;
   }
                              //set death face
   bool b;
   b = pmenu[plr_id]->face_img.Init(driver, "tables\\face.prj", "faces\\face_dth.png");
   assert(b);
   pmenu[plr_id]->SetFace();
}

//----------------------------

void C_game_menu::SetHealth(dword plr_id, float f){

   assert(pmenu[plr_id]);
   if(!pmenu[plr_id])
      return;
   pmenu[plr_id]->SetHealth(f);
}

//----------------------------
//----------------------------
                           
bool C_game_menu::HitCheck(float pos_x, float pos_y, int &plr_id, int &sub_id){
                              //check all players
   int i;
   for(i=0; i<MAX_PLAYERS; i++){
      C_plr_menu *pm = pmenu[plr_id=i];
      if(!pm)
         continue;
      if(pm->HitTest(pos_x, pos_y, sub_id))
         return true;
   }
   return false;
}

//----------------------------

void C_game_menu::SetPrgKeyColor(dword plr_id, dword key_id, dword color){

   assert(pmenu[plr_id]);
   pmenu[plr_id]->SetPrgKeyColor(key_id, color);
}

//----------------------------

dword C_game_menu::GetPrgKeyColor(dword plr_id, dword key_id){

   assert(pmenu[plr_id]);
   return pmenu[plr_id]->GetPrgKeyColor(key_id);
}

//----------------------------

void C_game_menu::ScrollTimeAxis(int time){

                              //compute new offset
   int nto = time_offset - time;
   nto = Max(0, Min(TIME_AXIS_LENGTH-TIME_RANGE, nto));
   if(time_offset==nto)
      return;
   time_offset = nto;
                              //scroll all player's axes
   int i;
   for(i=0; i<MAX_PLAYERS; i++)
   if(pmenu[i])
      pmenu[i]->TimeAxisScrolled();

   axis_nums.clear();
   CreateAxisNums();
}

//----------------------------
//----------------------------
