#include "all.h"
#include "net.h"
#include "tables.h"
#include "loader.h"
#include "actors.h"
#include <insanity\controller.h>
#include "winstuff.h"
#include "mainmenu.h"
#include <sortlist.hpp>
#include <advanced.h>

                              //version of binary file
#define VERSION_MAJOR 1
#define VERSION_MINOR 5


#define ANIM_CACHE_SIZE 64

//#define DEBUG_AUT_SET_MAP     //set map mode as mission starts (at GameBegin)

#define LOD_OPTIMAL_FP_LO 30  //lower fps bound under which we set worse LOD
#define LOD_OPTIMAL_FP_HI 50  //upper fps bound beyond which we set better LOD
#define LOD_OPTIMAL_FPS 30    //fps which is optimal for default LOD scale
#define MAX_LOD_SCALE 3.0f    //maximal scale for LOD (worse quality)
#define MIN_LOD_SCALE .5f     //minimal scale for LOD (better quality)

//----------------------------

bool LoadSoldierTable(LPC_table tab, int man_indx);
extern const char *actor_names[];

//----------------------------

                           //actor identification over network
class C_net_id_map{
public:
   word next_net_id;          //next proposed id
   word net_id_base;
   bool *net_id_use;          //allocation of unique ID
                              //ID to actor mapping
   map<word, LPC_actor, less<word> >id_map;

   C_net_id_map(word base):
      next_net_id(0),
      net_id_base(base)
   {
   }
};

//----------------------------

struct S_screen_effect{
   C_game_mission::E_SCREEN_EFFECT_MODE mode;
   S_vector color;
   float alpha;
   int come_time, leave_time;
   int come_count, stay_count, leave_count;
   enum{
      PHASE_COME,
      PHASE_STAY,
      PHASE_LEAVE,
   } phase;
   S_screen_effect():
      mode(C_game_mission::SCREEN_EFFECT_ADD),
      phase(PHASE_COME)
   {}

   bool Tick(int time){

      if(come_count > 0 && time){
         come_count -= time;
         time = 0;
         if(come_count < 0){
            time = -come_count;
            come_count = 0;
         }
      }
      if(stay_count > 0 && time){
         stay_count -= time;
         time = 0;
         if(stay_count < 0){
            time = -stay_count;
            stay_count = 0;
         }
      }
      if(leave_count > 0 && time){
         leave_count -= time;
         if(leave_count < 0){
            leave_count = 0;
         }
      }
      return (!come_count && !stay_count && !leave_count);
   }

   float GetPower() const{

      float f = 1.0f;
      if(come_count){
         f = 1.0f - ((float)come_count / (float)come_time);
      }else
      if(!stay_count){
         f = ((float)leave_count / (float)leave_time);
      }
      return f;
   }

   void SetColor(dword c){
      color.x = (float)((c>>16)&255) / 255.0f;
      color.y = (float)((c>> 8)&255) / 255.0f;
      color.z = (float)((c>> 0)&255) / 255.0f;
      alpha =   (float)((c>>24)&255) / 255.0f;
   }
};
typedef multimap<int, S_screen_effect> t_screen_effects;

//----------------------------
//----------------------------

C_game_mission::C_game_mission():
   cheat_pos(0),
   net_id_map(NULL),
   game_cam(scene),
   game_menu(texts),
   net_chat(NULL),
   map_active(false),
   aim_line_valid(false),
   sync_menu(NULL),
   is_in_game(false),
   want_state(MS_NULL),
   debug_dump(DEBUG_NO),
   fps_counter(12),
   cp_grid_valid(false),
   con_objectives(.007f, .0225f, .033f, .034f),
   screen_effects(new t_screen_effects)
{
   texts.Init();
   AddAnimDir("Anims");
   AddAnimDir(C_fstring("Anims\\%s", game_configuration.censored ? "censored" : "nocensor"));

   game_cam.InitFirst(*this);

   {
                              //init scene for aim model rendering
      aim_render_scene = driver->CreateScene();
      aim_render_scene->Release();
      PI3D_light lp;
      lp = driver->CreateLight();
      aim_render_scene->AddFrame(lp);
      aim_render_scene->GetPrimarySector()->AddLight(lp);
      lp->SetLightType(I3DLIGHT_AMBIENT);
      lp->Release();

      PI3D_camera cam = driver->CreateCamera();
      aim_render_scene->AddFrame(cam);
      cam->SetRange(CAM_NEAR_RANGE, 10.0f);
      cam->SetPos(S_vector(0, 0, 0));
      cam->SetDir(S_vector(0, 0, 1));
      cam->Update();
      aim_render_scene->SetActiveCamera(cam);
      cam->Release();
   }
                              //camera volume for active human
#if 1
   cam_vol = driver->CreateVolume();
   if(cam_vol){
      cam_vol->Release();
      cam_vol->SetVolumeType(I3DVOLUME_SPHERE);
      cam_vol->SetScale(CAM_COL_SPHERE);
      cam_vol->SetPos(S_vector(.0f, .0f, .0f));
      cam_vol->SetOn(false);
      cam_vol->SetCollisionMaterial(HUMAN_SPHERE_MAT);
   }
#endif

   memset(cheat_buf, 0, MAX_CHEAT_LEN);
   spr_cross = new C_poly_sprite(.5f, .375f, .02f, &icons_img, 2);
   spr_cross->SetOn(false);
   spr_cross->Release();

   net_sync = CreateNetSync(0, NM_SYNC, 16);
   net_sync->Release();

   scene->SetCallback(cbScene_thunk, SCNCALLBACK_SECTSWITCH);
}

//----------------------------

C_game_mission::~C_game_mission(){

                              //must clear actors before destructing anything else (they call this' functions)
   actors.clear();
   Close();
   delete net_chat;
   delete (t_screen_effects*)screen_effects;
   if(sync_menu)
      sync_menu->Release();
   debug_text = NULL;
}

//----------------------------

void C_game_mission::cbScene(SCN_CALLBACK_MESSAGE msg, dword prm1, dword prm2){

   switch(msg){
   case CBM_SECTORSWITCH:
      {
         PI3D_sector sct_to = (PI3D_sector)prm2;
         snow_rain_on = (sct_to->IsPrimary());
      }
      break;
   }
}

//----------------------------

void C_game_mission::ResetScreen(){

   C_mission::ResetScreen();
                              //reset menus, etc
   game_menu.ResetScreen();
   aim_render_scene->SetViewport(0, 0, igraph->Scrn_sx(), igraph->Scrn_sy(), false);
   for(int i=icon_info_list.size(); i--; ){
      icon_info_list[i]->ResetScreen();
   }
}

//----------------------------

void C_game_mission::DisplayDebugText(){

   C_string str;
   switch(debug_dump){
   case DEBUG_FPS:
      {
         float fps = fps_counter.GetCount();
         float avg = fps_counter.GetAverageFPS();
         str = C_fstring("FPS: %.1f\nAVG: %.1f", fps, avg);
      }
      break;
   case DEBUG_COORDS:
      {
         S_vector v(0, 0, 0);
         for(int i=actors.size(); i--; ){
            LPC_actor a = actors[i];
            if(a->GetType()==ACTOR_PLAYER && a->IsActive()){
               v = *a->GetPos();
               break;
            }
         }
         str = C_fstring("Player position: X: %.1f, Y: %.1f, Z: %.1f", v.x, v.y, v.z);
      }
      break;
   case DEBUG_NET_STATS:
      if(net){
         dword in, out_g, out_ng;
         net->GetStats(&in, &out_g, &out_ng);
         str = C_fstring(
            "Net stats:\n"
            "%s\n"
            "In: %i\n"
            "Out (g) : %i\n"
            "Out (ng): %i",
            net->IsHost() ? "host" : "join",
            in,
            out_g,
            out_ng
            );
      }
      break;
   }
   debug_text = NULL;
   if(str.Size()){
      debug_text = texts.CreateText(str, -1, .01f, .02f, .02f);
      debug_text->Release();
   }
}

//----------------------------

void C_game_mission::CreateScreenEffect(dword c, dword come_time, dword stay_time, dword leave_time,
   E_SCREEN_EFFECT_MODE mode, int priority){

   S_screen_effect se;
   se.mode = mode;
   se.SetColor(c);
   se.come_count = se.come_time = come_time;
   se.stay_count = stay_time;
   se.leave_count = se.leave_time = leave_time;
   ((t_screen_effects*)screen_effects)->insert(t_screen_effects::value_type(priority, se));
}


//----------------------------

void C_game_mission::Tick(const S_tick_context &tc){

   fps_counter.Tick(tc.time);

                              //manage LOD change
   {
      float fps = fps_counter.GetCount();
      if(fps){
         float scale = 1.0f;
         if(fps < LOD_OPTIMAL_FP_LO){
            scale = LOD_OPTIMAL_FP_LO / fps;
            scale = Min(MAX_LOD_SCALE, scale);
         }else
         if(fps > LOD_OPTIMAL_FP_HI){
            scale = LOD_OPTIMAL_FP_HI / fps;
            scale = Max(MIN_LOD_SCALE, scale);
         }
         driver->SetState(RS_LOD_SCALE, FloatAsInt(scale));
      }
   }

   if(net){
      net_msg->Ascend(NM_GAME);

      dword winner_id, event;
      switch(net_sync->GetState(&winner_id, &event)){
      case C_net_sync::SS_RESOLVED:
         {
            if(sync_menu){
               sync_menu->AddAllCheckMarks();
               RenderScreen();

               sync_menu->Release();
               sync_menu = NULL;
            }

            net_sync->Reset();
            switch(event){
            case NS_GAME_BEGIN:
               break;
            case NS_GAME_EXIT:
                     //check if this is us who's leaving...
               if(winner_id==net->GetPID()){
#ifndef EDITOR
                  GameEnd();
                  igraph->ClearViewport(); igraph->UpdateScreen();
                  net_msg->Descend();
                  //in_msg->Descend();
                  NetCloseAll();

                  game_configuration.net_host = false;
                  game_configuration.net_join = false;

                  tick_class = CreateMainMenu();
                  if(tick_class)
                     tick_class->Release();
                  return;
#endif//EDITOR
               }
               break;

            case NS_DONE_SUCCESS:
            case NS_DONE_FAIL:
               {
                  //in_msg->Descend();
                  net_msg->Descend();
                  NetSend(true);
                  want_state = event==NS_DONE_SUCCESS ? MS_DONE_SUCCESS : MS_DONE_FAIL;
                  //Release();
               }
               return;

            case NS_TO_MAP:
               {
                  net_msg->Descend();
                  NetSend(true);
                  //char buf[256];
                  //net->GetPlayerName(winner_id, buf, sizeof(buf));
                  //OUTTEXT(C_fstring("Map mode invoked by '%s'", buf));
                  SwitchToMap();
               }
               return;

            case NS_TO_BRIEFING:
               {
                  net_msg->Descend();
#ifndef EDITOR
                  NetSend(true);

                  C_string name = mission_name;
                  name[(int)name.Size()-5] = 0;

                  C_class_to_be_ticked_hd *brf = CreateBriefing(name, this);
                  if(brf){
                     scene->AttenuateSounds();
                     tick_class = brf;
                     brf->Release();
                  }
                  return;
#endif
               }
               break;

            default:
               assert(0);
            }
         }
         break;

      case C_net_sync::SS_PROGRESS:
         {
#ifndef EDITOR
                              //waiting for sync event
            if(!sync_menu){
                              //check what we synchronizing
               int tid;
               switch(event){
               case NS_GAME_BEGIN: tid = 20300; break;
               case NS_GAME_EXIT:
               case NS_DONE_SUCCESS:
               case NS_DONE_FAIL:
                  tid = 20302;
                  break;
               case NS_TO_MAP: tid = 20303; break;
               case NS_FROM_MAP: tid = 20304; break;
               case NS_TO_BRIEFING: tid = 20305; break;
               default: assert(0);
               }
               sync_menu = CreateSyncMenu(tid);
            }
                              //detect changes
            vector<dword> pids;
            net_sync->GetAnsweredPIDs(pids);
            sync_menu->AddCheckMarks(pids.begin(), pids.size());
#endif
            net_msg->Descend();
            return;
         }
         break;
      }
   }

   AddRef();

   if(map_active){
      map_mgr->Tick(tc);
      game_menu.Tick(tc);
   }else{
      con_objectives.Tick(tc.time);
#ifdef EDITOR
      if(editor->IsActive()){
                                 //tick only selected actors
         static const bool tick_flags[] = {
            false, false, false, false,
            false, true, false, false,
            false, false, false, false,
            false, false, false, false,
            false, true, true
         };
                                 //tick selected actors
         for(int i=0; i<actors.size(); ){
            C_actor *ap = actors[i];
            dword type = ap->GetType();
            if(type<sizeof(tick_flags))
            if(tick_flags[type])
               ap->Tick(tc);
            if(ap==actors[i])
               i++;
         }
      }else
#endif//EDITOR
      {
         S_tick_context tc_hlp;
         const S_tick_context *tcp = &tc;

         if(net && tc.p_ctrl && net_chat->IsChatMode()){
            net_chat->Tick(tc);
            tc_hlp = tc;
            tc_hlp.ClearInput();
            tcp = &tc_hlp;
         }

                                 //tick all actors
         for(int i=0; i<actors.size(); ){
            C_actor *ap = actors[i];
            ap->Tick(*tcp);
            if(ap==actors[i])
               i++;
         }
         TickSnowRain(tc.time);

         switch(tc.key){
         case K_F1:
            {
               DisplayObjectives();
            }
            break;
         case K_F2:
            {
#ifndef EDITOR
                                 //show briefing
               if(!stricmp(&mission_name[(int)mission_name.Size()-5], "\\game")){
                  C_string name = mission_name;
                  name[(int)name.Size()-5] = 0;
                  if(!net){
                     C_class_to_be_ticked_hd *brf = CreateBriefing(name, this);
                     if(brf){
                        scene->AttenuateSounds();
                        tick_class = brf;
                        brf->Release();
                        Release();
                        return;
                     }
                     //OUTTEXT(C_fstring("Can't show briefing for mission '%s'.", (const char*)mission_name));
                  }else{
                     net_sync->BeginEvent(NS_TO_BRIEFING);
                  }
               }
#endif
            }
            break;

         case K_F3:
                                 //toggle aim line
            game_configuration.draw_aim_line = !game_configuration.draw_aim_line;
            game_configuration.WriteSysConfig();
            break;

         case K_F4:
                                 //cycle through aim modes
            ++game_configuration.aim_mode %= 3;
            game_configuration.WriteSysConfig();
            BroadcastMessage(CB_SETFOCUS, 2);
            break;

         case K_F5:
            if(net){
               assert(net_chat);
               net_chat->Open();
            }
            break;

#ifdef FINAL
         case K_F6:
#else
         case K_F9:
#endif
            if(!net)
               SaveGame();
            break;

         case K_F10:
            if(!net){
                              //check if saved game available
               S_mission_state::S_save_game &save_game = game_profile.mission_states[game_profile.MissionIndex()].save_game;
               if(save_game.len){

                  RenderAllBackbuffers();
                  if(LoadGame()){
                     //DisplayObjectives();
                     igraph->ProcessWinMessages();
                     igraph->ReadKey();
                  }else{
#ifdef EDITOR
                     //tick_class = NULL;
#else
                     tick_class = CreateMainMenu();
                     if(tick_class)
                        tick_class->Release();
#endif
                     Release();
                     return;
                  }
               }
            }
            break;
         }
         if(tcp->p_ctrl){
            if(tcp->p_ctrl->Get(GKEY_ESCAPE)){
                              //ESC pressed
               tcp->p_ctrl->IgnoreValue(GKEY_ESCAPE);
#ifndef EDITOR
               if(net && net_chat->IsChatMode()){
                  net_chat->Close();
               }else{

                              //go to in-game menu
                  tick_class = CreateMainMenu(this);
                  if(tick_class)
                     tick_class->Release();
               }
#endif
            }else
            if(tcp->p_ctrl->Get(GKEY_MAP)){
               tcp->p_ctrl->IgnoreValue(GKEY_MAP);
                              //switch to map
               if(!net){
                  SwitchToMap();
                  Release();
                  return;
               }
               net_sync->BeginEvent(NS_TO_MAP);
            }
         }
         if(debug_dump)
            DisplayDebugText();

         ProcessCheat(tc.key);
         //if(tick_class!=this) return;
      }
      C_mission::Tick(tc);

                              //check all played sounds
      for(int i=play_sounds.size(); i--; ){
         if(!play_sounds[i]->IsPlaying()){
            play_sounds[i] = play_sounds.back(); play_sounds.pop_back();
         }
      }
      game_menu.Tick(tc);
      game_cam.Tick(tc.time);

                              //tick screen effects
                              //update screen effects
      t_screen_effects *effects = ((t_screen_effects*)screen_effects);
      t_screen_effects::iterator it, it_next;
      for(it=effects->begin(); it!=effects->end(); ){
         it_next = it; ++it_next;
         S_screen_effect &se = (*it).second;
         if(se.Tick(tc.time))
            effects->erase(it);
         it = it_next;
      }
   }

   if(net)
      net_msg->Descend();

   if(want_state!=MS_NULL)
      EndMission();

   Release();
}

//----------------------------

void C_game_mission::Render(){

   if(map_active){
      map_mgr->Render();
      //texts.Render(); //do not render (icon info text, etc)
      return;
   }

   driver->BeginScene();

   C_mission::Render();
   if(is_in_game){
      if(aim_line_valid){
         scene->SetRenderMatrix(GetIdentityMatrix());
         /*
         scene->DrawLine(aim_line[0], aim_line[1],
            tab_game_cfg->ItemV(TAB_F_GAME_AIMLINE_COLOR), tab_game_cfg->ItemI(TAB_I_GAME_AIMLINE_ALPHA));
            */
         const S_matrix &m = scene->GetActiveCamera()->GetMatrix();
         S_vector right = (aim_line[0]-m(3)).Cross(aim_line[0] - aim_line[1]);
         right.Normalize();
         right *= .01f;
         S_vector v[4];
         v[0] = aim_line[0] - right;
         v[1] = aim_line[1] - right;
         v[2] = v[0] + right * 2.0f;
         v[3] = v[1] + right * 2.0f;
         driver->SetTexture(NULL);

         static const word indx[] = {0, 1, 2, 2, 1, 3};
         const S_vector &color = tab_game_cfg->ItemV(TAB_F_GAME_AIMLINE_COLOR);
         byte a = tab_game_cfg->ItemI(TAB_I_GAME_AIMLINE_ALPHA);
         scene->DrawTriangles(v, 4, indx, 6, color, sizeof(S_vector), a | I3DDRAW_NOCULL);

         aim_line_valid = false;
      }
                              //render aim model
      PI3D_frame aim_model = game_cam.GetAimModel();
      if(aim_model && aim_model->IsOn() && game_cam.GetDistanceMode()==0){
         S_vector save_pos = aim_model->GetPos();
         S_quat save_rot = aim_model->GetRot();
         //PI3D_frame save_link = aim_model->GetParent();
         assert(!aim_model->GetParent() || aim_model->GetParent()==aim_render_scene->GetPrimarySector());

         PI3D_camera a_cam = scene->GetActiveCamera();
         const S_matrix &m_ic = a_cam->GetInvMatrix();
         S_vector pos = save_pos * m_ic;
         S_quat rot = save_rot * S_quat(m_ic);

         aim_model->LinkTo(aim_render_scene->GetPrimarySector());
         aim_model->SetPos(pos);
         aim_model->SetRot(rot);
         aim_model->Update();

         aim_render_scene->GetActiveCamera()->SetFOV(a_cam->GetFOV());

         bool was_cl = driver->GetState(RS_CLEAR);
         if(was_cl) driver->SetState(RS_CLEAR, false);
         aim_render_scene->Render();
         if(was_cl) driver->SetState(RS_CLEAR, true);

                              //must link it back due to shooting
         aim_model->SetPos(save_pos);
         aim_model->SetRot(save_rot);
         aim_model->Update();
         //aim_model->LinkTo(save_link);
      }
      if(sync_menu)
         sync_menu->Render();

      game_menu.Render();

      bool was_zb = driver->GetState(RS_USEZB);
      if(was_zb) driver->SetState(RS_USEZB, false);

      for(int i=icon_info_list.size(); i--; )
         icon_info_list[i]->Render();

      if(spr_cross->IsOn()){
         if(!aim_cross_centered){
                              //compute position of cross now
            S_vector v_scr;

            S_vectorw v4;
            scene->TransformPoints(&aim_line[1], &v4, 1);

            v4.y /= (screen_aspect_ratio * 1.33333333f);

            spr_cross->SetScreenPos(v4.x, v4.y);
         }
         spr_cross->Render();
         spr_cross->SetOn(false);
      }

      if(spr_use_drop)
         spr_use_drop->Render();

                              //render screen effects
      t_screen_effects *effects = ((t_screen_effects*)screen_effects);
      if(effects->size()){
         S_vector color(0, 0, 0);
         float alpha = 0.0f;
         t_screen_effects::iterator it, it_next;
         float alpha_sum = 0.0f;
         for(it=effects->begin(); it!=effects->end(); it++){
            const S_screen_effect &se = (*it).second;
            float p = se.GetPower();
            float a = se.alpha * p;
            color += se.color * a;
            alpha_sum += a;
            alpha += a;
         }
         if(!IsMrgZero(alpha_sum)){
            color /= alpha_sum;

            static const word indx[] = { 0, 1, 2, 2, 1, 3};
            driver->SetTexture(NULL);

            S_vectorw tv[4];
            float sx = igraph->Scrn_sx();
            float sy = igraph->Scrn_sy();
            for(int j = 0; j < 4; j++){
               tv[j].x = .0f;
               tv[j].y = .0f;
               if(j&1) tv[j].x += sx;
               if(j&2) tv[j].y += sy;
               tv[j].z = .0f;
               tv[j].w = 1.0f;
            }
            scene->DrawTriangles(tv, 4, indx, 6, color, sizeof(S_vectorw),
               I3DDRAW_TRANSFORMED | Max(1, 255 - (int)(alpha*255.0f)));
         }
      }

      if(was_zb) driver->SetState(RS_USEZB, true);
   }

   if(tick_class==this){
      texts.Render();
      con_objectives.texts.Render();
   }

#ifdef EDITOR
   if(editor) editor->Render();
#endif
   driver->EndScene();
}

//----------------------------

#define II_POS_X .96f //1.28f
#define II_POS_Y .1425f //.19f
#define II_SPACE_Y .071f//.095f

void C_game_mission::AddIconInfo(LPC_icon_info ii){

   int num = icon_info_list.size();
   ii->SetPos(II_POS_X, II_POS_Y + (float)num*II_SPACE_Y);
   icon_info_list.push_back(ii);
}

//----------------------------

void C_game_mission::DelIconInfo(LPC_icon_info ii){

   for(int i=icon_info_list.size(); i--; ){
      if(icon_info_list[i]==ii){
         icon_info_list.erase(&icon_info_list[i]);
         for(; i<icon_info_list.size(); i++)
            icon_info_list[i]->SetPos(II_POS_X, II_POS_Y + (float)i*II_SPACE_Y);
         return;
      }
   }
   assert(0);
}

//----------------------------

PI3D_sound C_game_mission::GetActionSound(dword mat_id, dword act_id){

   if(!isound)
      return NULL;
   if(mat_id >= MAX_MATERIALS)
      return NULL;
   if(act_id >= mat_table->ArrayLen(TAB_S12_SND_SNDNAME))
      return NULL;
   /*
   if(mat_id && !*(((char*)mat_table->Item(TAB_S12_SND_SNDNAME))+12*(MAX_ACTIONS*mat_id+act_id))){
      OUTTEXT(C_fstring("no sound defined for vol. %i, action %i", mat_id, act_id));
   }
   */
   const char *fname = ((char*)mat_table->Item(TAB_S12_SND_SNDNAME)) + 12 * (MAX_ACTIONS * mat_id + act_id);
   if(!*fname){
                              //no sound defined for this material, try default mat
      fname = ((char*)mat_table->Item(TAB_S12_SND_SNDNAME)) + 12* act_id;
      if(!*fname)
         return NULL;
      mat_id = 0;
   }
   C_fstring str("action\\%s", fname);

   PI3D_sound snd = CreateSound(str);
   if(snd){
      int vol_perc = ((int*)mat_table->Item(TAB_I_SND_VOLUME))[MAX_ACTIONS*mat_id + act_id];
      float vol = (float)vol_perc/100.0f;
      //vol *= .5f;
      snd->SetVolume(vol);
   }
   return snd;
}

//----------------------------

PI3D_sound C_game_mission::CreateSound(const C_string &fname, bool err_report){

   if(!isound)
      return NULL;

   PI3D_sound snd = scene->CreateSound();
   ISND_RESULT ir;
   C_string fn = fname;

   snd->SetSoundType(I3DSOUND_POINT);
   ir = sound_cache.Open(snd, fn, scene, 0);

   if(ISND_FAIL(ir)){
      if(err_report)
         ErrReport(C_fstring("Failed to open sound: '%s'", (const char*)fname), editor);
      snd->Release();
      return NULL;
   }
   return snd;
}

//----------------------------

bool C_game_mission::PlaySoundGlobal(const char *snd_name){

   if(!isound)
      return true;
                              //check if it is sound clip
   bool is_clip = C_string(snd_name).Matchi("clips\\*");
   if(is_clip){
      if(last_clip_sound){
         if(last_clip_sound->IsPlaying())
            return false;
         last_clip_sound = NULL;
      }
   }
                              //get sound size
   int len = -1;
   for(int i=0; i<sound_cache.NumDirs(); i++){
      const char *snd_dir = sound_cache.GetDir(i);
      len = GetFileSize(C_fstring("%s%s.ogg", snd_dir, snd_name));
      if(len==-1)
         len = GetFileSize(C_fstring("%s%s.wav", snd_dir, snd_name));
      if(len!=-1)
         break;
   }
   if(len==-1){
      ErrReport(C_fstring("can't play sound: %s", (const char*)snd_name), editor);
      return false;
   }
   bool streaming = false;
                              //assume mono sounds

                              //don't consider frequency
   float snd_time = (float)len / (float)(22050 * 2);
   if(snd_time>2.0f) streaming = true;

   if(!allow_stream_snd)
      streaming = false;

   PI3D_sound snd = scene->CreateSound();
#ifdef _DEBUG
   snd->SetName(snd_name);
#endif

   dword open_flags = 0;

   open_flags |= I3DLOAD_SOUND_SOFTWARE;
   if(streaming)
      snd->SetStreaming(true);
   snd->SetSoundType(I3DSOUND_AMBIENT);

   if(ISND_FAIL(sound_cache.Open(snd, snd_name, scene, open_flags))){
      snd->Release();
      ErrReport(C_fstring("can't play sound: %s", (const char*)snd_name), editor);
      return false;
   }
   if(is_clip){
      float vol = game_configuration.volume_music / 100.0f;
                              //we must set inversed volume of global sound vol
      float gv = Max(.01f, game_configuration.volume_effects / 100.0f);
      vol /= gv;
      snd->SetVolume(vol);
   }
   snd->LinkTo(scene->GetPrimarySector());
   snd->SetLoop(false);
   snd->SetOn(true);
   snd->Update();

   if(is_clip)
      last_clip_sound = snd;
   else{
      AddPlaySound(snd);
   }
   snd->Release();
   return true;
}

//----------------------------

bool C_game_mission::PlaySound(const char *snd_name, PI3D_frame lnk, float min, float max,
   float vol, bool loop){

   PI3D_sound snd = CreateSound(snd_name);
   if(snd){
      if(!lnk) lnk = scene->GetPrimarySector();
      snd->LinkTo(lnk);
      snd->SetRange(min, max);

      if(vol!=1.0f) snd->SetVolume(vol);

      //snd->Play(0);
      snd->SetLoop(loop);
      snd->SetOn(true);
      AddPlaySound(snd);
      snd->Release();
      return true;
   }
   return false;
}

//----------------------------

void C_game_mission::PlaySound(const char *snd_name, const S_vector &pos, float min, float max){

   PI3D_sound snd = CreateSound(snd_name);
   if(snd){
      //snd->LinkTo(scene);
      snd->LinkTo(scene->GetPrimarySector());
      snd->SetPos(pos);
      snd->SetRange(min, max);
      snd->SetLoop(false);
      snd->SetOn(true);
      snd->Update();
      AddPlaySound(snd);
      snd->Release();
   }
}

//----------------------------

static const struct S_cheat{
   const char *name;
   int size;
} cheats1[] = {
   "t0a1e2h3c4t5n6a7w8i9", 10,//iwantcheat
   NULL
}, cheats2[] = {
                              //0
   "t0a1e2h3c4t5n6a7w8i9", 10,//iwantcheat
   "e0f1i2l3w4e5n6", 7,       //newlife
   "d0a1e2h3y4n5n6u7f8", 9,   //funnyhead
   "s0p1f2w3o4h5s6", 7,       //showfps
                              //4
   "d0n1e2e3h4t5w6o7h8s9", 10,//showtheend
   "n0o1i2s3s4i5m6p7i8k9s0", 11, //skipmission
   "n0o1d2n3a4b5a6", 7,       //abandon
   "c0i1g2a3m4r5o6o7d8", 9,   //doormagic
                              //8
   "n0a1m2n3o4r5i6", 7,       //ironman
   "s0d1n2a3h4l5l6u7f8", 9,   //fullhands
   "s0u1c2o3f4", 5,           //focus
   "t0r1o2p3g4u5b6e7d8", 9,   //debugport
                              //12
   "s0t1a2t3s4t5e6n7", 8,     //netstats
   "s0d1r2o3o4c5", 6,         //coords
   "y0t1i2l3a4t5r6o7m8m9i0", 11, //immortality
   "s0e1m2u3l4o5v6g7u8b9e0d1", 12,  //debugvolumes
                              //16
   "e0r1i2w3w4a5r6d7g8u9b0e1d2", 13, //debugdrawwire
   "e0d1i2c3i4u5s6", 7,       //suicide
   "s0r1a2t3s4e5u6l7b8", 9,   //bluestars
   NULL
},
#ifndef EDITOR                //release version needs to unlock cheats before use
   //**cheats = cheats1;
   *cheats = cheats1;
#else
   //**cheats = cheats2;
   *cheats = cheats2;
#endif

//----------------------------

void C_game_mission::ProcessCheat(char k){

   if(k==K_NOKEY) return;
   char ac = igraph->GetAscii(k);
   if(!isalnum(ac)) return;
   cheat_buf[cheat_pos] = ac;
   if(++cheat_pos==MAX_CHEAT_LEN)
      cheat_pos = 0;
   for(int i=0; ; i++){
      const char *str = cheats[i].name;
      if(!str)
         break;
      int size = cheats[i].size * 2;
      int k = cheat_pos;
      for(int j=0; j<size; j += 2){
         if(!k--) k = MAX_CHEAT_LEN-1;
         if(str[j]!=cheat_buf[k])
            break;
      }
      if(j==size){
         PlaySoundGlobal("depleted");

         switch(i){

         case 0:           //uncock
            cheats = cheats2;
            break;

         case 3:              //toggle fps counter
            debug_dump = DEBUG_FPS;
            break;

         case 13:
            debug_dump = DEBUG_COORDS;
            break;

         case 12:
            debug_dump = DEBUG_NET_STATS;
            break;

         case 1:           //zombie
         case 2:           //bighead
         case 8:           //goodhealth
         case 9:           //allammo
         case 14:          //noplayerhits
#if defined EDITOR || defined _DEBUG
         case 17:          //suicide
         case 11:          //debugport
#endif
            BroadcastMessage(CB_CHEAT, i, 0);
            break;

         case 10:             //setfocus
            {
                                    //select actor to focus on
               vector<char> item_list;
               vector<LPC_actor> sorted_actors;
               sorted_actors.reserve(actors.size());
               for(int i=actors.size(); i--; ){
                  LPC_actor act = actors[i];
                  if(!act->IsHuman())
                     continue;
                  sorted_actors.push_back(act);
               }
               struct S_hlp{
                  static bool cbSort(LPC_actor a1, LPC_actor a2){
                     return (strcmp(a1->GetName(), a2->GetName()) < 0);
                  }
               };
               sort(sorted_actors.begin(), sorted_actors.end(), S_hlp::cbSort);
               int curr_index = -1;
               for(i=0; i<sorted_actors.size(); i++){
                  LPC_actor act = sorted_actors[i];
                  if(act->cbProc(CB_IS_FOCUSED))
                     curr_index = i; 
                  const char *name = act->GetName();
                  item_list.insert(item_list.end(), name, name+strlen(name)+1);
               }
                                    //insert terminator
               item_list.push_back(0);
               igraph->FlipToGDI();
               int indx = WinSelectItem("Select human to focus.", item_list.begin(), curr_index);
               if(indx==-1 || indx==curr_index)
                  break;
                                    //set focus on selected human
               //game_cam.SetDistanceMode(sorted_actors[indx]->IsPlayer() ? 0.0f : 3.0f);
               if(curr_index!=-1)
                  sorted_actors[curr_index]->cbProc(CB_SETFOCUS, false);
               sorted_actors[indx]->cbProc(CB_SETFOCUS, true);
            }
            break;

         case 4:           //end game mode
#ifdef FINAL
            if(!net){
               tick_class = CreateTheEnd();
               if(tick_class)
                  tick_class->Release();
            }
#endif
            break;

         case 5:           //mission successful
            SetGameState(MS_DONE_SUCCESS);
            break;

         case 6:           //mission unsuccessful
            SetGameState(MS_DONE_FAIL);
            break;

         case 7:           //openalldoor
            BroadcastMessage(CB_USE_DOOR, 1);
            break;

         case 15:
            //driver->SetState(RS_DRAWVOLUMES, !driver->GetState(RS_DRAWVOLUMES));
            break;

         case 16:
            driver->SetState(RS_WIREFRAME, !driver->GetState(RS_WIREFRAME));
            break;
         }
         break;
      }
   }
}

//----------------------------

void C_game_mission::GameBegin(bool setup_team, bool net_debug){

   want_state = MS_NULL;
   assert(!is_in_game);
   is_in_game = true;

   int active_man_index = 0;  //by default take 1st man from available
   int net_index;

                              //all net players (including us) sorted by PID
   vector<S_net_player> net_players;

                              //1. determine unique index among all net machines
   if(net){
      dword our_pid = net->GetPID();

      MakeSortedNetList(net_players, true);

                              //find on which position our PID has finished - that's our index
      for(net_index=net_players.size(); net_index--; )
         if(net_players[net_index].pid == our_pid)
            break;
                              //init our net ID base
      SetNetIdBase(0x4000 + net_index*0x2000);

      active_man_index = net_index;
                              //begin game by syncing
      net_msg->Ascend(NM_GAME);
      net_sync->BeginEvent(NS_GAME_BEGIN);
      net_msg->Descend();

      if(!net_chat)
         net_chat = new C_net_chat(NM_CHAT);
   }

   if(setup_team){
                              //2. spread and setup players in team
      S_mission_state &mstate = game_profile.mission_states[game_profile.MissionIndex()];
      int num_team4_players = 0;
      for(int i=0; i<4; i++){
         mstate.team_4[i].in_mission = false;
         if(mstate.team_4[i]!=-1)
            ++num_team4_players;
      }
                              //if we don't get player to control, at least wat
      if(active_man_index>=num_team4_players)
         active_man_index = -1;
                              //collect all players in mission
      vector<LPC_actor> plr_list;
      struct S_hlp{
         static bool cbEnum(LPC_actor a, dword c){
            if(a->GetType()==ACTOR_PLAYER){
               vector<LPC_actor> *plr_list = (vector<LPC_actor>*)c;
               plr_list->push_back(a);
            }
            return true;
         }
         static bool cbPlrLess(LPC_actor a1, LPC_actor a2){
            return (strcmp(a1->GetName(), a2->GetName()) < 0);
         }
      };
      EnumActors(S_hlp::cbEnum, (dword)&plr_list);
                              //sort players by name
      sort(plr_list.begin(), plr_list.end(), S_hlp::cbPlrLess);
                              //destroy players' menus, they'll be re-created only on existing ones,
                              // in proper order
      for(i=plr_list.size(); i--; )
         plr_list[i]->cbProc(CB_DESTROY_PLR_MENU);

      if(net_debug){
         if(active_man_index >= plr_list.size())
            active_man_index = -1;
      }

      int team_4_indx = 0;
      for(i=0; i<plr_list.size(); i++){
         if(!net_debug){
                              //find next man in team_4
            while(team_4_indx<4 && mstate.team_4[team_4_indx]==-1)
               team_4_indx++;
         }

         LPC_actor a = plr_list[i];
         if(net_debug || team_4_indx<4){
            plr_list[i]->cbProc(CB_CREATE_PLR_MENU);

            if(!net_debug){
                              //setup this man
               LPC_table tab = a->GetTable(0);
                              //load table because of character values
               int t4_man = mstate.team_4[team_4_indx];
               LoadSoldierTable(tab, t4_man);
                              //change face + apply table values
               tab->ItemI(TAB_I_HUM_FACE) = t4_man+1;
               a->TableUpdate(0, true);
                              //setup inventory
               a->cbProc(CB_SETINVENTORY, (dword)&mstate.team_4[team_4_indx].item_list);

               mstate.team_4[team_4_indx].in_mission = true;

               ++team_4_indx;
            }
            if(active_man_index==-1 && i==0){
                              //set focus on 1st player
               a->SetActive(true);
               a->cbProc(CB_SET_NETWORK_ACTOR, net_players[i]. pid);
            }else
            if(active_man_index==i){
                              //go into that man
               a->SetActive(true);
               a->cbProc(CB_SET_NETWORK_ACTOR, 0);
               if(net)
                  a->cbProc(CB_SHOW_PLAYER_NAME);
            }else
            if(net && i<net_players.size()){
                              //this player belongs to some other machine, thus can't be controlled by this
               a->cbProc(CB_SET_NETWORK_ACTOR, net_players[i].pid);
               a->cbProc(CB_SHOW_PLAYER_NAME);
            }
         }else{
                              //no more men in team_4, delete this actor
            DestroyActor(a);
         }
      }
                              //create automatic detector for 'all players dead' detection
      if(!net_debug)
         MakeEndDetector();
   }

   if(net){
                              //3. spread enemies among machines in network game
      int count = 0;
      for(int i=actors.size(); i--; ){
         LPC_actor a = actors[i];
         if(a->GetType()==ACTOR_ENEMY){
            bool local_ai = (count==net_index);
            a->cbProc(CB_SET_NETWORK_ACTOR,
               local_ai ? 0 : net_players[count].pid);
            ++count %= net_players.size();
#if defined _DEBUG && 1
            NET_LOG(C_fstring("Actor %s: network_actor = %i", a->GetName(), (int)(!local_ai)));
#endif
         }
      }
   }

                              //create actor ID mapping for network gameplay
   if(net)
      NetActorIDMake();

   BroadcastMessage(CB_APPLY_DIFFICULTY);

                              //save profile with current setting now
   if(!net || net->IsHost())
      game_profile.Save();

#ifdef EDITOR
   if(editor){
      bool acquired = editor->FindPlugin("Mission")->Action(E_MISSION_GCAM_IS_ACQUIRED);
      game_cam.Acquire(acquired);
   }
#endif

   game_cam.Init(scene->GetActiveCamera());
   game_cam.SetToIdealPosition();
#if 1
   if(cam_vol)
      cam_vol->SetScene(scene);
#endif
                              //reset statistics
   stat_ammo_spent = 0;
   stat_num_hits = 0;
   stat_enm_killed = 0;
   stat_civil_killed = 0;

   InitSnowRain();

                              //GameBegin on actors
   for(int i=actors.size(); i--; )
      actors[i]->GameBegin();

#ifdef DEBUG_AUT_SET_MAP
   map_mgr = CreateMap(*this);
   if(map_mgr){
      map_mgr->Release();
      map_active = true;
      map_mgr->Activate();
      game_menu.SetState(C_game_menu::STATE_EXPAND);
   }
#endif
}

//----------------------------

void C_game_mission::GameEnd(vector<int> *plr_killed, bool write_back){

   assert(is_in_game);

                              //reset driver's LOD
   driver->SetState(RS_LOD_SCALE, FloatAsInt(1.0f));

                              //GameBegin on actors
   for(int i=actors.size(); i--; )
      actors[i]->GameEnd();

   game_cam.Close();
#if 1
   if(cam_vol)
      cam_vol->SetScene(NULL);
#endif

   ((t_screen_effects*)screen_effects)->clear();

   if(write_back){
      S_mission_state *mstate = &game_profile.mission_states[game_profile.MissionIndex()];
      for(int i=0; i<4; i++){
         mstate->team_4[i].in_mission = (mstate->team_4[i] != -1);
      }
   }

   if(plr_killed){
      struct S_hlp{
         vector<int> *plr_killed;
         int count;
         bool write_back;

         static bool cbEnum(LPC_actor a, dword c){
            if(a->GetType()!=ACTOR_PLAYER)
               return true;
            S_hlp *hp = (S_hlp*)c;
            //int &count = *(int*)c;
            int &count = hp->count;
            S_mission_state *mstate = &game_profile.mission_states[game_profile.MissionIndex()];
                                 //get nth man in team_4
            for(int j=count+1, i=0; i<4; i++){
               if(mstate->team_4[i].in_mission && !--j)
                  break;
            }
                                 //shouldn't happen in final
            if(j)
               return true;
            bool alive = a->IsAlive();

            if(!alive)
               hp->plr_killed->push_back(mstate->team_4[i]);

            if(hp->write_back){
               mstate->team_4[i].item_list.clear();
               if(!alive){          //man is dead, his items are lost
                  mstate->team_4[i] = -1;
               }else{
                  a->cbProc(CB_STOREINVENTORY, (dword)&mstate->team_4[i].item_list);
               }
            }
            ++count;
            return true;
         }
      } hlp;

      hlp.plr_killed = plr_killed;
      hlp.count = 0;
      hlp.write_back = write_back;
      //EnumActors(S_hlp::cbEnum, (dword)&hlp);

                              //collect all players in mission
      vector<LPC_actor> plr_list;
      struct S_hlp1{
         static bool cbEnum(LPC_actor a, dword c){
            if(a->GetType()==ACTOR_PLAYER){
               vector<LPC_actor> *plr_list = (vector<LPC_actor>*)c;
               plr_list->push_back(a);
            }
            return true;
         }
         static bool cbPlrLess(LPC_actor a1, LPC_actor a2){
            return (strcmp(a1->GetName(), a2->GetName()) < 0);
         }
      };
      EnumActors(S_hlp1::cbEnum, (dword)&plr_list);
                              //sort players by name
      sort(plr_list.begin(), plr_list.end(), S_hlp1::cbPlrLess);

      for(i=0; i<plr_list.size(); i++){
         S_hlp::cbEnum(plr_list[i], (dword)&hlp);
      }
   }

                              //clear objectives text
   con_objectives.Clear();

   spr_use_drop = NULL;

   model_cache.Clear();
   sound_cache.Clear();

   is_in_game = false;
}

//----------------------------

void C_game_mission::ExitMapMode(){

   assert(map_active);
   if(!map_active)
      return;
   map_active = false;
   map_mgr->Suspend();
   game_menu.SetState(C_game_menu::STATE_COMPACT);
   game_cam.SetToIdealPosition();
   game_cam.Tick(8000); //(move to correct position if ideal position is behind wall)
}

//----------------------------

void C_game_mission::EmitListen(S_listen_source &ls, float max_distance1, bool net_send,
   LPC_actor emmitor, S_vector *shoot_ray){

   struct S_hlp{
      LPC_actor emmitor;
      float max_distance;
      S_listen_source *ls;
      S_vector *shoot_ray;
      PI3D_frame src_sector;
      bool net_send;
      C_game_mission *mission;
      static bool cbEnum(LPC_actor a, dword c){
         S_hlp *hp = (S_hlp*)c;
         switch(a->GetType()){
         case ACTOR_PLAYER:
                              //human-controlled player doesn't need to listen
            if(a->IsActive())
               return true;
                              //because we won't keep shooter with flare
                              // (sooter can die while flare is in the air and pointer invalidate),
                              // we skip sending flare sound to player ai.
            if(hp->ls->type == LISTEN_FLARE)
               return false;
                              //flow...
         case ACTOR_ENEMY:
            if(!a->IsAlive())
               return true;
            if(hp->emmitor){
               if(!a->IsEnemy(hp->emmitor))
                  return true;
            }
            break;
         case ACTOR_DETECTOR:
            if(!a->MayListen())
               return true;
            break;
         default:
            return true;
         }
         const S_vector *pos = a->GetHeadPos();
         if(!pos)
            pos = a->GetPos();
         const S_listen_source &ls = *hp->ls;
                              //check distance
         float dist, u;
         S_vector dir = (*pos) - ls.pos;
         if(!hp->shoot_ray){  //classical
            dist = dir.Magnitude();
         }else{               //ray mode
            u = hp->shoot_ray->Dot(*hp->shoot_ray);
            if(u<MRG_ZERO) return true;
            u = hp->shoot_ray->Dot((*pos)-ls.pos) / u;
            if(u<0.0f || u>1.05f) return true;
            S_vector p = ls.pos + (*hp->shoot_ray)*u;
            dist = (p-(*pos)).Magnitude();
         }
         if(dist<hp->max_distance){
                              //if between-sector distance, check collisions
            if(dist>.5f){
               PI3D_frame sct = a->GetFrame();
               while(sct && sct->GetType()!=FRAME_SECTOR) sct = sct->GetParent();
               if(hp->src_sector!=sct){
                  S_vector rpos;
                  bool b;

                  {
                     I3D_collision_data cd(ls.pos, dir, 0, hp->emmitor ? hp->emmitor->GetFrame() : NULL);
                     b = hp->mission->GetScene()->TestCollision(cd);
                     if(b){
                        //rpos = ls.pos + dir*cd.GetClosestHit() / dir.Magnitude();
                        rpos = cd.GetHitPosition();
                     }
                  }
                  if(b){
                     dir.Normalize();
                     rpos -= dir*.05f;
#ifdef EDITOR
                     if(tab_game_cfg->ItemB(TAB_B_GAME_DEBUG_SHOW_LISTEN))
                        DebugLine(ls.pos, rpos, 2, 0xff8080ff, 1000);
#endif
                        //check where is the collision point
//                     if(scene->GetSector(rpos)!=sct) return true;

                     PI3D_sector sc = hp->mission->GetScene()->GetSector(rpos);
                     if(sc){
                     if(sc!=sct)
                        return true;
                     }
                  }
#ifdef EDITOR
                  else
                     if(tab_game_cfg->ItemB(TAB_B_GAME_DEBUG_SHOW_LISTEN))
                        DebugLine(ls.pos, *pos, 2, 0xff8080ff, 1000);
#endif
               }
#ifdef EDITOR
               else
                  if(tab_game_cfg->ItemB(TAB_B_GAME_DEBUG_SHOW_LISTEN))
                     DebugLine(ls.pos, ls.pos + dir*hp->max_distance/dist, 2, 0xff8080ff, 1000);
#endif
            }
                        //ok to emit this source
            S_listen_source lsi = ls;
            if(hp->shoot_ray) lsi.pos += (*hp->shoot_ray)*u*.6f;
            lsi.power *= (hp->max_distance-dist)/hp->max_distance;
            lsi.power = Min(lsi.power, 2.0f);
            a->ReceiveListen(lsi, hp->net_send);
         }
         return true;
      }
   } hlp;
   hlp.emmitor = emmitor;
   hlp.max_distance = max_distance1;
   hlp.ls = &ls;
   hlp.shoot_ray = shoot_ray;
   hlp.src_sector = NULL;
   hlp.net_send = net_send;
   hlp.mission = this;
   if(emmitor){
      hlp.src_sector = emmitor->GetFrame();
      while(hlp.src_sector && hlp.src_sector->GetType()!=FRAME_SECTOR) hlp.src_sector = hlp.src_sector->GetParent();
   }
   if(!hlp.src_sector) hlp.src_sector = scene->GetSector(ls.pos);
   EnumActors(S_hlp::cbEnum, (dword)&hlp);
}

//----------------------------

bool C_game_mission::LoadChunk(CK_TYPE ck_t, S_load_context &lc){

   switch(ck_t){

   case CT_CHECKPOINTS:
      while(lc.ck.Size()){
         if(!lc.ck.RAscend(ck_t))
            return false;
         lc.ShowProgress();
         switch(ck_t){
         case CT_CPOINT:
            {
               LPC_checkpoint cp = CreateCheckpoint();
               while(lc.ck.Size()){
                  if(!lc.ck.RAscend(ck_t))
                     return false;
                  switch(ck_t){
                  case CT_NAME:
                     {
                        C_string str = lc.ck.RStringChunk();
                        cp->SetName(str);
                     }
                     break;
                  case CT_POSITION:
                     cp->SetPos(lc.ck.RVectorChunk());
                     break;
                  case CT_CP_TYPE:
                     {
                        word ct;
                        lc.ck.Read(&ct, sizeof(ct));
                        cp->SetType((E_CHECKPOINT_TYPE)ct);
                        lc.ck.Descend();
                     }
                     break;
                  case CT_CP_CONNECTION:
                     {
                        LPC_checkpoint last_cp = NULL;
                        while(lc.ck.Size()){
                           if(!lc.ck.RAscend(ck_t))
                              return false;
                           switch(ck_t){
                                                      //old style:
                           case CT_NAME:
                              {
                                 C_string str = lc.ck.RStringChunk();
                                 last_cp = FindCP(str);
                                 if(last_cp) cp->Connect(last_cp);
                              }
                              break;

                                                      //new style:
                           case CT_CP_CON_ID:
                              {
                                 word index;
                                 lc.ck.Read(&index, sizeof(word));
                                 assert(index<checkpoints.size());
                                 if(index<checkpoints.size()){
                                    last_cp = checkpoints[index];
                                    cp->Connect(last_cp);
                                 }
                              }
                              lc.ck.Descend();
                              break;

                           case CT_CP_CONTYPE:
                              if(last_cp){
                                 dword data;
                                 lc.ck.Read(&data, sizeof(dword));
                                 cp->SetConnectionData(cp->GetConnectionIndex(last_cp), data);
                                 if(data & ~(CP_CON_USE_LADDER | CP_CON_USE_DOOR)){
                                    const S_vector &p1 = cp->GetPos();
                                    const S_vector &p2 = last_cp->GetPos();
                                    S_vector pos = (p1 + p2) * .5f;
                                    DebugLine(p1, p2, 0, 0xff0000ff);
                                    lc.WarnLog(C_fstring("CP connection has invalid flags type (%.2f, %.2f, %.2f)", pos.x, pos.y, pos.z), NULL);
                                 }
                              }
                              lc.ck.Descend();
                              break;

                           case CT_CP_CONEXPENCE:
                              if(last_cp){
                                 float data;
                                 lc.ck.Read(&data, sizeof(float));
                                 cp->SetConExpence1(cp->GetConnectionIndex(last_cp), data);
                              }
                              lc.ck.Descend();
                              break;
                           }
                        }
                     }
                     lc.ck.Descend();
                     break;
                  default:
                     lc.ck.Descend();
                  }
               }
               lc.ck.Descend();
            }
            break;
         default:
            lc.ck.Descend();
         }
      }
      {
                           //create internal indicies
         for(int i=checkpoints.size(); i--; )
            checkpoints[i]->uin = i;
      }
      break;

   case CT_ACTOR:
      {
         word w;
         lc.ck.Read(&w, sizeof(word));
         E_ACTOR_TYPE type = (E_ACTOR_TYPE)w;
         LPC_actor act = NULL;
         while(lc.ck.Size()){
            if(!lc.ck.RAscend(ck_t))
               return false;
            switch(ck_t){
            case CT_NAME:
               {
                  C_string str = lc.ck.RStringChunk();
                              //try to find actor
                  map<C_string, PI3D_frame>::const_iterator it_fmap = lc.frame_map.find(str);
                  PI3D_frame frm = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;

                  if(frm){
                              //check if this frame has actor already
                     act = (LPC_actor)frm->GetData();
                     if(act){
                        lc.ErrLog(C_fstring("Warning: actor duplication on frame '%s' - deleting old", (const char*)str), frm);
                        DestroyActor(act);
                     }
                     act = CreateActor(type);
                     if(act){
                        act->SetFrame(frm);
#ifdef EDITOR
                        switch(type){
                        case ACTOR_TEXTURE_SHIFT:
                           lc.ErrLog(C_fstring("Warning: actor type '%s' (set on frame '%s') is obsolete!",
                              actor_names[type], (const char*)str), NULL);
                           break;
                        }
#endif
                     }else
                        lc.ErrLog(C_fstring("Warning: cannot create actor '%s' on frame '%s'",
                           actor_names[type], (const char*)str), NULL);
                  }else{
                     lc.ErrLog(C_fstring("Warning: cannot create actor '%s' on frame '%s' - not found",
                        actor_names[type], (const char*)str), NULL);
                  }
               }
               break;

            case CT_TABLE:
               if(act)
                  act->MissionLoad(&lc.ck, 0);
               lc.ck.Descend();
               break;

            case CT_TABLE_EX:
               if(act){
                  short id;
                  lc.ck.Read(&id, sizeof(short));
                  act->MissionLoad(&lc.ck, id);
               }
               lc.ck.Descend();
               break;
            }
         }
         if(lc.dump_chunks){
            lc.WarnLog(C_fstring("  Actor (%s), name: '%s'", actor_names[type], act->GetName()), NULL);
         }
      }
      break;

   case CT_CAMERA:
      {
         PI3D_camera cam = scene->GetActiveCamera();
         if(!cam)
            break;
         //lc.WarnLog("Importing camera settings", NULL);
         while(lc.ck.Size()){
            if(!lc.ck.RAscend(ck_t))
               return false;
            switch(ck_t){
            case CT_POSITION:
               cam->SetPos(lc.ck.RVectorChunk());
               break;
            case CT_DIRECTION:
               cam->SetDir(lc.ck.RVectorChunk(), 0.0f);
               break;
                           //game cam
            case CT_GC_MODE_DIST:
               game_cam.SetDistanceMode(lc.ck.RIntChunk());
               break;
            case CT_GC_MODE_ANGLE:
               game_cam.SetAngleMode(lc.ck.RIntChunk());
               break;
            default:
               lc.ck.Descend();
            }
         }
         cam->Update();
      }
      break;

   case CT_MODEL_ON_MAP:
      {
         C_string str;
         lc.ck.ReadString(str);
         map<C_string, PI3D_frame>::const_iterator it_fmap = lc.frame_map.find(str);
         PI3D_frame frm = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
         if(frm && frm->GetType()==FRAME_MODEL){
            dword flgs = frm->GetUserFlags();
            flgs |= 1;
            frm->SetUserFlags(flgs);
         }else{
            lc.ErrLog(C_fstring("Cannot set model to map: '%s' - not found", (const char*)str), NULL);
         }
      }
      break;

   default:
      return C_mission::LoadChunk(ck_t, lc);
   }
   return true;
}

//----------------------------

bool C_game_mission::IsAnyPlayerAlive() const{

                              //if any player alive, return true (allow save)
   for(int i = actors.size(); i--; ){
      LPC_actor actor = actors[i];
      if(actor && (actor->GetType() == ACTOR_PLAYER) && actor->IsAlive())
         return true;
   }
   return false;
}

//----------------------------

#ifdef EDITOR

E_MISSION_IO C_game_mission::Save(const char *bin_name){

   int i;

   C_fstring bin_str("missions\\%s\\scene.bin", bin_name);
                              //check if we have right to save to this file
   if(OsIsFileReadOnly(bin_str))
      return MIO_ACCESSFAIL;

                              //backup old file
   OsMoveFile(bin_str, C_fstring("missions\\_backup\\%s\\scene.bin", bin_name));

   C_chunk ck(256*1024);
   if(!ck.WOpen(bin_str)){
      if(!OsCreateDirectoryTree(bin_str))
         return MIO_ACCESSFAIL;
      if(!ck.WOpen(bin_str))
         return MIO_ACCESSFAIL;
   }
   ck.WAscend(CT_BASECHUNK);
   {                          //version + copyright
      ck.WAscend(CT_VERSION);
      i = VERSION_MAJOR;
      ck.Write(&i, 2);
      i = VERSION_MINOR;
      ck.Write(&i, 2);
      static const char copyright[] = " - Mission file - Copyright (c) 2001  Lonely Cat Games - ";
      ck.Write(copyright, sizeof(copyright)+1);
      ck.Descend();
   }
   {
                              //save table
      ck.WAscend(CT_TABLE_EX);
      tab_property->Save((dword)ck.GetHandle(), TABOPEN_FILEHANDLE);
      ck.Descend();
   }
                              //scene properties
   {
      S_vector v = scene->GetBgndColor();
      ck.WVectorChunk(CT_SCENE_BGND_COLOR, v);
      PI3D_camera cam = scene->GetActiveCamera();
      if(cam){
         ck.WFloatChunk(CT_CAMERA_FOV, cam->GetFOV());
         float rn, rf;
         cam->GetRange(rn, rf);
         ck.WFloatChunk(CT_CAMERA_RANGE, rf);
         if(cam->GetOrthogonal()){
            ck.WAscend(CT_CAMERA_ORTHOGONAL);
            ck.Descend();
         }
         ck.WFloatChunk(CT_CAMERA_ORTHO_SCALE, cam->GetOrthoScale());
                              //backdrop
         {
            float r[2];
            scene->GetBackdropRange(r[0], r[1]);
            ck.WAscend(CT_SCENE_BACKDROP);
            ck.Write((char*)r, sizeof(r));
            ck.Descend();
         }
      }
   }
#ifdef EDITOR
   editor->BroadcastAction(EDIT_ACTION_MISSION_SAVE_0, &ck);
   editor->BroadcastAction(EDIT_ACTION_MISSION_SAVE_1, &ck);
   editor->BroadcastAction(EDIT_ACTION_MISSION_SAVE_2, &ck);
#endif
                              //write checkpoints
   if(NumCheckpoints()){
                              //create internal indicies
      for(i=checkpoints.size(); i--; )
         checkpoints[i]->uin = i;

      ck.WAscend(CT_CHECKPOINTS);
      for(i=0; i<NumCheckpoints(); i++){
         LPC_checkpoint cp = checkpoints[i];
         word ct = cp->GetType();
         ck.WAscend(CT_CPOINT);
                              //name (if any)
         if(cp->GetName().Size())
            ck.WStringChunk(CT_NAME, cp->GetName());
                              //position
         ck.WVectorChunk(CT_POSITION, cp->GetPos());
                              //type
         if(ct != 0){
            ck.WAscend(CT_CP_TYPE);
            ck.Write(&ct, sizeof(ct));
            ck.Descend();
         }

         bool connect = false;
                              //connections
         for(int j=cp->NumConnections(); j--; ){
            LPC_checkpoint cp1 = cp->GetCPoint(j);
            const int &k = cp1->uin;
            if(k<i){
               if(!connect){
                  ck.WAscend(CT_CP_CONNECTION);
                  connect=true;
               }
               {
                  ck.WAscend(CT_CP_CON_ID);
                  word w = cp1->uin;
                  ck.Write(&w, sizeof(word));
                  ck.Descend();
               }
               {
                  dword data=cp->GetConnectionData(j);
                  if(data != 0){
                     ck.WAscend(CT_CP_CONTYPE);
                     ck.Write(&data, sizeof(dword));
                     ck.Descend();
                  }
               }
               {
                  float data = cp->GetConExpence(j);
                  if(data != 1.0f){
                     ck.WAscend(CT_CP_CONEXPENCE);
                     ck.Write(&data, sizeof(float));
                     ck.Descend();
                  }
               }
            }
         }
         if(connect)
            ck.Descend();
         ck.Descend();
      }
      ck.Descend();
   }

                              //write actors
   for(i=0; i<actors.size(); i++){
      LPC_actor act = actors[i];
                              //we're saving just several types
      switch(act->GetType()){
      case ACTOR_PLAYER:
      case ACTOR_ENEMY:
      case ACTOR_CANNON:
      case ACTOR_RESOURCE:
      case ACTOR_DOOR:
      case ACTOR_AUTOMOBIL:
      case ACTOR_AUTO_CANNON:
      case ACTOR_RANDOM_GEN:
      case ACTOR_TEXTURE_SHIFT:
      case ACTOR_VERSION:
      case ACTOR_DETECTOR:
      case ACTOR_LADDER:
      case ACTOR_MINE:
      case ACTOR_KILL_ANIM:
         break;
      default: continue;
      }

      {
         PI3D_frame frm = act->GetFrame();
         if(frm){
            ck.WAscend(CT_ACTOR);
            word w = act->GetType();
            ck.Write(&w, sizeof(word));
            ck.WStringChunk(CT_NAME, frm->GetName());
            act->MissionSave(&ck);
            ck.Descend();
         }
      }
   }

                              //frame flags
   {
      struct S_hlp{
         static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

            dword flgs = frm->GetUserFlags();
            if(flgs&1){
               C_chunk &ck = *(C_chunk*)c;
               ck.WStringChunk(CT_MODEL_ON_MAP, frm->GetName());
            }
            return I3DENUMRET_OK;
         }
      };
      scene->EnumFrames(S_hlp::cbEnum, (dword)&ck, ENUMF_MODEL);
   }


                              //save camera
   {
      ck.WAscend(CT_CAMERA);
                           //game cam
      ck.WIntChunk(CT_GC_MODE_DIST, game_cam.GetDistanceMode());
      ck.WIntChunk(CT_GC_MODE_ANGLE, game_cam.GetAngleMode());
      ck.Descend();
   }
   ck.Close();

   GetMissionCRC(mission_name, mission_crc);

   return MIO_OK;
}
#endif//EDITOR

//----------------------------

bool C_game_mission::SaveGame() const{

   int i;

#if 1
                              //don't allow save if all players are dead.
   if(!IsAnyPlayerAlive())
      return false;
#endif

   S_mission_state::S_save_game &save_game = game_profile.mission_states[game_profile.MissionIndex()].save_game;
   save_game.Close();

   C_chunk ck;
   if(!ck.WOpen(&save_game.mem, &save_game.len))
      return false;

   ck.WAscend(CT_SAVE_GAME);

   {
                              //write camera - before actors, because C_player uses its state at LoadGame
      PI3D_camera cam = scene->GetActiveCamera();
      if(cam){
         ck.WAscend(CT_SG_CAMERA);
         /*
         ck.WVectorChunk(CT_POSITION, cam->GetPos());
         ck.WVectorChunk(CT_DIRECTION, cam->GetDir());
         */
                                 //game cam
         ck.WIntChunk(CT_GC_MODE_DIST, game_cam.GetDistanceMode());
         ck.WIntChunk(CT_GC_MODE_ANGLE, game_cam.GetAngleMode());
         ck.Descend();
      }
   }

                              //write all non-human actors
   for(i=actors.size(); i--; ){
      LPC_actor a = actors[i];
      if(a->IsHuman())
         continue;
                              //several actors are not to be written
      if(!a->MaySaveGame())
         continue;

      ck.WAscend(CT_SG_ACTOR);
      ck.WStringChunk(CT_NAME, a->GetName());
      E_ACTOR_TYPE type = a->GetType();
      ck.Write(&type, sizeof(byte));
      a->SaveGame(&ck);

      ck.Descend();
   }
                              //write all human actors
   for(i=actors.size(); i--; ){
      LPC_actor a = actors[i];
      if(!a->IsHuman())
         continue;
                              //several actors are not to be written
      if(!a->MaySaveGame())
         continue;

      ck.WAscend(CT_SG_ACTOR);
      ck.WStringChunk(CT_NAME, a->GetName());
      E_ACTOR_TYPE type = a->GetType();
      ck.Write(&type, sizeof(byte));
      a->SaveGame(&ck);

      ck.Descend();
   }
   {                          //write stats
      ck.WAscend(CT_SG_STATS);
      ck.Write(&stat_ammo_spent, sizeof(int));
      ck.Write(&stat_num_hits, sizeof(int));
      ck.Write(&stat_enm_killed, sizeof(int));
      ck.Write(&stat_civil_killed, sizeof(int));
      ck.Descend();
   }

   ck.Descend();

   ck.Close();
                              //get mission crc (will be saved with mission state)
   {
      dword &miss_crc = game_profile.mission_states[game_profile.MissionIndex()].mission_crc;
      C_fstring mname("%s\\game", game_profile.GetCurrMissionName());
      GetMissionCRC(mname, miss_crc);
   }

   game_profile.Save();

                              //clear objectives, so that texts do not overlap
   ((C_game_mission*)this)->con_objectives.Clear();
   console_main.OutText(all_txt[20200], -1, all_txt.IsWide(20200));

   return true;
}

//----------------------------

bool C_game_mission::LoadGame(){

   S_mission_state::S_save_game &save_game = game_profile.mission_states[game_profile.MissionIndex()].save_game;
   if(!save_game.len)
      return false;

   C_chunk ck;
   if(!ck.ROpen(save_game.mem, save_game.len))
      return false;
   CK_TYPE t;
   if(!ck.RAscend(t) || t!=CT_SAVE_GAME)
      return false;

                              //re-init mission
   C_string mname = C_fstring("%s\\game", (const char*)game_profile.GetCurrMissionName());
#ifdef EDITOR
   mname = mission_name;
#elif defined _DEBUG
   if(game_configuration.net_cmdline_mission.Size())
      mname = game_configuration.net_cmdline_mission;
#endif
   if(is_in_game)
      GameEnd();
   Close();
   E_MISSION_IO mio = Open(mname, 0);
   assert(mio==MIO_OK);
   if(mio!=MIO_OK)
      return false;

   GameBegin(false);

                              //collect all players in mission
   vector<LPC_actor> plr_list;
   struct S_hlp{
      static bool cbEnum(LPC_actor a, dword c){
         if(a->GetType()==ACTOR_PLAYER){
            vector<LPC_actor> *plr_list = (vector<LPC_actor>*)c;
            plr_list->push_back(a);
         }
         return true;
      }
      static bool cbPlrLess(LPC_actor a1, LPC_actor a2){
         return (strcmp(a1->GetName(), a2->GetName()) < 0);
      }
   };
   EnumActors(S_hlp::cbEnum, (dword)&plr_list);
                             //sort players by name
   sort(plr_list.begin(), plr_list.end(), S_hlp::cbPlrLess);
                              //destroy players' menus, they'll be re-created only on existing ones,
                              // in proper order
   for(int i=plr_list.size(); i--; )
      plr_list[i]->cbProc(CB_DESTROY_PLR_MENU);

   for(i=0; i<plr_list.size(); i++)
      plr_list[i]->cbProc(CB_CREATE_PLR_MENU);

   vector<C_smart_ptr<C_actor> > used_actors;
   C_string str;

   while(ck.Size()){
      if(!ck.RAscend(t))
         goto fail;

      switch(t){
      case CT_SG_ACTOR:
         {
            ck.ReadString(CT_NAME, str);
            E_ACTOR_TYPE type = ACTOR_NULL;
            ck.Read(&type, sizeof(byte));
                              //try to find actor of that name and type
            LPC_actor a = NULL;
            if(str.Size()){
                              //find actor in scene - this may fail for run-time created and names actors
               a = FindActor(str);
               //assert(a);
               if(!a){
                  switch(type){
                  case ACTOR_TIME_BOMB:
                  case ACTOR_EXPLOSION:
                     break;
                  default: assert(0);
                  }
                  //ErrReport(C_fstring("LoadGame: Can't find actor: '%s'", (const char*)str), editor);
               }
            }
            if(a){
                              //apply to existing actor
               assert(a->GetType()==type);
               bool b;
               b = a->LoadGame(&ck);
               if(!b)
                  goto fail;
            }else{
                              //create new actor
               a = CreateActor(type);
               bool st = a->LoadGame(&ck);
               if(!st){
                              //init failed
                  DestroyActor(a);
                  a = NULL;
               }
            }
            if(a)
               used_actors.push_back(a);
            ck.Descend();
         }
         break;

      case CT_SG_CAMERA:
         {
            PI3D_camera cam = scene->GetActiveCamera();
            assert(cam);
            if(cam){
               while(ck.Size()){
                  ck.RAscend(t);
                  switch(t){
                     /*
                  case CT_POSITION:
                     {
                        S_vector pos = ck.RVectorChunk();
                        cam->SetPos(pos);
                     }
                     break;

                  case CT_DIRECTION:
                     {
                        S_vector dir = ck.RVectorChunk();
                        cam->SetDir(dir, 0.0f);
                     }
                     break;
                     */
                              //game cam
                  case CT_GC_MODE_DIST:
                     {
                        int pos = ck.RIntChunk();
                        //pos = Max(1, pos);
                        game_cam.SetDistanceMode(pos);
                     }
                     break;

                  case CT_GC_MODE_ANGLE:
                     {
                        int amode = ck.RIntChunk();
                        game_cam.SetAngleMode(amode);
                     }
                     break;

                  default:
                     assert(0);
                     ck.Descend();
                  }
               }
               cam->Update();
            }
            ck.Descend();
         }
         break;

      case CT_SG_STATS:
         {
            ck.Read(&stat_ammo_spent, sizeof(int));
            ck.Read(&stat_num_hits, sizeof(int));
            ck.Read(&stat_enm_killed, sizeof(int));
            ck.Read(&stat_civil_killed, sizeof(int));
            ck.Descend();
         }
         break;
      default:
         assert(0);
         ck.Descend();
      }
   }
                              //release all old actors
   //for(i=actors.size(); i--; ) actors[i]->Release();
   actors = used_actors;
   ck.Close();

   MakeEndDetector();

   game_cam.SetToIdealPosition();
                              //make sure we don't snow/rain inside
   if(snow_rain_on){
      PI3D_sector cam_sct = scene->GetSector(scene->GetActiveCamera()->GetWorldPos());
      if(!cam_sct->IsPrimary())
         EnableSnowRain(false);
   }


   {
      S_tick_context tc;
      memset(&tc, 0, sizeof(tc));
      tc.time = 1;
      Tick(tc);
   }
#ifdef EDITOR
   OUTTEXT("Game loaded");
#endif
   return true;

fail:
   ck.Close();
   return false;
}

//----------------------------

E_MISSION_IO C_game_mission::Open(const char *dir, dword open_flags){

   E_MISSION_IO mio = C_mission::Open(dir, open_flags);
   if(mio==MIO_OK){
      PI3D_camera cam = scene->GetActiveCamera();
      if(!cam){
         cam = driver->CreateCamera();
         scene->AddFrame(cam);
         scene->SetActiveCamera(cam);
         cam->SetFOV(cam_fov);
         cam->SetRange(CAM_NEAR_RANGE, cam_range);
         cam->Release();
      }

      for(int i=actors.size(); i--; )
         actors[i]->TableUpdate(-1, false);
   }
   return mio;
}

//----------------------------

void C_game_mission::Close(){

   if(is_in_game)
      GameEnd();

   m_snow_rain = NULL;
   snow_rain_on = false;

   map_mgr = NULL;
   map_active = false;

   actors.clear();

   SetNetIdBase(0);

   anims.clear();

                              //stop played sounds
   play_sounds.clear();
   last_clip_sound = NULL;

                              //release checkpoints
   checkpoints.clear();

   C_mission::Close();
}

//----------------------------

void C_game_mission::EnumActors(bool(*proc)(LPC_actor, dword), dword c){

   for(int i=0; i<actors.size(); ){
      LPC_actor actor = actors[i];
      if(!proc(actor, c))
         return;
      if(actors[i]==actor)
         i++;
   }
}

//----------------------------

LPC_actor C_game_mission::FindActor(const char *name){

   for(int i=0; i<actors.size(); i++)
   if(actors[i]->frame && !strcmp(actors[i]->frame->GetName(), name))
      return actors[i];
   return NULL;
}

//----------------------------

void C_game_mission::BroadcastMessage(dword msg, dword prm1, dword prm2){

   for(int i=actors.size(); i--; )
      actors[i]->cbProc(msg, prm1, prm2);
}

//----------------------------

void C_game_mission::DestroyActor(LPC_actor a){

   for(int i=actors.size(); i--; )
   if(actors[i]==a){
      if(net && a->net_id && net_id_map){
                              //is this ID valid?
         //map<word, LPC_actor, less<int> >::const_iterator it = net_id_map->id_map.find(a->net_id);
         //if(it!=net_id_map->id_map.end());
         assert(NetFindActor(a->net_id)==a);
         net_id_map->id_map.erase(a->GetNetID());
         assert(!NetFindActor(a->net_id));
      }
      a->Close();
      //actors.erase(&actors[i]);
      actors[i] = actors.back(); actors.pop_back();
      //a->Release();
      break;
   }
}

//----------------------------

void C_game_mission::SetNetIdBase(word w){

   if(w){
      assert(!net_id_map);
      net_id_map = new C_net_id_map(w);
   }else{
      if(net_id_map){
         delete net_id_map;
         net_id_map=NULL;
      }
   }
}

//----------------------------

word C_game_mission::CreateActorNetID(LPC_actor a){

   assert(net_id_map);
   word id;
                              //make unique network ID
   do{
      id = net_id_map->net_id_base + net_id_map->next_net_id;
      ++net_id_map->next_net_id %= 0x2000;
   }while(NetFindActor(id));

   a->SetNetID(id);
   net_id_map->id_map[id] = a;

   NET_LOG(C_fstring("Actor %s: net_id = 0x%x", a->GetName(), id));

   return id;
}

//----------------------------

                              //identify all currently created players
void C_game_mission::NetActorIDMake(){

   assert(net_id_map);
   for(int i=0; i<actors.size(); i++){
      word id = 0x2000 + i;
      actors[i]->SetNetID(id);
      net_id_map->id_map[id] = actors[i];
      NET_LOG(C_fstring("Actor %s: net_id = 0x%x", actors[i]->GetName(), id));
   }
   NET_LOG("Network actor mapping created");
}

//----------------------------

LPC_actor C_game_mission::NetFindActor(word actor_id) const{

   if(!actor_id) return NULL;
                              //find out whom belongs this ID
   map<word, LPC_actor, less<word> >::const_iterator it = net_id_map->id_map.find(actor_id);

                              //is this ID valid?
   if(it==net_id_map->id_map.end())
      return NULL;
   return (*it).second;
}

//----------------------------

void C_game_mission::EndMission(){

   switch(want_state){
   case MS_DONE_SUCCESS:
   case MS_DONE_FAIL:
      {
#ifndef EDITOR
         RenderAllBackbuffers();
         S_menu_debriefing_info di;
                              //ok, moving to debriefing screen
         di.ammo_spent = stat_ammo_spent;
         di.num_hits = stat_num_hits;
         di.enemy_killed = stat_enm_killed;
         di.civil_killed = stat_civil_killed;

         bool save_back = false;
         int curr_mi = game_profile.MissionIndex();
                              //move-on progress
         if(want_state==MS_DONE_SUCCESS){
            save_back = true;
            if(game_profile.curr_mission+1 != game_profile.NumMissionsInCamp()){
                              //mission done
               int i = ++game_profile.curr_mission;
               assert(i < game_profile.NumMissionsInCamp());

                              //increase max. mission we've come to
               if(game_profile.IsAddOn())
                  game_profile.max_addon_mission = Max(game_profile.max_addon_mission, i);
               else
               if(game_profile.curr_camp==game_profile.max_camp)
                  game_profile.max_mission = Max(game_profile.max_mission, i);

                              //copy profile
               S_mission_state &mstate = game_profile.mission_states[game_profile.MissionIndex()];
               mstate = game_profile.mission_states[curr_mi];
               mstate.save_game.Close();
               di.status = S_menu_debriefing_info::DEBRIEF_SUCCESS_MISSION;
            }else{
                              //campaign done
               if(!game_profile.IsAddOn()){
                  int i = game_profile.curr_camp + 1;
                  if(i==NUM_CAMPAIGNS){
                                       //game done
                     di.status = S_menu_debriefing_info::DEBRIEF_SUCCESS_GAME_DONE;
                     save_back = false;
                  }else{
                     game_profile.curr_camp = i;
                     game_profile.curr_mission = 0;
                     if(game_profile.max_camp < game_profile.curr_camp){
                        game_profile.max_camp = game_profile.curr_camp;
                        game_profile.max_mission = 0;
                     }
                                          //copy profile
                     S_mission_state &mstate = game_profile.mission_states[game_profile.MissionIndex()];
                     mstate = game_profile.mission_states[curr_mi];
                     mstate.save_game.Close();
                     mstate.item_list.clear();
                     for(i=0; i<4; i++)
                        mstate.team_4[i].item_list.clear();
                     di.status = S_menu_debriefing_info::DEBRIEF_SUCCESS_CAMPAIGN;
                  }
               }else{
                  di.status = S_menu_debriefing_info::DEBRIEF_SUCCESS_ADDON_DONE;
                  save_back = false;
               }
            }
         }else{
            di.status = S_menu_debriefing_info::DEBRIEF_FAIL;
         }
         GameEnd(&di.allies_killed, save_back);
         if(net) NetSend();

                              //save the profile now (due to crashes :)
         game_profile.Save();

         tick_class = CreateMenu(GAME_MENU_DEBRIEF, (dword)&di);
         if(tick_class)
            tick_class->Release();

         if(want_state==MS_DONE_SUCCESS){
                              //try to play after-game animation
            C_class_to_be_ticked_hd *canim = CreateCampAnim(C_fstring("Aftergame_%.2i", curr_mi), tick_class, false);
            if(canim){
               tick_class = canim;
               canim->Release();
            }
         }
#endif//EDITOR
      }
      break;

   case MS_EXIT:
#ifndef EDITOR
      game_configuration.net_host = false;
      game_configuration.net_join = false;
      tick_class = CreateMainMenu();
      if(tick_class)
         tick_class->Release();
#endif
      break;

   default:
      assert(0);
   }
}

//----------------------------

void C_game_mission::SwitchToMap(){

   if(!map_mgr){
      map_mgr = CreateMap(*this);
      if(!map_mgr)
         return;
      map_mgr->Release();
   }

   scene->AttenuateSounds();
   map_active = true;
   map_mgr->Activate();
   game_menu.SetState(C_game_menu::STATE_EXPAND);
   S_tick_context tc;
   memset(&tc, 0, sizeof(tc));
   tc.time = 1;
   map_mgr->Tick(tc);
   game_menu.Tick(tc);
}

//----------------------------

void C_game_mission::SetGameState(E_MISSION_STATE s, bool net_use_chunk){

   if(net){
      dword eid;
      switch(s){
      case MS_DONE_SUCCESS: eid = NS_DONE_SUCCESS; break;
      case MS_DONE_FAIL: eid = NS_DONE_FAIL; break;
      case MS_EXIT: eid = NS_GAME_EXIT; break;
      default: assert(0);
      }
      if(net_use_chunk)
         net_msg->Ascend(NM_GAME);
      net_sync->BeginEvent(eid);
      if(net_use_chunk)
         net_msg->Descend();
   }else
      want_state = s;
}

//----------------------------

void C_game_mission::NetIn(E_NETWORK_MESSAGE msg_id, PC_net_in_msg in_msg){

   AddRef();

   switch(msg_id){
   case NM_GAME:
      {
         assert(net_id_map);
         net_msg->Ascend(NM_GAME);
         in_msg->Ascend();
                                    //parse the packet
         word msg_code;
         while(msg_code = in_msg->GetID()){
            if(msg_code<0x100){
               word net_id = 0;
               LPC_actor a = NULL;
                                    //codes 0-255 are direct game codes
               switch(msg_code){
               case NM_GAME_CREATE_SHOOT:
                  {
                     S_gun_shoot_init si;
                     si.pos = in_msg->GetVector();
                     si.dir = in_msg->GetVector();
                     si.shoot_item = -1;
                     si.netw_actor = in_msg->GetSenderPID();
                     si.hit_dist = in_msg->GetFloat();
                     si.shooter = NetFindActor(in_msg->GetWord());
                     si.cannon_actor = NULL;
                     if(in_msg->GetBits(1)) si.cannon_actor = NetFindActor(in_msg->GetWord());
                     if(si.cannon_actor) si.gun_snd = (PI3D_sound)si.cannon_actor->cbProc(CB_GET_GUN_SOUND);
                     else
                     if(si.shooter) si.gun_snd = (PI3D_sound)si.shooter->cbProc(CB_GET_GUN_SOUND);
                     CreateActor(ACTOR_GUN_SHOOT, (dword)&si);
                  }
                  break;
               case NM_GAME_CREATE_ROCKET:
                  {
                     S_rocket_init ri;
                     ri.id = in_msg->GetByte();
                     ri.pos = in_msg->GetVector();
                     ri.dir = in_msg->GetVector();
                     ri.shooter = NetFindActor(in_msg->GetWord());
                     ri.netw_actor = in_msg->GetSenderPID();
                     a = CreateActor(ACTOR_ROCKET, (dword)&ri);
                  }
                  break;
               case NM_GAME_CREATE_EXPLOSION:
                  {
                     S_explosion_init ei;
                     ei.id = in_msg->GetByte();
                     ei.pos = in_msg->GetVector();
                     ei.netw_actor = in_msg->GetSenderPID();
                     a = CreateActor(ACTOR_EXPLOSION, (dword)&ei);
                  }
                  break;
               case NM_GAME_CREATE_TIME_BOMB:
                  {
                     net_id = in_msg->GetWord();
                     S_time_bomb_init ti;
                     ti.pos = in_msg->GetVector();
                     ti.netw_actor = in_msg->GetSenderPID();
                     a = CreateActor(ACTOR_TIME_BOMB, (dword)&ti);
                  }
                  break;
               case NM_GAME_CREATE_MINE:
                  {
                     S_mine_init mi;
                     mi.id = in_msg->GetByte();
                     net_id = in_msg->GetWord();
                     mi.pos = in_msg->GetVector();
                     mi.netw_actor = in_msg->GetSenderPID();
                     a = CreateActor(ACTOR_MINE, (dword)&mi);
                  }
                  break;
               case NM_GAME_CREATE_RESOURCE:
                  {
                     S_resource_init ri;
                     net_id = in_msg->GetWord();
                     ri.pos = in_msg->GetVector();
                     ri.inv_id = in_msg->GetByte();
                     ri.amount = in_msg->GetWord();
                     ri.netw_actor = in_msg->GetSenderPID();
                     a = CreateActor(ACTOR_RESOURCE, (dword)&ri);
                  }
                  break;
               case NM_GAME_CREATE_FLARE:
                  {
                     S_flare_init fi;
                     fi.pos = in_msg->GetVector();
                     fi.dir = in_msg->GetVector();
                     fi.netw_actor = in_msg->GetSenderPID();
                     fi.shooter = NetFindActor(in_msg->GetWord());
                     CreateActor(ACTOR_FLARE, (dword)&fi);
                  }
                  break;
               case NM_GAME_CREATE_GRENADE:
                  {
                     S_grenade_init gi;
                     net_id = in_msg->GetWord();
                     gi.pos = in_msg->GetVector();
                     gi.dir = in_msg->GetVector();
                     gi.inv_id = in_msg->GetByte();
                     gi.thrower = NetFindActor(in_msg->GetWord());
                     gi.netw_actor = in_msg->GetSenderPID();
                     a = CreateActor(ACTOR_GRENADE, (dword)&gi);
                  }
                  break;

               case NM_SYNC:
                  {
                     net_sync->NetIn(in_msg, false);

                              //probably something important, better tick mission now to update
                     net_msg->Descend();
                     S_tick_context tc;
                     tc.Zero();
                     Tick(tc);
                     if(!net){
                        in_msg->Descend();
                        Release();
                        return;
                     }
                     net_msg->Ascend(NM_GAME);
                  }
                  break;

               case NM_MAP:
                  assert(map_mgr);
                  map_mgr->NetIn((E_NETWORK_MESSAGE)msg_code, in_msg);
                  break;

               case NM_CHAT:
                  net_chat->NetIn(in_msg);
                  break;

               default:
                  assert(0);
               }
               if(a){
                  if(net_id){
                                    //map created actor to specified ID
                     a->SetNetID(net_id);
                     net_id_map->id_map[net_id] = a;
                  }
                  /*
                                    //hide latency - tick forward
                  S_tick_context tc;
                  tc.Zero();
                  tc.time = in_msg->GetLatency();
                  a->Tick(tc);
                  */
               }
            }else{
                                    //codes greater than 255 are player IDs
               LPC_actor actor = NetFindActor(msg_code);
               if(actor){
                  NET_LOG(C_fstring("C_game_mission::NetIn: message for actor: %s", actor->GetName()));
                  actor->NetIn(in_msg);
               }else{
                              //this actor doesn't exist (probably destroyed in mean-time),
                              // silently skip this message
                  //assert(0);
                  NET_LOG(C_fstring("C_mission::NetIn: unknown actor ID received: 0x%.4x, skipping chunk", msg_code));
                  in_msg->SkipChunk();
               }
            }
         }
         in_msg->Descend();
         net_msg->Descend();
      }
      break;

   case NM_MAP:
      assert(map_mgr);
      map_mgr->NetIn((E_NETWORK_MESSAGE)msg_id, in_msg);
      break;

   case NM_GAME_MENU+1: case NM_GAME_MENU+2: case NM_GAME_MENU+3: case NM_GAME_MENU+4:
   case NM_GAME_MENU+5: case NM_GAME_MENU+6: case NM_GAME_MENU+7: case NM_GAME_MENU+8:
                              //silently ignore
      in_msg->SkipChunk();
      break;

   default:
      assert(0);
      in_msg->SkipChunk();
   }
   Release();
}

//----------------------------

void C_game_mission::OnPlayerDisconnect(dword pid, const char *name){

   net_sync->OnPlayerDisconnect(pid);
   C_mission::OnPlayerDisconnect(pid, name);
   for(int i=actors.size(); i--; ){
      LPC_actor a = actors[i];
      a->NetPlrDisconnect(pid);
   }
   if(map_mgr)
      map_mgr->OnPlayerDisconnect(pid, name);
}

//----------------------------

void C_game_mission::OnSessionLost(){

   C_mission::OnSessionLost();
   for(int i=actors.size(); i--; ){
      LPC_actor a = actors[i];
      a->net_id = NULL;
      a->network_actor = 0;
   }
   SetNetIdBase(0);
   delete net_chat;
   net_chat = NULL;
}

//----------------------------

void C_game_mission::InitSnowRain(){

   m_snow_rain = I3DCAST_MODEL(scene->FindFrame("~05*", ENUMF_MODEL | ENUMF_USEMASK));
   if(m_snow_rain){
      m_snow_rain->SetOn(false);
      m_snow_rain->StopAnimation(0);
   }
   snow_rain_on = (m_snow_rain!=NULL);
   snow_rain_count = 0;

   if(snow_rain_on){
      PI3D_sector cam_sct = scene->GetSector(scene->GetActiveCamera()->GetWorldPos());
      if(!cam_sct->IsPrimary())
         EnableSnowRain(false);
   }
}

//----------------------------

PI3D_animation_set C_game_mission::GetAnim(const char *name){

   int i;
   PI3D_animation_set ret;
                              //look into loaded anims (cache)
   for(i=0; i<anims.size(); i++){
      if(anims[i].name==name){
         ret = anims[i].an;
         if(i!=anims.size()-1){
            anims.push_back(anims[i]);
            anims.erase(&anims[i]);
         }
         return ret;
      }
   }
   ret = driver->CreateAnimationSet();
   for(i=anims_dirs.size(); i--; ){
      I3D_RESULT ir;
      ir = ret->Open(anims_dirs[i]+name, I3DLOAD_ANIM_COMPATIBILITY | I3DLOAD_ERRORS,
         cbLoadScene);
      if(I3D_SUCCESS(ir))
         break;
   }
   if(i==-1){
                              //load error
      ret->Release();
      return NULL;
   }
   if(anims.size()>=ANIM_CACHE_SIZE){
                              //try to release some old animation
      anims.erase(anims.begin(), anims.begin()+anims.size()-ANIM_CACHE_SIZE);
   }
   anims.push_back(S_anim_cache());
   anims.back().an = ret;
   anims.back().name = name;
   ret->Release();
   return ret;
}

//----------------------------

void C_game_mission::TickSnowRain(int time){

   if(snow_rain_on && m_snow_rain && game_cam.GetCamera()){
      if((snow_rain_count -= time) <= 0){
         //snow_rain_count += 100 + S_int_random(100);
         snow_rain_count = Max(0, snow_rain_count + 100 + S_int_random(100));
                              //create new snow-rain element
         PI3D_model mod = driver->CreateModel();
         S_vector v = S_vector(0.0f, 3.0f, 3.0f) * game_cam.GetCamera()->GetMatrix();
         for(int i=3; i--; )
            v[i] += S_float_random(2.0f) - 1.0f;

         mod->Duplicate(m_snow_rain);
         mod->SetOn(true);

         mod->SetPos(v);
         //mod->SetRot(game_cam.GetCamera()->GetWorldRot());
         mod->Update();
#ifdef EDITOR
         mod->SetName("snow_rain_elem");
#endif
         mod->LinkTo(scene->GetPrimarySector());

         LPC_actor act = CreateActor(ACTOR_SNOWRAIN_ELEM);
         act->SetFrame(mod);

         mod->Release();
      }
   }
}

//----------------------------

                              //returns new id
int C_game_mission::PlayerSwitch(bool forward, int id){

   struct S_hlp{
      static bool cbGetPlrs(LPC_actor act, dword v){
         if(act->GetType()==ACTOR_PLAYER){
            vector<LPC_actor> *plrs=(vector<LPC_actor>*)v;
            plrs->push_back(act);
         }
         return true;
      }
   };
   vector<LPC_actor> plrs;
   EnumActors(S_hlp::cbGetPlrs, (dword)&plrs);
   int i;
   C_sort_list<LPC_actor> slist;
   for(i=0; i<plrs.size(); i++){
      if(plrs[i]->IsAlive() || plrs[i]->IsActive() || id != -1)
         slist.Add(plrs[i], plrs[i]->GetMenuID());
   }
   slist.Sort();
   bool show_switch_msg = (net!=NULL);
   if(id != -1){
      if(id >= slist.Count())
         return -1;
                              //can't switch to dead
      if(!slist[id]->IsAlive() || slist[id]->IsActive())
         return -1;

      //id = slist.size()-1-id;
      for(i=0; i<slist.Count(); i++)
      if(slist[i]->IsActive()){
         slist[i]->SetActive(false);
         break;
      }
      slist[id]->SetActive(true, show_switch_msg);
      return id;
   }
                              //find active
   for(i=0; i<slist.Count(); i++)
   if(slist[i]->IsActive()){
      slist[i]->SetActive(false);
      if(forward){
         if(!i--)
            i = slist.Count() - 1;
      }else
      if(++i==slist.Count())
         i = 0;
      slist[i]->SetActive(true, show_switch_msg);
      break;
   }
   if(!slist.Count())
      return -1;
   return slist.GetSortValue(i);
}

//----------------------------

void C_game_mission::MakeEndDetector(){

   LPC_actor act = CreateActor(ACTOR_DETECTOR);
   LPC_table tab = act->GetTable(0);
   tab->Open((dword)"tables\\det_fail.tab", TABOPEN_FILENAME | TABOPEN_UPDATE);
   PI3D_frame dummy = driver->CreateDummy();
   act->SetFrame(dummy);
   dummy->Release();
   act->TableUpdate(-1, false);
}

//----------------------------

LPC_checkpoint C_game_mission::FindCP(const char *cp){

   for(int i=checkpoints.size(); i--; )
      if(!strcmp(checkpoints[i]->GetName(), cp))
         return checkpoints[i];
   return NULL;
}

//----------------------------

LPC_checkpoint C_game_mission::CreateCheckpoint(){

   LPC_checkpoint cp = new C_checkpoint;
   checkpoints.push_back(cp);
   cp->Release();
   /*
   if(tab_game_cfg->ItemB(TAB_B_GAME_CPOINTS))
      cp->Show(true);
      */
   cp_grid_valid = false;
   return cp;
}

//----------------------------

void C_game_mission::DestroyCheckpoint(LPC_checkpoint cd, bool slct_erase){

   int i;
   for(i=checkpoints.size(); i--; ){
      if(checkpoints[i]==cd){
         checkpoints[i] = checkpoints.back(); checkpoints.pop_back();
         break;
      }
   }
   cp_grid_valid = false;
}

//----------------------------

const vector<C_cp_grid_element> &C_game_mission::GetCPGrid() const{

   if(!cp_grid_valid){
      cp_grid.clear();
      cp_grid.reserve(checkpoints.size());
      int i, j;
                              //reset indicies
      for(i=checkpoints.size(); i--; )
         checkpoints[i]->index = i;

      for(i=0; i<checkpoints.size(); i++){
         cp_grid.push_back(C_cp_grid_element());

         LPC_checkpoint cp = checkpoints[i];
         int numc = cp->NumConnections();

         cp_grid.back().cnts.reserve(numc);
         for(j=0; j<numc; j++){
                              //compute dist
            cp->cnts[j].dist = (cp->GetPos()-cp->cnts[j].cp->GetPos()).Magnitude();
            cp->cnts[j].dist *= cp->GetConExpence(j);

            cp_grid.back().cnts.push_back(C_cp_grid_element::S_connection());
            cp_grid.back().cnts.back().cp_index = cp->cnts[j].cp->index;
            cp_grid.back().cnts.back().dist = cp->cnts[j].dist;
            cp_grid.back().cnts.back().flags = cp->cnts[j].flags;
         }
      }
      cp_grid_valid = true;
   }
   return cp_grid;
}

//----------------------------

void C_game_mission::DisplayObjectives(){

                              //clear main console, so that they do not overlap
   console_main.Clear();

   con_objectives.Clear();

   vector<S_objective> obj_list;
                              //collect all objectives from all detectors
   struct S_hlp{
      static bool cbEnum(LPC_actor a, dword c){
         if(a->GetType()==ACTOR_DETECTOR)
            a->cbProc(CB_ADD_OBJECTIVES, c);
         return true;
      }
   };
   EnumActors(S_hlp::cbEnum, (dword)&obj_list);

   if(game_profile.IsDatadisk()){
      for(int i=obj_list.size(); i--; )
         obj_list[i].text_id += 1e6;
   }
                              //sort them by text id
   sort(obj_list.begin(), obj_list.end());
                              //display them all
   for(int i=0; i<obj_list.size(); i++){
      int tid = obj_list[i].text_id;
      C_string str = all_txt[tid];
      bool wide = all_txt.IsWide(tid);
      if(obj_list[i].completed){
         const C_string &str1 = all_txt[GT_GAME_OBJECTIVE_COMPLETED];
         if(!wide){
            str += str1;
         }else{
            int len = wcslen((wchar_t*)&str[0]) + wcslen((wchar_t*)&str1[0]) + 1;
            wchar_t *wp = new wchar_t[len];
            wcscpy(wp, (wchar_t*)&str[0]);
            wcscpy(&wp[wcslen((wchar_t*)&str[0])], (wchar_t*)&str1[0]);
            str.Assign((const char*)wp, len*sizeof(wchar_t));
            delete[] wp;
         }
      }
      con_objectives.OutText(str, OBJECTIVE_COLOR, wide);
   }
}

//----------------------------

LPC_actor C_game_mission::MakeParticle(const char *model_name, const char *table_name,
   const S_vector &disp, PI3D_frame link_to, const S_vector *dir){

   PI3D_model mp = driver->CreateModel();
   model_cache.Open(mp, model_name, scene, 0, ErrReport, editor);

   LPC_actor act = CreateActor(ACTOR_PARTICLE);
   mp->SetPos(disp);
   mp->LinkTo(link_to ? link_to : scene->GetSector(disp));
   act->SetFrame(mp);
   act->cbProc(CB_PAR_SETPARAMS, (dword)table_name, 0);
   if(dir) mp->SetDir(-(*dir), 0.0f);
   mp->Update();
   mp->Release();

   return act;
}

//----------------------------

/*
void C_game_mission::DisplayCross(dword color, int scrn_x, int scrn_y){

   spr_cross->SetOn(true);
   spr_cross->SetColor(color);
   spr_cross->SetScreenPos(scrn_x, scrn_y);
}
*/

//----------------------------

void C_game_mission::DisplayCross(dword color, bool centered){

   spr_cross->SetOn(true);
   spr_cross->SetColor(color);
   aim_cross_centered = centered;
   if(centered){
      float x = igraph->Scrn_sx()/2;
      float y = igraph->Scrn_sy() / 2;
      y /= (screen_aspect_ratio * 1.33333333f);

      spr_cross->SetScreenPos(x, y);
   }
}

//----------------------------

void C_game_mission::DisplayAimLine(const S_vector &from, const S_vector &to){

   aim_line[0] = from;
   aim_line[1] = to;
   aim_line_valid = true;
}

//----------------------------

C_game_mission *CreateGameMission(){

#if defined FINAL || 0
   if(!CDPresenceCheck())
      return NULL;
#endif
   return new C_game_mission;
}

//----------------------------
//----------------------------
