#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"


//----------------------------

#ifdef _DEBUG
//#define DEBUG_DRAW_FLY_PATH
#endif

//----------------------------

static void Bounce(S_vector &vel, const S_vector &hit_normal, float bounce){

   static const float STOP_EPSILON = .1f;

	float backoff = vel.Dot(hit_normal) * bounce;

   vel -= hit_normal * backoff;
	for(int i = 3 ; i--; ){
		if(vel[i] > -STOP_EPSILON && vel[i] < STOP_EPSILON)
			vel[i] = .0f;
	}	
}

//----------------------------
                              //grenades over network:
                              // - primary grenades:
                              //    - evaluation of explosion
                              // - network grenades:
                              //    - visual
                              //    - display counter
                              // - sent over network:
                              //    - word net_id
                              //    - S_vector position
                              //    - S_vector direction
                              //    - byte inv_id
                              //    - word thrower_net_id
//#define GR_NET_DEBUG(n) NET_DEBUG(n)
#define GR_NET_DEBUG(n)

#define GRENADE_TIME 5000
#define GRENADE_RANGE 10.0
#define GRENADE_POWER 1000
#define GRENADE_REACTION_DELAY 50
#define GRENADE_MAX_FLY_TIME 5000

class C_grenade: public C_actor{
   int countdown, fly_time;
   enum{
      MODE_FLY, MODE_LIE,
      //MODE_REACTION
   } mode;
   enum E_TYPE{
      G_TIME, G_COLLISION, G_BODYPART
   } type;
   C_smart_ptr<C_icon_info> ii;
   int col_free_time;
   LPC_actor thrower;

   void ShowTime(){
      if(ii){
         int tt = (countdown/1000) + 1;
         ii->SetText(C_fstring("%i:%.2i", tt/60, tt%60));
      }
   }
   bool DecTime(int t){
      int old_c = countdown;
      countdown -= t;
      if(countdown<=0) return true;
      if((countdown/1000) != (old_c/1000)) ShowTime();
      return false;
   }

   void NetDestroy(){
      if(net && !network_actor && net_id){
                              //send destroy code
         bool g = true;
         net_msg->Ascend(net_id);
         net_msg->AddBits(1, 2, g);
         net_msg->Descend();
      }
   }

   void Explode(){
                              //create explosion
      S_explosion_init ei;
      ei.pos = pos;
      ei.id = EXPLOSION_GRENADE;
      ei.actor_reason = thrower;
      mission.CreateActor(ACTOR_EXPLOSION, (dword)&ei);

      countdown = GRENADE_REACTION_DELAY;
      //mode = MODE_REACTION;
      NetDestroy();
      mission.DestroyActor(this);
   }

   void SetupAnim(){
      if(model){
         static const char *gr_anim[]={"gr_gb", "gr_ger", "body_fly"};
         PI3D_animation_set anim;
         anim = mission.GetAnim(C_fstring("%s.i3d", gr_anim[type]));
         if(anim)
            model->SetAnimation(0, anim, I3DANIMOP_LOOP);
      }
   }
public:
   float explode_range;
   int destruct_power;
   S_vector pos, dir;


   C_grenade::C_grenade(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_GRENADE),
      mode(MODE_FLY),
      countdown(GRENADE_TIME),
      explode_range(GRENADE_RANGE), destruct_power(GRENADE_POWER),
      type(G_TIME),
      col_free_time(0),
      thrower(NULL),
      fly_time(0)
   {
                              //allow init data to be NULL, used during LoadGame
      if(!init_data)
         return;

      const S_grenade_init &gi = *(S_grenade_init*)init_data;
                              //setup model
      if(gi.inv_id){
         model = driver->CreateModel();
         frame = model;
         const char *g_name = tab_inventory->ItemS(TAB_S16_INV_MODEL_S, gi.inv_id, 16);
         model_cache.Open(model, g_name, mission.GetScene(), 0, ErrReport, editor);
                              //setup type and range
         type = (E_TYPE)tab_inventory->ItemE(TAB_E_INV_GRN_TYPE, gi.inv_id);
         explode_range = tab_inventory->ItemF(TAB_F_INV_GRN_RANGE, gi.inv_id);
      }else
      if(gi.model){
         frame = gi.model;
         if(frame->GetType()!=FRAME_MODEL){
            model = driver->CreateModel();
            model->AddFrame(frame);
            frame->LinkTo(model);
            model->SetPos(frame->GetWorldPos());
            model->Update();
            frame->SetPos(S_vector(0, 0, 0));
            frame->Update();
            frame = model;
         }
         type = G_BODYPART;
      }
                              //setup pos/dir
      pos = gi.pos;
      dir = gi.dir;
      //dir *= 1.1f;
      frame->SetPos(pos);
      frame->Update();
      mission.GetScene()->SetFrameSector(frame);
                              //setup animation
      SetupAnim();
      switch(type){
      case G_BODYPART:
         col_free_time = 200;
         break;
      case G_TIME:
         ii = new C_icon_info(mission.texts, 49);
         ii->Release();
         mission.AddIconInfo(ii);
                              //flow...
      default:
         mission.PlaySound("gr_throw", model->GetWorldPos(), 3, 6);
      }
      thrower = gi.thrower;

      network_actor = gi.netw_actor;
                              //send over network
      if(net && !network_actor && gi.net_create){
                              //timed grenade is synced
         if(type==G_TIME)
            mission.CreateActorNetID(this);

         bool g = true;
         net_msg->AddWord(C_game_mission::NM_GAME_CREATE_GRENADE, g);
         net_msg->AddWord(net_id, g);
         net_msg->AddVector(pos, g);
         net_msg->AddVector(dir, g);
         net_msg->AddByte(gi.inv_id, g);
         net_msg->AddWord(thrower ? thrower->GetNetID() : 0, g);
      }
   }
   ~C_grenade(){
      if(ii)
         mission.DelIconInfo(ii);
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      switch(mode){
      case MODE_LIE:
      case MODE_FLY:
         {
            if(type==G_TIME && DecTime(tc.time)){
               if(!network_actor){
                  Explode();
               }
               return 1;
            }
            if(mode==MODE_LIE) return 2;
            float tsec = (float)tc.time*.001f;
            S_vector delta = dir*tsec;

            if(!network_actor){
                                 //check fly time overflow
               if((fly_time += tc.time) >= GRENADE_MAX_FLY_TIME){
                  //ReportActorError("fly time overflow, auto-destroying");
                  switch(type){
                  case G_BODYPART:
                     {
                        PI3D_frame body = model->GetChild(0);
                        if(body){
                           body->LinkTo(NULL);
                           body->SetOn(false);
                        }
                     }
                     break;
                  case G_TIME:
                     NetDestroy();
                     break;
                  }
                  mission.DestroyActor(this);
                  return 1;
               }

               if(!col_free_time){
                  I3D_collision_data cd(pos, delta, 0, NULL, FlyColResp);
                  if(mission.GetScene()->TestCollision(cd)){
                     const S_vector &ret_pos = cd.GetHitPosition();
                     const S_vector &ret_norm = cd.GetHitNormal();

                     switch(type){
                     case G_COLLISION:
                        {
                           pos = ret_pos + ret_norm * .02f;
                           Explode();
                        }
                        return 1;
                     case G_BODYPART:
                        {
                           PI3D_frame body = frame->GetChild(0);
                           if(body){
                              body->LinkTo(NULL, I3DLINK_UPDMATRIX);
                              //scene->SetFrameSector(body);
                              body->SetOn(false);
                           }
                           mission.DestroyActor(this);
                        }
                        return 1;
                     }
                                    //put few cm away of collision volume
                     pos = ret_pos + ret_norm * .02f;
                     float f = dir.Magnitude();
                     if(f<1.0f){
                        mode = MODE_LIE;
                     }else{
                        mission.PlaySound("gr_fall", frame, 2.0f, 10.0f, Min(1.0f, f/10.0f));

                        Bounce(dir, ret_norm, 1.8f);
                        dir *= tab_game_cfg->ItemF(TAB_F_GAME_GRENADE_JUMP);
                     }
                     if(net && !network_actor && net_id){
                                             //send net pos/dir
                        bool g = false;
                        net_msg->Ascend(net_id);
                        if(mode==MODE_LIE){
                           g = true;
                           //net_msg->AddByte(2, g);
                           net_msg->AddBits(2, 2, g);
                           net_msg->AddVector(pos, g);
                        }else{
                           //net_msg->AddByte(0, g);
                           net_msg->AddBits(0, 2, g);
                           net_msg->AddVector(pos, g);
                           net_msg->AddVector(dir, g);
                        }
                        net_msg->Descend();
                     }
                  }else{
#ifdef DEBUG_DRAW_FLY_PATH
                     DebugLine(pos, pos+delta, 0);
#endif
                     pos += delta;
                  }
               }else{
                  col_free_time = Max(0, col_free_time-tc.time);
                  pos += delta;
               }
            }else{
                                 //net actor - fly until direction changed
               pos += delta;
            }

            frame->SetPos(pos);
            frame->Update();
            mission.GetScene()->SetFrameSector(frame);
            dir.y -= tsec * 5.0f;
            if(model) model->Tick(tc.time);
         }
         break;

         /*
      case MODE_REACTION:
         if(DecTime(tc.time)){
                                 //send to all close
            struct S_h{
               static bool cbEnum(LPC_actor a, dword t){

                  C_grenade *gr = (C_grenade*)t;
                  if(gr==a)
                     return true;
                  if(!a->MayExplode())
                     return true;
                  const S_vector *vp = a->GetPos();
                  if(vp){
                     float dist=(*vp -
                        gr->pos).Magnitude();
                     float f = 1.0f - dist / gr->explode_range;
                     if(f>0.0f){
                        a->Explode((LPC_actor)t, (float)gr->destruct_power*f);
                     }
                  }
                  return true;
               }
            };
            mission.EnumActors(S_h::cbEnum, (dword)this);
            mission.DestroyActor(this);
         }
         break;
         */
      }
      return 0;
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      in_msg->Ascend();
      while(in_msg->GetChunkSize()){
         byte code = in_msg->GetBits(2);
         switch(code){
         case 0:              //new pos/dir received
            GR_NET_DEBUG("C_grenade: net - new pos/dir message received");
            pos = in_msg->GetVector();
            dir = in_msg->GetVector();
            break;
         case 1:              //destroy
            GR_NET_DEBUG("C_grenade: destroy net message received");
            mission.DestroyActor(this);
            break;
         case 2:              //set to LIE mode
            GR_NET_DEBUG("C_grenade: LIE net message received");
            mode=MODE_LIE;
            pos = in_msg->GetVector();
            frame->SetPos(pos);
            frame->Update();
            mission.GetScene()->SetFrameSector(frame);
            break;
         }
      }
      in_msg->Descend();
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      SG_SaveModelInfo(ck);
                           //write data
      ck->Write(&countdown, sizeof(countdown));
      ck->Write(&fly_time, sizeof(fly_time));
      ck->Write(&mode, sizeof(byte));
      ck->Write(&type, sizeof(byte));
      ck->Write(&col_free_time, sizeof(col_free_time));
      ck->Write(&explode_range, sizeof(explode_range));
      ck->Write(&destruct_power, sizeof(destruct_power));
      ck->Write(&pos, sizeof(pos));
      ck->Write(&dir, sizeof(dir));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      SG_LoadModelInfo(ck);
                           //read data
      ck->Read(&countdown, sizeof(countdown));
      ck->Read(&fly_time, sizeof(fly_time));
      ck->Read(&mode, sizeof(byte));
      ck->Read(&type, sizeof(byte));
      ck->Read(&col_free_time, sizeof(col_free_time));
      ck->Read(&explode_range, sizeof(explode_range));
      ck->Read(&destruct_power, sizeof(destruct_power));
      ck->Read(&pos, sizeof(pos));
      ck->Read(&dir, sizeof(dir));
      SetupAnim();

      return true;
   }

//----------------------------

   virtual bool MaySaveGame() const{ return (type!=G_BODYPART); }
};

typedef C_grenade *LPC_grenade;

//----------------------------

LPC_actor CreateGrenade(C_game_mission &gm, void *data){

   return new C_grenade(gm, data);
}

//----------------------------
