#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"
#include "smokeray.h"


//----------------------------

#define MAKE_HIT_PARTICLE     //create small particle at hit location

#define RAY_H_SCALE .05f
#define DEFAULT_SHOOT_DIST 300.0f   //used if nothing is hit

//----------------------------
                              //gunshots over network:
                              // - primary shot:
                              //    - evaluation of hit
                              //    - delaying
                              // - network shot:
                              //    - display ray
                              //    - play wheez
                              //    - emit listen sound
                              //    - display gun-smoke at hit position
                              // - sent over network:
                              //    - S_vector position
                              //    - S_vector direction
                              //    - float hit_dist
                              //    - word shooter_id
class C_gun_shoot: public C_actor{
   S_vector pos, dir;
   bool do_smoke;
   S_vector hit_dest, hit_norm;
   int shoot_item;
   float dist;
   float speed;               //m/sec
   LPC_actor hit_actor, shooter;
   PI3D_frame hit_frm;
   bool next_destroy;
   int delay_count;
   int power;
   C_smart_ptr<I3D_sound> gun_sound;
   LPC_actor cannon_actor;    //who created shot, may be NULL
   C_smoke_ray *smoke_ray;

                              //for delayed shots
   PI3D_frame weapon;
   bool weapon_inverse;
   S_vector shot_delta;
   bool aimed;
   bool is_fading;

//----------------------------

   void Finish(){

      if(hit_actor){
         assert(!network_actor);
                              //callback actor
         S_CB_hit sh;
         sh.sub_frame = hit_frm;
         sh.shooter = shooter;
         sh.hit_pos = hit_dest;
         sh.hit_norm = hit_norm;
         sh.hit_pos_ok = true;
         hit_actor->cbProc(CB_HIT, (dword)&sh, power);
      }

      bool good_material_for_smoke(false);
                              //try find material of frame which was hit
      if(hit_frm){
         int mat_id(-1);
         switch(hit_frm->GetType()){
         case FRAME_VOLUME:
            mat_id = I3DCAST_VOLUME(hit_frm)->GetCollisionMaterial();
            if(I3DCAST_VOLUME(hit_frm)->GetVolumeType() != I3DVOLUME_SPHERE)
               good_material_for_smoke = !mat_table->ItemB(TAB_B_MAT_AI_SEE_THROUGH, mat_id);
            break;
         case FRAME_VISUAL:
            mat_id = I3DCAST_VISUAL(hit_frm)->GetCollisionMaterial();
            good_material_for_smoke = true; //fullface always correct
            break;
         }
         if(mat_id != -1){
            const char *cp = mat_table->ItemS(TAB_S12_SND_GUNSHOOT, mat_id, 12);
            if(*cp)
               mission.PlaySound(cp, hit_dest, 2.0f, 8.0f);
         }
      }
                              //do smoke if allowed and is chance that is not in the air
      //if(do_smoke)
      if(do_smoke && good_material_for_smoke)
      {
         if(mission.fps_counter.GetCount() >= 20){
            LPC_actor a = mission.MakeParticle("gunsmoke", "gunsmoke.tab", hit_dest + hit_norm * .06f);
            if(a){
                                 //scale it
               PI3D_particle prt = (PI3D_particle)a->GetFrame()->FindChildFrame("*", ENUMF_VISUAL | ENUMF_USEMASK);
               if(prt && I3DCAST_VISUAL(prt)->GetVisualType()==I3D_VISUAL_PARTICLE){
                  prt->SetProperty(PRT_ELEMENT_SCALE, FloatAsInt(.3f));
               }
            }
         }
         /*
                              //make gun hole
         if((pos-hit_dest).Magnitude() < tab_game_cfg->ItemF(TAB_F_GAME_GUNHOLE_DIST)){
            PI3D_model model = driver->CreateModel();
            mission.LoadModel(model, "gun_hole.i3d");
            model->SetPos(hit_dest + hit_norm*.01f);
            model->SetDir(-hit_norm, Random(GetPI2()));
            model->Update();
            LPC_actor act = mission.CreateActor(ACTOR_GUNHOLE);
            act->SetFrame(model);
            model->LinkTo(hit_frm, I3DLINK_UPDMATRIX);
            model->Release();
         }
         */
      }
                              //try to find out which volume we've hit
      /*PI3D_volume vol = NULL;
      PI3D_frame frm = hit_frm;
      while(frm){
         vol = I3DCAST_VOLUME(frm->FindChildFrame(NULL, ENUMF_VOLUME));
         if(vol) break;
         frm = frm->GetParent();
      }
      if(vol){
                              //check material ID
         const char *cp = mat_table->ItemS(TAB_S12_SND_GUNSHOOT, vol->GetCollisionMaterial(), 12);
         if(*cp)
            mission.PlaySound(cp, hit_dest, 2.0f, 8.0f);
      }*/

      is_fading = true;
      smoke_ray->Finish();
   }

//----------------------------

   void Evaluate(){

      {
                              //check what we'll hit
         PI3D_frame obj(NULL);
         S_vector dest;
         PI3D_frame shooter_frame = shooter ? shooter->GetFrame() : NULL;
         I3D_collision_data cd(pos, dir, I3DCOL_RAY | I3DCOL_TEXTURE | I3DCOL_JOINT_VOLUMES, shooter_frame, FlyColRespGeo);

                              //reworked 21.12. - net actor gets dist, so doesn't compute collision
         if(!network_actor){
                              //compute collision - what will be hit
            if(mission.GetScene()->TestCollision(cd)){
                              //object hit, store for later processing
               obj = cd.GetHitFrm();
               do_smoke = true;
               hit_dest = dest = cd.GetHitPosition();
               hit_norm = cd.GetHitNormal();
               hit_frm = obj;
               if(hit_frm) hit_frm->AddRef();
               dist = (dest-pos).Magnitude();
            }else{
                              //no hit, just compute end of fire ray (for graphics and sound effects)
               do_smoke = false;
               dist = DEFAULT_SHOOT_DIST;
               dest = pos + dir*dist;
            }
         }else{
            dest = pos + dir * dist;
            hit_dest = dest;
            hit_norm = S_normal(-dir);
         }


         if(!network_actor){
                                 //check which actor will be hit
            PI3D_frame hit_obj = obj;
            if(hit_obj)
            do{
               LPC_actor a = (LPC_actor)hit_obj->GetData();
               if(a){
                  hit_actor = a;
                  hit_actor->AddRef();
                  break;
               }
            }while(hit_obj=hit_obj->GetParent());
         }
                              //draw shoot line
#ifdef EDITOR
         if(tab_game_cfg->ItemB(TAB_B_GAME_DEBUG_SHOW_SHOTS))
            DebugLine(pos, pos+dir*dist, 2, 0xffff0000, 3000);
#endif
                              //create ray
         {
            delete smoke_ray;
            S_smoke_ray_init si;
            si.scene = mission.GetScene();
            si.start_pos = pos;
            si.start_dir = dir;
            si.tail_fade_time = 80;
            si.tile_tile = 0.0f;
            si.dest_scale = 1.0f;
            si.tail_name = "tail01";
            smoke_ray = GreateSmokeRay(si);
         }
         {
                              //get closest dist to camera - make wheez sound
            const S_vector &cp = mission.game_cam.GetCamera()->GetWorldPos();
            S_vector dir(dest-pos);
            float u;
            u = dir.Dot(dir);
            if(u>MRG_ZERO){
               u = dir.Dot(cp-pos) / u;
               if(u>0.0f && u<1.05f){
                  S_vector p(pos+dir*u);
                  float dist = (p-cp).Magnitude();
                  if(dist<10.0f){
                     mission.PlaySound(C_fstring("wheez_%i", S_int_random(6)), p,
                        tab_game_cfg->ItemF(TAB_F_GAME_SNDR_GUNWHEEZ_MIN),
                        tab_game_cfg->ItemF(TAB_F_GAME_SNDR_GUNWHEEZ_MAX));
                  }
               }
            }
            S_listen_source ls;
            ls.type = LISTEN_WHEEZ;
            ls.pos = pos;
            ls.power = 5.0f;
                              //don't send over network - shadow shots will emit sound
            mission.EmitListen(ls, tab_game_cfg->ItemF(TAB_F_GAME_SNDR_GUNWHEEZ_MAX), false, shooter,
               &dir);
         }
                              //play sound
         if(gun_sound){
            if(gun_sound->IsPlaying()) gun_sound->SetCurrTime(0);
            else gun_sound->SetOn(true);
         }

                              //send over network
         if(net && !network_actor){
            bool g = false;
            net_msg->AddWord(C_game_mission::NM_GAME_CREATE_SHOOT, g);
            net_msg->AddVector(pos, g);
            net_msg->AddVector(dir, g);
            net_msg->AddFloat(dist, g);
            net_msg->AddWord(shooter ? shooter->GetNetID() : 0, g);
            net_msg->AddBits(bool(cannon_actor), 1, g);
            if(cannon_actor) net_msg->AddWord(cannon_actor->GetNetID(), g);
         }
      }
   }
public:
   C_gun_shoot(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_GUN_SHOOT),
      hit_actor(NULL),
      hit_frm(NULL),
      weapon(NULL),
      aimed(false),
      speed(tab_game_cfg->ItemI(TAB_I_GAME_SHOOT_SPEED)),
      dist(0.0f),
      next_destroy(false),
      is_fading(false),
      smoke_ray(NULL),
      weapon_inverse(false),
      delay_count(0)
   {
      assert(init_data);
      const S_gun_shoot_init &si = *(S_gun_shoot_init*)init_data;

      {
         PI3D_frame dum = driver->CreateDummy();
         SetFrame(dum);
         dum->Release();
      }

      shoot_item = si.shoot_item;

      pos = si.pos;
      dir = si.dir;
      dir.Normalize();
      shooter = si.shooter;
      cannon_actor = si.cannon_actor;
      delay_count = si.delay;
      gun_sound = si.gun_snd;
      power = si.power;
      if(!power && shoot_item!=-1) power = tab_inventory->ItemI(TAB_I_INV_WPN_POWER, shoot_item);
      if(delay_count){
         weapon = si.weapon;
         weapon_inverse = si.weapon_inverse;
         //assert(weapon);
         if(weapon){
            weapon->AddRef();
            if(si.aim_disp){
               shot_delta = *si.aim_disp;
               aimed = true;
            }else{
               S_matrix m(weapon->GetMatrix());
               m.Inverse();
               shot_delta = pos*m;
            }
         }
      }
      network_actor = si.netw_actor;
      if(network_actor){
         dist = si.hit_dist;

         if(shooter){
            PI3D_frame gun = (PI3D_frame)shooter->cbProc(CB_GET_GUN);
            if(gun){
               mission.MakeParticle("sh_flash", NULL, gun->GetWorldPos(), NULL, NULL);
            }
         }
      }
      if(!delay_count)
         Evaluate();
   }

//----------------------------

   ~C_gun_shoot(){
      if(hit_actor) hit_actor->Release();
      if(hit_frm) hit_frm->Release();
      if(weapon) weapon->Release();
      delete smoke_ray;
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      if(is_fading){
         smoke_ray->UpdateFade(tc.time);
         if(smoke_ray->IsOver())
            mission.DestroyActor(this);
         return 0;
      }

      int i;
      if(delay_count){
         if((delay_count -= tc.time) > 0)
            return 0;
         delay_count = 0;
         if(weapon){
            pos = shot_delta * weapon->GetMatrix();
            if(aimed){
               S_vector cpos, cdir;
               mission.game_cam.GetCameraPos(cpos, cdir);
               //dir = pos-cam->GetWorldPos();
               dir = pos - cpos;
               dir.Normalize();
               pos += dir * tab_inventory->ItemF(TAB_F_INV_WPN_SHOT_Z, shoot_item);
            }else{
               dir = weapon->GetWorldDir();
               if(weapon_inverse) dir.Invert();
            }
         }else{
            mission.game_cam.Tick(1);
            mission.game_cam.GetCameraPos(pos, dir);
            pos += dir * .5f;
         }
                              //adjust dir a little bit
         for(i=0; i<3; i++)
            //dir[i] += Random(.02f)-.01f;
            dir[i] += S_float_random() * .02f - .01f;
         dir.Normalize();

         Evaluate();
      }

      if(next_destroy){
                              //destroy actor
         Finish();
         return 0;
      }

      float tsec = ((float)tc.time) * .001f;

                              //move shoot
      float d1 = speed * tsec;
      float ray_len(d1);
      if((dist -= d1) <= 0.0f){
         next_destroy = true;
         ray_len += dist;
      }
      smoke_ray->UpdateTrajectory(pos, dir);
      smoke_ray->UpdateFade(tc.time);

      pos += dir * d1;
      return 0;
   }

//----------------------------

   virtual bool MaySaveGame() const{ return false; }
};

//----------------------------

LPC_actor CreateGunShoot(C_game_mission &gm, void *data){

   return new C_gun_shoot(gm, data);
}

//----------------------------
