#include <tabler2.h>
#include <insanity\texts.h>
#include <isound2.h>
#include "i3d_font.h"
#include <inet2.h>
#include "fade.h"
#include "FPS_counter.h"
#include <map>
#include <igraph2.h>
#include <insanity\os.h>
#ifdef EDITOR
#include <i3d\editor.h>
#endif
#include <i3d\insanity.h>

//#define JOYSTICK

#ifdef JOYSTICK
#include ".\ijoy.h"
#endif

//#define MAIN_ANIM_STAGE 4 // which stage is main for human animations, affect num stages too

#if defined _DEBUG && defined DEBUG_NEW_ALLOC
                              //memory allocation debugging
# include <crtdbg.h>
# define DEBUG_CLIENTBLOCK new(_NORMAL_BLOCK, __FILE__, __LINE__)
# define new DEBUG_CLIENTBLOCK
#endif //_DEBUG

//----------------------------
                              //constants

#define MAX_BASE_SOLDIERS 40
#define NUM_CAMPAIGNS 9       //number of campaigns
#define NUM_MISSIONS 32       //number of missions

#define MAX_NUM_ADDON_MISS 6  //maximal number of missions in an Add-on

#define CAM_NEAR_RANGE .2f
#define CAM_COL_SPHERE .3f
                              //special material which is ignored by bullets and rockets.
                              // set on human spheres at creation.
#define HUMAN_SPHERE_MAT 42
                              //special material - drowning humans in vehicle immediate
#define AUTO_KILL_WATER_MAT 43
                              //material for bushes -vehicle dide through, ai transparent.
#define BUSH_RIDE_FREE_MAT 44

//----------------------------

enum E_GAME_MENU{
   GAME_MENU_4_OF_8 = 1,
   GAME_MENU_8_OF_40,
   GAME_MENU_SOLDIER_INFO,
   GAME_MENU_CAMPAIGN_LOAD,
   GAME_MENU_CAMPAIGN_SELECT,
   GAME_MENU_MISSION_SELECT,
   GAME_MENU_MISSION_STATUS,
   GAME_MENU_DEBRIEF,
};

//----------------------------
                              //globals
extern PI3D_driver driver;
extern LPIGraph igraph;
extern const char APP_NAME[];
extern class C_sprite_image icons_img, font_img;
extern class ISND_driver *isound;
extern bool allow_stream_snd;
extern float screen_aspect_ratio;    //width/height
extern LPC_table tab_game_cfg, tab_campaigns;
extern LPC_table tab_inv[2], tab_human_anim, mat_table, tab_resistance, tab_sound_envs;
extern C_all_texts all_txt;

#ifdef EDITOR
#define tab_inventory (tab_inv[game_configuration.datadisk_ver])
#else
#define tab_inventory (tab_inv[game_profile.IsDatadisk()])
#endif

#ifdef JOYSTICK
extern LPC_JoyBase joy_base;
extern LPC_Joy joy;
extern LPC_FF_effect ff_effect;
#endif //JOYSTICK

//----------------------------
                              //functions
dword I3DAPI cbFrame(PI3D_frame, FRM_CALLBACK_MESSAGE, dword prm1, dword prm2);
I3D_LOAD_CB_PROC cbLoadScene;
I3DENUMRET I3DAPI cbSetAllCallback(PI3D_frame frm, dword mask);
bool CDPresenceCheck();

bool GetVoiceName(const char *voice, int human_id, C_string &str);

/*
struct S_model_load{
   PI3D_model model;
   PI3D_scene scene;
};

I3DENUMRET I3DAPI cbModelLoad(PI3D_frame frm, dword s1);
*/
I3DENUMRET I3DAPI cbAddVolumeToScene(PI3D_frame frm, dword c);

//----------------------------
// Render screen using currently active tick class, including game texts.
void RenderScreen();

//----------------------------
// Render screen (using RenderScreen function) into all backbuffers.
void RenderAllBackbuffers();

//----------------------------
                              //global game config
struct S_game_configuration{
                              //video:
   bool use_wbuffer, tripple, fullscreen, tx_compress, render_shadows;
   dword mode_x, mode_y;
   dword bits_per_pixel;
   int adapter_identifier;
   dword device_type;
   bool hw_vp;
   int brightness;            //0 ... 100
   enum E_FILTER{
      FILTER_LINEAR,
      FILTER_ANISO,
   } filter;

                              //sounds:
   bool use_sounds;
   int volume_music, volume_effects, volume_speech;    //0 ... 100
   bool sounds_use_3D_hw, sounds_use_EAX;
   bool safe_mode;

                              //network:
   IN_CONNECTION net_ctype;
   C_string net_tcpip_addr;
   C_string net_tcpip_port;
   C_string net_modem_str;
   C_string net_modem_num;
   C_string net_plr_name;
   C_string net_sess_name;
   bool enable_voice_chat;
   enum E_VOICE_COMPRESSION{  //this enum encodes voice compression Kb/sec
      VCC_1_2,
      VCC_3_2,
      VCC_6_4,
      VCC_8,
      VCC_13,
      VCC_32,
      VCC_64,
      VCC_LAST,
   } voice_chat_compression;

                              //other:
   bool use_dbase;

   bool draw_aim_line;
   byte aim_mode;             //0 = weapon, 1 = cross, 2 = empty

                              //installation:
   C_string inst_src, inst_dst;
   enum E_ISNTALL_TYPE{
      INSTALL_MINIMAL,
      INSTALL_MEDIUM,
      INSTALL_FULL,
      INSTALL_FULL_EDITOR
   } install_type;
   byte language;
   byte censored;
   byte datadisk_ver;         //0 = original, 1 = FFF

                              //debugging support (set from command-line only):
   bool net_debug_schedule_cpu;
   bool net_host;
   bool net_join;
   bool net_debug_log;
   int net_debug_numplayers;
   C_string net_cmdline_mission;
                              //command-line override
   C_string net_cmdline_plr_name;
   C_string net_cmdline_sess_name;
   C_string net_cmdline_tcpip_addr;
   C_string net_cmdline_tcpip_port;

//----------------------------
// Read/write config from/to system registry.
   bool ReadSysConfig();
   bool WriteSysConfig();

//----------------------------
// Read info about installation.
   bool ReadInstallationInfo();

//----------------------------
// Reset sound configuration defaults.
   void ResetSoundDefaults();

//----------------------------
// Reset video configuration defaults.
   void ResetVideoDefaults();

//----------------------------
// Reset video configuration defaults.
   void ResetNetworkDefaults();

   void ResetGameDefaults();

//----------------------------
// Apply current settings.
   void ApplySoundConfig();
   void ApplyVideoConfig();

   S_game_configuration();
};

extern S_game_configuration game_configuration;

//----------------------------

#ifdef EDITOR
extern LPC_editor editor;
#else
#define editor NULL
#endif

//----------------------------
// Error reporting function/empty macro for various problems - also to be passed into
// C_I3D_cache::Open and C_I3D_cache::Create functions.
// Reports specified error (text message) to Log window.
// Note: context parameter must be pointer to C_editor, or NULL.
//----------------------------

#ifdef EDITOR

void ErrReport(const char *cp, void *context);
void WarnReport(const char *cp, void *context);

#else //EDITOR

void __stdcall LogFunc(const char *cp);
inline void ErrReport(const char *cp, void*){ LogFunc(cp); }
inline void WarnReport(const char *cp, void*){ LogFunc(cp); }


#endif//!EDITOR

//----------------------------

class C_class_to_be_ticked_hd: public C_class_to_be_ticked{
public:
//----------------------------
   virtual void Notify(PI3D_frame, dword notify_key, dword anim_time){}
};

extern C_smart_ptr<C_class_to_be_ticked_hd> tick_class;

//----------------------------
                              //inline functions

//----------------------------
// Get 32-bit dword bitmask of float and vice-versa.
inline int FloatAsInt(float f){ return *(dword*)&f; }
inline float IntAsFloat(int i){ return *(float*)&i; }

//----------------------------
                              //debugging lines
#ifdef EDITOR
inline void DebugLine(const S_vector &from, const S_vector &to, int type, dword color = -1, int time = 500){
   LPC_editor_item ei = editor->FindPlugin("DebugLine");
   if(ei) ei->Action(E_DEBUGLINE_ADD + type, &S_add_line(from, to, color, time));
}
inline void DebugLineClear(){
   LPC_editor_item ei = editor->FindPlugin("DebugLine");
   if(ei) ei->Action(E_DEBUGLINE_CLEAR);
}
#else//EDITOR
inline void DebugLine(const S_vector &from, const S_vector &dir, int type, dword color = -1, int time = 500){}
inline void DebugLineClear(){}
#endif//!EDITOR

//----------------------------
//----------------------------

class C_console: public C_unknown{
   vector<struct S_out_text*> out_texts;
   float text_x, text_y;
   int out_text_time;
   float line_dist;
   float line_top, line_bottom;
   float line_size;
   dword scroll_time;
   dword justify;

   void OrderTexts();
public:
   C_poly_text texts;
   C_console(float x, float y, float line_size, float line_dist, dword scroll_time = 3000, float bottom = .75f, dword justify = 0);
   ~C_console();

   void OutText(const char *cp, dword color=-1, bool wide = false);
   void Clear();
   void Tick(int time);
};

extern C_console console_main, console_debug;

void OUTTEXT(const char *cp, dword color = -1);
void OUTTEXT(int);
void OUTTEXT(float);
void OUTTEXT(const S_vector&);

void DEBUG(const char *cp);
void DEBUG(int);
void DEBUG(float);
void DEBUG(const S_vector&);

//----------------------------

                              //human inventory
enum E_INV_ITEM{
                              //0
   INV_NULL, INV_WEAPON, INV_GRENADE, INV_MINE,
                              //4
   INV_HANDS, INV_FLAREGUN, INV_BAZOOKA, INV_PANZER,
                              //8
   INV_TIME_BOMB, INV_CAMERA, INV_UNIFORM, INV_KNIFE,
                              //12
   INV_BINOCULARS, INV_COLT, INV_KEY, INV_RADIO,

   INV_LAST,
};

//----------------------------

class C_inventory{
   class C_game_menu &game_menu;

   struct S_item: public C_unknown{
      int itm;
      C_smart_ptr<I3D_model> model;
      C_smart_ptr<I3D_interpolator> intp;
      C_smart_ptr<C_text> num;
      dword amount;
      dword bullets_in_stack;
      S_item(){}
      S_item(int i1, int am):
         itm(i1),
         amount(am), bullets_in_stack(0)
      {}
      S_item(const S_item &itm);
      S_item &operator =(const S_item&);
   };
   vector<C_smart_ptr<S_item> > items;
   C_smart_ptr<I3D_scene> inv_scene;   //when NULL, inventory is not visible
   int active_item;
   //bool visible;

   friend C_game_menu;

   void Tick(int time){
      for(int i=items.size(); i--; )
         items[i]->model->Tick(time);
   }
   void LoadModels(PI3D_scene, C_poly_text&);
   void ReleaseModels();
   void MakeOrder(bool place_models=true);
public:
   C_inventory(C_game_menu&);
   ~C_inventory();

   inline int NumItems() const{ return items.size(); }
   inline int GetItem(int i) const{ return items[i]->itm; }
   inline void ReplaceItem(int i, int new_itm){ items[i]->itm = new_itm; }
   inline dword GetItemAmount(int i) const{ return items[i]->amount; }
   void SetItemAmount(int i, dword a);

//----------------------------
// Add item into inventory. The returned value is index where item has been put.
   int AddItem(int itm, dword num = 1);

   void SetBullets(int i, dword nb);
   inline dword GetBullets(int i) const{ return items[i]->bullets_in_stack; }
   bool Reload(int i, bool set);

   void DeleteItem(int i);
   void DeleteAllItems();
   inline bool IsVisible() const{ return (inv_scene!=NULL); }
   inline int GetActiveItem() const{ return active_item; }
   void SetActiveItem(int i);
   int ItemIndex(int itm_id){
      for(int i=items.size(); i--; )
         if(items[i]->itm==itm_id)
            break;
      return i;
   }
   void Scroll(bool left);
   inline const vector<C_smart_ptr<S_item> > &GetItems() const{ return items; }

   static bool HasAmmo(int itm_id);
   static float GetItemWeight(int itm, int amnt = 1);
   float CountWeight() const{
      float ret = 0.0f;
      for(int i=NumItems(); --i; ) ret += GetItemWeight(GetItem(i), GetItemAmount(i));
      return ret;
   }
};

//----------------------------
                              //game camera
class C_game_camera{
   class C_game_mission *mission;
                              //user setting
   C_smart_ptr<I3D_frame> focus_dir, focus_pos;
   C_smart_ptr<I3D_frame> aim_model, aim_ref;
   S_vector delta_dir;
   C_smart_ptr<I3D_dummy> cam_hlp;  //for animation (explosions, shaking, etc.)
   class C_actor *user;
   bool reset;

   C_smart_ptr<I3D_interpolator> interp;
   C_smart_ptr<I3D_keyframe_anim> anim;
   C_smart_ptr<I3D_camera> real_cam;
   C_smart_ptr<I3D_model> debug_cam;
   C_smart_ptr<I3D_frame> cam;           //may be real_cam or debug_cam
   C_smart_ptr<I3D_model> m_snow_rain;
   C_smart_ptr<I3D_interpolator> intp_shake;
   S_vector last_dest_pos;
   byte mode_dist;            //0 = nearest
   byte mode_angle;           //0 = horizontal
   bool acquire;
   int subj_seen_count;       //Tick counts, CheckCol uses this for checking to see subject
                              //help
   S_vector look_at_pos;
   float base_fov, zoom_fov_mult;
   friend dword I3DAPI cbCam(PI3D_frame, FRM_CALLBACK_MESSAGE, dword, dword);
   void GetIdealPos(S_vector&, S_vector&, bool);
   inline void UpdateFPView(int t);
   inline void PrepareFlyAnim(const S_vector&, const S_vector&);
public:
   void SetDistanceMode(dword m, bool skip_fly = true);
   void SetAngleMode(dword m){
      if(m<3) mode_angle=m;
   }
   void SetAimModel(PI3D_frame aim_model, PI3D_frame aim_ref);
   PI3D_frame GetAimModel() const{ return aim_model; }

   byte GetDistanceMode() const{ return mode_dist; }
   byte GetAngleMode() const{ return mode_angle; }
   const S_vector &GetCameraPos() const;
   void SetCameraPos(const S_vector &pos, const S_vector &dir);
   void GetCameraPos(S_vector &pos, S_vector &dir) const;

   C_game_camera(PI3D_scene);
   ~C_game_camera();
   void InitFirst(C_game_mission&);
   void Init(PI3D_camera);
   void Close();
   void SetFocus(PI3D_frame fpos, PI3D_frame fdir, class C_actor *user = NULL);
   class C_actor *GetUser() const{ return user; }
   void SetZoom(float zoom_mult, bool force_upd = false);
   float GetZoom() const{ return zoom_fov_mult; }
   void Tick(int t);
   PI3D_frame GetFocus() const{ return focus_pos; }
   PI3D_frame GetFocusDir() const{ return focus_dir; }
   void SetDeltaDir(const S_vector &v){ delta_dir=v; }
   inline PI3D_frame GetCamera() const{ return cam; }
   void Acquire(bool);
   bool IsAcquired() const{ return acquire; }

   void CheckCol(PI3D_frame);

   void SetAnim(const char*, bool loop=false);

//----------------------------
// Reset camera to ideal position.
   void SetToIdealPosition();
};

//----------------------------

enum E_CHEAT_ID{
   CHEAT_ZOMBIE=1,
   CHEAT_BIGHEAD,
};
                              //actor class
enum E_ACTOR_TYPE{
   ACTOR_NULL,                //0
   ACTOR_PLAYER,
   ACTOR_ENEMY,
   ACTOR_GUN_SHOOT,
   ACTOR_CANNON,              //4
   ACTOR_PARTICLE,
   ACTOR_GUNHOLE,
   ACTOR_VERSION,
   ACTOR_AUTO_CANNON,         //8
   ACTOR_GRENADE,
   ACTOR_MINE,
   ACTOR_RESOURCE,
   ACTOR_DOOR,                //0xc
   ACTOR_FLARE,
   ACTOR_EXPLOSION,
   ACTOR_ROCKET,
   ACTOR_AUTOMOBIL,           //0x10
   ACTOR_RANDOM_GEN,
   ACTOR_TEXTURE_SHIFT,
   ACTOR_TIME_BOMB,
   ACTOR_DETECTOR,            //0x14
   ACTOR_LADDER,
   ACTOR_KILL_ANIM,
   ACTOR_SNOWRAIN_ELEM,

   ACTOR_LAST
};

//----------------------------

enum E_SIGNAL_TYPE{
   SIGNAL_TIME,               //from mission time
   SIGNAL_ALERT,
   SIGNAL_SHOOTING,
   SIGNAL_DETECTOR,           //singal from ACTOR_DETECTOR
   SIGNAL_PROGRAM,
};

//----------------------------

typedef struct S_signal_data{
   class C_actor *actor;      //emmitor
   E_SIGNAL_TYPE type;
   dword sub_type;
   S_vector origin;           //ALERT
   int sig_id;                //TIME
   S_signal_data(){ memset(this, 0, sizeof(S_signal_data)); }
} *LPS_signal_data;

//----------------------------

struct S_CB_hit{
   PI3D_frame sub_frame;
   class C_actor *shooter;
   S_vector hit_pos, hit_norm;
   bool hit_pos_ok;
   S_CB_hit(): hit_pos_ok(false){}
};

//----------------------------

struct S_prg_add{
   dword d[5];
   S_prg_add(){ memset(this, 0, sizeof(S_prg_add)); }
   S_prg_add(dword d0, dword d1 = 0, dword d2 = 0, dword d3 = 0, dword d4 = 0){
      d[0]=d0;
      d[1]=d1;
      d[2]=d2;
      d[3]=d3;
      d[4]=d4;
   }
};

//----------------------------

enum{                         //actor callback messages
   CB_SETPARAM,
   //CB_PRG_LOAD,
   //CB_SOUND,                  //(PI3D_frame, dword anim_notify)
   //CB_ANIM_NOTIFY,            //(PI3D_frame, dword anim_notify)
   CB_BEFORE_UPDATE,          //(PI3D_frame)
   CB_HIT,                    //(S_CB_hit*, int power), ret: true = dead
   //CB_MAY_EXPLODE,            //ret: (bool)
   //CB_EXPLODE,                //(LPC_actor caller, int power)
   CB_GET_EXPLODE_RATIO,      //(LPC_actor caller, (S_vector*)center); ret: (float)ratio 0.0 - 1.0
   //CB_GETPOS,                 //ret: S_vector*
   //CB_SETPOS,                 //(S_vector *pos, S_vector *dir = NULL)
   //CB_GETDIR,                 //ret: S_vector*
   //CB_GETHEADDIR,             //ret: S_vector*
   //CB_GETHEADPOS,             //ret: S_vector*
   //CB_GETCENTERPOS,           //(LPS_vector ret)
   //CB_GETBRESTPOS,            //ret: S_vector*
   //CB_GETTABLEINDEX,          //ret: int index (to be edited)
   //CB_GETTABLE,               //(int index); ret: LPC_table
   //CB_GETTEMPLATE,            //(int index); ret: LPC_table_template
   //CB_ISANIMATED,             //ret: bool
   //CB_TABLEUPDATE,            //(int index, bool not_load_prg)
   CB_APPLY_DIFFICULTY,
   CB_TABLE_MESSAGE,          //(int msg, int index = -1)
   //CB_SETMODEL,               //(PI3D_model)

   //CB_IS_FOR_USE,             //(LPC_actor asker) ret: int num_of_use_positions

   //CB_MAY_SAVE_GAME,          //ret: bool
   //CB_SAVE_GAME,              //(C_chunk*)
   //CB_LOAD_GAME,              //(C_chunk*); ret: bool

   //CB_SAVETABLE,              //(C_chunk*, LPC_table); use CT_TABLE
   //CB_SAVETABLE_EX,           //(C_chunk*, int tab_id); use CT_TABLE_EX

   //CB_MISSIONSAVE,            //(C_chunk*)
   //CB_MISSIONLOAD,            //(C_chunk*, int table_id)

   //CB_ISALIVE,                //ret: bool
   //CB_ISDESTROYED,            //ret: bool

   CB_SIGNAL,                 //(LPS_signal_data, float power) - give signal
   //CB_ISHUMAN,                //ret: bool

                              //player only:
   //CB_GET_MENU_ID,
   //CB_SETACTIVE,              //(bool)
   //CB_ISACTIVE,               //ret: bool
                              //player programming
   CB_PRG_SHOW,               //(bool pos_change_only, int comp_times <0=no, 1=all, 2=only first>) - (re)send prg list to menu

   //CB_DEL_PRG,                //int key; ret: bool success
   CB_GET_PRG,                //int key, ret: E_PRG_ITEM

   //CB_SET_PRG_DATA,           //set data for program
                              //int key, void *data; ret: bool success
                              //PRG_MOVE: data = (LPS_vector)dest, data2 = (bool)run, data3 = E_MOVE_REASON, data4 = (bool)force_smart
                              //PRG_GUARD: data = (LPS_vector)dest, [data2 = (LPC_checkpoint)aim_dest, data3 = countdown, data4 = (S_vector*)dest, data5 = (LPC_actor)subject]
                              //PRG_FOLLOW: data = (LPC_actor)subject
                              //PRG_ATTACK: data = (LPC_actor)subject, [data2 = attack_body_part_index, data3 = E_ATTACK_MODE]
                              //PRG_USE_INV: data = int inv_itm_id
                              //PRG_WAIT_SIGNAL: data = int signal_id
                              //PRG_STEP: data = (LPS_vector)dest, [data2 = (float)goto_min_dist, data3 = (S_vector*)turn_dir, data4 = (bool)run]
                              //PRG_SUSPEND: data = dword time
                              //PRG_WATCH: data = (LPS_vector)dest, [data2 = phase_count (def.=4)]
                              //PRG_USE: data = (LPC_actor)subject, [data2 = use_indx]
                              //PRG_WALK_CP: data = (LPC_checkpoint)start_CP, [data2 = (LPC_checkpoint)next, data3 = (bool)run]
                              //PRG_SETVISIBLE: data = (bool)on
                              //PRG_LADDER_CLIMB: data = (LPC_actor)ladder
                              //PRG_SEND_SIGNAL: data = int signal_id, data2 = (LPC_actor)subject (NULL=broadcast)
                              //PRG_DRIVE: data = (LPS_vector)dest, data2 = ideal_speed

   CB_PRG_PREPROCESS,         //(int mode<0=all, 1=process_time for key #0>, 2=last only)

                              //int key [|CB_SET_PRG_DATA], int property; ret: void *data
                              //PRG_MOVE: data = (S_vector*)dest
   CB_GET_PRG_DATA,
   CB_SET_PRG_TIME,           //int key, int suggest_time; ret: new time
   //CB_GET_PRG_TIME,           //int key; ret: time
   CB_GET_PRG_SIZE,           //ret: num_of_keys

                              //particle:
   CB_PAR_SETPARAMS,          //(char *table_name)
                              //cheating:
   CB_CHEAT,                  //(E_CHEAT_ID)
   //CB_DUPLICATE,              //(LPC_actor original)
   CB_IN_WATER,               //water collision detected
   CB_SETFACE,                //(int face_index, bool masked)
   CB_SETINVENTORY,           //(vector<S_item>*)
   CB_GETINVENTORY,           //ret: vector<S_item>*
   CB_STOREINVENTORY,         //(vector<S_item>*)

   CB_USE_DOOR,               //use door actor
                              //prm1 = message:
                              //0 - response to user: begin use, prm2=(LPC_actor)door
                              //1 - human to door: open! (LPC_actor who_is_opening)
                              //2 - check if closed; ret: bool
                              //3 - door to user: do you have a key?; ret: bool

   CB_SETFOCUS,               //(int <0=off, 1=on, 2=on if active>), ret: true = success
   CB_IS_FOCUSED,             //ret: bool

   CB_SET_NETWORK_ACTOR,      //(dword pid)
   CB_IS_NETWORK_ACTOR,       //ret: bool

   //CB_NET_GAME_LOST,
   CB_NET_PLAYER_LOST,        //dword pid

   //CB_GAME_BEFORE,

   CB_SHOW_PLAYER_NAME,       //(bool force, int time)
   CB_IS_NET_PLAYER,          //ret: bool

   CB_GET_GUN,                //ret: PI3D_frame





   CB_LAST_1
};

//----------------------------
                              //player program elements
enum E_PRG_ITEM{

   PRG_MOVE = 0,
   PRG_FOLLOW = 1,
   PRG_USE = 2,
   PRG_USE_INV = 3,

   PRG_ATTACK = 4,
   PRG_GUARD = 5,
   PRG_WATCH = 6,
   PRG_STAY_UP = 7,

   PRG_STAY_MID = 8,
   PRG_STAY_DOWN = 9,
   PRG_READY = 0xa,                 //auto at the end of list
   PRG_DEAD = 0xb,                  //auto if dead

                              //other (AI)
   PRG_RESTART = 0xc,               //for actors with fixed program
   PRG_PRG1 = 0xd,                  //debug
   PRG_WALK_CP = 0xe,               //walk on CP grid
   PRG_WAIT_SUSPENDED = 0xf,        //wait for signal freezed

   PRG_SETVISIBLE = 0x10,            //out of game forever
   PRG_SUICIDE = 0x11,               //... (development needs)
   PRG_STEP = 0x12,                  //make small step to dest pos
   PRG_SUSPEND = 0x13,               //suspend for a short time (allow grenade throw etc)

   PRG_LADDER_CLIMB = 0x14,          //climb ladder up or down
   PRG_DRIVE = 0x15,                 //drive vehicle
   PRG_SEND_SIGNAL = 0x16,           //communicate with others
   PRG_WAIT_SIGNAL = 0x17,           //wait for signal active

   PRG_LAST
};

//----------------------------
                              //AI - listening support

enum E_listen_type{           //types sorted by importance
   LISTEN_WHEEZ,
   LISTEN_SHOOT,
   LISTEN_EXPLOSION,
   LISTEN_HUMAN,
   LISTEN_VEHICLE,
   LISTEN_FLARE,
};

struct S_listen_source{
   E_listen_type type;
   S_vector pos;
   float power;
   int sub_type;              //voice command if LISTEN_HUMAN type
   S_listen_source():
      sub_type(0)
   {}
   S_listen_source(E_listen_type t, S_vector p1, float po1):
      type(t), pos(p1), power(po1),
      sub_type(0)
   {}
};

typedef vector<S_listen_source> t_listen_list;

//----------------------------
#ifdef EDITOR
                              //Insanity editor plugins:

                              //Plugin: "Mission"
                              //Dependence list: 
                              //Purpose: open/close edited scene
enum E_ACTION_MISSION_CONT{
   E_MISSION_OPEN = E_MISSION_LAST,
   E_MISSION_SAVE,            //(bool ask); ret: (bool)success
   E_MISSION_RELOAD,
   E_MISSION_MERGE,
   E_MISSION_LOG_TOGGLE,
   E_MISSION_SET_OPEN_PROC,   //(bool(*)(const char*))
   E_MISSION_PROPERTIES,      
   E_MISSION_AUTOSAVE_CONFIG,
   E_MISSION_SAVE_ASK_TOGGLE, // - togle save query messagebox
   E_MISSION_GET,
   E_MISSION_CHECK_IN,
   E_MISSION_CHECK_OUT,       //ret: 0=false, 1=ok, 2=ok and reloaded
   E_MISSION_CHECK_BSP,       //ret: false=action canceled, true = builded or user don't want build.

   E_MISSION_ACTOR_CREATE,    // - create actor on selection
   E_MISSION_ACTOR_DESTROY,   // - create actor on selection
   E_MISSION_ACTOR_DRAW_NAMES,// - toggle draw actor's names

   E_MISSION_GCAM_DIST_P,     
   E_MISSION_GCAM_DIST_N,
   E_MISSION_GCAM_ANGLE_P,
   E_MISSION_GCAM_ANGLE_N,
   E_MISSION_GCAM_ACQUIRE,
   E_MISSION_GCAM_IS_ACQUIRED,
   E_MISSION_DRAW_CAMPATH,    //toggle draw camera path

   E_MISSION_VERSION,         // - show current version
   E_MISSION_SET_MISSION,     //(const char *mission_name) - explicitly set mission name
   E_MISSION_SET_MODEL,       //(const char *mission_name) - set model for preview
   E_MISSION_ENABLE,          //(bool) - enable/disable mission functionality

                              // - set time scale:
   E_MISSION_TIME_0,          //0 - stop timer
   E_MISSION_TIME_5,          //5%
   E_MISSION_TIME_10,
   E_MISSION_TIME_20,   
   E_MISSION_TIME_50,   
   E_MISSION_TIME_100,  
   E_MISSION_TIME_200,  
   E_MISSION_TIME_500,
   E_MISSION_TIME_1000,

                              // - simulate fps
   E_MISSION_SIMFPS_OFF,      //off
   E_MISSION_SIMFPS_5,        //5fps
   E_MISSION_SIMFPS_10,
   E_MISSION_SIMFPS_15,
   E_MISSION_SIMFPS_20,
   E_MISSION_SIMFPS_50,


   E_MISSION_TIME_GET,        //get current time scale; ret: float
   E_MISSION_GET_MISSION,     //get editing mission; ret: LPC_game_mission

   E_MISSION_MODE_EDIT,       //go to edit mode
   E_MISSION_MODE_GAME,       //go to game mode

   E_MISSION_SET_ACTOR,       // - internal (undo)

                              //setting 'models on map' flag:
   E_MISSION_ADD_SEL_TO_MAP,  //mark selection
   E_MISSION_DEL_SEL_FROM_MAP,//clear mark on selection
   E_MISSION_SEL_MAP_MARKED,  //select all marked
   E_MISSION_FLASH_MAP_MARKED,//flash all marked

   E_MISSION_RT_ERR_MESSAGE,  //(const char*) - write run-time error message
};
#endif

//----------------------------

class C_actor: public C_unknown{
                              //for computing center position
   mutable S_vector center_delta;
   mutable bool center_comp;

   friend class C_game_mission;

protected:
   class C_game_mission &mission;
   E_ACTOR_TYPE type;
   word net_id;               //identification over network. all non-network actors have id = 0
   PI3D_model model;
   PI3D_frame frame;
   int resistance;
   bool init_ok;
   dword network_actor;       //pid of network player controlling this, or 0 if local actor

   t_listen_list listen_list; //listening support - for humans and detector
   void ListenTick(int time, float villigance = 1.0f);

   int last_mat_id;           //floor material id
   int last_mat_id_count;     //help counter
   int small_climb;           //disables falling when !0

   explicit C_actor(C_game_mission&, E_ACTOR_TYPE at);
   ~C_actor();
   void Close();
                              //save game support
   bool SG_SaveModelInfo(class C_chunk*, bool save_name = true) const;
   bool SG_LoadModelInfo(class C_chunk*, bool ascend = false);

//----------------------------
// Report error in actor to log, display actor's name and type, and error message.
   void ReportActorError(const char*) const;
public:
   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0);
   PI3D_model GetModel() const{ return model; }
   PI3D_frame GetFrame() const{ return frame; }
   word GetNetID() const{ return net_id; }
   virtual void SetNetID(word);

   virtual void SetVisible(bool on, bool vol_only = false, bool net_send = true);
   virtual void ReceiveListen(S_listen_source &ls, bool net_send = true);

   bool PlayAnim(int id_from, int id_to, bool force=false);

   virtual dword Tick(const S_tick_context&){ return 0; }
   inline E_ACTOR_TYPE GetType() const{ return type; }
   const char *GetName() const{
      assert(frame);
      if(!frame) return "";
      return frame->GetName();
   }

   virtual void NetIn(class C_net_in_msg *in_msg);
   virtual void NetCodeIn(byte msg_code, class C_net_in_msg *in_msg);
//----------------------------
// Network actor (specified by PID) was disconnected.
   virtual void NetPlrDisconnect(dword pid);

   virtual bool AddProgram(int pos, E_PRG_ITEM itm, const S_prg_add &pa, bool net_send){ return false; }
   virtual bool DelProgram(int pos, bool net_send){ return false; }

   virtual void SoundNotify(PI3D_frame, dword notify_id, dword stage);
   virtual bool AnimNotify(PI3D_frame, dword notify_id, dword stage){ return false; }
   virtual const S_vector *GetPos() const{ return frame ? &frame->GetWorldPos() : NULL; }
   virtual const S_vector *GetDir() const{ return frame ? &frame->GetWorldDir() : NULL; }
   virtual const S_vector *GetHeadPos() const{ return NULL; }
   virtual const S_vector *GetHeadDir() const{ return NULL; }
   virtual const S_vector *GetBrestPos() const{ return NULL; }
   virtual bool GetCenterPos(S_vector &pos) const{

      if(!frame)
         return false;
      if(!center_comp){
                              //try to use dummy 'center'
         PI3D_frame center = frame->FindChildFrame("center", ENUMF_DUMMY);
         if(center){
            center_delta = center->GetWorldPos() * frame->GetInvMatrix();
         }else{
            I3D_bbox bb;
            I3D_bound_volume bvol;
            frame->ComputeHRBoundVolume(&bvol);
            bb = bvol.bbox;
            center_delta = bb.min + (bb.max-bb.min)*.5f;
         }
         center_comp = true;
      }
      pos = center_delta * frame->GetMatrix();
      return true;
   }

   virtual int NumUsePositions(C_actor *asker = NULL) const{ return 0; }

//----------------------------
// Fill use info for particular index.
   virtual const S_vector GetUsePos(int indx, bool abs_index_mode = false) const{ return frame->GetWorldPos(); }
   virtual float GetUsePosThresh() const{ return 1.0f; }
   virtual bool GetUseDir(S_vector &ret, int indx, bool abs_index_mode = false) const{ return false; }
   virtual float GetUseDirThresh() const{ return PI*.1f; }
   virtual bool CheckUseFreeHands(int indx, bool abs_index_mode = false) const{ return false; }

//----------------------------
// Return stay mode from which item may be used. Returns E_STAY_MODE.
   virtual dword CheckUseStay(dword suggest_stay_mode, int indx) const{ return suggest_stay_mode; }

   virtual bool IsAlive() const{ return false; }
   virtual void SetActive(bool on, bool write_activate_msg = false){ }
   virtual bool IsActive() const{ return false; }
   virtual bool IsDestroyed() const{ return false; }
   virtual bool IsHuman() const{ return false; }
   virtual bool IsBoat() const{ return false; }

//----------------------------
   virtual void SetFrame(PI3D_frame);

//----------------------------
// prm: (use_pos<<16)|bool on [|0x8000 = all_index_based])
// ret: 0 = fail, 1 = ok, 2 = in progress
   virtual int Use(C_actor *user, dword prm, bool force_use_off = false){ return 0; }
   virtual bool MayListen() const{ return false; }
   virtual void ListenReset(){ listen_list.clear(); }
   virtual int GetEnemyGroup() const{ return 0; }
   virtual bool IsSeeing(C_actor*) const{ return false; }
   virtual bool ObstacleMove(const struct S_obstacle_move&){ return false; }
   virtual bool IsEnemy(C_actor*){ return false; }
   virtual bool IsOnMap() const{ return false; }
   virtual int GetMenuID() const{ return -1; }

   virtual int GetTableIndex() const{ return 0; }
   virtual LPC_table GetTable(int index) const{ return NULL; }
   virtual class C_table_template *GetTemplate(int index) const{ return NULL; }
   virtual void TableUpdate(int index, bool not_load_prg){}
   void SaveTable(C_chunk *ck, LPC_table tab) const;
   void SaveTableEx(C_chunk *ck, int tab_id) const;

   inline bool LoadTable(C_chunk *ck, LPC_table tab){ return tab->Open((dword)ck->GetHandle(), TABOPEN_FILEHANDLE | TABOPEN_UPDATE); }

//----------------------------
// Load/Save game (SaveGame).
   virtual bool MaySaveGame() const{ return true; }
   virtual void SaveGame(C_chunk*) const{}
   virtual bool LoadGame(C_chunk*){ return true; }

//----------------------------
// Load/Save mision (into binary).
   virtual void MissionSave(C_chunk *ck) const{}
   virtual bool MissionLoad(C_chunk *ck, int table_id){ return true; }

//----------------------------
// Explosion.
   virtual bool MayExplode() const{ return false; }
   virtual bool Explode(C_actor *caller, int power){ assert(0); return false; }

//----------------------------
// Return crash resistance.
   virtual float Crash(C_actor *who, float power){ return 1.0f; }

   virtual void GameBegin(){}
   virtual void GameEnd(){}
};

typedef C_actor *LPC_actor;

//----------------------------
//----------------------------
                              //checkpoint data

typedef class C_checkpoint *LPC_checkpoint;

                              //move path - collection of checkpoint indicies
typedef vector<int> t_move_path;


enum E_CHECKPOINT_TYPE{
   CP_PATH, CP_USER,
   CP_DRIVE,
   CP_LAST
};

//#define CP_CON_ONEWAY   1
//#define CP_CON_ONEWAYTO 2
#define CP_CON_USE_LADDER 4
#define CP_CON_USE_DOOR 8

class C_checkpoint: public C_unknown{
public:
   struct S_connection{
      LPC_checkpoint cp;
      float dist;
      float expence;
      dword flags;
      S_connection():
         flags(0),
         expence(1.0)
      {}
   };
private:
   vector<S_connection> cnts;
   E_CHECKPOINT_TYPE type;
   S_vector pos;
   C_string name;
   word uin;                  //keeps CP index

   friend class C_game_mission;
   friend class C_edit_checkpoint;
public:
#ifdef EDITOR
   bool help_visited;
#endif

   bool recurse;
   dword debug_flag;
                              //pathfinding help
   LPC_checkpoint prev;
   int index;
   int prev_index;
   float dist;

   C_checkpoint();
   ~C_checkpoint();

   LPC_checkpoint GetCPoint(int i) const{
      assert(i>=0 && i<cnts.size());
      if(i < 0 || i >= cnts.size())
         return cnts.front().cp;
      return cnts[i].cp;
   }
   dword NumConnections() const{ return cnts.size(); }

//----------------------------
// Get access to connection area.
   const S_connection &GetConnection(int i) const{
      assert(i>=0 && i<cnts.size());
      return cnts[i];
   }

   void Connect(LPC_checkpoint);
   void Disconnect(LPC_checkpoint);
   bool IsConnected(LPC_checkpoint) const;
   int GetConnectionIndex(LPC_checkpoint) const;
   void SetConnectionData(int i, dword d);
   dword GetConnectionData(int i) const{ return cnts[i].flags; }
   float GetConExpence(int i) const{ return cnts[i].expence; }
   void SetConExpence1(int i, float);

   void Reset1();
   const S_vector &GetPos() const{ return pos; }
   const C_string &GetName() const{ return name; }
   void SetPos(const S_vector &p){ pos = p; }
   void SetName(const C_string &n){ name = n; }

   void SetType(E_CHECKPOINT_TYPE t){ type = t; }
   inline E_CHECKPOINT_TYPE GetType() const{ return type; }

   void PFReset(int i){
      index=i;
      prev = NULL;
      prev_index = -1;
   }
   dword GetIndex() const{ return uin; }
};

//----------------------------
                              //CheckPoints grid
class C_cp_grid_element{
   struct S_connection{
      int cp_index;
      float dist;
      dword flags;
   };
   vector<S_connection> cnts;

   friend C_game_mission;
public:
   bool valid_source;         //used for inherited grides to disable this CP as source
public:
   C_cp_grid_element(const C_cp_grid_element &ge){ operator =(ge); }

   C_cp_grid_element():
      valid_source(true)
   {}
   C_cp_grid_element &operator =(const C_cp_grid_element &ref){
      valid_source = ref.valid_source;
      cnts = ref.cnts;
      return *this;
   }

   void DisConnect(int cp_indx){
      for(int i=cnts.size(); i--; )
      if(cnts[i].cp_index == cp_indx){
         cnts[i] = cnts.back(); cnts.pop_back();
         break;
      }
   }
   int GetCPoint(int indx) const{ return cnts[indx].cp_index; }
   dword NumConnections() const{ return cnts.size(); }
   dword GetConnectionData(int i) const{ return cnts[i].flags; }
   float GetDistance(int indx) const{ return cnts[indx].dist; }
};

//----------------------------
//----------------------------

class C_icon_info: public C_unknown{
   int sprite_mask;

   C_poly_text &texts;
   C_smart_ptr<C_text> text;

   C_smart_ptr<C_poly_sprite> spr;
   C_string str_text;

   float pos[2];
   float text_size;
public:
   C_icon_info(C_poly_text&, int sprite_mask);
   void GetPos(float &x, float &y) const;
   void SetPos(float x, float y);
   void SetText(const char*);
   inline int GetID() const{ return sprite_mask; }
   void Render();
   void ResetScreen();
};

typedef C_icon_info *LPC_icon_info;

//----------------------------
//----------------------------
                              //collision volume types
                              //type of shape
#define FRAME_CHECKPOINT FRAME_USER

//----------------------------

enum E_MISSION_STATE{
   MS_NULL,
   MS_DONE_SUCCESS,
   MS_DONE_FAIL,
   MS_EXIT,
};

//----------------------------
                              //mission loader - options
enum E_MISSION_OPEN_FLAGS{
   OPEN_MERGE_MODE = 1,       //merge with current scene
                              //created frames:
   OPEN_MERGE_SOUNDS = 2,          
   OPEN_MERGE_LIGHTS = 4,
   OPEN_MERGE_MODELS = 8,
   OPEN_MERGE_VOLUMES = 0x10,
   OPEN_MERGE_DUMMYS = 0x20,
   OPEN_MERGE_CAMERAS = 0x40,
   OPEN_MERGE_TARGETS = 0x80,
   OPEN_MERGE_OCCLUDERS = 0x100,
                              
   OPEN_MERGE_LIGHTMAPS = 0x10000, //lightmap info
   OPEN_MERGE_SCRIPTS = 0x20000,   
   OPEN_MERGE_ACTORS = 0x40000,   
   OPEN_MERGE_CHECKPOINTS = 0x80000,

   OPEN_MERGE_SETTINGS = 0x100000,     //scene settings (ranges, FOV, abckground)

   OPEN_NO_SHOW_PROGRESS = 0x200000,   //don't show progress indicator
   OPEN_NO_EDITOR = 0x400000,          //don't load editor info
   OPEN_LOG = 0x800000,                //log loading
   OPEN_NO_OPEN_FAIL_MSG = 0x1000000,  //don't display message if scene opening fails
   OPEN_USE_SCENE_CAMERA = 0x2000000,  //find camera in scene and make it active camera
};

enum E_MISSION_IO{            //file I/O statuse, used for C_mission::Open and Save
   MIO_OK,                    //operation succeeded
   MIO_CORRUPT,               //file corrupted
   MIO_ACCESSFAIL,            //access failed (e.g. write operation on read-only file)
   MIO_NOFILE,                //cannot find specified file
   MIO_CANCELLED,             //cancelled by user
};

//----------------------------

struct S_objective{
   int text_id;
   bool completed;
   
   bool operator<(const S_objective &o) const{
      int i, i1;
      bool b;
      if(game_configuration.language!=5){
         b = sscanf(all_txt[text_id], "%i", &i);
         if(b) b = sscanf(all_txt[o.text_id], "%i", &i1);
      }else{
         b = swscanf((wchar_t*)(const char*)all_txt[text_id], (wchar_t*)"%\0i\0\0", &i);
         if(b) b = swscanf((wchar_t*)(const char*)all_txt[o.text_id], (wchar_t*)"%\0i\0\0", &i1);
      }
      if(!b) return (text_id<o.text_id);
      return (i<i1);
   }
};

#define OBJECTIVE_COLOR             0xffffff80
#define OBJECTIVE_COLOR_COMPLETED   0x80ffff80

//----------------------------

#define MAX_PLAYERS 4

struct S_prg_item{
   E_PRG_ITEM prg_item;
   int show_time;
   bool active;
};

class C_game_menu{
public:
   enum E_MENU_STATE{
      STATE_COMPACT, STATE_EXPAND
   };
private:
   C_smart_ptr<C_poly_sprite> spr_compass;
   float curr_comp_angle;

   class C_plr_menu *pmenu[MAX_PLAYERS];
   dword num_players;
   E_MENU_STATE state;

   C_poly_text &texts;
   vector<C_smart_ptr<C_text> > axis_nums;

   C_inventory *curr_inv;
   int time_offset;
   bool CreateAxisNums();
   C_fade<S_fade_color> time_num_fade;
public:
                              //inventory stuff
   C_smart_ptr<I3D_scene> inv_scene;     //NULL = not init

   C_game_menu(C_poly_text&);
   ~C_game_menu();

   void SetState(E_MENU_STATE);
   void ResetScreen();
                              //player submenu
   dword AddPlayerMenu(dword face_id);
   void SetPlayerFace(dword plr_id, dword face_id);
   void DestroyPlayerMenu(dword plr_id);
   void SetPrgList(dword plr_id, const vector<S_prg_item>&, bool pos_only=false);
   void SetActive(dword plr_id);

   void SetDeathFace(dword plr_id);

   void SetHealth(dword plr_id, float f);
   void SetInventory(C_inventory*);
                              //prg key flashing
   void SetPrgKeyColor(dword plr_id, dword key_id, dword color);
   dword GetPrgKeyColor(dword plr_id, dword key_id);

   void SetVisible(bool on, bool comp_on=true);
   void Tick(const S_tick_context&);
   void Render();

   void SetCompassAngle(float a);
   bool HitCheck(float pos_x, float pos_y, int &plr_id, int &sub_id);
   bool IconX(int, float&);
   int GetTimeOffset() const{ return time_offset; }
   void ScrollTimeAxis(int time);
};

//----------------------------

#define CHECKCOL_NORESPONSE   1

                              //mission class
class C_mission: public C_class_to_be_ticked_hd{
protected:
   bool open_ok;

   dword GraphCallback(dword msg, dword par1, dword par2);
   static GRAPHCALLBACK GraphCallback_thunk;

                              //3d stuff
   C_string mission_name;

   PI3D_scene scene;

public:

   inline PI3D_scene GetScene() const{ return scene; }

protected:

   dword level_time;          //since mission loaded
   int next_time_value;       //next waited time key in tab_property
   vector<bool> used_time_value;

                              //title texts
   friend class C_actor;

public:
   struct S_load_context{
      C_chunk ck;
      map<C_string, PI3D_frame> frame_map;
      bool merge_mode;
      dword open_flags;
      PI3D_camera cam;
      float ortho_scale;
      bool is_ortho;
      bool cancelled;         //set to true if user presses ESC
      bool dump_chunks;
      int bsp_file_pos;       //position of CT_BSP_FILE chunk in file, or -1 if uninitialized (for postponing loading)
                              //load callback
      I3D_LOAD_CB_PROC *load_cb_proc;
      void *load_cb_context;
#ifdef EDITOR
      LPC_editor_item e_log, e_undo, e_modify;
      vector<C_string> err_list;
#endif
      C_string open_dir;
      S_load_context():
#ifdef EDITOR
         e_log(NULL),
         e_undo(NULL),
         e_modify(NULL),
#endif
         bsp_file_pos(-1),
         dump_chunks(false),
         last_progress_pos(0),
         cancelled(false),
         cam(NULL),
         ck(256*1024)
      {}

      int last_progress_pos;

   //----------------------------
   // This func returns true is user wants to cancel loading.
      bool ShowProgress(){
         int pos = ck.GetHandle()->tellg();
         if(pos<last_progress_pos || (pos-last_progress_pos) > 128){
            last_progress_pos = pos;
            float f = (float)pos / (float)ck.GetHandle()->filesize();
            cancelled = (*load_cb_proc)((I3D_LOADMESSAGE)10000, *(dword*)&f, 1, load_cb_context);
            return cancelled;
         }
         return false;
      }
#ifdef EDITOR
      void Log(const char *t, PI3D_frame frm, dword color){
         if(e_log && find(err_list.begin(), err_list.end(), t)==err_list.end()){
            if(!err_list.size()){
               e_log->Action(E_LOG_ADD_RICH_TEXT,
                  &pair<const char*, dword>(
                  C_fstring("*** Loading scene: %s ***", (const char*)open_dir), color));
            }
            err_list.push_back(t);
            e_log->Action(E_LOG_ADD_RICH_TEXT, &pair<const char*, dword>(t, color));
            if(frm){
               LPC_editor_item e_slct = editor->FindPlugin("Selection");
               if(e_slct)
                  e_slct->Action(E_SELECTION_BLINK_FRAME, frm);
            }
         }
      }
      void ErrLog(const char *t, PI3D_frame frm){ Log(t, frm, 0xff0000); }
      void WarnLog(const char *t, PI3D_frame frm){ Log(t, frm, 0x008000); }
#else //EDITOR
      inline void ErrLog(const char*, PI3D_frame){}
      inline void WarnLog(const char*, PI3D_frame){}
#endif   //!EDITOR

   };
protected:
   dword mission_crc;
                              //camera settings
   float cam_range;
   float cam_fov;

   virtual bool LoadChunk(CK_TYPE, struct S_load_context&);

public:

//----------------------------
// Get CRC of mission on disk.
   static bool GetMissionCRC(const C_string &mission_name, dword &out_crc);

//----------------------------
// Get CRC of currently open mission.
   inline dword GetCRC() const{ return mission_crc; }
                              //tables
                              //property table
   LPC_table tab_property;

   C_mission();
   ~C_mission();

   virtual E_MISSION_IO Open(const char *dir, dword options);
   virtual void Close();
   bool IsOpen() const{ return open_ok; }

   virtual void Notify(PI3D_frame, dword notify_key, dword anim_time);

                              //mission name
   inline const C_string &GetMissionName() const{ return mission_name; }

   virtual dword GetID() const{ return 'MISS'; }

   virtual void BroadcastMessage(dword msg, dword prm1=0, dword prm2=0){}

   virtual void Render();
   virtual void Tick(const S_tick_context&);
   virtual void ResetScreen();

   virtual void NetIn(E_NETWORK_MESSAGE msg_id, class C_net_in_msg *in_msg){}

   void SetFrameSector(PI3D_frame);
   void LinkToFloor(PI3D_frame, const S_vector &pos, PI3D_frame frm_ignore = NULL, bool change_pos = false);
};

//----------------------------

class C_map_manager: public C_mission{
public:
//----------------------------
// Activate map - initialize actors, etc.
   virtual void Activate() = 0;

//----------------------------
// Suspend map - close actors, hints, etc.
   virtual void Suspend() = 0;
};

C_map_manager *CreateMap(C_game_mission&);

//----------------------------
#define MAX_CHEAT_LEN   16

class C_game_mission: public C_mission{
   char cheat_buf[MAX_CHEAT_LEN];
   int cheat_pos;

   bool is_in_game;

   C_smart_ptr<I3D_model> m_snow_rain;
   C_smart_ptr<I3D_scene> aim_render_scene;
   bool snow_rain_on;
   int snow_rain_count;

   vector<C_smart_ptr<C_actor> > actors;

   LPC_actor CreateActor1(int type, dword data = 0);  //internal

   void TickSnowRain(int time);

   E_MISSION_STATE want_state;

                              //free sounds
   vector<C_smart_ptr<I3D_sound> > play_sounds;
   C_smart_ptr<I3D_sound> last_clip_sound;  //keep this so that they don't play over each other

   vector<LPC_icon_info> icon_info_list;
   C_smart_ptr<C_poly_sprite> spr_cross;

   C_smart_ptr<C_map_manager> map_mgr;
   bool map_active;

                              //nice menu showing synchro items in network game
   class C_sync_menu *sync_menu;
                              //screen effects - hidden implementation
   void *screen_effects;

   S_vector aim_line[2];
   bool aim_line_valid, aim_cross_centered;

   C_smart_ptr<C_text> debug_text;
   enum{
      DEBUG_NO,
      DEBUG_FPS,
      DEBUG_COORDS,
      DEBUG_NET_STATS,
   } debug_dump;
   void DisplayDebugText();

   vector<C_string> anims_dirs;
                              //anim cache
   struct S_anim_cache{
      C_smart_ptr<I3D_animation_set> an;
      C_string name;
      S_anim_cache(){}
      S_anim_cache(const S_anim_cache &ac){ operator =(ac); }
      S_anim_cache &operator =(const S_anim_cache &ac){
         an = ac.an;
         name = ac.name;
         return *this;
      }
   };
   vector<S_anim_cache> anims;

   void cbScene(SCN_CALLBACK_MESSAGE msg, dword prm1, dword prm2);
   static dword I3DAPI cbScene_thunk(SCN_CALLBACK_MESSAGE msg, dword prm1, dword prm2, void *context){
      if(context)
         ((C_game_mission*)context)->cbScene(msg, prm1, prm2);
      return 0;
   }

   void SwitchToMap();
private:
                              //checkpoints
   vector<C_smart_ptr<C_checkpoint> > checkpoints;
   mutable vector<C_cp_grid_element> cp_grid;
   mutable bool cp_grid_valid;
public:
   C_smart_ptr<C_poly_sprite> spr_use_drop;  //usable by C_player
   C_console con_objectives;
   C_poly_text texts;
   C_fps_counter fps_counter;

   enum E_SCREEN_EFFECT_MODE{
      SCREEN_EFFECT_ADD,
      SCREEN_EFFECT_MUL,
   };
//----------------------------
// Create screen effect.
// Parameters:
// 'priority' - sorting ID, lower numbers are processed first (blending precedence)
   void CreateScreenEffect(dword color, dword come_time, dword stay_time, dword leave_time,
      E_SCREEN_EFFECT_MODE mode = SCREEN_EFFECT_ADD, int priority = 0);


                              //statistics
   int stat_ammo_spent, stat_num_hits, stat_enm_killed, stat_civil_killed;

   LPC_checkpoint CreateCheckpoint();
   void DestroyCheckpoint(LPC_checkpoint, bool slct_erase=true);
   C_smart_ptr<C_checkpoint> *GetCheckpoints(){ return checkpoints.begin(); }
   int NumCheckpoints() const{ return checkpoints.size(); }
   LPC_checkpoint FindCP(const char*);
   LPC_checkpoint GetCPData(PI3D_frame frm) const{
      return (LPC_checkpoint)frm->GetData();
   }
   void CPSetVisible(bool);
   void DrawCPConnections();
   void InvalidateCPGrid(){ cp_grid_valid=false; }
                              //CP grid
   const vector<C_cp_grid_element> &GetCPGrid() const;

                              //anims
   void AddAnimDir(const C_string &str){
      anims_dirs.push_back(str);
      if(anims_dirs.back()[(int)str.Size()-1]!='\\')
         anims_dirs.back() += "\\";
   }

   PI3D_animation_set GetAnim(const char *name);
   I3D_RESULT LoadAnimSet(PI3D_animation_set set, const char *name){
      I3D_RESULT ir=I3DERR_NOFILE;
      for(int i=0; i<anims_dirs.size(); i++){
         ir=set->Open(anims_dirs[i]+name);
         if(I3D_SUCCESS(ir)) break;
      }
      return ir;
   }
   void AddIconInfo(LPC_icon_info);
   void DelIconInfo(LPC_icon_info);

   PI3D_sound GetActionSound(dword mat_id, dword act_id);
   PI3D_sound CreateSound(const C_string &name, bool err_report = true);
   inline void AddPlaySound(PI3D_sound snd){ play_sounds.push_back(snd); }
   bool PlaySoundGlobal(const char *snd_name);

   inline dword GetLevelTime() const{ return level_time; }

   C_game_mission(const C_game_mission&);
   C_game_mission &operator =(const C_game_mission&);
public:
   C_game_menu game_menu;
   C_game_camera game_cam;
   C_smart_ptr<I3D_volume> cam_vol; //used for keep active human's head out of walls
   inline PI3D_volume GetCamVolume() const{ return cam_vol; }

   C_game_mission();
   ~C_game_mission();

   virtual dword GetID() const{ return 'GMIS'; }

   virtual void Tick(const S_tick_context&);
   virtual void Render();
   E_MISSION_IO Save(const char *bin_name);

   virtual bool LoadChunk(CK_TYPE ck_t, S_load_context &lc);

   bool SaveGame() const;
   bool LoadGame();

   bool LoadGame(const char*, struct S_savegame *info);
   bool SaveGame(const char*, int slot);

   void ProcessCheat(char k);

   void GameBegin(bool setup_team = true, bool net_debug = false);
   void GameEnd(vector<int> *plr_killed=NULL, bool write_back=false);

   void EndMission();

   virtual E_MISSION_IO Open(const char *dir, dword options);
   virtual void Close();

   void InitSnowRain();
   void EnableSnowRain(bool b){ snow_rain_on = b; }
                              //network support
private:
   class C_net_id_map *net_id_map;
   class C_net_chat *net_chat;
public:
   enum E_NETWORK_MESSAGE_GAME{
                              //modes:
      NM_GAME_CREATE_SHOOT = 1,
         //S_vector pos
         //S_vector dir
         //float hit_dist
         //word shooter_id
         //? word cannon_id
      NM_GAME_CREATE_ROCKET = 2,
         //byte id
         //S_vector pos
         //S_vector dir
         //word shooter_id
      NM_GAME_CREATE_EXPLOSION = 3,
         //byte id
         //S_vector pos
      NM_GAME_CREATE_TIME_BOMB = 4,
         //word net_id
         //S_vector pos
                              //time-bomb private messages:
            //bits:2
            //0 - activate
            //1 - set time
               //word time (in sec)
            //2 - cancel using
      NM_GAME_CREATE_MINE = 5,
         //byte id
         //word net_id
         //S_vector pos
                              //mine private messages:
            //bits:1
            //0 - explosion query
               //dword collision_time
            //1 - query answer (may explode)
      NM_GAME_CREATE_RESOURCE = 6,
         //word net_id
         //S_vector pos
         //byte itm_id
         //word amount
      NM_GAME_CREATE_FLARE = 7,
         //S_vector pos
         //S_vector dir
      NM_GAME_CREATE_GRENADE = 8,
         //word net_id (may be 0)
         //S_vector pos
         //S_vector dir
         //byte id
         //word thrower_net_id
                              //grenade private messages (timed only):
            //bits:2
            //0 - wall collision
               //S_vector new_pos
               //S_vector new_dir
            //1 - destroy
            //2 - set to LIE mode
               //S_vector new_pos
      NM_SYNC = 0x10,
         //sync class data

      NM_MAP = 0x11,
         //map class' data

      NM_CHAT = 0x12,
         //C_net_chat data
   };

   enum E_NET_SYNC_EVENT{
      NS_NULL,
      NS_GAME_BEGIN,
      NS_GAME_EXIT,
      NS_TO_MAP,
      NS_FROM_MAP,
      NS_TO_BRIEFING,
      NS_DONE_SUCCESS,
      NS_DONE_FAIL,
   };
                              //for syncing game exit, menu mode, etc
   C_smart_ptr<class C_net_sync> net_sync;

   word CreateActorNetID(LPC_actor);
   void SetNetIdBase(word);
   void NetActorIDMake();
   LPC_actor NetFindActor(word id) const;

                              //actors
   LPC_actor CreateActor(int type, dword data = 0);
   void DestroyActor(LPC_actor a);
   void EnumActors(bool(*)(LPC_actor, dword), dword);
   LPC_actor FindActor(const char *name);
   const vector<C_smart_ptr<C_actor> > &GetActors() const{ return actors; }
   virtual void BroadcastMessage(dword msg, dword prm1=0, dword prm2=0);

   void EmitListen(S_listen_source &ls, float max_distance1, bool net_send = true, LPC_actor emmitor = NULL,
      S_vector *shoot_ray = NULL);

   virtual void NetIn(E_NETWORK_MESSAGE msg_id, class C_net_in_msg *in_msg);
   virtual void OnPlayerDisconnect(dword pid, const char *name);
   virtual void OnSessionLost();
   virtual void ResetScreen();

   int PlayerSwitch(bool forward, int id = -1);
   void MakeEndDetector();

   void DisplayObjectives();

   LPC_actor MakeParticle(const char *model_name, const char *table_name,
      const S_vector &disp, PI3D_frame link_to = NULL, const S_vector *dir = NULL);

//----------------------------
// Manage game state. If 'net_ascend' is true, net synchro is put into NM_GAME chunk.
   void SetGameState(E_MISSION_STATE s, bool net_use_chunk = false);

   bool PlaySound(const char *snd_name, PI3D_frame lnk, float min, float max,
      float vol = 1.0f, bool loop = false);

   void PlaySound(const char *snd_name, const S_vector &pos, float min, float max);

//----------------------------
// Exit map mode - function called from inside of map mode to cancel it.
   void ExitMapMode();

//----------------------------
// Display aim cross in next frame, with specified color.
   //void DisplayCross(dword color, int scrn_x, int scrn_y);
   void DisplayCross(dword color, bool centered);

//----------------------------
// Setup aim line to be drawn in next frame.
   void DisplayAimLine(const S_vector &from, const S_vector &to);

//----------------------------
// Return true if all players are dead.
   bool IsAnyPlayerAlive() const;
};

C_game_mission *CreateGameMission();

//----------------------------

extern C_I3D_sound_cache sound_cache;

class C_I3D_model_cache_special: public C_I3D_model_cache{
public:
   virtual I3D_RESULT OpenFile(PI3D_model, const char *dir, const char *filename, dword flags,
      I3D_LOAD_CB_PROC *cb_proc, void *cb_context);
};
extern C_I3D_model_cache_special model_cache;

//----------------------------
//----------------------------

enum E_GAME_KEY{
   GKEY_LEFT,
   GKEY_FORWARD,
   GKEY_RIGHT,
   GKEY_BACKWARD,
   GKEY_RUN,
   GKEY_STRAIFE,
   GKEY_FIRE,
   GKEY_UNUSED1,
   GKEY_MOVE_LEFT,
   GKEY_JUMP,
   GKEY_MOVE_RIGHT,
   //GKEY_UP,
   //GKEY_DOWN,
   GKEY_ITEM_USE = 13,

   GKEY_UP=32,
   GKEY_DOWN,
   GKEY_USE,
   //GKEY_ITEM_USE,
   GKEY_UNUSED2,

   GKEY_INV_SCROLL_LEFT,
   GKEY_INV_SCROLL_RIGHT,
   GKEY_MAP,
                              //map mode
   GKEY_MAP_FLOOR_UP,

   GKEY_MAP_FLOOR_DOWN,
   GKEY_PLR_SWITCH,
   GKEY_CAM_ZOOM_IN,
   GKEY_CAM_ZOOM_OUT,

   GKEY_CAM_VIEW_UP,
   GKEY_CAM_VIEW_DOWN,
   GKEY_CENTER,
   GKEY_RELOAD,
   GKEY_CMD_FOLLOW,
   GKEY_CMD_STOP,

   GKEY_CMD_MOVE,
   GKEY_CMD_HEY,
   GKEY_CMD_HOLD_FIRE,
   GKEY_CMD_PRG_CLR,

   GKEY_PLR_SWITCH_1,
   GKEY_PLR_SWITCH_2,
   GKEY_PLR_SWITCH_3,
   GKEY_PLR_SWITCH_4,

      GKEY_LAST_USER = GKEY_PLR_SWITCH_4,

   GKEY_ESCAPE,

   GKEY_SET_ITEM_0,
   GKEY_SET_ITEM_1,
   GKEY_SET_ITEM_2,
   GKEY_SET_ITEM_3,

   GKEY_ZOOM,

   GKEY_DEBUG,

   GKEY_LAST
};

#define GK_EDIT_MODIFY  0x40000000  //CapsLock
#define GK_EDIT_SELECT  0x80000000  //TAB

//----------------------------
#define MUSIC_FADE_TIME 1500

class C_game_music{
   PISND_sound snd;
   float fade_vol;            //volume to which we're fading
   float fade_base_vol;       //volume from which we're starting to fade
   float curr_vol;            //current volume
   int fade_count;            //0 = not fading
public:

   C_game_music():
      fade_count(0),
      snd(NULL),
      curr_vol(1.0f)
   {}

   ~C_game_music(){
      Close();
   }

   bool Play(const char *name, bool loop = false){

      if(!isound) return false;
      if(!snd){
         isound->CreateSound(&snd);
         if(!snd) return false;
      }
      curr_vol = 1.0f;
      PISND_source src;
      isound->CreateSource(&src);
      dword open_flags = ISND_NOPANNING;
      if(allow_stream_snd)
         open_flags |= ISND_STREAMING;
      C_fstring name_base("Music\\%s", name);
      ISND_RESULT ir = src->Open(C_fstring("%s.ogg", (const char*)name_base), open_flags);
      if(ISND_FAIL(ir))
         ir = src->Open(C_fstring("%s.wav", (const char*)name_base), open_flags);
      if(ISND_SUCCESS(ir)){
         snd->Open(src, ISND_SYSMEM);
         SetVolume();
         snd->Play(loop ? ISND_LOOP : 0);
         fade_count = 0;

         src->Release();
         return true;
      }
      ErrReport(C_fstring("Failed to open music %s.wav", (const char*)name_base), editor);
      src->Release();
      return false;
   }

   bool IsPlaying() const{
      return (snd && snd->IsPlaying());
   }

   void FadeVolume(float vol){
      if(IsPlaying()){
         fade_vol = vol;
         fade_base_vol = curr_vol;
         if(!fade_count) fade_count = MUSIC_FADE_TIME;
      }
   }

   void Stop(){
      if(IsPlaying()) FadeVolume(0.0f);
   }

   void SetVolume(){
      if(snd){
         float vol = curr_vol * game_configuration.volume_music / 100.0f;
                              //we must set inversed volume of global sound vol
         float gv = Max(.01f, game_configuration.volume_effects / 100.0f);
         vol /= gv;
         snd->SetVolume(vol);
      }
   }

   void Tick(int time){
      if(fade_count){
         fade_count = Max(0, fade_count-time);
         if(fade_count){
            float vol_delta = (fade_base_vol - fade_vol);
            curr_vol = fade_vol + vol_delta * fade_count / MUSIC_FADE_TIME;
            SetVolume();
         }else{
            curr_vol = fade_vol;
            if(snd){
               if(!curr_vol)
                  snd->Close();
               else
                  SetVolume();
            }
         }
      }
   }

   void Close(){
      if(snd){
         snd->Release();
         snd = NULL;
      }
   }
};

extern C_game_music game_music;

//----------------------------
//----------------------------
                              //all texts
enum E_GAME_TEXT{
   GT_MAP_DEST_OUT=0,
   GT_MAP_BAD_DEST,
   GT_MAP_NO_PATH,
   GT_MAP_DEST_REQ,
   GT_MAP_ACTIVE_REQ,
   GT_MAP_NO_ACTIVE,
   GT_MAP_NOT_FOR_USE,
   GT_MAP_CANCELED,
   GT_MAP_INVENTORY,
   GT_MAP_CANNOT_DO_THIS,

   GT_ERR_RUN_SETUP = 10,
   GT_MENU_UNDEF_KEY = 11,

   GT_MENU_DESCRIPTION = 50,

   GT_PRG_HINT_BASE = 100,

   GT_MENU_DELETE_PROFILE = 120,
   GT_MENU_PROFILE_DUP = 125,
   GT_MENU_VERSION_SELECT = 130,
   GT_MENU_MAIN_BASE = 200,
   GT_MENU_DIFFICULTY = 210,
   GT_MENU_QUIT = 215,
   GT_MENU_OPTIONS = 219,
   GT_GAME_NEW_PROFILE = 225,
   GT_MENU_GAME_PROFILE = 230,
      GT_MENU_EMPTY_SLOT,
      GT_MENU_DELETE_INFO,
   GT_MENU_SOUNDOPS = 235,
   GT_MENU_CONTROLLS = 240,
   GT_MENU_NET_BASE = 280,
   GT_MENU_NET_SETTING = 285,
   GT_MENU_NET_TCPIP = 290,
   GT_MENU_NET_MODEM = 295,
   GT_MENU_NET_EMPTY = 300,
   GT_MENU_NET_ENUM_FAIL = 301,
   GT_MENU_NET_CONCT_FAIL = 303,
   GT_MENU_NET_SESSIONS = 310,
   GT_MENU_NET_PLAYERS = 315,
   GT_MENU_NET_QUIT = 705,

   GT_DEBRIEF_TITLE = 320,
   GT_DEBRIEF_STATS = 325,

   GT_MENU_MOUSE_SETUP = 340,
   GT_NO = 348,
   GT_YES = 349,

   GT_MISSION_NAMES = 350,

   GT_GAME_MENU_BUTTONS = 380,
   GT_GAME_MENU_CAMP_NAME = 400,
   GT_GAME_MENU_CAMP_TEXT = 410,
   GT_GAME_MENU_CAMP_DESC = 420,
   GT_GAME_MENU_MISS_TEXT = 430,

   GT_GAME_MENU_SOLDIER_TITL = 460,
   GT_GAME_MENU_SOLDIER_NAME = 510,
   GT_GAME_MENU_SOLDIER_DESC = 560,
   GT_GAME_MENU_SOLDIER_BAR = 610,
   GT_GAME_MENU_ITEM_NAME = 620,
   GT_GAME_MENU_ITEM_DESC = 660,

   GT_NET_PLAYER_CONTROLL = 700,
   GT_MENU_PROFILE_DIFFICULTY = 710,

   GT_GAME_OBJECTIVE_COMPLETED = 749,

   GT_NET_ERR_GAMELOST = 800,
   GT_MENU_IN_GAME = 810,

   GT_THE_END = 900,
   GT_KEY_NAME = 10000,
};

//----------------------------

struct S_item{
   int itm;
   dword pieces;
   dword ammo;
   S_item(){}
   S_item(int i1, int am, int pcs = 1):
      itm(i1), pieces(pcs), ammo(am)
   {}
   S_item(const S_item &itm){ operator =(itm); }
   S_item &operator =(const S_item &i1){
      itm = i1.itm;
      ammo = i1.ammo;
      pieces = i1.pieces;
      return *this;
   }
   bool operator ==(const S_item &i1) const{
      return (itm==i1.itm && pieces==i1.pieces && ammo==i1.ammo);
   }
};
//----------------------------
                              //player profiles, mission states, ...

struct S_mission_state{
                              //false if dead, true if alive
   byte base_soldier_avl[MAX_BASE_SOLDIERS];
   short team_8[8];
   struct S_man_4{
      short id;
      vector<S_item> item_list;
      bool in_mission;        //used to indicate if man is taken into mission - in some missions less than 4 players are participating
      S_man_4():
         id(-1)
      {}
      S_man_4(const S_man_4 &m4){ operator =(m4); }
      S_man_4 &operator =(const S_man_4 &m4){
         id = m4.id;
         item_list = m4.item_list;
         in_mission = m4.in_mission;
         return *this;
      }
      operator short &(){ return id; }
      void Clear(){
         id = -1;
         item_list.clear();
      }
   };
   S_man_4 team_4[4];

   vector<S_item> item_list;
                              //saved game - residing in memory
   struct S_save_game{
      byte *mem;
      dword len;
      S_save_game():
         mem(NULL), len(0)
      {}
      ~S_save_game(){
         //delete[] mem;
         if(mem){
            C_cache ck;
            ck.FreeMem(mem);
         }
      }
      S_save_game(const S_save_game &sg): mem(NULL), len(0){ operator =(sg); }
      S_save_game &operator =(const S_save_game &sg){
         Close();
         len = sg.len;
         if(len){
            C_cache ck;
            mem = ck.AllocMem(len);
            //mem = new byte[len];
            memcpy(mem, sg.mem, len);
         }
         return (*this);
      }
      void Close(){
         if(mem){
            C_cache ck;
            ck.FreeMem(mem);
            mem = NULL;
         }
         len = 0;
      }
      inline bool IsOpen() const{ return (len); }
   } save_game;
   dword mission_crc;

   S_mission_state(){
      Close();
   }
   S_mission_state(const S_mission_state &ms){
      Close();
      operator =(ms);
   }
   S_mission_state &operator =(const S_mission_state&);

   void Move4To8(int indx);

   bool Save(class C_chunk &ck);
   bool Open(C_chunk &ck);
   void Close();

//----------------------------
// Network transfer
   bool NetSend() const;
   bool NetImport(PC_net_in_msg in_msg);
};

//----------------------------

struct S_game_profile{
                              //mission indicies are based on campaign
   int max_camp, max_mission;
   int curr_camp, curr_mission;

   enum E_DIFFICULTY{
      DIF_EASY,
      DIF_MEDIUM,
      DIF_HARD,
   } difficulty;
   C_string name;
   C_string file_name;

                              //Add-on support
   C_string addon_name;
   int max_addon_mission, num_addon_missions;
   dword addon_crc;
   C_smart_ptr<C_dta_read> addon_dta;

   int MissionIndex(int camp = -1, int miss = -1) const;

//----------------------------
// Get name of current mission.
   const char *GetCurrMissionName() const;

   int NumMissionsInCamp(int camp = -1) const;

                              //game-saves for each mission
   S_mission_state mission_states[NUM_MISSIONS + MAX_NUM_ADDON_MISS];

   S_game_profile():
      max_addon_mission(0),
      num_addon_missions(0),
      addon_crc(0)
   {
      Reset();
   }
   S_game_profile(int mc, int mm, int cc, int cm):
      max_camp(mc), max_mission(mm),
      curr_camp(cc), curr_mission(cm)
   {
   }

//----------------------------
// Check if we're in one of Add-on missions.
   inline bool IsAddOn() const{ return (curr_camp==9); }

//----------------------------
// Check if we're in one of datadisk (FFF) missions.
   inline bool IsDatadisk() const{ return (curr_camp>=6 && !IsAddOn()); }

   bool Save();
   bool Open(const char*);
   void Reset(){
      max_camp = 0, max_mission = 0, curr_camp = 0, curr_mission = 0;
      for(int i=0; i<sizeof(mission_states)/sizeof(S_mission_state); i++)
         mission_states[i].Close();
      name = NULL;
      file_name = NULL;
      difficulty = DIF_MEDIUM;
   }

//----------------------------
// Open specified Add-on. 
   bool OpenAddOn(const char *fname);

//----------------------------
// Close current Add-on.
   void CloseAddOn();

//----------------------------
   void ResetAddOn(){
      CloseAddOn();
      addon_name = NULL;
      max_addon_mission = 0;
      num_addon_missions = 0;
      addon_crc = NULL;
   }

//----------------------------
// Get CRC of currently opened Add-on.
   bool GetAddOnCRC(dword &crc);

//----------------------------
// Send game profile over network.
   bool NetSend() const;

//----------------------------
// Receive game profile from network.
   bool NetImport(class C_net_in_msg *in_msg);
};

extern S_game_profile game_profile;

//----------------------------
                              //structs associated with some menus, passed as context

struct S_menu_debriefing_info{//GAME_MENU_DEBRIEF
   enum E_status{
      DEBRIEF_SUCCESS_MISSION,
      DEBRIEF_FAIL,
      DEBRIEF_SUCCESS_CAMPAIGN,
      DEBRIEF_SUCCESS_GAME_DONE,
      DEBRIEF_SUCCESS_ADDON_DONE,
   } status;
   int ammo_spent, num_hits, enemy_killed, civil_killed; //some stats
   vector<int> allies_killed; //IDs of killed allies
};

struct S_menu_soldier_info{
   int soldier_id;            //ID of soldier to show info about
   C_class_to_be_ticked_hd *save_tc;   //tick class to return when back button pressed
};

C_class_to_be_ticked_hd *CreateMenu(E_GAME_MENU id, dword context = 0);

//----------------------------

C_class_to_be_ticked_hd *CreateBriefing(const char *dir, C_class_to_be_ticked_hd *save_tc = NULL);

//----------------------------

C_class_to_be_ticked_hd *CreateCampAnim(const char *anim_name, C_class_to_be_ticked_hd *save_tc = NULL, bool load_progress = true);

//----------------------------

C_class_to_be_ticked_hd *CreateTheEnd();

//----------------------------

class C_scrolltext: public C_class_to_be_ticked_hd{
public:
//----------------------------
// Check if scrolltext is finished.
   virtual bool IsEnd() const = 0;
};

C_scrolltext *CreateScrolltext(class C_poly_text&, struct S_scrolltext_style *cs, const C_string *ct,
   bool wide, float clip_y = .0375f);

//----------------------------

class C_hint_base: public C_unknown{
public:
                              //returns false when hint is finished and should be destroyed
   virtual bool Tick(int time) = 0;
   virtual void FadeOff() = 0;
   //virtual const S_vector2 &GetPos() const = 0;
   virtual void SetPos(const S_vector2&) = 0;
   virtual void SetOn(bool) = 0;
};

C_hint_base *CreateHint(const char *t1, bool wide, class C_poly_text *texts, float x, float y, float size, dword color = -1,
   bool flt_up = false, int count = 0, dword text_flags = TEXT_JUSTIFY_CLIP_TO_SCREEN | TEXT_JUSTIFY_CENTER | TEXT_JUSTIFY_CONVERT_EOL);

//----------------------------

struct S_scrolltext_style{
   float size;
   dword color;
   float italic;
   byte align;
   int pause;
};

struct S_scrolltext_text{
   char *text;
   int style;
   int pause;                 //0 = default
};

//----------------------------
                              //game profiles
struct S_logon_account{
   C_string file_name;  
   C_string user_name;
   S_date_time dt;
   S_game_profile::E_DIFFICULTY difficulty;

   S_logon_account(){}
   S_logon_account(const S_logon_account &p){ operator =(p); }
   S_logon_account &operator =(const S_logon_account &p){
      file_name = p.file_name;
      user_name = p.user_name;
      dt = p.dt;
      difficulty = p.difficulty;
      return *this;
   }
   bool operator <(S_logon_account &p) const{
      if(dt.year < p.dt.year) return true;
      if(dt.year > p.dt.year) return false;
      if(dt.month < p.dt.month) return true;
      if(dt.month > p.dt.month) return false;
      if(dt.day < p.dt.day) return true;
      if(dt.day > p.dt.day) return false;
      if(dt.hour < p.dt.hour) return true;
      if(dt.hour > p.dt.hour) return false;
      if(dt.minute < p.dt.minute) return true;
      if(dt.minute > p.dt.minute) return false;
      if(dt.second < p.dt.second) return true;
      return false;
   }
};

//----------------------------
// Function searching disk for user accounts, returning found ones in vector.
// The returned accounts are sorted by last access time.
void SearchAccounts(vector<S_logon_account>&);

//----------------------------

#ifdef _WINDOWS_
#error "windows.h cannot be included!"
#endif

//----------------------------
