#include "all.h"
#include "tables.h"
#include "loader.h"
#include "winstuff.h"
#include "net.h"

//----------------------------
// Get angle among 2 axes.
// Value is <0, 2*PI>.
float GetAngle(float x, float y){

   if(!x) return y>0 ? PI/2.0f : 3*PI/2.0f;
   float val;
   if(x > 0.0f){
      val=atan(y/x);
      if(val < 0.0f) val += GetPI2();
      return val;
   }else{
      val=PI+atan(-y/-x);
      if(val < 0.0f) val += GetPI2();
      return val;
   }
}

//----------------------------
// Get angle of 2D vector. [0, 1] is 0, [-1, 0] is PI/2.
float Angle2D(const S_vector2 &v1){

   float a;
   float fx = fabs(v1.x);
   float fy = fabs(v1.y);

   if(fy>fx){
      a = fy<MRG_ZERO ? 0.0f : atan(-v1.x/v1.y);
      if(v1.y<0.0f) a += GetPI();
   }else{
      a = fx>=MRG_ZERO ?
         atan(-v1.y/-v1.x) + PI*.5f :
         v1.x<0.0f ? PI*.5f : -PI*.5f;
      if(v1.x>0.0f) a += GetPI();
   }
   return a;
}

//----------------------------
// Get angle among 2 2D vectors.
float Angle2D(const S_vector2 &v1, const S_vector2 &v2){

   float f1, f2;
   f1 = Angle2D(v1);
   f2 = Angle2D(v2);
   float f = f2 - f1;
   if(f>GetPI()) f -= GetPI2();
   else
   if(f <= - GetPI()) f += GetPI2();
   return f;
}

//----------------------------

float SubAngle(float src, float sub){

   src -= sub;
   if(src < 0.0f)
      src += GetPI2();
   return src;
}

//----------------------------

dword I3DAPI cbDriver(SCN_CALLBACK_MESSAGE msg, dword prm1, dword prm2, void *context){

   switch(msg){

#if defined EDITOR
   case CBM_MESSAGE:
      {
         if(editor){
            LPC_editor_item ei = editor->FindPlugin("Log");
            if(ei) ei->Action(E_LOG_ADD_TEXT, (void*)(const char*)prm1);
         }
      }
      break;

   case CBM_DEBUG_MESSAGE:
      {
         const char *msg = (const char*)prm1;
         switch(prm2){
         case 0:
            DEBUG(msg);
            break;
         case 1:
            OUTTEXT(msg);
            break;
         case 2:
            if(editor)
               editor->FindPlugin("Mission")->Action(E_MISSION_RT_ERR_MESSAGE, (void*)msg);
            break;
         case 3:
            ErrReport(msg, editor);
            break;
         }
      }
      break;

   case CBM_DEBUG_CODE:
      switch(prm1){
      case 0:                 //(S_vector[2]) - debug line
      case 1:
      case 2:
         {
            LPC_editor_item ei = editor->FindPlugin("DebugLine");
            if(ei){
               S_vector *v = (S_vector*)prm2;
               ei->Action(E_DEBUGLINE_ADD + prm1, &S_add_line(v[0], v[1]));
            }
         }
         break;
      case 10:
      case 11:
      case 12:
         {
            LPC_editor_item ei = editor->FindPlugin("DebugLine");
            if(ei){
               S_vector *v = (S_vector*)prm2;
               ei->Action(E_DEBUGLINE_ADD + prm1-10, &S_add_line(v[0], v[1], 0xffffff | (dword(v[2].x*255.0f)<<24)));
            }
         }
         break;
      case 20:
      case 21:
      case 22:
         {
            LPC_editor_item ei = editor->FindPlugin("DebugLine");
            if(ei){
               S_vector *v = (S_vector*)prm2;
               ei->Action(E_DEBUGLINE_ADD + prm1-20, &S_add_line(v[0], v[1], 0xff000000 |
                  (((int)(v[2][0]*255.0f))<<16) |
                  (((int)(v[2][1]*255.0f))<<8) |
                  (((int)(v[2][2]*255.0f))<<0)
                  ));
            }
         }
         break;
      }
      break;
#endif //EDITOR

   case CBM_ANIM_NOTE:
      {
         /*
         const I3D_note_callback *nc = (const I3D_note_callback*)prm1;
         script_man.RunFunction(nc->frm, "OnNote", NULL, (dword)(const char*)nc->note);
         */
      }
      break;
   }

   return 0;
}

//----------------------------

dword I3DAPI cbFrame(PI3D_frame frm, FRM_CALLBACK_MESSAGE msg, dword prm1, dword prm2){

   switch(msg){

   case CBM_ANIMNOTE:
      {
                              //convert string to number
         const I3D_note_callback *nc = (I3D_note_callback*)prm1;
         prm2 = nc->note_time;
         prm1 = atoi(nc->note);
         if(prm1<30){
                                 //reserved for sounds
                                 //find model parent
            PI3D_frame prnt = frm;
            do{
               prnt = prnt->GetParent();
            }while(prnt && prnt->GetType()!=FRAME_MODEL);
            if(prnt){
               LPC_actor act = (LPC_actor)prnt->GetData();
               if(act){
                                 //callback actor
                  //fd->actor->cbProc(CB_SOUND, (dword)frm, prm1);
                              //accept notify only from main anim, not others which fading out
                  //bool is_human = (act->GetType() == ACTOR_PLAYER) || (act->GetType() == ACTOR_ENEMY);
                  //if((nc->stage == MAIN_ANIM_STAGE) || !is_human)
                  act->SoundNotify(frm, prm1, nc->stage);
               }
            }
            break;
         }else
         if(prm1<50){
                                 //reserved for actor actions
                                 //find model parent
            PI3D_frame prnt=frm;
            do{
               prnt=prnt->GetParent();
            }while(prnt && prnt->GetType()!=FRAME_MODEL);
            if(prnt){
               LPC_actor act = (LPC_actor)prnt->GetData();
               if(act){
                                 //callback actor
                  //fd->actor->cbProc(CB_ANIM_NOTIFY, (dword)frm, prm1);
                              //accept notify only from main anim, not others which fading out

                  //bool is_human = (act->GetType() == ACTOR_PLAYER) || (act->GetType() == ACTOR_ENEMY);
                  //if((nc->stage == MAIN_ANIM_STAGE) || !is_human)
                  act->AnimNotify(frm, prm1, nc->stage);
               }
            }
            break;
         }else
         if(prm1<99){
            tick_class->Notify(frm, prm1, prm2);
         }else
         if(prm1==99){
            /*
                                 //find model
            PI3D_frame prnt=frm;
            while((prnt=prnt->GetParent()) && prnt->GetType()!=FRAME_MODEL);
            if(prnt){
               prnt->LinkTo(NULL);
               scene->DeleteFrame(prnt);
            }
            */
            assert(0);
         }
      }
      break;
   }
   return 0;
}

//----------------------------

I3DENUMRET I3DAPI cbSetAllCallback(PI3D_frame frm, dword mask){

   frm->SetCallback(cbFrame, mask);
   return I3DENUMRET_OK;
}

//----------------------------
#ifdef EDITOR

void ErrReport(const char *cp, void *context){

                              //context is pointer to C_editor class
   LPC_editor ed = (LPC_editor)context;
   if(ed){
      LPC_editor_item ei_log = ed->FindPlugin("Log");
      if(ei_log) ei_log->Action(E_LOG_ADD_TEXT, (void*)cp);
   }
}

//----------------------------

void WarnReport(const char *cp, dword color, void *context){

   LPC_editor ed = (LPC_editor)context;
   if(ed){
      LPC_editor_item ei_log = ed->FindPlugin("Log");
      if(ei_log) ei_log->Action(E_LOG_ADD_RICH_TEXT, &pair<const char*, dword>(cp, color));
   }
}
#endif

//----------------------------

bool FreeCameraMove(S_vector &dir, const S_vector &limit_dir, int rx, int ry, float limit_angle, float look_speed){

#if 0
   DebugLine(S_vector(0, 0, 0), dir, 2, 0xffff0000, 10);
   DebugLine(S_vector(0, 0, 0), limit_dir, 2, 0xff0000ff, 10);
#endif

   if(!(rx || ry)) return false;
   if(dir.IsNull()) return false;

   float zx_len = sqrt(dir.x*dir.x + dir.z*dir.z);
   float zx_angle = GetAngle(dir.z, -dir.x);
   if(rx){
      zx_angle = SubAngle(zx_angle, rx*look_speed);
      if(zx_angle>GetPI()) zx_angle -= GetPI2();
   }
   if(ry){
      float angle = GetAngle(zx_len, dir.y);
      angle = SubAngle(angle, ry*look_speed);
      if(angle>GetPI()) angle -= GetPI2();
      dir.y = sin(angle);
      zx_len = cos(angle);
   }
   dir.x = -sin(zx_angle)*zx_len;
   dir.z = cos(zx_angle)*zx_len;
                              //limit
   while(true){
      float a = dir.AngleTo(limit_dir);
      if(a<=limit_angle) break;
      dir.Normalize();
      float f = limit_angle / a -.01f;
      dir = limit_dir + (dir-limit_dir)*f;
   }
   return true;
}

//----------------------------

enum{
   TAB_CV_LIGHT_COLOR,
   TAB_F_LIGHT_POWER,
   TAB_I_LIGHT_RANGE_N,
   TAB_F_LIGHT_RANGE_F,
   TAB_I_LIGHT_CONE_IN,
   TAB_I_LIGHT_CONE_OUT,
   TAB_B_LIGHT_SPECULAR,
   TAB_F_LIGHT_FALLOFF,
};

void TableToLight(LPC_table tp, PI3D_light lp){

   lp->SetColor(tp->ItemV(TAB_CV_LIGHT_COLOR));
   lp->SetPower(tp->ItemF(TAB_F_LIGHT_POWER));
   float f = tp->ItemF(TAB_F_LIGHT_RANGE_F);
   float n = (float)tp->ItemI(TAB_I_LIGHT_RANGE_N) * .01f;
   if(lp->GetLightType()!=I3DLIGHT_FOG)
      n *= f;
   lp->SetRange(n, f);
   int ci = tp->ItemI(TAB_I_LIGHT_CONE_IN),
       co = tp->ItemI(TAB_I_LIGHT_CONE_OUT);
   lp->SetCone((float)ci*PI/180.0f, (float)co*PI/180.0f);
   lp->Update();
}

//----------------------------

void TableToSound(LPC_table tp, PI3D_sound sp, PI3D_scene scene){

   const char *cp = tp->ItemS(TAB_S_SOUND_FILENAME);
   bool strm = tp->ItemB(TAB_B_SOUND_STREAM);
   bool amb = tp->ItemB(TAB_B_SOUND_NO3D);

   C_string str = cp;
   {
      for(int i=str.Size(); i--; ){
         if(str[i]=='.'){
            str[i] = 0;
            break;
         }
      }
                              //convert slashes
      for(i=str.Size(); i--; ){
         if(str[i]=='/')
            str[i] = '\\';
      }
   }

   //if(!allow_stream_snd)
   strm = false;

   dword open_flags = 0;
   I3D_RESULT ir;

   if(strm) sp->SetStreaming(true);
   //if(amb) open_flags |= I3DLOAD_SOUND_SOFTWARE;
   if(amb) sp->SetSoundType(I3DSOUND_AMBIENT);
   if(editor)
      ir = sound_cache.Open(sp, str, scene, open_flags, NULL, NULL);
   else
      ir = sound_cache.Open(sp, str, scene, open_flags, ErrReport, editor);

   sp->SetLoop(tp->ItemB(TAB_B_SOUND_LOOP));
   sp->SetOn(tp->ItemB(TAB_B_SOUND_ON));
   sp->SetVolume((float)(tp->ItemI(TAB_I_SOUND_VOLUME)) / 100.0f);
   sp->SetOutVol((float)(tp->ItemI(TAB_I_SOUND_OUTVOL)) / 100.0f);
   sp->SetRange(tp->ItemF(TAB_F_SOUND_RANGE_N), tp->ItemF(TAB_F_SOUND_RANGE_F));

   int ci = tp->ItemI(TAB_I_SOUND_CONE_IN),
       co = tp->ItemI(TAB_I_SOUND_CONE_OUT);
   sp->SetCone((float)ci*PI/180.0f, (float)co*PI/180.0f);
   sp->Update();
}

//----------------------------

void TableToParticle(LPC_table tab, PI3D_particle prt){

   S_vector base_dir, dest_dir;
   for(int i=0; i<Min(28, tab->NumItems()); i++){
      if(i != PRT_EMIT_BASE)
      switch(i){
      case PRT_ROT_SPEED_MIN:
      case PRT_ROT_SPEED_MAX:
         {
            float f = tab->ItemF(i) * PI / 180.0f;
            prt->SetProperty(i, FloatAsInt(f));
         }
         break;
      case PRT_BASE_DIR_X: case PRT_BASE_DIR_Y: case PRT_BASE_DIR_Z:
         base_dir[i-PRT_BASE_DIR_X] = tab->ItemF(i);
         break;
      case PRT_DEST_DIR_X: case PRT_DEST_DIR_Y: case PRT_DEST_DIR_Z:
         dest_dir[i-PRT_DEST_DIR_X] = tab->ItemF(i);
         break;
      default:
         switch(tab->SizeOf(i)){
         case 1: prt->SetProperty(i, tab->ItemB(i)); break;
         case 4: prt->SetProperty(i, tab->ItemI(i)); break;
         }
      }
   }
                              //apply directions
   for(i=0; i<3; i++){
      prt->SetProperty(PRT_BASE_DIR_X+i, FloatAsInt(base_dir[i]));
      prt->SetProperty(PRT_DEST_DIR_X+i, FloatAsInt(dest_dir[i]));
   }

   /*
   if(tab->ItemB(PRT_EMIT_BASE))
      prt->SetProperty(PRT_EMIT_BASE, 0);
      */
}

//----------------------------
//----------------------------

#define MIN_SPLASH_TIME_DELTA 50

bool I3DAPI cbLoadScene(I3D_LOADMESSAGE msg, dword context, dword prm2, void*){

   static float last_bit;
   static int last_bit_i;
   static int last_render_time;
   static C_smart_ptr<IImage> back_img;     //saved copy of backbuffer image
   static C_poly_text texts;

   switch(msg){

   case CBM_PROGRESS:
   case 10000:                //during mission load
      float p;
      p = *(float*)&context;
      if(msg==CBM_PROGRESS){
                              //1st half of progress bar - reserved to scene loader
         if(prm2==2) prm2=1;  //don't let it close the progress bar
         p *= .5f;
      }else{
                              //2nd half - reserved to mission loader
         p = .5f + p*.5f;
      }

      bool drawn;
      drawn = false;
      if(prm2 == 0){
         assert(!back_img);
         bool was_zb = driver->GetState(RS_USEZB);
         if(was_zb) driver->SetState(RS_USEZB, false);

         int sx = igraph->Scrn_sx(), sy = igraph->Scrn_sy();
         {
            driver->BeginScene();

            struct S_vertex{ 
               S_vectorw v;
               dword diffuse;
               //dword specular;   //padding to 32 bytes
               I3D_text_coor tx;
            } v[4];
                              //draw semi-dransparent triangle over backbuffer
            static const I3D_triface faces[] = {I3D_triface(0, 1, 2), I3D_triface(2, 1, 3)};
            PI3D_material mat = driver->CreateMaterial();

            I3D_face_group fg;
            fg.SetMaterial(mat);
            fg.base_index = 0;
            fg.num_faces = 2;
            mat->Release();

            {
               PI3D_texture txt;
               I3D_CREATETEXTURE ct;
               memset(&ct, 0, sizeof(ct));
               ct.opt_name = "loading.png";
               ct.flags = TEXTMAP_OPACITY | TEXTMAP_NOMIPMAP | TEXTMAP_TRUECOLOR;
               driver->CreateTexture(&ct, &txt);
               fg.GetMaterial()->SetTexture(txt);

               float tx = (float)sx / 32.0f + 1.0f / 128.0f;
               float ty = (float)sy / 32.0f + 1.0f / 128.0f;

               for(int i=4; i--; ){
                  v[i].v.Zero();
                  v[i].v.w = 1.0f;
                  v[i].tx.u = 0.0f;
                  v[i].tx.v = 0.0f;
                  v[i].diffuse = 0x80000000;

                  if(i&1){
                     v[i].v.x = sx;
                     v[i].tx.u = tx;
                  }
                  if(i&2){
                     v[i].v.y = sy;
                     v[i].tx.v = ty;
                  }
               }
               bool lf = driver->GetState(RS_LINEARFILTER);
               if(lf) driver->SetState(RS_LINEARFILTER, false);

               PI3D_visual_dynamic dvis = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);
               dvis->Build(v, 4, I3DVC_XYZRHW | I3DVC_DIFFUSE | /*I3DVC_SPECULAR | */(1<<I3DVC_TEXCOUNT_SHIFT),
                  faces, 2, &fg, 1, 0);
               dvis->Render();
               dvis->Release();
              
               if(lf) driver->SetState(RS_LINEARFILTER, true);

               if(txt){
                  txt->Release();
               }
            }
                              //draw logo texture to corner
            {
               PI3D_texture txt;
               I3D_CREATETEXTURE ct;
               memset(&ct, 0, sizeof(ct));
               ct.text_name = "HDlogosmall.png";
               ct.opt_name = "HDlogosmall_o.png";
               ct.flags = TEXTMAP_DIFFUSE | TEXTMAP_OPACITY | TEXTMAP_NOMIPMAP | TEXTMAP_TRUECOLOR;
               driver->CreateTexture(&ct, &txt);
               if(txt){
                  for(int i=4; i--; ){
                     v[i].v.Zero();
                     v[i].v.w = 1.0f;
                     v[i].tx.u = 0;
                     v[i].tx.v = 0;
                     v[i].diffuse = 0xffffffff;

                     v[i].v.x = (float)sx * .0f;
                     if(i&1){
                        v[i].v.x += (float)sx * .2f;
                        v[i].tx.u = 1;
                     }
                     v[i].v.y = (float)sy * .82f;
                     if(i&2){
                        v[i].v.y += (float)sy*.133f;
                        v[i].tx.v = 1;
                     }
                  }
                  fg.GetMaterial()->SetTexture(txt);
                  PI3D_visual_dynamic dvis = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);
                  dvis->Build(v, 4, I3DVC_XYZRHW | I3DVC_DIFFUSE | /*I3DVC_SPECULAR | */(1<<I3DVC_TEXCOUNT_SHIFT),
                     faces, 2, &fg, 1, 0);
                  dvis->Render();
                  dvis->Release();

                  txt->Release();
               }
            }
            driver->EndScene();
         }
         back_img = igraph->CreateBackbufferCopy();
         if(back_img)
            back_img->Release();
         last_bit_i = 0;
         last_render_time = igraph->ReadTimer() - MIN_SPLASH_TIME_DELTA;
         texts.Init();

         if(was_zb) driver->SetState(RS_USEZB, true);
         drawn = true;
      }else{
         int percent = p * 100.0f;
         if(last_bit_i!=percent){
            assert(percent>last_bit_i);
            int timer = igraph->ReadTimer();
            int delta = timer - last_render_time;
            if(delta > MIN_SPLASH_TIME_DELTA){
               last_render_time = timer;
               drawn = true;
               if(back_img){

                  back_img->Draw();
                  driver->BeginScene();

                              //draw text into this
                  const C_string &txt_loading = (!all_txt.IsWide(20201) && all_txt.IsDefined(20201)) ? all_txt[20201] : "Loading...";
                  C_fstring str("%s %i%%", (const char*)txt_loading, percent);
                  const float x = .0338f,
                     y = .71f,
                     shd = .003f,
                     sz = .0225f;
                  float italic = .2f;
                  LPC_text txt_shd = texts.CreateText(str, 0x80000000, x+shd, y+shd, sz, 0, false,
                     NULL, NULL, NULL, NULL, 0, italic);
                  LPC_text txt = texts.CreateText(str, -1, x, y, sz, 0, false,
                     NULL, NULL, NULL, NULL, 0, italic);
                  texts.Render();
                  txt->Release();
                  txt_shd->Release();

                  igraph->UpdateScreen();
                  driver->EndScene();
               }
            }
            last_bit_i = percent;
         }
         if(prm2 == 2){
            back_img = NULL;
            texts.Close();
            last_bit_i = 100;
         }
      }
#ifdef EDITOR
      if(editor && drawn)
         editor->Message(C_fstring("Loading... %i%%", last_bit_i), 0, EM_MESSAGE, true);
#endif
      break;

   case CBM_ERROR:
#ifdef EDITOR
      if(editor){
         dword color = 0;
         switch(prm2){
         case 0:              //error
            color = 0xff0000;
            break;
         case 1:              //warning
            color = 0x008000;
            break;
         }
         editor->FindPlugin("Log")->Action(E_LOG_ADD_RICH_TEXT, &pair<const char*, dword>((const char*)context, color));
      }
#else
      ErrReport((const char*)context, NULL);
#endif
      break;
   }
   return false;
}

//----------------------------
//----------------------------
// Voice manager class - initializes itself as necessary and keeps
// info about voices till game shutdown.
class C_voice_mgr{
   struct S_voice_info{
      union{
         char voice[4];
         dword voice_id;
      };
      int human_id;
      int num_vars;
   };
   vector<S_voice_info> voice_info;
public:

   int NumVars(const char *voice, int human_id){
                              //scan to see if we've got this entry
      dword voice_id = *(dword*)voice;
      int i;
      for(i=voice_info.size(); i--; )
      if(voice_info[i].human_id==human_id && voice_info[i].voice_id==voice_id)
         return voice_info[i].num_vars;
                              //no entry, scan for this voice
      for(i=0; i<100; i++){
         int h;
         for(int ii=0; ii<sound_cache.NumDirs(); ii++){
            const char *snd_dir = sound_cache.GetDir(ii);
            C_fstring base("%svoices\\%.2i_%.3s_%.1i", snd_dir, human_id, voice, i);
            h = dtaOpen(base+".ogg");
            if(h!=-1){
               dtaClose(h);
               break;
            }
            h = dtaOpen(base+".wav");
            if(h!=-1){
               dtaClose(h);
               break;
            }
         }
         if(h==-1) break;
      }
      voice_info.push_back(S_voice_info());
      voice_info.back().voice_id = voice_id;
      voice_info.back().human_id = human_id;
      voice_info.back().num_vars = i;
      return i;
   }
} voice_mgr;

//----------------------------
                              //determine how many variations for this voice are on disk
                              //get file name with random variation
bool GetVoiceName(const char *voice, int human_id, C_string &str){

   int numv = voice_mgr.NumVars(voice, human_id);
   str = C_fstring("voices\\%.2i_%.3s_%.1i", human_id, voice,
      (numv ? (int)S_int_random(numv) : 0));
   if(!numv){
      ErrReport(C_fstring("Missing voice %s", (const char*)str), editor);
      return false;
   }
   return true;
}

//----------------------------

bool ScaleObject(PI3D_frame frm, PI3D_mesh_level ml_src, const S_vector &scl){

   if(frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_OBJECT){

      PI3D_object obj = I3DCAST_OBJECT(frm);
      PI3D_mesh_level ml1 = driver->CreateMeshLevel();

      assert(ml_src->GetSizeOfVertex()==sizeof(I3D_vertex_mesh));
      const I3D_vertex_mesh *vp = (const I3D_vertex_mesh*)ml_src->LockVertices();
      const I3D_triface *fc = ml_src->LockFaces();
      dword numv = ml_src->NumVertices();
      dword numfg = ml_src->NumFGroups();
      I3D_vertex_mesh *vmp = new I3D_vertex_mesh[numv];
      I3D_face_group *fgrps = new I3D_face_group[numfg];
      for(int i=numfg; i--; )
         fgrps[i] = *ml_src->GetFGroup(i);
      for(i=numv; i--; ){
         vmp[i] = vp[i];
         vmp[i].xyz = vp[i].xyz * scl;
      }

      ml1->Build2(vmp, numv, fc, ml_src->NumFaces(), fgrps, numfg);

      delete[] vmp;
      delete[] fgrps;
      ml_src->UnlockFaces();
      ml_src->UnlockVertices();

      obj->GetMeshObject()->SetLOD(ml1, 0);
      obj->SetMeshObject(obj->GetMeshObject());

      ml1->Release();
      return true;
   }
   return false;
}

//----------------------------
//----------------------------

static I3DENUMRET I3DAPI cbMakeVols(PI3D_frame frm, dword c){

   PI3D_model model = (PI3D_model)c;

   PI3D_dummy dum = I3DCAST_DUMMY(frm);
   const I3D_bbox *bbox = dum->GetBBox();
   float scl = frm->GetScale();

   PI3D_volume vol;

   const char *str = frm->GetOrigName();
   switch(str[3]){
   case '0':                  //sphere
      {
         vol = driver->CreateVolume();
         vol->SetVolumeType(I3DVOLUME_SPHERE);

         float f = (bbox->max.x-bbox->min.x) * scl / 2;
         vol->SetScale(f);
      }
      break;
   case '1':                  //box
      {
         vol = driver->CreateVolume();
         vol->SetVolumeType(I3DVOLUME_BOX);
         vol->SetNUScale((bbox->max-bbox->min) * scl * .5f);
      }
      break;
   default:
      return I3DENUMRET_OK;
   }

   vol->SetName(frm->GetOrigName());
   vol->LinkTo(frm->GetParent());
   vol->SetPos(frm->GetPos());
   vol->SetRot(frm->GetRot());
   //vol->SetScale(frm->GetScale() * vol->GetScale());
   vol->Update();

   vol->SetOwner(model);

   model->DeleteFrame(frm);
   model->AddFrame(vol);
   vol->Release();

   return I3DENUMRET_OK;
}

//----------------------------

I3D_RESULT C_I3D_model_cache_special::OpenFile(PI3D_model mod, const char *dir, const char *filename, dword flags,
   I3D_LOAD_CB_PROC *cb_proc, void *cb_context){

                              //load file with extension
   I3D_RESULT ir = C_I3D_model_cache::OpenFile(mod, dir,
      C_fstring("%s.i3d", filename),
      flags | I3DLOAD_ANIM_COMPATIBILITY, cb_proc, cb_context);
   if(I3D_SUCCESS(ir)){
      mod->StoreFileName(C_fstring("%s%s", dir, filename));
                              //init volume boxes
      mod->EnumFrames(cbMakeVols, (dword)mod, ENUMF_DUMMY | ENUMF_USEMASK, "~0*");
   }
   return ir;
}

//----------------------------
//----------------------------

S_mission_state &S_mission_state::operator =(const S_mission_state &ms){

   int i;
   memcpy(base_soldier_avl, ms.base_soldier_avl, sizeof(base_soldier_avl));
   for(i=0; i<8; i++) team_8[i] = ms.team_8[i];
   for(i=0; i<4; i++) team_4[i] = ms.team_4[i];
   item_list = ms.item_list;
   save_game = ms.save_game;
   return *this;
}

//----------------------------

void S_mission_state::Move4To8(int indx){

   if(team_4[indx]==-1) return;
                              //find free place in team 8
   for(int i=0; i<8; i++) if(team_8[i]==-1) break;
   assert(i<8);
   team_8[i] = (short)team_4[indx];
                              //remove all ammo from the man and put it onto common heap
   for(i=0; i<team_4[indx].item_list.size(); i++){
      int j;
      for(j=item_list.size(); j--; ) if(item_list[j].itm==team_4[indx].item_list[i].itm) break;
      if(j!=-1){
         item_list[j].pieces += team_4[indx].item_list[i].pieces;
         item_list[j].ammo += team_4[indx].item_list[i].ammo;
      }else
         item_list.push_back(team_4[indx].item_list[i]);
   }
   team_4[indx].Clear();
}

//----------------------------

bool S_mission_state::Save(C_chunk &ck){

   int i;
                              //put everything into one chunk
   ck.WAscend(CT_PROF_MISSION_STATE);
   {
                              //write available men
      ck.WAscend(CT_PROF_BASE_SOLDIERS);
      ck.Write(base_soldier_avl, tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS));
      ck.Descend();
                              //write team 8
      ck.WAscend(CT_PROF_TEAM_8);
      ck.Write(team_8, 8*sizeof(short));
      ck.Descend();
                              //write team 4
      for(i=0; i<4; i++){
         ck.WAscend(CT_PROF_TEAM_4);
                              //write id
         ck.Write(&team_4[i].id, sizeof(short));
                              //write player's items
         for(int j=0; j<team_4[i].item_list.size(); j++){
            ck.WAscend(CT_PROF_ITEM);
            ck.Write(&team_4[i].item_list[j].itm, sizeof(word));
            ck.Write(&team_4[i].item_list[j].ammo, sizeof(word));
            ck.Write(&team_4[i].item_list[j].pieces, sizeof(word));
            ck.Descend();
         }
         ck.Descend();
      }
                              //write mission items
      for(i=0; i<item_list.size(); i++){
         ck.WAscend(CT_PROF_ITEM);
         ck.Write(&item_list[i].itm, sizeof(word));
         ck.Write(&item_list[i].ammo, sizeof(word));
         ck.Write(&item_list[i].pieces, sizeof(word));
         ck.Descend();
      }
                              //write save game
      if(save_game.len){
         ck.WAscend(CT_PROF_SAVE_GAME);
                              //write version
         word sg_version = CT_SAVE_GAME_VERSION;
         ck.Write(&sg_version, sizeof(word));

         ck.Write(save_game.mem, save_game.len);
         ck.Descend();
      }
                              //save mission crc
      {
         ck.WAscend(CT_PROF_MISSION_CRC);
         ck.Write(&mission_crc, sizeof(dword));
         ck.Descend();
      }

   }
   ck.Descend();
   return true;
}

//----------------------------

bool S_mission_state::Open(C_chunk &ck){

   Close();
   int curr_man_4(0);
   CK_TYPE t;
   while(ck.Size()){
      if(!ck.RAscend(t)) goto fail;
      switch(t){

      case CT_PROF_BASE_SOLDIERS:
         ck.Read(base_soldier_avl, tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS));
         break;

      case CT_PROF_TEAM_8:
         ck.Read(team_8, 8*sizeof(short));
         break;

      case CT_PROF_TEAM_4:
         ck.Read(&team_4[curr_man_4].id, sizeof(short));
         CK_TYPE t;
         while(ck.Size()){
            if(!ck.RAscend(t)) goto fail;
            switch(t){
            case CT_PROF_ITEM:
               {
                  word w[3];
                  ck.Read(w, 3*sizeof(word));
                  if(w[2]){
                     team_4[curr_man_4].item_list.push_back(S_item(w[0], w[1], w[2]));
                  }
               }
               break;
            }
            ck.Descend();
         }
         ++curr_man_4;
         break;

      case CT_PROF_ITEM:
         {
            word w[3];
            ck.Read(w, 3*sizeof(word));
            if(w[2]){
               item_list.push_back(S_item(w[0], w[1], w[2]));
            }
         }
         break;
      case CT_PROF_SAVE_GAME:
         {
                              //read version
            word sg_version;
            ck.Read(&sg_version, sizeof(word));
            if(sg_version == CT_SAVE_GAME_VERSION){
               save_game.len = ck.Size();
               //save_game.mem = new byte[save_game.len];
               save_game.mem = ck.GetHandle()->AllocMem(save_game.len);
               ck.Read(save_game.mem, save_game.len);
            }
         }
         break;

      case CT_PROF_MISSION_CRC:
         ck.Read(&mission_crc, sizeof(dword));
         break;
      }
      ck.Descend();
   }
fail:
   return true;
}

//----------------------------

enum E_GPROF_NET_MSG_ID{
                              //single mission state:

   NM_GP_SOLDIER_INFO = 0,
      //bits:1 avl_list[40]
      //bits:6 team_8[8]
   NM_GP_TEAM_4 = 1,
      //bits:6 id
   NM_GP_PLR_ITEM = 2,
      //byte itm_id
      //word ammo
      //word pcs
   NM_GP_MISS_ITEM = 3,
      //byte itm_id
      //word ammo
      //word pieces

                              //game profile:
   NM_GP_MISSION_STATE = 0x2000, //+32
      //mission state data
   NM_GP_POSITIONS = 0x2100,
      //byte max_camp, max_miss, curr_camp, curr_miss
   NM_GP_DIFFICULTY = 0x2200,
      //byte E_DIFFICULTY
   NM_GP_ADDON_NAME = 2300,
      //string
   NM_GP_ADDON_CRC = 2310,
      //dword
   NM_GP_ADDON_STATE = 2320,
      //byte max_addon_mission
};

//----------------------------

bool S_mission_state::NetSend() const{

   int i;
   {
      bool g = true;
                              //write soldier info
      net_msg->AddBits(NM_GP_SOLDIER_INFO, 2, g);
      for(i=0; i<40; i++)
         net_msg->AddBits(base_soldier_avl[i], 1, g);
      for(i=0; i<8; i++){
         int id = team_8[i];
         if(id==-1)  id = (1<<6)-1;
         assert(id < (1<<6));
         net_msg->AddBits(id, 6, g);
      }

                              //write team 4
      for(i=0; i<4; i++){
         net_msg->AddBits(NM_GP_TEAM_4, 2, g);
                              //write id
         int id = team_4[i].id;
         if(id==-1)  id = (1<<6)-1;
         assert(id < (1<<6));
         net_msg->AddBits(id, 6, g);
                              //write player's items
         for(int j=0; j<team_4[i].item_list.size(); j++){
            net_msg->AddBits(NM_GP_PLR_ITEM, 2, g);

            net_msg->AddByte(team_4[i].item_list[j].itm, g);
            net_msg->AddWord(team_4[i].item_list[j].ammo, g);
            net_msg->AddWord(team_4[i].item_list[j].pieces, g);
         }
      }
                              //write mission items
      for(i=0; i<item_list.size(); i++){
         net_msg->AddBits(NM_GP_MISS_ITEM, 2, g);

         net_msg->AddByte(item_list[i].itm, g);
         net_msg->AddWord(item_list[i].ammo, g);
         net_msg->AddWord(item_list[i].pieces, g);
      }
   }
   return true;
}

//----------------------------

bool S_mission_state::NetImport(PC_net_in_msg im){

   int i;
   int t4_pos = -1;
   word end = im->GetWord();
   end += im->GetPos();

   while(im->GetPos()!=end){
      switch(im->GetBits(2)){

      case NM_GP_SOLDIER_INFO:
         for(i=0; i<40; i++)
            base_soldier_avl[i] = im->GetBits(1);
         for(i=0; i<8; i++){
            int id = im->GetBits(6);
            if(id == (1<<6)-1)  id = -1;
            team_8[i] = id;
         }
         break;

      case NM_GP_TEAM_4:
         {
            int id = im->GetBits(6);
            if(id == (1<<6)-1)  id = -1;
            team_4[++t4_pos].id = id;
         }
         break;

      case NM_GP_PLR_ITEM:
         {
            byte id = im->GetByte();
            word ammo = im->GetWord();
            word pieces = im->GetWord();
            team_4[t4_pos].item_list.push_back(S_item(id, ammo, pieces));
         }
         break;

      case NM_GP_MISS_ITEM:
         {
            byte id = im->GetByte();
            word ammo = im->GetWord();
            word pieces = im->GetWord();
            item_list.push_back(S_item(id, ammo, pieces));
         }
         break;

      default:
         assert(0);
      }
   }
   return true;
}

//----------------------------

void S_mission_state::Close(){
   int i;
   item_list.clear();
   for(i=0; i<4; i++) team_4[i].Clear();
   for(i=0; i<MAX_BASE_SOLDIERS; i++) base_soldier_avl[i]=true;
   for(i=0; i<8; i++) team_8[i]=-1;
   save_game.Close();
   mission_crc = 0;
}

//----------------------------
//----------------------------

bool S_game_profile::Save(){

   C_chunk ck;
   if(!ck.WOpen(file_name)){
      OsCreateDirectoryTree(file_name);
                              //try again
      if(!ck.WOpen(file_name))
         return false;
   }

   ck.WAscend(CT_PROFILE);
   {                          //version + copyright
      ck.WAscend(CT_VERSION);
      word major = 1, minor = 5;
      ck.Write(&major, 2);
      ck.Write(&minor, 2);
      static const char copyright[] =
         " - Hidden and Dangerous - Player profile - Copyright (c) 2001  Lonely Cat Games - ";
      ck.Write(copyright, sizeof(copyright)+1);
      ck.Descend();
   }
                              //profile name
   ck.WStringChunk(CT_PROF_NAME, name);
   {
                              //write achieved positions
      ck.WAscend(CT_PROF_POSITIONS);
      ck.Write(&max_camp, 2);
      ck.Write(&max_mission, 2);
      ck.Write(&curr_camp, 2);
      ck.Write(&curr_mission, 2);
      ck.Descend();
   }
   {
                              //write difficulty
      ck.WAscend(CT_PROF_DIFFICULTY);
      ck.Write(&difficulty, 1);
      ck.Descend();
   }
   {
      ck.WStringChunk(CT_ADDON_NAME, addon_name);
      ck.WWordChunk(CT_ADDON_MISSION, max_addon_mission);
      ck.WIntChunk(CT_ADDON_CRC, addon_crc);
   }
                              //write all mission states
   int num_save = sizeof(mission_states)/sizeof(S_mission_state);
   for(int i=0; i<num_save; i++)
      mission_states[i].Save(ck);

   return ck.Close();
}

//----------------------------

bool S_game_profile::Open(const char *fname){

   Reset();
   ResetAddOn();

   int num_miss_read = 0;
   C_chunk ck;
   if(!ck.ROpen(fname))
      return false;
   CK_TYPE t;
   if(!ck.RAscend(t))
      goto fail;

   if(t==CT_PROFILE)
   while(ck.Size()){
      if(!ck.RAscend(t))
         goto fail;

      switch(t){
      case CT_PROF_POSITIONS:
         ck.Read(&max_camp, 2);
         ck.Read(&max_mission, 2);
         ck.Read(&curr_camp, 2);
         ck.Read(&curr_mission, 2);
         ck.Descend();
         break;

      case CT_PROF_DIFFICULTY:
         ck.Read(&difficulty, 1);
         ck.Descend();
         break;

      case CT_PROF_MISSION_STATE:
         {
            mission_states[num_miss_read++].Open(ck);
            ck.Descend();
#ifndef _DEBUG
                              //check crc
            int miss_index = num_miss_read-1;
            dword miss_crc;
            C_fstring mname("%s\\game", tab_campaigns->ItemS(TAB_S12_CM_MISS_DIR, miss_index, 12));
            if(!C_mission::GetMissionCRC(mname, miss_crc) || (miss_crc != mission_states[miss_index].mission_crc) ){
               mission_states[miss_index].save_game.Close();
            }
#endif
         }
         break;

      case CT_PROF_NAME:
         name = ck.RStringChunk();
         break;

      case CT_VERSION:
         {
            word major = 0, minor = 0;
            ck.Read(&major, 2);
            ck.Read(&minor, 2);
            if(major!=1 || minor!=5)
               return false;
            ck.Descend();
         }
         break;

      case CT_ADDON_NAME:
         addon_name = ck.RStringChunk();
         break;

      case CT_ADDON_MISSION:
         max_addon_mission = ck.RWordChunk();
         break;

      case CT_ADDON_CRC:
         addon_crc = ck.RIntChunk();
         break;

      default:
         assert(0);
         ck.Descend();
      }
   };
   ck.Descend();
   file_name = fname;
   return ck.Close();
fail:
   ck.Close();
   return false;
}

//----------------------------

bool S_game_profile::NetSend() const{

   net_msg->Ascend(NM_GAME_PROFILE);
   {
      const bool g = true;
                              //write achieved positions
      net_msg->AddID(NM_GP_POSITIONS, g);
         net_msg->AddByte(max_camp, g);
         net_msg->AddByte(max_mission, g);
         net_msg->AddByte(curr_camp, g);
         net_msg->AddByte(curr_mission, g);

                              //write difficulty
      net_msg->AddID(NM_GP_DIFFICULTY, g);
         net_msg->AddByte(difficulty, g);

                              //add-on mission data
      net_msg->AddID(NM_GP_ADDON_NAME, g);
         net_msg->AddString(addon_name, g);

      net_msg->AddID(NM_GP_ADDON_CRC, g);
         net_msg->AddDword(addon_crc, g);

      net_msg->AddID(NM_GP_ADDON_STATE, g);
         net_msg->AddByte(max_addon_mission, g);;
   }
                              //write mission states till now
   //int num_save_missions = MissionIndex(max_camp, max_mission) + 1;
   //assert(num_save_missions < sizeof(mission_states)/sizeof(S_mission_state));
   int num_save_missions = sizeof(mission_states)/sizeof(S_mission_state);

   for(int i=0; i<num_save_missions; i++){
      net_msg->Ascend(NM_GP_MISSION_STATE+i);
      mission_states[i].NetSend();
      net_msg->Descend();
   }
   net_msg->Descend();
   return true;
}

//----------------------------

bool S_game_profile::NetImport(PC_net_in_msg im){

   Reset();
   int num_miss_read = 0;

   word end = im->GetWord();
   end += im->GetPos();
   while(im->GetPos()!=end){

      word nm = im->GetID();
      switch(nm){
      case NM_GP_POSITIONS:
         max_camp       = im->GetByte();
         max_mission    = im->GetByte();
         curr_camp      = im->GetByte();
         curr_mission   = im->GetByte();
         break;

      case NM_GP_DIFFICULTY:
         difficulty = (E_DIFFICULTY)im->GetByte();
         break;

      case NM_GP_ADDON_NAME:
         {
            char buf[256];
            im->GetString(buf, sizeof(buf));
            addon_name = buf;
         }
         break;

      case NM_GP_ADDON_CRC:
         addon_crc = im->GetDword();
         break;

      case NM_GP_ADDON_STATE:
         max_addon_mission = im->GetByte();
         break;

      default:
         assert((nm-NM_GP_MISSION_STATE) == num_miss_read);
         mission_states[num_miss_read++].NetImport(im);
         break;
      }
   }
   return true;
}

//----------------------------

int S_game_profile::MissionIndex(int camp, int miss) const{

   if(camp==-1) camp = curr_camp;
   if(miss==-1) miss = curr_mission;
   int ret = 0;
   for(int i=0; i<camp; i++)
      ret += NumMissionsInCamp(i);
   return ret + miss;
}

//----------------------------

const char *S_game_profile::GetCurrMissionName() const{
   return tab_campaigns->ItemS(TAB_S12_CM_MISS_DIR, MissionIndex(), 12);
}

//----------------------------

int S_game_profile::NumMissionsInCamp(int camp) const{

   if(camp==-1)
      camp = curr_camp;
   if(camp==9)
      return num_addon_missions;
   return tab_campaigns->ItemI(TAB_I_CM_NUM_MISSIONS, camp);
}

//----------------------------

bool S_game_profile::GetAddOnCRC(dword &crc){

   crc = 0;
   for(int i=0; i<num_addon_missions; i++){
      const char *mname = tab_campaigns->ItemS(TAB_S12_CM_MISS_DIR, NUM_MISSIONS+i, 12);
      C_fstring fname("Missions\\%s\\Game\\scene.bin", mname);
      int h = dtaOpen(fname);
      if(h==-1)
         return false;
      dword ftime[2];
      dtaSeek(h, 0, DTA_SEEK_END);
      crc ^= dtaSeek(h, 0, DTA_SEEK_CUR);
      dtaGetTime(h, ftime);
      crc ^= ftime[0] ^ ftime[1];

      dtaClose(h);
   }
   return true;
}

//----------------------------

bool S_game_profile::OpenAddOn(const char *fname){

   if(fname[0]!='.'){
      addon_dta = dtaCreate(fname);
      if(!addon_dta)
         return false;
      addon_dta->Release();
   }
                              //determine number of missions
   for(num_addon_missions=0; num_addon_missions<6; num_addon_missions++){
      const char *mname = tab_campaigns->ItemS(TAB_S12_CM_MISS_DIR, NUM_MISSIONS+num_addon_missions, 12);
      C_fstring fname("Missions\\%s\\Game\\scene.bin", mname);
      int h = dtaOpen(fname);
      if(h==-1)
         break;
                              //check if the file is really located in our DTA
      bool this_dta = (dtaGetDtaPack(h)==addon_dta);
      dtaClose(h);
      if(!this_dta)
         break;
   }
   if(!num_addon_missions){
      addon_dta = NULL;
      return false;
   }
                              //install texts
   all_txt.AddFile("Text.mod\\AddOn.txt");

   bool addon_match = false;
   dword this_crc;
   GetAddOnCRC(this_crc);
   if(addon_name==fname){
                              //check CRC
      addon_match = (this_crc==addon_crc);
   }
   if(!addon_match){
      for(int i=0; i<MAX_NUM_ADDON_MISS; i++)
         mission_states[NUM_MISSIONS+i].Close();
      max_addon_mission = 0;
   }
   addon_name = fname;
   addon_crc = this_crc;
   return true;
}

//----------------------------

void S_game_profile::CloseAddOn(){

   if(num_addon_missions){
      num_addon_missions = 0;
      all_txt.DelTexts(100000, 100000);
      addon_dta = NULL;
   }
}

//----------------------------
//----------------------------

void SearchAccounts(vector<S_logon_account> &account_list){

   account_list.clear();

                              //build list of savegame names from savegame directory
   vector<C_string> fnames;
   OsCollectFiles("Savegame", "*.bin", fnames);

   for(int i=fnames.size(); i--; ){
      S_game_profile gp;
      const char *fname = fnames[i];
      if(gp.Open(fname)){
         account_list.push_back(S_logon_account());
         S_logon_account &la = account_list.back();
         la.file_name = fnames[i];
         la.user_name = gp.name;
         la.difficulty = gp.difficulty;
         OsGetFileTimeDate(fname, la.dt);
      }
   }
   sort(account_list.begin(), account_list.end());
   reverse(account_list.begin(), account_list.end());
}

//----------------------------

C_string MakeGameText(const C_string &str){

   if(game_configuration.language!=5)
      return str;

   C_string ret;
                              //make wide-char text
   wchar_t *wp = new wchar_t[str.Size()+1];
   for(int i=0; i<str.Size(); i++) wp[i] = str[i];
   wp[i] = 0;
   ret.Assign((const char*)wp, (str.Size()+1)*sizeof(wchar_t));
   delete[] wp;
   return ret;
}

//----------------------------

C_string AddGameText(const C_string &str1, const C_string &str2){

   C_string ret;
   if(game_configuration.language!=5){
      ret = str1 + str2;
   }else{
      int size1 = str1.Size() ? wcslen((wchar_t*)(const char*)str1) : 0;
      int size2 = str2.Size() ? wcslen((wchar_t*)(const char*)str2) : 0;
      int len = size1 + size2 + 1;
      wchar_t *wp = new wchar_t[len];
      if(size1) wcscpy(wp, (wchar_t*)&str1[0]);
      if(size2) wcscpy(&wp[size1], (wchar_t*)&str2[0]);
      wp[size1+size2] = 0;

      ret.Assign((const char*)wp, len*sizeof(wchar_t));
      delete[] wp;
   }
   return ret;
}

//----------------------------
//----------------------------
