#include "all.h"
#include "i3d_font.h"

//----------------------------
//----------------------------

struct S_text_vertex{
   S_vector pos;
   float rhw;
   dword diffuse;
   //dword specular;            //padding to 32 bytes
   I3D_text_coor tx;
};

//----------------------------

C_poly_sprite::C_poly_sprite(float x, float y, float spr_size,
   LPC_sprite_image img1, dword indx, int z1, float in_aspect):
   on(true),
   sort_z(z1),
   img(img1),
   vis_contents_dirty(true),
   sprite_index(indx),
   color(0xffffffff)
{
   assert(img);
   const vector<C_sprite_image::S_spr_rect> &rects = img->GetRects();
   assert(indx<rects.size());
   const C_sprite_image::S_spr_rect &rc = rects[indx];
   float sx = rc.sx * spr_size / rc.sy * in_aspect;

   float scrn_sx = igraph->Scrn_sx();

   size.x = sx;
   size.y = spr_size;

   hotspot.x = rc.cx * sx;
   hotspot.y = rc.cy * spr_size;
                              //correct scale problem
   sx *= img->GetAspectRatio();

   for(int i=0; i<4; i++){
      tlv[i].v.z = 1.0f;
      tlv[i].rhw = 1.0f;
      tlv[i].diffuse = 0xffffffff;
      S_vector2 &v = screen_rect[i];
      v.x = -rc.cx * sx * scrn_sx;
      v.y = -rc.cy * spr_size * scrn_sx;
      if(i&1) v.x += sx * scrn_sx;
      if(i&2) v.y += spr_size * scrn_sx;
                              //setup texture coordinates
      tlv[i].tx.u = rc.l;
      tlv[i].tx.v = rc.t;
      if(i&1) tlv[i].tx.u += rc.sx;
      if(i&2) tlv[i].tx.v += rc.sy;
   }

   SetPos(x, y);
   PrepareDraw();
}

//----------------------------

void C_poly_sprite::PrepareDraw(){

   float aspect = screen_aspect_ratio * 1.33333333f;
                              //deform by resolution
   for(int j=0; j<4; j++){
      tlv[j].v.x = (screen_pos.x + screen_rect[j].x);
      tlv[j].v.y = (screen_pos.y + screen_rect[j].y) * aspect;
   }
   vis_contents_dirty = true;
}

//----------------------------

void C_poly_sprite::Shear(float angle, float base_y){

   float dy = base_y - screen_pos.y;
   {
      float sy1 = -screen_rect[0].y;
      sy1 += dy;
      float dx = sy1 * tan(angle);
      screen_rect[0].x += dx;
      screen_rect[1].x += dx;
   }
   {
      float sy1 = -screen_rect[2].y;
      sy1 += dy;
      float dx = sy1 * tan(angle);
      screen_rect[2].x += dx;
      screen_rect[3].x += dx;
   }
   PrepareDraw();
}

//----------------------------

void C_poly_sprite::SetPos(float x1, float y1){

   pos.x = x1;
   pos.y = y1;

   float scrn_sx = igraph->Scrn_sx();
   SetScreenPos((int)(scrn_sx*pos.x), (int)(scrn_sx*pos.y));
}

//----------------------------

void C_poly_sprite::SetScreenPos(float x1, float y1){

   screen_pos.x = x1;
   screen_pos.y = y1;
   PrepareDraw();
}

//----------------------------

void C_poly_sprite::SetScreenSize(float sx, float sy){

   screen_rect[1].x = screen_rect[3].x = screen_rect[0].x + sx * (float)igraph->Scrn_sx();
   screen_rect[2].y = screen_rect[3].y = screen_rect[0].y + sy * (float)igraph->Scrn_sx();
   size.x = sx;
   size.y = sy;
   PrepareDraw();
}

//----------------------------

void C_poly_sprite::Rotate(float angle){

   float s = sin(angle), c = cos(angle);
   for(int i=0; i<4; i++){
      tlv[i].v.x = screen_pos.x + screen_rect[i].x*c - screen_rect[i].y*s;
      tlv[i].v.y = screen_pos.y + screen_rect[i].x*s + screen_rect[i].y*c;
   }
   vis_contents_dirty = true;
}

//----------------------------

void C_poly_sprite::SetTileX(float s, float e){

   tlv[1].tx.u = (tlv[0].tx.u = s) + e;
   tlv[3].tx.u = (tlv[2].tx.u = s) + e;
}

//----------------------------

void C_poly_sprite::SetColor(dword c){

   color = c;

   for(int i=4; i--; )
      tlv[i].diffuse = c;
   vis_contents_dirty = true;
}

//----------------------------
static const word indx[] = {0, 1, 2, 2, 1, 3};

void C_poly_sprite::Render(){

   if(!on)
      return;

//----------------------------
                              //prepare dynamic visual, if not yet
   if(vis_contents_dirty){
      if(!dyn_vis){
         dyn_vis = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);
         dyn_vis->Release();
      }

      I3D_triface fc[2];
      fc[0][0] = 2;
      fc[0][1] = 0;
      fc[0][2] = 1;
      fc[1][0] = 2;
      fc[1][1] = 1;
      fc[1][2] = 3;

      I3D_face_group fg;
      PI3D_material mat = driver->CreateMaterial();

      fg.base_index = 0;
      fg.num_faces = 2;

      fg.SetMaterial(mat);
      mat->SetTexture(img->GetTexture());
      mat->Release();

      dyn_vis->Build(tlv, 4, I3DVC_XYZRHW | I3DVC_DIFFUSE | /*I3DVC_SPECULAR | */(1<<I3DVC_TEXCOUNT_SHIFT),
         fc, 2, &fg, 1);

      vis_contents_dirty = false;
   }
   dyn_vis->Render();
}

//----------------------------

void C_poly_sprite::GetExtent(float &l, float &t, float &r, float &b) const{

   l = pos.x - hotspot.x;
   t = pos.y - hotspot.y;
   r = l + size.x;
   b = t + size.y;
}

//----------------------------
// Call with param diff_name relative to dir "Maps\\".
// 'font_image' is name of image to be used for building font spacing table.
//----------------------------
bool C_sprite_image::Init(PI3D_driver driver, const char *prjname, const char *diff_name,
   const char *opt_name, dword txt_create_flags,
   bool use_new_prj_format, float border_offset){

   int i;
   C_cache cache;
   if(!cache.open(prjname, CACHE_READ))
      return false;

                              //load texture
   I3D_RESULT ir;
   I3D_CREATETEXTURE ct;
   memset(&ct, 0, sizeof(ct));

   ct.flags = txt_create_flags | TEXTMAP_DIFFUSE;
                              //if mipmapping not requested, disable it
   if(!(txt_create_flags&TEXTMAP_MIPMAP))
      ct.flags |= TEXTMAP_NOMIPMAP;

   //if(!diff_name) diff_name = header.img_name;
   assert(diff_name);

   ct.text_name = diff_name;
   ct.opt_name = opt_name;

   if(opt_name) ct.flags |= TEXTMAP_OPACITY;
   PI3D_texture tp1;
   ir = driver->CreateTexture(&ct, &tp1);
   if(I3D_FAIL(ir)){
      cache.close();
      Close();
      return false;
   }
   tp = tp1;
   tp1->Release();

   dword img_sx, img_sy; 
   byte bpp; 
   dword id;

   bool b = igraph->GetImageInfo(C_fstring("Maps\\%s", ct.text_name), img_sx, img_sy, bpp, id);
   if(!b){
      b = igraph->GetImageInfo(C_fstring("Maps\\%s", ct.text_name), img_sx, img_sy, bpp, id);
      if(!b){
         cache.close();
         Close();
         return false;
      }
   }
   image_aspect_ratio = (float)img_sx / (float)img_sy;

                              //read rectangle definitions
   if(!use_new_prj_format){
      struct{
         dword id;
         word img_num;
         word spr_num;
         dword res[2];
         char img_name[16];
      }header;
      cache.read((char*)&header, sizeof(header));
      spr_rects.reserve(header.spr_num);
      cache.seekp(header.img_num * 16 + 16);

      struct{
         word x, y, sx, sy;
         short cx, cy;
         dword res;
      } rc;

      for(i=0; i<header.spr_num; i++){
         cache.read((char*)&rc, sizeof(rc));
         spr_rects.push_back(S_spr_rect(
            (float(rc.x) - border_offset)/float(img_sx),
            (float(rc.y) - border_offset)/float(img_sy),
            (float(rc.sx) + border_offset * 2.0f)/float(img_sx),
            (float(rc.sy) + border_offset * 2.0f)/float(img_sy),
            (float)rc.cx/(float)rc.sx,
            (float)rc.cy/(float)rc.sy));
         S_spr_rect &src = spr_rects.back();
         src.ix = rc.x;
         src.iy = rc.y;
         src.isx = rc.sx;
         src.isy = rc.sy;
         src.icx = rc.cx;
         src.icy = rc.cy;
      }
   }else{
      int line_num = 0;
      while(!cache.eof()){
         ++line_num;
                              //get entire line
         char line[512];
         cache.getline(line, sizeof(line));
         int line_size = strlen(line);
                              //skip white-space
         for(int i=0; i<line_size; i++){
            if(!isspace(line[i]))
               break;
         }
                              //skip empty or commented lines
         if(!line[i] || line[i]==';')
            continue;
                              //decode line
         int sn, x, y, sx, sy, cx, cy;
         int num_scanned = sscanf(line, "%i: %i, %i, %i, %i, %i, %i", &sn, &x, &y, &sx, &sy, &cx, &cy);
         if(num_scanned!=7){
            ErrReport(C_fstring("Sprite def file '%s' (%i): syntax error.", prjname, line_num), editor);
            continue;
         }
         if(sn < spr_rects.size()){
            ErrReport(C_fstring("Sprite def file '%s' (%i): sprite %i already defined.", prjname, line_num, sn), editor);
         }
                              //scan ok, add to sprites
         while(spr_rects.size() < sn){
            spr_rects.push_back(S_spr_rect());
            memset(&spr_rects.back(), 0, sizeof(S_spr_rect));
         }
         spr_rects.push_back(S_spr_rect(
            (float(x) - border_offset)/float(img_sx),
            (float(y) - border_offset)/float(img_sy),
            (float(sx) + border_offset * 2.0f)/float(img_sx),
            (float(sy) + border_offset * 2.0f)/float(img_sy),
            (float)cx/(float)sx,
            (float)cy/(float)sy));

         S_spr_rect &rc = spr_rects.back();
         rc.ix = x;
         rc.iy = y;
         rc.isx = sx;
         rc.isy = sy;
         rc.icx = cx;
         rc.icy = cy;
      }
   }
   cache.close();

   return true;
}

//----------------------------

bool C_poly_text::Init(){

   const char *font_image = "gamefont.png";
   {
      LPIImage font_img1;
      driver->GetGraphInterface()->CreateImage(&font_img1);
      S_pixelformat pf;
      memset(&pf, 0, sizeof(pf));
      pf.bytes_per_pixel = 1;
      pf.flags = PIXELFORMAT_PALETTE;

      if(!font_img1->Open(C_fstring("Maps\\%s", font_image),
         IMGOPEN_SYSMEM,
         0, 0, &pf)){

         font_img1->Release();
         return false;
      }
                              //build font spacing table
                              //get access to bitmap data
      byte *font_data;
      dword pitch;
      byte bpp;
      font_img1->Lock((void**)&font_data, &pitch, &bpp);

      font_size_offsets.clear();
      const vector<C_sprite_image::S_spr_rect> &rects = font_img.GetRects();
      int num_letters = rects.size();
      font_size_offsets.reserve(num_letters);
      font_sizes.clear();
                              //build all letters
      for(int i = 0; i<num_letters; i++){
         const C_sprite_image::S_spr_rect &rc = rects[i];
                              //get address of letter
         byte *mem = font_data + rc.iy * pitch + rc.ix;
                              //assign pointer
         font_size_offsets.push_back(font_sizes.size());

         dword sx = rc.isx, sy = rc.isy;
                              //process all letter's lines
         for(int j=0; j<sy; j++, mem += pitch){
            font_sizes.push_back(S_font_size());
            S_font_size &fs = font_sizes.back();

                              //find left
            fs.l = 0;
            byte *bm = mem;
            while(!*bm && fs.l<sx)
               ++bm, ++fs.l;
                              //find right
            fs.r = sx;
            bm = mem + sx;
            while(!*--bm && fs.r)
               --fs.r;
         }
      }

      font_img1->Unlock();
      font_img1->Release();
   }
   return true;
}

//----------------------------

void C_sprite_image::Close(){

   tp = NULL;
   spr_rects.clear();
}

//----------------------------
//----------------------------

C_sprite_group::C_sprite_group():
   resort(false)
{
   pos.Zero();
}

//----------------------------

static bool spr_less(C_poly_sprite *s1, C_poly_sprite *s2){ return s1->GetZ()<s2->GetZ(); }

//----------------------------

void C_sprite_group::AddSprite(LPC_poly_sprite spr){

   sprs.push_back(spr);
   if(spr->GetZ() != 0)
      resort = true;
}

//----------------------------

void C_sprite_group::RemoveSprite(C_poly_sprite *sp){

   for(int i=sprs.size(); i--; )
   if(sprs[i]==sp){
      sprs[i] = sprs.back(); sprs.pop_back();
      resort = true;
      break;
   }
}

//----------------------------

void C_sprite_group::SetColor(dword c){

   for(int i=0; i<sprs.size(); i++)
      sprs[i]->SetColor(c);
}

//----------------------------

void C_sprite_group::SetOn(bool b){
   for(int i=0; i<sprs.size(); i++) sprs[i]->SetOn(b);
}

//----------------------------

void C_sprite_group::Render(){

   if(resort){
      sort(sprs.begin(), sprs.end(), spr_less);
      resort = false;
   }
   for(int i=0; i<sprs.size(); i++)
      sprs[i]->Render();
}

//----------------------------
//----------------------------

C_text::~C_text(){

   if(texts)
      texts->RemoveText(this);
}

//----------------------------

void C_text::GetExtent(float &l, float &t, float &r, float &b) const{

   l = 1.0f;
   t = 1.0f;
   r = 0.0f;
   b = 0.0f;

   if(poly_text){
      const vector<C_smart_ptr<C_poly_sprite> > &sprs = poly_text->GetSPrites();
      for(int i=sprs.size(); i--; ){
         S_vector2 pos = sprs[i]->GetPos();
         const S_vector2 &size = sprs[i]->GetSize();
         const S_vector2 &hotspot = sprs[i]->GetHotSpot();
         pos -= hotspot;
         l = Min(l, pos.x);
         t = Min(t, pos.y);
         r = Max(r, pos.x + size.x);
         b = Max(b, pos.y + size.y);
      }
   }else
   if(wide_text){
      wide_text->GetExtent(l, t, r, b);
   }else{
      assert(0);
   }
}

//----------------------------

void C_text::SetColor(dword c){

   color = c;
   if(poly_text){
      poly_text->SetColor(c);

      if(dyn_vis){
         S_text_vertex *v = (S_text_vertex*)dyn_vis->LockVertices(I3DLOCK_DIFFUSE);
         if(v){
            for(int i=dyn_vis->NumVertices(); i--; )
               v[i].diffuse = c;
            dyn_vis->UnlockVertices();
         }
      }
   }else
      wide_text->SetColor(c);
}

//----------------------------

void C_text::SetPos(float x, float y){

   if(poly_text){
      float dx, dy;
      GetPos(dx, dy);

      dx = x - dx;
      dy = y - dy;

      for(int i=0; i<poly_text->sprs.size(); i++){
         const S_vector2 &pos = poly_text->sprs[i]->GetPos();
         poly_text->sprs[i]->SetPos(pos.x + dx, pos.y + dy);
      }
      poly_text->pos.x = x;
      poly_text->pos.y = y;

      float scrn_sx = igraph->Scrn_sx();
      dx *= scrn_sx;
      dy *= scrn_sx;

      S_text_vertex *v = (S_text_vertex*)dyn_vis->LockVertices(I3DLOCK_XYZ);
      dword numv = dyn_vis->NumVertices();
      for(i=numv; i--; ){
         v[i].pos.x += dx;
         v[i].pos.y += dy;
      }
      dyn_vis->UnlockVertices();
   }else{
      wide_text->SetPos(x, y);
   }
}

//----------------------------

void C_text::GetPos(float &x, float &y){

   if(poly_text){
      const S_vector2 &pos = poly_text->GetPos();
      x = pos.x;
      y = pos.y;
   }else{
      wide_text->GetPos(x, y);
   }
}

//----------------------------

void C_text::Render() const{

   if(!on)
      return;
   if(poly_text)
      dyn_vis->Render();
   else
      wide_text->Render();
}

//----------------------------

void C_text::CreateVisual(){

   dyn_vis = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);
   dyn_vis->Release();

   vector<S_text_vertex> v;
   vector<I3D_triface> f;
   I3D_face_group fg;
   PI3D_material mat = driver->CreateMaterial();

   if(poly_text){
      const vector<C_smart_ptr<C_poly_sprite> > &sprs = poly_text->GetSPrites();
      if(sprs.size()){
         v.reserve(sprs.size()*4);
         f.reserve(sprs.size()*2);
         mat->SetTexture(sprs[0]->img->GetTexture());
      }

      for(int i=sprs.size(); i--; ){
                              //construct 2 faces of single letter
         I3D_triface fc;
         fc[0] = v.size() + 2;
         fc[1] = fc[0] - 2;
         fc[2] = fc[0] - 1;
         f.push_back(fc);
         fc[1] = fc[0] - 1;
         fc[2] = fc[0] + 1;
         f.push_back(fc);
                              //contruct vertices
         const C_poly_sprite *spr = sprs[i];
         for(int j=0; j<4; j++){
            const C_poly_sprite::S_vertex &sv = spr->tlv[j];
            S_text_vertex vx;
            vx.pos = sv.v;
            vx.rhw = sv.rhw;
            vx.diffuse = color;
            vx.tx = sv.tx;
            v.push_back(vx);
         }
      }
   }else{
      assert(0);
   }

   fg.base_index = 0;
   fg.num_faces = f.size();

   fg.SetMaterial(mat);
   mat->Release();

   dyn_vis->Build(v.begin(), v.size(), I3DVC_XYZRHW | I3DVC_DIFFUSE | /*I3DVC_SPECULAR | */(1<<I3DVC_TEXCOUNT_SHIFT),
      f.begin(), f.size(), &fg, 1);
}

//----------------------------
//----------------------------
                              //national support
/*--------------------------------------------------------
   punktation marks:
   0x80  -  a' (?)   a`'
   0x81  -  cv       c`v
   0x82  -  d'       d`|
   0x83  -  a: (?)   a`:
   0x84  -  e^ (?)   e`^
   0x85  -  ao (?)   a`o
   0x86  -  c, (?)   c`,
   0x87  -  c. (?)   c`.
   0x88  -  a` (?)   a``
   0x89  -  n~ (?)   n`~
--------------------------------------------------------*/
#define DLZEN     0x80
#define MAKCEN    0x81
#define MAKCEN1   0x82
#define DVOJBODKA 0x83
#define BOKAN     0x84
#define KRUZOK    0x85
#define PL1       0x86
#define PL2       0x87
#define DLZEN_L   0x88
#define VLNOVKA   0x89
#define BODKA     0x8a
#define STRIKE1   0x8b
#define STRIKE2   0x8c
#define GERMAN_S  's'

static const struct S_national_table{
   byte orig;
   char base;
   byte mark;
} ntab_1[] = {
   0x8a, 'S', MAKCEN,
   0x8c, 'S', DLZEN,
   0x8d, 'T', MAKCEN,
   0x8e, 'Z', MAKCEN,

   0x92, '\'', 0,
   0x9a, 's', MAKCEN,
   0x9c, 's', DLZEN,
   0x9d, 't', MAKCEN1,
   0x9e, 'z', MAKCEN,

   0xbc, 'L', MAKCEN,
   0xbe, 'l', MAKCEN1,

   0xc0, 'R', DLZEN,
   0xc1, 'A', DLZEN,
   0xc4, 'A', DVOJBODKA,
   0xc5, 'L', DLZEN,
   0xc6, 'C', DLZEN,
   0xc8, 'C', MAKCEN,
   0xc9, 'E', DLZEN,
   0xcb, 'E', DVOJBODKA,
   0xcc, 'E', MAKCEN,
   0xcd, 'I', DLZEN,
   0xcf, 'D', MAKCEN1,

   0xd1, 'N', DLZEN,
   0xd2, 'N', MAKCEN,
   0xd3, 'O', DLZEN,
   0xd6, 'O', DVOJBODKA,
   0xd8, 'R', MAKCEN,
   0xd9, 'U', KRUZOK,
   0xda, 'U', DLZEN,
   0xdc, 'U', DVOJBODKA,
   0xdd, 'Y', DLZEN,
   0xdf, 's', GERMAN_S,       //german s converted to ss

   //0xe0, 'r', DLZEN,
   0xe0, 'a', DLZEN_L,
   0xe1, 'a', DLZEN,
   0xe2, 'a', BOKAN,
   0xe4, 'a', DVOJBODKA,
   0xe5, 'l', DLZEN,
   0xe6, 'c', DLZEN,
   0xe7, 'c', PL1,
   //0xe8, 'c', MAKCEN,
   0xe8, 'e', DLZEN_L,
   0xe9, 'e', DLZEN,
   0xea, 'e', BOKAN,
   0xeb, 'e', DVOJBODKA,
   //0xec, 'e', MAKCEN,
   0xec, 127, DLZEN_L,
   0xed, 127, DLZEN,
   0xee, 127, BOKAN,
   0xef, 'd', MAKCEN1,

   0xf1, 'n', VLNOVKA,
   //0xf2, 'n', MAKCEN,
   0xf2, 'o', DLZEN_L,
   0xf3, 'o', DLZEN,
   0xf4, 'o', BOKAN,
   0xf6, 'o', DVOJBODKA,
   0xf8, 'r', MAKCEN,
   //0xf9, 'u', KRUZOK,
   0xf9, 'u', DLZEN_L,
   0xfa, 'u', DLZEN,
   0xfb, 'u', BOKAN,
   0xfc, 'u', DVOJBODKA,
   0xfd, 'y', DLZEN,
   0
}, ntab_2[] = {               //Czech, Polish - special conversion table
   0x8a, 'S', MAKCEN,
   0x8c, 'S', DLZEN,
   0x8d, 'T', MAKCEN,
   0x8e, 'Z', MAKCEN,
   0x8f, 'Z', DLZEN,

   0x92, '\'', 0,
   0x9a, 's', MAKCEN,
   0x9c, 's', DLZEN,
   0x9d, 't', MAKCEN1,
   0x9e, 'z', MAKCEN,
   0x9f, 'z', DLZEN,

   0xa3, 'L', STRIKE1,
   0xa5, 'A', PL2,
   0xaf, 'Z', BODKA,

   0xb3, 'l', STRIKE1,
   0xb9, 'a', PL2,
   0xbc, 'L', MAKCEN,
   0xbe, 'l', MAKCEN1,
   0xbf, 'z', BODKA,

   0xc0, 'R', DLZEN,
   0xc1, 'A', DLZEN,
   0xc4, 'A', DVOJBODKA,
   0xc5, 'L', DLZEN,
   0xc6, 'C', DLZEN,
   0xc8, 'C', MAKCEN,
   0xc9, 'E', DLZEN,
   0xca, 'E', PL2,
   0xcb, 'E', DVOJBODKA,
   0xcc, 'E', MAKCEN,
   0xcd, 'I', DLZEN,
   0xcf, 'D', MAKCEN1,

   0xd1, 'N', DLZEN,
   0xd2, 'N', MAKCEN,
   0xd3, 'O', DLZEN,
   0xd6, 'O', DVOJBODKA,
   0xd8, 'R', MAKCEN,
   0xd9, 'U', KRUZOK,
   0xda, 'U', DLZEN,
   0xdc, 'U', DVOJBODKA,
   0xdd, 'Y', DLZEN,
   0xdf, 's', GERMAN_S,       //german s converted to ss

   0xe0, 'r', DLZEN,
   //0xe0, 'a', DLZEN_L,
   0xe1, 'a', DLZEN,
   0xe2, 'a', BOKAN,
   0xe4, 'a', DVOJBODKA,
   0xe5, 'l', DLZEN,
   0xe6, 'c', DLZEN,
   0xe7, 'c', PL1,
   0xe8, 'c', MAKCEN,
   //0xe8, 'e', DLZEN_L,
   0xe9, 'e', DLZEN,
   0xea, 'e', PL2,
   0xeb, 'e', DVOJBODKA,
   0xec, 'e', MAKCEN,
   //0xec, 127, DLZEN_L,
   0xed, 127, DLZEN,
   0xee, 127, BOKAN,
   0xef, 'd', MAKCEN1,

   0xf1, 'n', DLZEN,
   0xf2, 'n', MAKCEN,
   //0xf2, 'o', DLZEN_L,
   0xf3, 'o', DLZEN,
   0xf4, 'o', BOKAN,
   0xf6, 'o', DVOJBODKA,
   0xf8, 'r', MAKCEN,
   0xf9, 'u', KRUZOK,
   //0xf9, 'u', DLZEN_L,
   0xfa, 'u', DLZEN,
   0xfb, 'u', BOKAN,
   0xfc, 'u', DVOJBODKA,
   0xfd, 'y', DLZEN,
   0
};

//----------------------------

C_poly_text::~C_poly_text(){
   Close();
}

//----------------------------

void C_poly_text::Close(){

   //font_img = NULL;
   assert(!texts.size());
   /*
   while(texts.size())
      delete texts.front();
      */
}

//----------------------------

                              //find closest position for two letters
word C_poly_text::LetterFit(byte l, byte r){

   l -= ' ', r -= ' ';
   if(l>102 || r>102)
      return 0;
   int xdist = 0;

   const C_sprite_image::S_spr_rect *font_rects = font_img.GetRects().begin();;

   S_font_size *lfs = &font_sizes[font_size_offsets[l]],
      *rfs = &font_sizes[font_size_offsets[r]];
   int lcy = font_rects[l].icy, rcy = font_rects[r].icy;
   int lsy = font_rects[l].isy, rsy = font_rects[r].isy;
   int lsy1 = lsy;
   S_font_size *lfs1 = lfs;
                              //adjust top
   if(lcy>rcy)
      lfs += (lcy-rcy), lsy += (rcy-lcy);
   else
      rfs += (rcy-lcy), rsy += (lcy-rcy);
                              //get compare height
   int sy = Min(lsy, rsy);
   if(sy<=0){
                              //no overlapping, use hardspace
      while(lsy1-- > 0){
                              //compare line
         xdist = Max(xdist, (int)lfs1->r);
         ++lfs1;
      }
   }else{
      int average = 0;
      int max = 0;
      for(int i=sy; i--; ++lfs, ++rfs){
                              //compare line
         int distance = lfs->r - rfs->l;
         average += distance;
         max = Max(max, Max(distance, 0));
      }
      average /= sy;
      const float smooth_ratio = 0.4f;
      xdist = max - (max - average) * smooth_ratio;
   }
   return xdist;
}

//----------------------------

void C_poly_text::BuildSpaceTable(const char *text, short *tab, int tablen){

   int i;
   dword x = 0;
   char last = ' ';
   bool first = true;
   const C_sprite_image::S_spr_rect *font_rects = font_img.GetRects().begin();;

   int space_width = font_rects[0].isx / 2;
   int space_height = font_rects[0].isy;

   for(i=0; ; i++){
      byte c = text[i];       //get one character
      if(!c) break;           //end of text
      short *tp = &tab[i*2];

      if(c<' '){
         if(text[i]=='\n'){
            tp[0] = x + font_rects[last-' '].isx;
            x = 0;
         }else
            tp[0] = 0;
         tp[1] = 0;
         first = true;
      }else
      if(c>=128 && c<0x90){
         switch(c){
         case DLZEN:
         case MAKCEN:
         case DVOJBODKA:
         case BOKAN:
         case KRUZOK:
         case DLZEN_L:
         case VLNOVKA:
         case BODKA:
                              //put mark onto center of letter
            tp[0] = x + 1 + font_rects[last-' '].isx / 2;
            tp[1] = font_rects[last-' '].icy + 1;
            break;
         case PL1:
         case PL2:
                              //put mark below center of letter
            tp[0] = x + 1 + font_rects[last-' '].isx / 2;
            tp[1] = -1;
            break;
         case STRIKE1:
         case STRIKE2:
            tp[0] = x + 1 + font_rects[last-' '].isx / 2;
            tp[1] = 0;
            break;
         case MAKCEN1:
            tp[0] = x + font_rects[last-' '].isx - 2;
            tp[1] = font_rects[last-' '].icy;
            x += 3;
            break;
         }
      }else{
         switch(c){
         case ' ':
            x += space_height / 4;
            break;
         default:
            if(first)
               first = false;
            else
               x += LetterFit(last, c) + space_width;
            last = c;
         }
         tp[0] = x;
         tp[1] = 0;
      }
   }
   tab[i*2] = x + font_rects[last-' '].isx;
   tab[i*2+1] = 0;
}

//----------------------------
// Converts specially formatted string into native formats. It includes
// national characters adds punktations, '\' into EOL and
// special national encoding.
static dword FormatString(const char *in_str, char *out_str, dword len, bool convert_eol,
   const S_national_table *ntab){

   int i;
   dword out_len = 0;

   for(i=0; i<len; i++){
      byte c = in_str[i];
      if(c>=128){
                              //try to find in national table
         for(int ii = 0; ntab[ii].orig; ii++){
            if(ntab[ii].orig==c)
               break;
         }
         if(ntab[ii].orig){
            out_str[out_len++] = ntab[ii].base;
            if(ntab[ii].mark)
               out_str[out_len++] = ntab[ii].mark;
         }
      }else
      if(c=='\\' && convert_eol){
         out_str[out_len++] = '\n';
      }else
      if(c=='`'){
         if(out_str[out_len-1]=='i')
            out_str[out_len-1] = 127;
         char punkt_code = in_str[++i];
         switch(punkt_code){
         case '\'': out_str[out_len++] = (char)0x80; break;
         case 'v': out_str[out_len++] = (char)0x81; break;
         case '|': out_str[out_len++] = (char)0x82; break;
         case ':': out_str[out_len++] = (char)0x83; break;
         case '^': out_str[out_len++] = (char)0x84; break;
         case 'o': out_str[out_len++] = (char)0x85; break;
         case ',': out_str[out_len++] = (char)0x86; break;
         case '.': out_str[out_len++] = (char)0x87; break;
         case '`': out_str[out_len++] = (char)0x88; break;
         case '~': out_str[out_len++] = (char)0x89; break;
         }
      }else
         out_str[out_len++] = c;
   }
   out_str[out_len] = 0;

   return out_len;
}

//----------------------------

C_text *C_poly_text::CreateText(const void *tp1, dword color, float x, float y, float size,
   dword justify, bool wide, float fmt_rect[4], float *xr, float *yr, int *num_lines, int z,
   float italic, float in_scale_x, float in_scale_y){

   if(wide){
      C_wide_text *wt = new C_wide_text((const wchar_t*)tp1, color, x, y, size, justify, driver, fmt_rect,
         xr, yr, num_lines);
      C_text *ret = new C_text(wt, this);
      wt->Release();
      texts.push_back(ret);
      ret->color = color;
      ret->z = z;
      resort = true;
      return ret;
   }

   if(!IsInit())
      return NULL;

   const char *in_str = (const char*)tp1;
   int i;
   int len = strlen(in_str);

   if(num_lines) *num_lines = 0;

   char *cpa = new char[len*2+1];

   const S_national_table *ntab = ntab_1;
   switch(game_configuration.language){
   case 7:
   case 8:
   case 9:
      ntab = ntab_2;
      break;
   }

   len = FormatString(in_str, cpa, len, (justify&TEXT_JUSTIFY_CONVERT_EOL), ntab);

   C_sprite_group *pg = new C_sprite_group;
   pg->pos.x = x;
   pg->pos.y = y;

   short *spct = new short[(len+1)*2];    //x, y

   BuildSpaceTable(cpa, spct, (len+1)*2);
   float y_scale = size / font_img.GetRects()[0].isy * in_scale_y;

   if(xr) *xr = x;
   if(yr) *yr = y;
   float x_clip_delta = 0.0f;

   float scrn_sx = igraph->Scrn_sx();
   float r_pixel_size_x = 1.0f / scrn_sx;

   const char *curr_char = cpa;
   short *spct1 = spct;
   while(true){
      if(num_lines)
         ++(*num_lines);
                              //get width of line
      const char *end_of_line = strchr(curr_char, '\n');
      bool last_line = !end_of_line;
      if(last_line){
         end_of_line = curr_char + strlen(curr_char);
      }
      int line_len = end_of_line - curr_char;

      float line_width = (spct1[line_len*2] - spct1[0]) * y_scale;
      float x_base = x;
      
      switch(justify&TEXT_JUSTIFY_MASK){
      case TEXT_JUSTIFY_CENTER:
         x_base -= line_width/2;
         if(xr) *xr -= line_width/2;
         break;
      case TEXT_JUSTIFY_RIGHT:
         x_base -= line_width;
         if(xr) *xr -= line_width;
         break;
      }

                              //format text to fit in rectanlge (optional)
      if(fmt_rect){
         if(x_base<fmt_rect[0]){
            if(xr) *xr -= (x_base - fmt_rect[0]);
            x_base = fmt_rect[0];
         }
         float fmt_width = fmt_rect[2] - fmt_rect[0];
         if(line_width > fmt_width){
                              //remove some words from end
            int last_beg = line_len;
            while(true){
               for(i=last_beg; i--; ){
                  if(curr_char[i]==' ')
                     break;
               }
               if(i==-1){
                  i = last_beg;
                  break;
               }
               line_width = (spct1[i*2]-spct1[0]) * y_scale;
               if(line_width <= fmt_width)
                  break;
               last_beg = i;
            };
            last_line = (line_len==i);
            line_len = i;
         }
      }

      if(justify&TEXT_JUSTIFY_CLIP_TO_SCREEN){
         float new_x_base = x_base;
                              //clip onto screen
         if((x_base+line_width) > 1.0f)
            new_x_base = 1.0f-line_width;
         new_x_base = Max(0.0f, new_x_base);
         if(fabs(x_clip_delta)<fabs(new_x_base-x_base))
            x_clip_delta = new_x_base-x_base;
         x_base = new_x_base;
      }
      float scrn_y = y * scrn_sx;

      float spct_base = spct1[0];
      for(i=0; i<line_len; i++, ++curr_char, spct1+=2){
         int indx = ((byte)*curr_char) - ' ';
         if(indx <= 0)
            continue;
         const C_sprite_image::S_spr_rect &rc = font_img.GetRects()[indx];
         float fx = x_base+(float)(spct1[0]-spct_base) * y_scale * in_scale_x;
         float fy = y - ((float)spct1[1]*y_scale);
         fx -= fmod(fx, r_pixel_size_x);
         fy -= fmod(fy, r_pixel_size_x);

         LPC_poly_sprite spr = new C_poly_sprite(fx, fy, rc.isy*y_scale,
            &font_img, indx, z, in_scale_x / in_scale_y);

         spr->SetColor(color);
         if(italic!=0.0f)
            spr->Shear(italic, scrn_y);

         pg->AddSprite(spr);
         spr->Release();
      }
      if(last_line)
         break;
                           //skip eol
      ++curr_char;
      spct1 += 2;

      y += size;
   }

   if(xr) *xr += x_clip_delta;

   delete[] spct;
   delete[] cpa;

   LPC_text txt = new C_text(pg, this);
   txt->color = color;
   texts.push_back(txt);
   texts.back()->z = z;

   pg->Release();
   txt->CreateVisual();

   resort = true;

   return txt;
}

//----------------------------

bool C_poly_text::RemoveText(C_text *tp){

   for(int i=texts.size(); i--; ){
      if(texts[i]==tp){
         texts[i] = texts.back(); texts.pop_back();
         resort = true;
         return true;
      }
   }
   assert(0);              //text not found???
   return false;
}

//----------------------------

void C_poly_text::Render(){

   if(resort){
      sort(texts.begin(), texts.end(), TextLess);
      resort = false;
   }
   bool use_zb = driver->GetState(RS_USEZB);
   if(use_zb) driver->SetState(RS_USEZB, false);

   for(int i=0; i<texts.size(); i++){
      LPC_text tp = texts[i];
      tp->Render();
   }
   if(use_zb) driver->SetState(RS_USEZB, true);
}

//----------------------------
//----------------------------

