#include "all.h"
#include "resource.h"
#include <rules.h>
#include <win_reg.h>
#include <c_string.hpp>
#include <vector>
#include <algorithm>
#include <process.h>
#include <c_cache.h>
#include <iostream>
#include <insanity\texts.h>

using namespace std;

#define REG_BASE "Software\\Lonely Cat Games\\Hidden and Dangerous Deluxe\\"
static const char REGDIR[] = REG_BASE"1.50";
static const char REGKEY_CONFIG[] = REG_BASE"Config";

static const char APP_NAME[] = "Insanity 3D setup";

//#define ALLOW_FREQUENCY

C_all_texts all_txt;

//----------------------------

#ifndef NDEBUG
void D3D_Fatal(const char *text, dword hr);
#define CHECK_D3D_RESULT(text, hr) if(FAILED(hr)) D3D_Fatal(text, hr);
#else
#define D3D_Fatal(a, b)
#define CHECK_D3D_RESULT(text, hr)
#endif

//----------------------------

#ifndef NDEBUG
static void D3D_Fatal(const char *text, dword hr){

   char *msgp;
   switch(hr){

   case D3D_OK: msgp = "No error occurred."; break;
   case D3DERR_CONFLICTINGRENDERSTATE: msgp = "The currently set render states cannot be used together."; break;
   case D3DERR_CONFLICTINGTEXTUREFILTER: msgp = "The current texture filters cannot be used together."; break;
   case D3DERR_CONFLICTINGTEXTUREPALETTE: msgp = "The current textures cannot be used simultaneously. This generally occurs when a multitexture device requires that all palletized textures simultaneously enabled also share the same palette."; break;
   case D3DERR_DEVICELOST: msgp = "The device is lost and cannot be restored at the current time, so rendering is not possible."; break;
   case D3DERR_DEVICENOTRESET: msgp = "The device cannot be reset."; break;
   case D3DERR_DRIVERINTERNALERROR: msgp = "Internal driver error."; break;
   case D3DERR_INVALIDCALL: msgp = "The method call is invalid. For example, a method's parameter may have an invalid value."; break;
   case D3DERR_INVALIDDEVICE: msgp = "The requested device type is not valid."; break;
   case D3DERR_MOREDATA: msgp = "There is more data available than the specified buffer size can hold."; break;
   case D3DERR_NOTAVAILABLE: msgp = "This device does not support the queried technique."; break;
   case D3DERR_NOTFOUND: msgp = "The requested item was not found."; break;
   case D3DERR_OUTOFVIDEOMEMORY: msgp = "Direct3D does not have enough display memory to perform the operation."; break;
   case D3DERR_TOOMANYOPERATIONS: msgp = "The application is requesting more texture-filtering operations than the device supports."; break;
   case D3DERR_UNSUPPORTEDALPHAARG: msgp = "The device does not support a specified texture-blending argument for the alpha channel."; break;
   case D3DERR_UNSUPPORTEDALPHAOPERATION: msgp = "The device does not support a specified texture-blending operation for the alpha channel."; break;
   case D3DERR_UNSUPPORTEDCOLORARG: msgp = "The device does not support a specified texture-blending argument for color values."; break;
   case D3DERR_UNSUPPORTEDCOLOROPERATION: msgp = "The device does not support a specified texture-blending operation for color values."; break;
   case D3DERR_UNSUPPORTEDFACTORVALUE: msgp = "The device does not support the specified texture factor value."; break;
   case D3DERR_UNSUPPORTEDTEXTUREFILTER: msgp = "The device does not support the specified texture filter."; break;
   case D3DERR_WRONGTEXTUREFORMAT: msgp = "The pixel format of the texture surface is not valid."; break;
   case E_FAIL: msgp = "An undetermined error occurred inside the Direct3D subsystem."; break;
   case E_INVALIDARG: msgp = "An invalid parameter was passed to the returning function."; break;
   //case E_INVALIDCALL: msgp = "The method call is invalid. For example, a method's parameter may have an invalid value."; break;
   case E_OUTOFMEMORY: msgp = "Direct3D could not allocate sufficient memory to complete the call."; break;
   //default: DD_Fatal(text, hr); return;
   }

   char buf[16];
   itoa(hr, buf, 16);

   MessageBox(NULL,
      (C_string)text + (char const*)": " + (char const*)msgp +
      (char const*)" 0x" + (char const*)buf,
      "Fatal Error", MB_OK);
}
#endif

//----------------------------

static bool GetInstallationInfo(C_string &language){

   int i = RegkeyOpen(REGDIR, E_REGKEY_LOCAL_MACHINE);
   if(i==-1)
      return false;

   char buf[512];
   if(RegkeyRdata(i, "Language", (byte*)buf, sizeof(buf))==-1)
      buf[0] = 0;
   RegkeyClose(i);
   language = buf;

   return true;
}

//----------------------------

static void InitDlg(HWND hwnd){
                              //center dlg and flip to GDI
   RECT rc;
   GetWindowRect(hwnd, &rc);
   int sx = GetSystemMetrics(SM_CXSCREEN);
   int sy = GetSystemMetrics(SM_CYSCREEN);
   SetWindowPos(hwnd, NULL,
      (sx-(rc.right-rc.left))/2,
      (sy-(rc.bottom-rc.top))/2,
      0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

//----------------------------

struct S_config{
   bool use_wbuffer, tripple, fullscreen, tx_compress;

   dword mode_x, mode_y;
   dword bits_per_pixel;
   int frequency;
   int adapter_identifier;
   dword device_type;
   bool use_dbase;
   bool use_sounds;
   enum E_FILTER{
      FILTER_LINEAR,
      FILTER_ANISO,
   } filtering;
} config = {
   true,                      //use_wbuffer
   true,                      //tripple
   true,                      //fullscreen
   false,                     //compressed textures
   800, 600, 16,              //resolution
   0,                         //frequency
   0,                         //adapter id
   D3DDEVTYPE_HAL,
   true,                      //use dbase
   true,                      //use sounds
   S_config::FILTER_ANISO
};

static bool first_adapter_init = true, first_dev_init = true;

#ifdef NDEBUG
static bool all_features = false;
static int limit_mode_x = 512, limit_mode_y = 384;
#else
static bool all_features = true;
static int limit_mode_x = 0, limit_mode_y = 0;
#endif

//----------------------------

static const char *key_name[] = {
                              //0
   "Display width",
   "Display height",
   "Display bitdepth",
   "Display frequency",
                              //4
   "Fullscreen",
   "TrippleBuf",
   "Adapter ID",
   "Device type",
                              //8
   "Database cache",
   "Sounds",
   "W-buffer",
   "Compressed textures",
                              //12
   "Filtering",
};

//----------------------------

static void ReadConfig(){

   int i = RegkeyOpen(REGKEY_CONFIG, E_REGKEY_LOCAL_MACHINE);
   if(i!=-1){
      RegkeyRdata(i, key_name[0], (byte*)&config.mode_x, sizeof(config.mode_x));
      RegkeyRdata(i, key_name[1], (byte*)&config.mode_y, sizeof(config.mode_y));
      RegkeyRdata(i, key_name[2], (byte*)&config.bits_per_pixel, sizeof(config.bits_per_pixel));
      RegkeyRdata(i, key_name[3], (byte*)&config.frequency, sizeof(config.frequency));

      RegkeyRdata(i, key_name[4], (byte*)&config.fullscreen, sizeof(config.fullscreen));
      RegkeyRdata(i, key_name[5], (byte*)&config.tripple, sizeof(config.tripple));
      RegkeyRdata(i, key_name[6], (byte*)&config.adapter_identifier, sizeof(config.adapter_identifier));
      RegkeyRdata(i, key_name[7], (byte*)&config.device_type, sizeof(config.device_type));

      RegkeyRdata(i, key_name[8], (byte*)&config.use_dbase, sizeof(config.use_dbase));
      RegkeyRdata(i, key_name[9], (byte*)&config.use_sounds, sizeof(config.use_sounds));
      //RegkeyRdata(i, key_name[10], (byte*)&config.use_wbuffer, sizeof(config.use_wbuffer));
      RegkeyRdata(i, key_name[11], (byte*)&config.tx_compress, sizeof(config.tx_compress));

      RegkeyRdata(i, key_name[12], (byte*)&config.filtering, sizeof(config.filtering));

      RegkeyClose(i);

      first_dev_init = false;
      first_adapter_init = false;

      if(!all_features){
                              //setup some features hard way
         config.use_dbase = true;
      }
   }
};

//----------------------------

static void WriteConfig(){

   int i = RegkeyCreate(REGKEY_CONFIG, E_REGKEY_LOCAL_MACHINE);
   if(i!=-1){
      RegkeyWdword(i, key_name[0], config.mode_x);
      RegkeyWdword(i, key_name[1], config.mode_y);
      RegkeyWdword(i, key_name[2], config.bits_per_pixel);
      RegkeyWdword(i, key_name[3], config.frequency);

      RegkeyWdword(i, key_name[4], config.fullscreen);
      RegkeyWdword(i, key_name[5], config.tripple);
      RegkeyWdword(i, key_name[6], config.adapter_identifier);
      RegkeyWdword(i, key_name[7], config.device_type);

      RegkeyWdword(i, key_name[8], config.use_dbase);
      RegkeyWdword(i, key_name[9], config.use_sounds);
      //RegkeyWdword(i, key_name[10], config.use_wbuffer);
      RegkeyWdword(i, key_name[11], config.tx_compress);

      RegkeyWdword(i, key_name[12], config.filtering);

      RegkeyClose(i);
   }
};

//----------------------------

class C_DisplayMode{
public:
   word width, height;
#ifdef ALLOW_FREQUENCY
   int freq;
#endif

   vector<int> bit_depths;

   C_DisplayMode(){}
   C_DisplayMode(word w, word h
#ifdef ALLOW_FREQUENCY
      , int f
#endif
      ):
      width(w), height(h)
#ifdef ALLOW_FREQUENCY
         , freq(f)
#endif
   {}

                              //func for sorting display modes
   bool operator <(const C_DisplayMode &dm) const{

      if((width*height) < (dm.width*dm.height))
         return true;
      if((width*height) > (dm.width*dm.height))
         return false;

#ifdef ALLOW_FREQUENCY
      if(freq < dm.freq)
         return true;
#endif
      return false;
   }

   void AddBitDepth(int d){
      bit_depths.push_back(d);
      sort(bit_depths.begin(), bit_depths.end());
   }

   bool HasBitDepth(int d) const{

      for(int i=bit_depths.size(); i--; ){
         if(bit_depths[i]==d)
            return true;
      }
      return false;
   }
};

//----------------------------
//----------------------------

class C_D3DDevice{
   C_string name;
public:
   D3DDEVTYPE type;

   C_D3DDevice(){}
   C_D3DDevice(D3DDEVTYPE dt, const char *n):
      name(n),
      type(dt)
   {}
   C_D3DDevice& operator =(const C_D3DDevice& d){
      name = d.name;
      type = d.type;
      return *this;
   }
   C_D3DDevice(const C_D3DDevice &ref){
      *this = ref;
   }

   const C_string &Name(){ return name; }

   int GetHardwareMode(){
      switch(type){
      case D3DDEVTYPE_HAL:
         return 2;
      case D3DDEVTYPE_SW:
         return 1;
      default:
         return 0;
      }
   }

   dword GetRenderDepths(){

      /*
      dword dw = 0;
      D3DDEVICEDESC7 &curr = (hal.dwDeviceRenderBitDepth&(DDBD_8|DDBD_16|DDBD_24|DDBD_32)) ?
         hal : hel;
      if(curr.dwDeviceRenderBitDepth&DDBD_8)  dw |= 1;
      if(curr.dwDeviceRenderBitDepth&DDBD_16) dw |= 2;
      if(curr.dwDeviceRenderBitDepth&DDBD_24) dw |= 4;
      if(curr.dwDeviceRenderBitDepth&DDBD_32) dw |= 8;
      return dw;
      */
      return 0xf;
   }
};

//----------------------------
//----------------------------

class C_Adapter{
   C_string name;
   bool render_in_window;

   vector<C_DisplayMode> display_modes;
   vector<C_D3DDevice> d3d_devices;
public:
   int adapter_identifier;

   C_Adapter(){}

   C_Adapter(const char *name, IDirect3D8 *d3d8, int adapter_index);

   C_Adapter& operator =(const C_Adapter& d){
      name = d.name;
      render_in_window = d.render_in_window;
      display_modes = d.display_modes;
      d3d_devices = d.d3d_devices;
      adapter_identifier = d.adapter_identifier;
      return *this;
   }
   C_Adapter(const C_Adapter &ref){
      *this = ref;
   }

   const C_string &Name() const{ return name; }

   inline C_DisplayMode* GetModes(){ return display_modes.begin(); }

   inline int NumModes() const{ return display_modes.size(); }

   C_D3DDevice* Get3DDevices(){ return d3d_devices.begin(); }

   inline int Num3DDevices() const{ return d3d_devices.size(); }

   inline bool Match(int ai) const{ return adapter_identifier == ai; }

   inline bool IsWindowed() const{ return render_in_window; }
};

//----------------------------

static const char *device_names[] = {
   "Hardware acceleration",
   "Reference rasterizer",
   "Software rasterizer",
};

//----------------------------

C_Adapter::C_Adapter(const char *n,
   IDirect3D8 *d3d8, int adapter_index):
   adapter_identifier(adapter_index),
   name(n),
   render_in_window(true)
{
   int i;

   for(i=D3DDEVTYPE_HAL; i<=D3DDEVTYPE_SW; i++){
      D3DCAPS8 dc;
      HRESULT hr = d3d8->GetDeviceCaps(adapter_index, (D3DDEVTYPE)i, &dc);
      //CHECK_D3D_RESULT("GetDeviceCaps", hr);
      if(SUCCEEDED(hr)){
         d3d_devices.push_back(C_D3DDevice((D3DDEVTYPE)i, device_names[i - D3DDEVTYPE_HAL]));

         if(i==D3DDEVTYPE_HAL){
            render_in_window = (dc.Caps2&D3DCAPS2_CANRENDERWINDOWED);
         }
      }
   }

                              //enum display modes of this device
   for(i=0; i<d3d8->GetAdapterModeCount(adapter_index); i++){
      D3DDISPLAYMODE dd;
      d3d8->EnumAdapterModes(adapter_index, i, &dd);

      dword bpp = 0;
                              //allow only rgb modes
      switch(dd.Format){
      case D3DFMT_R5G6B5:
      case D3DFMT_X1R5G5B5:
         bpp = 16;
         break;
      case D3DFMT_R8G8B8:
         bpp = 24;
         break;
      case D3DFMT_X8R8G8B8:
         bpp = 32;
         break;
      }
                              //check if resolution acceptable
      if(dd.Width>=limit_mode_x && dd.Height>=limit_mode_y && bpp >= 16){
#ifndef ALLOW_FREQUENCY
                              //don't include same mode more times
         for(int j=display_modes.size(); j--; ){
            if(display_modes[j].width==dd.Width && display_modes[j].height==dd.Height){
               if(!display_modes[j].HasBitDepth(bpp))
                  display_modes[j].AddBitDepth(bpp);
               break;
            }
         }
         if(j!=-1)
            continue;
         display_modes.push_back(C_DisplayMode(dd.Width, dd.Height));
         display_modes.back().AddBitDepth(bpp);
#else
         display_modes.push_back(C_DisplayMode(dd.Width, dd.Height, bpp, dd.RefreshRate));
#endif
      }
   }
                              //add current mode, if not yet
   for(int j=display_modes.size(); j--; ){
      if(display_modes[j].width==config.mode_x && display_modes[j].height==config.mode_y)
         break;
   }
   if(j==-1)
      display_modes.push_back(C_DisplayMode(config.mode_x, config.mode_y));

   sort(display_modes.begin(), display_modes.end());
}

//----------------------------

class C_Enum{
   vector<C_Adapter*> adapters;
public:
   HRESULT EnumAll(C_string&);

   void Clear(){
      for(int i=0; i<adapters.size(); i++)
         delete adapters[i];
      adapters.clear();
   }

   inline int Size() const{
      return adapters.size();
   }

   inline C_Adapter& operator [](int i) const{
      return *adapters[i];
   }
};

//----------------------------

HRESULT C_Enum::EnumAll(C_string &err){

   {
                              //Load D3D8 library
      HINSTANCE hi_d3d8 = LoadLibrary("d3d8.dll");
      if(!hi_d3d8){
         err = "unable to load d3d8.dll";
         goto fail;
      }

      FARPROC fp = GetProcAddress(hi_d3d8, "Direct3DCreate8");
      if(!fp){
         FreeLibrary(hi_d3d8);
         err = "unable to find entry point 'Direct3DCreate8' in d3d8.dll";
         goto fail;
      }
      
      typedef IDirect3D8 *WINAPI t_Direct3DCreate8(UINT SDKVersion);
      t_Direct3DCreate8 *create_proc = (t_Direct3DCreate8*)fp;
                              //fake D3D it's compiled with version 8.0 (until 8.1 is officially out)
      IDirect3D8 *d3d8 = (*create_proc)(D3D_SDK_VERSION);
      //IDirect3D8 *d3d8 = (*create_proc)(120);
      if(!d3d8){
         FreeLibrary(hi_d3d8);
         err = "unable to create Direct3D object";
         goto fail;
      }

      int num_adapters = d3d8->GetAdapterCount();
      if(!num_adapters){
         d3d8->Release();
         FreeLibrary(hi_d3d8);
         err = "Direct3D returned zero adapter count";
         goto fail;
      }

                                 //enum all adapters
      for(int i=0; i<num_adapters; i++){

         D3DADAPTER_IDENTIFIER8 ai;
         d3d8->GetAdapterIdentifier(i, 0, &ai);

         adapters.push_back(new C_Adapter(/*&ai.DeviceIdentifier, ai.Driver, */ai.Description,
            d3d8, i));
      }

      d3d8->Release();

      FreeLibrary(hi_d3d8);
   }

   return D3D_OK;

fail:
   return -1;
}

//----------------------------

static int SelectAdapter(HWND hwnd, C_Enum* dde){

   for(int i=0; i<dde->Size(); i++){
      if(first_adapter_init){
         if((*dde)[i].adapter_identifier==0){
            config.adapter_identifier = (*dde)[i].adapter_identifier;
            break;
         }
      }else
         if((*dde)[i].Match(config.device_type)) break;
   }
   first_adapter_init = false;
   if(i==dde->Size()) i = 0;
   SendDlgItemMessage(hwnd, IDC_COMBO_DEVICES, CB_SETCURSEL, i, 0);
   return i;
}

//----------------------------

static void SetModes(HWND hwnd, C_Adapter *dp){
   
   C_DisplayMode *modes = dp->GetModes();
   SendDlgItemMessage(hwnd, IDC_COMBO_MODES, CB_RESETCONTENT, 0, 0);

   for(int i=0; i<dp->NumModes(); i++){
      C_string str;
      {
#ifdef ALLOW_FREQUENCY
         str = C_fstring("%ix%i %i bit (%i Hz)", modes[i].width,
            modes[i].height, modes[i].bits_per_pixel, modes[i].freq);
#else
         str = C_fstring("%i x %i", modes[i].width, modes[i].height);
#endif
         SendDlgItemMessage(hwnd, IDC_COMBO_MODES, CB_ADDSTRING,
            0, (LPARAM)(const char*)str);
      }
   }
}

//----------------------------

static void SelectMode(C_Adapter *dp, int res_i){

   const C_DisplayMode &mode = dp->GetModes()[res_i];
   config.mode_x = mode.width;
   config.mode_y = mode.height;
}

//----------------------------

static void SetDepths(HWND hwnd, const C_DisplayMode &mode, dword flags){

   SendDlgItemMessage(hwnd, IDC_COMBO_DEPTHS, CB_RESETCONTENT, 0, 0);

   for(int i=0, j=0; i<mode.bit_depths.size(); i++){
      int bpp = mode.bit_depths[i];
      if(flags&(1<<((bpp/8)-1))){
         C_fstring str("%i bit", bpp);
         SendDlgItemMessage(hwnd, IDC_COMBO_DEPTHS, CB_ADDSTRING,
            0, (LPARAM)(const char*)str);
         if(config.bits_per_pixel==bpp)
            SendDlgItemMessage(hwnd, IDC_COMBO_DEPTHS, CB_SETCURSEL, j, 0);
         ++j;
      }
   }
}

//----------------------------

static void SelectDepth(int res_i, int dep_i, C_Adapter *dp, dword flags){

   C_DisplayMode &mode = dp->GetModes()[res_i];
   for(int j=0; j<mode.bit_depths.size(); j++){
      int bpp = mode.bit_depths[j];
      if(flags&(1<<((bpp/8)-1))){
         if(!dep_i--){
            config.bits_per_pixel = bpp;
            return;
         }
      }
   }
}

//----------------------------

static void SelectMode(HWND hwnd, C_Adapter *dp, C_Enum* dde, dword flags){

   int best_mode = -1;
   const C_DisplayMode *modes = dp->GetModes();
   for(int i=0, j=0; i<dp->NumModes(); i++){
      const C_DisplayMode *mode = &dp->GetModes()[i];
      if(config.mode_x==mode->width &&
         config.mode_y==mode->height){
         if(best_mode==-1)
            best_mode = j;
      }
      ++j;
   }
   if(i==dp->NumModes())
      j=best_mode==-1 ? 0 : best_mode;
   SelectMode(dp, j);

   SendDlgItemMessage(hwnd, IDC_COMBO_MODES, CB_SETCURSEL, j, 0);
   SetDepths(hwnd, modes[best_mode], flags);
}

//----------------------------

static void SetDevices(HWND hwnd, C_Adapter *dp){

   C_D3DDevice *devs=dp->Get3DDevices();
   SendDlgItemMessage(hwnd, IDC_COMBO_3DDEVICES, CB_RESETCONTENT, 0, 0);

   for(int i=0; i<dp->Num3DDevices(); i++){
      char buf[64];
      SendDlgItemMessage(hwnd, IDC_COMBO_3DDEVICES, CB_ADDSTRING,
         0, (LPARAM)(const char*)devs[i].Name());
   }
   EnableWindow(GetDlgItem(hwnd, IDC_CHECK_FULLSCREEN), dp->IsWindowed());
   CheckDlgButton(hwnd, IDC_CHECK_FULLSCREEN, dp->IsWindowed() ? config.fullscreen : true);

   EnableWindow(GetDlgItem(hwnd, IDC_CHECK_TRIPPLEBUF), (!dp->IsWindowed() || config.fullscreen));
   EnableWindow(GetDlgItem(hwnd, IDC_COMBO_DEPTHS), (!dp->IsWindowed() || config.fullscreen));
}

//----------------------------

static int SelectDevice(HWND hwnd, C_Adapter *dp, C_Enum* dde){

   int best_hw_mode = 0;
   int best_i = dp->Num3DDevices();
   for(int i=0; i<dp->Num3DDevices(); i++){
      C_D3DDevice *dev = &dp->Get3DDevices()[i];
      if(first_dev_init){
         int hw_mode = dev->GetHardwareMode();
         if(hw_mode && hw_mode>best_hw_mode){
            best_hw_mode = hw_mode;
            config.device_type = dev->type;
            best_i = i;
         }
      }else
         if(dev->type == config.device_type) break;
   }
   if(first_dev_init)
      i = best_i;
   first_dev_init = false;
                              //couldn't happen
   if(i==dp->Num3DDevices()){
      if(!i) i = -1;
      else i = 0;
   }
   SendDlgItemMessage(hwnd, IDC_COMBO_3DDEVICES, CB_SETCURSEL, i, 0);
   return i;
}

//----------------------------

static void SaveChanges(HWND hwnd, C_Enum *dde){

   int i, j;

   //config.use_wbuffer = IsDlgButtonChecked(hwnd, IDC_CHECK_WBUF);
   config.tx_compress = IsDlgButtonChecked(hwnd, IDC_CHECK_TXT_COMPR);
   config.filtering = IsDlgButtonChecked(hwnd, IDC_RADIO_ANISO) ? S_config::FILTER_ANISO : S_config::FILTER_LINEAR;

   config.tripple    = IsDlgButtonChecked(hwnd, IDC_CHECK_TRIPPLEBUF);
   config.fullscreen = IsDlgButtonChecked(hwnd, IDC_CHECK_FULLSCREEN);
   config.use_dbase  = IsDlgButtonChecked(hwnd, IDC_CHECK_DBASE);
   config.use_sounds = IsDlgButtonChecked(hwnd, IDC_CHECK_SOUNDS);

   i = SendDlgItemMessage(hwnd, IDC_COMBO_DEVICES, CB_GETCURSEL, 0, 0);
   C_Adapter *dp = &(*dde)[i];
                              //save 3d device
   j = SendDlgItemMessage(hwnd, IDC_COMBO_3DDEVICES, CB_GETCURSEL, 0, 0);
   if(j!=-1){
      C_D3DDevice *dev = &dp->Get3DDevices()[j];
      config.device_type = dev->type;

      WriteConfig();
   }
}

//----------------------------

                              //items which have context help defined
static int context_help_info[] = {
   IDC_CHECK_WBUF,
   0                          //last
};

//----------------------------

static const char font_name[] = "MS Gothic";

static HFONT japCreateFontJ(int charx, int chary){

   static HFONT hfont = CreateFont(
      chary,                  //logical height of font
      charx/2,
      0,                      //angle of escapement
      0,                      //base-line orientation angle
      FW_NORMAL,              //font weight
      //FW_THIN,              //font weight
      FALSE,                  //italic attribute flag
      FALSE,                  //underline attribute flag
      FALSE,                  //strikeout attribute flag
      SHIFTJIS_CHARSET,       //character set identifier
      OUT_DEFAULT_PRECIS,     //output precision
      CLIP_DEFAULT_PRECIS,    //clipping precision
      DEFAULT_QUALITY,        //output quality
      FIXED_PITCH | FF_DONTCARE, //pitch and family
      (const char*)font_name);//pointer to typeface name string);

   return hfont;
}

//----------------------------

static BOOL CALLBACK DlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

   int i, j;

   switch(uMsg){
   case WM_INITDIALOG:
      {
         InitDlg(hwnd);
         C_Enum *dde = (C_Enum*)lParam;
         SetWindowLong(hwnd, GWL_USERDATA, lParam);

                              //init controls' texts
         struct S_ctrl_text{
            int ctrl_id;
            int text_id;
         } ctrl_texts[] = {
            IDC_DISPLAY, 1,
            IDC_3DDEV, 2,
            IDC_TXT_RES, 3,
            IDC_TXT_DEPTH, 4,
            IDOK, 5,
            IDCANCEL, 6,
            ID_APPLY, 7,
            IDC_TXT_SETTINGS, 8,
            IDC_CHECK_FULLSCREEN, 9,
            IDC_CHECK_TRIPPLEBUF, 10,
            IDC_CHECK_SOUNDS, 11,
            IDC_CHECK_TXT_COMPR, 12,
            IDC_CHECK_DBASE, 13,
            IDC_TXT_TXFILTER, 14,
            IDC_RADIO_LINEAR, 15,
            IDC_RADIO_ANISO, 16,
            0
         };
         for(int i=0; ctrl_texts[i].ctrl_id; i++){
            int tid = ctrl_texts[i].text_id;
            bool wide = all_txt.IsWide(tid);
            int id = ctrl_texts[i].ctrl_id;
            if(!wide){
               SetDlgItemText(hwnd, id, all_txt[tid]);
            }else{
               HFONT hfnt = japCreateFontJ(12, 12);
               SendDlgItemMessage(hwnd, id, WM_SETFONT, (WPARAM)hfnt, true);
               SetDlgItemTextW(hwnd, id, (wchar_t*)(const char*)all_txt[tid]);
            }
         }

         CheckDlgButton(hwnd, IDC_CHECK_TXT_COMPR, config.tx_compress);
         CheckDlgButton(hwnd, IDC_RADIO_LINEAR, (config.filtering==S_config::FILTER_LINEAR));
         CheckDlgButton(hwnd, IDC_RADIO_ANISO, (config.filtering==S_config::FILTER_ANISO));

         CheckDlgButton(hwnd, IDC_CHECK_TRIPPLEBUF, config.tripple);
         CheckDlgButton(hwnd, IDC_CHECK_FULLSCREEN, config.fullscreen);
         CheckDlgButton(hwnd, IDC_CHECK_DBASE, config.use_dbase);
         CheckDlgButton(hwnd, IDC_CHECK_SOUNDS, config.use_sounds);
         EnableWindow(GetDlgItem(hwnd, ID_APPLY), first_adapter_init);

         if(!all_features){
            EnableWindow(GetDlgItem(hwnd, IDC_CHECK_DBASE), false);
         }

                              //feed all devices
         for(i=0; i<dde->Size(); i++){
            C_Adapter *ddp=&(*dde)[i];
            SendDlgItemMessage(hwnd, IDC_COMBO_DEVICES, CB_ADDSTRING,
               0, (LPARAM)(const char*)ddp->Name());
         }
         i = SelectAdapter(hwnd, dde);
         C_Adapter *dp = &(*dde)[i];

         SetDevices(hwnd, dp);
         i = SelectDevice(hwnd, dp, dde);
         if(i != -1){
            dword flags = dp->Get3DDevices()[i].GetRenderDepths();
            SetModes(hwnd, dp);
            SelectMode(hwnd, dp, dde, flags);
         }

         ShowWindow(hwnd, SW_SHOW);
         UpdateWindow(hwnd);
      }
      return 1;

   case WM_COMMAND:
      switch(LOWORD(wParam)){
      case ID_APPLY:
         {
            C_Enum *dde = (C_Enum*)GetWindowLong(hwnd, GWL_USERDATA);
            SaveChanges(hwnd, dde);
            EnableWindow(GetDlgItem(hwnd, LOWORD(wParam)), 0);
            SetFocus(GetDlgItem(hwnd, IDOK));
         }
         break;

      case IDC_COMBO_3DDEVICES:
         switch(HIWORD(wParam)){
         case CBN_SELCHANGE:
            {
               C_Enum *dde = (C_Enum*)GetWindowLong(hwnd, GWL_USERDATA);
               i = SendDlgItemMessage(hwnd, IDC_COMBO_DEVICES, CB_GETCURSEL, 0, 0);
               C_Adapter *dp = &(*dde)[i];
               i = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
               dword flags = dp->Get3DDevices()[i].GetRenderDepths();
               C_D3DDevice *dev = &dp->Get3DDevices()[i];
               config.device_type = dev->type;
               SetModes(hwnd, dp);
               SelectMode(hwnd, dp, dde, flags);
               EnableWindow(GetDlgItem(hwnd, ID_APPLY), 1);
            }
            break;
         }
         break;

      case IDC_COMBO_MODES:
         switch(HIWORD(wParam)){
         case CBN_SELCHANGE:
            {
               C_Enum *dde = (C_Enum*)GetWindowLong(hwnd, GWL_USERDATA);
               i = SendDlgItemMessage(hwnd, IDC_COMBO_DEVICES, CB_GETCURSEL, 0, 0);
               C_Adapter *dp = &(*dde)[i];
               i = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
               dword flags = dp->Get3DDevices()[i].GetRenderDepths();
               SelectMode(dp, i);
               SelectMode(hwnd, dp, dde, flags);
               EnableWindow(GetDlgItem(hwnd, ID_APPLY), 1);
            }
            break;
         }
         break;

      case IDC_COMBO_DEPTHS:
         switch(HIWORD(wParam)){
         case CBN_SELCHANGE:
            {
               C_Enum *dde = (C_Enum*)GetWindowLong(hwnd, GWL_USERDATA);
               i = SendDlgItemMessage(hwnd, IDC_COMBO_DEVICES, CB_GETCURSEL, 0, 0);
               C_Adapter *dp = &(*dde)[i];
               i = SendDlgItemMessage(hwnd, IDC_COMBO_3DDEVICES, CB_GETCURSEL, 0, 0);
               dword flags = dp->Get3DDevices()[i].GetRenderDepths();
               int res_i = SendDlgItemMessage(hwnd, IDC_COMBO_MODES, CB_GETCURSEL, 0, 0);
               int dep_i = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
               SelectDepth(res_i, dep_i, dp, flags);
               EnableWindow(GetDlgItem(hwnd, ID_APPLY), 1);
            }
            break;
         }
         break;

      case IDC_CHECK_FULLSCREEN:
         {
            bool b = IsDlgButtonChecked(hwnd, IDC_CHECK_FULLSCREEN);
            EnableWindow(GetDlgItem(hwnd, IDC_CHECK_TRIPPLEBUF), b);
            EnableWindow(GetDlgItem(hwnd, IDC_COMBO_DEPTHS), b);

            EnableWindow(GetDlgItem(hwnd, ID_APPLY), true);
         }
         break;

      default:
         EnableWindow(GetDlgItem(hwnd, ID_APPLY), true);
         break;

      case IDOK:
         {
            C_Enum *dde = (C_Enum*)GetWindowLong(hwnd, GWL_USERDATA);
            SaveChanges(hwnd, dde);
            EndDialog(hwnd, 1);
         }
         break;

      case IDCANCEL:
         EndDialog(hwnd, 0);
         break;

      case IDC_COMBO_DEVICES:
         switch(HIWORD(wParam)){
         case CBN_SELCHANGE:
            {
               C_Enum *dde = (C_Enum*)GetWindowLong(hwnd, GWL_USERDATA);
               i = SendDlgItemMessage(hwnd, IDC_COMBO_DEVICES, CB_GETCURSEL, 0, 0);
               C_Adapter *dp = &(*dde)[i];
                              //save dd device
               config.adapter_identifier = dp->adapter_identifier;

               SetDevices(hwnd, dp);
               i = SelectDevice(hwnd, dp, dde);
               if(i != -1){
                  dword flags = dp->Get3DDevices()[i].GetRenderDepths();
                  SetModes(hwnd, dp);
                  SelectMode(hwnd, dp, dde, flags);
               }
               EnableWindow(GetDlgItem(hwnd, ID_APPLY), 1);
            }
            break;
         }
         break;
      }
      break;
   }

   return 0;
}

//----------------------------

int WinMain(HINSTANCE hInstance, HINSTANCE, LPSTR lpCmdLine1, int nCmdShow){

   int i, j;
   C_Enum dd_enum;

                              //command-line processing
   if(*lpCmdLine1=='a'){
                              //allow all modes
      limit_mode_x = 0;
      limit_mode_y = 0;
      all_features = true;
   }

                              //read localized texts
   C_string language;
   if(!GetInstallationInfo(language))
      language = "english";

   HRSRC hsrc;
   hsrc = FindResource(hInstance, language, "TEXTS");
   if(!hsrc)
      hsrc = FindResource(hInstance, "english", "TEXTS");
   bool txt_ok = false;
   if(hsrc){
      HGLOBAL hgl;
      hgl = LoadResource(hInstance, hsrc);
      if(hgl){
         void *vp = LockResource(hgl);
         if(vp){
            size_t sz = SizeofResource(hInstance, hsrc);
            C_cache is;
            if(is.open((byte**)&vp, &sz, CACHE_READ_MEM)){
               if(all_txt.AddFile(is))
                  txt_ok = true;
               is.close();
            }
         }
      }
   }
   if(!txt_ok){
      MessageBox(NULL, "Failed to open text file.", APP_NAME, MB_OK);
      return 1;
   }

   ReadConfig();

   int ret;
   C_string err;
   if(FAILED(dd_enum.EnumAll(err))){
      MessageBox(NULL,
         C_fstring("Unable to query DirectX 8.\nPlease install newer version of DirectX.\n\nError details:\n%s.", (const char*)err),
         APP_NAME, MB_OK);
      ret = 1;
   }else{
      ret = !DialogBoxParam(hInstance, "IDD_DIALOG_MAIN", NULL, (DLGPROC)DlgProc,
         (LPARAM)&dd_enum);
   }

   dd_enum.Clear();
   return ret;
}

//----------------------------
