#include "all.h"
#include "commandline.h"
#include "winstuff.h"
#include <insanity\controller.h>
#include "tables.h"
#include <isound2.h>
#include <advanced.h>
#include <win_reg.h>
#include <process.h>

//----------------------------

#define MODEL_CACHE_SIZE 48
#define SOUND_CACHE_SIZE 8

//----------------------------

extern LPC_controller controller;
extern const char *lang_names[];

//----------------------------

extern const char DATABASE_FILENAME[];
const char DATABASE_FILENAME[] = "dbase.bin";
#define DATABASE_SIZE (32*1024*1024)      //first number means MB

#define REG_BASE "Software\\Lonely Cat Games\\Hidden and Dangerous Deluxe\\"
static const char REGDIR[] = REG_BASE"1.50";
static const char REGKEY_CONFIG[] = REG_BASE"Config";
static const char REGKEY_CONTROLS[] = REG_BASE"Controls";

static const char INSTALL_TYPE_KEY[] = "Install type";
static const char INSTALL_DIR_KEY[] = "Install source dir";
static const char INSTALL_DEST_DIR_KEY[] = "Install target dir";
static const char INSTALL_LANG_KEY[] = "Language";

static const char INSERT_CD_TEXT[] = "Please insert the \"Hidden and Dangerous\" CD into drive %s.";
static const char NO_CD_TEXT[] = "The game was installed from CD-ROM drive %s.\nThis CD-ROM drive cannot be found. Please re-install from valid CD-ROM drive.\nThe game will now exit.";
static const char TEXT_BAD_INSTALL[] = "The game is not installed properly. Please re-install.";

static const char LOG_FILE_NAME[] = "h&d.log";

//----------------------------

static FILE *f_os_log;


//----------------------------

/*
                              //uncomment this to enable checking of multiple deleted blocks
#include <malloc.h>
#include <set>
static set<void*> mem_freed;  //freed memory blocks for multiple-deletion checkking
static bool is_inside;

void __cdecl operator delete(void *p){
   if(!p)
      return;
   if(is_inside){
      free(p);
      return;
   }
   is_inside = true;
   if(mem_freed.find(p) != mem_freed.end()){
      OsMessageBox("Mem freed twice.", "Fatal memory error!", WMB_OK);
      //exit(1);
      assert(0);
   }
                              //fill memory by garbage
   size_t sz = _msize(p);
   memset(p, 0xfa, sz);

   mem_freed.insert(p);
   is_inside = false;
}
/**/

//----------------------------

/*
#include <new.h>
#define MEM_CHECK_BOUNDARY 128

#undef new
void *__cdecl operator new(size_t sz){
   dword debug_alloc_size = sz + MEM_CHECK_BOUNDARY * 2 + sizeof(dword);
   byte *ptr = (byte*)malloc(debug_alloc_size);
   *((dword*&)ptr)++ = sz;
   memset(ptr, 0xcd, MEM_CHECK_BOUNDARY);
   ptr += MEM_CHECK_BOUNDARY;
   memset(ptr + sz, 0xce, MEM_CHECK_BOUNDARY);
   return ptr;
}
#define new DEBUG_CLIENTBLOCK

void __cdecl operator delete(void *p){
   if(!p)
      return;
   byte *orig_mem = (byte*)p - MEM_CHECK_BOUNDARY - sizeof(dword);
   dword orig_size = *(dword*)orig_mem;
   byte *base_1 = orig_mem + sizeof(dword);
   byte *base_2 = base_1 + MEM_CHECK_BOUNDARY + orig_size;
   for(int i=MEM_CHECK_BOUNDARY/sizeof(dword); i--; ){
      if(((dword*)base_1)[i] != 0xcdcdcdcd ||
         ((dword*)base_2)[i] != 0xcececece){

         char buf[256];
         sprintf(buf, "Mem corrupted: 0x%x, size %i.", p, orig_size);
         OsMessageBox(buf, "Total memory error!", WMB_OK);
         exit(1);
      }
   }
   free(orig_mem);
}
/**/

//----------------------------

#pragma pack(push, 1)

static char game_keys[] = {
   0, 0, 0, 0,
   0, 0, 0, 0,
   0, 0, 0, 0,
   0, 0, 0, 0,
   0, 0, 0, 0,
   0, 0, 0, 0,
   0, 0, 0, 0,
   0, 0, K_CAPSLOCK, K_TAB
};

#pragma pack(pop)

//----------------------------

inline int F2I(float f){ return *(dword*)&f; }
inline float I2F(int i){ return *(float*)&i; }

struct S_controller_item{
                              //fixed values
   int slot;
   C_controller::E_VALUE_MODE value_mode;
   struct{
      C_controller::E_INPUT_DEVICE input_device;
      dword data;
      //dword dw_range;
      float conversion;
   } input[2];
};

static const S_controller_item controller_init[] = {
                              //turn left
   //0, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_A, 0, C_controller::ID_JOY_AXIS, 0x80000000, .4f},
   0, C_controller::VM_FLOAT, {C_controller::ID_KEYBOARD, K_A, 1},//, C_controller::ID_JOY_AXIS, 0x00000000},
                              //move forward
   1, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_CURSORUP, 0},//, C_controller::ID_JOY_AXIS, 1, .4f},
                              //turn right
   //2, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_D, 0, C_controller::ID_JOY_AXIS, 0, .4f},
   2, C_controller::VM_FLOAT, {C_controller::ID_KEYBOARD, K_D, 1},//, C_controller::ID_JOY_AXIS, 0x80000000},
                              //back
   3, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_CURSORDOWN, 0},//, C_controller::ID_JOY_AXIS, 0x80000001, .4f},
                              //run mode
   4, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_SHIFT, 0},//, C_controller::ID_JOY_BUTTON, 0},
                              //strafe mode
   5, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_ALT, 0},//, C_controller::ID_JOY_BUTTON, 2},
                              //fire
   6, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_CTRL, 0, C_controller::ID_MOUSE_BUTTON, 0},
                              //fire 1
   //7, C_controller::VM_BOOL, {C_controller::ID_JOY_BUTTON, 0},
                              //left
   8, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_CURSORLEFT, 0},//, C_controller::ID_JOY_AXIS, 0x80000002, .4f},
                              //jump
   9, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_X, 0},//, C_controller::ID_JOY_BUTTON, 3},
                              //right
   10, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_CURSORRIGHT, 0},//, C_controller::ID_JOY_AXIS, 2, .4f},
                              //use inventory
   13, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_ENTER, 0},//, C_controller::ID_JOY_BUTTON, 4},
                              //up
   32, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_W, 0, C_controller::ID_MOUSE_AXIS, 2},
                              //down
   33, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_S, 0, C_controller::ID_MOUSE_AXIS, 0x80000002},
                              //use
   34, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_U},
                              //inv scroll left
   36, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_5B, 0},//, C_controller::ID_JOY_BUTTON, 5},
                              //inv scroll left
   37, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_5D, 0},//, C_controller::ID_JOY_BUTTON, 6},
                              //map mode
   38, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_SPACE},
                              //map floor up
   39, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_PAGEUP},
                              //map floor down
   40, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_PAGEDOWN},
                              //player switch
   41, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_TAB},
                              //cam zoom in
   42, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_3D, 0},
                              //cam zoom out
   43, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_2D, 0, C_controller::ID_MOUSE_BUTTON, 2},
                              //cam up
   44, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_5C},
                              //cam down
   45, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_BACKSPACE},
                              //center
   46, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_C},
                              //reload
   47, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_R, 0},//, C_controller::ID_JOY_BUTTON, 1},
                              //follow
   48, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_HOME},
                              //stop
   49, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_END},
                              //move
   50, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_PAGEUP},
                              //hey
   51, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_INS},
                              //hold fire
   52, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_PAGEDOWN},
                              //prg clr
   53, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_DEL},
                              //player switch
   54, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_1},
   55, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_2},
   56, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_3},
   57, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_4},
   57, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_4},
   63, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_Z, 0, C_controller::ID_MOUSE_BUTTON, 1},

                              //click in menus
   //64, C_controller::VM_BOOL, {C_controller::ID_MOUSE_BUTTON, 0, 0},
   //65, C_controller::VM_BOOL, {C_controller::ID_MOUSE_BUTTON, 1, 0},
   //66, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_ENTER},
                              //escape
   GKEY_ESCAPE, C_controller::VM_BOOL, {C_controller::ID_KEYBOARD, K_ESC},

   -1
};

//----------------------------

void SetupControllerDefaults(int slot){

   struct S_hlp{
      static void SetupSingleSlot(int slot_index){
         const S_controller_item &ci = controller_init[slot_index];
         for(int j=0; j<2; j++){
            controller->SetupSlotInput(ci.slot, j, ci.value_mode, ci.input[j].input_device,
               ci.input[j].data);
            controller->SetConversionValue(ci.slot, j, *(dword*)&ci.input[j].conversion);
         }
      }
   };
   if(slot==-1){
                              //setup default controller values - all slots
      for(int i=0; controller_init[i].slot!=-1; i++)
         S_hlp::SetupSingleSlot(i);
      
      controller->SetConfigValue(C_controller::CFG_ALWAYS_RUN, false);
      controller->SetConfigValue(C_controller::CFG_INVERT_MOUSE_Y, false);
      controller->SetConfigValue(C_controller::CFG_MOUSE_SENSITIVITY, 100);
   }else{
      S_hlp::SetupSingleSlot(slot);
   }
}

//----------------------------

#define CONTROLLER_REG_VERSION 0x02

bool WriteControllerConfig(){

   int rkey = RegkeyCreate(REGKEY_CONTROLS, E_REGKEY_LOCAL_MACHINE);
   if(rkey!=-1){
      dword dw;
      dw = CONTROLLER_REG_VERSION;
      RegkeyWdata(rkey, "Version", (const byte*)&dw, sizeof(dword));
                              //write all slots
      for(int i=0; controller_init[i].slot!=-1; i++){
         S_controller_item slot_data;
         int slot = controller_init[i].slot;
         for(int j=0; j<2; j++){
            C_controller::E_VALUE_MODE vm_ignore;
            controller->GetSlotSetting(slot, j, vm_ignore,
               slot_data.input[j].input_device, slot_data.input[j].data);
            slot_data.input[j].conversion = controller_init[i].input[j].conversion;
         }
         RegkeyWdata(rkey, C_fstring("%.3i", slot), (const byte*)&slot_data.input, sizeof(slot_data.input));
      }
                              //write additional values
      dw = controller->GetConfigValue(C_controller::CFG_ALWAYS_RUN);
      RegkeyWdata(rkey, "Always run", (const byte*)&dw, sizeof(dword));

      dw = controller->GetConfigValue(C_controller::CFG_INVERT_MOUSE_Y);
      RegkeyWdata(rkey, "Invert mouse", (const byte*)&dw, sizeof(dword));

      dw = controller->GetConfigValue(C_controller::CFG_MOUSE_SENSITIVITY);
      RegkeyWdata(rkey, "Mouse sensitivity", (const byte*)&dw, sizeof(dword));

      RegkeyClose(rkey);
      return true;
   }
   return false;
}

//----------------------------

bool ReadControllerConfig(){

   int rkey = RegkeyOpen(REGKEY_CONTROLS, E_REGKEY_LOCAL_MACHINE);
   if(rkey!=-1){
      dword dw;
                              //check version
      dw = 0;
      if(!(RegkeyRdata(rkey, "Version", &dw, sizeof(dword)) == sizeof(dword) && dw==CONTROLLER_REG_VERSION)){
         RegkeyClose(rkey);
         return false;
      }
                              //read all slots
      for(int i=0; controller_init[i].slot!=-1; i++){
         S_controller_item slot_data;
         int slot = controller_init[i].slot;
         if(RegkeyRdata(rkey, C_fstring("%.3i", slot), (byte*)&slot_data.input, sizeof(slot_data.input)) == sizeof(slot_data.input)){
            for(int j=0; j<2; j++){
               controller->SetupSlotInput(slot, j, C_controller::VM_NULL,
                  slot_data.input[j].input_device, slot_data.input[j].data);
               slot_data.input[j].conversion = controller_init[i].input[j].conversion;
            }
         }
      }
                              //read additional values
      if(RegkeyRdata(rkey, "Always run", &dw, sizeof(dword)) == sizeof(dword))
         controller->SetConfigValue(C_controller::CFG_ALWAYS_RUN, dw);

      if(RegkeyRdata(rkey, "Invert mouse", &dw, sizeof(dword)) == sizeof(dword))
         controller->SetConfigValue(C_controller::CFG_INVERT_MOUSE_Y, dw);

      if(RegkeyRdata(rkey, "Mouse sensitivity", &dw, sizeof(dword)) == sizeof(dword))
         controller->SetConfigValue(C_controller::CFG_MOUSE_SENSITIVITY, dw);

      RegkeyClose(rkey);
      return true;
   }
   return false;
}

//----------------------------
bool IsSpecialKeyDown();

bool CDPresenceCheck(){

#if defined FINAL || 0

   C_string drive = game_configuration.inst_src;
   if(drive.Size()>1 && drive[(int)drive.Size()-1]=='\\'){
      drive[(int)drive.Size()-1] = 0;
      drive[0] = toupper(drive[0]);
   }

#if (defined NDEBUG && !defined BETAVERSION) && 1
                              //check if valid CD-ROM drive is present
   if(!OsIsCDROMDrive(drive)){
      OsMessageBox(C_fstring(NO_CD_TEXT, (const char*)drive), APP_NAME, MBOX_OK);
      if(!IsSpecialKeyDown())
         return false;
   }
#endif

   bool cd_ok = true;
   bool win_hidden = false;
   while(true){
                        //try to open small file on CD
      int i = Load(C_fstring("%sdata\\data.dta", (const char*)game_configuration.inst_src), NULL, 0);
      if(i!=-1)
         break;
      if(igraph){
         if(igraph->GetFlags()&IG_FULLSCREEN){
            //igraph->FlipToGDI();
            win_hidden = true;
            OsShowWindow(igraph->GetHWND(), false);
         }
      }
      i = OsMessageBox(C_fstring(INSERT_CD_TEXT, (const char*)drive),
         APP_NAME, MBOX_OKCANCEL);
      if(i==MBOX_RETURN_CANCEL){
         cd_ok = false;
         break;
      }
   }
   if(win_hidden && cd_ok){
      OsShowWindow(igraph->GetHWND(), true);
      igraph->ProcessWinMessages();
   }

   return cd_ok;
#else
   return true;
#endif
}
//----------------------------

static vector<C_smart_ptr<C_dta_read> > dta_list;

enum{
   DTA_INSTALL_FULL = 1,
   DTA_INSTALL_MEDIUM = 2,
   DTA_INSTALL_MINIMAL = 4,
   DTA_INSTALL_CONTITIONAL = 8,
   DTA_INSTALL_TEST = 0x10,
   DTA_INSTALL_SOUND = 0x20,
   DTA_INSTALL_LOCALIZED = 0x40,
};

static struct{
   char *filename;
   dword inst_flags;
} dta_files[] = {
#if defined FINAL
   { "data\\data.dta", DTA_INSTALL_TEST}, //test for CD presence in minimal installation
#endif
   { "data\\data.dta", DTA_INSTALL_FULL | DTA_INSTALL_MEDIUM},
   { "data\\maps.dta", DTA_INSTALL_FULL | DTA_INSTALL_MEDIUM},
   { "data\\missions.dta", DTA_INSTALL_FULL | DTA_INSTALL_MEDIUM},
   { "data\\models.dta", DTA_INSTALL_FULL | DTA_INSTALL_MEDIUM},
   { "data\\anims.dta", DTA_INSTALL_FULL | DTA_INSTALL_MEDIUM},
   { "data\\sounds.dta", DTA_INSTALL_FULL | DTA_INSTALL_MEDIUM | DTA_INSTALL_SOUND},
   { "data\\sounds_%s.dta", DTA_INSTALL_FULL | DTA_INSTALL_SOUND | DTA_INSTALL_LOCALIZED},
   {NULL}
};

//----------------------------

void DtaCloseAll(){

   dta_list.clear();
}

//----------------------------
// Open all data files, either in local directory or in remote directory -
// dependent on installation type.
bool DtaOpenAll(){

   DtaCloseAll();

   if(!CDPresenceCheck())
      return false;

                              //open data files
   dword inst_mask = 0;
   switch(game_configuration.install_type){
   case S_game_configuration::INSTALL_MINIMAL: inst_mask = DTA_INSTALL_MINIMAL; break;
   case S_game_configuration::INSTALL_MEDIUM: inst_mask = DTA_INSTALL_MEDIUM; break;
   case S_game_configuration::INSTALL_FULL: //flow...
   case S_game_configuration::INSTALL_FULL_EDITOR: inst_mask = DTA_INSTALL_FULL; break;
   }

   for(int i=0; dta_files[i].filename; i++){
                              //determine if file is installed on local disk
      dword flags = dta_files[i].inst_flags;
      bool installed = (flags&inst_mask);

                              //if sound file, determine whether we're going to use streaming sounds
      if(flags&DTA_INSTALL_SOUND){
         allow_stream_snd = installed;
      }

      C_string full_name = installed ?
         C_fstring("%s%s", (const char*)game_configuration.inst_dst, dta_files[i].filename) :
         C_fstring("%s%s", (const char*)game_configuration.inst_src, dta_files[i].filename);
      if(flags&DTA_INSTALL_LOCALIZED)
         full_name = C_fstring(full_name, lang_names[game_configuration.language]);
      LPC_dta_read dta = dtaCreate(full_name, DTACREATE_SHOWINFO, APP_NAME);
      if(!dta){
#ifdef EDITOR
         continue;
#else
                              //is this required?
         if(flags&DTA_INSTALL_CONTITIONAL)
            continue;

         if(!installed){
            if(igraph) igraph->FlipToGDI();
            int id = OsMessageBox(C_fstring(INSERT_CD_TEXT, (const char*)game_configuration.inst_src), APP_NAME, MBOX_OKCANCEL);
            if(id==MBOX_RETURN_CANCEL){
               dta_list.clear();
               return false;
            }
         }else{
            if(igraph) igraph->FlipToGDI();
            int id = OsMessageBox(C_fstring("Cannot open data file %s.", (const char*)full_name),
               APP_NAME, MBOX_RETRYCANCEL);
            if(id==MBOX_RETURN_CANCEL){
               dta_list.clear();
               return false;
            }
         }
         --i;
#endif
      }else{
         if(!(flags&DTA_INSTALL_TEST)){
            dta_list.push_back(dta);
         }
         dta->Release();
      }
   }
   return true;
}
//----------------------------

bool OpenAllTexts(int lang_id){

   all_txt.Close();

   const char *lang_name = lang_names[lang_id];
   C_fstring dir_common("Text\\%s\\", lang_name);
   C_fstring dir_dd0("Text\\%s.0\\", lang_name);
   C_fstring dir_dd1("Text\\%s.1\\", lang_name);

   bool b = all_txt.AddFile(dir_common + "Common.txt");
   b = b && all_txt.AddFile(dir_common + "Gold.txt");
   b = b && all_txt.AddFile(dir_common + "Credits.txt");
   b = b && all_txt.AddFile(dir_dd0 + "GameMenu.txt");
   b = b && all_txt.AddFile(dir_dd0 + "Missions.txt");
                              //add datadisk texts with offset 1mio
   b = b &&
      all_txt.AddFile(dir_dd1 + "GameMenu.txt", 1e6);
   b = b &&
      all_txt.AddFile(dir_dd1 + "Missions.txt", 1e6);

#ifdef EDITOR
                              //Add-on texts during editing
   all_txt.AddFile("Text.mod\\AddOn.txt");
#endif
   return b;
}

//----------------------------

void __stdcall LogFunc(const char *cp){
   if(f_os_log){
      fwrite(cp, 1, strlen(cp), f_os_log);
      fwrite("\n", 1, 1, f_os_log);
                              //re-open, so that we have immediate feedback
      fclose(f_os_log);
      f_os_log = fopen(LOG_FILE_NAME, "at");
   }
};

//----------------------------

#ifdef EDITOR

extern const char *actor_names[];
//----------------------------
// Function used by editor - retrieve user-friendly ACTOR name from a frame.
const char *GetFrameActor(PI3D_frame frm){

   switch(frm->GetType()){
   case FRAME_VOLUME:
   case FRAME_CHECKPOINT:
   case FRAME_CAMERA:
      break;
   default:
      {
         LPC_actor act = (LPC_actor)frm->GetData();
         if(act)
            return actor_names[act->GetType()];
      }
   }
   return NULL;
}

#endif

//----------------------------

#ifdef EDITOR
void InitMissionPlugin(C_game_mission*, LPC_editor);
void InitSysConfigPlugin(C_game_mission*, LPC_editor);
void InitCheckpointPlugin(C_game_mission*, LPC_editor);
#endif

I3D_SCNCALLBACK cbDriver;

//----------------------------

extern const char *lang_names[];
void ApplyMaterialTable();
void ApplySoundEnvTable();

//----------------------------

static const char GRAPH_INIT_FAIL_TEXT[] = {
   "Unable to initialize graphics. This program requires DirectX 8 or greater.\n"
   "Please make sure you have latest DirectX installed, you can obtain it from www.microsoft.com.\n"
   "Also install latest drivers for your graphics card.\n"
   "Finally, you may need to run IConfig.exe and select appropriate video mode."
};

//----------------------------
void ApplySysSetting(LPC_table);

bool InitSystem(void *hi, const C_command_line &cmd_line){

   if(!CDPresenceCheck())
      return false;

   int i;
   I3D_RESULT ir;

   f_os_log = fopen(LOG_FILE_NAME, "wt");

   int res_x = cmd_line.res_x ? cmd_line.res_x : game_configuration.mode_x;
   int res_y = cmd_line.res_y ? cmd_line.res_y : game_configuration.mode_y;

#ifndef EDITOR                //force 4:3 screen ratio if windowed mode
   if(!game_configuration.fullscreen)
      game_configuration.mode_x = game_configuration.mode_y * 4 / 3;
#endif

                              //initialize graphics window
   IG_INIT ig;
   ig.size_x = res_x;
   ig.size_y = res_y;
   ig.bits_per_pixel = game_configuration.bits_per_pixel;
   ig.flags = 0;
#ifdef _DEBUG
   ig.flags |= IG_WARN_OUT;
#endif
#if defined EDITOR || defined _DEBUG
   ig.flags |= IG_NOHIDEMOUSE;
#endif
   if(!game_configuration.safe_mode)
      ig.flags |= IG_HW_TNL;
                              //allow full-screen mode only in non-editor
#ifndef EDITOR
   if(game_configuration.tripple)
      ig.flags |= IG_TRIPPLEBUF;
   if(game_configuration.fullscreen)
      ig.flags |= IG_FULLSCREEN;
#endif
#if defined EDITOR || defined _DEBUG
   //if(IsProgrammersPC())
   {
      ig.flags |= IG_DEBUGMOUSE;
   }
#endif
   ig.flags |= IG_DEBUGMOUSE | IG_NOHIDEMOUSE;
   ig.flags |= IG_SUSPENDINACTIVE;
#ifdef EDITOR
   ig.flags |= IG_RESIZEABLE;
#endif

   ig.hinst = hi;
   ig.log_func = LogFunc;

   ig.adapter_id = game_configuration.adapter_identifier;
   ig.d3d_device_type = game_configuration.device_type;
   if(cmd_line.dump_graph_caps) ig.flags |= IG_VERBOSE;

   if(cmd_line.pos_x!=0x80000000 && cmd_line.pos_y!=0x80000000){
      ig.pos_x = cmd_line.pos_x;
      ig.pos_y = cmd_line.pos_y;
      ig.flags |= IG_INITPOSITION;
   }

                              //initialize graphics
   igraph = IGraphCreate(&ig,
#ifndef NDEBUG
      !cmd_line.debug_use_release_dlls);
#else
      false);
#endif
   if(!igraph){
      OsMessageBox(GRAPH_INIT_FAIL_TEXT, APP_NAME, MBOX_OK);
      return false;
   }

                              //clear all buffers
   igraph->ClearViewport();
   igraph->UpdateScreen();
   if(game_configuration.fullscreen){
      igraph->ClearViewport();
      igraph->UpdateScreen();
      if(game_configuration.tripple){
         igraph->ClearViewport();
         igraph->UpdateScreen();
      }
   }
   igraph->SetAppName(APP_NAME);

   screen_aspect_ratio = (float)res_y / (float)res_x;
   //screen_aspect_ratio = .75f;


                              //initialize sounds
   if(game_configuration.use_sounds){
      ISND_INIT is;
      memset(&is, 0, sizeof(is));
      //is.flags = ISC_3DENABLE;
      is.mix_frequency = 22050;
      is.num_channels = 2;
      is.bits_per_second = 16;
      is.hwnd = igraph->GetHWND();
      is.guid = NULL;
#ifndef NDEBUG
      is.log_func = LogFunc;
      is.flags |= ISC_WARN_OUT;
#endif

      is.flags |= ISC_USE_EAX;

      if(cmd_line.dump_sound_caps)
         is.flags |= ISC_VERBOSE;

      isound = ISoundCreate(&is,
#ifndef NDEBUG
         !cmd_line.debug_use_release_dlls);
#else
         false);
#endif
      if(!isound){
         OsMessageBox(
            "Unable to initialize sound system.\n"
            "The application will now continue to run with sounds disabled."
            , APP_NAME, MBOX_OK);
      }else{
         game_configuration.ApplySoundConfig();
      }
   }

                              //initialize 3D engine
   I3DCreate2(&driver,
#ifndef NDEBUG
      !cmd_line.debug_use_release_dlls);
#else
      false);
#endif
   if(!driver){
      if(isound){
         isound->Release();
         isound = NULL;
      }
      igraph->Release();
      igraph = NULL;
      OsMessageBox(GRAPH_INIT_FAIL_TEXT, APP_NAME, MBOX_OK);
      return false;
   }

   dword i3d_init_flags = 0;
   if(cmd_line.dump_graph_caps)
      i3d_init_flags |= I3DINIT_VERBOSE;
#ifdef _DEBUG
   i3d_init_flags |= I3DINIT_WARN_OUT;
   if(cmd_line.debug_direct_trans)
      i3d_init_flags |= I3DINIT_DIRECT_TRANS;
#endif
   //if(game_configuration.hw_vp)
      //i3d_init_flags |= I3DINIT_DIRECT_TRANS;
   I3DINIT is = {
      igraph,
      i3d_init_flags,
      NULL,
      isound,
#ifdef _DEBUG
      LogFunc
#else
      NULL
#endif
   };
   ir = driver->Init(&is);
   if(I3D_SUCCESS(ir)){
#if !defined EDITOR && !defined _DEBUG
      OsDeleteFile(DATABASE_FILENAME);
#endif
      bool use_aniso = false;
      if(!game_configuration.safe_mode)
         use_aniso = (game_configuration.filter==S_game_configuration::FILTER_ANISO);
      driver->SetState(RS_ANISO_FILTERING, use_aniso);

//#if defined EDITOR || defined _DEBUG
      if(game_configuration.use_dbase && !game_configuration.safe_mode)
         driver->SetDataBase(DATABASE_FILENAME, DATABASE_SIZE);
//#endif
      //driver->SetState(RS_LMTRUECOLOR, false);
      driver->SetState(RS_TEXTUREDITHER, true);
      driver->SetState(RS_LMDITHER, true);
      driver->SetState(RS_TEXTURECOMPRESS, (!game_configuration.safe_mode && game_configuration.tx_compress));
      driver->SetState(RS_USESHADOWS, (!game_configuration.safe_mode && game_configuration.render_shadows));
      ApplySysSetting(tab_game_cfg);

      if(game_configuration.safe_mode){
         driver->SetState(RS_BUMPMAPPING, false);
         driver->SetState(RS_ENVMAPPING, false);
      }

      igraph->InitKeyboard();
      igraph->DefineKeys(game_keys, 32);

                              //register directories
                              //look in special directory for censored versions of textures
      {
         model_cache.SetCacheSize(MODEL_CACHE_SIZE);
         sound_cache.SetCacheSize(SOUND_CACHE_SIZE);

                              //common files
         driver->AddDir(I3DDIR_MAPS, "Maps");
                              //update - textures censored for all versions
         //driver->AddDir(I3DDIR_MAPS, C_fstring("Maps\\%s", game_configuration.censored ? "censored" : "nocensor"));
         driver->AddDir(I3DDIR_MAPS, C_fstring("Maps\\%s", "censored"));
         driver->AddDir(I3DDIR_MAPS, "Maps.mod");

         sound_cache.AddDir(C_fstring("Sounds\\%s", lang_names[game_configuration.language]));
         driver->AddDir(I3DDIR_SOUNDS, C_fstring("Sounds\\%s", lang_names[game_configuration.language]));

         //driver->AddDir(I3DDIR_SOUNDS, "Sounds\\Common");
         sound_cache.AddDir("Sounds\\Common");
         driver->AddDir(I3DDIR_SOUNDS, "Sounds.mod");
         sound_cache.AddDir("Sounds.mod");

         model_cache.AddDir("Models");
         model_cache.AddDir("Models.mod");

         driver->AddDir(I3DDIR_PROCEDURALS, "procedurals");

         driver->AddDir(I3DDIR_SOUNDS, "");  //we use cache, so fake driver to load from given dir
      }
      i = icons_img.Init(driver, "tables\\gamemenu.prj", "gamemenu.png", "menu_opt.png");
      if(i){
         i = font_img.Init(driver, "tables\\gamefont.prj", "gamefont.png", "font_opt.png");
         if(i){

                              //initialize joystick
#ifdef JOYSTICK
            i = CreateJoyBase(&joy_base, hi, igraph->GetHWND(), NULL);
            if(i){
               i = joy_base->GetDeviceCount();
               if(i){
                  joy_base->CreateJoy(&joy, 0);
                                       //setup joy
                  const float joy_dead_zone = .2f;
                  joy->SetDeadZone(JOY_X, joy_dead_zone);
                  joy->SetDeadZone(JOY_Y, joy_dead_zone);
                  joy->SetDeadZone(JOY_Z, joy_dead_zone);
                  joy->SetDeadZone(JOY_RX, joy_dead_zone);
                  joy->SetDeadZone(JOY_RY, joy_dead_zone);
                  joy->SetDeadZone(JOY_RZ, joy_dead_zone);

                  /*
                  bool b = joy->CreateEffect(&ff_effect);
                  if(b && ff_effect){
                     ff_effect->SetEffect(FF_SPRING, 0, 0, .5f);
                     S_condition sx, sy;
                     sx.offset = 0;
                     sx.pos_coeff = .5f;
                     sx.neg_coeff = .5f;
                     sx.pos_sat = 1;
                     sx.pos_sat = 1;
                     sx.dead_band = 0;
                     sy = sx;
                     ff_effect->SetCondition(&sx, &sy, false);
                     ff_effect->UpdateEffect();
                     ff_effect->Play(0, true);
                  }
                  */
               }
            }
            controller->Init(igraph, joy);
#else
            controller->Init(igraph, NULL, 65);
#endif
                        //first setup defaults, to initialize data values
            SetupControllerDefaults();
                        //read controller (may fail)
            ReadControllerConfig();

#ifdef EDITOR
            C_game_mission *mission = CreateGameMission();

            editor = CreateEditor(driver, mission->GetScene(), &model_cache, &sound_cache);
            if(editor){
               InitMissionPlugin(mission, editor);
               InitSysConfigPlugin(mission, editor);
               InitCheckpointPlugin(mission, editor);

               editor->LoadState("editor\\editor");
               LPC_editor_item ei = editor->FindPlugin("Selection");
               if(ei){
                  ei->Action(E_SELECTION_SET_GET_ACT, GetFrameActor);
               }
            }
            mission->Release();

            ApplyMaterialTable();
            ApplySoundEnvTable();
#else
            igraph->MouseInit();
#endif   //EDITOR
            driver->SetCallback(cbDriver, 0);
            driver->SetState(RS_LOADMIPMAP, true);

                              //apply brightness now
            float f = game_configuration.brightness * .01f;
            igraph->SetBrightness(f, f, f);

            return true;
//----------------
         }else{
            OsMessageBox("Internal error (unable to initialize sprite image).", APP_NAME, MBOX_OK);
         }
         icons_img.Close();
      }else{
         OsMessageBox("Internal error (unable to initialize sprite image).", APP_NAME, MBOX_OK);
      }
   }else{
      const char *cp = "unknown error";
      switch(ir){
      case I3DERR_OUTOFMEM:
         cp = "not enough video memory";
         break;
      }
      OsMessageBox(C_fstring("Unable to initialize 3D graphics (%s).", cp), APP_NAME, MBOX_OK);
      OsMessageBox(GRAPH_INIT_FAIL_TEXT, APP_NAME, MBOX_OK);
   }
   if(isound){
      isound->Release(); isound = NULL;
   }
   driver->Release(); driver = NULL;
   igraph->Release(); igraph = NULL;
   return false;
}

//----------------------------

void CloseSystem(){

#ifdef EDITOR
   if(editor){
      editor->SaveState("editor\\editor");
      C_smart_ptr<C_editor> ed = editor;  //keep ref until global 'editor' gets NULL
      editor->Release();
      editor = NULL;
   }
#endif//EDITOR

   model_cache.Clear();
   sound_cache.Clear();

   console_main.Clear();
   console_debug.Clear();
   icons_img.Close();
   font_img.Close();
   if(driver){
      driver->Close();
      driver->Release();
      driver = NULL;
   }
#ifdef JOYSTICK
   if(joy_base){
      if(ff_effect){
         ff_effect->Release();
         ff_effect = NULL;
      }
      if(joy){
         joy->Release();
         joy = NULL;
      }
      joy_base->Release();
      joy_base = NULL;
   }
#endif
   WriteControllerConfig();
   controller->Close();

   game_music.Close();
   if(isound){
      isound->Release();
      isound = NULL;
   }
   if(igraph){
                              //store current setting to config
      game_configuration.mode_x = igraph->Scrn_sx();
      game_configuration.mode_y = igraph->Scrn_sy();
      game_configuration.bits_per_pixel = igraph->GetPixelFormat()->bytes_per_pixel * 8;
      game_configuration.fullscreen = (igraph->GetFlags()&IG_FULLSCREEN);

      igraph->Release();
      igraph = NULL;
   }
   //NetClose();
#if !defined EDITOR && !defined _DEBUG
   OsDeleteFile(DATABASE_FILENAME);
#endif

   if(f_os_log){
      fclose(f_os_log);
      f_os_log = NULL;
#if defined _DEBUG || defined EDITOR || defined BETAVERSION
                              //show log, if it contains something
      if(GetFileSize(LOG_FILE_NAME) > 0){
         OsShell(LOG_FILE_NAME);
      }
#endif
   }
}

//----------------------------
//----------------------------

S_game_configuration::S_game_configuration():
   net_debug_schedule_cpu(false),
   net_debug_log(false),
   net_host(false),
   net_join(false),
   safe_mode(false),
   filter(FILTER_LINEAR)
{
                              //reset defaults of certain groups,
                              // for case we don't read this from registry
   ResetSoundDefaults();
   ResetVideoDefaults();
   ResetNetworkDefaults();
   ResetGameDefaults();
}

//----------------------------

bool S_game_configuration::ReadInstallationInfo(){

                              //read installation info
#if defined FINAL || 0
   int i = RegkeyOpen(REGDIR, E_REGKEY_LOCAL_MACHINE);
   if(i==-1){
#ifndef _DEBUG
      OsMessageBox(TEXT_BAD_INSTALL, APP_NAME, MBOX_OK);
      return false;
#else
      return true;
#endif
   }

   install_type = INSTALL_FULL;
   RegkeyRdata(i, INSTALL_TYPE_KEY, (byte*)&install_type, sizeof(E_ISNTALL_TYPE));

   char buf[512];
   if(RegkeyRdata(i, INSTALL_DIR_KEY, (byte*)buf, sizeof(buf))==-1)
      buf[0] = 0;
   else
   if(buf[strlen(buf)-1] != '\\')
      *(word*)(&buf[strlen(buf)]) = '\\';
   inst_src = buf;

   if(RegkeyRdata(i, INSTALL_DEST_DIR_KEY, (byte*)buf, sizeof(buf))==-1)
      buf[0] = 0;
   else
   if(buf[strlen(buf)-1] != '\\')
      *(word*)(&buf[strlen(buf)]) = '\\';
   inst_dst = buf;

   if(RegkeyRdata(i, INSTALL_LANG_KEY, (byte*)buf, sizeof(buf))==-1)
      buf[0] = 0;
   RegkeyClose(i);
   for(i=0; lang_names[i]; i++){
      if(!stricmp(lang_names[i], buf)){
         language = i;
         break;
      }
   }
#endif
   return true;
}

//----------------------------

static const char *key_name[] = {
                              //0
   "Display width",
   "Display height",
   "Display bitdepth",
   "Render shadows",
                              //4
   "Fullscreen",
   "TrippleBuf",
   "Adapter ID",
   "Device type",
                              //8
   "Database cache",
   "Sounds",
   "Filtering",
   "Compressed textures",
                              //12
   "Brightness",
   "Volume music",
   "Volume effects",
   "Volume speech",
                              //16
   "Sounds use 3D hardware",
   "Sounds use EAX",

   "Network connection type",
   "Network TCP IP address",
   "Network TCP IP port",
   "Network Modem string",
   "Network Modem number",
   "Network Player name",
                              //24
   "Draw aim line",
   "HW vertex processing",
   "",
   "Aim mode",
                              //28
   "Voice chat",
   "Voice compression",
};

//----------------------------

bool S_game_configuration::ReadSysConfig(){

   int i = RegkeyOpen(REGKEY_CONFIG, E_REGKEY_LOCAL_MACHINE);
   if(i!=-1){
      RegkeyRdata(i, key_name[0], (byte*)&mode_x, sizeof(mode_x));
      RegkeyRdata(i, key_name[1], (byte*)&mode_y, sizeof(mode_y));
      RegkeyRdata(i, key_name[2], (byte*)&bits_per_pixel, sizeof(bits_per_pixel));
      RegkeyRdata(i, key_name[3], (byte*)&render_shadows, sizeof(render_shadows));

      RegkeyRdata(i, key_name[4], (byte*)&fullscreen, sizeof(fullscreen));
      RegkeyRdata(i, key_name[5], (byte*)&tripple, sizeof(tripple));
      RegkeyRdata(i, key_name[6], (byte*)&adapter_identifier, sizeof(adapter_identifier));
      RegkeyRdata(i, key_name[7], (byte*)&device_type, sizeof(device_type));

      RegkeyRdata(i, key_name[8], (byte*)&use_dbase, sizeof(use_dbase));
      RegkeyRdata(i, key_name[9], (byte*)&use_sounds, sizeof(use_sounds));
      RegkeyRdata(i, key_name[10], (byte*)&filter, sizeof(filter));
      RegkeyRdata(i, key_name[11], (byte*)&tx_compress, sizeof(tx_compress));

      RegkeyRdata(i, key_name[12], (byte*)&brightness, sizeof(brightness));
      RegkeyRdata(i, key_name[13], (byte*)&volume_music, sizeof(volume_music));
      RegkeyRdata(i, key_name[14], (byte*)&volume_effects, sizeof(volume_effects));
      RegkeyRdata(i, key_name[15], (byte*)&volume_speech, sizeof(volume_speech));

      RegkeyRdata(i, key_name[16], (byte*)&sounds_use_3D_hw, sizeof(sounds_use_3D_hw));
      RegkeyRdata(i, key_name[17], (byte*)&sounds_use_EAX, sizeof(sounds_use_EAX));

      RegkeyRdata(i, key_name[18], (byte*)&net_ctype, sizeof(net_ctype));
      char buf[256];
      buf[sizeof(buf)-1] = 0;
      if(RegkeyRdata(i, key_name[19], buf, sizeof(buf)-1)!=-1){ net_tcpip_addr = buf; }
      if(RegkeyRdata(i, key_name[20], buf, sizeof(buf)-1)!=-1){ net_tcpip_port = buf; }
      if(RegkeyRdata(i, key_name[21], buf, sizeof(buf)-1)!=-1){ net_modem_str = buf; }
      if(RegkeyRdata(i, key_name[22], buf, sizeof(buf)-1)!=-1){ net_modem_num = buf; }
      if(RegkeyRdata(i, key_name[23], buf, sizeof(buf)-1)!=-1){ net_plr_name = buf; }

      hw_vp = false;
      RegkeyRdata(i, key_name[24], &draw_aim_line, sizeof(draw_aim_line));
      //RegkeyRdata(i, key_name[25], &hw_vp, sizeof(hw_vp));
      RegkeyRdata(i, key_name[27], &aim_mode, sizeof(aim_mode));

      RegkeyRdata(i, key_name[28], &enable_voice_chat, sizeof(enable_voice_chat));
      RegkeyRdata(i, key_name[29], &voice_chat_compression, sizeof(voice_chat_compression));

      RegkeyClose(i);
      return true;
   }

   return false;
}

//----------------------------

bool S_game_configuration::WriteSysConfig(){

   int i = RegkeyCreate(REGKEY_CONFIG, E_REGKEY_LOCAL_MACHINE);
   if(i!=-1){
      RegkeyWdword(i, key_name[0], mode_x);
      RegkeyWdword(i, key_name[1], mode_y);
      RegkeyWdword(i, key_name[2], bits_per_pixel);
      RegkeyWdword(i, key_name[3], render_shadows);

      RegkeyWdword(i, key_name[4], fullscreen);
      RegkeyWdword(i, key_name[5], tripple);
      RegkeyWdword(i, key_name[6], adapter_identifier);
      RegkeyWdword(i, key_name[7], device_type);

      RegkeyWdword(i, key_name[8], use_dbase);
      RegkeyWdword(i, key_name[9], use_sounds);
      RegkeyWdword(i, key_name[10], filter);
      RegkeyWdword(i, key_name[11], tx_compress);

      RegkeyWdword(i, key_name[12], brightness);
      RegkeyWdword(i, key_name[13], volume_music);
      RegkeyWdword(i, key_name[14], volume_effects);
      RegkeyWdword(i, key_name[15], volume_speech);

      RegkeyWdword(i, key_name[16], sounds_use_3D_hw);
      RegkeyWdword(i, key_name[17], sounds_use_EAX);

      RegkeyWdword(i, key_name[18], net_ctype);
      RegkeyWtext(i, key_name[19], net_tcpip_addr);
      RegkeyWtext(i, key_name[20], net_tcpip_port);
      RegkeyWtext(i, key_name[21], net_modem_str);
      RegkeyWtext(i, key_name[22], net_modem_num);
      RegkeyWtext(i, key_name[23], net_plr_name);

      RegkeyWdword(i, key_name[24], draw_aim_line);
      //RegkeyWdword(i, key_name[25], hw_vp);
      RegkeyWdword(i, key_name[27], aim_mode);

      RegkeyWdword(i, key_name[28], enable_voice_chat);
      RegkeyWdword(i, key_name[29], voice_chat_compression);

      RegkeyClose(i);

      return true;
   }
   return false;
}

//----------------------------

void S_game_configuration::ResetSoundDefaults(){

   use_sounds = true;
   volume_music = 100;
   volume_effects = 100;
   volume_speech = 100;
   sounds_use_3D_hw = true;
   sounds_use_EAX = true;
}

//----------------------------

void S_game_configuration::ResetVideoDefaults(){

   //filter = FILTER_LINEAR;
   tripple = true;
   fullscreen = false;
   tx_compress = true;
   render_shadows = true;
   brightness = 100;
}

//----------------------------

void S_game_configuration::ResetNetworkDefaults(){

   net_ctype = INCON_TCPIP;
   net_tcpip_addr = NULL;
   net_tcpip_port = NULL;
   net_modem_str = NULL;
   net_modem_num = NULL;
   OsGetComputerName(net_plr_name);
   enable_voice_chat = false;
   voice_chat_compression = VCC_3_2;
   net_debug_numplayers = 1;
}

//----------------------------

void S_game_configuration::ResetGameDefaults(){

   draw_aim_line = false;
   aim_mode = 0;
}

//----------------------------

void S_game_configuration::ApplySoundConfig(){

   if(isound){
      float gv = Max(.01f, volume_effects * .01f);
      isound->SetVolume(gv);
      game_music.SetVolume();
   }
}

//----------------------------

void S_game_configuration::ApplyVideoConfig(){

   igraph->UpdateParams(game_configuration.mode_x, game_configuration.mode_y,
      game_configuration.bits_per_pixel,
      (fullscreen ? IG_FULLSCREEN : 0) |
      (tripple ? IG_TRIPPLEBUF : 0),
      IG_FULLSCREEN | IG_TRIPPLEBUF);
   float f = game_configuration.brightness * .01f;
   igraph->SetBrightness(f, f, f);
}

//----------------------------
//----------------------------
