/*******************************************************
* File: c_except.hpp
* Purpose: exception handling - classes declaration
* Created by: Michal Bacik
*
*******************************************************/


#ifndef __C_EXCEPT_H
#define __C_EXCEPT_H

#include <stdio.h>
#include <c_string.hpp>

class C_except{
public:
   virtual const char *what() const = 0;
};

//----------------------------

class C_except_general: public C_except{
   C_string str;
public:
   C_except_general(){}
   C_except_general(const char *cp):
      str(cp)
   {}
   C_except_general(const C_string &s1):
      str(s1)
   {}
   C_except_general &operator << (const char *cp){
      str += cp;
      return *this;
   }
   C_except_general &operator << (int i){
      char buf[64];
      sprintf(buf, "%i", i);
      str += buf;
      return *this;
   }
   C_except_general &operator << (float f){
      char buf[64];
      sprintf(buf, "%f", f);
      str += buf;
      return *this;
   }
   virtual const char *what() const{
      return str.size() ? (const char*)str : "general exception";
   }
};

//----------------------------

class C_except_mem: public C_except_general{
public:
   C_except_mem():
      C_except_general("memory allocation failed")
   {}
   C_except_mem(const char *desc):
      C_except_general(C_fstring("memory allocation failed: %s", desc))
   {}
};

//----------------------------

class C_except_file: public C_except_general{
public:
   C_except_file(const char *fname):
      C_except_general(C_fstring("failed to open file: %s", fname))
   {}
   C_except_file():
      C_except_general("failed to open file")
   {}
};

//----------------------------

class C_except_math;

#ifdef EXCEPT_DEFINE_MATH

#error
class C_except_math: public C_except{
   C_string str;
public:
   C_except_math(_exception *err_info)
   {
      char *cp;
      switch(err_info->type){
      case DOMAIN:    cp="domain";    break;
      case SING:      cp="sing";      break;
      case OVERFLOW:  cp="owerflow";  break;
      case UNDERFLOW: cp="underflow"; break;
      case TLOSS:     cp="tloss";     break;
      case PLOSS:     cp="ploss";     break;
      default:        cp="unknown";
      }
      char buf[128];
      sprintf(buf, "%s error in %s [%f] [%f]", cp, err_info->name,
         err_info->arg1, err_info->arg2);
      str=buf;
   }
   virtual const char *what() const{ return str; }
};

int matherr(_exception *err_info){
   throw(C_except_math(err_info));
   return 0;
}

double _matherr(_exception *err_info){
   throw(C_except_math(err_info));
   return 0.0;
}

#endif   //EXCEPT_DEFINE_MATH

#endif   //__C_EXCEPT_H
