#ifndef __INSANITY_OS_H_
#define __INSANITY_OS_H_

#include <c_string.hpp>
#include <vector>

#ifdef _MSC_VER
using namespace std;

#ifdef _DEBUG
#pragma comment(lib, "insanity_d")
#else
#pragma comment(lib, "insanity")
#endif

#endif

//----------------------------

struct WIN_GUID{
   byte data[16];
};

//----------------------------

struct S_date_time{
   word year;
   byte month;                //1 ... 12
   byte day;                  //1 ... 31
   byte hour;                 //0 ... 23
   byte minute;               //0 ... 59
   byte second;               //0 ... 59
};

//----------------------------
                              //messagebox button styles
enum MSGBOX_STYLE{
   MBOX_OK,                   //OK button
   MBOX_YESNO,                //Yes/No buttons
   MBOX_YESNOCANCEL,          //Yes/No/Cancel buttons
   MBOX_RETRYCANCEL,          //Retry/Cancel buttons
   MBOX_OKCANCEL,             //Ok/Cancel buttons
};

                              //message box return value
enum MSGBOX_RETURN{
   MBOX_RETURN_NO,            //No button pressed
   MBOX_RETURN_YES,           //Yes/Ok/Retry button pressed
   MBOX_RETURN_CANCEL,        //Cancel button pressed
};

//----------------------------
// Display message box, and wait until user clicks one of buttons.
MSGBOX_RETURN OsMessageBox(const char *txt, const char *title, MSGBOX_STYLE);

//----------------------------
// Create specified directory.
bool OsCreateDirectory(const char *dir);

//----------------------------
// Create directory tree (ignoring file on top of path).
bool OsCreateDirectoryTree(const char *path);

//----------------------------
// Delete specified file
bool OsDeleteFile(const char *filename);

//----------------------------
// Get name of computer.
bool OsGetComputerName(C_string &str);

//----------------------------
// Move file from source to destination.
bool OsMoveFile(const char *src, const char *dst);

//----------------------------
// Destroy specified window.
void OsDestroyWindow(void *hwnd);

//----------------------------
// Show/hide specified window.
void OsShowWindow(void *hwnd, bool on);

//----------------------------
// Center specified window 'whnd' relative to another window 'hwnd_other'.
void OsCenterWindow(void *hwnd, void *hwnd_other);

//----------------------------
// Check if specified file exists and has read-only attribute set.
bool OsIsFileReadOnly(const char *filename);

//----------------------------
// Clear read-only attribute from specified file.
// The returned value is false if file doesn't exists, or it is already writable, or attributes cannot be modified.
bool OsMakeFileReadOnly(const char *filename, bool rdonly);

//----------------------------
// Sleep for specified time.
void OsSleep(int time);

//----------------------------
// Run system shell - perform specified cmd (usually open the file with associated program).
bool OsShell(const char *cmd);

//----------------------------
// Check if provided path is on a CD-ROM drive.
bool OsIsCDROMDrive(const char *path);

//----------------------------
// Collect all matching names of files in specified directory.
void OsCollectFiles(const char *path, const char *wildmask, vector<C_string> &ret);

//----------------------------
// Create unique filename from provided string. This means the name contains only characters allowable in filenames,
// and doesn't collide with existing file in provided directory.
// In this context, provided string is just a hint, and resulting name may be quite different from original.
// Upon return, 'str' contains new unique name.
void OsCreateUniqueFileName(const char *path, const char *ext, C_string &str);

//----------------------------
// Get date and time.
void OsGetCurrentTimeDate(struct S_date_time&);
bool OsGetFileTimeDate(const char *fname, struct S_date_time&);

//----------------------------
// Get version info about specified executable.
bool OsGetExecutableVersion(const char *fname, int &major, int &minor);

//----------------------------
// Create sytem font (HFONT). Returns handle to created font.
void *OsCreateFont(int size);

//----------------------------
// Destroy system font (HFONT).
void OsDestroyFont(void *handle);

//----------------------------

#endif

