#ifndef __EASE_HPP_
#define __EASE_HPP_
/*
   Ease interpolator.
   Written by Michal Bacik, spring 1998.

   Abstract:

      Creates interpolation in time range 0.0f ... 1.0f. In and out easiness
   may be set up independently.
      Complies with 3D Studio v.4.0 and MAX easy interpolators.
*/
#pragma once

#include <math.h>

#ifndef PI
 #define PI 3.1415926536L
#endif
                              //interpolators
class C_ease_interpolator{
   float in_power, out_power;
   float i_in_power, i_out_power;
public:
                              //Default constructor - sets maximal interpolation
                              //for both 'from' and 'to' ends.
   C_ease_interpolator(){
      Setup(1.0f, 1.0f);
   }
   C_ease_interpolator(float in, float out){
      Setup(in, out);
   }
                              //Setup of interpolator - sets 'from' and 'to'
                              // ends. Value 0.0f means linear interpolation,
                              // value 1.0f means lin. acceleration interpolation.
   void Setup(float ip, float op){
      in_power = ip;
      out_power = op;
      i_in_power = 1.0f - ip;
      i_out_power = 1.0f - op;
   }
                              //Getting interpolated value on time t. 't' must be
                              //in range 0.0f to 1.0f.
   float operator[](float t) const{
      if(t<.5f)
         return in_power * ((float)cos(PI+t*PI)+1.0f) * .5f + i_in_power * t;
      t = 1.0f - t;
      return 1.0f - (out_power*((float)cos(PI+t*PI)+1.0f) * .5f + i_out_power * t);
   }
};

#endif
