#ifndef __SMARTPTR_H_
#define __SMARTPTR_H_
/*
   Smart pointer class for use by classes using reference counting.
   Written by Michal Bacik 06.01.1999
   Requirements: AddRef and Release methods defined in class.
*/

template<class T> class C_smart_ptr{
   T *ptr;
public:
   inline C_smart_ptr(): ptr(0){}
   inline C_smart_ptr(T *tp): ptr(tp){ if(ptr) ptr->AddRef(); }
   inline C_smart_ptr(const C_smart_ptr<T> &sp): ptr(sp.ptr){ if(ptr) ptr->AddRef(); }
   inline ~C_smart_ptr(){ if(ptr) ptr->Release(); }
   inline void Release(){ if(ptr){ ptr->Release(); ptr = 0; } }
   inline C_smart_ptr &operator =(T *tp){
      if(ptr!=tp){
         if(tp) tp->AddRef();
         if(ptr) ptr->Release();
         ptr = tp;
      }
      return *this;
   }
   inline C_smart_ptr &operator =(const C_smart_ptr &sp){
      if(ptr!=sp.ptr){
         if(sp.ptr) sp.ptr->AddRef();
         if(ptr) ptr->Release();
         ptr = sp.ptr;
      }
      return *this;
   }
   inline operator T *() const{ return ptr; }
   inline T &operator *() const{ return *ptr; }
   inline T *operator ->() const{ return ptr; }

   inline bool operator !(){ return (!ptr); }
   inline bool operator ==(T *tp){ return (ptr==tp); }
   inline bool operator !=(T *tp){ return (ptr!=tp); }
};

#endif
