#ifndef __TABBLER2_H
#define __TABBLER2_H

#include <rules.h>

#pragma comment(lib,"itabler2")

#define TABAPI __stdcall
#define TABMETHOD_(type,method) virtual type TABAPI method

//----------------------------

                              //table element types
enum E_TABLE_ELEMENT_TYPE{
   TE_NULL,                   //must be last member of template
                              //basic types
   TE_BOOL,
   TE_INT,
   TE_FLOAT,
   TE_ENUM,
   TE_STRING,
   TE_COLOR_RGB,              //byte[3]   (0 - 255)
   TE_COLOR_VECTOR,           //float[3]  (0.0 - 1.0)
                              //special types (template-only)
   TE_BRANCH,
   TE_ARRAY,

   TE_LAST
};

typedef class C_table *LPC_table;

                              //editor callback messages
enum{
   TCM_MODIFY = 1,            //value modified (dword user, int item_index, int array_index)
   TCM_UPDATE = 2,            //value modified (dword user, 0)
   TCM_CLOSE  = 3,            //dialog is closing (dword user, LPRECT pos_size)
   TCM_OPEN   = 4,            //dialog is opening (dword user, LPRECT pos_size)
   TCM_FOCUS  = 5,            //focus changed (dword user, (short old_index)<<16 | short new_index)
};

                              //editor callback
typedef void (*table_callback)(LPC_table, dword msg, dword cb_user, dword prm2, dword prm3);
                              //table editor flags
#define TABEDIT_CENTER        1
#define TABEDIT_EXACTHEIGHT   2
#define TABEDIT_MANUALUPDATE  4
#define TABEDIT_MODELESS      8
#define TABEDIT_EDITDEFAULTS  0x10
#define TABEDIT_IMPORT        0x20
#define TABEDIT_EXPORT        0x40
#define TABEDIT_INFO          0x80
#define TABEDIT_HIDDEN        0x100
#define TABEDIT_CONTROL       0x200

//----------------------------
                              //table interface

#define TABOPEN_TEMPLATE   1  //create empty table from template
#define TABOPEN_FILENAME   2  //load from file
#define TABOPEN_FILEHANDLE 4  //load from open file
#define TABOPEN_DUPLICATE  8  //duplicate given table
#define TABOPEN_TEMPLTABLE 0x10  //table containing template

#define TABOPEN_UPDATE     0x10000  //only fill-in matching mebers


#ifndef TABBLER_FULL

class C_table{
public:
   TABMETHOD_(dword,AddRef)() = 0;
   TABMETHOD_(dword,Release)() = 0;

   TABMETHOD_(bool,Open)(dword dw_open, dword flags) = 0;
   TABMETHOD_(bool,Save)(dword dw_save, dword flags) = 0;
   TABMETHOD_(bool,IsOpen)() const = 0;

   TABMETHOD_(int,NumItems)() const = 0;
   TABMETHOD_(int,ArrayLen)(int index) const = 0;
   TABMETHOD_(int,StringSize)(int index) const = 0;
   TABMETHOD_(int,SizeOf)(int index) const = 0;

//----------------------------
// Return address of generic item on specified index. If index is out of range,
// or no such item exists, the returned value is NULL.
   TABMETHOD_(void*,Item)(int index) = 0;

//----------------------------
// Get various items - this is basically reference to generic item.
   TABMETHOD_(bool&,ItemB)(int index) = 0;
   TABMETHOD_(int&,ItemI)(int index) = 0;
   TABMETHOD_(float&,ItemF)(int index) = 0;
   TABMETHOD_(byte&,ItemE)(int index) = 0;
   TABMETHOD_(char*,ItemS)(int index) = 0;
   TABMETHOD_(struct S_vector&,ItemV)(int index) = 0;
//----------------------------
// Get various items from arrays.
   TABMETHOD_(bool&,ItemB)(int index, int ai) = 0;
   TABMETHOD_(int&,ItemI)(int index, int ai) = 0;
   TABMETHOD_(float&,ItemF)(int index, int ai) = 0;
   TABMETHOD_(byte&,ItemE)(int index, int ai) = 0;
   TABMETHOD_(char*,ItemS)(int index, int ai, int str_size) = 0;
   TABMETHOD_(struct S_vector&,ItemV)(int index, int ai) = 0;

   TABMETHOD_(E_TABLE_ELEMENT_TYPE,GetItemType)(int index) const = 0;
   TABMETHOD_(void*,GetGUID)() = 0;

                              //table editor - returns HWND
   TABMETHOD_(void*,Edit)(const class C_table_template*, void *hwnd_parent,
      table_callback cb_proc, dword cb_user, dword create_flags = 0,
      const char *def_table = NULL, int *pos_size = NULL) = 0;

   TABMETHOD_(void*,Edit)(C_table*, void *hwnd_parent,
      table_callback cb_proc, dword cb_user, dword create_flags = 0,
      const char *def_table = NULL, int *pos_size = NULL) = 0;
   TABMETHOD_(bool,IsModified)() const = 0;
};

#endif

//----------------------------
                              //table template element
class C_table_element{
public:
   E_TABLE_ELEMENT_TYPE type;
   int tab_index;
   char *caption;
   union{
      int flt_min;
      int int_min;

      int branch_depth;       //TE_BRANCH
      char *enum_list;        //TE_ENUM
      int min_array_size;     //TE_ARRAY
      int default_string_size;//TE_STRING
   };
   union{
      int flt_max;
      int int_max;
                              //formatted text for TE_BRANCH
                              //format is: %[x] where x is index of sub-item
                              //this is automamically included into string in its proper format
      char *branch_text;      //TE_BRANCH
                              //if this is zero (or less than min_array_size) no resizing is allowed
                              //set to -1 to allow unlimited table expanding
      int max_array_size;     //TE_ARRAY (optional)
   };
   union{
                              //defaults:
      int int_default;
      int flt_default;
      char *string_default;
   };
};
typedef C_table_element *LPC_table_element;

#ifndef TABBLER_FULL
                              //table template declaration
class C_table_template{
public:
   struct{ dword d1; word d2; word d3; byte d4[8]; }guid;
   char *caption;
   char *help_text;
#pragma warning(disable: 4200)//using non-standard extension
   C_table_element te[];
#pragma warning(default: 4200)
};
#endif

typedef class C_table_template *LPC_table_template;

//----------------------------
                              //table creation
extern"C"{
LPC_table __declspec(dllexport) TABAPI CreateTable();
}

//----------------------------

#endif
