#include <windows.h>
#include <winable.h>
#include <debug.h>


//----------------------------

#define USE_SYNCHRONIZATION

//----------------------------

static int vk_codes[] = {
   VK_SCROLL,
   VK_CAPITAL,
   VK_NUMLOCK
};

//----------------------------

                              //struct used to simulate key up and key down
static INPUT input[2];

static byte orig_state, curr_state;

static bool is_nt_platform;

#ifdef USE_SYNCHRONIZATION
static CRITICAL_SECTION csect;
#endif

//----------------------------

int flash_count[3];

//----------------------------

static class C_dled_auto_init{
public:
   C_dled_auto_init(){
                              //detect if OS is NT-based
      {
         OSVERSIONINFO oi;
         oi.dwOSVersionInfoSize = sizeof(oi);
         GetVersionEx(&oi);
         is_nt_platform = (oi.dwPlatformId == VER_PLATFORM_WIN32_NT);
      }

      if(is_nt_platform){
         for(int i=0; i<2; i++){
            input[i].type = INPUT_KEYBOARD;
            input[i].ki.dwFlags = KEYEVENTF_EXTENDEDKEY;
         }
         input[1].ki.dwFlags |= KEYEVENTF_KEYUP;
      }
      curr_state = 0;
      for(int i=0; i<3; i++){
         if(GetKeyState(vk_codes[i]))
            curr_state |= (1<<i);
      }
      orig_state = curr_state;
#ifdef USE_SYNCHRONIZATION
      InitializeCriticalSection(&csect);
#endif
   }

   ~C_dled_auto_init(){
      Led(orig_state);
#ifdef USE_SYNCHRONIZATION
      DeleteCriticalSection(&csect);
#endif
   }
} dled_auto_init;

//----------------------------

void Led(dword data, dword mask){

#ifdef USE_SYNCHRONIZATION
   EnterCriticalSection(&csect);
#endif

   dword change = (curr_state ^ data) & mask;
   if(change){

      curr_state ^= change;
      if(is_nt_platform){
         for(int i=3; i--; ){
            if(change & (1<<i)){
               input[0].ki.wVk = input[0].ki.wScan = vk_codes[i];
               input[1].ki.wVk = input[1].ki.wScan = vk_codes[i];
               SendInput(2, input, sizeof(INPUT));
            }
         }
      }else{
         __asm{
            mov al, 0xed
            out 0x60, al
            mov al, curr_state
            mov ah, al
            and al, 1
            shl ah, 1
            and ah, 4
            or al, ah
            mov ah, curr_state
            shr ah, 1
            and ah, 2
            or al, ah
            out 0x60, al
         }
      }
   }
#ifdef USE_SYNCHRONIZATION
   LeaveCriticalSection(&csect);
#endif
}

//----------------------------

dword CALLBACK ThreadProc(void *vp){

   dword dw = (dword)vp;
   int indx = dw>>16;
   int count = dw & 0xffff;
   
   Sleep(count);

#ifdef USE_SYNCHRONIZATION
   EnterCriticalSection(&csect);
#endif

   if(!--flash_count[indx])
      Led(0, 1<<indx);

#ifdef USE_SYNCHRONIZATION
   LeaveCriticalSection(&csect);
#endif

   ExitThread(0);
   return 0;
}

//----------------------------

void LedFlash(int indx, int count){

   if(!count || indx<0 || indx>2)
      return;

#ifdef USE_SYNCHRONIZATION
   EnterCriticalSection(&csect);
#endif

                              //flash this led
   if(!flash_count[indx])
      Led(7, 1<<indx);
   ++flash_count[indx];

   dword tid;
   CreateThread(NULL,
      0,
      ThreadProc,
      (void*)((indx<<16) | count),
      0,
      &tid);
#ifdef USE_SYNCHRONIZATION
   LeaveCriticalSection(&csect);
#endif
}

//----------------------------
