#include <windows.h>
#include <c_string.hpp>
#include <set>
#include <malloc.h>

using namespace std;

#pragma warning(disable:4073)
#pragma init_seg(lib)

//----------------------------
#define USE_SYNCHRONIZATION

#define FILL_MASK 0xffcecece  //float NAN, integer garbage, pointer garbage
#define MEM_GUARD_DWORDS 2    //number of guard dwords on both sides on allocated memory



//----------------------------

void __declspec(dllimport) __stdcall UserException(const char *msg);
#pragma comment(lib, "iexcpt")

//----------------------------

static bool in_alloc;
static bool de_init;
typedef set<void*> t_blocks;
static t_blocks *blocks;
static dword alloc_id;
static dword break_alloc_id = -1;
static void *break_alloc_ptr = NULL;


struct S_mem_header{
   dword alloc_id;            //-1 = internal block
   dword size;
};


#ifdef USE_SYNCHRONIZATION
static CRITICAL_SECTION csect;
#endif

static class C_auto_init{
public:
   C_auto_init(){
#ifdef USE_SYNCHRONIZATION
      InitializeCriticalSection(&csect);
#endif
      _set_sbh_threshold(1016);
   }
   ~C_auto_init(){
#ifdef USE_SYNCHRONIZATION
      EnterCriticalSection(&csect);
#endif
      if(break_alloc_id!=-2)
      if(blocks){
         if(blocks->size()){
            t_blocks::const_iterator it;
            dword sz = 0;
            C_string ids;
            int num_ids = 0;
            for(it=blocks->begin(); it!=blocks->end(); it++){
               void *vp = (*it);
               vp = (byte*)vp - sizeof(S_mem_header) - MEM_GUARD_DWORDS*4;
               const S_mem_header &header = *(S_mem_header*)vp;
               assert(header.alloc_id != -1);
               sz += header.size;
               if(num_ids<8){
                  if(ids.Size())
                     ids += ", ";
                  ids += C_fstring("%i", header.alloc_id);
               }
               ++num_ids;
            }
            C_fstring msg("Memory leaks detected - %i block(s), %i bytes\r\nAllocation IDs: ", blocks->size(), sz);

            msg += ids;
            UserException(msg);
         }
         delete blocks;
      }
      de_init = true;
#ifdef USE_SYNCHRONIZATION
      LeaveCriticalSection(&csect);
      DeleteCriticalSection(&csect);
#endif
   }

} auto_init_mem_debug;

//----------------------------
extern"C"{
void __declspec(dllexport) *__cdecl DebugAlloc(size_t);
void __declspec(dllexport) __cdecl DebugFree(void*);
void __declspec(dllexport) *__cdecl DebugRealloc(void*, size_t);
void __declspec(dllexport) __cdecl SetBreakAlloc(dword);
void __declspec(dllexport) __cdecl SetBreakAllocPtr(void*);
}

//----------------------------

void __declspec(dllexport) __cdecl SetBreakAlloc(dword id){
   break_alloc_id = id;
}

//----------------------------

void __declspec(dllexport) __cdecl SetBreakAllocPtr(void *vp){
   break_alloc_ptr = vp;
}

//----------------------------

                              //fill allocated and freed mem by garbage
                              // requires:
                              // non-debug memory allocation functions


void __declspec(dllexport) *__cdecl DebugAlloc(size_t sz){

   if(!sz)
      return NULL;

#ifdef USE_SYNCHRONIZATION
   EnterCriticalSection(&csect);
#endif

   void *vp;

   if(in_alloc){
      vp = malloc(sz + sizeof(S_mem_header) + MEM_GUARD_DWORDS*4);
      S_mem_header &header = *(S_mem_header*)vp;
      header.alloc_id = -1;
      vp = (byte*)vp + sizeof(S_mem_header) + MEM_GUARD_DWORDS*4;
   }else{
      in_alloc = true;

      vp = malloc(sz + sizeof(S_mem_header) + MEM_GUARD_DWORDS*4*2);
      S_mem_header &header = *(S_mem_header*)vp;
      header.alloc_id = ++alloc_id;
      if(header.alloc_id == break_alloc_id){
         break_alloc_id = -2;
         __asm int 3
      }
      header.size = sz;
      vp = (byte*)vp + sizeof(S_mem_header);

      __asm{
         push ecx

         mov eax, FILL_MASK
         mov edi, vp

         mov ecx, sz
         add ecx, MEM_GUARD_DWORDS*4
         push ecx
         shr ecx, 2

         jz new_no_fill
         rep stosd

      new_no_fill:
         pop ecx
         and ecx, 3
         jz new_no_fill1
         rep stosb

      new_no_fill1:

         mov ecx, MEM_GUARD_DWORDS
         rep stosd

         pop ecx
      }
      vp = (byte*)vp + MEM_GUARD_DWORDS*4;
      if(vp==break_alloc_ptr){
         __asm int 3
      }
      if(!blocks)
         blocks = new t_blocks;
      blocks->insert(vp);
      in_alloc = false;
   }
#ifdef USE_SYNCHRONIZATION
   LeaveCriticalSection(&csect);
#endif
   return vp;
}

//----------------------------

void __declspec(dllexport) __cdecl DebugFree(void *vp1){

   if(de_init)
      return;

#ifdef USE_SYNCHRONIZATION
   EnterCriticalSection(&csect);
#endif

   void *vp = vp1;
   if(vp){
      vp = (byte*)vp - sizeof(S_mem_header) - MEM_GUARD_DWORDS*4;
      const S_mem_header &header = *(S_mem_header*)vp;

      if(header.alloc_id != -1){

         in_alloc = true;

         bool ok_mem = false;
         if(blocks){
            t_blocks::iterator it = blocks->find(vp1);
            if(it!=blocks->end()){
               blocks->erase(it);
               ok_mem = true;
            }
         }
         if(ok_mem){

            size_t sz = header.size;
            void *mem = (byte*)vp + sizeof(S_mem_header);
            byte corrupt = 0;
            __asm{
               push ecx

               mov eax, FILL_MASK
                                 //check guard bytes
               mov edi, mem
               mov ecx, MEM_GUARD_DWORDS
               rep scasd
               jz ok_mem1
               mov corrupt, 1
               jmp del_no_fill1

            ok_mem1:
               mov ecx, MEM_GUARD_DWORDS
               mov edi, mem
               add edi, sz
               add edi, MEM_GUARD_DWORDS*4
               rep scasd
               jz ok_mem2
               mov corrupt, 2
               jmp del_no_fill1

            ok_mem2:
                                 //fill by garbage again
               mov edi, vp
               add edi, MEM_GUARD_DWORDS*4
               mov ecx, sz
               shr ecx, 2
               jz del_no_fill
               rep stosd
            del_no_fill:

               mov ecx, sz
               and ecx, 3
               jz del_no_fill1
               rep stosb

            del_no_fill1:
               pop ecx
            }
            if(corrupt){
               mem = (byte*)mem + MEM_GUARD_DWORDS*4;
            
               C_fstring msg(
                  "Memory corrupted: 0x%x, size: %i (written %s block)\r\n"
                  "Memory dump:"
                  ,
                  (byte*)mem, sz,
                  (corrupt==1) ? "before" : "past");
               for(int i=0; i<Min((int)sz, 32); i += 8){
                  C_string bytes;
                  char chars[9];
                  const byte *bp = (const byte*)mem + i;
                  for(int x=0; x<8; x++){
                     byte c = bp[x];
                     bytes += C_fstring("%.2x ", c);
                     if(!isprint(c))
                        c = '.';
                     chars[x] = c;
                  }
                  chars[8] = 0;
                  msg += C_fstring("\r\n%.2x   %s  %s", i, (const char*)bytes, chars);
               }
               UserException(msg);
               vp = NULL;
            }
         }else{
            UserException(C_fstring("Freeing non-allocated memory: 0x%x", vp1));
            vp = NULL;
         }
         in_alloc = false;
      }
   }
   free(vp);
#ifdef USE_SYNCHRONIZATION
   LeaveCriticalSection(&csect);
#endif
}

//----------------------------

void __declspec(dllexport) *__cdecl DebugRealloc(void *vp_old, size_t sz){

   assert(vp_old && sz);
   void *vp = DebugAlloc(sz);
   if(!vp)
      return NULL;

   int copy_size;
   {
      void *vp1 = (byte*)vp_old - sizeof(S_mem_header) - MEM_GUARD_DWORDS*4;
      const S_mem_header &header = *(S_mem_header*)vp1;
      copy_size = Min((dword)sz, header.size);
   }
   memcpy(vp, vp_old, copy_size);
   DebugFree(vp_old);

   return vp;
}

//----------------------------



