#include <windows.h>
#include <rules.h>
#include <assert.h>

//----------------------------

#pragma warning(disable:4073)
#pragma init_seg(lib)

//----------------------------

#define FILL_MASK 0xffcecece  //float NAN, integer garbage, pointer garbage

//----------------------------
                              //fill local vars by garbage
                              // requires:
                              // - Gh compiler switch
                              // - disable optimizations (Debug build)

extern"C"
void __cdecl _penter();

__declspec(naked) void __cdecl _penter(){

   __asm{

      push edi
      push ecx
      push eax

      mov edi, esp
      mov ecx, 0xfc           //keep upper 24 bits clear
                              //try to extract size of local vars
      mov eax, [esp+12]
      cmp word ptr[eax+3], 0xec83   //check for sub esp, imm8
      jne no_sz_1
      mov cl, [eax+5]
      jmp ok_sz

   no_sz_1:
      cmp word ptr[eax+3], 0xec81   //check for sub esp, imm32
      jnz no_sz_2
      mov ecx, [eax+5]
      jmp ok_sz

   no_sz_2:
                              //check for mov eax, imm32 / call
      cmp byte ptr[eax+3], 0xb8
      jne no_sz_3
      cmp byte ptr[eax+8], 0xe8
      jne no_sz_3
      mov ecx, [eax+4]
                              //for now, limit fill size to 4KB (todo later: implement __chkstk)
      cmp ecx, 0xff0
      jb no_sz_3
      mov ecx, 0xff0

      jmp ok_sz

   no_sz_3:
      jmp skip_fill

   ok_sz:
      sub edi, ecx
      shr ecx, 2
      mov eax, FILL_MASK
      rep stosd

   skip_fill:
      pop eax
      pop ecx
      pop edi
      mov dword ptr [esp-4], FILL_MASK
      mov dword ptr [esp-8], FILL_MASK
      mov dword ptr [esp-12], FILL_MASK

      ret
   }
}

//----------------------------

typedef void __declspec(dllexport) *__cdecl t_DebugAlloc(size_t);
typedef void __declspec(dllexport) __cdecl t_DebugFree(void*);
typedef void __declspec(dllexport) *__cdecl t_DebugRealloc(void*, size_t);
typedef void __declspec(dllexport) __cdecl t_SetBreakAlloc(dword);
typedef void __declspec(dllexport) __cdecl t_SetBreakAllocPtr(void*);

t_DebugAlloc *_DebugAlloc;
t_DebugFree *_DebugFree;
t_DebugRealloc *_DebugRealloc;
t_SetBreakAlloc *_SetBreakAlloc;
t_SetBreakAllocPtr *_SetBreakAllocPtr;

//----------------------------

static class C_auto_init{
public:
   C_auto_init(){

      HINSTANCE hi = LoadLibrary("DebugMem.dll");
      if(hi){
         _DebugAlloc = (t_DebugAlloc*)GetProcAddress(hi, "DebugAlloc");
         assert(_DebugAlloc);
         _DebugFree = (t_DebugFree*)GetProcAddress(hi, "DebugFree");
         assert(_DebugFree);
         _DebugRealloc = (t_DebugRealloc*)GetProcAddress(hi, "DebugRealloc");
         assert(_DebugRealloc);
         _SetBreakAlloc = (t_SetBreakAlloc*)GetProcAddress(hi, "SetBreakAlloc");
         assert(_SetBreakAlloc);
         _SetBreakAllocPtr = (t_SetBreakAllocPtr*)GetProcAddress(hi, "SetBreakAllocPtr");
         assert(_SetBreakAllocPtr);
      }
   }
} auto_init;

//----------------------------

void *__cdecl operator new(size_t sz){

   if(!sz)
      return NULL;
   return _DebugAlloc(sz);
}

//----------------------------

void __cdecl operator delete(void *vp){

   if(!vp)
      return;
   _DebugFree(vp);
}

//----------------------------

void *DebugAlloc(size_t sz){
   if(!sz)
      return NULL;
   return _DebugAlloc(sz);
}

//----------------------------

void DebugFree(void *vp){

   if(!vp)
      return;
   _DebugFree(vp);
}

//----------------------------

void *DebugRealloc(void *vp, size_t sz){

   if(!vp)
      return DebugAlloc(sz);
   if(!sz){
      DebugFree(vp);
      return NULL;
   }
   return _DebugRealloc(vp, sz);
}

//----------------------------

void SetBreakAlloc(dword id){
   _SetBreakAlloc(id);
}

//----------------------------

void SetBreakAllocPtr(void *vp){
   _SetBreakAllocPtr(vp);
}

//----------------------------



