#ifndef __CAMERA_H
#define __CAMERA_H

#include "frame.h"
#include "sector.h"


#ifdef _MSC_VER
using namespace std;
#endif

//----------------------------

class I3D_camera: public I3D_frame{

#define CAMF_RESET_PROJ    2
#define CAMF_RESET_VIEW    4
#define CAMF_ORTHOGONAL    8

   S_matrix m_proj;
   S_matrix m_view;
                              //perspective transform
   float fov_a, ncp, fcp;     //fov angle & clipping planes
                              //orthogonal transform
   float ortho_scale;

   dword cam_flags;
   C_smart_ptr<I3D_sector> curr_sector;  //current sector this camera is in

   I3D_camera(PI3D_driver);
   ~I3D_camera();

   friend I3D_driver;        //creation
public:
   inline const S_matrix &GetViewMatrix1() const{ return m_view; }
   inline const S_matrix &GetProjectionMatrix1() const{ return m_proj; }
   inline void GetRange1(float &n, float &f) const{
      n = ncp;
      f = fcp;
   }
   inline float GetNCP() const{ return ncp; }
   inline float GetFCP() const{ return fcp; }
   inline void ResetProjectionMatrix(){ cam_flags |= CAMF_RESET_PROJ; }
   inline void ResetViewMatrix(){ cam_flags |= CAMF_RESET_VIEW; }
                              //sectors
   inline void SetCurrSector(PI3D_sector s){ curr_sector = s; }
   inline PI3D_sector GetCurrSector() const{ return curr_sector; }
   inline float GetFOV1() const{ return fov_a; }
   inline float GetOrthoScale1() const{ return ortho_scale; }
//----------------------------
// Update camera's matrices - m_proj and m_view.
   void UpdateMatrices(float aspect_ratio);

//----------------------------
// Mirror camera - put behind mirror plane.
   void Mirror(const S_plane&);

   void Draw1(PI3D_scene, bool strong) const;
public:
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD(DebugDraw)(PI3D_scene) const;
   virtual void SetMatrixDirty(){
      ResetViewMatrix();
      I3D_frame::SetMatrixDirty();
   }
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
                              //setup:
   I3DMETHOD(SetFOV)(float);
   I3DMETHOD(SetRange)(float near1, float far1);
   I3DMETHOD(SetOrthoScale)(float scale);
   I3DMETHOD(SetOrthogonal)(bool);

   I3DMETHOD_(float,GetFOV)() const{ return fov_a; }
   I3DMETHOD_(void,GetRange)(float &near1, float &far1) const;
   I3DMETHOD_(float,GetOrthoScale)() const{ return ortho_scale; }
   I3DMETHOD_(bool,GetOrthogonal)() const{ return (cam_flags&CAMF_ORTHOGONAL); }

   I3DMETHOD_(void,Update)();

   I3DMETHOD_(const S_matrix&,GetViewMatrix)() const;
   I3DMETHOD_(const S_matrix&,GetProjectionMatrix)() const;
   I3DMETHOD_(PI3D_sector,GetSector)() const;
};

//----------------------------

#endif
