#ifndef __COMMON_H
#define __COMMON_H

#include <sortlist.hpp>
#include <vector>

//----------------------------

//#define USE_EAX
#define USE_D3DX

//#define USE_PREFETCH

//----------------------------

#ifdef USE_EAX
                              //override some EAX includes
#define _LPCWAVEFORMATEX_DEFINED
#define __DSOUND_INCLUDED__
//#define LPWAVEFORMATEX void*
#define LPDIRECTSOUND void*
//#define _WAVEFORMATEX_
#include <eax.h>
#endif   //USE_EAX

#ifdef USE_D3DX
#include <D3dx8math.h>
#pragma comment(lib, "D3dx8")
#endif


#ifdef _MSC_VER
namespace std{}
using namespace std;
#endif


#ifdef _DEBUG
void D3D_Fatal(const char *text, dword hr);
#define CHECK_D3D_RESULT(text, hr) if(FAILED(hr)) D3D_Fatal(text, hr);
#else
#define CHECK_D3D_RESULT(text, hr)
#endif

//----------------------------

struct I3D_texel_collision{
   float u, v;
   class I3D_texture_base *tb;
   PI3D_material mat;
};

//----------------------------
                              //sorting bits for render primitives
                              //bits 29 - 31 - alpha blending type (signed)
#define PRIM_SORT_OPAQUE         0x80000000
#define PRIM_SORT_CKEY_ALPHA     0xa0000000
#define PRIM_SORT_ALPHA_ZWRITE   0xc0000000
#define PRIM_SORT_ALPHA_NOZWRITE 0xe0000000
#define PRIM_SORT_ALPHA_NOZ      0x00000000
                              //bits 15-28 for - material sorting ID
#define PRIM_SORT_MAT_MASK    0x1fffffff
#define PRIM_SORT_MAT_SHIFT   0
                              //bits 0-14 - distance (not combined with material - alpha only)
#define PRIM_SORT_DIST_MASK   0x1fffffff
#define PRIM_SORT_DIST_SHIFT  0

//----------------------------

enum{                         
   I3DCLIP_LEFT = 1,
   I3DCLIP_RIGHT = 2,
   I3DCLIP_BOTTOM = 4,
   I3DCLIP_TOP = 8,
   I3DCLIP_SIDES = I3DCLIP_LEFT | I3DCLIP_TOP | I3DCLIP_RIGHT | I3DCLIP_BOTTOM,
   I3DCLIP_BACK = 0x10,
   I3DCLIP_FRONT = 0x20,
   I3DCLIP_ALL = I3DCLIP_SIDES | I3DCLIP_FRONT | I3DCLIP_BACK
};

//----------------------------

#ifndef NDEBUG
extern int debug_count[4];
#endif

//----------------------------

#if defined _MSC_VER && 1

inline int FloatToInt(float f){
   __asm{
      fld f
      fistp f 
   }
   return *(int*)&f;
}
#else
inline int FloatToInt(float f){
   return (int)f;
}
#endif

inline int FloatAsInt(float f){
   return *(dword*)&f;
}

inline float IntAsFloat(int i){
   return (*(float*)&i);
}

                              //fast float cheats
#define CHECK_ZERO_GREATER(f) (!((*(dword*)&(f)) & 0x80000000))
#define CHECK_ZERO_LESS(f) ((*(dword*)&(f)) & 0x80000000)
#define CHECK_ZERO(f) (!*(dword*)&(f))
#define CHECK_SAME_SIGNS(f1, f2) (!(((*(dword*)&(f1)) ^ (*(dword*)&(f2))) & 0x80000000))
#define FLOAT_BITMASK(f) (*(int*)&(f))

//----------------------------
#ifdef USE_PREFETCH
                              //prefetching function
typedef void __fastcall t_Prefetch1(const void *ptr);
typedef void __fastcall t_Prefetch2(const void*, const void*);
typedef void __fastcall t_Prefetch3(const void*, const void*, const void*);
typedef void __fastcall t_Prefetch64(const void*);
extern t_Prefetch1 *Prefetch1;
extern t_Prefetch2 *Prefetch2ptr;
extern t_Prefetch3 *Prefetch3ptr;
extern t_Prefetch64 *Prefetch64Bytes;
#endif//USE_PREFETCH

//----------------------------

enum E_VIEW_FRUSTUM_SIDE{     //side indicied
   VF_LEFT,
   VF_TOP,
   VF_RIGHT,
   VF_BOTTOM,
   VF_BACK,
   VF_FRONT
};

//----------------------------

#define MAX_FRUSTUM_CLIP_PLANES 32

struct S_view_frustum{
   S_plane clip_planes[MAX_FRUSTUM_CLIP_PLANES];
   int num_clip_planes;
};

//----------------------------

struct S_expanded_view_frustum{
   float nx[MAX_FRUSTUM_CLIP_PLANES];
   float ny[MAX_FRUSTUM_CLIP_PLANES];
   float nz[MAX_FRUSTUM_CLIP_PLANES];
   float d[MAX_FRUSTUM_CLIP_PLANES];
   int num_clip_planes;

//----------------------------
// Make expanded frustum from normal one.
   void Make(const S_view_frustum&);
};

//----------------------------

enum{
   E_PREP_NO_VISUALS = 1,     //no visuals in this branch are being added to render list
   E_PREP_HR_NO_CLIP = 2,     //hierarchy is 100% on screen - no clipping, no bbox testing
   E_PREP_TEST_OCCLUDERS = 4, //test against occluders
   E_PREP_ADD_SHD_CASTERS = 8,//add shadow casters list
   E_PREP_SET_HR_LIGHT_DIRTY = 0x10,//hierarchically set FRMFLAGS_HR_LIGHT_RESET flag on processed children
};

                              //mode how we preprocess, add primitives and draw primitives
enum E_RENDERVIEW_MODE{
   RV_NORMAL,                 //normal mode
   RV_MIRROR,                 //from behind mirror
   RV_SHADOW_CASTER,          //shadow caster from light's position
   RV_CAM_TO_TEXTURE,         //CamView visual rendering into texture
};

struct S_render_primitive{
   PI3D_visual vis;
   dword sort_value;
   bool clip;
   dword blend_mode;
   PI3D_sector sector;       //current sector
   PI3D_scene scene;         //current scene
   dword user;
   int curr_auto_lod;

   S_render_primitive(int auto_lod):
      curr_auto_lod(auto_lod)
   {}
};

//----------------------------

struct S_preprocess_context{

                              //counts:
   dword opaque;              //opaque/colorkey
   dword ckey_alpha;          //1-bit alpha, z-writes, alpharef
   dword alpha_zwrite;        //alpha with z-writes
   dword alpha_nozwrite;      //alpha, no z-writes
   dword alpha_noz;           //alpha, no z-buffering

                              //read-only values
   bool clip;
   PI3D_scene scene;
   void *render_context;      //value passed to I3D_scene::Render
   vector<S_render_primitive> prim_list;
   C_sort_list<dword> prim_sort_list;
   PI3D_sector curr_sector;
   PI3D_frame curr_mirror;
   S_vector viewer_pos;
                              //frames which may cast shadows
   vector<PI3D_model> shadow_casters;
                              //frames which may potentionally receive shadow
   vector<PI3D_visual> shadow_receivers;

   dword render_view_id;      //ID of rendered view (unique for each call to RenderView)
   dword render_flags;

   float LOD_factor;          //1.0f is default under FOV=65 deg and ideal conditions

//----------------------------
// Sort primitives by distance.
   void SortByDistance();

//----------------------------
// Sort primitives by sort value.
   void Sort();

   void Reset();

   inline void Reserve(int num_items){
      prim_list.reserve(num_items);
      shadow_casters.reserve(32);
      shadow_receivers.reserve(num_items/2);
   }

                              //list of occluders, built by PreProcess,
                              // used in PreprocessChildren
   vector<class I3D_occluder*> occ_list;

                              //view frustum volume
   S_view_frustum view_frustum;
   //S_expanded_view_frustum exp_frustum;
   I3D_bsphere vf_sphere;
   E_RENDERVIEW_MODE mode;

   S_preprocess_context(PI3D_scene s, dword rid, const S_vector &vp):
      scene(s),
      viewer_pos(vp),
      render_view_id(rid),
      mode(RV_NORMAL)
   {
      Reset();
   }
};

//----------------------------

#if defined _MSC_VER && 0
bool __stdcall SphereCollide(const I3D_bsphere &vf_bs, const I3D_bsphere &bs2);
#else
inline bool SphereCollide(const I3D_bsphere &vf_bs, const I3D_bsphere &bs2){
   S_vector dir = vf_bs.pos - bs2.pos;
   float dist_2 = dir.Dot(dir);
   float radius_sum_2 = vf_bs.radius + bs2.radius; radius_sum_2 *= radius_sum_2;
   return (dist_2 < radius_sum_2);
}
#endif

//----------------------------

#ifdef _MSC_VER

inline dword Rol(dword dw, byte b){
   __asm{
      push ecx
      mov cl, b
      rol dw, cl
      pop ecx
   }
   return dw;
}
inline dword Ror(dword dw, byte b){
   __asm{
      push ecx
      mov cl, b
      ror dw, cl
      pop ecx
   }
   return dw;
}

inline int FindLastBit(dword val){
   dword rtn;
   __asm{
      mov eax, val
      bsr eax, eax
      jnz ok
      mov eax, -1
ok:
      mov rtn, eax
   }
   return rtn;
}

#endif

//----------------------------

dword CountBits(dword val);

//----------------------------

typedef bool __stdcall t_SphereInFrustum(const S_expanded_view_frustum&, const I3D_bsphere&, bool &clip);
extern t_SphereInFrustum *SphereInFrustum;

//----------------------------
// Check if sphere (in world coords) is inside of viewing frustum. If not, return also
// clip flag.
#ifdef _MSC_VER

bool __stdcall SphereInVF(const S_view_frustum&, const I3D_bsphere&, bool &clip);

#else //_MSC_VER

bool SphereInVF(const S_view_frustum&, const I3D_bsphere&, bool &clip);

#endif//!_MSC_VER

//----------------------------
// Check if all of input vertices are inside of viewing frustum.
bool AreAllVerticesInVF(const S_view_frustum&, const S_vector *verts, dword num_verts);

//----------------------------
// Check if any of input vertices is inside of viewing frustum.
bool IsAnyVertexInVF(const S_view_frustum&, const S_vector *verts, dword num_verts);

//----------------------------
// Compute contour of axis-aligned bounding-box, tranformed by 'tm',
// as seen from 'look'.
// If 'orthogonal' is false, the 'look' is a position point, otherwise it is direction vector.
// On return, 'contour_points' contain contour points in world coordinates,
// in clockwise order, and 'num_cpts' contains number of points.
// If number of points is 0, look_pos is inside of bound box, otherwise number of points
// is eigher 4 (one side visible) or 6 (two or three sides visible).
void ComputeContour(const I3D_bbox &bb_in, const S_matrix &tm, const S_vector &look,
   S_vector contour_points[6], dword &num_cpts, bool orthogonal = false);

//----------------------------
// Detect collision of frustum with convex (planar) face.
bool CheckFrustumIntersection(const S_plane *planes, dword num_planes,
   const S_vector *verts, dword num_verts);

//----------------------------
                              //plugin registration
struct S_visual_property{
   I3D_PROPERTYTYPE prop_type;
   const char *prop_name;
};

typedef PI3D_visual t_CreateVisual(PI3D_driver drv);
void RegisterVisual(t_CreateVisual *create_proc, dword visual_type,
   const char *friendly_name, const S_visual_property* = NULL, dword num_props = 0);

//----------------------------

void TransformVertexArray(const S_vector *in_verts, dword in_stride, dword num_verts,
   S_vector *verts_out, dword out_stride, const S_matrix &m);

//----------------------------
// Compute size of vertex from given FVF code.
dword GetSizeOfVertex(dword fvf_flags);

//----------------------------

//----------------------------
// Find specified 32-bit pointer in array of pointers,
// return index of pointer, or -1 if pointer not found
//----------------------------
#ifdef _MSC_VER

#pragma warning(disable:4035)
inline int FindPointerInArray(void **vp, int array_len, void *what){
   __asm{
      push ecx
      mov edi, vp
      mov ecx, array_len
      mov eax, what
      mov edx, ecx
      repne scasd
      jz ok
      mov edx, ecx
   ok:
      sub edx, ecx
      lea eax, [edx-1]
      pop ecx
   }
}
#pragma warning(default:4035)

#else

inline int FindPointerInArray(void **vp, int array_len, void *what){
   for(int i(array_len); i--; ) if(vp[i]==what) break;
   return i;
}

#endif

//----------------------------
                              //enumeration - internal
#define ENUMF_MODEL_ORIG   0x80000000  //use model's original name

//----------------------------
// Template class for locking and unlocking D3D vertex buffer.
template<class T>
class D3D_lock{
   T *ptr;                    //NULL if buffer not locked
   IDirect3DVertexBuffer8 *vb;
   dword lock_index, num_items;
   dword size_of_item;
   dword flags;
public:
   inline operator T *(){ return ptr; }
   inline operator T &(){ return *ptr; }
   inline T &operator [](int i){ return ptr[i]; }

//----------------------------
// Initialize lock - save values for later lock.
   inline D3D_lock(IDirect3DVertexBuffer8 *vbi, dword li, dword nv, dword sz, dword fl):
      ptr(NULL), vb(vbi), lock_index(li), num_items(nv), size_of_item(sz), flags(fl)
   {
   }

   inline ~D3D_lock(){ Unlock(); }

//----------------------------
// Lock, if not yet, using saved values.
   inline bool AssureLock(){
      if(!ptr){
         HRESULT hr;
         hr = vb->Lock(
               //0,       //offset (in bytes)
               lock_index * size_of_item,
               //0,             //size (in bytes)
               num_items * size_of_item,
               (byte**)&ptr,
               flags);
         CHECK_D3D_RESULT("Lock", hr);
         /*
         if(SUCCEEDED(hr)){
            (byte*&)ptr += lock_index * size_of_data;
         }
         */
      }
      return (ptr != NULL);
   }

//----------------------------
// Unlock (possibly) locked buffer.
   inline void Unlock(){
      if(ptr){
         HRESULT hr;
         hr = vb->Unlock();
         CHECK_D3D_RESULT("Unlock", hr);
         ptr = NULL;
      }
   }
};

//----------------------------

struct I3D_face{
   dword num_points;
   word *points;
   I3D_face(): num_points(0), points(NULL){}
   I3D_face(dword num): num_points(num), points(num ? new word[num] : NULL){}
   I3D_face(const I3D_face &f): points(NULL){ operator =(f); }
   ~I3D_face(){ delete[] points; }
   I3D_face &operator =(const I3D_face &f){
      delete[] points;
      points = new word[num_points = f.num_points];
      memcpy(points, f.points, num_points*sizeof(word));
      return *this;
   }
   void Reserve(dword num){
      if(num_points!=num){
         delete[] points;
         points = new word[num_points = num];
      }
   }
   inline word &operator [](int i){ return points[i]; }
   inline const word &operator [](int i) const{ return points[i]; }
};

//----------------------------

struct I3D_cylinder{
   S_vector pos;              //position of the base
   S_vector dir;              //direction and length of main axis
   float radius;              //radius
};

//----------------------------
                              //we use smartpointers for all collision containers
                              // so that volumes are not gone during tests
typedef vector<C_smart_ptr<I3D_volume> > t_volume_container;

//----------------------------// collisions //----------------------------//

//----------------------------
// Detect collision of ray with bounding box.
// Parameters:
//    mat - matrix for bbox transformation
//    bb - bounding box
//    from, norm_dir - ray
//    len - length of ray used for testing
//    hit_dist_ret - when not NULL, it is filled with collision distance if collision detected
// Return value:
//    true if collision was detected
// Note:
//    If 'from' is inside of bounding box, collision is detected and closest distance is set to 0.0
bool pickBBoxTest(const S_matrix &mat, const I3D_bbox &bb,
   const S_vector &from, const S_vector &norm_dir, const float len,
   float *hit_dist_ret = NULL);

//----------------------------
inline bool pickBSphereTest(const I3D_bsphere &bs1, const S_vector &from,
   const S_vector &normalized_dir, float closest_hit){

                              //check closest point on line
   float u = (bs1.pos-from).Dot(normalized_dir);
   if(u < - bs1.radius || u > (closest_hit+bs1.radius))
      return false;

                              //determine closest point to line
   S_vector point_on_line = from + normalized_dir * u;
   S_vector dir = point_on_line - bs1.pos;
   float dist_2 = dir.Dot(dir);
   if(dist_2 >= (bs1.radius*bs1.radius))
      return false;

   return true;
}

//----------------------------

//bool __stdcall FastTest_S_S(const S_vector &from, const S_vector &to, float &radius,
//   const S_vector &sphere_pos, float &sphere_radius);

//----------------------------

inline bool FastTest_L_S(struct I3D_collision_data &cd, const S_vector &sphere_pos, float &radius, float radius_2){

   S_vector &norm_dir = cd.normalized_dir;
                              //get closest point of sphere to line
   float u = norm_dir.Dot(sphere_pos - cd.from);
                              //check out-of-range
   if(u < -radius)
      return false;
   if(u > cd.GetClosestHit()+radius)
      return false;

                              //get point on line
   S_vector point_on_line = cd.from + norm_dir * u;
   S_vector dir_to_point = point_on_line - sphere_pos;
   float dist_2 = dir_to_point.Dot(dir_to_point);
   return (FLOAT_BITMASK(dist_2) < FLOAT_BITMASK(radius_2));
}

//----------------------------
// Check if point is in triangle. Triangle is specified by 3 points and its normal (non-normalized works).
// The point in question doesn't need to lie on the triangle, algorithm just checks if the point is
// out of edge planes (edge plane specified as perpendicular to edge and triangle's normal).
inline bool IsPointInTriangle(const S_vector &pt, const S_vector &t0, const S_vector &t1, const S_vector &t2,
   const S_vector &face_normal){

   S_vector n;
   float d;

   float x, y, z;
   {
      x = t0.x - t1.x;
      y = t0.y - t1.y;
      z = t0.z - t1.z;
      n.x = face_normal.y * z - face_normal.z * y;
      n.y = face_normal.z * x - face_normal.x * z;
      n.z = face_normal.x * y - face_normal.y * x;
      d = t0.Dot(n) - pt.Dot(n);
      if(CHECK_ZERO_LESS(d))
         return false;
   }

   {
      x = t1.x - t2.x;
      y = t1.y - t2.y;
      z = t1.z - t2.z;
      n.x = face_normal.y * z - face_normal.z * y;
      n.y = face_normal.z * x - face_normal.x * z;
      n.z = face_normal.x * y - face_normal.y * x;
      d = t1.Dot(n) - pt.Dot(n);
      if(CHECK_ZERO_LESS(d))
         return false;
   }
   {
      x = t2.x - t0.x;
      y = t2.y - t0.y;
      z = t2.z - t0.z;
      n.x = face_normal.y * z - face_normal.z * y;
      n.y = face_normal.z * x - face_normal.x * z;
      n.z = face_normal.x * y - face_normal.y * x;
      d = t2.Dot(n) - pt.Dot(n);
      if(CHECK_ZERO_LESS(d))
         return false;
   }
   return true;
}

inline bool IsPointInTriangle(const S_vector &pt, const S_vector &t0, const S_vector &t1, const S_vector &t2,
   const S_vector &face_normal, float radius){

   S_vector n;
   float d;

   {
      S_vector v(t0-t1);
      n.x = face_normal.y * v.z - face_normal.z * v.y;
      n.y = face_normal.z * v.x - face_normal.x * v.z;
      n.z = face_normal.x * v.y - face_normal.y * v.x;
      d = t0.Dot(n) - pt.Dot(n);
      if(d < -radius)
         return false;
   }

   {
      S_vector v(t1-t2);
      n.x = face_normal.y * v.z - face_normal.z * v.y;
      n.y = face_normal.z * v.x - face_normal.x * v.z;
      n.z = face_normal.x * v.y - face_normal.y * v.x;
      d = t1.Dot(n) - pt.Dot(n);
      if(d < -radius)
         return false;
   }
   {
      S_vector v(t2-t0);
      n.x = face_normal.y * v.z - face_normal.z * v.y;
      n.y = face_normal.z * v.x - face_normal.x * v.z;
      n.z = face_normal.x * v.y - face_normal.y * v.x;
      d = t2.Dot(n) - pt.Dot(n);
      if(d < -radius)
         return false;
   }
   return true;
}

//----------------------------

struct S_sound_env_properties{
                              //EAX properties:
   float room_level_lf;       //room effect level - low  freqencies, ratio (0.0 ... 1.0)
   float room_level_hf;       //room effect level - high freqencies, ratio (0.0 ... 1.0)
   int decay_time;            //reverberation decay time, in msec
   float decay_hf_ratio;      //HF to LF decay time, ratio (0.0 ... 1.0)
   float reflections;         //early reflections level - relative to room effect, ratio (0.0 ... 1.0)
   int reflections_delay;     //initial reflection delay time, in msec
   float reverb;              //late reverberation level - relative to room effect, (0.0 ... 1.0)
   int reverb_delay;          //late reverberation delay time relative to initial reflection, in msec
   float env_diffusion;       //environment diffusion, ratio (0.0 ... 1.0)

                              //other properties:
   //int switch_fade_time;      //time of volume fade between swithing
   C_string name;
   float emulation_volume;    //volume of sector if EAX is not in use

//----------------------------
// Setup default values
//----------------------------
   void SetupDefaults();

//----------------------------
// Compute blend between 2 input data. Parameter 'blend' is in range 0.0f (full ssp1)
// to 1.0f (full ssp2).
//----------------------------
   void MakeBlend(const S_sound_env_properties &ssp1, const S_sound_env_properties &ssp2,
      float blend, bool use_eax = true);

   S_sound_env_properties(){
      SetupDefaults();
   }
   S_sound_env_properties(const S_sound_env_properties &ep){ operator =(ep); }
   S_sound_env_properties &operator =(const S_sound_env_properties &ep){
      room_level_lf = ep.room_level_lf;
      room_level_hf = ep.room_level_hf;
      decay_time = ep.decay_time;
      decay_hf_ratio = ep.decay_hf_ratio;
      reflections = ep.reflections;
      reflections_delay = ep.reflections_delay;
      reverb = ep.reverb;
      reverb_delay = ep.reverb_delay;
      env_diffusion = ep.env_diffusion;
      name = ep.name;
      emulation_volume = ep.emulation_volume;
      return *this;
   }
};

//----------------------------

                              //located in DLL:
extern"C"{
I3D_RESULT __declspec(dllexport) __stdcall I3DCreate(PI3D_driver *drvp, const char *dll_filename);
}

//----------------------------
//----------------------------

#endif
