#ifndef __CONTAINER_H
#define __CONTAINER_H


//----------------------------

struct S_container_vcall{
   typedef I3D_RESULT (I3DAPI C_unknown::*t_Open)(const char*, dword, PI3D_LOAD_CB_PROC, void*, PI3D_scene);
   typedef void (I3DAPI C_unknown::*t_Close)();
   typedef I3D_RESULT (I3DAPI C_unknown::*t_AddFrame)(PI3D_frame);
   typedef I3D_RESULT (I3DAPI C_unknown::*t_DeleteFrame)(PI3D_frame);

   t_Open cbOpen;
   t_Close cbClose;
   t_AddFrame cbAddFrame;
   t_DeleteFrame cbDeleteFrame;
};

//----------------------------
                              //container - keeping references to frames and interpolators,
                              // allows for loading scene files
class I3D_container{
   vector<C_smart_ptr<I3D_interpolator> > intps;

   I3D_container &operator =(const I3D_container&);

   friend class I3D_scene;
   friend class I3D_model;

                              //callbacks to master interface
   C_unknown *master_interface;
   const S_container_vcall *vcall;
protected:
   vector<C_smart_ptr<I3D_frame> > frames;
   C_string filename;         //loaded file

//----------------------------
// Clear contents of animator.
   void ClearInterpolators(){
      intps.clear();
   }

//----------------------------
// Open scene internal.
   I3D_RESULT Open(S_load_context &lc);
public:
   I3D_container():
      master_interface(NULL)
   {}

   inline dword NumInterpolators1() const{ return intps.size(); }
   inline const PI3D_interpolator *GetInterpolators1() const{ return (PI3D_interpolator*)intps.begin(); }

   inline const vector<C_smart_ptr<I3D_interpolator> > &GetInterpolatorVector() const{ return intps; }
   inline dword NumFrames1() const{ return frames.size(); }

   const vector<C_smart_ptr<I3D_frame> > &GetFrameVector() const{ return frames; }
   I3D_RESULT AddAnimFrame(PI3D_frame frm);
   I3D_RESULT DeleteAnimFrame(PI3D_frame frm);

public:
   I3DMETHOD_(I3D_RESULT,Open)(const char *fname, dword flags, PI3D_LOAD_CB_PROC cb_proc,
      void *context, PI3D_scene scene_hlp){
      return (master_interface->*vcall->cbOpen)(fname, flags, cb_proc, context, scene_hlp);
   }
   I3DMETHOD_(void,Close)(){
      (master_interface->*vcall->cbClose)();
   }
   I3DMETHOD_(const char*,GetFileName)() const{ return filename; }
   I3DMETHOD_(void,StoreFileName)(const char *fn){ filename = fn; }

   I3DMETHOD_(I3D_RESULT,AddFrame)(PI3D_frame frm){ return (master_interface->*vcall->cbAddFrame)(frm); }
   I3DMETHOD_(I3D_RESULT,DeleteFrame)(PI3D_frame frm){ return (master_interface->*vcall->cbDeleteFrame)(frm); }

   I3DMETHOD_(dword,NumFrames)() const{ return frames.size(); }
   I3DMETHOD_(const PI3D_frame*,GetFrames)() const{ return (const PI3D_frame*)frames.begin(); }

   I3DMETHOD_(I3D_RESULT,AddInterpolator)(PI3D_interpolator);
   I3DMETHOD_(I3D_RESULT,DeleteInterpolator)(PI3D_interpolator);
   I3DMETHOD_(dword,NumInterpolators)() const{ return NumInterpolators1(); }
   I3DMETHOD_(const PI3D_interpolator*,GetInterpolators)() const{ return GetInterpolators1(); }

   I3DMETHOD_(I3D_RESULT,Tick)(int time);
};

//----------------------------

#endif
