#include <rules.h>
#include <c_string.hpp>
#include <c_cache.h>
#include <vector>
#include <map>
#include <new.h>

using namespace std;

//----------------------------

struct S_db_header{
   dword ID;
   dword num_blocks;
   dword dir_block;
   dword dir_size;
};

//----------------------------

struct S_record_handle{
   C_cache cache;          //cache returned to user
   byte *mem;
   size_t mem_size;
   byte *alloc_mem;
   C_string name;
   bool write;
   dword file_time[2];     //valid only when 'write' == true

   S_record_handle():
      alloc_mem(NULL)
   {}
   S_record_handle(const S_record_handle&);
   S_record_handle &operator =(const S_record_handle&);
   ~S_record_handle(){
      delete[] alloc_mem;
   }
};

//----------------------------

class C_database{
   int file_handle;           //-1 = not opened
   dword curr_file_ptr;
   vector<dword> alloc_map;
   vector<S_record_handle*> open_handles;
   dword dir_block;           //first block of directory
   dword dir_size;            //in bytes, as saved on disk
   void *_directory;          //hidden implementation of directory
   bool alloc_map_dirty;
   bool directory_dirty;
   bool in_directory_write;

   dword AllocSingleBlock(dword prev_block);
   bool FreeBlocks(dword first_block);
   dword WriteSingleBlock(dword prev_block, const void *mem, dword block_size);

//----------------------------
// Read single block of data. The returned value is an index of next block.
   dword ReadSingleBlock(dword block, void *mem, dword block_size);

   bool WriteAllocMap();
   bool CreateDirectoryEntry(const C_string &name, dword first_block, dword size, const dword file_time[2]);
   bool ReadDirectory(int max_keep_days);
   bool WriteDirectory();
   bool EraseRecord(const C_string&, bool write_back);

//----------------------------
// Read 'mem_size' bytes beginning on specified 'block' into 'mem'.
   bool ReadData(dword block, void *mem, dword mem_size);
//----------------------------
// Write 'mem_size' bytes beginning on 'mem' into file, return beg block in 'beg_block'.
   bool WriteData(const void *mem, dword mem_size, dword *beg_block);
public:
   C_database();
   ~C_database();

                              //open or create dbase file
   bool Open(const char *fname, dword max_size, int max_keep_days);
                              //close dbase
   bool Close();
   inline bool IsOpen() const{ return (file_handle!=-1); }
                              //flush in-memory data on disk
   void Flush();

                              //open record in dbase
   C_cache *OpenRecord(const C_string&, const dword file_time[2]);
                              //create new record (overwrite if exists)
   C_cache *CreateRecord(const C_string&, const dword file_time[2]);
                              //erase existing record
   bool EraseRecord(const C_string&);
                              //close opened record, flush data to disk
   bool CloseRecord(C_cache*);

   void GetStats(dword &bytes_total, dword &bytes_used);
};

//----------------------------