#ifndef __FLARE_H
#define __FLARE_H

#include "visual.h"


//----------------------------

class I3D_flare: public I3D_visual{

#define FLRF_COLLIDING        1        //currently some occlusion between flare and camera

   dword flare_flags;

                              //parameters:
   C_smart_ptr<I3D_material> mat;
   float scale_begin_dist;    //distance when scaling and fading starts
   float rotation_ratio;      //ratio of rotation, multiplied by distance and 2*PI gives rotation angle
   float fade_normal_angle;   //angle, when normal-based rotation fades out
   int occlusion_fade_time;   //fade time used for occlusion fading, if zero, occlusion tests are disabled

   int last_check_col_time;   //last time we checked collisions
   int occlusion_fade_count;  //set to 0 for invisible (occluded) flare, set to -1 for fully visible (not occluded) flare
   float pivot;               //pivot displacement along the view Z axis

                              
   struct S_user_context{     //struct passed between calls to AddPrimitives, DrawPrimitive
      S_matrix g_matrix;
      S_vectorw diffuse;
   };
public:
   I3D_flare(PI3D_driver d);

public:
                              //specialized methods of I3D_visual:
   virtual bool ComputeBounds();
   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
   I3DMETHOD(Duplicate)(CPI3D_frame);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD(SetMaterial)(PI3D_material mat1){
      mat = mat1;
      if(!mat)
         SetOn(false);
      return I3D_OK;
   }
   I3DMETHOD(SetScaleDisance)(float f){ scale_begin_dist = f; return I3D_OK; }
   I3DMETHOD(SetRotationRatio)(float f){ rotation_ratio = f; return I3D_OK; }
   I3DMETHOD(SetNormalAngle)(float f){ fade_normal_angle = f; return I3D_OK; }
   I3DMETHOD(SetOcclusionFade)(int i){ occlusion_fade_time = i; return I3D_OK; }
   I3DMETHOD(SetPivot)(float f){ pivot = f; return I3D_OK; }

   I3DMETHOD_(PI3D_material,GetMaterial)() const{ return mat; }
   I3DMETHOD_(float,GetScaleDisance)() const{ return scale_begin_dist; }
   I3DMETHOD_(float,GetRotationRatio)() const{ return rotation_ratio; }
   I3DMETHOD_(float,GetNormalAngle)() const{ return fade_normal_angle; }
   I3DMETHOD_(int,GetOcclusionFade)() const{ return occlusion_fade_time; }
   I3DMETHOD_(float,GetPivot)() const{ return pivot; }
};

typedef I3D_flare *PI3D_flare;

//----------------------------

#endif