#ifndef __JOINT_H
#define __JOINT_H

#include "frame.h"


//----------------------------

class I3D_joint: public I3D_frame{

#define JOINTF_BBOX_VALID        1     //bounding-box is valid
#define JOINTF_VBOX_VALID    2         //volume-box is valid
#define JOINTF_FIXED_POSITION    4     //position cannot be adjusted

   dword joint_flags;

   S_matrix bbox_matrix;      //bounding box' matrix relative to this joint (i.e. local matrix)
   I3D_bbox bbox;             //bounding-box used for SM vertex assignment
   float icon_size;           //size used to draw icon of this joint

   S_matrix volume_box_matrix;
   I3D_bbox volume_box;      //volume-box - used for collision detection

   I3D_joint(PI3D_driver drv);
   ~I3D_joint();

   friend I3D_driver;        //creation

public:
   void SetBBox(const I3D_bbox &bb1, const S_matrix &m1);

   void SetVolumeBox(const I3D_bbox &bb1, const S_matrix &m1){
      volume_box = bb1;
      volume_box_matrix = m1;
      joint_flags |= JOINTF_VBOX_VALID;
   }

   inline bool IsBBoxValid() const{ return (joint_flags&JOINTF_BBOX_VALID); }
   inline const I3D_bbox &GetBBox1() const{ return bbox; }
   inline const S_matrix &GetBBoxMatrix1() const{ return bbox_matrix; }

   inline bool IsVolumeBoxValid() const{ return (joint_flags&JOINTF_VBOX_VALID); }
   inline const I3D_bbox &GetVolumeBox1() const{ return volume_box; }
   inline const S_matrix &GetVolumeBoxMatrix1() const{ return volume_box_matrix; }

   inline float GetIconSize() const{ return icon_size; }

   void LockPosition(bool b){
      joint_flags &= JOINTF_FIXED_POSITION;
      if(b) joint_flags |= JOINTF_FIXED_POSITION;
   }

   void Draw1(PI3D_scene, bool strong) const;

   I3DMETHOD(DebugDraw)(PI3D_scene) const;
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD(SetPos)(const S_vector &v);
   virtual void UpdateMatrixInternal();
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
};

//----------------------------

#endif