#include <vector>
# include <d3d8.h>
#include "visual.h"


#define MAX_DYNLIGHT_RECTS 17 //maximal number of dynamically-lit rectangles

//----------------------------
                              //use 1 weight for rectangle selection (index)
#define FVF_VERTEX_LMAP (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1)

struct I3D_vertex_lmap{
   S_vector xyz;
   S_vector normal;
   I3D_text_coor tex[1];
};

//----------------------------

                              //morphing mesh
class I3D_mesh_lmap: public I3D_mesh_base{

public:
   I3D_mesh_lmap(PI3D_driver d):
      I3D_mesh_base(d, FVF_VERTEX_LMAP)
   {
   }

public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
};

//----------------------------

struct S_lmap_pixel{
   byte b, g, r;

   void Zero(){
      r = 0;
      g = 0;
      b = 0;
   };
   void Add(byte r1, byte g1, byte b1){
      r = Min(255, r+r1);
      g = Min(255, g+g1);
      b = Min(255, b+b1);
   }
   bool operator ==(const S_lmap_pixel &lmp){
      return (r==lmp.r && g==lmp.g && b==lmp.b);
   }
};

//----------------------------

class I3D_lit_object: public I3D_visual{

#define LMAPF_CONSTRUCTED     1     //set when LM rectanges are created and lit
#define LMAPF_DYNLIGHT_VALID  2     //dynamic light prepared properly (between AddPrims/DrawPrim calls)
#define LMAPF_DRAW_LIGHT_MAP  0x10  //light-map should be drawn, in addiction to vertex lighting

   dword lmap_flags;
   C_smart_ptr<I3D_mesh_base> mesh;

                              //light-mapping:
   struct S_lmap_rectangle: public C_unknown{
                              //matrix for rectangle representation,
                              // in lmap object's local coords
      S_matrix m;
      S_matrix inv_m;
      float z_delta;

      bool dirty;             //need re-loading into cached lmap

      dword size_x, size_y;
      S_lmap_pixel *dst_map;  //RGB light-map representation

      S_lmap_rectangle():
         dst_map(NULL),
         z_delta(0.0f),
         dirty(true)
      {}
      S_lmap_rectangle(const S_lmap_rectangle &l1);
      S_lmap_rectangle &operator =(const S_lmap_rectangle&);
      ~S_lmap_rectangle(){
         delete[] dst_map;
      }
   };
   vector<C_smart_ptr<S_lmap_rectangle> > lm_rects;
   vector<vector<word> > face_lmap_use;
   bool any_rect_dirty;       //set if any of lm_rects is dirty

   class C_LM_alloc_block *lm_alloc_block; //light-map, run-time allocated
   float curr_resolution;

   I3D_lit_object& operator =(const I3D_lit_object&);
   friend I3D_visual *CreateLitObject(PI3D_driver);

//----------------------------
// Allocate lightmap textures and upload info there.
   void UploadLMaps();

   void MapVertices();
   void CheckUV();
   I3D_RESULT MakeLight(PI3D_scene scene, dword flags, PI3D_LOAD_CB_PROC cb_proc, void *context);
protected:
   I3D_lit_object(PI3D_driver);
public:
   ~I3D_lit_object();

   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD_(void,Update)();
   I3DMETHOD(SetProperty)(dword index, dword value);
   I3DMETHOD_(dword,GetProperty)(dword index) const;
   I3DMETHOD_(I3D_mesh_base*,GetMesh)() const{ return mesh; }

   virtual bool ComputeBounds();
   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
   virtual void RenderSolidMesh(PI3D_scene, bool, const I3D_cylinder*);
   void SetMesh(PI3D_mesh_base);
   virtual void PrepareDestVB(I3D_mesh_base*, dword = 2);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD(Construct)(PI3D_scene, dword flags = 0, PI3D_LOAD_CB_PROC = NULL, void *context = NULL);
   I3DMETHOD(Destruct)(dword flags = 0);
   I3DMETHOD(Load)(int handle);
   I3DMETHOD(Save)(int handle);
   I3DMETHOD_(bool,IsConstructed)() const{
      return (lmap_flags&LMAPF_CONSTRUCTED);
   }
};

typedef I3D_lit_object *PI3D_lit_object;

//----------------------------
