/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Load_3ds.cpp
   Content: File loader for .3ds and .i3d files.
--------------------------------------------------------*/

#include "all.h"
#include <dta_read.h>
#include <c_cache.h>
#include <malloc.h>           //alloca
#include "object.h"
#include "material.h"
#include "scene.h"
#include "light.h"
#include "soundi3d.h"
#include "camera.h"
#include "sector.h"
#include "volume.h"
#include "visual.h"
#include "model.h"
#include "dummy.h"
#include "target.h"
#include "user.h"
#include "texture.h"
#include "particle.h"
#include "lmap.h"
#include "load_3ds.h"
#include "sng_mesh.h"
#include "occluder.h"
#include "flare.h"
#include "procedural.h"



#define USE_ALLOCA            //use alloca for local storage instead of new[] / delete[]
#define USE_DATABASE          //try to cache computation-expensive info in database
#define INIT_TEXTURE_ANIMATION   //load animated texture sequence
//#define CREATE_COLLISION_DATA //create collision data immediately
//#define REPORT_EMPTY_MESH_ERROR  //report meshes with no vertices or faces
//#define REPORT_INVALID_FACES  //report meshes with invalid (on line) faces
//#define REPORT_UNUSED_VERTS  //report meshes with unused vertices
#define FORCE_MIPMAPPING      //force using mipmapping, unless it is explicitly disabled
#define AUTO_SMOOTH_ANGLE (50.0f*PI/180.0f)  //angle of autosmoothing - used when smoothgroups are not present

                              //debug:
//#define PAINT_SMOOTH_NORMALS  //paint results of scene smoothing

//----------------------------
                              //3ds import - animation fps constant
                              //3DS 4 uses film time (24 fps), we use
                              //100 fps
#define ANIM_FPS 100

//----------------------------

template<class T>
static bool operator ==(const T &v1, const T &v2){
   return !memcmp(&v1, &v2, sizeof(T));
}

//----------------------------

PI3D_material S_mat_info::GetMaterial(const S_load_context &lc){

   if(!txt_created){
      if(ct_flags&(TEXTMAP_DIFFUSE|TEXTMAP_OPACITY)){

         I3D_CREATETEXTURE ct;
         memset(&ct, 0, sizeof(ct));
         ct.flags = ct_flags;

         if(ct_flags&TEXTMAP_DIFFUSE) ct.text_name = filename[0];
         if(ct_flags&TEXTMAP_OPACITY) ct.opt_name = filename[1];

         PI3D_texture tp;
         I3D_RESULT ir;
         ir = lc.driver->CreateTexture1(&ct, &tp, (lc.load_flags&I3DLOAD_ERRORS) ? lc.load_cb_proc : NULL, lc.load_cb_context);
         const char *cp, *txt_name = ct.text_name;
         switch(ir){
         case I3DERR_OUTOFMEM: cp = "not enough memory"; break;
         case I3DERR_NOFILE: cp = "file not found"; break;
         case I3DERR_NOFILE1:
            cp = "opacity map: file not found";
            txt_name = ct.opt_name;
            break;
         default: cp = NULL;
         }
         if(cp){
            lc.REPORT_ERR(C_fstring("cannot load texture '%s' (%s) (mat '%s')", txt_name, cp, mat->GetName()));
         }

#ifdef INIT_TEXTURE_ANIMATION
                              //check if animated texture requested
         if(tp && (load_flags&LOADF_TXT_DIFF_ANIM)){
            I3D_animated_texture *atp = CreateAnimatedTexture(filename[0], filename[1], lc,
               tp, ct_flags, diffuse_anim_speed);
            if(atp){
               tp->Release();
               tp = NULL;
               mat->SetTexture1(MTI_DIFFUSE, atp);
               atp->Release();
            }
         }
#endif                        //INIT_TEXTURE_ANIMATION
         if(tp){
            if(env_effect==EE_ALPHABLEND){
                              //use opacity map for environment blending, rather than for
                              // diffuse blending - clear alpha flags from texture
               tp->SetTxtFlags(tp->GetTxtFlags() & ~(TXTF_ALPHA|TXTF_SELF_ALPHA));
            }
            mat->SetTexture(tp);
            tp->Release();
         }
      }
      if(filename[4].Size()){
         PI3D_procedural_base pb = lc.driver->GetProcedural(filename[4], lc, NULL, mat->GetName());
         if(pb){
            switch(pb->GetID()){
            case PROCID_TEXTURE:
               {
                  I3D_procedural_texture *pt = (I3D_procedural_texture*)pb;
                  I3D_MATERIAL_TEXTURE_INDEX mti;
                  switch(pt->GetMode()){
                  case I3D_procedural_texture::MODE_BUMP:
                     mti = MTI_BUMP;
                     mat->SetBumpOpacity(.1f);
                     break;
                  default:
                     {
                        //LPI3D_texture curr_diffuse = 
                        mti = MTI_DIFFUSE;
                     }
                  }
                  mat->SetTexture1(mti, pt);
               }
               break;
            default:
                              //no, thanks, can't use this procedural for texture processing
               lc.REPORT_ERR(C_fstring("Invalid procedural for texture: '%s' (mat '%s')", (const char*)filename[0],
                  mat->GetName()));
            }
            pb->Release();
         }
      }
      if(load_flags&LOADF_TXT_CAMERA){
         I3D_texture_base *tp;
         tp = new I3D_camera_texture(lc.driver);
         mat->SetTexture(tp);
         /*
         I3D_CREATETEXTURE ct;
         memset(&ct, 0, sizeof(ct));
         ct.flags = ct_flags;

         if(ct_flags&TEXTMAP_DIFFUSE) ct.text_name = filename[0];
         if(ct_flags&TEXTMAP_OPACITY) ct.opt_name = filename[1];

         I3D_RESULT ir;
         ir = lc.driver->CreateTexture1(&ct, &tp, (lc.load_flags&I3DLOAD_ERRORS) ? lc.load_cb_proc : NULL, lc.load_cb_context);
         const char *cp, *txt_name = ct.text_name;
         switch(ir){
         case I3DERR_OUTOFMEM: cp = "not enough memory"; break;
         case I3DERR_NOFILE: cp = "file not found"; break;
         case I3DERR_NOFILE1:
            cp = "opacity map: file not found";
            txt_name = ct.opt_name;
            break;
         default: cp = NULL;
         }
         if(cp){
            lc.REPORT_ERR(C_fstring("cannot load texture '%s' (%s, frame '%s')", txt_name, cp,
               (const char*)C_fstring("\"%s\"", frame_name ? frame_name : "unknown")));
         }
         */
      }
      if(env_effect != EE_NULL){
         if(filename[2].Size()){
                              //load environment map
            I3D_CREATETEXTURE ct;
            memset(&ct, 0, sizeof(ct));
            ct.flags = TEXTMAP_DIFFUSE;
            ct.text_name = filename[2];

            PI3D_texture tp;
            I3D_RESULT ir;
            ir = lc.driver->CreateTexture1(&ct, &tp, (lc.load_flags&I3DLOAD_ERRORS) ? lc.load_cb_proc : NULL, lc.load_cb_context);
            const char *cp, *txt_name = ct.text_name;
            switch(ir){
            case I3DERR_OUTOFMEM: cp = "not enough memory"; break;
            case I3DERR_NOFILE: cp = "file not found"; break;
            case I3DERR_NOFILE1:
               cp = "reflection map: file not found";
               txt_name = ct.opt_name;
               break;
            default: cp = NULL;
            }
            if(cp){
               lc.REPORT_ERR(C_fstring("cannot load texture '%s' (%s) (mat '%s')", txt_name, cp, mat->GetName()));
            }

                              //check if animated texture requested
            if(tp && (load_flags&LOADF_TXT_REFL_ANIM)){
               I3D_animated_texture *atp = CreateAnimatedTexture(filename[2], NULL, lc,
                  tp, TEXTMAP_DIFFUSE, diffuse_anim_speed);
               if(atp){
                  tp->Release();
                  tp = NULL;
                  mat->SetTexture1(MTI_ENVIRONMENT, atp);
                  atp->Release();
               }
            }
            if(tp){
               mat->SetTexture1(MTI_ENVIRONMENT, tp);
               tp->Release();
            }
         }
         mat->SetEnvOpacity(environment_power);
         mat->SetEnvironmetEffect(env_effect);
      }
      if(filename[3].Size()){
                              //check conflict with alpha channel
         if(mat->IsCkeyAlpha1() || mat->IsTextureAlpha()){
            lc.REPORT_ERR(C_fstring("bumpmap conflict with texture alpha (mat '%s')",
               mat->GetName()));
         }else
         if(lc.driver->GetFlags()&DRVF_BUMPMAPPING){
                           //load bump map
            I3D_CREATETEXTURE ct;
            memset(&ct, 0, sizeof(ct));
            ct.flags = TEXTMAP_BUMPMAP;
            ct.bump_name = filename[3];

            PI3D_texture tp;
            I3D_RESULT ir;
            ir = lc.driver->CreateTexture1(&ct, &tp, (lc.load_flags&I3DLOAD_ERRORS) ? lc.load_cb_proc : NULL, lc.load_cb_context);
            const char *cp, *txt_name = ct.text_name;
            switch(ir){
            case I3DERR_OUTOFMEM: cp = "not enough memory"; break;
            case I3DERR_NOFILE: cp = "file not found"; break;
            case I3DERR_NOFILE1:
               cp = "bump map: file not found";
               txt_name = ct.opt_name;
               break;
            case I3DERR_TEXTURESNOTSUPPORTED: cp = "cannot find bumpmap texture format"; break;
            default: cp = NULL;
            }
            if(cp)
               lc.REPORT_ERR(C_fstring("cannot load texture '%s' (%s) (mat '%s')", txt_name,
               cp, mat->GetName()));

                              //check if animated texture requested
            if(tp && bump_anim_speed){
               I3D_animated_texture *atp = CreateAnimatedTexture(filename[3], NULL, lc,
                  tp, TEXTMAP_BUMPMAP, bump_anim_speed);
               if(atp){
                  tp->Release();
                  tp = NULL;
                  mat->SetTexture1(MTI_BUMP, atp);
                  atp->Release();
               }
            }
            if(tp){
               mat->SetTexture1(MTI_BUMP, tp);
               tp->Release();
            }
            mat->SetBumpOpacity(bump_power);
         }
      }

      txt_created = true;
   }
   return mat;
}

//----------------------------

I3D_animated_texture *S_mat_info::CreateAnimatedTexture(const char *fname0, const char *fname1,
   const S_load_context &lc, PI3D_texture tp_last, dword ct_flags,
   int anim_speed){

   I3D_animated_texture *atp = NULL;
                     
   const C_string &tn = fname0;
                              //find extension index
   for(int ei=tn.Size(); ei--; ){
      if(tn[ei] == '.') break;
   }
                              //no extension? ok, we'll use entire filename
   if(ei==-1) ei = tn.Size();
                              //find digit index
   for(int di=ei; di-- && isdigit(tn[di]); );
   ++di;


   const C_string &o_tn = fname1;
                              //find extension index
   for(int o_ei=o_tn.Size(); o_ei--; ){
      if(o_tn[o_ei] == '.') break;
   }
                              //no extension? ok, we'll use entire filename
   if(o_ei==-1) o_ei = o_tn.Size();
                              //find digit index
   for(int o_di=o_ei; o_di-- && isdigit(o_tn[o_di]); );
   ++o_di;

   if(di==ei){
                              //error - no digit found in filename
      lc.REPORT_ERR(C_fstring("animated texture '%s' (mat '%s') - no digits found at end of filename", fname0, mat->GetName()));
   }else{
      int max_num;
      int i = sscanf(&tn[di], (const char*)C_fstring("%%%iu", ei - di), &max_num);
      if(i==1 && max_num!=0){
                              //digit scanned successfully
         atp = new I3D_animated_texture(lc.driver);
         atp->SetAnimSpeed(anim_speed);

         I3D_CREATETEXTURE ct;
         memset(&ct, 0, sizeof(ct));
         ct.flags = ct_flags;
                              //load texture sequence
         C_string new_name = tn, o_new_name = o_tn;
         for(i=0; i<max_num; i++){
            sprintf(&new_name[di], C_fstring("%%.%ii", ei - di), i);
            new_name[ei] = tn[ei];

            if(ct_flags&TEXTMAP_DIFFUSE) ct.text_name = new_name;
            else
            if(ct_flags&TEXTMAP_BUMPMAP) ct.bump_name = new_name;

                              //make (optionally) opacity name
            if((ct_flags&TEXTMAP_OPACITY)){
               if(o_ei != o_di){
                  sprintf(&o_new_name[o_di], C_fstring("%%.%ii", o_ei - o_di), i);
                  o_new_name[o_ei] = o_tn[o_ei];
                  ct.opt_name = o_new_name;
               }else
                  ct.opt_name = o_tn;
            }

            PI3D_texture tp;
            I3D_RESULT ir =
               lc.driver->CreateTexture1(&ct, &tp, (lc.load_flags&I3DLOAD_ERRORS) ? lc.load_cb_proc : NULL, lc.load_cb_context);
            const char *cp, *txt_name = new_name;
            switch(ir){
            case I3DERR_OUTOFMEM: cp = "not enough memory"; break;
            case I3DERR_NOFILE: cp = "file not found"; break;
            case I3DERR_NOFILE1:
               cp = "opacity map: file not found";
               txt_name = ct.opt_name;
               break;
            default: cp = NULL;
            }
            if(cp){
               lc.REPORT_ERR(C_fstring("cannot load texture '%s' (%s, mat '%s')", txt_name, cp, mat->GetName()));
            }
            if(tp){
               atp->AddTexture(tp);
               tp->Release();
            }
         }
         atp->AddTexture(tp_last);
      }
   }
   return atp;
}

//----------------------------

enum E_PROP_KEYWORD{          //frame properties keyword IDs
   PROP_UNKNOWN,
                              //casting:
   PROP_CAST_SNGM,
   PROP_CAST_PRTC,
   PROP_CAST_LMAP,
   PROP_CAST_BBRD,
   PROP_CAST_OCCL,
   PROP_CAST_SECT,
   PROP_CAST_MRPH,
   PROP_CAST_FLRE,
   PROP_CAST_OUVS,

   PROP_CAST_JTVB,
   PROP_CAST_JTBB,
   PROP_REGION,

   PROP_CAST_VOLS,
   PROP_CAST_VOLR,
   PROP_CAST_VOLB,

                              //modifiers:
   PROP_HELP,
   PROP_SIMPLE_SECTOR,
   PROP_SMOOTH,
   PROP_AUTO_LOD,
   PROP_ENVIRONMENT,
   PROP_HIDE,
   PROP_MATERIAL,
   PROP_PROCEDURAL,
};

static struct S_PROP_keyword{ //properties keywords
   E_PROP_KEYWORD id;
   const char *keyword;
} prop_keywords[] = {
                              //casts:
   {PROP_CAST_SNGM, "SNGM"},
   {PROP_CAST_PRTC, "PRTC"},
   {PROP_CAST_LMAP, "LMAP"},
   {PROP_CAST_BBRD, "BBRD"},
   {PROP_CAST_OCCL, "OCCL"},
   {PROP_CAST_SECT, "SECT"},
   {PROP_CAST_MRPH, "MRPH"},
   {PROP_CAST_FLRE, "FLRE"},
   {PROP_CAST_OUVS, "OUVS"},

   {PROP_CAST_VOLS, "VOLS"},
   {PROP_CAST_VOLR, "VOLR"},
   {PROP_CAST_VOLB, "VOLB"},

   {PROP_CAST_JTBB, "JTBB"},
   {PROP_CAST_JTVB, "JTVB"},
   {PROP_REGION, "region"},
                              //modifiers:
   {PROP_HELP, "help"},
   {PROP_SIMPLE_SECTOR, "simple"},
   {PROP_SMOOTH, "smooth"},
   {PROP_AUTO_LOD, "auto_lod"},
   {PROP_HIDE, "hide"},
   {PROP_ENVIRONMENT, "environment"},
   {PROP_MATERIAL, "material"},
   {PROP_PROCEDURAL, "procedural"},

   {PROP_UNKNOWN}             //must be last!
};

//----------------------------
// Get next keyword from property stream, return it's type and return also
// new pointer into property stream (optional),                              
//----------------------------
static E_PROP_KEYWORD GetPropertyKeyword(const char *cp, const char **ptr_out){

   if(ptr_out) *ptr_out = NULL;
   if(!cp){                   //NULL properties
      return PROP_UNKNOWN;
   }
   int slen;
   char keyword[256];
                              //take care of 1st comma
   for(slen=0; slen<sizeof(keyword) && cp[slen]; slen++){
      keyword[slen] = cp[slen];
      if(keyword[slen]==',' || isspace(keyword[slen])){
         keyword[slen] = 0;
         break;
      }
   }
   keyword[slen] = 0;
   if(ptr_out){
      *ptr_out = cp + slen;
      while(isspace(**ptr_out) || (**ptr_out)==',') ++(*ptr_out);
      if(!(**ptr_out)) *ptr_out = NULL;    //end of input
   }
   if(slen<3) return PROP_UNKNOWN;
   for(S_PROP_keyword *pk = prop_keywords; pk->id!=PROP_UNKNOWN; ++pk)
      if(!strnicmp(pk->keyword, cp, slen)) break;
   return pk->id;
}

//----------------------------
// Convert keyword back to string.
//----------------------------
static const char *GetKeywordString(E_PROP_KEYWORD id){

   for(int i=0; prop_keywords[i].id!=PROP_UNKNOWN; i++)
   if(prop_keywords[i].id==id){
      return prop_keywords[i].keyword;
   }
   return NULL;
}

//----------------------------
// Get first keyword in frame's properties, or PROP_UNKNOWN if not found.
// 'rtn' contains info for getting next data.
// To get further keywords, use GetPropertiesNext.
//----------------------------
static bool GetPropetriesBegin(const C_string &str, C_string &rtn, E_PROP_KEYWORD &kw){

                              //check properties
   const void *mem;
   dword size;
   //frm->GetPropertyBuffer(&mem, &size);
   mem = str;
   size = str.Size();
   if(size < 3){
      kw = PROP_UNKNOWN;
      return false;
   }

   const char *cp = (const char*)mem;
   if(cp && cp[0] == '{')//} <- leave it there!
   {
      ++cp;
      while(isspace(cp[0])) ++cp;
      rtn = cp;
                              //find and erase terminator
      for(int j=0; j<rtn.Size(); j++)  //{ <- leave it there!
      if(rtn[j]=='}'){
         rtn[j] = 0;
         break;
      }
   }
   kw = GetPropertyKeyword(rtn, &cp);
   rtn = cp;

   return true;
}

//----------------------------
// Get next keyword in property buffer (str).
//----------------------------
static bool GetPropertiesNext(C_string &str, E_PROP_KEYWORD &kw){

   if(!str.Size())
      return false;
                              //skip white-space
   for(int i=0; isspace(str[i]); i++);

   const char *cp;
                              //decode
   kw = GetPropertyKeyword(&str[i], &cp);
   str = cp;

   return true;
}

//----------------------------
// Get value from properties buffer specified by format string.
// Note: format string must have second parameter '%n'.
//----------------------------
static bool GetPropertiesValue(C_string &str, const char *fmt_string, void *retp){

   if(!str.Size())
      return false;
                              //skip white-space
   for(int i=0; isspace(str[i]) || str[i]==','; i++);
   int numc = 0;
   if(sscanf(&str[i], fmt_string, retp, &numc) == 1){
      str = &str[i+numc];
      return true;
   }
   return false;
}

//----------------------------
// Read float value from property buffer, if available.
static bool GetPropertiesFloat(C_string &str, float &ret){
   return GetPropertiesValue(str, "%f%n", &ret);
}

//----------------------------
// Read int value from property buffer, if available.
static bool GetPropertiesInt(C_string &str, int &ret){
   return GetPropertiesValue(str, "%i%n", &ret);
}

//----------------------------
// Read string value from property buffer, if available.
static bool GetPropertiesString(C_string &str, C_string &ret){
   char buf[256];
   if(!GetPropertiesValue(str, "%256s%n", buf))
      return false;
   ret = buf;
   return true;
}

//----------------------------

typedef bool(T_chunk_read)(S_load_context &lc);

//----------------------------
                              //???
static bool UnknownReader(C_cache *is, dword p, int ind){

   union{
      char c[100];
      byte b[100];
      word w[50];
      dword dw[25];
      int i[25];
      float f[25];
   }u;
   is->read(u.c, sizeof(u));
   return true;
}

//----------------------------
                              //0002
static bool VersionReader(S_load_context &lc){

   word b[2];
   lc.is->read((char*)b, sizeof(b));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%i.%i", b[0], b[1]));
   }
   return true;
}

//----------------------------
                              //0010
static bool VectorReader(S_load_context &lc){

   lc.is->read((char*)&lc.recent_vector, sizeof(S_vector));
   if(lc.load_flags&I3DLOAD_LOG)
      lc.LOG(C_fstring("x: %.3f, y: %.2f, z: %.3f", lc.recent_vector[0], lc.recent_vector[1], lc.recent_vector[2]));
   return true;
}

//----------------------------
                              //0011
static bool RGBBReader(S_load_context &lc){

   lc.is->read((char*)&lc.recent_rgb, sizeof(byte)*3);
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("r: %i, g: %i, b: %i", lc.recent_rgb[0], lc.recent_rgb[1], lc.recent_rgb[2]));
   }
   return true;
}

//----------------------------
                              //0012
static bool Lin24Reader(S_load_context &lc){

   byte c[3];
   lc.is->read(c, sizeof(c));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("r: %i, g: %i, b: %i", c[0], c[1], c[2]));
   }
   return true;
}

//----------------------------
                              //0013
static bool LinFloatReader(S_load_context &lc){

   float f[3];
   lc.is->read((char*)f, sizeof(f));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("r: %.2f, g: %.2f, b: %.2f", f[0], f[1], f[2]));
   }
   return true;
}

//----------------------------
                              //0030
static bool PercentageReader(S_load_context &lc){

   short s;
   lc.is->read((char*)&s, sizeof(s));
   lc.recent_percentage = s;
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%i", s));
   }
   return true;
}

//----------------------------
/*
                              //0100
static bool MasterScaleReader(S_load_context &lc){

   float f;
   lc.is->read((char*)&f, sizeof(f));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Master scale: %f", f));
   }
   return true;
}

//----------------------------
                              //1100
static bool EnvBitmapReader(S_load_context &lc){

   lc.ReadString(lc.recent_string);
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(lc.recent_string);
   }
   return true;
}

//----------------------------
                              //1200
static bool EnvBgndColorReader(S_load_context &lc){

   if(!lc.scene)
      return true;

   lc.recent_vector.Zero();
   bool b = lc.ChunkReader();
   if(b)
      lc.scene->SetBgndColor(lc.recent_vector);
   return b;
}

//----------------------------
                              //2100
static bool EnvAmbientReader(S_load_context &lc){

   if(!lc.scene) return true;

   bool b = lc.ChunkReader();
   return b;
   return true;
}

//----------------------------
                              //2200
static bool EnvFogReader(S_load_context &lc){
   return true;
}

//----------------------------
                              //2201
static bool EnvUseFogReader(S_load_context &lc){
   return true;
}

//----------------------------
                              //2210
static bool FogBgndReader(S_load_context &lc){

   return true;
}

//----------------------------
                              //3D3E
static bool MeshVersionReader(S_load_context &lc){

   word b[2];
   lc.is->read((char*)b, sizeof(b));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%i.%i", b[0], b[1]));
   }
   return true;
}
*/

//----------------------------

void S_load_context::UV_Check(){

   int i;
                              
   float minu = 1e+6f, minv = 1e+6f, maxu = -1e+6f, maxv = -1e+6f;
   for(i = recent_vertices.size(); i--; ){
      minu = Min(minu, recent_vertices[i].tex[0].u);
      minv = Min(minv, recent_vertices[i].tex[0].v);
      maxu = Max(maxu, recent_vertices[i].tex[0].u);
      maxv = Max(maxv, recent_vertices[i].tex[0].v);
   }
                              //get middle
   minu = floor(minu);
   minv = floor(minv);
   maxu = ceil(maxu);
   maxv = ceil(maxv);
                              //center around 0
   float midu = minu + floor((maxu - minu) * .5f);
   float midv = minv + floor((maxv - minv) * .5f);

   for(i = recent_vertices.size(); i--; ){
      recent_vertices[i].tex[0].u =  recent_vertices[i].tex[0].u - midu;
      recent_vertices[i].tex[0].v =  recent_vertices[i].tex[0].v - midv;
   }

   maxu -= minu;
   maxv -= minv;

   if(maxu > 128.0f || maxv > 128.0f){
      allow_cache = false;
      REPORT_ERR(C_fstring("Warning: mesh '%s': UV range overflow (U = %.1f, V = %.1f) (%s)",
         (const char*)recent_string, maxu, maxv,
         (const char*)C_fstring("'%s'", recent_frame ? recent_frame->GetName() : "unknown")));
   }
}

//----------------------------
// Check if UV coordinates are almost equal.
//----------------------------
static bool AlmostEqual(const I3D_text_coor &txt1, const I3D_text_coor &txt2){

   return (((float)fabs(txt1.u - txt2.u) < 1e-4f) && ((float)fabs(txt1.v - txt2.v) < 1e-4f));
}

//----------------------------

void S_load_context::CleanMesh(){

   int i, j;

   assert(!recent_uv_faces[1].size() || (recent_uv_faces[1].size() == recent_uv_faces[0].size()));

                              //mark used vertices
   vector<word> v_map(recent_vertices.size());
   vector<bool> v_use(recent_vertices.size(), false);

   int new_face_count = recent_faces.size();
                              //clean invalid faces
   for(i=new_face_count; i--; ){
                              //thresh = minimal distance each vertex in triangle must have from adjacent edge
      const float thresh = 5e-4f;

      const I3D_triface &fc = recent_faces[i];
      float d = recent_vertices[fc[0]].xyz.DistanceToLine(recent_vertices[fc[1]].xyz, recent_vertices[fc[2]].xyz - recent_vertices[fc[1]].xyz);
      if(d >= thresh)
         d = recent_vertices[fc[1]].xyz.DistanceToLine(recent_vertices[fc[2]].xyz, recent_vertices[fc[0]].xyz - recent_vertices[fc[2]].xyz);
      if(d >= thresh)
         d = recent_vertices[fc[2]].xyz.DistanceToLine(recent_vertices[fc[0]].xyz, recent_vertices[fc[1]].xyz - recent_vertices[fc[0]].xyz);
      bool valid = (d >= thresh);

      if(!valid){
                              //face from back is moved forward
         --new_face_count;
         recent_faces[i] = recent_faces[new_face_count];
         for(j=MAX_UV_CHANNELS; j--; ){
            if(recent_uv_faces[j].size())
               recent_uv_faces[j][i] = recent_uv_faces[j][new_face_count];
         }
         if(recent_smooth_groups.size())
            recent_smooth_groups[i] = recent_smooth_groups[new_face_count];
                              //remap face-groups
         for(j=recent_face_groups.size(); j--; ){
            S_face_group &fg = recent_face_groups[j];
            for(int k=fg.face_indicies.size(); k--; ){
               if(fg.face_indicies[k] == i){
                  fg.face_indicies[k] = fg.face_indicies.back(); fg.face_indicies.pop_back();
               }else
               if(fg.face_indicies[k] == new_face_count){
                  fg.face_indicies[k] = i;
               }
            }
         }

      }else{
         v_use[fc[0]] = true;
         v_use[fc[1]] = true;
         v_use[fc[2]] = true;
      }
   }

   if(new_face_count != recent_faces.size()){
                              //some face were removed
#ifdef REPORT_INVALID_FACES
      allow_cache = false;
      REPORT_ERR(C_fstring("Warning: mesh '%s': %i invalid face(s)",
         (const char*)recent_string, recent_faces.size() - new_face_count));
#endif

                              //strip verts
      recent_faces.erase(recent_faces.begin() + new_face_count, recent_faces.end());
      for(j=MAX_UV_CHANNELS; j--; ){
         if(recent_uv_faces[j].size())
            recent_uv_faces[j].erase(recent_uv_faces[j].begin() + new_face_count, recent_uv_faces[j].end());
      }
      if(recent_smooth_groups.size())
         recent_smooth_groups.erase(recent_smooth_groups.begin() + new_face_count, recent_smooth_groups.end());
   }

   int new_vertex_count = recent_vertices.size();
                              //process all verts, remove unused
   for(i=0; i<new_vertex_count; ){
      v_map[i] = i;           //by default vertex stays where it is
      if(!v_use[i]){
                              //vertex from back is moved forward
         --new_vertex_count;
         v_map[new_vertex_count] = i;
         recent_vertices[i] = recent_vertices[new_vertex_count];
         v_use[i] = v_use[new_vertex_count];
      }else
         ++i;
   }

   if(new_vertex_count != recent_vertices.size()){
                              //some vertex was removed
#ifdef REPORT_UNUSED_VERTS
      lc.allow_cache = false;
      REPORT_ERR(C_fstring("Warning: mesh '%s': %i unused vertices",
         (const char*)recent_string, recent_vertices.size() - new_vertex_count));
#endif
                              //strip verts
      recent_vertices.erase(recent_vertices.begin() + new_vertex_count, recent_vertices.end());
                              //remap faces
      for(i=recent_faces.size(); i--; ){
         for(int j=3; j--; )
            recent_faces[i][j] = v_map[recent_faces[i][j]];
      }
   }
}

//----------------------------

void S_load_context::AssignUVCoordinates(int num_orig_verts, vector<word> &vertex_map){

   if(!recent_uv_verts[0].size() || !recent_uv_faces[0].size())
      return;

   assert(recent_faces.size() == recent_uv_faces[0].size());

   vector<bool> vertex_used(num_orig_verts, false);

   int i, j;
   for(i=recent_faces.size(); i--; ){
      I3D_triface &fc = recent_faces[i];
      const I3D_triface &uv_fc = recent_uv_faces[0][i];
      for(j=3; j--; ){
         word vi = fc[j];
         word uv_vi = uv_fc[j];
         assert(uv_vi<recent_uv_verts[0].size());
         const I3D_text_coor &txt = recent_uv_verts[0][uv_vi];
         if(!vertex_used[vi]){
            recent_vertices[vi].tex[0] = txt;
            vertex_used[vi] = true;
         }else
         if(!AlmostEqual(recent_vertices[vi].tex[0], txt)){
                           //check if we may map it onto any other duplicated vertex
            for(int ii=num_orig_verts; ii<recent_vertices.size(); ii++){
               if(vertex_map[ii] == vi && AlmostEqual(recent_vertices[ii].tex[0], txt))
                  break;
            }
            if(ii < recent_vertices.size()){
                           //vertex found, use this
               fc[j] = ii;
            }else
            {
               vertex_map.push_back(vi);
               fc[j] = recent_vertices.size();
                           //duplicate vertex
               recent_vertices.push_back(recent_vertices[vi]);
               recent_vertices.back().tex[0] = txt;
            }
         }
      }
   }
}

//----------------------------

struct S_shade_face{
   I3D_triface face;          //indicies to unrolled vertices
   I3D_triface orig_face;     //indicies to original vertices
   byte smooth_group;
   S_vector face_normal;
   bool valid;
};

//----------------------------

inline dword *FindElement(dword *mem, int array_len, dword element){
#ifdef _MSC_VER
   __asm{
      push ecx

      mov edi, mem
      mov ecx, array_len
      mov eax, element
      repne scasd
      jz ok
      mov edi, 4
   ok:
      sub edi, 4
      mov mem, edi

      pop ecx
   }
   return mem;
#else
   while(array_len--){
      if(*mem==element) return mem;
      ++mem;
   }
   return NULL;
#endif
}

//----------------------------

void S_load_context::MakeVertexNormals(int num_orig_verts, vector<word> &vertex_map){

   int i, j, k;

                              //keeps unrolled faces
   S_shade_face *face_info = new S_shade_face[recent_faces.size()];

                              //keep this array to mark used verts,
                              // later for normal count on vertex,
                              // even later for index remapping
   vector<int> v_use(recent_vertices.size(), 0);

                              //map of vertices to adjacent face
                              // (first) is vertex
                              // (second) is connected face index
   multimap<word, word> connect_vertex;
   bool sgroups_valid = (recent_smooth_groups.size()!=0);

                              //generate help info for faces
   for(i=recent_faces.size(); i--; ){
      S_shade_face &sf = face_info[i];
      const I3D_triface &fc = recent_faces[i];
      sf.face = fc;
      sf.smooth_group = sgroups_valid ? recent_smooth_groups[i] : 0;
      sf.valid = false;

                              //compute face normal
      sf.face_normal.GetNormal(recent_vertices[fc[0]].xyz, recent_vertices[fc[1]].xyz, recent_vertices[fc[2]].xyz);

                              //check = too small normal - small area of face
      float f = sf.face_normal.Dot(sf.face_normal);
      if(f > 1e-12f){
         f = SqrtFast(f);
         sf.face_normal /= f;
         sf.valid = true;
      }
                              //create unique vertex for each face
      for(k=0; k<3; k++){
         int vi = fc[k];
         word orig_vertex = vertex_map[vi];
         sf.orig_face[k] = orig_vertex;
         if(v_use[vi]){
                              //add vertex (internally)
            sf.face[k] = recent_vertices.size();
            v_use.push_back(1);
            vertex_map.push_back(vertex_map[vi]);
            recent_vertices.push_back(recent_vertices[vi]);
            vi = sf.face[k];
         }else{
            v_use[vi] = 1;
         }
         if(sf.valid){
                              //store edge info
                              //store vertex connection info
            connect_vertex.insert(pair<word, word>(orig_vertex, i));

            recent_vertices[vi].normal = sf.face_normal;
                              //power of normal is affected by angle
            const I3D_triface &fc = sf.face;
            float a = (recent_vertices[fc[(k+1)%3]].xyz - recent_vertices[vi].xyz).AngleTo(recent_vertices[fc[(k+2)%3]].xyz - recent_vertices[vi].xyz);
            recent_vertices[vi].normal *= a;
         }else
            recent_vertices[vi].normal.Zero();
      }
   }

                              //perform shading
   for(i = recent_faces.size(); i--; ){
      const S_shade_face &sf = face_info[i];
      if(!sf.valid)
         continue;

      const I3D_triface &fc = sf.orig_face;

                              //check all adjacent faces and smooth with them
                              // if smoothgroups match
      for(j=3; j--; ){
         dword vi = fc[j];
         for(multimap<word, word>::const_iterator it = connect_vertex.find(vi);
            it != connect_vertex.end() && ((*it).first == vi);
            ++it)
         {
            dword connected_face_i = (*it).second;
                              //don't smooth with ourselves
            if(connected_face_i == i)
               continue;

            const S_shade_face &sf1 = face_info[connected_face_i];
                              //check if smoothgroups match
            bool sg_match;
            if(sgroups_valid){
               sg_match = (sf.smooth_group&sf1.smooth_group);
            }else{
               sg_match = (sf.face_normal.AngleTo(sf1.face_normal) <= AUTO_SMOOTH_ANGLE);
            }
            if(sg_match){
                              //faces adjacent - spread normals among both
               for(k=3; k--; ){
                  if(sf1.orig_face[k] == vi)
                     break;
               }
               assert(k!=-1);
               const I3D_triface &fc1 = sf1.face;
               dword vi1 = fc1[k];
               const S_vector &v_k = recent_vertices[vi1].xyz;
               float a = (recent_vertices[fc1.vi[(k+1)%3]].xyz - v_k).AngleTo(recent_vertices[fc1.vi[(k+2)%3]].xyz - v_k);
               recent_vertices[sf.face[j]].normal += sf1.face_normal * a;
            }
         }
      }
   }

                              //4. normalize normals and remove duplicates
                              // initialize v_use to be used for
                              // remapping of removed vertices

   int num_verts = recent_vertices.size();

                              //normalize normals
   for(i=recent_vertices.size(); i--; ){
      recent_vertices[i].normal.Normalize();
   }


   for(i=recent_vertices.size(); i--; ){
#ifdef PAINT_VERTEX_NORMALS
      {                       //debug - paint normals
         S_vector v[2] = { vert_list[i].xyz, vert_list[i].xyz + vert_list[i].normal};
         drv->DEBUG_CODE(0, (dword)v);
      }
#endif

#define NORMAL_COLLAPSE_DIFF .1f

      v_use[i] = i;
                              //find if it matches any other vertex from original ones
      int v_origin_i = vertex_map[i];
      if(v_origin_i==i)
         continue;
      const I3D_vertex_mesh &vi = recent_vertices[i];

      if(i>=num_orig_verts && v_origin_i<i){
                              //first try to join original vertex
         const I3D_vertex_mesh &vj = recent_vertices[v_origin_i];
         if(NORMAL_COLLAPSE_DIFF > (fabs(vi.normal.x-vj.normal.x) + fabs(vi.normal.y-vj.normal.y) + fabs(vi.normal.z-vj.normal.z)) &&
            (vi.tex[0]==vj.tex[0])){
                              //point this vertex to new location
            v_use[i] = v_origin_i;
            dword *dwp;

            for(k=i+1, dwp = (dword*)&v_use[k]; k<num_verts; ){
               dword *dwp1 = FindElement(dwp, num_verts-k, i);
               if(!dwp1)
                  break;
               *dwp1++ = v_origin_i;
               k += dwp1-dwp;
               dwp = dwp1;
            }
                              //move vertex from back to our location
            int rem_i = recent_vertices.size() - 1;
            if(rem_i!=i){
               recent_vertices[i] = recent_vertices.back();
               vertex_map[i] = vertex_map.back();
                              //adjust all indicies to this
               for(k=i+1, dwp = (dword*)&v_use[k]; k<num_verts; ){
                  dword *dwp1 = FindElement(dwp, num_verts-k, rem_i);
                  if(!dwp1) break;
                  *dwp1++ = i;
                  k += dwp1-dwp;
                  dwp = dwp1;
               }
            }
            recent_vertices.pop_back();
            vertex_map.pop_back();
            continue;
         }
      }
                              //now try all other vertices
      for(j=0; j<i; j++){
                              //check if verts the same
         const I3D_vertex_mesh &vj = recent_vertices[j];
                              //check only verts which were originally duplicated
         if(v_origin_i==vertex_map[j])
         if(NORMAL_COLLAPSE_DIFF > (fabs(vi.normal.x-vj.normal.x) + fabs(vi.normal.y-vj.normal.y) + fabs(vi.normal.z-vj.normal.z)))
         if(vi.tex[0]==vj.tex[0])
         {
                              //point this vertex to new location
            v_use[i] = j;
            dword *dwp;

            for(k=i+1, dwp = (dword*)&v_use[k]; k<num_verts; ){
               dword *dwp1 = FindElement(dwp, num_verts-k, i);
               if(!dwp1) break;
               *dwp1++ = j;
               k += dwp1-dwp;
               dwp = dwp1;
            }
                              //move vertex from back to our location
            int rem_i = recent_vertices.size() - 1;
            if(rem_i!=i){
               recent_vertices[i] = recent_vertices.back();
               vertex_map[i] = vertex_map.back();
                              //adjust all indicies to this
               for(k=i+1, dwp = (dword*)&v_use[k]; k<num_verts; ){
                  dword *dwp1 = FindElement(dwp, num_verts-k, rem_i);
                  if(!dwp1) break;
                  *dwp1++ = i;
                  k += dwp1-dwp;
                  dwp = dwp1;
               }
            }
            recent_vertices.pop_back();
            vertex_map.pop_back();
            break;
         }
      }
   }

                              //remap original faces
   for(i=recent_faces.size(); i--; ){
      for(j=0; j<3; j++){
         recent_faces[i][j] = v_use[face_info[i].face[j]];
#if 0
         {                    //perform consistency check
            assert(faces[j][k] < vertices.size());
         }
#endif
      }
   }

   delete[] face_info;
}

//----------------------------
                              //4000
static bool ObjBlockReader(S_load_context &lc){

   if(!lc.root_frame)         //true when animation is loaded
      return true;

   lc.ReadString(lc.recent_string);
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Object name: '%s'", (const char*)lc.recent_string));
   }

   bool ok = false;
#ifdef USE_DATABASE
   {                          //try using cached version
      C_fstring cache_name("%s+%s", (const char*)lc.file_name, (const char*)lc.recent_string);
      C_cache *ck = lc.driver->dbase.OpenRecord(cache_name, lc.file_time);
      if(ck){
                              //create mesh and try to load
         PI3D_mesh_level ml = lc.driver->CreateMeshLevel();
         assert(ml);
         ok = ml->LoadCachedInfo(ck, lc);
         lc.driver->dbase.CloseRecord(ck);
         if(ok){
                              //cached version loaded ok
            PI3D_mesh_object mo = lc.driver->CreateMesh();
            mo->SetLOD(ml);
            lc.mesh_list[lc.recent_string] = mo;
            mo->Release();
            lc.mesh_db_valid.insert(lc.recent_string);
#ifdef CREATE_COLLISION_DATA
            ml->CreateCollisionData();
#endif
         }
         ml->Release();
      }
   }
#endif

   if(!ok){
      ok = lc.ChunkReader();
   }
   if(ok){
      if(lc.ShowProgress())
         return false;
   }
   return ok;
}

//----------------------------
                              //4100
static bool NewMeshReader(S_load_context &lc){

   lc.allow_cache = true;

   int i, j;

   lc.recent_vertices.clear();
   lc.recent_faces.clear();
   lc.recent_smooth_groups.clear();
   for(i=MAX_UV_CHANNELS; i--; ){
      lc.recent_uv_verts[i].clear();
      lc.recent_uv_faces[i].clear();
   }
   lc.recent_face_groups.clear();

   C_smart_ptr<I3D_mesh_level> ml = lc.driver->CreateMeshLevel(); ml->Release();

   bool b = lc.ChunkReader();

   lc.CleanMesh();

   int num_orig_verts = lc.recent_vertices.size();
   vector<word> vertex_map(num_orig_verts);
   for(i=num_orig_verts; i--; ) vertex_map[i] = i;

                              //process uv coordinates, duplicate vertices as necessary
   lc.AssignUVCoordinates(num_orig_verts, vertex_map);

   lc.UV_Check();

   lc.MakeVertexNormals(num_orig_verts, vertex_map);

   if(lc.recent_uv_verts[1].size() && lc.recent_uv_faces[1].size()){
      C_map_channel *mc2 = new C_map_channel;
      mc2->uv_verts = lc.recent_uv_verts[1];
      mc2->uv_faces = lc.recent_uv_faces[1];
                              //save (possible) name of environment map
      assert(lc.recent_face_groups.size());
      PI3D_material mat = lc.recent_face_groups.begin()->mat;
      if(mat){
         map<C_string, S_mat_info>::const_iterator it = lc.mat_list.find(mat->GetName());
         if(it != lc.mat_list.end()){
            mc2->bitmap_name = (*it).second.filename[2];
         }
      }

      ml->SetMapChannel2(mc2);
      mc2->Release();
   }

                              //don't keep empty meshes
   if(!lc.recent_faces.size() || !lc.recent_vertices.size()){
#ifdef REPORT_EMPTY_MESH_ERROR
      lc.REPORT_ERR(C_fstring("warning: mesh has no vertices or faces: '%s'",
         (const char*)lc.recent_string));
#endif
      return true;
   }

   vector<I3D_triface> ordered_faces(lc.recent_faces.size());
   vector<byte> ordered_smooth_groups(lc.recent_smooth_groups.size());
   vector<I3D_face_group> fgroups;

                              //spread faces into groups
   bool corruption_warned = false;
   dword face_count = 0;
   vector<bool> face_used(lc.recent_faces.size(), false);
   for(i=0; i<lc.recent_face_groups.size(); i++){
      const S_load_context::S_face_group &fg = lc.recent_face_groups[i];
      int face_num = fg.face_indicies.size();
      if(!face_num)           //we could have cleaned this face group
         continue;

      fgroups.push_back(I3D_face_group());
      I3D_face_group &fgroup = fgroups.back();
      assert(fg.mat);
      fgroup.SetMaterial(fg.mat);
      fgroup.base_index = face_count;
      fgroup.num_faces = face_num;
                              //copy listed faces to group
      for(j=0; j<face_num; j++){
         word face_index = fg.face_indicies[j];
         if(!face_used[face_index]){
            //face_used[face_count] = true;
            face_used[face_index] = true;
            ordered_faces[face_count] = lc.recent_faces[face_index];
            if(lc.recent_smooth_groups.size())
               ordered_smooth_groups[face_count] = lc.recent_smooth_groups[face_index];
            ++face_count;
         }else{
                              //how can be face added in 2 face groups? export plugin bug?
            if(!corruption_warned){
               lc.REPORT_ERR(C_fstring("warning: mesh is invalid: '%s' - re-export file",
                  (const char*)lc.recent_string));
               corruption_warned = true;
            }
            --fgroup.num_faces;
            //return false;
         }
      }
      if(!fgroup.num_faces){
                              //corrupted mesh - remove this group
         fgroups.pop_back();
      }
   }

   if(face_count != lc.recent_faces.size()){
                              //put trifaces with default mat into list
      j = lc.recent_faces.size() - face_count;
      lc.allow_cache = false;
      lc.REPORT_ERR(C_fstring("Warning: mesh '%s' has default material", (const char*)lc.recent_string));

                              //create special group with default material
      fgroups.push_back(I3D_face_group());
      I3D_face_group &to_face = fgroups.back();
      {
         PI3D_material mat = lc.driver->CreateMaterial();
         mat->SetAmbient(S_vector(1, 1, 1));
         mat->SetDiffuse(S_vector(1, 1, 1));
         to_face.SetMaterial(mat);
         mat->Release();
         to_face.base_index = face_count;
      }
      to_face.num_faces = j;

      for(i=0, j=0; i<lc.recent_faces.size(); i++){
         if(!face_used[i]){
            ordered_faces[face_count] = lc.recent_faces[i];
            if(lc.recent_smooth_groups.size())
               ordered_smooth_groups[face_count] = lc.recent_smooth_groups[i];
            ++face_count;
         }
      }
   }
   assert(face_count==lc.recent_faces.size());

   ml->Build2(lc.recent_vertices.begin(), lc.recent_vertices.size(),
      ordered_faces.begin(), ordered_faces.size(),
      fgroups.begin(), fgroups.size());
   ml->SetSmoothGroups(ordered_smooth_groups);

   if(vertex_map.size() > num_orig_verts)
      ml->vertex_buffer.SetVertexSpreadMap(vector<word>(&vertex_map[num_orig_verts], vertex_map.end()));

#ifdef USE_DATABASE
                              //store into database, if no warnings issued
   if(lc.allow_cache)
      lc.mesh_db_save.insert(lc.recent_string);
   /*
   if(lc.allow_cache){
      C_fstring cache_name("%s+%s", (const char*)lc.file_name, (const char*)lc.recent_string);
      C_cache *ck = lc.driver->dbase.CreateRecord(cache_name, lc.file_time);
      if(ck){
         ml->SaveCachedInfo(ck);
         lc.driver->dbase.CloseRecord(ck);
      }
   }
   */
#endif

                              //save mesh into list
   PI3D_mesh_object mo = (PI3D_mesh_object)lc.driver->CreateMesh();
   mo->SetLOD(ml);
   lc.mesh_list[lc.recent_string] = mo;
   mo->Release();

#ifdef CREATE_COLLISION_DATA
   ml->CreateCollisionData();
#endif

   return b;
};

//----------------------------
                              //4110
static bool VertListReader(S_load_context &lc){

   int i;
                              //get # of vertices
   word vert_num;
   lc.is->read((char*)&vert_num, sizeof(vert_num));

#ifndef USE_ALLOCA
   S_vector *fp = new S_vector[vert_num];
#else
   S_vector *fp = (S_vector*)alloca(sizeof(S_vector)*vert_num);
#endif
   lc.is->read((char*)fp, sizeof(S_vector)*vert_num);

   lc.recent_vertices.assign(vert_num);
   memset(lc.recent_vertices.begin(), 0, sizeof(I3D_vertex_mesh) * vert_num);

   S_vector v;

   for(i=0; i<vert_num; i++){
                              //swap y and z
      v.x = fp[i].x;
      v.y = fp[i].z;
      v.z = fp[i].y;
      lc.recent_vertices[i].xyz = v;
   }
#ifndef USE_ALLOCA
   delete[] fp;
#endif

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Num vertices: %i", vert_num));
   }
   return true;
}

//----------------------------
/*
                              //4111
static bool VertSlctReader(C_cache *is, dword to_pos, int tab){
   word nv;
   word c;
   if(fread(&nv,sizeof(nv),1,f)!=1) return;
   int i=0;
   while(nv--){
      if(fread(&c,sizeof(c),1,f)!=1) return;
      recent_mesh->verts[i].slct=0;
      if(c&0x8000) recent_mesh->verts[i].slct|=SEL_A;
      if(c&0x4000) recent_mesh->verts[i].slct|=SEL_B;
      if(c&0x2000) recent_mesh->verts[i].slct|=SEL_C;
      if(verb){
         printf("%*sVertex %u selection: (%x) ",ind,"",i,(int)c);
         if(c & 0x8000) printf("A");
         if(c & 0x4000) printf("B");
         if(c & 0x2000) printf("C");
         printf("\n");
      }
      ++i;
   }
}
*/

//----------------------------
                              //4120
static bool FaceListReader(S_load_context &lc){

                              //read # of faces
   word face_num;
   lc.is->read((char*)&face_num, sizeof(face_num));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Num faces: %i", face_num));
   }
                              //read face descriptors
   struct S_face_desc{
      word index[3];
      word flag;
   } *fd;
#ifndef USE_ALLOCA
   fd = new S_face_desc[face_num];
#else
   fd = (S_face_desc*)alloca(sizeof(S_face_desc)*face_num);
#endif
   lc.is->read((char*)fd, sizeof(S_face_desc)*face_num);

   lc.recent_faces.assign(face_num);

#ifdef _DEBUG
   int max_vertex = lc.recent_vertices.size();
#endif

   for(int i=0; i<face_num; i++){
      I3D_triface &fc = lc.recent_faces[i];
                              //copy - also convert CCW to CW culling
      fc[0] = fd[i].index[0];
      fc[1] = fd[i].index[2];
      fc[2] = fd[i].index[1];

#ifdef _DEBUG
      assert(fc[0] < max_vertex);
      assert(fc[1] < max_vertex);
      assert(fc[2] < max_vertex);
#endif
                              //copy flags
      //byte &flg = face_flags[num_tri_faces];
      //flg = fd[i].flag & 7;
      /*
                              //swap bits of flags
      bool b[2]={flg&2, flg&4 };
      flg &= 1;
      if(b[0]) flg |= 4;
      if(b[1]) flg |= 2;
      */
   }

#ifndef USE_ALLOCA
   delete[] fd;
#endif

   return lc.ChunkReader();
}

//----------------------------
                              //4130
static bool FaceMatReader(S_load_context &lc){
                              //get material name
   C_string s;
   lc.ReadString(s);

   word face_num;
   lc.is->read((char*)&face_num, sizeof(face_num));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Material: '%s'", (const char*)s));
      lc.LOG(C_fstring("Face count: %i", face_num));
   }
   if(!face_num)
      return true;

   lc.recent_face_groups.push_back(S_load_context::S_face_group());
   S_load_context::S_face_group &fg = lc.recent_face_groups.back();
   fg.face_indicies.assign(face_num);
   lc.is->read((char*)fg.face_indicies.begin(), sizeof(word) * face_num);
   fg.mat = lc.mat_list[s].GetMaterial(lc);

   return true;
}

//----------------------------
                              //4140
static bool MapCoorsReader(S_load_context &lc){

   lc.REPORT_ERR(C_fstring("Mesh '%s': old-style maping channel used - reexport file", (const char*)lc.recent_string));
   lc.allow_cache = false;
   /*
   word count;
   lc.is->read((char*)&count, sizeof(count));

                              //read u/v info for all vertices
#ifndef USE_ALLOCA
   I3D_text_coor *tuv  = new I3D_text_coor[count];
#else
   I3D_text_coor *tuv  = (I3D_text_coor*)alloca(sizeof(I3D_text_coor)*count);
#endif
   lc.is->read((char*)tuv, sizeof(I3D_text_coor)*count);

   assert(lc.recent_vertices.size() == count);

                              //copy to vertices, also swap 'v'
   for(int i=count; i--; ){
      lc.recent_vertices[i].tex[0].u = tuv[i].u;
      lc.recent_vertices[i].tex[0].v = -tuv[i].v;
   }

#ifndef USE_ALLOCA
   delete[] tuv;
#endif
   */

   return true;
}

//----------------------------
                              //4150
static bool SmooListReader(S_load_context &lc){

   int num_faces = lc.recent_faces.size();

   dword *sgl = new dword[num_faces];
   lc.is->read((char*)sgl, sizeof(dword)*num_faces);
                              //convert dwords to bytes
   lc.recent_smooth_groups.assign(num_faces);
   dword sg_intersection = 0;
   for(int i=num_faces; i--; ){
      lc.recent_smooth_groups[i] = sgl[i];
      sg_intersection |= sgl[i];
   }

   if(sg_intersection&0xffffff00){
                              //convert smoothgroups
      byte conv_tab[32];
      byte dst_bit = 1;
      for(i=0; i<32; i++){
         conv_tab[i] = 0;
         if(sg_intersection&(1<<i)){
            conv_tab[i] = dst_bit;
            dst_bit <<= 1;
                              //override bug with more than 8 smoothgroups
            if(dst_bit==0x100)
               dst_bit = 1;
         }
      }
      for(i=lc.recent_smooth_groups.size(); i--; ){
         byte &dst = lc.recent_smooth_groups[i];
         dword src = sgl[i];
         dst = 0;

         dword test_mask = 1;
         for(int j=0; j<32; j++, test_mask <<= 1){
            if(src&test_mask){
               dst |= conv_tab[j];
            }
         }
#ifdef _DEBUG
         if(CountBits(sg_intersection) < 8){
            assert(CountBits(src) == CountBits(dst));
         }
#endif
      }
      /*
      if(CountBits(sg_intersection) >= 8){
         lc.REPORT_ERR(C_fstring("Error: mesh '%s' - too many smoothgroups.", (const char*)C_fstring("'%s'", (const char*)lc.recent_string)));
      }
      */
   }
   delete[] sgl;

   return true;
}

//----------------------------
                              //4160
static bool TrMatrixReader(S_load_context &lc){

   S_vector v[4];
   lc.is->read((char*)v, sizeof(v));

   S_matrix m;
   m.Identity();
                              //read matrix - it is 3x4 matrix, assuming
                              // last column is [0 0 0 1]
                              //note: since Y and Z are swapped in 3DS,
                              // we need to adjust matrix to consider it
                              // by swapping rows and columns
   for(int i=4; i--; ){
      m(i) = v[i];
      swap(m(i).y, m(i).z);
   }
   swap(m(1), m(2));

                              //multiply all vertices by inverse
                              //transformation matrix
   m.Invert();

   for(i=lc.recent_vertices.size(); i--; )
      lc.recent_vertices[i].xyz *= m;

   return true;
}

//----------------------------
                              //4165
static bool MeshColorReader(S_load_context &lc){

   return true;
}

//----------------------------
                              //4200
static bool MapChannelReader(S_load_context &lc){

                              //read channel index
   int channel;
   lc.is->read((char*)&channel, sizeof(channel));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("channel: %i", channel));
   }

   channel -= 1;

   if(channel >= MAX_UV_CHANNELS){
      lc.REPORT_ERR(C_fstring("Mesh '%s': invalid map channel used: %i", (const char*)lc.recent_string, channel+1));
      return true;
   }

                              //read uv vertices
   word num_uv_verts;
   lc.is->read((char*)&num_uv_verts, sizeof(num_uv_verts));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("number of uv verts: %i", num_uv_verts));
   }

   lc.recent_uv_verts[channel].assign(num_uv_verts);
   lc.is->read((char*)lc.recent_uv_verts[channel].begin(), num_uv_verts * sizeof(I3D_text_coor));

                              //negate 'v'
   for(int i=num_uv_verts; i--; ){
      lc.recent_uv_verts[channel][i].v = -lc.recent_uv_verts[channel][i].v;
   }


                              //read channel faces
   word num_faces;
   lc.is->read((char*)&num_faces, sizeof(num_faces));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("number of uv faces: %i", num_faces));
   }
   lc.recent_uv_faces[channel].assign(num_faces);
   lc.is->read((char*)lc.recent_uv_faces[channel].begin(), num_faces * sizeof(I3D_triface));

                              //swap culling for uv faces
   for(i=num_faces; i--; ){
      swap(lc.recent_uv_faces[channel][i][1], lc.recent_uv_faces[channel][i][2]);
   }

   return true;
}

//----------------------------
                              //4600
static bool LightReader(S_load_context &lc){

   if(!lc.root_frame)
      return true;
                              //read position
   S_vector v;
   lc.is->read((char*)&v, sizeof(v));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("x: %.2f, y: %.2f, z: %.2f", v[0], v[1], v[2]));
   }
   swap(v[1], v[2]);

   lc.recent_light = lc.driver->CreateLight();
   lc.recent_light->SetLightType(I3DLIGHT_POINT);
   lc.recent_light->SetRange(0.0, 1e+16f);
   lc.recent_light->SetPos(S_vector(v[0], v[1], v[2]));
   lc.recent_light->name = lc.recent_string;
   lc.recent_light->SetOn(true);

   lc.recent_vector.Zero();
   lc.recent_light_mult = 1.0f;

   bool b = lc.ChunkReader();
   if(lc.recent_light){
      switch(lc.recent_light->GetLightType()){
      case I3DLIGHT_FOG:
         break;
      }
      lc.recent_light->SetPower(lc.recent_light_mult);
      lc.recent_light->SetColor(lc.recent_vector);
      lc.recent_light->Update();
      lc.light_list.push_back(lc.recent_light);
      lc.recent_light->Release(); lc.recent_light = NULL;
   }
   return b;
}

//----------------------------
                              //4610
static bool SpotLightReader(S_load_context &lc){

   /*                         //spotlights unsupported
   struct S_dummy{
      S_vector target;
      float hotspot, falloff;
   }c;
   lc.is->read((char*)&c, sizeof(c));
   swap(c.target.y, c.target.z);
                              //make direction zero-based (normal)
   S_vector pos;
   if(lc.recent_light) pos = lc.recent_light->GetPos();
   else
   if(lc.recent_sound) pos = lc.recent_sound->GetPos();
   else
   pos.Zero();

   c.target -= pos;
   c.target.Normalize();
   if(lc.recent_light){
      lc.recent_light->SetDir(c.target, 0.0f);
      if(lc.recent_light->GetName()[0]=='|'){
         lc.recent_light->SetLightType(I3DLIGHT_DIRECTIONAL);
      }else{
         lc.recent_light->SetLightType(I3DLIGHT_SPOT);
         lc.recent_light->SetCone(c.hotspot*PI/180.0f, c.falloff*PI/180.0f);
      }
      lc.recent_light->Update();
   }else
   if(lc.recent_sound){
      lc.recent_sound->SetDir(c.target, 0.0f);
      lc.recent_sound->SetCone(c.hotspot*PI/180.0f, c.falloff*PI/180.0f);
   }
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Target x: %.2f, y: %.2f, z: %.2f, hotspot: %.2f, falloff: %.2f",
         c.target[0], c.target[1], c.target[2], c.hotspot, c.falloff));
   }
   */
   return true;
}

//----------------------------
                              //4620
static bool LightOffReader(S_load_context &lc){

   /*
   if(recent_light) recent_light->SetOn(false);
   else
   if(recent_sound) recent_sound->SetOn(false);
   */
   return true;
}

//----------------------------
                              //4625
static bool LightRangeReader(S_load_context &lc){

   if(lc.recent_light){
      lc.recent_light->SetRange(lc.recent_light_range[0], lc.recent_light_range[1]);
   }else
   if(lc.recent_sound){
      lc.recent_sound->SetRange(lc.recent_light_range[0], lc.recent_light_range[1], .8f);
   }
   return true;
}

//----------------------------
                              //4659
static bool LightStartReader(S_load_context &lc){

   lc.is->read((char*)&lc.recent_light_range[0], sizeof(float));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%i", lc.recent_light_range[0]));
   }
   return true;
}

//----------------------------
                              //465A
static bool LightEndReader(S_load_context &lc){

   lc.is->read((char*)&lc.recent_light_range[1], sizeof(float));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%i", lc.recent_light_range[1]));
   }
   return true;
}

//----------------------------
                              //465B
static bool LightMultReader(S_load_context &lc){

   lc.is->read((char*)&lc.recent_light_mult, sizeof(float));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%i", lc.recent_light_mult));
   }
   return true;
}

//----------------------------
                              //4700
static bool CameraReader(S_load_context &lc){

   if(!lc.root_frame)
      return true;

#pragma pack(push,1)
   struct S_dummy{
      S_vector pos;
      S_vector target;
      float bang, focal;
   } c;
#pragma pack(pop)
   lc.is->read((char*)&c, sizeof(c));
   swap(c.pos.y, c.pos.z);
   swap(c.target.y, c.target.z);

   float fov = (2400*PI/180.0f)/c.focal;

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Pos x: %.2f, y: %.2f, z: %.2f", c.pos.x, c.pos.y, c.pos.z));
      lc.LOG(C_fstring("Dir x: %.2f, y: %.2f, z: %.2f", c.target.x, c.target.y, c.target.z));
      lc.LOG(C_fstring("FOV: %.2f", fov));
   }
   bool b = lc.ChunkReader();

   PI3D_camera cam = lc.driver->CreateCamera();
   lc.container->AddFrame(cam);
   cam->LinkTo(lc.root_frame);
   cam->name = lc.recent_string;
   cam->SetPos(c.pos);
   c.target -= c.pos;

   if(!c.target.IsNull()) c.target.Normalize();
   else c.target=S_vector(0.0, 0.0, 1.0);

   cam->SetDir(c.target, 0);
   cam->SetFOV(fov);
   cam->SetRange(.5f, lc.recent_camera_range[1]);
   cam->Update();
   cam->Release();

   return b;
}

//----------------------------
                              //4720
static bool CameraRangeReader(S_load_context &lc){

   lc.is->read((char*)lc.recent_camera_range, sizeof(lc.recent_camera_range));

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Inner: %.2f, outer: %.2f", lc.recent_camera_range[0], lc.recent_camera_range[1]));
   }

   return true;
}

//----------------------------
                              //4904
static bool NURBsReader(S_load_context &lc){

   return false;
   /*
   lc.recent_nurbs = (PI3D_nurbs)lc.driver->CreateVisual(I3D_VISUAL_NURBS);
   if(!lc.recent_nurbs) return true;
   lc.recent_nurbs->SetName(lc.recent_string);

   bool b = lc.ChunkReader();
   if(b){
      lc.nurbs_list.push_back(lc.recent_nurbs);
   }
   lc.recent_nurbs->Release(); lc.recent_nurbs = NULL;

   return b;
   */
}

//----------------------------
                              //4905
static bool NURBsUVOrderReader(S_load_context &lc){

   int uv_order[2];
   lc.is->read((char*)uv_order, sizeof(uv_order));

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("U: %i, V: %i", uv_order[0], uv_order[1]));
   }
   return true;
}

//----------------------------
                              //4906
static bool NURBsCVsReader(S_load_context &lc){

   return false;
   /*
   if(!lc.recent_nurbs) return true;

   int res[2];
   lc.is->read((char*)res, sizeof(res));
   if(res[0]<2 || res[0]>4 || res[1]<2 || res[1]>4){
      lc.REPORT_ERR(C_fstring("unsupported NURBs CV resolution: %ix%i", res[0], res[1]));
      return true;
   }
   lc.recent_nurbs->SetResolution(res[0], res[1]);

   struct S_nurb_vertex{
      S_vector v;
      float weight;
   };
   int num = res[0] * res[1];
   S_nurb_vertex *vp = new S_nurb_vertex[num];
   lc.is->read((char*)vp, sizeof(S_nurb_vertex)*num);
   for(int x=0, i=0; x<res[1]; x++){
      for(int y=0; y<res[0]; y++, ++i){
         S_vector &v = vp[i].v;
         swap(v.y, v.z);
         lc.recent_nurbs->SetCV(x, y, v, vp[i].weight, 0);
      }
   }
   delete[] vp;

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Resolution: x =  %i, y = %i", res[0], res[1]));
   }

   return true;
   */
}

//----------------------------
                              //4907
static bool NURBsKnotsReader(S_load_context &lc){

   int num_uv_knots[2];
   lc.is->read((char*)num_uv_knots, sizeof(num_uv_knots));

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("U: %i, V: %i", num_uv_knots[0], num_uv_knots[1]));
   }

   return true;
}

//----------------------------
                              //4908
static bool NURBsTMReader(S_load_context &lc){

   /*
   S_vector v[4];
   is->read((char*)v, sizeof(v));

   S_matrix m;
   m.Identity();
                              //read matrix - it is 3x4 matrix, assuming
                              //last column is [0 0 0 1]
   for(int i=4; i--; ) m(i) = v[i];
   */

   return true;
}

//----------------------------
                              //4909
static bool NURBsMaterialReader(S_load_context &lc){

   return false;
   /*
   assert(lc.recent_nurbs);
   C_string s;
   lc.ReadString(s);

   PI3D_material mp = lc.mat_list[s].GetMaterial(lc, NULL);
   lc.recent_nurbs->SetMaterial(mp);

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Material: '%s'", (const char*)s));
   }
   return true;
   */
}

//----------------------------
                              //490a
static bool NURBsFlipNormReader(S_load_context &lc){

   bool b;
   lc.is->read((char*)&b, sizeof(bool));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%s", b ? "true" : "false"));
   }

   return true;
}

//----------------------------
                              //490b
static bool NURBsUVMapReader(S_load_context &lc){

   return false;
   /*
   assert(lc.recent_nurbs);
#pragma pack(push,1)
   struct S_map{
      S_vector2 offset;
      S_vector2 tile;
      I3D_text_coor uv[2][2];
   } map;
#pragma pack(pop)
   lc.is->read((char*)&map, sizeof(map));
   for(int y=0; y<2; y++){
      for(int x=0; x<2; x++){
         lc.recent_nurbs->SetUV(x, y, map.uv[y][x]);
      }
   }

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("offset: %.2f, %.2f", map.offset.x, map.offset.y));
      lc.LOG(C_fstring("tile: %.2f, %.2f", map.tile.x, map.tile.y));
   }
   return true;
   */
}

//----------------------------
                              //7001
//static bool ViewportLayoutReader(C_cache *is, dword to_pos, int tab){}

//----------------------------
                              //AFFF
static bool NewMatReader(S_load_context &lc){

   if(!lc.root_frame)
      return true;

   lc.recent_self_illum = 0;
   lc.env_effect = EE_NULL;
                              //read all material data
   bool b = lc.ChunkReader();

   dword ct_flags = 0;

   if(lc.diffuse_map || lc.opacity_map){
      if(lc.driver->GetState(RS_TEXTURECOMPRESS))
         ct_flags |= TEXTMAP_COMPRESS;
      if(lc.recent_txt_flags&LOADF_TXT_TRANSP){
         ct_flags |= TEXTMAP_TRANSP;
         if(lc.recent_txt_flags&LOADF_TXT_TRANSP_ALPHA)
            lc.recent_mat->SetMatGlags(lc.recent_mat->GetMatGlags() | MATF_CKEY_ZERO_REF);
      }
      if(lc.diffuse_map) ct_flags |= TEXTMAP_DIFFUSE;
      if(lc.recent_txt_flags&LOADF_TXT_TRUECOLOR) ct_flags |= TEXTMAP_TRUECOLOR;

      if(lc.opacity_map){
         ct_flags |= TEXTMAP_OPACITY;
         if(lc.opacity_map->inverse)
            ct_flags |= TEXTMAP_INVALPHA;
      }
      switch(lc.recent_text_mipmap){
      case 1: ct_flags |= TEXTMAP_MIPMAP; break;
      case 2: ct_flags |= TEXTMAP_NOMIPMAP; break;
      }
      if(!(lc.recent_txt_flags&LOADF_TXT_DIFFUSE_KEEP)){
                              //set diffuse color to white
         float percentage = lc.diffuse_map ? ((float)lc.diffuse_map->percentage/100.0f) : 1.0f;
         assert(lc.recent_mat);
         S_vector col = lc.recent_mat->GetDiffuse();
         S_vector up = S_vector(1, 1, 1) - col;
         lc.recent_mat->SetDiffuse(col+up*percentage);
         lc.recent_mat->SetAmbient(S_vector(1.0f, 1.0f, 1.0f));
      }
   }
   
   lc.recent_mat->SetEmissive(lc.recent_mat->GetDiffuse() * lc.recent_self_illum);
   lc.recent_mat->Update();

   {
      S_mat_info &mi = (lc.mat_list[lc.recent_mat->GetName1()] = S_mat_info());
      mi.ct_flags = ct_flags;
      mi.mat = lc.recent_mat;
      mi.load_flags = lc.recent_txt_flags;
      //mi.txt_anim = (lc.recent_txt_flags&LOADF_TXT_DIFF_ANIM);
      //mi.reflection_anim = (lc.recent_txt_flags&LOADF_TXT_REFL_ANIM);
      mi.diffuse_anim_speed = 0;
      if(lc.diffuse_map){
         mi.filename[0] = lc.diffuse_map->s;
         mi.diffuse_anim_speed = lc.diffuse_map->anim_speed;
      }
      if(lc.txt_proc_name.Size()){
         mi.filename[4] = lc.txt_proc_name;
      }

      if(lc.opacity_map){
         mi.filename[1] = lc.opacity_map->s;
      }
      if(lc.environment_map){
         mi.filename[2] = lc.environment_map->s;
         mi.environment_power = (float)lc.environment_map->percentage * .01f;
      }
      if(lc.bump_map){
         mi.filename[3] = lc.bump_map->s;
         mi.bump_power = (float)lc.bump_map->percentage * .01f;
         mi.bump_anim_speed = !(lc.recent_txt_flags&LOADF_TXT_BUMP_ANIM) ? 0 : lc.bump_map->anim_speed;
      }
      mi.env_effect = lc.env_effect;
   }

   lc.recent_mat->Release(); lc.recent_mat = NULL;

   delete lc.diffuse_map; lc.diffuse_map = NULL;
   delete lc.opacity_map; lc.opacity_map = NULL;
   delete lc.environment_map; lc.environment_map = NULL;
   delete lc.bump_map; lc.bump_map = NULL;

   if(lc.ShowProgress())
      return false;

   return b;
};

//----------------------------
                              //A000
static bool MatNameReader(S_load_context &lc){
                             ///read ASCIIZ material name
   C_string str;
   lc.ReadString(str);

   lc.recent_txt_flags = 0;
#ifdef FORCE_MIPMAPPING
   lc.recent_text_mipmap = 1;
#else
   lc.recent_text_mipmap = 0;
#endif
   lc.txt_proc_name = NULL;

   //bool mipmap_flag = false;
   int i = 0;
   char c;
   while(c=str[i++]){
      switch(c){

      case '!':
         lc.recent_txt_flags |= LOADF_TXT_TRANSP;
         if(str[i]=='!'){
            lc.recent_txt_flags |= LOADF_TXT_TRANSP_ALPHA;
            ++i;
         }
         break;
      case '$':
         if(str[i]=='$'){
            lc.recent_text_mipmap = 2;  //disable mipmap
            ++i;
         }else{
            lc.recent_text_mipmap = 1;  //enable mipmap
            //mipmap_flag = true;
         }
         break;

      case '+':
         lc.recent_txt_flags |= LOADF_TXT_TRUECOLOR;
         break;

      case '|':
         if(str[i]=='|'){
            ++i;
            lc.recent_txt_flags |= LOADF_TXT_REFL_ANIM;
         }else
            lc.recent_txt_flags |= LOADF_TXT_DIFF_ANIM;
         break;

      case '^':
         lc.recent_txt_flags |= LOADF_TXT_BUMP_ANIM;
         break;

      case '#':               //unusable - MAX uses this
      case '-':
         break;

      case '%':
                              //read environment effect type
         {
            /*
            int effect_i = 0, num = 0;
            sscanf(&str[i], "%i%n", &effect_i, &num);
            i += num;

            //if(effect_i <= EE_NULL || effect_i >= EE_LAST)
            if(effect_i!=EE_HILIGHT2X)
            {
               lc.REPORT_ERR(C_fstring("Material '%s': invalid environment effect: %i", (const char*)str, effect_i));
            }
            */
            lc.env_effect = EE_HILIGHT2X;
               //(E_ENVIROMENTAL_EFFECT)Max(EE_NULL+1, Min(EE_LAST-1, effect_i));
         }
         break;

      case '&':               //procedural texture
         {
            char pname[128];
            pname[0] = 0;
            int num = 0;
            sscanf(&str[i], "%128s%n", pname, &num);
            i += num;
            if(!pname[0]){
               lc.REPORT_ERR(C_fstring("Material '%s': invalid procedural name", (const char*)str));
            }else{
               lc.txt_proc_name = pname;
            }
         }
         break;

      case '@':               //keep diffuse color
         lc.recent_txt_flags |= LOADF_TXT_DIFFUSE_KEEP;
         break;

      case '~':
         lc.recent_txt_flags |= LOADF_TXT_CAMERA;
         break;

      default:
         if(isalnum(c) || c=='_' || c==' ');
         else lc.REPORT_ERR(C_fstring("invalid material name: '%s'", (const char*)str));
      }
   }
   lc.recent_mat = lc.driver->CreateMaterial();
   lc.recent_mat->name = str;

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(str);
   }
   return true;
}

//----------------------------
                              //A010
static bool AmbientReader(S_load_context &lc){

   lc.recent_rgb = 0;
   bool b = lc.ChunkReader();
   if(b){
      lc.recent_mat->SetAmbient(lc.recent_rgb.ToVector());
   }
   return b;
};

//----------------------------
                              //A020
static bool DiffuseReader(S_load_context &lc){

   lc.recent_rgb = 0;
   bool b = lc.ChunkReader();
   if(b){
      lc.recent_mat->SetDiffuse(lc.recent_rgb.ToVector());
   }
   return b;
};

//----------------------------
                              //A030
static bool SpecularReader(S_load_context &lc){

   return true;
   /*
                              //unsupported
   lc.recent_rgb = 0;
   bool b = lc.ChunkReader();
   if(b){
      lc.recent_mat->SetSpecularColor(lc.recent_rgb.ToVector());
   }
   return b;
   */
};

//----------------------------
                              //A040
static bool ShininessReader(S_load_context &lc){

   return true;
   /*
                              //unsupported
   bool b = lc.ChunkReader();
   if(b){
      float f = (float)(100 - lc.recent_percentage) * .01f;;
      lc.recent_mat->SetSpecularAngle(f*PI*.5f);
   }
   return b;
   */
};

//----------------------------
#define SPECULAR_MIN_POWER_THRESH .1f
                              //A041
static bool ShinStrReader(S_load_context &lc){

   return true;
   /*
                              //unsupported
   bool b = lc.ChunkReader();
   if(b){
      float power = lc.recent_percentage * .01f;
      if(power < SPECULAR_MIN_POWER_THRESH) power = 0.0f;
      lc.recent_mat->SetSpecularPower(power);
   }
   return b;
   */
};

//----------------------------
                              //A050
static bool TransparencyReader(S_load_context &lc){

   lc.recent_percentage = 0;
   bool b = lc.ChunkReader();
   if(b)
      if(lc.recent_percentage){
         byte alpha_percentage = 100 - lc.recent_percentage;
         if(!alpha_percentage){
                              //zero opacity, issue warning
            lc.REPORT_ERR(C_fstring("Error: material '%s' has zero opacity", lc.recent_mat->GetName()));
         }
         lc.recent_mat->SetAlpha((float)alpha_percentage / 100.0f);
      }
   return b;
};

//----------------------------
                              //A052
static bool XPFallReader(S_load_context &lc){

   return lc.ChunkReader();
};

//----------------------------
                              //A053
static bool RefBlurReader(S_load_context &lc){

   return lc.ChunkReader();
};

//----------------------------
                              //A081
static bool Mat2SideReader(S_load_context &lc){

   lc.recent_mat->Set2Sided();
   return true;
};

//----------------------------
                              //A084
static bool SelfIllumReader(S_load_context &lc){
                              //expected percentage
   lc.recent_percentage = 0;
   bool b = lc.ChunkReader();
   if(b)
      lc.recent_self_illum = (float)lc.recent_percentage / 100.0f;
   return b;
};

//----------------------------
                              //A085
static bool WireMatReader(S_load_context &lc){

   return true;
};

//----------------------------
                              //A087
static bool WireSizeReader(S_load_context &lc){

   float f;
   lc.is->read((char*)&f, sizeof(f));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%.2f", f));
   }
   return true;
};

//----------------------------
                              //A08A
static bool XPFallInReader(S_load_context &lc){

   return true;
};

//----------------------------
                              //A08C
static bool PhongSoftReader(S_load_context &lc){

   return true;
};

//----------------------------
                              //A100
static bool MatTypeReader(S_load_context &lc){

   /*
   word w;
   static const struct{
      I3D_SHADETYPE type;
      char name[12];
   }surf[]={
      SHD_FLAT,    "???",
      SHD_FLAT,    "Flat",
      SHD_GOURAUD, "Gouraud",
      SHD_GOURAUD, "Phong",
      SHD_GOURAUD, "Metal"
   };
   lc.is->read((char*)&w, sizeof(w));
   if(w>4) w = 0;
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(surf[w].name);
   }
   */
   return true;
}

//----------------------------
                              //A200
static bool TextMapReader(S_load_context &lc){

   lc.curr_map = lc.diffuse_map = new S_load_context::S_map;
   lc.recent_percentage = 0;
   lc.recent_text_blur = 0.0f;
   bool b = lc.ChunkReader();
   if(b){
      lc.curr_map->percentage = lc.recent_percentage;
      lc.curr_map->anim_speed = FloatToInt(lc.recent_text_blur * 1000.0f);
      lc.curr_map = NULL;
   }
   return b;
}

//----------------------------
                              //A204
static bool SpecularMapReader(S_load_context &lc){

   return true;
}

//----------------------------
                              //A210
static bool OptMapReader(S_load_context &lc){

   lc.curr_map = lc.opacity_map = new S_load_context::S_map;
   bool b = lc.ChunkReader();
   lc.curr_map = NULL;
   return b;
}

//----------------------------
                              //A220
static bool ReflMapReader(S_load_context &lc){

   assert(!lc.environment_map);
   lc.curr_map = lc.environment_map = new S_load_context::S_map;
   lc.recent_percentage = 0;

   bool b = lc.ChunkReader();
   if(b){
      lc.curr_map->percentage = lc.recent_percentage;
   }
   lc.curr_map = NULL;
   return b;
}

//----------------------------
                              //A230
static bool BumpMapReader(S_load_context &lc){

   assert(!lc.bump_map);
   lc.curr_map = lc.bump_map = new S_load_context::S_map;
   lc.recent_percentage = 0;
   bool b = lc.ChunkReader();
   if(b){
      lc.curr_map->percentage = lc.recent_percentage;
      lc.curr_map->anim_speed = FloatToInt(lc.recent_text_blur * 1000.0f);
   }
   lc.curr_map = NULL;
   return b;
};

//----------------------------
                              //A300
static bool MapFileReader(S_load_context &lc){

   if(!lc.curr_map) return true;

   lc.ReadString(lc.curr_map->s);

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(lc.curr_map->s);
   }
   return true;
}

//----------------------------
                              //A351
static bool MapParams(S_load_context &lc){

   word w;
   lc.is->read((char*)&w, sizeof(w));
   if(lc.curr_map)
      lc.curr_map->inverse = (w&8);

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%x", w));
   }
   return true;
}

//----------------------------
                              //A353
static bool TextBlurReader(S_load_context &lc){

   float f;
   lc.is->read((char*)&f, sizeof(f));
                              //this chunk is used for various purposes
   lc.recent_text_blur = 1.0f + f;

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%.2f", f));
   }
   return true;
}

//----------------------------

static const S_vector unit_vector(1.0f, 1.0f, 1.0f);
                              //B002
static bool ObjNodeReader(dword block_id, S_load_context &lc){

   PI3D_object temp_obj = I3DCAST_OBJECT(lc.driver->CreateVisual(I3D_VISUAL_OBJECT));
   lc.recent_frame = temp_obj;
   lc.recent_properties = NULL;
   lc.recent_pivot.Zero();
   lc.recent_nu_scale = unit_vector;
   lc.recent_dummy_bbox.Invalidate();

   bool b = lc.ChunkReader();

   if(b){
      lc.ValidateName(lc.recent_instance_name, false);
      if(lc.root_frame){
                              //check properties
         C_string prop_data;
         E_PROP_KEYWORD kw_cast_to;
         GetPropetriesBegin(lc.recent_properties, prop_data, kw_cast_to);

         PI3D_mesh_object mesh = lc.mesh_list[lc.recent_string];

         if(mesh){
            bool db_loaded = false;
#ifdef USE_DATABASE
                              //build LODs only if we haven't loaded the mesh from cache
            db_loaded = (lc.mesh_db_valid.find(lc.recent_string)!=lc.mesh_db_valid.end());
#endif
            if(!db_loaded)
               mesh->GetLOD(0)->ApplyPivotAndNUScale(lc.recent_pivot, lc.recent_nu_scale);

            PI3D_object obj = temp_obj;
            lc.recent_frame = obj;
            obj->SetMeshObject(mesh);

            switch(kw_cast_to){
            case PROP_CAST_SNGM:
            case PROP_CAST_PRTC:
            case PROP_CAST_LMAP:
            case PROP_CAST_BBRD:
            case PROP_CAST_MRPH:
            case PROP_CAST_FLRE:
            case PROP_CAST_OUVS:
               {
                  const char *cp_cast_to = GetKeywordString(kw_cast_to);
                  dword dw_cast_type = *(dword*)cp_cast_to;

                  PI3D_frame new_visual = lc.driver->CreateVisual(dw_cast_type);
                  if(new_visual){

                     new_visual->Duplicate(obj);
                     obj->Release();
                     obj = NULL;
                     lc.recent_frame = new_visual;
                                 //initialize visual (if possible)
                     switch(kw_cast_to){

                     case PROP_CAST_PRTC:
                        {
                           PI3D_particle prt = (PI3D_particle)new_visual;
                           prt->SetMaterial(mesh->GetLOD1()->GetFGroups1()->mat);
                        }
                        break;

                     case PROP_CAST_LMAP:
                        {
                           PI3D_lit_object lmap = (PI3D_lit_object)new_visual;
                           lmap->SetMesh(mesh->GetLOD1());
                        }
                        break;

                     case PROP_CAST_FLRE:
                        {
                           PI3D_flare flr = (PI3D_flare)new_visual;

                              //read additional properties
                           float f;
                           int i;
                           const char *err = NULL;
                           if(GetPropertiesFloat(prop_data, f)){
                              flr->SetScaleDisance(f);
                              if(GetPropertiesFloat(prop_data, f)){
                                 flr->SetNormalAngle(f * PI / 180.0f);
                                 if(GetPropertiesFloat(prop_data, f)){
                                    flr->SetRotationRatio(f);
                                    if(GetPropertiesInt(prop_data, i)){
                                       flr->SetOcclusionFade(i);
                                       if(GetPropertiesFloat(prop_data, f)){
                                          flr->SetPivot(f);
                                       }//else
                                    }//else
                                       //err = "occlusion fade time expected";
                                 }//else
                                    //err = "rotation ratio expected";
                              }//else
                                 //err = "normal fade angle expected";
                           }//else
                              //err = "scale distance expected";

                           if(err){
                              lc.REPORT_ERR(C_fstring("Scanning properties on flare '%s': %s",
                                 (const char*)lc.recent_instance_name, err));
                           }
                        }
                        break;
                     }
                  }else{
                     lc.REPORT_ERR(C_fstring("%s: cannot cast visual object: '%s'",
                        (const char*)lc.recent_instance_name, cp_cast_to));
                  }
               }
               break;

            case PROP_CAST_OCCL:
               {              //create and init occluder
                  PI3D_occluder occ = I3DCAST_OCCLUDER(lc.driver->CreateFrame(FRAME_OCCLUDER));
                  occ->LinkTo(lc.recent_frame);

                  lc.container->AddFrame(occ);
                  occ->SetName(C_fstring("occl %s", (const char*)lc.recent_instance_name));

                  PI3D_mesh_base mb = mesh->GetLOD1();
                  const I3D_vertex_mesh *verts = (const I3D_vertex_mesh*)mb->vertex_buffer.Lock(D3DLOCK_READONLY);
                  occ->Build(&verts->xyz, mb->vertex_buffer.NumVertices(), sizeof(I3D_vertex_mesh));
                  mb->vertex_buffer.Unlock();
                  occ->Release();
               }
               break;
            }

            if(lc.recent_frame){
                                //process modifiers
               C_string str_prop;
               E_PROP_KEYWORD kw;
               if(GetPropetriesBegin(lc.recent_properties, str_prop, kw))
               do{
                  switch(kw){
                  case PROP_AUTO_LOD:
                     if(lc.recent_frame->GetType1()==FRAME_VISUAL &&
                        I3DCAST_VISUAL(lc.recent_frame)->GetVisualType1()==I3D_VISUAL_SINGLEMESH)
                        break;
                     {
                              //build LODs only if we haven't loaded the mesh from cache
                        if(db_loaded)
                           break;

                        PI3D_visual vis = I3DCAST_VISUAL(lc.recent_frame);
                              //read additional properties
                        float min_dist = 5.0f;
                        float max_dist = 50.0f;
                        int num_parts = 10;
                        float lowest_quality = .1f;

                        const char *err = NULL;

                        if(GetPropertiesFloat(str_prop, min_dist)){
                           if(GetPropertiesFloat(str_prop, max_dist)){
                              if(GetPropertiesInt(str_prop, num_parts)){
                                 if(!GetPropertiesFloat(str_prop, lowest_quality))
                                    err = "lowest quality expected";
                                 {};
                              }else
                                 err = "number of parts expected";
                           }else
                              err = "maximal distance expected";
                        }else
                           err = "minimal distance expected";

                        if(err)
                           lc.REPORT_ERR(C_fstring("Scanning properties on flare '%s': auto_lod - %s",
                              (const char*)lc.recent_instance_name, err));

                        if(vis->GetMesh()){
                           vis->GetMesh()->CreateLODs(lowest_quality, lowest_quality,
                              (1.0f-lowest_quality) / num_parts, min_dist, max_dist);
                        }
                     }
                     break;

                  case PROP_HIDE:
                     if(lc.recent_frame->GetType1()==FRAME_VISUAL){
                        PI3D_visual vis = I3DCAST_VISUAL(lc.recent_frame);
                        float hide_dist;
                        if(!GetPropertiesFloat(str_prop, hide_dist)){
                           lc.REPORT_ERR(C_fstring("Scanning properties on '%s': hide - distance expected",
                              (const char*)lc.recent_instance_name));
                           break;
                        }
                        vis->SetHideDistance(hide_dist);
                     }
                     break;

                  case PROP_MATERIAL:
                     {
                        int mat_id = 0;
                        if(!GetPropertiesInt(str_prop, mat_id))
                           lc.REPORT_ERR(C_fstring("Scanning properties on '%s': expected material number",
                              (const char*)lc.recent_instance_name));

                        if(lc.recent_frame->GetType1()==FRAME_VISUAL){
                           PI3D_visual vis = I3DCAST_VISUAL(lc.recent_frame);
                           vis->SetCollisionMaterial(mat_id);
                           vis->SetFlags(I3D_FRMF_STATIC_COLLISION, I3D_FRMF_STATIC_COLLISION);
                        }else{
                           lc.REPORT_ERR(C_fstring("Scanning properties on '%s': can't specify material ID for non-visual frame",
                              (const char*)lc.recent_instance_name));
                        }
                     }
                     break;

                  case PROP_PROCEDURAL:
                     {
                              //get and attach procedural to frame
                        C_string proc_name;
                        if(!GetPropertiesString(str_prop, proc_name)){
                           lc.REPORT_ERR(C_fstring("Scanning properties on '%s': expected procedure name",
                              (const char*)lc.recent_instance_name));
                           break;
                        }
                        I3D_procedural_base *pb = lc.driver->GetProcedural(proc_name, lc, lc.recent_instance_name, NULL);
                        if(pb){
                           if(lc.recent_frame->GetType1()==FRAME_VISUAL){
                                          //try to attach this procedural to visual object
                              bool st = I3DCAST_VISUAL(lc.recent_frame)->AttachProcedural(pb);
                              if(!st){
                                 lc.REPORT_ERR(C_fstring("Visual '%s': failed to attach procedural '%s'",
                                    (const char*)lc.recent_instance_name, (const char*)proc_name));
                              }
                           }else{
                              assert(0);  //report err - attaching procedural to non-visual
                           }
                           pb->Release();
                        }
                     }
                     break;
                  }
               }while(GetPropertiesNext(str_prop, kw));
            }
#ifdef USE_DATABASE
                              //store mesh into cache, if not yet

            if(lc.recent_frame->GetType1()==FRAME_VISUAL){
               PI3D_visual vis = I3DCAST_VISUAL(lc.recent_frame);
               if(vis->GetVisualType1()!=I3D_VISUAL_SINGLEMESH)
               if(vis->GetMesh() && lc.mesh_db_save.find(lc.recent_string)!=lc.mesh_db_save.end()){

                  C_fstring cache_name("%s+%s", (const char*)lc.file_name, (const char*)lc.recent_string);
                  C_cache *ck = lc.driver->dbase.CreateRecord(cache_name, lc.file_time);
                  if(ck){
                     vis->GetMesh()->SaveCachedInfo(ck);
                     lc.driver->dbase.CloseRecord(ck);
                  }
               }
            }
#endif
         }else{
            {

               bool make_dummy = true;
               if(block_id == 0xb00e){
                              //bone, automatically cast to joint
                  lc.recent_frame = lc.driver->CreateFrame(FRAME_JOINT);
                  lc.recent_frame->Duplicate(temp_obj);
                  make_dummy = false;

                  if(prop_data && !strncmp(prop_data, "lock", 4)){
                     I3DCAST_JOINT(lc.recent_frame)->LockPosition(true);
                  }
               }else{
                  switch(kw_cast_to){
                  case PROP_CAST_JTBB:
                  case PROP_CAST_JTVB:
                     {
                                 //bounding-box for parent joint or SM
                        PI3D_frame prnt = NULL;
                        if(lc.recent_parent_node_id!=-1){
                           PI3D_frame frm = lc.GetFrame(lc.recent_parent_node_id);
                           if(frm && (frm->GetType1()==FRAME_JOINT || (frm->GetType1()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_SINGLEMESH)))
                              //jnt = I3DCAST_JOINT(frm);
                              prnt = frm;
                        }
                        if(prnt){
                           I3D_bbox bb = lc.recent_dummy_bbox;
                                 //scale bbox by nu-scale
                           bb.min *= lc.recent_nu_scale;
                           bb.max *= lc.recent_nu_scale;
                                 //setup bounding-box of joint
                           temp_obj->Update();
                           switch(kw_cast_to){
                           case PROP_CAST_JTBB:
                              if(prnt->GetType1()==FRAME_JOINT)
                                 I3DCAST_JOINT(prnt)->SetBBox(bb, temp_obj->GetFrameMatrix());
                              break;
                           case PROP_CAST_JTVB:
                              switch(prnt->GetType1()){
                              case FRAME_JOINT:
                                 I3DCAST_JOINT(prnt)->SetVolumeBox(bb, temp_obj->GetFrameMatrix());
                                 break;
                              case FRAME_VISUAL:
                                 ((PI3D_object_singlemesh)prnt)->SetVolumeBox(bb, temp_obj->GetFrameMatrix());
                                 break;
                              }
                              break;
                           }

                           make_dummy = false;
                           lc.recent_frame = NULL;
                        }else
                           lc.REPORT_ERR(C_fstring("dummy '%s': cannot find parent joint or single-mesh object",
                              (const char*)lc.recent_instance_name));
                     }
                     break;

                  case PROP_REGION:
                     {
                              //region for parent visual
                        PI3D_visual vis = NULL;
                        if(lc.recent_parent_node_id!=-1){
                           PI3D_frame frm = lc.GetFrame(lc.recent_parent_node_id);
                           if(frm && frm->GetType1()==FRAME_VISUAL)
                              vis = I3DCAST_VISUAL(frm);
                        }
                        if(vis){
                           I3D_bbox bb = lc.recent_dummy_bbox;
                                       //scale bbox by nu-scale
                           bb.min *= lc.recent_nu_scale;
                           bb.max *= lc.recent_nu_scale;
                                       //setup bounding-box
                           temp_obj->Update();

                           int indx = -1;
                           GetPropertiesInt(prop_data, indx);

                           bool ok = vis->AddRegion(bb, temp_obj->GetFrameMatrix(), indx);
                           if(!ok){
                              lc.REPORT_ERR(C_fstring("Failed to set region on visual '%s'",
                                 vis->GetName()));
                           }
                           make_dummy = false;
                           lc.recent_frame = NULL;
                        }else{
                           lc.REPORT_ERR(C_fstring("Region '%s': cannot find parent visual",
                              (const char*)lc.recent_instance_name));
                        }
                     }
                     break;

                  case PROP_CAST_VOLS:
                  case PROP_CAST_VOLB:
                  case PROP_CAST_VOLR:
                     {
                              //create collision volume
                        PI3D_volume vol = lc.driver->CreateVolume();

                        lc.recent_frame = vol;
                        lc.recent_frame->Duplicate(temp_obj);

                        I3D_bbox bb = lc.recent_dummy_bbox;
                              //scale bbox by nu-scale
                        bb.min *= lc.recent_nu_scale;
                        bb.max *= lc.recent_nu_scale;

                        const S_vector &base_scale = bb.max;

                        float avg_scale = base_scale.Sum() / 3.0f;
                        lc.recent_frame->SetScale(avg_scale * lc.recent_frame->GetScale());
                        S_vector nu_scale = base_scale / avg_scale;
                        vol->SetNUScale(nu_scale);

                        switch(kw_cast_to){
                        case PROP_CAST_VOLS:
                           {
                              vol->SetVolumeType(I3DVOLUME_SPHERE);
                           }
                           break;
                        case PROP_CAST_VOLB:
                           {
                              vol->SetVolumeType(I3DVOLUME_BOX);
                           }
                           break;
                        case PROP_CAST_VOLR:
                           {
                              vol->SetVolumeType(I3DVOLUME_RECTANGLE);
                           }
                           break;
                        }
                        vol->SetScene(lc.scene);
                        vol->Update();
      
                        make_dummy = false;

                                //process modifiers
                        C_string str_prop;
                        E_PROP_KEYWORD kw;
                        if(GetPropetriesBegin(lc.recent_properties, str_prop, kw))
                        do{
                           switch(kw){
                           case PROP_MATERIAL:
                              {
                                 int mat_id = 0;
                                 if(!GetPropertiesInt(str_prop, mat_id))
                                    lc.REPORT_ERR(C_fstring("Scanning properties on '%s': expected material number",
                                       (const char*)lc.recent_instance_name));
                                 vol->SetCollisionMaterial(mat_id);
                              }
                              break;
                           }
                        }while(GetPropertiesNext(str_prop, kw));
                     }
                     break;
                  }
               }
               if(make_dummy){
                              //create dummy frame
                  PI3D_dummy dum = lc.driver->CreateDummy();

                  if(lc.recent_dummy_bbox.IsValid()){
                              //scale bbox by nu-scale
                     lc.recent_dummy_bbox.min *= lc.recent_nu_scale;
                     lc.recent_dummy_bbox.max *= lc.recent_nu_scale;
                     dum->SetBBox(lc.recent_dummy_bbox);
                  }
                  lc.recent_frame = dum;
                  lc.recent_frame->Duplicate(temp_obj);
               }
            }
            temp_obj->Release(); temp_obj = NULL;
         }
         if(lc.recent_frame){
            lc.container->AddFrame(lc.recent_frame);

            lc.recent_frame->name = lc.recent_instance_name;

            if(lc.recent_anim && !lc.anim_set && lc.scene)
               lc.scene->AddAnimation(lc.recent_anim, lc.recent_frame, I3DANIMOP_LOOP);

            lc.AddFrameID(lc.recent_node_id, lc.recent_frame);
                              //add to hierarchy
            if(lc.recent_parent_node_id==-1){
               lc.recent_frame->LinkTo(lc.root_frame);
            }else{
               PI3D_frame frm = lc.GetFrame(lc.recent_parent_node_id);
               if(frm) lc.recent_frame->LinkTo(frm);
               else
                  lc.REPORT_ERR(C_fstring("frame '%s': cannot find parent object", (const char*)lc.recent_instance_name));
            }
            lc.recent_frame->Update();
         }
      }
      if(lc.anim_set && lc.recent_anim && lc.recent_frame)
         lc.anim_set->AddAnimation(lc.recent_anim, lc.recent_instance_name);
   }

   if(lc.recent_anim){
      lc.recent_anim->Release(); lc.recent_anim = NULL;
   }
   lc.recent_instance_name = NULL;
   if(lc.recent_frame){
                              //save properties for later use (during loading)
      if(lc.recent_properties.Size())
         lc.frm_properties[lc.recent_frame] = lc.recent_properties;

      lc.recent_frame->Release(); lc.recent_frame = NULL;
   }

   if(b){
      if(lc.ShowProgress())
         return false;
   }

   return b;
}

//----------------------------
                              //B002
static bool ObjNodeReader(S_load_context &lc){
   return ObjNodeReader(0xb002, lc);
}

//----------------------------
                              //B00E
static bool BoneNodeReader(S_load_context &lc){
   return ObjNodeReader(0xb00e, lc);
}

//----------------------------
                              //B003
static bool CamNodeReader(S_load_context &lc){

   lc.recent_frame = lc.driver->CreateUserFrame();

   bool b = lc.ChunkReader();

   if(b){
      const PI3D_frame *frames = lc.container->GetFrames();
      for(int i=lc.container->NumFrames1(); i--; ){
         if(lc.recent_string.Match(frames[i]->GetName()))
            break;
      }
      if(i!=-1 && frames[i]->GetType1()==FRAME_CAMERA){
         PI3D_camera cam = I3DCAST_CAMERA(frames[i]);
         lc.ValidateName(lc.recent_string, false);

         cam->Duplicate(lc.recent_frame);
         cam->name = lc.recent_instance_name;

         lc.AddFrameID(lc.recent_node_id, cam);
                                    //add to hierarchy
         if(lc.recent_parent_node_id!=-1){
            PI3D_frame frm = lc.GetFrame(lc.recent_parent_node_id);
            if(frm) cam->LinkTo(frm);
            else
               lc.REPORT_ERR(C_fstring("camera '%s': cannot find parent object", (const char*)lc.recent_instance_name));
         }
         cam->Update();
      }
   }
   if(lc.recent_anim){
      lc.recent_anim->Release(); lc.recent_anim = NULL;
   }
   lc.recent_frame->Release(); lc.recent_frame = NULL;
   return b;
}

//----------------------------
                              //B005 - omni
                              //B007 - spot
static bool LightNodeReader(S_load_context &lc){

   //if(!lc.root_frame) return true;
   lc.recent_frame = lc.driver->CreateUserFrame();

   bool b = lc.ChunkReader();

   if(b){
      lc.ValidateName(lc.recent_string, false);
      if(lc.root_frame){
         PI3D_frame frm = lc.FindStockItem(lc.light_list, lc.recent_string);
         if(frm){
            lc.container->AddFrame(frm);
            frm->SetPos(lc.recent_frame->GetPos());
            frm->SetScale(lc.recent_frame->GetScale());

            lc.AddFrameID(lc.recent_node_id, frm);

            if(lc.recent_parent_node_id==-1){
               frm->LinkTo(lc.root_frame);
            }else{
               PI3D_frame frm1a = lc.GetFrame(lc.recent_parent_node_id);
               if(frm1a) frm->LinkTo(frm1a);
               else
                  lc.REPORT_ERR(C_fstring("light '%s': cannot find parent object", (const char*)lc.recent_instance_name));
            }
            frm->Update();

            if(lc.recent_anim){
               if(lc.anim_set)
                  lc.anim_set->AddAnimation(lc.recent_anim, lc.recent_instance_name);
               else
               if(lc.scene)
                  lc.scene->AddAnimation(lc.recent_anim, frm, I3DANIMOP_LOOP);
            }
         }
      }
   }
   if(lc.recent_anim){
      lc.recent_anim->Release(); lc.recent_anim = NULL;
   }
   lc.recent_frame->Release(); lc.recent_frame = NULL;

   lc.recent_instance_name = NULL;

   return b;
}

//----------------------------
                              //B004
                              //B006
static bool TargetNodeReader(S_load_context &lc){

   //if(!lc.root_frame) return true;

   PI3D_target tp;
   lc.recent_frame = tp = lc.driver->CreateTarget();

   bool b = lc.ChunkReader();

   if(b){
      lc.ValidateName(lc.recent_string, true);
      if(lc.root_frame){
                              //setup target
         PI3D_frame target = lc.FindStockItem(lc.light_list, lc.recent_string);
         if(!target){
            const PI3D_frame *frames = lc.container->GetFrames();
            for(int i=lc.container->NumFrames1(); i--; ){
               if(lc.recent_string.Match(frames[i]->GetName())){
                  target = frames[i];
                  break;
               }
            }
         }
         if(target){
            tp->SetFrame(target);
            lc.container->AddFrame(lc.recent_frame);

            lc.recent_frame->name = lc.recent_instance_name + C_string("_target");

            if(lc.recent_anim && !lc.anim_set && lc.scene)
               lc.scene->AddAnimation(lc.recent_anim, lc.recent_frame, I3DANIMOP_LOOP);

            lc.AddFrameID(lc.recent_node_id, lc.recent_frame);
                                       //add to hierarchy
            if(lc.recent_parent_node_id==-1){
               lc.recent_frame->LinkTo(lc.root_frame);
            }else{
               PI3D_frame frm = lc.GetFrame(lc.recent_parent_node_id);
               if(frm) lc.recent_frame->LinkTo(frm);
               else
                  lc.REPORT_ERR(C_fstring("target '%s': cannot find parent object", (const char*)lc.recent_instance_name));
            }
            lc.recent_frame->Update();
            if(lc.anim_set && lc.recent_anim)
               lc.anim_set->AddAnimation(lc.recent_anim, lc.recent_instance_name);
         }
      }
   }
   if(lc.recent_anim){
      lc.recent_anim->Release(); lc.recent_anim = NULL;
   }
   lc.recent_frame->Release(); lc.recent_frame = NULL;
   lc.recent_instance_name = NULL;

   return b;
}

//----------------------------
                              //B008
static bool FramesReader(S_load_context &lc){

   struct S_anim_header{
      dword frame_start, frame_end;
   } anim_header;
   lc.is->read((char*)&anim_header, sizeof(anim_header));
   lc.recent_max_anim_time = anim_header.frame_end*1000/ANIM_FPS;

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Start: %i, end: %i (%.2f sec)", anim_header.frame_start,
         anim_header.frame_end, (float)anim_header.frame_end/(float)ANIM_FPS));
   }
   return true;
}

//----------------------------
                              //B009
static bool KFCurrTimeReader(S_load_context &lc){

   dword dw;
   lc.is->read((char*)&dw, sizeof(dw));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%u", dw));
   }
   return true;
}

//----------------------------
                              //B00A
static bool KFHeaderReader(S_load_context &lc){

   word w;
   lc.is->read((char*)&w, sizeof(w));
   C_string s;
   lc.ReadString(s);
   struct S_kf_header{
      word anim_size;
      word unk1;
   } kf_header;
   lc.is->read((char*)&kf_header, sizeof(kf_header));
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("ver. %u \"%s\" anim. size = %i, unk = %i", w, (const char*)s, kf_header.anim_size, kf_header.unk1));
   }
   return true;
}

//----------------------------
                              //B00B (I3D MAX custom export)
static bool ObjInfoReader(S_load_context &lc){

   C_string str;
   lc.ReadString(str);
   int size = str.Size();
   if(size){
      size = dtaBin2Text(&str[0], size);
      str[size] = 0;
      assert(lc.recent_frame);
      //lc.recent_frame->SetPropertyBuffer(&str[0], size+1);
      lc.recent_properties = str;
      if(lc.load_flags&I3DLOAD_LOG) lc.LOG(str);
   }
   return true;
}

//----------------------------
                              //B010
static bool NodeHeaderReader(S_load_context &lc){

   lc.ReadString(lc.recent_string);

   struct{
      word flags;
      word flags2;
      short parent_id;
   } linfo;
   lc .is->read((char*)&linfo, sizeof(linfo));
                              //for case object is not instance
   if(lc.recent_instance_name) lc.recent_instance_name = lc.recent_string;

   lc.recent_parent_node_id = linfo.parent_id;
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%s, %i, %i, parent id: %i", (const char*)lc.recent_string,
         linfo.flags, linfo.flags2, linfo.parent_id));
   }
   return true;
}

//----------------------------
                              //B011
static bool InstNameReader(S_load_context &lc){

   lc.ReadString(lc.recent_instance_name);
   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(lc.recent_instance_name);
   }
   return true;
}

//----------------------------
                              //B013
static bool NodePivotReader(S_load_context &lc){

   if(!lc.recent_frame)
      return true;

   S_vector v;
   lc.is->read((char*)&v, sizeof(v));
   swap(v.y, v.z);

   lc.recent_pivot = v;

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("x: %.2f, y: %.2f, z: %.2f", v.x, v.y, v.z));
   }
   return true;
}

//----------------------------
                              //B014
static bool BoundBoxReader(S_load_context &lc){

   lc.is->read((char*)&lc.recent_dummy_bbox, sizeof(I3D_bbox));
   swap(lc.recent_dummy_bbox.min.y, lc.recent_dummy_bbox.min.z);
   swap(lc.recent_dummy_bbox.max.y, lc.recent_dummy_bbox.max.z);

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("v[0]: %.2f, %.2f, %.2f", lc.recent_dummy_bbox.min.x, lc.recent_dummy_bbox.min.y, lc.recent_dummy_bbox.min.z));
      lc.LOG(C_fstring("v[1]: %.2f, %.2f, %.2f", lc.recent_dummy_bbox.max.x, lc.recent_dummy_bbox.max.y, lc.recent_dummy_bbox.max.z));
   }
   return true;
}

//----------------------------
#pragma pack(push, 1)

struct S_track_header{
   word unk[3];
   word max_frame;            //? # of total frames?
   word unk2;                 //0
   word num_keys;
   word unk3;                 //0
};

struct S_key_header{
   short time;
   word unk;
   word flags;
};

struct S_key_tcb{
   float tension;             //-1.0 to 1.0
   float continuity;          //-1.0 to 1.0
   float bias;                //-1.0 to 1.0
   float easy_from;           //0.0 to 1.0
   float easy_to;             //0.0 to 1.0
   float unk[2];

   void Read(C_cache *is, dword flags){
      if(flags&1) is->read((char*)&tension, sizeof(float));
      else tension=0.0f;
      if(flags&2) is->read((char*)&continuity, sizeof(float));
      else continuity=0.0f;
      if(flags&4) is->read((char*)&bias, sizeof(float));
      else bias=0.0f;
      if(flags&8) is->read((char*)&easy_to, sizeof(float));
      else easy_to=0.0f;
      if(flags&0x10) is->read((char*)&easy_from, sizeof(float));
      else easy_from=0.0f;
      if(flags&0x20) is->read((char*)&unk[0], sizeof(float));
   }
};

#pragma pack(pop)

struct S_position_key{
   S_key_header header;
   S_key_tcb add_data;
   S_vector data;

   void Read(C_cache *is){
      is->read((char*)&header, sizeof(header));
      add_data.Read(is, header.flags);
      is->read((char*)&data, sizeof(data));
   }
};

                              //B020
static bool PosTrackReader(S_load_context &lc){

   S_track_header header;
   lc.is->read((char*)&header, sizeof(header));

   dword num_keys=Max(1ul, (dword)header.num_keys);

   I3D_anim_pos_tcb *akeys;

   bool ani_track = (num_keys>1);
   if(ani_track){
#ifndef USE_ALLOCA
      akeys = new I3D_anim_pos_tcb[num_keys];
#else
      akeys = (I3D_anim_pos_tcb*)alloca(sizeof(I3D_anim_pos_tcb)*num_keys);
#endif
   }
   S_position_key key;

   for(int i=0; i<num_keys; i++){
      key.Read(lc.is);
      swap(key.data.y, key.data.z);

      if(ani_track){
         /*
                              //detect redundant keys 
         if(!lc.anim_set && i && i!=num_keys-1){
            const S_vector &last_key = akeys[i-1].v;
            if(IsAbsMrgZero(key.data.x-last_key.x) &&
               IsAbsMrgZero(key.data.y-last_key.y) &&
               IsAbsMrgZero(key.data.z-last_key.z)){
                              //check if key is not redundant
               lc.REPORT_ERR(C_fstring("Frame: '%s': redundant animation position key on time %i.",
                  (const char*)lc.recent_instance_name, key.header.time));
            }
         }
         */
         akeys[i].time = key.header.time*1000/ANIM_FPS;
         akeys[i].v = key.data;
         akeys[i].tension = key.add_data.tension;
         akeys[i].continuity = key.add_data.continuity;
         akeys[i].bias = key.add_data.bias;
         akeys[i].easy_from = key.add_data.easy_from;
         akeys[i].easy_to = key.add_data.easy_to;
      }
      if(!i){
         if(lc.recent_frame) lc.recent_frame->SetPos(key.data);

         if(lc.load_flags&I3DLOAD_LOG){
            lc.LOG(C_fstring("Number of keys: %i", header.num_keys));
            lc.LOG(C_fstring("Base translation x: %.2f, y: %.2f, z: %.2f", 
               key.data.x, key.data.y, key.data.z));
         }
      }
   }

   if(ani_track){
      if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
      lc.recent_anim->SetEndTime(Max((dword)key.header.time*1000/ANIM_FPS, lc.recent_anim->GetEndTime()));
      lc.recent_anim->SetPositionKeys(akeys, num_keys);
#ifndef USE_ALLOCA
      delete[] akeys;
#endif
   }
   return true;
}

//----------------------------

struct S_rotation_key{
   S_key_header header;
   S_key_tcb add_data;
   struct S_data{
      float angle;
      S_vector axis;
   } data;

   void Read(C_cache *is){
      is->read((char*)&header, sizeof(header));
      add_data.Read(is, header.flags);
      is->read((char*)&data, sizeof(data));
   }
};

//----------------------------
                              //B021
static bool RotTrackReader(S_load_context &lc){

   S_track_header header;
   lc.is->read((char*)&header, sizeof(header));
   dword num_keys = Max(1ul, (dword)header.num_keys);

   I3D_anim_rot *akeys;
   if(num_keys>1){
#ifndef USE_ALLOCA
      akeys = new I3D_anim_rot[num_keys];
#else
      akeys = (I3D_anim_rot*)alloca(sizeof(I3D_anim_rot)*num_keys);
#endif
   }

   int i;

   S_vector axis;
   S_rotation_key key;

   for(i=0; i<num_keys; i++){
      key.Read(lc.is);
      axis = S_vector(-key.data.axis.x, -key.data.axis.z, -key.data.axis.y);

      if(num_keys>1){
         /*
                              //detect redundant keys 
         if(!lc.anim_set && i && i!=num_keys-1){
            if(IsAbsMrgZero(key.data.angle)){
                              //check if key is not redundant
               lc.REPORT_ERR(C_fstring("Frame: '%s': redundant animation rotation key on time %i",
                  (const char*)lc.recent_instance_name, key.header.time));
            }
         }
         */
         akeys[i].time = (key.header.time)*1000/ANIM_FPS;
         akeys[i].axis = axis;
         akeys[i].angle = key.data.angle;
         akeys[i].easy_from = key.add_data.easy_from;
         akeys[i].easy_to = key.add_data.easy_to;
      }
      if(!i){
         if(lc.recent_frame){
            S_quat q;
            q.Make(axis, key.data.angle);
            lc.recent_frame->SetRot(q);
         }
         if(lc.load_flags&I3DLOAD_LOG){
            lc.LOG(C_fstring("Number of keys: %i", header.num_keys));
            lc.LOG(C_fstring("%i, %i, %i, %i, %i", header.unk[0],
               header.unk[1], header.unk[2], header.unk2, header.unk3));
            lc.LOG(C_fstring("Base rotation angle: %.2f, x: %.2f, y: %.2f, z: %.2f", 
               key.data.angle, axis.x, axis.y, axis.z));
         }
      }
   }
   if(num_keys>1){
      if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
      lc.recent_anim->SetEndTime(Max((dword)key.header.time*1000/ANIM_FPS, lc.recent_anim->GetEndTime()));
      lc.recent_anim->SetRotationKeys(akeys, num_keys);
#ifndef USE_ALLOCA
      delete[] akeys;
#endif
   }
   return true;
}

//----------------------------

struct S_scale_key{
   S_key_header header;
   S_key_tcb add_data;
   S_vector data;

   void Read(C_cache *is){
      is->read((char*)&header, sizeof(header));
      add_data.Read(is, header.flags);
      is->read((char*)&data, sizeof(data));
   }
};

inline bool IsUniformScale(const S_vector &scale){
   return ((fabs(scale.x-scale.y) + fabs(scale.y-scale.z)) < 1e-4f);
}

//----------------------------

static bool ScaleTrackReaderOld(S_load_context &lc){

   S_track_header header;
   lc.is->read((char*)&header, sizeof(header));
   dword num_keys = Max(1ul, (dword)header.num_keys);
   dword num_scale_keys = 0;
   dword num_note_keys = 0;
   dword num_vis_keys = 0;

   I3D_anim_scale_tcb *skeys;
   I3D_anim_visibility *vkeys;
   if(num_keys>1){
      skeys = new I3D_anim_scale_tcb[num_keys];
      vkeys = new I3D_anim_visibility[num_keys];
   }

   I3D_anim_note *nkeys;
   nkeys = new I3D_anim_note[num_keys];
   char *cp = new char[num_keys*3], *cp_curr = cp;
   bool is_uni_scale = true;
   S_vector base_scale;

   S_scale_key key;
   for(int i=0; i<num_keys; i++){
      key.Read(lc.is);
      swap(key.data.y, key.data.z);

                              //use continuity values in range 20.0 to 30.0
                              //multiplied by 10 for notify data
                              //this gives total 101 notify codes
                              //note: we convert float value to dword,
                              //so precision is limited
      float f = (key.add_data.continuity + 1.0f) * 250.0f;
      dword key_data = FloatToInt(floor(f + .5f));
      if(key_data<100){       //notify
         nkeys[num_note_keys].time=(key.header.time)*1000/ANIM_FPS;
         nkeys[num_note_keys].note = cp_curr;
         sprintf(cp_curr, "%i", key_data);
         cp_curr += strlen(cp_curr) + 1;
         ++num_note_keys;
      }else
      if(key_data>=499){      //visibility
         if(num_keys>1){
            vkeys[num_vis_keys].time = (key.header.time)*1000/ANIM_FPS;
            vkeys[num_vis_keys].on = (key_data==500);
            ++num_vis_keys;
         }
         if(key.header.time<0){
            if(lc.recent_frame && key_data==499)
               lc.recent_frame->SetOn(false);
         }
      }else{                  //scale
         is_uni_scale = (is_uni_scale && IsUniformScale(key.data));
         if(num_keys>1){
            skeys[num_scale_keys].time = (key.header.time)*1000/ANIM_FPS;
            skeys[num_scale_keys].v = key.data;
            skeys[num_scale_keys].tension = key.add_data.tension;
            skeys[num_scale_keys].continuity = key.add_data.continuity;
            skeys[num_scale_keys].bias = key.add_data.bias;
            skeys[num_scale_keys].easy_from = key.add_data.easy_from;
            skeys[num_scale_keys].easy_to = key.add_data.easy_to;
            ++num_scale_keys;
         }
      }
      if(!i)
         base_scale = key.data;
   }
   if(lc.recent_frame){
      if(is_uni_scale){
         lc.recent_frame->SetScale(base_scale.x);
      }else{
         float avg_scale = base_scale.Sum() / 3.0f;
         lc.recent_frame->SetScale(avg_scale);
         lc.recent_nu_scale = base_scale / avg_scale;
      }
   }

   if(num_keys>1){
      if(num_scale_keys || num_vis_keys){
         if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
         lc.recent_anim->SetEndTime(Max((dword)key.header.time*1000/ANIM_FPS, lc.recent_anim->GetEndTime()));
         if(num_scale_keys>1){
            if(!is_uni_scale){
                              //set scale keys
               lc.recent_anim->SetScaleKeys(skeys, num_scale_keys);
            }else{
                              //set power keys
               I3D_anim_power *pkeys = (I3D_anim_power*)alloca(sizeof(I3D_anim_power)*num_scale_keys);
               for(i=0; i<num_scale_keys; i++){
                  pkeys[i].time = skeys[i].time;
                  pkeys[i].power = skeys[i].v.x;
               }
               lc.recent_anim->SetPowerKeys(pkeys, num_scale_keys);
            }
         }
         if(num_vis_keys)
            lc.recent_anim->SetVisibilityKeys(vkeys, num_vis_keys);
      }
      delete[] skeys;
      delete[] vkeys;
   }
   if(num_note_keys){
      if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
      lc.recent_anim->SetEndTime(Max((dword)key.header.time*1000/ANIM_FPS, lc.recent_anim->GetEndTime()));
      lc.recent_anim->SetNoteKeys(nkeys, num_note_keys);
   }
   delete[] nkeys;
   delete[] cp;
   return true;
}

//----------------------------
                              //B022
                              //scale - currently used for
                              // - anim notifications
static bool ScaleTrackReader(S_load_context &lc){

   if(lc.load_flags&I3DLOAD_ANIM_COMPATIBILITY)
      return ScaleTrackReaderOld(lc);

   S_track_header header;
   lc.is->read((char*)&header, sizeof(header));
   dword num_keys = Max(1ul, (dword)header.num_keys);

   I3D_anim_scale_tcb *skeys;
   if(num_keys>1){
#ifndef USE_ALLOCA
      skeys = new I3D_anim_scale_tcb[num_keys];
#else
      skeys = (I3D_anim_scale_tcb*)alloca(sizeof(I3D_anim_scale_tcb)*num_keys);
#endif
   }

   S_scale_key key;
   bool scale_err_reported = false;

   bool is_uni_scale = true;
   S_vector base_scale;

   for(int i=0; i<num_keys; i++){
      key.Read(lc.is);
      swap(key.data.y, key.data.z);

      if(num_keys>1){
                              //detect redundant keys 
         if(!lc.anim_set && i){
            const S_vector &last_key = skeys[i-1].v;
            if(IsAbsMrgZero(key.data.x-last_key.x) &&
               IsAbsMrgZero(key.data.y-last_key.y) &&
               IsAbsMrgZero(key.data.z-last_key.z)){
                              //check if key is not redundant
               lc.REPORT_ERR(C_fstring("Frame: '%s': redundant animation scale key on time %i",
                  (const char*)lc.recent_instance_name, key.header.time));
            }
         }

         skeys[i].time = (key.header.time)*1000/ANIM_FPS;
         skeys[i].v = key.data;
         skeys[i].tension = key.add_data.tension;
         skeys[i].continuity = key.add_data.continuity;
         skeys[i].bias = key.add_data.bias;
         skeys[i].easy_from = key.add_data.easy_from;
         skeys[i].easy_to = key.add_data.easy_to;
      }
                              //negative scale check
      if(key.data.x<0.0f || key.data.y<0.0f || key.data.z<0.0f){
         if(!scale_err_reported){
            scale_err_reported = true;
            lc.REPORT_ERR(C_fstring("Frame: '%s': negative scale", (const char*)lc.recent_instance_name));
         }
      }

      is_uni_scale = (is_uni_scale && IsUniformScale(key.data));
      if(!i){
         base_scale = key.data;
         if(lc.load_flags&I3DLOAD_LOG){
            lc.LOG(C_fstring("Number of keys: %i", header.num_keys));
            lc.LOG(C_fstring("%i, %i, %i, %i, %i", header.unk[0],
               header.unk[1], header.unk[2], header.unk2, header.unk3));
            lc.LOG(C_fstring("Base scale x: %.2f, y: %.2f, z: %.2f", 
               key.data.x, key.data.y, key.data.z));
            lc.LOG(C_fstring("Time: %i, T: %.2f, C: %.2f, B: %.2f, Easy from: %.2f, to: %.2f", 
               key.header.time, key.add_data.tension, key.add_data.continuity,
               key.add_data.bias, key.add_data.easy_from, key.add_data.easy_to));
         }
      }
   }
   if(lc.recent_frame){
      if(is_uni_scale){
         lc.recent_frame->SetScale(base_scale.x);
      }else{
         float avg_scale = base_scale.Sum() / 3.0f;
         lc.recent_frame->SetScale(avg_scale);
         lc.recent_nu_scale = base_scale / avg_scale;
      }
   }

   if(num_keys>1){
      if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
      lc.recent_anim->SetEndTime(Max((dword)key.header.time*1000/ANIM_FPS, lc.recent_anim->GetEndTime()));

      if(!is_uni_scale){
                              //set scale keys
         lc.recent_anim->SetScaleKeys(skeys, num_keys);
      }else{
                              //set power keys
         I3D_anim_power *pkeys = (I3D_anim_power*)alloca(sizeof(I3D_anim_power)*num_keys);
         for(i=0; i<num_keys; i++){
            pkeys[i].time = skeys[i].time;
            pkeys[i].power = skeys[i].v.x;
         }
         lc.recent_anim->SetPowerKeys(pkeys, num_keys);
      }
#ifndef USE_ALLOCA
      delete[] skeys;
#endif
   }
   return true;
}

//----------------------------
                              //B023
static bool FOVTrackReader(S_load_context &lc){
//   UnknownReader(is, to_pos, tab);
   return true;
}

//----------------------------
                              //B024
static bool RollTrackReader(S_load_context &lc){
//   UnknownReader(is, to_pos, tab);
   return true;
}

//----------------------------

struct S_color_key{
   S_key_header header;
   S_key_tcb add_data;
   S_vector color;

   void Read(C_cache *is){
      is->read((char*)&header, sizeof(header));
      add_data.Read(is, header.flags);
      is->read((char*)&color, sizeof(color));
   }
};

                              //B025
static bool ColorTrackReader(S_load_context &lc){

   S_track_header header;
   lc.is->read((char*)&header, sizeof(header));

   dword num_keys = Max(1ul, (dword)header.num_keys);

   I3D_anim_color *keys;
   if(num_keys>1){
#ifndef USE_ALLOCA
      keys = new I3D_anim_color[num_keys];
#else
      keys = (I3D_anim_color*)alloca(sizeof(I3D_anim_color)*num_keys);
#endif

      S_color_key key;

      for(int i=0; i<num_keys; i++){
         key.Read(lc.is);
         keys[i].time=(key.header.time)*1000/ANIM_FPS;
         keys[i].color = key.color;
      }
      if(!i)
      {
         if(lc.load_flags&I3DLOAD_LOG){
            lc.LOG(C_fstring("Number of keys: %i", header.num_keys));
            lc.LOG(C_fstring("R: %.2f, G: %.2f, B: %.2f", keys[num_keys-1].color.x, keys[num_keys-1].color.y, keys[num_keys-1].color.z));
         }
      }
      if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
      lc.recent_anim->SetEndTime(Max((dword)key.header.time*1000/ANIM_FPS, lc.recent_anim->GetEndTime()));
      lc.recent_anim->SetColorKeys(keys, num_keys);
#ifndef USE_ALLOCA
      delete[] keys;
#endif
   }
   return true;
}

//----------------------------
                              //B027
static bool HotspotTrackReader(S_load_context &lc){
//   UnknownReader(is, to_pos, tab);
   return true;
}

//----------------------------
                              //B028
static bool FalloffTrackReader(S_load_context &lc){
//   UnknownReader(is, to_pos, tab);
   return true;
}

//----------------------------
                              //B029
static bool HideTrackReader(S_load_context &lc){

   int num_keys;
   lc.is->read((char*)&num_keys, sizeof(num_keys));

   if(num_keys){
      int *keys = NULL;
      I3D_anim_visibility *vkeys;
#ifndef USE_ALLOCA
      keys = new int[num_keys];
      vkeys = new I3D_anim_visibility[num_keys + 1];
#else
      keys = (int*)alloca(sizeof(int)*num_keys);
      vkeys = (I3D_anim_visibility*)alloca(sizeof(I3D_anim_visibility)*(num_keys + 1));
#endif
      lc.is->read((char*)keys, sizeof(int)*num_keys);

      int j = 0;
      bool on = true;
      if(keys[0]!=0){
                              //put 'on' key on position 0
         vkeys[0].time = 0;
         vkeys[0].on = true;
         ++j;
      }else{
         lc.recent_frame->SetOn(false);
      }
      if(j || num_keys>1){
         for(int i=0; i<num_keys; i++, j++){
            on = !on;
            vkeys[j].time = keys[i] * 1000/ANIM_FPS;
            vkeys[j].on = on;
         }
         if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
         lc.recent_anim->SetEndTime(Max((dword)keys[num_keys-1]*1000/ANIM_FPS, lc.recent_anim->GetEndTime()));
         lc.recent_anim->SetVisibilityKeys(vkeys, j);
      }
#ifndef USE_ALLOCA
      delete[] vkeys;
      delete[] keys;
#endif
   }

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("Number of keys: %i", num_keys));
   }
   return true;
}

//----------------------------
                              //B02A
static bool NoteTrackReader(S_load_context &lc){

                              //read # of note tracks
   int num_tracks;
   lc.is->read((char*)&num_tracks, sizeof(num_tracks));

   if(!num_tracks) return true;

   int num_keys;
   lc.is->read((char*)&num_keys, sizeof(num_keys));

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("keys: %i", num_keys));
   }
   if(!num_keys){
      lc.REPORT_ERR(C_fstring("Frame: '%s': empty note track", (const char*)lc.recent_instance_name));
      return true;
   }

   vector<C_string> str_list(num_keys, C_string());
   vector<I3D_anim_note> note_list(num_keys, I3D_anim_note());

   int i;
   for(i=0; i<num_keys; i++){
                              //read time
      int time;
      lc.is->read((char*)&time, sizeof(time));
      time *= 1000/ANIM_FPS;
                              //read note
      lc.ReadString(str_list[i]);
      dtaBin2Text(&str_list[i][0], str_list[i].Size());

      I3D_anim_note &note = note_list[i];
      note.time = time;
      note.note = str_list[i];

      if(lc.load_flags&I3DLOAD_LOG){
         lc.LOG(C_fstring("note, time %i \"%s\"", i, time, str_list[i]));
      }
   }
   if(!lc.recent_anim) lc.recent_anim = (PI3D_keyframe_anim)lc.driver->CreateAnimation(I3DANIM_KEYFRAME);
   lc.recent_anim->SetEndTime(Max((dword)note_list[num_keys-1].time, lc.recent_anim->GetEndTime()));
   lc.recent_anim->SetNoteKeys(note_list.begin(), note_list.size());

   return true;
}

//----------------------------
                              //B030
static bool NodeIDReader(S_load_context &lc){

   lc.is->read((char*)&lc.recent_node_id, sizeof(lc.recent_node_id));

   if(lc.load_flags&I3DLOAD_LOG){
      lc.LOG(C_fstring("%i", lc.recent_node_id));
   }
   return true;
}

//----------------------------
                              //* = I3d export chunks (not 3DS compatible)



enum{
   CHUNK_NULL      = 0x0000,
   CHUNK_VERSION   = 0x0002,
   CHUNK_RGBF      = 0x0010,
   CHUNK_RGBB      = 0x0011,
   LIN_COLOR_24    = 0x0012,
   LIN_COLOR_F     = 0x0013,
   INT_PERCENTAGE  = 0x0030,
   //FLOAT_PERCENTAGE= 0x0031,
   //MASTER_SCALE    = 0x0100,

   CHUNK_MAIN      = 0x4D4D,
      CHUNK_OBJMESH   = 0x3D3D,
         //CHUNK_MESHVERSION  = 0x3D3E,
         //BIT_MAP         = 0x1100,
         //CHUNK_BKGCOLOR  = 0x1200,
//         V_GRADIENT             = 0x1300,
//         USE_V_GRADIENT         = 0x1301,

         //CHUNK_AMBCOLOR  = 0x2100,
//      LO_SHADOW_BIAS         = 0x1400,
//HI_SHADOW_BIAS         = 0x1410,
//SHADOW_MAP_SIZE        = 0x1420,
//SHADOW_SAMPLES         = 0x1430,
//SHADOW_RANGE           = 0x1440,
//SHADOW_FILTER          = 0x1450,
//RAY_BIAS               = 0x1460,
//   O_CONSTS               = 0x1500,



//         USE_SOLID_BGND  = 0x2101,
         //CHUNK_FOG    = 0x2200,
         //CHUNK_USEFOG = 0x2201,
         //FOGBGND      = 0x2210,
         //DISTANCE_CUE           = 0x2300,
//   USE_DISTANCE_CUE       = 0x2301,
         LAYER_FOG              = 0x2302,
//   USE_LAYER_FOG          = 0x2303,
//   DCUE_BGND              = 0x2310,
         CHUNK_OBJBLOCK  = 0x4000,
         NURBS           = 0x4101,                       //*
//   OBJ_HIDDEN             = 0x4010,
//   OBJ_FROZEN             = 0x4016,
            CHUNK_TRIMESH   = 0x4100,
               CHUNK_VERTLIST  = 0x4110,
               CHUNK_VERTSLCT  = 0x4111,
               CHUNK_FACELIST  = 0x4120,
                  CHUNK_FACEMAT   = 0x4130,
                  CHUNK_MAPCOORS  = 0x4140,
                  CHUNK_SMOOLIST  = 0x4150,
               CHUNK_TRMATRIX  = 0x4160,
               MESH_COLOR      = 0x4165,
               CHUNK_MAP_CHANNEL = 0x4200,

            CHUNK_LIGHT     = 0x4600,
               CHUNK_SPOTLIGHT    = 0x4610,
               CHUNK_LIGHT_OFF    = 0x4620,
               CHUNK_SPOTRANGE    = 0x4625,
//   DL_SPOT_OVERSHOOT      = 0x4652,
               CHUNK_LIGHT_START  = 0x4659,
               CHUNK_LIGHT_END    = 0x465A,
               CHUNK_LIGHT_MULT   = 0x465B,
            CHUNK_CAMERA = 0x4700,
//               CAM_SEECONE  = 0x4720,
               CAM_RANGES   = 0x4720,
//   HIERARCHY              = 0x4F00,
//   PARENT_OBJECT          = 0x4F10,
//   PIVOT_OBJECT           = 0x4F20,
//   PIVOT_LIMITS           = 0x4F30,
//   PIVOT_ORDER            = 0x4F40,
//   XLATE_RANGE            = 0x4F50,
      VIEWPORT_LAYOUT = 0x7001,

      CHUNK_NURBS     = 0x4904,
      CHUNK_NURBS_UV_ORDER = 0x4905,
      CHUNK_NURBS_CVS = 0x4906,
      CHUNK_NURBS_KNOTS = 0x4907,
      CHUNK_NURBS_TRANSFORM_MATRIX = 0x4908,
      CHUNK_NURBS_MAT = 0x4909,
      CHUNK_NURBS_FLIP_NORMALS = 0x490a,
      CHUNK_NURBS_UV_MAP = 0x490b,

      CHUNK_MATERIAL  = 0xAFFF,
         CHUNK_MATNAME   = 0xA000,
         CHUNK_AMBIENT   = 0xA010,
         CHUNK_DIFFUSE   = 0xA020,
         CHUNK_SPECULAR  = 0xA030,
         CHUNK_SHININESS = 0xA040,
         CHUNK_SHINSTRENGTH = 0xA041,
//   MAT_SHIN3PCT         = 0xA042,
         CHUNK_TRANSPARENCY = 0xA050,
         MAT_XPFALL   = 0xA052,
         MAT_REFBLUR  = 0xA053,
//   MAT_SELF_ILLUM         = 0xA080,
         CHUNK_MAT2SIDE  = 0xA081,
//   MAT_DECAL              = 0xA082,
//   MAT_ADDITIVE           = 0xA083,
         CHUNK_SELFILUM  = 0xA084,
         CHUNK_MAT_WIRE        = 0xA085,
//   MAT_SUPERSMP      = 0xA086,
         CHUNK_WIRESIZE  = 0xA087,
//   MAT_FACEMAP       = 0xA088,
         MAT_XPFALLIN    = 0xA08A,
         MAT_PHONGSOFT   = 0xA08c,
         CHUNK_MATSHADE  = 0xA100,
         CHUNK_TEXTURE   = 0xA200,
         MAT_SPECMAP     = 0xA204,
         MAT_OPACMAP     = 0xA210,
         MAT_REFLMAP     = 0xA220,
         CHUNK_BUMPMAP   = 0xA230,
//   MAT_USE_XPFALL         = 0xA240,
//   MAT_USE_REFBLUR        = 0xA250,
//         MAT_BUMP_PERCENT       = 0xA252,
            CHUNK_MAPFILE   = 0xA300,
//   MAT_ACUBIC             = 0xA310,
//   MAT_TEX2MAP       = 0xA33A,
//         MAT_SHINMAP       = 0xA33C,
//   MAT_SELFIMAP      = 0xA33D,
//   MAT_TEXMASK       = 0xA33E,
//   MAT_TEX2MASK      = 0xA340,
//   MAT_OPACMASK      = 0xA342,
//   MAT_BUMPMASK      = 0xA344,
//   MAT_SHINMASK      = 0xA346,
//   MAT_SPECMASK      = 0xA348,
//   MAT_SELFIMASK     = 0xA34A,
//   MAT_REFLMASK      = 0xA34C,
      MAT_MAP_PARAMS  = 0xA351,
      MAT_MAP_TEXBLUR = 0xA353,
//   MAT_MAP_USCALE    = 0xA354,
//   MAT_MAP_VSCALE    = 0xA356,
//   MAT_MAP_UOFFSET      = 0xA358,
//   MAT_MAP_VOFFSET      = 0xA35A,
//   MAT_MAP_ANG       = 0xA35C,
      CHUNK_KEYFRAMER = 0xB000,
         OBJECT_NODE_TAG = 0xB002,
         CAMERA_NODE_TAG = 0xB003,
         TARGET_NODE_TAG = 0xB004,
         LIGHT_NODE_TAG  = 0xB005,
         L_TARGET_NODE_TAG  = 0xB006,
         SPOTLIGHT_NODE_TAG = 0xB007,
         CHUNK_FRAMES    = 0xB008,
         KFCURTIME       = 0xB009,
         KFHDR           = 0xB00A,
         CHUNK_OBJ_INFO  = 0xB00B,                       // *
         CHUNK_BONE_NODE = 0xB00E,                       // *
         NODE_HDR        = 0xB010,
         INSTANCE_NAME   = 0xB011,
         PIVOT           = 0xB013,
         BOUNDBOX        = 0xB014,
         POS_TRACK_TAG   = 0xB020,
         ROT_TRACK_TAG   = 0xB021,
         SCL_TRACK_TAG   = 0xB022,
         FOV_TRACK_TAG   = 0xB023,
         ROLL_TRACK_TAG  = 0xB024,
         COL_TRACK_TAG   = 0xB025,
         HOT_TRACK_TAG   = 0xB027,
         FALL_TRACK_TAG  = 0xB028,
         HIDE_TRACK_TAG  = 0xB029,
         NOTE_TRACK_TAG  = 0xB02A,                       //*
         NODE_ID         = 0xB030,
//   AMBIENT_NODE_TAG       = 0xB001,

//   PRESCALE               = 0xB012,
//   MORPH_SMOOTH           = 0xB015,
//   MORPH_TRACK_TAG        = 0xB026,

};

//----------------------------
                              //supported chunk table
                              //strictly sorted by id for fast look-up
static const struct S_chk_name{
   dword id;
   const char* name;
   T_chunk_read* func;
}chunk_names[]={
   {CHUNK_VERSION,      "0002 - Version",          VersionReader},
   {CHUNK_RGBF,         "0010 - vector",           VectorReader},
   {CHUNK_RGBB,         "0011 - RGB byte",         RGBBReader},
   {LIN_COLOR_24,       "0012 - Lin. 24-bit color",Lin24Reader},
   {LIN_COLOR_F,        "0013 - Lin. float color", LinFloatReader},
   {INT_PERCENTAGE,     "0030 - Percentage",       PercentageReader},
   //{MASTER_SCALE,       "0100 - Master scale",     MasterScaleReader},
   //{BIT_MAP,            "1100 - Bit-map",          EnvBitmapReader},
   //{CHUNK_BKGCOLOR,     "1200 - Background color", EnvBgndColorReader},
   //{CHUNK_AMBCOLOR,     "2100 - Ambient color",    EnvAmbientReader},
   //{CHUNK_FOG,          "2200 - Fog",              EnvFogReader},
   //{CHUNK_USEFOG,       "2201 - Use fog",          EnvUseFogReader},
   //{FOGBGND,            "2210 - Fog background",   FogBgndReader},

   {CHUNK_OBJMESH,      "3D3D - Object mesh",      NULL},
   //{CHUNK_MESHVERSION,  "3D3E - Mesh version",     MeshVersionReader},
   {CHUNK_OBJBLOCK,     "4000 - Object block",     ObjBlockReader},
   {CHUNK_TRIMESH,      "4100 - Tri-Mesh",         NewMeshReader},
   {NURBS,              "4101 - *NURBs",           NURBsReader}, //*
   {CHUNK_VERTLIST,     "4110 - Vertex list",      VertListReader},
//   {CHUNK_VERTSLCT,    "4111 - Vertex selection", VertSlctReader},
   {CHUNK_FACELIST,     "4120 - Face list",        FaceListReader},
   {CHUNK_FACEMAT,      "4130 - Face material",    FaceMatReader},
   {CHUNK_MAPCOORS,     "4140 - Mappings coors",   MapCoorsReader},
   {CHUNK_SMOOLIST,     "4150 - Smoothings",       SmooListReader},
   {CHUNK_TRMATRIX,     "4160 - Matrix",           TrMatrixReader},
   {MESH_COLOR,         "4165 - Mesh color",       MeshColorReader},

   {CHUNK_MAP_CHANNEL,  "4200 - Map channel",      MapChannelReader},

   {CHUNK_LIGHT,        "4600 - Light",            LightReader},
   {CHUNK_SPOTLIGHT,    "4610 - Spotlight",        SpotLightReader},
   {CHUNK_LIGHT_OFF,    "4620 - Light off",        LightOffReader},
   {CHUNK_SPOTRANGE,    "4625 - Spot range",       LightRangeReader},
   {CHUNK_LIGHT_START,  "4659 - Light range start",LightStartReader},
   {CHUNK_LIGHT_END,    "465A - Light range end",  LightEndReader},
   {CHUNK_LIGHT_MULT,   "465B - Light multiplier", LightMultReader},
   {CHUNK_CAMERA,       "4700 - Camera",           CameraReader},
   {CAM_RANGES,         "4720 - Camera ranges",    CameraRangeReader},

   {CHUNK_NURBS,        "4904 - NURBs",            NURBsReader},
   {CHUNK_NURBS_UV_ORDER,  "4905 - NURBs UV order",NURBsUVOrderReader},
   {CHUNK_NURBS_CVS,    "4906 - NURBs CVs",        NURBsCVsReader},
   {CHUNK_NURBS_KNOTS,  "4907 - NURBs Knots",      NURBsKnotsReader},
   {CHUNK_NURBS_TRANSFORM_MATRIX,"4908 - NURBs trans. matrix", NURBsTMReader},
   {CHUNK_NURBS_MAT,    "4909 - NURBs material",   NURBsMaterialReader},
   {CHUNK_NURBS_FLIP_NORMALS, "490A - NURBs flip normals",  NURBsFlipNormReader},
   {CHUNK_NURBS_UV_MAP, "490B - NURBs UV map",     NURBsUVMapReader},

   {CHUNK_MAIN,         "4D4D - Main chunk",       NULL},

   //{VIEWPORT_LAYOUT,   "7001 - Viewport layout",  ViewportLayoutReader},

   {CHUNK_MATNAME,      "A000 - Material name",    MatNameReader},
   {CHUNK_AMBIENT,      "A010 - Ambient color",    AmbientReader},
   {CHUNK_DIFFUSE,      "A020 - Diffuse color",    DiffuseReader},
   {CHUNK_SPECULAR,     "A030 - Specular color",   SpecularReader},
   {CHUNK_SHININESS,    "A040 - Shininess",        ShininessReader},
   {CHUNK_SHINSTRENGTH, "A041 - Shin strength",    ShinStrReader},
   {CHUNK_TRANSPARENCY, "A050 - Transparency",     TransparencyReader},
   {MAT_XPFALL,         "A052 - Xp fall",          XPFallReader},
   {MAT_REFBLUR,        "A053 - Reflection blur",  RefBlurReader},
   {CHUNK_MAT2SIDE,     "A081 - Material 2-side",  Mat2SideReader},
   {CHUNK_SELFILUM,     "A084 - Self illumin.",    SelfIllumReader},
   {CHUNK_MAT_WIRE,     "A085 - Wire material",    WireMatReader},
   {CHUNK_WIRESIZE,     "A087 - Wire size",        WireSizeReader},
   {MAT_XPFALLIN,       "A08A - Xp fall in",       XPFallInReader},
   {MAT_PHONGSOFT,      "A08C - Phong soft",       PhongSoftReader},
   {CHUNK_MATSHADE,     "A100 - Material shading", MatTypeReader},
   {CHUNK_TEXTURE,      "A200 - Texture map",      TextMapReader},
   {MAT_SPECMAP,        "A204 - Specular map",     SpecularMapReader},
   {MAT_OPACMAP,        "A210 - Opacity map",      OptMapReader},
   {MAT_REFLMAP,        "A220 - Reflection map",   ReflMapReader},
   {CHUNK_BUMPMAP,      "A230 - Bump map",         BumpMapReader},
   //{MAT_BUMP_PERCENT,   "A252 - Bump percentage",  PercentageReader},
   {CHUNK_MAPFILE,      "A300 - Map filename",     MapFileReader},
   //{MAT_SHINMAP,        "A33C - Specular texture", SpecularMapReader},
   {MAT_MAP_PARAMS,     "A351 - Map parameters",   MapParams},
   {MAT_MAP_TEXBLUR,    "A353 - Map texture blur", TextBlurReader},
   {CHUNK_MATERIAL,     "AFFF - Material",         NewMatReader},

   {CHUNK_KEYFRAMER,    "B000 - Keyframer data",   NULL},
   {OBJECT_NODE_TAG,    "B002 - Object node",      ObjNodeReader},
   {CAMERA_NODE_TAG,    "B003 - Camera node",      CamNodeReader},
   {TARGET_NODE_TAG,    "B004 - Camera target",    TargetNodeReader},
   {LIGHT_NODE_TAG,     "B005 - Light node",       LightNodeReader},
   {L_TARGET_NODE_TAG,  "B006 - Light target node",TargetNodeReader},
   {SPOTLIGHT_NODE_TAG, "B007 - Spotlight node",   LightNodeReader},
   {CHUNK_FRAMES,       "B008 - Frames",           FramesReader},
   {KFCURTIME,          "B009 - Current time",     KFCurrTimeReader},
   {KFHDR,              "B00A - KF header",        KFHeaderReader},
   {CHUNK_OBJ_INFO,     "B00B - *Object info",     ObjInfoReader}, //*
   {CHUNK_BONE_NODE,    "B00E - *Bone node",       BoneNodeReader}, //*
   {NODE_HDR,           "B010 - Node header",      NodeHeaderReader},
   {INSTANCE_NAME,      "B011 - Instance name",    InstNameReader},
   {PIVOT,              "B013 - Pivot",            NodePivotReader},
   {BOUNDBOX,           "B014 - Bounding box",     BoundBoxReader},
   {POS_TRACK_TAG,      "B020 - Position track",   PosTrackReader},
   {ROT_TRACK_TAG,      "B021 - Rotation track",   RotTrackReader},
   {SCL_TRACK_TAG,      "B022 - Scale track",      ScaleTrackReader},
   {FOV_TRACK_TAG,      "B023 - FOV track",        FOVTrackReader},
   {ROLL_TRACK_TAG,     "B024 - Roll track",       RollTrackReader},
   {COL_TRACK_TAG,      "B025 - Color track",      ColorTrackReader},
   {HOT_TRACK_TAG,      "B027 - Hotspot track",    HotspotTrackReader},
   {FALL_TRACK_TAG,     "B028 - Falloff track",    FalloffTrackReader},
   {HIDE_TRACK_TAG,     "B029 - Hide track",       HideTrackReader},
   {NOTE_TRACK_TAG,     "B02A - *Note track",       NoteTrackReader},
   {NODE_ID,            "B030 - Node ID",          NodeIDReader},
   //{CHUNK_NULL}
};

//----------------------------

inline const S_chk_name *FindChunk(dword id){

   dword min = 0, rng = sizeof(chunk_names)/sizeof(S_chk_name);
   do{
      dword mid = min + rng / 2;
      if(chunk_names[mid].id==id) return &chunk_names[mid];
      else
      if(chunk_names[mid].id<id) min = mid + (rng&1);
   }while(rng /= 2);
   return NULL;
}

//----------------------------

bool S_load_context::ChunkReader(){
                              //packet header
#pragma pack(push, 1)
   struct{
      word id;
      dword len;
   } h;
#pragma pack(pop)

   long pc;

   dword curr_to_pos = to_pos;
   while((pc=is->tellg()) < curr_to_pos){
                              //read chunk header
      size_t hdr_len = is->read((char*)&h, sizeof(h));
      if(is->bad() || hdr_len<sizeof(h))
         return false;
      pc += h.len;
      if(h.len<sizeof(h) || pc > file_length)
         return false;

      const S_chk_name *ck = FindChunk(h.id);
      if(!ck){                //unknown chunk, skip
         if(load_flags&I3DLOAD_LOG){
            LOG(C_fstring("skipped: 0x%.4x, size: %i", h.id, h.len-sizeof(h)));
         }
      }else{
         if(load_flags&I3DLOAD_LOG){
            LOG(C_fstring("%s, size: %i", ck->name, h.len-sizeof(h)));
         }
         to_pos = pc;
         ++tab;
         bool st;
         if(ck->func){        //read this chunk
            st = ck->func(*this);
         }else{               //read sub-chunk
            st = ChunkReader();
         }
         --tab;
         if(!st)
            return false;
      }
      is->seekg(pc);
      if(is->bad())
         return false;
   }
   return true;
}

//----------------------------

struct S_al_help{
   PI3D_frame root_frame;
   PI3D_driver drv;
   S_load_context *lc;
   vector<PI3D_frame> frames_to_delete;
};

//----------------------------

static I3DENUMRET I3DAPI cbAfterLoad(PI3D_frame frm, dword context){

   switch(frm->GetType1()){
   case FRAME_VISUAL:
      {
         S_al_help *alh = (S_al_help*)context;

         PI3D_visual vis = I3DCAST_VISUAL(frm);
         switch(vis->GetVisualType1()){
         case I3D_VISUAL_OBJECT:
         case I3D_VISUAL_BILLBOARD:
            {
               //I3DCAST_OBJECT(frm)->GetMesh1()->ComputeBoundingVolume();
               return I3DENUMRET_SKIPCHILDREN;
            }
            break;
         case I3D_VISUAL_SINGLEMESH:
            {
               PI3D_object_singlemesh sm = (PI3D_object_singlemesh)vis;
               PI3D_singlemesh smesh = sm->GetSingleMesh();

#ifdef USE_DATABASE
               C_fstring cache_name("%s+%s+\"SM\"", (const char*)alh->lc->file_name, vis->GetName());
               C_cache *ck = alh->lc->driver->dbase.OpenRecord(cache_name, alh->lc->file_time);
               bool ok = false;
               if(ck){
                              //try to load SM info
                  ok = smesh->LoadCachedInfo(ck, *alh->lc);
                  alh->lc->driver->dbase.CloseRecord(ck);
               }
               if(!ok){
                  I3D_singlemesh::E_IBS_RESULT ibsr = smesh->InitBoneSegments(sm);
                  if(ibsr!=I3D_singlemesh::IBS_OK){
                     const char *err;
                     switch(ibsr){
                     case I3D_singlemesh::IBS_NO_VERTICES: err = "no vertices"; break;
                     case I3D_singlemesh::IBS_NO_FACES: err = "no faces"; break;
                     case I3D_singlemesh::IBS_NO_JOINTS: err = "no joints children"; break;
                     default: assert(0);
                     }
                     alh->lc->REPORT_ERR(C_fstring("invalid single-mesh: '%s' (%s)", sm->GetName(), err));
                  }
                  if(ibsr==I3D_singlemesh::IBS_OK){
                                //process modifiers
                     C_string str_prop;
                     E_PROP_KEYWORD kw;

                     map<PI3D_frame, C_string>::const_iterator it = alh->lc->frm_properties.find(frm);
                     if(it!=alh->lc->frm_properties.end())
                     if(GetPropetriesBegin((*it).second, str_prop, kw))
                     do{
                        switch(kw){
                        case PROP_AUTO_LOD:
                           {
                                                //read additional properties
                              float min_dist = 5.0f;
                              float max_dist = 50.0f;
                              int num_parts = 10;
                              float lowest_quality = .1f;

                              if(GetPropertiesFloat(str_prop, min_dist)){
                                 if(GetPropertiesFloat(str_prop, max_dist)){
                                    if(GetPropertiesInt(str_prop, num_parts)){
                                       if(!GetPropertiesFloat(str_prop, lowest_quality))
                                          alh->lc->REPORT_ERR(C_fstring("Scanning properties on '%s': auto_lod - lowest quality expected", (const char*)frm->GetName()));
                                       {};
                                    }else
                                       alh->lc->REPORT_ERR(C_fstring("Scanning properties on '%s': auto_lod - number of parts expected", (const char*)frm->GetName()));
                                 }else
                                    alh->lc->REPORT_ERR(C_fstring("Scanning properties on '%s': auto_lod - maximal distance expected", (const char*)frm->GetName()));
                              }else
                                 alh->lc->REPORT_ERR(C_fstring("Scanning properties on '%s': auto_lod - minimal distance expected", (const char*)frm->GetName()));
                              smesh->CreateLODs(lowest_quality, lowest_quality, (1.0f-lowest_quality) / num_parts, min_dist, max_dist);
                           }
                           break;
                        }
                     }while(GetPropertiesNext(str_prop, kw));
                  }
                  if((ibsr==I3D_singlemesh::IBS_OK)){
                     C_cache *ck = alh->lc->driver->dbase.CreateRecord(cache_name, alh->lc->file_time);
                     if(ck){
                        ok = smesh->SaveCachedInfo(ck);
                        alh->lc->driver->dbase.CloseRecord(ck);
                     }
                  }
               }
#else                         //USE_DATABASE
               I3D_singlemesh::E_IBS_RESULT ibsr = smesh->InitBoneSegments(sm);
               if(ibsr==I3D_singlemesh::IBS_OK)
                  smesh->CreateLODs(LOD_MIN_RATIO, LOD_MIN_AREA_RATIO, LOD_STEP_CHANGE, 5, 50);
#endif                        //!USE_DATABASE
               sm->SetMesh(smesh);
            }
            break;
         }
      }
      break;

   case FRAME_LIGHT:
      {
         S_al_help *alh = (S_al_help*)context;
         PI3D_light lp = I3DCAST_LIGHT(frm);
         PI3D_frame frm1a = frm->GetParent();
         while(frm1a){
            if(frm1a->GetType1()==FRAME_SECTOR) break;
            else frm1a = frm1a->GetParent();
         }
         (frm1a ? (PI3D_sector)frm1a : alh->lc->scene->GetPrimarySector1())->AddLight(lp);
      }
      return I3DENUMRET_SKIPCHILDREN;
   }
   return I3DENUMRET_OK;
}

//----------------------------

inline bool VectorsApproxEqual(const S_vector &v0, const S_vector &v1){

   S_vector dir = v0 - v1;
   float dist_2 = dir.Dot(dir);
   return (dist_2 < 1e-2f);
}

//----------------------------

struct S_smooth_vertex_info{
   S_vector v_world;
   vector<PI3D_vertex_mesh> v_srcs;
};

//----------------------------

static void SmoothSceneNormals(PI3D_scene scene, const S_load_context &lc){

   vector<PI3D_visual> vis_list;           //list of all visuals being shaded

   struct S_hlp{
      vector<PI3D_visual> *vis_list;
      const S_load_context *lc;

      static I3DENUMRET I3DAPI cbShade(PI3D_frame frm, dword c){

         if(I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_OBJECT){

            S_hlp *hp = (S_hlp*)c;
                              //check if shading requested
            C_string str_prop;
            E_PROP_KEYWORD kw;

            map<PI3D_frame, C_string>::const_iterator it = hp->lc->frm_properties.find(frm);
            if(it!=hp->lc->frm_properties.end())
            if(GetPropetriesBegin((*it).second, str_prop, kw))
            do{
               switch(kw){
               case PROP_SMOOTH:
                  hp->vis_list->push_back(I3DCAST_OBJECT(frm));
                  break;
               }
            }while(GetPropertiesNext(str_prop, kw));
         }
         return I3DENUMRET_OK;
      }
   } hlp;
   hlp.vis_list = &vis_list;
   hlp.lc = &lc;

                              //collect all visuals to be shaded
   scene->EnumFrames(S_hlp::cbShade, (dword)&hlp, ENUMF_VISUAL);

   if(vis_list.size()<2)
      return;

   vector<PI3D_vertex_mesh> vp_src;    //source vertex arrays
   vector<const I3D_vertex_mesh*> vp;        //shadow vertex arrays
   vector<I3D_bbox> world_bboxes;      //bound-boxes
   vector<vector<word> >used_verts;    //indicies of used verts of meshes
   vector<vector<S_vector> >world_verts;  //world-transformed used verts of meshes

   int i, j;
   vp_src.reserve(vis_list.size());
   vp.reserve(vis_list.size());
   world_verts.reserve(vis_list.size());
   world_bboxes.reserve(vis_list.size());
   used_verts.reserve(vis_list.size());

                              //lock all meshes, transform vertices,
                              // create AABB
   for(i=0; i<vis_list.size(); i++){
      PI3D_mesh_object mo = I3DCAST_OBJECT(vis_list[i])->GetMesh1();
      PI3D_mesh_level ml = mo->GetLOD1();

      const I3D_vertex_mesh *vptr = ml->LockVertices2();
      assert(vptr);
      vp.push_back(vptr);

      PI3D_vertex_mesh verts = (PI3D_vertex_mesh)vptr;
      vp_src.push_back(verts);

      used_verts.push_back(vector<word>());
      vector<word> &used_verts1 = used_verts.back();
      {                       //collect border vertices
         int num_faces = ml->NumFaces();
         vector<I3D_edge> edges; edges.reserve(num_faces);
         vector<I3D_edge> edges_mapped; edges_mapped.reserve(num_faces);

         const vector<word> &v_map = ml->vertex_buffer.GetVertexSpreadMap();
         int num_orig_verts = ml->vertex_buffer.NumVertices() - v_map.size();

         const I3D_triface *fcs = ml->LockFaces();
         for(int fi=ml->NumFaces(); fi--; ){
            const I3D_triface &fc = fcs[fi];

            word v0 = fc.vi[0];
            for(j=3; j--; ){
               word v1 = v0;
               v0 = fc.vi[j];

               I3D_edge e_mapped(v0, v1);
               if(v0 >= num_orig_verts) e_mapped[0] = v_map[v0 - num_orig_verts];
               if(v1 >= num_orig_verts) e_mapped[1] = v_map[v1 - num_orig_verts];

                           //add or split edge
               int ei = FindEdge(edges_mapped.begin(), edges_mapped.size(), I3D_edge(e_mapped.v[1], e_mapped.v[0]));
               if(ei!=-1){
                  edges_mapped[ei] = edges_mapped.back(); edges_mapped.pop_back();
                  edges[ei] = edges.back(); edges.pop_back();
               }else{
                  edges_mapped.push_back(e_mapped);
                  edges.push_back(I3D_edge(v0, v1));
               }
            }
         }
         ml->UnlockFaces();

         used_verts1.assign(edges.size());
         for(j=edges.size(); j--; ){
            used_verts1[j] = edges[j].v[0];
         }
      }

                              //transform verts
      world_verts.push_back(vector<S_vector>());
      vector<S_vector> &tv1 = world_verts.back();
      const S_matrix &m = vis_list[i]->GetMatrix();

      world_bboxes.push_back(I3D_bbox());
      I3D_bbox &bbox = world_bboxes.back();
      bbox.Invalidate();

      int numv = used_verts1.size();
      tv1.assign(numv);
      for(j=0; j<numv; j++){
         S_vector v = verts[used_verts1[j]].xyz * m;
         tv1[j] = v;
         bbox.min.Minimal(v);
         bbox.max.Maximal(v);
      }
   }

   vector<S_smooth_vertex_info> smooth_info;

                              //generate info for shared vertices
   for(i = vis_list.size(); i--; ){
      S_vector *wv_i = world_verts[i].begin();
      const I3D_bbox &bb_i = world_bboxes[i];

      for(j=i; j--; ){
                              //check if those boxes overlap
         const I3D_bbox &bb_j = world_bboxes[j];
         if(bb_i.min.x > bb_j.max.x || bb_i.max.x < bb_j.min.x ||
            bb_i.min.y > bb_j.max.y || bb_i.max.y < bb_j.min.y ||
            bb_i.min.z > bb_j.max.z || bb_i.max.z < bb_j.min.z){
            continue;
         }

         S_vector *wv_j = world_verts[j].begin();
         int verts_i_count = world_verts[i].size();

                              //check all vertices
         int num_verts_j = world_verts[j].size();
         while(verts_i_count--){
            for(int verts_j_count = num_verts_j; verts_j_count--; ){
               if(VectorsApproxEqual(wv_i[verts_i_count], wv_j[verts_j_count])){
                              //create new share info
                  for(int k=smooth_info.size(); k--; )
                     if(VectorsApproxEqual(smooth_info[k].v_world, wv_i[verts_i_count]))
                        break;
                  if(k==-1){
                     k = smooth_info.size();
                     smooth_info.push_back(S_smooth_vertex_info());
                     smooth_info[k].v_world = wv_i[verts_i_count];
                  }
                  vector<PI3D_vertex_mesh> &v_ptr = smooth_info[k].v_srcs;
                  PI3D_vertex_mesh vp;

                  vp = &vp_src[i][used_verts[i][verts_i_count]];
                  if(find(v_ptr.begin(), v_ptr.end(), vp)==v_ptr.end())
                     v_ptr.push_back(vp);

                  vp = &vp_src[j][used_verts[j][verts_j_count]];
                  if(find(v_ptr.begin(), v_ptr.end(), vp)==v_ptr.end())
                     v_ptr.push_back(vp);
               }
            }
         }
      }
   }

                              //adjust all vertices
   for(j=smooth_info.size(); j--; ){
      const vector<PI3D_vertex_mesh> &v_ptrs = smooth_info[j].v_srcs;
                              //add all normals together
      S_vector n; n.Zero();
      for(i=v_ptrs.size(); i--; )
         n += v_ptrs[i]->normal;
      n.Normalize();
                              //spread resulting normals to all participating verts
      for(i=v_ptrs.size(); i--; ){
         v_ptrs[i]->normal = n;
#ifdef PAINT_SMOOTH_NORMALS   //note: uses local coords!
         {           //debug - draw vertex and normal
            S_vector v[2] = {v_ptrs[i]->xyz, v_ptrs[i]->xyz + n};
            scene->GetDriver1()->DEBUG_CODE(0, (dword)v);
         }
#endif
      }
   }

                              //unlock all meshes
   for(i=vis_list.size(); i--; ){
      PI3D_mesh_level ml = I3DCAST_OBJECT(vis_list[i])->GetMesh1()->GetLOD1();
      ml->UnlockVertices();
   }
}

//----------------------------

static void BuildSectors(S_load_context &lc, PI3D_frame root, PI3D_scene scene,
   PI3D_model model = NULL){
                           //build sectors
   struct S_hlp{
      PI3D_scene scene;
      PI3D_model model;      //if !NULL, it's considered to be owner of frames
      S_load_context *lc;

      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

         const S_hlp *hp = (S_hlp*)c;
         assert(frm->GetType1()==FRAME_VISUAL);
         do{
                           //check if its father or grand-father is already sector
            if(frm->GetParent1()){
               if(frm->GetParent1()->GetType1()==FRAME_SECTOR){
                  if(!I3DCAST_SECTOR(frm->GetParent1())->IsPrimary())
                     break;
               }
               if(frm->GetParent1()->GetParent1() &&
                  frm->GetParent1()->GetParent1()->GetType1()==FRAME_SECTOR){
                  break;
               }
            }

            C_string prop_data;
            E_PROP_KEYWORD kw_cast_to;

            map<PI3D_frame, C_string>::const_iterator it = hp->lc->frm_properties.find(frm);
            if(it==hp->lc->frm_properties.end())
               break;
            if(!GetPropetriesBegin((*it).second, prop_data, kw_cast_to))
               break;

            while(kw_cast_to != PROP_CAST_SECT && GetPropertiesNext(prop_data, kw_cast_to));
            if(kw_cast_to!=PROP_CAST_SECT)
               break;

            PI3D_sector sct = hp->scene->CreateSector();
            sct->SetName(C_fstring("sector %s", frm->GetName()));
                           //collect all walls forming sector
            vector<C_smart_ptr<I3D_frame> > walls;
            //walls.push_back(frm);

            bool root_help = false;
            bool simple_sector = false;

                           //check additional properties
            E_PROP_KEYWORD kw;

            it = hp->lc->frm_properties.find(frm);
            if(it!=hp->lc->frm_properties.end())
            if(GetPropetriesBegin((*it).second, prop_data, kw))
            do{
               switch(kw){
               case PROP_HELP:
                  root_help = true;
                  break;

               case PROP_SIMPLE_SECTOR:
                  simple_sector = true;
                  break;

                  /*
               case PROP_ENVIRONMENT:
                  {
                     char buf[256];
                     bool b = GetPropertiesValue(prop_data, "%s%n", buf);
                     if(b){
                        S_mat_info mi;
                        mi.mat = hp->lc->driver->CreateMaterial(); mi.mat->Release();
                        mi.filename[0] = buf;
                        mi.ct_flags = TEXTMAP_DIFFUSE | TEXTMAP_MIPMAP;
                        mi.GetMaterial(*hp->lc, sct->GetName());
                        sct->SetEnvMaterial(mi.mat);
                     }else
                        hp->lc->REPORT_ERR(C_fstring("Warning: sector '%s' - keyword 'environment' - expecting name of texture", sct->GetName()));
                  }
                  break;
                  */
               }
            } while(GetPropertiesNext(prop_data, kw));
                              //check additional properties
            struct S_hlp{

               static void cbAddWalls(PI3D_frame root, vector<C_smart_ptr<I3D_frame> > &walls, bool root_help,
                  PI3D_sector sct, S_load_context &lc){

                  walls.push_back(root);

                  for(int i=root->NumChildren1(); i--; ){
                     PI3D_frame f1 = root->GetChild1(i);

                     C_string prop_data;
                     E_PROP_KEYWORD kw_cast_to;

                     map<PI3D_frame, C_string>::const_iterator it = lc.frm_properties.find(f1);
                     if(it==lc.frm_properties.end())
                        continue;
                     if(!GetPropetriesBegin((*it).second, prop_data, kw_cast_to))
                        continue;
                     while(kw_cast_to != PROP_CAST_SECT && GetPropertiesNext(prop_data, kw_cast_to));
                     if(kw_cast_to!=PROP_CAST_SECT)
                        continue;

                     if(root_help){
                                    //re-link to sector
                        f1->LinkTo(sct, I3DLINK_UPDMATRIX);
                     }
                     cbAddWalls(f1, walls, root_help, sct, lc);

                     E_PROP_KEYWORD kw;

                     if(it!=lc.frm_properties.end())
                     if(GetPropetriesBegin((*it).second, prop_data, kw))
                     do{
                        switch(kw){
                        case PROP_HELP:
                           {
                                                      //link all childrens of f1 to frm
                              while(f1->NumChildren()){
                                 f1->GetChild(0)->LinkTo(root, I3DLINK_UPDMATRIX);
                              }
                              I3D_RESULT ir;
                              ir = lc.container->DeleteFrame(f1);
                              assert(I3D_SUCCESS(ir));
                           }
                           break;
                        case PROP_SIMPLE_SECTOR:
                           lc.REPORT_ERR(C_fstring("Warning: keyword 'simple' ignored on linked wall: '%s'", f1->GetName()));
                           break;
                        }
                     } while(GetPropertiesNext(prop_data, kw));
                  }
               }
            };
            S_hlp::cbAddWalls(frm, walls, root_help, sct, *hp->lc);


#ifdef USE_DATABASE
            C_fstring cache_name("%s+%s:sector", (const char*)hp->lc->file_name, frm->GetName());
            C_cache *ck = NULL;
            ck = hp->lc->driver->dbase.OpenRecord(cache_name, hp->lc->file_time);
            bool ok = false;
            if(ck){
               ok = sct->LoadCachedInfo(ck);
               hp->lc->driver->dbase.CloseRecord(ck);
            }
            if(!ok){
#endif

               sct->Build((PI3D_frame*)walls.begin(), walls.size(), simple_sector);
#ifdef USE_DATABASE
               C_cache *ck = hp->lc->driver->dbase.CreateRecord(cache_name, hp->lc->file_time);
               if(ck){
                  sct->SaveCachedInfo(ck);
                  hp->lc->driver->dbase.CloseRecord(ck);
               }
            }
#endif
            sct->Connect(hp->scene);
            sct->LinkTo(frm->GetParent());
            frm->LinkTo(sct);
            //if(hp->model) hp->model->AddFrame(sct);
            //else hp->scene->AddFrame(sct);
            assert(!hp->model);
            hp->lc->container->AddFrame(sct);
            sct->Release();

            if(root_help){    //delete root
                              //link all childrens of f1 to frm
               while(frm->NumChildren()){
                  frm->GetChild(0)->LinkTo(frm->GetParent(), I3DLINK_UPDMATRIX);
               }
               I3D_RESULT ir;
               ir = hp->lc->container->DeleteFrame(frm);
               assert(I3D_SUCCESS(ir));
            }
                              //nested sectors unsupported now
            //return I3DENUMRET_SKIPCHILDREN;
         }while(false);    //allow break-out from block

         //return I3DENUMRET_OK;
         return I3DENUMRET_SKIPCHILDREN;
      }
   } hlp = {scene, model, &lc};

   root->EnumFrames(S_hlp::cbEnum, (dword)&hlp, ENUMF_VISUAL);
}

//----------------------------
//----------------------------
                              //animation set loader

I3D_RESULT I3D_animation_set::Open(const char* fname, dword flags, PI3D_LOAD_CB_PROC cb_proc,
   void *cb_context, PI3D_scene scene_hlp){

   C_cache icache(32*1024);
                              //open file
   if(!icache.open(fname, CACHE_READ))
      return I3DERR_NOFILE;

   if(!cb_proc)
      flags &= ~I3DLOAD_LOG;

   S_load_context lc(drv);
   lc.is = &icache;
   lc.load_flags = flags;
   lc.load_cb_proc = cb_proc;
   lc.load_cb_context = cb_context;
   lc.anim_set = this;
   lc.file_name = fname;
   lc.file_name.ToLower();
   lc.file_length = lc.to_pos = icache.filesize();
   dtaGetTime(icache.GetHandle(), lc.file_time);

   bool b = lc.ChunkReader();
   I3D_RESULT ir = b ? I3D_OK : lc.cancelled ? I3DERR_CANCELED : I3DERR_FILECORRUPTED;
   icache.close();

   if(I3D_SUCCESS(ir)){
                              //sync all times
      dword max_time=0;
      int i;
      for(i=0; i<anims.size(); i++) max_time = Max(max_time, ((PI3D_keyframe_anim)(PI3D_animation_base)anims[i].anim)->GetEndTime());
      for(i=0; i<anims.size(); i++) ((PI3D_keyframe_anim)(PI3D_animation_base)anims[i].anim)->SetEndTime(max_time);
   }
   return ir;
}

//----------------------------
//----------------------------

I3D_RESULT I3D_container::Open(S_load_context &lc){

   filename = lc.file_name;

   Close();
   C_cache icache(128*1024);
                              //open file
   if(!icache.open(lc.file_name, CACHE_READ))
      return I3DERR_NOFILE;

   if(!lc.load_cb_proc){
      lc.load_flags &= ~(I3DLOAD_LOG | I3DLOAD_PROGRESS);
   }

   if(lc.load_flags&I3DLOAD_PROGRESS)
      lc.load_cb_proc(CBM_PROGRESS, FloatAsInt(0.0f), 0, lc.load_cb_context);

   lc.is = &icache;
   lc.file_length = lc.to_pos = icache.filesize();
   dtaGetTime(icache.GetHandle(), lc.file_time);

   bool b = lc.ChunkReader();
   I3D_RESULT ir = b ? I3D_OK : lc.cancelled ? I3DERR_CANCELED : I3DERR_FILECORRUPTED;
   icache.close();

   if(lc.load_flags&I3DLOAD_PROGRESS)
      lc.load_cb_proc(CBM_PROGRESS, FloatAsInt(1.0f), 2, lc.load_cb_context);

   if(I3D_FAIL(ir))
      Close();

   return ir;
}

//----------------------------
                              //scene loader
I3D_RESULT I3D_scene::Open(const char* fname, dword flags, PI3D_LOAD_CB_PROC cb_proc,
   void *cb_context, PI3D_scene scene_hlp){

   S_load_context lc(drv);
   lc.load_flags = flags;
   lc.load_cb_proc = cb_proc;
   lc.load_cb_context = cb_context;
   lc.scene = this;
   lc.container = &cont;
   lc.root_frame = GetPrimarySector1();
   lc.file_name = fname;
   lc.file_name.ToLower();

   I3D_RESULT ir = cont.Open(lc);
   if(I3D_SUCCESS(ir)){

      BuildSectors(lc, GetPrimarySector1(), this);

      S_al_help alh;
      alh.root_frame = GetPrimarySector1();
      alh.drv = drv;
      alh.lc = &lc;
      EnumFrames(cbAfterLoad, (dword)&alh, ENUMF_VISUAL | ENUMF_LIGHT | ENUMF_DUMMY);
                              //relese frames
      for(int i=alh.frames_to_delete.size(); i--; )
         DeleteFrame(alh.frames_to_delete[i]);

      SmoothSceneNormals(this, lc);
   }
   return ir;
}

//----------------------------
//----------------------------

I3D_RESULT I3D_model::Open(const char *fname, dword flags, I3D_LOAD_CB_PROC *cb_proc,
   void *cb_context, PI3D_scene scene_hlp){

   S_load_context lc(drv);
   lc.load_flags = flags;
   lc.load_cb_proc = cb_proc;
   lc.load_cb_context = cb_context;
   lc.root_frame = this;
   lc.scene = scene_hlp;
   lc.container = &cont;
   lc.anim_set = drv->CreateAnimationSet();
   lc.file_name = fname;
   lc.file_name.ToLower();

   I3D_RESULT ir = cont.Open(lc);
   if(I3D_SUCCESS(ir)){
      if((flags&I3DLOAD_BUILD_SECTORS) && scene_hlp)
         BuildSectors(lc, this, scene_hlp, this);

                              //mark all children models
      for(int i=0; i<cont.frames.size(); i++)
         cont.frames[i]->frm_flags |= FRMFLAGS_SUBMODEL;

      S_al_help alh;
      alh.root_frame = this;
      alh.drv = drv;
      alh.lc = &lc;
      EnumFrames(cbAfterLoad, (dword)&alh, ENUMF_VISUAL | ENUMF_DUMMY, NULL);
                              //relese frames
      for(i=alh.frames_to_delete.size(); i--; )
         DeleteFrame(alh.frames_to_delete[i]);
      SetName(name);
   }
   if(lc.anim_set->NumLinks())
      anim_set = lc.anim_set;
   lc.anim_set->Release();

   return ir;
}

//----------------------------
//----------------------------
