#ifndef __LOAD_3DS_H
#define __LOAD_3DS_H

#include "light.h"
#include "soundi3d.h"
#include "mesh.h"

//----------------------------

#define MAX_UV_CHANNELS    2  //number of uv channels we may load

//----------------------------

#pragma pack(push,1)
class C_rgb{
   byte color[3];
public:
   C_rgb(dword val=0){
      color[0]=val>>16;
      color[1]=val>>8;
      color[2]=val;
   }
   C_rgb(byte r, byte g, byte b){
      color[0]=r;
      color[1]=g;
      color[2]=b;
   }
   byte operator[](int i){
      return color[i];
   }

   S_vector ToVector() const{
      return S_vector((float)color[0] / 255.0f,
         (float)color[1] / 255.0f, (float)color[2] / 255.0f);
   }
};
#pragma pack(pop)

//----------------------------

#define LOADF_TXT_TRANSP 1          //colorkeyed
#define LOADF_TXT_TRUECOLOR 2
#define LOADF_TXT_DIFF_ANIM 4
#define LOADF_TXT_REFL_ANIM 8
#define LOADF_TXT_BUMP_ANIM 0x10
#define LOADF_TXT_DIFFUSE_KEEP 0x20
#define LOADF_TXT_CAMERA 0x40
#define LOADF_TXT_TRANSP_ALPHA 0x80 //colorkeyed with alpharef at zero

//----------------------------
                              //internal stock lists for object keeping
                              // between creation and hierarchy build
struct S_mat_info{
   C_smart_ptr<I3D_material> mat;
   C_string filename[5];      //[diffuse, opacity, environment, bump, procedural]
   dword ct_flags;
   int diffuse_anim_speed;
   E_ENVIROMENTAL_EFFECT env_effect;
   float environment_power, bump_power;
   bool txt_created;
   dword load_flags;
   int bump_anim_speed;
   S_mat_info():
      load_flags(0),
      txt_created(false),
      env_effect(EE_NULL)
   {}
   S_mat_info(PI3D_material m1, dword ct_flags1, const C_string &sd, const C_string &so,
      const C_string &ss, const C_string &sb):
      env_effect(EE_NULL),
      mat(m1), ct_flags(ct_flags1)
   {
      filename[0] = sd;
      filename[1] = so;
      filename[2] = ss;
      filename[3] = sb;
   }

   PI3D_material GetMaterial(const struct S_load_context &lc);
   I3D_animated_texture *CreateAnimatedTexture(const char *fname0, const char *fname1,
      const S_load_context &lc, PI3D_texture tp_last,
      dword ct_flags, int anim_speed);
};

//----------------------------
                              //structure passed to each chunk loading function
struct S_load_context{
   C_cache *is;               //file cache
   dword to_pos;              //position to which we may read from file
   dword file_length;
   int tab;                   //tab for log
   dword load_flags;

   PI3D_driver driver;
   PI3D_scene scene;          //valid during scene loading
   PI3D_frame root_frame;     //valid during scene or model loading
   PI3D_container container;  //valid during scene or model loading
   PI3D_animation_set anim_set; //valid during animation loading

                              //loader callback function
   I3D_LOAD_CB_PROC *load_cb_proc;
   void *load_cb_context;

                              //recent values
   C_rgb recent_rgb;
   S_vector recent_vector;
   byte recent_percentage;
   C_string recent_string;
   PI3D_material recent_mat;
   PI3D_light recent_light;
   PI3D_sound recent_sound;
   float recent_light_mult;
   float recent_light_range[2];
   float recent_camera_range[2];
   C_string recent_instance_name;
   short recent_node_id;
   short recent_parent_node_id;
   PI3D_frame recent_frame;
   C_string recent_properties;
   PI3D_keyframe_anim recent_anim;
   int recent_max_anim_time;
   I3D_bbox recent_dummy_bbox;
   S_vector recent_pivot;
   S_vector recent_nu_scale;
   set<C_string> mesh_db_valid;  //meshes listed here were loaded by LoadCachedInfo
   set<C_string> mesh_db_save;   //meshes listed here will be saved into cache
   map<PI3D_frame, C_string> frm_properties;

   vector<I3D_vertex_mesh> recent_vertices;
   vector<I3D_triface> recent_faces;
   vector<byte> recent_smooth_groups;
   vector<I3D_text_coor> recent_uv_verts[MAX_UV_CHANNELS];
   vector<I3D_triface> recent_uv_faces[MAX_UV_CHANNELS];
   struct S_face_group{
      vector<word> face_indicies;
      C_smart_ptr<I3D_material> mat;
      S_face_group(){}
      S_face_group(const S_face_group &fg){ operator =(fg); }
      S_face_group &operator =(const S_face_group &fg){ face_indicies = fg.face_indicies; mat = fg.mat; return *this; }
   };
   vector<S_face_group> recent_face_groups;
   bool allow_cache;          //helper - subchunks may clear it to avoid upper chunk to write to cache (if there's warning)

                              //for caching
   C_string file_name;
   dword file_time[2];

   struct S_map{
      C_string s;
      bool inverse;
      byte percentage;
      float scale;
      int anim_speed;         //in msec

      S_map():
         inverse(false),
         percentage(100),
         scale(1.0)
      {}
   } *diffuse_map, *opacity_map, *environment_map, *bump_map, *curr_map;
   E_ENVIROMENTAL_EFFECT env_effect;
   C_string txt_proc_name;

   dword recent_txt_flags;

   byte recent_text_mipmap;   //0 - default(no), 1 - yes, 2 - disable
   float recent_self_illum;
   float recent_text_blur;
   bool cancelled;

   S_load_context(PI3D_driver in_driver):
      driver(in_driver),
      cancelled(false),
      scene(NULL),
      root_frame(NULL),
      container(NULL),
      anim_set(NULL),
      recent_mat(NULL),
      recent_light(NULL),
      recent_sound(NULL),
      //recent_nurbs(NULL),
      recent_frame(NULL),
      recent_anim(NULL),
      diffuse_map(NULL), opacity_map(NULL), environment_map(NULL), bump_map(NULL), curr_map(NULL),
      is(NULL), 
      tab(0), 
      load_flags(0),
      load_cb_proc(NULL), load_cb_context(NULL)
   {
   }

//----------------------------
// If this function returns true, it is treated as a request to interrupt loading.
   bool ShowProgress(){
      if((load_flags&I3DLOAD_PROGRESS) && load_cb_proc){
         //driver->GetGraphInterface()->ProcessWinMessages(true);
         float f = (float)is->tellg() / (float)file_length;
         cancelled = load_cb_proc(CBM_PROGRESS, FloatAsInt(f), 1, load_cb_context);
         return cancelled;
      }
      return false;
   }

   void REPORT_ERR(const char *cp) const{
      if((load_flags&I3DLOAD_ERRORS) && load_cb_proc){
         load_cb_proc(CBM_ERROR,
            (dword)(const char*)C_fstring("%s (file '%s')", cp, (const char*)file_name),
            0, load_cb_context);
      }
   }

   void LOG(const char *cp){
      char buf[512];
      memset(buf, ' ', Min(256, tab*2));
      strcpy(&buf[tab*2], cp);
      load_cb_proc(CBM_LOG, (dword)buf, 0, load_cb_context);
   }

   vector<C_smart_ptr<I3D_light> > light_list;
   map<C_string, C_smart_ptr<I3D_mesh_object> > mesh_list;
   map<C_string, S_mat_info> mat_list;

   static PI3D_light FindStockItem(const vector<C_smart_ptr<I3D_light> > &lst, const C_string &name){
      for(int i=lst.size(); i--; )
      if(lst[i]->GetName1()==name) return lst[i];
      return NULL;
   }

   static PI3D_sound FindStockItem(const vector<C_smart_ptr<I3D_sound> > &lst, const C_string &name){
      for(int i=lst.size(); i--; ) if(lst[i]->GetName1()==name) return lst[i];
      return NULL;
   }
                              //scene frames
private:
                              //id-based array of frames
   vector<PI3D_frame> frame_id_list; 
public:
   PI3D_frame GetFrame(int id){
      return id<frame_id_list.size() ? frame_id_list[id] : NULL;
   }

   bool AddFrameID(int id, PI3D_frame frm){
      if(frame_id_list.size() < (id+1)) frame_id_list.resize(id+256, NULL);
      frame_id_list[id] = frm;
      return true;
   }


   void ReadString(C_string &s){

      char buf[512], *bp = buf;
      int i = sizeof(buf)-1;
      do{
         is->read(bp, 1);
         if(!--i) *(++bp) = 0;
      }while(*bp++);
      s = buf;
   }

   void ValidateName(C_string &str, bool can_be_dot){

      for(int i=str.Size(); i--; ){
         if(!isalnum(str[i])){
            switch(str[i]){
            case '.':
               if(can_be_dot){
                  str[i] = '_';
                  break;
               }
                              //flow...
            case '<':
                              //all others are invalid
               REPORT_ERR(C_fstring("invalid object name: '%s' (%s)", (const char*)str,
                  (const char*)C_fstring("file \"%s\"", (const char*)file_name)));
               return;
            }
         }
      }
   }
   bool ChunkReader();

//----------------------------
// Clean mesh - remove invalid faces, unused vertices, remaps faces.
// Operates on:
//    - recent_vertices
//    - recent_faces
//    - recent_uv_faces
//    - recent_face_groups
//    - recent_smooth_groups
   void CleanMesh();

//----------------------------
// Assigns uv coordinates to vertices (from uv-map faces), duplicates vertices as necessary.
// Operates on:
//    - recent_vertices
//    - recent_faces
// Uses:
//    - recent_uv_verts
//    - recent_uv_faces
   void AssignUVCoordinates(int num_orig_verts, vector<word> &vertex_map);

//----------------------------
// Put all UV coors into near zero range, graphics hardware don't like too big float numbers
// in texture coordinates.
// Operates on:
//    - recent_vertices
   void UV_Check();

//----------------------------
// Generate normals on set of vertices, using info from smooth groups.
// Operates on:
//    - recent_vertices
// Uses:
//    - recent_faces
//    - recent_smooth_groups
   void MakeVertexNormals(int num_orig_verts, vector<word> &vertex_map);
};

//----------------------------
#endif