/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Material.cpp
   Content: All about materials.
--------------------------------------------------------*/

#include "all.h"
#include "material.h"


//----------------------------


                              //material class
I3D_material::I3D_material(PI3D_driver drv1):
   ref(1),
   env_effect(EE_NULL),
   refl_opacity(1.0f),
   bump_opacity(1.0f),
   mat_flags(0),
   sort_id(0),
   alpharef(0x80),
   drv(drv1)
{
   drv->AddCount(I3D_CLID_MATERIAL);

   memset(&md, 0, sizeof(md));
   md.diffuse = S_vector(1.0f, 1.0f, 1.0f);
   md.diffuse_alpha = 1.0f;
   //md.ambient_alpha = 1.0f;
   //md.emissive_alpha = 1.0f;
}

//----------------------------

I3D_material::~I3D_material(){

   SetTexture(NULL);
   //if(drv->last_render_mat==this) drv->last_render_mat = NULL;
   drv->DecCount(I3D_CLID_MATERIAL);
}

//----------------------------

I3D_RESULT I3D_material::SetAmbient(const S_vector &v){

   md.ambient = v;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_material::SetDiffuse(const S_vector &v){

   md.diffuse = v;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_material::SetSpecularColor(const S_vector &v){

   //md.specular = v;
   //return I3D_OK;
   return I3DERR_UNSUPPORTED;
}

//----------------------------

I3D_RESULT I3D_material::SetSpecularPower(float f){

   //md.specular_power = f;
   //return I3D_OK;
   return I3DERR_UNSUPPORTED;
}

//----------------------------

I3D_RESULT I3D_material::SetSpecularAngle(float f){

   //md.specular_angle = f;
   //return I3D_OK;
   return I3DERR_UNSUPPORTED;
}

//----------------------------

I3D_RESULT I3D_material::SetEmissive(const S_vector &v){

   md.emissive = v;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_material::SetAlpha(float a){

   a = Max(0.0f, Min(1.0f, a));

   //md.ambient_alpha  = a;
   md.diffuse_alpha  = a;
   //md.emissive_alpha = a;

   mat_flags &= ~MATF_DIFFUSE_ALPHA;
   if(a<1.0f)
      mat_flags |= MATF_DIFFUSE_ALPHA;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_material::SetTexture(PI3D_texture_base tp1){

   alpharef = 0x80;

   tp[MTI_DIFFUSE] = tp1;
   mat_flags &= ~(MATF_CKEY_ALPHA | MATF_TXT_ALPHA | MATF_SELF_ALPHA);
   if(tp1){
      if(tp1->GetTxtFlags()&TXTF_CKEY_ALPHA){
         mat_flags |= MATF_CKEY_ALPHA;
         if(mat_flags&MATF_CKEY_ZERO_REF){
                              //minimal alpha ref requested for colorkeys
            alpharef = 0x20;
         }else{
                              //automatic alpha ref
            alpharef = 0xe0;
            if(tp1->GetTxtFlags()&TXTF_MIPMAP)
               alpharef = 0xa0;
         }
      }
      if(tp1->GetTxtFlags()&TXTF_ALPHA){
         mat_flags |= MATF_TXT_ALPHA;
         alpharef = 0xe0;
      }
      if(tp1->GetTxtFlags()&TXTF_SELF_ALPHA){
         mat_flags |= TXTF_ALPHA;
      }

      sort_id = tp1->GetSortID() & PRIM_SORT_MAT_MASK;
      if(GetAlpha() != 1.0f)
         mat_flags &= ~(TXTF_CKEY_ALPHA);
   }else{
      sort_id = 0;
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_material::Set2Sided(){

   //mat_flags |= MATF_2SIDED;
   md.two_sided = true;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_material::Duplicate(CPI3D_material mp){

   if(mp==this)
      return I3D_OK;
   name = mp->name;
   md = mp->md;
   mat_flags = mp->mat_flags;
   refl_opacity = mp->refl_opacity;
   bump_opacity = mp->bump_opacity;
   env_effect = mp->env_effect;

   for(int i=0; i<MTI_LAST; i++){
      tp[i] = mp->tp[i];
   }

   Update();
   return I3D_OK;
}

//----------------------------
//----------------------------
