#ifndef __MODEL_H
#define __MODEL_H

#include "anim.h"
#include "frame.h"
#include "container.h"

//----------------------------

struct S_shadow_info{
                              //last light used to majke shadow
                              // this pointer is not ref-counted, so it can't be referenced!
   C_smart_ptr<I3D_light> last_light;

                              //during transition, this is last successive dir of 'last_light'
   S_normal trans_base_dir;

                              //this holds direction used in rendering of previous frame
   S_normal last_dir;

                              //counter of transition, 0 if not animating
   dword transition_count;

   int shadow_select_count;   //counter for re-selecting shadow light

   S_shadow_info():
      shadow_select_count(0)
   {
      Reset();
   }

   void Reset(){
      last_light = NULL;
      transition_count = 0;
   }
};

//----------------------------

class I3D_model: public I3D_frame{

   I3D_container cont;
   C_smart_ptr<I3D_animation_set> anim_set;    //loaded anim set

   C_bound_volume hr_bound;

   I3D_container *in_container;//container which ticks ours animations

   friend I3D_driver;
   I3D_model(PI3D_driver);
   ~I3D_model();

   S_shadow_info *shadow_info;

   friend class I3D_scene;
public:

   inline PI3D_container GetContainer1(){ return &cont; }
                              //some over-riding
   I3DMETHOD(EnumFrames)(I3D_ENUMPROC*p, dword user, dword flags = ENUMF_ALL,
      const char *mask = NULL) const{
      return I3D_frame::EnumFrames(p, user, flags | ENUMF_MODEL_ORIG, mask);
   }
   I3DMETHOD_(PI3D_frame,FindChildFrame)(const char *name, dword flags) const{
      return I3D_frame::FindChildFrame(name, flags | ENUMF_MODEL_ORIG);
   }
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD_(void,SetName)(const char *cp);
   I3DMETHOD_(dword,SetFlags)(dword new_flags, dword flags_mask);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD_(const I3D_bound_volume&,GetHRBoundVolume)() const{
      if(!(frm_flags&FRMFLAGS_HR_BOUND_VALID))
         ((PI3D_model)this)->hr_bound.MakeHierarchyBounds((PI3D_frame)this);
      return hr_bound.bound_local; 
   }

   I3DMETHOD_(I3D_RESULT,SetAnimation)(dword stage, PI3D_animation_set,
      dword options, float scale_factor, float blend_factor, float time_scale);
   I3DMETHOD_(I3D_RESULT,SetAnimScaleFactor)(dword stage, float scale_factor);
   I3DMETHOD_(I3D_RESULT,SetAnimBlendFactor)(dword stage, float blend_factor);
   I3DMETHOD_(I3D_RESULT,SetAnimSpeed)(dword stage, float speed);
   I3DMETHOD(StopAnimation)(dword stage);
   I3DMETHOD_(PI3D_animation_set,GetAnimationSet)() const{ return anim_set; }

   I3D_CONTAINER(;)
   I3DMETHOD_(I3D_container*,GetContainer)(){ return &cont; }
};

//----------------------------

#endif