#ifndef __OBJECT_H
#define __OBJECT_H

#include <vector>
#include "mesh.h"
#include "visual.h"
#include "camera.h"
#include "scene.h"


//----------------------------

class I3D_object: public I3D_visual{

   friend I3D_visual *CreateObject(PI3D_driver);

//----------------------------
// Make clone from specified mesh, keep in LOD #0.
   void Clone(I3D_mesh_base *mb){
      mesh = drv->CreateMesh();
      PI3D_mesh_level ml = drv->CreateMeshLevel();
      mesh->SetLOD(ml);
      ml->Clone(mb);
      ml->Release();
      mesh->Release();

      vertex_buffer.DestroyD3DVB(drv);
      vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID | VISF_BOUNDS_VALID);
      frm_flags &= ~FRMFLAGS_HR_BOUND_VALID;
   }
protected:

   C_smart_ptr<I3D_mesh_object> mesh;
   I3D_object(PI3D_driver);
public:
   I3D_object(const I3D_object&);               //not implemented
   I3D_object& operator =(const I3D_object&);   //not implemented

   inline PI3D_mesh_object GetMesh1() const{ return mesh; }

public:
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD_(I3D_mesh_base*,GetMesh)() const{
      if(!mesh)
         return NULL;
      return mesh->GetLOD(0);
   }

   virtual bool ComputeBounds();
   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
   virtual void RenderSolidMesh(PI3D_scene, bool, const I3D_cylinder*);
   virtual void PrepareDestVB(I3D_mesh_base*, dword = 1);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD(SetMeshObject)(PI3D_mesh_object);
   I3DMETHOD_(I3D_mesh_object*,GetMeshObject)() const{ return mesh; }

   I3DMETHOD_(void,SetUpdate)();
   I3DMETHOD(SetMesh)(PI3D_mesh_level);
};

//----------------------------

#endif
