/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Object.cpp
   Content: Object frame.
--------------------------------------------------------*/

#include "all.h"
#include "object.h"
#include "scene.h"
#include "camera.h"



                              //debug:
//#define DEBUG_NO_CACHE_LIGHT  //don't cache light info in vertices

//----------------------------

static I3D_visual *CreateObject(PI3D_driver drv){
   return new I3D_object(drv);
}

static struct S_auto_init_object{
   S_auto_init_object(){
      RegisterVisual(CreateObject, I3D_VISUAL_OBJECT, "Object");
   }
} auto_init_object;


//----------------------------
//----------------------------
                              //object frame class
I3D_object::I3D_object(PI3D_driver d):
   I3D_visual(d)
   //curr_lod(0)
{
   visual_type = I3D_VISUAL_OBJECT;
}

//----------------------------

I3D_RESULT I3D_object::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()!=FRAME_VISUAL)
      return I3D_frame::Duplicate(frm);

   PI3D_visual vis = I3DCAST_VISUAL(frm);
   switch(vis->GetVisualType()){
   case I3D_VISUAL_OBJECT:
      SetMeshObject(I3DCAST_OBJECT(frm)->GetMesh1());
      break;
   default:
      {
         mesh = NULL;
         I3D_mesh_base *mb = vis->GetMesh();
         if(mb){
            PI3D_mesh_level ml = drv->CreateMeshLevel();
            ml->Clone(mb);
            SetMesh(ml);
            ml->Release();

         }
      }
   }
   return I3D_visual::Duplicate(vis);
}

//----------------------------

I3D_RESULT I3D_object::SetMesh(PI3D_mesh_level ml){

   mesh = drv->CreateMesh();
   mesh->SetLOD(ml);
   mesh->Release();

   vertex_buffer.DestroyD3DVB(drv);
   vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID | VISF_BOUNDS_VALID);
   frm_flags &= ~FRMFLAGS_HR_BOUND_VALID;

   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_object::SetMeshObject(PI3D_mesh_object m){

   mesh = m;

   vertex_buffer.DestroyD3DVB(drv);
   vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID | VISF_BOUNDS_VALID);
   frm_flags &= ~FRMFLAGS_HR_BOUND_VALID;
   return I3D_OK;
}

//----------------------------

void I3D_object::PrepareDestVB(I3D_mesh_base *mb, dword num_txt_sets){

   mb->ComputeBoundingVolume();

   I3D_visual::PrepareDestVB(mb, num_txt_sets);
   vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
}

//----------------------------

void I3D_object::SetUpdate(){

   vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
}

//----------------------------
//----------------------------

bool I3D_object::ComputeBounds(){

   if(!mesh){
      bound.bound_local.bbox.Invalidate();
      bound.bound_local.bsphere.pos.Zero();
      bound.bound_local.bsphere.radius = 0.0f;
      return true;
   }

   PI3D_mesh_base mb = mesh->GetLOD1();

   mb->ComputeBoundingVolume();
   bound.bound_local = mb->GetBoundingVolume();

   S_vector bbox_diag_half = (bound.bound_local.bbox.max - bound.bound_local.bbox.min) * .5f;
   bound.bound_local.bsphere.pos = (bound.bound_local.bbox.min + bbox_diag_half);
   bound.bound_local.bsphere.radius = bbox_diag_half.Magnitude();

   vis_flags |= VISF_BOUNDS_VALID;
   frm_flags &= ~FRMFLAGS_BSPHERE_TRANS_VALID;

   return true;
}

//----------------------------

extern const dword vs_decl_vertex_mesh[];
const dword vs_decl_vertex_mesh[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT3),   //position
   D3DVSD_REG(2, D3DVSDT_FLOAT3),   //normal
   D3DVSD_REG(1, D3DVSDT_FLOAT2),   //UV coordinates
   D3DVSD_END()
};

//----------------------------

void I3D_object::AddPrimitives(S_preprocess_context &pc){

   if(!mesh)
      return;
   I3D_mesh_base *mb = mesh->GetLOD(0);
   AddPrimitives1(mb, pc);
}

//----------------------------

void I3D_object::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   I3D_mesh_base *mb = mesh->I3D_mesh_object::GetLOD1();

   drv->SetClipping(rp.clip);
                              //preprocess lights - once per frame for PV, each time when HWVP
   if(dest_vb_trans_id!=pc.render_view_id || drv->IsDirectTransform()){
      dest_vb_trans_id = pc.render_view_id;

      if(last_auto_lod != rp.curr_auto_lod){
         last_auto_lod = rp.curr_auto_lod;
         vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
      }

      dword vertex_count;
      const vector<C_auto_lod> &auto_lods = mb->GetAutoLODs();
      if(rp.curr_auto_lod < 0){
         vertex_count = mb->vertex_buffer.NumVertices();
      }else{
         vertex_count = auto_lods[rp.curr_auto_lod].vertex_count;
      }
                              //transform vertices
      dword prep_flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX;

      if(drv->IsDirectTransform()){
         if(pc.mode!=RV_SHADOW_CASTER)
            prep_flags |= VSPREP_MAKELIGHTING;
         prep_flags |= VSPREP_COPY_UV;
      }else{
         if(pc.mode!=RV_SHADOW_CASTER){
#ifdef DEBUG_NO_CACHE_LIGHT
            vis_flags &= ~VISF_DEST_LIGHT_VALID;  
#endif
            if(!(vis_flags&VISF_DEST_LIGHT_VALID)){
               prep_flags |= VSPREP_MAKELIGHTING;
               vis_flags |= VISF_DEST_LIGHT_VALID;
            }
         }
         if(!(vis_flags&VISF_DEST_UV0_VALID)){
            prep_flags |= VSPREP_COPY_UV;
            vis_flags |= VISF_DEST_UV0_VALID;
         }
      }
      if(pc.mode==RV_SHADOW_CASTER)
         prep_flags |= VSPREP_NO_ENVMAP | VSPREP_NO_FOG;

      I3D_driver::S_vs_shader_entry_in se(vs_decl_vertex_mesh);
      PrepareVertexShader(mb->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);

      if(!drv->IsDirectTransform()){
         drv->DisableTextureStage(1);

         drv->SetStreamSource(mb->vertex_buffer.GetD3DVertexBuffer(), sizeof(I3D_vertex_mesh));
         HRESULT hr;
         hr = drv->GetDevice1()->ProcessVertices(mb->vertex_buffer.D3D_vertex_buffer_index,
            vertex_buffer.D3D_vertex_buffer_index, vertex_count,
            vertex_buffer.GetD3DVertexBuffer(), D3DPV_DONOTCOPYDATA);
         CHECK_D3D_RESULT("ProcessVertices", hr);
      }
   }
   DrawPrimitive1(mb, pc, rp);
}

//----------------------------

void I3D_object::RenderSolidMesh(PI3D_scene scene, bool debug, const I3D_cylinder *vf){

   I3D_driver::S_vs_shader_entry_in se(vs_decl_vertex_mesh);
   se.AddFragment(I3D_driver::VSF_TRANSFORM);
   se.AddFragment(I3D_driver::VSF_TEXT_PROJECT);
   if(debug)
      se.AddFragment(I3D_driver::VSF_SIMPLE_DIR_LIGHT);

   const I3D_driver::S_vs_shader_entry *se1 = drv->GetVSHandle(se);
   drv->SetVertexShader(se1->vs_handle);

   I3D_visual::RenderSolidMesh(scene, debug, vf);
}

//----------------------------
//----------------------------


