#include "all.h"
#include "object.h"


//----------------------------
//----------------------------

class I3D_object_uv_shift: public I3D_object{
                              //texture shifting
   struct{
      float add[2];
      float reserved[2];
   } curr_vals;
   S_vector2 uv_add;
   enum E_MODE{
      MODE_ANIMATE,
      MODE_DELTA,
   } mode;

   void ComputeShift();
public:
   I3D_object_uv_shift(PI3D_driver d1):
      I3D_object(d1),
      mode(MODE_ANIMATE)
   {
      visual_type = I3D_VISUAL_UV_SHIFT;
      uv_add.Zero();
      memset(&curr_vals, 0, sizeof(curr_vals));
   }

   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);

   virtual I3D_RESULT I3DAPI SetProperty(dword index, dword value){
      switch(index){
      case 0: uv_add.x = IntAsFloat(value); break;
      case 1: uv_add.y = IntAsFloat(value); break;
      case 2:
         if(value>MODE_DELTA)
            return I3DERR_INVALIDPARAMS;
         mode = (E_MODE)value;
         break;
      default:
         return I3DERR_UNSUPPORTED;
      }
      vis_flags &= ~VISF_DEST_UV0_VALID;
      return I3D_OK;
   }

   virtual dword I3DAPI GetProperty(dword index) const{
      switch(index){
      case 0: return FloatAsInt(uv_add.x);
      case 1: return FloatAsInt(uv_add.y);
      case 2: return mode;
      }
      return 0;
   }
};

//----------------------------

void I3D_object_uv_shift::ComputeShift(){

   int time_delta = drv->GetRenderTime() - last_render_time;
   float tsec = (float)time_delta * .001f;

   curr_vals.add[0] += uv_add.x * tsec;
   curr_vals.add[1] += uv_add.y * tsec;
                              //clamp to range -1.0 ... 1.0
   curr_vals.add[0] = fmod(curr_vals.add[0], 1.0f);
   curr_vals.add[1] = fmod(curr_vals.add[1], 1.0f);
}

//----------------------------

extern const dword vs_decl_vertex_mesh[];

//----------------------------

void I3D_object_uv_shift::AddPrimitives(S_preprocess_context &pc){

   if(!mesh)
      return;
   
   PI3D_mesh_base mb = mesh->GetLOD1();
   AddPrimitives1(mb, pc);

   if(pc.mode!=RV_SHADOW_CASTER && mode==MODE_ANIMATE)
      ComputeShift();
}

//----------------------------

void I3D_object_uv_shift::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   PI3D_mesh_base mb = mesh->I3D_mesh_object::GetLOD1();

   drv->SetClipping(rp.clip);
   HRESULT hr;

   if(dest_vb_trans_id!=pc.render_view_id || drv->IsDirectTransform()){
      dest_vb_trans_id = pc.render_view_id;

      LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();

      if(last_auto_lod != rp.curr_auto_lod){
         last_auto_lod = rp.curr_auto_lod;
         vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
      }

      dword vertex_count;
      const vector<C_auto_lod> &auto_lods = mb->GetAutoLODs();
      if(rp.curr_auto_lod < 0){
         vertex_count = mb->vertex_buffer.NumVertices();
      }else{
         vertex_count = auto_lods[rp.curr_auto_lod].vertex_count;
      }

      I3D_driver::S_vs_shader_entry_in se(vs_decl_vertex_mesh);

      dword prep_flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX | VSPREP_NO_COPY_UV;
      if(pc.mode!=RV_SHADOW_CASTER){
         if(drv->IsDirectTransform()){
            prep_flags |= VSPREP_MAKELIGHTING;
         }else{
            if(!(vis_flags&VISF_DEST_LIGHT_VALID)){
               prep_flags |= VSPREP_MAKELIGHTING;
               vis_flags |= VISF_DEST_LIGHT_VALID;
            }
         }
         se.AddFragment(I3D_driver::VSF_SHIFT_UV_0);
         {
            switch(mode){
            case MODE_ANIMATE:
               break;
            case MODE_DELTA:
               vis_flags |= VISF_DEST_UV0_VALID;
               curr_vals.add[0] = uv_add[0];
               curr_vals.add[1] = uv_add[1];
               break;
            default:
               assert(0);
            }
                                 //feed UV constants
         }
         hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_UV_SHIFT, &curr_vals, 1);
         CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
      }else{
         prep_flags |= VSPREP_NO_ENVMAP | VSPREP_NO_FOG;
      }
   
      PrepareVertexShader(mb->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);

      if(!drv->IsDirectTransform()){
         drv->DisableTextureStage(1);

         drv->SetStreamSource(mb->vertex_buffer.GetD3DVertexBuffer(), sizeof(I3D_vertex_mesh));
         hr = lpDev8->ProcessVertices(mb->vertex_buffer.D3D_vertex_buffer_index,
            vertex_buffer.D3D_vertex_buffer_index, vertex_count,
            vertex_buffer.GetD3DVertexBuffer(), D3DPV_DONOTCOPYDATA);
         CHECK_D3D_RESULT("ProcessVertices", hr);
      }
   }
   DrawPrimitive1(mb, pc, rp);
}

//----------------------------
//----------------------------

static const S_visual_property props[] = {
   I3DPROP_FLOAT, "Anim U", I3DPROP_FLOAT, "Anim V",
   I3DPROP_ENUM, "Mode\0Animate\0Delta\0"
};

static I3D_visual *CreateObject_uv_shift(PI3D_driver drv){
   return new I3D_object_uv_shift(drv);
}

static struct S_auto_init_object_uvs{
   S_auto_init_object_uvs(){
      RegisterVisual(CreateObject_uv_shift, I3D_VISUAL_UV_SHIFT, "Object (UV shift)",
         props, sizeof(props)/sizeof(S_visual_property));
   }
} auto_init_object_uv_shift;


//----------------------------
//----------------------------

