/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Render.cpp
   Content: Rendering pipeline.
--------------------------------------------------------*/

#include "all.h"
#include <isound2.h>
#include "light.h"
#include "camera.h"
#include "scene.h"
#include "model.h"
#include "visual.h"
#include "occluder.h"
#include "target.h"
#include "joint.h"
#include "volume.h"
#include <malloc.h>


#define MAX_COMPUTED_OCCLUDERS 14

#define SHADOW_LENGTH 8.0f    //length of cast shadows

                              //debugging:
#define DEBUG_FORCE_CLIP      //always pass clip flags to AddPrimitives - D3D VB bug?!

//----------------------------

void S_preprocess_context::Sort(){

   int i = prim_list.size();
   prim_sort_list.Reserve(i);
   while(i--)
      prim_sort_list.Add(i, prim_list[i].sort_value);
   prim_sort_list.Sort();
}

//----------------------------

void S_preprocess_context::Reset(){

   opaque = 0;
   ckey_alpha = 0;
   alpha_zwrite = 0;
   alpha_nozwrite = 0;
   alpha_noz = 0;
   prim_list.clear();
   prim_sort_list.Clear();
   shadow_casters.clear();
   shadow_receivers.clear();

   LOD_factor = scene->GetActiveCamera1()->GetFOV1() / (65.0f * PI / 180.0f);
   LOD_factor *= scene->GetDriver1()->lod_scale;
}

//----------------------------

void S_preprocess_context::SortByDistance(){
   int i = prim_list.size();
   prim_sort_list.Reserve(i);
   while(i--) prim_sort_list.Add(i, -FloatToInt(prim_list[i].vis->GetPos().Magnitude()*65536.0f));
   prim_sort_list.Sort();
}

//----------------------------
// check sphere intersection with viewing frustum
// return false if sphere is outside frustum
// set clip flag if sphere collides with any plane
//----------------------------

#ifdef _MSC_VER

/*
__declspec(naked) bool __stdcall SphereInFrustum_x86(const S_expanded_view_frustum &vf, const I3D_bsphere &bsphere, bool &clip){
   __asm{
      push ecx
      mov ecx, [esp+8+4]
      //call Prefetch1
      push ebx
      push edx
      push esi
      push edi
                              //registers:

                              //eax - free
                              //ebx - sphere_radius bitmask
                              // dl - ret
                              // dh - clip
                              //esi - vf*
                              //edi - loop counter

                              //st(0) .. st(2) - sphere_pos
                              //st(3) - -sphere_radius
      fld dword ptr [ecx+0xc]
      fchs

      mov ebx, [ecx+0xc]
      fld dword ptr [ecx+8]
      fld dword ptr [ecx+4]
      fld dword ptr [ecx+0]

      //mov esi, vf
      mov esi, [esp+4+20]
      mov edx, 0x0001
                              //get number of planes
      mov edi, [esi + (SIZE S_plane) * MAX_FRUSTUM_CLIP_PLANES]

   main_loop:

      //lea ecx, [esi+16]
      //call Prefetch1
                              //Dot
      fld dword ptr [esi + MAX_FRUSTUM_CLIP_PLANES*0]
      fmul st, st(1)
      fld dword ptr [esi + MAX_FRUSTUM_CLIP_PLANES*4]
      fmul st, st(3)
      faddp st(1), st
      fld dword ptr [esi + MAX_FRUSTUM_CLIP_PLANES*8]
      fmul st, st(4)
      faddp st(1), st

                              //add view_frustum_d
      fadd dword ptr [esi + MAX_FRUSTUM_CLIP_PLANES*12]


                              //compare against negative radius
      fcom st(4)

      add esi, 4
                              //compare against sphere_radius
      fstp dword ptr [esp+4+20]
      cmp [esp+4+20], ebx
      jge skip1

                              //compare against negative radius (cont.)
      fstsw ax
      sahf
      jb skip2
      mov dh, 1
   skip2:
                              //next side

      dec edi
      jnz main_loop
                              //simulate cmp eax, dword (2 bytes left)
      _emit 0x3d
      _emit 0
      _emit 0
   skip1:                     //test failed
      mov dl, 0               //2 bytes long

                              //remove stuff from FPU
      fstp dword ptr [esp+4+20]
      fstp dword ptr [esp+4+20]
      fstp dword ptr [esp+4+20]
      xor eax, eax            //return value in dword
      fstp dword ptr [esp+4+20]
                              //store values
      mov edi, [esp+12+20]
      mov al, dl
      mov [edi], dh

      pop edi
      pop esi
      pop edx
      pop ebx
      pop ecx

      ret 12
   }
}

//----------------------------

static __int64 sign_64 = 0x8000000080000000;

__declspec(naked) bool __stdcall SphereInFrustum_3DNow(const S_expanded_view_frustum &vf, const I3D_bsphere &bsphere, bool &clip){

   __int64 tmp0, tmp1;
   __asm{
      femms
                              //eax - free
                              //esi - vf*
                              //edi - loop counter
      mov eax, [esp+8]        //get bsphere
      prefetch [eax]

      push ebx
      push esi
      push edi

      mov esi, [esp+4+12]     //get expanded view frustum

                              //MM0 - 2X bsphere.pos.x
                              //MM1 - 2X bsphere.pos.y
                              //MM2 - 2X bsphere.pos.z
                              //MM3 - 2X bsphere.radius
                              //mm6 - clip flags
      mov ebx, [eax+0]
      mov edi, [eax+4]
      mov dword ptr[tmp0+0], ebx
      mov dword ptr[tmp0+4], ebx
      mov dword ptr[tmp1+0], edi
      mov dword ptr[tmp1+4], edi
      movq mm0, tmp0
      movq mm1, tmp1

      mov ebx, [eax+8]
      mov edi, [eax+12]
      mov dword ptr[tmp0+0], ebx
      mov dword ptr[tmp0+4], ebx
      mov dword ptr[tmp1+0], edi
      mov dword ptr[tmp1+4], edi
      movq mm2, tmp0
      movq mm3, tmp1

                              //default clipping = no
      movq mm6, sign_64

                              //get number of planes
      mov edi, [esi + (SIZE S_plane) * MAX_FRUSTUM_CLIP_PLANES]
      inc edi
                              //process 2 planes simultaneously
      shr edi, 1
      jz out1
   loop1:
                              //make x
      movq mm4, [esi + MAX_FRUSTUM_CLIP_PLANES * 0]
      pfmul mm4, mm0
                              //make y
      movq mm5, [esi + MAX_FRUSTUM_CLIP_PLANES * 4]
      pfmul mm5, mm1
                              //add x + y
      pfadd mm4, mm5
                              //make z
      movq mm5, [esi + MAX_FRUSTUM_CLIP_PLANES * 8]
      pfmul mm5, mm2
                              //add (x + y) + z
      pfadd mm4, mm5
                              //add d
      pfadd mm4, [esi + MAX_FRUSTUM_CLIP_PLANES * 12]
                              //distances to 2 planes are in MM4

                              //update clip flags
      movq mm5, mm3
      pfadd mm5, mm4
      pand mm6, mm5
                              //check if out
      pfsub mm4, mm3
      pand mm4, sign_64
      //pcmpeq mm4, sign_64
      jnz false1

                              //move to next four floats
      add esi, 8
      dec edi
      jnz loop1
   out1:
                              //store clipping
      mov eax, [esp+12+12]
      cmp ebx, 0xf
      setnz [eax]
                              //Z = 0
      cmp ebx, ebx
   false1:
      setz al
   
      pop edi
      pop esi
      pop ebx

      and eax, 1

      femms

      ret 12
   }
}

//----------------------------

__declspec(naked) bool __stdcall SphereInFrustum_SSE(const S_expanded_view_frustum &vf, const I3D_bsphere &bsphere, bool &clip){

   __asm{
                                 ;eax - free
                                 ;ebx - clip flags
                                 ;esi - vf*
                                 ;edi - loop counter
      mov eax, [esp+8]           ;get bsphere
      prefetcht0 [eax]

      push ebx
      push esi
      push edi

      mov esi, [esp+4+12]        ;get expanded view frustum

                                 ;XMM0 - 4X bsphere.pos.x
                                 ;XMM1 - 4X bsphere.pos.y
                                 ;XMM2 - 4X bsphere.pos.z
                                 ;XMM3 - 4X bsphere.radius
      movups xmm0, [eax]
      prefetcht0 [esi]
      movaps xmm1, xmm0
      movaps xmm2, xmm0
      movaps xmm3, xmm0
      shufps xmm0, xmm0, 0x00
      shufps xmm1, xmm1, 0x55
      shufps xmm2, xmm2, 0xaa
      shufps xmm3, xmm3, 0xff

                                 ;default clipping = no
      mov ebx, 0xf

                                 ;get number of planes
      mov edi, [esi + (SIZE S_plane) * MAX_FRUSTUM_CLIP_PLANES]
      add edi, 3
      shr edi, 2
      jz out2
   loop2:
      prefetcht0 [esi + 16]
                                 ;make x
      movups xmm4, [esi + 0]  
      prefetcht0 [esi + 16 + MAX_FRUSTUM_CLIP_PLANES*4]
      mulps xmm4, xmm0
                                 ;make y
      prefetcht0 [esi + 16 + MAX_FRUSTUM_CLIP_PLANES*8]
      movups xmm5, [esi + MAX_FRUSTUM_CLIP_PLANES*4]
      prefetcht0 [esi + 16 + MAX_FRUSTUM_CLIP_PLANES*12]
      mulps xmm5, xmm1
                                 ;add x + y
      addps xmm4, xmm5
                                 ;make z
      movups xmm5, [esi + MAX_FRUSTUM_CLIP_PLANES*8]
      mulps xmm5, xmm2
                                 ;add (x + y) + z
      addps xmm4, xmm5
                                 ;add d
      movups xmm5, [esi + MAX_FRUSTUM_CLIP_PLANES*12]
      addps xmm4, xmm5
                                 ;distances to 4 planes are in xmm4

                                 ;update clip flags
      movaps xmm5, xmm3
      addps xmm5, xmm4
      movmskps eax, xmm5
      and ebx, eax
                                 ;check if out
      subps xmm4, xmm3
      movmskps eax, xmm4
      cmp eax, 0xf
      jnz false2

                                 ;move to next four floats
      add esi, 16             
      dec edi
      jnz loop2
   out2:
                                 ;store clipping
      mov eax, [esp+12+12]
      cmp ebx, 0xf
      setnz [eax]
                                 ;Z = 0
      cmp ebx, ebx
   false2:
      setz al
   
      pop edi
      pop esi
      pop ebx

      and eax, 1

      ret 12

   }
}
*/

//----------------------------

#pragma warning(disable:4035)
__declspec(naked) bool __stdcall SphereInVF(const S_view_frustum &vf, const I3D_bsphere &bsphere, bool &clip){

   __asm{
      push ecx
      mov ecx, [esp+8+4]
      //call Prefetch1
      push ebx
      push edx
      push esi
      push edi
                              //registers:

                              //eax - free
                              //ebx - sphere_radius bitmask
                              // dl - ret
                              // dh - clip
                              //esi - vf*
                              //edi - loop counter

                              //st(0) .. st(2) - sphere_pos
                              //st(3) - -sphere_radius
      fld dword ptr [ecx+0xc]
      fchs

      mov ebx, [ecx+0xc]
      fld dword ptr [ecx+8]
      fld dword ptr [ecx+4]
      fld dword ptr [ecx+0]

      //mov esi, vf
      mov esi, [esp+4+20]
      mov edx, 0x0001
                              //get number of planes
      mov edi, [esi + (SIZE S_plane) * MAX_FRUSTUM_CLIP_PLANES]

   main_loop:

      lea ecx, [esi+16]
      //call Prefetch1
                              //Dot
      fld dword ptr [esi+0]
      fmul st, st(1)
      fld dword ptr [esi+4]
      fmul st, st(3)
      faddp st(1), st
      fld dword ptr [esi+8]
      fmul st, st(4)
      faddp st(1), st

                              //add view_frustum_d
      fadd dword ptr [esi+12]


                              //compare against negative radius
      fcom st(4)

      add esi, 16
                              //compare against sphere_radius
      fstp dword ptr [esp+4+20]
      cmp [esp+4+20], ebx
      jge skip1

                              //compare against negative radius (cont.)
      fstsw ax
      sahf
      jb skip2
      mov dh, 1
   skip2:
                              //next side

      dec edi
      jnz main_loop
                              //simulate cmp eax, dword (2 bytes left)
      //_emit 0x3d
      //_emit 0
      //_emit 0
      jmp skip3
   skip1:                     //test failed
      mov dl, 0               //2 bytes long
   skip3:

                              //remove stuff from FPU
      fstp dword ptr [esp+4+20]
      fstp dword ptr [esp+4+20]
      fstp dword ptr [esp+4+20]
      xor eax, eax            //return value in dword
      fstp dword ptr [esp+4+20]
                              //store values
      mov edi, [esp+12+20]
      mov al, dl
      mov [edi], dh

      pop edi
      pop esi
      pop edx
      pop ebx
      pop ecx

      ret 12
   }
}
#pragma warning(default:4035)

//----------------------------
#else //_MSC_VER

bool SphereInVF(const S_view_frustum &vf, const I3D_bsphere &bsphere, bool &clip){

   clip = false;
   float neg_radius = -bsphere.radius;
   const S_plane *pp = vf.clip_plane;

   int i = vf.num_clip_planes;
   do{
      float d = pp->d + pp->normal.Dot(bsphere.pos);

      if(d >= bsphere.radius) return false;
      if(d >= neg_radius) clip = true;
   }while(++pp, --i);

   return true;
}

#endif   //!_MSC_VER

//----------------------------

bool IsAnyVertexInVF(const S_view_frustum &vf, const S_vector *verts, dword num_verts){

                              //transform input vertices

#if defined _MSC_VER && 0

   byte rtn = false;
   __asm{
      push ecx
                              //eax = current status
                              //ebx = current vertex
                              //ecx = vertex count
                              //edx = plane counter
                              //esi = current vertex
                              //edi = current plane
      push eax                //local variable
      mov esi, verts
      mov ecx, num_verts
   l:
                              //check all planes
      mov edi, vf
      mov edx, [edi + SIZE S_plane * MAX_FRUSTUM_CLIP_PLANES]
   ll1:
                              //compute dot-product
      fld dword ptr[esi + 0]
      fmul dword ptr[edi + 0]
      fld dword ptr[esi + 4]
      fmul dword ptr[edi + 4]
      faddp st(1), st
      fld dword ptr[esi + 8]
      fmul dword ptr[edi + 8]
      faddp st(1), st
      fadd dword ptr[edi + 12]
                              //compare to zero
      fstp dword ptr[esp]
      add edi, SIZE S_plane
      test byte ptr[esp+3], 0x80
      jz ok1

      mov rtn, 1
      jmp end
   ok1:
      dec edx
      jnz ll1

      add esi, SIZE S_vector
      dec ecx
      jnz l
   end:

      pop eax                 //clean stack

      pop ecx
   }
   return rtn;

#else

   for(int i=0; i<num_verts; i++){
      const S_plane *pp = vf.clip_planes;
      int j = vf.num_clip_planes;
      do{
         float d = pp->d + pp->normal.Dot(verts[i]);
         if(d >= 0.0f)
            goto next;
      }while(++pp, --j);
      return true;
   next:{}
   }

#endif
   return false;
}

//----------------------------

void I3D_scene::PreprocessSounds(int time){

   PI3D_sound *sptr = sounds.begin();
   for(int i=sounds.size(); i--; ){
      PI3D_sound snd = *sptr++;
#ifdef USE_PREFETCH
      if(i) Prefetch1(*sptr);
#endif
      snd->UpdateSound(time);
   }
}

//----------------------------

void I3D_scene::PreprocessOccluders(const S_preprocess_context &pc, PI3D_frame root,
   C_sort_list<PI3D_occluder> &collect_list, PI3D_sector curr_sector){

   const vector<PI3D_frame> &frm_list = root->children;
   const PI3D_frame *frms = frm_list.begin();
   int i = frm_list.size();         //can't be zero!
   do{
      PI3D_frame frm = *(frms++);

      switch(frm->type){
      case FRAME_SECTOR:
         {
            PI3D_sector sct = I3DCAST_SECTOR(frm);
            if(sct->IsVisible() && frm->NumChildren1()){
               PreprocessOccluders(pc, frm, collect_list, sct);
            }
         }
         continue;

      case FRAME_OCCLUDER:
         {
            if(frm->frm_flags&FRMFLAGS_CHANGE_MAT){
                              //re-compute world matrix
               frm->matrix.Make(frm->f_matrix, root->matrix);
               frm->frm_flags &= ~(FRMFLAGS_INV_MAT_VALID | FRMFLAGS_CHANGE_MAT | FRMFLAGS_BSPHERE_TRANS_VALID);
            }
                              //determine if it's visible
            PI3D_occluder occ = I3DCAST_OCCLUDER(frm);

            occ->bound.GetBoundSphereTrans(occ, FRMFLAGS_BSPHERE_TRANS_VALID);
            bool clip_vf;
            bool is_visible =
               curr_sector->IsBoundVolumeVisible(occ->bound, frm->matrix,
                  current_camera->GetCurrSector(), current_camera->matrix(3),
                  pc, &clip_vf,
                  NULL);
            if(!is_visible)
               break;

            {                 //put into temporary list, compute frustum when it's definite
               assert(frm->frm_flags&FRMFLAGS_BSPHERE_TRANS_VALID);
               const S_vector &occ_pos = occ->bound.bsphere_world.pos;
               //float dist_to_cam = (current_camera->matrix(3) - occ_pos).Magnitude();
               float dist_to_cam = (current_camera->matrix(3) - occ_pos).Square();
               occ->SetWorkSector(curr_sector);
               collect_list.Add(occ, FloatToInt(dist_to_cam * 10.0f));
            }
            if(!frm->num_hr_occl)
               continue;
         }
         break;

      case FRAME_VISUAL:
      case FRAME_MODEL:
      case FRAME_DUMMY:
         if(!frm->IsOn1())
            continue;
                              //flow...
      default: 
         {
                              //don't update matrix if no occluders are linked to this frame
            if(!frm->num_hr_occl)
               continue;
            if(frm->frm_flags&FRMFLAGS_CHANGE_MAT){
                                    //update matrix
               frm->matrix.Make(frm->f_matrix, root->matrix);
               frm->frm_flags &= ~(FRMFLAGS_INV_MAT_VALID | FRMFLAGS_CHANGE_MAT | FRMFLAGS_BSPHERE_TRANS_VALID);
            }
         }
      }
      if(frm->NumChildren1())
         PreprocessOccluders(pc, frm, collect_list, curr_sector);
   } while(--i);
}

//----------------------------

#define SAVE_LIGHT_MARK if(prep_flags&E_PREP_SET_HR_LIGHT_DIRTY) frm->frm_flags |= FRMFLAGS_HR_LIGHT_RESET;
#define PROPAGATE_LIGHT_MARK if(frm->frm_flags&FRMFLAGS_HR_LIGHT_RESET){\
   new_prep_flags |= E_PREP_SET_HR_LIGHT_DIRTY;\
   frm->frm_flags &= ~FRMFLAGS_HR_LIGHT_RESET; }

void I3D_scene::PreprocessChildren(PI3D_frame root, S_preprocess_context &pc, dword prep_flags){

   const vector<PI3D_frame> &frm_list = root->children;
   const PI3D_frame *frms = frm_list.begin();

   int num_hr_vis = root->num_hr_vis;
   assert(num_hr_vis);

   do{
      dword new_prep_flags = prep_flags;
      PI3D_frame frm = *frms++;
                              //stop as soon as we exhaust frames containing visuals
      num_hr_vis -= frm->num_hr_vis;

#ifdef _DEBUG
      //++drv->scene_stats.num_vis_computed;
#endif
                              //make proper frame matrix
      bool change_mat_loc = (frm->frm_flags&FRMFLAGS_CHANGE_MAT);

      if(change_mat_loc){
         frm->matrix.Make(frm->f_matrix, root->matrix);
         frm->frm_flags &= ~(FRMFLAGS_INV_MAT_VALID | FRMFLAGS_CHANGE_MAT | FRMFLAGS_BSPHERE_TRANS_VALID);
      }

                              //propagate light-reset on this frame
      switch(frm->type){

      case FRAME_MODEL:
         {
            if(!(frm->frm_flags&FRMFLAGS_ON) || !frm->num_hr_vis || (prep_flags&E_PREP_NO_VISUALS)){
               SAVE_LIGHT_MARK;
               break;
            }

            bool is_caster = (frm->frm_flags&I3D_FRMF_SHADOW_CAST) &&
               (prep_flags&E_PREP_ADD_SHD_CASTERS);

            if(frm->frm_flags&FRMFLAGS_ON_CONDITIONAL){
               if(pc.mode==RV_NORMAL && !is_caster){
                  SAVE_LIGHT_MARK;
                  break;
               }
            }

            PI3D_model mod = I3DCAST_MODEL(frm);

            bool do_hr_test = (frm->num_hr_vis >= 2 || is_caster);
            
                              //visibility tests
            if(do_hr_test){
               bool off_sector = (pc.curr_sector!=current_camera->GetCurrSector());
               if(off_sector || !(new_prep_flags&E_PREP_HR_NO_CLIP)){
                  if(!(frm->frm_flags&FRMFLAGS_HR_BOUND_VALID))
                     mod->hr_bound.MakeHierarchyBounds(frm);
                  mod->hr_bound.GetBoundSphereTrans(frm, FRMFLAGS_HR_BSPHERE_VALID);

                  bool clip_vf, clip_occ;
                  bool is_visible =
                     pc.curr_sector->IsBoundVolumeVisible(mod->hr_bound, frm->matrix,
                        current_camera->GetCurrSector(), current_camera->matrix(3),
                        pc,
                        !(new_prep_flags&E_PREP_HR_NO_CLIP) ? &clip_vf : NULL,
                        (new_prep_flags&E_PREP_TEST_OCCLUDERS) ? &clip_occ : NULL);
                  if(!is_visible){
                     if(is_caster){
                              //shadow casters may cast shadow even if their bounding box is out of view
                              // brief check - expanded bounding sphere, enclosing any possible shadow from the model
                        I3D_bsphere bsphere;
                        bsphere.pos = frm->matrix(3);
                        bsphere.radius = Max(mod->hr_bound.bsphere_world.radius, SHADOW_LENGTH);

                        bool shd_visible = pc.curr_sector->IsBoundSphereVisible(bsphere,
                           current_camera->GetCurrSector(), current_camera->matrix(3),
                           pc, &clip_vf, (new_prep_flags&E_PREP_TEST_OCCLUDERS) ? &clip_occ : NULL);
                        if(shd_visible){
                           mod->frm_flags |= FRMFLAGS_SHD_CASTER_NONVIS;
                           pc.shadow_casters.push_back(mod);
                           new_prep_flags &= ~E_PREP_ADD_SHD_CASTERS;
                        }
                     }
                     SAVE_LIGHT_MARK;
                     break;
                  }
                  if(!(new_prep_flags&E_PREP_HR_NO_CLIP) && !clip_vf)
                     new_prep_flags |= E_PREP_HR_NO_CLIP;
                  if((new_prep_flags&E_PREP_TEST_OCCLUDERS) && !clip_occ)
                     new_prep_flags &= ~E_PREP_TEST_OCCLUDERS;
               }
            }

            if(is_caster){
               mod->frm_flags &= ~FRMFLAGS_SHD_CASTER_NONVIS;
               pc.shadow_casters.push_back(mod);
               new_prep_flags &= ~E_PREP_ADD_SHD_CASTERS;
               if(pc.mode==RV_NORMAL && (frm->frm_flags&FRMFLAGS_ON_CONDITIONAL)){
                  SAVE_LIGHT_MARK;
                  break;
               }
            }
            PROPAGATE_LIGHT_MARK;
            PreprocessChildren(frm, pc, new_prep_flags);
         }
         break;

      case FRAME_VISUAL:
         {
            --num_hr_vis;
            PI3D_visual vis = I3DCAST_VISUAL(frm);
                              //if matrix changed, schedule to recompute lights
            if(change_mat_loc) 
               vis->vis_flags &= ~VISF_DEST_LIGHT_VALID;

            if(!(frm->frm_flags&FRMFLAGS_ON) || (prep_flags&E_PREP_NO_VISUALS)){
               SAVE_LIGHT_MARK;
               break;
            }
                              //visibility tests

                              //hide distance test
            if(FloatAsInt(vis->hide_dist)!=0){
               float d = (current_camera->matrix(3) - frm->matrix(3)).Magnitude();
               d *= pc.LOD_factor;
               if(d > vis->hide_dist){
                  SAVE_LIGHT_MARK;
                  break;
               }
            }
            
                              //hierarchy bounding test
            if(frm->num_hr_vis >= 2){
               if(!(new_prep_flags&E_PREP_HR_NO_CLIP) ||
                  (new_prep_flags&E_PREP_TEST_OCCLUDERS) ||
                  pc.curr_sector!=current_camera->GetCurrSector()){

                  if(!(frm->frm_flags&FRMFLAGS_HR_BOUND_VALID))
                     vis->hr_bound.MakeHierarchyBounds(frm);

                  vis->hr_bound.GetBoundSphereTrans(frm, FRMFLAGS_HR_BSPHERE_VALID);
                  bool clip_vf, clip_occ;
                  bool is_visible =
                     pc.curr_sector->IsBoundVolumeVisible(vis->hr_bound, frm->matrix,
                        current_camera->GetCurrSector(), current_camera->matrix(3),
                        pc, !(new_prep_flags&E_PREP_HR_NO_CLIP) ? &clip_vf : NULL,
                        (new_prep_flags&E_PREP_TEST_OCCLUDERS) ? &clip_occ : NULL);
                  if(!is_visible){
                     SAVE_LIGHT_MARK;
                     break;
                  }
                  if(!(new_prep_flags&E_PREP_HR_NO_CLIP) && !clip_vf)
                     new_prep_flags |= E_PREP_HR_NO_CLIP;
                  if((new_prep_flags&E_PREP_TEST_OCCLUDERS) && !clip_occ)
                     new_prep_flags &= ~E_PREP_TEST_OCCLUDERS;
               }
            }
                              //propagate light reset and set to visual
            if((frm->frm_flags&FRMFLAGS_HR_LIGHT_RESET) || (prep_flags&E_PREP_SET_HR_LIGHT_DIRTY)){
               new_prep_flags |= E_PREP_SET_HR_LIGHT_DIRTY;\
               frm->frm_flags &= ~FRMFLAGS_HR_LIGHT_RESET;
               vis->vis_flags &= ~VISF_DEST_LIGHT_VALID;
            }

            if(!(vis->vis_flags&VISF_BOUNDS_VALID))
               vis->ComputeBounds();
            vis->bound.GetBoundSphereTrans(frm, FRMFLAGS_BSPHERE_TRANS_VALID);

            bool clip_vf, clip_occ, *clipp = &clip_vf;
            if(new_prep_flags&E_PREP_HR_NO_CLIP){
               clip_vf = false;
               clipp = NULL;
            }
            bool is_visible =
               pc.curr_sector->IsBoundVolumeVisible(vis->bound, frm->matrix,
                  current_camera->GetCurrSector(), current_camera->matrix(3),
                  pc, clipp,
                  (new_prep_flags&E_PREP_TEST_OCCLUDERS) ? &clip_occ : NULL);
            if(!is_visible)
               goto not_visible;

                              //visual is visible
#ifdef DEBUG_FORCE_CLIP
            clip_vf = true;
#endif
            //++debug_count[clip_vf];
            pc.clip = clip_vf;
            vis->AddPrimitives(pc);
            
                              //stats
            ++drv->scene_stats.frame_count[FRAME_VISUAL];
            drv->scene_stats.num_vis_computed += (int)clip_vf;

            if(pc.mode!=RV_SHADOW_CASTER){
               vis->last_render_time = last_render_time;
               if(frm->frm_flags&I3D_FRMF_SHADOW_RECEIVE)
                  pc.shadow_receivers.push_back(vis);
            }
not_visible:
            if(frm->num_hr_vis)
               PreprocessChildren(frm, pc, new_prep_flags);
         }
         break;

      case FRAME_SECTOR:
         {
            PI3D_sector sct = I3DCAST_SECTOR(frm);
                              //if sector is not visible, children are also not
            if(!sct->IsVisible())
               break;
            if(sct->IsSimple()){
                              //it's visible if parent sector is visible
               if(prep_flags&E_PREP_NO_VISUALS)
                  break;
            }

            ++drv->scene_stats.frame_count[FRAME_SECTOR];
                              //save sector & global light settings
            PI3D_sector save_sct = pc.curr_sector;
                              //set our global settings
            pc.curr_sector = sct;
            new_prep_flags &= (E_PREP_ADD_SHD_CASTERS | E_PREP_TEST_OCCLUDERS);

                              //update all lights
            if(sct->frm_flags&FRMFLAGS_HR_LIGHT_RESET){
               sct->UpdateAllLights();
               new_prep_flags |= E_PREP_SET_HR_LIGHT_DIRTY;
            }

            assert(frm->num_hr_vis);
            PreprocessChildren(frm, pc, new_prep_flags);
                              //restore global settings
            pc.curr_sector = save_sct;
         }
         break;

      case FRAME_DUMMY:
         if(!(frm->frm_flags&FRMFLAGS_ON)){
            SAVE_LIGHT_MARK;
            break;
         }
                              //flow...
      default:
         assert(frm->num_hr_vis);
         PROPAGATE_LIGHT_MARK;
         PreprocessChildren(frm, pc, new_prep_flags);
      }
      assert(num_hr_vis >= 0);
   } while(num_hr_vis);
}

//----------------------------

void I3D_scene::Preprocess(S_preprocess_context &pc, dword prep_flags){

   assert(!pc.occ_list.size());

   if(!primary_sector->NumChildren1())
      return;

   if(drv->GetFlags2()&DRVF2_USE_OCCLUSION)
   if(primary_sector->num_hr_occl){
                              //collect all visible occluders
      C_sort_list<PI3D_occluder> occ_list1(primary_sector->num_hr_occl);
      pc.curr_sector = primary_sector;
      PreprocessOccluders(pc, primary_sector, occ_list1, primary_sector);
      pc.curr_sector = NULL;
                              //sort them by distance to camera
      occ_list1.Sort();
      pc.occ_list.reserve(occ_list1.Count());
                              //process sorted occluders back-to-front, try to hide occluded ones
      for(int i=0; i<occ_list1.Count(); i++)    //this loop must go upwards
      {
         PI3D_occluder occ1 = occ_list1[i];

         bool occluded = false;
         for(int j=0; j<pc.occ_list.size(); j++){
            PI3D_occluder occ = pc.occ_list[j];

                              //test if occ1 is occluded by occ
            bool clip;

            occ1->bound.GetBoundSphereTrans(occ1, FRMFLAGS_BSPHERE_TRANS_VALID);
            occluded = occ->IsOccluding(occ1->bound, occ1->matrix, clip);
            if(occluded)
               break;
         }
         if(!occluded){
                              //compute frustum now
            bool b;
                              //we could make ComputeClipPlanes virtual method
                              // of I3D_occluder_base, but is would make mismatch
                              // with user's include files, as they omit I3D_occluder_base
            b = occ1->ComputeOccludingFrustum(this);
            if(b){
                              //definitely add to list
                              //set debug flags
               occ1->occ_flags |= OCCF_DEBUG_OCCLUDING;
               ++drv->scene_stats.frame_count[FRAME_OCCLUDER];
               pc.occ_list.push_back(occ1);
            }
         }else{
                              //invisible occluder, if it's sector, set it invisible
            if(occ1->GetType1()==FRAME_SECTOR){
               PI3D_sector sct = I3DCAST_SECTOR(occ1);
               sct->SetSctFlags(sct->GetSctFlags() & ~SCTF_VISIBLE);
            }
         }
      }
      if(pc.occ_list.size()){
                              //limit number of occluders
         if(pc.occ_list.size() > MAX_COMPUTED_OCCLUDERS)
            pc.occ_list.erase(pc.occ_list.begin()+MAX_COMPUTED_OCCLUDERS, pc.occ_list.end());

                              //hide portals behind occluders
         //current_camera->GetCurrSector()->SolveOcclusion(pc.occ_list);
      }
   }
   pc.curr_sector = primary_sector;

   if(pc.occ_list.size())
      prep_flags |= E_PREP_TEST_OCCLUDERS;
   if(pc.curr_sector->num_hr_vis){
      PreprocessChildren(pc.curr_sector, pc, prep_flags);
   }
   //PreprocessSectors(pc.curr_sector, pc);
   assert(pc.prim_list.size() == (pc.opaque + pc.ckey_alpha +
      pc.alpha_zwrite + pc.alpha_nozwrite + pc.alpha_noz));
}

//----------------------------

void I3D_scene::DrawList(const S_preprocess_context &pc){

   int i = 0;
   int j;

                              // --opaque--
   for(j=pc.opaque; i<j; ++i){
      const S_render_primitive &ct = pc.prim_list[pc.prim_sort_list[i]];
                              //setup fog
      PI3D_light lf = ct.sector->GetFogLight();
      if(lf) drv->SetFogColor(lf->dw_color);
      ct.vis->DrawPrimitive(pc, ct);
   }
                              
                              // --color-keyed--
   if(pc.ckey_alpha){
      drv->SetupAlphaTest(true);
      i = pc.ckey_alpha;
      do{
         --i;
         const S_render_primitive &ct = pc.prim_list[pc.prim_sort_list[j+i]];
                              //setup fog
         PI3D_light lf = ct.sector->GetFogLight();
         if(lf) drv->SetFogColor(lf->dw_color);
         ct.vis->DrawPrimitive(pc, ct);
      }while(i);
      drv->SetupAlphaTest(false);
      j += pc.ckey_alpha;
   }
                         
                              // --alpha with z-writes--
   if(pc.alpha_zwrite){
      i = pc.alpha_zwrite;
      do{
         --i;
         const S_render_primitive &ct = pc.prim_list[pc.prim_sort_list[j+i]];
                              //setup fog
         PI3D_light lf = ct.sector->GetFogLight();
                              //avoid blending ADD mode with fog
         if(lf){
            bool add_mode = ((ct.blend_mode&0xffff) == I3DBLEND_ADD);
            drv->SetFogColor(add_mode ? 0 : lf->dw_color);
         }
         ct.vis->DrawPrimitive(pc, ct);
      }while(i);
      j += pc.alpha_zwrite;
   }

   if((drv->GetFlags()&DRVF_USESHADOWS) &&
      pc.shadow_casters.size() && pc.shadow_receivers.size() &&
      (drv->GetFlags()&DRVF_CANRENDERSHADOWS)){

      RenderShadows(pc);
   }

                              // --alpha without z-writes--
   if(pc.alpha_nozwrite){
      drv->EnableZWrite(false);
      i = pc.alpha_nozwrite;
      do{
         --i;
         const S_render_primitive &ct = pc.prim_list[pc.prim_sort_list[j+i]];
                              //setup fog
         PI3D_light lf = ct.sector->GetFogLight();
                              //avoid blending ADD mode with fog
         if(lf){
            bool add_mode = ((ct.blend_mode&0xffff) == I3DBLEND_ADD);
            drv->SetFogColor(add_mode ? 0 : lf->dw_color);
         }
         ct.vis->DrawPrimitive(pc, ct);
      }while(i);
      j += pc.alpha_nozwrite;
      drv->EnableZWrite(true);
   }

                              // --alpha with no z-buffer--
   if(pc.alpha_noz){
      bool use_zb = (drv->GetFlags()&DRVF_USEZB);
      if(use_zb)
         drv->SetState(RS_USEZB, false);
      i = pc.alpha_noz;
      do{
         --i;
         const S_render_primitive &ct = pc.prim_list[pc.prim_sort_list[j+i]];
                              //setup fog
         PI3D_light lf = ct.sector->GetFogLight();
                              //avoid blending ADD mode with fog
         if(lf){
            bool add_mode = ((ct.blend_mode&0xffff) == I3DBLEND_ADD);
            drv->SetFogColor(add_mode ? 0 : lf->dw_color);
         }
         ct.vis->DrawPrimitive(pc, ct);
      }while(i);
      j += pc.alpha_noz;
      if(use_zb)
         drv->SetState(RS_USEZB, true);
   }

   drv->EnableFog(false);
   drv->DisableTextureStage(1);
}

//----------------------------

PI3D_light I3D_scene::SelectShadowLight(PI3D_sector sct, const S_vector &visual_pos,
   PI3D_frame last_light, float dist_to_cam, float &opacity) const{

   PI3D_light lp = NULL;
   float lp_power;
   float amb_power = 0.0f;

   const vector<PI3D_light> &sct_lights = sct->GetLights_shadow();
   const PI3D_light *lptr = sct_lights.begin();
   for(int li=sct_lights.size(); li--; ){
      PI3D_light lp1 = *lptr++;
      if(!lp1->IsOn1())
         continue;
#ifdef USE_PREFETCH
      if(li){
         PI3D_light next = *lptr;
         Prefetch2ptr(next, &next->ltype);
      }
#endif

      const S_vector &c = lp1->color;

      if(lp1->GetLightType1() == I3DLIGHT_AMBIENT){
         float power = (c[0] * .3f + c[1] * .6f + c[2] * .1f) * lp1->power;
         amb_power += power;
         continue;
      }

      {
                        //compute this light's power
         float power = (c[0] * .3f + c[1] * .6f + c[2] * .1f) * lp1->power;

         switch(lp1->GetLightType1()){
         case I3DLIGHT_POINT:
            {
               const S_matrix &m_light = lp1->GetMatrixDirect();

               const float nr = lp1->range_n_scaled;
               const float fr = lp1->range_f_scaled;
               float ldist = (m_light(3) - visual_pos).Magnitude();
                        //consider attenuation
               if(ldist >= fr)
                  continue;
               if(ldist >= nr)
                  power *= (fr - ldist) / (fr - nr);
            }
            break;
         case I3DLIGHT_DIRECTIONAL:
            power *= .2f;
            break;
         default:
            continue;
         }
                        //give current light greater priority
         if(lp1==last_light)
            power *= 1.2f;

         if(lp){
                           //check if it's stronger than light we already have
            if(lp_power >= power)
               continue;
         }
         lp = lp1;
         lp_power = power;
      }
   }
   if(!lp)
      return NULL;

                              //apply fog intensity
   const vector<PI3D_light> &sct_fogs = sct->GetFogs();
   lptr = sct_fogs.begin();
   for(li=sct_fogs.size(); li--; ){
      PI3D_light lp1 = *lptr++;
      if(!lp1->IsOn1())
         continue;
      if(lp1->GetLightType1() == I3DLIGHT_FOG){
                        //compute fog on per-object basis
         float fog_range_f = lp1->range_f;
         if(IsAbsMrgZero(fog_range_f))
            fog_range_f = current_camera->GetFCP();
         float fog_range_n = lp1->range_n * fog_range_f;
         if(dist_to_cam >= fog_range_f)
            break;
         if(dist_to_cam >= fog_range_n)
            opacity *= (fog_range_f - dist_to_cam) / (fog_range_f - fog_range_n);
      }
   }
   if(li!=-1)
      return NULL;

                        //apply ambient - consider its relative ratio to shadow light
   float f = lp_power + amb_power * .333f;
   if((float)fabs(f) > .001f){
      float amb_ratio = lp_power / f;
      opacity *= amb_ratio;
   }
   if(IsAbsMrgZero(opacity))
      return NULL;
   return lp;
}

//----------------------------

void I3D_scene::RenderShadows(const S_preprocess_context &pc){

   assert(drv->tp_shadow);
   shd_camera->Duplicate(current_camera);

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();

   int debug_show_pos = 0;
   drv->EnableFog(false);

   C_smart_ptr<IDirect3DSurface8> save_rt, save_zs;
   drv->SaveRenderTarget(save_rt, save_zs);

   for(int i=pc.shadow_casters.size(); i--; ){
      PI3D_model mod_sc = pc.shadow_casters[i];

      const S_matrix &tm = mod_sc->GetMatrix();
      if(!(mod_sc->frm_flags&FRMFLAGS_HR_BOUND_VALID))
         mod_sc->hr_bound.MakeHierarchyBounds(mod_sc);
      const I3D_bbox &bb = mod_sc->hr_bound.bound_local.bbox;
      if(!bb.IsValid())
         continue;

      float opacity = 1.0f;
      S_vector bb_center = ((bb.max + bb.min) * .5f) * tm;
                              //get distance to camera
      float dist_to_cam = (bb_center - pc.viewer_pos).Magnitude();
                              //beyond the max range?
      if(dist_to_cam >= drv->shadow_range_f)
         continue;
                              //fade shadow depending on distance
      if(dist_to_cam >= drv->shadow_range_n)
         opacity *= (drv->shadow_range_f - dist_to_cam) / (drv->shadow_range_f - drv->shadow_range_n);

      assert(mod_sc->shadow_info);
      S_shadow_info &si = *mod_sc->shadow_info;

                              //find its sector
      PI3D_frame sct_prnt = mod_sc;
      while((sct_prnt=sct_prnt->GetParent()) && sct_prnt->GetType1()!=FRAME_SECTOR);
      assert(sct_prnt);
      PI3D_sector sct = I3DCAST_SECTOR(sct_prnt);

                              //select light
      /*
      PI3D_light lp = si.last_light;
      {
         dword last_select_delta = last_render_time - si.shadow_select_count;
         if(last_select_delta > 400){
            si.shadow_select_count = last_render_time;
            lp = SelectShadowLight(sct, bb_center, lp, dist_to_cam, opacity);
         }
      }
      */
      PI3D_light lp = SelectShadowLight(sct, bb_center, lp, dist_to_cam, opacity);
                              //no such light?
      if(!lp){
         si.Reset();
         continue;
      }
      lp->UpdateLight();

      S_normal ldir;
      switch(lp->GetLightType1()){
      case I3DLIGHT_DIRECTIONAL: ldir = lp->GetWorldDir(); break;
      default: ldir = bb_center - lp->GetMatrixDirect()(3);
      }

                              //check if we're going to make transition
      if(si.last_light!=lp){
         if(si.last_light){
                              //setup dir vector to what it was last frame
            si.trans_base_dir = si.last_dir;
            //si.trans_base_length = si.last_length;
                              //setup starting of transition
            si.transition_count = last_render_time;
         }
         si.last_light = lp;
      }

      if(si.transition_count){
         const int transition_count = 1000;
                              //make smooth transition
         int render_time_delta = last_render_time - si.transition_count;
         if(render_time_delta >= transition_count){
                              //end of transition
            si.transition_count = 0;
         }else{
            float f = (float)render_time_delta / (float)transition_count;
            float fi = 1.0f - f;
            ldir = ldir*f + si.trans_base_dir*fi;
            //shadow_length = shadow_length*f + si.trans_base_length*fi;
         }
      }

                              //save current values
      si.last_dir = ldir;
      //si.last_length = shadow_length;

      int i;
                              //get bbox contour
      S_vector contour_points[6];
      dword num_cpts;
      ComputeContour(bb, tm, ldir, contour_points, num_cpts, true);
      assert(num_cpts);
                              //find matrix for smallest rectangle
         
                              //get projection plane, going through object's center
      S_plane pl;
      pl.normal = ldir;
      pl.d = -bb_center.Dot(pl.normal);

                              //find bbox point fahrtest from the plane
      S_vector bb_full[8], bb_trans[8];
      bb.Expand(bb_full);
      TransformVertexArray(bb_full, sizeof(S_vector), 8, bb_trans, sizeof(S_vector), mod_sc->GetMatrix());
      float d_bb_min = 1e+16f;
      for(i=8; i--; ){
         const S_vector &v = bb_trans[i];
         float d = v.DistanceToPlane(pl);
         if(d_bb_min > d){
            d_bb_min = d;
         }
      }
      d_bb_min = pl.d - d_bb_min;
                              //construct contour frustum
      S_view_frustum vf;
      vf.num_clip_planes = 0;
      for(i=num_cpts; i--; ){
         S_plane &cpl = vf.clip_planes[vf.num_clip_planes];
         cpl.normal.GetNormal(contour_points[i], contour_points[(i+1)%num_cpts],
            contour_points[i] + ldir);
         if(cpl.normal.Dot(cpl.normal) < .01f)
            continue;
         cpl.normal.Normalize();
         cpl.d = -(contour_points[i].Dot(cpl.normal));
         ++vf.num_clip_planes;
      }
                              //add front and back clipping planes
      vf.clip_planes[vf.num_clip_planes].normal = -pl.normal;
      vf.clip_planes[vf.num_clip_planes].d = -d_bb_min;
      ++vf.num_clip_planes;
      vf.clip_planes[vf.num_clip_planes] = pl;
      vf.clip_planes[vf.num_clip_planes].d -= SHADOW_LENGTH;
      ++vf.num_clip_planes;

      vector<PI3D_visual> hit_receivers;
      hit_receivers.reserve(pc.shadow_receivers.size());

                              //determine which receivers are really hit
      {
         const PI3D_visual *recs = pc.shadow_receivers.begin();
         for(i=pc.shadow_receivers.size(); i--; ){
            PI3D_visual vis = *recs++;
            /*
            if(i){
               PI3D_visual next = *recs;
               Prefetch2ptr(&next->vis_flags, &next->frm_flags);
            }
            */
            if(!(vis->vis_flags&VISF_BOUNDS_VALID))
               vis->ComputeBounds();
            const I3D_bsphere &bs = vis->bound.GetBoundSphereTrans(vis, FRMFLAGS_BSPHERE_TRANS_VALID);
            bool clip;
            bool in = SphereInVF(vf, bs, clip);
            if(!in)
               continue;
            if(clip){
                                 //detailed bbox test
               S_vector vis_countours[6];
               dword num_vis_cpts;
               ComputeContour(vis->bound.bound_local.bbox, vis->GetMatrix(), ldir,
                  vis_countours, num_vis_cpts, true);
               assert(num_vis_cpts);
                                 //check only with contour clip planes of vf
               in = CheckFrustumIntersection(vf.clip_planes, vf.num_clip_planes - 2,
                  vis_countours, num_vis_cpts);
               if(!in)
                  continue;
            }
            hit_receivers.push_back(vis);
         }
      }
                           //if there're no objects hit, skip this caster
      if(!hit_receivers.size())
         continue;

                              //project bounding box onto the plane
      S_vector projected_contour[6];
      for(i=num_cpts; i--; )
         pl.Intersection(contour_points[i], ldir, projected_contour[i]);

                           //find smallest rectangle
      float best_size = 1e+16f;
      I3D_bbox best_bbox;
      S_matrix best_mat;

                              //proposed transformation matrix
      S_matrix prop_matrix;
      prop_matrix.Identity();
      prop_matrix(2) = pl.normal;

      for(int i0=num_cpts; i0--; ){
         const S_vector &v0 = projected_contour[i0];
         for(int i1=i0; i1--; ){
            const S_vector &v1 = projected_contour[i1];
            S_vector dir_v0_v1 = v1 - v0;
                           //if verts are too close (identical), skip this vertex
            if(IsMrgZero(dir_v0_v1.Dot(dir_v0_v1)))
               continue;
                           //compute proposed matrix
            prop_matrix(0) = S_normal(dir_v0_v1);
            prop_matrix(1) = prop_matrix(2).Cross(prop_matrix(0));
      
            S_matrix prop_matrix_inv = ~prop_matrix;
                           //process all points, build bounds (projected to 2D)
            I3D_bbox rc_bbox;
            rc_bbox.Invalidate();

            S_vector v_trans[6];
            TransformVertexArray(contour_points, sizeof(S_vector), num_cpts,
               v_trans, sizeof(S_vector), prop_matrix_inv);

            for(int l=num_cpts; l--; ){
               const S_vector &vt = v_trans[l];
               rc_bbox.min.Minimal(vt);
               rc_bbox.max.Maximal(vt);
            }
                           //save result, if better than previous
            float curr_size = (rc_bbox.max.x-rc_bbox.min.x) * (rc_bbox.max.y-rc_bbox.min.y);
            if(best_size > curr_size){
               best_size = curr_size;
               best_bbox = rc_bbox;
               best_mat = prop_matrix;
            }
         }
      }
      float mod_scale = mod_sc->GetMatrix()(0).Magnitude();
      float bb_diag = (bb.max - bb.min).Magnitude() * mod_scale;
      shd_camera->SetRange(0.0f, bb_diag);
      shd_camera->SetOrthogonal(true);

                              //compute scale
      S_vector bb_size = best_bbox.max - best_bbox.min;
      float fside = Max(bb_size.x, bb_size.y); 
      fside *= 1.0f + 3.0f / drv->tp_shadow[1]->SizeX1();
      float ortho_scale = 2.0f / fside;
      shd_camera->SetOrthoScale(ortho_scale);

                              //re-compute texture matrix
                              // note: Y is swapped due to texture adressing from upper-left corner
      S_matrix m_txt; m_txt.Identity();
      m_txt(0) = best_mat(0) * fside;
      m_txt(1) = best_mat(1) * (-fside);
      m_txt(2) = pl.normal;
                              //position just in front of bbox from light's view dir
      m_txt(3) = bb_center + ldir * (pl.d - d_bb_min);


                              //create shadow affect cylinder
      I3D_cylinder vc;
      vc.pos = m_txt(3);
      vc.dir = m_txt(2) * SHADOW_LENGTH;
      vc.radius = 0.0f;
      for(i=num_cpts; i--; )
         vc.radius = Max(vc.radius, contour_points[i].DistanceToLine(vc.pos, vc.dir));

      if(mod_sc->frm_flags&FRMFLAGS_SHD_CASTER_NONVIS){
                              //caster wasn't rendered, do detailed check if shadow will really be visible
         I3D_bsphere bsphere;
         bsphere.pos = vc.pos + vc.dir*.5f;
         float d = vc.dir.Magnitude()*.5f;
         float r = vc.radius;
         bsphere.radius = SqrtFast(d*d + r*r);

         PI3D_frame sct = mod_sc;
         while((sct=sct->GetParent1()) && sct->GetType1()!=FRAME_SECTOR);
         assert(sct);

         bool clip_vf;
         bool shd_visible = I3DCAST_SECTOR(sct)->IsBoundSphereVisible(bsphere,
            current_camera->GetCurrSector(), current_camera->matrix(3),
            pc, &clip_vf, NULL);
         if(!shd_visible)
            continue;
      }
      ++drv->scene_stats.dyn_shd_casters;


      S_matrix m_txt_inv = ~m_txt;
                              //add half-texel to correct rounding errors
      float half_texel = .5f / (float)drv->tp_shadow[1]->SizeX1();

      m_txt_inv(3, 0) += .5f + half_texel;
      m_txt_inv(3, 1) += .5f + half_texel;
                              //Z's origin close to bbox center, but slightly towards the light,
                              // so that shadow is not cut too much, but is cut on objects in light direction
      m_txt_inv(3, 2) -= bb_diag * .35f;
      const float r_sl = 1.0f / SHADOW_LENGTH;
      for(i=4; i--; ){
         m_txt_inv(i, 2) *= r_sl;
      }
#if 0
      {
                              //debug!
         for(i=0; i<num_cpts; i++){
            drv->DEBUG(projected_contour[i] * m_txt_inv);
            drv->DEBUG(contour_points[i] * m_txt_inv);
         }
      }
#endif

      if(drv->GetFlags()&DRVF_DEBUGDRAWSHADOWS){
         SetRenderMatrix(GetIdentityMatrix());
                              //render bbox contour
         DebugDrawContour(contour_points, num_cpts, S_vector(1, 0, 0));

         S_vector v[4] = { m_txt(3) - m_txt(0)*.5f - m_txt(1)*.5f, v[0] + m_txt(0), v[1] + m_txt(1), v[0] + m_txt(1) };
         DebugDrawContour(v, 4, S_vector(0, 1, 0));
                              //volume
         for(int i=num_cpts; i--; ){
            S_vector v = projected_contour[i] + ldir * (pl.d - d_bb_min);
            DrawLine(v, v + ldir * (SHADOW_LENGTH + (d_bb_min - pl.d)), S_vector(0, 1, 1), 0x80);
         }
                              //axes
         DrawLine(v[2] - m_txt(1)*.5f, v[2] - m_txt(1)*.5f + m_txt(0)*.2f, S_vector(1, 0, 0), 0);
         DrawLine(v[2] - m_txt(0)*.5f, v[2] - m_txt(0)*.5f + m_txt(1)*.2f, S_vector(0, 1, 0), 0);
         DrawLine((v[2] + v[0]) * .5f, (v[2] + v[0]) * .5f + m_txt(2)*.2f, S_vector(0, 0, 1), 0);
      }
                              //prepare view and preprocess shadow caster
      shd_camera->SetPos(m_txt(3));
      shd_camera->SetDir1(m_txt(2), -m_txt(1));
      shd_camera->Update();
      shd_camera->UpdateMatrices(1.0f);
      shd_camera->SetCurrSector(sct);
      SetMViewProj(shd_camera->GetViewMatrix1(), shd_camera->GetProjectionMatrix1());
      HRESULT hr;
                              //update matrices
      hr = lpDev8->SetTransform(D3DTS_PROJECTION, (const D3DMATRIX*)&shd_camera->GetProjectionMatrix1());
      CHECK_D3D_RESULT("SetTransform", hr);

      C_smart_ptr<I3D_camera> save_curr_cam = current_camera;
      SetActiveCamera(shd_camera);

      bool was_zb = (drv->GetFlags()&DRVF_USEZB);
      if(was_zb) drv->SetState(RS_USEZB, false);
      drv->tp_shadow[0]->Manage(I3D_texture::MANAGE_CREATE_ONLY);

                              //switch render target
      IDirect3DSurface8 *rt;
      hr = drv->tp_shadow[0]->GetD3DTexture()->GetSurfaceLevel(0, &rt);
      CHECK_D3D_RESULT("GetSurfaceLevel", hr);
      drv->SetRenderTarget(rt, drv->zb_shadow);
      rt->Release();

      drv->UpdateViewport(0, 0, drv->tp_shadow[0]->SizeX1(), drv->tp_shadow[0]->SizeY1());
      hr = lpDev8->Clear(0, NULL, D3DCLEAR_TARGET, 0, 1.0f, 0);
      CHECK_D3D_RESULT("Clear", hr);
      drv->UpdateViewport(1, 1, drv->tp_shadow[0]->SizeX1() - 1, drv->tp_shadow[0]->SizeY1() - 1);

      S_preprocess_context pc_shd(this, drv->GetRenderviewID(), pc.viewer_pos);
      pc_shd.render_context = pc.render_context;
      pc_shd.render_flags = pc.render_flags;
      pc_shd.curr_sector = sct;
      pc_shd.LOD_factor = pc.LOD_factor * 2.0f;
      pc_shd.mode = RV_SHADOW_CASTER;
      PrepareViewFrustum(pc_shd);
      PreprocessChildren(mod_sc, pc_shd, 0);

                              //render into the texture
      pc_shd.Sort();
#if 0
      {                       //render shadow viewport in shadow texture
         SetRenderMatrix(I3D_identity_matrix);

         S_vector v[4] = { m_txt(3) - m_txt(0)*.5f - m_txt(1)*.5f, v[0] + m_txt(0), v[1] + m_txt(1), v[0] + m_txt(1) };
         DebugDrawContour(v, 4, S_vector(0, 1, 0));
                              //render bbox contour
         DebugDrawContour(contour_points, num_cpts, S_vector(1, 0, 0));
      }
#endif
      int shd_tp_index = 0;

      drv->DisableTextureStage(1);
      drv->SetupBlend(I3DBLEND_OPAQUE);
      drv->SetupTextureStage(0, D3DTOP_SELECTARG1);
      hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
      byte a = FloatToInt(opacity*255.0f);
      drv->SetTextureFactor(0xff000000 | (a<<16) | (a<<8) | a);

      drv->SetTexture(NULL);
      DrawList(pc_shd);

                              //blur the texture
#if 1
      if(dist_to_cam < drv->shadow_range_n*.5f){
         bool was_dt = (drv->GetFlags()&DRVF_DITHER);
         if(was_dt) drv->SetState(RS_DITHER, false);

         bool was_lf = (drv->GetFlags()&DRVF_LINFILTER);
         //if(was_lf) drv->SetState(RS_LINEARFILTER, false);
         if(!was_lf) drv->SetState(RS_LINEARFILTER, true);
         drv->tp_shadow[1]->Manage(I3D_texture::MANAGE_CREATE_ONLY);

         C_smart_ptr<IDirect3DSurface8> tmp_rt, tmp_zs;
         HRESULT hr;
         hr = drv->tp_shadow[1]->GetD3DTexture()->GetSurfaceLevel(0, &rt);
         CHECK_D3D_RESULT("GetSurfaceLevel", hr);
         drv->SetRenderTarget(rt, drv->zb_shadow);
         rt->Release();

         drv->UpdateViewport(0, 0, drv->tp_shadow[1]->SizeX1(), drv->tp_shadow[1]->SizeY1());
         hr = lpDev8->Clear(0, NULL, D3DCLEAR_TARGET, 0, 1.0f, 0);
         CHECK_D3D_RESULT("Clear", hr);
         drv->UpdateViewport(1, 1, drv->tp_shadow[1]->SizeX1() - 1, drv->tp_shadow[1]->SizeY1() - 1);

         drv->SetTexture1(0, drv->tp_shadow[0]);

         drv->SetStreamSource(drv->vb_blur, sizeof(I3D_driver::S_blur_vertex));
         hr = lpDev8->SetIndices(drv->ib_blur, 0);
         CHECK_D3D_RESULT("SetIndices", hr);
         drv->SetVertexShader(D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1);

         drv->SetupBlend(I3DBLEND_ADDALPHA | I3DBLEND_VERTEXALPHA);

         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
         CHECK_D3D_RESULT("SetTextureStageState", hr);

         if(drv->GetFlags()&DRVF_WIREFRAME)
            lpDev8->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

         drv->SetClipping(true);
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0,
            I3D_driver::NUM_SHADOW_BLUR_SAMPLES * 4,
            0, I3D_driver::NUM_SHADOW_BLUR_SAMPLES * 2);
         CHECK_D3D_RESULT("DrawIndexedPrimitive", hr);
         drv->SetState(RS_LINEARFILTER, was_lf);

         if(was_dt) drv->SetState(RS_DITHER, true);
         if(drv->GetFlags()&DRVF_WIREFRAME)
            lpDev8->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

         shd_tp_index = 1;
      }else{
         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
      }
#else
      hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
#endif
                              //restore rt back
      drv->SetRenderTarget(save_rt, save_zs);

      if(was_zb) drv->SetState(RS_USEZB, true);
      drv->UpdateViewport(viewport[0], viewport[1], viewport[2], viewport[3]);

      SetActiveCamera(save_curr_cam);
      SetMViewProj(current_camera->GetViewMatrix1(), current_camera->GetProjectionMatrix1());
      hr = lpDev8->SetTransform(D3DTS_PROJECTION, (const D3DMATRIX*)&current_camera->GetProjectionMatrix1());
      CHECK_D3D_RESULT("SetTransform", hr);

                              //use multi-texturing for shadow fading
                              // stage 1 contains 1-D clamped texture
      drv->SetTexture1(0, drv->tp_shadow[shd_tp_index]);
      if(!(drv->GetFlags()&DRVF_DEBUGDRAWSHADOWS)){
         drv->SetupBlend(I3DBLEND_INVBLEND);
                              //setup shadow opacity
         //byte a = FloatToInt(opacity*255.0f);
         //drv->SetTextureFactor((a<<16) | (a<<8) | a);
#if 1
                              //enable blending with stage 1 (clip to positive direction only)
         drv->SetTexture1(1, drv->tp_shadow[2]);
         drv->SetupTextureStage(1, D3DTOP_MODULATE);
         drv->DisableTextureStage(2);
#else
         drv->DisableTextureStage(1);
#endif
      }else{
         drv->SetupBlend(I3DBLEND_ALPHABLEND);
         drv->SetupTextureStage(0, D3DTOP_MODULATE);
         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TFACTOR);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         drv->SetTextureFactor(0x8000ff00);
         drv->DisableTextureStage(1);
      }
      for(i=0; i<2; i++){
         hr = lpDev8->SetTextureStageState(i, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         hr = lpDev8->SetTextureStageState(i, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
      }
      //drv->SetFogColor(0x80808080);
      //drv->EnableNoCull(false);
      const D3DCAPS8 *caps = drv->GetCaps();
      if(caps->RasterCaps&D3DPRASTERCAPS_ZBIAS)
         lpDev8->SetRenderState(D3DRS_ZBIAS, 1);

                              //re-render all receivers using this shadow
      for(int ri=hit_receivers.size(); ri--; ){
         PI3D_visual vis = hit_receivers[ri];
                              //compute texture transformation matrix
         S_matrix tm = vis->GetMatrix() * m_txt_inv;
         tm.Transpose();
                              //re-use last row for light direction, in local coords
         tm(3) = ldir.RotateByMatrix(vis->GetInvMatrix1());
         hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_1, &tm, 4);
         CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

         //PR_BEG; //for(int i=100; i--; )
         vis->RenderSolidMesh(this, false, &vc);
         //vis->RenderSolidMesh(this, false);
         //PR_END;
      }

      if(caps->RasterCaps&D3DPRASTERCAPS_ZBIAS)
         lpDev8->SetRenderState(D3DRS_ZBIAS, 0);

      for(i=0; i<2; i++){
         hr = lpDev8->SetTextureStageState(i, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         hr = lpDev8->SetTextureStageState(i, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
      }

      if(drv->GetFlags()&DRVF_DEBUGDRAWSHADOWS){
         hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
         CHECK_D3D_RESULT("SetTextureStageState", hr);

                              //render shadow texture
         drv->DisableTextureStage(1);
         drv->EnableFog(false);
         drv->SetVertexShader(D3DFVF_XYZRHW | D3DFVF_TEX1);
         const int NUM_X = 8, NUM_Y = 6;
         float sx = (float)drv->GetGraphInterface()->Scrn_sx() / NUM_X;
         float sy = (float)drv->GetGraphInterface()->Scrn_sy() / NUM_Y;
         int x = debug_show_pos % NUM_X;
         int y = debug_show_pos / NUM_X;
         struct S_vertex{
            float x, y, z, w;
            float u, v;
         } v[4] = {
            x*sx,       y*sy,       0, 1,    0, 0,
            x*sx+sx-1,  y*sy,       0, 1,    1, 0,
            x*sx,       y*sy+sy-1,  0, 1,    0, 1,
            x*sx+sx-1,  y*sy+sy-1,  0, 1,    1, 1,
         };
         drv->SetupBlend(I3DBLEND_OPAQUE);
         drv->SetupTextureStage(0, D3DTOP_SELECTARG1);
         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE | D3DTA_COMPLEMENT);
         CHECK_D3D_RESULT("SetTextureStageState", hr);

         bool use_zb = (drv->GetFlags()&DRVF_USEZB);
         if(use_zb) drv->SetState(RS_USEZB, false);
         bool is_wire = (drv->GetFlags()&DRVF_WIREFRAME);
         if(is_wire) drv->SetState(RS_WIREFRAME, false);

         hr = lpDev8->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, v, sizeof(S_vertex));
         CHECK_D3D_RESULT("DrawPrimitiveUP", hr);

         if(use_zb) drv->SetState(RS_USEZB, true);
         if(is_wire) drv->SetState(RS_WIREFRAME, true);
         drv->ResetStreamSource();
         ++debug_show_pos;

         {
            S_matrix m;
            m.SetDir(vc.dir, 0.0f);
            swap(m(2), m(1));
            m(3) = vc.pos;
            DebugDrawCylinder(m, vc.radius, vc.radius, SHADOW_LENGTH, S_vector(0, 0, 1), 0x80);
            SetRenderMatrix(GetIdentityMatrix());
            DrawLine(vc.pos, vc.pos+vc.dir, S_vector(1, 0, 0), 0x0);
         }
         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
      }
   }
}

//----------------------------

struct S_dd_context{
   I3D_FRAME_TYPE ft;
   PI3D_scene scene;
   bool sector_visible;
   dword enum_mask;
   const vector<PI3D_occluder> *occ_list;
   const S_view_frustum *vf;
   dword render_time;
   dword drv_flags, drv_flags2;
   int debug_draw_mat;
};


//----------------------------

I3DENUMRET I3DAPI I3D_scene::cbDebugDraw(PI3D_frame frm, dword c){

   S_dd_context *ddc = (S_dd_context*)c;
   switch(frm->GetType1()){

   case FRAME_MODEL:
      {
         PI3D_model mod = I3DCAST_MODEL(frm);
         if((ddc->drv_flags&DRVF_DRAWHRBBOX) && (frm->num_hr_vis) &&
            frm->IsOn1()){
            if(!(frm->GetFlags()&FRMFLAGS_HR_BOUND_VALID))
               mod->hr_bound.MakeHierarchyBounds(frm);
            mod->hr_bound.GetBoundSphereTrans(frm, FRMFLAGS_HR_BSPHERE_VALID);
            mod->hr_bound.DebugDrawBounds(ddc->scene, frm, FRMFLAGS_HR_BSPHERE_VALID);
         }
      }
      break;

   case FRAME_VISUAL:
      {
         HRESULT hr;
         PI3D_visual vis = I3DCAST_VISUAL(frm);
         if(vis->last_render_time != ddc->render_time)
            break;
         if(ddc->drv_flags&(DRVF_DRAWBBOX | DRVF_DRAWHRBBOX)){
            if(ddc->drv_flags&DRVF_DRAWBBOX){
               if(!(vis->vis_flags&VISF_BOUNDS_VALID)) vis->ComputeBounds();
               vis->bound.DebugDrawBounds(ddc->scene, frm, FRMFLAGS_BSPHERE_TRANS_VALID);
            }
            if((ddc->drv_flags&DRVF_DRAWHRBBOX) && frm->num_hr_vis){
               if(!(frm->frm_flags&FRMFLAGS_HR_BOUND_VALID))
                  vis->hr_bound.MakeHierarchyBounds(frm);
               vis->hr_bound.DebugDrawBounds(ddc->scene, frm, FRMFLAGS_HR_BSPHERE_VALID);
            }
         }
         if(((ddc->drv_flags&DRVF_DEBUGDRAWSHDRECS) && (frm->frm_flags&I3D_FRMF_SHADOW_RECEIVE)) ||
             ((ddc->drv_flags2&DRVF2_DEBUGDRAWSTATIC) && (frm->frm_flags&I3D_FRMF_STATIC_COLLISION)) ||
             (ddc->debug_draw_mat==vis->GetCollisionMaterial() && (vis->GetFrameFlags()&I3D_FRMF_STATIC_COLLISION)))
         {
            bool b_shadow = ((ddc->drv_flags&DRVF_DEBUGDRAWSHDRECS) && (frm->frm_flags&I3D_FRMF_SHADOW_RECEIVE));
            bool b_static = ((ddc->drv_flags2&DRVF2_DEBUGDRAWSTATIC) && (frm->frm_flags&I3D_FRMF_STATIC_COLLISION));
            bool b_debug_mat = (ddc->debug_draw_mat==vis->GetCollisionMaterial() && (vis->GetFrameFlags()&I3D_FRMF_STATIC_COLLISION));
            dword texture_factor;

                              //debug_mat override static(becouse every visual with material IS static)
            if(b_shadow&&b_debug_mat)
               texture_factor = 0xc070a070;
            else
            if(b_shadow&&b_static)
               texture_factor = 0xc080f000;
            else
            if(b_shadow)
               texture_factor = 0xc0008000;
            else
            if(b_debug_mat)
               texture_factor = 0xc0800080;
            else
               texture_factor = 0xc0808000;

                              //render shadow receiver
            PI3D_driver drv = ddc->scene->GetDriver1();
            drv->SetTexture1(0, NULL);
            drv->DisableTextureStage(1);
            LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
            drv->SetupBlend(I3DBLEND_ALPHABLEND);
            drv->SetupTextureStage(0, D3DTOP_MODULATE);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TFACTOR);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            drv->SetTextureFactor(texture_factor);
                              //setup camera vector in visual's local coords
            {
               const S_matrix &m_cam = ddc->scene->GetActiveCamera1()->GetMatrix();
               S_vectorw vw = m_cam(2);
               vw.Invert();
               vw %= vis->GetInvMatrix();
               drv->GetDevice1()->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_1, &vw, 1);
            }

            vis->RenderSolidMesh(ddc->scene, true);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
         }
      }
      break;

   case FRAME_TARGET:
      {
                        //check if target is of requested type
         PI3D_frame f1 = I3DCAST_TARGET(frm)->GetFrame();
         if(f1 && f1->GetType1()==ddc->ft)
            I3DCAST_TARGET(frm)->Draw1(ddc->scene, false);
      }
      break;
   case FRAME_LIGHT:
      {
         PI3D_light lp = I3DCAST_LIGHT(frm);
         lp->Draw1(ddc->scene, ddc->vf, false);
      }
      break;
   case FRAME_SOUND:
      {
         PI3D_sound snd = I3DCAST_SOUND(frm);
         snd->Draw1(ddc->scene, ddc->vf, false);
      }
      break;

   case FRAME_CAMERA:
      I3DCAST_CAMERA(frm)->Draw1(ddc->scene, false);
      break;

   case FRAME_OCCLUDER:
      I3DCAST_OCCLUDER(frm)->Draw1(ddc->scene, false);
      break;

   case FRAME_JOINT:
      I3DCAST_JOINT(frm)->Draw1(ddc->scene, false);
      break;
   case FRAME_SECTOR:
      {
         PI3D_sector sct = I3DCAST_SECTOR(frm);
         bool save_visible = ddc->sector_visible;
         ddc->sector_visible = sct->IsVisible();
                        //optim: skip drawing children if not visible
         sct->EnumFrames(cbDebugDraw, c, ddc->enum_mask);
         ddc->sector_visible = save_visible;
         return I3DENUMRET_SKIPCHILDREN;
      }
      break;
   case FRAME_VOLUME:
      {
         PI3D_volume vol = I3DCAST_VOLUME(frm);

         /*
         if(vol->GetVolumeType() == I3DVOLUME_FACE)
            break;
            */
                              //check if this, or parent frames are visible
         bool on = vol->IsOn1();
         for(PI3D_frame prnt=vol; prnt=prnt->GetParent1(); ){
            if(!prnt->IsOn1()){
               on = false;
               break;
            }
         }
         if(on){
            vol->Prepare();

            bool clip;
                              //view frustum check
            bool out = !SphereInVF(*ddc->vf,
               vol->GetBoundSphere(), clip);
            if(out)
               break;
                              //occlusion test (front to back!)
            for(int i=0; i<ddc->occ_list->size(); i++){
               PI3D_occluder occ = (*ddc->occ_list)[i];

               bool occluded = occ->IsOccluding(vol->GetBoundSphere(), clip);
               if(occluded)
                  break;
            }
            if(i != ddc->occ_list->size())
               break;

            if(vol->GetCollisionMaterial()==ddc->debug_draw_mat){
               static const S_vector color(.5f, 0, .5f);
               vol->Draw1(ddc->scene, false, false, &color);
            }else
            if((ddc->drv_flags2&DRVF2_DRAWVOLUMES) ||
                              //draw dynamic volimes when DRVF_DEBUGDRAWDYNAMIC is set
               ((ddc->drv_flags&DRVF_DEBUGDRAWDYNAMIC) && !(frm->frm_flags&I3D_FRMF_STATIC_COLLISION)) ||
                              //draw static volumes when DRVF2_DEBUGDRAWSTATIC is set
               ((ddc->drv_flags2&DRVF2_DEBUGDRAWSTATIC) && (frm->frm_flags&I3D_FRMF_STATIC_COLLISION)) ){

               vol->Draw1(ddc->scene, false);
            }
         }
      }
      break;
   default:
      frm->DebugDraw(ddc->scene);
   }
   return I3DENUMRET_OK;
}

//----------------------------

void I3D_scene::DebugDrawFrames(I3D_FRAME_TYPE ft, dword enum_mask,
   const S_view_frustum &vf, const vector<PI3D_occluder> &occ_list){

   PI3D_driver drv = GetDriver1();

   bool zw_en = drv->IsZWriteEnabled();
   if(zw_en) drv->EnableZWrite(false);

   S_dd_context ddc;

   ddc.ft = ft;
   ddc.scene = this;
   ddc.sector_visible = GetPrimarySector1()->IsVisible();
   ddc.enum_mask = enum_mask;
   ddc.occ_list = &occ_list;
   ddc.vf = &vf;
   ddc.render_time = drv->GetRenderTime();
   ddc.drv_flags = drv->GetFlags();
   ddc.drv_flags2 = drv->GetFlags2();
   ddc.debug_draw_mat = drv->debug_draw_mat;

   EnumFrames(cbDebugDraw, (dword)&ddc, ENUMF_TARGET | enum_mask);

   if(zw_en) drv->EnableZWrite(true);
}

//----------------------------

void I3D_scene::RenderView(void *context, dword render_flags, E_RENDERVIEW_MODE rv_mode, PI3D_frame curr_mirror,
   const S_plane *add_clip_planes, dword num_add_clip_planes){

   HRESULT hr;

   current_camera->UpdateMatrices(aspect_ratio);

   S_preprocess_context pc(this, drv->GetRenderviewID(), current_camera->GetMatrixDirect()(3));
   pc.render_context = context;
   pc.render_flags = render_flags;
   pc.curr_sector = NULL;
   pc.curr_mirror = curr_mirror;
   pc.mode = rv_mode;

   drv->UpdateViewport(viewport[0], viewport[1], viewport[2], viewport[3]);

   PrepareViewFrustum(pc);
   for(int i=0; i<num_add_clip_planes; i++){
      pc.view_frustum.clip_planes[pc.view_frustum.num_clip_planes++] =
         add_clip_planes[i];
   }
                              //camera sector visible also without portal test
   if(!curr_mirror){
      SetupCameraSector(context);
   }
   SetMViewProj(current_camera->GetViewMatrix1(), current_camera->GetProjectionMatrix1());
   current_camera->GetCurrSector()->SetVisible(current_camera->GetWorldPos(), pc.view_frustum);

   bool render_backdrop = false;
   if(primary_sector->IsVisible()){
      render_backdrop = (backdrop_sector->IsOn1() && (drv->drv_flags&DRVF_CLEAR) && (backdrop_sector->NumChildren1()!=0));
   }
   
                              //clear background/depth/stencil
   if(rv_mode!=RV_MIRROR){
                              //setup projection matrix now (W-buffer clearing needs it?)
      drv->GetDevice1()->SetTransform(D3DTS_PROJECTION, (const D3DMATRIX*)&current_camera->GetProjectionMatrix1());
      dword clear_flags = 0;
      /*                      //Bug fix: always clear, because sector may be clipped by far clipping plane
      bool do_clear = false;
      if(drv->GetFlags()&DRVF_CLEAR){
         do_clear = (primary_sector->IsVisible() ||
                                 //clear background if wireframe on
            (drv->drv_flags&DRVF_WIREFRAME) ||
                                 //clear background if visuals not drawn
            !(drv->drv_flags&DRVF_DRAWVISUALS));
      }
      */
      bool do_clear = (drv->GetFlags()&DRVF_CLEAR);

      if(do_clear)
         clear_flags |= D3DCLEAR_TARGET;
      if(drv->drv_flags&DRVF_USEZB) clear_flags |= D3DCLEAR_ZBUFFER;
                              //clear also stencil buffer, if there's one
      if((drv->GetFlags()&DRVF_HASSTENCIL))
         clear_flags |= D3DCLEAR_STENCIL;
      if(clear_flags){
         hr = drv->GetDevice1()->Clear(0, NULL, clear_flags, dw_color_bgnd, 1.0f, 0);
         CHECK_D3D_RESULT("Clear", hr);
      }
   }

   const vector<C_smart_ptr<I3D_frame> > &frames = cont.GetFrameVector();

   if(render_backdrop){
                              //draw backdrop image
      bool was_zb = (drv->drv_flags&DRVF_USEZB);
                              //disable writting to z-buffer
      if(was_zb) drv->SetState(RS_USEZB, false);
      PI3D_camera bd_cam = backdrop_camera;
      bd_cam->SetFOV(current_camera->GetFOV1());
      bd_cam->UpdateMatrices(aspect_ratio);
      drv->GetDevice1()->SetTransform(D3DTS_PROJECTION, (const D3DMATRIX*)&bd_cam->GetProjectionMatrix1());
      {                       //update view matrix - rotation part only
         S_matrix &mdst = *(S_matrix*)&bd_cam->GetViewMatrix1();
         const S_matrix &msrc = current_camera->GetViewMatrix1();
         mdst(0) = msrc(0);
         mdst(1) = msrc(1);
         mdst(2) = msrc(2);
      }
      {
         S_matrix &mdst = *(S_matrix*)&bd_cam->matrix;
         const S_matrix &msrc = current_camera->matrix;
         mdst(0) = msrc(0);
         mdst(1) = msrc(1);
         mdst(2) = msrc(2);
      }
      SetMViewProj(bd_cam->GetViewMatrix1(), bd_cam->GetProjectionMatrix1());
      C_smart_ptr<I3D_camera> save_curr_cam = current_camera;
      SetActiveCamera(bd_cam);

      S_preprocess_context bd_pc(this, drv->GetRenderviewID(), S_vector(0, 0, 0));
      bd_pc.render_context = context;
      bd_pc.render_flags = render_flags;
      bd_pc.curr_sector = backdrop_sector;
      bd_pc.curr_mirror = curr_mirror;
      bd_pc.Reserve(frames.size() * 4);
      PrepareViewFrustum(bd_pc);

                              //transform all backdrop objects
      dword prep_flags = 0;

                              //update all lights
      if(backdrop_sector->frm_flags&FRMFLAGS_HR_LIGHT_RESET){
         backdrop_sector->UpdateAllLights();
         prep_flags |= E_PREP_SET_HR_LIGHT_DIRTY;
      }

      assert(backdrop_sector->NumChildren1());
      if(backdrop_sector->num_hr_vis)
         PreprocessChildren(backdrop_sector, bd_pc, prep_flags);
                              //sort face groups by pivot distance from [0,0,0]
      bd_pc.SortByDistance();
      DrawList(bd_pc);
                              //put everything back
      SetActiveCamera(save_curr_cam);
      if(was_zb) drv->SetState(RS_USEZB, true);
   }

   SetMViewProj(current_camera->GetViewMatrix1(), current_camera->GetProjectionMatrix1());

   if(num_add_clip_planes){
      assert(num_add_clip_planes==1);
      drv->SetClippingPlane(add_clip_planes[0], m_view_proj_hom);
   }
   //drv->SetClippingPlane(*(S_plane*)drv->debug_float, m_view_proj_hom);

                              //update matrices
   drv->GetDevice1()->SetTransform(D3DTS_PROJECTION, (const D3DMATRIX*)&current_camera->GetProjectionMatrix1());

   vector<PI3D_occluder> occ_list;
   if(drv->drv_flags&DRVF_DRAWVISUALS){

      dword prep_flags = 0;
                              //alloc enough members for storing
                              // pre-processed info
      pc.Reserve(frames.size() * 4);

      if(primary_sector->IsVisible()){
         pc.curr_sector = primary_sector;
         ++drv->scene_stats.frame_count[FRAME_SECTOR];
      }else{
         prep_flags |= E_PREP_NO_VISUALS;
      }

      if(primary_sector->frm_flags&FRMFLAGS_HR_LIGHT_RESET){
         primary_sector->UpdateAllLights();
         prep_flags |= E_PREP_SET_HR_LIGHT_DIRTY;
      }

      pc.shadow_receivers.reserve(proposed_num_shd_receivers * 2);
      Preprocess(pc, prep_flags | E_PREP_ADD_SHD_CASTERS);
      current_camera->GetCurrSector()->ResetRecursive();
      proposed_num_shd_receivers = pc.shadow_receivers.size();

      pc.Sort();
      DrawList(pc);
   }else{
      current_camera->GetCurrSector()->ResetRecursive();
   }

   drv->EnableNoCull(false);

   if((drv->GetFlags2() & (DRVF2_DRAWLIGHTS | DRVF2_DRAWSOUNDS | DRVF2_DRAWCAMERAS|
      DRVF2_DRAWDUMMYS | DRVF2_DRAWOCCLUDERS | DRVF2_DRAWJOINTS | DRVF2_DRAWVOLUMES |
      DRVF2_DEBUGDRAWSTATIC)) ||
      (drv->GetFlags() & (DRVF_DRAWHRBBOX | DRVF_DRAWBBOX | DRVF_DEBUGDRAWSHDRECS | DRVF_DEBUGDRAWDYNAMIC)) ||
      drv->debug_draw_mat!=-1){

      dword draw_flags = ENUMF_SECTOR;
      dword f = drv->GetFlags();
      dword f2 = drv->GetFlags2();
      if(f2&DRVF2_DRAWLIGHTS) draw_flags |= ENUMF_LIGHT;
      if(f2&DRVF2_DRAWSOUNDS) draw_flags |= ENUMF_SOUND;
      if(f2&DRVF2_DRAWCAMERAS) draw_flags |= ENUMF_CAMERA;
      if(f2&DRVF2_DRAWDUMMYS) draw_flags |= ENUMF_DUMMY;
      if(f2&DRVF2_DRAWOCCLUDERS) draw_flags |= ENUMF_OCCLUDER;
      if(f2&DRVF2_DRAWJOINTS) draw_flags |= ENUMF_JOINT;
      if(f2&DRVF2_DEBUGDRAWSTATIC) draw_flags |= ENUMF_VISUAL;
      if((f2&(DRVF2_DRAWVOLUMES|DRVF2_DEBUGDRAWSTATIC)) || (f&DRVF_DEBUGDRAWDYNAMIC)) draw_flags |= ENUMF_VOLUME;
      if(f&DRVF_DRAWBBOX) draw_flags |= ENUMF_VISUAL;
      if(f&DRVF_DRAWHRBBOX) draw_flags |= ENUMF_VISUAL | ENUMF_MODEL;
      if(f&DRVF_DEBUGDRAWSHDRECS) draw_flags |= ENUMF_VISUAL;
      if(drv->debug_draw_mat!=-1)
         draw_flags |= ENUMF_VISUAL | ENUMF_VOLUME;
      //if(draw_flags&ENUMF_VOLUME)
         //BuildVolTree();
      if(draw_flags&ENUMF_VOLUME)
         BuildDynamicVolTree();

      DebugDrawFrames(FRAME_NULL, draw_flags, pc.view_frustum, occ_list);
   }

   if(drv->GetFlags() &DRVF_DEBUGDRAWBSP){
      bsp_builder.DebugDraw(pc);
   }
   if(drv->GetFlags() &DRVF_DEBUGDRAWDYNAMIC){
      DebugDrawDynamicVolumes();
   }

   if(drv->drv_flags&(DRVF_DRAWSECTORS | DRVF_DRAWPORTALS)){
      for(int i=sectors.size(); i--; ){
         PI3D_sector sct = sectors[i];
         sct->Draw1(this, drv->drv_flags&DRVF_DRAWSECTORS,
            drv->drv_flags&DRVF_DRAWPORTALS);
      }
   }
   if(num_add_clip_planes)
      drv->GetDevice1()->SetRenderState(D3DRS_CLIPPLANEENABLE, 0);
   //drv->DisableClippingPlane();
}

//----------------------------

I3D_RESULT I3D_scene::Render(dword flags, void *context){

   if(!current_camera)
      return I3DERR_INVALIDPARAMS;
   //current_camera->SetCurrSector(NULL);

   dword curr_time = drv->GetGraphInterface()->ReadTimer();
   int render_time_delta = curr_time - drv->render_time;
   assert((render_time_delta >= 0) && (render_time_delta < 1000*60*60*72));
   /*
   render_time_delta = Max(1, Min(1000, render_time_delta));
   if(render_time_delta <= 0){
      curr_time = drv->render_time + 1;
      render_time_delta = 1;
   }
   */
                              //new render id
   drv->render_time = curr_time;
   last_render_time = curr_time;

                              //save due to sound update
   PI3D_sector old_cam_sector = current_camera->GetCurrSector();

   I3D_RESULT ir = drv->BeginScene();
   if(I3D_FAIL(ir))
      return ir;

   drv->DisableTextureStage(1);

                              //render the screen using current camera
   RenderView(context, flags, RV_NORMAL, NULL, NULL);

#if 0
         S_vectorw v[4];
         for(int i=0; i<4; i++){
            v[i].w = 1.0f;
            v[i].z = 0.0f;
            v[i].x = viewport[(i&1) << 1];
            v[i].y = viewport[(i&2) + 1];
         }
         static const word indx[] = {0, 1, 2, 2, 1, 3};
         drv->SetTexture1(0, NULL);
         DrawTriangles(v, 4, indx, 6, S_vector(1, 0, 0), sizeof(S_vectorw),
            I3DDRAW_TRANSFORMED | 0x40);
#endif
                              //re-compute sounds
   if(drv->GetSoundInterface()){
      //if(primary_sector->num_hr_snd)
         //PreprocessSounds(primary_sector, render_time_delta);
      PreprocessSounds(render_time_delta);
   }
   if(flags&I3DRENDER_UPDATELISTENER)
      UpdateSoundProperties(render_time_delta, old_cam_sector);

   if(drv->GetFlags()&DRVF_DRAWLMTEXTURES)
      drv->DrawLMaps(viewport);

#ifndef NDEBUG

   float f = PR_GET;
   if(f) drv->DEBUG(C_fstring("%.3f", f));
   if(debug_count[0] || debug_count[1] || debug_count[2] || debug_count[3]){
      for(int i=0; i<4; i++){
         drv->DEBUG(C_fstring("[%i]: %i", i, debug_count[i]));
         debug_count[i] = 0;
      }
   }
#endif

   drv->EndScene();
   return I3D_OK;
}

//----------------------------

