#ifndef __SCENE_H
#define __SCENE_H

#include "anim.h"
#include "sector.h"
#include "bsp.h"
#include "container.h"



typedef class I3D_volume *PI3D_volume;
//----------------------------

template<class T> class C_sort_list;

//----------------------------

class I3D_scene{

   dword ref;

   I3D_container cont;
#define SCNF_VIEWPORT_ASPECT  2        //use real viewport aspect ratio
#define SCNF_RESET_DYNAMIC_VOL_TREE 0x80  //volume hierarchy tree reset
#define SCNF_DYNAMIC_VOL_TREE_BUILT 0x100 //volume hierarchy tree has been built

   PI3D_driver drv;
   PI3D_sector primary_sector;
   mutable dword scn_flags;
                              //callback: flags, registered proc and context
   dword cb_flags;
   PI3D_SCNCALLBACK cb_proc;
                              //level map containers:
   vector<C_smart_ptr<I3D_sector> > sectors;
   vector<PI3D_sound> sounds;
   C_smart_ptr<I3D_camera> shd_camera; //camera for rendering shadows (so that we don't need to create one in runtime)

   S_vector color_bgnd;       //background color
   dword dw_color_bgnd;       //background color - precomputed
   PI3D_sector backdrop_sector;
   PI3D_camera backdrop_camera;
                              
   C_smart_ptr<I3D_camera> current_camera;
                              //transformations
   S_matrix m_view_proj_hom;  //result of view and projection matrices, 4x4, in homogeneous space (x and y in range -1 ... 1, z in range 0 .. 1)
   S_matrix mt_view_proj_hom;
   S_matrix m_view_proj;      //m_view_proj_hom scaled to viewport size
   S_matrix m_world_view_proj;        //result of m_view_proj and m_world matrices, 4x4

   dword last_render_time;    //counter set to current time (IGraph::ReadTimer) each time a scene is rendered

   int proposed_num_shd_receivers;  //number of shd recs from last rendering
public:
   inline const S_matrix &GetViewProjHomMat() const{ return m_view_proj_hom; }
   inline const S_matrix &GetViewProjHomMatTransposed() const{ return mt_view_proj_hom; }
private:
                              //sector sound properties fading
   S_sound_env_properties ssp_fade;
   int ssp_count_down;        //0 = not fading

                              //recursive rendering funcs
   void Preprocess(S_preprocess_context&, dword prep_flags);
   void PreprocessOccluders(const S_preprocess_context&, PI3D_frame root, C_sort_list<class I3D_occluder*> &collect_list, PI3D_sector curr_sector);
   void PreprocessSectors(PI3D_frame root, S_preprocess_context&);
   void PreprocessChildren(PI3D_frame root, S_preprocess_context&, dword prep_flags);
   void PreprocessSounds(int time);
   void UpdateSoundProperties(int time, PI3D_sector prev_sector);

   void DrawList(const S_preprocess_context&);
   void RenderShadows(const S_preprocess_context&);

//----------------------------
// Select which light will be used for making shadow on given visual.
   PI3D_light SelectShadowLight(PI3D_sector sct, const S_vector &visual_pos,
      PI3D_frame last_light, float dist_to_cam, float &opacity) const;
   

   void SetupCameraSector(void *context);

   friend I3D_sector;
   friend I3D_driver;
   friend I3D_sound;

                              //viewing frustum tests - cache
   void PrepareViewFrustum(S_preprocess_context&);

   I3D_scene(PI3D_driver);
   bool Init(I3D_driver*);


                              //collisions:
   static I3DENUMRET I3DAPI cbPick(PI3D_frame frm, dword dw_pick_data);
   static void cbPick(PI3D_visual vis, PI3D_frame pick_frm,
      class I3D_mesh_base*, struct S_pick_help &pick_data);
   static I3DENUMRET cbGeoCol(PI3D_frame root, struct S_pick_help &pick_data);
   C_bsp_tree_builder bsp_builder;
   mutable t_volume_container dynamic_volume_tree;
   mutable t_volume_container dynamic_auto_volume_list;
   mutable t_volume_container root_volumes; //used for line test - contain all volumes which are not staic
   mutable dword last_dynamic_tree_build_time;
   typedef map<int, vector<CPI3D_volume> > t_spheres_space;
   mutable t_spheres_space spheres_space;
   void CheckColLineList(I3D_collision_data &cd, const PI3D_volume *vols, int num_vols, bool children) const;
   inline void CheckColLineDyn(I3D_collision_data &cd, const t_volume_container &vtree) const;
   PI3D_volume CheckColSlist(const S_trace_help &th, const PI3D_volume *vols, int num_vols) const;
   inline PI3D_volume CheckColSDyn(const S_trace_help &th, const t_volume_container &vtree) const;
public:
   inline void AddSphereToSpace(CPI3D_volume vol, int id_min, int id_max);
   inline void DeleteSphereFromSpace(CPI3D_volume vol, int id_min, int id_max);

   inline void ResetDynamicVolTree() const{ scn_flags |= SCNF_RESET_DYNAMIC_VOL_TREE; }
   void DestroyDynamicVolTree() const;
   inline dword GetBgndClearColor() const{ return dw_color_bgnd; }

   void SetMViewProj(const S_matrix &view, const S_matrix &proj);
//----------------------------
// Render single view from currently active camera.
   void RenderView(void *context, dword render_flags, E_RENDERVIEW_MODE, PI3D_frame curr_mirror = NULL,
      const S_plane *add_clip_planes = NULL, dword num_add_clip_planes = 0);

   static I3DENUMRET I3DAPI cbDebugDraw(PI3D_frame, dword);

   void DebugDrawFrames(I3D_FRAME_TYPE ft, dword enum_mask,
      const S_view_frustum&, const vector<PI3D_occluder> &occ_list);

   E_BSP_BUILD_RESULT GetStaticFrames(vector<PI3D_frame> &frm_list) const;

private:
   void BuildDynamicVolTree() const;
   void DebugDrawDynamicVolumes() const;

   bool TestPickEditFrames(struct I3D_collision_data&, const float test_length) const;
   void TestVolumeSM(I3D_collision_data &cd) const;
   static void cbEnumSMandJoints(PI3D_frame frm, I3D_collision_data &cd);

   PI3D_frame TestColBspLine(struct I3D_collision_data&) const;
   PI3D_volume TestColLineDynamic(struct I3D_collision_data &cd) const;
   bool TestColSphereDynamic(const S_trace_help &th) const;
   bool SphereMove(I3D_collision_data &cd, PI3D_frame root = NULL) const;

   friend I3DENUMRET I3DAPI cbFrmCol(PI3D_frame frm, dword hlp);
public:
   ~I3D_scene();
   int viewport[4];           //left, top, right, bottom

                              //viewport pre-process
   int fi_viewport[4];        //fake float viewport - bitmasks of float
   float viewport_center[2];
   float scrn_width2, scrn_height2;
   float aspect_ratio;

   inline int NumSectors() const{ return sectors.size(); }
   inline PI3D_sector *GetSectors(){ return (PI3D_sector*)sectors.begin(); }
   //inline PI3D_camera GetCurrentCamera() const{ return current_camera; }
   inline PI3D_sector GetPrimarySector1() const{ return primary_sector; }
   inline const int *GetViewport1() const{ return viewport; }
   inline PI3D_camera GetActiveCamera1() const{ return current_camera; }
   inline PI3D_container GetContainer1(){ return &cont; }

   void DelAnimation(PI3D_frame);

   void TransformVertsClipped(const S_vector *in_verts, dword num_verts,
      S_vectorw *verts_out, dword *clip_union1, dword *clip_intersection1) const;
   void TransformVertsNonClipped(const S_vector *in_verts, dword in_stride, dword num_verts,
      S_vectorw *verts_out, dword out_stride) const;

                              //may be NULL, which means identity matrix
   void SetWorldMatrix(const S_matrix*);
                              //debugging
   void DrawIcon(const S_vector &pos, int res_id, byte inv_alpha = 0, float size = 1.0f);
   void DebugDrawSphere(const S_matrix &m, float radius, const S_vector &color = S_vector(1, 1, 1), byte alpha = 0);
   void DebugDrawCylinder(const S_matrix &m, float radius1, float radius2, float height, const S_vector &color = S_vector(1, 1, 1), byte alpha = 0);
   void DebugDrawArrow(const S_matrix &m, float lenght, const S_vector &color = S_vector(1, 1, 1), byte alpha = 0, int axis = 2);
   void DebugDrawBBox(const S_vector *bbox_full, const S_vector &color = S_vector(1, 1, 1), byte alpha = 0);
   void DebugDrawContour(const S_vector *contour_points, dword num_cpts, const S_vector &color = S_vector(1, 1, 1), byte alpha = 0);

   inline PI3D_driver GetDriver1() const{ return drv; }

   void AddAnimation(PI3D_animation_base, PI3D_frame, dword options);

   PI3D_frame TestColGeoInternal(const S_vector &from, const S_vector &dir,
      dword flags, float *ret_dist = NULL,
      S_vector *ret_normal = NULL, PI3D_line_collision_response = NULL, 
      void *cb_context = NULL, PI3D_line_collision_response textel_resp = NULL) const;

public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
                              //general
   I3DMETHOD_(PI3D_driver,GetDriver)() const{ return (PI3D_driver)drv; }
   I3DMETHOD_(PI3D_SCNCALLBACK,SetCallback)(PI3D_SCNCALLBACK cb, dword cbf);
                              //rendering
   I3DMETHOD(Render)(dword flags, void *context);
   I3DMETHOD(SetViewport)(dword l, dword t, dword r, dword b, bool aspect=false);
   I3DMETHOD_(void,GetViewport)(dword &l, dword &t, dword &r, dword &b);
                              //frames
   I3DMETHOD(EnumFrames)(I3D_ENUMPROC*, dword user, dword flags=ENUMF_ALL, const char *mask=NULL) const;
   I3DMETHOD_(PI3D_frame,FindFrame)(const char *name, dword data=ENUMF_ALL) const;
                              //lights
   I3DMETHOD(SetGlobalColor)(const S_vector&);
   I3DMETHOD_(S_vector,GetLightness)(const S_vector &pos, const S_vector *norm, dword light_mode_flags) const;
                              //cameras
   I3DMETHOD(SetActiveCamera)(PI3D_camera);
   I3DMETHOD_(PI3D_camera,GetActiveCamera)() const{ return current_camera; }
   I3DMETHOD(UnmapScreenPoint)(int x, int y, S_vector &pos, S_vector &dir) const;
   I3DMETHOD(TransformPoints)(const S_vector *in, S_vectorw *out, dword num) const;
                              //sectors
   I3DMETHOD_(PI3D_sector,CreateSector)();
   I3DMETHOD_(PI3D_sound,CreateSound)();
   I3DMETHOD_(PI3D_sector,GetSector)(const S_vector&, PI3D_sector suggest_sector = NULL) const;
   I3DMETHOD_(PI3D_sector,GetPrimarySector)() const{ return primary_sector; }
   I3DMETHOD_(PI3D_sector,GetBackdropSector)() const{ return backdrop_sector; }
   I3DMETHOD_(void,SetBackdropRange)(float n1, float f1);
   I3DMETHOD_(void,GetBackdropRange)(float &near1, float &far1) const;
   I3DMETHOD(SetFrameSector)(PI3D_frame);
   I3DMETHOD(SetFrameSectorPos)(PI3D_frame, const S_vector &pos);
                              //background
   I3DMETHOD(SetBgndColor)(const S_vector&);
   I3DMETHOD_(const S_vector&,GetBgndColor)() const{ return color_bgnd; }

                              //collisions
   I3DMETHOD_(void, ResetDynamicVolTreeEx)();
   I3DMETHOD_(bool, TestCollision)(struct I3D_collision_data&) const;

   I3DMETHOD_(void,AttenuateSounds)();

   I3DMETHOD_(dword,GetLastRenderTime)() const{ return last_render_time; }

   I3DMETHOD_(I3D_RESULT,SetRenderMatrix)(const S_matrix&);
   I3DMETHOD(DrawLine)(const S_vector &v1, const S_vector &v2,
      const S_vector &color = S_vector(1, 1, 1), dword flags = 0);
   I3DMETHOD(DrawLines)(const S_vector *v, dword numv, const word *indx,
      dword numi, const S_vector &color, dword v_stride=sizeof(S_vector),
      dword flags = 0);
   I3DMETHOD(DrawTriangles)(const void *v, dword numv, const word *indx,
      dword numi, const S_vector &color, dword v_stride = sizeof(S_vector),
      dword flags = 0, PI3D_text_coor = NULL);
   I3DMETHOD(DrawSprite)(const S_vector &pos, PI3D_material, byte inv_alpha = 0, float size = 1.0f);

   I3DMETHOD(CreateBspTree)(I3D_BSP_BUILD_CB_PROC *cbP = NULL, void *context = NULL);
   I3DMETHOD_(void, DestroyBspTree)();

   I3DMETHOD_(bool, SaveBsp)(C_cache *cp);
   I3DMETHOD_(bool, LoadBsp)(C_cache *cp, I3D_LOAD_CB_PROC* = NULL, void *context = NULL, dword check_flags = 0);
   I3DMETHOD_(bool, IsBspBuild)();

   I3D_CONTAINER(;)
   I3DMETHOD_(I3D_container*,GetContainer)(){ return &cont; }
};

//----------------------------

#endif
