/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Sector.cpp
   Content: Sector frame.
--------------------------------------------------------*/

#include "all.h"
#include "sector.h"
#include "light.h"
#include "camera.h"
#include "scene.h"
#include "object.h"
#include "occluder.h"
#include <malloc.h>           //alloca

#define USE_VF_BSPHERE        //use bounding-sphere tests on viewing frustum

                              //distance under which verts are considered equal
                              // in sector and portal builtindg and portal connecting
#define VERTEXT_COLLAPSE .04f
#define ENABLE_CONTOUR_TEST   //enable testing object's contour agains vf/occluders/portals


                              //debugging:
//#define DEBUG_DRAW_FACE_NORMALS
//#define DEBUG_PAINT_BBOX_CONTOUR    //paint contour of bounding box

//----------------------------
                              //thunk into QHull library
#pragma comment(lib,"QHull\\qhull.lib")

bool __cdecl qhCreateConvexHull(const vector<S_vector> &verts, vector<I3D_triface> *tri_faces,
   vector<I3D_face> *faces = NULL);

//----------------------------

bool CheckFrustumIntersection(const S_plane *planes, dword num_planes,
   const S_vector *verts, dword num_verts){

   bool penetrated = false;
   dword clip_union = (1<<num_planes) - 1;
   dword clip_intersection = 0;
   for(int p2i=num_verts, p2i_next=0; p2i--; p2i_next=p2i){
      const S_vector &v_p2 = verts[p2i];
      const S_vector dir_p2 = verts[p2i_next] - v_p2;

                              //flag specifying that v_p2 is inside
      bool v_p2_in = true;

      float last_enter = -1e+16f, first_leave = 1e+16f;
                              //check this edge against all planes of poly1
      dword plane_in_flags = 0;
      dword plane_mask = 1;
      for(int p1i=num_planes; p1i--; plane_mask<<=1){
#ifdef USE_PREFETCH
         Prefetch1(&planes[p1i-1]);
#endif
         const S_plane &p1_edge_plane = planes[p1i];
         float d = v_p2.DistanceToPlane(p1_edge_plane);
                              //check if out of one plane
         //if(d < 0.0f)
         if(CHECK_ZERO_LESS(d)){
            plane_in_flags |= plane_mask;
         }else{
            clip_intersection |= plane_mask;
            v_p2_in = false;
         }
         if(!penetrated){
                              //get distance at which we penetrate it
            float f = p1_edge_plane.normal.Dot(dir_p2);
            if(!IsAbsMrgZero(f)){
               float r = -d / f;
               if(CHECK_ZERO_LESS(f)){
                  last_enter = Max(last_enter, r);
               }else{
                  first_leave = Min(first_leave, r);
               }
            }else
            if(CHECK_ZERO_GREATER(d)){
                              //missing the plane, can't hit poly
               last_enter = 1e+16f;
            }
         }
      }
      if(v_p2_in)
         return true;
      clip_union &= ~plane_in_flags;
      if(!penetrated){
         if(CHECK_ZERO_GREATER(last_enter) && first_leave <= 1.0f)
            penetrated = (last_enter < first_leave);
      }
   }
                              //if all verts are out of single plane, intersection = false
   if(clip_union)
      return false;
                              //if verts are around all planes, intersection = true
   if(clip_intersection == (1<<num_planes)-1)
      return true;
                              //now it depends if any side penetrated some plane
   return penetrated;
}

//----------------------------

#if defined _MSC_VER && 0

#pragma warning(disable:4035)
__declspec(naked) bool __stdcall SphereCollide(const I3D_bsphere &vf_bs, const I3D_bsphere &bs2){
   __asm{
      push edi
      mov eax, [esp+4+4]
      mov edi, [esp+8+4]
      push eax                //make stack variable
      fld dword ptr[eax+0]
      fsub dword ptr[edi+0]
      fmul st, st
      fld dword ptr[eax+4]
      fsub dword ptr[edi+4]
      fmul st, st
      faddp st(1), st
      fld dword ptr[eax+8]
      fsub dword ptr[edi+8]
      fmul st, st
      faddp st(1), st
      fld dword ptr[eax+12]
      fadd dword ptr[edi+12]
      fmul st, st
      fsubp st(1), st
      fstp dword ptr[esp]
      pop eax
      pop edi
      shr eax, 31
      ret 8
   }
}
#pragma warning(default:4035)

#endif

//----------------------------

                           //links to 2 edges for each edge
static const struct{
   byte link[2][2];        //[neg|pos][]
} edge_links[12] = {
   {4, 8, 5, 9},  //0
   {4, 10, 5, 11},//1
   {6, 8, 7, 9},  //2
   {6, 10, 7, 11},//3
   {0, 8, 1, 10}, //4
   {0, 9, 1, 11}, //5
   {2, 8, 3, 10}, //6
   {2, 9, 3, 11}, //7
   {0, 4, 2, 6},  //8
   {0, 5, 2, 7},  //9
   {1, 4, 3, 6},  //10
   {1, 5, 3, 7}   //11
};

static const struct{
   byte vi[2];             //[neg|pos]
} vertex_indicies[12] = {
   0, 1,    2, 3,    4, 5,    6, 7,
   0, 2,    1, 3,    4, 6,    5, 7,
   0, 4,    1, 5,    2, 6,    3, 7,
};

//----------------------------

void ComputeContour(const I3D_bbox &bb_in, const S_matrix &tm, const S_vector &look,
   S_vector contour_points[6], dword &num_cpts, bool orthogonal){

   num_cpts = 0;

   enum{
      EDGE_DIR_NEG = 0,       //edge pointing to negative direction
      EDGE_DIR_POS = 1,       //edge pointing to positive direction
      EDGE_VISIBLE = 2,       //edge is visible
      EDGE_DIR_MASK = 1,
   };
                              //edge check-in flags - used XOR, so that when edge is
                              // marked twice, its EDGE_VISIBLE flag is cleared
   union{
      byte edge_check_in[12];
      dword dw[3];
   };
   dw[0] = 0; dw[1] = 0; dw[2] = 0;
   
                              //transform edge bbox
   I3D_bbox bb_trans;
   bb_trans.min = bb_in.min * tm;
   bb_trans.max = bb_in.max * tm;

   if(!orthogonal){
      S_vector dir_min = look - bb_trans.min;
      S_vector dir_max = look - bb_trans.max;

      bool inside = true;

      if(dir_min.Dot(tm(0)) < 0.0f){
         edge_check_in[8]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[6]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[10] ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[4]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         inside = false;
      }else
      if(dir_max.Dot(tm(0)) >= 0.0f){
         edge_check_in[5]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[11] ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[7]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[9]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         inside = false;
      }

      if(dir_min.Dot(tm(1)) < 0.0f){
         edge_check_in[0]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[9]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[2]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[8]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         inside = false;
      }else
      if(dir_max.Dot(tm(1)) >= 0.0f){
         edge_check_in[10] ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[3]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[11] ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[1]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         inside = false;
      }

      if(dir_min.Dot(tm(2)) < 0.0f){
         edge_check_in[4]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[1]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[5]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[0]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         inside = false;
      }else
      if(dir_max.Dot(tm(2)) >= 0.0f){
         edge_check_in[2]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[7]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[3]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[6]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         inside = false;
      }
      if(inside)
         return;
   }else{
      if(look.Dot(tm(0)) < 0.0f){
         edge_check_in[8]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[6]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[10] ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[4]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
      }else{
         edge_check_in[5]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[11] ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[7]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[9]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
      }

      if(look.Dot(tm(1)) < 0.0f){
         edge_check_in[0]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[9]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[2]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[8]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
      }else{
         edge_check_in[10] ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[3]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[11] ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[1]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
      }

      if(look.Dot(tm(2)) < 0.0f){
         edge_check_in[4]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[1]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[5]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[0]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
      }else{
         edge_check_in[2]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[7]  ^= EDGE_VISIBLE | EDGE_DIR_POS;
         edge_check_in[3]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
         edge_check_in[6]  ^= EDGE_VISIBLE | EDGE_DIR_NEG;
      }
   }
                              //get any beginning edge
   for(int ei_beg=12; ei_beg--; ){
      if(edge_check_in[ei_beg]&EDGE_VISIBLE)
         break;
   }
   assert(ei_beg!=-1);

   S_vector bbd_t[3];         //three vector aligned with tm axes, scaled by bbox
   bbd_t[0] = tm(0) * (bb_in.max.x - bb_in.min.x);
   bbd_t[1] = tm(1) * (bb_in.max.y - bb_in.min.y);
   bbd_t[2] = tm(2) * (bb_in.max.z - bb_in.min.z);

   int ei = ei_beg;
   do{
      int vi = vertex_indicies[ei].vi[edge_check_in[ei]&EDGE_DIR_MASK];

      assert(num_cpts < 6);
      S_vector &cp = contour_points[num_cpts++];
      cp = bb_trans.min;
      if(vi&1) cp += bbd_t[0];
      if(vi&2) cp += bbd_t[1];
      if(vi&4) cp += bbd_t[2];

                              //move to the next visible edge
      int ei1 = edge_links[ei].link[edge_check_in[ei]&EDGE_DIR_MASK][0];
      ei = edge_links[ei].link[edge_check_in[ei]&EDGE_DIR_MASK][!(edge_check_in[ei1]&EDGE_VISIBLE)];
   } while(ei!=ei_beg);
}

//----------------------------
//----------------------------
                              //portal class
I3D_portal::I3D_portal(PI3D_sector s):
   ref(1),
   owning_sector(s),
   state(PS_UNKNOWN),
   connect_sector(NULL),
   drv(s->GetDriver1()),
   connect_portal(NULL),
   prt_flags(PRTF_ENABLED)
{
   drv->AddCount(I3D_CLID_PORTAL);
}

//----------------------------

I3D_portal::~I3D_portal(){
   drv->DecCount(I3D_CLID_PORTAL);
}

//----------------------------
                              //test if portals identical (but inversed)
bool I3D_portal::Equals(CPI3D_portal prt) const{

   int vnum = vertices.size();
                              //vertiex count must match
   if(vnum!=prt->vertices.size())
      return false;

   const S_matrix &m = owning_sector->GetMatrix();
   const S_matrix &m1 = prt->owning_sector->GetMatrix();
                              //normals must be opposite
                              //rotate normals
   S_vector n = plane.normal.RotateByMatrix(m);
   S_vector n1 = prt->plane.normal.RotateByMatrix(m1);
                              //don't bee too strict, it's just brief test
   if((float)fabs(n.AngleTo(n1)-GetPI()) > 1e-1f)
      return false;

   const S_vector *vp = vertices.begin();
   const S_vector *vp1 = prt->vertices.begin();
   S_vector *v = (S_vector*)alloca(vnum*sizeof(S_vector));
   S_vector *v1 = (S_vector*)alloca(vnum*sizeof(S_vector));
   int i;
                              //transform all his vertices
   for(i=0; i<vnum; i++)
      v1[i] = vp1[i] * m1;
                              //transform our 1st vertex & find one match
   v[0] = vp[0] * m;
   int j = vnum;
   while(j--)
      if((v[0]-v1[j]).Magnitude() <= VERTEXT_COLLAPSE)
         break;
                              
   if(j==-1)                  //no match found
      return false;
                              //transform rest of our vertices
   for(i=1; i<vnum; i++)
      v[i] = vp[i] * m;
                              //now traverse and compare rest of vertices
   for(i=1; i<vnum; i++){
      if(!j--)
         j = vnum-1;
      if((v[i]-v1[j]).Magnitude() > VERTEXT_COLLAPSE)
         break;
   }
   if(i!=vnum)
      return false;
                              //portals identical
   return true;
}

//----------------------------
                              
bool I3D_portal::Build(const vector<S_vector> &v){

   vertices = v;
   S_vector &n = plane.normal;
                              //get portal's normal
#if 1

                              //find 2 points with maximal distance
   int p0, p1, p2;
   float best_dist = 0.0f;
   for(int i=v.size(); i--; ){
      for(int j=i; j--; ){
         S_vector dir = v[i] - v[j];
         float d = dir.Dot(dir);
         if(best_dist<d){
            best_dist = d;
            p0 = i;
            p1 = j;
         }
      }
   }
                              //now find 3rd point with maximal distance to their line
   bool invert;
   S_vector dir = v[p1] - v[p0];
   best_dist = -1.0f;
   bool curr_invert = true;
   for(i = (p0+1)%v.size(); i!=p0; ++i %= v.size()){
      if(i==p1)
         curr_invert = false;
      else{
         float dtl = v[i].DistanceToLine(v[p0], dir);
         if(best_dist<dtl){
            best_dist = dtl;
            invert = curr_invert;
            p2 = i;
         }
      }
   }
   n.GetNormal(v[p0], v[p1], v[p2]);
   if(invert)
      n.Invert();
#else

   n.GetNormal(v[0], v[1], v[2]);
#endif
   if(IsMrgZero(n.Magnitude()))
      return false;
   n.Normalize();
   plane.d = -n.Dot(v[0]);
   return true;
}

//----------------------------

void I3D_portal::Duplicate(CPI3D_portal prt){

   prt_flags = prt->prt_flags;
   vertices = prt->vertices;
   plane = prt->plane;
}

//----------------------------

void I3D_portal::SetOn(bool b){

   prt_flags &= ~PRTF_ENABLED;
   if(b) prt_flags |= PRTF_ENABLED;
   state = PS_UNKNOWN;
                              //make the same for connected portal
   if(connect_portal){
      connect_portal->prt_flags &= ~PRTF_ENABLED;
      if(b) connect_portal->prt_flags |= PRTF_ENABLED;
      connect_portal->state = PS_UNKNOWN;
   }
}

//----------------------------

bool I3D_portal::PrepareFrustum(const S_vector &cam_pos){

   if(vertices.size() > MAX_FRUSTUM_CLIP_PLANES)
      return false;
                              //determine on which side of portal we are
   float d = cam_pos.DistanceToPlane(plane);
   bool in_front = (d >= 0.0f);
                              //build frustum
   vf.num_clip_planes = vertices.size();   
   assert(vf.num_clip_planes <= MAX_FRUSTUM_CLIP_PLANES);
   for(int i=vf.num_clip_planes, next_i=0; i--; next_i=i){
      const S_vector &v0 = vertices[i];
      const S_vector &v1 = vertices[next_i];
      S_plane &pl = vf.clip_planes[i];
      pl.normal.GetNormal(cam_pos, v0, v1);
      pl.normal.Normalize();
      pl.d = -cam_pos.Dot(pl.normal);
      if(!in_front){
         pl.normal.Invert();
         pl.d = -pl.d;
      }
   }
   return in_front;
}

//----------------------------
//----------------------------
                              //sector class
I3D_sector::I3D_sector(PI3D_scene s, bool primary):
   I3D_frame(s->GetDriver1()),
   fog_light(NULL),
   sound_env_id(0),
   sct_flags(0)
{
   drv->AddCount(I3D_CLID_SECTOR);
   type = FRAME_SECTOR;
   enum_mask = ENUMF_SECTOR;

   if(primary)
      sct_flags |= SCTF_PRIMARY;
}

//----------------------------

I3D_sector::~I3D_sector(){

   int i;
                              //disconnect lights from this
   for(int li=0; li<2; li++){
      vector<PI3D_light> &l = !li ? light_list : fog_light_list;

      for(i=0; i<l.size(); i++){
         PI3D_light lp = l[i];
         for(int j=lp->light_sectors.size(); j--; ){
            if(lp->light_sectors[j]==this){
               lp->light_sectors[j] = lp->light_sectors.back(); lp->light_sectors.pop_back();
               break;
            }
         }
         assert(j!=-1);
      }
      l.clear();
   }
                              //disconnect sounds from this
   for(i=0; i<sound_list.size(); i++){
      PI3D_sound sp = sound_list[i];
      for(int j=sp->sound_sectors.size(); j--; ){
         if(sp->sound_sectors[j]==this){
            sp->sound_sectors[j] = sp->sound_sectors.back(); sp->sound_sectors.pop_back();
            break;
         }
      }
      assert(j!=-1);
   }
   sound_list.clear();

   Disconnect(NULL);
   for(i=own_portals.size(); i--; ){
      own_portals[i]->SetOwningSector(NULL);
   }
   own_portals.clear();

   drv->DecCount(I3D_CLID_SECTOR);
}

//----------------------------

I3D_RESULT I3D_sector::Duplicate(CPI3D_frame frm){

   if(frm->GetType1()==FRAME_SECTOR){
      PI3D_sector sct = I3DCAST_SECTOR(frm);
      sct_flags = sct->sct_flags;
      vertices = sct->vertices;
      faces = sct->faces;
      bound_planes = sct->bound_planes;
      bbox = sct->bbox;
      sct_flags |= SCTF_LIGHT_LIST_RESET;

      I3D_RESULT ir = I3D_frame::Duplicate(frm);
      if(!IsSimple()){
         for(int i=sct->own_portals.size(); i--; ){
            PI3D_portal prt = new I3D_portal(this);
            own_portals.push_back(prt);
            prt->Duplicate(sct->own_portals[i]);
            prt->Release();
         }
      }
      return ir;
   }
   return I3DERR_INVALIDPARAMS;
}

//----------------------------

void I3D_sector::BuildHelperLightLists(){

   lights_vertex.clear();
   lights_dyn_lm.clear();
   lights_shadow.clear();

   int num_lights = light_list.size();
   const PI3D_light *lptr = light_list.begin();
                              //make counts
   int num_v = 0, num_lm = 0, num_s = 0;
   for(int i=num_lights; i--; ){
      PI3D_light lp = *lptr++;
      dword lflg = lp->GetLightFlags();
      if(lflg&I3DLIGHTMODE_VERTEX) ++num_v;
      if(lflg&I3DLIGHTMODE_DYNAMIC_LM) ++num_lm;
      if(lflg&I3DLIGHTMODE_SHADOW) ++num_s;
   }
                              //reserve storage
   lights_vertex.reserve(num_v);
   lights_dyn_lm.reserve(num_lm);
   lights_shadow.reserve(num_s);

                              //build lists
   lptr = light_list.begin();
   for(i=num_lights; i--; ){
      PI3D_light lp = *lptr++;
      dword lflg = lp->GetLightFlags();
      if(lflg&I3DLIGHTMODE_VERTEX) lights_vertex.push_back(lp);
      if(lflg&I3DLIGHTMODE_DYNAMIC_LM) lights_dyn_lm.push_back(lp);
      if(lflg&I3DLIGHTMODE_SHADOW) lights_shadow.push_back(lp);
   }

   sct_flags &= ~SCTF_LIGHT_LIST_RESET;
}

//----------------------------
// Collect meshes of multiple visuals, transform into world coordinates
// and put into contiguous list.
// Split vertices which are too close to each other.
//----------------------------
void CollectMeshes(PI3D_frame *walls, dword num_walls,
   vector<S_vector> &verts, vector<I3D_triface> &tri_faces){

   int j;
   while(num_walls--){
      CPI3D_frame frm = walls[num_walls];
      if(frm->GetType1()!=FRAME_VISUAL)
         continue;
      CPI3D_visual vis = I3DCAST_VISUAL(frm);
      I3D_mesh_base *mb = vis->GetMesh();
      if(!mb)
         continue;
                              //store vertices
      const S_matrix &mat = vis->GetMatrix();
      
      const S_vector *src_verts = (const S_vector*)mb->vertex_buffer.Lock(D3DLOCK_READONLY);
      dword vstride = mb->vertex_buffer.GetSizeOfVertex();
      int vertex_base = verts.size();
      for(j=0; j<mb->vertex_buffer.NumVertices(); j++, src_verts = (const S_vector*)(((const byte*)src_verts) + vstride))
         verts.push_back((*src_verts) * mat);
      mb->vertex_buffer.Unlock();

      PI3D_triface faces = mb->index_buffer.Lock(D3DLOCK_READONLY);
                              //store faces - flip them now
      for(j=mb->NumFaces1(); j--; ){
         tri_faces.push_back(*faces++);
         I3D_triface &fc = tri_faces.back();
         swap(fc[0], fc[1]);
         for(int l=0; l<3; l++)
            fc[l] += vertex_base;
      }
      mb->index_buffer.Unlock();
   }
                              //build vertex map and remap tri_faces
   vector<word> v_map(verts.size());
   MakeVertexMapping(verts.begin(), sizeof(S_vector),
      verts.size(), v_map.begin(), VERTEXT_COLLAPSE);

   for(int i=tri_faces.size(); i--; ){
      I3D_triface &fc = tri_faces[i];
      for(j=3; j--; ){
         fc[j] = v_map[fc[j]];
      }
                              //check if face is valid after remapping
      if(fc[0]==fc[1] || fc[1]==fc[2] || fc[2]==fc[0]){
         fc = tri_faces.back(); tri_faces.pop_back();
      }
   }
}

//----------------------------

#ifdef _DEBUG
static void PaintEdge(PI3D_driver drv, const vector<S_vector> &verts, const I3D_edge &e){
   S_vector vv[2] = {verts[e[0]], verts[e[1]]};
   drv->DEBUG_CODE(0, (dword)vv);
}

static void PaintNormal(PI3D_driver drv, const S_vector &v, const S_vector &n){
   S_vector vv[2] = {v, v+n};
   drv->DEBUG_CODE(0, (dword)vv);
}
#else
#define PaintEdge(a,b,c)
#endif

//----------------------------
// Build portal from set of vertices indicies, erase used from list.
void I3D_sector::BuildPortals(const vector<S_vector> &verts, const vector<word> &prt_verts){

#define BP_PLANAR_THRESH .02f


   int i, j;

   int best_num_behind = 0;
   int best_tri = -1;      //1st point of best triangle
   S_plane best_tri_pl;    //saved plane of best triangle, so that we don't compute it twice

                           //find 1st triangle, of which 1st edge has all verts
                           // behind, and which has most other verts behind
   int numv = prt_verts.size();
   for(i=numv; i--; ){
      const S_vector &v0 = verts[prt_verts[i]];
      const S_vector &v1 = verts[prt_verts[(i+1)%numv]];
      const S_vector &v2 = verts[prt_verts[(i+2)%numv]];

                           //get normal of triangle
      S_vector n_tri; n_tri.GetNormal(v0, v1, v2); n_tri.Normalize();

                           //get plane of triangle
      S_plane pl_tri(n_tri, -v0.Dot(n_tri));

                           //get normal of v01 edge
      S_vector n_edge01; n_edge01.GetNormal(v0, v1, v1+n_tri); n_edge01.Normalize();

                           //get plane of v01 edge
      S_plane pl01(n_edge01, -v0.Dot(n_edge01));

                           //check if all verts are behind edge's plane
      for(j=prt_verts.size(); j--; ){
         const S_vector &v = verts[prt_verts[j]];
         float d = v.DistanceToPlane(pl01);
         if(d > .001f)
            break;
      }
      if(j==-1){
                           //1st edge is ok
                           //check number of verts behind plane
         int num_behind = 0;
         for(j=prt_verts.size(); j--; ){
            const S_vector &v = verts[prt_verts[j]];
            float d = v.DistanceToPlane(pl_tri);
            if(d < BP_PLANAR_THRESH)
               ++num_behind;
         }
         if(best_num_behind < num_behind){
            best_num_behind = num_behind;
            best_tri = i;
            best_tri_pl = pl_tri;
         }
                           //optim (usual case for planar portals)
         if(num_behind == numv)
            break;
      }
   }
   if(best_tri == -1){
                           //error - too crazy shape for portal?
      assert(0);
      return;
   }

   int beg_i, end_i;
                           //collect all planar verts in both directions
   for(end_i = (best_tri+3)%numv; end_i != best_tri; (++end_i) %= numv){
      const S_vector &v = verts[prt_verts[end_i]];
      float d = v.DistanceToPlane(best_tri_pl);
      if(fabs(d) > BP_PLANAR_THRESH)
         break;
   }
   beg_i = best_tri;
   if(end_i != best_tri){
      beg_i = (best_tri + numv-1) % numv;
      do{
         const S_vector &v = verts[prt_verts[beg_i]];
         float d = v.DistanceToPlane(best_tri_pl);
         if(fabs(d) > BP_PLANAR_THRESH){
                           //this one is already wrong, go back
            ++beg_i %= numv;
            break;
         }
      }while(((beg_i += numv-1) %= numv) != end_i);
   }

                           //put existing vertices into contiguous list
   vector<word> new_prt_verts;
   int new_best_tri = -1;
   i = beg_i;
   do{
      if(i == best_tri)
         new_best_tri = new_prt_verts.size();
      new_prt_verts.push_back(prt_verts[i]);
   }while((++i %= numv) != end_i);
   int new_numv = new_prt_verts.size();

                           //process existing vertices, remove concave ones
   vector<S_vector> portal_points;

   int last_good_i = new_best_tri;
   portal_points.push_back(verts[new_prt_verts[last_good_i]]);

   for(i = new_best_tri; (++i %= new_numv) != new_best_tri; ){
      const S_vector &v0 = verts[new_prt_verts[last_good_i]];
      const S_vector &v1 = verts[new_prt_verts[i]];

                           //get normal of v01 edge
      S_vector n_edge01; n_edge01.GetNormal(v0, v1, v1+best_tri_pl.normal); n_edge01.Normalize();

                           //get plane of v01 edge
      S_plane pl01(n_edge01, -v0.Dot(n_edge01));

                           //check if all other verts are behind this plane
      for(j = new_numv; j--; ){
         const S_vector &v = verts[new_prt_verts[j]];
         float d = v.DistanceToPlane(pl01);
         if(d >= .001f)
            break;
      }
      if(j==-1){
         last_good_i = i;
         portal_points.push_back(verts[new_prt_verts[i]]);
      }
   }
   assert(portal_points.size() >= 3);

                           //build portal now
   PI3D_portal prt = new I3D_portal(this);
   if(prt->Build(portal_points))
      own_portals.push_back(prt);
   prt->Release();

   if(beg_i != end_i){
                           //put remaining points into new list and repeat process
      vector<word> new_prt_verts;

      int new_beg_i = (beg_i+1) % numv;
      for(i=(end_i+numv-1)%numv; i != new_beg_i; ++i %= numv){
         new_prt_verts.push_back(prt_verts[i]);
      }
      BuildPortals(verts, new_prt_verts);
   }

   return;
}

//----------------------------
// Collect edges, split any edges which are on line.
static void CollectEdges(const vector<S_vector> &verts, vector<I3D_triface> &tri_faces,
   vector<I3D_edge> &edges, PI3D_driver drv){

   assert(!edges.size());

   vector<S_vector> edge_dirs;

   const float angle_thresh = PI*.0005f;
   const float line_dist_thresh = .0075f;

#if 0                         //debugging support - create and paint one more face each run
   static int debug_count1 = 18;
   int debug_count = ++debug_count1;
#endif

   for(int i=tri_faces.size(); i--; ){
      const I3D_triface &fc = tri_faces[i];
      for(int j=0; j<3; j++){
         I3D_edge e(fc[(j+1)%3], fc[j]);

         int ei = FindEdge(edges.begin(), edges.size(), I3D_edge(e[1], e[0]));
         if(ei!=-1){
                              //remove existing edge
            edges[ei] = edges.back(); edges.pop_back();
            edge_dirs[ei] = edge_dirs.back(); edge_dirs.pop_back();
         }else{
            S_vector e_dir(verts[e[1]] - verts[e[0]]);
                              //compare with all other edges already in list
            for(int k=edges.size(); k--; ){
               S_vector &ed = edge_dirs[k];
               float angle = e_dir.AngleTo(ed);
               bool angle0 = (angle < angle_thresh);
               bool anglepi = ((float)fabs(angle-PI) < angle_thresh);
               if(angle0 || anglepi){
                  I3D_edge &ee = edges[k];
                              //get line segment of e on ee
                  float e_len = ed.Dot(ed);
                  assert(!IsMrgZero(e_len));
                  float u0 = (verts[e[0]] - verts[ee[0]]).Dot(ed) / e_len;
                  float u1 = (verts[e[1]] - verts[ee[0]]).Dot(ed) / e_len;
                              //first check if segments overlap
                  bool overlap;
                  if(angle0){
                     overlap = (IsAbsMrgZero(u0) || IsAbsMrgZero(1.0f-u0));
                  }else{
                     overlap = (u0 >= 0.0f && u1 <= 1.0f);
                  }
                  if(overlap){
                              //check if they're really on the same line
                     float dist = (S_vector(verts[ee[0]] + edge_dirs[k] * u0) - verts[e[0]]).Magnitude();
                     if(dist < line_dist_thresh){
                        if(angle0){
                           if(e[1]==ee[0]){
                              ee[0] = e[0];
                              ed = verts[ee[1]] - verts[ee[0]];
                              e.Invalidate();
                           }else
                           if(e[0]==ee[1]){
                              ee[1] = e[1];
                              ed = verts[ee[1]] - verts[ee[0]];
                              e.Invalidate();
                           }else
                              continue;
                           int eei = FindEdge(edges.begin(), edges.size(), I3D_edge(ee[1], ee[0]));
                           if(eei!=-1){
                              ee = edges.back(); edges.pop_back();
                              ed = edge_dirs.back(); edge_dirs.pop_back();
                              if(eei==edges.size()) eei = k;
                                                //remove existing edge
                              edges[eei] = edges.back(); edges.pop_back();
                              edge_dirs[eei] = edge_dirs.back(); edge_dirs.pop_back();
                           }
                           break;
                        }else{
#if 0                                              //debug draw new edges
                           PaintEdge(drv, verts, e);
                           PaintEdge(drv, verts, ee);
#endif
                                                   //split edges
                                                   // perform this by simply swapping indicies
                                                   // this works for all 4 possibilities of
                                                   // how these 2 edges could overlap
                                                   // (not assuming they share identical point, which shouldn't be possible due vertex-mapping made earlier)
                           if(u1 >= 1.0f)
                              swap(ee[1], e[1]);
                           else
                              swap(ee[0], e[0]);

                                                   //adjust edge dirs
                           ed = verts[ee[1]] - verts[ee[0]];
                           e_dir = verts[e[1]] - verts[e[0]];

                           if(e[0]==e[1]){
                              swap(e, ee);
                              swap(e_dir, ed);
                           }
                           if(ee[0]==ee[1]){
                              ee = edges.back(); edges.pop_back();
                              ed = edge_dirs.back(); edge_dirs.pop_back();
                           }
                           if(e[0]==e[1]){
                              e.Invalidate();
                              break;
                           }
                        }
                     }
                  }
               }
            }
            if(e.IsValid()){
               int ei = FindEdge(edges.begin(), edges.size(), I3D_edge(e[1], e[0]));
               if(ei!=-1){
                              //remove existing edge
                  edges[ei] = edges.back(); edges.pop_back();
                  edge_dirs[ei] = edge_dirs.back(); edge_dirs.pop_back();
               }else{


                  edges.push_back(e);
                  edge_dirs.push_back(e_dir);
#if 0                         //debug paint the edges
                  if(!--debug_count){
                     PaintEdge(drv, verts, e);
                  }
#endif
               }
            }
         }
      }
   }
}

//----------------------------
// Build sector and its portals from set of objects (walls).
// Phases:
//    1. collect mesh - vertices and tri_faces
//       includes also welding common vertices
//    2. collect all edges
//    3. build portals
I3D_RESULT I3D_sector::Build(PI3D_frame *walls, dword num_walls, bool simple){

   int i;
   
   vector<S_vector> verts;
   vector<I3D_triface> tri_faces;

                              //collect all edges and vertices
   CollectMeshes(walls, num_walls, verts, tri_faces);

   assert(!own_portals.size());
   if(!simple){
      sct_flags &= ~SCTF_SIMPLE;

      vector<I3D_edge> edges;

                              //build list of unique edges (those forming holes)
      CollectEdges(verts, tri_faces, edges, drv);

#if 0                         //debug paint edges
      {
         for(i=edges.size(); i--; )
            PaintEdge(drv, verts, edges[i]);
      }
#endif

//----------------------------
                              //build portals
                              // inputs:
                              //    - edges of hole(s)
      while(edges.size()){
         vector<word> prt_verts;
                              //choose starting edge
         I3D_edge e0 = edges.back(); edges.pop_back();
         word v_start = e0[0], v_curr = e0[1];
         prt_verts.push_back(v_start);
#if 0                         //debug paint edges
         {
            PaintEdge(drv, verts, e0);
         }
#endif
         S_vector start_dir(verts[v_curr] - verts[v_start]);
         S_vector last_dir = start_dir;

         const float PORTAL_ANGLE_THRESH = PI*.05f;
                              //collect chain
         for(i=edges.size(); i--; ){
            I3D_edge &e1 = edges[i];
            if(e1[0]==v_curr){
#if 0                         //debug paint edges
               {
                  PaintEdge(drv, verts, e1);
               }
#endif
                              //check angle to previous edge
               S_vector curr_dir(verts[e1[1]] - verts[v_curr]);
               float a = curr_dir.AngleTo(last_dir);
               if(a < PORTAL_ANGLE_THRESH || (float)fabs(a-PI) < PORTAL_ANGLE_THRESH){
                              //just do nothing (forget v_curr)
               }else{
                  last_dir = curr_dir;

                  prt_verts.push_back(v_curr);
               }
               v_curr = e1[1];
               e1 = edges.back(); edges.pop_back();
               if(v_curr==v_start)
                  i = 0;
               else
                  i = edges.size(); //start again
            }
         }

         assert(v_curr==v_start);   //error - incomplete hole found?
                              //check last angle
         {
            float a = start_dir.AngleTo(last_dir);
            if(a < PORTAL_ANGLE_THRESH || (float)fabs(a-PI) < PORTAL_ANGLE_THRESH){
                              //remove 1st vertex
               prt_verts[0] = prt_verts.back(); prt_verts.pop_back();
               if((prt_verts.size() >= 2) && (prt_verts[0] == prt_verts[1])){
                  prt_verts[1] = prt_verts.back(); prt_verts.pop_back();
                  prt_verts[0] = prt_verts.back(); prt_verts.pop_back();
               }
            }
         }
         if(!prt_verts.size())
            continue;
#if 0                         //debug paint edges
         {
            for(i=prt_verts.size(); i--; )
               PaintEdge(drv, verts, I3D_edge(prt_verts[i], prt_verts[(i+1)%prt_verts.size()]));
         }
#endif
         if(prt_verts.size() < 3)
            continue;

         BuildPortals(verts, prt_verts);
      }
   }else{
      sct_flags |= SCTF_SIMPLE;
   }

   CleanMesh(verts, tri_faces);

   bool b = qhCreateConvexHull(verts, NULL, &faces);
   if(!b){
      vertices.clear();
      tri_faces.clear();
      bbox.Invalidate();
      bound_planes.clear();
      return I3DERR_INVALIDPARAMS;
   }

   CleanMesh(verts, faces);

                              //save sector's mesh
   vertices = verts;
                              //build skeleton - boundbox and planes
   BuildBounds();
                              //connect to scene
   return I3D_OK;
}

//----------------------------

void I3D_sector::BuildBounds(){

                              //create bounding box
   int i;
                              //make bound box
   bbox.Invalidate();
   for(i=vertices.size(); i--; ){
      bbox.min.Minimal(vertices[i]);
      bbox.max.Maximal(vertices[i]);
   }

                              //build bound planes - keep only unique ones
   for(i=0; i < faces.size(); i++){
      const I3D_face &fc = faces[i];
      S_plane pl;

                              //get normal of plane
      if(fc.num_points == 3){
                              //simple and fast version - from 3 available points
         pl = S_plane(vertices[fc[0]], vertices[fc[1]], vertices[fc[2]]);
      }else{
                              //face with more than 3 points, we must be careful
                              // to choose best points for plane computation

                              //find 2 points with maximal distance
         int p0, p1, p2;
         float best_dist = 0.0f;
         for(int ii=fc.num_points; ii--; ){
            for(int jj=ii; jj--; ){
               S_vector dir = vertices[fc[ii]] - vertices[fc[jj]];
               float d = dir.Dot(dir);
               if(best_dist<d){
                  best_dist = d;
                  p0 = ii;
                  p1 = jj;
               }
            }
         }
                              //now find 3rd point with maximal distance to their line
         bool invert;
         S_vector dir = vertices[fc[p1]] - vertices[fc[p0]];
         best_dist = -1.0f;
         bool curr_invert = true;
         for(ii = (p0+1)%fc.num_points; ii!=p0; ++ii %= fc.num_points){
            if(ii==p1)
               curr_invert = false;
            else{
               float dtl = vertices[fc[ii]].DistanceToLine(vertices[fc[p0]], dir);
               if(best_dist<dtl){
                  best_dist = dtl;
                  invert = curr_invert;
                  p2 = ii;
               }
            }
         }
         pl.normal.GetNormal(vertices[fc[p0]], vertices[fc[p1]], vertices[fc[p2]]);
         pl.normal.Normalize();
         if(invert)
            pl.normal.Invert();
         pl.d = -vertices[fc[0]].Dot(pl.normal);
      }

      bound_planes.push_back(pl);
   }
}

//----------------------------
#define DBASE_VERSION 0x000f

bool I3D_sector::SaveCachedInfo(C_cache *ck) const{

                              //write version
   word version = DBASE_VERSION;
   ck->write((char*)&version, sizeof(version));

   {                          //write simple flag
      bool simple = (sct_flags&SCTF_SIMPLE);
      ck->write((char*)&simple, sizeof(byte));
   }
                              //write vertices
   {
      word num = vertices.size();
      ck->write((char*)&num, sizeof(num));
      ck->write((char*)vertices.begin(), sizeof(S_vector)*num);
   }
                              //write faces
   {
      word num = faces.size();
      ck->write((char*)&num, sizeof(num));
      for(int i=0; i<faces.size(); i++){
         const I3D_face &fc = faces[i];
         ck->write((char*)&fc.num_points, sizeof(byte));
         ck->write((char*)fc.points, fc.num_points*sizeof(word));
      }
   }
                              //write portals
   {
      word num = own_portals.size();
      ck->write((char*)&num, sizeof(num));
      while(num--){
         CPI3D_portal prt = own_portals[num];
         word numv = prt->NumVertices();
         ck->write((char*)&numv, sizeof(numv));
         ck->write((char*)prt->GetVertices(), sizeof(S_vector)*numv);
      }
   }
   return true;
}

//----------------------------

bool I3D_sector::LoadCachedInfo(C_cache *ck){

                              //read and check version
   word version = 0;
   ck->read((char*)&version, sizeof(version));
   if(version!=DBASE_VERSION) return false;
   
   {                          //read simple flag
      bool simple = false;
      ck->read((char*)&simple, sizeof(byte));
      sct_flags &= ~SCTF_SIMPLE;
      if(simple) sct_flags |= SCTF_SIMPLE;
   }

   {                          //read vertices
      word num = 0;
      ck->read((char*)&num, sizeof(num));
      vertices.assign(num);
      ck->read((char*)vertices.begin(), sizeof(S_vector)*num);
   }
                              //read faces
   {
      word num = 0;
      ck->read((char*)&num, sizeof(num));
      faces.assign(num);
      for(int i=0; i<num; i++){
         I3D_face &fc = faces[i];
         byte num_points;
         ck->read((char*)&num_points, sizeof(byte));
         fc.Reserve(num_points);
         ck->read((char*)fc.points, fc.num_points*sizeof(word));
      }
   }
   {
      word num = 0;
      ck->read((char*)&num, sizeof(num));
      while(num--){
         word numv = 0;
         ck->read((char*)&numv, sizeof(numv));
         vector<S_vector> v(numv);
         ck->read((char*)v.begin(), sizeof(S_vector)*numv);
         PI3D_portal prt = new I3D_portal(this);
         if(prt->Build(v))
            own_portals.push_back(prt);
         prt->Release();
      }
   }
                              //build skeleton - boundbox and planes
   BuildBounds();
   return true;
}

//----------------------------

bool I3D_sector::Connect(PI3D_scene scene){

   if(IsSimple())
      return false;

                              //find parent sector
   PI3D_frame prnt = this;
   while((prnt=prnt->GetParent1()) && prnt->GetType1()!=FRAME_SECTOR);
   PI3D_sector prnt_sect = I3DCAST_SECTOR(prnt);
                              //if no parent found, use scene's primary
   if(!prnt_sect) prnt_sect = scene->GetPrimarySector1();
                              //process all portals
   for(int i=own_portals.size(); i--; ){
      PI3D_portal prt = own_portals[i];
                              //already connected?
      if(prt->GetConnectedSector())   
         continue;
                              //look for inverse portal
      PI3D_sector *sp = scene->GetSectors();
      for(int j=scene->NumSectors(); j--; ){
         PI3D_sector sct = sp[j];
                              //don't test with itself
         if(sct==this) continue;
         const PI3D_portal *pp1 = sct->GetPortals();
         for(int k=sct->NumPortals(); k--; ){
            PI3D_portal prt1 = pp1[k];
            if(prt->Equals(prt1)){

               if(prt->GetConnectedSector())
                  prt->GetConnectedSector()->DelConnectedPortal(prt);
               if(prt1->GetConnectedSector())
                  prt1->GetConnectedSector()->DelConnectedPortal(prt1);

                              //connect portals together
               prt->ConnectTo(prt1);
               prt1->ConnectTo(prt);

               AddConnectedPortal(prt1);
               sct->AddConnectedPortal(prt);
               break;
            }
         }
         if(k!=-1)
            break;
      }
      if(j==-1){
                              //no sector found, connect with parent
         prnt_sect->AddConnectedPortal(prt);
      }
   }

   return true;
}

//----------------------------

void I3D_sector::Disconnect(PI3D_scene scene){

                              //clear connected portals
   for(int i=connected_portals.size(); i--; ){
      PI3D_portal prt = connected_portals[i];
      prt->SetConnectedSector(NULL);
      prt->ConnectTo(NULL);
   }
   connected_portals.clear();
                              //disconnect our portals
   for(i=own_portals.size(); i--; ){
      PI3D_portal prt = own_portals[i];
      prt->ConnectTo(NULL);
      if(prt->GetConnectedSector()){
         PI3D_sector sct = prt->GetConnectedSector();
         sct->DelConnectedPortal(prt);
         if(scene) sct->Connect(scene);
      }
   }
}

//----------------------------

void I3D_sector::AddConnectedPortal(PI3D_portal prt){

#ifdef _DEBUG
   int i = FindPointerInArray((void**)connected_portals.begin(), connected_portals.size(), prt);
   assert(i == -1);
#endif
   connected_portals.push_back(prt);
   prt->SetConnectedSector(this);
}

//----------------------------

void I3D_sector::DelConnectedPortal(PI3D_portal prt){

   int i = FindPointerInArray((void**)connected_portals.begin(), connected_portals.size(), prt);
   assert(i!=-1);
   prt->SetConnectedSector(NULL);
   connected_portals[i] = connected_portals.back(); connected_portals.pop_back();
}

//----------------------------

void I3D_sector::PreparePortals(const S_vector &cam_pos, const S_view_frustum &vf){

   assert(!(sct_flags&SCTF_PREPARED));
                              //prepare primary enabled portals
   for(int i=own_portals.size(); i--; ){
      PI3D_portal prt = own_portals[i];
      prt->SetState(PS_OUT_OF_VIEW);
      if(!prt->IsEnabled() || !prt->IsActive())
         continue;
                              //prepare its frustum
      bool in_front = prt->PrepareFrustum(cam_pos);
                              //check if portal is inside of view frustum
      const vector<S_vector> &prt_verts = prt->GetVertexPool();
      bool in = CheckFrustumIntersection(vf.clip_planes, 4,
         prt_verts.begin(), prt_verts.size());
      if(!in)
         continue;
      prt->SetState(in_front ? PS_FACED : PS_NOT_FACED);
   }
   sct_flags |= SCTF_PREPARED;
}

//----------------------------

bool I3D_sector::CheckPortalCollision(CPI3D_portal prt1) const{

   const vector<S_vector> &prt_verts = prt1->GetVertexPool();

   for(int phase=0; phase<2; phase++){
      const vector<C_smart_ptr<I3D_portal> > &v_portals = (!phase ? own_portals : connected_portals);
      PORTAL_STATE ps = (!phase ? PS_FACED : PS_NOT_FACED);
      CPI3D_portal *prts = (CPI3D_portal*)v_portals.begin();
      for(int pi=v_portals.size(); pi--; ){
         CPI3D_portal prt = *prts++;
         if(prt->GetState()==ps && prt->IsEnabled() && prt!=prt1 && prt->IsActive()){
            {

               const S_view_frustum &prt_vf = prt->GetFrustum();
               if(CheckFrustumIntersection(prt_vf.clip_planes, prt_vf.num_clip_planes,
                  prt_verts.begin(), prt_verts.size())){
            
                  return true;
               }
            }
         }
      }
   }
   return false;
}

//----------------------------

struct S_set_vis_link{
   S_set_vis_link *prev;      //previous link, NULL for dest sector
   PI3D_sector sct;           //sector of the link
};

//----------------------------

void I3D_sector::SetVisibleRec(const S_vector &cam_pos, const S_view_frustum &vf, S_set_vis_link *prev_link){

   assert(!IsSimple());

   sct_flags |= SCTF_VISIBLE | SCTF_DRAW_TRIGGER;

                              //transform our portals
   if(!(sct_flags&SCTF_PREPARED))
      PreparePortals(cam_pos, vf);
                              //transform all sectors of our connected portals
   for(int pi=connected_portals.size(); pi--; ){
      CPI3D_portal prt = connected_portals[pi];
      if(!prt->IsActive() || !prt->IsEnabled())
         continue;
      PI3D_sector prt_sct = prt->GetOwningSector();
      if(!(prt_sct->sct_flags&SCTF_PREPARED))
         prt_sct->PreparePortals(cam_pos, vf);
   }

   S_set_vis_link link;
   link.prev = prev_link;
   link.sct = this;
                              //set all seen sectors visible
   for(int phase=2; phase--; ){
      const PORTAL_STATE ps = !phase ? PS_NOT_FACED : PS_FACED;
      const vector<C_smart_ptr<I3D_portal> > &p_list = !phase ? own_portals : connected_portals;
      const PI3D_portal *prts = (PI3D_portal*)p_list.begin();
      for(int pi=p_list.size(); pi--; ){
         PI3D_portal prt = *prts++;
                              //only one portal of two connected is actively processed
         if(!prt->IsActive()) continue;
                              //check if it's seen into
         if(prt->GetState() != ps) continue;

         PI3D_sector next_sect = !phase ? prt->GetConnectedSector() : prt->GetOwningSector();
                              //check if not visible yet
         if(!next_sect->IsVisible())
         {
                              //check if portal seen through all linked sectors
                              // back to camera sector
            const S_set_vis_link *link_up = &link;
            while(link_up->prev){
               if(!link_up->sct->CheckPortalCollision(prt))
                  break;
               link_up = link_up->prev;
            }
                              //if we came back to camera sector, the portal is visible,
                              // and also the connected sector is
            if(!link_up->prev){
                              //recursively do the same on this
               if(!next_sect->IsVisible())
                  next_sect->SetVisibleRec(cam_pos, vf, &link);
            }else{
                              //this portal is not visible from camera,
                              // don't consider it further
               prt->SetState(PS_OUT_OF_VIEW);
            }
         }
      }
   }
}

//----------------------------
#define SCTF_RECURSE1 0x80000000

void I3D_sector::DisablePortalsRec(PI3D_sector dest_sect){

   sct_flags |= SCTF_RECURSE1;

   for(int phase=2; phase--; ){
      const vector<C_smart_ptr<I3D_portal> > &p_list = !phase ? own_portals : connected_portals;
      for(int pi=p_list.size(); pi--; ){
         PI3D_portal prt = p_list[pi];
                              //only one portal of two connected is actively processed
         if(!prt->IsActive())
            continue;

         PI3D_sector cons = !phase ? prt->GetConnectedSector() : prt->GetOwningSector();
                              //check if it's seen into
         if(prt->GetState() != PS_OUT_OF_VIEW){
            if(cons!=dest_sect && this!=dest_sect){
               const vector<S_vector> &vpool = prt->GetVertexPool();
               if(!cons->CheckIntersectionRec(vpool.begin(), vpool.size(), dest_sect) &&
                  !CheckIntersectionRec(vpool.begin(), vpool.size(), dest_sect)){
                  prt->SetState(PS_OUT_OF_VIEW);
               }
            }
         }
         if((cons->sct_flags&SCTF_PREPARED) && !(cons->sct_flags&SCTF_RECURSE1))
            cons->DisablePortalsRec(dest_sect);
      }
   }
   sct_flags &= ~SCTF_RECURSE1;
}

//----------------------------

void I3D_sector::SetVisible(const S_vector &cam_pos, const S_view_frustum &vf){

                              //recursively enable all sectors
   SetVisibleRec(cam_pos, vf, NULL);
}

//----------------------------

void I3D_sector::SolveOcclusion(const vector<class I3D_occluder*> &occ_list){

   sct_flags |= SCTF_RECURSE;
                              //for all connected visible sectors, solve visibility
   for(int phase=2; phase--; ){
      const PORTAL_STATE ps = !phase ? PS_NOT_FACED : PS_FACED;
      const vector<C_smart_ptr<I3D_portal> > &p_list = !phase ? own_portals : connected_portals;
      const PI3D_portal *prts = (PI3D_portal*)p_list.begin();
      for(int pi=p_list.size(); pi--; ){
         PI3D_portal prt = *prts++;
                              //only one portal of two connected is actively processed
         if(!prt->IsActive())
            continue;
                              //check if it's seen into
         if(prt->GetState() != ps)
            continue;
         PI3D_sector next_sect = !phase ? prt->GetConnectedSector() : prt->GetOwningSector();
         if(next_sect->IsVisible() && !(next_sect->sct_flags&SCTF_RECURSE)){
            next_sect->SolveOcclusion(occ_list);

            for(int oi=occ_list.size(); oi--; ){
               CPI3D_occluder occ = occ_list[oi];
                              //occluder works only in its sector
               if(occ->GetWorkSector() != this)
                  continue;
               if(occ->GetOccluderType1()!=I3DOCCLUDER_MESH)
                  continue;
               const S_view_frustum &of = occ->GetViewFrustum();
               const vector<S_vector> &pverts = prt->GetVertexPool();

               bool b = AreAllVerticesInVF(of, pverts.begin(), pverts.size());
               if(b){
                              //portal successfully hidden
                  prt->SetState(PS_OUT_OF_VIEW);
               }
            }
         }
      }
   }
   sct_flags &= ~SCTF_RECURSE;
}

//----------------------------

void I3D_sector::ResetRecursive(){

   assert(sct_flags&(SCTF_VISIBLE|SCTF_PREPARED));
   sct_flags &= ~(SCTF_VISIBLE | SCTF_PREPARED);
                              //reset all our linked visible sectors
   int i;
   for(i=own_portals.size(); i--; ){
      CPI3D_portal prt = own_portals[i];
      PI3D_sector con_sct = prt->GetConnectedSector();
      assert(con_sct && con_sct!=this);
      con_sct->sct_flags &= ~SCTF_PREPARED;
      //if(con_sct->IsVisible())
      if(con_sct->sct_flags&(SCTF_VISIBLE|SCTF_PREPARED))
         con_sct->ResetRecursive();
   }
   for(i=connected_portals.size(); i--; ){
      CPI3D_portal prt = connected_portals[i];
      PI3D_sector con_sct = prt->GetOwningSector();
      assert(con_sct && con_sct!=this);
      con_sct->sct_flags &= ~SCTF_PREPARED;
      //if(con_sct->IsVisible())
      if(con_sct->sct_flags&(SCTF_VISIBLE|SCTF_PREPARED))
         con_sct->ResetRecursive();
   }
}

//----------------------------

bool I3D_sector::IsBoundSphereVisible(const I3D_bsphere &bs_world, CPI3D_sector dest_sct,
   const S_vector &cam_pos, const S_preprocess_context &pc, bool *clip_vf, bool *clip_occ) const{

                              //bsphere check against view frustum
   if(clip_vf){
                              //perform bounding-sphere test
      bool in;
#ifdef USE_VF_BSPHERE
      in = SphereCollide(pc.vf_sphere, bs_world);
      if(!in)
         return false;
#endif
#if defined _DEBUG && 0
      in = SphereInFrustum(pc.exp_frustum, bs_world, *clip_vf);
      //assert(in==SphereInVF(pc.view_frustum, bs_world, *clip_vf));
#else
      in = SphereInVF(pc.view_frustum, bs_world, *clip_vf);
#endif
      if(!in)
         return false;
   }

                              //bsphere check against occluders
   if(clip_occ){
      *clip_occ = false;
      CPI3D_occluder *occs = (CPI3D_occluder*)pc.occ_list.begin();
      for(int oi=pc.occ_list.size(); oi--; ){
         CPI3D_occluder occ = *occs++;
                              //occluder works only in its sector
         if(occ->GetWorkSector() != this) continue;

         bool clip;
         bool occluded = occ->IsOccluding(bs_world, clip);
         if(occluded)
            return false;
         (*(byte*)clip_occ) |= (byte)clip;
      }
   }

                              //bsphere check against portals
   if(dest_sct!=this && !IsSimple()){
      bool in = CheckIntersectionRec(bs_world, dest_sct);
      if(!in)
         return false;
   }
   return true;
}

//----------------------------

bool I3D_sector::IsBoundVolumeVisible(const C_bound_volume &bvol, const S_matrix &tm,
   CPI3D_sector dest_sct, const S_vector &cam_pos,
   const S_preprocess_context &pc, bool *clip_vf, bool *clip_occ) const{

   if(!IsBoundSphereVisible(bvol.bsphere_world, dest_sct, cam_pos, pc, clip_vf, clip_occ))
      return false;

#ifdef ENABLE_CONTOUR_TEST
                              //bounding sphere tests failed to give clear results
                              // check if it's worth to expand bounding box
   if(!(clip_vf && *clip_vf) &&
      !(clip_occ && *clip_occ) &&
      (dest_sct==this)){
                              //no way to get out of this, we're visible
      return true;
   }

                              //expand contour of bbox as seen from camera position
   S_vector contour_points[6];
   dword num_cpts;

   ComputeContour(bvol.bound_local.bbox, tm, cam_pos, contour_points, num_cpts);
   if(!num_cpts)
      return true;

#ifdef DEBUG_PAINT_BBOX_CONTOUR
   {
      pc->render_list.scene->SetRenderMatrix(I3D_identity_matrix);
      pc->render_list.scene->DebugDrawContour(contour_points, num_cpts, S_vector(1, 0, 0));
   }
#endif

                              //contour check against view frustum
                              // only if requested, and bsphere was clipped
   if(clip_vf && *clip_vf){
      bool in = CheckFrustumIntersection(pc.view_frustum.clip_planes, 4,
         contour_points, num_cpts);
      if(!in)
         return false;
   }

                              //contour check against occluders
   if(clip_occ && *clip_occ){
      CPI3D_occluder *occs = (CPI3D_occluder*)pc.occ_list.begin();
      for(int oi=pc.occ_list.size(); oi--; ){
         CPI3D_occluder occ = *occs++;

                              //occluder works only in its sector
         if(occ->GetWorkSector() != this) continue;

         if(occ->GetOccluderType1() != I3DOCCLUDER_MESH)
            continue;

         const S_view_frustum &vf = occ->GetViewFrustum();
         bool in = AreAllVerticesInVF(vf, contour_points, num_cpts);
         if(in)
            return false;
      }
   }

                              //contour check against portals
   if(dest_sct!=this && !IsSimple()){
      bool in = CheckIntersectionRec(contour_points, num_cpts, dest_sct);
      if(!in)
         return false;
   }
#endif//ENABLE_CONTOUR_TEST

   return true;
}

//----------------------------

#ifdef ENABLE_CONTOUR_TEST

bool I3D_sector::CheckIntersectionRec(const S_vector *verts,
   dword num_verts, CPI3D_sector dest_sct) const{

   sct_flags |= SCTF_RECURSE;

   for(int phase=2; phase--; ){
      const vector<C_smart_ptr<I3D_portal> > &v_portals =
         (!phase ? own_portals : connected_portals);
      PORTAL_STATE ps = (!phase ? PS_FACED : PS_NOT_FACED);
      CPI3D_portal *prts = (CPI3D_portal*)v_portals.begin();
      for(int pi=v_portals.size(); pi--; ){
         CPI3D_portal prt = *prts++;
         if(!prt->IsActive())
            continue;
         if(prt->GetState() != ps)
            continue;
         /*
         CPI3D_sector connect_sct;
         if(!(prt->GetOwningSector()->sct_flags&SCTF_RECURSE))
            connect_sct = prt->GetOwningSector();
         else
            connect_sct = prt->GetConnectedSector();
         assert(connect_sct && connect_sct!=this);
         */
         CPI3D_sector connect_sct = phase ? prt->GetOwningSector() : prt->GetConnectedSector();
         assert(connect_sct && connect_sct!=this);
         if(!(connect_sct->sct_flags&SCTF_VISIBLE) || (connect_sct->sct_flags&SCTF_RECURSE))
            continue;

                              //check intersection on portal
         const S_view_frustum &prt_vf = prt->GetFrustum();
         if(CheckFrustumIntersection(prt_vf.clip_planes, prt_vf.num_clip_planes,
            verts, num_verts)){
                              //check if we came to dest sector
            if(connect_sct==dest_sct){
               sct_flags &= ~SCTF_RECURSE;
               return true;
            }
                              //recursively check on this sector
            if(connect_sct->CheckIntersectionRec(verts, num_verts, dest_sct)){
               sct_flags &= ~SCTF_RECURSE;
               return true;
            }
         }
      }
   }
   sct_flags &= ~SCTF_RECURSE;
   return false;
}
#endif//ENABLE_CONTOUR_TEST

//----------------------------

bool I3D_sector::CheckIntersectionRec(const I3D_bsphere &bsphere, CPI3D_sector dest_sct) const{

   sct_flags |= SCTF_RECURSE;

   for(int phase=2; phase--; ){
      const vector<C_smart_ptr<I3D_portal> > &v_portals =
         (!phase ? own_portals : connected_portals);
      PORTAL_STATE ps = (!phase ? PS_FACED : PS_NOT_FACED);
      CPI3D_portal *prts = (CPI3D_portal*)v_portals.begin();
      for(int pi=v_portals.size(); pi--; ){
         CPI3D_portal prt = *prts++;
         if(!prt->IsActive())
            continue;
         if(prt->GetState() != ps)
            continue;
         /*
         CPI3D_sector connect_sct;
         if(!(prt->GetOwningSector()->sct_flags&SCTF_RECURSE))
            connect_sct = prt->GetOwningSector();
         else
            connect_sct = prt->GetConnectedSector();
         assert(connect_sct && connect_sct!=this);
         */
         CPI3D_sector connect_sct = phase ? prt->GetOwningSector() : prt->GetConnectedSector();
         assert(connect_sct && connect_sct!=this);
         if(!(connect_sct->sct_flags&SCTF_VISIBLE) || (connect_sct->sct_flags&SCTF_RECURSE))
            continue;

                              //check intersection on portal
         const S_view_frustum &prt_vf = prt->GetFrustum();
         bool clip;
         bool in = SphereInVF(prt_vf, bsphere, clip);
         if(in){
                              //check if we came to dest sector
            if(connect_sct==dest_sct){
               sct_flags &= ~SCTF_RECURSE;
               return true;
            }
                              //recursively check on this sector
            if(connect_sct->CheckIntersectionRec(bsphere, dest_sct)){
               sct_flags &= ~SCTF_RECURSE;
               return true;
            }
         }
      }
   }
   sct_flags &= ~SCTF_RECURSE;
   return false;
}

//----------------------------

bool I3D_sector::CheckPoint(const S_vector &v){
                              //check if point is in this sector
                              //always true for primary sector
   if(IsPrimary1())
      return true;
   if(IsSimple())
      return false;

                              //put point to local coords
   S_vector v_loc = v * GetInvMatrix1();

                              //bounding-box test first
   int i;
   const S_vector &min = bbox.min;
   const S_vector &max = bbox.max;
   if(v_loc.x<min.x || v_loc.x>=max.x || v_loc.y<min.y || v_loc.y>=max.y || v_loc.z<min.z || v_loc.z>=max.z)
      return false;
                              //detailed face check
                              //sectors are convex, so we only check,
                              //if our vertex lies in front of all faces
   for(i=bound_planes.size(); i--; ){
      float d = v_loc.DistanceToPlane(bound_planes[i]);
      if(CHECK_ZERO_GREATER(d))
         return false;
   }
   return true;
}

//----------------------------

I3D_RESULT I3D_sector::AddLight(PI3D_light lp){

   vector<PI3D_light> &l = !lp->IsFogType() ? light_list : fog_light_list;
   int i = FindPointerInArray((void**)l.begin(), l.size(), lp);
   if(i!=-1)
      return I3D_OK;

   l.push_back(lp);
   lp->light_sectors.push_back(this);
   if(!lp->IsFogType())
      sct_flags |= SCTF_LIGHT_LIST_RESET;

   switch(lp->GetLightType()){
   case I3DLIGHT_FOG:
      fog_light = lp;
      break;
   case I3DLIGHT_LAYEREDFOG:
      if(!fog_light)
         fog_light = lp;
      break;
   }
   lp->MarkDirty();
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_sector::DeleteLight(PI3D_light lp){

   vector<PI3D_light> &l = !lp->IsFogType() ? light_list : fog_light_list;
   int i = FindPointerInArray((void**)l.begin(), l.size(), lp);
   if(i!=-1){
      for(int j=0; j<lp->light_sectors.size(); j++)
      if(lp->light_sectors[j] == this){
         lp->MarkDirty();
         frm_flags |= FRMFLAGS_HR_LIGHT_RESET;
         if(!lp->IsFogType())
            sct_flags |= SCTF_LIGHT_LIST_RESET;

         l[i] = l.back(); l.pop_back();
         lp->light_sectors[j] = lp->light_sectors.back(); lp->light_sectors.pop_back();
         if(lp->IsFogType()){
            if(fog_light==lp){
               fog_light = NULL;
                              //find new fog light
               for(int i=l.size(); i--; ){
                  PI3D_light lp = l[i];
                  switch(lp->GetLightType1()){
                  case I3DLIGHT_FOG:
                     fog_light = lp;
                     i = 0;
                     break;
                  case I3DLIGHT_LAYEREDFOG:
                     fog_light = lp;
                     break;
                  }
               }
            }
         }
         return I3D_OK;
      }
   }
   return I3DERR_OBJECTNOTFOUND;
}

//----------------------------

void I3D_sector::UpdateAllLights(){

   for(int j=light_list.size(); j--; )
      light_list[j]->UpdateLight();
   for(j=fog_light_list.size(); j--; )
      fog_light_list[j]->UpdateLight();

   frm_flags &= ~FRMFLAGS_HR_LIGHT_RESET;
}

//----------------------------

I3D_RESULT I3D_sector::AddSound(PI3D_sound sp){

   int i;

   i = FindPointerInArray((void**)sound_list.begin(), sound_list.size(), sp);
   if(i!=-1)
      return I3D_OK;

   sound_list.push_back(sp);
   sp->GetSectorsVector().push_back(this);

   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_sector::DeleteSound(PI3D_sound sp){

   int i = FindPointerInArray((void**)sound_list.begin(), sound_list.size(), sp);
   if(i!=-1){
      int j = FindPointerInArray((void**)sp->GetSectorsVector().begin(), sp->GetSectorsVector().size(), this);
      assert(j!=-1);

      sound_list[i] = sound_list.back(); sound_list.pop_back();
      sp->GetSectorsVector()[j] = sp->GetSectorsVector().back(); sp->GetSectorsVector().pop_back();
      return I3D_OK;
   }
   return I3DERR_OBJECTNOTFOUND;
}

//----------------------------

void I3D_sector::SetEnvironmentID(dword id){
   sound_env_id = id;
   drv->MakeEnvSettingDirty();
}

//----------------------------

void I3D_sector::Draw1(PI3D_scene scene, bool draw_skeleton, bool draw_portals) const{

   scene->SetRenderMatrix(GetIdentityMatrix());
   drv->SetTexture(NULL);

   const S_vector &cpos = scene->GetActiveCamera1()->GetWorldPos() * GetInvMatrix1();

   bool zw_en = drv->IsZWriteEnabled();
   if(zw_en) drv->EnableZWrite(false);
   drv->EnableNoCull(true);

   if(draw_skeleton && vertices.size()){
                              //visible/invisible
      static const S_vector colors[] = {
         S_vector(.5f, .5f, 1.0f), S_vector(.5f, .5f, .5f)
      };
      static const byte transps[] = { 0xe0, 0xf0 };
      bool cam_s = (scene->GetActiveCamera1()->GetCurrSector()==this);
      bool col_i = !(sct_flags&SCTF_DRAW_TRIGGER);
      sct_flags &= ~SCTF_DRAW_TRIGGER;

      {
                              //solid fill
         vector<I3D_triface> tri_faces;
         for(int i=faces.size(); i--; ){
            const I3D_face &fc = faces[i];
            I3D_triface tfc;
            tfc[0] = fc[0];
            for(int j=1; j<fc.num_points-1; j++){
               tfc[1] = fc[j];
               tfc[2] = fc[j+1];
               tri_faces.push_back(tfc);
            }
         }

         assert(tri_faces.size());
         scene->DrawTriangles(vertices.begin(), vertices.size(),
            &tri_faces[0][0], tri_faces.size() * 3,
            colors[col_i], sizeof(S_vector),
            transps[col_i]);
      }

      vector<I3D_edge> edges;
      vector<byte> e_flags;   //0 = back, 1 = contour, 2 = front
      int j;

      for(j=faces.size(); j--; ){
         const I3D_face &fc = faces[j];
         S_vector fdir = vertices[fc[0]] - cpos;
         S_vector fnormal;
         fnormal.GetNormal(vertices[fc[0]], vertices[fc[1]], vertices[fc[2]]);

         float d = fnormal.Dot(fdir);
         bool visible = (d<0.0f);
         if(cam_s) visible = !visible;

         for(int k=fc.num_points; k--; ){
            I3D_edge e(fc[k], fc[(k+1)%fc.num_points]);
            int ei = FindEdge(edges.begin(), edges.size(), I3D_edge(e[1], e[0]));
            if(ei==-1){
               edges.push_back(e);
               e_flags.push_back(visible);
            }else
               e_flags[ei] += visible;
         }

#ifdef DEBUG_DRAW_FACE_NORMALS                              
         {                    //draw normal of face
            S_vector sum(0, 0, 0);
            for(int k=fc.num_points; k--; ){
               sum += vertices[fc[k]];
            }
            sum /= fc.num_points;
            drv->DrawLine(sum, sum + bound_planes[j].normal * .5f, colors[col_i],
               j==14 ? 0 : 0x80);
         }
#endif   //DEBUG_DRAW_FACE_NORMALS
      }
                              //separate edges by their visibility
      vector<I3D_edge> separate_edges[3];
      for(j=edges.size(); j--; )
         separate_edges[e_flags[j]].push_back(edges[j]);

      byte a = transps[col_i];
                              //back
      scene->DrawLines(vertices.begin(), vertices.size(),
         &separate_edges[0].front()[0], separate_edges[0].size()*2, colors[col_i],
         sizeof(S_vector), a);
                              //front
      scene->DrawLines(vertices.begin(), vertices.size(),
         &separate_edges[2].front()[0], separate_edges[2].size()*2, colors[col_i], sizeof(S_vector),
         a*5/6);
                              //contour
      scene->DrawLines(vertices.begin(), vertices.size(),
         &separate_edges[1].front()[0], separate_edges[1].size()*2, colors[col_i], sizeof(S_vector),
         a*4/6);
   }

   if(draw_portals){
      for(int i=own_portals.size(); i--; ){ 
         CPI3D_portal prt = own_portals[i];
         if(!prt->IsActive())
            continue;

         static const S_vector colors[] = {
                              //passive
            S_vector(0, 0, 1), S_vector(1, 0, 1),
         };
         const S_vector *vp = prt->GetVertices();
                                 //fill inside
         int numv = prt->GetVertexPool().size();
         word *poly_indx = (word*)alloca(sizeof(word)*(numv-2)*3);

         for(int k=0; k<numv-2; k++){
            poly_indx[k*3+0] = 0;
            poly_indx[k*3+1] = k + 1;
            poly_indx[k*3+2] = k + 2;
         }
         bool connected = (prt->GetConnectedPortal()!=NULL);
         scene->DrawTriangles(vp, numv, poly_indx, (numv-2)*3,
            colors[!connected], sizeof(S_vector),
            0xe0);
                              //outline
         word *line_indx = (word*)alloca(sizeof(word)*numv*2);
         for(k=0; k<numv; k++){
            line_indx[k*2] = k;
            line_indx[k*2+1] = k + 1;
         }
         line_indx[k*2-1] = 0;
         scene->DrawLines(vp, numv, line_indx, k*2,
            colors[!connected], sizeof(S_vector), 0x80);

         if(!connected/* && seen*/){
                              //draw normal vector
            S_vector mid(0, 0, 0);
            for(int i=numv; i--; ) mid += vp[i];
            mid /= numv;

            scene->DrawLine(mid, mid + prt->GetPlane().normal*.25f, colors[!connected], 0);
         }
      }
   }
   if(zw_en) drv->EnableZWrite(true);
}

//----------------------------

I3D_RESULT I3D_sector::DebugDraw(PI3D_scene scene) const{

   Draw1(scene, true, true);
   return I3D_OK;
}

//----------------------------
//----------------------------
