/*--------------------------------------------------------
   Copyright (c) 1999, 2000 Michal Bacik. 
   All rights reserved.

   File: Target.cpp
   Content: Target frame.
--------------------------------------------------------*/

#include "all.h"
#include "target.h"
#include "scene.h"


//----------------------------
                              //user frame class
I3D_target::I3D_target(PI3D_driver d):
   I3D_frame(d)
{
   type = FRAME_TARGET;
   enum_mask = ENUMF_TARGET;
   drv->AddCount(I3D_CLID_TARGET);
}

//----------------------------

I3D_target::~I3D_target(){

   drv->DecCount(I3D_CLID_TARGET);
}

//----------------------------

I3D_RESULT I3D_target::Duplicate(CPI3D_frame frm){

   return I3D_frame::Duplicate(frm);
}

//----------------------------

void I3D_target::Update(){

   I3D_frame::Update();
                              //make target frame look at us
   if(frame){
      S_vector dir = GetMatrix()(3) - frame->GetWorldPos();
                              //we have world direction, rotate by inverse
                              //camera parent's matrix
      PI3D_frame target_parent = frame->GetParent1();
      if(target_parent){
         const S_matrix &m = target_parent->GetInvMatrix1();
         dir %= m;
      }
      frame->SetDir(dir, 0.0f);
      frame->Update();
   }
}

//----------------------------

void I3D_target::Draw1(PI3D_scene scene, bool strong) const{

   static const S_vector color(.9f, .9f, .9f);
   byte alpha = strong ? 0 : 0xa0;

   scene->DrawIcon(GetWorldPos(), 2, alpha, .4f);
                              //draw connection to frame
   if(frame){
      byte alpha = strong ? 0x80 : 0xd0;
      const S_vector &v0 = GetWorldPos();
      const S_vector &v1 = frame->GetWorldPos();
      scene->SetRenderMatrix(GetIdentityMatrix());
      scene->DrawLine(v0, v1, color, alpha);
   }
}

//----------------------------

I3D_RESULT I3D_target::DebugDraw(PI3D_scene scene) const{

   Draw1(scene, true);
   return I3D_OK;
}

//----------------------------
//----------------------------
