/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Texture.cpp
   Content: Texture class, loader.
--------------------------------------------------------*/

#include "all.h"
#include "texture.h"
#include <conv_rgb.hpp>

#define NONPOW2_WARNING_ON    //issue warning when textures size isn't pow2

                              //problem: VooDoo cards
#define UPLOAD_IMMEDIATELY    //load textures to vidmem as they are created

#define USE_DATABASE          //try to cache computation-expensive info in database


//#define DEBUG_LIMIT_TEXTURE_SIZE 512

//----------------------------

I3D_texture_base::I3D_texture_base(PI3D_driver d1):
   drv(d1),
   txt_flags(0),
   sort_id(d1->GetTextureSortID() & 0xffff)
{
}

//----------------------------

                              //texture class
I3D_texture::I3D_texture(PI3D_driver d):
   ref(1),
   I3D_texture_base(d),
   img_open_flags(0),
   last_render_time(0),
   create_flags(0)
{
   drv->RegisterManagedTexture(this);
   drv->AddCount(I3D_CLID_TEXTURE);
}

//----------------------------

I3D_texture::~I3D_texture(){

   drv->UnregisterManagedTexture(this);
   drv->DecCount(I3D_CLID_TEXTURE);
   //if(palette_num!=-1) drv->FreePaletteNumber(palette_num);

   /*
   if(d3d_tex){
      drv->gone_textures.push_back(I3D_driver::S_D3D_texture_gone());
      I3D_driver::S_D3D_texture_gone &gt = drv->gone_textures.back();
      gt.d3d_tex = d3d_tex;
      gt.frame_count = 5;
   }
   */
}

//----------------------------

IDirect3DTexture8 *I3D_texture::GetD3DTexture(){

   if(txt_flags&TXTF_UPLOAD_NEEDED)
      Manage(MANAGE_UPLOAD);
   last_render_time = drv->GetRenderTime();
   return d3d_tex;
}

//----------------------------

bool I3D_texture::Match(const I3D_CREATETEXTURE *ct){

                              //flags must match
   const dword match_flags = TEXTMAP_DIFFUSE | TEXTMAP_OPACITY | TEXTMAP_BUMPMAP |
      TEXTMAP_TRANSP | TEXTMAP_INVALPHA |
      TEXTMAP_NOMIPMAP | TEXTMAP_COMPRESS | TEXTMAP_HINTDYNAMIC | TEXTMAP_TRUECOLOR;
   if((create_flags&match_flags) != (ct->flags&match_flags))
      return false;

                              //check if diffuse name matches
   if(create_flags&TEXTMAP_DIFFUSE){
      if(create_flags&TEXTMAP_MEM_PTR){
         if(filenames[0] != ct->text_mem)
            return false;
      }else{
         if(stricmp(filenames[0], ct->text_name))
            return false;
      }
   }
                              //check if opacity name matches
   if(create_flags&TEXTMAP_OPACITY){
      if(stricmp(filenames[1], ct->opt_name))
         return false;
   }
                              //check if bump name matches
   if(create_flags&TEXTMAP_BUMPMAP){
      if(stricmp(filenames[2], ct->bump_name))
         return false;
   }
                              //all params match
   return true;
}

//----------------------------

bool I3D_texture::Reload(){

   if(!(create_flags&(TEXTMAP_DIFFUSE|TEXTMAP_OPACITY|TEXTMAP_BUMPMAP)))
      return false;
   if(create_flags&TEXTMAP_USEPIXELFORMAT)
      return false;
   I3D_CREATETEXTURE ct;
   memset(&ct, 0, sizeof(ct));
   ct.flags = create_flags;
   if(create_flags&TEXTMAP_DIFFUSE){
      if(create_flags&TEXTMAP_MEM_PTR)
         return false;
      else
         ct.text_name = filenames[0];
   }else
   if(create_flags&TEXTMAP_BUMPMAP){
      ct.bump_name = filenames[2];
   }
   if(create_flags&TEXTMAP_OPACITY)
      ct.opt_name = filenames[1];
   I3D_RESULT ir = Open(&ct, NULL, NULL);
   return I3D_SUCCESS(ir);
}

//----------------------------

static bool GetImageInfo(LPIGraph igraph, const C_string &s,
   dword *sx1, dword *sy1, byte *bpp1, dword *id1){

   dword sx, sy;
   byte bpp;
   dword id;
   if(igraph->GetImageInfo(s,
      sx1 ? *sx1 : sx,
      sy1 ? *sy1 : sy,
      bpp1 ? *bpp1 : bpp,
      id1 ? *id1 : id)){

      return true;
   }
   return false;
}

//----------------------------
// Find specified bitmap file in registered directories, if found,
// obtain information about it.
static bool FindBitmap(LPIGraph igraph, const vector<C_string>&mdir,
   const char *fname, C_string &s, dword &last_maps_dir,
   dword *sx1 = NULL, dword *sy1 = NULL, byte *bpp1 = NULL, dword *id1 = NULL
   ){

   if(last_maps_dir<mdir.size()){
      s = C_fstring("%s\\%s", (const char*)mdir[last_maps_dir], fname);
      if(GetImageInfo(igraph, s, sx1, sy1, bpp1, id1))
         return true;
   }

   for(int i=0; i<mdir.size(); i++){
      if(i==last_maps_dir) continue;
      s = C_fstring("%s\\%s", (const char*)mdir[i], fname);
      if(GetImageInfo(igraph, s, sx1, sy1, bpp1, id1)){
         last_maps_dir = i;
         return true;
      }
   }
   if(GetImageInfo(igraph, fname, sx1, sy1, bpp1, id1)){
      s = fname;
      return true;
   }
   return false;
}

//----------------------------

I3D_RESULT I3D_texture::Open(const I3D_CREATETEXTURE *ct, I3D_LOAD_CB_PROC *cb_load, void *cb_context){

   const char *fname = ct->text_name, *fname1 = ct->opt_name;
   dword flags = ct->flags;
   
   const D3DCAPS8 *caps = drv->GetCaps();

                              //check if textures supported by hardware
   if(!(caps->DevCaps&D3DDEVCAPS_TEXTURESYSTEMMEMORY) && !(caps->DevCaps&D3DDEVCAPS_TEXTUREVIDEOMEMORY) &&
      !(caps->DevCaps&D3DDEVCAPS_TEXTURENONLOCALVIDMEM))
      return I3DERR_TEXTURESNOTSUPPORTED;

                              //check mipmapping
   if(flags&(TEXTMAP_MIPMAP|TEXTMAP_NOMIPMAP)){
      if((drv->GetFlags()&DRVF_LOADMIPMAP) &&
         (caps->TextureCaps&D3DPTEXTURECAPS_MIPMAP)){
      }else{
         flags &= ~TEXTMAP_MIPMAP;
      }
   }
   bool empty = !(flags&(TEXTMAP_DIFFUSE|TEXTMAP_OPACITY|TEXTMAP_BUMPMAP));

                              //close previous
   d3d_tex = NULL;
   vidmem_image = NULL;
   sysmem_image = NULL;
   /*
   if(palette_num!=-1){
      drv->FreePaletteNumber(palette_num);
      palette_num = -1;
   }
   */
   txt_flags = 0;

   IDirect3DResource8 *lpRes8;
   int i;

                              //get info about image
   dword img_sx, img_sy;
   byte bpp2;

   C_string sfname_a;
   const char *sfname = NULL;
   if(!empty){
      if(flags&TEXTMAP_MEM_PTR){    //create from memory
         BITMAPINFO *bi = (BITMAPINFO*)fname;
         img_sx = bi->bmiHeader.biWidth;
         img_sy = bi->bmiHeader.biHeight;
         bpp2 = bi->bmiHeader.biBitCount/8;
         sfname = fname;
      }else{
         bool use_name0 = (flags&(TEXTMAP_DIFFUSE|TEXTMAP_BUMPMAP));
         i = FindBitmap(drv->GetGraphInterface(), drv->GetDirs(I3DDIR_MAPS),
            (use_name0 ? fname : fname1),
            sfname_a, drv->last_dir_index[I3DDIR_MAPS],
            &img_sx, &img_sy, &bpp2);
         if(!i)
            return use_name0 ? I3DERR_NOFILE : I3DERR_NOFILE1;
         sfname = sfname_a;
      }
   }else{
      img_sx = ct->size_x;
      img_sy = ct->size_y;
      bpp2 = 2;
      assert(img_sx && img_sy);
   }

#ifdef NONPOW2_WARNING_ON
   {
                              //check if dimensions power of 2
      if(cb_load && (CountBits(img_sx)!=1 || CountBits(img_sy)!=1)){
         cb_load(CBM_ERROR, (dword)(const char*)
            C_fstring(
            "Warning: texture dimensions not power of 2: '%s' (%ix%i)",
            (flags&TEXTMAP_DIFFUSE) ? fname : fname1, img_sx, img_sy),
            1, cb_context);
      }
   }
#endif

   C_string sfname1;

   if(flags&TEXTMAP_OPACITY){
                              //check if opacity map present
      if(flags&TEXTMAP_MEM_PTR){ //create from memory
         sfname1 = fname1;
      }else{
         i = FindBitmap(drv->GetGraphInterface(), drv->GetDirs(I3DDIR_MAPS), fname1, sfname1,
            drv->last_dir_index[I3DDIR_MAPS]);
         if(!i)
            return I3DERR_NOFILE1;
      }
   }

   S_pixelformat pf;
   memset(&pf, 0, sizeof(pf));

                              //choose texture format
   if(!(flags&TEXTMAP_USEPIXELFORMAT)){
      dword find_texture_flags = 0;

      if(flags&TEXTMAP_RENDERTARGET)
         find_texture_flags |= FINDTF_RENDERTARGET;

      if(flags&TEXTMAP_BUMPMAP){
         find_texture_flags |= FINDTF_BUMPMAP;
      }else{
         if(flags&TEXTMAP_OPACITY){
                              //check if texture alpha and alpha-blending
                              //supported
            if((caps->TextureCaps&D3DPTEXTURECAPS_ALPHA) &&
               (caps->SrcBlendCaps&D3DPBLENDCAPS_SRCALPHA) &&
               (caps->DestBlendCaps&D3DPBLENDCAPS_INVSRCALPHA))
            {
                              //check if same texture is used as opacity map
               if((flags&TEXTMAP_DIFFUSE) && !(flags&TEXTMAP_MEM_PTR) && !strcmp(fname, fname1)){
                  if(bpp2==1)
                     find_texture_flags = FINDTF_PALETIZED;
                  txt_flags |= TXTF_SELF_ALPHA;
               }else{
                  find_texture_flags = FINDTF_ALPHA;
               }
            }else{ 
               if(bpp2==1)
                  find_texture_flags = FINDTF_PALETIZED;
            }
         }else{
            if(bpp2==1)
               find_texture_flags = FINDTF_PALETIZED;
         }
         if(flags&TEXTMAP_TRUECOLOR){
            find_texture_flags |= FINDTF_TRUECOLOR;
                              //forget compressed format
            flags &= ~TEXTMAP_COMPRESS;
         }
                              //check if texture compression required
         if(flags&TEXTMAP_COMPRESS){
                              //may be compressed if both sides greater than 4
            if(img_sx>=4 && img_sy>=4)
               find_texture_flags |= FINDTF_COMPRESSED;
            else
               flags &= ~TEXTMAP_COMPRESS;
         }
      }

      {
         I3D_RESULT ir = drv->FindTextureFormat(pf, find_texture_flags);
         if(I3D_FAIL(ir))
            return ir;
                              //try find alpha format
                              //for colorkey substitution
         if(!(pf.flags&PIXELFORMAT_BUMPMAP))    //can't be bumpmap
         if(!(pf.flags&PIXELFORMAT_COMPRESS))   //can't be compressed format
         if(pf.bytes_per_pixel>=2)              //can't trade paletized for this
         if((flags&TEXTMAP_TRANSP) && !(pf.flags&PIXELFORMAT_ALPHA) &&
            (caps->TextureCaps&D3DPTEXTURECAPS_ALPHA)){
                              //try to find alpha-format,
                              // 1-bit for non-mipmaps, wider for mipmaps
            S_pixelformat pf1;
            ir = drv->FindTextureFormat(pf1, (flags&TEXTMAP_MIPMAP) ? FINDTF_ALPHA : FINDTF_ALPHA1);
            if(SUCCEEDED(ir) && pf1.bytes_per_pixel<=pf.bytes_per_pixel &&
               (pf1.flags&PIXELFORMAT_ALPHA) && !(pf1.flags&PIXELFORMAT_PALETTE)){

                              //use alpha-format for colorkeying
               pf = pf1;
            }
         }
                              //compressed mipmaps don't go below side size of 4 texels
         if((flags&TEXTMAP_MIPMAP) && (flags&TEXTMAP_COMPRESS)){
            if(img_sx<=4 || img_sy<=4)
               flags &= ~TEXTMAP_MIPMAP;
         }
                              //no color-keyed mipmaps
         if((flags&TEXTMAP_TRANSP) &&
            !(pf.flags&PIXELFORMAT_ALPHA) &&
            (flags&TEXTMAP_MIPMAP)){
                              //try find alpha format
                              //for colorkey substitution
            S_pixelformat pf1;
            ir = drv->FindTextureFormat(pf1,
               ((flags&TEXTMAP_MIPMAP) ? FINDTF_ALPHA : FINDTF_ALPHA1) |
               ((flags&TEXTMAP_COMPRESS) ? FINDTF_COMPRESSED : 0)
               );
            if(SUCCEEDED(ir) && //pf1.bytes_per_pixel==2 &&
               (pf1.flags&PIXELFORMAT_ALPHA) &&
               !(pf1.flags&PIXELFORMAT_PALETTE))
               pf = pf1;
            else
               flags &= ~TEXTMAP_MIPMAP;
         }
      }
   }else{
      if(!ct->pf)
         return I3DERR_INVALIDPARAMS;
      pf = *ct->pf;
   }

   bool compressed = (pf.flags&PIXELFORMAT_COMPRESS);
                              //compressed mipmaps are created in non-remapped images, put to mipmap at the end
   bool compress_postpone = (compressed && (flags&TEXTMAP_MIPMAP));

                              //get texture min/max info
   dword mintx = 1;
   dword minty = 1;
   dword maxtx = caps->MaxTextureWidth  ? caps->MaxTextureWidth  : 256;
   dword maxty = caps->MaxTextureHeight ? caps->MaxTextureHeight : 256;
#ifdef DEBUG_LIMIT_TEXTURE_SIZE
   maxtx = maxty = DEBUG_LIMIT_TEXTURE_SIZE;
#endif

   /*
                              //check if wide surfaces supported
   if(!(drv->GetDDCaps()->dwCaps2&DDCAPS2_WIDESURFACES)){
                              //limit max texture width/height to width of current screen mode
      LPIGraph igraph = drv->GetGraphInterface();
      dword sx = igraph->Scrn_sx();
      dword max_width = 1;
      while(max_width<sx) max_width <<= 1;
      max_width >>= 1;
      maxtx = Min(maxtx, max_width);
      maxty = Min(maxty, max_width);
   }
   */

   img_open_flags = 0;
   dword req_sx = img_sx, req_sy = img_sy;
                              //compute destination resolution
   if(caps->TextureCaps&D3DPTEXTURECAPS_POW2){
      size_x = mintx;
      while(size_x < req_sx){
         if(size_x >= req_sx - size_x/2) break;
         if(size_x >= maxtx) break;
         size_x *= 2;
      }
      size_y = minty;
      while(size_y < req_sy){
         if(size_y >= req_sy - size_y/2) break;
         if(size_y >= maxty) break;
         size_y *= 2;
      }
   }else{
      size_x = Max(mintx, Min(maxtx, req_sx));
      size_y = Max(minty, Min(maxty, req_sy));
   }

   if(caps->TextureCaps&D3DPTEXTURECAPS_SQUAREONLY){

      dword more = Max(size_x, size_y);
      dword less = Min(size_x, size_y);
      while(less < more){
         less *= 2;
         more = Max(more/2, 1ul);
      }
      size_y = size_x = more;
   }

   dword num_mip_levels = 0;
   if(flags&TEXTMAP_MIPMAP){
      img_open_flags |= IMGOPEN_MIPMAP;
      dword min_size = Max(mintx, minty);
      if(pf.flags&PIXELFORMAT_COMPRESS)
         min_size = Max(4, (int)min_size);
      for(dword sx=size_x>>1, sy=size_y>>1; sx >= min_size && sy >= min_size; ){
         ++num_mip_levels;
         sx >>= 1;
         sy >>= 1;
      }
   }else
   if(flags&TEXTMAP_NOMIPMAP){
      img_open_flags |= IMGOPEN_MIPMAP;
   }
                              //open from memory
   if(flags&TEXTMAP_MEM_PTR)
      img_open_flags |= IMGOPEN_MEM_PTR;

                              //if stretching necessary, postpone compression
   if(compressed){
      if(size_x!=img_sx || size_y!=img_sy)
         compress_postpone = true;
   }

   S_pixelformat pf_compress;
   if(compressed && compress_postpone){
                              //setup 8-8-8-8 format
      pf_compress.bytes_per_pixel = 4;
      pf_compress.r_mask = 0x00ff0000;
      pf_compress.g_mask = 0x0000ff00;
      pf_compress.b_mask = 0x000000ff;

      pf_compress.a_mask = 0xff000000;
      pf_compress.flags = PIXELFORMAT_ALPHA;
   }
   if(flags&TEXTMAP_HINTDYNAMIC)
      img_open_flags |= IMGOPEN_HINTDYNAMIC;

   if(flags&TEXTMAP_TRANSP)
      img_open_flags |= IMGOPEN_INIT_CKEY;

   i = drv->GetGraphInterface()->CreateImage((IImage**)&sysmem_image);
   if(!i)
      return I3DERR_OUTOFMEM;

   if(flags&TEXTMAP_BUMPMAP){
                              //open empty bumpmap texture
      i = sysmem_image->Open(NULL, IMGOPEN_EMPTY | IMGOPEN_TEXTURE | IMGOPEN_SYSMEM | img_open_flags,
         size_x, size_y, &pf, num_mip_levels);
      if(!i){
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
                              //stretch source image into proper size
      LPIImage tmp_pic;
      i = drv->GetGraphInterface()->CreateImage(&tmp_pic);
      if(!i){
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
                              //open temp image
      i = tmp_pic->Open(sfname,
         (img_open_flags&(IMGOPEN_NOREMAP | IMGOPEN_MEM_PTR)) |
         IMGOPEN_NOREMAP |
         IMGOPEN_SYSMEM,
         0, 0, NULL);
      if(!i){
         tmp_pic->Release();
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
      sysmem_image->ConvertBumpMap(tmp_pic);

      tmp_pic->Release();
      tmp_pic = NULL;
   }else
   if(!(flags&TEXTMAP_DIFFUSE)){
                              //empty image
      i = sysmem_image->Open(sfname,
         IMGOPEN_EMPTY | IMGOPEN_TEXTURE | IMGOPEN_SYSMEM | img_open_flags,
         size_x, size_y,
         compress_postpone ? &pf_compress : &pf,
         num_mip_levels);
      if(!i){
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
      if(flags&TEXTMAP_OPACITY){
                              //fill texture by white
         byte *area;
         dword pitch;
         byte bpp;
         i = sysmem_image->Lock((void**)&area, &pitch, &bpp);
         if(!i){
            sysmem_image = NULL;
            return I3DERR_GENERIC;
         }
         if((compress_postpone ? pf_compress : pf).flags&PIXELFORMAT_COMPRESS){
                              //fill linear compressed surface
            memset(area, -1, pitch);
         }else{
                              //fill line-by-line
            for(int y=size_y; y--; area += pitch)
               memset(area, -1, bpp*size_x);
         }
         sysmem_image->Unlock();
      }
   }else
   if(size_x!=img_sx || size_y!=img_sy){
                              //stretch source image into proper size
      LPIImage tmp_pic;
      i = drv->GetGraphInterface()->CreateImage(&tmp_pic);
      if(!i){
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
                              //open temp image
      i = tmp_pic->Open(sfname,
         (img_open_flags&(IMGOPEN_NOREMAP | IMGOPEN_MEM_PTR)) |
         IMGOPEN_SYSMEM,
         0, 0,
         compress_postpone ? &pf_compress : &pf);
      if(!i){
         tmp_pic->Release();
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
      //color_key = tmp_pic->GetColorKey();
      if(compress_postpone)
         pf_compress = *tmp_pic->GetPixelFormat();
                              //open empty image
      i = sysmem_image->Open(NULL,
         IMGOPEN_EMPTY | IMGOPEN_TEXTURE | IMGOPEN_SYSMEM | img_open_flags, size_x, size_y,
         compress_postpone ? &pf_compress : &pf, num_mip_levels);
      if(!i){
         tmp_pic->Release();
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
      sysmem_image->CopyStretched(tmp_pic);

      sysmem_image->SetPal(tmp_pic->GetPal());
      tmp_pic->Release();
      tmp_pic = NULL;
   }else{
                              //open texture picture directly
      dword imgopen_flags = img_open_flags;
      if(!(flags&TEXTMAP_TRANSP) && (drv->GetFlags()&DRVF_TEXTDITHER))
         imgopen_flags |= IMGOPEN_DITHER;
      if(flags&TEXTMAP_TRANSP)
         imgopen_flags |= IMGOPEN_COLORKEY;
      const S_pixelformat *imgopen_pf = compress_postpone ? &pf_compress : &pf;
      i = sysmem_image->Open(sfname, IMGOPEN_TEXTURE | IMGOPEN_SYSMEM | imgopen_flags,
         0, 0, imgopen_pf, num_mip_levels);
      if(!i){
         sysmem_image = NULL;
         return I3DERR_OUTOFMEM;
      }
   }

   if(flags&TEXTMAP_OPACITY){
      if(txt_flags&TXTF_SELF_ALPHA){
         txt_flags |= TXTF_ALPHA;
      }else{
                              //mix opacity map into texture
         LPIImage pic_opt;
         i = drv->GetGraphInterface()->CreateImage(&pic_opt);
         if(i){
            i = pic_opt->Open(sfname1, IMGOPEN_NOREMAP | IMGOPEN_SYSMEM);
            if(i){
               if(pic_opt->SizeX()!=size_x || pic_opt->SizeY()!=size_y){
                              //stretch to required size
                  LPIImage tmp_pic;
                  i = drv->GetGraphInterface()->CreateImage(&tmp_pic);
                  if(!i){
                     sysmem_image = NULL;
                     pic_opt->Release();
                     return I3DERR_OUTOFMEM;
                  }
                  i = tmp_pic->Open(NULL,
                     IMGOPEN_SYSMEM | IMGOPEN_EMPTY | IMGOPEN_NOREMAP,
                     size_x, size_y, pic_opt->GetPixelFormat(), 0);
                  if(!i){
                     tmp_pic->Release();
                     sysmem_image = NULL;
                     pic_opt->Release();
                     return I3DERR_OUTOFMEM;
                  }
                  tmp_pic->CopyStretched(pic_opt);

                  tmp_pic->SetPal(pic_opt->GetPal());
                  pic_opt->Release();
                  pic_opt = tmp_pic;
                  tmp_pic = NULL;
               }
               if(pf.flags&PIXELFORMAT_ALPHA){
                  i = sysmem_image->AlphaBlend(pic_opt,
                     ((!(flags&TEXTMAP_TRANSP) && (drv->GetFlags()&DRVF_TEXTDITHER)) ? IMGBLEND_DITHER : 0) |
                     (flags&TEXTMAP_INVALPHA ? IMGBLEND_INVALPHA : 0));
                              //success, can be alpha
                  if(i)
                     txt_flags |= TXTF_ALPHA;
               }
            }else{
                              //some problem, report to user
               if(cb_load){
                  cb_load(CBM_ERROR, (dword)(const char*)
                     C_fstring("Failed to load texture: '%s'", (const char*)sfname1),
                     1, cb_context);
               }
            }
            pic_opt->Release();
            pic_opt = NULL;
         }
      }
   }
   lpRes8 = sysmem_image->GetDirecXTexture();

   if(flags&TEXTMAP_TRANSP){
      if((pf.flags&PIXELFORMAT_COMPRESS) || (pf.flags&PIXELFORMAT_ALPHA))
         txt_flags |= TXTF_CKEY_ALPHA;
   }

   if(flags&TEXTMAP_MIPMAP){
      txt_flags |= TXTF_MIPMAP;
      sysmem_image->MixMipmap(0);
   }

   if(compress_postpone){
      LPIImage tmp_img;
      i = drv->GetGraphInterface()->CreateImage(&tmp_img);
      if(!i) return I3DERR_OUTOFMEM;

      i = tmp_img->Open(sfname, IMGOPEN_EMPTY | IMGOPEN_TEXTURE | IMGOPEN_SYSMEM | img_open_flags,
         size_x, size_y, &pf, num_mip_levels);
      assert(i);

      IDirect3DTexture8 *src = sysmem_image->GetDirecXTexture();
      assert(src);
      IDirect3DTexture8 *dst = tmp_img->GetDirecXTexture();
      assert(dst);
                              //use convertor
      LPC_rgb_conversion rgb_conv;
      drv->GetGraphInterface()->CreateRGBConv(&rgb_conv);
      rgb_conv->Init(pf);
      int sx = size_x, sy = size_y;

      dword conv_flags = MCONV_DITHER;

      dword num_levels = src->GetLevelCount();
      for(int i=0; i<num_levels; i++){
         D3DLOCKED_RECT rc_src;
         D3DLOCKED_RECT rc_dst;

         HRESULT hr = src->LockRect(i, &rc_src, NULL, D3DLOCK_READONLY);
         CHECK_D3D_RESULT("LockRect", hr);
         if(FAILED(hr))
            break;
         hr = dst->LockRect(i, &rc_dst, NULL, 0);
         CHECK_D3D_RESULT("LockRect", hr);
         if(FAILED(hr)){
            src->UnlockRect(i);
            break;
         }

         conv_flags &= ~MCONV_COLORKEY;
         if(flags&TEXTMAP_TRANSP)
            conv_flags |= MCONV_COLORKEY;
         rgb_conv->Convert(rc_src.pBits, rc_dst.pBits, sx, sy,
            rc_src.Pitch, 0, sysmem_image->GetPixelFormat()->bytes_per_pixel, NULL, conv_flags,
            0);

         src->UnlockRect(i);
         dst->UnlockRect(i);

         sx >>= 1;
         sy >>= 1;
      }
      rgb_conv->Release();
      tmp_img->CopyColorkeyInfo(sysmem_image);
      sysmem_image = tmp_img;
      tmp_img->Release();
   }

                              //debug value
   if(flags&TEXTMAP_RENDERTARGET)
      img_open_flags |= IMGOPEN_RENDERTARGET;

   if(!(flags&TEXTMAP_HINTDYNAMIC))
      img_open_flags |= IMGOPEN_HINTSTATIC;

   /*
   if(sysmem_image->GetPixelFormat()->flags&PIXELFORMAT_PALETTE){
      assert(palette_num == -1);
      palette_num = drv->AllocatePaletteNumber();
      HRESULT hr;
      hr = drv->GetDevice1()->SetPaletteEntries(palette_num, (CONST PALETTEENTRY*)sysmem_image->GetPal());
      CHECK_D3D_RESULT("SetPaletteEntries", hr);
   }
   */

   SetDirty();
#ifdef UPLOAD_IMMEDIATELY
   if(!empty){
      Manage(MANAGE_UPLOAD_IF_FREE);   //try to get onto video card, if there's some space
   }
#endif

   if(flags&TEXTMAP_DIFFUSE){
      if(flags&TEXTMAP_MEM_PTR)
         filenames[0].Assign((const char*)&fname, sizeof(fname));
      else
         filenames[0] = fname;
   }
   filenames[1] = (flags&TEXTMAP_OPACITY) ? fname1 : NULL;
   filenames[2] = (flags&TEXTMAP_BUMPMAP) ? fname : NULL;
   create_flags = flags;
   for(i=3; i--; )
      filenames[i].ToLower();

   return I3D_OK;
}

//----------------------------

void I3D_texture::Manage(E_manage_cmd cmd){

   HRESULT hr;

   switch(cmd){
   case MANAGE_FREE:
      if(vidmem_image){
         vidmem_image = NULL;
         d3d_tex = NULL;
                              //mark source dirty, so that it is updated next time
         sysmem_image->GetDirecXTexture()->AddDirtyRect(NULL);
         SetDirty();
      }
      break;
   case MANAGE_UPLOAD:
   case MANAGE_UPLOAD_IF_FREE:
   case MANAGE_CREATE_ONLY:
      if(txt_flags&TXTF_UPLOAD_NEEDED){

         if(!vidmem_image){
            drv->GetGraphInterface()->CreateImage((IImage**)&vidmem_image);
                              //open the image
            int i = vidmem_image->Open(NULL, IMGOPEN_TEXTURE | IMGOPEN_EMPTY | IMGOPEN_VIDMEM |
               img_open_flags, size_x, size_y, sysmem_image->GetPixelFormat(),
               sysmem_image->GetDirecXTexture()->GetLevelCount()-1);
            if(!i){
                              //let it be, if there's not enough space
               if(cmd==MANAGE_UPLOAD_IF_FREE){
                  vidmem_image = NULL;
                  d3d_tex = NULL;
                  break;
               }
               const vector<PI3D_texture> &mts = drv->GetManagedTextures();
                              //failed to create empty image, unload some texture(s)
                              //1. build sorted list of current cached textures
               C_sort_list<PI3D_texture> slist;
               slist.Reserve(mts.size()-1); //for all except ours
               dword render_time = drv->GetRenderTime();
               for(i=mts.size(); i--; ){
                  PI3D_texture tp = mts[i];
                  if(!(tp->txt_flags&TXTF_UPLOAD_NEEDED) && tp!=this && tp->vidmem_image)
                     slist.Add(tp, render_time - tp->last_render_time);
               }
               slist.Sort();
                              //2. step through oldiest, try to reuse, or release as necessary
               for(i=slist.Count(); i--; ){
                  PI3D_texture tp = slist[i];
                              //check if we may use its surface directly
                  if(tp->vidmem_image &&
                     tp->size_x==size_x && tp->size_y==size_y &&
                     tp->img_open_flags==img_open_flags &&
                     tp->sysmem_image->GetDirecXTexture()->GetLevelCount()==sysmem_image->GetDirecXTexture()->GetLevelCount() &&
                     !memcmp(tp->sysmem_image->GetPixelFormat(), sysmem_image->GetPixelFormat(), sizeof(S_pixelformat))){
                              //great, that's our texture, reuse it
                     vidmem_image = tp->vidmem_image;
                     tp->vidmem_image = NULL;
                     d3d_tex = tp->d3d_tex;
                     tp->d3d_tex = NULL;

                     tp->SetDirty();
                     break;
                  }
                              //unload the texture
                  tp->Manage(MANAGE_FREE);
                              //try to open image now
                  bool b = vidmem_image->Open(NULL, IMGOPEN_TEXTURE | IMGOPEN_EMPTY | IMGOPEN_VIDMEM |
                     img_open_flags, size_x, size_y, sysmem_image->GetPixelFormat(),
                     sysmem_image->GetDirecXTexture()->GetLevelCount()-1);
                  if(b)
                     break;
               }
               if(i==-1){
                              //error: cannot find free space, we're without texture
                  vidmem_image = NULL;
                  d3d_tex = NULL;
                  txt_flags &= ~TXTF_UPLOAD_NEEDED;
                  break;
               }
            }
            if(!d3d_tex){
               d3d_tex = vidmem_image->GetDirecXTexture();
            }
                              //mark dirty rectangle (entire area)
            sysmem_image->GetDirecXTexture()->AddDirtyRect(NULL);
         }
         if(cmd!=MANAGE_CREATE_ONLY){
                              //Load surface(s)
            assert(sysmem_image && vidmem_image);
            hr = drv->GetDevice1()->UpdateTexture(sysmem_image->GetDirecXTexture(),
               vidmem_image->GetDirecXTexture());
            CHECK_D3D_RESULT("UpdateTexture", hr);
         }
         txt_flags &= ~TXTF_UPLOAD_NEEDED;
      }
      break;
   }
}

//----------------------------
//----------------------------

I3D_animated_texture::I3D_animated_texture(PI3D_driver d1):
   ref(1),
   I3D_texture_base(d1),
   last_render_time(0),
   anim_delay(250),
   anim_count(0),
   curr_phase(0)
{}

//----------------------------

IDirect3DTexture8 *I3D_animated_texture::GetD3DTexture(){

   dword curr_rt = drv->GetRenderTime();
   if(last_render_time!=curr_rt){
      int time_delta = curr_rt - last_render_time;
      anim_count += time_delta;
      while(anim_count >= anim_delay){
         anim_count -= anim_delay;
         ++curr_phase %= texture_list.size();
      }
      last_render_time = curr_rt;
   }
   return texture_list[curr_phase]->GetD3DTexture();
}

//----------------------------
//----------------------------

I3D_camera_texture::I3D_camera_texture(PI3D_driver d1):
   ref(1),
   I3D_texture_base(d1)
{
   S_pixelformat pf;
   if(I3D_SUCCESS(drv->FindTextureFormat(pf, FINDTF_RENDERTARGET))){
      LPIImage img;
      drv->GetGraphInterface()->CreateImage(&img);
      I3D_RESULT ir = img->Open(NULL, IMGOPEN_TEXTURE | IMGOPEN_VIDMEM | IMGOPEN_HINTDYNAMIC | IMGOPEN_RENDERTARGET |
         IMGOPEN_MIPMAP, 256, 256, 0);
      if(I3D_SUCCESS(ir)){
         vidmem_image = img;
      }
      img->Release();
   }
}

//----------------------------

IDirect3DTexture8 *I3D_camera_texture::GetD3DTexture(){

   if(!vidmem_image)
      return NULL;
   return vidmem_image->GetDirecXTexture();
}

//----------------------------
//----------------------------
