#include <vector>
#include <c_string.hpp>

//----------------------------
#define DISABLE_HIEARCHY_FOR_NON_STATIC
#define ENABLE_SPHERE_SPACE

#define FLOAT_2_SPHERE_SPACE(n) ( FloatToInt(n/ 4.0f))

//----------------------------
//----------------------------

class I3D_volume: public I3D_frame{

#define VOLF_AUTO_GEN      1  //automatically generated (box only)
#define VOLF_RESET         2  //internal cache reset
#define VOLF_DEBUG_TRIGGER 4  //debugging - set when collided, to be painted in special color
                              //include this volumes in dynamic testing, no matter it is static.
                              //this is used when volume is static but bsp tree stil was not rebuld.
#define VOLF_EDIT_FORCE_DYNAMIC 0x20  
                              


   mutable dword vol_flags;

   I3D_VOLUMETYPE volume_type;
   PI3D_frame owner;

   S_vector non_uniform_scale;   //valid for box and rectangle types
   mutable S_matrix g_matrix;    //resulting world geometry matrix = matrix * non_uniform_scale

                              //user-defined material ID
   dword material_id;

//----------------------------
                              //pre-computed cache:

                              //sphere volume:
   mutable I3D_bsphere w_sphere;
                              //sphere-only - squared values
   mutable float radius_2, sphere_pos_dot;

                              //box volume:
   mutable S_vector normal[3], side[3];
   mutable float side_size_2[3]; //side_size^2 - for edge calculation
   mutable float side_dot[3]; //side.Dot(side)
   mutable I3D_bbox bbox;     //valid for any type, for sphere it is axis-aligned
                              //d_min and d_max are box plane distances to
                              //coplanar plane through (0,0,0)
   mutable float d_max[3], d_min[3];
//----------------------------

   mutable float sphere_space_min, sphere_space_max; //exact space occupy.
   mutable int last_min, last_max;                   //indexes of spaces in which sphere reside(computed from above)

   I3D_volume(PI3D_driver);
   ~I3D_volume();

   friend I3D_driver;        //creation

   friend void cbEnumVol(PI3D_driver, const I3D_scene*, PI3D_frame, I3D_bbox&,
      t_volume_container&, t_volume_container&, t_volume_container&);
   friend class I3D_scene;
   friend struct S_bsp_triface;      //access to geomatrix  

//----------------------------
// Get geometry matrix, build if not valid.
   const S_matrix &GetGeoMatrix() const;

public:
   t_volume_container vol_tree;
   t_volume_container dyn_vol_tree;
   PI3D_scene hr_scene; //scene on which we're active in hierarchy tree - put by I3D_scene::BuildVolumeTree

   I3D_volume &operator =(const I3D_volume&);
   void Reset();
   bool Prepare() const;
   inline bool IsPassive() const{ return (vol_flags&VOLF_AUTO_GEN); }
   void Draw1(PI3D_scene, bool strong, bool force_bsp = false, const S_vector *force_color = NULL) const;

   inline I3D_bsphere &GetBoundSphere() const{ return w_sphere; }

//----------------------------
// Get volume's bounding box.
   void GetBoundBox(PI3D_bbox) const;
   void SetupAutoBoxDyn(const t_volume_container &vol_tree, PI3D_bbox bbox);
   inline void SetupInSphereSpace() const;

                              //line - sphere
   void CheckCol_L_S_dyn(struct I3D_collision_data &cd);
                              //line - box
   bool CheckCol_L_B_dyn(struct I3D_collision_data &cd, bool passive);
                              //line - rectangle
   void CheckCol_L_R_dyn(struct I3D_collision_data &cd);
                              //sphere - sphere
   bool CheckCol_S_S_dyn(const struct S_trace_help *th) const;
                              //sphere - box
   bool CheckCol_S_B_dyn(const struct S_trace_help *th, bool passive) const;
                              //sphere - rectangle
   bool CheckCol_S_R_dyn(const struct S_trace_help *th) const;

   inline I3D_VOLUMETYPE GetVolumeType1() const{ return volume_type; }
   inline PI3D_frame GetOwner1() const{ return owner; }
   inline dword GetVolFlags() const{ return vol_flags; }

                              //overriden virtual functions
   I3DMETHOD_(void,SetOn)(dword flg);
   I3DMETHOD_(void,Update)();
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD(DebugDraw)(PI3D_scene) const;

   virtual void SetMatrixDirty(){
      Reset();
      I3D_frame::SetMatrixDirty();
   }
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD(SetNUScale)(const S_vector &scl);
   I3DMETHOD_(const S_vector&,GetNUScale)(){ return non_uniform_scale; }

   I3DMETHOD(SetVolumeType)(I3D_VOLUMETYPE t){
      volume_type = t;
      return I3D_OK;
   }
   I3DMETHOD(SetOwner)(PI3D_frame o){
      owner = o;
      return I3D_OK;
   }
   I3DMETHOD(SetScene)(PI3D_scene);
   I3DMETHOD_(void,SetCollisionMaterial)(dword id){ material_id = id; }

   I3DMETHOD_(I3D_VOLUMETYPE,GetVolumeType)() const{ return volume_type; }
   I3DMETHOD_(PI3D_frame,GetOwner)() const{ return owner; }
   I3DMETHOD_(dword,GetCollisionMaterial)() const{ return material_id; }
   I3DMETHOD_(void, SetForceDynamic)(bool b){
      if(b)
         vol_flags |= VOLF_EDIT_FORCE_DYNAMIC;
      else
         vol_flags &= ~VOLF_EDIT_FORCE_DYNAMIC;
   }

};

//----------------------------
