
#include "all.h"
#include "visual.h"
#include "volume.h"
#include "object.h"

#include "bsp.h"

                              //file support:
#define BSP_VERSION 4
                              //which types of frame can be written in file
#define FRAME_BASE_ALLOWED_TYPES (ENUMF_ALL)
//#define FRAME_BASE_ALLOWED_TYPES (ENUMF_VISUAL | ENUMF_VOLUME)

                              //building:
#define TREE_SAFE_MAX_DEPTH 2000 //cancel build if tree depth overruns this
                              //tolerance between cosines of angles, cosidered as same
#define BSP_COS_ANGLE_TOLERANCE MRG_ZERO
                              //tolerance between distances, cosidered as same
#define BSP_DISTANCE_TOLERANCE MRG_ZERO
                              //points closer then this are considered as on partion plane
#define BSP_SPLIT_TOLERANCE .001f
                              //random type of best partion selection 
#define RANDOM_PLANES 20

#define USE_BALLANCE_DEPTH 7

                              //enable count collision tests on primitives
#define VOLUME_STATISTICS

//----------------------------//build log

#define BSP_ERR_LOG(cp) \
   FILE *f_log = fopen("BspBuild.log", "wt"); \
   if(f_log){ \
      fwrite((const char *)cp, 1, strlen((const char *)cp), f_log); fwrite("\n", 1, 1, f_log); \
      fclose(f_log); \
   }


//----------------------------//debug:

#if defined _DEBUG

//#define MS_F_DEBUG(n) scene->GetDriver1()->DEBUG(n)
#define MS_F_DEBUG(n)
#define SLIDE_DEBUG(n) scene->GetDriver1()->DEBUG(n)
//#define SLIDE_DEBUG(n) 

#else

#define MS_F_DEBUG(n)
#define SLIDE_DEBUG(n)

#endif //_DEBUG

//----------------------------
// box indicies
//----------------------------

static const S_vector box_verts[] = {
   S_vector(-1, -1, -1),
   S_vector(1, -1, -1),
   S_vector(-1, 1, -1),
   S_vector(1, 1, -1),
   S_vector(-1, -1, 1),
   S_vector(1, -1, 1),
   S_vector(-1, 1, 1),
   S_vector(1, 1, 1),
};

static const word box_fill_indicies[] = {
   1, 3, 7, 1, 7, 5,
   0, 4, 6, 0, 6, 2,
   2, 6, 7, 2, 7, 3,
   0, 1, 5, 0, 5, 4,
   0, 2, 3, 0, 3, 1,
   5, 7, 6, 5, 6, 4
};

static const word box_line_indicies[] = {
   0, 1, 0, 2, 0, 4, 1, 3, 1, 5, 2, 3, 2, 6,
   3, 7, 4, 5, 4, 6, 5, 7, 6, 7
};

static const S_vector rect_edges[4] = {
   S_vector(-1.0f, -1.0f, 0.0f),
   S_vector( 1.0f, -1.0f, 0.0f),
   S_vector( 1.0f,  1.0f, 0.0f),
   S_vector(-1.0f,  1.0f, 0.0f),
};

static const word rect_fill_indicies[] = {
   0, 1, 2, 0, 2, 3, 0, 2, 1, 0, 3, 2
};

static const word rect_line_indicies[] = {
   0, 1, 1, 2, 2, 3, 3, 0
};

//----------------------------
// file support
//----------------------------

struct S_frame_base{

   vector<PI3D_frame> base;                              
   map<PI3D_frame, int> base_map; //cross refenrece
   C_smart_ptr<I3D_scene> scene;

//----------------------------

   S_frame_base(PI3D_scene sc)
   {
      scene = sc;
   }

//----------------------------
                              //look into base for frame, return id if exist, othrewise -1
   inline int FindFrameId(PI3D_frame frm)const{

      assert(frm);

      map<PI3D_frame, int>::const_iterator map_it = base_map.find(frm);
      if(map_it != base_map.end())
         return (*map_it).second;
      else
         return -1;
   }

//----------------------------
                              //add frame in base (if not there yet) and return its id.
   inline int AddFrameToBase(PI3D_frame frm){

      assert(frm);

      int id = FindFrameId(frm);

      if(id == -1){
         assert(base.size() == base_map.size());  //(must correspond)
         id = base.size();
                              //make cross - reference
         base.push_back(frm);
         pair<PI3D_frame, int> new_pair(frm, id);
         base_map.insert(new_pair);
      }
      return id;
   }

//----------------------------
                              //return id of frame. if not in base return NULL.
   inline PI3D_frame GetFrame(int id) const{

      bool is_in_base = ((id >= 0) && (id < base.size()));
      assert(is_in_base);

      return (is_in_base) ? base[id] : NULL;
   }

//----------------------------//save 

   I3D_RESULT Save(C_cache *cp){

      int bsize = base.size();
                              //write num of frame names
      cp->write((char*)&bsize, sizeof(int));
      for(int i = 0; i < bsize; i++){
                              //write size of string
         const char *name = base[i]->GetName();
         int len = strlen(name) + 1;
         assert(len < 256);
         cp->write((char*)&len, sizeof(byte));
         cp->write((char*)name, len);
      }
      return I3D_OK;
   }

//----------------------------//load

   I3D_RESULT Load(C_cache *cp, I3D_LOAD_CB_PROC *cbP, void *context){

      base.clear();
                              //create name - frames map for speed up loading
      const map<C_string, PI3D_frame> help_frame_map;
      static struct S_hlp{
         static I3DENUMRET I3DAPI cbAddFramesToMap(PI3D_frame frm, dword c){
            map<C_string, PI3D_frame> &frame_map = *(map<C_string, PI3D_frame>*)c;
            frame_map[frm->GetName()] = frm;
            return I3DENUMRET_OK;
         }
      }hlp;
      scene->EnumFrames(S_hlp::cbAddFramesToMap, (dword)&help_frame_map, FRAME_BASE_ALLOWED_TYPES);
                              //read num of frame names
      int bsize;
      cp->read((char*)&bsize, sizeof(int));
      for(int i = 0; i < bsize; i++){
                              //read size of string
         byte len;
         char buf[256];
         cp->read(&len, sizeof(byte));

         cp->read((char*)&buf, len);
         PI3D_frame frm = NULL;
         map<C_string, PI3D_frame>::const_iterator map_it;
         map_it = help_frame_map.find(C_string(buf));
         if(map_it != help_frame_map.end()){
            frm = (*map_it).second;
         }
         if(!frm){
            if(cbP){
               (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: frame '%s' not found" , buf), 0, context);
            }
            return I3DERR_OBJECTNOTFOUND;
         }
         if(!(frm->GetFlags()&I3D_FRMF_STATIC_COLLISION)){
            if(cbP){
               (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: frame '%s' is not static: ", buf), 1, context);
            }
         }

         AddFrameToBase(frm);
      }
      return I3D_OK;
   }

//----------------------------//add info from node, call recursivly on childs

   void BuildNode(const C_bsp_tree &tree, const S_bsp_node &node){

      const BSP_index *pflist = node.GetFaces();
      //for(int i = node.face_list.size(); i--;)
      for(int i = node.NumFaces(); i--;)
      {
         const dword face_id = pflist[i];
         assert(face_id < tree.GetNumFaces());
         const S_bsp_triface &tf = tree.faces[face_id];
         AddFrameToBase(tf.origin.frame);
      }
      if(node.front)
         BuildNode(tree, *node.front);
      if(node.back)
         BuildNode(tree, *node.back);
   }

//----------------------------//create frame base

   void Build(const C_bsp_tree *tree){

      assert(tree);
      if(!tree) return;

      base.clear();
      if(!tree->root){
         return;
      }else{
         BuildNode(*tree, *tree->root);
      }
   }
};

//----------------------------
//----------------------------

struct S_bsp_header{
   byte version;
   int node_count;
   S_bsp_header() :
      version(0),
      node_count(-1) //invalid tree
   {
   }
};

//----------------------------
// S_bsp_triface
//----------------------------

bool S_bsp_triface::GetVertexIndexes(S_triindex &ti)const{

   assert(origin.frame);
   if(!origin.frame)
      return false;

   switch(origin.frame->GetType1()){

   case FRAME_VISUAL:
      {
         PI3D_visual vis = I3DCAST_VISUAL(origin.frame);
         I3D_mesh_base *mb = vis->GetMesh();
         if(!mb){
            return false;
         }
                        //get all verticies on first LOD and push them to list by tri verticies groups
         //const S_vector *pick_verts = (const S_vector*)mb->vertex_buffer.Lock(D3DLOCK_READONLY);
         //dword vstride = mb->vertex_buffer.GetSizeOfVertex();
         PI3D_triface faces = mb->index_buffer.Lock(D3DLOCK_READONLY);

         if(origin.GetFaceIndex() >= mb->NumFaces())
            return false;
         const I3D_triface &fc = faces[origin.GetFaceIndex()];

         for(int kk = 0; kk < 3; kk++)
            ti.id[kk] = fc[kk];

         //mb->vertex_buffer.Unlock();
         mb->index_buffer.Unlock();
         return true;
      }
      break;

   case FRAME_VOLUME:
      {   
         PI3D_volume vol = I3DCAST_VOLUME(origin.frame);
         switch(vol->GetVolumeType()){
         case I3DVOLUME_BOX:
            {
               if(origin.GetFaceIndex() >= 12)
                  return false;
               for(int j = 0; j < 3; j++)
                  ti.id[j] = box_fill_indicies[origin.GetFaceIndex()*3 + j];
               return true;
            }
            break;
         case I3DVOLUME_RECTANGLE:
            {
               if(origin.GetFaceIndex() >= 2)
                  return false;
               for(int j = 0; j < 3; j++)
                  ti.id[j] = rect_fill_indicies[origin.GetFaceIndex()*3 + j];
               return true;
            }
            break;
         default: 
            return false; 
         } //switch
      }
      break;
   }
   return false;
}

//----------------------------

bool S_bsp_triface::Expand(){

   assert(origin.frame);
   if(!origin.frame)
      return false;

   switch(origin.frame->GetType1()){

   case FRAME_VISUAL:
      {
         PI3D_visual vis = I3DCAST_VISUAL(origin.frame);
         I3D_mesh_base *mb = vis->GetMesh();
         if(!mb){
            BSP_ERR_LOG(C_fstring("Visual '%s' with no mesh, ignored.",
               (const char*)vis->GetName()));
            return false;                 
         }

                        //get all verticies on first LOD and push them to list by tri verticies groups
         const S_matrix &m = origin.frame->GetMatrix();
         const S_vector *pick_verts = (const S_vector*)mb->vertex_buffer.Lock(D3DLOCK_READONLY);
         dword vstride = mb->vertex_buffer.GetSizeOfVertex();
         PI3D_triface faces = mb->index_buffer.Lock(D3DLOCK_READONLY);

         if(origin.GetFaceIndex() >= mb->NumFaces())
            return false;
         const I3D_triface &fc = faces[origin.GetFaceIndex()];

         for(int kk = 0; kk < 3; kk++){
            const S_vector &vx = *(const S_vector*)(((byte*)pick_verts) + fc[kk] * vstride);
            verticies[kk] = S_vector(vx.RotateByMatrix(m) + m(3));
         }

         mb->vertex_buffer.Unlock();
         mb->index_buffer.Unlock();
      }
      break;

   case FRAME_VOLUME:
      {   
         PI3D_volume vol = I3DCAST_VOLUME(origin.frame);
         const S_matrix &vol_matrix = vol->GetGeoMatrix();
         switch(vol->GetVolumeType()){
         case I3DVOLUME_BOX:
            {
               if(origin.GetFaceIndex() >= 12)
                  return false;
               for(int j = 0; j < 3; j++){
                  const S_vector &vt = box_verts[box_fill_indicies[origin.GetFaceIndex()*3 + j]];
                  verticies[j] = vt*vol_matrix;
               }
            }
            break;
         case I3DVOLUME_RECTANGLE:
            {
               if(origin.GetFaceIndex() >= 2)
                  return false;
               for(int j = 0; j < 3; j++){
                  const S_vector &vt = rect_edges[rect_fill_indicies[origin.GetFaceIndex()*3 + j]];
                  verticies[j] = vt*vol_matrix;
               }
            }
            break;
         default: 
            return false; 
         } //switch
      }
      break;

   default: 
      return false;
   } //switch

   expanded = true;
   ComputeEdges();
   return true;
}
//----------------------------

void S_bsp_triface::ComputeEdges(){

   assert(expanded);
   S_vector face_normal; 
   face_normal.GetNormal(verticies[0], verticies[1], verticies[2]);
   face_normal.Normalize();
   for(int i = 3; i--;){
      int j2 = (i+1)%3;
      edges[i] = verticies[i] - verticies[j2];
      edge_planes[i] = S_plane(verticies[i], verticies[j2], verticies[j2] + face_normal);
   }
   edges_computed = true;
}

//----------------------------

void S_bsp_triface::DebugDraw(PI3D_scene sc, dword color, bool fill)const{

//PR_BEG;
   static const word triface_fill_indicies[] = {0, 1, 2 };

   S_vector clr(((color>>16)&255)/255.0f, ((color>>8)&255)/255.0f, ((color>>0)&255)/255.0f);
   byte alpha = 255 - ((color>>24)&255);

   if(highlight){
      for(int k = 3; k--;)
         clr[k] += .2f;
      highlight = false;
   }

   assert(sc);
   sc->SetRenderMatrix(GetIdentityMatrix());
//PR_BEG;
   for(int j = 3; j--;){
      const S_vector &pA = verticies[j];
      const S_vector &pB = verticies[(j+1)%3];
      sc->DrawLine(pA, pB, clr, alpha);
   }
//PR_END;
//PR_BEG;
   if(fill){
      sc->DrawTriangles(verticies, 3, triface_fill_indicies, sizeof(triface_fill_indicies)/sizeof(word),
         clr, sizeof(S_vector), alpha*3);
   }
//PR_END;
}

//----------------------------

bool S_bsp_triface::IsPointInTriangle(const S_vector &pt)const{


   assert(expanded);
   assert(edges_computed);
                        //check if the point is inside of triangle
   float u = pt.DistanceToPlane(edge_planes[0]);
   if(CHECK_ZERO_GREATER(u)){
      //L_F_DEBUG("Fast quit 4 - point on plane out of edge 0");
      return false;
   }
   u = pt.DistanceToPlane(edge_planes[1]);
   if(CHECK_ZERO_GREATER(u)){
      //L_F_DEBUG("Fast quit 4 - point on plane out of edge 1");
      return false;
   }
   u = pt.DistanceToPlane(edge_planes[2]);
   if(CHECK_ZERO_GREATER(u)){
      //L_F_DEBUG("Fast quit 4 - point on plane out of edge 2");
      return false;
   }
   return true;
}

//----------------------------

inline dword S_bsp_triface::GetVertexChecksum() const{

   assert(expanded);
   dword sum = 0;
   for(int i = 3; i--; ){
      for(int j = 3; j --; )
         sum += (dword)verticies[i].f[j];
   }
   return sum;
}

//----------------------------
// S_bsp_origin
//----------------------------

bool S_bsp_origin::Setup(PI3D_frame frm, dword face){

   assert(frm);
   frame = frm;
   face_index = face;
   return SetGroupMaterial();
}

//----------------------------

const S_bsp_origin &S_bsp_origin::operator =(const S_bsp_origin &org){

   frame = org.frame;
   face_index = org.face_index;
   mat = org.mat;
   return *this;
}

//----------------------------

inline bool S_bsp_origin::SetGroupMaterial(){

   mat = NULL;
   if(frame->GetType1() != FRAME_VISUAL)
      return true; //ok volumes has not material

   I3D_mesh_base *mb = I3DCAST_VISUAL(frame)->GetMesh();
   assert(mb);
   if(!mb) 
      return false; //something wrong, we always need geometry info for visuals

   const I3D_face_group *fgrps = mb->GetFGroups1();
                              //find triface in meshbase
   dword curr_base_id(0);
   for(int ii = 0; ii < mb->NumFGroups1(); ii++){
      const I3D_face_group &fg2 = fgrps[ii];
      curr_base_id += fg2.NumFaces1();
      if(curr_base_id > face_index){
         mat = fg2.GetMaterial1();
         break;
      }
   }
   return true;
}

//----------------------------
// S_bsp_polygon
//----------------------------

S_bsp_polygon::S_bsp_polygon():
   valid(false)
{
}

//----------------------------

S_bsp_polygon::S_bsp_polygon(const S_bsp_polygon &p){

   verticies = p.verticies;
   origin_face = p.origin_face;
   valid = p.valid;
}

//----------------------------
const S_bsp_polygon &S_bsp_polygon::operator =(const S_bsp_polygon &p){

   verticies.clear();
   verticies = p.verticies;
   origin_face = p.origin_face;
   valid = p.valid;
   return *this;
}

//----------------------------

void S_bsp_polygon::SetVerticies(const S_vector *vp, int num_v, BSP_index face_idx){

   verticies.assign(vp, vp+num_v);
   origin_face = face_idx;
   valid = true;
}

//----------------------------

S_bsp_polygon::~S_bsp_polygon(){

   verticies.clear();
   valid = false;
}

//----------------------------
// S_bsp_node
//----------------------------

I3D_RESULT S_bsp_node::Save(C_cache *cp)const{

                              //partion plane
   cp->write((char*)&plane_id, sizeof(BSP_index));
                              //count of faces in node
   word face_cnt = NumFaces();
   cp->write((char*)&face_cnt, sizeof(word));
                              //array of faces id
   cp->write((char*)GetFaces(), face_cnt*sizeof(BSP_index));

                              //sub tree
   bool has_front = (front != NULL);
   bool has_back = (back != NULL);
   cp->write((char*)&has_front, sizeof(bool));
   cp->write((char*)&has_back, sizeof(bool));

   if(has_front)
      front->Save(cp);
   if(has_back)
      back->Save(cp);
   
   return I3D_OK;
}

//----------------------------

I3D_RESULT S_bsp_node::Load(C_cache *cp){

                              //partion plane
   cp->read((char*)&plane_id, sizeof(BSP_index));

                              //count of faces in node
   word face_cnt;
   cp->read((char*)&face_cnt, sizeof(word));
   assert(!NumFaces());

   AllocFaces(face_cnt);
   if(!GetFaces())
      return I3DERR_OUTOFMEM;

   cp->read((char*)GetFaces(), face_cnt*sizeof(BSP_index));

                              //sub tree
   bool has_front;
   cp->read((char*)&has_front, sizeof(bool));
   bool has_back;
   cp->read((char*)&has_back, sizeof(bool));

   if(has_front){
      front = new S_bsp_node;
      front->Load(cp);
   }
   if(has_back){
      back = new S_bsp_node;
      back->Load(cp);
   }
   return I3D_OK;
}

//----------------------------

void S_bsp_node::AllocFaces(int num_faces1){

   facelist = new BSP_index[num_faces1];
   num_faces = facelist ? num_faces1 : 0;
}

//----------------------------

void S_bsp_node::SetFaces(const BSP_index *p_faces, int num_faces1){

   assert(p_faces);
   AllocFaces(num_faces1);
   if(num_faces1){
      memcpy(facelist, p_faces, num_faces * sizeof(BSP_index));
   }
}


//----------------------------

dword S_bsp_node::GetNumChildren()const{

   dword ret = 1;
   if(front)
      ret += front->GetNumChildren();
   if(back)
      ret += back->GetNumChildren();
   return ret;
}

//----------------------------

void S_bsp_node::GetStatsRec(I3D_stats_bsp &st) const{

   static dword curr_depth = 0;
   curr_depth++;
                              //nodes
   st.num_nodes++;
                              //depth
   st.depth = Max(st.depth, curr_depth);
                              //faces
   st.face_references += NumFaces();
                              //memory
   st.mem_used += sizeof(S_bsp_node);
   st.mem_used += NumFaces() * sizeof(BSP_index);
                              //children
   if(front)
      front->GetStatsRec(st);
   if(back)
      back->GetStatsRec(st);

   curr_depth--;
}

//----------------------------

S_bsp_node::S_bsp_node() : front(NULL), 
   back(NULL),
   facelist(NULL),
   num_faces(0)
{
}
//----------------------------

S_bsp_node::~S_bsp_node(){

   if(front){
      delete front;
   }
   if(back){
      delete back;
   }
   if(facelist){
      delete[] facelist;
      facelist = NULL;
   }
}

//----------------------------
// C_bsp_tree
//----------------------------

C_bsp_tree::C_bsp_tree() : 
   root(NULL), scene(NULL), valid(false)
{
}

//----------------------------

C_bsp_tree::~C_bsp_tree(){

   scene = NULL;
   Clear();
}

//----------------------------

I3D_RESULT C_bsp_tree::Save(C_cache *cp){

                              //header
   S_bsp_header hdr;
   hdr.version = BSP_VERSION;
   hdr.node_count = IsValid() ? GetNumNodes() : -1;

   cp->write((char*)&hdr, sizeof(S_bsp_header));

   if(!IsValid())             //invalid tree, save only header
      return I3DERR_GENERIC;
                              //save names of all used frames as table, first name has id 0
   assert(scene);
   S_frame_base frame_base(scene);
   frame_base.Build(this);

   I3D_RESULT ir1 = frame_base.Save(cp);
   if(I3D_FAIL(ir1))
      return ir1;
                              //planes array
   dword planes_num = planes.size();
   cp->write((char*)&planes_num, sizeof(dword));
   const S_plane *p_planes = planes.begin();
   for(dword i = 0; i < planes.size(); i++){
      const S_plane &curr_plane = p_planes[i];
      cp->write((char*)&curr_plane, sizeof(S_plane));
   }
                              //face array
   dword faces_num = faces.size();
   cp->write((char*)&faces_num, sizeof(dword));
   const S_bsp_triface *p_faces = faces.begin();
   for(int j = 0; j < faces_num; j++){
      const S_bsp_triface &tf = p_faces[j];
                              //face plane
      cp->write((char*)&tf.plane_id, sizeof(BSP_index));
                              //plane_side
      cp->write((char*)&tf.plane_side, sizeof(bool));
                              //origin
      int frame_id = frame_base.FindFrameId(tf.origin.frame);
      assert(frame_id != -1);      //must be in frame base now
      cp->write((char*)&frame_id, sizeof(int));
      cp->write((char*)&tf.origin.GetFaceIndex(), sizeof(dword));
                              //edges slide flags
      cp->write((char*)&tf.slide_flags, sizeof(byte));
                              //write checksum
      cp->write((char*)&tf.vertex_checksum, sizeof(dword));
   }

                              //tree structure
   I3D_RESULT ir;
   if(!root){
      ir = I3D_OK;
   }
   else{
      ir = root->Save(cp);
   }
   return ir;
}

//----------------------------

bool C_bsp_tree::CompareBspChecksum(I3D_LOAD_CB_PROC *cbP, void *context) const{

   const S_bsp_triface *p_faces = faces.begin();
   for(int i = faces.size(); i--; ){
      const S_bsp_triface &tf = p_faces[i];
      int delta = tf.GetVertexChecksum() - tf.vertex_checksum;
      if(abs(delta) > 2){
         if(cbP){
            const char *fail_name = tf.origin.frame ? tf.origin.frame->GetName() : "NULL";
            (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: checksum failed(frame %s).",
            fail_name), 0, context);
         }
         return false;
      }
   }
   return true;
}

//----------------------------

I3D_RESULT C_bsp_tree::Load(C_cache *cp, I3D_LOAD_CB_PROC *cbP, void *context, dword check_flags){

#ifdef _DEBUG
   /*
   dword fpu_cw = _control87(0, 0);
   assert((fpu_cw&_MCW_PC) == _PC_24);
   */
#endif
   _control87(_PC_24, _MCW_PC);
                              //header
   S_bsp_header hdr;

   cp->read((char*)&hdr, sizeof(S_bsp_header));

                              //check version
   if(hdr.version != BSP_VERSION){
      if(cbP)
         (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: different version detected (%i), rebuild tree.", hdr.version), 0, context);
      SetAbsentVolumesPickable(); //fail expected olny in editor, set volumes pickable
      return I3DERR_FILECORRUPTED;
   }
                              //invalid tree saved(should not happen)
   if(hdr.node_count == -1){
      SetAbsentVolumesPickable(); //fail expected olny in editor, set volumes pickable
      return I3DERR_FILECORRUPTED;
   }

   S_frame_base frame_base(scene);
                              //read frames names id table
   I3D_RESULT ir = frame_base.Load(cp, cbP, context);
   if(I3D_FAIL(ir)){
      SetAbsentVolumesPickable(); //fail expected olny in editor, set volumes pickable
      return ir;
   }
                              //planes array
   assert(!planes.size());
   dword planes_num;
   cp->read((char*)&planes_num, sizeof(dword));
   planes.assign(planes_num);
   cp->read((char*)planes.begin(), planes_num*sizeof(S_plane));
                              //face array
   dword faces_num;
   cp->read((char*)&faces_num, sizeof(dword));
   faces.reserve(faces_num);
   for(int j = 0; j < faces_num; j++){
      faces.push_back(S_bsp_triface());
      S_bsp_triface &tf = faces.back();
                              //face plane
      cp->read((char*)&tf.plane_id, sizeof(BSP_index));
                              //plane_side
      cp->read((char*)&tf.plane_side, sizeof(bool));
                              //origin
      int frame_id;
      cp->read((char*)&frame_id, sizeof(int));
      PI3D_frame frm1 = frame_base.GetFrame(frame_id);
      assert(frm1);
      if(!frm1){
         if(cbP)
            (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: missing frame, id %i.", frame_id), 0, context);
         return I3DERR_OBJECTNOTFOUND;
      }
      dword face1;

      cp->read((char*)&face1, sizeof(dword));
      if(!tf.origin.Setup(frm1, face1)){
         (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: origin.Setup failed, frame %s.", (const char*)frm1->GetName()), 0, context);
         return I3DERR_FILECORRUPTED;
      }
      bool b1 = tf.Expand();
      if(!b1){
         SetAbsentVolumesPickable(); //fail expected olny in editor, set volumes pickable
         return I3DERR_OBJECTNOTFOUND;
      }
                              //edges slide flags
      cp->read((char*)&tf.slide_flags, sizeof(byte));
               //check vertex sum (allowed only on flag, for support editor, but should not be set in final game, where bsp IS correct)
      cp->read((char*)&tf.vertex_checksum, sizeof(dword));
      /*if(check_flags&FBSPLOAD_CHECKSUM){
         //if(tf.GetVertexChecksum() != tf.vertex_checksum)
         int delta = tf.GetVertexChecksum() - tf.vertex_checksum;
         if(abs(delta) > 2){
            if(cbP)
               (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: checksum failed(frame %s).",
               (const char*)frm1->GetName()), 0, context);

            SetAbsentVolumesPickable(); //fail expected olny in editor, set volumes pickable
            return I3DERR_OBJECTNOTFOUND;
         }
      }*/
   }

   if(check_flags&FBSPLOAD_CHECKSUM){
      if(!CompareBspChecksum(cbP, context)){
         SetAbsentVolumesPickable(); //fail expected onlny in editor, set volumes pickable
         return I3DERR_OBJECTNOTFOUND;
      }
   }

                              //now all nodes
   if(!hdr.node_count){
      ir = I3D_OK;
   }
   else{
      root = new S_bsp_node;
      ir = root->Load(cp);
   }

   if(I3D_FAIL(ir)){
      SetAbsentVolumesPickable(); //fail expected olny in editor, set volumes pickable
      return ir;
   }

                        //check consistency(slowdown loading, don't set flag in final game)
   if(cbP && check_flags&FBSPLOAD_CONSISTENCE){
      ReportConsistancy(cbP, context, &frame_base);
   }

   valid = true;
   return I3D_OK;
}

//----------------------------

void C_bsp_tree::Clear(){

   valid = false;
   planes.clear();
   faces.clear();
                              //reset driver stats
   if(scene && scene->GetDriver1())
      memset(&scene->GetDriver1()->bsp_stats, 0, sizeof(scene->GetDriver1()->bsp_stats));
   if(root){
      delete root;
      //delete[] root;
      root = NULL;
   }
}

//----------------------------
// input:   plane index, ray interval
// output:  ray_origin near/far side(FRONT/BACK)
// Return signed distance along ray to plane.

float C_bsp_tree::ClassifyRayInterval(BSP_index plane_indx, float ray_min, float ray_max, E_PLANE_SIDE &near_side, E_PLANE_SIDE &far_side, const I3D_collision_data &cd) const{

   float signed_dist;         //ray-plane intersection distance.

   assert(plane_indx < planes.size());
   const S_plane &partion_plane = planes[plane_indx];
                              //coplanar test
   float costh = partion_plane.normal.Dot(cd.normalized_dir);
   bool ray_is_parallel = IsAbsMrgZero(costh);
   if(ray_is_parallel){
      signed_dist = -1e+16f;     //we need zero less value for future process
   }else{
      signed_dist =  -(partion_plane.normal.Dot(cd.from) + partion_plane.d) / costh;
   }

                              //determinate start point space against partion plane
   float point_side = cd.from.DistanceToPlane(partion_plane);
                              //store tree of half-space which contain ray_start
   if(point_side < -MRG_ZERO){
      near_side = PS_BACK;
      far_side = PS_FRONT;
   }else
   if(point_side > MRG_ZERO){
      near_side = PS_FRONT;
      far_side = PS_BACK;
   }else{
      near_side = PS_ON;
      if(!ray_is_parallel){
                              //angle between ray direction and partion_plane normal determinate 
                              //side where is any ohter point then ray_origin
          far_side = (- costh > MRG_ZERO) ? PS_FRONT : PS_BACK;
      }else{
                              //ray lay on plane and is parallel, select any one side
         far_side = PS_FRONT;
      }
   }       
   return signed_dist;
}

//----------------------------

inline bool IsChildOf(PI3D_frame frm, PI3D_frame root){

   do{
      if(frm == root)
         return true;
   }while(frm = frm->GetParent());
   return false;
}

//----------------------------

bool C_bsp_tree::FaceListItersection(const S_bsp_node &node, const float part_dist, I3D_collision_data &cd) const{

                              //we use one point for test, because faces are coincident with partion plane
                              //but, there is some error bounded by BSP_SPLIT_TOLERANCE
   const S_vector partion_intersection_point = cd.from + cd.normalized_dir * part_dist;

                              //proces all polygons
   const BSP_index *p_flist = node.GetFaces();
   const S_bsp_triface *p_faces = faces.begin();
   const S_plane *p_planes = planes.begin();
   for(int i = node.NumFaces(); i--;)
   {

#ifdef VOLUME_STATISTICS
      ++scene->GetDriver1()->col_stats[I3D_driver::DRV_COL_STATS_L_F];
#endif
      BSP_index curr_face_id = p_flist[i];
      const S_bsp_triface &tf = p_faces[curr_face_id];

      assert(tf.origin.frame);

      if((!(cd.flags&I3DCOL_INVISIBLE) && !tf.origin.frame->IsOn1()) ||
         (tf.origin.frame->GetFrameFlags()&I3D_FRMF_NOCOLLISION)){
         continue;
      }

      bool invert_side = !tf.plane_side;

                              //back face culling
      {
                              //lazy, can be improved (we alredy computed required information before)
         bool two_side_mat = false;
         assert(tf.plane_id < planes.size());
         const S_plane &f_plane = p_planes[tf.plane_id];
         if(tf.origin.frame->GetType1() == FRAME_VOLUME){
            if(I3DCAST_VOLUME(tf.origin.frame)->GetVolumeType() == I3DVOLUME_RECTANGLE)
               two_side_mat = true;
         }else
         {
            PI3D_material mat = tf.origin.GetMaterial();
            if(mat)
               two_side_mat = (mat->Is2Sided1() || (cd.flags&I3DCOL_FORCE2SIDE));
         }

         float f_side = cd.from.DistanceToPlane(f_plane);
         bool back_side = CHECK_ZERO_LESS(f_side);
         if((back_side && tf.plane_side) || (!back_side && !tf.plane_side)){
            if(!two_side_mat){
                                 //back face culling
               continue;
            }
            invert_side = !invert_side;
         }
      }

      const S_vector &intersection_point = partion_intersection_point;
      float dist = part_dist;

      tf.highlight = true;
      bool b = tf.IsPointInTriangle(intersection_point);

      if(b){

         if(cd.frm_ignored && IsChildOf(tf.origin.frame, cd.frm_ignored)){
            continue;
         }
                              //check texture transparent pixels, if requested
         if(cd.flags&I3DCOL_TEXTURE){
            if(!TextureIntersection(tf, dist, cd)){
               continue;
            }
         }
                              //return normal of hit face, not partion plane, altrhough they should be near same
         assert(tf.plane_id < planes.size());
         const S_plane face_plane = p_planes[tf.plane_id];

                              //check if we want this
         if(cd.cresp_line_old){
            bool user_accept = cd.cresp_line_old(face_plane.normal, dist, tf.origin.frame, cd.cresp_context, NULL);
            if(!user_accept){
               return false;
            }
         }
         if(cd.cresp_face){
                              //store last valid values
            S_vector last_normal = cd.hit_normal;
            PI3D_frame last_frame = cd.closest_frm;
            float last_dist  = cd.closest_hit;
                              //setup values for callback
            S_vector hit_norm = face_plane.normal;
            if(invert_side)
               hit_norm.Invert();
            cd.SetReturnValues(dist, tf.origin.frame, &hit_norm);
            bool user_accept = cd.cresp_face(cd);
            if(!user_accept){
                              //restore last values if user not accept this collision
               cd.SetReturnValues(last_dist, last_frame, &last_normal);
               return false;
            }
         }
                              //setup return values
         cd.closest_frm = tf.origin.frame;
         cd.closest_hit = dist;
         cd.hit_normal = face_plane.normal;
         if(invert_side)
            cd.hit_normal.Invert();
         return true;
      }
   }
   return false;
}

//----------------------------

bool C_bsp_tree::TextureIntersection(const S_bsp_triface &tf, const float hit_dist, I3D_collision_data &cd) const{

   bool rtn = true;
   bool invert_side = false;

   assert(tf.origin.frame);
   switch(tf.origin.frame->GetType1()){
   case FRAME_VISUAL:
      break;
   case FRAME_VOLUME:
      return true;            //no texture check on volumes
   default: 
      assert(0);
   }
   assert(tf.origin.frame->GetType1() == FRAME_VISUAL);
   PI3D_visual vis = I3DCAST_VISUAL(tf.origin.frame);
   assert(vis);
   I3D_mesh_base *mb = vis->GetMesh();
   if(!mb)
      return true;            //no mesh, skip test a considere it as sucess

                              //verticies, face groups, faces, texture coord
   const S_vector *pick_verts = (const S_vector*)mb->vertex_buffer.Lock(D3DLOCK_READONLY);
   dword vstride = mb->vertex_buffer.GetSizeOfVertex();
                              //assumption: texture coordinates are at end of vertex
   const I3D_text_coor *txt_coords = (const I3D_text_coor*)(((byte*)pick_verts) + vstride - sizeof(I3D_text_coor));

   PI3D_triface trifaces = mb->index_buffer.Lock(D3DLOCK_READONLY);

   PI3D_material mat = tf.origin.GetMaterial();
   assert(mat);
                     //detailed texture check
   PI3D_texture_base tb = mat->GetTexture1();
   if(tb && (mat->I3D_material::IsCkey() || mat->IsCkeyAlpha1() || cd.cresp_texel)){

      bool two_side_mat;
      two_side_mat = (mat->Is2Sided1() || (cd.flags&I3DCOL_FORCE2SIDE));
                                 //find triface in meshbase
      const I3D_triface &fc = trifaces[tf.origin.GetFaceIndex()];
      float u1, u2;

      const S_vector &v0 = *(const S_vector*)(((byte*)pick_verts) + fc[0] * vstride);
      const S_vector &v1 = *(const S_vector*)(((byte*)pick_verts) + fc[1] * vstride);
      const S_vector &v2 = *(const S_vector*)(((byte*)pick_verts) + fc[2] * vstride);
      S_vector edge_dir = v2 - v1;

//----------------------------
      const S_matrix &matrix = vis->GetMatrix();
      const S_matrix &m_inv = vis->GetInvMatrix1();

                                 //inverse-map vectors to local coords
      S_vector loc_from = cd.from * m_inv;
      S_vector loc_to = (cd.from + cd.normalized_dir * hit_dist) * m_inv;
      S_vector loc_dir_normalized = loc_to - loc_from;
      float loc_hit_dist = loc_dir_normalized.Magnitude();
      if(IsMrgZero(loc_hit_dist))
         return false;
      loc_dir_normalized /= loc_hit_dist;

                              //intersection point in local coord
      assert(tf.plane_id < planes.size());
//----------------------------
                              //compute plane in local coord
      S_plane face_plane;
      face_plane.normal.GetNormal(v0, v1, v2);
      face_plane.normal.Normalize();
      face_plane.d = -v0.Dot(face_plane.normal);

//----------------------------//computation from world plane
//      S_plane face_plane2 = planes[tf.plane_id];
//      if(!tf.plane_side)
//         face_plane2.Invert();
//      face_plane2 = face_plane2 * m_inv;
//----------------------------

      S_vector f_normal = face_plane.normal;
      
                              //intersection point in local coord
      float f_dist = loc_from.DistanceToPlane(face_plane);
      if(CHECK_ZERO_LESS(f_dist)){
         assert(two_side_mat); //backface culling performed in ray-face test, can' be single side material test
         invert_side = true; 
      }
      float u = f_normal.Dot(loc_dir_normalized);
      assert(!IsAbsMrgZero(u)); //parrallel tast performed in ray-face test, can't be parallel
      float curr_dist = - f_dist / u;
      S_vector point_on_plane = loc_from + loc_dir_normalized * curr_dist;
                              //computation from world point
      //S_vector point_on_plane2 = (*ipt) * m_inv; 

//----------------------------

      S_vector p_dir = point_on_plane - v0;
      LineIntersection(v0, p_dir, v1, edge_dir, u1, u2);
      if(!IsAbsMrgZero(u1)){
         S_vector pp = v0 + p_dir * u1;
                     //get uv on edge point
         float e_len = edge_dir.Sum();
         if(!IsAbsMrgZero(e_len)){
            u2 = (pp - v1).Sum() / e_len;
            I3D_text_coor tex;

            const I3D_text_coor &uv0 = *(const I3D_text_coor*)(((byte*)txt_coords) + fc[0] * vstride);
            const I3D_text_coor &uv1 = *(const I3D_text_coor*)(((byte*)txt_coords) + fc[1] * vstride);
            const I3D_text_coor &uv2 = *(const I3D_text_coor*)(((byte*)txt_coords) + fc[2] * vstride);

            tex.u = uv1.u + (uv2.u - uv1.u) * u2;
            tex.v = uv1.v + (uv2.v - uv1.v) * u2;

            u1 = 1.0f / u1;
            tex.u = uv0.u + (tex.u - uv0.u) * u1;
            tex.v = uv0.v + (tex.v - uv0.v) * u1;

            tex.u = fmod(tex.u, 1.0f);
            tex.v = fmod(tex.v, 1.0f);
            if(tex.u<0.0f) tex.u += 1.0f;
            if(tex.v<0.0f) tex.v += 1.0f;

            if(mat->I3D_material::IsCkey() || mat->IsCkeyAlpha1()){
                           //look on texel
               const byte *rle = tb->GetRLEMask();
               assert(rle);
               dword sx = tb->SizeX1(), sy = tb->SizeY1();
               int x = Min((int)(sx * tex.u), (int)sx-1);
               int y = Min((int)(sy * tex.v), (int)sy-1);
               const byte *mem = &rle[*(dword*)(&rle[y*4])];
               bool full = false;
               do{
                  byte b = *mem++;
                  if(!b){     //eol
                     full = false;
                     break;
                  }
                  full = (b&0x80);
                  x -= (b&0x7f);
               }while(x>=0);
               if(!full)
                  rtn = false;
                  goto before_return;
            }
                                    //ask user if we furher process this
            if(cd.cresp_texel_old){
               I3D_texel_collision tc = {tex.u, tex.v, tb, mat};
                                    //due to this, we need to put hit data
                                    // into world coords,
                                    // performance is not critical, because this
                                    // feature is expected to be called bu light-mapping only
               S_vector hit_normal = f_normal.RotateByMatrix(matrix);
               if(invert_side)
                  hit_normal.Invert();
               S_vector col_pt = (loc_from + loc_dir_normalized * curr_dist) * matrix;
               float hit_dist = (col_pt - cd.from).Magnitude();
               if(!cd.cresp_texel_old(hit_normal, hit_dist, tf.origin.frame, cd.cresp_context, &tc)){
                  rtn = false;
                  goto before_return;
               }
            }
            if(cd.cresp_texel){
               bool texel_accept = cd.cresp_texel(cd);
               if(!texel_accept){
                  rtn = false;
                  goto before_return;
               }
            }
         }
      }
   }
before_return:
   mb->vertex_buffer.Unlock();
   mb->index_buffer.Unlock();
   return rtn;
}

//----------------------------

bool C_bsp_tree::NodeIntersection(const S_bsp_node &node, float ray_min, float ray_max, I3D_collision_data &cd) const{

   const S_bsp_node *tree_near = NULL; //child of Node for half-space containing the origin of Ray
   const S_bsp_node *tree_far = NULL;  //the "other" child of Node -- i.e. not equal to near

                              //signed distance along Ray to plane defined by Node
   E_PLANE_SIDE near_side, far_side;
   float dist = ClassifyRayInterval(node.plane_id, ray_min, ray_max, near_side, far_side, cd);
   assert(near_side != far_side); //debug
   tree_near = (near_side == PS_FRONT) ? node.front : (near_side == PS_BACK) ? node.back : NULL;
   tree_far = (far_side == PS_FRONT) ? node.front : (far_side == PS_BACK) ? node.back : NULL;


   //if(CHECK_ZERO_LESS(dist) || (dist > ray_max))
   //if(CHECK_ZERO_LESS(dist) || (dist > ray_max+MRG_ZERO))
   if(CHECK_ZERO_LESS(dist) || (dist > ray_max+BSP_SPLIT_TOLERANCE)){
                              //whole interval is on near side
      if(tree_near){
         bool hit = NodeIntersection(*tree_near, ray_min, ray_max, cd);
         return hit;
      }
   }else
   if(dist < ray_min){
                              //whole interval is on far side
      if(tree_far){
         bool hit = NodeIntersection(*tree_far, ray_min, ray_max, cd);
         return hit;
      }
   }else{
                              //the interval intersects the plane
      if(tree_near){          //Test near side
         bool hit = NodeIntersection(*tree_near, ray_min, dist, cd);
         if(hit)                   // if there was a hit return pick_data
            return hit;
      }
                           //test faces in node (on partion plane)
      if(FaceListItersection(node, dist, cd))
         return true;
                              //Test far side
      if(tree_far){
         bool hit =  NodeIntersection(*tree_far, dist, ray_max, cd);
         return hit;
      }
   }
   return false;
}

//----------------------------

bool C_bsp_tree::RayIntersection(I3D_collision_data &cd) const{

   if(!root)
      return false;

   float test_length = (cd.flags&I3DCOL_RAY) ? 1e+16f : cd.dir_magnitude;

                              //optimizing: shrink line by last closest_hit, this speed up test in bsp tree
   if(cd.GetClosestHit() < test_length){
      test_length = cd.GetClosestHit();
   }

   return NodeIntersection(*root, .0f, test_length, cd);
}

//----------------------------
// SPHERE_SHIFTING
//----------------------------
// Classify sphere move against partion plane. th contain move information,
// in_dist, out_dist return distances on move line where sphere touch the plane, 
// near_side inform where origin of move lie. (far side is the other)
// return TRUE if move direction facing to plane
bool C_bsp_tree::Classify_MS(BSP_index plane_indx, const S_trace_help &th, float &in_dist, float &out_dist, E_PLANE_SIDE &near_side)const{

   assert(plane_indx < planes.size());
   const S_plane &plane = planes[plane_indx];

                              //distance of move origin to partion plane
   float orig_dist = th.from.DistanceToPlane(plane);
                              //facing of move dir and plane normal
   float cosh1 = plane.normal.Dot(th.dir_norm);

   if(orig_dist < -MRG_ZERO){
      near_side = PS_BACK;
   }
   else
   if(orig_dist > MRG_ZERO){
      near_side = PS_FRONT;
   }
   else{
      if(CHECK_ZERO_LESS(cosh1))
         near_side = PS_FRONT;
      else
         near_side = PS_BACK;
   }

                              //classify move
   if(IsAbsMrgZero(cosh1)){
                           //move is parallel to plane, check distance
      if(fabs(orig_dist) >= th.radius){
                           //never touch the plane
         in_dist = .0f;
         out_dist = .0f;
      }
      else{
         out_dist = -1.e+8f;
      }
   }
   else{
                           //shift plane by radius
      //float f = plane.normal.Dot(from1) + plane.d;
      const float &f = orig_dist;
      float min_d = -(f + th.radius) / cosh1;
      float max_d =  -(f - th.radius) / cosh1;
      if(min_d > max_d){
         swap(min_d, max_d);
      }
      in_dist = min_d;
      out_dist = max_d;
   }

   return CHECK_ZERO_LESS(cosh1);
}

//----------------------------

bool C_bsp_tree::FaceListItersection_MS(const S_bsp_node &node, float near_dist, float far_dist, const S_trace_help &th)const{

   bool ret = false;
   const BSP_index *pflist = node.GetFaces();
   const S_bsp_triface *pfaces = faces.begin();
                              //FIXME: from and to are tested as entire original
                              //segmet due risc that zero length dir could accure.
                              //this require recomputing sphere/plane move line clip
                              //which was actually computed in tree traversal,so
                              //can we handle it better?
                              //NOTE:while we re-testing original segment in CheckCol_MS_F
                              //we get as output valid in/out distaces, 
                              //do not add them to near_dist!
   for(int i = node.NumFaces(); i--;)
   {

#ifdef VOLUME_STATISTICS
      ++scene->GetDriver1()->col_stats[I3D_driver::DRV_COL_STATS_S_F];
#endif

      BSP_index curr_face_id = pflist[i];
      const S_bsp_triface &tf = pfaces[curr_face_id];
                              //todo: pass info from header to CheckCol_MS_F, avoid recomputation.
//PR_BEG;
      bool b = CheckCol_MS_F(tf, th, curr_face_id);
//PR_END;
      if(b){
         tf.highlight = true;
      }
      ret |= b;
   }
   return ret;
}

//----------------------------

enum E_CLIP_LINE{
   //CL_OUTSIDE = -1,
   CL_INSIDE_NOCLIP = 0,
   CL_INSIDE_CLIP_MIN = 1,
   CL_INSIDE_CLIP_MAX = 2,
   CL_OUTSIDE = 3,
};

//----------------------------
// Same as ClipLineByPlane but used precomputed values plane_dist and cosh1. Used in tree traverse.
// cosh1 = pl_norm.Dot(dir);
// plane_dist = pos_dot + pl_d;
inline E_CLIP_LINE ClipLineByPlaneCompact(float plane_dist, float cosh1, float &dmin, float &dmax){

#define EDGE_TOLERANCE .001f

   if(IsAbsMrgZero(cosh1)){
                              //line is parallel to plane
      if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, entire line inside, no clipping 
         return CL_INSIDE_NOCLIP;
      }else{
                              //origin outside, entire line invalid
         dmax = .0f;
         return CL_OUTSIDE;
      }
   }else{
      float dist = -(plane_dist) / cosh1;

      if(CHECK_ZERO_LESS(dist)){
                              //clip plane before origin, entire line behind plane
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, segment valid, no clipping
            return CL_INSIDE_NOCLIP;
         }else{
                              //origin outside, entire line invalid 
            dmax = .0f;
            return CL_OUTSIDE;
         }
      }else
      if(dist < dmin){
                              //clip plane between origin and min
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, entire segment outside, invalid
            dmax = .0f;
            return CL_OUTSIDE;
         }else{
                              //origin outside, entire segment valid
            return CL_INSIDE_NOCLIP;
         }
      }else
      if(dist < dmax){
                              //clip plane between min and max
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin and min is inside, clip max
            dmax = dist;
            return CL_INSIDE_CLIP_MAX;
         }else{
                              //origin and min is outside, clip min
            dmin = dist;
            return CL_INSIDE_CLIP_MIN;
         }
      }else{
                              //clip plane after max
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, entire line valid, no clipping
            return CL_INSIDE_NOCLIP;
         }else{
                              //origin outside, entire line invalid
            dmax = .0f;
            return CL_OUTSIDE;
         }
      }
   }
}

//----------------------------
// Return segment of line clipped by plane, ie part which is before plane.
// Line is defined by pos, dir and tested segment in which are we interested in(dmin, dmax).
// Plane is defined by pl_normal and pl_d.
// If line is clipped, dmin and dmax are adjuster as retunr values.
inline E_CLIP_LINE ClipLineByPlane(const S_vector &pos, const S_vector &dir, const S_vector &pl_norm, float pl_d, float &dmin, float &dmax){

   float cosh1 = pl_norm.Dot(dir);
   float plane_dist = pl_norm.Dot(pos) + pl_d;

   return ClipLineByPlaneCompact(plane_dist, cosh1, dmin, dmax);
}

//----------------------------

/*inline E_CLIP_LINE ClipLineByPlane(const S_vector &pos, const S_vector &dir, const S_vector &pl_norm, float pl_d, float &dmin, float &dmax){

#define EDGE_TOLERANCE .001f

   float cosh1 = pl_norm.Dot(dir);
   float plane_dist = pl_norm.Dot(pos) + pl_d;
   if(IsAbsMrgZero(cosh1)){
                              //line is parallel to plane
      if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, entire line inside, no clipping 
         return CL_INSIDE_NOCLIP;
      }else{
                              //origin outside, entire line invalid
         dmax = .0f;
         return CL_OUTSIDE;
      }
   }else{
      float dist = -(plane_dist) / cosh1;

      //if(dist < .0f)
      if(CHECK_ZERO_LESS(dist)){
                              //clip plane before origin, entire line behind plane
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, segment valid, no clipping
            return CL_INSIDE_NOCLIP;
         }else{
                              //origin outside, entire line invalid 
            dmax = .0f;
            return CL_OUTSIDE;
         }
      }else
      if(dist < dmin){
                              //clip plane between origin and min
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, entire segment outside, invalid
            dmax = .0f;
            return CL_OUTSIDE;
         }else{
                              //origin outside, entire segment valid
            return CL_INSIDE_NOCLIP;
         }
      }else
      if(dist < dmax){
                              //clip plane between min and max
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin and min is inside, clip max
            dmax = dist;
            return CL_INSIDE_CLIP_MAX;
         }else{
                              //origin and min is outside, clip min
            dmin = dist;
            return CL_INSIDE_CLIP_MIN;
         }
      }else{
                              //clip plane after max
         if(CHECK_ZERO_LESS(plane_dist)){
                              //origin inside, entire line valid, no clipping
            return CL_INSIDE_NOCLIP;
         }else{
                              //origin outside, entire line invalid
            dmax = .0f;
            return CL_OUTSIDE;
         }
      }
   }
}*/


//----------------------------
//----------------------------
// cosh1 = pl_norm.Dot(dir);
// plane_dist = pos_dot + pl_d;
bool C_bsp_tree::NodeIntersection_MS(const S_bsp_node &node, float ray_min, float ray_max, const S_trace_help &th)const{

   bool any_hit = false;
   const S_plane &plane = planes[node.plane_id];

   const float plane2dir_dot = plane.normal.Dot(th.dir_norm);
   const float pos_dot = plane.normal.Dot(th.from);

   //if(plane.normal.Dot(th.dir_norm) > MRG_ZERO)
   if(plane2dir_dot > MRG_ZERO){
      float shift_up_d = plane.d - th.radius;
      float dmin1 = ray_min;
      float dmax1 = ray_max;
      //E_CLIP_LINE cl1 = ClipLineByPlane(th.from, th.dir_norm, plane.normal, shift_up_d, dmin1, dmax1);
      E_CLIP_LINE cl1 = ClipLineByPlaneCompact(pos_dot + shift_up_d, plane2dir_dot, dmin1, dmax1);

      bool part_under = (cl1 != CL_OUTSIDE);
      if(part_under){
         if(node.back){
            bool b = NodeIntersection_MS(*node.back, dmin1, dmax1, th);
            any_hit |= b;
         }
      }
      float dmin2 = ray_min;
      float dmax2 = ray_max;
      float shift_down_d = -(plane.d + th.radius);
      //E_CLIP_LINE cl2 = ClipLineByPlane(th.from, th.dir_norm, -plane.normal, shift_down_d, dmin2, dmax2);
      E_CLIP_LINE cl2 = ClipLineByPlaneCompact(-pos_dot + shift_down_d, -plane2dir_dot, dmin2, dmax2);
      bool part_over = (cl2 != CL_OUTSIDE);
      if(part_over){
         if(node.front){
            bool b = NodeIntersection_MS(*node.front, dmin2, dmax2, th);
            any_hit |= b;
         }
      }

      if(part_under && part_over){
         bool b = FaceListItersection_MS(node, ray_min, ray_max, th);
         any_hit |= b;
      }
   }else{
      float shift_up_d = -plane.d - th.radius;
      float dmin1 = ray_min;
      float dmax1 = ray_max;
      //E_CLIP_LINE cl1 = ClipLineByPlane(th.from, th.dir_norm, -plane.normal, shift_up_d, dmin1, dmax1);
      E_CLIP_LINE cl1 = ClipLineByPlaneCompact(-pos_dot + shift_up_d, -plane2dir_dot, dmin1, dmax1);
      bool part_under = (cl1 != CL_OUTSIDE);
      if(part_under){
         if(node.front){
            bool b = NodeIntersection_MS(*node.front, dmin1, dmax1, th);
            any_hit |= b;
         }
      }
      float dmin2 = ray_min;
      float dmax2 = ray_max;
      float shift_down_d = plane.d - th.radius;
      //E_CLIP_LINE cl2 = ClipLineByPlane(th.from, th.dir_norm, plane.normal, shift_down_d, dmin2, dmax2);
      E_CLIP_LINE cl2 = ClipLineByPlaneCompact(pos_dot + shift_down_d, plane2dir_dot, dmin2, dmax2);
      bool part_over = (cl2 != CL_OUTSIDE);
      if(part_over){
         if(node.back){
            bool b = NodeIntersection_MS(*node.back, dmin2, dmax2, th);
            any_hit |= b;
         }
      }
      if(part_under && part_over){
         bool b = FaceListItersection_MS(node, ray_min, ray_max, th);
         any_hit |= b;
      }
   }
   return any_hit;
}

//----------------------------

inline bool MakeBevelPlane(const S_vector &norm1, const S_vector &norm2, const S_vector &ip, float radius, S_plane &bevel){

                              //bevel is neaded when planes intersects in angel greater than 90 deg
   float f = norm1.Dot(norm2);
   //if(CHECK_ZERO_LESS(f))
                              //put some tolerance here because we use this function also for edge beveling, which are above 90deg.
   if(f < .001f){             
                              //bevel normal is half-way between intersections planes normals
      bevel.normal = norm1+norm2;
      bevel.normal.Normalize();
                              //distance from ip to bevel plane must be equal to radius, so bevel's 'd' is determinated by DistanceToPlane formula
                              //radius == ip.Dot(pl.normal) + pl.d; planes facing outside of triangel, so distance to plane is negative.
      bevel.d = -radius - ip.Dot(bevel.normal);
      return true;
   }
   else{
      return false;
   }
}

//----------------------------

byte C_bsp_tree::GetSlideNormal(const S_bsp_triface &tf, const S_plane &pl, const S_vector &ip, S_vector &rn) const{

   byte clip_flags = 0;
                              //calculate closest point on face
   float dist = ip.DistanceToPlane(pl);
   S_vector point_on_plane = ip - pl.normal*dist;
                              //check point against edges
   bool out_edge[3] = {false, false, false};
   int num_clips = 0;
   int clip_index[2];
   for(int i = 3; i--; ){
      if(tf.slide_flags&(1<<i)){
         const S_plane &ep = tf.edge_planes[i];
         float d = point_on_plane.DistanceToPlane(ep);
         if(d > .0f){
            out_edge[i] = true;
            clip_flags |= (1<<i);
            assert(num_clips < 2);
            clip_index[num_clips] = i;
            num_clips++;
         }
      }
   }
   S_vector closest_point;
   if(clip_flags){
      assert(num_clips);
      if(num_clips == 1){
                              //recompute closest point - on the edge
         int cp_indx = clip_index[0];
         S_vector e_dir = -tf.edges[cp_indx];
         const S_vector &e_point = tf.verticies[cp_indx];
         float d = ip.PositionOnLine(e_point, e_dir);
         d = Max(.0f, Min(d, 1.0f));
         closest_point = e_point + e_dir * d;
      }else{
         int cp_indx;
         if(abs(clip_index[0]-clip_index[1])==1)   //they're (0 and 1) or (1 and 2) - use point of greater
            cp_indx = Max(clip_index[0], clip_index[1]);
         else        //they're 0 and 2
            cp_indx = 0;
         closest_point = tf.verticies[cp_indx];
      }
      rn = S_normal(ip-closest_point);
   }else
      rn = pl.normal;
   return clip_flags;
}

//----------------------------
// Face is id of currently tested triface(S_bsp_triface doesn't contain this information yet)
bool C_bsp_tree::CheckCol_MS_F(const S_bsp_triface &tf, const S_trace_help &th, BSP_index face_id) const{

#if 0     //debug
   if(tf.origin.GetFaceIndex() != 0)
      return false;
#endif

   assert(tf.origin.frame);
   if(!tf.origin.frame)
      return false;
                              //dont' test invisible frames
   if((!(th.flags&I3DCOL_INVISIBLE) && !tf.origin.frame->IsOn()) || 
      (tf.origin.frame->GetFrameFlags()&I3D_FRMF_NOCOLLISION))
      return false;

   bool two_sided = false;

   float in_dist = .0f;

   if(IsAbsMrgZero(th.dir_magn))
      return false;
   const S_vector &from1 = th.from;
   const S_vector &dir1 = th.dir_norm;
   float out_dist = th.dir_magn;
   
                              //clip segment by face plane, shifted by radius
   S_plane face_plane = planes[tf.plane_id];
   if(!tf.plane_side)
      face_plane.Invert();

                              //check side, make it facing to from point
   {
      float ff = from1.Dot(face_plane.normal) + face_plane.d;
      bool other_side = CHECK_ZERO_LESS(ff);
      if(other_side){
         if(th.flags&I3DCOL_FORCE2SIDE){
            two_sided = true;
         }else{
                              //make it two sided for rectangle volume
            if(tf.origin.frame->GetType1() == FRAME_VOLUME){
               if(I3DCAST_VOLUME(tf.origin.frame)->GetVolumeType()==I3DVOLUME_RECTANGLE)
                  two_sided = true;
            }else{
                              //make it two sided for two sided materials
               PI3D_material mat = tf.origin.GetMaterial();
               if(mat)
                  two_sided = mat->Is2Sided1();
            }
         }
         if(!two_sided){
            MS_F_DEBUG("sphere move 0.1: backface quit");
            return false;
         }
         face_plane.Invert();
      }
   }
                              //backface cull - don't stick in plane, which you are leaving
   {
      float f11 = dir1.Dot(face_plane.normal);
      if(CHECK_ZERO_GREATER(f11)){
         MS_F_DEBUG("sphere move 0.2: backface quit");
         return false;
      }
   }
                              //chop move segment to touch plane part only
   {
      float cosh1 = face_plane.normal.Dot(dir1);
      if(fabs(cosh1)<MRG_ZERO){
                              //move is parallel to plane, check distance
         float dist1 = from1.DistanceToPlane(face_plane);
         if(fabs(dist1) >= th.radius)
         {
            MS_F_DEBUG("sphere move 1: parallel to face plane, distance > radius");
            return false;
         }
      }
      else{
         float f = face_plane.normal.Dot(from1) + face_plane.d;
         float min_d = -(f + th.radius) / cosh1;
         float max_d =  -(f - th.radius) / cosh1;
         if(min_d > max_d){
            float tmp_dist = min_d;
            min_d = max_d;
            max_d = tmp_dist;
         }

         if((min_d) >= out_dist){
            MS_F_DEBUG("sphere move 1.1: move don't touch face plane");
            return false;
         }
         if(max_d <= in_dist){
            MS_F_DEBUG("sphere move 1.2: move don't touch face plane");
            return false;
         }

         in_dist = Max(min_d, in_dist);
         out_dist = Min(max_d, out_dist);
      }
   }
//----------------------------

                              //clip line by edges planes
   for(int i=3; i--; ){
                              //edges facing outside of triangle
      const S_plane &ed_plane = tf.edge_planes[i];
                              //shrink litte radius to prevent multiple collisions on edges
      float ed_plane_d = ed_plane.d - (th.radius-.001);
      E_CLIP_LINE cr = ClipLineByPlane(from1, dir1, ed_plane.normal, ed_plane_d, in_dist, out_dist);
      if(cr == CL_OUTSIDE){
                              //entire segment out of edge plane
         MS_F_DEBUG(C_fstring("sphere move 2: line segment out of edge plane %i", i));
         return false;
      }
   }/**/

   S_plane bevel;
                              //apply edges beveling
   for(i=3; i--; ){
      const S_vector &shared_point = tf.verticies[i];
      const S_vector &norm1 = tf.edge_planes[i].normal;
      bool b = MakeBevelPlane(face_plane.normal, norm1, shared_point, th.radius, bevel);
      if(b){
         E_CLIP_LINE cr = ClipLineByPlane(from1, dir1, bevel.normal, bevel.d, in_dist, out_dist);
         if(cr == CL_OUTSIDE){
            MS_F_DEBUG(C_fstring("sphere move 3: line segment clipped by edge bevel %i", i));
            return false;
         }
      }

   }/**/
                              //counterside edges beveling
   for(i=3; i--; ){
      const S_vector &shared_point = tf.verticies[i];
      const S_vector &norm1 = tf.edge_planes[i].normal;
      S_vector invert_face = -face_plane.normal;
      bool b = MakeBevelPlane(invert_face, norm1, shared_point, th.radius, bevel);
      if(b){
         E_CLIP_LINE cr = ClipLineByPlane(from1, dir1, bevel.normal, bevel.d, in_dist, out_dist);
         if(cr == CL_OUTSIDE){
            MS_F_DEBUG(C_fstring("sphere move 3.5: line segment clipped by counterside edge bevel %i", i));
            return false;
         }
      }

   }/**/

                              //apply corners points beveling
   for(i=3; i--; ){
      const S_vector &shared_point = tf.verticies[i];
      const S_vector &norm1 = tf.edge_planes[i].normal;
      const S_vector &norm2 = tf.edge_planes[(i+2)%3].normal;
      bool b = MakeBevelPlane(norm1, norm2, shared_point, th.radius, bevel);
      if(b){
         int cr = ClipLineByPlane(from1, dir1, bevel.normal, bevel.d, in_dist, out_dist);
         if(cr == CL_OUTSIDE){
            MS_F_DEBUG(C_fstring("sphere move 4: line segment clipped by point bevel %i", i));
            return false;
         }
      }
   }/**/

//----------------------------
                              //compute slide normal
   S_vector ipact_point = th.from + th.dir_norm*in_dist;
   S_vector slide_normal;
   byte slide = GetSlideNormal(tf, face_plane, ipact_point, slide_normal);
//----------------------------
                              //allow slide rectangle corner:
                              //if we collide with rectangle face which IS clipped,
                              //search for 2nd face and remove it if not clipped.
                              //if we collide with rectangle face which NOT clipped,
                              //push it only when 2nd face is not clipped or presented
   int replace_id = -1;
   if((tf.origin.frame->GetType() == FRAME_VOLUME) && (I3DCAST_VOLUME(tf.origin.frame)->GetVolumeType() == I3DVOLUME_RECTANGLE)){
      S_collision_info *p_col = th.collision_list.begin();
      for(int i1 = th.collision_list.size(); i1--;){       
                              //we need also check volume_source 'cos different spheres can collide with same face in different distance
         if((p_col[i1].hit_frm == tf.origin.frame) && (p_col[i1].vol_source == th.vol_source))
         {
            if(p_col[i1].face_id == face_id)
               return true; //we alredy have this face in list(should not happen)
            else{
               if(slide){
                              //if previous face was NOT clipped, replace it
                  if(!p_col[i1].slided)
                     replace_id = i1;
               }else{
                              //do not store non clipped face if clipped is already stored.
                  if(p_col[i1].slided)
                     return true;
               }
            }
            break; //only 2 different faces expected for rectangle, end search
         }
      }
   }

                              //collect all collisions in the way
   if(replace_id == -1){
      th.collision_list.push_back(S_collision_info());
   }
   S_collision_info &ci = (replace_id == -1) ? th.collision_list.back() : th.collision_list[replace_id];

   ci.plane.d = face_plane.d;
   ci.plane.normal = slide ? slide_normal : face_plane.normal;
   ci.face_normal = face_plane.normal;
   ci.modified_face_normal = true;

   ci.slided = slide;
                              //shift by radius
   ci.plane.d -= th.radius;
   ci.hit_frm = tf.origin.frame;
   ci.vol_source = th.vol_source;
   ci.plane_id = tf.plane_id;
   ci.face_id = face_id;
   ci.in_dist = in_dist;

   return true;
}

//----------------------------

void C_bsp_tree::SetAbsentVolumesPickable() const{

   vector<PI3D_frame> static_list;
   E_BSP_BUILD_RESULT br1 = scene->GetStaticFrames(static_list);
   if(br1 != BSP_BUILD_OK)
      return;

   S_frame_base temp_base(scene);
   temp_base.Build(this);

   const PI3D_frame *p_static_list = static_list.begin();
   for(int i = static_list.size(); i--; ){

      PI3D_frame frm = p_static_list[i];
      switch(frm->GetType1()){
      case FRAME_VOLUME:
         bool is_in_tree = (temp_base.FindFrameId(frm) != -1);
         I3DCAST_VOLUME(frm)->SetForceDynamic(!is_in_tree);
         break;
      }
   }
   scene->ResetDynamicVolTree(); //reset tree to make changes visible
}

//----------------------------

void C_bsp_tree::ReportConsistancy(I3D_LOAD_CB_PROC *cbP, void *context, const S_frame_base *hlp_base)const{

   assert(cbP);
   if(!cbP) return;

   assert(scene);
   if(!scene) return;

   vector<PI3D_frame> static_list;
   E_BSP_BUILD_RESULT br1 = scene->GetStaticFrames(static_list);
   if(br1 != BSP_BUILD_OK)
      return;

   S_frame_base *temp_base = NULL;

   if(!hlp_base){
      temp_base = new S_frame_base(scene);
      temp_base->Build(this);
   }

   const S_frame_base &fb_help = (hlp_base) ? *hlp_base : *temp_base;

   const PI3D_frame *p_static_list = static_list.begin();
   for(int i = static_list.size(); i--; ){

      if(fb_help.FindFrameId(p_static_list[i]) == -1){

         if(p_static_list[i]->GetType1() == FRAME_VOLUME){
            PI3D_volume vol = I3DCAST_VOLUME(p_static_list[i]);
            vol->SetForceDynamic(true);
         }
         if(cbP){
            const char *frm_name = p_static_list[i]->GetName();
            (*cbP)(CBM_ERROR, (dword)(const char*)C_fstring("bsp: static frame '%s' missing in tree", (const char*)frm_name), 1, context);
         }
      }
   }
   if(temp_base){
      delete[] temp_base;
   }
}

//----------------------------

void C_bsp_tree::GetStats(I3D_stats_bsp &st) const{

                              //memory:
   memset(&st, 0, sizeof(st));
                              //class size
   st.mem_used += sizeof(C_bsp_tree);
                              //face_cache
   st.mem_used += (face_cache.Size()/CELL_SIZE)*sizeof(dword); //(face_cache.alloc_size)
                              //planes
   st.mem_used += planes.size() * sizeof(S_plane);
                              //faces
   st.mem_used += faces.size() * sizeof(S_bsp_triface);
                              //nodes
   if(root)
      root->GetStatsRec(st);
                              //other:
   st.num_planes = planes.size();          
   st.num_faces = faces.size();         
}

//----------------------------
// C_bsp_tree_builder
//----------------------------

enum C_bsp_tree_builder::E_PARTION_RESULT{

   PR_COINCIDE,
   PR_FRONT,
   PR_BACK,
   PR_INTERSECT,
};

//----------------------------

void C_bsp_tree_builder::S_build_stat::Reset(){

   polygon_rest = 0;
   curr_depth = 0;
}

//----------------------------

void C_bsp_tree_builder::AddFacePlane(S_bsp_triface &tf, dword c){

   static const float map_trunc_mask = 1e4f;

   assert(c);
   multimap<long int, int> &hmap = *(multimap<long int, int>*) c;
   assert(tf.IsValid());
   S_plane face_plane(tf.verticies[0], tf.verticies[1], tf.verticies[2]);

   assert(!IsAbsMrgZero(face_plane.normal.Magnitude()));

                              //check if equal plane exist in planes list

   bool same_side;            //if face normal is oriented as plane normal; valid only if will be found plane.
   int find_id = -1;
   const S_plane *plane_p = planes.begin();

   long int search_d = (long int)(face_plane.d*map_trunc_mask);
                              //optimising: find equal d sequence in map
   for(int second_test = 0; second_test < 2; second_test++){
      if(second_test){
                              //second test other side
         search_d = - search_d;
      }
      for(multimap<long int, int>::const_iterator it_plus = hmap.find(search_d); it_plus != hmap.end(); it_plus++){
         if((*it_plus).first > search_d)
            break;               //out of tolerance
         int curr_id = (*it_plus).second;
         const S_plane &curr_plane = plane_p[curr_id];
                                 //check angle
         float cosh1 = face_plane.normal.Dot(curr_plane.normal);

         float cos_zero_delta = fabs(cosh1)-1.0f;
         if(fabs(cos_zero_delta) > BSP_COS_ANGLE_TOLERANCE)
            continue;
         same_side = CHECK_ZERO_GREATER(cosh1);
         float d_delta = (same_side) ? (curr_plane.d - face_plane.d) : (curr_plane.d + face_plane.d);

         if(fabs(d_delta) < BSP_DISTANCE_TOLERANCE)
         {
                                 //this test check if face is really cooplanar with plane
            S_bsp_polygon pp;
            pp.SetVerticies(tf.verticies, 3, -1);
            E_PARTION_RESULT pr = ClassifyPolygon(pp, curr_id);
            if(pr!=PR_COINCIDE)
               continue;
            find_id = curr_id;
            break;               //plane found
         }
      }
      if(find_id != -1)
         break;
   }

   if(find_id == -1){
      tf.plane_id = planes.size();
      planes.push_back(face_plane);
      tf.plane_side = true;
      search_d = -search_d;   //value was negated in second test, return sign back
      pair<long int, int> hp(search_d, tf.plane_id);      
      hmap.insert(hp);
   }
   else{
      tf.plane_id = find_id;
      tf.plane_side = same_side;
   }
}

//----------------------------

E_BSP_BUILD_RESULT I3D_scene::GetStaticFrames(vector<PI3D_frame> &frm_list) const{

   static struct S_hlp{

      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

         if(!(frm->GetFlags()&I3D_FRMF_STATIC_COLLISION))
            return I3DENUMRET_OK;

         vector<PI3D_frame> &frmlist = *(vector<PI3D_frame> *)c;

         switch(frm->GetType1()){
         case FRAME_VOLUME:
            {

               PI3D_volume vol = I3DCAST_VOLUME(frm);
               switch(vol->GetVolumeType()){
               case I3DVOLUME_BOX:
               case I3DVOLUME_RECTANGLE:
                  {
                     frmlist.push_back(vol);
                  }
                  break;
               default:
                  break;     //unsupport type for build in bsp
               }
            }
            break;
         case FRAME_VISUAL:
            frmlist.push_back(frm);
            break;
         default:
            assert(0);
         }
         return I3DENUMRET_OK;
      }
   }hlp;
   frm_list.clear();
   EnumFrames(S_hlp::cbEnum, (dword)&frm_list, ENUMF_VOLUME | ENUMF_VISUAL);

   return BSP_BUILD_OK;
}

//----------------------------
// Add face to tree faces list and scene_polugons list for build.
// If face is invalid, it will not be added and false is returned.
bool C_bsp_tree_builder::AddFace(PI3D_frame origin_frm, int face_id){

   faces.push_back(S_bsp_triface());
   S_bsp_triface &tf = faces.back();
   tf.origin.Setup(origin_frm, face_id);
                           //push transformed verticies
   tf.Expand();
                          //check for degenerate face
   S_plane face_plane(tf.verticies[0], tf.verticies[1], tf.verticies[2]);
   if(IsAbsMrgZero(face_plane.normal.Magnitude()))
   {
      faces.pop_back();
      BSP_ERR_LOG(C_fstring("Invalid face normal: frame '%s', face %i. Face ignored.",
         (const char*)origin_frm->GetName(), face_id));
      return false;
   }
                           //also push to build list 
   scene_polygons.push_back(S_bsp_polygon());
   scene_polygons.back().SetVerticies(tf.verticies, 3, faces.size() -1);
   return true;
}

//----------------------------

E_BSP_BUILD_RESULT C_bsp_tree_builder::AddPolygons(vector<PI3D_frame> &frm_list, I3D_BSP_BUILDER_CB_PROC *cbP, void *context){

   multimap<long int, int> help_map;

   const PI3D_frame *p_frm = frm_list.begin();
   for(int i = 0; i < frm_list.size(); i++){
      const PI3D_frame frm = p_frm[i];

      switch(frm->GetType1()){
      case FRAME_VISUAL:
         {

            PI3D_visual vis = I3DCAST_VISUAL(frm);
            I3D_mesh_base *mb = vis->GetMesh();
            if(!mb)              //skip visuals vithout mesh
               continue;

            int num_fgroups = mb->NumFGroups1();
            if(!num_fgroups)
               continue;         //skip visuals without faces

                           //get all verticies on first LOD and push them to list by tri verticies groups
            for(int ii = mb->NumFaces(); ii--; ){
               if(AddFace(vis, ii))
                  AddFacePlane(faces.back(), (dword)&help_map);
            }
         }
         break;
      case FRAME_VOLUME:
         {
            PI3D_volume vol = I3DCAST_VOLUME(frm);
            switch(vol->GetVolumeType()){
            case I3DVOLUME_BOX:
               {
                              //push 12 faces in polygon list
                  for(int i = 0; i < 12; i++){
                     if(AddFace(frm, i))
                        AddFacePlane(faces.back(), (dword)&help_map);
                  }
               }
               break;
            case I3DVOLUME_RECTANGLE:
               {
#if 1                         //assure that rectangle faces will have same plane
                  for(int i = 0; i < 2; i++){
                     if(!AddFace(frm, i))
                        break;
                     if(!i)
                        AddFacePlane(faces.back(), (dword)&help_map);
                     else{
                        faces.back().plane_id = faces[faces.size()-2].plane_id;
                        faces.back().plane_side = faces[faces.size()-2].plane_side;
                     }
                  }
#else
                  for(int i = 0; i < 2; i++){
                     if(AddFace(frm, i))
                        AddFacePlane(faces.back(), (dword)&help_map);
                  }
#endif
               }
               break;
            }
         }
         break;
      }
                              //check cancellation
      {                       
         LPIGraph igraph = scene->GetDriver1()->GetGraphInterface();
         char k = igraph->ReadKey(true);
         if(k==K_ESC){
            return BSP_BUILD_CANCEL;
         }
      }

      if(cbP!=NULL){
         (*cbP)(planes.size(), 1, context);
      }   
   }
   return BSP_BUILD_OK;
}

//----------------------------

E_BSP_BUILD_RESULT C_bsp_tree_builder::BuilScenePolygons(I3D_BSP_BUILDER_CB_PROC *cbP, void *context){

   assert(scene);
   if(!scene)
      return BSP_BUILD_FAIL;

   PI3D_frame prim_sector = scene->GetPrimarySector1();
   assert(prim_sector);
   if(!prim_sector)
      return BSP_BUILD_FAIL;
                              //clear tree lists
   faces.clear();
   planes.clear();
                              //collect candidate list of frames, which should be in bsp tree
   vector<PI3D_frame> cadidate_list;

   E_BSP_BUILD_RESULT br1 = scene->GetStaticFrames(cadidate_list);
   if(br1 != BSP_BUILD_OK)
      return br1;

   E_BSP_BUILD_RESULT ret = AddPolygons(cadidate_list, cbP, context);

   build_stats.polygon_rest = scene_polygons.size();

   return ret;
}

//----------------------------
// Make axis alignated box aroud all point in list
void C_bsp_tree_builder::MakePolygonsVolume(I3D_bbox &bb, const vector<S_bsp_polygon> &poly_list){

   static const float EXPAND_VALUE = .001f;
   bb.Invalidate();

   for(int kk = poly_list.size(); kk--;){
      const S_bsp_polygon &poly = poly_list[kk];
      assert(poly.valid);
      for (int ii=poly
         .verticies.size(); ii--;){
         const S_vector &point = poly.verticies[ii];
                              //update min and max
         for (int i=3; i--;){
            if (point[i] < bb.min[i]) bb.min[i] = point[i];
            if (point[i] > bb.max[i]) bb.max[i] = point[i];
         }
      }
   }
                              //expand the bounding box slightly
   for(int jj = 3; jj--;){
      bb.min[jj] -= EXPAND_VALUE;
      bb.max[jj] += EXPAND_VALUE;
   }
}

//----------------------------

C_bsp_tree_builder::E_PARTION_RESULT C_bsp_tree_builder::ClassifyPolygon(const S_bsp_polygon &poly,
   BSP_index plane_id){

   assert(plane_id < planes.size());
   const S_plane &plane = planes[plane_id];
   E_PARTION_RESULT poly_class = PR_COINCIDE;
   
   assert(poly.valid);

#if 0

   const S_vector *p_vert = poly.verticies.begin();
   for(int i = poly.verticies.size(); i--; p_vert++){
      float f = p_vert->DistanceToPlane(plane);
      if(f > BSP_SPLIT_TOLERANCE){
         if(poly_class == PR_COINCIDE)
            poly_class = PR_FRONT;
         if(poly_class == PR_BACK)
            poly_class = PR_INTERSECT;
      }else
      if(f < -BSP_SPLIT_TOLERANCE){
         if(poly_class == PR_COINCIDE)
            poly_class = PR_BACK;
         if(poly_class == PR_FRONT)
            poly_class = PR_INTERSECT;
      }
   }

#else

   const S_vector *p_vert = poly.verticies.begin();
   for(int i = poly.verticies.size(); i--; ){
      float f = p_vert->DistanceToPlane(plane);
      p_vert++;
      if(f > BSP_SPLIT_TOLERANCE){
         poly_class = PR_FRONT;
         break;
      }else
      if(f < -BSP_SPLIT_TOLERANCE){
         poly_class = PR_BACK;
         break;
      }
   }
   if(i!=-1){
      if(poly_class == PR_BACK){
         while(i--){
            float f = p_vert->DistanceToPlane(plane);
            p_vert++;
            if(f > BSP_SPLIT_TOLERANCE)
               return PR_INTERSECT;
         }
      }else{
         assert(poly_class == PR_FRONT);
         while(i--){
            float f = p_vert->DistanceToPlane(plane);
            p_vert++;
            if(f < -BSP_SPLIT_TOLERANCE)
               return PR_INTERSECT;
         }
      }
   }
#endif
   return poly_class;
}

//----------------------------

C_bsp_tree_builder::E_PARTION_RESULT C_bsp_tree_builder::Split_Polygon(const S_bsp_polygon &poly,
   const S_plane &partion, vector<S_bsp_polygon> &front, vector<S_bsp_polygon> &back)
{

   assert(poly.valid);
   int   count = poly.verticies.size();
   E_PARTION_RESULT ret_res = PR_COINCIDE;

   assert(count > 2);

   S_vector ptA, ptB;
   vector<S_vector> out_pts;
   vector<S_vector> in_pts;

   float	sideA, sideB;
   ptA = poly.verticies[count - 1];
   sideA = ptA.DistanceToPlane(partion);
   for (short i = -1; ++i < count;)
   {
      ptB = poly.verticies[i];
      sideB = ptB.DistanceToPlane(partion);
      if (sideB > BSP_SPLIT_TOLERANCE)
      {
         if (sideA < -BSP_SPLIT_TOLERANCE)
         {
            // compute the intersection point of the line from 
            // point A to point B with the partition
            // plane. This is a simple ray-plane intersection.
            S_vector v = ptB - ptA;
            S_vector sect_point;
            bool b;
            b = partion.Intersection(ptA, v, sect_point);
            assert(b);
            out_pts.push_back(sect_point); in_pts.push_back(sect_point);
            ret_res = PR_INTERSECT;
         }
         out_pts.push_back(ptB);

         if(ret_res == PR_COINCIDE)
            ret_res = PR_FRONT;
         if(ret_res == PR_BACK)
            ret_res = PR_INTERSECT;
      }
      else if (sideB < -BSP_SPLIT_TOLERANCE)
      {
         if (sideA > BSP_SPLIT_TOLERANCE)
         {
            // compute the intersection point of the line
            // from point A to point B with the partition
            // plane. This is a simple ray-plane intersection.
            S_vector v = ptB - ptA;
            S_vector sect_point;
            bool b;
            b = partion.Intersection(ptB, v, sect_point);
            assert(b);
            out_pts.push_back(sect_point); in_pts.push_back(sect_point);
            ret_res = PR_INTERSECT;
         }
         in_pts.push_back(ptB);
         if(ret_res == PR_COINCIDE)
            ret_res = PR_BACK;
         if(ret_res == PR_FRONT)
            ret_res = PR_INTERSECT;
      }
      else{
         out_pts.push_back(ptB); in_pts.push_back(ptB);
      }
      ptA = ptB;
      sideA = sideB;
   }

   if(ret_res == PR_INTERSECT){
      front.push_back(S_bsp_polygon());
      front.back().SetVerticies(out_pts.begin(), out_pts.size(), poly.origin_face);
      back.push_back(S_bsp_polygon());
      back.back().SetVerticies(in_pts.begin(), in_pts.size(), poly.origin_face);
   }

   return  ret_res;
}

//----------------------------

void C_bsp_tree_builder::ClassifyPartionPlane(BSP_index plane_id, const vector<S_bsp_polygon> &poly_list, S_partion_rating &rating){

   rating.Reset();
   for(int j = poly_list.size(); j--;){
      const S_bsp_polygon &pp = poly_list[j];

      assert(pp.valid);

      E_PARTION_RESULT pr = ClassifyPolygon(pp, plane_id);

      switch(pr){
      case PR_COINCIDE:
         rating.on++;
         break;
      case PR_FRONT:
         rating.front++;
         break;
      case PR_BACK:
         rating.back++;
         break;
      case PR_INTERSECT:
         rating.split++;
         break;
      }
   }
}

//----------------------------

BSP_index C_bsp_tree_builder::PickPartionPlaneRandom(const vector<S_bsp_polygon> &poly_list){

   vector<dword> rest_rnd;
   dword best_split = 9999999;
   float best_ballance = -1.0f;
   BSP_index best_plane_id;

   bool exclusive_ones = poly_list.size() < (RANDOM_PLANES *3);
   if(exclusive_ones){
                              //prepare rnd array
      for(int i = poly_list.size(); i--;){
         rest_rnd.push_back(i);
      }
   }

   const S_bsp_triface *p_faces = faces.begin();
   for(int rest_num = RANDOM_PLANES; rest_num--;){
      int rnd_polygon_id;
      if(exclusive_ones){
         if(rest_rnd.size()==0)
            break;
         int rnd_index = S_int_random(rest_rnd.size());
         rnd_polygon_id = rest_rnd[rnd_index];
         assert(rnd_index <rest_rnd.size());
         rest_rnd[rnd_index] = rest_rnd.back(); rest_rnd.pop_back();
      }
      else{
         rnd_polygon_id = S_int_random(poly_list.size());
      }

      S_partion_rating rating;
      assert(rnd_polygon_id < poly_list.size());
      assert(poly_list[rnd_polygon_id].valid);
      BSP_index rnd_face_id = poly_list[rnd_polygon_id].origin_face;
      assert(rnd_face_id < faces.size());
      const S_bsp_triface &trf = p_faces[rnd_face_id];
      assert(trf.IsValid());
      BSP_index plane_id = trf.plane_id;
      ClassifyPartionPlane(plane_id, poly_list, rating);

      if(build_stats.curr_depth < USE_BALLANCE_DEPTH){
         float greater = Max(rating.back, rating.front);
         float smaller = Min(rating.back, rating.front);
         float curr_ballace = greater ? smaller / greater : .0f;
         if(best_ballance < curr_ballace){
            best_ballance = curr_ballace;
            best_plane_id = plane_id;
         }
      }else
      if(best_split > rating.split){
         best_split = rating.split;
         best_plane_id = plane_id;
      }
   }
   rest_rnd.clear();
   return best_plane_id;
}

//----------------------------

E_BSP_BUILD_RESULT C_bsp_tree_builder::BuildNode(S_bsp_node *node, vector<S_bsp_polygon> &poly_list,
   I3D_BSP_BUILDER_CB_PROC *cbP, void *context){

   build_stats.curr_depth++;

   //MakePolygonsVolume(node->bbox, poly_list);
   BSP_index prt_plane = PickPartionPlaneRandom(poly_list);

                              //save plane id
   node->plane_id = prt_plane;
   const S_plane &partion_plane = planes[prt_plane];
                              //divide polygon list to polygons on plane, front list and back lists
   vector<S_bsp_polygon> poly_front;
   vector<S_bsp_polygon> poly_back;
   vector<BSP_index> coincident_list;

   for(int i = poly_list.size(); i--; ){

      S_bsp_polygon &pp = poly_list[i];
      assert(pp.valid);

      E_PARTION_RESULT pr = Split_Polygon(pp, partion_plane, poly_front, poly_back);

      switch(pr){
      case PR_COINCIDE:
                              //store at vector, alloc in node at end (save memory in final tree)
         coincident_list.push_back(pp.origin_face);
         build_stats.polygon_rest--;
         break;
      case PR_FRONT:
                              //add to temporary front list
         poly_front.push_back(pp);
         break;
      case PR_BACK:
                              //add to temporary back list
         poly_back.push_back(pp);
         break;
      case PR_INTERSECT:
         build_stats.polygon_rest++;
         break;
      }
      poly_list.pop_back();
   }

   assert(coincident_list.size()); //every time must be pushed at least one face - this of which plane was used as partion
   if(!coincident_list.size()){
      return BSP_BUILD_FAIL;
   }
   node->SetFaces(coincident_list.begin(), coincident_list.size());

                              //let check cancellation
   {
      LPIGraph igraph = scene->GetDriver1()->GetGraphInterface();
      char k = igraph->ReadKey(true);
      if(k==K_ESC){
         return BSP_BUILD_CANCEL;
      }
   }

   if(cbP!=NULL){
      (*cbP)(build_stats.polygon_rest, 0, context);
      //(*cbP)(build_stats.depth, 0, context);
   }

                              //assure that divided list was cleared by children
   assert(!poly_list.size());

   assert(build_stats.curr_depth < TREE_SAFE_MAX_DEPTH);
   if(build_stats.curr_depth > TREE_SAFE_MAX_DEPTH){
      return BSP_BUILD_FAIL;
   }

   if(poly_front.size()){
      if(node->front == NULL)
         node->front = new S_bsp_node;
      E_BSP_BUILD_RESULT br1 = BuildNode(node->front, poly_front, cbP, context);
      assert(poly_front.size() == 0);
      if(br1 != BSP_BUILD_OK)
         return br1;
   }

   if(poly_back.size()){
      if(node->back == NULL)
         node->back = new S_bsp_node;
      E_BSP_BUILD_RESULT br2 = BuildNode(node->back, poly_back, cbP, context);
      assert(poly_back.size() == 0);
      if(br2 != BSP_BUILD_OK)
         return br2;
   }

   build_stats.curr_depth--;

   return BSP_BUILD_OK;                                                         
}

//----------------------------

I3D_RESULT C_bsp_tree_builder::Build(I3D_BSP_BUILDER_CB_PROC *cbP, void *context){

   Clear();
   build_stats.Reset();
   E_BSP_BUILD_RESULT br = BuilScenePolygons(cbP, context);
   if((br == BSP_BUILD_OK) && scene_polygons.size()){
      root = new S_bsp_node;
      br = BuildNode(root, scene_polygons, cbP, context);
   }

   I3D_RESULT ret;

   switch(br){
   case BSP_BUILD_OK:
      {
         valid = true;
         ret = I3D_OK;
         ComputeFacesChecksum();
         BuildSlideFlags();

                              //remove all builded frames from force dynamic selcetion
         /*vector<PI3D_frame> statical_list;
         scene->GetStaticFrames(statical_list);
         PI3D_frame *p_list = statical_list.begin();
         for(int i = statical_list.size(); i--;){
            if(p_list[i]->GetType1() == FRAME_VOLUME){
               I3DCAST_VOLUME(p_list[i])->SetForceDynamic(false);
            }
         }
                              //reset tree to make changes visible
         scene->ResetDynamicVolTree();*/
      }
      break;
   case BSP_BUILD_CANCEL:
      ret = I3DERR_CANCELED;
      break;
   default:
      ret = I3DERR_OBJECTNOTFOUND;
   }
   SetAbsentVolumesPickable();   //make pickable volumes, which are not in bsp tree                        
   return ret;

}

//----------------------------

void C_bsp_tree_builder::ComputeFacesChecksum(){

   S_bsp_triface *p_faces = faces.begin();
   for(int i = faces.size(); i--;)
      p_faces[i].vertex_checksum = p_faces[i].GetVertexChecksum();
}

//----------------------------

struct S_facelist{
   vector<BSP_index> face_list;
};

void C_bsp_tree_builder::BuildSlideFlags(){

   S_bsp_triface *p_faces = faces.begin();
   typedef map<PI3D_frame, S_facelist> M_FRM_FACE;

   M_FRM_FACE help_map;

   for(int ii = faces.size(); ii--; ){

      PI3D_frame frm = p_faces[ii].origin.frame;
      M_FRM_FACE::iterator map_it = help_map.find(frm);
      if(map_it == help_map.end()){
         pair<PI3D_frame, S_facelist> new_pair(frm, S_facelist());
         pair<M_FRM_FACE::iterator, bool> result_it;
         result_it = help_map.insert(new_pair);
         map_it = result_it.first;
      }
      S_facelist &fl = (*map_it).second;
      fl.face_list.push_back(ii);
   }

   for(M_FRM_FACE::iterator it = help_map.begin(); it != help_map.end(); it++){

      vector<BSP_index> &list = (*it).second.face_list;
      BSP_index *p_list = list.begin();  
      for(int i = list.size(); i--; ){
         S_bsp_triface &tf = p_faces[p_list[i]];
         S_triindex ti;
         bool b1 = tf.GetVertexIndexes(ti);
         assert(b1);
         if(!b1) continue;
         for(int j = list.size(); j--; ){
            if(i == j)
               continue;
            assert(p_list[i] != p_list[j]);
            S_bsp_triface &tf2 = p_faces[p_list[j]];
            assert(tf2.origin.frame == tf.origin.frame);

            S_triindex ti2;
            bool b2 = tf2.GetVertexIndexes(ti2);
            assert(b2);
            if(!b2) continue;
            bool shared_v1[3] = {false, false, false};
            bool shared_v2[3] = {false, false, false};
            byte share_vert_cnt = 0;
            for(int i1 = 3; i1--; ){
               for(int i2 = 3; i2--; ){
                  if(ti.id[i1] == ti2.id[i2]){
                      shared_v1[i1] = shared_v2[i2] = true;
                      share_vert_cnt++;
                  }
               }
            }
            //assert(share_vert_cnt < 3); //comment out: we can share all verticies if we have 2 faces on same verticies(simulating 2 sided material)
            if(share_vert_cnt != 2)
               continue;
                                 //check angle between faces
            if(tf.plane_id != tf2.plane_id){
               S_vector norm_1 = planes[tf.plane_id].normal;
               if(!tf.plane_side)
                  norm_1.Invert();
               S_vector norm_2 = planes[tf2.plane_id].normal;
               if(!tf2.plane_side)
                  norm_2.Invert();

               float cosh1 = norm_1.Dot(norm_2);
               float cos_zero_delta = fabs(cosh1)-1.0f;
               if(fabs(cos_zero_delta) > .001)
                  continue; //not near cooplanar
            }
                                 //find index of edge
            byte edges1 = 0; byte edges2 = 0;
            for(int k = 3; k--; ){
               if(shared_v1[k] && shared_v1[(k+1)%3]){
                  tf.slide_flags &= ~(1 << k);
                  edges1++;
               }
               if(shared_v2[k] && shared_v2[(k+1)%3]){
                  tf2.slide_flags &= ~(1 << k);
                  edges2++;
               }
            }
            assert((edges1 == 1) && (edges2 == 1));
         }
      }
   }
}

//----------------------------

void C_bsp_tree_builder::DrawEdges(const S_preprocess_context &pc)const{

   vector<S_vector> edge_verts;
   vector<word> edge_indexes;

   dword cur_rend_time = scene->GetDriver1()->GetRenderTime();
   const S_bsp_triface *p_faces = faces.begin();
   for(int i = faces.size(); i--;){
      const S_bsp_triface &tf = p_faces[i];

      assert(tf.origin.frame);
                              //optim: reject if render time different
      if((tf.origin.frame->GetType1() == FRAME_VISUAL) && (I3DCAST_VISUAL(tf.origin.frame)->GetLastRenderTime() != cur_rend_time))
         continue;
      if(!tf.slide_flags)
         continue;
      if(!CheckFrustumIntersection(pc.view_frustum.clip_planes, pc.view_frustum.num_clip_planes, tf.verticies, 3))
         continue;

      word first_index = edge_verts.size();
      for(int kk = 0; kk < 3; kk++){
         if(tf.slide_flags&(1<<kk)){
            edge_indexes.push_back(edge_verts.size());
            edge_indexes.push_back(first_index+((kk+1)%3));
         }
         edge_verts.push_back(tf.verticies[kk]);
      }
   }
   if(edge_verts.size()){

      scene->GetDriver1()->SetTexture(NULL);

      scene->SetRenderMatrix(GetIdentityMatrix());
      //static const S_vector slide_clr(.0f, .5f, 1.0f);
      static const S_vector slide_clr(.0f, .0f, 1.0f);
      const S_vector *p_vert = edge_verts.begin();
      const word *p_line_idx = edge_indexes.begin();
      scene->DrawLines(p_vert, edge_verts.size(), p_line_idx, edge_indexes.size(),
         slide_clr, sizeof(S_vector));
   }
}

//----------------------------

void C_bsp_tree_builder::GetFrames(vector<PI3D_frame> &frm_list) const{

   if(!faces.size())
      return;
   if(frm_list.size())
      frm_list.clear();

   typedef map<PI3D_frame, bool> M_FRAMES;
   M_FRAMES draw_map;
                              //collect all frames in bsp
   const S_bsp_triface *p_faces = faces.begin();
   for(int i = faces.size(); i--;){
      const S_bsp_triface &tf = p_faces[i];
      M_FRAMES::iterator map_it = draw_map.find(tf.origin.frame);
      if(map_it == draw_map.end()){
         pair<PI3D_frame, bool> new_pair(tf.origin.frame, false);
         draw_map.insert(new_pair);
      }
   }
   for(M_FRAMES::iterator it = draw_map.begin(); it != draw_map.end(); it++)
      frm_list.push_back((*it).first);
}

//----------------------------

void C_bsp_tree_builder::DebugDraw(const S_preprocess_context &pc) const{

   if(!IsValid())
      return;

   DrawEdges(pc);
   HRESULT hr;

   if(!tree_frames.size()){
      GetFrames(tree_frames);
   }
   dword cur_rend_time = scene->GetDriver1()->GetRenderTime();
   PI3D_frame *p_frames = tree_frames.begin();
   for(int ii = tree_frames.size(); ii--; )
   {

      PI3D_frame frm = p_frames[ii];
      assert(frm);
      switch(frm->GetType1()){
      case FRAME_VOLUME:
         I3DCAST_VOLUME(frm)->Draw1(scene, false, true);
         break;
      case FRAME_VISUAL:
         {

            PI3D_visual vis = I3DCAST_VISUAL(frm);
            if(vis->GetLastRenderTime() != cur_rend_time)
               continue;
//----------------------------
            dword texture_factor = 0xa00000a0;
                              //render shadow receiver
            PI3D_driver drv = scene->GetDriver1();
            drv->SetTexture1(0, NULL);
            drv->DisableTextureStage(1);
            LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
            drv->SetupBlend(I3DBLEND_ALPHABLEND);
            drv->SetupTextureStage(0, D3DTOP_MODULATE);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TFACTOR);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            drv->SetTextureFactor(texture_factor);
                              //setup camera vector in visual's local coords
            {
               const S_matrix &m_cam = scene->GetActiveCamera1()->GetMatrix();
               S_vectorw vw = m_cam(2);
               vw.Invert();
               vw %= vis->GetInvMatrix();
               drv->GetDevice1()->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_1, &vw, 1);
            }

            vis->RenderSolidMesh(scene, true);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
//----------------------------
         }
         break;
      }
   }
   DebugHighlight(pc);
}

//----------------------------

void C_bsp_tree_builder::DebugHighlight(const S_preprocess_context &pc) const{

   vector<S_vector> trg_verticies;
   vector<word> trg_indexes;

   dword cur_rend_time = scene->GetDriver1()->GetRenderTime();
   const S_bsp_triface *p_faces = faces.begin();

   for(int i = faces.size(); i--;){
      const S_bsp_triface &tf = p_faces[i];

      if(!tf.highlight)
         continue;

      tf.highlight = false;

      assert(tf.origin.frame);
                              //optim: reject if render time different
      if((tf.origin.frame->GetType1() == FRAME_VISUAL) && (I3DCAST_VISUAL(tf.origin.frame)->GetLastRenderTime() != cur_rend_time))
         continue;
      if(!CheckFrustumIntersection(pc.view_frustum.clip_planes, pc.view_frustum.num_clip_planes, tf.verticies, 3))
         continue;
      for(int kk = 0; kk < 3; kk++){
         trg_indexes.push_back(trg_verticies.size());
         trg_verticies.push_back(tf.verticies[kk]);
      }
   }
   if(trg_verticies.size()){

      scene->GetDriver1()->SetTexture(NULL);
      scene->SetRenderMatrix(GetIdentityMatrix());

      static const S_vector clr(.0f, .0f, 1.0f);
      if(trg_verticies.size()){
         const S_vector *p_vert = trg_verticies.begin();
         const word *p_tri_idx = trg_indexes.begin();

         scene->DrawTriangles(p_vert, trg_verticies.size(), p_tri_idx, trg_indexes.size(),
            clr, sizeof(S_vector), 180);
      }
   }
}

//----------------------------
//----------------------------