#include "all.h"
#include "common.h"

//----------------------------

static bool WinGetComputerName(C_string &str){

   char buf[255];
   dword sz = sizeof(buf);
   if(!GetComputerName(buf, &sz)){
      str = NULL;
      return false;
   }
   str = buf;
   return true;
}

//----------------------------

static bool IsMyPC(){
   
   C_string str;
   WinGetComputerName(str);
   return !stricmp(str, "mike");
}

//----------------------------

class C_edit_Access: public C_editor_item{
   virtual const char *GetName() const{ return "Access"; }
   C_smart_ptr<C_table> tab_supervisor;
   C_smart_ptr<C_table> tab_users;
   HWND hwnd_edit;

   static void cbTabCfg1(LPC_table tab, dword msg, dword cb_user, dword prm2, dword prm3){
      switch(msg){
      case TCM_CLOSE:
         {
            C_edit_Access *ei = (C_edit_Access*)cb_user;
            tab->Save((dword)"tables\\i3d_supr.tab", TABOPEN_FILENAME);
            ei->ed->GetIGraph()->DeleteDlgHWND(ei->hwnd_edit);
            ei->hwnd_edit = NULL;
         }
         break;
      }
   }

   static void cbTabCfg2(LPC_table tab, dword msg, dword cb_user, dword prm2, dword prm3){
      switch(msg){
      case TCM_CLOSE:
         {
            C_edit_Access *ei = (C_edit_Access*)cb_user;
            tab->Save((dword)"tables\\i3d_user.tab", TABOPEN_FILENAME);
            ei->ed->GetIGraph()->DeleteDlgHWND(ei->hwnd_edit);
            ei->hwnd_edit = NULL;
         }
         break;
      }
   }

   static C_table_template *GetTemplate1(){
                           //template
      static C_table_template templ = {
      // {105471C0-C907-11d3-92AE-0000B49B8E8B}
         { 0x105471c0, 0xc907, 0x11d3, { 0x92, 0xae, 0x0, 0x0, 0xb4, 0x9b, 0x8e, 0x8b } },
         "Supervisor access rights",
         NULL,
         {
            {TE_ARRAY, 0, "List", 64},
               {TE_STRING, 0, 0, 32},
            {TE_NULL}
         }
      };
      return &templ;
   }

   static C_table_template *GetTemplate2(){
                           //template
      static C_table_template templ = {
      // {105471C0-C907-11d3-92AE-0000B49B8E8B}
         { 0x105471c0, 0xc907, 0x11d3, { 0x92, 0xae, 0x0, 0x0, 0xb4, 0x9b, 0x8e, 0x8b } },
         "User access rights",
         NULL,
         {
            {TE_ARRAY, 0, "Standard users", 64},
               {TE_STRING, 0, 0, 32},
            {TE_ARRAY, 1, "Programmers", 64},
               {TE_STRING, 1, 0, 32},
            {TE_NULL}
         }
      };
      return &templ;
   }

   void DestroyEdit(){
      if(hwnd_edit){
         ed->GetIGraph()->DeleteDlgHWND(hwnd_edit);
         DestroyWindow(hwnd_edit);
         hwnd_edit = NULL;
      }
   }
public:
   C_edit_Access(): hwnd_edit(NULL)
      {}
   virtual bool Init(){
                              //create & read tables
      tab_supervisor = CreateTable();
      tab_supervisor->Open((dword)GetTemplate1(), TABOPEN_TEMPLATE);
      tab_supervisor->Open((dword)"tables\\i3d_supr.tab", TABOPEN_FILENAME | TABOPEN_UPDATE);
      tab_supervisor->Release();

      tab_users = CreateTable();
      tab_users->Open((dword)GetTemplate2(), TABOPEN_TEMPLATE);
      tab_users->Open((dword)"tables\\i3d_user.tab", TABOPEN_FILENAME | TABOPEN_UPDATE);
      tab_users->Release();

      if(IsMyPC()){
         ed->AddShortcut(this, 1000, "%0 &File\\Access\\Supervisor", 0, 0);
      }
                              //for one of supervisors, create advanced menu
      C_string str;
      WinGetComputerName(str);
      for(int i=tab_supervisor->ArrayLen(0); i--; )
         if(str.Matchi(tab_supervisor->ItemS(0, i, 32))) break;
      if(i!=-1)
         ed->AddShortcut(this, 1001, "%0 &File\\Access\\Users", 0, 0);
      return true;
   }
   virtual void Close(){ 
      DestroyEdit();
   }
   //virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){}
   virtual dword Action(int id, void *context){

      switch(id){
                              //determine access right
                              // return value:
                              // 0 = no access
                              // 1 = standard user
                              // 2 = advanced user (possibly programmer?)
                              // 3 = supervisor
      case 0:                 
         {
            C_string str;
            WinGetComputerName(str);
            int i;
                              //check supervisors
            for(i=tab_supervisor->ArrayLen(0); i--; )
               if(str.Matchi(tab_supervisor->ItemS(0, i, 32)))
                  return 3;
                              //check advanced users
            for(i=tab_users->ArrayLen(1); i--; )
               if(str.Matchi(tab_users->ItemS(1, i, 32)))
                  return 2;
                              //check standard users
            for(i=tab_users->ArrayLen(0); i--; )
               if(str.Matchi(tab_users->ItemS(0, i, 32)))
                  return 1;
                              //no access
            return 0;
         }
      case 1000:
         {
            if(hwnd_edit) DestroyEdit();
            hwnd_edit = (HWND)tab_supervisor->Edit(GetTemplate1(), ed->GetIGraph()->GetHWND(), 
               cbTabCfg1, (dword)this,
               TABEDIT_MODELESS | TABEDIT_IMPORT | TABEDIT_EXPORT,
               NULL, NULL);
            if(hwnd_edit) ed->GetIGraph()->AddDlgHWND(hwnd_edit);
         }
         break;
      case 1001:
         {
            if(hwnd_edit) DestroyEdit();
            hwnd_edit = (HWND)tab_users->Edit(GetTemplate2(), ed->GetIGraph()->GetHWND(),
               cbTabCfg2, (dword)this,
               TABEDIT_MODELESS | TABEDIT_IMPORT | TABEDIT_EXPORT,
               NULL, NULL);
            if(hwnd_edit) ed->GetIGraph()->AddDlgHWND(hwnd_edit);
         }
         break;
      }
      return 0;
   }
   //virtual void Render(){}
   virtual bool LoadState(class C_chunk *ck){ 
      return false; 
   }
   virtual bool SaveState(class C_chunk *ck) const{ 
      return false; 
   }
};

//----------------------------

void CreateAccess(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Access;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
