
#include "all.h"
//#include "common.h"

#define BSP_TOOL_VERSION 1

//----------------------------
// Editor pluggin for bsp tree builder.
//----------------------------

#pragma pack(push,1)
   struct S_bsp_static_undo_struct{
      bool static_on;
      char name[1];
   };
#pragma pack(pop)

class C_edit_bsp_tool: public C_editor_item{

   bool draw_tree;

//----------------------------

   virtual const char *GetName() const{ return "BSP-tools"; }

public:

//----------------------------
   virtual bool Init(){

      draw_tree = false;

#define MENU_BASE "%90 &BSP-tree\\"

      ed->AddShortcut(this, E_BSP_TOGGLE_STATIC_SELECTION,  MENU_BASE"%05 Toggle selection &static \tCtrl+V", K_V, SKEY_CTRL);
      ed->AddShortcut(this, E_BSP_CREATE,                   MENU_BASE"%10 &Create \tCtrl+B", K_B, SKEY_CTRL);

      ed->AddShortcut(this, E_BSP_DESTROY,                  MENU_BASE"%i %50 Clear", 0, 0);
      return true;
   }

//----------------------------

   virtual void Close(){}

//----------------------------

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){
   }

//----------------------------

   inline void RegisterDrawTree(bool en){
      if(en)
         ed->RegForRender(this, 1120);
      else
         ed->UnregForRender(this);
   }

   inline void SaveUndo(PI3D_frame frm){

      const char *frm_name = frm->GetName();
      int len1 = strlen(frm_name);            
      S_bsp_static_undo_struct *us = new(new byte[sizeof(S_bsp_static_undo_struct) + len1 + 1]) S_bsp_static_undo_struct;
      us->static_on = (frm->GetFlags()&I3D_FRMF_STATIC_COLLISION);
      strcpy(us->name, frm_name);
      S_edit_Undo_2 eu(GetName(), E_BSP_UNDO_TOGGLE_STATIC, us, sizeof(S_bsp_static_undo_struct) + len1 + 1);
      ed->FindPlugin("Undo")->Action(E_UNDO_SAVE, &eu);
      delete (byte*)us;
   }

//----------------------------

   void ToggleChildrenStatic(PI3D_frame frm, int &toggled_on, int &toggled_off, bool volume_only){

      bool cur_static = (frm->GetFlags()&I3D_FRMF_STATIC_COLLISION);
      bool skip_child = false;
      switch(frm->GetType()){
      case FRAME_VISUAL:
                              //for visual selection don't toggle children
         if(volume_only) break;
         else skip_child = true;
                              //..flow
      case FRAME_VOLUME:
         {
                              //save undo
            SaveUndo(frm);
            frm->SetFlags(!cur_static ? I3D_FRMF_STATIC_COLLISION : 0, I3D_FRMF_STATIC_COLLISION);
            if(cur_static)
               toggled_off++;
            else
               toggled_on++;
         }
         break;
      default: 
         break;
      }
      if(!skip_child)
      for(int i = 0; i < frm->NumChildren(); i++){
         PI3D_frame child_frm = frm->GetChild(i);
         ToggleChildrenStatic(child_frm, toggled_on, toggled_off, volume_only);
      }
   }

//----------------------------

   static bool I3DAPI cbErrLog(I3D_LOADMESSAGE msg, dword prm1, dword prm2, void *context){

      if(context){
         LPC_editor ed = (LPC_editor)context;
         switch(msg){
         case CBM_ERROR:
            {
               dword color = 0;
               switch(prm2){
               case 0:              //error
                  color = 0xff0000;
                  break;
               case 1:              //warning
                  color = 0x008000;
                  break;
               }
               ed->FindPlugin("Log")->Action(E_LOG_ADD_RICH_TEXT, &pair<const char*, dword>((const char*)prm1, color));
            }
            break;
         }
      }
      return false;
   }

//----------------------------

   virtual dword Action(int id, void *context){

      dword ret = 0;          //dafault retunr value

      switch(id){
      case E_BSP_CREATE:
         {
            if(!ed->CanModify())
               return false;
                              //let it run also in background
            ed->GetIGraph()->UpdateParams(0, 0, 0, 0, IG_SUSPENDINACTIVE);

            dword start_time = ed->GetIGraph()->ReadTimer();
            I3D_RESULT br = ed->GetScene()->CreateBspTree(cpBspBuild, ed);
            dword time_delta = ed->GetIGraph()->ReadTimer() - start_time;

            ed->GetIGraph()->UpdateParams(0, 0, 0, IG_SUSPENDINACTIVE, IG_SUSPENDINACTIVE);
            ed->SetModified();

            ed->CheckMenu(this, E_BSP_CREATE, I3D_SUCCESS(br));
            if(I3D_SUCCESS(br)){
               int mins = time_delta / 60000;
               int secs = (time_delta /  1000) % 60;
               int decs = (time_delta / 100) % 10;
               ed->Message(C_fstring("Tree constructed in %i min %2i.%i sec",
                  mins, secs, decs));
               return true;
            }else{
               switch(br){
               case I3DERR_CANCELED:
                  ed->Message(C_string("Bsp build canceled! Tree cleared."));
                  break;
               default:
                  ed->Message(C_string("Failed to build bsp tree."));
               }
               ed->GetScene()->DestroyBspTree();
               return false;
            }
         }
         break;

      case E_BSP_TOGGLE_SHOW:
         draw_tree = !draw_tree;
         ed->CheckMenu(this, E_BSP_TOGGLE_SHOW, draw_tree);
         ed->Message(C_fstring("Draw space occupy %s.", draw_tree ? "on" : "off"));
         RegisterDrawTree(draw_tree);
         break;

      case E_BSP_DESTROY:
         if(!ed->CanModify())
            return false;
         ed->GetScene()->DestroyBspTree();
         ed->SetModified();
         ed->Message("Bsp tree cleared.");
         ed->CheckMenu(this, E_BSP_CREATE, false);
         break;

      case E_BSP_TOGGLE_STATIC_SELECTION:
         if(ed->CanModify())
         {
            int togg_on(0); int togg_off(0); 
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ed->FindPlugin("Selection")->Action(E_SELECTION_GETVECTOR);
            for(int i = sel_list.size(); i--; ){

               PI3D_frame frm = sel_list[i];
               assert(frm);
                              //visual can be set only directly, when frame IS visual. 
                              //for other selection(dummy, models, sectors) change only volumes.
               bool volume_only = (frm->GetType() != FRAME_VISUAL);
               ToggleChildrenStatic(frm, togg_on, togg_off, volume_only);
            }
            ed->SetModified();
            ed->Message(C_fstring("%i frames set static, %i non static.",togg_on, togg_off));
         }
         break;

      case E_BSP_UNDO_TOGGLE_STATIC:
         {            
            S_bsp_static_undo_struct *us = (S_bsp_static_undo_struct*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(us->name);
            if(frm){
               SaveUndo(frm); //redo
               frm->SetFlags(us->static_on ? I3D_FRMF_STATIC_COLLISION : 0, I3D_FRMF_STATIC_COLLISION);
            }
            else
               ed->Message("Toggle static(E_BSP_UNDO_TOGGLE_STATIC): can't find frame.", 0, EM_ERROR);
         }
         break;

      case EDIT_ACTION_AFTER_LOAD:
         ed->CheckMenu(this, E_BSP_CREATE, ed->GetScene()->IsBspBuild());
         break;

      case EDIT_ACTION_MISSION_SAVE_1:
         {
            C_chunk &ck = *(C_chunk*)context;
                              //don't save invalid tree
                              //comented out: save for load varning in editor
            //if(ed->GetScene()->IsBspBuild())
            {
               ck.WAscend(CT_BSP_FILE);
               ed->GetScene()->SaveBsp(ck.GetHandle());
               ck.Descend();
            }
         }
         break;
#if 0
      case EDIT_ACTION_VALIDATE:
         {
                              //currenly not used. validity is checked with loading, any change in geometry after load
                              //is detected by modify pluggin but not by this test(tree use for test its ovn copy of geometry)
                              //and after build validity IS ok.
            bool err = !ed->GetScene()->CompareBspChecksum(cbErrLog, NULL);
            return err;
         }
         break;
#endif
      }
      return ret;
   }

//----------------------------

   virtual bool LoadState(class C_chunk *ck){

                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version != BSP_TOOL_VERSION)
         return false;
                              //read variables
      ck->Read(&draw_tree, sizeof(bool));

      RegisterDrawTree(draw_tree);
      ed->CheckMenu(this, E_BSP_TOGGLE_SHOW, draw_tree);
      return true; 
   }

//----------------------------

   virtual bool SaveState(class C_chunk *ck) const{
   
                              //write version
      byte version = BSP_TOOL_VERSION;
      ck->Write(&version, sizeof(byte));
                              //write variables
      ck->Write(&draw_tree, sizeof(bool));
      return true;
   }

//----------------------------

   static bool I3DAPI cpBspBuild(dword prm1, dword prm2, void *context){

      LPC_editor ed = (LPC_editor)context;
      switch(prm2){
      case 0:
         ed->Message(C_fstring("Polygon rest: %i", prm1), 0, EM_MESSAGE, true);
         break;
      case 1:
         ed->Message(C_fstring("Getting polygons: %i", prm1), 0, EM_MESSAGE, true);
         break;
      }
      //ed->GetScene()->Render();
      //ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
      return true;
   }

//----------------------------

};

//----------------------------
//----------------------------

void CreateBSPTool(LPC_editor ed){

   C_edit_bsp_tool *ei = new C_edit_bsp_tool(); 
   ed->InstallPlugin(ei); 
   ei->Release();
}

//----------------------------
