#include "all.h"
#include "common.h"


//----------------------------

void AddVolumesToScene(PI3D_frame frm, PI3D_scene scene){

   struct S_hlp{
      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
         I3DCAST_VOLUME(frm)->SetScene((PI3D_scene)c);
         return I3DENUMRET_OK;
      }
   };
   frm->EnumFrames(S_hlp::cbEnum, (dword)scene, ENUMF_VOLUME);
}

//----------------------------

void SetupVolumesStatic(PI3D_frame frm, bool on){

   struct S_hlp{
      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
         frm->SetFlags(c ? I3D_FRMF_STATIC_COLLISION : 0, I3D_FRMF_STATIC_COLLISION);
         return I3DENUMRET_OK;
      }
   };
   frm->EnumFrames(S_hlp::cbEnum, on, ENUMF_VOLUME);
}

//----------------------------

C_string FloatStrip(const C_string &str){
   for(int i=str.Size(); i--; ) if(str[i]=='.') break;
   if(i==-1) return str;
   C_string ret = str;
   for(i=ret.Size(); i--; ) 
      if(ret[i]!='0') break;
   if(ret[i]=='.') ++i;
   ret[i+1] = 0;
   return ret;
}

//----------------------------

C_string GetDlgItemText(HWND hwnd, int dlg_id){
   char buf[256];
   return GetDlgItemText(hwnd, dlg_id, buf, sizeof(buf)) ?
      buf : NULL;
}

//----------------------------

bool IsNameInScene(PI3D_scene scene, const C_string &str1){

   struct S_nc{
      static I3DENUMRET I3DAPI cbNameCheck(PI3D_frame frm, dword str1){
         C_string **str=(C_string**)str1;
         if(strcmp(frm->GetName(), **str)) return I3DENUMRET_OK;
         *str = NULL;
         return I3DENUMRET_CANCEL;
      }
   };
   const C_string *str = &str1;
   scene->EnumFrames(S_nc::cbNameCheck, (dword)&str,
      ENUMF_ALL | ENUMF_BACKDROP);
   return !str;
}

//----------------------------

PI3D_visual CastVisual(LPC_editor ed, PI3D_visual vis, dword visual_type,
   LPC_editor_item e_modify, LPC_editor_item e_slct){

   PI3D_visual new_vis = ed->GetDriver()->CreateVisual(visual_type);
   bool ret = false;
   if(new_vis){
      I3D_RESULT ir = new_vis->Duplicate(vis);
      if(I3D_SUCCESS(ir)){
         new_vis->LinkTo(vis->GetParent());
         while(vis->NumChildren()) vis->GetChild(0)->LinkTo(new_vis);

         new_vis->Update();
         vis->LinkTo(NULL);

         pair<PI3D_frame, dword> p(vis, E_MODIFY_FLG_VISUAL);
         e_modify->Action(E_MODIFY_ADD_FLAGS, &p);
         e_modify->Action(E_MODIFY_REPLACE_FRAME, &pair<PI3D_frame, PI3D_frame>(vis, new_vis));

         e_slct->Action(E_SELECTION_ADD_FRAME, new_vis);

                           //check if scene or model owns it
         for(PI3D_frame model_root=new_vis; (model_root=model_root->GetParent()) && model_root->GetType()!=FRAME_MODEL; );
         if(model_root){
            I3DCAST_MODEL(model_root)->DeleteFrame(vis);
            I3DCAST_MODEL(model_root)->AddFrame(new_vis);
         }else{
            ed->GetScene()->DeleteFrame(vis);
            ed->GetScene()->AddFrame(new_vis);
         }

         ed->SetModified();
         ret = true;
      }
      new_vis->Release();
   }
   return ret ? new_vis : NULL;
}

//----------------------------

HINSTANCE GetHInstance(){
   static HINSTANCE hInstance = NULL;
   if(!hInstance){
      hInstance = GetModuleHandle("IEditor.dll");
   }
   return hInstance;
}

//----------------------------

void InitDlg(LPIGraph igraph, void *hwnd){
                              //center dlg and flip to GDI
   //igraph->FlipToGDI();
   RECT rc;
   GetWindowRect((HWND)hwnd, &rc);
   int x = rc.left + ((int)igraph->Scrn_sx() - (rc.right  - rc.left)) / 2;
   int y = rc.top  + ((int)igraph->Scrn_sy() - (rc.bottom - rc.top )) / 2;
   x = Max(0, x);
   y = Max(0, y);
   SetWindowPos((HWND)hwnd, NULL, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

//----------------------------

void MakeSceneName(PI3D_scene scene, char *buf, bool suffix_hint){

                              //try to extract number
   if(!suffix_hint && !IsNameInScene(scene, buf))
      return;

                              //collect all scene names into a list
   set<C_string> name_list;
   struct S_hlp{
      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

         set<C_string> *name_list = (set<C_string>*)c;
         (*name_list).insert(frm->GetName());
         return I3DENUMRET_OK;
      }
   };
   scene->EnumFrames(S_hlp::cbEnum, (dword)&name_list, ENUMF_ALL | ENUMF_BACKDROP);

   int i;
   i = strlen(buf);
   while(i-- && isdigit(buf[i]));
   ++i;
   int num = 0;

   sprintf(&buf[i], num<10 ? "%.2i" : "%i", num);
   if(name_list.find(buf) == name_list.end())
      return;
   while(true){
      ++num;
      sprintf(&buf[i], num<10 ? "%.2i" : "%i", num);
      if(name_list.find(buf) == name_list.end())
         return;
   }
   sprintf(&buf[i], num<10 ? "%.2i" : "%i", num);
}

//----------------------------

static bool WinGetComputerName(C_string &str){

   char buf[255];
   dword sz = sizeof(buf);
   if(!GetComputerName(buf, &sz)){
      str = NULL;
      return false;
   }
   str = buf;
   return true;
}

//----------------------------

bool IsProgrammersPC(LPC_editor ed){

   LPC_editor_item ei = ed->FindPlugin("Access");
   if(ei)
      return (ei->Action(0) >= 2);
   return false;
}

//----------------------------

PI3D_frame TrySelectTopModel(PI3D_frame frm){

   PI3D_frame f1 = frm;
   while(f1 && (f1=f1->GetParent())){
      switch(f1->GetType()){
      case FRAME_MODEL:
         frm = f1;
         break;
      }
   }
   return frm;
}

//----------------------------

int ChooseItemFromList(LPIGraph igraph, const char *title, const char *item_list, int curr_sel){

   struct S_hlp{
      const char *title;
      const char *item_list;
      int curr_sel;
      LPIGraph igraph;

      static BOOL CALLBACK dlgListSelect(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

         switch(uMsg){
         case WM_INITDIALOG:
            {
               S_hlp *hp = (S_hlp*)lParam;
               InitDlg(hp->igraph, hwnd);
               SetWindowText(hwnd, hp->title);
               const char *cp = hp->item_list;
               while(*cp){
                  SendDlgItemMessage(hwnd, IDC_ITEM_SELECT1, LB_ADDSTRING, 0, (LPARAM)cp);
                  cp += strlen(cp) + 1;
               }
               SendDlgItemMessage(hwnd, IDC_ITEM_SELECT1, LB_SETCURSEL, hp->curr_sel, 0);
               ShowWindow(hwnd, SW_SHOW);
            }
            return 1;
         case WM_COMMAND:
            switch(LOWORD(wParam)){
            case IDOK:
               {
                  EndDialog(hwnd, SendDlgItemMessage(hwnd, IDC_ITEM_SELECT1, LB_GETCURSEL, 0, 0));
               }
               break;
            case IDCANCEL: EndDialog(hwnd, -1); break;
            }
            break;
         }
         return 0;
      }
   } hlp;
   hlp.title = title;
   hlp.item_list = item_list;
   hlp.curr_sel = curr_sel;
   hlp.igraph = igraph;

   return DialogBoxParam(GetHInstance(), "IDD_LISTSELECT", (HWND)igraph->GetHWND(),
      (DLGPROC)S_hlp::dlgListSelect, (LPARAM)&hlp);
}

//----------------------------
//----------------------------
