#include <c_string.hpp>

C_string FloatStrip(const C_string &str);
C_string GetDlgItemText(HWND hwnd, int dlg_id);

//----------------------------
// Check if name is already used on some frame in scene.
bool IsNameInScene(PI3D_scene scene, const C_string &str1);

PI3D_visual CastVisual(LPC_editor ed, PI3D_visual vis, dword visual_type,
   LPC_editor_item e_modify, LPC_editor_item e_slct);
HINSTANCE GetHInstance();
void InitDlg(LPIGraph igraph, void *hwnd);

//----------------------------
// Create unique frame name. 'buf' contain base of name, if it contains digits at the end,
// those are stripped and replaced by new digits.
void MakeSceneName(PI3D_scene scene, char *buf, bool suffix_hint = true);

//----------------------------
bool IsProgrammersPC(LPC_editor ed);

//----------------------------
// Add all children volumes of frame to the scene hierarchy.
void AddVolumesToScene(PI3D_frame, PI3D_scene);

//----------------------------
// Setup all children volumes of frame to be static or dynamic.
void SetupVolumesStatic(PI3D_frame, bool);

//----------------------------

struct S_load_err{              
   LPC_editor_item e_log;
   bool first;
   const char *msg;
   S_load_err(LPC_editor_item ei, const char *cp):
      e_log(ei), msg(cp), first(true)
   {}
   static void OpenErrFnc(const char *cp, void *context){
      S_load_err *ep = (S_load_err*)context;
      if(ep->first){
         ep->first = false;
         ep->e_log->Action(E_LOG_ADD_TEXT, 
            (void*)(const char*)C_fstring("*** %s load errors ***", ep->msg));
      }
      ep->e_log->Action(E_LOG_ADD_TEXT, (void*)cp);
   }
};

#define OPEN_ERR_LOG(ed, name) S_load_err::OpenErrFnc, (void*)&S_load_err(ed->FindPlugin("Log"), name)

//----------------------------
                              //find specified pointer in array of 32-bit pointers (asm optimized)
#ifdef _MSC_VER

inline int FindPointerIndex(void **vp, int array_len, void *what){
   int rtn;
   __asm{
      push ecx
      mov edi, vp
      mov edx, array_len
      mov eax, what
      mov ecx, edx
      repne scasd
      jz ok
      mov edx, ecx
   ok:
      sub edx, ecx
      dec edx
      mov rtn, edx
      pop ecx
   }
   return rtn;
}
   
#else

inline int FindPointerIndex(void **vp, int array_len, void *what){
   for(int i(array_len); i--; ) if(vp[i]==what) break;
   return i;
}

#endif

//----------------------------

inline int FloatAsInt(float f){
   return *(dword*)&f;
}

inline float IntAsFloat(int i){
   return (*(float*)&i);
}

//----------------------------

inline void DebugLine(LPC_editor ed, const S_vector &from, const S_vector &to, int type, dword color = -1, int time = 500){
   LPC_editor_item ei = ed->FindPlugin("DebugLine");
   if(ei) ei->Action(E_DEBUGLINE_ADD + type, &S_add_line(from, to, color, time));
}
inline void DebugLineClear(LPC_editor ed){
   LPC_editor_item ei = ed->FindPlugin("DebugLine");
   if(ei) ei->Action(E_DEBUGLINE_CLEAR);
}

//----------------------------
// Try to traverse hierarchy and select topmost parent model frame of given frame.
// If 'frm' is not child of model, the returned value is 'frm'.
PI3D_frame TrySelectTopModel(PI3D_frame frm);

//----------------------------
// Let user visually select one item from a list of items. The returned value is
// index of item, or -1 if selection is cancelled.
// The input parameter points to list of strings to choose from, terminated by '\0'.
int ChooseItemFromList(LPIGraph, const char *title, const char *item_list, int curr_sel = -1);

//----------------------------
