#include "all.h"
#include "common.h"

//----------------------------

struct S_edit_Create_1{
   S_vector pos;
   S_quat rot;
   float scale;
   S_vector nu_scale;
   dword frm_type;
   dword modify_flags;
   char name_buf[1];          //[frame_name, link_name, add_data] - separated by '\0'
};

//----------------------------

class C_edit_Create: public C_editor_item{
   virtual const char *GetName() const{ return "Create"; }
   C_string last_model_path;
   C_string last_mission_path;
   C_string last_sound_path;
   LPC_editor_item e_undo, e_slct, e_modify;
   bool single_volume;
   bool create_static_vol;

//----------------------------

   void SaveDeleteUndo(PI3D_frame frm) const{
                              //store to undo class
      S_edit_Undo_2 eu(GetName(), E_CREATE_DELETE_NAMED, (void*)frm->GetName(), strlen(frm->GetName())+1);
      e_undo->Action(E_UNDO_SAVE, &eu);
   }

//----------------------------

   void SaveCreateUndo(PI3D_frame frm, dword modify_flags, bool model_child,
      void *add_data = NULL, int add_data_len = 0){

                              //create info for re-generation
      dword buf_len = sizeof(S_edit_Create_1) + strlen(frm->GetName()) +
         (frm->GetParent() ? strlen(frm->GetParent()->GetName()) : 0) + 1 + add_data_len;
      byte *tmp_buf = new byte[buf_len];
      S_edit_Create_1 *nm = (S_edit_Create_1*)tmp_buf;
      nm->modify_flags = modify_flags;
      nm->pos = frm->GetPos();
      nm->rot = frm->GetRot();
      if(nm->modify_flags&E_MODIFY_FLG_SCALE) nm->scale = frm->GetScale();
      if(nm->modify_flags&E_MODIFY_FLG_NU_SCALE){
         switch(frm->GetType()){
         //case FRAME_VISUAL: nm->nu_scale = I3DCAST_VISUAL(frm)->GetNUScale(); break;
         case FRAME_VOLUME: nm->nu_scale = I3DCAST_VOLUME(frm)->GetNUScale(); break;
         }
      }
      nm->frm_type = frm->GetType();
                              //store names
      strcpy(nm->name_buf, frm->GetName());
      int pos = strlen(frm->GetName()) + 1;
      if(frm->GetParent()){
         strcpy(&nm->name_buf[pos], frm->GetParent()->GetName());
         pos += strlen(frm->GetParent()->GetName()) + 1;
      }else{
         nm->name_buf[pos] = 0;
         pos += 1;
      }
      memcpy(&nm->name_buf[pos], add_data, add_data_len);
      pos += add_data_len;
                              //store to undo class
      S_edit_Undo_2 eu(GetName(),
         model_child ? E_CREATE_CHILD_NAMED : E_CREATE_FRAME_NAMED, 
         nm, pos + sizeof(S_edit_Create_1) - 1);
      e_undo->Action(E_UNDO_SAVE, &eu); 

      delete[] tmp_buf;
   }

//----------------------------

   bool DeleteFrame(PI3D_frame frm){

      bool del = false;
                              //keep one ref of deletef frame
      C_smart_ptr<I3D_frame> keep_ref = frm;
                              //broadcast delete action
      ed->BroadcastAction(EDIT_ACTION_FRAME_DELETE, frm);

                              //check if this frame was explicitly created
      dword modify_flags = e_modify->Action(E_MODIFY_GET_FLAGS, frm);
      bool model_child = false;
      if(!(modify_flags&E_MODIFY_FLG_CREATE)){
                              //we're able to delete child of model (forever)
         PI3D_frame prnt = frm;
         while((prnt=prnt->GetParent()) && prnt->GetType()!=FRAME_MODEL);
         if(!prnt) return false;
         PI3D_model mod = I3DCAST_MODEL(prnt);
         const PI3D_frame *frmp = mod->GetFrames();
         for(int i=mod->NumFrames(); i--; )
            if(frmp[i]==frm) break;
         if(i==-1) return false;

                              //ask user if it's ok to delete this frame
         int st = MessageBox((HWND)ed->GetIGraph()->GetHWND(), C_fstring("Object %s is part of model.\nDo you really want to delete it?",
            frm->GetName()), "", MB_YESNOCANCEL);
         if(st!=IDYES)
            return false;

         model_child = true;

                              //add to list of deleted frames
         e_modify->Action(E_MODIFY_ADD_MODEL_DEL_CHILD,
            &pair<PI3D_frame, const char*>(mod, frm->GetOrigName()));
      }

      {
                              //save undo info
         void *add_data = NULL;
         dword add_data_len = 0;
         byte *tmp_mem = NULL;   //allocated by frame if needed, deleted automatically

         switch(frm->GetType()){
         case FRAME_MODEL:
            {
               PI3D_model model = I3DCAST_MODEL(frm);
               C_string str = model->GetFileName();
               ed->GetModelCache().GetRelativeDirectory(&str[0]);
               add_data = tmp_mem = new byte[add_data_len = str.Size()+1];
               strcpy((char*)tmp_mem, str);
            }
            break;

         case FRAME_LIGHT:
            {
               PI3D_light lp = I3DCAST_LIGHT(frm);
               add_data_len = sizeof(S_edit_Modify_undo_light::S_frm);
               
                           //count length of names of sectors
               PI3D_sector const *spp = lp->GetLightSectors();
               for(int i=lp->NumLightSectors(); i--; )
                  add_data_len += strlen(spp[i]->GetName()) + 1;

               add_data = tmp_mem = new byte[add_data_len];
               S_edit_Modify_undo_light::S_frm &fi = *(S_edit_Modify_undo_light::S_frm*)tmp_mem;
               fi.ltype = lp->GetLightType();
               fi.color = lp->GetColor();
               fi.power = lp->GetPower();
               fi.mode = lp->GetMode();
               lp->GetCone(fi.cone_i, fi.cone_o);
               lp->GetRange(fi.range_n, fi.range_f);

                           //store names of sectors
               int j;
               for(i=lp->NumLightSectors(), j=0; i--; ){
                  strcpy(fi.frm_names+j, spp[i]->GetName());
                  j += strlen(spp[i]->GetName()) + 1;
               }
               fi.frm_names[j] = 0;
            }
            break;

         case FRAME_SOUND:
            {
               PI3D_sound sp = I3DCAST_SOUND(frm);
               add_data_len = sizeof(S_edit_Modify_undo_sound::S_frm);

                           //count length of names of sectors
               PI3D_sector const *spp = sp->GetSoundSectors();
               for(int i=sp->NumSoundSectors(); i--; )
                  add_data_len += strlen(spp[i]->GetName()) + 1;

               add_data_len += strlen(sp->GetFileName()) + 1;
               add_data = tmp_mem = new byte[add_data_len];
               S_edit_Modify_undo_sound::S_frm &fi = *(S_edit_Modify_undo_sound::S_frm*)tmp_mem;
               fi.stype = sp->GetSoundType();
               fi.vol = sp->GetVolume();
               fi.out_vol = sp->GetOutVol();
               fi.loop = sp->IsLoop();
               fi.streaming = sp->IsStreaming();
               fi.on = sp->IsOn();
               sp->GetCone(fi.cone_i, fi.cone_o);
               sp->GetRange(fi.range_n, fi.range_f, fi.fade_ratio);
               strcpy(fi.frm_names, sp->GetFileName());

                           //store names of sectors
               int j = strlen(sp->GetFileName()) + 1;
               for(i=sp->NumSoundSectors(); i--; ){
                  strcpy(fi.frm_names+j, spp[i]->GetName());
                  j += strlen(spp[i]->GetName()) + 1;
               }
               fi.frm_names[j] = 0;

               sp->SetOn(false);
            }
            break;

         case FRAME_DUMMY:
            {
               PI3D_dummy dum = I3DCAST_DUMMY(frm);
               add_data = tmp_mem = new byte[add_data_len = sizeof(I3D_bbox)];
               memcpy(add_data, dum->GetBBox(), sizeof(I3D_bbox));
            }
            break;

         case FRAME_VOLUME:
            {
               PI3D_volume vol = I3DCAST_VOLUME(frm);
               add_data_len = sizeof(S_edit_Modify_undo_volume::S_frm);
               add_data = tmp_mem = new byte[add_data_len];
               S_edit_Modify_undo_volume::S_frm &fi = *(S_edit_Modify_undo_volume::S_frm*)tmp_mem;
               fi.vtype = vol->GetVolumeType();
            }
            break;

         case FRAME_OCCLUDER:
            {
               PI3D_occluder occ = I3DCAST_OCCLUDER(frm);
               dword numv = occ->NumVertices();
               add_data_len = numv * sizeof(S_vector) + 2 * sizeof(dword);
               add_data = tmp_mem = new byte[add_data_len];
               ((dword*)add_data)[0] = numv;
               ((dword*)add_data)[1] = occ->GetOccluderType();
               S_vector *vp = occ->LockVertices();
               memcpy(tmp_mem + 2*sizeof(dword), vp, numv*sizeof(S_vector));
               occ->UnlockVertices();
            }
            break;
         }

                              //perform action on specialized frames to un-link from
                              // world
                              //this must be done after frame has saved its undo data
         struct S_hlp{
            LPC_editor_item e_modify;
            LPC_editor_item e_slct;
            static I3DENUMRET I3DAPI cbDelFrame(PI3D_frame frm, dword c){

               S_hlp *hp = (S_hlp*)c;
               hp->e_modify->Action(E_MODIFY_DEL_FRAME, frm);
               switch(frm->GetType()){
               case FRAME_LIGHT:
                  {
                     PI3D_light lp = I3DCAST_LIGHT(frm);
                     while(lp->NumLightSectors())
                        lp->GetLightSectors()[0]->DeleteLight(lp);
                  }
                  break;

               case FRAME_SOUND:
                  {
                     PI3D_sound sp = I3DCAST_SOUND(frm);
                     while(sp->NumSoundSectors())
                        sp->GetSoundSectors()[0]->DeleteSound(sp);
                  }
                  break;
               }
                              //remove child of deleted item from selection
               hp->e_slct->Action(E_SELECTION_DEL_FRAME, frm);
               return I3DENUMRET_OK;
            }
         } hlp;
         hlp.e_modify = e_modify;
         hlp.e_slct = e_slct;
         S_hlp::cbDelFrame(frm, (dword)&hlp);
         frm->EnumFrames(S_hlp::cbDelFrame, (dword)&hlp, ENUMF_ALL, NULL);

         SaveCreateUndo(frm, modify_flags, model_child, add_data, add_data_len);
         delete[] tmp_mem;
      }
      frm->LinkTo(NULL);
      ed->GetScene()->DeleteFrame(frm);
      del = true;
      ed->SetModified();

      return del;
   }

//----------------------------
   struct S_dlg_name{
      LPIGraph igraph;
      const char *frm_name;
   };

   static BOOL CALLBACK dlgName(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

      switch(uMsg){
      case WM_INITDIALOG:
         {
            S_dlg_name *hlp = (S_dlg_name*)lParam;
            InitDlg(hlp->igraph, hwnd);
            SetDlgItemText(hwnd, IDC_EDIT, hlp->frm_name);
            ShowWindow(hwnd, SW_SHOW);
            SetWindowLong(hwnd, GWL_USERDATA, lParam);
         }
         return 1;
      case WM_COMMAND:
         switch(LOWORD(wParam)){
         case IDCANCEL: EndDialog(hwnd, 0); break;
         case IDOK:
            {
               S_dlg_name *hlp = (S_dlg_name*)GetWindowLong(hwnd, GWL_USERDATA);
               SendDlgItemMessage(hwnd, IDC_EDIT, WM_GETTEXT, 256, (LPARAM)hlp->frm_name);
               EndDialog(hwnd, 1);
            }
            break;
         }
         break;
      }
      return 0;
   }

//----------------------------

   dword dlgColType(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

      int i;

      switch(uMsg){
      case WM_INITDIALOG:
         {
            InitDlg(ed->GetIGraph(), hwnd);
            CheckDlgButton(hwnd, IDC_RADIO_COLBOX, true);
                              //get selection and determine if we're going
                              // to offer user uncheck single volume
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            EnableWindow(GetDlgItem(hwnd, IDC_CHECK_VOL_SINGLE), (sel_list.size()>1));
            bool check = (sel_list.size()>1) ? single_volume : true;
            CheckDlgButton(hwnd, IDC_CHECK_VOL_SINGLE, check ? BST_CHECKED : BST_UNCHECKED);
            CheckDlgButton(hwnd, IDC_CHECK_VOL_STATIC, create_static_vol ? BST_CHECKED : BST_UNCHECKED);

            ShowWindow(hwnd, SW_SHOW);
         }
         return 1;

      case WM_COMMAND:
         switch(LOWORD(wParam)){
         case IDOK:
            {
               static const int check_id[] = {
                  IDC_RADIO_COLBOX, IDC_RADIO_COLSHP, IDC_RADIO_COLRECT//, IDC_RADIO_COLFACE
               };
               i = sizeof(check_id)/4;
               while(i--)
               if(IsDlgButtonChecked(hwnd, check_id[i])){
                  EndDialog(hwnd, i);
                  break;
               }
            }
            break;

         case IDC_CHECK_VOL_SINGLE:
            single_volume = IsDlgButtonChecked(hwnd, LOWORD(wParam));
            break;

         case IDC_CHECK_VOL_STATIC:
            create_static_vol = IsDlgButtonChecked(hwnd, LOWORD(wParam));
            break;

         case IDCANCEL:
            EndDialog(hwnd, -1);
            break;
         }
         break;
      }
      return 0;
   }

//----------------------------

   static BOOL CALLBACK dlgColType_thunk(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

      switch(uMsg){
      case WM_INITDIALOG:
         SetWindowLong(hwnd, GWL_USERDATA, lParam);
      }
      C_edit_Create *ec = (C_edit_Create*)GetWindowLong(hwnd, GWL_USERDATA);
      if(ec)
         return ec->dlgColType(hwnd, uMsg, wParam, lParam);
      return 0;
   }

//----------------------------

   static BOOL CALLBACK dlgOccType(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){
      int i;

      switch(uMsg){
      case WM_INITDIALOG:
         {
            C_edit_Create *ec = (C_edit_Create*)lParam;
            //SetWindowLong(hwnd, GWL_USERDATA, lParam);

            InitDlg(ec->ed->GetIGraph(), hwnd);
            CheckDlgButton(hwnd, IDC_RADIO_OCCMESH, true);
            ShowWindow(hwnd, SW_SHOW);
         }
         return 1;
      case WM_COMMAND:
         switch(LOWORD(wParam)){
         case IDOK:
            {
               static const int check_id[] = {
                  IDC_RADIO_OCCMESH, IDC_RADIO_OCCSHP
               };
               for(i = sizeof(check_id)/4; i--; )
               if(IsDlgButtonChecked(hwnd, check_id[i])){
                  EndDialog(hwnd, i);
                  break;
               }
            }
            break;
         case IDCANCEL: EndDialog(hwnd, -1); break;
         }
         break;
      }
      return 0;
   }

//----------------------------
                              //vars for repeat duplicate
   S_vector last_dup_pos;
   C_string last_dup_frm;

//----------------------------

   bool Duplicate(PI3D_frame frm, S_vector *new_pos = NULL){

      PI3D_frame new_frm = NULL;
      dword sub_type = 0;
      switch(frm->GetType()){
      case FRAME_VISUAL:
         //sub_type = I3DCAST_VISUAL(frm)->GetVisualType();
         //break;
         return false;
      }
      new_frm = ed->GetDriver()->CreateFrame(frm->GetType(), sub_type);

      PI3D_scene scene = ed->GetScene();
      if(new_frm){
                        //prompt name
         int i;
         char buf[256];

         strcpy(buf, frm->GetName());

                              //try to get new name
         MakeSceneName(scene, buf);
         if(!new_pos){        //don't ask name if repeated duplication
            while(true){
               S_dlg_name hlp = {ed->GetIGraph(), buf};
               i = DialogBoxParam(GetHInstance(), "IDD_MODELNAME", (HWND)ed->GetIGraph()->GetHWND(),
                  dlgName, (LPARAM)&hlp);
               if(!i) break;
                              //check name duplication
               if(!IsNameInScene(scene, buf))
                  break;
            }
         }else i = true;
         if(!i){
            new_frm->Release();
            new_frm = NULL;
         }else{
            I3D_RESULT ir =
               new_frm->Duplicate(frm);
            if(I3D_FAIL(ir)){
               new_frm->Release();
               new_frm = NULL;
            }else{
                           //can't duplicate user value
               struct S_hlp{
                  static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword){
                     frm->SetData(0);
                     return I3DENUMRET_OK;
                  }
               };
               S_hlp::cbEnum(new_frm, 0);
               new_frm->EnumFrames(S_hlp::cbEnum, 0);

               new_frm->SetName(buf);
               new_frm->LinkTo(frm->GetParent());
               if(new_pos){
                  if(frm->GetParent()){
                     S_matrix m = frm->GetParent()->GetMatrix();
                     m.Inverse();
                     *new_pos = *new_pos * m;
                  }
                  new_frm->SetPos(*new_pos);
               }
               new_frm->Update();

                        //add this frame to selection
               e_slct->Action(E_SELECTION_ADD_FRAME, new_frm);

               ed->SetModified();
               SaveDeleteUndo(new_frm);

               dword modify_flags = E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION |
                  E_MODIFY_FLG_LINK | E_MODIFY_FLG_CREATE;
               if(new_frm->GetScale()!=1.0f) modify_flags |= E_MODIFY_FLG_SCALE;
               switch(new_frm->GetType()){
               //case FRAME_VISUAL: if(I3DCAST_VISUAL(new_frm)->GetNUScale()!=S_vector(1, 1, 1)) modify_flags |= E_MODIFY_FLG_NU_SCALE; break;
               case FRAME_VOLUME:
                  {
                     PI3D_volume vol1 = I3DCAST_VOLUME(new_frm);
                     if(vol1->GetNUScale()!=S_vector(1, 1, 1)) modify_flags |= E_MODIFY_FLG_NU_SCALE; break;
                     vol1->SetForceDynamic(true);

                  }
               case FRAME_MODEL:
                  {
                     I3DCAST_MODEL(new_frm)->SetAnimation(0, NULL, I3DANIMOP_LOOP);
                              //set same flags on duplicated children than those of original
                     struct S_hlp{
                        PI3D_model mod;
                        LPC_editor_item e_modify;
                        static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                           S_hlp *hp = (S_hlp*)c;
                           PI3D_frame f_orig = hp->mod->FindChildFrame(frm->GetOrigName());
                           assert(f_orig);
                           dword flags = hp->e_modify->Action(E_MODIFY_GET_FLAGS, f_orig);
                           if(flags)
                              hp->e_modify->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(frm, flags));

                           return I3DENUMRET_OK;
                        }
                     } hlp = {I3DCAST_MODEL(frm), e_modify};
                     new_frm->EnumFrames(S_hlp::cbEnum, (dword)&hlp);
                  }
                  break;
               }

               pair<PI3D_frame, dword> mp(new_frm, modify_flags);
               e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);

               scene->AddFrame(new_frm);

                              //broadcast duplicate action
               ed->BroadcastAction(EDIT_ACTION_FRAME_DUPLICATE, &pair<PI3D_frame, PI3D_frame>(frm, new_frm));
            }
         }
      }
      if(!new_frm){
         ed->Message("Duplication failed", 0, EM_WARNING);
         return false;
      }else{
         last_dup_pos = frm->GetWorldPos();
         last_dup_frm = new_frm->GetName();
         new_frm->Release();
         return true;
      }
   }

//----------------------------

/*
   bool GetOcclPickInfo(int scrn_x, int scrn_y, PI3D_visual *close_frm,
      PI3D_mesh_base *close_mesh, int *closest_point){

      PI3D_scene scene = ed->GetScene();
      PI3D_camera cam = scene->GetActiveCamera();
      if(!cam) return false;
      S_vector from, dir;
      if(I3D_FAIL(scene->UnmapScreenPoint(scrn_x, scrn_y, from, dir)))
         return false;

      I3D_collision_data cd;
      cd.from = from;
      cd.dir = dir;
      cd.flags = I3DGEOCOL_TEXTURE | I3DGEOCOL_RAY | I3DCOL_EXACT;
      ed->GetScene()->TestCollision(cd);
      PI3D_frame frm = cd.GetHitFrm();
      if(!frm)
         return false;
      S_vector ret_pos = from + S_normal(dir) * cd.GetClosestHit();
                              //get mesh from frame
      PI3D_mesh_base mb = NULL;
      switch(frm->GetType()){
      case FRAME_VISUAL:
         mb = I3DCAST_VISUAL(frm)->GetMesh();
         break;
      }
      if(!mb)
         return false;
      *close_frm = I3DCAST_VISUAL(frm);
      *close_mesh = mb;
                              //determine which vertex is closest to the line
      //S_matrix m_inv = ~I3DCAST_VISUAL(frm)->GetGeoMatrix();
      const S_matrix &m_inv = ~frm->GetMatrix();
      from *= m_inv;
      dir = dir.RotateByMatrix(m_inv);
      //const I3D_vertex_mesh *vp = ml->LockVertices();
      const S_vector *vp = mb->LockVertices();
      if(!vp)
         return false;
      dword vstride = mb->GetSizeOfVertex();

      *closest_point = -1;
      float best_dist = (cam->GetWorldPos() - ret_pos).Magnitude() * .15f;
      for(int i=mb->NumVertices(); i--; ){
         //const S_vector &v = vp[i].xyz;
         const S_vector &v = *(const S_vector*)(((byte*)vp) + i * vstride);
         float dist = v.DistanceToLine(from, dir);
         if(best_dist>dist){
            best_dist = dist;
            *closest_point = i;
         }
      }
      mb->UnlockVertices();
      return true;
   }

//----------------------------

   struct S_occl_pick_vertex{
      C_smart_ptr<I3D_visual> vis;
      int vertex_index;
      S_occl_pick_vertex(){}
      S_occl_pick_vertex(const S_occl_pick_vertex &opv){ operator =(opv); }
      S_occl_pick_vertex &operator =(const S_occl_pick_vertex &opv){
         vis = opv.vis;
         vertex_index = opv.vertex_index;
         return *this;
      }
   };
   vector<S_occl_pick_vertex> occl_pick_verts;

//----------------------------

   void CancelOcclPickMode(bool save_undo = true){

      if(save_undo){
                              //save undo info (vertex selection)
         dword undo_size = 1; //size of terminating '\0' character
         for(int i=occl_pick_verts.size(); i--; ){
            const S_occl_pick_vertex &opv = occl_pick_verts[i];
            undo_size += strlen(opv.vis->GetName()) + 1 + sizeof(opv.vertex_index);
         }
                              //save selected verts info undo buffer
         char *mem = new char[undo_size];
         int j = 0;
         for(i=occl_pick_verts.size(); i--; ){
            const S_occl_pick_vertex &opv = occl_pick_verts[i];
            strcpy(mem+j, opv.vis->GetName());
            j += strlen(mem+j) + 1;
            *(dword*)(mem+j) = opv.vertex_index;
            j += sizeof(opv.vertex_index);
         }
         mem[j] = 0;
         S_edit_Undo_2 eu(GetName(), E_CREATE_VERT_SELECTION, mem, undo_size);
         e_undo->Action(E_UNDO_SAVE, &eu);
         delete[] mem;
      }

      occl_pick_verts.clear();
      ed->UnregForRender(this);

      LPC_editor_item e_medit = ed->FindPlugin("MouseEdit");
      if(e_medit && e_medit->Action(E_MOUSE_GET_USER_PICK)==(dword)this){
         e_medit->Action(E_MOUSE_SET_USER_PICK, NULL);
      }
   }
   */

public:
   bool mod_show_prv, mod_anim_prv, snd_show_prv, snd_loop_prv;

   C_edit_Create():
      single_volume(false),
      create_static_vol(true),
      mod_show_prv(true), mod_anim_prv(true), snd_show_prv(true), snd_loop_prv(true),
      e_undo(NULL),
      e_modify(NULL),
      e_slct(NULL)
   {}

//----------------------------

   virtual bool Init(){
                              //find required plugins
      e_undo = ed->FindPlugin("Undo");
      e_slct = ed->FindPlugin("Selection");
      e_modify = ed->FindPlugin("Modify");
      if(!e_undo || !e_slct || !e_modify){
         e_undo = NULL;
         e_slct = NULL;
         e_modify = NULL;
         return false;
      }
      e_undo->AddRef();
      e_slct->AddRef();
      e_modify->AddRef();

      ed->AddShortcut(this, E_CREATE_MODEL, "%25 &Create\\%0 3D object\\&Model\tShift+M", K_M, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_LIGHT, "&Create\\3D object\\&Light\tShift+L", K_L, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_SOUND, "&Create\\3D object\\&Sound\tShift+S", K_S, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_VOLUME, "&Create\\3D object\\&Volume\tShift+V", K_V, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_CAMERA, "&Create\\3D object\\&Camera\tShift+C", K_C, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_DUMMY, "&Create\\3D object\\&Dummy\tShift+D", K_D, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_TARGET, "&Create\\3D object\\&Target\tShift+T", K_T, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_OCCLUDER, "&Create\\3D object\\&Occluder\tShift+O", K_O, SKEY_SHIFT);

      ed->AddShortcut(this, E_CREATE_DUPLICATE_SELECTION, "&Create\\%80 %i &Duplicate\tIns", K_INS, 0);
      ed->AddShortcut(this, E_CREATE_DUPLICATE_REPEAT, "&Create\\%81Repeat duplicate\tShift+Ins", K_INS, SKEY_SHIFT);
      ed->AddShortcut(this, E_CREATE_DELETE_SELECTION, "&Create\\%81 &Delete\tDel", K_DEL, 0);

      {                       //receive notifications from selection
         pair<LPC_editor_item, dword> p(this, E_CREATE_NOTIFY_SELECTION);
         e_slct->Action(E_SELECTION_ADD_NOTIFY, &p);
      }
      return true;
   }

//----------------------------

   virtual void Close(){
      if(e_slct)
         e_slct->Action(E_SELECTION_DEL_NOTIFY, this);

      if(e_undo) e_undo->Release();
      if(e_slct) e_slct->Release();
      if(e_modify) e_modify->Release();
      e_undo = NULL;
      e_slct = NULL;
      e_modify = NULL;

      //CancelOcclPickMode(false);
   }

//----------------------------

   virtual dword Action(int id, void *context){    

      switch(id){

      case E_CREATE_MODEL:
      case E_CREATE_LIGHT:
      case E_CREATE_SOUND:
      case E_CREATE_CAMERA:
      case E_CREATE_VOLUME:
      case E_CREATE_DUMMY:
      case E_CREATE_TARGET:
      case E_CREATE_OCCLUDER:
         {
            if(!ed->CanModify())
               break;
            int i;
            PI3D_frame frm = NULL;
            char frm_name[256];
            PI3D_driver driver = ed->GetDriver();
            bool setup_pos = true;
            dword modify_flags = E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_CREATE;

                              //create and initialize frame
            switch(id){
            case E_CREATE_MODEL:
               {
                  char buf[256];
                  if(Action(E_CREATE_BROWSE_MODEL, buf)){
                     C_string file_name = buf;
                                    //create and load model
                     PI3D_model model;
                     model = driver->CreateModel();
                     I3D_RESULT ir = ed->GetModelCache().Open(model, file_name,
                        ed->GetScene(), 0, OPEN_ERR_LOG(ed, file_name));
                     if(I3D_SUCCESS(ir)){
                        frm = model;
                        model->SetAnimation(0, NULL, I3DANIMOP_LOOP);
                        AddVolumesToScene(model, ed->GetScene());
                        SetupVolumesStatic(model, true);

                              //create suggested frame name
                        for(i=file_name.Size(); i--; ) if(file_name[i]=='\\') break;
                        strcpy(frm_name, &file_name[i+1]);
                        for(i=strlen(frm_name); i--; ) if(frm_name[i]=='.') break;
                        if(i!=-1) frm_name[i] = 0;
                     }else{
                                    //load failed
                        model->Release();
                        ed->Message("Load model failed", 0, EM_WARNING);
                     }
                  }
               }
               break;

            case E_CREATE_SOUND:
               {
                  char buf[256];
                  if(Action(E_CREATE_BROWSE_SOUND, buf)){
                     C_string file_name = buf;
                     PI3D_sound snd = ed->GetScene()->CreateSound();
                     I3D_RESULT ir = ed->GetSoundCache().Open(snd, file_name, ed->GetScene(), 0,
                        OPEN_ERR_LOG(ed, file_name));
                     if(I3D_SUCCESS(ir)){
                        snd->SetSoundType(I3DSOUND_POINT);
                        snd->SetLoop(true);
                        frm = snd;
                              //create suggested frame name
                        for(i=file_name.Size(); i--; ) if(file_name[i]=='\\') break;
                        strcpy(frm_name, &file_name[i+1]);
                        for(i=strlen(frm_name); i--; ) if(frm_name[i]=='.') break;
                        if(i!=-1) frm_name[i] = 0;
                     }else{
                                    //load failed
                        snd->Release();
                        ed->Message("Load sound failed", 0, EM_WARNING);
                     }
                  }
               }
               break;

            case E_CREATE_LIGHT:
               {
                  strcpy(frm_name, "light");
                  PI3D_light lp = driver->CreateLight();
                  lp->SetLightType(I3DLIGHT_POINT);
                  frm = lp;
               }
               break;

            case E_CREATE_CAMERA:
               {
                  strcpy(frm_name, "camera");
                  frm = driver->CreateCamera();
               }
               break;

            case E_CREATE_VOLUME:
               {
                  struct S_hlp{
                     vector<PI3D_frame> sel_frm;
                     int type;
                  } *hlp = (S_hlp*)context;
                  const vector<PI3D_frame> &sel_list =
                     hlp ? hlp->sel_frm :
                     *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
                  if(!sel_list.size()){
                     ed->Message("Selection required.", 0, EM_WARNING);
                     break;
                  }
                  int i, j;
                  if(!hlp){
                     j = DialogBoxParam(GetHInstance(), "IDD_COLTYPE", (HWND)ed->GetIGraph()->GetHWND(),
                        dlgColType_thunk, (LPARAM)this);
                     if(j==-1) break;
                  }else{
                     j = hlp->type;
                  }
                  if(sel_list.size()>1 && !single_volume){
                     vector<PI3D_frame> sl = sel_list;
                     e_slct->Action(E_SELECTION_CLEAR, 0);

                     S_hlp hlp;
                     hlp.type = j;
                     for(i=0; i<sl.size(); i++){
                        hlp.sel_frm.clear();
                        hlp.sel_frm.push_back(sl[i]);
                        Action(id, &hlp);
                     }
                     break;
                  }

                  I3D_bbox bbox;
                  PI3D_frame sel0 = sel_list[0];
                  S_matrix mat0 = sel0->GetMatrix();
                  mat0.Inverse();
                  for(i=0; i<sel_list.size(); i++){
                     PI3D_frame sf = sel_list[i];
                     I3D_bbox bbox1;
                     switch(sf->GetType()){
                     case FRAME_VISUAL:
                        bbox1 = I3DCAST_VISUAL(sf)->GetBoundVolume().bbox;
                        break;
                     default: 
                        {
                           I3D_bound_volume bvol;
                           sf->ComputeHRBoundVolume(&bvol);
                           bbox1 = bvol.bbox;
                        }
                     }
                     if(i){
                        if(bbox1.IsValid()){
                           S_matrix m = sf->GetMatrix() * mat0;
                           for(int j=0; j<8; j++){
                              S_vector v =
                                 S_vector(bbox1[j&1].x, bbox1[(j&2)/2].y, bbox1[(j&4)/4].z);
                              v = v * m;
                              bbox.min.Minimal(v);
                              bbox.max.Maximal(v);
                           }
                        }
                     }else
                        bbox = bbox1;
                  }
                  if(!bbox.IsValid()){                              
                     bbox.min = S_vector(-1, -1, -1);
                     bbox.max = S_vector(1, 1, 1);
                  }

                  I3D_VOLUMETYPE vt = I3DVOLUME_NULL;
                  PI3D_volume vol = driver->CreateVolume();
                              //force all created volumes dynamic until tree will be rebuilded
                  vol->SetForceDynamic(true);
                  vol->SetScene(ed->GetScene());
                  if(create_static_vol)
                     vol->SetFlags(I3D_FRMF_STATIC_COLLISION, I3D_FRMF_STATIC_COLLISION);

                  S_vector diag = bbox.max - bbox.min;
                  vol->SetPos(bbox.min +diag*.5f);
                  switch(j){
                  case 0:
                     {
                        vt = I3DVOLUME_BOX;

                        S_vector scale(Max(.001f, (bbox.max.x-bbox.min.x)/2.0f),
                           Max(.001f, (bbox.max.y-bbox.min.y)/2.0f),
                           Max(.001f, (bbox.max.z-bbox.min.z)/2.0f));
                        vol->SetNUScale(scale);
                        modify_flags |= E_MODIFY_FLG_NU_SCALE;
                     }
                     break;
                  case 1: 
                     {
                        vt = I3DVOLUME_SPHERE;
                        float bs = diag.Magnitude() * .5f;
                        vol->SetScale(bs);
                        modify_flags |= E_MODIFY_FLG_SCALE;
                     }
                     break;
                  case 2:
                     {
                        vt = I3DVOLUME_RECTANGLE;

                        S_vector scale(Max(.001f, (bbox.max.x-bbox.min.x)/2.0f),
                           Max(.001f, (bbox.max.y-bbox.min.y)/2.0f),
                           Max(.001f, (bbox.max.z-bbox.min.z)/2.0f));
                        if(scale.x<scale.z || scale.y<scale.z){
                           if(scale.x<scale.y){
                                             //x's smallest
                                             //turn PI/2 by y and swap x with z
                              scale.x = scale.z;
                              S_quat q;
                              q.Make(S_vector(0, 1, 0), PI/2.0f);
                              vol->SetRot(q);
                           }else{
                                             //y's smallest
                                             //turn PI/2 by x and swap y with z
                              scale.y = scale.z;
                              S_quat q;
                              q.Make(S_vector(1, 0, 0), PI/2.0f);
                              vol->SetRot(q);
                           }
                        }
                        scale.z = 1.0f;
                        vol->SetNUScale(scale);
                        modify_flags |= E_MODIFY_FLG_NU_SCALE;
                     }
                     break;
                  }
                  vol->SetVolumeType(vt);
                  vol->LinkTo(sel0);
                  modify_flags |= E_MODIFY_FLG_LINK;
                  vol->Update();

                  while(sel0 && sel0->GetType()!=FRAME_MODEL) sel0 = sel0->GetParent();
                  vol->SetOwner(sel0);
                  //vol->SetHierarchical(true);

                  setup_pos = false;

                  strcpy(frm_name, "volume");
                  frm = vol;
               }
               break;

            case E_CREATE_DUMMY:
               {
                  strcpy(frm_name, "dummy");
                  PI3D_dummy dum;
                  frm = dum = driver->CreateDummy();
                  dum->SetBBox(I3D_bbox(S_vector(-.5f, -.5f, -.5f), S_vector(.5f, .5f, .5f)));
               }
               break;

            case E_CREATE_TARGET:
               {
                  strcpy(frm_name, "target");
                  frm = driver->CreateTarget();
               }
               break;

            case E_CREATE_OCCLUDER:
               {
                  int i;
                  vector<S_vector> v_pool;
                  I3D_OCCLUDERTYPE ot;

                  /*
                              //if we're already in pick mode, skip type dialog
                  LPC_editor_item e_medit = ed->FindPlugin("MouseEdit");
                  if(e_medit && e_medit->Action(E_MOUSE_GET_USER_PICK)==(dword)this){
                              //collect selected vertices into single pool
                     for(i=occl_pick_verts.size(); i--; ){
                        const S_occl_pick_vertex &opv = occl_pick_verts[i];
                        PI3D_mesh_base mb = opv.vis->GetMesh();
                        const S_vector *vp = mb->LockVertices();
                        const S_vector &v = *(const S_vector*)(((byte*)vp) + opv.vertex_index * mb->GetSizeOfVertex());
                        const S_matrix &m = opv.vis->GetMatrix();
                        v_pool.push_back(v * m);
                        mb->UnlockVertices();
                     }
                     CancelOcclPickMode();
                     ot = I3DOCCLUDER_MESH;
                  }else*/
                  {
                                 //let user select the type of occluder
                     i = DialogBoxParam(GetHInstance(), "IDD_OCCTYPE", (HWND)ed->GetIGraph()->GetHWND(),
                        dlgOccType, (LPARAM)this);
                     if(i==-1)
                        break;
                     ot = (I3D_OCCLUDERTYPE)i;

                     switch(ot){
                     case I3DOCCLUDER_MESH:
                        {
                              //collect all vertices of current selection, in world coordinates
                           const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
                           for(i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()!=FRAME_VISUAL)
                                 continue;
                              PI3D_visual vis = I3DCAST_VISUAL(frm);
                              const S_matrix &mat = vis->GetMatrix();

                              PI3D_mesh_base mb = vis->GetMesh();
                              if(mb){
                                 const S_vector *vp = mb->LockVertices();
                                 dword vs = mb->GetSizeOfVertex();
                                 for(int i=0; i<mb->NumVertices(); i++, (byte*&)vp += vs){
                                    v_pool.push_back((*vp) * mat);
                                 }
                                 mb->UnlockVertices();
                              }
                           }
                        }
                        break;
                     }
                  }

                  bool ok = true;
                  S_vector occ_pos;
                  float scale;
                  bool setup_scale = false;

                  switch(ot){
                  case I3DOCCLUDER_MESH:
                     if(v_pool.size()){
                              //get geometric center of points
                        I3D_bbox bb;
                        bb.Invalidate();
                        for(i=v_pool.size(); i--; ){
                           bb.min.Minimal(v_pool[i]);
                           bb.max.Maximal(v_pool[i]);
                        }
                        occ_pos = bb.min + (bb.max-bb.min) * .5f;
                        for(i=v_pool.size(); i--; )
                           v_pool[i] -= occ_pos;
                        setup_pos = false;
                     }else{
                        /*
                              //setup vertex picking mode
                        S_MouseEdit_pick mep = {
                           this, E_CREATE_PICK_OCCL_VERTEX, LoadCursor(GetHInstance(), "IDC_MAKE_OCCLUDER")
                        };
                        ed->FindPlugin("MouseEdit")->Action(E_MOUSE_SET_USER_PICK, &mep);
                        ed->RegForRender(this, 1000);
                        ok = false;
                        */
                        v_pool.push_back(S_vector(-1, 0, -1));
                        v_pool.push_back(S_vector( 1, 0, -1));
                        v_pool.push_back(S_vector( 1, 0,  1));
                        v_pool.push_back(S_vector(-1, 0,  1));
                        v_pool.push_back(S_vector( 0, 1,  0));
                        setup_pos = true;
                     }
                     break;
                  case I3DOCCLUDER_SPHERE:
                     {
                        const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);

                        I3D_bbox bbox;
                        bbox.Invalidate();
                        for(i=0; i<sel_list.size(); i++){
                           PI3D_frame sf = sel_list[i];
                           I3D_bbox bbox1;
                           switch(sf->GetType()){
                           case FRAME_VISUAL:
                              bbox1 = I3DCAST_VISUAL(sf)->GetBoundVolume().bbox;
                              break;
                           default: 
                              {
                                 I3D_bound_volume bvol;
                                 sf->ComputeHRBoundVolume(&bvol);
                                 bbox1 = bvol.bbox;
                              }
                           }
                           if(bbox1.IsValid()){
                              const S_matrix &m = sf->GetMatrix();
                              for(int j=0; j<8; j++){
                                 S_vector v =
                                    S_vector(bbox1[j&1].x, bbox1[(j&2)/2].y, bbox1[(j&4)/4].z);
                                 v = v * m;
                                 bbox.min.Minimal(v);
                                 bbox.max.Maximal(v);
                              }
                           }
                        }
                        if(bbox.IsValid()){
                           S_vector diag = bbox.max - bbox.min;

                           scale = diag.Magnitude() * .5f;
                           setup_scale = true;
                           modify_flags |= E_MODIFY_FLG_SCALE;

                           occ_pos = bbox.min + diag*.5f;
                           setup_pos = false;
                        }
                     }
                     break;
                  }

                  if(ok){
                              //create occluder
                     strcpy(frm_name, "occluder");
                     frm = driver->CreateFrame(FRAME_OCCLUDER);
                     PI3D_occluder occ = I3DCAST_OCCLUDER(frm);

                     occ->SetOccluderType(ot);

                     switch(ot){
                     case I3DOCCLUDER_MESH:
                        {
                                    //build occluder
                           I3D_RESULT ir;
                           ir = occ->Build(v_pool.begin(), v_pool.size());
                           //assert(I3D_SUCCESS(ir));
                           if(I3D_FAIL(ir)){
                              ed->Message("Failed to create occluder from given data!");
                              occ->Release();
                              occ = NULL;
                              frm = NULL;
                           }
                        }
                        break;
                     }
                     if(!occ)
                        break;

                     if(!setup_pos){
                              //no automatic positioning, setup to center of geometry
                        occ->SetPos(occ_pos);
                     }
                     if(setup_scale){
                        occ->SetScale(scale);
                     }
                     occ->Update();

                     ed->GetScene()->SetFrameSector(frm);
                  }
               }
               break;

            }
            if(!frm) break;
                  
            MakeSceneName(ed->GetScene(), frm_name);
                              //prompt user for name
            while(true){
               S_dlg_name hlp = {ed->GetIGraph(), frm_name};
               bool b = DialogBoxParam(GetHInstance(), "IDD_MODELNAME", (HWND)ed->GetIGraph()->GetHWND(),
                  dlgName, (LPARAM)&hlp);
               if(!b){
                  frm->Release();
                  frm = NULL;
                  break;
               }
                              //check name duplication
               if(!IsNameInScene(ed->GetScene(), frm_name))
                  break;
            }
            if(frm){   
               frm->SetName(frm_name);

               if(setup_pos){
                              //put somewhere in front of camera
                  PI3D_camera cam = ed->GetScene()->GetActiveCamera();
                  assert(cam);
                  const S_vector &cdir = cam->GetWorldDir();
                  frm->SetPos(cam->GetWorldPos() + cdir * 2.0f);
                  frm->SetDir(S_vector(cdir.x, 0.0f, cdir.z));
                  frm->Update();
                  ed->GetScene()->SetFrameSector(frm);
                  LPC_editor_item e_medit = ed->FindPlugin("MouseEdit");
                  if(e_medit)
                     e_medit->Action(E_MOUSE_POLAR_ALIGN_FRAME, frm);
               }
               switch(frm->GetType()){
               case FRAME_LIGHT:
                  {
                              //set light to current sector
                     PI3D_frame sct = frm->GetParent();
                     if(sct && sct->GetType()==FRAME_SECTOR){
                        I3DCAST_SECTOR(sct)->AddLight(I3DCAST_LIGHT(frm));
                     }
                  }
                  break;
               }

                              //clear selection and select this model
               if(!context)
                  e_slct->Action(E_SELECTION_CLEAR, 0);
               e_slct->Action(E_SELECTION_ADD_FRAME, frm);

               ed->SetModified();
               SaveDeleteUndo(frm);

               pair<PI3D_frame, dword> mp(frm, modify_flags);
               e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);

               ed->GetScene()->AddFrame(frm);
               frm->Release();
            }
         }
         break;

      case E_CREATE_FRAME_NAMED:
      case E_CREATE_CHILD_NAMED:
         {
            if(!ed->CanModify()) break;
            S_edit_Create_1 &nm = *(S_edit_Create_1*)context;
            PI3D_frame frm = NULL;
            const void *add_data;   //pointer to data saved in DeleteFrame - frame-specific

            const char *link_to_name = nm.name_buf + strlen(nm.name_buf) + 1;
            PI3D_frame link_to = NULL;
            if(*link_to_name)
               link_to = ed->GetScene()->FindFrame(link_to_name, ENUMF_BACKDROP | ENUMF_ALL);

            {
               const char *cp = nm.name_buf + strlen(nm.name_buf)+1;
               cp = cp + strlen(cp)+1;
               add_data = cp;
            }

            switch(nm.frm_type){
            case FRAME_MODEL:
               {
                              //create and load model
                  PI3D_model model;
                  model = ed->GetDriver()->CreateModel();
                  const char *mname = (const char*)add_data;
                  I3D_RESULT ir = ed->GetModelCache().Open(model, mname, ed->GetScene(), 0, OPEN_ERR_LOG(ed, mname));
                  if(I3D_SUCCESS(ir)){
                     frm = model;
                     model->SetAnimation(0, NULL, I3DANIMOP_LOOP);
                     AddVolumesToScene(model, ed->GetScene());
                     SetupVolumesStatic(model, true);
                  }else{
                     model->Release();
                     ed->Message("Create(E_CREATE_FRAME_NAMED): cannot load model", 0, EM_WARNING);
                  }
               }
               break;

            case FRAME_LIGHT:
               {
                  PI3D_light lp = ed->GetDriver()->CreateLight();
                  const S_edit_Modify_undo_light::S_frm &fi = *(S_edit_Modify_undo_light::S_frm*)add_data;
                  lp->SetLightType(fi.ltype);
                  lp->SetColor(fi.color);
                  lp->SetPower(fi.power);
                  lp->SetMode(fi.mode);
                  lp->SetCone(fi.cone_i, fi.cone_o);
                  lp->SetRange(fi.range_n, fi.range_f);
                  lp->Update();
                  frm = lp;
                              //init light sectors
                  int j = 0;
                  while(fi.frm_names[j]){
                     PI3D_sector sct = I3DCAST_SECTOR(ed->GetScene()->FindFrame(fi.frm_names+j, ENUMF_SECTOR | ENUMF_BACKDROP));
                     if(sct) sct->AddLight(lp);
                     j += strlen(fi.frm_names+j) + 1;
                  }
               }
               break;

            case FRAME_SOUND:
               {
                  PI3D_sound sp = ed->GetScene()->CreateSound();
                  const S_edit_Modify_undo_sound::S_frm &fi = *(S_edit_Modify_undo_sound::S_frm*)add_data;
                  sp->SetOn(false);
                  sp->SetSoundType(fi.stype);
                  sp->SetVolume(fi.vol);
                  sp->SetOutVol(fi.out_vol);
                  sp->SetLoop(fi.loop);
                  sp->SetCone(fi.cone_i, fi.cone_o);
                  sp->SetRange(fi.range_n, fi.range_f, fi.fade_ratio);
                  sp->Open(fi.frm_names, 0);
                  sp->Update();
                  if(fi.on) sp->SetOn(true);
                  frm = sp;
                              //init sound sectors
                  int j = strlen(fi.frm_names) + 1;
                  while(fi.frm_names[j]){
                     PI3D_sector sct = I3DCAST_SECTOR(ed->GetScene()->FindFrame(fi.frm_names+j, ENUMF_SECTOR | ENUMF_BACKDROP));
                     if(sct) sct->AddSound(sp);
                     j += strlen(fi.frm_names+j) + 1;
                  }
               }
               break;

            case FRAME_DUMMY:
               {
                  PI3D_dummy dum;
                  frm = dum = ed->GetDriver()->CreateDummy();
                  dum->SetBBox(*(PI3D_bbox)add_data);
               }
               break;

            case FRAME_VOLUME:
               {
                  PI3D_volume vol;
                  frm = vol = ed->GetDriver()->CreateVolume();
                  const S_edit_Modify_undo_volume::S_frm &fi = *(S_edit_Modify_undo_volume::S_frm*)add_data;
                  vol->SetScene(ed->GetScene());
                  vol->SetVolumeType(fi.vtype);
               }
               break;

            case FRAME_OCCLUDER:
               {
                  PI3D_occluder occ;
                  frm = ed->GetDriver()->CreateFrame(FRAME_OCCLUDER);
                  occ = I3DCAST_OCCLUDER(frm);
                  dword numv = ((dword*)add_data)[0];
                  I3D_OCCLUDERTYPE ot = (I3D_OCCLUDERTYPE)((dword*)add_data)[1];
                  occ->SetOccluderType(ot);
                  S_vector *vp = (S_vector*)((byte*)add_data + 2*sizeof(dword));
                  occ->Build(vp, numv);
               }
               break;

            case FRAME_CAMERA:
            case FRAME_TARGET:
               {
                  frm = ed->GetDriver()->CreateFrame(nm.frm_type);
               }
               break;

            case FRAME_VISUAL:
               {
                  assert(id==E_CREATE_CHILD_NAMED);
                  assert(link_to);
                              //find parent model (which used to be our parent)
                  for(PI3D_frame prnt = link_to; prnt && prnt->GetType()!=FRAME_MODEL; prnt=prnt->GetParent());
                  assert(prnt);
                  if(!prnt) break;
                  PI3D_model mod = I3DCAST_MODEL(prnt);
                              //re-load original model
                  PI3D_model orig = ed->GetDriver()->CreateModel();
                  C_string fname = mod->GetFileName();
                  ed->GetModelCache().GetRelativeDirectory(&fname[0]);
                  ed->GetModelCache().Open(orig, fname, 0, 0);
                  for(int i=strlen(nm.name_buf); i && nm.name_buf[i-1]!='.'; i--);
                  PI3D_frame frm_orig = orig->FindChildFrame(nm.name_buf+i);
                  assert(frm_orig && frm_orig->GetType()==FRAME_VISUAL);
                  if(frm_orig && frm_orig->GetType()==FRAME_VISUAL){
                     frm = ed->GetDriver()->CreateFrame(FRAME_VISUAL,
                        I3DCAST_VISUAL(frm_orig)->GetVisualType());
                     assert(frm);
                     if(frm)
                        frm->Duplicate(frm_orig);
                  }
                  orig->Release();
               }
               break;

            default: 
               assert(0);
            }

            if(frm){          //creation ok
                              //make frame's name
               char buf[256];
               strcpy(buf, nm.name_buf);
               MakeSceneName(ed->GetScene(), buf, false);
               frm->SetName(buf);

                              //put to specified location
               frm->SetPos(nm.pos);
               frm->SetRot(nm.rot);
               frm->Update();

                              //link to specified frame
                              //if failed to locate link frame - link to primary
               if(!link_to) link_to = ed->GetScene()->GetPrimarySector();
               frm->LinkTo(link_to);

                              //add to modify
               if(nm.modify_flags){
                  pair<PI3D_frame, dword> mp(frm, nm.modify_flags);
                  e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);
               }

               if(nm.modify_flags&E_MODIFY_FLG_SCALE)
                  frm->SetScale(nm.scale);
               if(nm.modify_flags&E_MODIFY_FLG_NU_SCALE){
                  switch(frm->GetType()){
                  //case FRAME_VISUAL: I3DCAST_VISUAL(frm)->SetNUScale(nm.nu_scale); break;
                  case FRAME_VOLUME: I3DCAST_VOLUME(frm)->SetNUScale(nm.nu_scale); break;
                  }
               }
               if(nm.modify_flags&(E_MODIFY_FLG_SCALE|E_MODIFY_FLG_NU_SCALE))
                  frm->Update();
                              //clear selection and select this model
               e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

               if(id==E_CREATE_CHILD_NAMED){
                  for(PI3D_frame prnt = link_to; prnt && prnt->GetType()!=FRAME_MODEL; prnt=prnt->GetParent());
                  assert(prnt);
                  I3DCAST_MODEL(prnt)->AddFrame(frm);
                  bool b;
                  b = e_modify->Action(E_MODIFY_DEL_MODEL_DEL_CHILD,
                     &pair<PI3D_frame, const char*>(prnt, frm->GetOrigName()));
                  assert(b);
               }else{
                  ed->GetScene()->AddFrame(frm);
               }

               ed->SetModified();
               SaveDeleteUndo(frm);

               //ei->Action(E_SELECTION_ADD_FRAME, frm);

               ed->Message(C_fstring("Create: %s", frm->GetName()));
               frm->Release();
            }
         }
         break;

      case E_CREATE_DUPLICATE_SELECTION:
         {
            vector<PI3D_frame> sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            e_slct->Action(E_SELECTION_CLEAR);

            for(int i=0; i<sel_list.size(); i++){
               if(!Duplicate(sel_list[i])) break;
            }
            ed->Message(C_fstring("Duplicated %i frame(s)", i), 0, EM_MESSAGE);
         }
         break;

      case E_CREATE_DUPLICATE_REPEAT:
         {
            PI3D_frame frm = NULL;
            if(last_dup_frm.Size()) frm = ed->GetScene()->FindFrame(last_dup_frm, ENUMF_ALL | ENUMF_BACKDROP);
            if(frm){
               e_slct->Action(E_SELECTION_CLEAR);
                              //compute new position
               S_vector pos = frm->GetWorldPos() + (frm->GetWorldPos() - last_dup_pos);
               Duplicate(frm, &pos);
            }else
               ed->Message("Cannot repeat duplication", 0, EM_WARNING);
         }
         break;

      case E_CREATE_BROWSE_MODEL:
      case E_CREATE_BROWSE_SOUND:
         {
            char *buf = (char*)context;
            buf[0] = 0;

            C_string *last_path;
            const char *title = NULL;
            const char *extensions = NULL;
            const char *def_ext = NULL;

            OPENFILENAME on;
            memset(&on, 0, sizeof(on));
            on.lStructSize = sizeof(on);

            struct S_hlp{
               LPIGraph igraph;
               PI3D_driver driver;
               PI3D_scene ed_scene;
               C_edit_Create *ec;
               LPC_editor_item e_mission;
               char *buf;
               bool mission_model;
               const char *last_mission_path;
               C_string cwd;
               bool in;       //re-entering block

               enum E_mode{
                  MODE_MODEL, MODE_SOUND
               } mode;
                              //MODE_MODEL:
               PI3D_scene scene;
               PI3D_camera cam;
               PI3D_model mod;
               C_I3D_cache_base<I3D_model> *cache_model;
               bool timer_on;
               float cam_orbit_angle;

                              //MODE_SOUND:
               C_I3D_cache_base<I3D_sound> *sound_cache;
               PI3D_sound snd;
               PISND_sound play_snd;

               S_hlp():
                  timer_on(false),
                  in(false),
                  mission_model(false),
                  e_mission(NULL)
               {}

               void Render(HWND hwnd, bool clear = false){
                  if(!clear){
                     if(mod){
                        S_quat q;
                        q.Make(S_vector(0, 1, 0), cam_orbit_angle);
                        mod->SetRot(q);
                        mod->Update();
                     }
                     scene->Render(0);
                  }
                  hwnd = GetDlgItem(hwnd, IDC_PREVIEW);

                  if(clear){
                     igraph->ClearViewport();
                  }
                  igraph->UpdateScreen2(0, hwnd);
               }

               static UINT CALLBACK cbOpen(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){
                  switch(uMsg){
                  case WM_INITDIALOG:
                     {
                        OPENFILENAME *ofn = (OPENFILENAME*)lParam;
                        S_hlp *hp = (S_hlp*)ofn->lCustData;
                        SetWindowLong(hwnd, GWL_USERDATA, (LPARAM)hp);
                        InitDlg(hp->igraph, hwnd);

                        switch(hp->mode){
                        case MODE_MODEL:
                           {
                              HWND hwnd_prv = GetDlgItem(hwnd, IDC_PREVIEW);
                              RECT rc;
                              GetClientRect(hwnd_prv, &rc);
                              //hp->scene->SetViewport(0, 0, rc.right-rc.left-4, rc.bottom-rc.top-4);

                              CheckDlgButton(hwnd, IDC_CHECK_PRV_SHOW, hp->ec->mod_show_prv ? BST_CHECKED : BST_UNCHECKED);
                              CheckDlgButton(hwnd, IDC_CHECK_PRV_ANIMATE, hp->ec->mod_anim_prv ? BST_CHECKED : BST_UNCHECKED);

                              if(!hp->ec->mod_show_prv) EnableWindow(GetDlgItem(hwnd, IDC_CHECK_PRV_ANIMATE), false);
                              if(!hp->e_mission)
                                 ShowWindow(GetDlgItem(hwnd, IDC_MODEL_MISSION), SW_HIDE);
                           }
                           break;
                        case MODE_SOUND:
                           {
                              CheckDlgButton(hwnd, IDC_CHECK_SND_PRV_SHOW, hp->ec->snd_show_prv ? BST_CHECKED : BST_UNCHECKED);
                              CheckDlgButton(hwnd, IDC_CHECK_SND_PRV_LOOP, hp->ec->snd_loop_prv ? BST_CHECKED : BST_UNCHECKED);
                              if(!hp->ec->snd_show_prv) EnableWindow(GetDlgItem(hwnd, IDC_CHECK_SND_PRV_LOOP), false);
                           }
                           break;
                        }
                        return 1;
                     }
                     break;

                  case WM_COMMAND:
                     switch(LOWORD(wParam)){
                     case IDC_CHECK_PRV_SHOW:
                     case IDC_CHECK_PRV_ANIMATE:
                     case IDC_CHECK_SND_PRV_SHOW:
                     case IDC_CHECK_SND_PRV_LOOP:
                        {
                           S_hlp *hp = (S_hlp*)GetWindowLong(hwnd, GWL_USERDATA);
                           switch(LOWORD(wParam)){
                           case IDC_CHECK_PRV_SHOW:
                              hp->ec->mod_show_prv = IsDlgButtonChecked(hwnd, LOWORD(wParam));
                              if(!hp->ec->mod_show_prv){
                                 if(hp->timer_on){
                                    KillTimer(hwnd, 1);
                                    hp->timer_on = false;
                                 }
                                 hp->Render(hwnd, true);
                              }
                              EnableWindow(GetDlgItem(hwnd, IDC_CHECK_PRV_ANIMATE), hp->ec->mod_show_prv);
                              break;
                           case IDC_CHECK_PRV_ANIMATE:
                              hp->ec->mod_anim_prv = IsDlgButtonChecked(hwnd, LOWORD(wParam));
                              if(hp->ec->mod_anim_prv){
                                 if(!hp->timer_on){
                                    hp->timer_on = SetTimer(hwnd, 1, 50, NULL);
                                 }
                              }else{
                                 if(hp->timer_on){
                                    KillTimer(hwnd, 1);
                                    hp->timer_on = false;
                                    hp->cam_orbit_angle = 0.0f;
                                    hp->Render(hwnd);
                                 }
                              }
                              break;
                           case IDC_CHECK_SND_PRV_SHOW:
                              hp->ec->snd_show_prv = IsDlgButtonChecked(hwnd, LOWORD(wParam));
                              if(!hp->ec->snd_show_prv){
                                 if(hp->play_snd) hp->play_snd->Stop();
                              }
                              EnableWindow(GetDlgItem(hwnd, IDC_CHECK_SND_PRV_LOOP), hp->ec->snd_show_prv);
                              break;
                           case IDC_CHECK_SND_PRV_LOOP:
                              hp->ec->snd_loop_prv = IsDlgButtonChecked(hwnd, LOWORD(wParam));
                              if(hp->play_snd){
                                 PISND_source src = hp->snd->GetSoundSource();
                                 hp->play_snd->PlaySource(src, ISND_NOVOLUME | 
                                    (hp->ec->snd_loop_prv ? ISND_LOOP : 0));
                              }
                              break;
                           }
                        }
                        break;

                     case IDC_MODEL_MISSION:
                        {
                           S_hlp *hp = (S_hlp*)GetWindowLong(hwnd, GWL_USERDATA);
                           strcpy(hp->buf, hp->last_mission_path);
                           if(hp->e_mission->Action(E_MISSION_BROWSE, &pair<char*,HWND>(hp->buf, hwnd))){
                              hp->mission_model = true;
                              //PostMessage(GetParent(hwnd), WM_COMMAND, MAKELPARAM(IDABORT, BN_CLICKED), (LPARAM)GetDlgItem(GetParent(hwnd), IDABORT));
                              EndDialog(GetParent(hwnd), true);
                           }
                        }
                        break;
                     }
                     break;

                  case WM_NOTIFY:
                     {
                        OFNOTIFY *on = (OFNOTIFY*)lParam;
                        switch(on->hdr.code){
                        case CDN_SELCHANGE:
                           {
                              S_hlp *hp = (S_hlp*)on->lpOFN->lCustData;
                              switch(hp->mode){
                              case MODE_MODEL:
                                 {
                                    if(!hp->ec->mod_show_prv || hp->in) break;
                                    hp->in = true;

                                    char buf[MAX_PATH];
                                    SendMessage(GetParent(hwnd), CDM_GETFILEPATH, MAX_PATH, (LPARAM)&buf);
                                    hp->cache_model->GetRelativeDirectory(buf);
                                    int len = strlen(buf);
                                    if(len>=4 && !stricmp(&buf[len-4], ".i3d"))
                                       buf[len-4] = 0;
                                                   //try to open the file and render scene
                                    {
                                       if(hp->mod){
                                          hp->mod->LinkTo(NULL);
                                          hp->mod->Release();
                                          hp->mod = NULL;
                                       }
                                       hp->Render(hwnd, true);
                                       hp->mod = hp->driver->CreateModel();
                                       hp->mod->LinkTo(hp->scene->GetPrimarySector());
                                       _chdir(hp->cwd);
                                       SetCursor(LoadCursor(NULL, IDC_WAIT));
                                       I3D_RESULT ir = hp->cache_model->Open(hp->mod, buf, hp->ed_scene, 0, NULL, NULL);
                                       SetCursor(LoadCursor(NULL, IDC_ARROW));
                                       if(I3D_SUCCESS(ir)){
                                          hp->mod->SetAnimation(0, NULL, I3DANIMOP_LOOP);

                                          I3D_bound_volume bvol;
                                          hp->scene->GetPrimarySector()->ComputeHRBoundVolume(&bvol);
                                          const I3D_bbox &bbox = bvol.bbox;
                                          float cam_dist;
                                          S_vector cam_look_at;
                                          if(bbox.IsValid()){
                                             S_vector diag = bbox.max - bbox.min;
                                             cam_look_at = bbox.min + diag*.5f;
                                             cam_dist = diag.Magnitude() * 1.2f;
                                             cam_dist = Max(cam_dist, 2.0f);
                                          }else{
                                             cam_look_at.Zero();
                                             cam_dist = 0.0f;
                                          }
                                    
                                          S_vector pos = cam_look_at + S_vector(.8f, .4f, -.8f) * cam_dist;
                                          hp->cam->SetPos(pos);
                                          hp->cam->SetDir(cam_look_at - pos);
                                          hp->cam->Update();

                                          hp->cam_orbit_angle = 0.0f;

                                          hp->Render(hwnd);

                                          if(hp->ec->mod_anim_prv && !hp->timer_on){
                                             hp->timer_on = SetTimer(hwnd, 1, 50, NULL);
                                          }
                                       }else{
                                          if(hp->timer_on){
                                             KillTimer(hwnd, 1);
                                             hp->timer_on = false;
                                          }
                                       }
                                    }
                                    hp->in = false;
                                 }
                                 break;
                              case MODE_SOUND:
                                 {
                                    if(hp->in) break;
                                    hp->in = true;

                                    char buf[MAX_PATH];
                                    //SendMessage(GetParent(hwnd), CDM_GETSPEC, MAX_PATH, (LPARAM)&buf);
                                    SendMessage(GetParent(hwnd), CDM_GETFILEPATH, MAX_PATH, (LPARAM)&buf);
                                    hp->sound_cache->GetRelativeDirectory(buf);
                                                      //strip extension
                                    for(int i=strlen(buf); i--; ){
                                       if(buf[i]=='.'){
                                          buf[i] = 0;
                                          break;
                                       }
                                    }

                                    if(hp->snd){
                                       hp->snd->LinkTo(NULL);
                                       hp->snd->Release();
                                       hp->snd = NULL;
                                    }
                                    hp->snd = hp->ed_scene->CreateSound();
                                    C_string str;

                                    _chdir(hp->cwd);
                                    I3D_RESULT ir = hp->sound_cache->Open(hp->snd, buf, hp->ed_scene);
                                    if(I3D_SUCCESS(ir)){
                                       PISND_source src = hp->snd->GetSoundSource();
                                       if(src){
                                          const S_wave_format &fmt = *src->GetFormat();

                                          str = C_fstring(
                                             "Channels: %i\nBits: %i\nFrequency: %i Hz\nLength: %.2f s",
                                             fmt.num_channels,
                                             fmt.bytes_per_sample*8,
                                             fmt.samples_per_sec,
                                             (float)fmt.size / (fmt.samples_per_sec*fmt.bytes_per_sample));
                                          if(hp->ec->snd_show_prv && hp->play_snd){
                                             hp->play_snd->PlaySource(src, ISND_NOVOLUME | 
                                                (hp->ec->snd_loop_prv ? ISND_LOOP : 0));
                                          }
                                       }
                                    }
                                    if(!str.Size()){
                                       if(hp->play_snd) hp->play_snd->Stop();
                                    }
                                    SetDlgItemText(hwnd, IDC_SOUND_DETAILS, str);

                                    hp->in = false;
                                 }
                                 break;
                              }
                           }
                           break;

                        case CDN_FILEOK:
                           {
                              S_hlp *hp = (S_hlp*)on->lpOFN->lCustData;
                              if(hp->mode==MODE_MODEL)      //hide our rendering
                                 hp->ed_scene->Render(0);
                           }
                           break;
                        }
                     }
                     break;

                  case WM_TIMER:
                     {
                        S_hlp *hp = (S_hlp*)GetWindowLong(hwnd, GWL_USERDATA);
                        if(hp->timer_on && !hp->in){
                           hp->in = true;
                           hp->cam_orbit_angle -= PI*.025f;
                           hp->mod->Tick(50);
                           hp->Render(hwnd);
                           hp->in = false;
                        }
                     }
                     break;
                  }
                  return 0;
               }
            };

            S_hlp hlp;
            hlp.igraph = ed->GetIGraph();
            hlp.ed_scene = ed->GetScene();
            hlp.ec = this;
            hlp.buf = buf;
            hlp.last_mission_path = last_mission_path;
            {
               char buf[MAX_PATH];
               _getcwd(buf, sizeof(buf));
               hlp.cwd = buf;
            }

            switch(id){
            case E_CREATE_BROWSE_MODEL:
               last_path = &last_model_path; 
               if(!last_model_path.Size())
                  last_model_path = ed->GetModelCache().NumDirs() ? ed->GetModelCache().GetDir(0) : "";
               title = "Open model";
               extensions =
                  "Insanity 3D files (*.i3d)\0*.i3d\0"
                  "All files\0*.*\0";

               hlp.mode = S_hlp::MODE_MODEL;
               {              //init rendering
                  hlp.cache_model = &ed->GetModelCache();
                  hlp.mod = NULL;

                  PI3D_driver drv = ed->GetDriver();
                  hlp.driver = drv;
                  hlp.scene = drv->CreateScene();
                  hlp.scene->SetBgndColor(S_vector(.2f, .2f, .4f));
                  //hlp.scene->SetViewport(0, 0, 128, 128);

                  hlp.cam = drv->CreateCamera();
                  hlp.cam->SetFOV(PI*.4f);
                  hlp.scene->AddFrame(hlp.cam);
                  hlp.scene->SetActiveCamera(hlp.cam);
                  hlp.cam->Release();

                  PI3D_light lp = drv->CreateLight();
                  lp->SetLightType(I3DLIGHT_AMBIENT);
                  lp->SetColor(S_vector(1, 1, 1));
                  lp->SetPower(.4f);
                  hlp.scene->AddFrame(lp);
                  hlp.scene->GetPrimarySector()->AddLight(lp);
                  lp->Release();

                  lp = drv->CreateLight();
                  lp->SetLightType(I3DLIGHT_DIRECTIONAL);
                  lp->SetColor(S_vector(1, 1, 1));
                  lp->SetPower(2.0f);
                  lp->SetDir(S_vector(1, -1, 1));
                  lp->Update();
                  hlp.scene->AddFrame(lp);
                  hlp.scene->GetPrimarySector()->AddLight(lp);
                  lp->Release();
               }
               on.lpTemplateName = "IDD_OPEN_MODEL";
               hlp.e_mission = ed->FindPlugin("Mission");
               break;
            case E_CREATE_BROWSE_SOUND: 
               last_path = &last_sound_path; 
               if(!last_sound_path.Size())
                  last_sound_path = ed->GetSoundCache().NumDirs() ? ed->GetSoundCache().GetDir(0) : "";
               title = "Open sound";
               extensions = "Wave files (*.wav, *.ogg)\0*.wav;*.ogg\0""All files\0*.*\0";
               //def_ext = "wav";

               {
                  PI3D_driver drv = ed->GetDriver();
                  hlp.driver = drv;
                  hlp.mode = S_hlp::MODE_SOUND;
                  hlp.sound_cache = &ed->GetSoundCache();
                  hlp.snd = NULL;
                  hlp.play_snd = NULL;
                  PISND_driver isound = drv->GetSoundInterface();
                  if(isound) isound->CreateSound(&hlp.play_snd);
               }
               on.lpTemplateName = "IDD_OPEN_SOUND";
               ed->GetScene()->AttenuateSounds();
               break;
            }

            on.hwndOwner = (HWND)ed->GetIGraph()->GetHWND();
            on.lpstrFilter = extensions;
            on.nFilterIndex = 1;
            on.lpstrFile = buf;
            buf[0] = 0;
            on.nMaxFile = 256;
            on.lpstrInitialDir = *last_path;
            on.lpstrTitle = title;
            on.Flags |= OFN_HIDEREADONLY | OFN_NOCHANGEDIR | OFN_ENABLETEMPLATE | OFN_ENABLEHOOK | OFN_EXPLORER;
            on.lpstrDefExt = def_ext;

            on.lpfnHook = S_hlp::cbOpen;
            on.hInstance = GetHInstance();
            on.lCustData = (LPARAM)&hlp;

            bool b = GetOpenFileName(&on);

            switch(hlp.mode){
            case S_hlp::MODE_MODEL:
               hlp.scene->Release();
               if(hlp.mod) hlp.mod->Release();
               break;
            case S_hlp::MODE_SOUND:
               if(hlp.snd) hlp.snd->Release();
               if(hlp.play_snd) hlp.play_snd->Release();
               break;
            }
            if(b){
               if(!hlp.mission_model){
                  (*last_path) = buf;
                  for(int i=last_path->Size(); i--; )
                     if((*last_path)[i]=='\\')
                        break;
                  (*last_path)[i+1] = 0;

                  switch(id){
                  case E_CREATE_BROWSE_MODEL:
                     if(!ed->GetModelCache().GetRelativeDirectory(buf))
                        b = false;
                     break;
                  case E_CREATE_BROWSE_SOUND:
                     if(!ed->GetSoundCache().GetRelativeDirectory(buf))
                        b = false;
                     break;
                  }
               }else{
                  last_mission_path = buf;
                  C_fstring tmp("+%s", buf);
                  strcpy(buf, tmp);
               }
               if(b && !hlp.mission_model){
                           //strip extension
                  for(int i=strlen(buf); i--; ){
                     if(buf[i]=='.'){
                        buf[i] = 0;
                        break;
                     }
                  }
               }
            }
            return b;
         }
         break;

      case E_CREATE_DELETE_NAMED:
         {
            if(!ed->CanModify()) break;
            const char *cp = (const char*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(cp, ENUMF_BACKDROP | ENUMF_ALL);
            if(frm){
                              //remove from selection
               e_slct->Action(E_SELECTION_DEL_FRAME, frm);

               e_slct->Action(E_SELECTION_FLASH_FRAME, frm);
               DeleteFrame(frm);
            }else
               ed->Message("Create(E_CREATE_DELETE_NAMED): cannot find specified frame");
         }
         break;

      case E_CREATE_DELETE_FRAME:
         {
            if(!ed->CanModify()) break;
            PI3D_frame frm = (PI3D_frame)context;
                              //remove from selection
            e_slct->Action(E_SELECTION_DEL_FRAME, frm);

            return DeleteFrame(frm);
         }
         break;

      case E_CREATE_DELETE_SELECTION:
         {
            if(!ed->CanModify()) break;

            LPC_editor_item e_medit = ed->FindPlugin("MouseEdit");
            const S_MouseEdit_user_mode *umode = (const S_MouseEdit_user_mode*)e_medit->Action(E_MOUSE_GET_USER_MODE);
            if(umode){
               umode->ei->Action(umode->modes[S_MouseEdit_user_mode::MODE_DELETE].action_id);
               break;
            }

            vector<PI3D_frame> sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            int num = 0;
            for(int i=0; i<sel_list.size(); i++){
               if(DeleteFrame(sel_list[i]))
                  ++num;
            }
            ed->Message(C_fstring("%i object(s) deleted", num));
         }
         break;

      case E_CREATE_NOTIFY_SELECTION:
         {
            vector<PI3D_frame> *sel_list = (vector<PI3D_frame>*)context;
            //ed->EnableMenu(this, E_CREATE_DUPLICATE_SELECTION, (sel_list->size()==1));
            ed->EnableMenu(this, E_CREATE_DELETE_SELECTION, (sel_list->size()!=0));
         }
         break;

         /*
      case E_CREATE_PICK_OCCL_VERTEX:
         {
            if(!ed->CanModify()) break;
            PI3D_visual close_vis;
            PI3D_mesh_base close_mesh;
            int closest_point;
            bool b = GetOcclPickInfo(ed->GetIGraph()->Mouse_x(), ed->GetIGraph()->Mouse_y(),
               &close_vis, &close_mesh, &closest_point);
            if(b && closest_point!=-1){
                              //toggle this vertex
               for(int i=occl_pick_verts.size(); i--; ){
                  S_occl_pick_vertex &opv = occl_pick_verts[i];
                  if(opv.vis==close_vis && opv.vertex_index==closest_point){
                     opv = occl_pick_verts.back(); occl_pick_verts.pop_back();
                     break;
                  }
               }
               if(i==-1){
                  occl_pick_verts.push_back(S_occl_pick_vertex());
                  occl_pick_verts.back().vis = close_vis;
                  occl_pick_verts.back().vertex_index = closest_point;
               }
            }
            ed->Message(C_fstring("Vertices: %i", occl_pick_verts.size()), 2);
            return true;
         }
         break;

      case E_CREATE_VERT_SELECTION:
         {
            if(!ed->CanModify()) break;
            if(occl_pick_verts.size())
               break;

            const char *mem = (const char*)context;

            for(int i=0; mem[i]; ){
               PI3D_visual vis = I3DCAST_VISUAL(ed->GetScene()->FindFrame(&mem[i], ENUMF_VISUAL));
               i += strlen(mem+i) + 1;
               dword vi = *(dword*)(mem+i);
               i += sizeof(dword);
               if(vis){
                  occl_pick_verts.push_back(S_occl_pick_vertex());
                  occl_pick_verts.back().vis = vis;
                  occl_pick_verts.back().vertex_index = vi;
               }
            }
            if(occl_pick_verts.size()){
               S_MouseEdit_pick mep = {
                  this, E_CREATE_PICK_OCCL_VERTEX, LoadCursor(GetHInstance(), "IDC_MAKE_OCCLUDER")
               };
               ed->FindPlugin("MouseEdit")->Action(E_MOUSE_SET_USER_PICK, &mep);
               ed->RegForRender(this, 1000);
            }

            e_undo->Action(E_UNDO_SAVE, &S_edit_Undo_2(GetName(), E_CREATE_DEL_VERT_SEL));
         }
         break;
         */

      case E_CREATE_DEL_VERT_SEL:
         if(!ed->CanModify()) break;
         //CancelOcclPickMode();
         break;

      case EDIT_ACTION_BEFORE_FREE:
         //CancelOcclPickMode(false);
         break;
      }
      return 0;
   }

//----------------------------

   void Render(){

      /*
      LPC_editor_item e_medit = ed->FindPlugin("MouseEdit");
      if(!e_medit) return;
      if(e_medit->Action(E_MOUSE_GET_USER_PICK)!=(dword)this){
         CancelOcclPickMode();
         return;
      }
      PI3D_driver drv = ed->GetDriver();
      PI3D_scene scene = ed->GetScene();
      
      PI3D_material occl_pick_mat = drv->CreateMaterial();
      occl_pick_mat->SetDiffuse(S_vector(0, 1, 0));

      S_vector draw_offset = scene->GetActiveCamera()->GetWorldDir() * -.05f;

                              //check if we're in edit mode
      if(e_medit->Action(E_MOUSE_GET_MODE)==0){
                              //draw current selection as well as close vertices
         PI3D_visual close_vis;
         PI3D_mesh_base close_mesh;
         int closest_point;
         bool b = GetOcclPickInfo(ed->GetIGraph()->Mouse_x(), ed->GetIGraph()->Mouse_y(),
            &close_vis, &close_mesh, &closest_point);
         if(b){

                              //draw vertices of pointed frame
            S_matrix m = close_vis->GetMatrix();
            m(3) += draw_offset;

            const S_vector *vp = close_mesh->LockVertices();
            dword vs = close_mesh->GetSizeOfVertex();
            assert(vp);
            for(int i=close_mesh->NumVertices(); i--; ){
               const S_vector &v1 = *(const S_vector*)(((byte*)vp) + vs * i);
               S_vector v = v1 * m;
               byte alpha = 0xc0;
               if(i==closest_point) alpha = 0;
               scene->DrawSprite(v, occl_pick_mat, alpha, .2f);
            }
            close_mesh->UnlockVertices();
         }
      }
                              //draw all selected verts in a special color
      occl_pick_mat->SetDiffuse(S_vector(1, 0, 0));
      for(int i=occl_pick_verts.size(); i--; ){
         const S_occl_pick_vertex &opv = occl_pick_verts[i];
         PI3D_mesh_base mb = opv.vis->GetMesh();
         const S_vector *vp = mb->LockVertices();
         const S_matrix &m = opv.vis->GetMatrix();
         S_vector v = *(const S_vector*)(((byte*)vp) + mb->GetSizeOfVertex() * opv.vertex_index);
         v = v * m + draw_offset;
         scene->DrawSprite(v, occl_pick_mat, 0, .1f);
         mb->UnlockVertices();
      }
      occl_pick_mat->Release();
      */
   }

//----------------------------
#define SAVE_VERSION 0x03

   virtual bool LoadState(class C_chunk *ck){
                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version!=SAVE_VERSION)
         return false;
                              //read other variables
      ck->Read(&mod_show_prv, sizeof(byte));
      ck->Read(&mod_anim_prv, sizeof(byte));
      ck->Read(&snd_show_prv, sizeof(byte));
      ck->Read(&snd_loop_prv, sizeof(byte));
      ck->Read(&single_volume, sizeof(byte));
      ck->Read(&create_static_vol, sizeof(byte));
      ck->ReadString(CT_NAME, last_model_path);
      ck->ReadString(CT_NAME, last_mission_path);
      ck->ReadString(CT_NAME, last_sound_path);

      return true; 
   }

//----------------------------

   virtual bool SaveState(class C_chunk *ck) const{

                              //write version
      byte version = SAVE_VERSION;
      ck->Write(&version, sizeof(byte));
                              //write other variables
      ck->Write(&mod_show_prv, sizeof(byte));
      ck->Write(&mod_anim_prv, sizeof(byte));
      ck->Write(&snd_show_prv, sizeof(byte));
      ck->Write(&snd_loop_prv, sizeof(byte));
      ck->Write(&single_volume, sizeof(byte));
      ck->Write(&create_static_vol, sizeof(byte));
      ck->WStringChunk(CT_NAME, last_model_path);
      ck->WStringChunk(CT_NAME, last_mission_path);
      ck->WStringChunk(CT_NAME, last_sound_path);
      return true; 
   }
};

//----------------------------

void CreateCreate(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Create;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
