#include "all.h"
#include "common.h"

//----------------------------

class C_edit_DebugLine: public C_editor_item{
   virtual const char *GetName() const{ return "DebugLine"; }

   struct S_db_line_timed{
      S_vector from, to;
      int time, init_time;
      S_vector color;
      byte alpha;
   };
   vector<S_db_line_timed> db_lines, db_lines_s, db_lines_t;

   void Update(int time){
                              //timed lines are fading
      for(int i=db_lines_t.size(); i--; ){
         if((db_lines_t[i].time -= time) <= 0){
            db_lines_t[i] = db_lines_t.back();
            db_lines_t.pop_back();
         }
      }
   }
   C_smart_ptr<I3D_frame> col_from, col_to;

   enum E_COL_TYPE{
      CT_NO,
      CT_STANDARD,
      CT_EXACT,
      CT_SPHERE,
      CT_STATIC_SPHERE,
   } line_col_test_type;

   void CheckColTypeMenu(){
      ed->CheckMenu(this, E_DEBUGLINE_COLTEST_EXACT, (line_col_test_type==CT_EXACT));
      ed->CheckMenu(this, E_DEBUGLINE_COLTEST_STANDARD, (line_col_test_type==CT_STANDARD));
      ed->CheckMenu(this, E_DEBUGLINE_COLTEST_SPHERE, (line_col_test_type==CT_SPHERE));
      ed->CheckMenu(this, E_DEBUGLINE_COLTEST_STATIC_SPHERE, (line_col_test_type==CT_STATIC_SPHERE));
   }
public:
   C_edit_DebugLine():
      line_col_test_type(CT_NO)                                         
   {}
   virtual bool Init(){
      ed->AddShortcut(this, E_DEBUGLINE_CLEAR, "%30 &Debug\\%50 Clear lines\tCtrl+C", K_C, SKEY_CTRL);
      ed->AddShortcut(this, E_DEBUGLINE_COLTEST_STANDARD, "%30 &Debug\\%50 &Standard collision", 0, 0);
      ed->AddShortcut(this, E_DEBUGLINE_COLTEST_EXACT, "%30 &Debug\\%50 &Exact collision", 0, 0);
      ed->AddShortcut(this, E_DEBUGLINE_COLTEST_SPHERE, "%30 &Debug\\%50 &Sphere collision", 0, 0);
      ed->AddShortcut(this, E_DEBUGLINE_COLTEST_STATIC_SPHERE, "%30 &Debug\\%50 &Static sphere collision", 0, 0);

      ed->RegForRender(this, 1120);

      CheckColTypeMenu();
      return true;
   }
   virtual void Close(){ }

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){
      Update(time);

                              //debug collision testing
      if(line_col_test_type!=CT_NO){
         if(line_col_test_type == CT_STATIC_SPHERE){
            PI3D_frame sfrm = ed->GetScene()->FindFrame("debugsphere", ENUMF_VOLUME);
            if(sfrm){
               I3D_collision_data col_data;
               col_data.flags |= I3DCOL_STATIC_SPHERE;
               col_data.from = sfrm->GetWorldPos();
               col_data.radius = sfrm->GetWorldScale();
               ed->GetScene()->TestCollision(col_data);
            }
            return;
         }
         if(!col_from)
            col_from = ed->GetScene()->FindFrame("from");
         if(!col_to)
            col_to = ed->GetScene()->FindFrame("to");
         if(col_from && col_to){
            const S_vector &from = col_from->GetWorldPos(), to = col_to->GetWorldPos();
            float dist;
            S_vector dir = to-from;
            S_vector c_norm;
            S_vector col_pt = to;

            PI3D_frame cf = NULL;
            PI3D_frame sfrm = NULL;
            switch(line_col_test_type){
            case CT_SPHERE:
               sfrm = ed->GetScene()->FindFrame("debugsphere", ENUMF_VOLUME);
                              //flow...
            case CT_EXACT:
            case CT_STANDARD:
               {
                  I3D_collision_data col_data;
                  col_data.from = from;
                  col_data.dir = dir;
                  col_data.flags = I3DCOL_TEXTURE;
                  col_data.frm_ignored = sfrm;
                  switch(line_col_test_type){
                  case CT_EXACT: col_data.flags |= I3DCOL_EXACT; break;
                  case CT_SPHERE:
                     {
                        //col_data.flags |= I3DCOL_SPHERE | I3DCOL_SOLVE_SLIDE;
                        col_data.flags |= I3DCOL_SPHERE;
                        if(sfrm)
                           col_data.radius = sfrm->GetWorldScale();
                        else
                           col_data.radius = 1.0f;
                     }
                     break;
                  case CT_STANDARD:
                     col_data.flags |= I3DCOL_JOINT_VOLUMES;
                     break;
                  }
                  bool b;
                  //for(int i=1000; i--; )
                  b = ed->GetScene()->TestCollision(col_data);
                  if(b){
                     cf = col_data.GetHitFrm();
                     dist = col_data.GetClosestHit();
                     c_norm = col_data.GetHitNormal();                     
                  }
                  if(line_col_test_type==CT_SPHERE){
                     if(sfrm){
                        sfrm->SetPos(col_data.GetDestinationPos());
                        sfrm->Update();
                     }
                     if(cf){
                        LPC_editor_item ei = ed->FindPlugin("DebugLine");
                        ei->Action(E_DEBUGLINE_ADD_SINGLE, &S_add_line(
                           col_data.GetDestinationPos(), col_data.GetHitPosition(), 0xffffff00));
                        col_pt = col_data.GetHitPosition();
                     }
                  }else{
                     float mag = dir.Magnitude();
                     if(!IsMrgZero(mag))
                        col_pt = from + dir * dist / mag;
                  }
               }
               break;
            }

            LPC_editor_item ei = ed->FindPlugin("DebugLine");
            assert(ei);
            if(!cf){
                           //single line
               ei->Action(E_DEBUGLINE_ADD_SINGLE, &S_add_line(from, to, 0x80ff8080));
               ed->Message("<no collision>");
            }else{
               //S_vector col_pt = from + dir*dist/dir.Magnitude();
                           //paint line in 2 parts
               ei->Action(E_DEBUGLINE_ADD_SINGLE, &S_add_line(from, col_pt, 0xffff0000));
               ei->Action(E_DEBUGLINE_ADD_SINGLE, &S_add_line(col_pt, to, 0x80ff8080));
                           //paint also normal
               ei->Action(E_DEBUGLINE_ADD_SINGLE, &S_add_line(col_pt, col_pt+c_norm*.25f, 0xffff00ff));
               ed->Message(C_fstring("Hit: %s", cf->GetName()));
            }
         }
      }
   }

   virtual dword Action(int id, void *context){

      switch(id){
      case E_DEBUGLINE_ADD:
      case E_DEBUGLINE_ADD_SINGLE:
      case E_DEBUGLINE_ADD_TIMED:
         {
            S_add_line &al = *(S_add_line*)context;
            byte a = ((al.color>>24)&255);
            if(!a) break;     //don't add fully translucent line
            S_db_line_timed *l;
            switch(id){
            case E_DEBUGLINE_ADD:
               db_lines.push_back(S_db_line_timed());
               l = &db_lines.back();
               break;
            case E_DEBUGLINE_ADD_SINGLE:
               db_lines_s.push_back(S_db_line_timed());
               l = &db_lines_s.back();
               break;
            case E_DEBUGLINE_ADD_TIMED:
               if(!al.time) return 0;
               db_lines_t.push_back(S_db_line_timed());
               l = &db_lines_t.back();
               l->init_time = l->time = al.time;
               break;
            }
            l->from = al.v0;
            l->to = al.v1;
            l->color = S_vector(((al.color>>16)&255)/255.0f, ((al.color>>8)&255)/255.0f, ((al.color>>0)&255)/255.0f);
            l->alpha = 255 - a;
         }
         break;

      case E_DEBUGLINE_CLEAR:
         {
            db_lines.clear();
            db_lines_s.clear();
            db_lines_t.clear();
         }
         break;

      case E_DEBUGLINE_COLTEST_EXACT:
         line_col_test_type = (line_col_test_type==CT_EXACT) ? CT_NO : CT_EXACT;
         CheckColTypeMenu();
         break;

      case E_DEBUGLINE_COLTEST_STANDARD:
         line_col_test_type = (line_col_test_type==CT_STANDARD) ? CT_NO : CT_STANDARD;
         CheckColTypeMenu();
         break;

      case E_DEBUGLINE_COLTEST_SPHERE:
         line_col_test_type = (line_col_test_type==CT_SPHERE) ? CT_NO : CT_SPHERE;
         CheckColTypeMenu();
         break;

      case E_DEBUGLINE_COLTEST_STATIC_SPHERE:
         line_col_test_type = (line_col_test_type==CT_STATIC_SPHERE) ? CT_NO : CT_STATIC_SPHERE;
         CheckColTypeMenu();
         break;

      case EDIT_ACTION_BEFORE_FREE:
         Action(E_DEBUGLINE_CLEAR, 0);
         col_from = NULL;
         col_to = NULL;
         break;
      }
      return 0;
   }
   virtual void Render(){

      int i;
      PI3D_scene scene = ed->GetScene();
      scene->SetRenderMatrix(GetIdentityMatrix());
      for(i = db_lines.size(); i--; )
         scene->DrawLine(db_lines[i].from, db_lines[i].to, db_lines[i].color, db_lines[i].alpha);
      for(i = db_lines_s.size(); i--; )
         scene->DrawLine(db_lines_s[i].from, db_lines_s[i].to, db_lines_s[i].color, db_lines_s[i].alpha);
      db_lines_s.clear();
      for(i = db_lines_t.size(); i--; ){
         const S_db_line_timed &lt = db_lines_t[i];
         int a = lt.alpha;
         a = 255 - (255-a) * ((float)lt.time/(float)lt.init_time);
         scene->DrawLine(lt.from, lt.to, lt.color, a);
      }
   }

   virtual bool LoadState(class C_chunk *ck){
      
      ck->Read(&line_col_test_type, sizeof(line_col_test_type));
      CheckColTypeMenu();

      return true; 
   }

   virtual bool SaveState(class C_chunk *ck) const{

      ck->Write(&line_col_test_type, sizeof(line_col_test_type));

      return true; 
   }
};

//----------------------------

void CreateDebugLine(LPC_editor ed){
   LPC_editor_item ei = new C_edit_DebugLine;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
