#include "all.h"
#include "common.h"
#include <advanced.h>         //Load
#ifdef _DEBUG
# include <crtdbg.h>
#endif

//----------------------------

#define FULL_ACCESS           //don't make access checking
//#define LIMITED_FUNCTIONALITY //allow only basic functions for viewing scene

//----------------------------

#define DEBUG_NO_LEAK_WARNING //no warning about unreleased memory

//----------------------------

BOOL WINAPI DllMain(HINSTANCE hinstDLL, dword fdwReason, void *lpvReserved){

   switch(fdwReason){
   case DLL_THREAD_ATTACH:
      break;

   case DLL_THREAD_DETACH:
      break;

   case DLL_PROCESS_ATTACH:
      {
                              //this func switches to MSVC mem allocation, instead of system allocation
         _set_sbh_threshold(1016);
#if defined _MSC_VER && defined _DEBUG && !defined DEBUG_NO_LEAK_WARNING
                              //report memory leaks at exit
         _CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_LEAK_CHECK_DF);

                              //report mode
         _CrtSetReportMode(_CRT_ERROR, _CRTDBG_MODE_WNDW);
         _CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_WNDW);
         _CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_WNDW);
         struct S_hlp{
            static int __cdecl CrtDumpClient(int report_type, char *msg, int *ret_val){

               static bool called;
               if(!called){
                  _CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) & ~_CRTDBG_LEAK_CHECK_DF);
                  called = true;
               }
               if(report_type==_CRT_WARN){
                                          //display message that some errors occured
                  int i = MessageBox(NULL,
                     msg,
                     "IEditor.dll", MB_ABORTRETRYIGNORE);
                  switch(i){
                  case IDABORT:
                     abort();
                     break;
                  case IDRETRY:
                     *ret_val = 1;
                     break;
                  default:
                     *ret_val = 0;
                  }
               }
               return true;
            }
         };
         _CrtSetReportHook(S_hlp::CrtDumpClient);
         //_CrtSetBreakAlloc(50);
#endif
      }
      break;

   case DLL_PROCESS_DETACH:
      break;
   }
   return true;
}

//----------------------------

class I3D_driver_internal: public I3D_driver{
public:
   I3DMETHOD_(void,Dummy)() = 0;
   I3DMETHOD(GetVersion)(dword&,dword&) = 0;
};

//----------------------------

static bool GetFileName(LPIGraph igraph, const char *title, const char *filter,
   const char *def_ext, char *buf1, dword bufsize, const char *init_dir){

   //igraph->FlipToGDI();
   OPENFILENAME on;
   memset(&on, 0, sizeof(on));
   on.lStructSize = sizeof(on);
   on.hwndOwner = (HWND)igraph->GetHWND();
   on.lpstrFilter = filter;
   on.nFilterIndex = 1;
   on.lpstrFile = buf1;
   buf1[0] = 0;
   on.nMaxFile = bufsize;
   on.lpstrInitialDir = init_dir;
   on.lpstrTitle = title;
   on.Flags = OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
   on.lpstrDefExt = def_ext;
   return (GetOpenFileName(&on)!=0);
}

//----------------------------
//----------------------------
                              //editor implementation
class C_editor_imp: public C_editor{
   dword ref;
   struct S_edit_render_item{
      LPC_editor_item itm;
      int order;
   };
   vector<C_smart_ptr<C_editor_item> > plugin_list;   //installed plugins
   list<S_edit_render_item> render_list;        //plugins registered for rendering
   
   struct S_status_bar_msg{
      C_string str;
      int pos;
      bool modified;
      S_status_bar_msg():
         modified(false)
      {
      }
   };
   vector<S_status_bar_msg> status_bar_strings;
   
   PI3D_driver driver;
   PI3D_scene scene;

                              //status bar - it's derived from S_menu_item to be notified about screen changes
   class C_edit_StatusBar: public C_editor_item{
      virtual const char *GetName() const;
   public:
      void *hwnd_sb;
      virtual dword Action(int id, void *context = NULL){ 
         switch(id){
         case EDIT_ACTION_WINDOW_RESET:
            WinInit((LPIGraph)context);
            break;
         case E_STATUSBAR_GET_HWND:
            return (dword)hwnd_sb;
         }
         return 0;
      }
      void WinInit(LPIGraph);
      virtual bool Init(){ return true; }
   } edit_StatusBar;
   void RefreshStatusBar();

   bool re_enter_block;       //flag set when inside of editor, to avoid re-entering
   bool scene_modified;
   int sb_modify;             //status bar column for modify mark
   bool active;
   bool menu_redraw;          //menu needs to be redrawn
   int last_menu_id;

   C_I3D_cache_base<I3D_model> *model_cache;
   C_I3D_cache_base<I3D_sound> *sound_cache;

//----------------------------

   dword GenerateMenuID(){
      return ++last_menu_id;
   }

//----------------------------
   static dword I2DAPI cbGraph(dword msg, dword par1, dword par2, void *context){

      switch(msg){
      case CM_COMMAND:
         {
            C_editor_imp *ed = (C_editor_imp*)context;
            ed->MenuHit(par1, K_NOKEY, 0);
         }
         break;
      }
      return 0;
   }

//----------------------------

   bool Init(PI3D_driver, PI3D_scene, C_I3D_cache_base<I3D_model> *model_cache,
      C_I3D_cache_base<I3D_sound> *sound_cache);
   virtual void Close();
   friend C_editor __declspec(dllexport) *CreateEditor(PI3D_driver, PI3D_scene, C_I3D_cache_base<I3D_model> *model_cache,
      C_I3D_cache_base<I3D_sound> *sound_cache);
public:
   C_editor_imp():
      ref(1),
      last_menu_id(0),
      driver(NULL),
      scene(NULL),
      scene_modified(false),
      re_enter_block(false),
      active(true),
      menu_redraw(false)
      //sb_modify(0)
   {
      //edit_StatusBar.hwnd_sb = NULL;
   }
   virtual ~C_editor_imp(){
      Close();
   }

   virtual dword AddRef(){ return ++ref; }
   virtual dword Release(){ if(--ref) return ref; Close(); delete this; return 0; }

   virtual void MenuHit(int menu_id, char key, byte key_modifier);
   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt);
   virtual void Render();
   virtual int CreateStatusIndex(int width);
   virtual void Message(const char *str, int sb_index = 0, E_EDIT_MESSAGE = EM_MESSAGE, bool immediate = false);
   virtual void BroadcastAction(E_EDIT_ACTION, void *context);
   virtual void EnumPlugins(bool(*)(LPC_editor_item, void *context), void *context);

   virtual bool LoadState(const char *bin_name);
   virtual bool SaveState(const char *bin_name) const;

//----------------------------
// Check/uncheck modify flag.
   virtual void SetModified(bool);

//----------------------------
// Check if edited scene is modified.
   virtual bool IsModified() const{ return scene_modified; }

//----------------------------
// Query if modifications can be made.
   virtual bool CanModify(bool quiet);

   virtual bool InstallPlugin(LPC_editor_item);
   virtual bool RemovePlugin(LPC_editor_item);
   virtual bool RegForRender(LPC_editor_item, int render_order = 1000);
   virtual bool UnregForRender(LPC_editor_item);

                              //called by plugins during insitialization
   virtual bool AddShortcut(LPC_editor_item ei, int id, const char *menu_item, char key, byte key_modifier);
   virtual bool DelShortcut(LPC_editor_item ei, int id);

   virtual bool IsInit() const{ return (driver!=NULL); }
   virtual LPIGraph GetIGraph() const{ return driver->GetGraphInterface(); }
   virtual PI3D_driver GetDriver() const{ return driver; }
   virtual PI3D_scene GetScene() const{ return scene; }

   virtual LPC_editor_item FindPlugin(const char *plugin_name) const;
                              //functions to be called bu plug-ins
   virtual void CheckMenu(LPC_editor_item, int id, bool);
   virtual void EnableMenu(LPC_editor_item, int id, bool);

   virtual void SetActive(bool);
   virtual bool IsActive() const{ return active; }

   virtual C_I3D_cache_base<I3D_model> &GetModelCache(){ return *model_cache; }
   virtual C_I3D_cache_base<I3D_sound> &GetSoundCache(){ return *sound_cache; }
};

//----------------------------

const char *C_editor_imp::C_edit_StatusBar::GetName() const{ return "StatusBar"; }

//----------------------------

void C_editor_imp::C_edit_StatusBar::WinInit(LPIGraph igraph){

   dword sx, sy;
   RECT rc2;
   GetWindowRect((HWND)igraph->GetHWND(), &rc2);
   sx = rc2.right-rc2.left;
   sy = rc2.bottom-rc2.top;

   RECT rc_sb;
   GetWindowRect((HWND)hwnd_sb, &rc_sb);
   if(!(igraph->GetFlags()&IG_FULLSCREEN)){
                           //enlarge window for status bar
      sy += (rc_sb.bottom - rc_sb.top);
   }else{
      sy = Min(sy, igraph->Scrn_sy());
   }
   SetWindowPos((HWND)igraph->GetHWND(), NULL, 0, 0, sx, sy,
      SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER);
   RECT rc1;
   GetWindowRect((HWND)igraph->GetHWND(), &rc1);
   OffsetRect(&rc_sb, rc1.left, rc1.top);

   SetWindowPos((HWND)hwnd_sb, NULL, rc_sb.left, rc_sb.top, 0, 0,
      SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | SWP_SHOWWINDOW);
}

//----------------------------

#ifndef FULL_ACCESS
static bool access_enabled = true;
#else
static bool access_enabled = true;
#endif

//----------------------------

void CreateProperties(LPC_editor);
void CreateLog(LPC_editor);
void CreateView(LPC_editor);
void CreateUndo(LPC_editor);
void CreateGrid(LPC_editor);
void CreateSelection(LPC_editor);
void CreateCreate(LPC_editor);
void CreateMouseEdit(LPC_editor);
void CreateHide(LPC_editor);
void CreateScreen(LPC_editor);
void CreateLight(LPC_editor);
void CreateLightMap(LPC_editor);
void CreateHelp(LPC_editor);
void CreateModify(LPC_editor);
void CreateValidity(LPC_editor);
void CreateStats(LPC_editor);
void CreateScreenShot(LPC_editor);
void CreateDebugLine(LPC_editor);
void CreateExit(LPC_editor);
void CreateSeed(LPC_editor);
void CreateGravitation(LPC_editor);
void CreateBSPTool(LPC_editor);
void CreateSubEdit(LPC_editor);
void CreateAccess(LPC_editor);
void CreateVideoGrab(LPC_editor);

//----------------------------

bool C_editor_imp::Init(PI3D_driver d1, PI3D_scene s1, C_I3D_cache_base<I3D_model> *mc,
   C_I3D_cache_base<I3D_sound> *sc){

   {
      I3D_driver_internal *di = (I3D_driver_internal*)d1;
      dword v[2];
      I3D_RESULT ir = di->GetVersion(v[0], v[1]);
      if(I3D_FAIL(ir)) return false;
      if(v[0]!=0x35fe9d7c || v[1]!=0x0ac576b3){
         return false;
      }
   }

   driver = d1;
   driver->AddRef();
   scene = s1;
   scene->AddRef();

   CreateHelp(this);
//#ifndef FULL_ACCESS
   CreateAccess(this);
//#endif
                              //init status bar
   {
      InitCommonControls();
      edit_StatusBar.hwnd_sb = CreateStatusWindow(
         WS_CHILD,
         NULL,
         (HWND)GetIGraph()->GetHWND(),
         0
         );
      //GetIGraph()->ProcessWinMessages();
      edit_StatusBar.WinInit(GetIGraph());

      InstallPlugin(&edit_StatusBar);

                              //alloc first area for standard messages
      CreateStatusIndex(350);
                              //second for modify flag
      sb_modify = CreateStatusIndex(32);
   }
                              //force copy mode in igraph
   //GetIGraph()->UpdateParams(0, 0, 0, IG_UPDATE_USE_COPY, IG_UPDATE_USE_COPY);
   GetIGraph()->AddCallback(cbGraph, this);

   model_cache = mc;
   sound_cache = sc;

   CreateUndo(this);
   CreateSelection(this);
   CreateModify(this);

   CreateMouseEdit(this);
#ifndef LIMITED_FUNCTIONALITY
   CreateGrid(this);
   CreateCreate(this);

   CreateValidity(this);
   CreateHide(this);
   CreateScreen(this);
   CreateLight(this);
   CreateLightMap(this);
   CreateStats(this);
   CreateView(this);
   CreateProperties(this);
   CreateScreenShot(this);
   CreateLog(this);
   CreateDebugLine(this);
   CreateSeed(this);
   CreateGravitation(this);
   CreateBSPTool(this);
   CreateSubEdit(this);
   CreateVideoGrab(this);
   
#endif
   CreateExit(this);

   /*
                              //initialize other plugins from directory
   {
      WIN32_FIND_DATA fd;
      memset(&fd, 0, sizeof(fd));
      //fd.dwFileAttributes =
      HANDLE h = FindFirstFile("Plugins\\*.dll", &fd);
      if(h!=INVALID_HANDLE_VALUE){
         do{
            HINSTANCE hi = LoadLibrary(C_fstring("Plugins\\%s", fd.cFileName));
            if(hi){
               typedef void __stdcall tInitialize(LPC_editor);
               tInitialize *ip = (tInitialize*)GetProcAddress(hi, "_Initialize@4");
               if(ip){
                  (*ip)(this);
                  //FindPlugin("Log")->Action(E_LOG_ADD_TEXT, (void*)(const char*)C_fstring("Initialized plugin: %s", fd.cFileName));
               }else
                  FreeLibrary(hi);
            }
         }while(FindNextFile(h, &fd));
         FindClose(h);
      }
   }
   */
   
   return true;
}

//----------------------------

void C_editor_imp::Close(){

   if(!IsInit()) return;

   for(int i=plugin_list.size(); i--; ){
      plugin_list[i]->Close();
      plugin_list.pop_back();
   }
   plugin_list.clear();

                              //no copy mode
   if(GetIGraph()){
      //GetIGraph()->UpdateParams(0, 0, 0, 0, IG_UPDATE_USE_COPY);
      GetIGraph()->RemoveCallback(cbGraph, this);
   }

   if(scene){
      scene->Release();
      scene = NULL;
   }
   if(driver){
      driver->Release();
      driver = NULL;
   }
   //FreeLibrary((HINSTANCE)hInstance);
   //hInstance = NULL;

   if(edit_StatusBar.hwnd_sb){
      DestroyWindow((HWND)edit_StatusBar.hwnd_sb);
      edit_StatusBar.hwnd_sb = NULL;
   }
}

//----------------------------

bool C_editor_imp::InstallPlugin(LPC_editor_item ei){

   ei->ed = this;
                              //access rights
   if(!access_enabled)
      return false;
#ifndef FULL_ACCESS
   LPC_editor_item itm = FindPlugin("Access");
   if(itm){
      if(itm->Action(0) == 0)
         return false;
   }else{
                              //allow installation of Access and Help plugins without access
      if(strcmp(ei->GetName(), "Access") && strcmp(ei->GetName(), "Help")) return false;
   }
#endif                        //FULL_ACCESS

   if(!GetIGraph()) return false;
   if(ei->Init()){
      plugin_list.push_back(ei);
      return true;
   }
   return false;
}

//----------------------------

bool C_editor_imp::RemovePlugin(LPC_editor_item ei){

   for(int i=plugin_list.size(); i--; ) 
      if(plugin_list[i]==ei) break;
   if(i==-1)
      return false;
   plugin_list.erase(plugin_list.begin()+i);
   return true;
}

//----------------------------

void C_editor_imp::MenuHit(int menu_id, char key, byte key_modifier){

   if(re_enter_block) return;
   re_enter_block = true;
   for(int i=plugin_list.size(); i--; ){
      LPC_editor_item ei = plugin_list[i];
      for(int j=ei->menu_items.size(); j--; ){
         const C_editor_item::S_menu_item &itm = ei->menu_items[j];
         if(itm.enabled)
         if((menu_id && itm.menu_id == menu_id) || (key!=K_NOKEY && itm.key==key && itm.key_modifier==key_modifier)){
            ei->Action(itm.id);
            re_enter_block = false;
            return;
         }
      }
   }
   re_enter_block = false;
}

//----------------------------

void C_editor_imp::RefreshStatusBar(){

   if(edit_StatusBar.hwnd_sb){
      for(int i=status_bar_strings.size(); i--; )
      if(status_bar_strings[i].modified){
         status_bar_strings[i].modified = false;
         SendMessage((HWND)edit_StatusBar.hwnd_sb, SB_SETTEXT, i, (LPARAM)(const char*)status_bar_strings[i].str);
      }
   }
}

//----------------------------

void C_editor_imp::Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){

   if(re_enter_block) return;
   re_enter_block = true;
   for(int i=plugin_list.size(); i--; ){
      LPC_editor_item ei = plugin_list[i];
      ei->Tick(skeys, time, mouse_rx, mouse_ry, mouse_rz, mouse_butt);
   }

   RefreshStatusBar();
   re_enter_block = false;

   if(menu_redraw){
      DrawMenuBar((HWND)GetIGraph()->GetHWND());
      menu_redraw = false;
   }
}

//----------------------------

LPC_editor_item C_editor_imp::FindPlugin(const char *plugin_name) const{

   for(int i=plugin_list.size(); i--; ){
      LPC_editor_item ei = plugin_list[i];
      if(!strcmp(ei->GetName(), plugin_name))
         return ei;
   }
   return NULL;
}

//----------------------------

void C_editor_imp::SetModified(bool b){

   assert(!b || CanModify(false));
   if(scene_modified!=b){
      scene_modified = b;
      Message(b ? "MOD" : "", sb_modify);
   }
}

//----------------------------

bool C_editor_imp::CanModify(bool quiet){

                              //ask all plugins
   for(int i=plugin_list.size(); i--; ){
      LPC_editor_item ei = plugin_list[i];
      int st = ei->Action(EDIT_ACTION_QUERY_MODIFY, (void*)quiet);
      if(st){
         Message(st==1 ? "RD" : "LOCK", sb_modify);
         return false;
      }
   }
   return true;
}

//----------------------------

bool C_editor_imp::RegForRender(LPC_editor_item ei, int render_order){

   UnregForRender(ei);
   list<S_edit_render_item>::iterator it;
   for(it=render_list.begin(); it!=render_list.end(); it++){
      if((*it).order<render_order) break;
   }
   render_list.insert(it, S_edit_render_item());
   --it;
   (*it).itm = ei;
   (*it).order = render_order;
   return true;
}

//----------------------------

bool C_editor_imp::UnregForRender(LPC_editor_item ei){

   list<S_edit_render_item>::iterator it;
   for(it=render_list.begin(); it!=render_list.end(); it++){
      if((*it).itm==ei){
         render_list.erase(it);
         break;
      }
   }
   return false;
}

//----------------------------

void C_editor_imp::Render(){

   //if(!active) return;
   if(re_enter_block) return;
   re_enter_block = true;
   list<S_edit_render_item>::const_iterator it, it1;
   for(it=render_list.begin(); it!=render_list.end(); ){
                              //plugin may be erased from list during call, so save it now
      it1 = it++;
      (*it1).itm->Render();
   }
   re_enter_block = false;
}

//----------------------------

bool C_editor_imp::LoadState(const char *bin_name){
   C_chunk ck;

   if(!ck.ROpen(C_fstring("%s.bin", bin_name)))
      return false;
   CK_TYPE t;
   if(!ck.RAscend(t)) goto fail;
   if(t==CT_EDITOR_STATE)
   while(ck.Size()){
      if(!ck.RAscend(t)) goto fail;
      switch(t){
      case CT_EDITOR_PLUGIN_STATE:
                              //read name of plugin
         {
            C_string str;
            ck.ReadString(CT_NAME, str);
            LPC_editor_item ep = FindPlugin(str);
            if(ep){
               ep->LoadState(&ck);
            }
         }
         break;
      };
      ck.Descend();
   }
   ck.Descend();
   return ck.Close();
fail:
   ck.Close();
   return false;
}

//----------------------------

bool WinCreateDirectoryTree(const char *dst){
   
   SECURITY_ATTRIBUTES sa;
   memset(&sa, 0, sizeof(sa));
   sa.nLength = sizeof(sa);
   int cpos, npos = 0;
   C_string str = dst;
   while(true){
      cpos = npos;
      while(str[++npos]) if(str[npos]=='\\') break;
      if(cpos){
         str[cpos] = 0;
         //bool st =
            CreateDirectory(str, &sa);
         /*
         if(!st)
            return false;
            */
         str[cpos] = '\\';
      }
      if(!str[npos])
         break;
   }
   return true;
}

//----------------------------

bool C_editor_imp::SaveState(const char *bin_name) const{

                              //backup old file
   //CopyFile(C_fstring("%s.bin", bin_name), C_fstring("%s.bak", bin_name), false);

   C_chunk ck;
   if(!ck.WOpen(C_fstring("%s.bin", bin_name))){
                              //try to build path
      bool b = WinCreateDirectoryTree(bin_name);
      if(b){
         b = ck.WOpen(C_fstring("%s.bin", bin_name));
      }
      if(!b){
         MessageBox((HWND)GetIGraph()->GetHWND(),
            C_fstring("Cannot save editor state: failed to open file \"%s.bin\" for writing.\nPlease make sure the file is not read-only.", bin_name),
            "Insanity editor", MB_OK);
         return false;
      }
   }
   
   ck.WAscend(CT_EDITOR_STATE);
   {                          //copyright
      ck.WAscend(CT_COPYRIGHT);
      static const char copyright[] = " - Editor configuration - Copyright (c) 2000  Lonely Cat Games - ";
      ck.Write(copyright, sizeof(copyright)+1);
      ck.Descend();
   }
                              //write all installed plugins
   for(int i=0; i<plugin_list.size(); i++){
      ck.WAscend(CT_EDITOR_PLUGIN_STATE);
      {
         ck.WStringChunk(CT_NAME, plugin_list[i]->GetName());
                              //let the plug-in save its data
         plugin_list[i]->SaveState(&ck);
      }
      ck.Descend();
   }
   ck.Descend();
   return ck.Close();
}

//----------------------------

int C_editor_imp::CreateStatusIndex(int width){

   if(!edit_StatusBar.hwnd_sb)
      return -1;
   int count = SendMessage((HWND)edit_StatusBar.hwnd_sb, SB_GETPARTS, 0, 0);
   int *ip = new int[count+1];
   SendMessage((HWND)edit_StatusBar.hwnd_sb, SB_GETPARTS, count, (LPARAM)ip);
   if(ip[count-1]==-1)
      --count;
   ip[count] = (count ? ip[count-1] : 0) + width;
   SendMessage((HWND)edit_StatusBar.hwnd_sb, SB_SETPARTS, count+1, (LPARAM)ip);
   delete[] ip;
   status_bar_strings.push_back(S_status_bar_msg());
   return count;
}

//----------------------------

void C_editor_imp::Message(const char *str, int index, E_EDIT_MESSAGE msg_type, bool immediate){

   /*
   LPC_editor_item e_stats = FindPlugin("Stats");
   if(e_stats)
      e_stats->Action(E_STATS_ADD_SB_TEXT, (void*)str);
   */
   if(index < status_bar_strings.size()){
      status_bar_strings[index].str = str;
      status_bar_strings[index].modified = true;
      if(immediate){
         status_bar_strings[index].modified = false;
         SendMessage((HWND)edit_StatusBar.hwnd_sb, SB_SETTEXT, index, (LPARAM)(const char*)status_bar_strings[index].str);
      }
   }
}

//----------------------------

void C_editor_imp::BroadcastAction(E_EDIT_ACTION ea, void *context){

   for(int i=plugin_list.size(); i--; ){
      LPC_editor_item ei = plugin_list[i];
      ei->Action(ea, context);
   }
}

//----------------------------

void C_editor_imp::EnumPlugins(bool(*proc)(LPC_editor_item, void *context), void *context){

   for(int i=plugin_list.size(); i--; ){
      LPC_editor_item ei = plugin_list[i];
      if(!(*proc)(ei, context))
         break;
   }
}

//----------------------------

bool C_editor_imp::AddShortcut(LPC_editor_item ei, int id, const char *menu_item, 
   char key, byte key_modifier){

   int i;
                              //check key duplication
   if(key!=K_NOKEY && key!=0){
                              //scan all plugins
      for(int j=-1; j<(int)plugin_list.size(); j++){
         LPC_editor_item ei1 = (j==-1) ? ei : plugin_list[j];
                              //scan all its commands
         for(i=ei1->menu_items.size(); i--; ){
            if(ei1->menu_items[i].key==key && ei1->menu_items[i].key_modifier==key_modifier){
               MessageBox((HWND)GetIGraph()->GetHWND(),
                  C_fstring("Plugin \"%s\": shortcut key already used in \"%s\", not installing.\n"
                     "\"%s\""
                  ,
                  (const char*)ei->GetName(),
                  (const char*)ei1->GetName(), menu_item),
                  "Insanity editor", MB_OK);
               return false;
            }
         }
      }
   }
   

   ei->menu_items.push_back(C_editor_item::S_menu_item());
   C_editor_item::S_menu_item &mi = ei->menu_items.back();
   mi.menu_id = 0;
   mi.hmenu_parent = NULL;
   mi.enabled = true;

   if(menu_item){
      HMENU hm = (HMENU)GetIGraph()->GetMenu();
      if(hm){
         mi.menu_id = GenerateMenuID();
         int i = 0;
         int len = strlen(menu_item);
         int last_pos = 0;
         while(true){
            int j;
                              //process formatting
            int break_mode = 0;
            int want_pos = 50;
            while(i<len && menu_item[i]=='%'){
               ++i;
               if(isalpha(menu_item[i])){
                  switch(menu_item[i++]){
                  case 'i': break_mode |= 1; break;   //insert break flag
                  case 'a': break_mode |= 2; break;   //add break flag
                  }
               }else{
                  int num = 0;
                  sscanf(&menu_item[i], "%i%n", &want_pos, &num);
                  i += num;
               }
               while(i<len && isspace(menu_item[i])) ++i;
            }
            last_pos = i;
            for(; i<len; i++) if(menu_item[i]=='\\') break;
            bool popup = (i<len);
            if(popup){
                                 //locate or create sub-menu
               for(j=GetMenuItemCount(hm); j--; ){
                  char buf[256];
                  GetMenuString(hm, j, buf, 255, MF_BYPOSITION);
                  bool b = !strncmp(menu_item+last_pos, buf, i-last_pos);
                                 //sub-menu located
                  if(b){
                     hm = GetSubMenu(hm, j);
                     break;
                  }
               }
               if(j!=-1){
                  ++i;
                  continue;
               }
            }
                                 //add menu item itself
            char buf[256];
            strncpy(buf, menu_item+last_pos, i-last_pos);
            buf[i-last_pos] = 0;

            MENUITEMINFO m;
            memset(&m, 0, sizeof(m));
            int cnt = GetMenuItemCount(hm);
            for(j=0; j<cnt; j++){
               m.cbSize = sizeof(m);
               m.fMask = MIIM_ID | MIIM_TYPE | MIIM_DATA;
               GetMenuItemInfo(hm, j, true, &m);
               //if(!(m.fType&MFT_SEPARATOR))
               {
                  if(m.dwItemData>want_pos)
                     break;
               }
            }
            if(/*j && */(break_mode&1)){
                              //check if not already break
               memset(&m, 0, sizeof(m));
               m.cbSize = sizeof(m);
               m.fMask = MIIM_ID | MIIM_TYPE | MIIM_DATA;
               GetMenuItemInfo(hm, j-1, true, &m);
               if(!(m.fType&MFT_SEPARATOR)){
                  m.fMask = MIIM_TYPE;
                  m.fType = MFT_SEPARATOR;
                  m.dwItemData = want_pos;
                  InsertMenuItem(hm, j, true, &m);
                  ++j;
               }
            }

            memset(&m, 0, sizeof(m));
            m.cbSize = sizeof(m);
            m.fMask |= MIIM_DATA | MIIM_TYPE;
            m.fType = MFT_STRING;
            if(!popup){
               m.fMask |= MIIM_ID | MIIM_ID;
               m.wID = mi.menu_id;
            }else{
               m.fMask |= MIIM_SUBMENU;
               m.hSubMenu = CreateMenu();
            }
            m.dwTypeData = buf;
            m.cch = strlen(buf);
            m.dwItemData = want_pos;
            InsertMenuItem(hm, j, true, &m);

            if(popup) hm = m.hSubMenu;

            if((break_mode&2)){
               ++j;
               MENUITEMINFO m;
               memset(&m, 0, sizeof(m));
               m.cbSize = sizeof(m);
               m.fMask = MIIM_ID | MIIM_TYPE | MIIM_DATA;
               GetMenuItemInfo(hm, j-1, true, &m);
               if(!(m.fType&MFT_SEPARATOR)){
                  m.fMask = MIIM_TYPE;
                  m.fType = MFT_SEPARATOR;
                  m.dwItemData = want_pos;
                  InsertMenuItem(hm, j, true, &m);
               }
            }
            if(!popup) break;
            last_pos = ++i;
         }
         mi.hmenu_parent = (void*)hm;
//----------------------------
         menu_redraw = true;
      }
   }
   mi.key = key;
   mi.key_modifier = key_modifier;
   mi.id = id;

   return true;
}

//----------------------------

bool C_editor_imp::DelShortcut(LPC_editor_item ei, int id){

   for(int i=ei->menu_items.size(); i--; ){
      C_editor_item::S_menu_item &mi = ei->menu_items[i];
      if(mi.id==id){
                              //delete menu item (if any)
         if(mi.menu_id && mi.hmenu_parent){
            HMENU hmenu_parent = (HMENU)mi.hmenu_parent;
            bool b;
            b = DeleteMenu(hmenu_parent, mi.menu_id, MF_BYCOMMAND);
            assert(b);
            /*
            while(hmenu_parent && !GetMenuItemCount(hmenu_parent)){
               DestroyMenu(hmenu_parent);
               hmenu_parent = NULL;
            }
            */
         }

         mi = ei->menu_items.back(); ei->menu_items.pop_back();
         return true;
      }
   }
   return false;
}

//----------------------------

void C_editor_imp::CheckMenu(LPC_editor_item ei, int id, bool b){
                              //find menu item
   for(int i=ei->menu_items.size(); i--; ){
      const C_editor_item::S_menu_item &itm = ei->menu_items[i];
      if(itm.id==id){
         CheckMenuItem((HMENU)itm.hmenu_parent, itm.menu_id, MF_BYCOMMAND |
            (b ? MF_CHECKED : MF_UNCHECKED));
         break;
      }
   }
}

//----------------------------

void C_editor_imp::EnableMenu(LPC_editor_item ei, int id, bool b){
                              //find menu item
   for(int i=ei->menu_items.size(); i--; ){
      C_editor_item::S_menu_item &itm = ei->menu_items[i];
      if(itm.id==id){
         EnableMenuItem((HMENU)itm.hmenu_parent, itm.menu_id, MF_BYCOMMAND |
            (b ? MF_ENABLED : MF_GRAYED));
         itm.enabled = b;
         break;
      }
   }
}

//----------------------------

void C_editor_imp::SetActive(bool b){

   if(active!=b){
      active = b;
      Message(active ? "Edit mode" : "Game mode");
      BroadcastAction(EDIT_ACTION_SET_ACTIVE, (void*)b);
      if(!active)             //last chance to do so
         RefreshStatusBar();
   }
}

//----------------------------
//----------------------------

static BOOL CALLBACK dlgSplash(HWND hwnd,  // handle to dialog box
   UINT uMsg,     // message  
   WPARAM wParam, // first message parameter
   LPARAM lParam  // second message parameter
   ){

   static igraph_init_flags;
   switch(uMsg){
   case WM_INITDIALOG:
      {
                              //center window
         LPIGraph igraph = (LPIGraph)lParam;
         SetWindowLong(hwnd, GWL_USERDATA, lParam);
         igraph_init_flags = igraph->GetFlags();
         igraph->UpdateParams(0, 0, 0, 0, IG_SUSPENDINACTIVE);

         RECT rc;
         //SetRect(&rc, 0, 0, 350, 230);
         SetRect(&rc, 0, 0, 298, 196);
         //rc.right += 4;
         //rc.bottom += 4;
         int scx = GetSystemMetrics(SM_CXSCREEN);
         int scy = GetSystemMetrics(SM_CYSCREEN);
         int x = //scx - (rc.right-rc.left) - (rc.right-rc.left) / 12;
            rc.left+(scx-(rc.right-rc.left))/2;
         int y = //(rc.bottom-rc.top) / 12;
            rc.top+(scy-(rc.bottom-rc.top))/2;
         SetWindowPos((HWND)hwnd, NULL, x, y, 
            rc.right-rc.left, rc.bottom-rc.top, SWP_NOZORDER);
         igraph->AddRef();
         igraph->AddDlgHWND(hwnd);
         SetTimer(hwnd, 1, 2000, NULL);

         ShowWindow(hwnd, SW_SHOW);
      }
      return 1;

   case WM_TIMER:
      KillTimer(hwnd, 1);
      DestroyWindow(hwnd);
      break;

   case WM_COMMAND:
      switch(LOWORD(wParam)){
      case IDCLOSE:
      case IDCANCEL:
         {
            EndDialog(hwnd, 0);
         }
         break;
      }
      break;
   case WM_DESTROY:
      {
         LPIGraph igraph = (LPIGraph)GetWindowLong(hwnd, GWL_USERDATA);
         igraph->UpdateParams(0, 0, 0, 
            igraph_init_flags&IG_SUSPENDINACTIVE, 
            IG_SUSPENDINACTIVE);
         igraph->DeleteDlgHWND(hwnd);
         igraph->Release();
      }
      break;
   }
   return 0;
}

//----------------------------

static void RegisterViewer(){
                           //register icon and viewer
   int rkey;
   {
      rkey = RegkeyCreate(".i3d", E_REGKEY_CLASSES_ROOT);
      if(rkey!=-1){
         RegkeyWtext(rkey, NULL, "Insanity.Viewer");
         RegkeyClose(rkey);
      }
      rkey = RegkeyCreate("Insanity.Viewer", E_REGKEY_CLASSES_ROOT);
      if(rkey!=-1){
         RegkeyWtext(rkey, NULL, "Insanity 3D scene");
         RegkeyClose(rkey);
      }
      rkey = RegkeyCreate("Insanity.Viewer\\DefaultIcon", E_REGKEY_CLASSES_ROOT);
      if(rkey!=-1){
         char buf[MAX_PATH];
         GetModuleFileName(GetModuleHandle("I3D_2.DLL"), buf, sizeof(buf));
         RegkeyWtext(rkey, NULL, C_fstring("%s,0", buf));
         RegkeyClose(rkey);
      }
      rkey = RegkeyCreate("Insanity.Viewer\\shell\\open\\command", E_REGKEY_CLASSES_ROOT);
      if(rkey!=-1){
         char buf[MAX_PATH];
         GetModuleFileName(NULL, buf, sizeof(buf));
         RegkeyWtext(rkey, NULL, C_fstring("%s model \"%%1\"", buf));
         RegkeyClose(rkey);
      }
      rkey = RegkeyCreate("Software\\Classes\\.I3D", E_REGKEY_LOCAL_MACHINE);
      if(rkey!=-1){
         RegkeyWtext(rkey, NULL, "Insanity.Viewer");
      }
   }
}

//----------------------------

LPC_editor __declspec(dllexport) CreateEditor(PI3D_driver driver, PI3D_scene scene, 
   C_I3D_cache_base<I3D_model> *mc, C_I3D_cache_base<I3D_sound> *sc){

   RegisterViewer();

   C_editor_imp *ed = new C_editor_imp;
   if(!ed->Init(driver, scene, mc, sc)){
      delete ed;
      ed = NULL;
      return NULL;
   }

   if(IsProgrammersPC(ed)){
#if defined _DEBUG && !defined DEBUG_NO_LEAK_WARNING
                              //report memory leaks at exit
      _CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_LEAK_CHECK_DF);
      _CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_WNDW);
#endif
   }else{
                              //show splash screen
      CreateDialogParam(GetHInstance(), "IDD_ABOUT1", (HWND)driver->GetGraphInterface()->GetHWND(), dlgSplash,
         (LPARAM)driver->GetGraphInterface());
   }
   return ed;
}

//----------------------------

