#include "all.h"
#include "common.h"

//----------------------------

#define GRAVITATION_VERSION 7

#define SPEED_SLIDER_MIN  4*100
#define SPEED_SLIDER_MAX  100*100 //in cm

//----------------------------

inline void DEBUGLINE(LPC_editor ed, const S_vector &from, const S_vector &to, int type, dword color = -1, int time = 500){
   LPC_editor_item ei = ed->FindPlugin("DebugLine");
   if(ei) ei->Action(E_DEBUGLINE_ADD + type, &S_add_line(from, to, color, time));
}

//----------------------------
                              //editor plugin: Gravitation
class C_edit_Gravitation: public C_editor_item{
   virtual const char *GetName() const{ return "Gravitation"; }

   enum E_fall_mode{
      FM_GRAVITATION,         //fall down by gravitation
      FM_OBJECT_Z,            //fall by object's local z axis
   } fall_mode;
   float fall_speed;          //meters/sec
   bool link_falled;          //true to link falled object to background

   enum E_collision_mode{
      CM_PIVOT,
      CM_VERTEX,
   } collision_mode;

   HWND hWndGrav;             //NULL if sheet is not displayed


   struct S_falling_object: public C_unknown{
      C_smart_ptr<I3D_frame> frm;
      C_smart_ptr<I3D_interpolator> intp;
      S_vector dest_pos;
   };                                                       

   vector<C_smart_ptr<S_falling_object> > fall_list;

#pragma pack(push,1)
   struct S_grav_undo_struct{
      S_vector pos;
      S_quat rot;
      float scale;
      char name_buf[1];       //frame_name, parent_name
   };
#pragma pack(pop)

//----------------------------

   static LRESULT CALLBACK GravCallbackProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){

      bool changed = false;
	   switch (message){
      case WM_INITDIALOG:
         {
            SetWindowLong(hDlg, GWL_USERDATA, lParam);
            //C_edit_Gravitation *eg = (C_edit_Gravitation*)GetWindowLong(hDlg, GWL_USERDATA);

                              //setup speed track-bar
            SendMessage(GetDlgItem(hDlg, IDC_SPEEDSLIDER), TBM_SETRANGEMIN, true, SPEED_SLIDER_MIN);
            SendMessage(GetDlgItem(hDlg, IDC_SPEEDSLIDER), TBM_SETRANGEMAX, true, SPEED_SLIDER_MAX);
            SendMessage(GetDlgItem(hDlg, IDC_SPEEDSLIDER), TBM_SETPOS, true, ((SPEED_SLIDER_MAX-SPEED_SLIDER_MIN)/2));
            SendMessage(GetDlgItem(hDlg, IDC_SPEEDSLIDER), TBM_SETLINESIZE, 0, (SPEED_SLIDER_MAX-SPEED_SLIDER_MIN) / 20);
            SendMessage(GetDlgItem(hDlg, IDC_SPEEDSLIDER), TBM_SETPAGESIZE, 0, (SPEED_SLIDER_MAX-SPEED_SLIDER_MIN) / 5);

            return true;
         }
         break;

      case WM_HSCROLL:
         {
            C_edit_Gravitation *eg = (C_edit_Gravitation*)GetWindowLong(hDlg, GWL_USERDATA);
            dword value = SendMessage(GetDlgItem(hDlg, IDC_SPEEDSLIDER), TBM_GETPOS, (WPARAM) 0, (LPARAM) 0);
            if(value<SPEED_SLIDER_MIN) value = SPEED_SLIDER_MIN;
            if(value>SPEED_SLIDER_MAX) value = SPEED_SLIDER_MAX;
            eg->fall_speed = value / 100.0f;
            SetDlgItemText(hDlg, IDC_SPEEDVALUE, C_fstring("%.2f", eg->fall_speed));
         }
         break;

      case WM_COMMAND:
         {
            switch(LOWORD(wParam)){

            case((WORD)IDC_HIDE):
               {
                     C_edit_Gravitation *eg = (C_edit_Gravitation*)GetWindowLong(hDlg, GWL_USERDATA);
                     if(HIWORD(wParam) == (WORD)BN_CLICKED)
                        if(eg->hWndGrav)
                           eg->Action(E_GRAV_TOGGLE_DIALOG, NULL);
               }
               break;

            case ((WORD)IDC_LINK):
            case ((WORD)IDC_FALL_Z):
            case ((WORD)IDC_COLM_PIVOT):
            case ((WORD)IDC_COLM_VERTEX):
               if(HIWORD(wParam) == (WORD)BN_CLICKED)
                  changed = true;
               break;
            }
         }
         break;
      }
      if(changed){

         C_edit_Gravitation *eg = (C_edit_Gravitation*)GetWindowLong(hDlg, GWL_USERDATA);

         eg->link_falled = (SendDlgItemMessage(hDlg, IDC_LINK, BM_GETCHECK, 0, 0) == BST_CHECKED);
         eg->fall_mode =
            (SendDlgItemMessage(hDlg, IDC_FALL_Z, BM_GETCHECK, 0, 0) == BST_CHECKED) ?
            FM_OBJECT_Z : FM_GRAVITATION;
         eg->collision_mode = (SendDlgItemMessage(hDlg, IDC_COLM_PIVOT, BM_GETCHECK, 0, 0) == BST_CHECKED) ?
            CM_PIVOT :
            CM_VERTEX;

         return true;
      }
      return 0;
   }

//----------------------------

   void SetDialogByParams(){

      if(!hWndGrav)
         return;

      SendDlgItemMessage(hWndGrav, IDC_SPEEDSLIDER, TBM_SETPOS, true, fall_speed * 100.0f);
      SetDlgItemText(hWndGrav, IDC_SPEEDVALUE, C_fstring("%.2f", fall_speed));
   
      SendDlgItemMessage(hWndGrav, IDC_LINK, BM_SETCHECK, !link_falled ? BST_UNCHECKED : BST_CHECKED, 0);
      SendDlgItemMessage(hWndGrav, IDC_FALL_Z, BM_SETCHECK, (fall_mode==FM_GRAVITATION) ? BST_UNCHECKED : BST_CHECKED, 0);
      SendDlgItemMessage(hWndGrav, IDC_COLM_PIVOT, BM_SETCHECK, (collision_mode==CM_PIVOT) ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hWndGrav, IDC_COLM_VERTEX, BM_SETCHECK, (collision_mode==CM_VERTEX) ? BST_CHECKED : BST_UNCHECKED, 0);
   }

//----------------------------

   void ToggleDialog(){
      if(hWndGrav){
         ed->FindPlugin("Properties")->Action(E_PROP_DEL_SHEET, hWndGrav);
         DestroyWindow(hWndGrav);
         hWndGrav = NULL;
      }else{
         hWndGrav = CreateDialogParam(GetHInstance(), "IDD_GRAVITATION", NULL, (DLGPROC)GravCallbackProc, (LPARAM)this);
         if(hWndGrav){
            ed->FindPlugin("Properties")->Action(E_PROP_ADD_SHEET, hWndGrav);
            SetDialogByParams();
            ed->FindPlugin("Properties")->Action(E_PROP_SHOW_SHEET, hWndGrav);
         }
      }                                                       
   }

//----------------------------

public:

   C_edit_Gravitation():
      fall_mode(FM_GRAVITATION),      
      collision_mode(CM_PIVOT),
      fall_speed((SPEED_SLIDER_MAX-SPEED_SLIDER_MIN)/200.0f),
      link_falled(false),
      hWndGrav(NULL)
   {}
   
//----------------------------

   virtual bool Init(){

      ed->AddShortcut(this, E_GRAV_SELECTED_DOWN, "&Edit\\&Gravitation\\Fall down\tCtrl+G", K_G, SKEY_CTRL);
      ed->AddShortcut(this, E_GRAV_TOGGLE_DIALOG, "&Edit\\&Gravitation\\Properties\tCtrl+R", K_R, SKEY_CTRL);
      ed->AddShortcut(this, E_GRAV_LINK_TO_BACKGROUND, "&Edit\\&Gravitation\\&Link to ground\tCtrl+L", K_L, SKEY_CTRL);

      return true;                                                     
   }

//----------------------------

   virtual void Close(){ 

                              //stop falling all
      fall_list.clear();

      if(hWndGrav)
         ToggleDialog();
   }

//----------------------------
   
   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){

      for(int i = fall_list.size(); i--; ){
         PI3D_interpolator in = fall_list[i]->intp;
         assert(in);
         I3D_RESULT ir = in->Tick(time);
         if(ir == I3D_DONE){
            fall_list[i] = fall_list.back(); fall_list.pop_back();
         }
      }
   }

//----------------------------

   virtual dword Action(int id, void *context);
                            
   virtual void Render(){}

//----------------------------

   virtual bool LoadState(class C_chunk *ck){ 

      dword version;
      ck->Read(&version, sizeof(version));
      if(GRAVITATION_VERSION != version){
         SetDialogByParams();
         return false;       
      }

      ck->Read(&fall_mode, sizeof(fall_mode));
      ck->Read(&collision_mode, sizeof(collision_mode));
      ck->Read(&link_falled, sizeof(link_falled));
      ck->Read(&fall_speed, sizeof(fall_speed));
      SetDialogByParams();
      return true; 
   }

//----------------------------

   virtual bool SaveState(class C_chunk *ck) const{
      
      dword version = GRAVITATION_VERSION;
      ck->Write(&version, sizeof(float));

      ck->Write(&fall_mode, sizeof(fall_mode));
      ck->Write(&collision_mode, sizeof(collision_mode));
      ck->Write(&link_falled, sizeof(link_falled));
      ck->Write(&fall_speed, sizeof(fall_speed));
      return true; 
   }
};

//----------------------------------

dword C_edit_Gravitation::Action(int id, void *context){

   switch(id){
         
   case E_GRAV_LINK_TO_BACKGROUND:
      {
         if(!ed->CanModify()) break;
         const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ed->FindPlugin("Selection")->Action(E_SELECTION_GETVECTOR);
         int num_linked = 0;

         for(int i = sel_list.size(); i--; ){

            PI3D_frame frm = sel_list[i];

            I3D_bound_volume frame_bbox;
            frm->ComputeHRBoundVolume(&frame_bbox);

                              //get center of frame, in world coords
            S_vector frame_centre = (frame_bbox.bbox.min + frame_bbox.bbox.max) * .5f;
            frame_centre *= frm->GetMatrix();

            S_vector link_dir;
            if(!ed->FindPlugin("MouseEdit")->Action(E_MOUSE_GET_POLAR_MODE)){
               link_dir = S_vector(.0f, -1.0f, .0f);
            }else{
               link_dir = -S_normal(frame_centre);
            }

                              //don't collide with itself
            bool is_on = frm->IsOn();
            frm->SetOn(false);

                              //find frame we're going link to
            I3D_collision_data cd;
            cd.from = frame_centre;
            cd.dir = link_dir;
            cd.flags = I3DGEOCOL_TEXTURE | I3DGEOCOL_RAY | I3DCOL_EXACT;
            ed->GetScene()->TestCollision(cd);
            PI3D_frame target = cd.GetHitFrm();

            frm->SetOn(is_on);
            if(!target)
               continue;

                              //check if this is part of model, if so, use model instead
            PI3D_frame mod_frm = target;
            while(mod_frm = mod_frm->GetParent()){
               if(mod_frm->GetType()==FRAME_MODEL)
                  break;
            }
            if(mod_frm)
               target = mod_frm;

                              //if already linked to target, skip
            if(target == frm->GetParent())
               continue;
                              
            {                 //save undo
               assert(frm->GetParent());
               const char *parent_name = frm->GetParent()->GetName();
               const char *frm_name = frm->GetName();
               int len1 = strlen(parent_name);
               int len2 = strlen(frm_name);

               char *buf = new char[len1 + len2 + 2];
               strcpy(buf, frm_name);
               strcpy(buf + len2 + 1, parent_name);

               S_edit_Undo_2 undostruct(GetName(), E_GRAV_SET_PARENT, buf, len1 + len2 + 2);
               ed->FindPlugin("Undo")->Action(E_UNDO_SAVE, &undostruct);
            }
                              //save modification
            LPC_editor_item e_mod = ed->FindPlugin("Modify");
            assert(e_mod);
            e_mod->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(frm, E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE | E_MODIFY_FLG_LINK));

                              //re-link
            frm->LinkTo(target, I3DLINK_UPDMATRIX);
            frm->Update();

            ++num_linked;
         }
         ed->Message(C_fstring("Linked: %i object(s).", num_linked));
      }
      break;

   case E_GRAV_SET_PARENT:
      {
         if(!ed->CanModify()) break;

         char *frame_name = (char *)context;
         char *parent_name = ((char *)context) + strlen(frame_name)+1;

         PI3D_frame frm = ed->GetScene()->FindFrame(frame_name);
         if(frm){
            PI3D_frame parent = ed->GetScene()->FindFrame(parent_name);

            {
               assert(frm->GetParent());
               const char *parent_name = frm->GetParent()->GetName();
               const char *frm_name = frm->GetName();
               int len1 = strlen(parent_name);
               int len2 = strlen(frm_name);

               char *buf = new char[len1 + len2 + 2];
               strcpy(buf, frm_name);
               strcpy(buf + len2 + 1, parent_name);

               S_edit_Undo_2 undostruct(GetName(), E_GRAV_SET_PARENT, buf, len1 + len2 + 2);
               ed->FindPlugin("Undo")->Action(E_UNDO_SAVE, &undostruct);
            }
                              //re-link
            frm->LinkTo(parent, I3DLINK_UPDMATRIX);
            frm->Update();

            ed->Message("Frame re-linked.");
         }else
            ed->Message("Gravitation(E_GRAV_SET_PARENT): can't find frame.", 0, EM_ERROR);
      }
      break;

   case E_GRAV_SELECTED_DOWN:
      {
         if(!ed->CanModify()) break;
         int num_falls = 0;

         const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ed->FindPlugin("Selection")->Action(E_SELECTION_GETVECTOR);

         for(int i = sel_list.size(); i--; ){

            PI3D_frame frm = sel_list[i];

                              //check if already in falling list, if yes, skip this
            for(int k = fall_list.size(); k--; )
            if(fall_list[k]->frm == frm)
               break;
            if(k != -1){
                              //put to destination position
               frm->SetPos(fall_list[k]->dest_pos);
               frm->Update();
                              //remove this interpolation
               fall_list[k] = fall_list.back(); fall_list.pop_back();
            }

                              //determine destination position of our fall
            const S_vector &w_from = frm->GetWorldPos();
            S_vector w_to;
            const S_matrix &mat = frm->GetMatrix();

                              //get direction
            S_vector dir;

            switch(fall_mode){

            case FM_GRAVITATION:
               if(!ed->FindPlugin("MouseEdit")->Action(E_MOUSE_GET_POLAR_MODE)){
                  dir = S_vector(.0f, -1.0f, .0f);
               }else{
                  dir = -mat(3);
               }
               break;

            case FM_OBJECT_Z:
               dir = mat(2);
               break;

            default:
               assert(0);
            }
            if(IsAbsMrgZero(dir.Magnitude()))
               continue;
            dir.Normalize();

            PI3D_frame target;
            float col_dist;
            S_vector add_thresh = dir * .001f;
                              //compute destination position
            switch(collision_mode){

            case CM_VERTEX:
               {
                  struct S_hlp{
                     static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                        PI3D_visual vis = I3DCAST_VISUAL(frm);
                        PI3D_mesh_base mb = vis->GetMesh();
                        if(mb){
                           const S_matrix &m = vis->GetMatrix();
                           vector<S_vector> &vert_list = *(vector<S_vector>*)c;
                           const S_vector *vp = mb->LockVertices();
                           dword vs = mb->GetSizeOfVertex();
                           for(int i=mb->NumVertices(); i--; ){
                              const S_vector &v = *(S_vector*)(((byte*)vp) + vs * i);
                              vert_list.push_back(v * m);
                           }
                           mb->UnlockVertices();
                        }
                        return I3DENUMRET_OK;
                     }
                  };
                              //collect all vertices, in world coordinates
                  vector<S_vector> vert_list;
                  if(frm->GetType() == FRAME_VISUAL)
                     S_hlp::cbEnum(frm, (dword)&vert_list);
                  frm->EnumFrames(S_hlp::cbEnum, (dword)&vert_list, ENUMF_VISUAL);

                  if(vert_list.size()){
                  
                     target = NULL;
                     bool hiden = frm->IsOn();
                     frm->SetOn(false);
                              //find closest collision
                     for(int i=vert_list.size(); i--; ){
                        float dist1;

                        const S_vector &from = vert_list[i] + add_thresh;
                        //DEBUGLINE(ed, from, from + dir, 0);

                        I3D_collision_data cd;
                        cd.from = from;
                        cd.dir = dir;
                        cd.flags = I3DGEOCOL_TEXTURE | I3DGEOCOL_RAY | I3DCOL_EXACT;
                        ed->GetScene()->TestCollision(cd);
                        PI3D_frame t1 = cd.GetHitFrm();
                        dist1 = cd.GetClosestHit();
                        if(t1){
                           dist1 += .001f;
                           if(!target){
                              target = t1;
                              col_dist = dist1;
                           }else
                           if(col_dist > dist1){
                              col_dist = dist1;
                              target = t1;
                           }
                        }
                     }
                     frm->SetOn(hiden);
                     break;
                  }
               }
                              //flow... (failcase to pivot mode if object has no vertices)
            case CM_PIVOT:
               {
                              //don't check with itself
                  bool hiden = frm->IsOn();
                  frm->SetOn(false);
                  const S_vector &from = w_from + add_thresh;

                  I3D_collision_data cd;
                  cd.from = from;
                  cd.dir = dir;
                  cd.flags = I3DGEOCOL_TEXTURE | I3DGEOCOL_RAY | I3DCOL_EXACT;
                  ed->GetScene()->TestCollision(cd);
                  target = cd.GetHitFrm();
                  col_dist = cd.GetClosestHit();

                  frm->SetOn(hiden);
                  if(target){
                     col_dist += .001f;
                  }
               }
               break;

            default:
               assert(0);
            }

            if(!target)
               continue;

            w_to = w_from + dir * col_dist;

                              //don't link to the same parent
            if(!link_falled || target == frm->GetParent())
               target = NULL;

            {                 //save undo
               const char *frm_name = frm->GetName();
               const char *prnt_name = target ? frm->GetParent()->GetName() : "";
               int len1 = strlen(frm_name), len2 = strlen(prnt_name);
               S_grav_undo_struct *us = new(new byte[sizeof(S_grav_undo_struct) + len1 + len2 + 1]) S_grav_undo_struct;
               us->pos = frm->GetPos();
               if(target){
                  us->rot = frm->GetRot();
                  us->scale = frm->GetScale();
               }
               strcpy(us->name_buf, frm_name);
               strcpy(us->name_buf+len1+1, prnt_name);

               S_edit_Undo_2 eu(GetName(), E_GRAV_UNDO_FALL, us, sizeof(S_grav_undo_struct) + len1 + len2 + 1);
               ed->FindPlugin("Undo")->Action(E_UNDO_SAVE, &eu);
               delete (byte*)us;
            }

            {                 //save modification
               pair<PI3D_frame, dword> m_pr(frm, E_MODIFY_FLG_POSITION);
               if(target)
                  m_pr.second |= E_MODIFY_FLG_LINK | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE;
               ed->FindPlugin("Modify")->Action(E_MODIFY_ADD_FLAGS, &m_pr);
            }

            {                 //setup falling
               S_falling_object *obj = new S_falling_object;

               obj->frm  = frm;

               if(target){
                  frm->LinkTo(target, I3DLINK_UPDMATRIX);
               }
               const S_matrix &imat = ~frm->GetParent()->GetMatrix();

               
               I3D_anim_pos_tcb poskeys[2];

               const S_vector &from = frm->GetPos();
               S_vector to = w_to * imat;
               int fall_time = (from - to).Magnitude() / fall_speed * 1000;
               obj->dest_pos = to;

               poskeys[0].Clear();
               poskeys[0].time = 0;
               poskeys[0].easy_from = 1.0f;
               poskeys[0].v = from;
               poskeys[1].Clear();
               poskeys[1].time = fall_time;
               poskeys[1].v = to;
               

               PI3D_interpolator intp = ed->GetScene()->GetDriver()->CreateInterpolator();

               intp->SetFrame(frm);
               {
                  PI3D_keyframe_anim anim = (PI3D_keyframe_anim)ed->GetScene()->GetDriver()->CreateAnimation(I3DANIM_KEYFRAME);
                  //anim->SetOptions(I3DANIM_SIMPLE);
                  anim->SetPositionKeys(poskeys, 2); 
                  anim->SetEndTime(fall_time);
                  intp->SetAnimation(0, anim);
                  anim->Release();
               }
               intp->SetCurrTime(0, 0);

               obj->intp = intp;
               intp->Release();

               fall_list.push_back(obj);
               obj->Release();

               ++num_falls;
            }
         }
         if(num_falls){
            ed->SetModified();
                              //computing collisions might take time,
                              // so reset counter now so that we don't lose ticks
            ed->GetIGraph()->GetTimer(1, 1);
         }

         ed->Message(C_fstring("Gravitation: falling %i frame(s).", num_falls));
      }
      break;

   case E_GRAV_UNDO_FALL:
      {
         if(!ed->CanModify()) break;
         S_grav_undo_struct *us = (S_grav_undo_struct*)context;
         PI3D_frame frm = ed->GetScene()->FindFrame(us->name_buf);
         if(frm){
                              //if frame is currently falling, stop it
            for(int k = fall_list.size(); k--; )
            if(fall_list[k]->frm == frm)
               break;
            if(k != -1){
               frm->SetPos(fall_list[k]->dest_pos);
               frm->Update();
                              //remove this interpolation
               fall_list[k] = fall_list.back(); fall_list.pop_back();
            }
            const char *link_to_name = us->name_buf + strlen(us->name_buf) + 1;

            {                 //save undo
               const char *frm_name = frm->GetName();
               const char *prnt_name = *link_to_name ? frm->GetParent()->GetName() : "";
               int len1 = strlen(frm_name), len2 = strlen(prnt_name);
               S_grav_undo_struct *us = new(new byte[sizeof(S_grav_undo_struct) + len1 + len2 + 1]) S_grav_undo_struct;
               us->pos = frm->GetPos();
               if(*prnt_name){
                  us->rot = frm->GetRot();
                  us->scale = frm->GetScale();
               }
               strcpy(us->name_buf, frm_name);
               strcpy(us->name_buf+len1+1, prnt_name);

               S_edit_Undo_2 eu(GetName(), E_GRAV_UNDO_FALL, us, sizeof(S_grav_undo_struct) + len1 + len2 + 1);
               ed->FindPlugin("Undo")->Action(E_UNDO_SAVE, &eu);
               delete (byte*)us;
            }

                              //set to original position
            frm->SetPos(us->pos);
            if(*link_to_name){
               PI3D_frame target = ed->GetScene()->FindFrame(link_to_name);
               if(target){
                  frm->SetRot(us->rot);
                  frm->SetScale(us->scale);
                  frm->LinkTo(target);
               }
            }
            frm->Update();
                              //flash the frame
            ed->FindPlugin("Selection")->Action(E_SELECTION_FLASH_FRAME, frm);

            ed->SetModified();
         }else
            ed->Message("Gravitation(E_GRAV_UNDO_FALL): can't find frame.", 0, EM_ERROR);
      }
      break;

   case E_GRAV_TOGGLE_DIALOG:
      ToggleDialog();   
      break;
   }
   return 0;
}

//----------------------------

void CreateGravitation(LPC_editor editor){
   C_edit_Gravitation *gr = new C_edit_Gravitation; editor->InstallPlugin(gr); gr->Release();
}

//----------------------------