#include "all.h"
#include "common.h"

//----------------------------

class C_edit_Grid: public C_editor_item{
   virtual const char *GetName() const{ return "Grid"; }

   bool enabled;
   enum E_line_type{
      LINE_MAIN,              //zero-based lines
      LINE_MAJOR,             //every n-th line
      LINE_MINOR              //rest of lines
   };
   byte draw_on_axis;         //first 3 bits used to enable each axe

   LPC_table config_tab;
   enum{
      TAB_F_CFG_SPACING,
      TAB_V_CFG_COLOR_0,
      TAB_V_CFG_COLOR_1,
      TAB_V_CFG_COLOR_2,
      TAB_F_CFG_ALPHA_0,
      TAB_F_CFG_ALPHA_1,
      TAB_F_CFG_ALPHA_2,
      TAB_F_CFG_CENTER_X,
      TAB_F_CFG_CENTER_Y,
      TAB_F_CFG_CENTER_Z,
   };

   static void cbTabCfg(LPC_table tab, dword msg, dword cb_user, dword prm2, dword prm3){
      switch(msg){
      case TCM_CLOSE:
         {
            C_edit_Grid *ei = (C_edit_Grid*)cb_user;
            ei->ed->GetIGraph()->DeleteDlgHWND(ei->hwnd_edit);
            ei->hwnd_edit = NULL;
         }
         break;
      case TCM_MODIFY:
         switch(prm2){
         case TAB_V_CFG_COLOR_0: case TAB_V_CFG_COLOR_1: case TAB_V_CFG_COLOR_2:
            {
               C_edit_Grid *ei = (C_edit_Grid*)cb_user;
               ei->ed->GetScene()->Render(0);
               ei->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
            }
            break;
         }
         break;
      }
   }
   static C_table_template *GetTemplate(){
                           //template
      static C_table_template templ_config={
         // {B1DE98E0-4090-11d3-B34C-0000B49B8E8B}
         { 0xb1de98e0, 0x4090, 0x11d3, { 0xb3, 0x4c, 0x0, 0x0, 0xb4, 0x9b, 0x8e, 0x8b } },

         "Grid configuration",
         NULL,
         {
            {TE_FLOAT, TAB_F_CFG_SPACING, "Spacing (m)", 0, 0, 1},
            {TE_BRANCH, 0, "Colors", 3},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_0, "Zero axes"},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_1, "Major lines"},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_2, "Other lines"},
            {TE_BRANCH, 0, "Alpha", 3},
               {TE_FLOAT, TAB_F_CFG_ALPHA_0, "Zero axes", 0, 1},
               {TE_FLOAT, TAB_F_CFG_ALPHA_1, "Major lines", 0, 1},
               {TE_FLOAT, TAB_F_CFG_ALPHA_2, "Other lines", 0, 1},
            {TE_BRANCH, 0, "Center", 3},
               {TE_FLOAT, TAB_F_CFG_CENTER_X, "X"},
               {TE_FLOAT, TAB_F_CFG_CENTER_Y, "Y"},
               {TE_FLOAT, TAB_F_CFG_CENTER_Z, "Z"},
            {TE_NULL}
         }
      };
      return &templ_config;
   }

   void UpdateParams(){
      int i;
      if(enabled) ed->RegForRender(this, 1120);
      else ed->UnregForRender(this);
      ed->CheckMenu(this, E_GRID_TOGGLE, enabled);
      for(i=0; i<3; i++) 
         ed->CheckMenu(this, E_GRID_TOGGLE_X+i, 
         (draw_on_axis&(1<<i))!=0);
   }

   HWND hwnd_edit;

   void DestroyEdit(){
      if(hwnd_edit){
         /*
         LPC_editor_item e_prop = ed->FindPlugin("Properties");
         if(e_prop) e_prop->Action(E_PROP_DEL_SHEET, hwnd_edit);
         else
         */
         {
            ed->GetIGraph()->DeleteDlgHWND(hwnd_edit);
            DestroyWindow(hwnd_edit);
         }
         hwnd_edit = NULL;
      }
   }
public:
   C_edit_Grid():
      enabled(true),
      config_tab(NULL),
      hwnd_edit(NULL),
      draw_on_axis(2)
   {
      config_tab = CreateTable();
      config_tab->Open((dword)GetTemplate(), TABOPEN_TEMPLATE);
                              //setup defaults
      config_tab->ItemV(TAB_V_CFG_COLOR_0) = S_vector(1, 1, 1);
      config_tab->ItemV(TAB_V_CFG_COLOR_1) = S_vector(1, 1, 1);
      config_tab->ItemV(TAB_V_CFG_COLOR_2) = S_vector(1, 1, 1);
      config_tab->ItemF(TAB_F_CFG_ALPHA_0) = 1.0f;
      config_tab->ItemF(TAB_F_CFG_ALPHA_1) = .5f;
      config_tab->ItemF(TAB_F_CFG_ALPHA_2) = .25f;
   }
   ~C_edit_Grid(){
      if(config_tab) config_tab->Release();
      DestroyEdit();
   }

   virtual bool Init(){
      ed->AddShortcut(this, E_GRID_TOGGLE, "%30 &View\\&Grid\\%a Enable\tG", K_G, 0);
      ed->AddShortcut(this, E_GRID_TOGGLE_X, "&View\\&Grid\\X\tCtrl+Alt+X", K_X, SKEY_CTRL|SKEY_ALT);
      ed->AddShortcut(this, E_GRID_TOGGLE_Y, "&View\\&Grid\\Y\tCtrl+Alt+Y", K_Y, SKEY_CTRL|SKEY_ALT);
      ed->AddShortcut(this, E_GRID_TOGGLE_Z, "&View\\&Grid\\Z\tCtrl+Alt+Z", K_Z, SKEY_CTRL|SKEY_ALT);
      ed->AddShortcut(this, E_GRID_CONFIG, "&View\\&Grid\\%i Configure", 0, 0);
      UpdateParams();

      return true;
   }
   virtual void Close(){
      if(config_tab){
         config_tab->Release();
         config_tab = NULL;
      }
   }

   virtual dword Action(int id, void*){

      switch(id){
      case E_GRID_TOGGLE:
         enabled = !enabled;
         UpdateParams();
         //*(byte*)NULL = 0;    //crash test
         ed->Message(C_fstring("Grid %s", enabled ? "on" : "off"));
         break;

      case E_GRID_TOGGLE_X:
      case E_GRID_TOGGLE_Y:           
      case E_GRID_TOGGLE_Z:           
         {
            byte mask = (1<<(id-E_GRID_TOGGLE_X));
            ed->CheckMenu(this, id, (draw_on_axis&mask)!=0);
            draw_on_axis ^= mask;
            ed->CheckMenu(this, id, (draw_on_axis&mask)!=0);
         }
         break;

      case E_GRID_CONFIG:
         {
            if(hwnd_edit) DestroyEdit();
            hwnd_edit = (HWND)config_tab->Edit(GetTemplate(), ed->GetIGraph()->GetHWND(), 
               cbTabCfg, (dword)this,
               TABEDIT_MODELESS | TABEDIT_IMPORT | TABEDIT_EXPORT,// | TABEDIT_HIDDEN,
               NULL, NULL);
            if(hwnd_edit){
               /*
               LPC_editor_item e_prop = ed->FindPlugin("Properties");
               if(e_prop) e_prop->Action(E_PROP_ADD_SHEET, hwnd_edit);
               else 
               */
                  ed->GetIGraph()->AddDlgHWND(hwnd_edit);
            }
         }
         break;
      }
      return 0;
   }

//----------------------------

   virtual void Render(){
      const int major_spacing = 10;
                              //setup rendering matrix
      ed->GetScene()->SetRenderMatrix(GetIdentityMatrix());
      PI3D_camera cam = ed->GetScene()->GetActiveCamera();
      if(!cam) return;
      const S_vector &cdir = cam->GetWorldDir();
      const S_vector &cpos = cam->GetWorldPos();
      float cam_near, cam_far;
      cam->GetRange(cam_near, cam_far);

      int i, j;
                              //get 4 corner points
      S_vector corner_pos[4];
      S_vector corner_dir[4];
      dword l, t, r, b;
      ed->GetScene()->GetViewport(l, t, r, b);
      for(i=0; i<4; i++){
         ed->GetScene()->UnmapScreenPoint(
            !(i&1) ? l : r-1,
            !(i&2) ? t : b-1,
            corner_pos[i], corner_dir[i]);
      }

                              //3 lists for all line types
      vector<S_vector> v_list[3];
      vector<word> index_list[3];
      bool ortho = cam->GetOrthogonal();

      int axis;               //loop through all 3 axes
      for(axis = 0; axis<3; axis++){
         if(!(draw_on_axis&(1<<axis))) continue;
         float center = config_tab->ItemF(TAB_F_CFG_CENTER_X+axis);
         float f = cdir[axis];
         if(fabs(f)<MRG_ZERO) continue;
         float dist_to_plane = (-cpos[axis] + center) / f;
         const float grid_border = 1000.0f;
         if(dist_to_plane<0.0f) dist_to_plane = grid_border;
         else dist_to_plane = Min(dist_to_plane, grid_border);

         //int limit_coef[2][2];//min/max coeficients for both axes [axe][min|max]
         float limit_coef[2][2];//min/max coeficients for both axes [axe][min|max]
         int axe[2];
         const float max_dist = 1000.0f;
         const int max_coef = grid_border;

         float max_corner_dist = 0.0f;
         bool corner_on_grid = false;

         for(int rest_axis=0; rest_axis<2; rest_axis++){
            axe[rest_axis] = (axis+1+rest_axis)%3;
                              //by default make it wide
            limit_coef[rest_axis][0] = max_coef;
            limit_coef[rest_axis][1] = -max_coef;
                              //check all 4 corners
            for(j=0; j<4; j++)
            {
               const S_vector &pos = corner_pos[j];
               const S_vector &dir = corner_dir[j];
               float f = dir[axis];
               if(fabs(f)<MRG_ZERO) continue;
               //float dist_to_plane = (-cpos[axis] + center) / f;
               float dist_to_plane = -pos[axis] / f;
               //int curr_coef;
               float curr_coef;
               if(ortho || dist_to_plane>=0.0f)
               {
                              //limit dist to camera range
                  S_vector pt_on_plane = pos + dir * dist_to_plane;
                  curr_coef = Min(pt_on_plane[axe[rest_axis]], cam_far);// / curr_spacing;
                              //clip coef
                  //curr_coef = Max(-max_coef, Min(max_coef, curr_coef));
                  curr_coef = Max(-(float)max_coef, Min((float)max_coef, curr_coef));
                  //max_corner_dist = Max(max_corner_dist, dist_to_plane);
                  //max_corner_dist += dist_to_plane;
                  max_corner_dist += Min(dist_to_plane, cam_far);
                  corner_on_grid = true;
               }else{
                  curr_coef = (dir[axe[rest_axis]] < 0.0f) ? -max_coef : max_coef;
                  //max_corner_dist = max_coef;
                  //max_corner_dist += max_coef;
                  max_corner_dist += Min(grid_border, cam_far);
               }
               //DEBUG(C_fstring("dist_to_plane = %.2f", dist_to_plane));
               //DEBUG(C_fstring("curr_coef = %i", curr_coef));
               limit_coef[rest_axis][0] = Min(limit_coef[rest_axis][0], curr_coef);
               limit_coef[rest_axis][1] = Max(limit_coef[rest_axis][1], curr_coef);
            }
         }
                              //grid not visible?
         if(!corner_on_grid) continue;
         max_corner_dist *= .125f;

         float curr_spacing = Max(.05f, config_tab->ItemF(TAB_F_CFG_SPACING));
         if(!ortho){
                              //reduce spacing if too many lines are to be visible
            if(max_corner_dist<0.0f) max_corner_dist = grid_border;
            else max_corner_dist = Min(max_corner_dist, grid_border);

            //float dist_to_plane1 = fabs(cpos[axis]);
            while(max_corner_dist>curr_spacing*60.0f) curr_spacing *= major_spacing;
         }else{
            float scl = cam->GetOrthoScale();
            float num_x = 1.0f / (curr_spacing*scl);
            while(num_x > 40){
               curr_spacing *= major_spacing;
               num_x /= major_spacing;
            }
         }


         for(i=0; i<2; i++)
         for(j=0; j<2; j++)
            limit_coef[i][j] /= curr_spacing;

                              //create all lines on one secondary axe
         for(int axis2 = 0; axis2<2; axis2++){
            for(i=limit_coef[axis2][0]; i<limit_coef[axis2][1]+1; i++)
            {
               E_line_type ltype = !i ? LINE_MAIN :
                  !(i%major_spacing) ? LINE_MAJOR :
                  LINE_MINOR;
               for(j=0; j<2; j++){    //each line has 2 ends
                  index_list[ltype].push_back(v_list[ltype].size());
                  v_list[ltype].push_back(S_vector());
                  S_vector &v = v_list[ltype].back();
                  //v[axis] = 0.0f;
                  v[axis] = center;
                  v[axe[axis2]] = i * curr_spacing;
                  //v[axe[1-axis2]] = !j ? -max_dist : max_dist;
                  float f = cpos[axe[1-axis2]];
                  f += !j ? -cam_far : cam_far;
                  f = Max(-max_dist, Min(max_dist, f));
                  v[axe[1-axis2]] = f;
               }
            }
         }
      }
                             //draw all line types
      for(i=0; i<3; i++){
         //DEBUG(C_fstring("Lines type %i: %i", i, index_list[i].size()/2));
         byte alpha = 255 - 255.0f * config_tab->ItemF(TAB_F_CFG_ALPHA_0+i);
         ed->GetScene()->DrawLines(&v_list[i][0], v_list[i].size(),
            &index_list[i][0], index_list[i].size(),
            *(S_vector*)config_tab->Item(TAB_V_CFG_COLOR_0+i),
            sizeof(S_vector), 
            alpha);
      }
   }

   virtual bool LoadState(C_chunk *ck){
                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version!=0) return false;
                              //read table
      config_tab->Open((dword)ck->GetHandle(), TABOPEN_FILEHANDLE | TABOPEN_UPDATE);
                              //write other variables
      byte en1;
      ck->Read(&en1, sizeof(byte));
      enabled = en1;
      ck->Read(&draw_on_axis, sizeof(byte));

      UpdateParams();
      return true;
   }

   virtual bool SaveState(C_chunk *ck) const{
                              //write version
      byte version = 0;
      ck->Write(&version, sizeof(byte));
                              //write table
      config_tab->Save((dword)ck->GetHandle(), TABOPEN_FILEHANDLE);
                              //write other variables
      ck->Write(&enabled, sizeof(byte));
      ck->Write(&draw_on_axis, sizeof(byte));
      return true;
   }
};

//----------------------------

void CreateGrid(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Grid;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
