#include "all.h"
#include "common.h"

//----------------------------

class C_edit_Hide: public C_editor_item{
   virtual const char *GetName() const{ return "Hide"; }
   LPC_editor_item e_undo, e_slct, e_modify;
public:
   C_edit_Hide():
      e_undo(NULL),
      e_slct(NULL),
      e_modify(NULL)
   {}
   virtual bool Init(){
      e_undo = ed->FindPlugin("Undo");
      e_slct = ed->FindPlugin("Selection");
      e_modify = ed->FindPlugin("Modify");
      if(!e_undo || !e_slct || !e_modify){
         e_undo = NULL;
         e_slct = NULL;
         e_modify = NULL;
      }
      e_undo->AddRef();
      e_slct->AddRef();
      e_modify->AddRef();

      ed->AddShortcut(this, E_HIDE_SELECTION, "%10 &Edit\\%60 %i &Hide\tH", K_H, 0);
      ed->AddShortcut(this, E_HIDE_UNHIDE_PROMPT, "&Edit\\%60 &Unhide by name\tCtrl+H", K_H, SKEY_CTRL);
      ed->AddShortcut(this, E_HIDE_UNHIDE_ALL, "&Edit\\%60 Unh&ide all\tCtrl+Alt+H", K_H, SKEY_CTRL | SKEY_ALT);

      {                       //receive notifications from selection
         pair<LPC_editor_item, dword> p(this, E_HIDE_NOTIFY_SELECTION);
         e_slct->Action(E_SELECTION_ADD_NOTIFY, &p);
      }
      return true;
   }

   virtual void Close(){
      if(e_slct)
         e_slct->Action(E_SELECTION_DEL_NOTIFY, this);

      if(e_undo) e_undo->Release();
      if(e_slct) e_slct->Release();
      if(e_modify) e_modify->Release();
      e_undo = NULL;
      e_slct = NULL;
      e_modify = NULL;
   }

   virtual dword C_edit_Hide::Action(int id, void *context){

      switch(id){
      case E_HIDE_SELECTION:
         {
            if(!ed->CanModify()) break;
            bool on = false;//(id==0);

            vector<PI3D_frame> sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            pair<PI3D_frame, dword>mp(NULL, E_MODIFY_FLG_HIDE);

            if(!sel_list.size()) break;

            dword num = 0;
            for(int i=0; i<sel_list.size(); i++){
               PI3D_frame frm = sel_list[i];
               switch(frm->GetType()){
               case FRAME_VISUAL:
               case FRAME_DUMMY:
               case FRAME_MODEL:
                  if(on!=(bool)frm->IsOn()){
                     if(!on){
                        mp.first = frm;
                        e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);
                        e_slct->Action(E_SELECTION_DEL_FRAME, frm);
                              //save undo info
                        const char *cp = frm->GetName();
                        S_edit_Undo_2 eu(GetName(), E_HIDE_UNHIDE_NAMED, (void*)cp, strlen(cp)+1);
                        e_undo->Action(E_UNDO_SAVE, &eu);
                     }
                     frm->SetOn(on);
                     ++num;
                     ed->SetModified();
                  }
                  break;
               }
            }
            ed->Message(C_fstring("%s: %i object(s)", (on ? "Unhidden" : "Hidden"), num));
         }
         break;

      case E_HIDE_HIDE_NAMED:
      case E_HIDE_UNHIDE_NAMED:
         {
            const char *cp = (const char*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(cp, ENUMF_BACKDROP | ENUMF_ALL);
            if(frm){
               bool on = (id==E_HIDE_UNHIDE_NAMED);
               frm->SetOn(on);
                              //update flags
               pair<PI3D_frame, dword> mp(frm, E_MODIFY_FLG_HIDE);
               e_modify->Action(on ? E_MODIFY_DEL_FLAGS : E_MODIFY_ADD_FLAGS, &mp);
                              //save undo
               S_edit_Undo_2 eu(GetName(), on ? E_HIDE_HIDE_NAMED : E_HIDE_UNHIDE_NAMED, (void*)cp, strlen(cp)+1);
               e_undo->Action(E_UNDO_SAVE, &eu);

               e_slct->Action(E_SELECTION_FLASH_FRAME, frm);
            }else
               ed->Message(C_fstring("Hide(E_HIDE_HIDE_NAMED): can't find frame '%s'", cp));
         }
         break;

      case E_HIDE_UNHIDE_PROMPT:
         {
                              //collect list of hidden frames
            vector<PI3D_frame> h_list;
            int i;
            
            struct S_hlp{
               static void cbEnum(PI3D_frame frm, dword flags, dword c){
                  if(flags&E_MODIFY_FLG_HIDE){
                     ((vector<PI3D_frame>*)c)->push_back(frm);
                  }
               }
            };
            e_modify->Action(E_MODIFY_ENUM, &pair<t_modify_enum*, dword>(S_hlp::cbEnum, (dword)&h_list));
            if(!h_list.size())
               break;
                              //perform for selection
            i = e_slct->Action(E_SELECTION_PROMPT, &h_list);
            if(!i || !h_list.size())
               break;

            for(i=h_list.size(); i--; ){
               PI3D_frame frm = h_list[i];
                              //unhide this
               frm->SetOn(true);                
               pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_HIDE);
               e_modify->Action(E_MODIFY_DEL_FLAGS, &p);

               e_slct->Action(E_SELECTION_ADD_FRAME, frm);

                              //save undo info
               const char *cp = p.first->GetName();
               S_edit_Undo_2 eu(GetName(), E_HIDE_HIDE_NAMED, (void*)cp, strlen(cp)+1);
               e_undo->Action(E_UNDO_SAVE, &eu);
            }
            ed->Message(C_fstring("UnHidden %i object(s)", h_list.size()));
         }
         break;

      case E_HIDE_UNHIDE_ALL:
         {
            e_slct->Action(E_SELECTION_CLEAR);

            dword num_unhidden = 0;
            vector<PI3D_frame> unhide_list;
            int i;

            struct S_hlp{
               static void cbEnum(PI3D_frame frm, dword flags, dword c){
                  if(flags&E_MODIFY_FLG_HIDE){
                     ((vector<PI3D_frame>*)c)->push_back(frm);
                  }
               }
            };
            e_modify->Action(E_MODIFY_ENUM, &pair<t_modify_enum*, dword>(S_hlp::cbEnum, (dword)&unhide_list));

            for(i=unhide_list.size(); i--; ){
               PI3D_frame frm = unhide_list[i];
               frm->SetOn(true);                
               pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_HIDE);
               e_modify->Action(E_MODIFY_DEL_FLAGS, &p);

               e_slct->Action(E_SELECTION_ADD_FRAME, frm);

                              //save undo info
               const char *cp = frm->GetName();
               S_edit_Undo_2 eu(GetName(), E_HIDE_HIDE_NAMED, (void*)cp, strlen(cp)+1);
               e_undo->Action(E_UNDO_SAVE, &eu);

               ++num_unhidden;
            }
            if(num_unhidden)
               ed->Message(C_fstring("Unhidden %i object(s)", num_unhidden));
         }
         break;

      case E_HIDE_NOTIFY_SELECTION:
         {
            vector<PI3D_frame> *sel_list = (vector<PI3D_frame>*)context;
            ed->EnableMenu(this, E_HIDE_SELECTION, (sel_list->size()!=0));
         }
         break;
      }
      return 0;
   }
};

//----------------------------

void CreateHide(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Hide;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
