#ifndef __I3D_FONT_H
#define __I3D_FONT_H

#include <vector>
#include <c_unknwn.hpp>
#include <i3d\i3d2.h>

#ifdef _MSC_VER
using namespace std;
#endif

//----------------------------

                              //polygone sprite system
class C_poly_sprite: public C_unknown{
   bool on;
   friend class C_poly_group;
   int z;
   float fx, fy;              //relative coors
   float sx, sy;              //screen coors
   S_vector2 pt[4];
   struct{
      int x, y;
      float tu, tv;
   } tlv[4];
   struct S_vertex{           //param to DrawPrimitive
      float sx, sy, sz;
      float rhw;
      //dword diffuse;
      I3D_text_coor tx;
   } tlv1[4];
   bool tlv_reset;
   void PrepareDraw();
   dword color;
   friend class C_poly_text;
public:
   C_poly_sprite(int z1):
      on(true), color(-1),
      z(z1)
   {}
   void SetOn(bool b){ on=b; }
   //void SetScreenPos(float x1, float y1, int scrn_sx, int scrn_sy);
   void SetScreenPos(float x1, float y1);
   void SetScreenSize(float x, float y, int scrn_sx, int scrn_sy);
   void SetTileX(float start, float end);
   void SetTileY(float n);
   void SetPos(float x1, float y1, int scrn_sx, int scrn_sy);
   void GetPos(float &x1, float &y1);
   //void SetZ(int i){ z=i; }
   int GetZ() const{ return z; }
   S_vector2 GetPos() const{ return S_vector2(fx, fy); }
   void MakeRect(float angle=0);
   void MakeItalic(float angle=0);
   void MakeItalic(float angle=0, float base_y=0);
   void SetColor(dword c){ color=c; }
   dword GetColor() const{ return color; }
   void GetScreenRect(int*) const;
   void GetScreenPos(float *x, float *y) const{ *x = sx, *y = sy; }
};

typedef C_poly_sprite *LPC_poly_sprite;

//----------------------------

class C_poly_group: public C_unknown{
   vector<LPC_poly_sprite> sprs;
   bool resort;
   virtual ~C_poly_group(){
      for(int i=0; i<sprs.size(); i++) sprs[i]->Release();
   }
   friend class C_poly_text;
public:
   C_poly_group();

   LPC_poly_sprite AddSprite(float x, float y, float sy,
      class C_sprite_image *img, dword indx, int scrn_sx, int scrn_sy, int z = 0);
   void DeleteSprite(LPC_poly_sprite);

   void Draw(PI3D_scene);

   void SetColor(dword c);
   dword GetColor() const;
   void SetOn(bool b);
   LPC_poly_sprite GetSpr(int i){
      //Assert(i<sprs.size());
      return sprs[i];
   }
   dword NumSprs() const{ return sprs.size(); }
};

//----------------------------

class C_sprite_image{

   PI3D_texture tp;
   struct S_spr_rect{
      int ix, iy, isx, isy, icx, icy;
      float l, t, sx, sy, cx, cy;
      word wsx, wsy;
      S_spr_rect(){}
      S_spr_rect(float l1, float t1, float sx1, float sy1,
         float cx1, float cy1, word wsx1, word wsy1):
         l(l1),
         t(t1),
         sx(sx1),
         sy(sy1),
         cx(cx1),
         cy(cy1),
         wsx(wsx1),
         wsy(wsy1)
      {}
   };
   vector<S_spr_rect> spr_rects;
   friend class C_poly_text;
   friend class C_poly_group;
public:
   C_sprite_image():
      tp(NULL)
   {}
   bool Init(PI3D_driver driver, const char *prjname, const char *diff_name = NULL,
      const char *opt_name = NULL, dword flags = 0, dword img_sx = 256, dword img_sy = 256,
      void *hInstance = NULL);
   void Close();
   void Setup(PI3D_driver driver);
};

//----------------------------

class C_poly_text{
   PI3D_driver driver;
                              //font:

   struct S_font_size{
      byte l, r;
   };
   vector<S_font_size> font_sizes;
   vector<int> font_size_offsets;

   word LetterFit(byte l, byte r);
   void BuildSpaceTable(const char *text, short *tab, int tablen);

   C_sprite_image img;
   vector<C_poly_group*> texts;

   void AddText(C_poly_group *pg){
      texts.push_back(pg);
      //pg->AddRef();
   }
public:
   ~C_poly_text(){
      Close();
   }
   bool Init(PI3D_driver driver, const char *prjname, const char *bitmap_name = NULL,
      const char *opt_name = NULL,
      dword flags=0, void *hInstance = NULL);
   bool IsInit() const{ return (font_sizes.size()!=0); }
   void Close(){
      img.Close();
      for(int i=0; i<texts.size(); i++) texts[i]->Release();
      texts.clear();
   }
   C_poly_group *CreateText(const char *cp, float x, float y, float size, dword justify=0,
      float *xret=NULL, float *yret=NULL, float fmt_rect[4] = NULL, int *num_lines=NULL);
   void SetPos(C_poly_group*, float x, float y);
   void GetPos(C_poly_group*, float &x, float &y);
   void SetScreenPos(C_poly_group*, float x, float y, int scrn_sx, int scrn_sy);
   void SetItalic(C_poly_group*, float angle);
   void DeleteText(C_poly_group *pg){
      for(int i=0; i<texts.size(); i++)
      if(texts[i]==pg){
         texts[i]=texts.back();
         texts.pop_back();
         pg->Release();
         return;
      }
   }
   void Draw(PI3D_scene scene){
      img.Setup(driver);
      for(int i=0; i<texts.size(); i++) texts[i]->Draw(scene);
   }
};

//----------------------------

#endif