#include "all.h"

//----------------------------

void C_poly_sprite::PrepareDraw(){

   for(int j=0; j<4; j++){
      tlv1[j].sx = (sx + tlv[j].x);
      tlv1[j].sy = sy + tlv[j].y;
      tlv1[j].sz = 0.0f;
      tlv1[j].rhw = 1.0f;
      //tlv1[j].diffuse = color;
      tlv1[j].tx.u = tlv[j].tu;
      tlv1[j].tx.v = tlv[j].tv;
   }
   tlv_reset = false;
}

//----------------------------

void C_poly_sprite::MakeRect(float angle){

   float s = sinf(angle), c = cosf(angle);
   for(int i=0; i<4; i++){
      tlv[i].x = c*pt[i].x - s*pt[i].y;
      tlv[i].y = s*pt[i].x + c*pt[i].y;
   }
   tlv_reset = true;
}

//----------------------------

void C_poly_sprite::MakeItalic(float angle){

   float sy = pt[2].y - pt[0].y;
   float dx = sy * tanf(angle);
   tlv[0].x = pt[0].x + dx;
   tlv[1].x = pt[3].x + dx;
   tlv_reset = true;
}

//----------------------------

void C_poly_sprite::SetScreenPos(float x1, float y1){

   sx = x1;
   sy = y1;
}

//----------------------------

void C_poly_sprite::MakeItalic(float angle, float base_y){

   float dy = base_y - sy;
   {
      float sy1 = -pt[0].y;
      sy1 += dy;
      float dx = sy1 * tanf(angle);
      tlv[0].x += dx;
      tlv[1].x += dx;
   }
   {
      float sy1 = -pt[2].y;
      sy1 += dy;
      float dx = sy1 * tanf(angle);
      tlv[2].x += dx;
      tlv[3].x += dx;
   }
   tlv_reset = true;
}

//----------------------------

void C_poly_sprite::GetScreenRect(int *rc) const{
   
   rc[0] = sx + Min(tlv[0].x, tlv[2].x);
   rc[1] = sy + Min(tlv[0].y, tlv[1].y);
   rc[2] = sx + Max(tlv[1].x, tlv[3].x);
   rc[3] = sy + Max(tlv[2].y, tlv[3].y);
}

//----------------------------

void C_poly_sprite::SetScreenSize(float sx, float sy, int scrn_sx, int scrn_sy){

   float sizex = (pt[1].x - pt[0].x) / scrn_sy;
   float rx = sx / sizex;
   float sizey = (pt[2].y - pt[0].y) / scrn_sy;
   float ry = sy / sizey;
   for(int i=0; i<4; i++){
      pt[i].x *= rx;
      pt[i].y *= ry;
   }
   tlv_reset = true;
}

//----------------------------

void C_poly_sprite::SetTileX(float s, float e){
   
   tlv[1].tu = (tlv[0].tu = s) + e;
   tlv[3].tu = (tlv[2].tu = s) + e;
   tlv_reset = true;
}

//----------------------------

void C_poly_sprite::SetTileY(float n){
   
   tlv[2].tv = tlv[0].tv + n;
   tlv[3].tv = tlv[1].tv + n;
   tlv_reset = true;
}

//----------------------------

void C_poly_sprite::SetPos(float x1, float y1, int scrn_sx, int scrn_sy){

   fx = x1;
   fy = y1;

   //SetScreenPos((int)(scrn_sy*fx), (int)(scrn_sy*fy));
   SetScreenPos((int)(scrn_sx * .75f * fx), (int)(scrn_sx * .75f * fy));
   tlv_reset = true;
}

//----------------------------

void C_poly_sprite::GetPos(float &x1, float &y1){

   x1 = fx;
   y1 = fy;
}

//----------------------------
//----------------------------

bool C_sprite_image::Init(PI3D_driver driver, const char *prjname, const char *diff_name,
   const char *opt_name, dword flags, dword img_sx, dword img_sy, void *hInstance){

   int i;

   C_cache cache;
   if(!hInstance){
      if(!cache.open(prjname, CACHE_READ))
         return false;
   }else{
      HRSRC hrsrc = FindResource((HINSTANCE)hInstance, "gamefont.prj", "BINARY");
      HGLOBAL hg = LoadResource((HINSTANCE)hInstance, hrsrc);
      void *vp = LockResource(hg);
      size_t len = SizeofResource((HINSTANCE)hInstance, hrsrc);

      if(!cache.open((byte**)&vp, &len, CACHE_READ_MEM))
         return false;
   }

   struct{
      dword id;
      word img_num;
      word spr_num;
      dword res[2];
      char img_name[16];
   }header;
   cache.read((char*)&header, sizeof(header));


   I3D_RESULT ir;
   I3D_CREATETEXTURE ct;
   memset(&ct, 0, sizeof(ct));

   ct.flags = flags | TEXTMAP_DIFFUSE | TEXTMAP_NOMIPMAP;
   if(!diff_name) diff_name = header.img_name;
   if(!hInstance){
      ct.text_name = diff_name;
      ct.opt_name = opt_name;
   }else{
      HGLOBAL hg = LoadResource((HINSTANCE)hInstance, 
         FindResource((HINSTANCE)hInstance, diff_name, RT_BITMAP));
      if(hg){
         ct.text_name = (const char*)hg;
      }

      if(opt_name){
         HGLOBAL hg = LoadResource((HINSTANCE)hInstance, 
            FindResource((HINSTANCE)hInstance, opt_name, RT_BITMAP));
         if(hg) ct.opt_name = (const char*)hg;
      }
      ct.flags |= TEXTMAP_MEM_PTR;
   }
   if(opt_name) ct.flags |= TEXTMAP_OPACITY;
   bool low_det = driver->GetState(RS_TEXTURELOWDETAIL);
   if(low_det) driver->SetState(RS_TEXTURELOWDETAIL, false);
   ir = driver->CreateTexture(&ct, &tp);
   if(low_det) driver->SetState(RS_TEXTURELOWDETAIL, true);
   if(I3D_FAIL(ir)){
      Close();
      return false;
   }

   spr_rects.reserve(header.spr_num);
   cache.seekp(header.img_num*16+16);
   struct{
      word x, y, sx, sy;
      short cx, cy;
      dword res;
   } rc;
   for(i=0; i<header.spr_num; i++){
      cache.read((char*)&rc, sizeof(rc));
      spr_rects.push_back(S_spr_rect(
         (float(rc.x)-.5f) / float(img_sx),
         (float(rc.y)-.5f) / float(img_sy),
         (float(rc.sx)+1.0f) / float(img_sx),
         (float(rc.sy)+1.0f) / float(img_sy),
         (float)rc.cx / (float)rc.sx,
         (float)rc.cy / (float)rc.sy,
         rc.sx,
         rc.sy));

      spr_rects.back().ix = rc.x;
      spr_rects.back().iy = rc.y;
      spr_rects.back().isx = rc.sx;
      spr_rects.back().isy = rc.sy;
      spr_rects.back().icx = rc.cx;
      spr_rects.back().icy = rc.cy;
   }
   cache.close();

   return true;
}

//----------------------------

bool C_poly_text::Init(PI3D_driver drv, const char *prjname, 
   const char *bitmap_name, const char *opt_name, dword flags,
   void *hInstance){

   driver = drv;
   bool b = img.Init(driver, prjname, bitmap_name, opt_name, flags, 256, 256, hInstance);

   if(b){
      C_cache cache;
      if(!hInstance){
         cache.open(prjname, CACHE_READ);
      }else{
         HRSRC hrsrc = FindResource((HINSTANCE)hInstance, "gamefont.prj", "BINARY");
         HGLOBAL hg = LoadResource((HINSTANCE)hInstance, hrsrc);
         void *vp = LockResource(hg);
         size_t len = SizeofResource((HINSTANCE)hInstance, hrsrc);
         cache.open((byte**)&vp, &len, CACHE_READ_MEM);
      }

      struct{
         dword id;
         word img_num;
         word spr_num;
         dword res[2];
         char img_name[16];
      } header;
      cache.read((char*)&header, sizeof(header));
      cache.close();

      LPIImage font_img;
      driver->GetGraphInterface()->CreateImage(&font_img);
      S_pixelformat pf;
      memset(&pf, 0, sizeof(pf));
      pf.bytes_per_pixel = 1;
      pf.flags = PIXELFORMAT_PALETTE;
      if(!bitmap_name) bitmap_name = header.img_name;
      if(!hInstance){
         if(!font_img->Open(C_fstring("editor\\%s", bitmap_name),
            IMGOPEN_SYSMEM,
            0, 0, &pf)){

            img.Close();
            font_img->Release();
            return false;
         }
      }else{
         HGLOBAL hg = LoadResource((HINSTANCE)hInstance, 
            FindResource((HINSTANCE)hInstance, bitmap_name, RT_BITMAP));
         if(!hg) return false;
         /*
         {
            BITMAPINFO *bi = (BITMAPINFO*)hg;
            bi->bmiHeader.biXPelsPerMeter = 0;
         }
         */
         font_img->Open((const char*)hg, IMGOPEN_SYSMEM | IMGOPEN_MEM_PTR,
            0, 0, &pf);
      }
                              //build font spacing table
                              //get access to bitmap data
      byte *font_data;
      dword pitch;
      byte bpp;
      font_img->Lock((void**)&font_data, &pitch, &bpp);

      font_size_offsets.clear();
      font_size_offsets.reserve(header.spr_num);
      font_sizes.clear();
                              //build all letters
      for(int i = 0; i<header.spr_num; i++){
                              //get address of letter
         byte *mem = font_data + img.spr_rects[i].iy * pitch + img.spr_rects[i].ix;
                              //assign pointer
         font_size_offsets.push_back(font_sizes.size());

         dword sx = img.spr_rects[i].isx, sy = img.spr_rects[i].isy;
                              //process all letter's lines
         for(int j=0; j<sy; j++, mem += pitch){
            font_sizes.push_back(S_font_size());
            S_font_size &fs = font_sizes.back();

                              //find left
            fs.l = 0;
            byte *bm = mem;
            while(!*bm && fs.l<sx) ++bm, ++fs.l;
                              //find right
            fs.r = sx;
            bm = mem + sx;
            while(!*--bm && fs.r) --fs.r;
         }
      }

      font_img->Unlock();
      font_img->Release();
   }
   return b;
}

//----------------------------

void C_sprite_image::Close(){

   if(tp){
      tp->Release();
      tp = NULL;
   }
   spr_rects.clear();
}

//----------------------------

void C_sprite_image::Setup(PI3D_driver driver){

   driver->SetTexture(tp);
}

//----------------------------
//----------------------------

C_poly_group::C_poly_group():
   resort(false)
{}

//----------------------------
static bool spr_less(C_poly_sprite *s1, C_poly_sprite *s2){ return s1->GetZ()<s2->GetZ(); }

C_poly_sprite *C_poly_group::AddSprite(float x, float y, float sy,
   C_sprite_image *img, dword indx, int scrn_sx, int scrn_sy, int z){

   C_poly_sprite *spr = new C_poly_sprite(z);
   sprs.push_back(spr);
   if(z!=0)
      resort = true;

   const C_sprite_image::S_spr_rect &rc=img->spr_rects[indx];
   float sx = rc.sx*sy/rc.sy;

   //float ratio = scrn_sy;
   float ratio = scrn_sx * .75f;

   for(int i=0; i<4; i++){
      spr->pt[i].x = -rc.cx*sx * ratio;
      spr->pt[i].y = -rc.cy*sy * ratio;
      if(i&1) spr->pt[i].x += sx * ratio;
      if(i&2) spr->pt[i].y += sy * ratio;

      spr->tlv[i].tu=rc.l;
      spr->tlv[i].tv=rc.t;
      if(i&1) spr->tlv[i].tu += rc.sx;
      if(i&2) spr->tlv[i].tv += rc.sy;
   }
   spr->SetPos(x, y, scrn_sx, scrn_sy);
   spr->MakeRect();

   return spr;
}

//----------------------------

void C_poly_group::DeleteSprite(C_poly_sprite *sp){

   for(int i=0; i<sprs.size(); i++)
   if(sprs[i]==sp){
      sp->Release();
      sprs[i] = sprs.back();
      sprs.pop_back();
      break;
   }
   resort = true;
}

//----------------------------

void C_poly_group::SetColor(dword c){
   for(int i=0; i<sprs.size(); i++) sprs[i]->SetColor(c);
}

//----------------------------

void C_poly_group::SetOn(bool b){
   
   for(int i=0; i<sprs.size(); i++)
      sprs[i]->SetOn(b);
}

//----------------------------

dword C_poly_group::GetColor() const{
   
   return sprs.size() ? sprs[0]->GetColor() : 0;
}

//----------------------------

void C_poly_group::Draw(PI3D_scene scene){

   int i;
   if(resort){
      sort(sprs.begin(), sprs.end(), spr_less);
      resort=false;
   }

                              //draw back to front
   for(i=0; i<sprs.size(); i++){
      LPC_poly_sprite spr = sprs[i];
      if(!sprs[i]->on) continue;
      if(spr->tlv_reset) spr->PrepareDraw();

      dword c = spr->color;
      S_vector color;
      color[2] = float(c&0xff) / 255.0f;
      color[1] = float((c>>8)&0xff) / 255.0f;
      color[0] = float((c>>16)&0xff) / 255.0f;
      static const word indx[] = {0, 1, 2, 2, 1, 3};
      scene->DrawTriangles((S_vector*)spr->tlv1, 4, indx, sizeof(indx)/sizeof(word), color,
         sizeof(C_poly_sprite::S_vertex),
         I3DDRAW_TRANSFORMED | Max(1, 255-(int)(c>>24)),
         &spr->tlv1[0].tx);
   }
}

//----------------------------
                              //national support
/*--------------------------------------------------------
                              //(unused)
   marks:
   0x80  -  a' (?)   a`'
   0x81  -  cv       c`v
   0x82  -  d'       d`|
   0x83  -  a: (?)   a`:
   0x84  -  e^ (?)   e`^
   0x85  -  ao (?)   a`o
   0x86  -  c, (?)   c`,
   0x87  -  c. (?)   c`.
   0x88  -  a` (?)   a``
   0x89  -  n~ (?)   n`~
--------------------------------------------------------*/
#define DLZEN     0x80
#define MAKCEN    0x81
#define MAKCEN1   0x82
#define DVOJBODKA 0x83
#define BOKAN     0x84
#define KRUZOK    0x85
#define PL1       0x86
#define PL2       0x87
#define DLZEN_L   0x88
#define VLNOVKA   0x89
#define GERMAN_S  's'

static struct S_national_table{
   byte orig;
   char base;
   byte mark;
} char_tab_default[]={
   0x8a, 'S', MAKCEN,
   0x8c, 'S', DLZEN,
   0x8d, 'T', MAKCEN,
   0x8e, 'Z', MAKCEN,

   0x92, '\'', 0,
   0x9a, 's', MAKCEN,
   0x9c, 's', DLZEN,
   0x9d, 't', MAKCEN1,
   0x9e, 'z', MAKCEN,

   0xbc, 'L', MAKCEN,
   0xbe, 'l', MAKCEN1,

   0xc0, 'R', DLZEN,
   0xc1, 'A', DLZEN,
   0xc4, 'A', DVOJBODKA,
   0xc5, 'L', DLZEN,
   0xc6, 'C', DLZEN,
   0xc8, 'C', MAKCEN,
   0xc9, 'E', DLZEN,
   0xcb, 'E', DVOJBODKA,
   0xcc, 'E', MAKCEN,
   0xcd, 'I', DLZEN,
   0xcf, 'D', MAKCEN1,

   0xd1, 'N', DLZEN,
   0xd2, 'N', MAKCEN,
   0xd3, 'O', DLZEN,
   0xd6, 'O', DVOJBODKA,
   0xd8, 'R', MAKCEN,
   0xd9, 'U', KRUZOK,
   0xda, 'U', DLZEN,
   0xdc, 'U', DVOJBODKA,
   0xdd, 'Y', DLZEN,
   0xdf, 's', GERMAN_S,       //german s converted to ss

   //0xe0, 'r', DLZEN,
   0xe0, 'a', DLZEN_L,
   0xe1, 'a', DLZEN,
   0xe2, 'a', BOKAN,
   0xe4, 'a', DVOJBODKA,
   0xe5, 'l', DLZEN,
   0xe6, 'c', DLZEN,
   0xe7, 'c', PL1,
   //0xe8, 'c', MAKCEN,
   0xe8, 'e', DLZEN_L,
   0xe9, 'e', DLZEN,
   0xea, 'e', BOKAN,
   0xeb, 'e', DVOJBODKA,
   //0xec, 'e', MAKCEN,
   0xec, 127, DLZEN_L,
   0xed, 127, DLZEN,
   0xee, 127, BOKAN,
   0xef, 'd', MAKCEN1,

   0xf1, 'n', VLNOVKA,
   //0xf2, 'n', MAKCEN,
   0xf2, 'o', DLZEN_L,
   0xf3, 'o', DLZEN,
   0xf4, 'o', BOKAN,
   0xf6, 'o', DVOJBODKA,
   0xf8, 'r', MAKCEN,
   //0xf9, 'u', KRUZOK,
   0xf9, 'u', DLZEN_L,
   0xfa, 'u', DLZEN,
   0xfb, 'u', BOKAN,
   0xfc, 'u', DVOJBODKA,
   0xfd, 'y', DLZEN,
   0
}, *national_tables[]={
   char_tab_default,          //english
   char_tab_default,          //german
   char_tab_default,          //french
};

//----------------------------

                              //find closest position for two letters
word C_poly_text::LetterFit(byte l, byte r){

   l -= ' ', r -= ' ';
   if(l>102 || r>102) return 0;
   int xdist = 0;

   S_font_size *lfs = &font_sizes[font_size_offsets[l]],
      *rfs = &font_sizes[font_size_offsets[r]];
   int lcy = img.spr_rects[l].icy, rcy = img.spr_rects[r].icy;
   int lsy = img.spr_rects[l].isy, rsy = img.spr_rects[r].isy;
   int lsy1 = lsy;
   S_font_size *lfs1 = lfs;
                              //adjust top
   if(lcy>rcy) lfs -= (rcy-lcy), lsy += (rcy-lcy);
   else rfs -= (lcy-rcy), rsy += (lcy-rcy);
                              //get compare height
   int sy = Min(lsy, rsy);
   if(sy<=0){
                              //no overlapping, use hardspace
      while(lsy1-- > 0){
                              //compare line
         xdist = Max(xdist, (int)lfs1->r);
         ++lfs1;
      }
   }else
   while(sy-- > 0){
                              //compare line
      xdist = Max(xdist, Max(lfs->r-rfs->l, 0));
      ++lfs, ++rfs;
   }
   return xdist;
}

//----------------------------

void C_poly_text::BuildSpaceTable(const char *text, short *tab, int tablen){

   int i;
   dword x = 0;
   char last = ' ';
   bool first = true;
   for(i=0; ; i++){
      byte c = text[i];       //get one character
      if(!c) break;           //end of text
      short *tp = tab+i*2;

      if(c<' '){
         if(text[i]=='\n'){
            tp[0] = x + img.spr_rects[last-' '].sx;
            x = 0;
         }else
            tp[0] = 0;
         tp[1] = 0;
         first = true;
      }else
      if(c>=128 && c<0x90){
         switch(c){
         case 0x80:
         case 0x81:
         case 0x83:
         case 0x84:
         case 0x85:
         case 0x88:
         case 0x89:
                              //put mark onto center of letter
            tp[0] = x + 1 + img.spr_rects[last-' '].isx/2;
            tp[1] = img.spr_rects[last-' '].icy + 1;
            break;
         case 0x86:
         case 0x87:
                              //put mark below center of letter
            tp[0] = x + 1 + img.spr_rects[last-' '].isx/2;
            tp[1] = -1;
            break;
         case 0x82:
            tp[0] = x + 3 + img.spr_rects[last-' '].isx;
            tp[1] = img.spr_rects[last-' '].icy;
            x += 3;
            break;
         }
      }else{
         if(c==' '){
            x += 4 * img.spr_rects[0].isx;
         }else{
            if(first) first = false;
            else
               x += LetterFit(last, c) + img.spr_rects[0].isx;
            last = c;
         }
         tp[0] = x;
         tp[1] = 0;
      }
   }
   tab[i*2] = x + img.spr_rects[last-' '].isx;
   tab[i*2+1] = 0;
}

//----------------------------

C_poly_group *C_poly_text::CreateText(const char *cp1, float x, float y, float size,
   dword just, float *xr, float *yr, float fmt_rect[4], int *num_lines){

   if(!IsInit())
      return NULL;

   int i, j;
   int len = strlen(cp1);
   bool scr_clip = (just&0x10000);
   bool eol_char = (just&0x20000);
   just &= 0xffff;

   if(num_lines)
      *num_lines = 0;

   char *cpa = new char[len*2+1],
      *cp = cpa;
   S_national_table *national_table = char_tab_default;

   for(i=0, j=0; i<len; i++){
      byte c(cp1[i]);
      if(c>=128){
                              //try to find in national table
         //for(int ii=sizeof(national_table)/sizeof(S_national_table); ii--; )
         for(int ii = 0; national_table[ii].orig; ii++)
            if(national_table[ii].orig==c) break;
         if(national_table[ii].orig){
            //if(national_table[ii].base)
            cp[j++] = national_table[ii].base;
            if(national_table[ii].mark)
               cp[j++] = national_table[ii].mark;
         }
      }else
      if(c=='\\' && eol_char){
         cp[j++] = '\n';
      }else
      if(c=='`'){
         if(cp[j-1]=='i') cp[j-1]=127;
         switch(cp1[++i]){
         case '\'': cp[j++] = (char)0x80; break;
         case 'v': cp[j++] = (char)0x81; break;
         case '|': cp[j++] = (char)0x82; break;
         case ':': cp[j++] = (char)0x83; break;
         case '^': cp[j++] = (char)0x84; break;
         case 'o': cp[j++] = (char)0x85; break;
         case ',': cp[j++] = (char)0x86; break;
         case '.': cp[j++] = (char)0x87; break;
         case '`': cp[j++] = (char)0x88; break;
         case '~': cp[j++] = (char)0x89; break;
         }
      }else
         cp[j++]=c;
   }
   cp[len=j] = 0;

   C_poly_group *pg = new C_poly_group;

   short *spct = new short[(len+1)*2];    //x, y

   BuildSpaceTable(cp, spct, (len+1)*2);
   float y_scale = size / img.spr_rects[0].isy;
   float x_scale = y_scale * .75f;

   if(xr) *xr = x;
   if(yr) *yr = y;
   float x_clip_delta = 0.0f;

   short *spct1 = spct;
   while(true){
      if(num_lines) ++(*num_lines);
                              //get width of line
      const char *eol = strchr(cp, '\n');
      bool end = !eol;
      if(end) eol=cp+strlen(cp);
      int line_len(eol-cp);

      float line_width((spct1[line_len*2]-spct1[0]) * x_scale);
      float x_base = x;
                              //format text to fit in rectanlge (optional)
      if(fmt_rect){
         x_base = Max(x_base, fmt_rect[0]);
         //float x_delta = x_base-fmt_rect[0];
         if(line_width/*+x_delta*/ > fmt_rect[2]){
                              //remove some words from end
                              //1. find beg if another word
            int last_beg(line_len);
            while(true){
               for(i=last_beg; i--; )
               if(cp[i]==' ') break;
               if(i==-1){
                  i = last_beg;
                  break;
               }
               line_width = (spct1[i*2]-spct1[0]) * x_scale;
               if(line_width/*+x_delta*/ <= fmt_rect[2]) break;
               last_beg = i;
            };
            line_len = i;
         }
      }

      switch(just){
      case 1:              //center
         x_base -= line_width/2;
         if(xr) *xr -= line_width/2;
         break;
      case 2:              //right
         x_base -= line_width;
         if(xr) *xr -= line_width;
         break;
      }
      if(scr_clip){
         float new_x_base = x_base;
                              //clip onto screen
         if((x_base+line_width) > 1.3333f-.02f) new_x_base = 1.3333f-.02f-line_width;
         new_x_base = Max(0.0f, new_x_base);
         if(fabs(x_clip_delta)<fabs(new_x_base-x_base))
            x_clip_delta = new_x_base-x_base;
         x_base = new_x_base;
      }

      float spct_base = spct1[0];
      for(i=0; i<line_len; i++, ++cp, spct1+=2){
         int indx = ((byte)*cp) - ' ';
         if(indx <= 0) continue;

         const C_sprite_image::S_spr_rect &rc = img.spr_rects[indx];
         pg->AddSprite(x_base+(float)(spct1[0]-spct_base)*x_scale,
            y-((float)spct1[1]*y_scale),
            rc.wsy*y_scale, &img, indx, 
            driver->GetGraphInterface()->Scrn_sx(), driver->GetGraphInterface()->Scrn_sy());
      }
      if(!*cp) break;
                           //skip eol
      ++cp, spct1+=2;

      //if(end) break;
      y += size;
   }

   if(xr) *xr += x_clip_delta;

   delete[] spct;
   delete[] cpa;

   AddText(pg);
   return pg;
}

//----------------------------

void C_poly_text::SetPos(C_poly_group *pg, float x, float y){

   if(!pg || !pg->sprs.size()) return;
   float dx, dy;
   GetPos(pg, dx, dy);

   dx = x - dx;
   dy = y - dy;
   for(int i=0; i<pg->sprs.size(); i++){
      pg->sprs[i]->SetPos(pg->sprs[i]->fx+dx, pg->sprs[i]->fy+dy,
         driver->GetGraphInterface()->Scrn_sx(), driver->GetGraphInterface()->Scrn_sy());
   }
}

//----------------------------

void C_poly_text::GetPos(C_poly_group *pg, float &x, float &y){

   pg->sprs[0]->GetPos(x, y);
   x = 1000.0f;
                              //get left-most pos
   for(int i=pg->sprs.size(); i--; ) x=Min(x, pg->sprs[i]->fx);


}

//----------------------------

void C_poly_text::SetScreenPos(C_poly_group *pg, float x, float y, int scrn_sx, int scrn_sy){

   if(!pg->sprs.size()) return;
   float dx = x - pg->sprs[0]->sx;
   float dy = y - pg->sprs[0]->sy;
   for(int i=0; i<pg->sprs.size(); i++){
      pg->sprs[i]->SetScreenPos(pg->sprs[i]->sx+dx, pg->sprs[i]->sy+dy);
   }
}

//----------------------------

void C_poly_text::SetItalic(C_poly_group *pg, float angle){

   if(!pg->sprs.size()) return;
   float base_y = pg->sprs[0]->sy;
   for(int i=0; i<pg->sprs.size(); i++) pg->sprs[i]->MakeItalic(angle, base_y);
}

//----------------------------