#include "all.h"
#include "common.h"

//----------------------------

class C_edit_Light: public C_editor_item{
   virtual const char *GetName() const{ return "Light"; }
   PI3D_light help_light[2]; //directional, ambient
   bool status;               //status on/off
   bool light_on;             //real lights

   void AddHelpLight(){
      DelHelpLight();
      for(int i=0; i<2; i++){
         help_light[i] = ed->GetDriver()->CreateLight();
         help_light[i]->SetName("Helping light");
         if(!i){
                              //directional
            help_light[i]->SetLightType(I3DLIGHT_DIRECTIONAL);
            help_light[i]->SetDir(S_vector(1, -1, 1), 0.0);
            help_light[i]->SetColor(S_vector(1, 1, 1));
         }else{
                              //ambient
            help_light[i]->SetLightType(I3DLIGHT_AMBIENT);
            help_light[i]->SetColor(S_vector(.5, .5, .5));
         }
         help_light[i]->SetMode(I3DLIGHTMODE_VERTEX | I3DLIGHTMODE_DYNAMIC_LM);
         help_light[i]->SetPos(S_vector(0, 0, 0));
         help_light[i]->SetOn(true);
         help_light[i]->Update();
                              //add to all sectors
         struct S_hlp{
            static I3DENUMRET I3DAPI cbAdd(PI3D_frame frm, dword l){
               I3DCAST_SECTOR(frm)->AddLight((PI3D_light)l);
               return I3DENUMRET_OK;
            }
         };
         ed->GetScene()->GetPrimarySector()->AddLight(help_light[i]);
         ed->GetScene()->EnumFrames(S_hlp::cbAdd, (dword)help_light[i], ENUMF_SECTOR);
      }
      light_on = true;
   }
   void DelHelpLight(){
      if(light_on){
         for(int i=0; i<2; i++){
                              //delete of all sectors
            PI3D_sector const *lscts = help_light[i]->GetLightSectors();
            for(int j=help_light[i]->NumLightSectors(); j--; )
            lscts[j]->DeleteLight(help_light[i]);

            help_light[i]->Release();
         }
         light_on = false;
      }
   }
public:
   C_edit_Light():
      status(false),
      light_on(false)
   {}
   ~C_edit_Light(){ Close(); }

   virtual bool Init(){
      ed->AddShortcut(this, E_LIGHT_TOGGLE_HELP, "%30 &Debug\\%0 Helping light\tF10", K_F10, 0);
      return true;
   }
   virtual void Close(){
      DelHelpLight();
   }

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){
      //Action(E_LIGHT_TOGGLE_HELP, 0);
   }

   virtual dword C_edit_Light::Action(int id, void *context){

      switch(id){
      case E_LIGHT_DESTROY_HELP: 
         DelHelpLight();
         status = false;
         break;

      case E_LIGHT_MAKE_HELP:  
         AddHelpLight();
         status = true;
         break;

      case E_LIGHT_TOGGLE_HELP:
         {
            if(status){
               DelHelpLight();
               status = false;
            }else{
               AddHelpLight();
               status = true;
            }
            ed->CheckMenu(this, id, status);
            ed->Message(C_string("Help light ")+(status ? "created" : "destroyed"));
         }
         break;

      case EDIT_ACTION_BEFORE_FREE:
         DelHelpLight();
         break;

      case EDIT_ACTION_AFTER_LOAD:
         if(status) AddHelpLight();
         break;
         /*
                              //ignored commands:
      case EDIT_ACTION_PICK: case EDIT_ACTION_MISSION_SAVE_0: case EDIT_ACTION_MISSION_SAVE_1: case EDIT_ACTION_MISSION_SAVE_2: case EDIT_ACTION_MISSION_LOAD: case EDIT_ACTION_WINDOW_RESET: case EDIT_ACTION_FRAME_DELETE: case EDIT_ACTION_FRAME_DUPLICATE: case EDIT_ACTION_SET_ACTIVE:
      case EDIT_ACTION_QUERY_MODIFY: case EDIT_ACTION_VALIDATE:
         break;
      default:
         todo(C_fstring("Plugin %s: unknown command: %i", GetName(), id));
         */
      }
      return 0;
   }

//----------------------------

   virtual bool LoadState(C_chunk *ck){
      byte b;
      ck->Read(&b, sizeof(byte));
      status = b;
      ed->CheckMenu(this, 1002, status);
      return true;
   }

   virtual bool SaveState(C_chunk *ck) const{
      ck->Write(&status, sizeof(byte));
      return true;
   }
};

//----------------------------

void CreateLight(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Light;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
