#include "all.h"
#include "common.h"

//----------------------------

class C_edit_LightMap: public C_editor_item{
   virtual const char *GetName() const{ return "LightMap"; }
   C_smart_ptr<C_editor_item> e_undo, e_slct, e_modify;
                              //creation params
   bool lm_blur;
   int lm_aa_ratio;

   bool SaveUndo(PI3D_lit_object lm, bool constr){

      bool ret = false;
      const char *cp = lm->GetName();
      int nlen = strlen(cp) + 1;
      if(constr){
         C_cache c;
         byte *mem = NULL;
         size_t len = 0;
         if(c.open(&mem, &len, CACHE_WRITE_MEM)){
            c.write(cp, nlen);
                              //reserve space for len
            c.write((const char*)&nlen, sizeof(dword));
            if(I3D_SUCCESS(lm->Save((int)&c))){
               c.close();
               dword *dwp = (dword*)&mem[nlen];
               *dwp = len - nlen - sizeof(dword);
               S_edit_Undo_2 eu(GetName(), E_LMAP_CONSTRUCT_NAMED, mem, len);
               e_undo->Action(E_UNDO_SAVE, &eu);
               ret = true;
            }
            c.FreeMem(mem);
         }
      }else{
         S_edit_Undo_2 eu(GetName(), E_LMAP_DESTRUCT_NAMED, (void*)cp, nlen);
         e_undo->Action(E_UNDO_SAVE, &eu);
         ret = true;
      }
      return ret;
   }

//----------------------------

   struct S_light_map_help{   //info used during light-map construction
      dword num_all_lmaps;
      dword curr_lmap;
      C_edit_LightMap *e_lmap;
      PI3D_frame frm;        //frame being constructed
   };

   static bool I3DAPI cbLightMap(I3D_LOADMESSAGE msg, dword prm1, dword prm2, void *context){

      switch(msg){
      case CBM_PROGRESS:
         {
            const S_light_map_help &lmh = *(S_light_map_help*)context;
            const float &f = *(float*)&prm1;
            lmh.e_lmap->ed->Message(C_fstring(
               "LM: %i/%i - %s (%i%%)",
               lmh.curr_lmap + 1,
               lmh.num_all_lmaps,
               lmh.frm->GetName(),
               (int)(f*100.0f)),
               0, EM_MESSAGE, true);

            if(prm2==2){
                              //show user what he's done
               lmh.e_lmap->ed->GetScene()->Render();
               lmh.e_lmap->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
            }
         }
         break;

      case CBM_ERROR:
         {
            const S_light_map_help &lmh = *(S_light_map_help*)context;
            const char *cp = (const char*)prm1;
            LPC_editor_item e_log = lmh.e_lmap->ed->FindPlugin("Log");
            if(e_log)
               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
               C_fstring("Constructing lightmap on '%s': %s", lmh.frm->GetName(), cp));
         }
         break;
      }
      return false;
   }

//----------------------------

public:
   C_edit_LightMap():
      lm_blur(true),
      lm_aa_ratio(1)
   {}

//----------------------------

   virtual bool Init(){
      e_undo = ed->FindPlugin("Undo");
      e_slct = ed->FindPlugin("Selection");
      e_modify = ed->FindPlugin("Modify");
      if(!e_undo || !e_slct || !e_modify){
         e_undo = NULL;
         e_slct = NULL;
         e_modify = NULL;
      }

#define MB "%80 Light-&map\\"
      ed->AddShortcut(this, E_LMAP_SELECT_NOTC, MB"Select &unconstructed", 0, 0);
      ed->AddShortcut(this, E_LMAP_CONSTRUCT_SEL, MB"Construct &selection\tCtrl+F1", K_F1, SKEY_CTRL);
      ed->AddShortcut(this, E_LMAP_CONSTRUCT_ALL, MB"%a Construct &all", 0, 0);
      ed->AddShortcut(this, E_LMAP_TOGGLE_BLUR, MB"&Blur", 0, 0);
      ed->AddShortcut(this, E_LMAP_AA_1, MB"Ant&ialiasing\\&1x1 (draft)", 0, 0);
      ed->AddShortcut(this, E_LMAP_AA_2, MB"Ant&ialiasing\\&2x2", 0, 0);
      ed->AddShortcut(this, E_LMAP_AA_3, MB"Ant&ialiasing\\&3x3", 0, 0);
      ed->AddShortcut(this, E_LMAP_AA_4, MB"Ant&ialiasing\\&4x4 (production)", 0, 0);
      ed->AddShortcut(this, E_LMAP_AA_5, MB"Ant&ialiasing\\&5x5", 0, 0);
      ed->AddShortcut(this, E_LMAP_AA_6, MB"Ant&ialiasing\\&6x6 (waste of time)", 0, 0);
      ed->AddShortcut(this, E_LMAP_DESTRUCT_SEL, MB"&Destruct selection", 0, 0);
      ed->AddShortcut(this, E_LMAP_DESTRUCT_ALL, MB"D&estruct all", 0, 0);

      ed->CheckMenu(this, E_LMAP_TOGGLE_BLUR, lm_blur);
      ed->GetDriver()->SetState(RS_LM_AA_RATIO, lm_aa_ratio);
      ed->CheckMenu(this, E_LMAP_AA_1 + lm_aa_ratio - 1, true);
      return true;
   }

   virtual void Close(){ 
      e_undo = NULL;
      e_slct = NULL;
      e_modify = NULL;
   }

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){}

   virtual dword C_edit_LightMap::Action(int id, void *context){

      int i;
      switch(id){
      case E_LMAP_SELECT_NOTC:   //select all not constructed
         {
            e_slct->Action(E_SELECTION_CLEAR);

            struct S_hlp{
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  if(vis->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
                     PI3D_lit_object lm = I3DCAST_LIT_OBJECT(vis);
                     if(!lm->IsConstructed()){
                        ((LPC_editor_item)c)->Action(E_SELECTION_ADD_FRAME, lm);
                     }
                  }
                  return I3DENUMRET_OK;
               }
            };
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)(LPC_editor_item)e_slct, ENUMF_VISUAL);
         }
         break;

      case E_LMAP_CONSTRUCT_SEL:
      case E_LMAP_CONSTRUCT_ALL:
         {
            if(!ed->CanModify()) break;
            vector<PI3D_lit_object> lm_list;
                              //collect all
            switch(id){
            case E_LMAP_CONSTRUCT_SEL:
               {
                  const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
                  for(i=sel_list.size(); i--; ){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_LIT_OBJECT)
                        lm_list.push_back((PI3D_lit_object)frm);
                  }
               }
               break;
            case E_LMAP_CONSTRUCT_ALL:
               {
                  struct S_hlp{
                     static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                        PI3D_visual vis = I3DCAST_VISUAL(frm);
                        if(vis->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
                           vector<PI3D_lit_object> &lm_list = *(vector<PI3D_lit_object>*)c;
                           lm_list.push_back(I3DCAST_LIT_OBJECT(vis));
                        }
                        return I3DENUMRET_OK;
                     }
                  };
                  ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&lm_list, ENUMF_VISUAL);
               }
               break;
            }
                              //sort objects by name (at least by something)
            {
               struct S_hlp{
                  static bool cbSort(PI3D_frame f1, PI3D_frame f2){
                     return (strcmp(f1->GetName(), f2->GetName()) < 0);
                  }
               };
               sort(lm_list.begin(), lm_list.end(), S_hlp::cbSort);
            }

            S_light_map_help lmh = { lm_list.size(), 0, this };
            I3D_RESULT ir = I3D_OK;
            C_string cancel_name;

                              //let it run also in background
            ed->GetIGraph()->UpdateParams(0, 0, 0, 0, IG_SUSPENDINACTIVE);
                              //show hourglass
            long save_cursor = GetClassLong((HWND)ed->GetIGraph()->GetHWND(), GCL_HCURSOR);
            SetCursor(LoadCursor(NULL, IDC_WAIT));
            SetClassLong((HWND)ed->GetIGraph()->GetHWND(), GCL_HCURSOR, (long)LoadCursor(NULL, IDC_WAIT));
                              //count time
            dword start_time = ed->GetIGraph()->ReadTimer();

            for(i=lm_list.size(); i--; ){
               PI3D_lit_object lm = lm_list[i];
               SaveUndo(lm, lm->IsConstructed());

               lmh.frm = lm;

               ir = lm->Construct(ed->GetScene(), 
                  (lm_blur ? I3D_LIGHT_BLUR : 0) |
                  I3D_LIGHT_GEO_COLLISIONS |
                  //I3D_LIGHT_VOL_COLLISIONS |
                  I3D_LIGHT_ALLOW_CANCEL,
                  cbLightMap, &lmh);

               pair<PI3D_frame, dword> mp(lm, E_MODIFY_FLG_VISUAL);
               e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);

               if(I3D_FAIL(ir)) break;
               ++lmh.curr_lmap;
            }
                              //restore cursor
            SetCursor((HCURSOR)save_cursor);
            SetClassLong((HWND)ed->GetIGraph()->GetHWND(), GCL_HCURSOR, save_cursor);

            ed->GetIGraph()->UpdateParams(0, 0, 0, IG_SUSPENDINACTIVE, IG_SUSPENDINACTIVE);

            ed->SetModified();
            if(I3D_SUCCESS(ir)){
               dword time_delta = ed->GetIGraph()->ReadTimer() - start_time;
               int mins = time_delta / 60000;
               int secs = (time_delta /  1000) % 60;
               int decs = (time_delta / 100) % 10;
               ed->Message(C_fstring("Lightmaps constructed: %i in %i:%.2i.%i sec",
                  lm_list.size(),
                  mins, secs, decs));
            }else
               ed->Message(C_fstring("Canceled on lightmap %i (%s)", lmh.curr_lmap+1,
                  lm_list[i]->GetName()));
         }
         break;

      case E_LMAP_DESTRUCT_SEL:
         {
            if(!ed->CanModify()) break;
            int num_done = 0;
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            for(i=sel_list.size(); i--; ){
               PI3D_frame frm = sel_list[i];
               if(frm->GetType()==FRAME_VISUAL){
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  if(vis->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
                     PI3D_lit_object lm = I3DCAST_LIT_OBJECT(vis);
                     if(lm->IsConstructed()){
                        SaveUndo(lm, true);

                        lm->Destruct();
                        ++num_done;
                     }
                  }
               }
            }
            if(num_done) ed->SetModified();
            ed->Message(C_fstring("Lightmaps destructed: %i", num_done));
         }
         break;

      case E_LMAP_DESTRUCT_ALL:
         {
            if(!ed->CanModify()) break;
            vector<PI3D_lit_object> lm_list;
            struct S_hlp{
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  if(vis->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
                     PI3D_lit_object lm = I3DCAST_LIT_OBJECT(vis);
                     if(lm->IsConstructed()){
                        vector<PI3D_lit_object> &lm_list = *(vector<PI3D_lit_object>*)c;
                        lm_list.push_back(lm);
                     }
                  }
                  return I3DENUMRET_OK;
               }
            };
                              //collect contructed
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&lm_list, ENUMF_VISUAL);
            for(int i=lm_list.size(); i--; ){
               PI3D_lit_object lm = lm_list[i];
               SaveUndo(lm, true);
               lm->Destruct();
            }
            if(lm_list.size()) ed->SetModified();
            ed->Message(C_fstring("Lightmaps destructed: %i", lm_list.size()));
         }
         break;

      case E_LMAP_DESTRUCT_NAMED:
         {
            if(!ed->CanModify()) break;
            const char *cp = (const char*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(cp, ENUMF_BACKDROP | ENUMF_ALL);
            if(frm){
               if(frm->GetType()==FRAME_VISUAL){
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  if(vis->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
                     PI3D_lit_object lm = I3DCAST_LIT_OBJECT(vis);
                     if(lm->IsConstructed()){   //should always pass
                        SaveUndo(lm, true);
                     }
                     lm->Destruct();
                  }
               }
            }else
               ed->Message(C_fstring("LightMap(E_LMAP_DESTRUCT_NAMED): can't find frame '%s'", cp));
         }
         break;

      case E_LMAP_CONSTRUCT_NAMED:
         {
            if(!ed->CanModify()) break;
            const char *cp = (const char*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(cp, ENUMF_BACKDROP | ENUMF_ALL);
            if(frm){
               if(frm->GetType()==FRAME_VISUAL){
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  if(vis->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
                     PI3D_lit_object lm = I3DCAST_LIT_OBJECT(vis);
                     SaveUndo(lm, lm->IsConstructed());
                              //load from memory
                     C_cache c;
                     int nlen = strlen(cp)+1;
                     dword *dwp = (dword*)&cp[nlen];
                     byte *mem = (byte*)&dwp[1];
                     size_t mem_size = *dwp;
                     if(c.open(&mem, &mem_size, CACHE_READ_MEM)){
                        lm->Load((int)&c);
                     }
                  }
               }
            }else
               ed->Message(C_fstring("LightMap(E_LMAP_CONSTRUCT_NAMED): can't find frame '%s'", cp));
         }
         break;

      case E_LMAP_TOGGLE_BLUR:
         lm_blur = !lm_blur;
         ed->CheckMenu(this, E_LMAP_TOGGLE_BLUR, lm_blur);
         break;

      case E_LMAP_AA_1:
      case E_LMAP_AA_2:
      case E_LMAP_AA_3:
      case E_LMAP_AA_4:
      case E_LMAP_AA_5:
      case E_LMAP_AA_6:
         ed->CheckMenu(this, E_LMAP_AA_1 + lm_aa_ratio - 1, false);
         lm_aa_ratio = id - E_LMAP_AA_1 + 1;
         ed->GetDriver()->SetState(RS_LM_AA_RATIO, lm_aa_ratio);
         ed->CheckMenu(this, E_LMAP_AA_1 + lm_aa_ratio - 1, true);
         break;
      }
      return 0;
   }

#define VERSION 5

   virtual bool LoadState(C_chunk *ck){
                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version!=VERSION)
         return false;

      ed->CheckMenu(this, E_LMAP_AA_1 + lm_aa_ratio - 1, false);
                              //read variables
      ck->Read(&lm_blur, sizeof(byte));
      ck->Read(&lm_aa_ratio, sizeof(int));

      ed->CheckMenu(this, E_LMAP_TOGGLE_BLUR, lm_blur);
      ed->GetDriver()->SetState(RS_LM_AA_RATIO, lm_aa_ratio);
      ed->CheckMenu(this, E_LMAP_AA_1 + lm_aa_ratio - 1, true);
      return true;
   }

   virtual bool SaveState(C_chunk *ck) const{
                              //write version
      byte version = VERSION;
      ck->Write(&version, sizeof(byte));
                              //write variables
      ck->Write(&lm_blur, sizeof(byte));
      ck->Write(&lm_aa_ratio, sizeof(int));
      return true;
   }
};

//----------------------------

void CreateLightMap(LPC_editor ed){
   LPC_editor_item ei = new C_edit_LightMap;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
