#include <stddef.h>
#include <i3d\insanity.h>
#include "loader.h"

#pragma pack(push, 1)
struct S_ckhdr{
   word type;
   size_t size;
};
#pragma pack(pop)

//----------------------------
                              //read
bool C_chunk::ROpen(const char *fname){
   if(!IsInit()){
      cache.open(fname, CACHE_READ);
      if(!cache.fail()){
         out=false;
         return init=true;
      }
   }
   return false;
}

//----------------------------

bool C_chunk::ROpen(byte *mem, dword size){

   if(!IsInit()){
      cache.open((byte**)&mem, (size_t*)&size, CACHE_READ_MEM);
      if(!cache.fail()){
         out=false;
         return init=true;
      }
   }
   return false;
}

//----------------------------

bool C_chunk::RAscend(CK_TYPE &t){
   Assert(!out);
   if(IsInit()){
      S_ckhdr chkdr;
      size_t rl;
      rl=cache.read((char*)&chkdr, sizeof(S_ckhdr));
      if(rl<sizeof(S_ckhdr)) return false;
      level.push_back(chkdr.size+cache.tellg()-sizeof(S_ckhdr));
      t=chkdr.type;
      return true;
   }
   return false;
}

//----------------------------
                              //write
bool C_chunk::WOpen(const char *fname){
   if(!IsInit()){
      cache.open(fname, CACHE_WRITE);
      if(!cache.fail()){
         out=true;
         return init=true;
      }
   }
   return false;
}

//----------------------------

bool C_chunk::WOpen(byte **rmem, dword *rsize){
   if(!IsInit()){
      cache.open(rmem, (size_t*)rsize, CACHE_WRITE_MEM);
      if(!cache.fail()){
         out=true;
         return init=true;
      }
   }
   return false;
}

//----------------------------

bool C_chunk::WAscend(CK_TYPE ckt){
   Assert(out);
   if(IsInit()){
      S_ckhdr chkdr;
      chkdr.type=ckt;
                              //write partial header
      level.push_back(cache.tellg());
      cache.write((char*)&chkdr, sizeof(S_ckhdr));
      return true;
   }
   return false;
}

//----------------------------
                              //common
bool C_chunk::Descend(){
   if(IsInit()){
      if(!out){               //in
                              //skip to rest of chunk
         cache.seekp(level.back());
         level.pop_back();
      }else{                  //out
                              //write chunk header
         size_t curr_pos=cache.tellg();
         size_t chk_pos=level.back();
         Assert(level.size());
         level.pop_back();
         size_t chk_size=curr_pos-chk_pos;
         cache.seekp(chk_pos+offsetof(S_ckhdr, size));
         cache.write((char*)&chk_size, sizeof(size_t));
         cache.seekp(curr_pos);
      }
      return true;
   }
   return false;
}

//----------------------------

bool C_chunk::Close(){
   if(IsInit()){
      while(level.size()) Descend();
      cache.close();
      init=false;
      return true;
   }
   return false;
}

//----------------------------
//----------------------------
