#include <rules.h>
#include <c_vector.hpp>
#include <c_string.hpp>
#include <i3d\i3d.h>
#include <c_cache.h>
#include <debug.h>

typedef word CK_TYPE;

//----------------------------
                              //H&D binary file chunks

enum E_CHUNKTYPE{

   //CT_BASECHUNK   =0x4c53,    //mission data

      CT_SCRIPTS  =0x1000,    //list of script names
         //number_of (2)
         //name [...]
      CT_SCRCON   =0x1100,
         //scr_id (2), name, [...]

//----------------------------   *** COMPATIBILITY ***
      //CT_CAMERA   =0x3000,
         CT_GC_MODE_DIST   =0x3103,
         CT_GC_MODE_ANGLE,
         CT_GC_MODE_DISP,

      CT_SCRIPT = 0x4100,
         //CT_NAME frame_name, empty string means master script
         CT_SCRIPT_NAME    =0x4101,
            //text file_name
};

//----------------------------

class C_chunk{
   C_vector<size_t> level;
   C_cache cache;
   bool out, init;
public:
   C_chunk(): init(false){}
   ~C_chunk(){ Close(); }
                              //read
   bool ROpen(const char *fname);
   bool ROpen(byte *mem, dword size);
   bool RAscend(CK_TYPE &t);
   inline bool Read(void *data, size_t sz){
      Assert(!out);
      if(IsInit() && !out){
         if(level.size()){    //don't read data after the block size
            size_t bl_left = Max(0, (int)(level.back() - cache.tellg()));
            if(sz > bl_left) return false;
         }
         if(cache.read((char*)data, sz)<sz) return false;
         //return cache.good();
         return true;
      }
      return false;
   }
   inline size_t Size() const{
      //return IsInit() ? level.back()-cache.tellg() : 0;
      if(!IsInit()) return 0;
      Assert(level.size());
      return level.back()-cache.tellg();
   }
                              //write
   bool WOpen(const char *fname);
   bool WOpen(byte **rmem, dword *rsize);
   bool WAscend(CK_TYPE);
   bool Write(const void *data, size_t sz){
      Assert(out);
      if(IsInit() && out){
         cache.write((char*)data, sz);
         return cache.good();
      }
      return false;
   }

   C_cache *GetHandle(){ return &cache; }
                              //common
   bool Descend();
   bool Close();
   inline bool IsInit() const{ return init; }

                              //specialized I/O functions
                              //write
   void WriteText(dword ct, const char *cp){
      WAscend(ct);
      Write(cp, strlen(cp)+1);
      Descend();
   }
   void WriteVector(dword ct, const S_vector &v){
      WAscend(ct);
      Write(&v, sizeof(S_vector));
      Descend();
   }
   void WriteRot(const S_vector &axis, float angle){
      WAscend(CT_ROTATION);
      Write(&axis, sizeof(S_vector));
      Write(&angle, sizeof(float));
      Descend();
   }
   void WriteRot(const S_quat &q){
      WAscend(CT_QUATERNION);
      Write(&q, sizeof(q));
      Descend();
   }
   void WriteFloat(dword ct, float f){
      WAscend(ct);
      Write(&f, sizeof(float));
      Descend();
   }
   void WriteInt(dword ct, int i){
      WAscend(ct);
      Write(&i, sizeof(int));
      Descend();
   }

                              //read
   bool ReadName(dword ct, C_string &name){
      CK_TYPE t;
      if(RAscend(t)){
         if(t==ct){
            char buf[256];
            Read(buf, Min(sizeof(buf), Size()));
            name = buf;
            Descend();
            return true;
         }
         Descend();
      }
      name = NULL;
      return false;
   }
   void ReadText(C_string &str){
      char buf[256];
      Read(buf, Min(sizeof(buf), Size()));
      str = buf;
      Descend();
   }
   S_vector ReadVector(){
      S_vector v;
      Read(&v, sizeof(S_vector));
      Descend();
      return v;
   }
   S_quat ReadQuaternion(){
      S_quat q;
      Read(&q, sizeof(q));
      Descend();
      return q;
   }
   float ReadFloat(){
      float f;
      Read(&f, sizeof(f));
      Descend();
      return f;
   }
   int ReadInt(){
      int i;
      Read(&i, sizeof(i));
      Descend();
      return i;
   }
};

//----------------------------
