#include "all.h"
#include "common.h"

//----------------------------

class C_edit_Log: public C_editor_item{
   virtual const char *GetName() const{ return "Log"; }
   C_smart_ptr<C_editor_item> e_property;

   dword last_color;

   struct S_text_queue{
      //C_string text;
      vector<char> text;      //including new-line and null terminator
      int action_id;
      dword color;

      S_text_queue(){}
      S_text_queue(const S_text_queue &at){ operator =(at); }
      void operator =(const S_text_queue &at){
         text = at.text;
         action_id = at.action_id;
         color = at.color;
      }
   };
   vector<S_text_queue> text_queue;

//----------------------------

   struct S_hlp{
      const char *cp;
      int len;
      ofstream *os;           //used at stream-out
      static DWORD CALLBACK StreamIn(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb){

         S_hlp *hp = (S_hlp*)dwCookie;
         int trans = Min((int)cb, hp->len);
         memcpy(pbBuff, hp->cp, trans);
         hp->cp += trans;
         hp->len -= trans;
         *pcb = trans;
         return 0;
      }
      static DWORD CALLBACK StreamOut(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb){

         S_hlp *hp = (S_hlp*)dwCookie;
         assert(hp->os);
         hp->os->write((const char*)pbBuff, cb);
         *pcb = cb;
         return 0;
      }
   };


//----------------------------

   struct S_add_text{
      const char *cp;
      dword color;
   };

//----------------------------

   static BOOL CALLBACK dlgLog(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

      switch(uMsg){
      case WM_INITDIALOG:
         SetWindowLong(hwnd, GWL_USERDATA, lParam);
         return 1;

      case WM_APP + 100:
         {
            const S_add_text *at = (S_add_text*)lParam;
            const char *cp = at->cp;
            HWND hwnd_edit = GetDlgItem(hwnd, IDC_EDIT_LOG1);
            dword len = SendMessage(hwnd_edit, WM_GETTEXTLENGTH, 0, 0);
            CHARRANGE cr = {len, len};
            SendMessage(hwnd_edit, EM_EXSETSEL, 0, (LPARAM)&cr);

            {
                              //change default color
               union{
                  dword color;
                  byte b[4];
               }u;
               u.color = at->color;
               swap(u.b[0], u.b[2]);
               CHARFORMAT cf;
               cf.cbSize = sizeof(cf);
               SendMessage(hwnd_edit, EM_GETCHARFORMAT, 0, (LPARAM)&cf);
               cf.dwEffects &= ~CFE_AUTOCOLOR;
               cf.crTextColor = u.color;
               SendMessage(hwnd_edit, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&cf);
            }
                              //stream-in text
            S_hlp hlp = {cp, strlen(hlp.cp) };
            EDITSTREAM es = {(dword)&hlp, 0, S_hlp::StreamIn };
            SendMessage(hwnd_edit, EM_STREAMIN, SF_TEXT | SFF_SELECTION, (LPARAM)&es);
            /*
            {                 //line-break
               S_hlp hlp = {"\n", 1 };
               EDITSTREAM es = {(dword)&hlp, 0, S_hlp::StreamIn };
               SendMessage(hwnd_edit, EM_STREAMIN, SF_TEXT | SFF_SELECTION, (LPARAM)&es);
            }
            */
            
            SendMessage(hwnd_edit, EM_SCROLLCARET, 0, 0);
         }
         break;

      case WM_COMMAND:
         switch(LOWORD(wParam)){
         case IDC_BUTTON_LOG_CLOSE:
            {
               C_edit_Log *el = (C_edit_Log*) GetWindowLong(hwnd, GWL_USERDATA);
               el->e_property->Action(E_PROP_DEL_SHEET, el->hwnd_log);
               el->log_visible = false;

               el->ClearLogContents();
            }
            break;

         case IDC_BUTTON_DUMP_CLOSE:
            {
               C_edit_Log *el = (C_edit_Log*) GetWindowLong(hwnd, GWL_USERDATA);
               el->e_property->Action(E_PROP_DEL_SHEET, el->hwnd_dump);
               el->dump_visible = false;

               el->ClearDumpContents();
            }
            break;

         case IDC_BUTTON_LOG_SAVE:
            {            
               char buf[MAX_PATH];
               buf[0] = 0;
               OPENFILENAME on;
               memset(&on, 0, sizeof(on));
               on.lStructSize = sizeof(on);
               on.hwndOwner = hwnd;
               on.lpstrFilter = "Text files\0*.txt\0All files\0*.*\0";
               on.lpstrFile = buf;
               on.lpstrFileTitle = NULL;
               on.nMaxFile = MAX_PATH;
               on.nMaxFileTitle = 0;//MAX_PATH;
               on.lpstrTitle = "Enter name";
               on.lpstrDefExt = "txt";
               on.Flags = OFN_EXPLORER | OFN_NOREADONLYRETURN | OFN_OVERWRITEPROMPT | OFN_NOCHANGEDIR;

               bool b = GetSaveFileName(&on);
               if(b){
                  {
                     S_hlp hlp = {NULL, 0, new ofstream(buf, ios::out | ios::trunc) };
                     EDITSTREAM es = {(dword)&hlp, 0, S_hlp::StreamOut };
                     SendDlgItemMessage(hwnd, IDC_EDIT_LOG1, EM_STREAMOUT, SF_TEXT, (LPARAM)&es);
                     delete hlp.os;
                  }
               }
            }
            break;
         case IDCANCEL:
            SendMessage(GetParent(hwnd), uMsg, wParam, lParam);
            break;
         }
         break;
      }
      return 0;
   }

   HWND hwnd_log, hwnd_dump;
   bool log_visible, dump_visible;

   void InitLogWindow(){
      LoadLibrary("RICHED32.DLL");
      hwnd_log = CreateDialogParam(GetHInstance(), "IDD_LOG", 
         (HWND)ed->GetIGraph()->GetHWND(),
         dlgLog, (LPARAM)this);
   }

   void CloseLogWindow(){
      if(hwnd_log){
         if(log_visible){
            e_property->Action(E_PROP_DEL_SHEET, hwnd_log);
            log_visible = false;
         }
         e_property->Action(E_PROP_DEL_SHEET, hwnd_log);
         DestroyWindow(hwnd_log);
         hwnd_log = NULL;
      }
   }

   void ClearLogContents(){
                              //clear contents
      S_hlp hlp = {NULL, 0 };
      EDITSTREAM es = {(dword)&hlp, 0, S_hlp::StreamIn };
      HWND hwnd_edit = GetDlgItem(hwnd_log, IDC_EDIT_LOG1);
      SendMessage(hwnd_edit, EM_SETSEL, 0, -1);
      SendMessage(hwnd_edit, EM_STREAMIN, SF_TEXT | SFF_SELECTION, (LPARAM)&es);
   }

   void InitDumpWindow(){
      hwnd_dump = CreateDialogParam(GetHInstance(), "IDD_DUMP",
         (HWND)ed->GetIGraph()->GetHWND(),
         dlgLog, (LPARAM)this);
   }

   void CloseDumpWindow(){
      if(hwnd_dump){
         if(dump_visible){
            e_property->Action(E_PROP_DEL_SHEET, hwnd_dump);
            dump_visible = false;
         }
         e_property->Action(E_PROP_DEL_SHEET, hwnd_dump);
         DestroyWindow(hwnd_dump);
         hwnd_dump = NULL;
      }
   }

   void ClearDumpContents(){

      SendDlgItemMessage(hwnd_dump, IDC_DUMP, WM_SETTEXT, 0, (LPARAM)"");
   }


   //bool new_log_text;
   bool new_dump_text;
public:
   C_edit_Log():
      //new_log_text(false),
      new_dump_text(false),
      hwnd_log(NULL), hwnd_dump(NULL),
      log_visible(false), dump_visible(false),
      last_color(0)
   {}

//----------------------------

   virtual bool Init(){

      e_property = ed->FindPlugin("Properties");
      if(!e_property){
         return false;
      }
      InitLogWindow();
      InitDumpWindow();
      return true;
   }

//----------------------------

   virtual void Close(){ 

      CloseLogWindow();
      CloseDumpWindow();
      e_property = NULL;
   }

//----------------------------

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){

      if(text_queue.size()){
         if(!log_visible)
            log_visible = e_property->Action(E_PROP_ADD_SHEET, hwnd_log);
         bool new_log_text = false; 

                              //don't make it more than 32K (rich-text control can't handle greater blocks)
         const int buf_size = 32000;
         vector<char> tmp_buf(buf_size);

         for(int i=0; i<text_queue.size(); i++){
            const S_text_queue &tq = text_queue[i];
            S_add_text at;
                              //try to concanetate
            if(i!=text_queue.size()-1 && (tq.color==text_queue[i+1].color) &&
               ((tq.text.size()+text_queue[i+1].text.size()-2) < (buf_size-4))){

               at.cp = tmp_buf.begin();
               int len = tq.text.size()-1;
               memcpy(tmp_buf.begin(), tq.text.begin(), len);
               do{
                  ++i;
                  const S_text_queue &tq1 = text_queue[i];
                  int len1 = tq1.text.size()-1;
                  memcpy(tmp_buf.begin()+len, tq1.text.begin(), len1);
                  len += len1;

                  if(tq1.action_id!=E_LOG_ADD_TEXT_QUIET)
                     new_log_text = true;

               }while(i!=text_queue.size()-1 && text_queue[i+1].color==tq.color &&
                  (len+text_queue[i+1].text.size()-1) < (buf_size-4));
               tmp_buf[len] = 0;
            }else{
               at.cp = tq.text.begin();
            }
            at.color = tq.color;
            SendMessage(hwnd_log, WM_APP + 100, 0, (LPARAM)&at);

            if(tq.action_id!=E_LOG_ADD_TEXT_QUIET)
               new_log_text = true;
         }
         text_queue.clear();
         if(new_log_text)
            e_property->Action(E_PROP_SHOW_SHEET, hwnd_log);
      }

      if(new_dump_text){
         new_dump_text = false;
         e_property->Action(E_PROP_SHOW_SHEET, hwnd_dump);
      }
   }

//----------------------------

   virtual dword Action(int id, void *context){

      switch(id){
      case E_LOG_ADD_TEXT:
      case E_LOG_ADD_TEXT_QUIET:
      case E_LOG_ADD_RICH_TEXT:
         {
                              //test NULL string
            if(!context)
               break;
                              //put into queue, process in Tick
            text_queue.push_back(S_text_queue());
            S_text_queue &tq = text_queue.back();
            tq.action_id = id;
            const char *cp;
            if(tq.action_id==E_LOG_ADD_RICH_TEXT){
               const pair<const char*, dword> *p = (const pair<const char*, dword>*)context;
               cp = p->first;
               tq.color = p->second & 0xffffff;
            }else{
               cp = (const char*)context;
               tq.color = 0;
            }
            int len = strlen(cp);
            tq.text.assign(len+2);
            memcpy(tq.text.begin(), cp, len);
            tq.text[len] = '\n';
            tq.text[len+1] = 0;
         }
         break;

      case E_LOG_SET_DUMP_TEXT:
         {
            if(!context)
               break;
            if(!dump_visible){
               dump_visible = e_property->Action(E_PROP_ADD_SHEET, hwnd_dump);
            }
            SendDlgItemMessage(hwnd_dump, IDC_DUMP, WM_SETREDRAW, false, 0);
            SendDlgItemMessage(hwnd_dump, IDC_DUMP, WM_SETTEXT, 0, (LPARAM)context);
            SendDlgItemMessage(hwnd_dump, IDC_DUMP, WM_SETREDRAW, true, 0);
            UpdateWindow(GetDlgItem(hwnd_dump, IDC_DUMP));
            new_dump_text = true;
         }
         break;

      case E_LOG_CLOSE:
         if(log_visible){
            e_property->Action(E_PROP_DEL_SHEET, hwnd_log);
            log_visible = false;
            ClearLogContents();
         }
         break;

      case E_LOG_CLOSE_DUMP:
         if(dump_visible){
            e_property->Action(E_PROP_DEL_SHEET, hwnd_dump);
            dump_visible = false;
            ClearDumpContents();
         }
         break;
      }
      return 0;
   }
   //virtual void Render(){}
   //virtual bool LoadState(class C_chunk*){ return false; }
   //virtual bool SaveState(class C_chunk*) const{ return false; }
};

//----------------------------

void CreateLog(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Log;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
