#include "all.h"
#include "common.h"

//----------------------------

#define STARTUP_ORBIT_DIST 20.0f

//----------------------------

class C_edit_MouseEdit: public C_editor_item{
   virtual const char *GetName() const{ return "MouseEdit"; }

   LPC_editor_item e_undo, e_slct, e_modify;
   int mx, my;
   enum E_mode{
      MODE_EDIT, MODE_VIEW, MODE_SELECT
   } mode;
   int slct_rect[2];          //used in MODE_SELECT, beginning of selection
   bool mouse_init;
   int rotate_cumulate;
   bool uniform_scale;
   bool loc_system;
   bool collisions;
   bool polar_coords;         //use polar coordinate system
   int undo_cache_count;

   S_vector orbit_pos;
   bool orbit_ok, orbit_init;

   static const float speeds[10];
   enum{
      AXIS_X = 1, AXIS_Y = 2, AXIS_Z = 4,
   };
   byte rot_axis;
   float speed;

   C_smart_ptr<I3D_camera> edit_cam, curr_cam;
   C_smart_ptr<I3D_interpolator> cam_interp;

   HMENU igraph_menu;         //we're auto-hiding menu of igraph in full-screen mode, thus we keep this here
   bool menu_on;

                              //undo structs:
   struct S_edit_Mouse_undo_cam{
      S_vector pos;
      S_quat rot;
      char cam_name[1];
   };

   enum E_CURSOR_INDEX{
      CURSOR_DEFAULT, CURSOR_SELECT, CURSOR_LINK, CURSOR_PICK_SPEC, CURSOR_USER,
      CURSOR_LAST
   };
   HCURSOR cursor[CURSOR_LAST];
   /*
   static bool SubModelWarning(PI3D_frame frm, const char *op){

                              //check if model's child
      switch(frm->GetType()){
      case FRAME_USER:
      case FRAME_VOLUME:
         return false;
      }
      PI3D_frame prnt=frm;
      while(prnt=prnt->GetParent())
      if(prnt->GetType()==FRAME_MODEL){
         if(!mission.GetModifyFlags(frm)){
                              //show warning
            igraph->FlipToGDI();
            int i=MessageBox((HWND)igraph->GetHWND(), C_fstring("Object %s is part of model.\nDo you really want to %s it?",
               frm->GetName(), op), APP_NAME, MB_YESNO);
            return (i!=IDYES);
         }
         break;
      }
      return false;
   }
   */

   static float GetAngle(float x, float y){

      if(!x) return y>0 ? PI/2.0f : 3*PI/2.0f;
      float val;
      if(x > 0.0f){
         val = atanf(y/x);
         if(val < 0.0f) val += GetPI2();
         return val;
      }else{
         val = PI + atanf(-y/-x);
         if(val < 0.0f) val += GetPI2();
         return val;
      }
   }

//----------------------------

   void SetFrameSector(PI3D_frame frm){
                              //set by frame's center point
      I3D_bound_volume bvol;
      frm->ComputeHRBoundVolume(&bvol);
      const I3D_bbox &bbox = bvol.bbox;

      S_vector pos = frm->GetWorldPos();
      if(bbox.IsValid()) pos += bbox.min + (bbox.max-bbox.min)*.5f;
      ed->GetScene()->SetFrameSectorPos(frm, pos);
   }

//----------------------------
// Get translation vector.
   S_vector GetTranslateVector(int rx, int ry, byte mode){

      PI3D_camera cam = curr_cam;
      S_vector delta;
      switch(mode){
      case 0:              //move in X/Z
         if(!polar_coords){
            const S_vector &cdir = cam->GetMatrix()(2);
            float zx_angle = GetAngle(cdir.z, -cdir.x);
            delta = S_vector((cosf(zx_angle)*rx+sin(zx_angle)*ry) * speed / 200.0f,
               0, (sinf(zx_angle)*rx - cosf(zx_angle)*ry) * speed / 200.0f);
         }else{
            const S_vector &world_pos = cam->GetWorldPos();
            const S_vector &up = world_pos;
            S_vector right; right.GetNormal(cam->GetWorldDir(), S_vector(0, 0, 0), up);
            //S_vector right; right.GetNormal(up, S_vector(0, 0, 0), cam->GetWorldPos() - world_pos);
            right.Normalize();
            S_vector front; front.GetNormal(right, S_vector(0, 0, 0), up); front.Normalize();
            delta = right * rx * speed / 200.0f +
               front * ry * speed / 200.0f;
         }
         break;
      case 1:              //move in Y
         if(!polar_coords){
            delta = S_vector(0, -ry * speed / 200.0f, 0);
         }else{
            S_normal dir(cam->GetWorldPos());
            delta = dir * -ry * speed / 200.0f;
         }
         break;
      default:
         assert(0);
      }
      return delta;
   }

//----------------------------

   void TranslateObjects(int rx, int ry, byte mode){

      if(user_mode.ei){
         S_vector delta = GetTranslateVector(rx, ry, mode);
         user_mode.ei->Action(user_mode.modes[S_MouseEdit_user_mode::MODE_MOVE_Y].action_id, &delta);
         return;
      }

      const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
      if(!sel_list.size()) return;

                              //check if we enable undo info; init undo buffer
      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = undo_reset || (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_TRANSLATE_NAMED);
      vector<byte> undo_info;
      if(undo_on) for(int i=0; i<4; i++) undo_info.push_back(0);

      pair<PI3D_frame, dword> modify_pair;
      modify_pair.second = E_MODIFY_FLG_POSITION;

      for(int ii=0; ii<sel_list.size(); ii++){
         PI3D_frame frm = sel_list[ii];

         dword save_flags = e_modify->Action(E_MODIFY_GET_FLAGS, frm);

         modify_pair.first = frm;
         e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);

         const S_vector from = frm->GetPos();
         S_vector to = from;

         S_vector delta = GetTranslateVector(rx, ry, mode);

         const S_matrix *m_prnt;
         S_matrix m_prnt_i;
         if(frm->GetParent()){
            m_prnt = &frm->GetParent()->GetMatrix();
            m_prnt_i = *m_prnt;
            m_prnt_i.Inverse();
            delta = delta.RotateByMatrix(m_prnt_i);
         }else{
            m_prnt = &GetIdentityMatrix();
         }
         to += delta;

         if(collisions){
            I3D_collision_data cd;
            cd.from = from * (*m_prnt);
            cd.dir = (to * (*m_prnt)) - from;
            cd.flags = I3DCOL_SPHERE_GROUP | I3DCOL_SOLVE_SLIDE;
            cd.frm_ignored = frm;
            cd.frm_root = frm;
            if(ed->GetScene()->TestCollision(cd)){
               if(frm->GetParent())
                  to = cd.GetDestinationPos() * m_prnt_i;
            }
         }

         I3D_RESULT ir = frm->SetPos(to);
         if(I3D_SUCCESS(ir)){
            frm->Update();

            switch(frm->GetType()){
            case FRAME_MODEL:
               {
                  PI3D_frame prnt=frm->GetParent();
                  if(!prnt || prnt->GetType()==FRAME_SECTOR)
                     SetFrameSector(frm);
               }
               break;
            }
                              //save undo info
            if(undo_on){
               int len = strlen(frm->GetName()) + 1 + sizeof(S_vector);
               for(int i=0; i<len; i++) undo_info.push_back(0);
               S_edit_Modify_undo_pos::S_frm &ui = *(S_edit_Modify_undo_pos::S_frm*)(undo_info.end() - len);
               ui.pos = from;
               strcpy(ui.frm_name, frm->GetName());
               S_edit_Modify_undo_pos &upi = *(S_edit_Modify_undo_pos*)&undo_info.front();
               ++upi.num_frames;
            }
            ed->SetModified();
         }else{
                              //failed to change value, restore flags
            if(!(save_flags&E_MODIFY_FLG_POSITION))
               e_modify->Action(E_MODIFY_DEL_FLAGS, &modify_pair);
         }
      }
      if(undo_on){
         S_edit_Modify_undo_pos &upi = *(S_edit_Modify_undo_pos*)&undo_info.front();
         if(upi.num_frames){
                              //save undo info
            S_edit_Undo_2 eu("Modify", E_MODIFY_TRANSLATE_NAMED, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
            undo_reset = false;
         }
      }
   }

//----------------------------

   void ScaleObjects(int rx){

      const float scl_delta = 1.0f + (float)rx * speed * .001f;

      if(user_mode.ei){
         user_mode.ei->Action(user_mode.modes[S_MouseEdit_user_mode::MODE_SCALE].action_id, (void*)FloatAsInt(scl_delta));
         return;
      }

      const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
      if(!sel_list.size()) return;
                              //check if we enable undo info; init undo buffer
      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = undo_reset || (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_SCALE_NAMED);
      vector<byte> undo_info;
      if(undo_on) for(int i=0; i<4; i++) undo_info.push_back(0);

      pair<PI3D_frame, dword> modify_pair;

      for(int ii=0; ii<sel_list.size(); ii++){
         PI3D_frame frm = sel_list[ii];

         //if(SubModelWarning(frm, "scale")) break;

         bool upd = true;

         float scale = frm->GetScale();
         float orig_scale = scale;
         S_vector nu_scale, orig_nu_scale;
         bool can_nu_scale = false;
         if(!uniform_scale){
            switch(frm->GetType()){
            case FRAME_VOLUME:
               nu_scale = I3DCAST_VOLUME(frm)->GetNUScale();
               orig_nu_scale = nu_scale;
               can_nu_scale = true;
               break;
            }
         }
         if(uniform_scale){
            modify_pair.second = E_MODIFY_FLG_SCALE;
            modify_pair.first = frm;
            e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);

            scale = Max(.001f, scale*scl_delta);
            scale = Min(scale, I3D_MAX_FRAME_SCALE);
            frm->SetScale(scale);
         }else
         if(can_nu_scale){
            modify_pair.second = E_MODIFY_FLG_NU_SCALE;
            modify_pair.first = frm;
            e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);

            if(rot_axis&AXIS_X){
               nu_scale.x = Max(.001f, nu_scale.x*scl_delta);
            }
            if(rot_axis&AXIS_Y){
               nu_scale.y = Max(.001f, nu_scale.y*scl_delta);
            }
            if(rot_axis&AXIS_Z){
               nu_scale.z = Max(.001f, nu_scale.z*scl_delta);
            }
            switch(frm->GetType()){
            case FRAME_VOLUME: I3DCAST_VOLUME(frm)->SetNUScale(nu_scale); break;
            }
         }else
            upd = false;

         if(upd){
            frm->Update();
            ed->SetModified();
                              //save undo info
            if(undo_on){
               int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_scl::S_frm);
               for(int i=0; i<len; i++) undo_info.push_back(0);
               S_edit_Modify_undo_scl::S_frm &ui = *(S_edit_Modify_undo_scl::S_frm*)(undo_info.end() - len);
               if(uniform_scale){
                  ui.scl[0] = orig_scale;
               }else
                  ui.scl = orig_nu_scale;
               ui.use_non_uniform = !uniform_scale;
               strcpy(ui.frm_name, frm->GetName());
               S_edit_Modify_undo_scl &upi = *(S_edit_Modify_undo_scl*)&undo_info.front();
               ++upi.num_frames;
            }
         }
      }
      if(undo_on){
         S_edit_Modify_undo_scl &upi = *(S_edit_Modify_undo_scl*)&undo_info.front();
         if(upi.num_frames){
                              //save undo info
            S_edit_Undo_2 eu("Modify", E_MODIFY_SCALE_NAMED, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
            undo_reset = false;
         }
      }
   }

//----------------------------
// Align frame to polar coords.
   void AlignFrameToPolar(PI3D_frame frm){

      const S_matrix &fmat = frm->GetMatrix();
      const S_vector &up_dir = fmat(3);
      S_vector front; front.GetNormal(up_dir, S_vector(0, 0, 0), fmat(0));
      S_matrix m_inv = ~frm->GetParent()->GetMatrix();
      frm->SetDir1(front.RotateByMatrix(m_inv), up_dir.RotateByMatrix(m_inv));

      frm->Update();
   }

//----------------------------
// Align frame to specified axis.
   void AlignFrameToAxis(PI3D_frame frm, const S_vector &align_to, int which_axis){

      const S_matrix &fmat = frm->GetMatrix();
      //const S_vector &up_dir = fmat(3);
      S_vector front; front.GetNormal(align_to, S_vector(0, 0, 0), fmat(!which_axis ? 1 : 0));
      S_matrix m_inv = ~frm->GetParent()->GetMatrix();
      frm->SetDir1(front.RotateByMatrix(m_inv), align_to.RotateByMatrix(m_inv));

      frm->Update();
   }


//----------------------------

   enum E_ROTATE_MODE{
      RM_ANGLE,               //rotate by specified angle
      RM_POLAR_ALIGN,         //align to polar coordinates
      RM_AXIS_ALIGN,          //align with specified axis
   };

//----------------------------
// Rotate selected objects by specified angle.
   void RotateObjects(E_ROTATE_MODE rm, float angle, const S_vector *axis = NULL,
      int align_axis = 0){

      if(user_mode.ei){
         user_mode.ei->Action(user_mode.modes[S_MouseEdit_user_mode::MODE_ROTATE].action_id, 0);
         return;
      }

      const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
      if(!sel_list.size())
         return;
      bool single_sel = (sel_list.size()==1);

      dword undo_action = single_sel ? E_MODIFY_ROTATE_NAMED : E_MODIFY_ROT_TRANS_NAMED;
      bool undo_on = false;

      vector<byte> undo_info;
      {
                              //check if we enable undo info; init undo buffer
         const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
         undo_on = undo_reset || (!top_name || strcmp(top_name, "Modify") ||
            e_undo->Action(E_UNDO_GET_TOP_ACTION)!=undo_action);
         if(undo_on) for(int i=0; i<sizeof(int); i++) undo_info.push_back(0);
      }

      pair<PI3D_frame, dword> modify_pair;
      modify_pair.second = E_MODIFY_FLG_ROTATION;

      S_vector pivot_all;
      pivot_all.Zero();

      if(!single_sel){
         modify_pair.second |= E_MODIFY_FLG_POSITION;
                              //compute center point
         I3D_bbox bbox_all;      
         bbox_all.Invalidate();

         for(int i=sel_list.size(); i--; ){
            PI3D_frame frm = sel_list[i];
            pivot_all += frm->GetWorldPos();
         }
         pivot_all /= sel_list.size();
      }

      S_quat q1;              //to be computed only once
      S_matrix mat_pos;

      for(int ii=0; ii<sel_list.size(); ii++){
         PI3D_frame frm = sel_list[ii];

         modify_pair.first = frm;
         e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);

         S_quat q = frm->GetRot();

                              //save undo info
         if(undo_on){
            int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_rot::S_frm);
            for(int i=0; i<len; i++) undo_info.push_back(0);
            S_edit_Modify_undo_rot::S_frm &ui = *(S_edit_Modify_undo_rot::S_frm*)(undo_info.end() - len);
            ui.rot = q;
            ui.use_pos = !single_sel; 
            if(ui.use_pos) ui.pos = frm->GetPos();
            strcpy(ui.frm_name, frm->GetName());
            S_edit_Modify_undo_rot &upi = *(S_edit_Modify_undo_rot*)&undo_info.front();
            ++upi.num_frames;
         }

         switch(rm){
         case RM_ANGLE:
            {
               if(ii==0){
                  S_vector axis;
                  axis.Zero();
                  if(rot_axis&AXIS_X) axis.x = 1.0f;
                  else
                  if(rot_axis&AXIS_Y) axis.y = 1.0f;
                  else
                  if(rot_axis&AXIS_Z) axis.z = 1.0f;

                  if(loc_system && single_sel){
                                    //rotate by frame's matrix
                     const S_matrix &m = frm->GetFrameMatrix();
                     S_vector axis1 = axis.RotateByMatrix(m);
                     if(!axis1.IsNull()) axis = axis1;
                  }else
                                    //put to local coord system
                  if(frm->GetParent()){
                     S_matrix m = frm->GetParent()->GetMatrix();
                     m.Inverse();
                     axis = axis.RotateByMatrix(m);
                  }

                  axis.Normalize();
                  q1.Make(axis, angle);
                  if(!single_sel)
                     mat_pos = S_matrix(q1);
               }

               frm->SetRot(q * q1);

               if(!single_sel){
                  PI3D_frame prnt = frm->GetParent();
                  if(prnt){
                     S_vector pos = frm->GetWorldPos();
                     S_vector delta_pos = pos - pivot_all;
                     delta_pos = delta_pos * mat_pos;
                     S_matrix m = prnt->GetMatrix();
                     m.Inverse();
                     frm->SetPos((pivot_all + delta_pos) * m);
                  }
               }
               frm->Update();
            }
            break;

         case RM_POLAR_ALIGN:
            AlignFrameToPolar(frm);
            break;

         case RM_AXIS_ALIGN:
            AlignFrameToAxis(frm, *axis, align_axis);
            break;

         default: assert(0);
         }
         ed->SetModified();
      }
                              //save undo info
      if(undo_on){
         S_edit_Modify_undo_rot &upi = *(S_edit_Modify_undo_rot*)&undo_info.front();
         if(upi.num_frames){
                              //save undo info
            S_edit_Undo_2 eu("Modify", undo_action, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
            undo_reset = false;
         }
      }
   }

//----------------------------

   bool LinkTo(PI3D_frame link_to, const vector<PI3D_frame> &sel_list){

      if(!ed->CanModify()) return false;
      /*
                              //check if we're linking to model's child
      {
         PI3D_frame prnt = link_to;
         while(prnt=prnt->GetParent())
         if(prnt->GetType()==FRAME_MODEL){
            ed->GetIGraph()->FlipToGDI();
            int i = MessageBox((HWND)ed->GetIGraph()->GetHWND(), C_fstring("Object %s is part of model.\nDo you really want to link to it?",
               link_to->GetName()), "", MB_YESNO);
            if(i==IDNO) return false;
         }
      }
      */
      int i;

                              //init undo info
      vector<byte> undo_info;
      for(i=0; i<4; i++) undo_info.push_back(0);

      for(i=sel_list.size(); i--; ){
         PI3D_frame frm = sel_list[i];
         PI3D_frame curr_prnt = frm->GetParent();
                              //check if already has this parent
         if(curr_prnt==link_to)
            continue;

                              //check if we're re-linking child of a model
         {
            int i = IDOK;
            PI3D_frame prnt = frm;
            while(prnt=prnt->GetParent())
            if(prnt->GetType()==FRAME_MODEL){
               i = MessageBox((HWND)ed->GetIGraph()->GetHWND(), C_fstring("Object %s is part of model.\nDo you really want to re-link it?",
                  frm->GetName()), "", MB_YESNOCANCEL);
               break;
            }
            if(i==IDNO) continue;
            if(i==IDCANCEL) break;
         }
         bool add_mod = true;
         bool was_mod = false;

         /*
         int indx = e_modify->Action(E_MODIFY_GET_ITEM_INDEX, frm);
         if(indx!=-1){
            const char *orig_ling = (const char*)e_modify->Action(E_MODIFY_GET_ITEM_LINK, (void*)indx);
         */
            const char *orig_ling = (const char*)e_modify->Action(E_MODIFY_GET_ITEM_LINK, frm);
            if(orig_ling){
               was_mod = true;
               if(!strcmp(orig_ling, link_to->GetName())){
                  add_mod = false;
                              //delete modification
                  pair<PI3D_frame, dword> mp(frm, E_MODIFY_FLG_LINK);
                  e_modify->Action(E_MODIFY_DEL_FLAGS, &mp);
               }
            }
         //}
         if(add_mod){
            pair<PI3D_frame, dword> mp(frm, E_MODIFY_FLG_LINK | E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE);
            e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);
         }

                              //save undo info
         S_edit_Modify_undo_lnk::S_frm ui;
         {
            ui.orig_pos = frm->GetPos();
            ui.orig_rot = frm->GetRot();
            ui.orig_scl = frm->GetScale();
         }

         I3D_RESULT ir = frm->LinkTo(link_to, I3DLINK_UPDMATRIX);
         if(I3D_SUCCESS(ir)){
            int nlen = strlen(frm->GetName()) + 1;
            int len = nlen + sizeof(S_edit_Modify_undo_lnk::S_frm);
            if(curr_prnt) len += strlen(curr_prnt->GetName());
            for(int i=0; i<len; i++) undo_info.push_back(0);
            S_edit_Modify_undo_lnk::S_frm &ui1 = *(S_edit_Modify_undo_lnk::S_frm*)(undo_info.end() - len);
            ui1 = ui;
            strcpy(ui1.frm_names, frm->GetName());
            if(curr_prnt) strcpy(&ui1.frm_names[nlen], curr_prnt->GetName());
            S_edit_Modify_undo_lnk &upi = *(S_edit_Modify_undo_lnk*)&undo_info.front();
            ++upi.num_frames;

            e_slct->Action(E_SELECTION_FLASH_FRAME, frm);
         }else{
                              //link failed
            if(was_mod){
               pair<PI3D_frame, dword> mp(frm, E_MODIFY_FLG_LINK | E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE);
               e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);
            }else
            if(add_mod){
               pair<PI3D_frame, dword> mp(frm, E_MODIFY_FLG_LINK | E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE);
               e_modify->Action(E_MODIFY_DEL_FLAGS, &mp);
            }
         }
         ed->SetModified();
      }
      ed->Message(C_fstring("%i object(s) linked to %s", sel_list.size(), link_to->GetName()));

      {                       //save undo info
         S_edit_Modify_undo_lnk &upi = *(S_edit_Modify_undo_lnk*)&undo_info.front();
         if(upi.num_frames){
                              //save undo info
            S_edit_Undo_2 eu("Modify", E_MODIFY_LINK_NAMED, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
            e_slct->Action(E_SELECTION_FLASH_FRAME, link_to);
         }
      }
      return true;
   }

//----------------------------

   void SetPolarCoords(bool b){

      polar_coords = b; 

      PI3D_camera cam = curr_cam;

      S_vector cdir = cam->GetWorldDir();

      S_matrix inv_cmat;
      if(cam->GetParent()){
         inv_cmat = cam->GetParent()->GetMatrix();
         inv_cmat.Inverse();
         cdir = cdir.RotateByMatrix(inv_cmat);
      }

      if(!polar_coords){
         cam->SetDir(cdir, 0.0f);
      }else{
         S_vector up_v = cam->GetWorldPos();
         if(cam->GetParent())
            up_v = up_v.RotateByMatrix(inv_cmat);
         cam->SetDir1(cdir, up_v);
      }
      cam->Update();
   }     

//----------------------------

   void MoveByMouse(int rx, int ry, int rz, byte mbut, int time, bool orbit_mode){

      MouseInit();
      if(cam_interp)
         return;

      bool upd_pos = false;
      bool upd = false;
      PI3D_camera cam = curr_cam;

      const S_matrix &cmat = cam->GetMatrix();
      S_vector cdir = cmat(2);

      float tsec = time*.001f;

                              //position
      bool allow_dir = true;

      S_vector cpos = cmat(3);
      S_vector new_pos = cpos;
      switch(mbut){
      case 1:                 //forward
         new_pos += cdir * speed * tsec;
         upd_pos = true;
         break;
      case 2:                 //backward
         new_pos -= cdir * speed * tsec;
         upd_pos = true;
         break;
      case 3:                 //move in x/y camera's plane
         if(rx || ry){
            float x =  speed * rx / 200.0f;
            float y = -speed * ry / 200.0f;
            new_pos += cmat(0) * x + cmat(1) * y;
            upd_pos = true;
         }
         allow_dir = false;
         break;
      case 4:
         orbit_mode = true;
         break;
      }
      
      bool upd_dir = false;
      S_vector new_dir = cdir;

      if(allow_dir && orbit_mode){

         allow_dir = false;
         if(!orbit_ok){
            float dist;
            float nr, fr;
            cam->GetRange(nr, fr);
            S_vector cpos1 = cpos + cdir * nr;

            I3D_collision_data cd;
            cd.from = cpos1;
            cd.dir = cdir;
            cd.flags = I3DGEOCOL_EDIT_FRAMES | I3DGEOCOL_RAY | I3DCOL_EXACT;
            ed->GetScene()->TestCollision(cd);
            PI3D_frame frm = cd.GetHitFrm();
            dist = cd.GetClosestHit();

            if(frm){
               orbit_pos = cpos1 + cdir*dist;
               orbit_init = true;
            }else{            //deduct point
               float len = orbit_init ? (cpos-orbit_pos).Magnitude() : STARTUP_ORBIT_DIST;
               orbit_pos = cpos + cdir*len;
            }
            orbit_ok = true;
         }
         if(orbit_ok && (rx || ry)){

            S_vector dir;

            S_vector old_dir = new_pos - orbit_pos;
            float sz = old_dir.Magnitude();
            dir = old_dir / sz;
            dir = dir - cmat(0) * rx*PI/500.0f +
               cmat(1) * ry*PI/500.0f;
            dir.Normalize();
            dir *= sz;

            new_dir = -dir;
            upd_dir = true;

            new_pos = orbit_pos + dir;
            upd_pos = true;
         }
      }else
         orbit_ok = false;

      bool make_dir = (allow_dir && (rx || ry));
                              //direction
      if(make_dir){
         new_dir.Zero();

         float x_angle = rx * PI/500.0f;
         new_dir.x = sinf(x_angle);
         new_dir.z = cosf(x_angle);

         float y_angle = -ry * PI/500.0f;
         new_dir.y = sinf(y_angle);
         new_dir.z += cosf(y_angle);

         new_dir = new_dir.RotateByMatrix(cmat);

         {
                           //check if we would turn up-head
            const S_vector &up_dir = !polar_coords ? S_vector(0, 1, 0) : cpos;
            S_vector up_plane;
            up_plane.GetNormal(up_dir, S_vector(0, 0, 0), cmat(0));
            up_plane.Normalize();

            float new_dot = new_dir.Dot(up_plane);
            if(new_dot < PI*.01f){
                              //limit angle, don't go further
               new_dir = cmat(2);
            }
         }

         upd_dir = true;
         upd = true;
      }

      if(upd_dir){

         if(cam->GetParent()){
                              //convert from world coords to local coords
            S_matrix inv_cmat = cam->GetParent()->GetMatrix();
            inv_cmat.Inverse();
            new_dir = new_dir.RotateByMatrix(inv_cmat);
         }

         if(!polar_coords){
            cam->SetDir(new_dir, 0.0f);
         }else{
            //S_vector up_dir = cpos;
            S_vector up_dir = new_pos;
            if(cam->GetParent()){
               S_matrix inv_cmat = cam->GetParent()->GetMatrix();
               inv_cmat.Inverse();
               up_dir = up_dir.RotateByMatrix(inv_cmat);
            }
            cam->SetDir1(new_dir, up_dir);
         }
      }

      if(upd || upd_pos){
         if(upd_pos){
                              //check camera collisions
            if(collisions){
               float cam_near_range, cfr;
               cam->GetRange(cam_near_range, cfr);

               I3D_collision_data cd;
               cd.from = cpos;
               cd.dir = new_pos-cpos;
               cd.flags = I3DCOL_SPHERE | I3DCOL_SOLVE_SLIDE;
               cd.radius = cam_near_range * 1.5f;
               if(ed->GetScene()->TestCollision(cd))
                  new_pos = cd.GetDestinationPos();
            }

            if(polar_coords){
                              //check if we would turn up-head
               const S_vector &up_dir = !polar_coords ? S_vector(0, 1, 0) : new_pos;
               S_vector up_plane;
               up_plane.GetNormal(up_dir, S_vector(0, 0, 0), cmat(0));
               up_plane.Normalize();

               float new_dot = new_dir.Dot(up_plane);
               if(new_dot < PI*.01f){
                                 //limit angle, don't go further
                  new_pos = cmat(3);
               }
            }

            if(cam->GetParent()){
               S_matrix inv_cmat = cam->GetParent()->GetMatrix();
               inv_cmat.Inverse();
               new_pos = new_pos * inv_cmat;
            }
            cam->SetPos(new_pos);
         }

         if(curr_cam != edit_cam){
            if(ed->CanModify()){
                              //save modifications of non-edit camera to UNDO buffer
               if(!undo_cache_count){
                  int len = strlen(curr_cam->GetName()) + sizeof(S_edit_Mouse_undo_cam);
                  byte tmp[256];
                  S_edit_Mouse_undo_cam *uc = (S_edit_Mouse_undo_cam*)tmp;
                  uc->pos = cpos;
                  uc->rot = cam->GetRot();
                  strcpy(uc->cam_name, curr_cam->GetName());
                  S_edit_Undo_2 eu(GetName(), E_MOUSE_CAMERA_MOVE, uc, len);
                  e_undo->Action(E_UNDO_SAVE, &eu);
               }
               undo_cache_count = 1000; 
                                 //since this is not editing camera, scene is changed
               pair<PI3D_frame, dword> mp(curr_cam, E_MODIFY_FLG_POSITION|E_MODIFY_FLG_ROTATION);
               e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);
               ed->SetModified();
            }else{
               ed->Message("Can't modify scene!");
            }
         }
         if(!upd_dir && polar_coords){
            S_vector up_dir = new_pos;
            if(cam->GetParent()){
               S_matrix inv_cmat = cam->GetParent()->GetMatrix();
               inv_cmat.Inverse();
               up_dir = up_dir.RotateByMatrix(inv_cmat);
            }
            cam->SetDir1(new_dir, up_dir);
         }
         cam->Update();
      }else
         undo_cache_count = Max(0, undo_cache_count-time);
   }

//----------------------------

   void UpdateParams(){
      ed->CheckMenu(this, 1000, rot_axis&AXIS_X);
      ed->CheckMenu(this, 1001, rot_axis&AXIS_Y);
      ed->CheckMenu(this, 1002, rot_axis&AXIS_Z);
      ed->CheckMenu(this, 1003, loc_system);
      ed->CheckMenu(this, 1004, uniform_scale);
      for(int i=10; i--; ){
         ed->CheckMenu(this, 1010+i, (speeds[i]==speed));
      }
      ed->CheckMenu(this, 1020, collisions);
      ed->CheckMenu(this, 1101, polar_coords);
   }

//----------------------------

   PI3D_frame PickObject(S_MouseEdit_picked::E_PICK_ACTION pa, S_vector *ppos = NULL,
      S_vector *pdir = NULL, float *pdist = NULL, S_vector *pnorm = NULL){

      PI3D_camera cam = curr_cam;
      S_vector pos, dir;
      ed->GetScene()->UnmapScreenPoint(mx, my, pos, dir);
      float cam_near_range, cfr;
      cam->GetRange(cam_near_range, cfr);
                              //start from near clipping plane
      float near_dist = cam_near_range/cosf(dir.AngleTo(cam->GetWorldDir()));
      pos += dir*near_dist;

      if(ppos) *ppos = pos;
      if(pdir) *pdir = dir;

      if(user_mode.ei)
         pa = S_MouseEdit_picked::PA_DEFAULT;

      switch(pa){
      case S_MouseEdit_picked::PA_VOLUME:
         {
            struct S_hlp{
               static bool I3DAPI cbCol(I3D_collision_data &cd){
                  return (cd.GetHitFrm()->GetType()==FRAME_VOLUME);
               }
            };
            I3D_collision_data cd;
            cd.from = pos;
            cd.dir = dir*1000.0f;
            cd.cresp_face = S_hlp::cbCol;
            ed->GetScene()->TestCollision(cd);
            return cd.GetHitFrm();
         }
         break;

      case S_MouseEdit_picked::PA_EDIT_ONLY:
         {
            I3D_collision_data cd;
            cd.from = pos;
            cd.dir = dir*1000.0f;
            cd.flags = I3DCOL_EDIT_FRMS_ONLY;
            ed->GetScene()->TestCollision(cd);
            return cd.GetHitFrm();
         }
         break;

      default:
         I3D_collision_data cd;
         cd.from = pos;
         cd.dir = dir;
         cd.flags = I3DGEOCOL_TEXTURE | I3DGEOCOL_EDIT_FRAMES |
            I3DCOL_EXACT |
            I3DGEOCOL_RAY;
         if(pa==S_MouseEdit_picked::PA_NONPICKABLE)
            cd.flags |= I3DCOL_EDIT_NONPICKABLE;
         if(ed->GetScene()->TestCollision(cd)){
            *pdist = cd.GetClosestHit();
            *pnorm = cd.GetHitNormal();
         }
         return cd.GetHitFrm();
      }
   }

   byte last_mouse_but;
   bool cursor_reset;
   bool undo_reset;           //force adding into undo buffer

                              //double-click detection:
   int last_click_time;
   byte last_click_but;

                              //user-pick mode:
   LPC_editor_item ei_user_pick; //plugin to be notified (NULL if not in that mode)
   dword user_pick_action;    //action to be called
   void CancelUserPickMode(){
      if(ei_user_pick){
         ei_user_pick->Release();
         ei_user_pick = NULL;
         cursor_reset = true;
      }
   }

                              //subobject-edit mode:
   S_MouseEdit_user_mode user_mode;
   void CancelUserMode(){
      if(user_mode.ei){
         user_mode.ei->Action(EDIT_ACTION_SUBOBJ_CANCEL);
         user_mode.ei = NULL;
         cursor_reset = true;
         LoadCursors();
      }
   }

//----------------------------

   void ManageMouseButtons(int mx1, int my1, byte mbut, dword gk, S_MouseEdit_picked::E_PICK_ACTION action = S_MouseEdit_picked::PA_DEFAULT){

      const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);

      if(last_mouse_but != mbut){
         last_mouse_but = mbut;

         LPIGraph igraph = ed->GetIGraph();
         bool double_click = false;
         if(last_click_but==last_mouse_but){
            int curr_time = igraph->ReadTimer();
            double_click = ((curr_time-last_click_time) < GetDoubleClickTime());
            if(action != S_MouseEdit_picked::PA_DEFAULT)
               double_click = false;
         }

         switch(mbut){
         case 0:              //no button
            switch(mode){
            case MODE_SELECT:
               {  
                  mode = MODE_EDIT;
                  ed->UnregForRender(this);
                  SetCapture(NULL);
                              //evaluate selection
                  int mx = igraph->Mouse_x();
                  int my = igraph->Mouse_y();
                  if(mx!=slct_rect[0] && my!=slct_rect[1]){
                     //todo("eval selection");
                  }
               }
               break;
            }
            break;

         case 1:              //LMB
         case 2:              //RMB
            {
               S_vector ppos, pdir, pnorm;
               float pdist;
               if((mbut == 2) && action==S_MouseEdit_picked::PA_VOLUME)
                  action = S_MouseEdit_picked::PA_EDIT_ONLY;
               else
               if(double_click)
                  action = S_MouseEdit_picked::PA_NONPICKABLE;
               PI3D_frame sel_frm = PickObject(action, &ppos, &pdir, &pdist, &pnorm);
               switch(action){
               case S_MouseEdit_picked::PA_VOLUME:
                  break;
               default: 
                  if(ei_user_pick)
                     break;
                  if(mbut==1 && sel_frm){
                              //go up to find if it's higher type (model, etc)
                     sel_frm = TrySelectTopModel(sel_frm);
                  }
               }
                              //broadcast picking
               S_MouseEdit_picked pi;
               pi.frm_picked = sel_frm;
               pi.screen_x = mx, pi.screen_y = my;
               pi.pick_from = ppos;
               pi.pick_dir = pdir;
               pi.pick_dist = pdist;
               pi.pick_norm = pnorm;
               pi.pick_action = (S_MouseEdit_picked::E_PICK_ACTION)action;

               ed->BroadcastAction(EDIT_ACTION_PICK, &pi);
               action = pi.pick_action;
               sel_frm = pi.frm_picked;

               if(ei_user_pick){
                  if(mbut==2){
                     CancelUserPickMode();
                  }else{
                              //call registered plugin
                     if(!ei_user_pick->Action(user_pick_action, &pi))
                        CancelUserPickMode();
                  }
                  break;
               }
               if(user_mode.ei){
                  dword id;
                  switch(action){
                  case S_MouseEdit_picked::PA_DEFAULT: id = S_MouseEdit_user_mode::MODE_DEFAULT; break;
                  case S_MouseEdit_picked::PA_TOGGLE: id = S_MouseEdit_user_mode::MODE_SELECT; break;
                  case S_MouseEdit_picked::PA_LINK: id = S_MouseEdit_user_mode::MODE_LINK; break;
                  case S_MouseEdit_picked::PA_VOLUME: id = S_MouseEdit_user_mode::MODE_PICK_SPEC; break;
                  case S_MouseEdit_picked::PA_NONPICKABLE: id = S_MouseEdit_user_mode::MODE_DOUBLECLICK; break;
                  default: assert(0);
                  }
                  user_mode.ei->Action(user_mode.modes[id].action_id, &pi);
                  break;
               }
               if(action==S_MouseEdit_picked::PA_NULL)
                  break;

               if(sel_frm){
                  int i;
                  switch(action){
                  case S_MouseEdit_picked::PA_DEFAULT:
                  case S_MouseEdit_picked::PA_VOLUME:
                  case S_MouseEdit_picked::PA_NONPICKABLE:
                     if(sel_frm!=(PI3D_frame)e_slct->Action(E_SELECTION_GET_SINGLE)){
                        e_slct->Action(E_SELECTION_CLEAR);
                        e_slct->Action(E_SELECTION_ADD_FRAME, sel_frm);
                        ed->Message(C_fstring("Object selected: '%s'", sel_frm->GetName()));
                     }
                     break;
                  case S_MouseEdit_picked::PA_TOGGLE:
                              //check if object is in selection
                     for(i=sel_list.size(); i--; )
                        if(sel_list[i]==sel_frm) break;
                     if(i==-1){
                        e_slct->Action(E_SELECTION_ADD_FRAME, sel_frm);
                        ed->Message(C_fstring("Object selected: '%s'", sel_frm->GetName()));
                     }else{
                        e_slct->Action(E_SELECTION_DEL_FRAME, sel_frm);
                        ed->Message(C_fstring("Object unselected: '%s'", sel_frm->GetName()));
                     }
                     break;
                  case S_MouseEdit_picked::PA_LINK:
                     LinkTo(sel_frm, sel_list);
                     break;
                  }
               }else{
                  switch(action){
                  case S_MouseEdit_picked::PA_DEFAULT:
                  case S_MouseEdit_picked::PA_VOLUME:
                     if(sel_list.size()){
                        e_slct->Action(E_SELECTION_CLEAR);
                        ed->Message("Selection cleared");
                     }
                     break;
                  case S_MouseEdit_picked::PA_LINK:
                     LinkTo(ed->GetScene()->GetPrimarySector(), sel_list);
                     break;
                  }
               }
            }
            break;

         case 4:
            break;
         }
         if(last_mouse_but){
            if(double_click) last_click_but = 0;
            else{
               last_click_but = last_mouse_but;
               last_click_time = igraph->ReadTimer();
            }
         }
      }
   }

//----------------------------

   int sb_mode_index;         //status-bar index for mode
   int curr_message;

   void SetMessage(int i){

      static const char *messages[] = {
         "Edit mode", "View mode", "Lasso", "Move Y",
         "Scale", "Rotate", "Link", "Add/Remove",
         "Sel. Volume", "Move XZ", "User pick",
      };
      if(curr_message!=i){
         curr_message = i;
         ed->Message(messages[i], sb_mode_index);
      }
   }

   void SetIGMenu(bool b){
      if(menu_on!=b){
         menu_on = b;
         ed->GetIGraph()->SetMenu(b ? igraph_menu : NULL);
      }
   }

   void MouseClose(){
      if(mouse_init){
         ed->GetIGraph()->MouseClose();
         ed->GetIGraph()->SetMousePos(mx, my);
         mouse_init = false;
      }
   }

   void MouseInit(){
      if(!mouse_init){
         ed->GetIGraph()->MouseInit();
         mouse_init = true;
      }
   }

//----------------------------
// Load MouseEdit cursors.
   void LoadCursors(){
                              //pre-load cursors, so that they're ready
      cursor[CURSOR_SELECT] = LoadCursor(GetHInstance(), "IDC_CURSOR_CROSS_PLUS");
      cursor[CURSOR_LINK] = LoadCursor(GetHInstance(), "IDC_CURSOR_CROSS_LINK");
      cursor[CURSOR_PICK_SPEC] = LoadCursor(GetHInstance(), "IDC_CURSOR_CROSS_VOLUME");
      cursor[CURSOR_DEFAULT] = LoadCursor(GetHInstance(), "IDC_CURSOR_CROSS");

                              //change default cursor
      SetClassLong((HWND)ed->GetIGraph()->GetHWND(), GCL_HCURSOR, (long)cursor[CURSOR_DEFAULT]);
   }
public:
   C_edit_MouseEdit():
      rotate_cumulate(0),
      undo_cache_count(0),
      uniform_scale(true),
      loc_system(true),
      last_mouse_but(0),
      rot_axis(AXIS_Y),
      last_click_but(0),
      mode(MODE_EDIT),
      orbit_ok(false),
      orbit_init(false),
      mouse_init(false),
      cursor_reset(false),
      curr_message(-1),
      undo_reset(false),
      collisions(true),
      igraph_menu(NULL),
      menu_on(true),
      polar_coords(false),
      ei_user_pick(NULL),
      e_undo(NULL),
      e_slct(NULL),
      e_modify(NULL),
      speed(5.0f)
   {
      memset(&user_mode, 0, sizeof(user_mode));
   }

   ~C_edit_MouseEdit(){
      Close();
   }

   virtual bool Init(){
      e_undo = ed->FindPlugin("Undo");
      e_slct = ed->FindPlugin("Selection");
      e_modify = ed->FindPlugin("Modify");
      if(!e_undo || !e_slct || !e_modify){
         e_undo = NULL;
         e_slct = NULL;
         e_modify = NULL;
      }
      e_undo->AddRef();
      e_slct->AddRef();
      e_modify->AddRef();

      mx = ed->GetIGraph()->Mouse_x();
      my = ed->GetIGraph()->Mouse_y();
      ed->AddShortcut(this, E_MOUSE_TOGGLE_MODE, NULL, K_SPACE, 0);

      ed->AddShortcut(this, 1000, "%20 A&xis\\%0 &X\t(Shift+) X", K_X, 0);
      ed->AddShortcut(this, 1001, "A&xis\\%0 &Y\t(Shift+) Y", K_Y, 0);
      ed->AddShortcut(this, 1002, "A&xis\\%0 &Z\t(Shift+) Z", K_Z, 0);
      ed->AddShortcut(this, 1101, "A&xis\\%10 &Polar coordinates\tCtrl+Alt+Shift+P", K_P, SKEY_CTRL | SKEY_ALT | SKEY_SHIFT);
      ed->AddShortcut(this, E_MOUSE_POLAR_ALIGN, "A&xis\\%11 Ali&gn by Y\tCtrl+P", K_P, SKEY_CTRL);
      ed->AddShortcut(this, E_MOUSE_ALIGN_TO_NORMAL, "A&xis\\%11 &Align to normal\tShift+A", K_A, SKEY_SHIFT);
      
      ed->AddShortcut(this, 1200, NULL, K_X, SKEY_SHIFT);
      ed->AddShortcut(this, 1201, NULL, K_Y, SKEY_SHIFT);
      ed->AddShortcut(this, 1202, NULL, K_Z, SKEY_SHIFT);

      ed->AddShortcut(this, 1003, "A&xis\\%20 %i &Local\t`", K_60, 0);
      ed->AddShortcut(this, 1004, "A&xis\\%20 &Uniform scale\tCtrl+U", K_U, SKEY_CTRL);

      ed->AddShortcut(this, 1100, "%10 &Edit\\%88 %i %a Link &to...", 0, 0);
      ed->AddShortcut(this, E_MOUSE_ROT_45_LEFT,  "%10 &Edit\\Rotate 45 degrees\\Left\t,", K_2C, 0);
      ed->AddShortcut(this, E_MOUSE_ROT_45_RIGHT, "%10 &Edit\\Rotate 45 degrees\\Right\t.", K_2E, 0);

      ed->AddShortcut(this, E_MOUSE_EDIT_SUBOBJECT, "%10 &Edit\\Subobject edit\tB", K_B, 0);

                              //edit/fly speed
#define SB "&Edit\\&Speed\\"
      ed->AddShortcut(this, 1010, "%10 &Edit\\%90 &Speed\\1. 20 cm\t1", K_1, 0);
      ed->AddShortcut(this, 1011, SB"2. 50 cm\t2", K_2, 0);
      ed->AddShortcut(this, 1012, SB"3. 1 m\t3", K_3, 0);
      ed->AddShortcut(this, 1013, SB"4. 2 m\t4", K_4, 0);
      ed->AddShortcut(this, 1014, SB"5. 5 m\t5", K_5, 0);
      ed->AddShortcut(this, 1015, SB"6. 10 m\t6", K_6, 0);
      ed->AddShortcut(this, 1016, SB"7. 20 m\t7", K_7, 0);
      ed->AddShortcut(this, 1017, SB"8. 50 m\t8", K_8, 0);
      ed->AddShortcut(this, 1018, SB"9. 100 m\t9", K_9, 0);
      ed->AddShortcut(this, 1019, SB"0. 200 m\t0", K_0, 0);

                              //camera control
#define CB "&View\\Camera\\"
      ed->AddShortcut(this, 1020, "%10 &Edit\\%80 Collisions\tShift+F7", K_F7, SKEY_SHIFT);
      ed->AddShortcut(this, 1040, CB"Go to &next\t]", K_5D, 0);
      ed->AddShortcut(this, 1041, CB"Go to previous\t[", K_5B, 0);
      ed->AddShortcut(this, 1042, CB"Go to &editing\tCtrl+[", K_5B, SKEY_CTRL);
      ed->AddShortcut(this, 1043, CB"Go to &selected\tCtrl+]", K_5D, SKEY_CTRL);
      ed->AddShortcut(this, 1044, CB"%i Find selected object(s)\tCtrl+F", K_F, SKEY_CTRL);

      ed->AddShortcut(this, E_MOUSE_CANCEL, NULL, K_ESC, 0);
      UpdateParams();

      ed->GetIGraph()->MouseUpdate();
      mx = ed->GetIGraph()->Mouse_x();
      my = ed->GetIGraph()->Mouse_y();

      sb_mode_index = ed->CreateStatusIndex(70);

      edit_cam = ed->GetDriver()->CreateCamera();
      edit_cam->SetName("Edit camera");
      curr_cam = edit_cam;
      edit_cam->Release();
      ed->GetScene()->SetActiveCamera(edit_cam);

      LoadCursors();
      /*
      igraph_menu = (HMENU)igraph->GetMenu();
      if(igraph->GetFlags()&IG_FULLSCREEN)
         SetIGMenu(false);
         */
      return true;
   }

//----------------------------

   virtual void Close(){ 
      if(e_undo) e_undo->Release();
      if(e_slct) e_slct->Release();
      if(e_modify) e_modify->Release();
      e_undo = NULL;
      e_slct = NULL;
      e_modify = NULL;

      edit_cam = NULL;
      curr_cam = NULL;
      CancelUserPickMode();
      MouseClose();
   }

//----------------------------

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){

      if(cam_interp){
         if(cam_interp->Tick(time)==I3D_DONE)
            cam_interp = NULL;
      }

      LPIGraph igraph = ed->GetIGraph();
      if(!mouse_init){
         mx = igraph->Mouse_x();
         my = igraph->Mouse_y();
         /*
                              //show/hide menu in full-screen mode
         if(igraph->GetFlags()&IG_FULLSCREEN){
            if(my==0) SetIGMenu(true);
            else
            if(menu_on && my>GetSystemMetrics(SM_CYMENU)*2) SetIGMenu(false);
         }
         ed->Message(C_fstring("%i", my));
         */
      }

                              //adjust speed by mouse wheel
      if(mouse_rz){
         float k = (float)mouse_rz * .5f;
         if(mouse_rz<0) k *= .666666f;
         speed *= (1.0f + k);
         speed = Max(speeds[0], Min(speeds[9], speed));
         /*
         if(mouse_rz<0){
            speed = Max(speeds[0], speed*.666f);
         }else
         if(mouse_rz>0){
            speed = Min(speeds[9], speed*1.5f);
         }
         */
         ed->Message(C_fstring("Edit speed: %.1f m/sec", speed), false);
      }

      if(ei_user_pick){
         if(mode!=MODE_VIEW){
            MouseClose();
            SetMessage(10);
            SetCursor(cursor[CURSOR_USER]);
            cursor_reset = true;

            S_MouseEdit_picked::E_PICK_ACTION pa = S_MouseEdit_picked::PA_DEFAULT;
            switch(skeys){
            case SKEY_CTRL: pa = S_MouseEdit_picked::PA_TOGGLE; break;
            case SKEY_CTRL + SKEY_SHIFT: pa = S_MouseEdit_picked::PA_LINK; break;
            }
            ManageMouseButtons(mouse_rx, mouse_ry, mouse_butt, igraph->ReadKeys(), pa);
         }else
            MoveByMouse(mouse_rx, mouse_ry, mouse_rz, mouse_butt, time, (skeys==SKEY_CTRL));
      }else
      switch(skeys){
      case SKEY_CTRL:         //toggle selection / orbit mode
         if(mode!=MODE_VIEW){
            SetMessage(7);
            MouseClose();
                              //toggle selection mode
            SetCursor(cursor[CURSOR_SELECT]);
            cursor_reset = true;
            ManageMouseButtons(mouse_rx, mouse_ry, mouse_butt, igraph->ReadKeys(), S_MouseEdit_picked::PA_TOGGLE);
         }else{
            MoveByMouse(mouse_rx, mouse_ry, mouse_rz, mouse_butt, time, true);
         }
         break;

      case SKEY_CTRL | SKEY_SHIFT:  //link
         if(mode!=MODE_VIEW){
            SetMessage(6);
            MouseClose();

            SetCursor(cursor[CURSOR_LINK]);
            cursor_reset = true;
            ManageMouseButtons(mouse_rx, mouse_ry, mouse_butt, igraph->ReadKeys(), S_MouseEdit_picked::PA_LINK);
         }
         break;

      case SKEY_CTRL | SKEY_ALT | SKEY_SHIFT:   //select volume
         if(mode!=MODE_VIEW){
            SetMessage(8);
            MouseClose();
                              //toggle selection mode
            SetCursor(cursor[CURSOR_PICK_SPEC]);
            cursor_reset = true;
            ManageMouseButtons(mouse_rx, mouse_ry, mouse_butt, igraph->ReadKeys(), S_MouseEdit_picked::PA_VOLUME);
         }
         break;

      case SKEY_ALT:          //move in Y
         SetMessage(3);
         MouseInit();
         if(mouse_ry){
            if(ed->CanModify())
               TranslateObjects(mouse_rx, mouse_ry, 1);
         }
         break;

      case SKEY_SHIFT:        //move in XZ
         SetMessage(9);
         MouseInit();
         if(mouse_rx || mouse_ry){
            if(ed->CanModify())
               TranslateObjects(mouse_rx, mouse_ry, 0);
         }
         break;

      case SKEY_CTRL | SKEY_ALT:    //scale
         SetMessage(4);
         MouseInit();
         if(mouse_rx){
            if(ed->CanModify())
               ScaleObjects(mouse_rx);
         }
         break;

      case SKEY_SHIFT | SKEY_ALT:   //rotate
         SetMessage(5);
         MouseInit();
         if(mouse_rx){
            if(ed->CanModify())
               RotateObjects(RM_ANGLE, PI/1000.0f * (float)mouse_rx);
         }
         break;

      default:                //select
         undo_reset = true;
                              //manage buttons
         if(mode!=MODE_VIEW) MouseClose();
         if(cursor_reset){
            if(!mouse_init){
               SetCursor(cursor[CURSOR_DEFAULT]);
            }
            cursor_reset = false;
         }
         SetMessage(mode);
         switch(mode){
         case MODE_VIEW:
            MoveByMouse(mouse_rx, mouse_ry, mouse_rz, mouse_butt, time, false);
            break;
         case MODE_EDIT:
         case MODE_SELECT:
            ManageMouseButtons(mouse_rx, mouse_ry, mouse_butt, igraph->ReadKeys());
            break;
         }
      }
   }

//----------------------------

   virtual dword C_edit_MouseEdit::Action(int id, void *context){

      switch(id){

      case E_MOUSE_GET_LOC_SYSTEM: return loc_system;
      case E_MOUSE_GET_UNIFORM: return uniform_scale;
      case E_MOUSE_GET_ROT_AXIS: return rot_axis;
      case E_MOUSE_GET_SPEED: return *(dword*)&speed;

      case E_MOUSE_SET_MODE:
         {
            mode = (E_mode)(dword)context;
            switch(mode){
            case MODE_VIEW:
               MouseInit();
               break;
            case MODE_EDIT:
               MouseClose();
               break;
            }
         }
         break;

      case E_MOUSE_GET_MODE:
         return mode;

      case E_MOUSE_TOGGLE_MODE:
         switch(mode){
         case MODE_EDIT: Action(E_MOUSE_SET_MODE, (void*)MODE_VIEW); break;
         case MODE_VIEW: Action(E_MOUSE_SET_MODE, (void*)MODE_EDIT); break;
         }
         break;

      case E_MOUSE_SET_USER_PICK:
         {
            CancelUserPickMode();
            CancelUserMode();
            if(context){
               S_MouseEdit_pick &p = *(S_MouseEdit_pick*)context;
               ei_user_pick = p.ei;
               ei_user_pick->AddRef();
               user_pick_action = p.action_id;
               cursor[CURSOR_USER] = (p.hcursor) ? (HCURSOR)p.hcursor : LoadCursor(GetHInstance(), "IDC_CURSOR_CROSS_USER");

               cursor_reset = true;
            }
         }
         break;

      case E_MOUSE_GET_USER_PICK:
         return (dword)ei_user_pick;

      case E_MOUSE_CANCEL:
         if(ei_user_pick){
            CancelUserPickMode();
            break;
         }
         if(user_mode.ei){
            CancelUserMode();
            break;
         }
         break;

      case E_MOUSE_SET_USER_MODE:
         {
            CancelUserPickMode();
            CancelUserMode();
            if(!context)
               break;
            memcpy(&user_mode, context, sizeof(user_mode));
            if(user_mode.modes[S_MouseEdit_user_mode::MODE_DEFAULT].hcursor){
               cursor[CURSOR_DEFAULT] = (HCURSOR)user_mode.modes[S_MouseEdit_user_mode::MODE_DEFAULT].hcursor;
               SetClassLong((HWND)ed->GetIGraph()->GetHWND(), GCL_HCURSOR, (long)cursor[CURSOR_DEFAULT]);
            }
            if(user_mode.modes[S_MouseEdit_user_mode::MODE_SELECT].hcursor)
               cursor[CURSOR_SELECT] = (HCURSOR)user_mode.modes[S_MouseEdit_user_mode::MODE_SELECT].hcursor;
            if(user_mode.modes[S_MouseEdit_user_mode::MODE_LINK].hcursor)
               cursor[CURSOR_LINK] = (HCURSOR)user_mode.modes[S_MouseEdit_user_mode::MODE_LINK].hcursor;
            if(user_mode.modes[S_MouseEdit_user_mode::MODE_PICK_SPEC].hcursor)
               cursor[CURSOR_PICK_SPEC] = (HCURSOR)user_mode.modes[S_MouseEdit_user_mode::MODE_PICK_SPEC].hcursor;
         }
         break;

      case E_MOUSE_GET_USER_MODE:
         return user_mode.ei ? (dword)&user_mode : NULL;

      case E_MOUSE_ROT_45_LEFT:
      case E_MOUSE_ROT_45_RIGHT:
         {
            RotateObjects(RM_ANGLE, (id==E_MOUSE_ROT_45_LEFT) ? PI*.25f : -PI*.25f);
            ed->Message(C_fstring("Selection rotated 45 deg %s",
               (id==E_MOUSE_ROT_45_LEFT) ? "left" : "right"));
         }
         break;

         /*
      case 1000:              //x
      case 1001:              //y
      case 1002:              //z
         {
            dword a1, a2, a3;
            switch(id){
            case 1000: a1 = AXIS_X, a2 = AXIS_Y, a3 = AXIS_Z; break;
            case 1001: a1 = AXIS_Y, a2 = AXIS_X, a3 = AXIS_Z; break;
            case 1002: a1 = AXIS_Z, a2 = AXIS_X, a3 = AXIS_Y; break;
            }
            if(rot_axis&a1){
               if(rot_axis&a2) rot_axis = a1 | a3;
               else
               if(rot_axis&a3) rot_axis = a1;
               else rot_axis = a1 | a2;
            }else rot_axis = a1;

            ed->Message(C_fstring("%s%s%s axis",
               (rot_axis&AXIS_X) ? "X" : "",
               (rot_axis&AXIS_Y) ? "Y" : "",
               (rot_axis&AXIS_Z) ? "Z" : ""), false);
            ed->CheckMenu(this, 1000, rot_axis&AXIS_X);
            ed->CheckMenu(this, 1001, rot_axis&AXIS_Y);
            ed->CheckMenu(this, 1002, rot_axis&AXIS_Z);
         }
         break;
         */
      case 1000:              //x
      case 1001:              //y
      case 1002:              //z
      case 1200:              //add x
      case 1201:              //add y
      case 1202:              //add z
         {
            switch(id){
            case 1000: rot_axis = AXIS_X; break;
            case 1001: rot_axis = AXIS_Y; break;
            case 1002: rot_axis = AXIS_Z; break;
            case 1200: rot_axis ^= AXIS_X; break;
            case 1201: rot_axis ^= AXIS_Y; break;
            case 1202: rot_axis ^= AXIS_Z; break;
            }
            ed->Message(C_fstring("%s%s%s axis",
               (rot_axis&AXIS_X) ? "X" : "",
               (rot_axis&AXIS_Y) ? "Y" : "",
               (rot_axis&AXIS_Z) ? "Z" : ""), false);
            ed->CheckMenu(this, 1000, rot_axis&AXIS_X);
            ed->CheckMenu(this, 1001, rot_axis&AXIS_Y);
            ed->CheckMenu(this, 1002, rot_axis&AXIS_Z);
         }
         break;

      case 1101:              //toggle polar mode
         SetPolarCoords(!polar_coords);
         ed->CheckMenu(this, id, polar_coords);
         ed->Message(C_fstring("Polar coordinate system %s", polar_coords ? "on" : "off"), false);
         break;

      case 1044:
         {
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            if(!sel_list.size()){
               ed->Message("Selection is empty!");
               break;
            }
            S_vector center(0, 0, 0), dir(0, 0, 0);
            for(int i=sel_list.size(); i--; ){
               CPI3D_frame frm = sel_list[i];
               switch(frm->GetType()){
               case FRAME_VISUAL:
                  {
                     const I3D_bound_volume &bvol = I3DCAST_VISUAL(frm)->GetBoundVolume();
                     S_vector p = ((bvol.bbox.min + bvol.bbox.max) * .5f) * frm->GetMatrix();
                     center += p;
                  }
                  break;
               case FRAME_MODEL:
                  {
                     const I3D_bound_volume &bvol = I3DCAST_MODEL(frm)->GetHRBoundVolume();
                     S_vector p = ((bvol.bbox.min + bvol.bbox.max) * .5f) * frm->GetMatrix();
                     center += p;
                  }
                  break;
               default:
                  center += frm->GetWorldPos();
               }
               const S_vector &fdir = frm->GetWorldDir();
               dir += S_normal(S_vector(fdir.x, 0.0f, fdir.z));
            }
            center /= sel_list.size();
            dir.y = 0.0f;
            dir.Normalize();

            S_vector dest = center;
            dest -= dir * 5.0f;
            dest.y += 5.0f;
            S_quat rot;
            rot.SetDir(center - dest, 0.0f);

            float dist = (curr_cam->GetPos() - dest).Magnitude();
            int speed = 200 + (int)(dist * 4.0f);

            PI3D_keyframe_anim anim = (PI3D_keyframe_anim)ed->GetDriver()->CreateAnimation(I3DANIM_KEYFRAME);
            {
               I3D_anim_pos_tcb pkeys[2];
               memset(pkeys, 0, sizeof(pkeys));
               pkeys[0].v = curr_cam->GetPos();
               pkeys[0].easy_from = 1.0f;
               pkeys[1].v = dest;
               pkeys[1].time = speed;
               anim->SetPositionKeys(pkeys, 2);
            }
            {
               I3D_anim_rot rkeys[2];
               memset(rkeys, 0, sizeof(rkeys));
               const S_quat &curr_rot = curr_cam->GetRot();
               curr_rot.Inverse(rkeys[0].axis, rkeys[0].angle);
               rkeys[0].easy_from = 1.0f;
               rot = -curr_rot * rot;
               rot.Inverse(rkeys[1].axis, rkeys[1].angle);
               rkeys[1].time = speed;
               anim->SetRotationKeys(rkeys, 2);
            }
            anim->SetEndTime(speed);
            cam_interp = ed->GetDriver()->CreateInterpolator();
            cam_interp->Release();
            cam_interp->SetFrame(curr_cam);
            cam_interp->SetAnimation(0, anim);
            anim->Release();
         }
         break;

      case E_MOUSE_POLAR_MODE:
         SetPolarCoords((bool)context);
         ed->CheckMenu(this, 1101, polar_coords);
         break;

      case E_MOUSE_GET_POLAR_MODE:
         return polar_coords;

      case E_MOUSE_POLAR_ALIGN:
         {
            if(polar_coords){
               RotateObjects(RM_POLAR_ALIGN, 0.0f);
            }else{
               RotateObjects(RM_AXIS_ALIGN, 0.0f, &S_vector(0, 1, 0), 1);
            }
         }
         break;

      case E_MOUSE_POLAR_ALIGN_FRAME:
         {
            if(polar_coords){
               PI3D_frame frm = (PI3D_frame)context;
               AlignFrameToPolar(frm);
            }
         }
         break;

      case E_MOUSE_ALIGN_TO_NORMAL:
         {
                              //setup pick mode
            S_MouseEdit_pick mp = {
               this, E_MOUSE_ALIGN_PICK,
               LoadCursor(GetHInstance(), "IDC_CURSOR_ALIGN")}
            ;
            Action(E_MOUSE_SET_USER_PICK, &mp);
         }
         break;

      case E_MOUSE_ALIGN_PICK:
         {
            /*
                              //ask for axis
            struct S_hlp{
               static BOOL CALLBACK cbDlg(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

                  static const dword axis_ctrls[3] = {IDC_ALIGN_X, IDC_ALIGN_Y, IDC_ALIGN_Z};
                  switch(uMsg){
                  case WM_INITDIALOG:
                     CheckDlgButton(hwnd, axis_ctrls[lParam], BST_CHECKED);
                     return 1;
                  case WM_COMMAND:
                     switch(LOWORD(wParam)){
                     case IDOK:
                        {
                           for(int i=3; i--; ){
                              if(IsDlgButtonChecked(hwnd, axis_ctrls[i]))
                                 break;
                           }
                           assert(i!=-1);
                           EndDialog(hwnd, i);
                        }
                        break;
                     case IDCANCEL:
                        EndDialog(hwnd, -1);
                        break;
                     }
                     break;
                  }
                  return 0;
               }
            };
            int i = DialogBoxParam(GetHInstance(), "IDD_ALIGNMODE", (HWND)ed->GetIGraph()->GetHWND(),
               (DLGPROC)S_hlp::cbDlg, 0);
            if(i==-1)
               break;
               */
            const S_MouseEdit_picked *pick = (const S_MouseEdit_picked*)context;
            if(!pick->frm_picked)
               return true;
            RotateObjects(RM_AXIS_ALIGN, 0, &pick->pick_norm, 1);
            return true;
         }
         break;

      case 1003:              //local
         loc_system = !loc_system;
         ed->CheckMenu(this, id, loc_system);
         ed->Message(C_string(loc_system ? "Local" : "World")+" coordinate system", false);
         break;
      case 1004:              //uniform scale
         uniform_scale = !uniform_scale;
         ed->CheckMenu(this, id, uniform_scale);
         ed->Message(C_string("Uniform scale ") + (uniform_scale ? "on" : "off"), false);
         break;
      case 1010:
      case 1011:
      case 1012:
      case 1013:
      case 1014:
      case 1015:
      case 1016:
      case 1017:
      case 1018:
      case 1019:
         {
            for(int i=10; i--; ) if(speeds[i]==speed) break;
            if(i!=-1) ed->CheckMenu(this, 1010+i, false);
            speed = speeds[id-1010];
            ed->CheckMenu(this, id, true);
            ed->Message(C_fstring("Edit speed: %.1f m/sec", speed), false);
         }
         break;
      case 1020:              //toggle collisions
         collisions = !collisions;
         ed->CheckMenu(this, id, collisions);
         ed->Message(C_fstring("Collisions %s", collisions ? "on" : "off"));
         break;

      case 1040:              //next
      case 1041:              //prev
      case 1042:              //go to default camera
      case 1043:              //go to selected camera
      case E_MOUSE_SET_CURR_CAMERA:
         {
            PI3D_camera want_cam = NULL;
            switch(id){
            case 1040:              //next
            case 1041:              //prev
               {
                                    //collect all cameras in scene
                  static struct S_hlp{
                     static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                        vector<PI3D_camera> &cl = *(vector<PI3D_camera>*)c;
                        cl.push_back(I3DCAST_CAMERA(frm));
                        return I3DENUMRET_OK;
                     }
                     static bool pr(PI3D_camera c1, PI3D_camera c2){
                        return (strcmp(c1->GetName(), c2->GetName()) < 0);
                     }
                  };
                  vector<PI3D_camera> cl;
                  ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&cl, ENUMF_CAMERA);
                  cl.push_back(edit_cam);
                  if(cl.size()<=1) break;
                                    //sort by name
                  sort(cl.begin(), cl.end(), S_hlp::pr);
                                    //find our camera
                  for(int i=cl.size(); i--; )
                     if(cl[i]==curr_cam) break;
                                    //go to next
                  i += (id==1040 ? 1 : -1);
                  i += cl.size();   //don't go to negative!
                  i %= cl.size();
                  want_cam = cl[i];
               }
               break;
            case 1042: want_cam = edit_cam; break;
            case 1043:
               {
                  PI3D_frame frm = (PI3D_frame)e_slct->Action(E_SELECTION_GET_SINGLE);
                  if(!frm || frm->GetType()!=FRAME_CAMERA && curr_cam!=frm) break;
                  want_cam = I3DCAST_CAMERA(frm);
               }
               break;
            case E_MOUSE_SET_CURR_CAMERA:
               {
                  PI3D_frame frm = (PI3D_frame)context;
                  if(frm && frm->GetType()==FRAME_CAMERA){
                     want_cam = I3DCAST_CAMERA(frm);
                  }else
                     ed->Message("MouseEdit(E_MOUSE_SET_CURR_CAMERA): need a camera parameter");
               }
               break;
            }
            if(!want_cam || curr_cam == want_cam)
               break;
                              //copy params
            want_cam->SetFOV(curr_cam->GetFOV());
            float n, f;
            curr_cam->GetRange(n, f);
            want_cam->SetRange(n, f);
            want_cam->SetOrthoScale(curr_cam->GetOrthoScale());
            want_cam->SetOrthogonal(curr_cam->GetOrthogonal());

            ed->GetScene()->SetActiveCamera(curr_cam = want_cam);
            ed->Message(C_fstring("Current camera: '%s'", curr_cam->GetName()));
         }
         break;

      case 1100:
         {
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            if(!sel_list.size()){
               ed->Message("Can't link - empty selection.");
               break;
            }
                              //ask user for dest frame
            vector<PI3D_frame> all_frames;
            struct S_hlp{
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                  vector<PI3D_frame> &all_frames = *(vector<PI3D_frame>*)c;
                  all_frames.push_back(frm);
                  return I3DENUMRET_OK;
               }
               static bool cbSortFrames(PI3D_frame f1, PI3D_frame f2){
                  return (strcmp(f1->GetName(), f2->GetName()) < 0);
               }
            };
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&all_frames, ENUMF_ALL | ENUMF_BACKDROP);
            all_frames.push_back(ed->GetScene()->GetPrimarySector());
            all_frames.push_back(ed->GetScene()->GetBackdropSector());
            sort(all_frames.begin(), all_frames.end(), S_hlp::cbSortFrames);
            int i = e_slct->Action(E_SELECTION_PROMPT, &all_frames);
            if(!i || !all_frames.size())
               break;
            if(all_frames.size() != 1){
               ed->Message("Can't link to multiple frames.");
               break;
            }
            LinkTo(all_frames.front(), sel_list);
         }
         break;

      case EDIT_ACTION_AFTER_LOAD:
         curr_cam = ed->GetScene()->GetActiveCamera();
         Action(1042, 0);
                              //update camera if polar coordinates
         if(polar_coords)
            SetPolarCoords(true);
         break;

      case EDIT_ACTION_BEFORE_FREE:
         cam_interp = NULL;
         curr_cam = edit_cam;
         MouseClose();
         CancelUserMode();
         break;

      case E_MOUSE_GET_EDIT_CAMERA:
         return (dword)(PI3D_camera)edit_cam;

      case EDIT_ACTION_MISSION_SAVE_1:
         {
            C_chunk &ck = *(C_chunk*)context;
            ck.WAscend(CT_EDITOR_PLUGIN);
            ck.WStringChunk(CT_NAME, GetName());
                              //save camera
            {
               ck.WAscend(CT_CAMERA);
               {
                  ck.WVectorChunk(CT_POSITION, edit_cam->GetPos());
                  ck.WVectorChunk(CT_DIRECTION, edit_cam->GetDir());
               }
               ck.Descend();
            }
            ck.Descend();
         }
         break;

      case EDIT_ACTION_MISSION_LOAD:
         {
            C_chunk &ck = *(C_chunk*)context;
            while(ck.Size()){
               CK_TYPE t;
               if(!ck.RAscend(t)) break;
               switch(t){
               case CT_CAMERA:
                  while(ck.Size()){ 
                     if(!ck.RAscend(t)) break;
                     switch(t){
                     case CT_POSITION: edit_cam->SetPos(ck.RVectorChunk()); break;
                     case CT_DIRECTION: edit_cam->SetDir(ck.RVectorChunk(), 0.0f); break;
                     default: ck.Descend();
                     }
                  }
                  ck.Descend();
                  break;
               }
            }
            edit_cam->Update();
         }
         break;

      case E_MOUSE_CAMERA_MOVE:
         {
            S_edit_Mouse_undo_cam *ui = (S_edit_Mouse_undo_cam*)context;
                              //find camera
            PI3D_camera cam = I3DCAST_CAMERA(ed->GetScene()->FindFrame(ui->cam_name, ENUMF_BACKDROP | ENUMF_CAMERA));
            if(cam){
               S_vector p = cam->GetPos();
               S_quat r = cam->GetRot();
               cam->SetPos(ui->pos);
               cam->SetRot(ui->rot);
               cam->Update();
                              //save undo info
               ui->pos = p;
               ui->rot = r;
               int len = strlen(ui->cam_name) + sizeof(S_edit_Mouse_undo_cam);
               S_edit_Undo_2 eu(GetName(), E_MOUSE_CAMERA_MOVE, ui, len);
               e_undo->Action(E_UNDO_SAVE, &eu);
            }else
               ed->Message(C_fstring("MouseEdit(E_MOUSE_CAMERA_MOVE): can't find camera '%s'", ui->cam_name));
         }
         break;

      case E_MOUSE_EDIT_SUBOBJECT:
         {
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            if(sel_list.size()!=1){
               ed->Message("Single selection required");
               break;
            }
                              //broadcast validation message
            struct S_hlp{
               bool ok;
               PI3D_frame frm;
               static bool cbEnum(LPC_editor_item ei, void *c){
                  S_hlp *hp = (S_hlp*)c;
                  if(ei->Action(EDIT_ACTION_SUBOBJ_EDIT, hp->frm)){
                     hp->ok = true;
                     return false;
                  }
                  return true;
               }
            } hlp;
            hlp.ok = false;
            hlp.frm = sel_list.front();
            ed->EnumPlugins(S_hlp::cbEnum, &hlp);
            if(!hlp.ok){
               ed->Message(C_fstring("Can't sub-edit '%s'", hlp.frm->GetName()));
               break;
            }
         }
         break;

         /*
                              //ignored commands:
      case EDIT_ACTION_PICK: case EDIT_ACTION_MISSION_SAVE_0: case EDIT_ACTION_MISSION_SAVE_2: case EDIT_ACTION_WINDOW_RESET: case EDIT_ACTION_FRAME_DELETE: case EDIT_ACTION_FRAME_DUPLICATE: case EDIT_ACTION_SET_ACTIVE:
      case EDIT_ACTION_QUERY_MODIFY: case EDIT_ACTION_VALIDATE:
         break;
      default:
         todo(C_fstring("Plugin %s: unknown command: %i", GetName(), id));
         */
      }
      return 0;
   }

//----------------------------

   virtual void Render(){

      LPIGraph igraph = ed->GetIGraph();
      assert(mode==MODE_SELECT);
      int mx = igraph->Mouse_x();
      int my = igraph->Mouse_y();
      S_vector rect[4];
      PI3D_camera cam = edit_cam;
      float nr, fr;
      cam->GetRange(nr, fr);
      S_plane cam_ncp;
      cam_ncp.normal = cam->GetWorldDir();
      cam_ncp.d = -cam_ncp.normal.Dot(cam->GetWorldPos() + cam_ncp.normal*nr);

      int i;
      for(i=0; i<4; i++){
         S_vector pos, dir;
         int x = !(i&1) ? slct_rect[0] : mx;
         int y = !(i&2) ? slct_rect[1] : my;
         ed->GetScene()->UnmapScreenPoint(x, y, pos, dir);
                              //get intersection with ncp
         cam_ncp.Intersection(pos, dir, rect[i]);
         rect[i] += dir*(nr*.1f);
      }
      static const word indx[] = {0, 1, 1, 3, 3, 2, 2, 0};
      ed->GetScene()->SetRenderMatrix(GetIdentityMatrix());
      ed->GetScene()->DrawLines(rect, 4, indx, sizeof(indx)/2, S_vector(1, 1, 1));
   }

//----------------------------

   virtual bool LoadState(C_chunk *ck){
                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version!=2) return false;
                              //read other variables
      ck->Read(&mx, sizeof(mx));
      ck->Read(&my, sizeof(my));
      byte mode1;
      ck->Read(&mode1, sizeof(mode1));
      Action(E_MOUSE_SET_MODE, (void*)mode1);
      ck->Read(&collisions, sizeof(byte));
      ck->Read(&speed, sizeof(speed));
      ck->Read(&polar_coords, sizeof(polar_coords));

      UpdateParams();

      SetPolarCoords(polar_coords);
      return true;
   }

   virtual bool SaveState(C_chunk *ck) const{
                              //write version
      byte version = 2;
      ck->Write(&version, sizeof(byte));
                              //write other variables
      ck->Write(&mx, sizeof(mx));
      ck->Write(&my, sizeof(my));
      ck->Write(&mode, sizeof(byte));
      ck->Write(&collisions, sizeof(byte));
      ck->Write(&speed, sizeof(speed));
      ck->Write(&polar_coords, sizeof(polar_coords));
      return true;
   }
};

//----------------------------

const float C_edit_MouseEdit::speeds[10] = { 
   .2f, .5f, 1.0f, 2.0f, 5.f, 10, 20, 50, 100, 200.0f
};

//----------------------------

void CreateMouseEdit(LPC_editor ed){
   LPC_editor_item ei = new C_edit_MouseEdit;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
