#include "all.h"
#include "common.h"
//#include <eax.h>


//----------------------------

                              //debug:
//#define DEBUG_SELECT_LAST_SHEET  //select last active sheet by default

#define IDC_VIS_PROP_GROUP 20000

//----------------------------

static void MapDlgUnits(HWND hwnd_dlg, int *x, int *y){

   RECT rc;
   SetRect(&rc, x ? *x : 0, y ? *y : 0, x ? *x : 0, y ? *y : 0);
   MapDialogRect(hwnd_dlg, &rc);
   if(x) *x=rc.left;
   if(y) *y=rc.top;
}

//----------------------------

int GetPropertySize(I3D_PROPERTYTYPE ptype, dword value){

   switch(ptype){
   case I3DPROP_NULL: return 0;
   case I3DPROP_BOOL: return 1;
   case I3DPROP_INT: return sizeof(int);
   case I3DPROP_FLOAT: return sizeof(float);
   case I3DPROP_COLOR: return sizeof(S_vector);
   case I3DPROP_VECTOR3: return sizeof(S_vector);
   case I3DPROP_VECTOR4: return sizeof(S_vectorw);
   case I3DPROP_STRING: return strlen((const char*)value) + 1;
   case I3DPROP_ENUM: return 4;
   }
   assert(0);
   return 0;
}

//----------------------------

static const struct{
   const char *name;
   I3D_LIGHTTYPE lt;
} light_type_list[] = {
   "Ambient", I3DLIGHT_AMBIENT,
   "Point", I3DLIGHT_POINT,
   "Directional", I3DLIGHT_DIRECTIONAL,
   "PointAmbient", I3DLIGHT_POINTAMBIENT,
   "Fog", I3DLIGHT_FOG,
   //"PointFog", I3DLIGHT_POINTFOG,
   "LayeredFog", I3DLIGHT_LAYEREDFOG,
   NULL
};

static const struct{
   const char *name;
   I3D_SOUNDTYPE st;
} sound_type_list[] = {
   "Ambient", I3DSOUND_AMBIENT,
   "Point", I3DSOUND_POINT,
   "Spot", I3DSOUND_SPOT,
   "PointAmbient", I3DSOUND_POINTAMBIENT,
   NULL
};

static const struct{
   const char *name;
   I3D_VOLUMETYPE vt;
} volume_type_list[] = {
   "Sphere", I3DVOLUME_SPHERE,
   "Rectangle", I3DVOLUME_RECTANGLE,
   "Box", I3DVOLUME_BOX,
   NULL
};



//----------------------------

#define WM_APP_SHOW_TAB (WM_APP + 100)
#define WM_APP_INIT_TAB (WM_APP + 100)

                              //editor plugin: Properties
class C_edit_Properties: public C_editor_item{
   virtual const char *GetName() const{ return "Properties"; }
   LPC_editor_item e_slct;
   LPC_editor_item e_modify;
   LPC_editor_item e_undo;
   C_smart_ptr<C_editor_item> e_mouseedit;

   dword igraph_init_flags;

   HWND hwnd, hwnd_tc;        //hwnd of property dialog and tab-control
   enum{
      SHEET_FRAME = 1,
      SHEET_LIGHT = 2,
      SHEET_SOUND = 4,
      SHEET_CAMERA = 8,
      SHEET_MODEL = 0x10,
      SHEET_TARGET = 0x20,
      SHEET_LSECTORS = 0x40,
      SHEET_VOLUME = 0x80,
      SHEET_VISUAL = 0x100,
      SHEET_VISUAL_PROPERTY = 0x200,   //special use - child of visual sheet
      SHEET_UNUSED = 0x400,
      SHEET_SSECTORS = 0x800,
      SHEET_SECTOR = 0x1000,
      SHEET_USER = 0x2000,     //last
   };
   dword curr_sheets;
   int curr_sheet_index;

#define NUM_SHEETS 13
   vector<HWND> hwnd_sheet;     //sheets are loaded by their name, which is "IDD_PROP_SHEET%.2i"

   int vis_property_ground_pos[2];  //pos of visual's properties, init'ed at startup
   int vis_property_ground_size[2]; //size of visual's properties, init'ed at startup

   bool on;
   bool slct_reset;           //flag set when selection notify received - used in Tick to avoid multiple changes per frame
   bool sheet_reset;          //flag set when modify notify received    - ''
   bool in_sheet_modify;      //set true when user is modifying data in sheet

                              //material table
   map<int, C_string> mat_table;
   map<int, C_string> env_table;

//----------------------------

   void RenameFrame(PI3D_frame frm, const char *new_name){

      C_string old_name = frm->GetName();
                              //check if we may concanetate with top undo entry
      const char *top_undo = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      if(top_undo && !strcmp(top_undo, GetName()) &&
         e_undo->Action(E_UNDO_GET_TOP_ACTION)==E_PROP_RENAME_FRAME){

         const char *cp = (const char*)e_undo->Action(E_UNDO_GET_TOP_CONTEXT);
         old_name = &cp[strlen(cp)+1];
         e_undo->Action(E_UNDO_POP_TOP);
      }

                              //save undo info, reverse = E_PROP_RENAME_FRAME
      int len_old = old_name.Size();
      int len_new = strlen(new_name);
      char *buf = new char[len_old + len_new + 2];
      strcpy(buf, new_name);
      strcpy(buf+len_new+1, old_name);
      S_edit_Undo_2 eu(GetName(), E_PROP_RENAME_FRAME, buf, len_old + len_new + 2);
      e_undo->Action(E_UNDO_SAVE, &eu);
      delete[] buf;

      frm->SetName(new_name);

      ed->SetModified();
   }

//----------------------------

   void SaveUndoLights(const vector<PI3D_frame> &sel_list){

      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_LIGHT);
      if(!undo_on) return;

      int i, j;

      vector<byte> undo_info;
      for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

      for(i=0; i<sel_list.size(); i++){
         PI3D_frame frm = sel_list[i];
         if(frm->GetType()!=FRAME_LIGHT) continue;
         PI3D_light lp = I3DCAST_LIGHT(frm);

         int len = strlen(frm->GetName()) + 1 + sizeof(S_edit_Modify_undo_light::S_frm);

                              //count len of sectors' names
         PI3D_sector const *spp = lp->GetLightSectors();
         for(j=lp->NumLightSectors(); j--; )
            len += strlen(spp[j]->GetName()) + 1;

         for(j=0; j<len; j++) undo_info.push_back(0);
         S_edit_Modify_undo_light::S_frm &ui = *(S_edit_Modify_undo_light::S_frm*)(undo_info.end() - len);

         ui.ltype = lp->GetLightType();
         ui.color = lp->GetColor();
         ui.power = lp->GetPower();
         lp->GetCone(ui.cone_i, ui.cone_o);
         lp->GetRange(ui.range_n, ui.range_f);
         ui.mode = lp->GetMode();
         ui.specular_color = lp->GetSpecularColor();
         ui.specular_power = lp->GetSpecularPower();
         //ui.specular_angle = lp->GetSpecularAngle();

         strcpy(ui.frm_names, frm->GetName());

                              //store names of sectors
         j = strlen(frm->GetName()) + 1;
         for(int ii=lp->NumLightSectors(); ii--; ){
            strcpy(ui.frm_names+j, spp[ii]->GetName());
            j += strlen(spp[ii]->GetName()) + 1;
         }
         ui.frm_names[j] = 0;

         S_edit_Modify_undo_light &upi = *(S_edit_Modify_undo_light*)&undo_info.front();
         ++upi.num_frames;
      }
      dword num = *(dword*)&undo_info.front();
      if(num){
                              //save undo info
         S_edit_Undo_2 eu("Modify", E_MODIFY_LIGHT, undo_info.begin(), undo_info.size());
         e_undo->Action(E_UNDO_SAVE, &eu);
      }
   }

//----------------------------

   void SaveUndoSounds(const vector<PI3D_frame> &sel_list){

      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_SOUND);
      if(!undo_on) return;

      int i, j;

      vector<byte> undo_info;
      for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

      for(i=0; i<sel_list.size(); i++){
         PI3D_frame frm = sel_list[i];
         if(frm->GetType()!=FRAME_SOUND) continue;
         PI3D_sound sp = I3DCAST_SOUND(frm);
         int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_sound::S_frm);

                              //count len of sectors' names
         PI3D_sector const *spp = sp->GetSoundSectors();
         for(j=sp->NumSoundSectors(); j--; )
            len += strlen(spp[j]->GetName()) + 1;

         C_string fname = sp->GetFileName();
                              //strip extension
         for(int ei=fname.Size(); ei--; ){
            if(fname[ei]=='.'){
               fname[ei] = 0;
               break;
            }
         }
         ed->GetSoundCache().GetRelativeDirectory(&fname[0]);
         len += fname.Size() + 1;
         for(int j=0; j<len; j++) undo_info.push_back(0);
         S_edit_Modify_undo_sound::S_frm &ui = *(S_edit_Modify_undo_sound::S_frm*)(undo_info.end() - len);

         ui.stype = sp->GetSoundType();
         ui.vol = sp->GetVolume();
         ui.out_vol = sp->GetOutVol();
         sp->GetCone(ui.cone_i, ui.cone_o);
         sp->GetRange(ui.range_n, ui.range_f, ui.fade_ratio);
         ui.loop = sp->IsLoop();
         ui.streaming = sp->IsStreaming();
         ui.on = sp->IsOn();

         strcpy(ui.frm_names, frm->GetName());
         j = strlen(frm->GetName()) + 1;
         strcpy(ui.frm_names+j, fname);
         j += strlen(ui.frm_names+j) + 1;

                              //store names of sectors
         for(int ii=sp->NumSoundSectors(); ii--; ){
            strcpy(ui.frm_names+j, spp[ii]->GetName());
            j += strlen(spp[ii]->GetName()) + 1;
         }
         ui.frm_names[j] = 0;

         S_edit_Modify_undo_sound &upi = *(S_edit_Modify_undo_sound*)&undo_info.front();
         ++upi.num_frames;
      }
      dword num = *(dword*)&undo_info.front();
      if(num){
                              //save undo info
         S_edit_Undo_2 eu("Modify", E_MODIFY_SOUND, undo_info.begin(), undo_info.size());
         e_undo->Action(E_UNDO_SAVE, &eu);
      }
   }

//----------------------------

   void SaveUndoVolumes(const vector<PI3D_frame> &sel_list){

      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_VOLUME);
      if(!undo_on) return;

      int i;

      vector<byte> undo_info;
      for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

      for(i=0; i<sel_list.size(); i++){
         PI3D_frame frm = sel_list[i];
         if(frm->GetType()!=FRAME_VOLUME) continue;
         PI3D_volume vp = I3DCAST_VOLUME(frm);
         int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_volume::S_frm);

         for(int j=0; j<len; j++) undo_info.push_back(0);
         S_edit_Modify_undo_volume::S_frm &ui = *(S_edit_Modify_undo_volume::S_frm*)(undo_info.end() - len);

         ui.vtype = vp->GetVolumeType();
         ui.material_id = vp->GetCollisionMaterial();
         strcpy(ui.frm_name, frm->GetName());

         S_edit_Modify_undo_volume &upi = *(S_edit_Modify_undo_volume*)&undo_info.front();
         ++upi.num_frames;
      }
      dword num = *(dword*)&undo_info.front();
      if(num){
                              //save undo info
         S_edit_Undo_2 eu("Modify", E_MODIFY_VOLUME, undo_info.begin(), undo_info.size());
         e_undo->Action(E_UNDO_SAVE, &eu);
      }
   }

//----------------------------

   void SaveUndoVisuals(const vector<PI3D_frame> &sel_list){

      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_VISUAL);
      if(!undo_on) return;
      int i;

      vector<byte> undo_info;
      for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

      for(i=0; i<sel_list.size(); i++){
         PI3D_frame frm = sel_list[i];
         if(frm->GetType()!=FRAME_VISUAL) continue;
         PI3D_visual vis = I3DCAST_VISUAL(frm);

         int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_visual::S_frm);
         dword vtype = vis->GetVisualType();
                              //compute size of all properties
         int num_props = ed->GetDriver()->NumProperties(vtype);
         for(int pi=num_props; pi--; ){
            dword value = vis->GetProperty(pi);
            len += GetPropertySize(ed->GetDriver()->GetPropertyType(vtype, pi), value);
         }

                              //reserve size
         for(int j=0; j<len; j++) undo_info.push_back(0);
         S_edit_Modify_undo_visual::S_frm &ui = *(S_edit_Modify_undo_visual::S_frm*)(undo_info.end() - len);

         ui.visual_type = vtype;
         ui.material_id = vis->GetCollisionMaterial();
         ui.data_size = 0;
                              //store name
         strcpy(&ui.data[ui.data_size], frm->GetName());
         ui.data_size += strlen(&ui.data[ui.data_size]) + 1;
                              //store properties
         for(pi=0; pi<num_props; pi++){
            I3D_PROPERTYTYPE ptype = ed->GetDriver()->GetPropertyType(vtype, pi);
            dword value = vis->GetProperty(pi);
            dword len = GetPropertySize(ptype, value);
            switch(ptype){
            case I3DPROP_STRING:
            case I3DPROP_VECTOR3:
            case I3DPROP_VECTOR4:
            case I3DPROP_COLOR:
               memcpy(&ui.data[ui.data_size], (const void*)value, len);
               break;
            default:
               memcpy(&ui.data[ui.data_size], &value, len);
            }
            ui.data_size += len;
         }

         S_edit_Modify_undo_visual &upi = *(S_edit_Modify_undo_visual*)&undo_info.front();
         ++upi.num_frames;
      }
      dword num = *(dword*)&undo_info.front();
      if(num){
                              //save undo info
         S_edit_Undo_2 eu("Modify", E_MODIFY_VISUAL, undo_info.begin(), undo_info.size());
         e_undo->Action(E_UNDO_SAVE, &eu);
      }
   }

//----------------------------

   void SaveUndoSectors(const vector<PI3D_frame> &sel_list){

      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_SECTOR);
      if(!undo_on) return;
      int i;

      vector<byte> undo_info;
      for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

      for(i=0; i<sel_list.size(); i++){
         PI3D_frame frm = sel_list[i];
         if(frm->GetType()!=FRAME_SECTOR) continue;
         PI3D_sector sct = I3DCAST_SECTOR(frm);

         int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_sector::S_frm);
                              //reserve size
         for(int j=0; j<len; j++) undo_info.push_back(0);
         S_edit_Modify_undo_sector::S_frm &ui = *(S_edit_Modify_undo_sector::S_frm*)(undo_info.end() - len);

         ui.env_id = sct->GetEnvironmentID();;
                              //store name
         strcpy(ui.frm_name, frm->GetName());

         S_edit_Modify_undo_sector &upi = *(S_edit_Modify_undo_sector*)&undo_info.front();
         ++upi.num_frames;
      }
      dword num = *(dword*)&undo_info.front();
      if(num){
                              //save undo info
         S_edit_Undo_2 eu("Modify", E_MODIFY_SECTOR, undo_info.begin(), undo_info.size());
         e_undo->Action(E_UNDO_SAVE, &eu);
      }
   }

//----------------------------

   void ModifyVisualsProperty(const vector<PI3D_frame> &sel_list,
      int prop_index, dword value){

      dword vtype = 0;
                              //modify all visuals
      for(int i=sel_list.size(); i--; ){
         PI3D_frame frm = sel_list[i];
         assert(frm->GetType()==FRAME_VISUAL);
         if(frm->GetType()!=FRAME_VISUAL)
            continue;
         PI3D_visual vis = I3DCAST_VISUAL(frm);
         if(!vtype){
            vtype = vis->GetVisualType();
            assert(prop_index >= 0 && prop_index < ed->GetDriver()->NumProperties(vtype));
         }else{
            assert(vis->GetVisualType() == vtype);
         }
               //modify visual's property
         vis->SetProperty(prop_index, value);

         pair<PI3D_frame, dword> p(vis, E_MODIFY_FLG_VISUAL);
         e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

         ed->SetModified();
      }
   }

//----------------------------

   void SaveUndoFrameFlags(const vector<PI3D_frame> &sel_list){

      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = (!top_name || strcmp(top_name, "Modify") ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_MODIFY_FRMFLAGS);
      if(!undo_on) return;
      int i;

      vector<byte> undo_info;
      for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

      for(i=0; i<sel_list.size(); i++){
         PI3D_frame frm = sel_list[i];
         int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_frmflags::S_frm);

         for(int j=0; j<len; j++) undo_info.push_back(0);
         S_edit_Modify_undo_frmflags::S_frm &ui = *(S_edit_Modify_undo_frmflags::S_frm*)(undo_info.end() - len);

         ui.frm_flags = frm->GetFlags();
         strcpy(ui.frm_name, frm->GetName());

         S_edit_Modify_undo_frmflags &upi = *(S_edit_Modify_undo_frmflags*)&undo_info.front();
         ++upi.num_frames;
      }
      dword num = *(dword*)&undo_info.front();
      if(num){
                              //save undo info
         S_edit_Undo_2 eu("Modify", E_MODIFY_FRMFLAGS, undo_info.begin(), undo_info.size());
         e_undo->Action(E_UNDO_SAVE, &eu);
      }
   }

//----------------------------

   static UINT APIENTRY ccHook(HWND hwnd, UINT uiMsg, WPARAM wParam, LPARAM lParam,
      bool specular){

      switch(uiMsg){
      case WM_INITDIALOG:
         {
            LPCHOOSECOLOR cc = (LPCHOOSECOLOR)lParam;
            SetWindowLong(hwnd, GWL_USERDATA, cc->lCustData);
         }
         break;
      case WM_COMMAND:
         switch(LOWORD(wParam)){
         case 706: case 707: case 708:
                              //hook RGB edit controlls
            switch(HIWORD(wParam)){
            case EN_CHANGE:
               C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
               if(ep){
                  const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                  ep->SaveUndoLights(sel_list);
                  S_vector v(GetDlgItemInt(hwnd, 706, NULL, false)/255.0f, 
                     GetDlgItemInt(hwnd, 707, NULL, false)/255.0f, 
                     GetDlgItemInt(hwnd, 708, NULL, false)/255.0f);
                  for(int i=0; i<sel_list.size(); i++){
                     if(sel_list[i]->GetType()==FRAME_LIGHT){
                        PI3D_light lp = I3DCAST_LIGHT(sel_list[i]);
                        if(specular)
                           lp->SetSpecularColor(v);
                        else
                           lp->SetColor(v);
                        lp->Update();
   
                        pair<PI3D_frame, dword> p(lp, E_MODIFY_FLG_LIGHT);
                        ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                        ep->ed->SetModified();
                        ep->ed->GetScene()->Render();
                        ep->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
                     }
                  }
               }
               break;
            }
            break;
         }
         break;
      }
      return 0;
   }

   static UINT APIENTRY ccHookColor(HWND hwnd, UINT uiMsg, WPARAM wParam, LPARAM lParam){
      return ccHook(hwnd, uiMsg, wParam, lParam, false);
   }

//----------------------------

   struct S_sheet_init{
      int sheet_index;
      C_edit_Properties *ep;
   };

   class C_property_control: public C_unknown{
   public:
      HWND hwnd;
      int base_x, base_y;

      C_property_control():
         hwnd(NULL)
      {
      }
      C_property_control(HWND h1, int bx1, int by1):
         hwnd(h1),
         base_x(bx1),
         base_y(by1)
      {
      }
      ~C_property_control(){
         if(hwnd){
            DestroyWindow(hwnd);
         }
      }
   };

   vector<C_smart_ptr<C_property_control> > prop_controls;
   int vis_prop_scroll_y, vis_prop_scroll_max_y;

//----------------------------
// Scroll all visual's properties to align with slide-bar.
   void ScrollVisualProperties(){
                     //update all controls
      for(int i=prop_controls.size(); i--; ){
         C_property_control *pc = prop_controls[i];
         if(pc->base_y==0x80000000)
            continue;
         SetWindowPos(pc->hwnd, NULL, pc->base_x, pc->base_y - vis_prop_scroll_y,
            0, 0, SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_NOZORDER | SWP_NOSIZE);
      }
   }

//----------------------------
// Replace model - load new file. Delete all frames from modify list,
// modify all common frames the same way as original.
   bool ReplaceModel(PI3D_model mod, const C_string &filename){


      C_string old_name = mod->GetFileName();
      ed->GetModelCache().GetRelativeDirectory(&old_name[0]);
      for(int i=strlen(old_name); i--; ){
         if(old_name[i]=='.'){
            old_name[i] = 0;
            break;
         }
      }

      if(filename.Matchi(old_name))
         return false;

      {
                              //save undo
         int len = strlen(mod->GetName()) + old_name.Size() + 2;
         char *cp = new char[len];
         strcpy(cp, mod->GetName());
         strcpy(cp+strlen(mod->GetName())+1, old_name);
         S_edit_Undo_2 eu(GetName(), E_PROP_RELOAD_MODEL, cp, len);
         e_undo->Action(E_UNDO_SAVE, &eu);
         delete[] cp;
      }

                              //save old params
      struct S_hlp{
         static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

            vector<C_smart_ptr<I3D_frame> > *frm_save = (vector<C_smart_ptr<I3D_frame> >*)c;
            frm_save->push_back(frm);
            return I3DENUMRET_OK;
         }
      };
      vector<C_smart_ptr<I3D_frame> > frm_save;
      mod->EnumFrames(S_hlp::cbEnum, (dword)&frm_save);
      for(int ii=frm_save.size(); ii--; ){
         dword flags = e_modify->Action(E_MODIFY_GET_FLAGS, (PI3D_frame)frm_save[ii]);;
         if(!flags){
            frm_save[ii] = frm_save.back();
            frm_save.pop_back();
         }
      }

      ed->GetModelCache().Open(mod, filename, ed->GetScene(), 0, OPEN_ERR_LOG(ed, filename));

                              //now apply changes from saved frames
                              // and remove old frames from modify list
      for(ii=frm_save.size(); ii--; ){
         PI3D_frame frm = frm_save[ii];
         PI3D_frame frm_new = mod->FindChildFrame(frm->GetOrigName());
         if(frm_new){
            dword flags = e_modify->Action(E_MODIFY_GET_FLAGS, frm);;
            flags &= (E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE);
            if(flags){
               e_modify->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(frm_new, flags));
               if(flags&E_MODIFY_FLG_POSITION)
                  frm_new->SetPos(frm->GetPos());
               if(flags&E_MODIFY_FLG_ROTATION)
                  frm_new->SetRot(frm->GetRot());
               if(flags&E_MODIFY_FLG_SCALE)
                  frm_new->SetScale(frm->GetScale());
               frm_new->Update();
            }
         }
         e_modify->Action(E_MODIFY_DEL_FRAME, frm);
      }

                              //set its own anim
      mod->SetAnimation(0, NULL, I3DANIMOP_LOOP);
      ed->SetModified();

      return true;
   }

//----------------------------
// Copy properties of visual frame from 'from' visual to all selected visuals which are
// of the same type. Saves UNDO for all changes, sets MODIFY property of editor.
   bool CopyProperties(PI3D_visual from){

      if(!ed->CanModify())
         return false;

      const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
      dword vtype = from->GetVisualType();
      dword num_props = ed->GetDriver()->NumProperties(vtype);

      SaveUndoVisuals(sel_list);

                        //change all compatible visuals
      int count = 0;
      for(int i=sel_list.size(); i--; ){
         PI3D_frame frm = sel_list[i];
         if(frm->GetType()!=FRAME_VISUAL)
            continue;
         PI3D_visual vis = I3DCAST_VISUAL(frm);
         if(vis->GetVisualType() != vtype || vis==from)
            continue;
                        //copy all properties
         for(int pi=0; pi<num_props; pi++){
            vis->SetProperty(pi, from->GetProperty(pi));
         }
         e_modify->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(frm, E_MODIFY_FLG_VISUAL));
         ++count;
      }
      SendMessage(hwnd_sheet[8], WM_APP_INIT_TAB + 8, 0, (LPARAM)&sel_list);
      if(count)
         ed->SetModified();
      ed->Message(C_fstring("Visuals updated: %i.", count), 0, EM_WARNING);
      return true;
   }

//----------------------------

   static BOOL CALLBACK dlgSheet(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

      switch(uMsg){
      case WM_INITDIALOG:
         {
            S_sheet_init *si = (S_sheet_init*)lParam;
            switch(si->sheet_index){

            case 1:           //light sheet
               {
                  HWND hwnd_t = GetDlgItem(hwnd, IDC_COMBO_LIGHT_TYPE);
                  for(int i=0; light_type_list[i].name; i++){
                     int indx = SendMessage(hwnd_t, CB_ADDSTRING, 0, (LPARAM)light_type_list[i].name);
                     SendMessage(hwnd_t, CB_SETITEMDATA, indx, (LPARAM)light_type_list[i].lt);
                  }
               }
               break;

            case 2:           //sound sheet
               {
                  HWND hwnd_t = GetDlgItem(hwnd, IDC_COMBO_SOUND_TYPE);
                  for(int i=0; sound_type_list[i].name; i++){
                     int indx = SendMessage(hwnd_t, CB_ADDSTRING, 0, (LPARAM)sound_type_list[i].name);
                     SendMessage(hwnd_t, CB_SETITEMDATA, indx, (LPARAM)sound_type_list[i].st);
                  }
               }
               break;

            case 6:           //light sectors sheet
            case 11:          //sound sectors sheet
               {
                  HWND hwnd_lc = NULL;
                  switch(si->sheet_index){
                  case 6:  hwnd_lc = GetDlgItem(hwnd, IDC_LIST_LSECTORS); break;
                  case 11: hwnd_lc = GetDlgItem(hwnd, IDC_LIST_SSECTORS); break;
                  }
                  SendMessage(hwnd_lc, LVM_SETEXTENDEDLISTVIEWSTYLE, 
                     LVS_EX_CHECKBOXES | LVS_EX_FULLROWSELECT, 
                     LVS_EX_CHECKBOXES | LVS_EX_FULLROWSELECT);

                  RECT rc;
                  GetClientRect(hwnd_lc, &rc);

                  LVCOLUMN lvc;
                  memset(&lvc, 0, sizeof(lvc));
                  lvc.mask = LVCF_WIDTH;
                  lvc.cx = (rc.right - rc.left);
                  SendMessage(hwnd_lc, LVM_INSERTCOLUMN, 0, (LPARAM)&lvc);
               }
               break;

            case 7:           //volume sheet
               {
                  HWND hwnd_t = GetDlgItem(hwnd, IDC_COMBO_VOLUME_TYPE);
                  for(int i=0; volume_type_list[i].name; i++){
                     int indx = SendMessage(hwnd_t, CB_ADDSTRING, 0, (LPARAM)volume_type_list[i].name);
                     SendMessage(hwnd_t, CB_SETITEMDATA, indx, (LPARAM)volume_type_list[i].vt);
                  }
               }
               break;

            case 8:           //visual sheet
               {
                  struct S_hlp{
                     static I3DENUMRET I3DAPI cbEnum(dword visual_type,
                        const char *friendly_name, dword context){

                        int i = SendDlgItemMessage((HWND)context, IDC_COMBO_VISUAL_TYPE,
                           CB_ADDSTRING, 0, (LPARAM)friendly_name);
                        SendDlgItemMessage((HWND)context, IDC_COMBO_VISUAL_TYPE,
                           CB_SETITEMDATA, i, (LPARAM)visual_type);
                        return I3DENUMRET_OK;
                     }
                  };
                  si->ep->ed->GetDriver()->EnumVisualTypes(S_hlp::cbEnum, (dword)hwnd);
               }
               break;

            }
            SetWindowLong(hwnd, GWL_USERDATA, (LPARAM)si->ep);
         }
         return 1;

      case WM_DRAWITEM:
         switch(wParam){
         case IDC_BUTTON_LIGHT_COLOR:
         //case IDC_BUTTON_LIGHT_SPECULAR:
            {                 //draw colored button
               LPDRAWITEMSTRUCT di = (LPDRAWITEMSTRUCT)lParam;
               HBRUSH hbr = CreateSolidBrush(GetWindowLong(GetDlgItem(hwnd, wParam), GWL_USERDATA));
               SelectObject(di->hDC, hbr);
               Rectangle(di->hDC, di->rcItem.left, di->rcItem.top, di->rcItem.right, di->rcItem.bottom);
               DeleteObject(hbr);
            }
            break;
         }
         break;

      case WM_VSCROLL:
         {
            C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
            HWND hwnd_sb = (HWND)lParam;
                              //we have only one SB - for visuals' props
            assert(ep->prop_controls.size() && ep->prop_controls.back()->hwnd==hwnd_sb);
            int sb_pos = ep->vis_prop_scroll_y;
            switch(LOWORD(wParam)){
            case SB_BOTTOM: sb_pos = ep->vis_prop_scroll_max_y; break;
            case SB_TOP: sb_pos = 0;
            case SB_LINEDOWN: sb_pos = Min(sb_pos+10, ep->vis_prop_scroll_max_y); break;
            case SB_LINEUP: sb_pos = Max(sb_pos-10, 0); break;
            case SB_PAGEDOWN: sb_pos = Min(sb_pos+100, ep->vis_prop_scroll_max_y); break;
            case SB_PAGEUP: sb_pos = Max(sb_pos-100, 0); break;
            case SB_THUMBTRACK: sb_pos = HIWORD(wParam); break;
            }
            if(sb_pos != ep->vis_prop_scroll_y){
               ep->vis_prop_scroll_y = sb_pos;
               ep->ScrollVisualProperties();
               SendMessage(hwnd_sb, SBM_SETPOS, sb_pos, true);
               RedrawWindow(GetParent(hwnd_sb), NULL, NULL, RDW_INVALIDATE);
               UpdateWindow(GetParent(hwnd_sb));
            }
            return 0;
         }
         break;

                              //Init:
      case WM_APP_INIT_TAB + 0:    //frame
      case WM_APP_INIT_TAB + 1:    //light
      case WM_APP_INIT_TAB + 2:    //sound
      case WM_APP_INIT_TAB + 4:    //model
      case WM_APP_INIT_TAB + 5:    //target
      case WM_APP_INIT_TAB + 6:    //lsectors
      case WM_APP_INIT_TAB + 7:    //volume
      case WM_APP_INIT_TAB + 8:    //visual
      case WM_APP_INIT_TAB + 9:    //occluder
      case WM_APP_INIT_TAB + 10:   //(currently unused)
      case WM_APP_INIT_TAB + 11:   //sound sectors
      case WM_APP_INIT_TAB + 12:   //sector
         {
            C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
            SetWindowLong(hwnd, GWL_USERDATA, 0);
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)lParam;
            int i, j;

            switch(uMsg){
            case WM_APP_INIT_TAB + 0:      //frame
               {
                  if(!sel_list.size()) break;
                  PI3D_frame frm0 = sel_list.front();
                              //name
                  if(sel_list.size()==1){
                     SetDlgItemText(hwnd, IDC_EDIT_FRM_NAME, frm0->GetName());
                              //enable for created
                     bool created = (ep->e_modify->Action(E_MODIFY_GET_FLAGS, frm0)&E_MODIFY_FLG_CREATE);
                     HWND hwnd_name = GetDlgItem(hwnd, IDC_EDIT_FRM_NAME);
                     EnableWindow(hwnd_name, created);
                  }else{
                     SetDlgItemText(hwnd, IDC_EDIT_FRM_NAME, "");
                     EnableWindow(GetDlgItem(hwnd, IDC_EDIT_FRM_NAME), false);
                  }
                              //P/R/S
                  const S_vector &pos = frm0->GetPos();
                  const S_quat &rot = frm0->GetRot();
                  const float scale = frm0->GetScale();
                  bool valid[5] = {1, 1, 1, 1, 1};
                  for(i=1; i<sel_list.size(); i++){
                     PI3D_frame frm = sel_list[i];
                     const S_vector &p1 = frm->GetPos();
                     for(j=0; j<3; j++) valid[j] = (valid[j] && pos[j]==p1[j]);
                     const S_quat &r1 = frm->GetRot();
                     valid[3] = (valid[3] && rot==r1);
                     valid[4] = (valid[4] && frm->GetScale()==scale);
                  }
                  static const int id_map_pos[] = {IDC_EDIT_FRM_POS_X, IDC_EDIT_FRM_POS_Y, IDC_EDIT_FRM_POS_Z};
                  static const int id_map_rot[] = {IDC_EDIT_FRM_ROT_X, IDC_EDIT_FRM_ROT_Y, IDC_EDIT_FRM_ROT_Z};
                  for(j=0; j<3; j++)
                     SetDlgItemText(hwnd, id_map_pos[j], !valid[j] ? "" : FloatStrip(C_fstring("%f", pos[j])));
                  S_vector axis;
                  float angle;
                  if(valid[3]) rot.Inverse(axis, angle);
                  for(j=0; j<3; j++)
                     SetDlgItemText(hwnd, id_map_rot[j], !valid[3] ? "" : FloatStrip(C_fstring("%f", axis[j])));
                  SetDlgItemText(hwnd, IDC_EDIT_FRM_ROT_A, !valid[3] ? "" : FloatStrip(C_fstring("%f", angle*180.0f/PI)));
                  SetDlgItemText(hwnd, IDC_EDIT_FRM_SCALE_U, !valid[4] ? "" : FloatStrip(C_fstring("%f", scale)));
               }
               break;

            case WM_APP_INIT_TAB + 1:   //light
               {
                  if(!sel_list.size()) break;
                  PI3D_frame frm0 = sel_list.front();
                  if(frm0->GetType()!=FRAME_LIGHT) break;
                  PI3D_light lp0 = I3DCAST_LIGHT(frm0);
                  I3D_LIGHTTYPE ltype = lp0->GetLightType();
                  S_vector color = lp0->GetColor();
                  //S_vector specular = lp0->GetSpecularColor();
                  float power = lp0->GetPower();
                  //float specular_power = lp0->GetSpecularPower();
                  //float specular_angle = lp0->GetSpecularAngle();
                  bool mode_v = lp0->GetMode()&I3DLIGHTMODE_VERTEX;
                  bool mode_lm = lp0->GetMode()&I3DLIGHTMODE_LIGHTMAP;
                  bool mode_shd = lp0->GetMode()&I3DLIGHTMODE_SHADOW;
                  bool mode_dyn = lp0->GetMode()&I3DLIGHTMODE_DYNAMIC_LM;
                  //float cone_i, cone_o;
                  //lp0->GetCone(cone_i, cone_o);
                  float range_n, range_f;
                  lp0->GetRange(range_n, range_f);
                  bool valid[14] = {
                     true,
                     true,    //1 - color
                     1, 1, 1, 1, 1, 1, 1, 
                     true,    //9 - specular
                     true,    //10 - specular power
                     true,    //11 - specular angle
                     true,    //12 - shadow
                     true,    //13 - DynLM
                  };

                  for(int i=1; i<sel_list.size(); i++){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()!=FRAME_LIGHT) continue;
                     PI3D_light lp = I3DCAST_LIGHT(frm);
                     valid[0] = (valid[0] && (ltype==lp->GetLightType()));
                     valid[1] = (valid[1] && (color==lp->GetColor()));
                     valid[2] = (valid[2] && (power==lp->GetPower()));

                     //float ci, co;
                     //lp->GetCone(ci, co);
                     float rn, rf;
                     lp->GetRange(rn, rf);
                     //valid[3] = (valid[3] && (cone_i==ci));
                     //valid[4] = (valid[4] && (cone_o==co));
                     valid[5] = (valid[5] && (range_n==rn));
                     valid[6] = (valid[6] && (range_f==rf));
                     bool mv = (lp->GetMode()&I3DLIGHTMODE_VERTEX);
                     bool mlm = (lp->GetMode()&I3DLIGHTMODE_LIGHTMAP);
                     bool mshd = (lp->GetMode()&I3DLIGHTMODE_SHADOW);
                     bool mdyn = (lp->GetMode()&I3DLIGHTMODE_DYNAMIC_LM);
                     valid[7] = (valid[7] && (mode_v == mv));
                     valid[8] = (valid[8] && (mode_lm == mlm));
                     valid[12] = (valid[12] && (mode_shd==mshd));
                     valid[13] = (valid[13] && (mode_dyn==mdyn));
                     //valid[9] = (valid[9] && (specular==lp->GetSpecularColor()));
                     //valid[10] = (valid[10] && (specular_power==lp->GetSpecularPower()));
                     //valid[11] = (valid[11] && (specular_angle==lp->GetSpecularAngle()));
                  }
                              //type
                  {
                     int lti = -1;
                     if(valid[0]){
                        for(int i=0; light_type_list[i].name; i++){
                           if(light_type_list[i].lt==ltype){
                              lti = i;
                              break;
                           }
                        }
                     }
                     SendDlgItemMessage(hwnd, IDC_COMBO_LIGHT_TYPE, CB_SETCURSEL, lti, 0);
                  }
                              //color (force redraw button)
                  SetWindowLong(GetDlgItem(hwnd, IDC_BUTTON_LIGHT_COLOR), GWL_USERDATA, 
                     !valid[1] ? GetSysColor(COLOR_BTNFACE) :
                     RGB(color[0]*255.0f, color[1]*255.0f, color[2]*255.0f));
                  ShowWindow(GetDlgItem(hwnd, IDC_BUTTON_LIGHT_COLOR), SW_HIDE);
                  ShowWindow(GetDlgItem(hwnd, IDC_BUTTON_LIGHT_COLOR), SW_SHOW);

                              //specular
                  bool specular_on = (valid[0] && ltype!=I3DLIGHT_FOG && ltype!=I3DLIGHT_NULL && ltype!=I3DLIGHT_AMBIENT);
                  if(!specular_on) valid[9] = valid[10] = valid[11] = false;
                  //EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_LIGHT_SPECULAR), specular_on);
                  //EnableWindow(GetDlgItem(hwnd, IDC_EDIT_LIGHT_SPECULAR_POWER), specular_on);
                  //EnableWindow(GetDlgItem(hwnd, IDC_EDIT_LIGHT_SPECULAR_SPECULAR_ANGLE), specular_on);
                  //SetDlgItemText(hwnd, IDC_EDIT_LIGHT_SPECULAR_POWER, !valid[10] ? "" : FloatStrip(C_fstring("%f", specular_power)));
                  //SetDlgItemText(hwnd, IDC_EDIT_LIGHT_SPECULAR_SPECULAR_ANGLE, !valid[11] ? "" : FloatStrip(C_fstring("%f", specular_angle*180.0f/PI)));
                              //specular color
                  /*
                  SetWindowLong(GetDlgItem(hwnd, IDC_BUTTON_LIGHT_SPECULAR), GWL_USERDATA, 
                     !valid[9] ? GetSysColor(COLOR_BTNFACE) :
                     RGB(specular[0]*255.0f, specular[1]*255.0f, specular[2]*255.0f));
                     */
                  //ShowWindow(GetDlgItem(hwnd, IDC_BUTTON_LIGHT_SPECULAR), SW_HIDE);
                  //ShowWindow(GetDlgItem(hwnd, IDC_BUTTON_LIGHT_SPECULAR), SW_SHOW);
                              
                              //power
                  bool power_on = (valid[0] && ltype!=I3DLIGHT_FOG && ltype!=I3DLIGHT_NULL);
                  if(!power_on) valid[2] = false;
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_LIGHT_POWER), power_on);
                  SetDlgItemText(hwnd, IDC_EDIT_LIGHT_POWER, !valid[2] ? "" : FloatStrip(C_fstring("%f", power)));
                              //cone
                  bool cone_on = (valid[0] && ltype==I3DLIGHT_SPOT);
                  if(!cone_on) valid[3] = valid[4] = false;
                  //EnableWindow(GetDlgItem(hwnd, IDC_EDIT_LIGHT_CONE_IN), cone_on);
                  //EnableWindow(GetDlgItem(hwnd, IDC_EDIT_LIGHT_CONE_OUT), cone_on);
                  //SetDlgItemText(hwnd, IDC_EDIT_LIGHT_CONE_IN, !valid[3] ? "" : FloatStrip(C_fstring("%f", cone_i*180.0f/PI)));
                  //SetDlgItemText(hwnd, IDC_EDIT_LIGHT_CONE_OUT, !valid[4] ? "" : FloatStrip(C_fstring("%f", cone_o*180.0f/PI)));
                              //range
                  bool range_on = (valid[0] && (ltype==I3DLIGHT_POINT || ltype==I3DLIGHT_SPOT
                     || ltype==I3DLIGHT_POINTAMBIENT || ltype==I3DLIGHT_FOG ||
                     //ltype==I3DLIGHT_DIRECTIONAL ||
                     ltype==I3DLIGHT_POINTFOG || ltype==I3DLIGHT_LAYEREDFOG));
                  bool range_on_n = (range_on || (ltype==I3DLIGHT_FOG));
                  if(!range_on) valid[6] = false;
                  if(!range_on && !range_on_n) valid[5] = false;
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_LIGHT_RANGE_N), range_on || range_on_n);
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_LIGHT_RANGE_F), range_on);
                  SetDlgItemText(hwnd, IDC_EDIT_LIGHT_RANGE_N, !valid[5] ? "" : FloatStrip(C_fstring("%f", range_n)));
                  SetDlgItemText(hwnd, IDC_EDIT_LIGHT_RANGE_F, !valid[6] ? "" : FloatStrip(C_fstring("%f", range_f)));
                              //mode
                  CheckDlgButton(hwnd, IDC_CHECK_LIGHT_MODE_V, !valid[7] ? BST_INDETERMINATE : mode_v ? BST_CHECKED : BST_UNCHECKED);
                  CheckDlgButton(hwnd, IDC_CHECK_LIGHT_MODE_LM, !valid[8] ? BST_INDETERMINATE : mode_lm ? BST_CHECKED : BST_UNCHECKED);
                  CheckDlgButton(hwnd, IDC_CHECK_LIGHT_MODE_SHD, !valid[12] ? BST_INDETERMINATE : mode_shd ? BST_CHECKED : BST_UNCHECKED);
                  CheckDlgButton(hwnd, IDC_CHECK_LIGHT_MODE_DYN, !valid[13] ? BST_INDETERMINATE : mode_dyn ? BST_CHECKED : BST_UNCHECKED);
               }
               break;

            case WM_APP_INIT_TAB + 2:   //sound
               {
                  if(!sel_list.size())
                     break;
                  PI3D_frame frm0 = sel_list.front();
                  if(frm0->GetType()!=FRAME_SOUND)
                     break;
                  PI3D_sound snd0 = I3DCAST_SOUND(frm0);
                  I3D_SOUNDTYPE stype = snd0->GetSoundType();
                  float vol = snd0->GetVolume();
                  float out_vol = snd0->GetOutVol();
                  bool loop = snd0->IsLoop();
                  bool strm = snd0->IsStreaming();
                  bool on = snd0->IsOn();
                  float cone_i, cone_o;
                  snd0->GetCone(cone_i, cone_o);
                  float range_n, range_f, fade;
                  snd0->GetRange(range_n, range_f, fade);
                  C_string fname = snd0->GetFileName();

                  set<PISND_source> snd_srcs;        //normal sound sources
                  dword sam_size = 0;

                  if(snd0->GetSoundSource()){
                     if(!snd0->IsStreaming())
                        snd_srcs.insert(snd0->GetSoundSource());
                     else{
                        const S_wave_format &wf = *snd0->GetSoundSource()->GetFormat();
                        sam_size += (((int)(wf.samples_per_sec*wf.bytes_per_sample*1.0f)) & -16);
                     }
                  }
                  bool valid[12] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1};

                  for(int i=1; i<sel_list.size(); i++){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()!=FRAME_SOUND) continue;
                     PI3D_sound snd = I3DCAST_SOUND(frm);
                     valid[0] = (valid[0] && (stype==snd->GetSoundType()));
                     valid[1] = (valid[1] && fname.Match(snd->GetFileName()));
                     float ci, co;
                     snd->GetCone(ci, co);
                     float rn, rf, f;
                     snd->GetRange(rn, rf, f);
                     valid[2] = (valid[2] && (fade==f));
                     valid[3] = (valid[3] && (cone_i==ci));
                     valid[4] = (valid[4] && (cone_o==co));
                     valid[5] = (valid[5] && (range_n==rn));
                     valid[6] = (valid[6] && (range_f==rf));
                     valid[7] = (valid[7] && (vol==snd->GetVolume()));
                     valid[8] = (valid[8] && (out_vol==snd->GetOutVol()));
                     valid[9] = (valid[9] && (loop==snd->IsLoop()));
                     valid[10] = (valid[10] && (on==(bool)snd->IsOn()));
                     valid[11] = (valid[11] && (strm==snd->IsStreaming()));

                     //sam_size += !snd->GetSoundSource() ? 0 : snd->GetSoundSource()->GetFormat()->size;
                     if(snd->GetSoundSource()){
                        if(!snd->IsStreaming())
                           snd_srcs.insert(snd->GetSoundSource());
                        else{
                           const S_wave_format &wf = *snd->GetSoundSource()->GetFormat();
                           sam_size += (((int)(wf.samples_per_sec*wf.bytes_per_sample*1.0f)) & -16);
                        }
                     }
                  }
                              //type
                  {
                     int sti = -1;
                     if(valid[0]){
                        for(int i=0; sound_type_list[i].name; i++){
                           if(sound_type_list[i].st==stype){
                              sti = i;
                              break;
                           }
                        }
                     }
                     SendDlgItemMessage(hwnd, IDC_COMBO_SOUND_TYPE, CB_SETCURSEL, sti, 0);
                  }

                  //SendDlgItemMessage(hwnd, IDC_COMBO_SOUND_TYPE, CB_SETCURSEL, valid[0] ? stype : -1, 0);
                              //filename
                  ep->ed->GetSoundCache().GetRelativeDirectory(&fname[0]);
                  SetDlgItemText(hwnd, IDC_EDIT_SOUND_FILE, !valid[1] ? "" : fname);

                  {           //sample size

                              //collect size of samples
                     for(set<PISND_source>::const_iterator it = snd_srcs.begin(); it!=snd_srcs.end(); ++it){
                        const S_wave_format *wf = (*it)->GetFormat();
                        if(wf)
                           sam_size += wf->size;
                     }
                     SetDlgItemInt(hwnd, IDC_EDIT_SOUND_SIZE, sam_size, true);
                  }
                              //range
                  bool range_on = (valid[0] && stype!=I3DSOUND_NULL && stype!=I3DSOUND_AMBIENT);
                  if(!range_on) valid[2] = valid[5] = valid[6] = false;
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_SOUND_RANGE_N), range_on);
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_SOUND_RANGE_F), range_on);
                  SetDlgItemText(hwnd, IDC_EDIT_SOUND_RANGE_N, !valid[5] ? "" : FloatStrip(C_fstring("%f", range_n)));
                  SetDlgItemText(hwnd, IDC_EDIT_SOUND_RANGE_F, !valid[6] ? "" : FloatStrip(C_fstring("%f", range_f)));
                              //cone
                  bool cone_on = (valid[0] && stype==I3DSOUND_SPOT);
                  if(!cone_on) valid[3] = valid[4] = false;
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_SOUND_CONE_IN), cone_on);
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_SOUND_CONE_OUT), cone_on);
                  SetDlgItemText(hwnd, IDC_EDIT_SOUND_CONE_IN, !valid[3] ? "" : FloatStrip(C_fstring("%f", cone_i*180.0f/PI)));
                  SetDlgItemText(hwnd, IDC_EDIT_SOUND_CONE_OUT, !valid[4] ? "" : FloatStrip(C_fstring("%f", cone_o*180.0f/PI)));
                              //volume
                  bool vol_on = (valid[0] && (stype!=I3DSOUND_NULL));
                  if(!vol_on) valid[7] = false;
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_SOUND_VOL), vol_on);
                  SetDlgItemText(hwnd, IDC_EDIT_SOUND_VOL, !valid[7] ? "" : FloatStrip(C_fstring("%f", vol)));
                              //out volume
                  bool out_vol_on = (valid[0] && (stype==I3DSOUND_SPOT));
                  if(!out_vol_on) valid[8] = false;
                  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_SOUND_VOL_OUT), out_vol_on);
                  SetDlgItemText(hwnd, IDC_EDIT_SOUND_VOL_OUT, !valid[8] ? "" : FloatStrip(C_fstring("%f", out_vol)));
                              //loop
                  CheckDlgButton(hwnd, IDC_CHECK_SOUND_MODE_LOOP, !valid[9] ? BST_INDETERMINATE : loop ? BST_CHECKED : BST_UNCHECKED);
                  CheckDlgButton(hwnd, IDC_CHECK_SOUND_MODE_STREAM, !valid[11] ? BST_INDETERMINATE : strm ? BST_CHECKED : BST_UNCHECKED);
                              //on
                  CheckDlgButton(hwnd, IDC_CHECK_SOUND_MODE_ON, !valid[10] ? BST_INDETERMINATE : on ? BST_CHECKED : BST_UNCHECKED);
               }
               break;

            case WM_APP_INIT_TAB + 4:   //model sheet
               {
                  if(!sel_list.size()) break;
                  PI3D_frame frm0 = sel_list.front();
                  bool no_shadow = (frm0->GetFlags()&I3D_FRMF_NOSHADOW);
                  bool shadow_cast = (frm0->GetFlags()&I3D_FRMF_SHADOW_CAST);
                  assert(frm0->GetType()==FRAME_MODEL);
                  PI3D_model mod = I3DCAST_MODEL(frm0);

                  bool valid[2] = {true, true};
                  C_string name = mod->GetFileName();

                  for(int i=1; i<sel_list.size(); i++){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()!=FRAME_MODEL) continue;

                     PI3D_model mod = I3DCAST_MODEL(frm);
                     if(!name.Match(mod->GetFileName())){
                        name = "";
                     }
                     bool no_shadow1 = (frm->GetFlags()&I3D_FRMF_NOSHADOW);
                     bool shadow_cast1 = (frm->GetFlags()&I3D_FRMF_SHADOW_CAST);
                     valid[0] = (valid[0] && (no_shadow==no_shadow1));
                     valid[1] = (valid[1] && (shadow_cast==shadow_cast1));
                  }
                  ep->ed->GetModelCache().GetRelativeDirectory(&name[0]);
                  SetDlgItemText(hwnd, IDC_EDIT_MODEL_NAME, name);
                              //set 'no shadow' checkbox
                  CheckDlgButton(hwnd, IDC_CHECK_NO_SHADOW, !valid[0] ? BST_INDETERMINATE : no_shadow ? BST_CHECKED : BST_UNCHECKED);
                  CheckDlgButton(hwnd, IDC_CHECK_SHADOW_CAST, !valid[1] ? BST_INDETERMINATE : shadow_cast ? BST_CHECKED : BST_UNCHECKED);
               }
               break;

            case WM_APP_INIT_TAB + 5:   //target sheet
               {
                  C_string name;
                  bool on = false;
                  if(sel_list.size()==1){
                     PI3D_frame frm = sel_list.front();
                     if(frm->GetType()==FRAME_TARGET){      //should always pass
                        PI3D_target tgt = I3DCAST_TARGET(frm);
                        on = true;
                        PI3D_frame tf = tgt->GetFrame();
                        if(tf) name = tf->GetName();
                     }
                  }
                  EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_TARGET_BROWSE), on);
                  SetDlgItemText(hwnd, IDC_EDIT_MODEL_NAME, name);
               }
               break;

            case WM_APP_INIT_TAB + 6:    //light's sectors
            case WM_APP_INIT_TAB + 11:   //sound's sectors
               {
                  HWND hwnd_lc = NULL;
                  I3D_FRAME_TYPE ftype;

                  switch(uMsg){
                  case WM_APP_INIT_TAB + 6:
                     hwnd_lc = GetDlgItem(hwnd, IDC_LIST_LSECTORS);
                     ftype = FRAME_LIGHT;
                     break;
                  case WM_APP_INIT_TAB + 11:
                     hwnd_lc = GetDlgItem(hwnd, IDC_LIST_SSECTORS);
                     ftype = FRAME_SOUND;
                     break;
                  }
                  assert(hwnd_lc);

                  SendMessage(hwnd_lc, LVM_DELETEALLITEMS, 0, 0);

                              //add all sectors
                  struct S_hlp{
                     static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                        vector<PI3D_sector> *sect_list = (vector<PI3D_sector>*)c;
                        sect_list->push_back(I3DCAST_SECTOR(frm));
                        return I3DENUMRET_OK;
                     }
                  };
                  vector<PI3D_sector> sect_list;
                  sect_list.push_back(ep->ed->GetScene()->GetPrimarySector());
                  sect_list.push_back(ep->ed->GetScene()->GetBackdropSector());

                  ep->ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&sect_list, ENUMF_SECTOR);
                              //use list: -1 = uninit, 0 = not in, 1 = in, 2 = indeterminate
                  vector<int> use_list(sect_list.size(), -1);

                  for(i=sel_list.size(); i--; ){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()==ftype){
                        for(int j=sect_list.size(); j--; ){
                           PI3D_sector sct = sect_list[j];
                           bool is_in;

                           switch(ftype){
                           case FRAME_LIGHT:
                              {
                                 PI3D_light lp = I3DCAST_LIGHT(frm);
                                 is_in = (1 == count(lp->GetLightSectors(), lp->GetLightSectors()+lp->NumLightSectors(), sct));
                              }
                              break;
                           case FRAME_SOUND:
                              {
                                 PI3D_sound sp = I3DCAST_SOUND(frm);
                                 is_in = (1 == count(sp->GetSoundSectors(), sp->GetSoundSectors()+sp->NumSoundSectors(), sct));
                              }
                              break;
                           }
                           if(is_in){
                              use_list[j] = (use_list[j]==-1 || use_list[j]==1) ? 1 : 2;
                           }else{
                              use_list[j] = (use_list[j]==-1 || use_list[j]==0) ? 0 : 2;
                           }
                        }
                     }
                  }

                  LVITEM li;
                  memset(&li, 0, sizeof(li));
                  li.stateMask = LVIS_STATEIMAGEMASK;

                  for(i=0; i<sect_list.size(); i++){
                     PI3D_sector sct = sect_list[i];
                     li.pszText = (char*)sct->GetName();
                     li.mask = LVIF_TEXT | LVIF_PARAM;
                     li.lParam = (LPARAM)sct;
                     int indx = SendMessage(hwnd_lc, LVM_INSERTITEM, 0, (LPARAM)&li);

                     li.mask = LVIF_STATE;
                     li.state = INDEXTOSTATEIMAGEMASK(1 + use_list[i]);
                     SendMessage(hwnd_lc, LVM_SETITEMSTATE, indx, (LPARAM)&li);
                  }
               }
               break;

            case WM_APP_INIT_TAB + 7:   //volume sheet
               {
                  if(!sel_list.size()) break;
                  PI3D_frame frm0 = sel_list.front();
                  if(frm0->GetType()!=FRAME_VOLUME) break;
                  PI3D_volume vol0 = I3DCAST_VOLUME(frm0);
                  I3D_VOLUMETYPE vtype = vol0->GetVolumeType();
                  int mat_id = vol0->GetCollisionMaterial();
                  bool col_static = (frm0->GetFlags()&I3D_FRMF_STATIC_COLLISION);
                  S_vector scl = vol0->GetNUScale();
                  bool valid[6] = {true, true, true, true, true, true};

                  bool scale_on = true;
                  switch(vtype){
                  case I3DVOLUME_SPHERE:
                     valid[1] = valid[2] = valid[3] = false;
                     scale_on = false;
                     break;
                  }

                  for(int i=1; i<sel_list.size(); i++){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()!=FRAME_VOLUME) continue;
                     PI3D_volume vol = I3DCAST_VOLUME(frm);
                     valid[0] = (valid[0] && (vtype==vol->GetVolumeType()));
                     const S_vector &s1 = vol->GetNUScale();
                     for(j=0; j<3; j++) valid[1+j] = (valid[1+j] && scl[j]==s1[j]);
                     valid[4] = (valid[4] && vol->GetCollisionMaterial()==mat_id);
                     bool col_static1 = (frm->GetFlags()&I3D_FRMF_STATIC_COLLISION);
                     valid[5] = (valid[5] && (col_static==col_static1));
                  }
                              //type
                  {
                     int vti = -1;
                     if(valid[0]){
                        for(int i=0; volume_type_list[i].name; i++){
                           if(volume_type_list[i].vt==vtype){
                              vti = i;
                              break;
                           }
                        }
                     }
                     SendDlgItemMessage(hwnd, IDC_COMBO_VOLUME_TYPE, CB_SETCURSEL, vti, 0);
                  }
                              //scale
                  static const int id_map_scl[] = {IDC_EDIT_VOL_NUSCALE_X, IDC_EDIT_VOL_NUSCALE_Y, IDC_EDIT_VOL_NUSCALE_Z};
                  for(j=0; j<3; j++){
                     EnableWindow(GetDlgItem(hwnd, id_map_scl[j]), scale_on);
                     SetDlgItemText(hwnd, id_map_scl[j], !valid[1+j] ? "" : FloatStrip(C_fstring("%f", scl[j])));
                  }
                  if(valid[4]){
                     int i = 0;
                     for(map<int, C_string>::const_iterator it=ep->mat_table.begin(); it!=ep->mat_table.end(); it++, i++)
                        if((*it).first==mat_id) break;
                     if(it==ep->mat_table.end()) i = -1;
                     SendDlgItemMessage(hwnd, IDC_COMBO_VOLUME_MAT, CB_SETCURSEL, i, 0);
                  }else{
                     SendDlgItemMessage(hwnd, IDC_COMBO_VOLUME_MAT, CB_SETCURSEL, -1, 0);
                  }
                              //set 'static collision' checkbox
                  CheckDlgButton(hwnd, IDC_CHECK_VOLUME_STATIC, !valid[5] ? BST_INDETERMINATE : col_static ? BST_CHECKED : BST_UNCHECKED);
               }
               break;

            case WM_APP_INIT_TAB + 8:   //visual
               {
                  if(!sel_list.size())
                     break;
                  int j;
                  PI3D_frame frm0 = sel_list.front();
                  if(frm0->GetType()!=FRAME_VISUAL) break;
                  PI3D_visual vis0 = I3DCAST_VISUAL(frm0);
                  dword vtype = vis0->GetVisualType();
                  int mat_id = vis0->GetCollisionMaterial();
                  bool no_shadow = (frm0->GetFlags()&I3D_FRMF_NOSHADOW);
                  bool no_col = (frm0->GetFlags()&I3D_FRMF_NOCOLLISION);
                  bool shadow_receive = (frm0->GetFlags()&I3D_FRMF_SHADOW_RECEIVE);
                  bool static_col = (frm0->GetFlags()&I3D_FRMF_STATIC_COLLISION);
                  bool valid[12] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1};

                  for(int i=1; i<sel_list.size(); i++){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()!=FRAME_VISUAL)
                        continue;
                     PI3D_visual vis = I3DCAST_VISUAL(frm);
                     valid[0] = (valid[0] && (vtype==vis->GetVisualType()));
                     bool no_shadow1 = (frm->GetFlags()&I3D_FRMF_NOSHADOW);
                     valid[7] = (valid[7] && (no_shadow==no_shadow1));
                     bool no_col1 = (frm->GetFlags()&I3D_FRMF_NOCOLLISION);
                     valid[8] = (valid[8] && (no_col==no_col1));
                     bool shadow_receive1 = (frm->GetFlags()&I3D_FRMF_SHADOW_RECEIVE);
                     valid[9] = (valid[9] && (shadow_receive==shadow_receive1));
                     valid[10] = (valid[10] && (vis->GetCollisionMaterial()==mat_id));
                     bool static_col1 = (frm->GetFlags()&I3D_FRMF_STATIC_COLLISION);;
                     valid[11] = (valid[11] && (static_col1 == static_col));
                  }
                              //hide all properties
                  ep->prop_controls.clear();
                              //type
                  if(!valid[0]){
                              //different visual types
                     SendDlgItemMessage(hwnd, IDC_COMBO_VISUAL_TYPE, CB_SETCURSEL, -1, 0);
                  }else{
                     for(j=SendDlgItemMessage(hwnd, IDC_COMBO_VISUAL_TYPE, CB_GETCOUNT, 0, 0); j--; ){
                        if(vtype==SendDlgItemMessage(hwnd, IDC_COMBO_VISUAL_TYPE, CB_GETITEMDATA, j, 0))
                           break;
                     }
                     SendDlgItemMessage(hwnd, IDC_COMBO_VISUAL_TYPE, CB_SETCURSEL, j, 0);

                     HWND hwnd_dlg = GetParent(hwnd);
                              //build property lists
                     int num_props = ep->ed->GetDriver()->NumProperties(vtype);
                     int sx = ep->vis_property_ground_size[0];
                     int sy = 10;
                     MapDlgUnits(hwnd_dlg, NULL, &sy);
                     int pos_x = 2, pos_y = 0;
                     MapDlgUnits(hwnd_dlg, &pos_x, NULL);
                     sx -= pos_x;
                     HWND hwnd_ground = GetDlgItem(hwnd, IDC_VIS_PROP_GROUP);
                     SetWindowLong(hwnd_ground, GWL_USERDATA, 0);

                     for(int pi=0; pi<num_props; pi++){
                        I3D_PROPERTYTYPE ptype = ep->ed->GetDriver()->GetPropertyType(vtype, pi);
                                          //ignore NULL
                        if(ptype==I3DPROP_NULL)
                           continue;

                        const char *pname = ep->ed->GetDriver()->GetPropertyName(vtype, pi);
                                       //determine state - check if all visuals are the same
                        dword value;
                        bool same = true;
                        for(int vi=0; vi<sel_list.size(); vi++){
                           PI3D_visual vis = I3DCAST_VISUAL(sel_list[vi]);
                           assert(vis->GetType()==FRAME_VISUAL && vis->GetVisualType()==vtype);
                           if(!vi)
                              value = vis->GetProperty(pi);
                           else
                           if(same){
                              switch(ptype){
                              case I3DPROP_STRING:
                                 same = !strcmp((const char*)vis->GetProperty(pi), (const char*)value);
                                 break;
                              case I3DPROP_BOOL:
                              case I3DPROP_INT:
                              case I3DPROP_FLOAT:
                              case I3DPROP_ENUM:
                                 same = (vis->GetProperty(pi) == value);
                                 break;
                              default:
                                 assert(0);
                              }
                           }
                        }
                                       //setup control state
                        const char *ctrl_type = NULL;
                        bool make_desc = true;
                        dword wstyle = WS_CHILD | WS_TABSTOP | WS_VISIBLE | ES_AUTOHSCROLL;
                        dword wsextyle = WS_EX_NOPARENTNOTIFY;
                        int curr_sy = sy;
                        int add_sy = sy;
                        switch(ptype){
                        case I3DPROP_BOOL:
                           ctrl_type = "BUTTON";
                           wstyle |= BS_3STATE;
                           break;
                        case I3DPROP_INT:
                        case I3DPROP_FLOAT:
                        case I3DPROP_STRING:
                           ctrl_type = "EDIT";
                           wsextyle |= WS_EX_CLIENTEDGE;
                           add_sy = curr_sy = (float)curr_sy * 1.2f;
                           break;
                        case I3DPROP_ENUM:
                           ctrl_type = "COMBOBOX";
                           wstyle |= CBS_DROPDOWNLIST;
                           add_sy = (float)curr_sy * 1.4f;
                           curr_sy = (float)curr_sy * 8.0f;
                           break;
                        }
                        if(ctrl_type){
                                       //create desired control
                           int x = pos_x, sx1 = sx;
                           if(make_desc){
                              int desc_size = (float)sx * .6f;
                              x += desc_size;
                              sx1 -= desc_size;
                           }
                           HWND h1 = CreateWindowEx(wsextyle,
                              ctrl_type,
                              NULL,
                              wstyle,
                              x, pos_y, sx1, curr_sy,
                              hwnd_ground,
                              (HMENU)(20000 + pi),
                              NULL,
                              NULL);
                           SendMessage(h1, WM_SETFONT, SendMessage(hwnd_dlg, WM_GETFONT, 0, 0), 0);
                           SetWindowLong(h1, GWL_USERDATA, (LPARAM)ptype);

                           C_property_control *ec = new C_property_control(h1, x, pos_y);
                           ep->prop_controls.push_back(ec);
                           ec->Release();

                           if(make_desc){
                              HWND h1 = CreateWindow(
                                 "STATIC",
                                 C_fstring("%s: ", pname),
                                 WS_CHILD | WS_VISIBLE | SS_RIGHT,
                                 pos_x, pos_y + 2, sx - sx1, add_sy,
                                 hwnd_ground,
                                 NULL,
                                 NULL,
                                 NULL);
                              SendMessage(h1, WM_SETFONT, SendMessage(hwnd_dlg, WM_GETFONT, 0, 0), 0);

                              C_property_control *ec = new C_property_control(h1, pos_x, pos_y);
                              ep->prop_controls.push_back(ec);
                              ec->Release();
                           }

                           pos_y += add_sy;

                                       //setup item
                           switch(ptype){
                           case I3DPROP_BOOL:
                              SendMessage(h1, BM_SETCHECK, !same ? BST_INDETERMINATE : value ? BST_CHECKED : BST_UNCHECKED, 0);
                              break;
                           case I3DPROP_INT:
                              if(same)
                                 SendMessage(h1, WM_SETTEXT, 0, (LPARAM)(const char*)C_fstring("%i", value));
                              break;
                           case I3DPROP_FLOAT:
                              if(same)
                                 SendMessage(h1, WM_SETTEXT, 0, (LPARAM)(const char*)FloatStrip(C_fstring("%f", *(float*)&value)));
                              break;
                           case I3DPROP_ENUM:
                              {
                                          //fill-in contents of box
                                 int num_items = 0;
                                 for(const char *cp=pname; cp += strlen(cp) + 1, *cp; ++num_items){
                                    SendMessage(h1, CB_ADDSTRING, 0, (LPARAM)cp);
                                 }
                                 if(same && value<num_items)
                                    SendMessage(h1, CB_SETCURSEL, value, 0);
                              }
                              break;
                           case I3DPROP_STRING:
                              if(same && value)
                                 SendMessage(h1, WM_SETTEXT, 0, (LPARAM)(const char*)value);
                              break;
                           default:
                              assert(0);
                           }
                        }
                     }
                              //now check if scrollbar is needed
                     if(pos_y >= ep->vis_property_ground_size[1]){
                        int sb_size = 8;
                        MapDlgUnits(hwnd_dlg, &sb_size, NULL);

                        HWND hwnd_sb = CreateWindow(
                           "SCROLLBAR",
                           NULL,
                           WS_CHILD | WS_VISIBLE | SBS_VERT,
                           ep->vis_property_ground_pos[0] + ep->vis_property_ground_size[0],
                           ep->vis_property_ground_pos[1],
                           10, ep->vis_property_ground_size[1],
                           hwnd,
                           NULL,
                           NULL,
                           NULL);
                        SendMessage(hwnd_sb, WM_SETFONT, SendMessage(hwnd_dlg, WM_GETFONT, 0, 0), 0);

                        C_property_control *ec = new C_property_control(hwnd_sb, 0x80000000, 0x80000000);
                        ep->prop_controls.push_back(ec);
                        ec->Release();

                        EnableWindow(hwnd_sb, true);
                        ep->vis_prop_scroll_max_y = pos_y - ep->vis_property_ground_size[1];
                              //retain previous scroll position
                        ep->vis_prop_scroll_y = Min(ep->vis_prop_scroll_y, ep->vis_prop_scroll_max_y);
                        ep->ScrollVisualProperties();
                        SCROLLINFO si;
                        si.cbSize = sizeof(si);
                        si.fMask = SIF_DISABLENOSCROLL | SIF_PAGE | SIF_POS | SIF_RANGE;
                        si.nMin = 0;
                        si.nMax = pos_y;
                        si.nPage = ep->vis_property_ground_size[1];
                        si.nPos = ep->vis_prop_scroll_y;
                        SendMessage(hwnd_sb, SBM_SETSCROLLINFO, true, (LPARAM)&si);
                     }else{
                        ep->vis_prop_scroll_y = 0;
                     }
                     SetWindowLong(hwnd_ground, GWL_USERDATA, (LPARAM)ep);
                  }
                              //set 'no shadow' checkbox
                  CheckDlgButton(hwnd, IDC_CHECK_NO_SHADOW, !valid[7] ? BST_INDETERMINATE : no_shadow ? BST_CHECKED : BST_UNCHECKED);
                              //set 'no collision' checkbox
                  CheckDlgButton(hwnd, IDC_CHECK_NO_COLL, !valid[8] ? BST_INDETERMINATE : no_col ? BST_CHECKED : BST_UNCHECKED);
                              //set 'receive shadow' checkbox
                  CheckDlgButton(hwnd, IDC_CHECK_SHADOW_REC, !valid[9] ? BST_INDETERMINATE : shadow_receive ? BST_CHECKED : BST_UNCHECKED);
                              //set 'static collision' checkbox
                  CheckDlgButton(hwnd, IDC_CHECK_VOLUME_STATIC, !valid[11] ? BST_INDETERMINATE : static_col ? BST_CHECKED : BST_UNCHECKED);
                              //material id
                  EnableWindow(GetDlgItem(hwnd, IDC_COMBO_VISUAL_MAT), static_col);

                  if(valid[10]){
                     int i = 0;
                     for(map<int, C_string>::const_iterator it=ep->mat_table.begin(); it!=ep->mat_table.end(); it++, i++)
                        if((*it).first==mat_id) break;
                     if(it==ep->mat_table.end()) i = -1;
                     SendDlgItemMessage(hwnd, IDC_COMBO_VISUAL_MAT, CB_SETCURSEL, i, 0);
                  }else{
                     SendDlgItemMessage(hwnd, IDC_COMBO_VISUAL_MAT, CB_SETCURSEL, -1, 0);
                  }
               }
               break;

            case WM_APP_INIT_TAB + 12:    //sector
               {
                  if(!sel_list.size())
                     break;
                  PI3D_frame frm0 = sel_list.front();
                  if(frm0->GetType()!=FRAME_SECTOR)
                     break;
                  PI3D_sector sct0 = I3DCAST_SECTOR(frm0);
                  dword env_id = sct0->GetEnvironmentID();
                  bool valid[1] = {1};

                  for(int i=1; i<sel_list.size(); i++){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()!=FRAME_SECTOR)
                        continue;
                     PI3D_sector sct = I3DCAST_SECTOR(frm);
                     valid[0] = (valid[0] && (env_id==sct->GetEnvironmentID()));
                  }
                  if(valid[0]){
                     int i = 0;
                     for(map<int, C_string>::const_iterator it=ep->env_table.begin(); it!=ep->env_table.end(); it++, i++)
                        if((*it).first==env_id)
                           break;
                     if(it==ep->env_table.end()) i = -1;
                     SendDlgItemMessage(hwnd, IDC_COMBO_SCT_ENV, CB_SETCURSEL, i, 0);
                  }else{
                     SendDlgItemMessage(hwnd, IDC_COMBO_SCT_ENV, CB_SETCURSEL, -1, 0);
                  }
               }
               break;
            }
            SetWindowLong(hwnd, GWL_USERDATA, (LPARAM)ep);
         }
         break;

      case WM_NOTIFY:
         switch(wParam){
         case IDC_LIST_LSECTORS:
         case IDC_LIST_SSECTORS:
            {
               I3D_FRAME_TYPE ftype;
               dword modify_flags;
               switch(wParam){
               case IDC_LIST_LSECTORS:
                  ftype = FRAME_LIGHT;
                  modify_flags = E_MODIFY_FLG_LIGHT;
                  break;
               case IDC_LIST_SSECTORS:
                  ftype = FRAME_SOUND;
                  modify_flags = E_MODIFY_FLG_SOUND;
                  break;
               }

               LPNMHDR hdr = (LPNMHDR)lParam;
               switch(hdr->code){
               case LVN_ITEMCHANGED:
                  {
                     C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                     if(!ep)
                        break;
                     if(!ep->ed->CanModify()) break;
                     ep->in_sheet_modify = true;
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                     LPNMLISTVIEW lvi = (LPNMLISTVIEW)hdr;
                     if(lvi->uChanged&LVIF_STATE){
                        bool was_on = ((lvi->uOldState&LVIS_STATEIMAGEMASK)==INDEXTOSTATEIMAGEMASK(2));
                        bool on = ((lvi->uNewState&LVIS_STATEIMAGEMASK)==INDEXTOSTATEIMAGEMASK(2));
                        if(was_on!=on){
                           switch(ftype){
                           case FRAME_LIGHT:
                              ep->SaveUndoLights(sel_list);
                              break;
                           case FRAME_SOUND:
                              ep->SaveUndoSounds(sel_list);
                              break;
                           }

                           PI3D_sector sct = (PI3D_sector)lvi->lParam;
                           ep->e_slct->Action(E_SELECTION_FLASH_FRAME, sct);

                           pair<PI3D_frame, dword> p(NULL, modify_flags);
                           for(int i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()==ftype){
                                 switch(ftype){
                                 case FRAME_LIGHT:
                                    {
                                       PI3D_light lp = I3DCAST_LIGHT(frm);
                                       if(on) sct->AddLight(lp);
                                       else sct->DeleteLight(lp);
                                    }
                                    break;
                                 case FRAME_SOUND:
                                    {
                                       PI3D_sound sp = I3DCAST_SOUND(frm);
                                       if(on) sct->AddSound(sp);
                                       else sct->DeleteSound(sp);
                                    }
                                    break;
                                 }

                                 p.first = frm;
                                 ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                                 ep->ed->SetModified();
                              }
                           }
                        }
                     }
                     ep->in_sheet_modify = false;
                  }
                  break;

               case NM_DBLCLK:
                  {
                     LPNMLISTVIEW nmlv = (LPNMLISTVIEW)lParam;
                     LVITEM lvi;
                     lvi.mask = LVIF_STATE;
                     lvi.stateMask = LVIS_STATEIMAGEMASK;
                     lvi.iItem = nmlv->iItem;
                     lvi.iSubItem = nmlv->iSubItem;
                     SendDlgItemMessage(hwnd, wParam, LVM_GETITEM, 0, (LPARAM)&lvi);
                     bool was_on = ((lvi.state&LVIS_STATEIMAGEMASK)==INDEXTOSTATEIMAGEMASK(2));
                     lvi.state &= ~LVIS_STATEIMAGEMASK; 
                     lvi.state = INDEXTOSTATEIMAGEMASK(was_on ? 1 : 2);
                     SendDlgItemMessage(hwnd, wParam, LVM_SETITEM, 0, (LPARAM)&lvi);
                  }
                  break;
               }
            }
            break;
         }
         break;
      
      case WM_COMMAND:
         switch(HIWORD(wParam)){
         case EN_CHANGE:
            {
                              //get current selection
               C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
               if(!ep) break; 
               if(!ep->ed->CanModify()) break;
               const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
               if(!sel_list.size()) break;
               ep->in_sheet_modify = true;

               switch(LOWORD(wParam)){
               case IDC_EDIT_FRM_NAME:
                  {
                     PI3D_frame frm0 = sel_list.front();
                     C_string new_name = GetDlgItemText(hwnd, LOWORD(wParam));
                     if(new_name != frm0->GetName() && !IsNameInScene(ep->ed->GetScene(), new_name))
                        ep->RenameFrame(frm0, new_name);
                  }
                  break;

               case IDC_EDIT_FRM_POS_X: case IDC_EDIT_FRM_POS_Y: case IDC_EDIT_FRM_POS_Z:
               case IDC_EDIT_FRM_ROT_X: case IDC_EDIT_FRM_ROT_Y: case IDC_EDIT_FRM_ROT_Z: case IDC_EDIT_FRM_ROT_A:
               case IDC_EDIT_FRM_SCALE_U:
               case IDC_EDIT_VOL_NUSCALE_X: case IDC_EDIT_VOL_NUSCALE_Y: case IDC_EDIT_VOL_NUSCALE_Z:
                  {
                     C_string num = GetDlgItemText(hwnd, LOWORD(wParam));
                     float f;
                     int i;
                     i = sscanf(num, "%f", &f);
                     if(i==1){
                              //modify selection
                        vector<byte> undo_info;

                        dword undo_action;
                        switch(LOWORD(wParam)){
                        case IDC_EDIT_FRM_POS_X: case IDC_EDIT_FRM_POS_Y: case IDC_EDIT_FRM_POS_Z:
                           undo_action = E_MODIFY_TRANSLATE_NAMED;
                           break;
                        case IDC_EDIT_FRM_ROT_X: case IDC_EDIT_FRM_ROT_Y: case IDC_EDIT_FRM_ROT_Z: case IDC_EDIT_FRM_ROT_A:
                           undo_action = E_MODIFY_ROTATE_NAMED;
                           break;
                        case IDC_EDIT_FRM_SCALE_U:
                        case IDC_EDIT_VOL_NUSCALE_X: case IDC_EDIT_VOL_NUSCALE_Y: case IDC_EDIT_VOL_NUSCALE_Z:
                           undo_action = E_MODIFY_SCALE_NAMED;
                           break;
                        }

                        const char *top_name = (const char*)ep->e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
                        bool undo_on = (!top_name || strcmp(top_name, "Modify") ||
                           ep->e_undo->Action(E_UNDO_GET_TOP_ACTION)!=undo_action);
                        if(undo_on) for(i=0; i<4; i++) undo_info.push_back(0);

                        pair<PI3D_frame, dword> p;

                        for(int ii=0; ii<sel_list.size(); ii++){
                           p.second = 0;

                           PI3D_frame frm = sel_list[ii];

                           p.first = frm;

                           switch(LOWORD(wParam)){
                           case IDC_EDIT_FRM_POS_X: case IDC_EDIT_FRM_POS_Y: case IDC_EDIT_FRM_POS_Z:
                              {
                                 S_vector pos = frm->GetPos();
                                             //save undo
                                 if(undo_on){
                                    int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_pos);
                                    for(i=0; i<len; i++) undo_info.push_back(0);
                                    S_edit_Modify_undo_pos::S_frm &ui = *(S_edit_Modify_undo_pos::S_frm*)(undo_info.end() - len);
                                    ui.pos = pos;
                                    strcpy(ui.frm_name, frm->GetName());
                                    S_edit_Modify_undo_pos &upi = *(S_edit_Modify_undo_pos*)&undo_info.front();
                                    ++upi.num_frames;
                                 }
                                             //modify position
                                 switch(LOWORD(wParam)){
                                 case IDC_EDIT_FRM_POS_X: pos.x = f; break;
                                 case IDC_EDIT_FRM_POS_Y: pos.y = f; break;
                                 default: pos.z = f; break;
                                 }

                                 p.second = E_MODIFY_FLG_POSITION;
                                 ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                                 frm->SetPos(pos);
                                 frm->Update();
                              }
                              break;

                           case IDC_EDIT_FRM_ROT_X: case IDC_EDIT_FRM_ROT_Y: case IDC_EDIT_FRM_ROT_Z: case IDC_EDIT_FRM_ROT_A:
                              {
                                                //get all values
                                 C_string num;
                                 S_vector axis;
                                 float angle;

                                 num = GetDlgItemText(hwnd, IDC_EDIT_FRM_ROT_X);
                                 i = sscanf(num, "%f", &axis.x);
                                 if(i==1){
                                    num = GetDlgItemText(hwnd, IDC_EDIT_FRM_ROT_Y);
                                    i = sscanf(num, "%f", &axis.y);
                                    if(i==1){
                                       num = GetDlgItemText(hwnd, IDC_EDIT_FRM_ROT_Z);
                                       i = sscanf(num, "%f", &axis.z);
                                       if(i==1){
                                          num = GetDlgItemText(hwnd, IDC_EDIT_FRM_ROT_A);
                                          i = sscanf(num, "%f", &angle);
                                          if(i==1){
                                             angle *= PI/180.0f;

                                             S_quat rot;
                                             rot.Make(axis, angle);
                                                         //save undo info
                                             if(undo_on){
                                                int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_rot::S_frm);
                                                for(int i=0; i<len; i++) undo_info.push_back(0);
                                                S_edit_Modify_undo_rot::S_frm &ui = *(S_edit_Modify_undo_rot::S_frm*)(undo_info.end() - len);
                                                ui.rot = frm->GetRot();
                                                strcpy(ui.frm_name, frm->GetName());
                                                S_edit_Modify_undo_rot &upi = *(S_edit_Modify_undo_rot*)&undo_info.front();
                                                ++upi.num_frames;
                                             }

                                             p.second = E_MODIFY_FLG_ROTATION;
                                             ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);
                                                         //modify rotation
                                             frm->SetRot(rot);
                                             frm->Update();

                                             
                                          }
                                       }
                                    }
                                 }
                              }
                              break;

                           case IDC_EDIT_VOL_NUSCALE_X: case IDC_EDIT_VOL_NUSCALE_Y: case IDC_EDIT_VOL_NUSCALE_Z:
                              if(frm->GetType()==FRAME_VOLUME) goto ok1;
                              break;

                           case IDC_EDIT_FRM_SCALE_U:
                        ok1:
                              {
                                 bool uni_scale = (LOWORD(wParam)==IDC_EDIT_FRM_SCALE_U);
                                 if(undo_on){
                                    int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_scl::S_frm);
                                    for(int i=0; i<len; i++) undo_info.push_back(0);
                                    S_edit_Modify_undo_scl::S_frm &ui = *(S_edit_Modify_undo_scl::S_frm*)(undo_info.end() - len);
                                    if(uni_scale) ui.scl[0] = frm->GetScale();
                                    else{
                                       switch(frm->GetType()){
                                       //case FRAME_VISUAL: ui.scl = I3DCAST_VISUAL(frm)->GetNUScale(); break;
                                       case FRAME_VOLUME: ui.scl = I3DCAST_VOLUME(frm)->GetNUScale(); break;
                                       }
                                    }
                                    ui.use_non_uniform = !uni_scale;
                                    strcpy(ui.frm_name, frm->GetName());
                                    S_edit_Modify_undo_scl &upi = *(S_edit_Modify_undo_scl*)&undo_info.front();
                                    ++upi.num_frames;
                                 }
                                 if(uni_scale){
                                    p.second = E_MODIFY_FLG_SCALE;
                                    ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                                    frm->SetScale(f);
                                 }else{
                                    p.second = E_MODIFY_FLG_NU_SCALE;
                                    ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                                    S_vector scl;
                                    switch(frm->GetType()){
                                    case FRAME_VOLUME: 
                                       scl = I3DCAST_VOLUME(frm)->GetNUScale(); 
                                       switch(LOWORD(wParam)){
                                       case IDC_EDIT_VOL_NUSCALE_X: scl.x = f; break;
                                       case IDC_EDIT_VOL_NUSCALE_Y: scl.y = f; break;
                                       default: scl.z = f; break;
                                       }
                                       I3DCAST_VOLUME(frm)->SetNUScale(scl); 
                                       break;
                                    }
                                 }
                                 frm->Update();
                              }
                              break;

                           }
                           if(p.second) ep->ed->SetModified();
                        }
                        if(undo_on){
                           dword num = *(dword*)&undo_info.front();
                           if(num){
                                             //save undo info
                              S_edit_Undo_2 eu("Modify", undo_action, undo_info.begin(), undo_info.size());
                              ep->e_undo->Action(E_UNDO_SAVE, &eu);
                           }
                        }
                     }
                  }
                  break;

               case IDC_EDIT_LIGHT_POWER: case IDC_EDIT_LIGHT_RANGE_N: case IDC_EDIT_LIGHT_RANGE_F:
               //case IDC_EDIT_LIGHT_CONE_IN: case IDC_EDIT_LIGHT_CONE_OUT:
               //case IDC_EDIT_LIGHT_SPECULAR_POWER: //case IDC_EDIT_LIGHT_SPECULAR_SPECULAR_ANGLE:
                  {
                     C_string num = GetDlgItemText(hwnd, LOWORD(wParam));
                     float f;
                     int i;
                     i = sscanf(num, "%f", &f);
                     if(i==1){
                        ep->SaveUndoLights(sel_list);
                        for(i=sel_list.size(); i--; ){
                           PI3D_frame frm = sel_list[i];
                           if(frm->GetType()!=FRAME_LIGHT) continue;
                           PI3D_light lp = I3DCAST_LIGHT(frm);
                           switch(LOWORD(wParam)){
                           case IDC_EDIT_LIGHT_POWER: lp->SetPower(f); break;
                           //case IDC_EDIT_LIGHT_SPECULAR_POWER: lp->SetSpecularPower(f); break;
                           //case IDC_EDIT_LIGHT_SPECULAR_SPECULAR_ANGLE: lp->SetSpecularAngle(f*PI/180.0f); break;
                              /*
                           case IDC_EDIT_LIGHT_CONE_IN:
                              {
                                 float c[2];
                                 lp->GetCone(c[0], c[1]);
                                 lp->SetCone(f*PI/180.0f, c[1]);
                              }
                              break;
                           case IDC_EDIT_LIGHT_CONE_OUT:
                              {
                                 float c[2];
                                 lp->GetCone(c[0], c[1]);
                                 lp->SetCone(c[0], f*PI/180.0f);
                              }
                              break;
                              */
                           case IDC_EDIT_LIGHT_RANGE_N:
                              {
                                 float r[2];
                                 lp->GetRange(r[0], r[1]);
                                 lp->SetRange(f, r[1]);
                              }
                              break;
                           case IDC_EDIT_LIGHT_RANGE_F:
                              {
                                 float r[2];
                                 lp->GetRange(r[0], r[1]);
                                 lp->SetRange(r[0], f);
                              }
                              break;
                           }
                           lp->Update();

                           pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_LIGHT);
                           ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                           ep->ed->SetModified();
                        }
                     }
                  }
                  break;

               case IDC_EDIT_SOUND_RANGE_N: case IDC_EDIT_SOUND_RANGE_F:
               case IDC_EDIT_SOUND_CONE_IN: case IDC_EDIT_SOUND_CONE_OUT:
               case IDC_EDIT_SOUND_VOL: case IDC_EDIT_SOUND_VOL_OUT:
                  {
                     C_string num = GetDlgItemText(hwnd, LOWORD(wParam));
                     float f;
                     int i;
                     i = sscanf(num, "%f", &f);
                     if(i==1){
                        ep->SaveUndoSounds(sel_list);
                        for(i=sel_list.size(); i--; ){
                           PI3D_frame frm = sel_list[i];
                           if(frm->GetType()!=FRAME_SOUND) continue;
                           PI3D_sound sp = I3DCAST_SOUND(frm);
                           switch(LOWORD(wParam)){
                           case IDC_EDIT_SOUND_VOL: sp->SetVolume(f); break;
                           case IDC_EDIT_SOUND_VOL_OUT: sp->SetOutVol(f); break;
                           case IDC_EDIT_SOUND_CONE_IN:
                              {
                                 float c[2];
                                 sp->GetCone(c[0], c[1]);
                                 sp->SetCone(f*PI/180.0f, c[1]);
                              }
                              break;
                           case IDC_EDIT_SOUND_CONE_OUT:
                              {
                                 float c[2];
                                 sp->GetCone(c[0], c[1]);
                                 sp->SetCone(c[0], f*PI/180.0f);
                              }
                              break;
                           case IDC_EDIT_SOUND_RANGE_N:
                              {
                                 float r[3];
                                 sp->GetRange(r[0], r[1], r[2]);
                                 sp->SetRange(f, r[1], r[2]);
                              }
                              break;
                           case IDC_EDIT_SOUND_RANGE_F:
                              {
                                 float r[3];
                                 sp->GetRange(r[0], r[1], r[2]);
                                 sp->SetRange(r[0], f, r[2]);
                              }
                              break;
                           }
                           sp->Update();

                           pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_SOUND);
                           ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                           ep->ed->SetModified();
                        }
                     }
                  }
                  break;
               default:
                  {
                     if(ep->curr_sheet_index!=8)
                        break;
                              //this may be one of visual property
                     int id = LOWORD(wParam);
                     HWND hwnd_ctrl = (HWND)lParam;
                              //make sure this is one of expected controls
                     for(int i=ep->prop_controls.size(); i--; ){
                        if(ep->prop_controls[i]->hwnd==hwnd_ctrl)
                           break;
                     }
                     assert(i!=-1);
                     int prop_index = id - 20000;

                     C_string str = GetDlgItemText(hwnd, id);
                     dword value = 0;
                              //get type of fiels
                     I3D_PROPERTYTYPE ptype = (I3D_PROPERTYTYPE)GetWindowLong(hwnd_ctrl, GWL_USERDATA);
                     switch(ptype){
                     case I3DPROP_INT:
                        sscanf(str, "%i", &value);
                        break;
                     case I3DPROP_FLOAT:
                        sscanf(str, "%f", &value);
                        break;
                     case I3DPROP_STRING:
                        value = (dword)(const char*)str;
                        break;
                     default:
                        assert(0);
                     }

                     ep->SaveUndoVisuals(sel_list);
                     ep->ModifyVisualsProperty(sel_list, prop_index,
                        value);
                  }
               }
               ep->in_sheet_modify = false;
            }
            break;

         case CBN_SELCHANGE:
            {
               C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
               if(!ep) break;
               if(!ep->ed->CanModify()) break;
               ep->in_sheet_modify = true;
               const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);

               switch(LOWORD(wParam)){
               case IDC_COMBO_LIGHT_TYPE:
               case IDC_COMBO_SOUND_TYPE:
               case IDC_COMBO_VOLUME_TYPE:
               case IDC_COMBO_VISUAL_TYPE:
               case IDC_COMBO_VOLUME_MAT:
               case IDC_COMBO_VISUAL_MAT:
               case IDC_COMBO_SCT_ENV:
                  {
                     switch(LOWORD(wParam)){

                     case IDC_COMBO_LIGHT_TYPE:
                        {
                           ep->SaveUndoLights(sel_list);
                           int lti = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
                           assert(lti!=-1);
                           I3D_LIGHTTYPE ltype = (I3D_LIGHTTYPE)SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETITEMDATA, lti, 0);
                           for(int i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()!=FRAME_LIGHT) continue;
                              PI3D_light lp = I3DCAST_LIGHT(frm);
                              lp->SetLightType(ltype);

                              pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_LIGHT);
                              ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                              ep->ed->SetModified();
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 1, 0, (LPARAM)&sel_list);
                        }
                        break;

                     case IDC_COMBO_SOUND_TYPE:
                        {
                           ep->SaveUndoSounds(sel_list);
                           int sti = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
                           assert(sti!=-1);
                           I3D_SOUNDTYPE stype = (I3D_SOUNDTYPE)SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETITEMDATA, sti, 0);

                           for(int i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()!=FRAME_SOUND) continue;
                              PI3D_sound sp = I3DCAST_SOUND(frm);
                              sp->SetSoundType(stype);

                              pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_SOUND);
                              ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                              ep->ed->SetModified();
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 2, 0, (LPARAM)&sel_list);
                        }
                        break;

                     case IDC_COMBO_VOLUME_TYPE:
                        {
                           ep->SaveUndoVolumes(sel_list);
                           int vti = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
                           assert(vti!=-1);
                           I3D_VOLUMETYPE vtype = (I3D_VOLUMETYPE)SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETITEMDATA, vti, 0);

                           for(int i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()!=FRAME_VOLUME) continue;
                              PI3D_volume vp = I3DCAST_VOLUME(frm);
                              vp->SetVolumeType(vtype);

                              pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_VOLUME);
                              ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                              ep->ed->SetModified();
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 7, 0, (LPARAM)&sel_list);
                        }
                        break;

                     case IDC_COMBO_VOLUME_MAT:
                        {
                           ep->SaveUndoVolumes(sel_list);
                           int i = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
                           int mat_id = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETITEMDATA, i, 0);
                           for(i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()!=FRAME_VOLUME) continue;
                              PI3D_volume vp = I3DCAST_VOLUME(frm);
                              vp->SetCollisionMaterial(mat_id);

                              pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_VOLUME);
                              ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                              ep->ed->SetModified();
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 7, 0, (LPARAM)&sel_list);
                        }
                        break;

                     case IDC_COMBO_VISUAL_TYPE:
                        {
                           int i;
                           dword vtype = SendDlgItemMessage(hwnd, LOWORD(wParam),
                              CB_GETITEMDATA, SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0), 0);
                                          //issue warning before casting into non-geometry types
                           static dword known_geom_vistypes[] = {
                              I3D_VISUAL_OBJECT, I3D_VISUAL_LIT_OBJECT, I3D_VISUAL_SINGLEMESH, I3D_VISUAL_BILLBOARD,
                              I3D_VISUAL_MIRROR, I3D_VISUAL_MORPH, I3D_VISUAL_DYNAMIC, I3D_VISUAL_UV_SHIFT,
                              I3D_VISUAL_CAMVIEW, I3D_VISUAL_MIRROR,
                              0
                           };
                           for(i=0; known_geom_vistypes[i]; i++){
                              if(vtype==known_geom_vistypes[i])
                                 break;
                           }
                           bool do_cast = true;
                           if(!known_geom_vistypes[i]){
                              do_cast = (MessageBox(hwnd,
                                 "You selected to cast to non-geometry visual type.\n"
                                 "Undo information won't contain geometry information.\n"
                                 "Are you sure to continue?",
                                 "Casting visual",
                                 MB_YESNOCANCEL)==IDYES);
                           }
                           if(do_cast){

                              ep->SaveUndoVisuals(sel_list);
                              vector<PI3D_frame> save_list = sel_list;
                              ep->e_slct->Action(E_SELECTION_CLEAR);

                              for(i=save_list.size(); i--; ){
                                 PI3D_frame frm = save_list[i];
                                 if(frm->GetType()!=FRAME_VISUAL) continue;
                                 PI3D_visual vis = I3DCAST_VISUAL(frm);

                                 if(vis->GetVisualType()==vtype) continue;

                                             //perform desired cast
                                 CastVisual(ep->ed, vis, vtype, ep->e_modify, ep->e_slct);
                              }
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 8, 0, (LPARAM)&sel_list);
                        }
                        break;

                     case IDC_COMBO_VISUAL_MAT:
                        {
                           ep->SaveUndoVisuals(sel_list);
                           int i = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
                           int mat_id = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETITEMDATA, i, 0);
                           for(i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()!=FRAME_VISUAL) continue;
                              PI3D_visual vis = I3DCAST_VISUAL(frm);
                              vis->SetCollisionMaterial(mat_id);

                              pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_VISUAL);
                              ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                              ep->ed->SetModified();
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 8, 0, (LPARAM)&sel_list);
                        }
                        break;

                     case IDC_COMBO_SCT_ENV:
                        {
                           ep->SaveUndoSectors(sel_list);
                           int i = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETCURSEL, 0, 0);
                           int env_id = SendDlgItemMessage(hwnd, LOWORD(wParam), CB_GETITEMDATA, i, 0);
                           for(i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()!=FRAME_SECTOR)
                                 continue;
                              PI3D_sector sct = I3DCAST_SECTOR(frm);
                              sct->SetEnvironmentID(env_id);

                              pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_SECTOR);
                              ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                              ep->ed->SetModified();
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 12, 0, (LPARAM)&sel_list);
                        }
                        break;
                     }
                  }
                  break;
               default:
                  {
                     if(ep->curr_sheet_index!=8)
                        break;
                              //this may be one of visual property
                     int id = LOWORD(wParam);
                     HWND hwnd_ctrl = (HWND)lParam;
                              //make sure this is one of expected controls
                     for(int i=ep->prop_controls.size(); i--; ){
                        if(ep->prop_controls[i]->hwnd==hwnd_ctrl)
                           break;
                     }
                     assert(i!=-1);
                     int prop_index = id - 20000;

                     ep->SaveUndoVisuals(sel_list);
                     ep->ModifyVisualsProperty(sel_list, prop_index,
                        SendDlgItemMessage(hwnd, id, CB_GETCURSEL, 0, 0));
                  }
               }
               ep->in_sheet_modify = false;
            }
            break;

         case BN_CLICKED:
            {
               C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
               if(!ep) break; //no modifications if it's not user's action
               if(!ep->ed->CanModify()) break;

               switch(LOWORD(wParam)){
               case IDC_BUTTON_MODEL_BROWSE:
                  {
                     LPC_editor_item e_create = ep->ed->FindPlugin("Create");
                     if(e_create){
                        char buf[256];
                        if(e_create->Action(E_CREATE_BROWSE_MODEL, buf)){
                           C_string str = buf;
                           const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                                                   //change all models
                           for(int i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()==FRAME_MODEL){
                                 PI3D_model mod = I3DCAST_MODEL(frm);
                                 ep->ReplaceModel(mod, str);
                              }
                           }
                           SendMessage(hwnd, WM_APP_INIT_TAB + 4, 0, (LPARAM)&sel_list);
                        }
                     }
                  }
                  break;

               case IDC_BUTTON_TARGET_BROWSE:
               case IDC_BUTTON_LIGHT_SAME_AS:
               case IDC_BUTTON_MODEL_SAME_AS:
               case IDC_BUTTON_VISUAL_SAME_AS:
                  {
                     S_MouseEdit_pick p = {ep, 0, NULL };
                     switch(LOWORD(wParam)){
                     case IDC_BUTTON_TARGET_BROWSE: p.action_id = E_PROP_NOTIFY_PICK_TARGET; break;
                     case IDC_BUTTON_LIGHT_SAME_AS: p.action_id = E_PROP_NOTIFY_PICK_LIGHT; break;
                     case IDC_BUTTON_MODEL_SAME_AS: p.action_id = E_PROP_NOTIFY_PICK_MODEL; break;
                     case IDC_BUTTON_VISUAL_SAME_AS: p.action_id = E_PROP_NOTIFY_PICK_VISUAL; break;
                     default: assert(0);
                     }
                     ep->e_mouseedit->Action(E_MOUSE_SET_USER_PICK, &p);
                  }
                  break;

               case IDC_BUTTON_VISUAL_SAME_AS2:
                  {
                     vector<PI3D_frame> frms;
                     struct S_hlp{
                        static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                           ((vector<PI3D_frame>*)c)->push_back(frm);
                           return I3DENUMRET_OK;
                        }
                     };
                     ep->ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&frms, ENUMF_VISUAL);
                     if(ep->e_slct->Action(E_SELECTION_PROMPT, &frms)){
                        if(frms.size()==1){
                           ep->CopyProperties(I3DCAST_VISUAL(frms.front()));
                        }else{
                           ep->ed->Message("Failed (single visual must be selected)");
                        }
                     }
                  }
                  break;

               case IDC_BUTTON_MODEL_RELOAD:
                  {
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                     int num_ok = 0;
                     for(int i=sel_list.size(); i--; ){
                        PI3D_frame frm = sel_list[i];
                        if(frm->GetType()==FRAME_MODEL){
                           PI3D_model mod = I3DCAST_MODEL(frm);
                           C_string fname = mod->GetFileName();
                           ep->ed->GetModelCache().GetRelativeDirectory(&fname[0]);
                           for(int i=strlen(fname); i--; ){
                              if(fname[i]=='.'){
                                 fname[i] = 0;
                                 break;
                              }
                           }
                           ep->ed->GetModelCache().Open(mod, fname, ep->ed->GetScene(), 0, NULL, 0);
                           ep->e_modify->Action(E_MODIFY_DEL_MODEL_DEL_CHILD,
                              &pair<PI3D_frame, const char*>(mod, NULL));
                           ++num_ok;
                        }
                     }
                     ep->ed->Message(C_fstring("Models reloaded: %i", num_ok));
                  }
                  break;

               case IDC_BUTTON_SOUND_BROWSE:
                  {
                     LPC_editor_item e_create = ep->ed->FindPlugin("Create");
                     if(e_create){
                        char buf[256];
                        if(e_create->Action(E_CREATE_BROWSE_SOUND, buf)){
                           C_string str = buf;
                           const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                           ep->SaveUndoSounds(sel_list);
                                       //change all 
                           for(int i=sel_list.size(); i--; ){
                              PI3D_frame frm = sel_list[i];
                              if(frm->GetType()==FRAME_SOUND){
                                 PI3D_sound snd = I3DCAST_SOUND(frm);
                                 C_string old_name = snd->GetFileName();
                                 ep->ed->GetSoundCache().GetRelativeDirectory(&old_name[0]);
                                 if(str.Matchi(old_name)) continue;
                                 
                                 int len = strlen(frm->GetName()) + old_name.Size() + 2;
                                 char *cp = new char[len];
                                 strcpy(cp, frm->GetName());
                                 strcpy(cp+strlen(frm->GetName())+1, old_name);
                                 S_edit_Undo_2 eu(ep->GetName(), E_PROP_RELOAD_SOUND, cp, len);
                                 ep->e_undo->Action(E_UNDO_SAVE, &eu);
                                 delete[] cp;

                                 ep->ed->GetSoundCache().Open(snd, str, ep->ed->GetScene(), 0,
                                    OPEN_ERR_LOG(ep->ed, str));
                                 ep->ed->SetModified();

                                 SendMessage(hwnd, WM_APP_INIT_TAB + 2, 0, (LPARAM)&sel_list);
                              }
                           }
                        }
                     }
                  }
                  break;

               case IDC_CHECK_LIGHT_MODE_V:
               case IDC_CHECK_LIGHT_MODE_LM:
               case IDC_CHECK_LIGHT_MODE_SHD:
               case IDC_CHECK_LIGHT_MODE_DYN:
                  {
                     C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                     if(!ep) break;
                     ep->in_sheet_modify = true;
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);

                     bool on = IsDlgButtonChecked(hwnd, LOWORD(wParam));
                     on = !on;

                     ep->SaveUndoLights(sel_list);
                     for(int i=sel_list.size(); i--; ){
                        PI3D_frame frm = sel_list[i];
                        if(frm->GetType()!=FRAME_LIGHT) continue;
                        PI3D_light lp = I3DCAST_LIGHT(frm);
                        dword mode = lp->GetMode();
                        
                        switch(LOWORD(wParam)){
                        case IDC_CHECK_LIGHT_MODE_V:
                           mode &= ~I3DLIGHTMODE_VERTEX;
                           if(on) mode |= I3DLIGHTMODE_VERTEX;
                           break;
                        case IDC_CHECK_LIGHT_MODE_LM:
                           mode &= ~I3DLIGHTMODE_LIGHTMAP;
                           if(on) mode |= I3DLIGHTMODE_LIGHTMAP;
                           break;
                        case IDC_CHECK_LIGHT_MODE_SHD:
                           mode &= ~I3DLIGHTMODE_SHADOW;
                           if(on) mode |= I3DLIGHTMODE_SHADOW;
                           break;
                        case IDC_CHECK_LIGHT_MODE_DYN:
                           mode &= ~I3DLIGHTMODE_DYNAMIC_LM;
                           if(on) mode |= I3DLIGHTMODE_DYNAMIC_LM;
                           break;
                        default: assert(0);
                        }
                        lp->SetMode(mode);
                        lp->Update();

                        pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_LIGHT);
                        ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                        ep->ed->SetModified();
                     }
                     CheckDlgButton(hwnd, LOWORD(wParam), !on ? BST_UNCHECKED : BST_CHECKED);
                     ep->in_sheet_modify = false;
                  }
                  break;

               case IDC_CHECK_SOUND_MODE_ON:
               case IDC_CHECK_SOUND_MODE_LOOP:
               case IDC_CHECK_SOUND_MODE_STREAM:
                  {
                     C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                     if(!ep) break; 
                     ep->in_sheet_modify = true;
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);

                     bool on = IsDlgButtonChecked(hwnd, LOWORD(wParam));
                     on = !on;

                     ep->SaveUndoSounds(sel_list);
                     for(int i=sel_list.size(); i--; ){
                        PI3D_frame frm = sel_list[i];
                        if(frm->GetType()!=FRAME_SOUND) continue;
                        PI3D_sound sp = I3DCAST_SOUND(frm);
                        bool loop = sp->IsLoop();
                        bool strm = sp->IsStreaming();
                        bool play = sp->IsOn();
                        
                        switch(LOWORD(wParam)){
                        case IDC_CHECK_SOUND_MODE_ON:
                           play = false;
                           if(on) play = true;
                           break;
                        case IDC_CHECK_SOUND_MODE_LOOP:
                           loop = false;
                           if(on) loop = true;
                           break;
                        case IDC_CHECK_SOUND_MODE_STREAM:
                           strm = false;
                           if(on) strm = true;
                           break;
                        }
                        sp->SetLoop(loop);
                        sp->SetStreaming(strm);
                        sp->SetOn(play);

                        pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_SOUND);
                        ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                        ep->ed->SetModified();
                     }
                     CheckDlgButton(hwnd, LOWORD(wParam), !on ? BST_UNCHECKED : BST_CHECKED);
                     ep->in_sheet_modify = false;
                  }
                  break;

               case IDC_CHECK_NO_SHADOW:
               case IDC_CHECK_NO_COLL:
               case IDC_CHECK_SHADOW_CAST:
               case IDC_CHECK_SHADOW_REC:
               case IDC_CHECK_VOLUME_STATIC:
                  {
                     C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                     if(!ep) break; 
                     ep->in_sheet_modify = true;
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);

                     bool on = IsDlgButtonChecked(hwnd, LOWORD(wParam));
                     on = !on;

                     dword flags;
                     switch(LOWORD(wParam)){
                     case IDC_CHECK_NO_SHADOW: flags = I3D_FRMF_NOSHADOW; break;
                     case IDC_CHECK_NO_COLL: flags = I3D_FRMF_NOCOLLISION; break;
                     case IDC_CHECK_SHADOW_CAST: flags = I3D_FRMF_SHADOW_CAST; break;
                     case IDC_CHECK_SHADOW_REC: flags = I3D_FRMF_SHADOW_RECEIVE; break;
                     case IDC_CHECK_VOLUME_STATIC:
                        {
                           flags = I3D_FRMF_STATIC_COLLISION;
                           ep->ed->GetScene()->ResetDynamicVolTreeEx();
                        }
                        break;
                     default: assert(0);
                     }

                     ep->SaveUndoFrameFlags(sel_list);
                     for(int i=sel_list.size(); i--; ){
                        PI3D_frame frm = sel_list[i];
                        frm->SetFlags(on ? flags : 0, flags);
                        ep->ed->SetModified();
                     }
                     CheckDlgButton(hwnd, LOWORD(wParam), !on ? BST_UNCHECKED : BST_CHECKED);
                     if(LOWORD(wParam)==IDC_CHECK_VOLUME_STATIC){
                        SendMessage(hwnd, WM_APP_INIT_TAB + 8, 0, (LPARAM)&sel_list);
                     }
                     ep->in_sheet_modify = false;
                  }
                  break;

               case IDC_BUTTON_LIGHT_COLOR:
               //case IDC_BUTTON_LIGHT_SPECULAR:
                  {
                     C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                     if(!ep) break; 
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                     //bool use_specular = (LOWORD(wParam)==IDC_BUTTON_LIGHT_SPECULAR);

                     static COLORREF c_cust[16];
                     CHOOSECOLOR cc;
                     memset(&cc, 0, sizeof(cc));
                     cc.lStructSize = sizeof(CHOOSECOLOR);
                     cc.hwndOwner = hwnd;
                     cc.lpCustColors = c_cust;
                     cc.Flags = CC_ANYCOLOR | CC_FULLOPEN;
                              //check if all lights has the same color
                     bool valid = false;
                     S_vector color;
                     if(sel_list.front()->GetType()==FRAME_LIGHT){   //should always pass
                        color = 
                           //(use_specular ? I3DCAST_LIGHT(sel_list.front())->GetSpecularColor() : I3DCAST_LIGHT(sel_list.front())->GetColor());
                           I3DCAST_LIGHT(sel_list.front())->GetColor();
                        valid = true;
                     }
                     for(int i=1; i<sel_list.size(); i++){
                        if(sel_list[i]->GetType()==FRAME_LIGHT)
                           if(valid){
                              PI3D_light lp = I3DCAST_LIGHT(sel_list[i]);
                              const S_vector &c1 =
                                 //(use_specular ? lp->GetSpecularColor() : lp->GetColor());
                                 lp->GetColor();
                              valid = (c1==color);
                           }
                     }
                     if(valid){
                        cc.rgbResult = RGB(color[0]*255.0f, color[1]*255.0f, color[2]*255.0f);
                        cc.Flags |= CC_RGBINIT;
                     }
                     cc.Flags |= CC_ENABLEHOOK;
                     cc.lpfnHook = ccHookColor;
                     cc.lCustData = (LPARAM)ep;

                     if(ChooseColor(&cc)){
                        ep->SaveUndoLights(sel_list);

                        union{
                           byte b[4];
                           dword dw;
                        } au;
                        au.dw = cc.rgbResult;
                        S_vector v(au.b[0]/255.0f, au.b[1]/255.0f, au.b[2]/255.0f);
                        for(i=0; i<sel_list.size(); i++){
                           if(sel_list[i]->GetType()==FRAME_LIGHT){
                              PI3D_light lp = I3DCAST_LIGHT(sel_list[i]);
                              /*
                              if(use_specular)
                                 lp->SetSpecularColor(v);
                              else
                              */
                                 lp->SetColor(v);
                              lp->Update();
   
                              pair<PI3D_frame, dword> p(lp, E_MODIFY_FLG_LIGHT);
                              ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                              ep->ed->SetModified();
                           }
                        }
                        SendMessage(hwnd, WM_APP_INIT_TAB + 1, 0, (LPARAM)&sel_list);
                     }
                  }
                  break;

               case IDC_BUTTON_LSECTORS_ALL1:
               case IDC_BUTTON_LSECTORS_NO1:
               case IDC_BUTTON_SSECTORS_ALL1:
               case IDC_BUTTON_SSECTORS_NO1:
                  {
                     C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                     if(!ep)
                        break;
                     ep->in_sheet_modify = true;
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);

                     I3D_FRAME_TYPE ftype;
                     dword modify_flags;
                     bool add;

                     switch(LOWORD(wParam)){
                     case IDC_BUTTON_LSECTORS_ALL1:
                     case IDC_BUTTON_LSECTORS_NO1:
                        ep->SaveUndoLights(sel_list);
                        ftype = FRAME_LIGHT;
                        modify_flags = E_MODIFY_FLG_LIGHT;
                        add = (LOWORD(wParam)==IDC_BUTTON_LSECTORS_ALL1);
                        break;
                     case IDC_BUTTON_SSECTORS_ALL1:
                     case IDC_BUTTON_SSECTORS_NO1:
                        ep->SaveUndoSounds(sel_list);
                        ftype = FRAME_SOUND;
                        modify_flags = E_MODIFY_FLG_SOUND;
                        add = (LOWORD(wParam)==IDC_BUTTON_SSECTORS_ALL1);
                        break;
                     }
                              //get all sectors in scene
                     struct S_hlp{
                        static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                           vector<PI3D_sector> *sect_list = (vector<PI3D_sector>*)c;
                           sect_list->push_back(I3DCAST_SECTOR(frm));
                           return I3DENUMRET_OK;
                        }
                     };
                     vector<PI3D_sector> sect_list;
                     sect_list.push_back(ep->ed->GetScene()->GetPrimarySector());
                     ep->ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&sect_list, ENUMF_SECTOR);
                     pair<PI3D_frame, dword> p(NULL, modify_flags);

                     for(int i=sel_list.size(); i--; ){
                        PI3D_frame frm = sel_list[i];
                        if(frm->GetType()==ftype){
                           switch(ftype){
                           case FRAME_LIGHT:
                              {
                                 PI3D_light lp = I3DCAST_LIGHT(frm);
                                 for(int j=sect_list.size(); j--; ){
                                    PI3D_sector sct = sect_list[j];
                                    if(add) sct->AddLight(lp);
                                    else sct->DeleteLight(lp);
                                 }
                              }
                              break;
                           case FRAME_SOUND:
                              {
                                 PI3D_sound sp = I3DCAST_SOUND(frm);
                                 for(int j=sect_list.size(); j--; ){
                                    PI3D_sector sct = sect_list[j];
                                    if(add) sct->AddSound(sp);
                                    else sct->DeleteSound(sp);
                                 }
                              }
                              break;
                           }
                           p.first = frm;
                           ep->e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                           ep->ed->SetModified();
                        }
                     }
                     switch(ftype){
                     case FRAME_LIGHT:
                        SendMessage(hwnd, WM_APP_INIT_TAB + 6, 0, (LPARAM)&sel_list);
                        break;
                     case FRAME_SOUND:
                        SendMessage(hwnd, WM_APP_INIT_TAB + 11, 0, (LPARAM)&sel_list);
                        break;
                     }
                     ep->in_sheet_modify = false;
                  }
                  break;

               case IDC_BUTTON_LSECTORS_PICK:
               case IDC_BUTTON_SSECTORS_PICK:
                  {
                              //pick sector
                     S_MouseEdit_pick p;
                     p.ei = ep;
                     p.hcursor = NULL;
                     p.hcursor = LoadCursor(GetHInstance(),
                        LOWORD(wParam)==IDC_BUTTON_LSECTORS_PICK ?
                        "IDC_CURSOR_LSECT_PICK" :
                        "IDC_CURSOR_SSECT_PICK"
                        );
                     switch(LOWORD(wParam)){
                     case IDC_BUTTON_LSECTORS_PICK:
                        p.action_id = E_PROP_NOTIFY_PICK_LSECT;
                        break;
                     case IDC_BUTTON_SSECTORS_PICK:
                        p.action_id = E_PROP_NOTIFY_PICK_SSECT;
                        break;
                     }
                     ep->e_mouseedit->Action(E_MOUSE_SET_USER_PICK, &p);
                  }
                  break;

                  /*
               case IDC_BUTTON_SCT_SND_PROPS:
                  {
                              //get all sector's properties - feed the table
                     struct S_hlp{
                        LPC_editor ed;
                        C_edit_Properties *ep;
                        LPC_editor_item e_slct;
                        LPC_editor_item e_modify;

                        static void cbTabCfg(LPC_table tab, dword msg, dword cb_user, dword prm2, dword prm3){

                           switch(msg){
                           case TCM_UPDATE:
                              {
                                                   //just send modify notify on all items
                                 for(int i=I3D_SSDI_LAST; i--; )
                                    cbTabCfg(tab, TCM_MODIFY, cb_user, i, -1);
                              }
                              break;
                           case TCM_MODIFY:
                              {
                                 S_hlp *hp = (S_hlp*)cb_user;
                                                   //update params on selection
                                 const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)hp->e_slct->Action(E_SELECTION_GETVECTOR);
                                 for(int i=sel_list.size(); i--; ){
                                    PI3D_frame frm = sel_list[i];
                                    if(frm->GetType()!=FRAME_SECTOR) continue;
                                    PI3D_sector sct = I3DCAST_SECTOR(frm);
                                    sct->SetSoundData((I3D_SECTOR_SOUND_DATA_INDEX)prm2, tab->ItemI(prm2));

                                    pair<PI3D_frame, dword> mp(sct, E_MODIFY_FLG_SECTOR);
                                    hp->e_modify->Action(E_MODIFY_ADD_FLAGS, &mp);
                                 }
                                 hp->ed->GetScene()->Render();
                                 hp->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
                              }
                              break;
                           }
                        }
                     } hlp = {ep->ed, ep, ep->e_slct, ep->e_modify};

                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                     if(sel_list.size() && sel_list.front()->GetType()==FRAME_SECTOR){

                              //init table by 1st sector in selection
                        PI3D_sector sct = I3DCAST_SECTOR(sel_list.front());
                        for(int i=I3D_SSDI_LAST; i--; ){
                           tab->ItemI(i) = sct->GetSoundData((I3D_SECTOR_SOUND_DATA_INDEX)i);
                        }


                        tab->Edit(&templ, hwnd, S_hlp::cbTabCfg,
                           (dword)&hlp,
                           TABEDIT_IMPORT | TABEDIT_EXPORT,
                           NULL, NULL);
                     }

                     tab->Release();
                  }
                  break;
                  */

               default:
                  {
                     if(ep->curr_sheet_index!=8)
                        break;
                     const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ep->e_slct->Action(E_SELECTION_GETVECTOR);
                              //this may be one of visual property
                     int id = LOWORD(wParam);
                     HWND hwnd_ctrl = (HWND)lParam;
                              //make sure this is one of expected controls
                     for(int i=ep->prop_controls.size(); i--; ){
                        if(ep->prop_controls[i]->hwnd==hwnd_ctrl)
                           break;
                     }
                     assert(i!=-1);
                     int prop_index = id - 20000;

                     bool on = IsDlgButtonChecked(hwnd, id);
                     on = !on;
                     CheckDlgButton(hwnd, id, !on ? BST_UNCHECKED : BST_CHECKED);

                     ep->SaveUndoVisuals(sel_list);
                     ep->ModifyVisualsProperty(sel_list, prop_index,
                        on);
                  }
               }
            }
            break;
         }
         break;
      }
      return 0;
   }

//----------------------------

   static BOOL CALLBACK dlgProp(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

      switch(uMsg){
      case WM_INITDIALOG:
         SetWindowLong(hwnd, GWL_USERDATA, lParam);
         return 1;

      case WM_VKEYTOITEM:
         return -1;

      case WM_ACTIVATE:
         {
            C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
            ep->ed->GetIGraph()->UpdateParams(0, 0, 0, 
               LOWORD(wParam)==WA_INACTIVE ? (ep->igraph_init_flags&IG_SUSPENDINACTIVE) : 0,
               IG_SUSPENDINACTIVE);
            if(ep->curr_sheet_index!=-1 && (ep->curr_sheets&(1<<ep->curr_sheet_index))){
               SendMessage(ep->hwnd_sheet[ep->curr_sheet_index], uMsg, wParam, lParam);
            }
         }
         break;

      case WM_NOTIFY:
         switch(wParam){
         case 1:
            {
               LPNMHDR hdr = (LPNMHDR)lParam;
               switch(hdr->code){
               case TCN_SELCHANGE:
                  {
                     SendMessage(hwnd, WM_APP_SHOW_TAB, 0, 0);
                     //SetActiveWindow(hwnd);
                     C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                     SetFocus(ep->hwnd_tc);
                  }
                  break;
               case TCN_KEYDOWN:
                  {
                     NMTCKEYDOWN *kd = (NMTCKEYDOWN*)lParam;
                     switch(kd->wVKey){
                     case VK_PRIOR:
                     case VK_NEXT:
                        {
                           C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                           int num_tabs = SendMessage(ep->hwnd_tc, TCM_GETITEMCOUNT, 0, 0);
                           if(num_tabs>1){
                              int curr_sel = SendMessage(ep->hwnd_tc, TCM_GETCURSEL, 0, 0);
                              (kd->wVKey==VK_PRIOR) ? --curr_sel : ++curr_sel;
                              if(curr_sel<0) curr_sel += num_tabs;
                              curr_sel %= num_tabs;
                              SendMessage(ep->hwnd_tc, TCM_SETCURSEL, curr_sel, 0);
                              SendMessage(hwnd, WM_APP_SHOW_TAB, 0, 0);
                           }
                        }
                        break;
                     }
                  }
                  break;
               }
            }
            break;
         }
         break;

      case WM_COMMAND:
         switch(LOWORD(wParam)){
         case IDCANCEL:
            {
               C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
               SetActiveWindow((HWND)ep->ed->GetIGraph()->GetHWND());
            }
            break;
         }
         break;

      case WM_APP_SHOW_TAB:  //show currently selected tab and hide previous one
         {
            C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
                              //hide current sheet and display new
            if(ep->curr_sheet_index!=-1)
               ShowWindow(ep->hwnd_sheet[ep->curr_sheet_index], SW_HIDE);
            int tab_indx = SendMessage(ep->hwnd_tc, TCM_GETCURSEL, 0, 0);
            if(tab_indx==-1){
               ep->curr_sheet_index = -1;
            }else{
               ep->curr_sheet_index = ep->TabIndexToSheetIndex(tab_indx);
               assert(ep->curr_sheet_index!=-1);
               ShowWindow(ep->hwnd_sheet[ep->curr_sheet_index], SW_SHOW);
            }
         }
         break;

      case WM_CLOSE:
         {
            C_edit_Properties *ep = (C_edit_Properties*)GetWindowLong(hwnd, GWL_USERDATA);
            ep->on = false;
            ep->UpdateWin();
         }
         break;
      }
      return 0;
   }

//----------------------------

   void UpdateWin(){
      if(on){
         ShowWindow(hwnd, SW_SHOW);
         SetActiveWindow(hwnd);
      }else{
         ShowWindow(hwnd, SW_HIDE);
      }
      ed->CheckMenu(this, E_PROP_TOGGLE, on);
   }

//----------------------------

   void AddWindow(int indx, HWND hwnd_sh){

      TCITEM tci;
      memset(&tci, 0, sizeof(tci));
      tci.mask = TCIF_TEXT;

      char buf[256];
      SendMessage(hwnd_sh, WM_GETTEXT, sizeof(buf), (LPARAM)buf);
      tci.pszText = buf;
      tci.lParam = (LPARAM)hwnd_sh;
      SendMessage(hwnd_tc, TCM_INSERTITEM, indx, (LPARAM)&tci);

      {
         RECT rc_tc;
         GetWindowRect(hwnd_tc, &rc_tc);
         rc_tc.right -= rc_tc.left;
         rc_tc.bottom -= rc_tc.top;
         {
            RECT rc;
            GetWindowRect(hwnd, &rc);
            rc_tc.left -= rc.left;
            rc_tc.top -= rc.top;
         }
         SendMessage(hwnd_tc, TCM_ADJUSTRECT, false, (LPARAM)&rc_tc);

         SetWindowPos(hwnd_sh, NULL, rc_tc.left, rc_tc.top, 
            (rc_tc.right-rc_tc.left), (rc_tc.bottom-rc_tc.top), 
            SWP_NOOWNERZORDER | SWP_NOZORDER | SWP_NOACTIVATE);
      }
   }

//----------------------------

   void DelItem(int indx){
      ShowWindow(hwnd_sheet[TabIndexToSheetIndex(indx)], SW_HIDE);

      SendMessage(hwnd_tc, TCM_DELETEITEM, indx, 0);
   }

//----------------------------

   static dword GetUpperBits(int i){
      dword ret = 0;
      while(i<32){
         ret |= (1<<i);
         ++i;
      }
      return ret;
   }

//----------------------------

   void InitTabs(dword mask, const vector<PI3D_frame> *sel_list = NULL, 
      bool del_from_list = false){

      int itm_indx = 0;
      for(int i=0; i<hwnd_sheet.size(); i++){
         dword bit = (1<<i);

         bool init = (mask&bit);

         if((curr_sheets^mask)&bit){
                              //some change
            if(mask&bit){
               curr_sheets |= bit;
                              //insert
               AddWindow(itm_indx, hwnd_sheet[i]);
            }else{
                              //remove
               DelItem(itm_indx);
               curr_sheets &= ~bit;
               if(del_from_list){
                              //detete sheet handle
                  hwnd_sheet.erase(&hwnd_sheet[i]);
                              //shift bits
                  dword save = curr_sheets & GetUpperBits(i+1);
                  curr_sheets &= ~GetUpperBits(i);
                  curr_sheets |= save>>1;

                  save = mask & GetUpperBits(i+1);
                  mask &= ~GetUpperBits(i);
                  mask |= save>>1;

                  if(curr_sheet_index==i)
                     curr_sheet_index = -1;

                  --i;
               }
            }
         }

         if(init){
                              //init sheet
            SendMessage(hwnd_sheet[i], WM_APP_INIT_TAB + i, 0, (WPARAM)sel_list);

            ++itm_indx;
         }
      }

      if(curr_sheet_index!=-1 && !(curr_sheets&(1<<curr_sheet_index)))
         curr_sheet_index = -1;

      if(curr_sheet_index==-1){
                              //select first active sheet
#ifdef DEBUG_SELECT_LAST_SHEET
         int tab_indx = SendMessage(hwnd_tc, TCM_GETITEMCOUNT, 0, 0) - 1;
#else
         int tab_indx = 0;
#endif
         SendMessage(hwnd_tc, TCM_SETCURSEL, tab_indx, 0);
         SendMessage(hwnd, WM_APP_SHOW_TAB, 0, 0);
      }else{
                              //keep selected active tab
         if(TabIndexToSheetIndex(SendMessage(hwnd_tc, TCM_GETCURSEL, 0, 0)) != curr_sheet_index){
            SendMessage(hwnd_tc, TCM_SETCURSEL, SheetIndexToTabIndex(curr_sheet_index), 0);
         }
      }
   }

//----------------------------

   void InitTabs(){
                              //get current selection
      const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
      dword mask = curr_sheets & ~(SHEET_USER-1);
      if(sel_list.size()){
         mask |= SHEET_FRAME;
         int i;
                              //check what have frames in common and display appropriate tab
         int frm_type = -1;
         int sub_type = -1;
         for(i=sel_list.size(); i-- && frm_type!=-2; ){
            PI3D_frame frm = sel_list[i];
            dword ft = frm->GetType();
            switch(frm_type){
            case -2: break;
            case -1: 
               frm_type = ft;
               if(ft==FRAME_VISUAL) sub_type = I3DCAST_VISUAL(frm)->GetVisualType();
               break;
            default: 
               if(frm_type != ft) frm_type = -2;
               else
               switch(sub_type){
               case -1: break;
               case -2: break;
               default:
                  if(ft==FRAME_VISUAL && sub_type != I3DCAST_VISUAL(frm)->GetVisualType())
                     sub_type = -2;
               }
            }
         }
                              //if same type, display frame-dependent sheet
         switch(frm_type){
         case FRAME_LIGHT:    mask |= SHEET_LIGHT | SHEET_LSECTORS; break;
         case FRAME_SOUND:    mask |= SHEET_SOUND | SHEET_SSECTORS; break;
         case FRAME_CAMERA:   mask |= SHEET_CAMERA; break;
         case FRAME_MODEL:    mask |= SHEET_MODEL;  break;
         case FRAME_TARGET:   mask |= SHEET_TARGET; break;
         case FRAME_VOLUME:   mask |= SHEET_VOLUME; break;
         case FRAME_SECTOR:   mask |= SHEET_SECTOR; break;
		   case FRAME_VISUAL:   mask |= SHEET_VISUAL; break;
         }
      }
      InitTabs(mask, &sel_list);
   }

//----------------------------

   void FeedMaterials(){

      assert(hwnd_sheet.size() >= NUM_SHEETS);
      HWND hwnd_mat = GetDlgItem(hwnd_sheet[7], IDC_COMBO_VOLUME_MAT);
      HWND hwnd_vis_mat = GetDlgItem(hwnd_sheet[8], IDC_COMBO_VISUAL_MAT);
      assert(hwnd_mat);
      SendMessage(hwnd_mat, CB_RESETCONTENT, 0, 0);
      assert(hwnd_vis_mat);
      SendMessage(hwnd_vis_mat, CB_RESETCONTENT, 0, 0);
      int i = 0;
      for(map<int, C_string>::const_iterator it=mat_table.begin(); it!=mat_table.end(); it++, i++){
         SendMessage(hwnd_mat, CB_ADDSTRING, 0, (LPARAM)(const char*)((*it).second));
         SendMessage(hwnd_mat, CB_SETITEMDATA, i, (*it).first);
         SendMessage(hwnd_vis_mat, CB_ADDSTRING, 0, (LPARAM)(const char*)((*it).second));
         SendMessage(hwnd_vis_mat, CB_SETITEMDATA, i, (*it).first);
      }
   }

//----------------------------

   void FeedEnvironments(){

      assert(hwnd_sheet.size() >= NUM_SHEETS);
      HWND hwnd_cb = GetDlgItem(hwnd_sheet[12], IDC_COMBO_SCT_ENV);
      assert(hwnd_cb);
      SendMessage(hwnd_cb, CB_RESETCONTENT, 0, 0);
      int i = 0;
      for(map<int, C_string>::const_iterator it=env_table.begin(); it!=env_table.end(); it++, i++){
         SendMessage(hwnd_cb, CB_ADDSTRING, 0, (LPARAM)(const char*)((*it).second));
         SendMessage(hwnd_cb, CB_SETITEMDATA, i, (*it).first);
      }
   }

//----------------------------
// Refresh sheets - reorganize tabs if selection changes, or selection is modified.
   void RefreshSheets(){
      if(on){
         if(slct_reset){
            slct_reset = false;
            sheet_reset = false;
            InitTabs();
                              //cancel any previous picking
            if(((LPC_editor_item)e_mouseedit->Action(E_MOUSE_GET_USER_PICK))==this){
               e_mouseedit->Action(E_MOUSE_SET_USER_PICK, NULL);
            }
         }else
         if(sheet_reset){
            sheet_reset = false;
            SendMessage(hwnd_sheet[curr_sheet_index], WM_APP_INIT_TAB + curr_sheet_index,
               0, e_slct->Action(E_SELECTION_GETVECTOR));
         }
      }
   }

public:
   C_edit_Properties():
      hwnd(NULL),
      e_slct(NULL),
      e_modify(NULL),
      e_undo(NULL),
      curr_sheets(0),
      curr_sheet_index(-1),
      vis_prop_scroll_y(0),
      slct_reset(false),
      sheet_reset(false),
      in_sheet_modify(false),
      on(true)
   {}

//----------------------------

   virtual bool Init(){
      e_slct = ed->FindPlugin("Selection");
      e_modify = ed->FindPlugin("Modify");
      e_undo = ed->FindPlugin("Undo");
      e_mouseedit = ed->FindPlugin("MouseEdit");
      if(!e_slct || !e_modify || !e_undo || !e_mouseedit){
         e_slct = NULL;
         e_modify = NULL;
         e_undo = NULL;
         e_mouseedit = NULL;
         return false;
      }
      e_slct->AddRef();
      e_modify->AddRef();
      e_undo->AddRef();

      igraph_init_flags = ed->GetIGraph()->GetFlags();
                              //create properties window
      hwnd = CreateDialogParam(GetHInstance(), "IDD_PROPERTIES",
         (HWND)ed->GetIGraph()->GetHWND(), dlgProp, (LPARAM)this);
      if(hwnd){
         RECT rc;
         GetWindowRect(hwnd, &rc);
                              //by default, put it into right down corner
         int x = GetSystemMetrics(SM_CXSCREEN);
         int y = GetSystemMetrics(SM_CYSCREEN);
         x -= (rc.right - rc.left);
         y -= (rc.bottom - rc.top);

         SetWindowPos(hwnd, NULL, x, y, 0, 0,
            SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_NOZORDER);

         UpdateWin();

         ed->GetIGraph()->AddDlgHWND(hwnd);
         SetActiveWindow((HWND)ed->GetIGraph()->GetHWND());

         hwnd_tc = GetDlgItem(hwnd, 1);

                              //load sheets into property window
         for(int i=0; i<NUM_SHEETS; i++){
            S_sheet_init si = {i, this};
            HWND hwnd_sh = CreateDialogParam(GetHInstance(), C_fstring("IDD_PROP_SHEET%.2i", i),
               hwnd, dlgSheet, (LPARAM)&si);
            SetParent(hwnd_sh, hwnd);
            hwnd_sheet.push_back(hwnd_sh);
         }
         {
            HWND hwnd_pg = GetDlgItem(hwnd_sheet[8], IDC_VIS_PROP_GROUP_BASE);
            assert(hwnd_pg);
            RECT rc;
            POINT pt = {0, 0};
            GetWindowRect(hwnd_pg, &rc);
            ClientToScreen(hwnd_sheet[8], &pt);
            vis_property_ground_pos[0] = rc.left - pt.x;
            vis_property_ground_pos[1] = rc.top - pt.y;
            vis_property_ground_size[0] = rc.right - rc.left;
            vis_property_ground_size[1] = rc.bottom - rc.top;

            hwnd_pg = hwnd_sheet[9];
            
            SetParent(hwnd_pg, hwnd_sheet[8]);
            SetWindowLong(hwnd_pg, GWL_ID, IDC_VIS_PROP_GROUP);
            SetWindowPos(hwnd_pg, NULL, vis_property_ground_pos[0], vis_property_ground_pos[1],
               vis_property_ground_size[0], vis_property_ground_size[1],
               SWP_SHOWWINDOW);
         }
      }

      ed->AddShortcut(this, E_PROP_TOGGLE, "%10 &Edit\\%100 %i &Properties\tAlt+Enter", K_ENTER, SKEY_ALT);

      {                       //be sensible on selection change
         pair<LPC_editor_item, dword> p(this, E_PROP_NOTIFY_SELECTION);
         e_slct->Action(E_SELECTION_ADD_NOTIFY, &p);
      }
      {                       //be sensible on modifications
         pair<LPC_editor_item, dword> p(this, E_PROP_NOTIFY_MODIFY);
         e_modify->Action(E_MODIFY_ADD_NOTIFY, &p);
      }
      return true;
   }

//----------------------------

   virtual void Close(){ 
      if(e_slct)
         e_slct->Action(E_SELECTION_DEL_NOTIFY, this);
      if(e_modify)
         e_modify->Action(E_MODIFY_DEL_NOTIFY, this);

      if(e_slct) e_slct->Release();
      if(e_modify) e_modify->Release();
      if(e_undo) e_undo->Release();
      e_slct = NULL;
      e_modify = NULL;
      e_undo = NULL;
      e_mouseedit = NULL;

      if(hwnd){
         ed->GetIGraph()->DeleteDlgHWND(hwnd);
         DestroyWindow(hwnd);
         hwnd = NULL;
      }
      InitTabs(0);
      for(int i=hwnd_sheet.size(); i--; ){
         DestroyWindow(hwnd_sheet[i]);
      }
      hwnd_sheet.clear();
   }

//----------------------------

   int TabIndexToSheetIndex(int ti) const{
      if(!curr_sheets) return -1;
      for(int i = ti+1, indx = 0; i; ++indx)
         if(curr_sheets&(1<<indx)) --i;
      --indx;
      return indx;
   }

//----------------------------

   int SheetIndexToTabIndex(int si) const{
      if(!(curr_sheets&(1<<si))) return -1;
      for(int i = 0, indx = -1; i<si; ++i)
         if(curr_sheets&(1<<i)) ++indx;
      ++indx;
      return indx;
   }

//----------------------------

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){

      RefreshSheets();
   }

//----------------------------

   virtual dword Action(int id, void *context){

      switch(id){
      case E_PROP_TOGGLE:
         on = true;
         UpdateWin();
         break;

      case E_PROP_NOTIFY_SELECTION:
         slct_reset = true;
         break;

      case E_PROP_ADD_SHEET:
         {
            RefreshSheets();

            HWND hwnd_sh = (HWND)context;
            if(!hwnd_sh) return false;

            dword ws = GetWindowLong(hwnd_sh, GWL_STYLE);
            ws &= ~(WS_POPUP | WS_BORDER | WS_CAPTION | WS_DLGFRAME | WS_SYSMENU | WS_THICKFRAME);
            ws |= WS_CHILD | WS_CLIPSIBLINGS;
            SetWindowLong(hwnd_sh, GWL_STYLE, ws);

            dword wsex = GetWindowLong(hwnd_sh, GWL_EXSTYLE);
            wsex &= ~(WS_EX_WINDOWEDGE);
            wsex |= WS_EX_CONTROLPARENT;
            SetWindowLong(hwnd_sh, GWL_EXSTYLE, wsex);

            SetParent(hwnd_sh, hwnd);
            SetActiveWindow((HWND)ed->GetIGraph()->GetHWND());

            hwnd_sheet.push_back(hwnd_sh);
            InitTabs(curr_sheets | (1<<(hwnd_sheet.size()-1)),
               (vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR));
            return true;
         }
         break;

      case E_PROP_DEL_SHEET:
         {
            HWND hwnd_sh = (HWND)context;
            for(int i=NUM_SHEETS; i<hwnd_sheet.size(); i++)
            if(hwnd_sheet[i]==hwnd_sh){
               if(e_slct){
                  InitTabs(curr_sheets & ~(1<<i),
                     (vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR),
                     true);
               }else
                  hwnd_sheet.erase(&hwnd_sheet[i]);
               return true;
            }
         }
         break;

      case E_PROP_SHOW_SHEET:
         {
            HWND hwnd_sh = (HWND)context;
            for(int i=NUM_SHEETS; i<hwnd_sheet.size(); i++)
            if(hwnd_sheet[i]==hwnd_sh){
               if(curr_sheet_index!=i){
                  SendMessage(hwnd_tc, TCM_SETCURSEL, SheetIndexToTabIndex(i), 0);
                  SendMessage(hwnd, WM_APP_SHOW_TAB, 0, 0);
               }
               break;
            }
         }
         break;

      case E_PROP_ACTIVATE:
         on = true;
         UpdateWin();
         break;

      case E_PROP_NOTIFY_MODIFY:
                              //check if immediate action is to be taken (sheet update)
         if(on && !slct_reset && !in_sheet_modify/* && curr_sheet==0 && (curr_sheets&SHEET_FRAME)*/){
            const pair<PI3D_frame, dword> &p = *(pair<PI3D_frame, dword>*)context;
                              //get flags of current sheet
            dword flg = 1<<curr_sheet_index;
            sheet_reset = ((flg&SHEET_FRAME) && (p.second&(E_MODIFY_FLG_POSITION|E_MODIFY_FLG_ROTATION|E_MODIFY_FLG_SCALE))) ||
               ((flg&(SHEET_LIGHT|SHEET_LSECTORS)) && (p.second&E_MODIFY_FLG_LIGHT)) ||
               ((flg&(SHEET_SOUND|SHEET_SSECTORS)) && (p.second&E_MODIFY_FLG_SOUND)) ||
               ((flg&SHEET_VISUAL) && (p.second&(E_MODIFY_FLG_NU_SCALE|E_MODIFY_FLG_VISUAL))) ||
               ((flg&SHEET_SECTOR) && (p.second&E_MODIFY_SECTOR)) ||
               ((flg&SHEET_VOLUME) && (p.second&(E_MODIFY_FLG_NU_SCALE|E_MODIFY_FLG_VOLUME)));
         }
         break;

      case E_PROP_RESET_SHEET:
         sheet_reset = true;
         break;

      case E_PROP_NOTIFY_PICK_TARGET:
         {                    //get single (target) selection
            PI3D_frame frm = (PI3D_frame)e_slct->Action(E_SELECTION_GET_SINGLE);
            if(frm && frm->GetType()==FRAME_TARGET){
               PI3D_target tgt = I3DCAST_TARGET(frm);
               tgt->SetFrame(((S_MouseEdit_picked*)context)->frm_picked);
                              //re-init sheet
               sheet_reset = (curr_sheet_index==5);
            }
         }
         break;

      case E_PROP_NOTIFY_PICK_LIGHT:
         {
            /*
            if(!ed->CanModify()) break;
            PI3D_frame frm = ((S_MouseEdit_picked*)context)->frm_picked;
            if(!frm || frm->GetType()!=FRAME_MODEL){
               ed->Message("Please click on light", 0, EM_WARNING);
               break;
            }
            */
            ed->Message("Function not implemented", 0, EM_WARNING);
         }
         break;

      case E_PROP_NOTIFY_PICK_MODEL:
         {
            if(!ed->CanModify()) break;
            PI3D_frame frm = ((S_MouseEdit_picked*)context)->frm_picked;
            frm = TrySelectTopModel(frm);
            if(!frm || frm->GetType()!=FRAME_MODEL){
               ed->Message("Please click on model", 0, EM_WARNING);
               break;
            }
            C_string new_name = I3DCAST_MODEL(frm)->GetFileName();
            ed->GetModelCache().GetRelativeDirectory(&new_name[0]);

            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
                              //change all models
            for(int i=sel_list.size(); i--; ){
               PI3D_frame frm = sel_list[i];
               if(frm->GetType()==FRAME_MODEL)
                  ReplaceModel(I3DCAST_MODEL(frm), new_name);
            }
            SendMessage(hwnd_sheet[4], WM_APP_INIT_TAB + 4, 0, (LPARAM)&sel_list);
         }
         break;

      case E_PROP_NOTIFY_PICK_VISUAL:
         {
            if(!ed->CanModify()) break;
            PI3D_frame frm = ((S_MouseEdit_picked*)context)->frm_picked;
            if(!frm || frm->GetType()!=FRAME_VISUAL){
               ed->Message("Please click on visual", 0, EM_WARNING);
               break;
            }
            CopyProperties(I3DCAST_VISUAL(frm));
         }
         break;

      case E_PROP_NOTIFY_PICK_LSECT:
      case E_PROP_NOTIFY_PICK_SSECT:
         {
            if(!ed->CanModify()) break;
                              //determine picked sector
            PI3D_frame frm = ((S_MouseEdit_picked*)context)->frm_picked;
            if(!frm) frm = ed->GetScene()->GetPrimarySector();
            while(frm && frm->GetType()!=FRAME_SECTOR)
               frm = frm->GetParent();
            if(!frm)
               break;
            PI3D_sector sct = I3DCAST_SECTOR(frm);

            in_sheet_modify = true;
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            if(!sel_list.size())
               break;

            switch(id){
            case E_PROP_NOTIFY_PICK_LSECT:
               {
                  if(sel_list[0]->GetType()!=FRAME_LIGHT)
                     break;
                  PI3D_light lp0 = I3DCAST_LIGHT(sel_list[0]);
                                 //determine if 1st light is currently in,
                                 // toggle based on 1st light
                  for(int i=sct->NumLights(); i--; )
                     if(sct->GetLight(i)==lp0) break;
                  bool set_into = (i==-1);
                  SaveUndoLights(sel_list);
                  pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_LIGHT);
                  for(i=sel_list.size(); i--; ){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()==FRAME_LIGHT){
                        PI3D_light lp = I3DCAST_LIGHT(frm);
                        if(set_into) sct->AddLight(lp);
                        else sct->DeleteLight(lp);

                        p.first = frm;
                        e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                        ed->SetModified();
                     }
                  }
                  SendMessage(hwnd_sheet[6], WM_APP_INIT_TAB + 6, 0, (LPARAM)&sel_list);
                  in_sheet_modify = false;
               }
               break;

            case E_PROP_NOTIFY_PICK_SSECT:
               {
                  if(sel_list[0]->GetType()!=FRAME_SOUND)
                     break;
                  PI3D_sound sp0 = I3DCAST_SOUND(sel_list[0]);
                                 //determine if 1st light is currently in,
                                 // toggle based on 1st light
                  for(int i=sct->NumSounds(); i--; )
                     if(sct->GetSounds()[i]==sp0) break;
                  bool set_into = (i==-1);
                  SaveUndoSounds(sel_list);
                  pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_SOUND);
                  for(i=sel_list.size(); i--; ){
                     PI3D_frame frm = sel_list[i];
                     if(frm->GetType()==FRAME_SOUND){
                        PI3D_sound sp = I3DCAST_SOUND(frm);
                        if(set_into) sct->AddSound(sp);
                        else sct->DeleteSound(sp);

                        p.first = frm;
                        e_modify->Action(E_MODIFY_ADD_FLAGS, &p);

                        ed->SetModified();
                     }
                  }
                  SendMessage(hwnd_sheet[11], WM_APP_INIT_TAB + 11, 0, (LPARAM)&sel_list);
                  in_sheet_modify = false;
               }
               break;
            }
                              //stay in pick mode
            return true;
         }
         break;

      case E_PROP_RENAME_FRAME:
         {
            if(!ed->CanModify()) break;
            const char *cp = (const char*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(cp, ENUMF_BACKDROP | ENUMF_ALL);
            if(frm){
               cp += strlen(cp) + 1;
               if(IsNameInScene(ed->GetScene(), cp))
                  break;
               RenameFrame(frm, cp);
                              //update sheet
               if(on)
                  sheet_reset = true;
               else
                  slct_reset = true;
            }else
               ed->Message("Properties(E_PROP_RENAME_FRAME): frame not found", 0, EM_ERROR);
         }
         break;

      case E_PROP_RELOAD_MODEL:
         {
            if(!ed->CanModify()) break;
            const char *frm_name = (const char*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(frm_name, ENUMF_BACKDROP | ENUMF_ALL);
            if(frm && frm->GetType()==FRAME_MODEL){
               int len = strlen(frm_name) + 1;
               const char *new_name = frm_name + len;

               PI3D_model mod = I3DCAST_MODEL(frm);
               ReplaceModel(mod, new_name);
               e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

               SendMessage(hwnd_sheet[4], WM_APP_INIT_TAB + 4, 0, e_slct->Action(E_SELECTION_GETVECTOR));
            }else
               ed->Message("Properties(E_PROP_RELOAD_MODEL): cannot find frame", 0, EM_ERROR);
         }
         break;

      case E_PROP_RELOAD_SOUND:
         {
            if(!ed->CanModify()) break;
            const char *frm_name = (const char*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(frm_name, ENUMF_BACKDROP | ENUMF_ALL);
            if(frm && frm->GetType()==FRAME_SOUND){
               PI3D_sound snd = I3DCAST_SOUND(frm);

               C_string old_name = snd->GetFileName();
               ed->GetSoundCache().GetRelativeDirectory(&old_name[0]);

               int len = strlen(frm_name) + 1;
               const char *file_name = frm_name + len;

               len += old_name.Size() + 1;
               char *cp = new char[len];
               strcpy(cp, frm_name);
               strcpy(cp+strlen(frm_name)+1, old_name);
               S_edit_Undo_2 eu(GetName(), E_PROP_RELOAD_SOUND, cp, len);
               e_undo->Action(E_UNDO_SAVE, &eu);
               delete[] cp;

               ed->GetSoundCache().Open(snd, file_name, ed->GetScene(), 0, OPEN_ERR_LOG(ed, file_name));

               ed->SetModified();
               e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

               SendMessage(hwnd_sheet[2], WM_APP_INIT_TAB + 2, 0, e_slct->Action(E_SELECTION_GETVECTOR));
            }else
               ed->Message("Properties(E_PROP_RELOAD_SOUND): cannot find frame", 0, EM_ERROR);
         }
         break;

      case EDIT_ACTION_BEFORE_FREE:
         InitTabs(~(SHEET_USER-1));
         break;

      case EDIT_ACTION_SET_ACTIVE:
         /*
         if(hwnd && on){
            ShowWindow(hwnd, context ? SW_SHOWNOACTIVATE : SW_HIDE);
         }
         */
         break;

      case EDIT_ACTION_FEED_MATERIALS:
         {
            pair<int, const char*> *pairs = (pair<int, const char*>*)context;
            mat_table.clear();
            for(int i=0; pairs[i].first!=-1; i++)
               mat_table[pairs[i].first] = pairs[i].second;
            FeedMaterials();
         }
         break;

      case EDIT_ACTION_FEED_ENVIRONMENTS:
         {
            pair<int, const char*> *pairs = (pair<int, const char*>*)context;
            env_table.clear();
            for(int i=0; pairs[i].first!=-1; i++)
               env_table[pairs[i].first] = pairs[i].second;
            FeedEnvironments();
            sheet_reset = (curr_sheet_index==12);
         }
         break;

      case E_PROP_SAVE_UNDO_VISUALS:
         SaveUndoVisuals(*(const vector<PI3D_frame>*)context);
         break;
      }
      return 0;
   }

//----------------------------

   //virtual void Render(){}

   virtual bool LoadState(C_chunk *ck){
                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version!=0) return false;
                              //read window pos/size
      RECT rc;
      ck->Read(&rc, sizeof(rc));
      SetWindowPos(hwnd, NULL, rc.left, rc.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_NOACTIVATE);
                              //read on/off status
      ck->Read(&on, sizeof(byte));

      //ck->Read(&curr_sheet_index, sizeof(curr_sheet_index));

                              //update params
      UpdateWin();
      if(on) SetActiveWindow((HWND)ed->GetIGraph()->GetHWND());
      return true;
   }

//----------------------------

   virtual bool SaveState(C_chunk *ck) const{
                              //write version
      byte version = 0;
      ck->Write(&version, sizeof(byte));
                              //write window pos/size
      RECT rc;
      GetWindowRect(hwnd, &rc);
      ck->Write(&rc, sizeof(rc));
                              //write on/off status
      ck->Write(&on, sizeof(byte));
      //ck->Write(&curr_sheet_index, sizeof(curr_sheet_index));

      return true;
   }
};

//----------------------------

void CreateProperties(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Properties;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------

