#include "all.h"
#include "common.h"


#define SEED_VERSION 12

//----------------------------

class C_edit_Seed: public C_editor_item{
   virtual const char *GetName() const{ return "Seed"; }

   C_smart_ptr<C_editor_item> e_slct;
   C_smart_ptr<C_editor_item> e_medit;

   HWND hWndSeed;

   HCURSOR hCursorPutModel;
   HCURSOR hCursorPickModel;

   bool z_align;

   bool sets_changed;

   struct S_set: public C_unknown{
      C_string name;
      vector<C_string> model_names;

      bool random_rotation;

//----------------------------

      int FindModelByName(const C_string &name) const{
         for(int i=model_names.size();i--;)
            if(!stricmp(model_names[i], name))
               return i;
         return -1;
      }

//----------------------------

      bool AddModel(const char *modelfilename){

         if(!*modelfilename)
            return false;

         if(ModelNameInSet(modelfilename))
            return false;

         model_names.push_back(modelfilename);
         return true;
      }

//----------------------------

      bool ModelNameInSet(const char *name) const{
         for(int i=model_names.size(); i--; )
            if(!stricmp(model_names[i], name))
               return true;
         return false;
      }

//----------------------------

      bool DeleteModel(int index){

         if(index < model_names.size()){
            model_names[index] = model_names.back(); model_names.pop_back();
            return true;
         }
         return false;
      }

//----------------------------

      S_set(const C_string &n1):
         name(n1),
         random_rotation(true)
      {
      }

      S_set(const S_set &ss):
         name(ss.name),
         model_names(ss.model_names)
      {
         assert(0);
      }
   };

   typedef S_set *LPS_set;

   vector<C_smart_ptr<S_set> > sets;

   int selected_set;          //index of currently selected set

//----------------------------

   static bool cbSetLess(LPS_set ss1, LPS_set ss2){
      return (stricmp(ss1->name, ss2->name) < 0);
   }

//----------------------------

   bool AddModel(PI3D_model mod){

      assert(mod);
      if(selected_set == -1){
         ed->Message("Please select set to add model to.");
         return false;
      }

      char buf[256];
      strcpy(buf, mod->GetFileName());
      ed->GetModelCache().GetRelativeDirectory(buf);

      if(!sets[selected_set]->AddModel(buf)){
         ed->Message("This model already exists in set.");
         return false;
      }
      return true;
   }

//----------------------------

   bool SetNameExists(const C_string &name) const{

      for(int i=sets.size();i--;)
         if(sets[i]->name == name)
            return true;
      return false;
   }

//----------------------------

   LPS_set AddSet(const C_string &name, bool sort1){

      if(SetNameExists(name)){
         ed->Message("Name of new set is already used.", 0, EM_WARNING);
         return NULL;
      }
      LPS_set newset = new S_set(name);
      sets.push_back(newset);
      if(sort1){
         sort(sets.begin(), sets.end(), cbSetLess);
         for(selected_set = sets.size(); selected_set--; ){
            if(sets[selected_set] == newset)
               break;
         }
         assert(selected_set != -1);
      }
      newset->Release();
      return newset;
   }

//----------------------------
// Deletes currently selected set.
   bool DeleteSet(){

      if(selected_set == -1){
         ed->Message("Please select set to delete.");
         return false;
      }
      sets[selected_set] = sets.back(); sets.pop_back();
      selected_set = Min(0, (int)sets.size() - 1);
      return true;
   }

//----------------------------
// Delete model from the currently selected set
   bool DeleteModel(int model_index){

      if(selected_set == -1){
         ed->Message("Please select set to delete model from.");
         return false;
      }
      assert(model_index < sets[selected_set]->model_names.size());
      return sets[selected_set]->DeleteModel(model_index);
   }

//----------------------------

   static LRESULT CALLBACK SeedCallbackProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){

      bool reset_dialog = false;

	   switch (message){

         case WM_INITDIALOG:
            {
                              //save 'this' pointer
               SetWindowLong(hDlg, GWL_USERDATA, lParam);
            }
            break;

         case WM_COMMAND:
            switch(LOWORD(wParam)){

               case IDC_SEED_SET_NAME:{
                  switch(HIWORD(wParam)){
                     case CBN_SELCHANGE:
                        {
                           C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                           es->selected_set = SendDlgItemMessage(hDlg,IDC_SEED_SET_NAME, CB_GETCURSEL, 0, 0);
                           reset_dialog = true;
                        }
                        break;
                     }
                  }
                  break;

               case IDC_SEED_ROTATION:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     if(es->selected_set != -1){
                        es->sets[es->selected_set]->random_rotation = !es->sets[es->selected_set]->random_rotation;
                     }
                  }
                  break;

               case IDC_SEED_ALIGN:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     es->z_align = !es->z_align;
                  }
                  break;

               case IDC_SEED_DELETE_SET:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     if(es->selected_set != -1){
                        reset_dialog = 
                           es->DeleteSet();
                        if(reset_dialog)
                           es->sets_changed = true;
                     }
                  }
                  break;

               case IDC_SEED_PICK_MODEL:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     S_MouseEdit_pick mousepick;
                     mousepick.ei = es;
                     mousepick.action_id = E_SEED_PICK_MODEL;
                     mousepick.hcursor = es->hCursorPickModel;
                     es->e_medit->Action(E_MOUSE_SET_USER_PICK,&mousepick);
                  }
                  break;

               case IDC_SEED_BROWSE_MODEL:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     char buf[256];
                     if(es->ed->FindPlugin("Create")->Action(E_CREATE_BROWSE_MODEL, buf)){
                        reset_dialog = es->sets[es->selected_set]->AddModel(buf);
                        if(reset_dialog) es->sets_changed = true;
                     }
                  }
                  break;

               case IDC_SEED_PUT:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     if(es->selected_set == -1){
                        es->ed->Message("Please select set to put from.");
                        break;
                     }
                     S_MouseEdit_pick mousepick;
                     mousepick.ei = es;
                     mousepick.action_id =E_SEED_PUT_MODEL;
                     mousepick.hcursor = es->hCursorPutModel;
                     es->e_medit->Action(E_MOUSE_SET_USER_PICK,&mousepick);
                  }
                  break;

               case IDC_SEED_DELETE_MODEL:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     int sel_i = SendDlgItemMessage(hDlg, IDC_SEED_MODELS, LB_GETCURSEL, 0, 0);
                     if(sel_i != -1){
                        int ii = SendDlgItemMessage(hDlg, IDC_SEED_MODELS, LB_GETITEMDATA, sel_i, 0);
                        reset_dialog = 
                           es->DeleteModel(ii);
                        if(reset_dialog) es->sets_changed = true;
                     }
                  }
                  break;

               case IDC_SEED_CREATE_SET:
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     char buf[256];
                     GetDlgItemText(hDlg, IDC_SEED_SET_NAME, buf, 256);
                     reset_dialog =
                        (es->AddSet(buf, true)!=NULL);
                     if(reset_dialog)
                        es->sets_changed = true;
                  }
                  break;

               case ((WORD)IDC_SEED_CLOSE):
                  {
                     C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
                     if(es->hWndSeed){
                        es->Action(E_SEED_TOGGLE_DIALOG, NULL);
                     }
                     return true;
                  }
                  break;
            }
            break;
	   }

      if(reset_dialog){
         C_edit_Seed *es = (C_edit_Seed*)GetWindowLong(hDlg, GWL_USERDATA);
         es->SetupDialogParams();
         return true;
      }
      return false;
   }

//----------------------------

   bool SetupDialogParams(){

                              //dialog must be present
      if(!hWndSeed)
         return false;

      SendDlgItemMessage(hWndSeed, IDC_SEED_ALIGN, BM_SETCHECK, z_align ? BST_CHECKED : BST_UNCHECKED, 0);

                              //clear set names
      SendDlgItemMessage(hWndSeed,IDC_SEED_SET_NAME, CB_RESETCONTENT, 0, 0);

                              //feed set names
      for(int i = sets.size(); i--; ){
         int ii = SendDlgItemMessage(hWndSeed, IDC_SEED_SET_NAME, CB_ADDSTRING, 0, (LPARAM)(LPCTSTR)sets[i]->name);
                              //store index as data
         SendDlgItemMessage(hWndSeed, IDC_SEED_SET_NAME, CB_SETITEMDATA, ii, i);
      }
   
      SendDlgItemMessage(hWndSeed, IDC_SEED_SET_NAME, CB_SETCURSEL, selected_set, 0);

      if(selected_set != -1){
         LPS_set ss = sets[selected_set];

         SendDlgItemMessage(hWndSeed, IDC_SEED_ROTATION, BM_SETCHECK, ss->random_rotation ? BST_CHECKED : BST_UNCHECKED, 0);

                              //feed model names
         SendDlgItemMessage(hWndSeed,IDC_SEED_MODELS, LB_RESETCONTENT, 0, 0);
         for(int j = 0; j<ss->model_names.size(); j++){
            int ii = SendDlgItemMessage(hWndSeed, IDC_SEED_MODELS, LB_ADDSTRING, 0, (LPARAM)(LPCTSTR)ss->model_names[j]);
            SendDlgItemMessage(hWndSeed, IDC_SEED_MODELS, LB_SETITEMDATA, ii, j);
         }

      }else{
         SendDlgItemMessage(hWndSeed, IDC_SEED_MODELS, LB_RESETCONTENT, 0, 0);
         SendDlgItemMessage(hWndSeed, IDC_SEED_ROTATION, BM_SETCHECK, BST_UNCHECKED, 0);
      }

      return true;
   }

//----------------------------

   int FindNearUpNameFromScene(char *buf, int start, int speed = 1){
      assert(buf);
      assert(start>=0);
      char *nbuf = new char [strlen(buf)+8];
      sprintf(nbuf,"%s%.2d", buf, start);

      if(!ed->GetScene()->FindFrame(nbuf)){
         if(start == 0)return start;
         sprintf(nbuf,"%s%.2d", buf, start-1);
         bool found = ed->GetScene()->FindFrame(nbuf);
         if(found) return start;
         return FindNearDownNameFromScene(buf, start-1);
      }
      return FindNearUpNameFromScene(buf, start+speed, speed*2);
   }

//----------------------------

   int FindNearDownNameFromScene(char *buf, int start, int speed = 1){
      assert(buf);
      assert(start>=0);

      char *nbuf = new char [strlen(buf)+8];
      sprintf(nbuf,"%s%.2d", buf, start);

      if(ed->GetScene()->FindFrame(nbuf)){
         sprintf(nbuf,"%s%.2d", buf, start+1);
         bool found = ed->GetScene()->FindFrame(nbuf);
         if(!found) return start+1;
         return FindNearUpNameFromScene(buf, start+1);
      }
      return FindNearDownNameFromScene(buf, start-speed, speed*2);
   }

//----------------------------

   void GetNewPosRot(PI3D_model model, const S_vector &from, const S_vector &pick_dir,
      bool rrot, float pick_dist, const S_vector &pick_norm, PI3D_frame pick_frm){

      S_vector dest_pos = from + pick_dir * pick_dist;
      model->SetPos(dest_pos);

      float roll = !rrot ? 0.0f : S_float_random() * PI * 2.0f;

      if(z_align){
         S_matrix m;
         if(!e_medit->Action(E_MOUSE_GET_POLAR_MODE)){
            m.SetDir(-pick_norm, 0.0f);
         }else{
            m.SetDir(-pick_norm, dest_pos);
         }
         model->SetRot(S_quat(m));
         /*
         S_quat resq, rollq, dirq;
         S_vector axis(0.0f, 1.0f, 0.0f);
         S_vector daxis(0.0f, 1.0f,0.0f);
         float dirc = 0.0f;

         float angle;

                              //-Y
         angle = pick_norm.AngleTo(S_vector(0.0f, -1.0f, 0.0f));
         if(angle<(PI*0.25f)){
            axis = S_vector(0.0f, -1.0f, 0.0f);
            dirc = PI;
            daxis = S_vector(0.0f, 0.0f, 1.0f);
         }else{
                              //+Z
            angle = pick_norm.AngleTo(S_vector(0.0f, 0.0f, 1.0f));
            if(angle<(PI*0.25f)){
               axis = S_vector(0.0f, -1.0f, 0.0f);
               dirc = -PI/2;
               daxis = S_vector(1.0f, 0.0f, 0.0f);
            }else{
                              //-Z
               angle = pick_norm.AngleTo(S_vector(0.0f, 0.0f, -1.0f));
               if(angle<(PI*0.25f)){
                  axis = S_vector(0.0f, -1.0f, 0.0f);
                  dirc = PI/2;
                  daxis = S_vector(1.0f, 0.0f, 0.0f);
               }else{
                              //+X
                  angle = pick_norm.AngleTo(S_vector(1.0f, 0.0f, 0.0f));
                  if(angle<(PI*0.25f)){
                     axis = S_vector(0.0f, -1.0f, 0.0f);
                     dirc = PI/2;
                     daxis = S_vector(0.0f, 0.0f, 1.0f);
                  }else{
                              //-X
                     angle = pick_norm.AngleTo(S_vector(-1.0f, 0.0f, 0.0f));
                     if(angle<(PI*0.25)){
                        axis = S_vector(0.0f, -1.0f, 0.0f);
                        dirc = -PI/2;
                        daxis = S_vector(0.0f, 0.0f, 1.0f);
                     }
                  }
               }
            }
         }
         rollq.Make(axis, roll);
         dirq.Make(daxis, dirc);
         model->SetRot(rollq * dirq);
         */
      }else{
         S_matrix m;
         m.SetDir(-pick_norm, roll);
         if(!e_medit->Action(E_MOUSE_GET_POLAR_MODE)){
            model->SetDir1(m(1), -m(2));
         }else{
            m.SetDir(m(1), -m(2));
            const S_vector &up_dir = dest_pos;
            S_vector front; front.GetNormal(up_dir, S_vector(0, 0, 0), m(0));
            model->SetDir1(front, up_dir);
         }
      }
      model->Update();
      model->LinkTo(pick_frm, I3DLINK_UPDMATRIX);
   }

//----------------------------
public:
   C_edit_Seed():
      selected_set(-1),
      sets_changed(false)
   {}

//----------------------------

   virtual bool Init(){
      e_slct = ed->FindPlugin("Selection");
      e_medit = ed->FindPlugin("MouseEdit");
      if(!e_slct || !e_medit){
         return false;
      }

      hWndSeed = NULL;
      ed->AddShortcut(this, E_SEED_TOGGLE_DIALOG, "&Create\\&Seed\tCtrl+E", K_E, SKEY_CTRL);
 
      hCursorPickModel = LoadCursor(GetHInstance(), MAKEINTRESOURCE(IDC_SEED_PICKCURSOR));
      hCursorPutModel =  LoadCursor(GetHInstance(), MAKEINTRESOURCE(IDC_SEED_PUTCURSOR));

      return true;
   }

//----------------------------

   virtual void Close(){ 

      if(hWndSeed){
         DestroyWindow(hWndSeed);
         hWndSeed = NULL;
      }
      if(hCursorPickModel){
         DestroyCursor(hCursorPickModel);
         hCursorPickModel = NULL;
      }
      if(hCursorPutModel){
         DestroyCursor(hCursorPutModel);
         hCursorPutModel = NULL;
      }
      sets.clear();

      e_slct = NULL;
      e_medit = NULL;
   }

//----------------------------

   void PutModel(S_MouseEdit_picked *pick_info);
   
   //virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){};

   virtual dword Action(int id, void *context);

   //virtual void Render(){}

//----------------------------

   virtual bool LoadState(class C_chunk *ck){

      dword version;

      ck->Read(&version, sizeof(version));

      if(version != SEED_VERSION){

         return false;
      }

      ck->Read(&z_align, sizeof(z_align));          
      ck->Read(&selected_set, sizeof(selected_set));

      {
         sets.clear();
         ifstream is;
         is.open("editor\\seeds.txt");
         if(is.good()){
            LPS_set last_set = NULL;
            do{
               char line[1024], *cp = line;
               is.getline(line, sizeof(line));
               if(line[0]==';' || !line[0])
                  continue;
               char kw[64];
               int n;
               if(sscanf(cp, "%s%n", kw, &n)!=1)
                  continue;
               cp += n+1;
               C_string str_kw = kw;
               char data[256];

               if(str_kw.Match("Set")){
                              //make new set
                  sscanf(cp, "\"%[^\"]s\"", data);
                  last_set = AddSet(data, false);
               }else
               if(str_kw.Match("Model")){
                              //add model to last set
                  if(last_set){
                     sscanf(cp, "\"%[^\"]s\"", data);
                     last_set->AddModel(data);
                  }
               }else
               if(str_kw.Match("random_rotation")){
                              //modify set flags
                  if(last_set){
                     sscanf(cp, "%s", data);
                     last_set->random_rotation = atoi(data);
                  }
               }
            }while(!is.eof());
            is.close();
            sort(sets.begin(), sets.end(), cbSetLess);
         }
      }

      selected_set = Min(selected_set, (int)sets.size() - 1);
      SetupDialogParams();
      return true; 
   }

//----------------------------

   virtual bool SaveState(class C_chunk *ck) const{

      dword version = SEED_VERSION;

      ck->Write(&version, sizeof(version));

      ck->Write(&z_align, sizeof(z_align));
      ck->Write(&selected_set, sizeof(selected_set));

      if(sets_changed){
                              //write in text format
         ofstream os;
         os.open("editor\\seeds.txt");
         if(os.good()){

            os<<"; This file was created by Insanity Editor.\n";
            os<<"; Do not modify this file manualy!\n";
            os<<"\n";
            for(int i=0; i<sets.size(); i++){
               const S_set &set = *sets[i];

               os<<(const char*)C_fstring("Set \"%s\"\n", (const char*)set.name);
               os<<(const char*)C_fstring("random_rotation %i\n", set.random_rotation);
               for(int j=set.model_names.size(); j--; ){
                  const C_string &mname = set.model_names[j];
                  os<<(const char*)C_fstring("Model \"%s\"\n", (const char*)mname);
               }
               os<<"\n";
            }
            os.close();
         }
      }
      return true; 
   }
};

//----------------------------

void C_edit_Seed::PutModel(S_MouseEdit_picked *pick_info){

   if(selected_set == -1){
      ed->Message("Please select set to put model from.", 0, EM_WARNING);
      return;
   }

   int num_models = sets[selected_set]->model_names.size();
   if(!num_models){
      ed->Message("Cannot choose model from empty set.", 0, EM_WARNING);
      return;
   }
   int model_index = S_int_random(num_models);

   PI3D_model mod = ed->GetScene()->GetDriver()->CreateModel();

   const C_string &model_name = sets[selected_set]->model_names[model_index];
   if(I3D_FAIL(ed->GetModelCache().Open(mod, model_name, ed->GetScene(), 0, NULL, NULL))){
      ed->Message(C_fstring("Can't open model: '%s'.", (const char*)model_name));
      return;
   }
   mod->SetAnimation(0, NULL, I3DANIMOP_LOOP);
   AddVolumesToScene(mod, ed->GetScene());
   SetupVolumesStatic(mod, true);

   GetNewPosRot(mod, pick_info->pick_from, pick_info->pick_dir,
      sets[selected_set]->random_rotation,
      pick_info->pick_dist, pick_info->pick_norm, pick_info->frm_picked);

   char buf[256];
   for(int i=model_name.Size(); i--; ){
      char c = model_name[i];
      if(c=='\\' || c=='+')
         break;
   }
   strcpy(buf, &model_name[i+1]);
   MakeSceneName(ed->GetScene(), buf);
   mod->SetName(buf);

   pair<PI3D_frame, dword> pr(mod,
      E_MODIFY_FLG_CREATE | E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE | E_MODIFY_FLG_LINK);
   ed->FindPlugin("Modify")->Action(E_MODIFY_ADD_FLAGS, &pr);

   ed->GetScene()->SetFrameSector(mod);

   ed->SetModified();

   {                          //save undo
      S_edit_Undo_2 undostruct("Create", E_CREATE_DELETE_NAMED, (void*)mod->GetName(), strlen(mod->GetName()) + 1);
      ed->FindPlugin("Undo")->Action(E_UNDO_SAVE, &undostruct);
   }
                                                          
   e_slct->Action(E_SELECTION_CLEAR, NULL);
   e_slct->Action(E_SELECTION_ADD_FRAME, mod);

   ed->GetScene()->AddFrame(mod);
   mod->Release();
}

//----------------------------

dword C_edit_Seed::Action(int id, void *context){

   switch(id){

   case E_SEED_PICK_MODEL:
      {
         S_MouseEdit_picked *picked = (S_MouseEdit_picked *)context;
         I3D_frame *frm = picked->frm_picked;
         frm = TrySelectTopModel(frm);
         /*
         while(frm && frm->GetType() != FRAME_MODEL)
            frm = frm->GetParent();
            */
         if(!frm || frm->GetType()!=FRAME_MODEL){
            ed->Message("Please click on model.");
            return true;
         }

         if(AddModel(I3DCAST_MODEL(frm))){
            sets_changed = true;
            SetupDialogParams();
         }
         return true;
      }
      break;

   case E_SEED_PUT_MODEL:
      {
         if(!ed->CanModify()) break;
         S_MouseEdit_picked *picked = (S_MouseEdit_picked *)context;
         if(picked->frm_picked)
            PutModel(picked);
         return true;
      }
      break;

   case E_SEED_TOGGLE_DIALOG:
      {
         LPC_editor_item e_prop = ed->FindPlugin("Properties");
         if(hWndSeed){
            e_prop->Action(E_PROP_DEL_SHEET, hWndSeed);
            DestroyWindow(hWndSeed);
            hWndSeed = NULL;
         }else{
            hWndSeed = CreateDialogParam(GetHInstance(), "IDD_SEED", NULL, (DLGPROC)SeedCallbackProc, (LPARAM)this);
            if(hWndSeed){
               e_prop->Action(E_PROP_ADD_SHEET, hWndSeed);
               e_prop->Action(E_PROP_SHOW_SHEET, hWndSeed);
               SetupDialogParams();
            }
         }
      }
      return true;
   }
   return false;
}

//----------------------------

void CreateSeed(LPC_editor editor){
   C_edit_Seed *sd = new C_edit_Seed; editor->InstallPlugin(sd); sd->Release();    
}

//----------------------------