#include "all.h"
#include "common.h"

//----------------------------


enum{                         //groups
   LM_VISUAL,
   LM_LIGHT,
   LM_SOUND,
   LM_MODEL,

   LM_TARGET,
   LM_VOLUME,
   LM_SECTOR,
   LM_DUMMY,

   LM_CAMERA,
   LM_OCCLUDER,
   LM_JOINT,
   LM_USER,
                              //other (modes)
   LM_HIERARCHY,
   LM_HIDE,

   LM_LAST
};

static bool list_mode[LM_LAST] = {
   1, 1, 1, 1,
   1, 1, 1, 1,
   1, 1, 1, 1,
   1, 0,
};

#define SELECTION_NUM_TABS 4

static struct S_objsel_column{
   const char *title;
   int image;                 //one-based, 0 = no image
} objsel_list[] = {
   {""},
   {"Name"},
   {"Actor"},
   {"Script"},
};

static int objsel_list_width_defaults[SELECTION_NUM_TABS] = {
   16,
   150,
   50,
   50
};

//----------------------------

class C_edit_Selection: public C_editor_item{
   virtual const char *GetName() const{ return "Selection"; }
   LPC_editor_item e_undo;

   t_GetFrameScript *LPGetFrameScript;
   t_GetFrameActor *LPGetFrameActor;

   int objsel_list_width[SELECTION_NUM_TABS];

//----------------------------

   enum E_FRM_TYPE_INDEX{
      FI_UNKNOWN, 
      FI_VISUAL,
      FI_LIGHT,
      FI_SOUND,
      FI_MODEL,
      FI_TARGET,
      FI_SECTOR,
      FI_DUMMY,
      FI_VOLUME,
      FI_CAMERA,
      FI_USER,
      FI_JOINT,
      FI_OCCLUDER,
   };

//----------------------------
                              //phase 1 - collect one hierarchy level
   static I3DENUMRET I3DAPI dlgObjAdd(PI3D_frame frm, dword hl){
      vector<C_smart_ptr<I3D_frame> > *hier_level = (vector<C_smart_ptr<I3D_frame> >*)hl;
      hier_level->push_back(frm);
      return list_mode[LM_HIERARCHY] ? I3DENUMRET_SKIPCHILDREN : I3DENUMRET_OK;
   }

//----------------------------

   static bool cbFrameLess(PI3D_frame frm1a, PI3D_frame frm2a){
      return stricmp(frm1a->GetName(), frm2a->GetName()) < 0;
   }

//----------------------------

   bool AddFrame(HWND hwnd, PI3D_frame frm, vector<C_smart_ptr<I3D_frame> > *sel_list,
      int &focused, byte slct_tab, bool also_hidden){

      bool add = false;
      switch(frm->GetType()){
      case FRAME_VISUAL: if(list_mode[LM_VISUAL]) add = also_hidden || frm->IsOn(); break;
      case FRAME_LIGHT: if(list_mode[LM_LIGHT]) add = true; break;
      case FRAME_SOUND: if(list_mode[LM_SOUND]) add = true; break;
      case FRAME_MODEL: if(list_mode[LM_MODEL]) add = also_hidden || frm->IsOn(); break;
      case FRAME_TARGET: if(list_mode[LM_TARGET]) add = true; break;
      case FRAME_SECTOR: if(list_mode[LM_SECTOR]) add = true; break;
      case FRAME_DUMMY: if(list_mode[LM_DUMMY]) add = also_hidden || frm->IsOn(); break;
      case FRAME_VOLUME: if(list_mode[LM_VOLUME]) add = true; break;
      case FRAME_CAMERA: if(list_mode[LM_CAMERA]) add = true; break;
      case FRAME_OCCLUDER: if(list_mode[LM_OCCLUDER]) add = true; break;
      case FRAME_JOINT: if(list_mode[LM_JOINT]) add = true; break;
      case FRAME_USER: if(list_mode[LM_USER]) add = true; break;
      }
      if(add){
         E_FRM_TYPE_INDEX ti = FI_UNKNOWN;

         switch(frm->GetType()){
         case FRAME_VISUAL: ti = FI_VISUAL; break;
         case FRAME_LIGHT: ti = FI_LIGHT; break;
         case FRAME_SOUND: ti = FI_SOUND; break;
         case FRAME_MODEL: ti = FI_MODEL; break;
         case FRAME_TARGET: ti = FI_TARGET; break;
         case FRAME_SECTOR: ti = FI_SECTOR; break;
         case FRAME_USER: ti = FI_USER; break;
         case FRAME_DUMMY: ti = FI_DUMMY; break;
         case FRAME_JOINT: ti = FI_JOINT; break;
         case FRAME_VOLUME: ti = FI_VOLUME; break;
         case FRAME_CAMERA: ti = FI_CAMERA; break;
         case FRAME_OCCLUDER: ti = FI_OCCLUDER; break;
         }
         {
            char buf[256];
            for(int j=0; j<slct_tab; j++) buf[j]=' ';
            strcpy(buf+j, frm->GetName());
            int count = SendDlgItemMessage((HWND)hwnd, IDC_LIST, LVM_GETITEMCOUNT, 0, 0);

            static LVITEM lvi = { LVIF_PARAM | LVIF_STATE | LVIF_IMAGE};
            lvi.iItem = count;
            lvi.lParam = (LPARAM)frm;
            lvi.state = 0;
            lvi.iImage = ti;
                              //check if this is selected
            for(j=sel_list->size(); j--; )
               if(frm==(*sel_list)[j]) break;
            if(j != -1){
               lvi.state = LVIS_SELECTED | LVIS_FOCUSED;
               focused = count;
            }
            SendDlgItemMessage((HWND)hwnd, IDC_LIST, LVM_INSERTITEM, 0, (LPARAM)&lvi);
                              //set item's text
            {
               static LVITEM lvi = { LVIF_TEXT, 0, 1};
               lvi.iItem = count;
               lvi.pszText = buf;
               lvi.cchTextMax = strlen(buf);
               SendDlgItemMessage((HWND)hwnd, IDC_LIST, LVM_SETITEM, 0, (LPARAM)&lvi);
            }
                              //set item's sctipt
            if(LPGetFrameScript){
               const char *scr_name = (*LPGetFrameScript)(frm);
               if(scr_name){
                  static LVITEM lvi = { LVIF_TEXT, 0, 3};
                  lvi.iItem = count;
                  lvi.pszText = (char*)scr_name;
                  lvi.cchTextMax = strlen(scr_name);
                  SendDlgItemMessage((HWND)hwnd, IDC_LIST, LVM_SETITEM, 0, (LPARAM)&lvi);
               }
            }
            if(LPGetFrameActor){
               const char *cp = (*LPGetFrameActor)(frm);
               if(cp){
                  static LVITEM lvi = { LVIF_TEXT, 0, 2};
                  lvi.iItem = count;
                  lvi.pszText = (char*)cp;
                  lvi.cchTextMax = strlen(cp);
                  SendDlgItemMessage((HWND)hwnd, IDC_LIST, LVM_SETITEM, 0, (LPARAM)&lvi);
               }
            }
         }
      }
      return add;
   }

//----------------------------

   void AddLevel(PI3D_frame root, HWND hwnd, vector<C_smart_ptr<I3D_frame> > *sel_list, int &focused,
      byte slct_tab=0){

      vector<C_smart_ptr<I3D_frame> > hl;
      root->EnumFrames(dlgObjAdd, (dword)&hl, ENUMF_ALL, NULL);
                                 //sort level by name
      sort(hl.begin(), hl.end(), cbFrameLess);

      for(int i=0; i<hl.size(); i++){
         PI3D_frame frm = hl[i];
         bool add = AddFrame(hwnd, frm, sel_list, focused, slct_tab, false);

         if(add || !list_mode[LM_HIDE])
         if(list_mode[LM_HIERARCHY]){
                                 //repeat on children
            vector<C_smart_ptr<I3D_frame> > hier_level;
            frm->EnumFrames(dlgObjAdd, (dword)&hier_level,
               ENUMF_ALL, NULL);

            AddLevel(frm, hwnd, sel_list, focused, slct_tab+2);
         }
      }
   }

//----------------------------

   void AddObjsToList(PI3D_scene scene, HWND hwnd,
      vector<C_smart_ptr<I3D_frame> > *in_list, vector<C_smart_ptr<I3D_frame> > *sel_list){

      SendDlgItemMessage(hwnd, IDC_LIST, LVM_DELETEALLITEMS, 0, 0);

      int focused = -1;

      if(!in_list){           //fill-in from scene hierarchy
                              //first is primary sector
         AddFrame(hwnd, scene->GetPrimarySector(), sel_list, focused, 0, false);
                              //then any other frames
         AddLevel(scene->GetPrimarySector(), hwnd, sel_list, focused, 0);
                              //second is backdrop sector
         if(scene->GetBackdropSector()->NumChildren()){
            AddFrame(hwnd, scene->GetBackdropSector(), sel_list, focused, 0, false);
            AddLevel(scene->GetBackdropSector(), hwnd, sel_list, focused, 0);
         }
      }else{
                              //fill-in from provided list
         for(int i=0; i<in_list->size(); i++){
            AddFrame(hwnd, (*in_list)[i], sel_list, focused, 0, true);
         }
      }

      if(focused!=-1)
         SendDlgItemMessage(hwnd, IDC_LIST, LVM_ENSUREVISIBLE, focused, false);
   }

//----------------------------

   static void ShowGroupButtons(HWND hwnd, bool enable_hierarchy){

      CheckDlgButton(hwnd, IDC_CHECK_VIS, list_mode[LM_VISUAL]);
      CheckDlgButton(hwnd, IDC_CHECK_LIGHT, list_mode[LM_LIGHT]);
      CheckDlgButton(hwnd, IDC_CHECK_MODEL, list_mode[LM_MODEL]);
      CheckDlgButton(hwnd, IDC_CHECK_SND, list_mode[LM_SOUND]);
      CheckDlgButton(hwnd, IDC_CHECK_TRGT, list_mode[LM_TARGET]);
      CheckDlgButton(hwnd, IDC_CHECK_OCC, list_mode[LM_OCCLUDER]);
      CheckDlgButton(hwnd, IDC_CHECK_VOL, list_mode[LM_VOLUME]);
      CheckDlgButton(hwnd, IDC_CHECK_DUM, list_mode[LM_DUMMY]);
      CheckDlgButton(hwnd, IDC_CHECK_SCT, list_mode[LM_SECTOR]);
      CheckDlgButton(hwnd, IDC_CHECK_CAM, list_mode[LM_CAMERA]);
      CheckDlgButton(hwnd, IDC_CHECK_JOINT, list_mode[LM_JOINT]);
      CheckDlgButton(hwnd, IDC_CHECK_USER, list_mode[LM_USER]);
      if(enable_hierarchy){
         EnableWindow(GetDlgItem(hwnd, IDC_CHECK_HIER), true);
         CheckDlgButton(hwnd, IDC_CHECK_HIER, list_mode[LM_HIERARCHY]);
      }else{
         EnableWindow(GetDlgItem(hwnd, IDC_CHECK_HIER), false);
         CheckDlgButton(hwnd, IDC_CHECK_HIER, false);
      }
      CheckDlgButton(hwnd, IDC_CHECK_HIDE, list_mode[LM_HIDE]);
      EnableWindow(GetDlgItem(hwnd, IDC_CHECK_HIDE), list_mode[LM_HIERARCHY]);
   }

//----------------------------

   static bool StringMatch(const char *cp1, int l1, const char *cp2, int l2){

      if(!l1 || !l2) return (l1==l2);
      for(; ; ++cp1, ++cp2){
         switch(*cp2){
         case 0: return !(*cp1);
         case '*': return true;
         case '?':
            if(!*cp1) return false;
            break;
         default: if(*cp1 != *cp2) return false;
         }
      }
   }

//----------------------------

   struct S_sel_help{
      LPIGraph igraph;
      PI3D_scene scene;
      vector<C_smart_ptr<I3D_frame> > *in_list;
      vector<C_smart_ptr<I3D_frame> > *out_list;
      C_edit_Selection *e_slct;
      S_sel_help(): in_list(NULL){}
   };

//----------------------------

   static BOOL CALLBACK dlgObjSel(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

      switch(uMsg){
      case WM_INITDIALOG:
         {
            S_sel_help *hlp = (S_sel_help*)lParam;
            InitDlg(hlp->igraph, hwnd);
            ShowGroupButtons(hwnd, (hlp->in_list==NULL));
      
            vector<C_smart_ptr<I3D_frame> > *sel_list = hlp->out_list;
            SetWindowLong(hwnd, GWL_USERDATA, (LPARAM)hlp);

            InitCommonControls();
            HINSTANCE hi = LoadLibraryEx("IEditor.dll", NULL, DONT_RESOLVE_DLL_REFERENCES);
            HIMAGELIST il;
            il = ImageList_LoadBitmap(hi, "objselect_list", 12, 1, 0x808000);
            FreeLibrary(hi);
            SendDlgItemMessage(hwnd, IDC_LIST, LVM_SETIMAGELIST, LVSIL_SMALL, (LPARAM)il);

            for(int i=0; i < (sizeof(objsel_list)/sizeof(S_objsel_column)); i++){
               LVCOLUMN lvc;
               memset(&lvc, 0, sizeof(lvc));
               lvc.mask = LVCF_WIDTH;
               lvc.cx = hlp->e_slct->objsel_list_width[i];
               if(objsel_list[i].title){
                  lvc.pszText = (char*)objsel_list[i].title;
                  lvc.cchTextMax = strlen(objsel_list[i].title);
                  lvc.mask |= LVCF_TEXT;
               }
               if(objsel_list[i].image){
                  lvc.mask |= LVCFMT_IMAGE;
                  lvc.mask |= LVCF_FMT;
                  lvc.fmt = LVCFMT_IMAGE;
                  lvc.iImage = 2;
               }
               SendDlgItemMessage(hwnd, IDC_LIST, LVM_INSERTCOLUMN, i , (dword)&lvc);
            }
            SendDlgItemMessage(hwnd, IDC_LIST, LVM_SETEXTENDEDLISTVIEWSTYLE, 
               LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES, 
               LVS_EX_FULLROWSELECT/* | LVS_EX_GRIDLINES*/);
               //LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES);

            /*
            SendDlgItemMessage(hwnd, IDC_LIST, LVM_SETEXTENDEDLISTVIEWSTYLE, 
               LVS_EX_SUBITEMIMAGES, LVS_EX_SUBITEMIMAGES);
               */
            SetDlgItemInt(hwnd, IDC_NUM_SEL, sel_list->size(), false);

            hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, sel_list);

            ShowWindow(hwnd, SW_SHOW);
         }
         return 1;

      case WM_NOTIFY:
         {
            NMHDR *nm = (NMHDR*)lParam;
            switch(wParam){
            case IDC_LIST:
               switch(nm->code){
                  /*
               case NM_CUSTOMDRAW:
                  {
                     NMLVCUSTOMDRAW *cd = (NMLVCUSTOMDRAW*)lParam;
                     //SetBkColor(cd->nmcd.hdc, 0);
                     //Rectangle(cd->nmcd.hdc, cd->nmcd.rc.left, cd->nmcd.rc.top, cd->nmcd.rc.right, cd->nmcd.rc.bottom);
                     HBRUSH hbr = CreateSolidBrush(0);
                     FillRect(cd->nmcd.hdc, &cd->nmcd.rc, hbr);
                     DeleteObject(hbr);
                  }
                  break;
                  */
               case LVN_COLUMNCLICK:
                  {
                     struct S_hlp1{
                        int sort_type;
                        C_edit_Selection *e_slct;
                        t_GetFrameScript *LPGetFrameScript;
                        t_GetFrameActor *LPGetFrameActor;

                        static int CALLBACK CompareFunc(LPARAM lParam1, LPARAM lParam2,
                           LPARAM lParamSort){

                           PI3D_frame f1 = (PI3D_frame)lParam1;
                           PI3D_frame f2 = (PI3D_frame)lParam2;
                           S_hlp1 *hp = (S_hlp1*)lParamSort;
                           switch(hp->sort_type){
                           case 0:     //type
                              {
                                 dword ft1 = f1->GetType();
                                 dword ft2 = f2->GetType();
                                 return (ft1<ft2 ? -1 : ft1==ft2 ? 0 : 1);
                              }
                              break;
                           case 2:     //actor
                              if(hp->LPGetFrameActor){
                                 const char *cp1 = (*(hp->LPGetFrameActor))(f1);
                                 const char *cp2 = (*(hp->LPGetFrameActor))(f2);
                                 if(!cp1 && !cp2) return stricmp(f1->GetName(), f2->GetName());
                                 if(!cp1) return -1;
                                 if(!cp2) return 1;
                                 int i = stricmp(cp1, cp2);
                                 if(!i) i = stricmp(f1->GetName(), f2->GetName());
                                 return i;
                              }
                              break;
                           case 3:     //script
                              if(hp->LPGetFrameScript){
                                 const char *scr_1 = (*(hp->LPGetFrameScript))(f1);
                                 const char *scr_2 = (*(hp->LPGetFrameScript))(f2);
                                 if(!scr_1 && !scr_2) return stricmp(f1->GetName(), f2->GetName());
                                 if(!scr_1) return -1;
                                 if(!scr_2) return 1;
                                 int i = stricmp(scr_1, scr_2);
                                 if(!i) i = stricmp(f1->GetName(), f2->GetName());
                                 return i;
                              }
                              break;
                           }
                           return 0;
                        }
                     } hlp1;
                     NMLISTVIEW *nml = (NMLISTVIEW*)lParam;
                     if(nml->iSubItem==1){   //sort by frame
                        S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                        hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
                     }else{
                        S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                        hlp1.sort_type = nml->iSubItem;
                        hlp1.e_slct = hlp->e_slct;
                        hlp1.LPGetFrameScript = hlp->e_slct->LPGetFrameScript;
                        hlp1.LPGetFrameActor = hlp->e_slct->LPGetFrameActor;

                        SendDlgItemMessage(hwnd, IDC_LIST, LVM_SORTITEMS, (LPARAM)&hlp1,
                           (dword)S_hlp1::CompareFunc);
                     }
                  }
                  break;
               }
               break;
            }
         }
         break;

      case WM_COMMAND:

         switch(HIWORD(wParam)){

         case EN_CHANGE:
            switch(LOWORD(wParam)){
            case IDC_INC_SEL:
               {
                  char buf[256];
                  SendDlgItemMessage(hwnd, IDC_INC_SEL, WM_GETTEXT, sizeof(buf)-1, (LPARAM)buf);
                  int len = strlen(buf);
                  if(!len)
                     break;
                              //space at the end is special, if present, the text typed
                              // is matched exactly (without the space), otherwise
                              // wild-cards are used
                  if(buf[len-1] == ' ')
                     buf[len-1] = 0;
                  else
                     buf[len++] = '*';
                  bool model_search = (buf[0]=='.');
                              //change contents of selection
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->out_list->clear();
                  int count = SendDlgItemMessage(hwnd, IDC_LIST, LVM_GETITEMCOUNT, 0, 0);
                  LVITEM lvi;
                  memset(&lvi, 0, sizeof(lvi));
                  lvi.mask = LVIF_PARAM | LVIF_STATE;
                  lvi.stateMask = LVIS_SELECTED | LVIS_FOCUSED;
                  int focus_index = -1;
                  int i;
                  for(i=0; i<count; i++){
                     lvi.iItem = i;
                     SendDlgItemMessage(hwnd, IDC_LIST, LVM_GETITEM, 0, (LPARAM)&lvi);
                     PI3D_frame frm = (PI3D_frame)lvi.lParam;
                     const char *fn = frm->GetName();
                     bool b;
                     if(model_search){
                        const char *mn = strchr(fn, '.');
                        if(mn)
                           fn = mn+1;
                        b = StringMatch(fn, strlen(fn), buf+1, len-1);
                     }else
                        b = StringMatch(fn, strlen(fn), buf, len);
                     bool curr_state = (lvi.state&LVIS_SELECTED);
                     if(b != curr_state){
                        lvi.state &= ~(LVIS_SELECTED | LVIS_FOCUSED);
                        if(b){
                           lvi.state |= LVIS_SELECTED;
                           if(focus_index==-1){
                              lvi.state |= LVIS_FOCUSED;
                              focus_index = i;
                           }
                        }
                        SendDlgItemMessage(hwnd, IDC_LIST, LVM_SETITEMSTATE, i, (LPARAM)&lvi);
                     }else
                     if(curr_state && focus_index==-1){
                        focus_index = i;
                     }
                     if(b) hlp->out_list->push_back(frm);
                  }
                  if(focus_index!=-1)
                     SendDlgItemMessage(hwnd, IDC_LIST, LVM_ENSUREVISIBLE, focus_index, false);
                  SetDlgItemInt(hwnd, IDC_NUM_SEL, hlp->out_list->size(), false);
               }
               break;
            }
            break;

         case BN_CLICKED:
            switch(LOWORD(wParam)){

            case IDC_BUTTON_GRP_ALL:
               {
                  for(int i=0; i<LM_HIERARCHY; i++) list_mode[i] = true;
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  ShowGroupButtons(hwnd, (hlp->in_list==NULL));
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;

            case IDC_BUTTON_GRP_NONE:
               {
                  for(int i=0; i<LM_HIERARCHY; i++) list_mode[i] = false;
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  ShowGroupButtons(hwnd, (hlp->in_list==NULL));
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;

            case IDC_BUTTON_SEL_ALL:
            case IDC_BUTTON_SEL_NONE:
            case IDC_BUTTON_SEL_INV:
               {
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->out_list->clear();
                  int count = SendDlgItemMessage(hwnd, IDC_LIST, LVM_GETITEMCOUNT, 0, 0);
                  LVITEM lvi;
                  memset(&lvi, 0, sizeof(lvi));
                  lvi.mask = LVIF_PARAM | LVIF_STATE;
                  lvi.stateMask = LVIS_SELECTED | LVIS_FOCUSED;
                  for(int i=0; i<count; i++){
                     lvi.iItem = i;
                     SendDlgItemMessage(hwnd, IDC_LIST, LVM_GETITEM, 0, (LPARAM)&lvi);
                     PI3D_frame frm = (PI3D_frame)lvi.lParam;
                     switch(LOWORD(wParam)){
                     case IDC_BUTTON_SEL_ALL:
                        lvi.state |= LVIS_SELECTED;
                        break;
                     case IDC_BUTTON_SEL_NONE:
                        lvi.state &= ~LVIS_SELECTED;
                        break;
                     case IDC_BUTTON_SEL_INV:
                        lvi.state ^= LVIS_SELECTED;
                        break;
                     }
                     SendDlgItemMessage(hwnd, IDC_LIST, LVM_SETITEMSTATE, i, (LPARAM)&lvi);
                     if(lvi.state&LVIS_SELECTED)
                        hlp->out_list->push_back(frm);
                  }
                  SetDlgItemInt(hwnd, IDC_NUM_SEL, hlp->out_list->size(), false);
               }
               break;

            case IDC_CHECK_VIS:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_VISUAL]=IsDlgButtonChecked(hwnd, IDC_CHECK_VIS);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_LIGHT:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_LIGHT]=IsDlgButtonChecked(hwnd, IDC_CHECK_LIGHT);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_MODEL:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_MODEL]=IsDlgButtonChecked(hwnd, IDC_CHECK_MODEL);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_SND:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_SOUND]=IsDlgButtonChecked(hwnd, IDC_CHECK_SND);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_TRGT:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_TARGET]=IsDlgButtonChecked(hwnd, IDC_CHECK_TRGT);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_OCC:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_OCCLUDER]=IsDlgButtonChecked(hwnd, IDC_CHECK_OCC);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_VOL:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_VOLUME]=IsDlgButtonChecked(hwnd, IDC_CHECK_VOL);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_DUM:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_DUMMY]=IsDlgButtonChecked(hwnd, IDC_CHECK_DUM);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_SCT:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_SECTOR]=IsDlgButtonChecked(hwnd, IDC_CHECK_SCT);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_CAM:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_CAMERA]=IsDlgButtonChecked(hwnd, IDC_CHECK_CAM);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
            case IDC_CHECK_JOINT:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_JOINT]=IsDlgButtonChecked(hwnd, IDC_CHECK_JOINT);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
            case IDC_CHECK_USER:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_USER]=IsDlgButtonChecked(hwnd, IDC_CHECK_USER);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
            case IDC_CHECK_HIER:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_HIERARCHY]=IsDlgButtonChecked(hwnd, IDC_CHECK_HIER);
                  EnableWindow(GetDlgItem(hwnd, IDC_CHECK_HIDE), list_mode[LM_HIERARCHY]);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;
            case IDC_CHECK_HIDE:
               if(HIWORD(wParam)==BN_CLICKED){
                  list_mode[LM_HIDE]=IsDlgButtonChecked(hwnd, IDC_CHECK_HIDE);
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                  hlp->e_slct->AddObjsToList(hlp->scene, hwnd, hlp->in_list, hlp->out_list);
               }
               break;

            case IDCANCEL: EndDialog(hwnd, 0); break;

            case IDOK:
               {
                  S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
                                    //collect selection and select
                  hlp->out_list->clear();
                  int count = SendDlgItemMessage(hwnd, IDC_LIST, LVM_GETITEMCOUNT, 0, 0);
                  LVITEM lvi;
                  memset(&lvi, 0, sizeof(lvi));
                  lvi.mask = LVIF_PARAM | LVIF_STATE;
                  lvi.stateMask = LVIS_SELECTED;
                  for(int i=0; i<count; i++){
                     lvi.iItem = i;
                     SendDlgItemMessage(hwnd, IDC_LIST, LVM_GETITEM, 0, (LPARAM)&lvi);
                     PI3D_frame frm = (PI3D_frame)lvi.lParam;
                     if(lvi.state&LVIS_SELECTED) hlp->out_list->push_back(frm);
                  }
                  EndDialog(hwnd, 1);
               }
               break;
            }
            break;
         }
         break;

      case WM_DESTROY:
         {
            S_sel_help *hlp = (S_sel_help*)GetWindowLong(hwnd, GWL_USERDATA);
            for(int i=0; i < (sizeof(objsel_list)/sizeof(S_objsel_column)); i++){
               hlp->e_slct->objsel_list_width[i] =
                  SendDlgItemMessage(hwnd, IDC_LIST, LVM_GETCOLUMNWIDTH, i , 0);
            }
         }
         break;
      }
      return 0;
   }

   struct S_flash_item{
      C_smart_ptr<I3D_frame> frm;
      int count_down;
      bool persistent;

      S_flash_item():
         persistent(false)
      {}
      S_flash_item(PI3D_frame f1):
         frm(f1), persistent(false)
      {
         ResetCount();
      }
      S_flash_item(const S_flash_item &fi):
         frm(fi.frm), persistent(fi.persistent)
      {
         ResetCount();
      }
      S_flash_item &operator =(const S_flash_item &fi){
         frm = fi.frm;
         count_down = fi.count_down;
         persistent = fi.persistent;
         return (*this);
      }
      void ResetCount(){
         count_down = 250;
      }
   };

   vector<C_smart_ptr<I3D_frame> > sel_list;
   vector<S_flash_item> flash_list;
   int sb_index;              //index in status bar

                              //list of plugins to which we send notifications
   vector<pair<LPC_editor_item, dword> > notify_list;

   void SendNotify(){
      for(int i=0; i<notify_list.size(); i++){
         notify_list[i].first->Action(notify_list[i].second, &sel_list);
      }
   }

   void Message(){
      ed->Message(C_fstring("Selected: %i", sel_list.size()), sb_index);

      bool any_sel = (sel_list.size()!=0);
      ed->EnableMenu(this, E_SELECTION_CLEAR2, any_sel);
      ed->EnableMenu(this, E_SELECTION_PARENTS, any_sel);
      ed->EnableMenu(this, E_SELECTION_CHILDREN, any_sel);
   }

   bool draw_links;
   /*
   C_smart_ptr<C_table> config_tab;
   enum{
      TAB_V_CFG_COLOR_SEL,
      TAB_V_CFG_COLOR_PIVOT,
      TAB_V_CFG_COLOR_AXE,
      TAB_V_CFG_COLOR_FLASH1,
      TAB_V_CFG_COLOR_FLASH2,

      TAB_F_CFG_ALPHA,
      TAB_B_CFG_DRAWLINKS,
   };

   static void cbTabCfg(LPC_table tab, dword msg, dword cb_user, dword prm2, dword prm3){
      switch(msg){
      case TCM_MODIFY:
         switch(prm2){
         case TAB_V_CFG_COLOR_SEL: 
            {
               C_edit_Selection *ei = (C_edit_Selection*)cb_user;
               ei->ed->GetScene()->Render(0);
               ei->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
            }
            break;
         }
         break;
      }
   }
   static C_table_template *GetTemplate(){
                           //template
      static C_table_template templ_config={
         // {A418B6A0-E54A-11d3-92AE-0000B49B8E8B}
         { 0xa418b6a0, 0xe54a, 0x11d3, { 0x92, 0xae, 0x0, 0x0, 0xb4, 0x9b, 0x8e, 0x8b } },
         "Selection configuration",
         NULL,
         {
            {TE_BRANCH, 0, "Colors", 6},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_SEL, "Selection"},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_PIVOT, "Pivot"},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_AXE, "Axes"},
               {TE_FLOAT, TAB_F_CFG_ALPHA, "Selection Alpha", 0, 1},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_FLASH1, "Flashing 1"},
               {TE_COLOR_VECTOR, TAB_V_CFG_COLOR_FLASH2, "Flashing 2"},
            {TE_BOOL, TAB_B_CFG_DRAWLINKS, "Draw links", 0, 0, 1},
            {TE_NULL}
         }
      };
      return &templ_config;
   }
   */

//----------------------------
// helper function - build list of names separated by '\0',
// last name is followed by two '\0' characters
   static void BuildNameList(const vector<C_smart_ptr<I3D_frame> > &frm_list, char **buf, dword *len){
      int i, nlen;
      for(i=0, nlen=0; i<frm_list.size(); i++){
         nlen += strlen(frm_list[i]->GetName()) + 1;
      }
      *buf = new char[nlen + 1];
      for(i=0, nlen=0; i<frm_list.size(); i++){
         const char *cp = frm_list[i]->GetName();
         strcpy(&(*buf)[nlen], cp);
         nlen += strlen(cp) + 1;
      }
      (*buf)[nlen] = 0;
      *len = nlen+1;
   }

public:
   C_edit_Selection():
      e_undo(NULL),
      LPGetFrameScript(NULL),
      LPGetFrameActor(NULL),
      draw_links(true)
   {
      memcpy(objsel_list_width, objsel_list_width_defaults, sizeof(objsel_list_width));

      /*
      config_tab = CreateTable(); 
      config_tab->Open((dword)GetTemplate(), TABOPEN_TEMPLATE);
                              //setup defaults
      config_tab->ItemV(TAB_V_CFG_COLOR_SEL) = S_vector(1, 1, 1);
      config_tab->ItemV(TAB_V_CFG_COLOR_PIVOT) = S_vector(1, 0, 0);
      config_tab->ItemV(TAB_V_CFG_COLOR_AXE) = S_vector(1, 1, 0);
      config_tab->ItemV(TAB_V_CFG_COLOR_FLASH1) = S_vector(0, 1, 0);
      config_tab->ItemV(TAB_V_CFG_COLOR_FLASH2) = S_vector(1, 0, 0);

      config_tab->ItemF(TAB_F_CFG_ALPHA) = 1.0f;
      config_tab->ItemB(TAB_B_CFG_DRAWLINKS) = true;

      config_tab->Release();
      */
   }

   virtual bool Init(){
      e_undo = ed->FindPlugin("Undo");
      if(!e_undo){
         e_undo = NULL;
      }
      e_undo->AddRef();

      ed->AddShortcut(this, E_SELECTION_BROWSE, "%10 &Edit\\%40 %i Select by &name\tN", K_N, 0);
      ed->AddShortcut(this, E_SELECTION_CLEAR2, "&Edit\\%40 Clear selection\tCtrl+N", K_N, SKEY_CTRL);
      ed->AddShortcut(this, E_SELECTION_INVERT, "&Edit\\%40 Invert selection", 0, 0);
      ed->AddShortcut(this, E_SELECTION_PARENTS, "&Edit\\%40 Select parent\tPgUp", K_PAGEUP, 0);
      ed->AddShortcut(this, E_SELECTION_CHILDREN, "&Edit\\%40 Select children\tPgDn", K_PAGEDOWN, 0);
      ed->AddShortcut(this, E_SELECTION_TOGGLE_LINK, "&Edit\\%40 Draw &links\tCtrl+Alt+L", K_L, SKEY_CTRL | SKEY_ALT);
      //ed->AddShortcut(this, E_SELECTION_CONFIG, "&Edit\\%40 Selection confi&g", 0, 0);
      ed->AddShortcut(this, E_SELECTION_CLEAR_FLASH, "%30 &Debug\\%51 Clear flashing", 0, 0);
      ed->AddShortcut(this, E_SELECTION_SEL_FLASH_PREV, "%30 &Debug\\%52 Select prev flashing\tAlt+[", K_5B, SKEY_ALT);
      ed->AddShortcut(this, E_SELECTION_SEL_FLASH_NEXT, "%30 &Debug\\%52 Select next flashing\tAlt+]", K_5D, SKEY_ALT);

      ed->CheckMenu (this, E_SELECTION_TOGGLE_LINK, draw_links);

      ed->RegForRender(this, 1100);
      sb_index = ed->CreateStatusIndex(80);
      return true;
   
   }
   virtual void Close(){ 
      sel_list.clear();
      flash_list.clear();
      if(e_undo) e_undo->Release();
      e_undo = NULL;

      int i;

      for(i=notify_list.size(); i--; ) notify_list[i].first->Release();
      notify_list.clear();
   }

   virtual dword Action(int id, void *context){
      switch(id){

      case E_SELECTION_GETVECTOR:
         return (dword)&sel_list;

      case E_SELECTION_CLEAR:
         if(sel_list.size())
         {
                              //save undo info, reverse = E_SELECTION_ADD_FRAMES
            S_edit_Undo_2 eu(GetName(), E_SELECTION_ADD_FRAMES);
            BuildNameList(sel_list, (char**)&eu.data, &eu.data_size);
            e_undo->Action(E_UNDO_SAVE, &eu);
            delete[] (char*)eu.data;
                              //clear selection
            sel_list.clear();
            Message();
            SendNotify();
         }
         break;

      case E_SELECTION_INVERT:
         {
                              //save undo info 
            const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
            if(top_name && !strcmp(top_name, GetName()) && e_undo->Action(E_UNDO_GET_TOP_ACTION)==E_SELECTION_INVERT){
               e_undo->Action(E_UNDO_POP_TOP);
            }else
               e_undo->Action(E_UNDO_SAVE, &S_edit_Undo_2(GetName(), E_SELECTION_INVERT));
                              //invert selection
            vector<C_smart_ptr<I3D_frame> > all_list;
            struct S_hlp{
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                  vector<C_smart_ptr<I3D_frame> > &sel_list = *(vector<C_smart_ptr<I3D_frame> >*)c;
                  sel_list.push_back(frm);
                  return I3DENUMRET_OK;
               }
            };
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&all_list, ENUMF_ALL | ENUMF_BACKDROP);
            for(int i=all_list.size(); i--; ){
               if(find(sel_list.begin(), sel_list.end(), all_list[i])!=sel_list.end()){
                  all_list[i] = all_list.back();
                  all_list.pop_back();
               }
            }
            sel_list = all_list;
            Message();
            SendNotify();
         }
         break;

      case E_SELECTION_ADD_FRAMES:
         {
            const char *cp = (const char*)context;
            while(*cp){
               PI3D_frame frm = ed->GetScene()->FindFrame(cp, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  if(FindPointerIndex((void**)sel_list.begin(), sel_list.size(), frm) == -1)
                     sel_list.push_back(frm);
               }
               cp += strlen(cp)+1;
            }
                              //save undo info, reverse = E_SELECTION_DEL_FRAMES
            S_edit_Undo_2 eu(GetName(), E_SELECTION_DEL_FRAMES, context, cp-(char*)context+1);
            e_undo->Action(E_UNDO_SAVE, &eu);

            Message();
            SendNotify();
         }
         break;

      case E_SELECTION_DEL_FRAMES:
         {
            const char *cp = (const char*)context;
            while(*cp){
               PI3D_frame frm = ed->GetScene()->FindFrame(cp, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  for(int j=sel_list.size(); j--; )
                     if(sel_list[j]==frm) break;
                  if(j!=-1){
                     sel_list[j] = sel_list.back();
                     sel_list.pop_back();
                  }
               }
               cp += strlen(cp)+1;
            }
                              //save undo info, reverse = E_SELECTION_ADD_FRAMES
            S_edit_Undo_2 eu(GetName(), E_SELECTION_ADD_FRAMES, context, cp-(char*)context+1);
            e_undo->Action(E_UNDO_SAVE, &eu);

            Message();
            SendNotify();
         }
         break;

      case E_SELECTION_GET_SINGLE:
         return (sel_list.size()==1) ? (dword)(PI3D_frame)sel_list.front() : NULL;

      case E_SELECTION_ADD_FRAME:
         {
            PI3D_frame frm = (PI3D_frame)context;
            if(!frm){
               ed->Message("Selection(E_SELECTION_ADD_FRAME): can't add NULL frame");
               break;
            }
            int j = FindPointerIndex((void**)sel_list.begin(), sel_list.size(), frm);
            if(j==-1){
               sel_list.push_back(frm);
                              //save undo info, reverse = E_SELECTION_DEL_FRAMES
               int len = strlen(frm->GetName());
               char *buf = new char[len+2];
               strcpy(buf, frm->GetName());
               buf[len++] = 0;
               buf[len++] = 0;
               S_edit_Undo_2 eu(GetName(), E_SELECTION_DEL_FRAMES, buf, len);
               e_undo->Action(E_UNDO_SAVE, &eu);
               delete[] buf;

               Message();

               SendNotify();
            }else
               ed->Message("Selection(E_SELECTION_ADD_FRAME): frame already in selection!");
         }
         break;

      case E_SELECTION_DEL_FRAME:
         {
            PI3D_frame frm = (PI3D_frame)context;
            int j = FindPointerIndex((void**)sel_list.begin(), sel_list.size(), frm);
            if(j!=-1){
               sel_list[j] = sel_list.back(); sel_list.pop_back();
                              //save undo info, reverse = E_SELECTION_ADD_FRAMES
               int len = strlen(frm->GetName());
               char *buf = new char[len+2];
               strcpy(buf, frm->GetName());
               buf[len++] = 0;
               buf[len++] = 0;
               S_edit_Undo_2 eu(GetName(), E_SELECTION_ADD_FRAMES, buf, len);
               e_undo->Action(E_UNDO_SAVE, &eu);
               delete[] buf;

               Message();
               SendNotify();
               return true;
            }
            return false;
         }
         break;
                              //menu items:
      case E_SELECTION_BROWSE:
         {
            int i;
            vector<C_smart_ptr<I3D_frame> > new_list;
            new_list = sel_list;

            S_sel_help hlp;
            hlp.igraph = ed->GetIGraph();
            hlp.scene = ed->GetScene();
            hlp.out_list = &new_list;
            hlp.e_slct = this;
            i = DialogBoxParam(GetHInstance(), "IDD_OBJSELECT", (HWND)ed->GetIGraph()->GetHWND(),
               (DLGPROC)dlgObjSel, (LPARAM)&hlp);
            if(i){            //unselect old and select new
                              //clear current selection
               Action(E_SELECTION_CLEAR, 0);

                              //save undo info, reverse = E_SELECTION_CLEAR
               S_edit_Undo_2 eu(GetName(), E_SELECTION_CLEAR);
               e_undo->Action(E_UNDO_SAVE, &eu);

               sel_list = new_list;

               Message();
               SendNotify();
            }
         }
         break;

      case E_SELECTION_PROMPT:
         {
            vector<PI3D_frame> *prompt_list = (vector<PI3D_frame>*)context;
            vector<C_smart_ptr<I3D_frame> > tmp_list;
            tmp_list.reserve(prompt_list->size());
            for(int i=prompt_list->size(); i--; )
               tmp_list.push_back((*prompt_list)[i]);
            vector<C_smart_ptr<I3D_frame> > new_list;

            S_sel_help hlp;
            hlp.igraph = ed->GetIGraph();
            hlp.scene = ed->GetScene();
            hlp.in_list = &tmp_list;
            hlp.out_list = &new_list;
            hlp.e_slct = this;

            bool b = DialogBoxParam(GetHInstance(), "IDD_OBJSELECT", (HWND)ed->GetIGraph()->GetHWND(),
               (DLGPROC)dlgObjSel, (LPARAM)&hlp);
            if(b){
               prompt_list->clear();
               prompt_list->reserve(new_list.size());
               for(i=new_list.size(); i--; )
                  prompt_list->push_back(new_list[i]);
            }
            return b;
         }
         break;

      case E_SELECTION_CLEAR2:
         if(sel_list.size()){
            Action(E_SELECTION_CLEAR, 0);
            ed->Message("Selection cleared");
         }
         break;

      case E_SELECTION_CLEAR_FLASH:
         flash_list.clear();
         break;

      case E_SELECTION_PARENTS:
         {
            vector<C_smart_ptr<I3D_frame> > new_list;
            for(int i=sel_list.size(); i--; ){
               PI3D_frame prnt = sel_list[i]->GetParent();
               if(prnt){
                  for(int j=new_list.size(); j--; ) if(new_list[j]==prnt) break;
                  if(j==-1) new_list.push_back(prnt);
               }
            }
                              //clear current selection
            Action(E_SELECTION_CLEAR, 0);

                              //save undo info, reverse = E_SELECTION_CLEAR
            S_edit_Undo_2 eu(GetName(), E_SELECTION_CLEAR);
            e_undo->Action(E_UNDO_SAVE, &eu);

            sel_list = new_list;

            Message();
            SendNotify();
         }
         break;

      case E_SELECTION_CHILDREN:
         {
            vector<C_smart_ptr<I3D_frame> > new_list;
            for(int i=sel_list.size(); i--; ){
               for(int ii=sel_list[i]->NumChildren(); ii--; ){
                  PI3D_frame chld = sel_list[i]->GetChild(ii);
                  for(int j=new_list.size(); j--; ) if(new_list[j]==chld) break;
                  if(j==-1) new_list.push_back(chld);
               }
            }
                              //clear current selection
            Action(E_SELECTION_CLEAR, 0);

                              //save undo info, reverse = E_SELECTION_CLEAR
            S_edit_Undo_2 eu(GetName(), E_SELECTION_CLEAR);
            e_undo->Action(E_UNDO_SAVE, &eu);

            sel_list = new_list;

            Message();
            SendNotify();
         }
         break;

      case E_SELECTION_SEL_FLASH_NEXT:
      case E_SELECTION_SEL_FLASH_PREV:
         {
            if(!flash_list.size()){
               ed->Message("Flash list is empty");
               break;
            }
            PI3D_frame curr = (sel_list.size()==1) ? sel_list.front() : NULL;
            for(int i=flash_list.size(); i--; ){
               if(flash_list[i].frm==curr)
                  break;
            }
            Action(E_SELECTION_CLEAR, 0);
                              //save undo info, reverse = E_SELECTION_CLEAR
            S_edit_Undo_2 eu(GetName(), E_SELECTION_CLEAR);
            e_undo->Action(E_UNDO_SAVE, &eu);

            switch(id){
            case E_SELECTION_SEL_FLASH_NEXT: i = (i+1)%flash_list.size(); break;
            case E_SELECTION_SEL_FLASH_PREV: i = (i+flash_list.size()-1)%flash_list.size(); break;
            default: assert(0);
            }
            sel_list.push_back(flash_list[i].frm);
            ed->Message(C_fstring("Selected flashing frame: '%s'", sel_list.front()->GetName()));

            Message();
            SendNotify();
         }
         break;

      case E_SELECTION_FLASH_FRAME:
      case E_SELECTION_BLINK_FRAME:
         {
            PI3D_frame frm = (PI3D_frame)context;
            for(int j=flash_list.size(); j--; )
               if(flash_list[j].frm==frm) break;
            if(j==-1){
               flash_list.push_back(S_flash_item(frm));
               if(id==E_SELECTION_BLINK_FRAME)
                  flash_list.back().persistent = true;
            }else
               flash_list[j].ResetCount();
         }
         break;

      case E_SELECTION_TOGGLE_LINK:
         {
            draw_links = !draw_links;
            ed->CheckMenu (this, E_SELECTION_TOGGLE_LINK, draw_links);
         }
         break;

      case EDIT_ACTION_MISSION_SAVE_1:
         {
                              //save current selection
            if(sel_list.size()){
               C_chunk &ck = *(C_chunk*)context;
               ck.WAscend(CT_EDITOR_PLUGIN);
               ck.WStringChunk(CT_NAME, GetName());
               for(int i=0; i<sel_list.size(); i++) 
                  ck.WStringChunk(CT_NAME, sel_list[i]->GetName());
               ck.Descend();
            }
         }
         break;

      case EDIT_ACTION_MISSION_LOAD:
         {
            sel_list.clear();
            C_chunk &ck = *(C_chunk*)context;
            while(ck.Size()){
               CK_TYPE t;
               if(!ck.RAscend(t)) break;
               switch(t){
               case CT_NAME:
                  {
                     C_string str = ck.RStringChunk();
                     PI3D_frame frm = ed->GetScene()->FindFrame(str, ENUMF_ALL | ENUMF_BACKDROP);
                     if(frm) 
                        sel_list.push_back(frm);
                  }
                  break;
               default: ck.Descend();
               }
            }
         }
         break;

      case E_SELECTION_SET_GET_SCR:
         LPGetFrameScript = (t_GetFrameScript*)context;
         break;

      case E_SELECTION_SET_GET_ACT:
         LPGetFrameActor = (t_GetFrameActor*)context;
         break;

         /*
      case E_SELECTION_CONFIG:
         {
            config_tab->Edit(GetTemplate(), ed->GetIGraph()->GetHWND(), cbTabCfg,
               (dword)this, 
               TABEDIT_IMPORT | TABEDIT_EXPORT | TABEDIT_INFO,
               NULL, NULL);
         }
         break;
         */

      case EDIT_ACTION_BEFORE_FREE:
         {
            sel_list.clear();
            flash_list.clear();
            Message();
         }
         break;

      case EDIT_ACTION_AFTER_LOAD:

         Message();
         SendNotify();
         break;

      case E_SELECTION_ADD_NOTIFY:
         {
            const pair<LPC_editor_item, dword> &np = *(pair<LPC_editor_item, dword>*)context;
            for(int i=notify_list.size(); i--; )
               if(notify_list[i].first==np.first) break;
            if(i==-1){
               notify_list.push_back(np);
               np.first->AddRef();
               return true;
            }
         }
         break;

      case E_SELECTION_DEL_NOTIFY:
         {
            LPC_editor_item ei = (LPC_editor_item)context;
            for(int i=notify_list.size(); i--; )
               if(notify_list[i].first==ei) break;
            if(i!=-1){
               ei->Release();
               notify_list[i] = notify_list.back();
               notify_list.pop_back();
            }
         }
         break;

      case EDIT_ACTION_FRAME_DELETE:
         {
                              //check if this frame (or its children) are in our flash list
            PI3D_frame frm = (PI3D_frame)context;
            struct S_hlp{
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                  vector<C_edit_Selection::S_flash_item> &flash_list = *(vector<C_edit_Selection::S_flash_item>*)c;
                  for(int i=flash_list.size(); i--; ){
                     if(flash_list[i].frm==frm && flash_list[i].persistent){
                        flash_list[i] = flash_list.back(); flash_list.pop_back();
                        break;
                     }
                  }
                  return I3DENUMRET_OK;
               }
            };
            S_hlp::cbEnum(frm, (dword)&flash_list);
            frm->EnumFrames(S_hlp::cbEnum, (dword)&flash_list);
         }
         break;
         /*
                              //ignored commands:
      case EDIT_ACTION_PICK: case EDIT_ACTION_MISSION_SAVE_0: case EDIT_ACTION_MISSION_SAVE_2: case EDIT_ACTION_WINDOW_RESET: case EDIT_ACTION_FRAME_DUPLICATE: case EDIT_ACTION_SET_ACTIVE:
      case EDIT_ACTION_QUERY_MODIFY: case EDIT_ACTION_VALIDATE:
         break;
      default:
         todo(C_fstring("Plugin %s: unknown command: %i", GetName(), id));
         */
      }
      return 0;
   }

//----------------------------

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){
                              //process flashing selection
      for(int i=flash_list.size(); i--; ){
         if(!flash_list[i].persistent){
            flash_list[i].count_down -= time;
            if(flash_list[i].count_down <= 0){
               flash_list[i] = flash_list.back();
               flash_list.pop_back();
            }
         }
      }
   }

//----------------------------

   virtual void Render(){

      if(!sel_list.size() && !flash_list.size())
         return;
      int i;

      PI3D_scene scene = ed->GetScene();
      I3D_bbox temp_bb;
                              //line indicies
      static word bbox_list[]={
         0, 1, 2, 3, 4, 5, 6, 7,
         0, 2, 1, 3, 4, 6, 5, 7,
         0, 4, 1, 5, 2, 6, 3, 7,
      }, pivot_list[] = {
         0, 1, 2, 3, 4, 5
      };

      if(sel_list.size()){

         bool loc_system = false;
         bool uniform_scale = true;
         byte rot_axis = 0;
         LPC_editor_item me = ed->FindPlugin("MouseEdit");
         if(me){
            loc_system = me->Action(E_MOUSE_GET_LOC_SYSTEM);
            uniform_scale = me->Action(E_MOUSE_GET_UNIFORM);
            rot_axis = me->Action(E_MOUSE_GET_ROT_AXIS);
         }
      
         S_matrix mi = GetIdentityMatrix();
         //static const S_vector color(1, 1, 1);
         //static const S_vector rot_color(1, 1, 0);
      
         bool single_sel = (sel_list.size()==1);

         S_vector pivot_all;
         pivot_all.Zero();

         for(i=sel_list.size(); i--; ){
            PI3D_frame frm = sel_list[i];

            const S_matrix &m = frm->GetMatrix();
            const I3D_bbox *lvp = &temp_bb;

            bool draw_pivot = false;

            switch(frm->GetType()){
            case FRAME_LIGHT: case FRAME_SOUND: case FRAME_VOLUME:
            case FRAME_TARGET: case FRAME_CAMERA: case FRAME_OCCLUDER:
            case FRAME_JOINT:
            case FRAME_SECTOR:
               frm->DebugDraw(scene);
               lvp = NULL;
               draw_pivot = true; 
               scene->SetRenderMatrix(m);
               break;
            case FRAME_DUMMY:
               lvp = I3DCAST_DUMMY(frm)->GetBBox();
               break;
            case FRAME_VISUAL:
               {
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  switch(vis->GetVisualType()){
                  case I3D_VISUAL_NURBS:
                  case I3D_VISUAL_SINGLEMESH:
                     vis->DebugDraw(scene);
                     break;
                  }
                  lvp = &vis->GetBoundVolume().bbox;
               }
               break;
            case FRAME_MODEL:
               {
                  PI3D_model mod = I3DCAST_MODEL(frm);
                  lvp = &mod->GetHRBoundVolume().bbox;
                  if(!lvp->IsValid()){
                     temp_bb.min=S_vector(-.25f, -.25f, -.25f);
                     temp_bb.max=S_vector(.25f, .25f, .25f);
                     lvp = &temp_bb;
                  }
               }
               break;
            default:
               temp_bb.Invalidate();
               {
                  I3D_bound_volume bvol;
                  frm->ComputeHRBoundVolume(&bvol);
                  temp_bb = bvol.bbox;
               }
               if(!temp_bb.IsValid()){
                  temp_bb.min=S_vector(-.1f, -.1f, -.1f);
                  temp_bb.max=S_vector(.1f, .1f, .1f);
               }
               break;
            }
            if(lvp){
                                 //check if this frame is in flash list
               {
                  for(int j=flash_list.size(); j--; )
                  if(flash_list[j].frm==frm){
                     if(flash_list[j].persistent && (ed->GetIGraph()->ReadTimer()%500) >= 250)
                        break;
                  }
                  if(j!=-1) continue;
               }

               scene->SetRenderMatrix(m);
               S_vector bbox[8];
               for(int ii=0; ii<8; ii++)
                  bbox[ii] = S_vector((*lvp)[ii&1].x, (*lvp)[(ii&2)/2].y, (*lvp)[(ii&4)/4].z);

               //byte alpha = 255 - 255.0f * config_tab->ItemF(TAB_F_CFG_ALPHA);
               byte alpha = 0;
               scene->DrawLines(bbox, 8, bbox_list, sizeof(bbox_list)/sizeof(word),
                  //config_tab->ItemV(TAB_V_CFG_COLOR_SEL),
                  S_vector(1, 1, 1),
                  sizeof(S_vector),
                  alpha);
               if(single_sel){
                  draw_pivot = true;
               }else{
                  pivot_all += frm->GetWorldPos();
               }
            }
            if(draw_pivot){
                           //draw pivot
               S_vector v[6];
               float scale = frm->GetScale();
               v[0] = S_vector(.2f, 0, 0) * scale;
               v[1] = S_vector(-.2f, 0, 0) * scale;
               v[2] = S_vector(0, .2f, 0) * scale;
               v[3] = S_vector(0, -.2f, 0) * scale;
               v[4] = S_vector(0, 0, .2f) * scale;
               v[5] = S_vector(0, 0, -.2f) * scale;
   
               //byte alpha = 255 - 255.0f * config_tab->ItemF(TAB_F_CFG_ALPHA);
               byte alpha = 0;
               /*
               driver->DrawLines(v, 6, pivot_list, sizeof(pivot_list)/sizeof(word),
                  config_tab->ItemV(TAB_V_CFG_COLOR_PIVOT),
                  sizeof(S_vector), alpha);
                  */
                              //draw in 3 colors (red, green, blue)
               const float c0 = .4f, c1 = .8f;
               scene->DrawLines(&v[0], 2, pivot_list, 2, S_vector(c1, c0, c0), sizeof(S_vector), alpha);
               scene->DrawLines(&v[2], 2, pivot_list, 2, S_vector(c0, c1, c0), sizeof(S_vector), alpha);
               scene->DrawLines(&v[4], 2, pivot_list, 2, S_vector(c0, c0, c1), sizeof(S_vector), alpha);
            }

            if(single_sel){
                                 //draw axes
               float len;
               if(!loc_system){
                  mi(3) = frm->GetWorldPos();
                  scene->SetRenderMatrix(mi);
                  len = lvp ? (lvp->max-lvp->min).Magnitude()/(2*1.732) : 1.0f;
               }else
                  scene->SetRenderMatrix(frm->GetMatrix());

               //byte alpha = 255 - 255.0f * config_tab->ItemF(TAB_F_CFG_ALPHA);
               byte alpha = 0;

               if(rot_axis&1){      //X
                  if(loc_system) len = lvp ? (lvp->max.x-lvp->min.x)/2 : 1.0f;
                  scene->DrawLine(S_vector(len*.5f, 0, 0), S_vector(len*1.5f, 0, 0), 
                     //config_tab->ItemV(TAB_V_CFG_COLOR_AXE),
                     S_vector(1, 1, 0),
                     alpha);
               }
               if(rot_axis&2){      //Y
                  if(loc_system) len = lvp ? (lvp->max.y-lvp->min.y)/2 : 1.0f;
                  scene->DrawLine(S_vector(0, len*.5f, 0), S_vector(0, len*1.5f, 0), 
                     //config_tab->ItemV(TAB_V_CFG_COLOR_AXE),
                     S_vector(1, 1, 0),
                     alpha);
               }
               if(rot_axis&4){      //Z
                  if(loc_system) len = lvp ? (lvp->max.z-lvp->min.z)/2 : 1.0f;
                  scene->DrawLine(S_vector(0, 0, len*.5f), S_vector(0, 0, len*1.5f),
                     //config_tab->ItemV(TAB_V_CFG_COLOR_AXE),
                     S_vector(1, 1, 0),
                     alpha);
               }
            }

            if(draw_links){
                              //transform 8 corner points of bbox and draw connection
                              // to center of parent bbox (if it's not primary sector
               PI3D_frame frm_parent = frm->GetParent();
               if(!(!frm_parent || (frm_parent->GetType()==FRAME_SECTOR && I3DCAST_SECTOR(frm_parent)->IsPrimary()))){
                  scene->SetRenderMatrix(GetIdentityMatrix());
                              //get center of parent frame
                  S_vector v_prnt;
                  switch(frm_parent->GetType()){
                  case FRAME_VISUAL:
                     {
                              //center to bbox
                        const S_matrix &m = frm_parent->GetMatrix();
                        const I3D_bound_volume &bvol = I3DCAST_VISUAL(frm_parent)->GetBoundVolume();
                        I3D_bbox bb;
                        bb.min = bvol.bbox.min * m;
                        bb.max = bvol.bbox.max * m;
                        v_prnt = bb.min + (bb.max - bb.min) * .5f;
                     }
                     break;
                  default:
                     v_prnt = frm_parent->GetWorldPos();
                  }



                  //byte alpha = 255 - 255.0f * config_tab->ItemF(TAB_F_CFG_ALPHA);
                  byte alpha = 0;
                  static const S_vector color(.9f, .6f, .0f);
                  if(lvp){
                     S_vector v_t[9];
                     (*lvp).Expand(v_t);
                     for(int i=8; i--; ) v_t[i] = v_t[i] * m;
                     v_t[8] = v_prnt;
                     static const word indx[] = {
                        0, 8, 1, 8, 2, 8, 3, 8, 4, 8, 5, 8, 6, 8, 7, 8,
                     };
                     scene->DrawLines(v_t, 9, indx, sizeof(indx)/sizeof(word),
                        color, sizeof(S_vector), alpha);
                  }else{
                     scene->DrawLine(v_prnt, frm->GetWorldPos(),
                        color, alpha);
                  }
               }
            }
         }

         if(!single_sel){
                                 //draw pivot
            pivot_all /= sel_list.size();
            S_vector v[6];
            v[0] = pivot_all + S_vector(.3f, 0, 0);
            v[1] = pivot_all + S_vector(-.3f, 0, 0);
            v[2] = pivot_all + S_vector(0, .3f, 0);
            v[3] = pivot_all + S_vector(0, -.3f, 0);
            v[4] = pivot_all + S_vector(0, 0, .3f);
            v[5] = pivot_all + S_vector(0, 0, -.3f);
            scene->SetRenderMatrix(GetIdentityMatrix());

            //byte alpha = 255 - 255.0f * config_tab->ItemF(TAB_F_CFG_ALPHA);
            byte alpha = 0;
            scene->DrawLines(v, 6, pivot_list, sizeof(pivot_list)/sizeof(word),
               //config_tab->ItemV(TAB_V_CFG_COLOR_PIVOT),
               S_vector(1, 0, 0),
               sizeof(S_vector), alpha);
         }
      }

                              //render flashing items
      for(i=flash_list.size(); i--; ){
                              //blink persistent
         if(flash_list[i].persistent && (ed->GetIGraph()->ReadTimer()%500) < 250) continue;
         PI3D_frame frm = flash_list[i].frm;

         const I3D_bbox *lvp;
         const S_matrix *m = &frm->GetMatrix();

         switch(frm->GetType()){
         case FRAME_VOLUME:
            {
               lvp = &temp_bb;
               PI3D_volume vol = I3DCAST_VOLUME(frm);
               switch(vol->GetVolumeType()){
               case I3DVOLUME_SPHERE:
                  {
                     temp_bb.min = S_vector(-1, -1, -1);
                     temp_bb.max = S_vector(1, 1, 1);
                  }
                  break;
               case I3DVOLUME_RECTANGLE:
               case I3DVOLUME_BOX:
                  {
                     const S_vector &s = vol->GetNUScale();
                     temp_bb.min = -s;
                     temp_bb.max = s;
                     if(vol->GetVolumeType()==I3DVOLUME_RECTANGLE){
                        temp_bb.min.z = -.01f;
                        temp_bb.max.z =  .01f;
                     }
                  }
                  break;
               default:
                  temp_bb.Invalidate();
               }
            }
            break;
         case FRAME_DUMMY:
            lvp = I3DCAST_DUMMY(frm)->GetBBox();
            break;
         case FRAME_VISUAL:
            {
               PI3D_visual vis = I3DCAST_VISUAL(frm);
               lvp = &vis->GetBoundVolume().bbox;
            }
            break;
         case FRAME_MODEL:
            {
               PI3D_model mod = I3DCAST_MODEL(frm);
               lvp = &mod->GetHRBoundVolume().bbox;
            }
            break;
         default:
            temp_bb.Invalidate();
            {
               I3D_bound_volume bvol;
               frm->ComputeHRBoundVolume(&bvol);
               temp_bb = bvol.bbox;
            }
            lvp = &temp_bb;
            break;
         }
         if(!lvp->IsValid()){
            temp_bb.min = S_vector(-.25f, -.25f, -.25f);
            temp_bb.max = S_vector(.25f, .25f, .25f);
            lvp = &temp_bb;
         }
         scene->SetRenderMatrix(*m);
         S_vector bbox[8];
         static const S_vector flash_color[] = {
            S_vector(0, 1, 0),
            S_vector(1, 0, 0),
         };
         for(int ii=0; ii<8; ii++)
            bbox[ii] = S_vector((*lvp)[ii&1].x, (*lvp)[(ii&2)/2].y, (*lvp)[(ii&4)/4].z);
         scene->DrawLines(bbox, 8, bbox_list, sizeof(bbox_list)/sizeof(word),
            //config_tab->ItemV(TAB_V_CFG_COLOR_FLASH1+flash_list[i].persistent),
            flash_color[flash_list[i].persistent],
            sizeof(S_vector));

         static const word bbox_list1[]={
            1, 3, 5,  5, 3, 7,  0, 4, 6,  0, 6, 2,
            3, 2, 6,  3, 6, 7,  0, 1, 4,  5, 4, 1,
            5, 7, 6,  5, 6, 4,  0, 2, 1,  1, 2, 3
         };

         scene->DrawTriangles(bbox, 8, bbox_list1, sizeof(bbox_list1)/sizeof(word),
            //config_tab->ItemV(TAB_V_CFG_COLOR_FLASH1 + flash_list[i].persistent) * .3f,
            flash_color[flash_list[i].persistent] * .3f,
            sizeof(S_vector), 
            0xc0);
      }
   }

//----------------------------
#define SAVE_VERSION 3

   virtual bool LoadState(class C_chunk *ck){

      byte version = 0;
      ck->Read((char*)&version, sizeof(version));
      if(version!=SAVE_VERSION)
         return false;

      //config_tab->Open((dword)ck->GetHandle(), TABOPEN_FILEHANDLE | TABOPEN_UPDATE);
      int i;
      ck->Read((char*)&i, sizeof(int));
      ck->Read((char*)objsel_list_width, sizeof(int)*i);
      ck->Read(&draw_links, sizeof(byte));

      ed->CheckMenu (this, E_SELECTION_TOGGLE_LINK, draw_links);
      return true;
   }

//----------------------------

   virtual bool SaveState(class C_chunk *ck) const{

      byte version = SAVE_VERSION;
      ck->Write((char*)&version, sizeof(version));
      //config_tab->Save((dword)ck->GetHandle(), TABOPEN_FILEHANDLE);
                              //save current widths
      int i = SELECTION_NUM_TABS;
      ck->Write((char*)&i, sizeof(int));
      ck->Write((char*)objsel_list_width, sizeof(int)*SELECTION_NUM_TABS);
      ck->Write(&draw_links, sizeof(byte));
      return true;
   }
};

//----------------------------

void CreateSelection(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Selection;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
