#include "all.h"
#include "common.h"
#include <fps_counter.h>

//----------------------------

class C_edit_Stats: public C_editor_item{
   virtual const char *GetName() const{ return "Stats"; }
   int stats_index;
   bool show_fps;

   int curr_count;            //used for state update

   C_string net_stats;        //saved network text

   C_fps_counter fps_cnt;

//----------------------------

   float curr_fps_count;

//----------------------------

   void DrawRect(float x, float y, float sx1, float sy1, bool bottom = false){

      PI3D_driver drv = ed->GetDriver();
      bool is_zb = drv->GetState(RS_USEZB);
      if(is_zb)
         drv->SetState(RS_USEZB, false);
      drv->SetTexture(NULL);
      dword vp[4];
      ed->GetScene()->GetViewport(vp[0], vp[1], vp[2], vp[3]);
      int sy = (vp[2] - vp[0]) * .75f;

      if(bottom){
         sy *= (float)ed->GetIGraph()->Scrn_sy() / ((float)ed->GetIGraph()->Scrn_sx() * .75f);
      }

      S_vectorw v[4];
      for(int i=0; i<4; i++){
         v[i].x = vp[0] + sy * x;
         v[i].y = vp[1] + sy * y;
         v[i].z = 0.0f;
         v[i].w = 1.0f;
         if(i&1) v[i].x += sy * sx1;
         if(i&2) v[i].y += sy * sy1;
      }
      static const word indx[] = {0, 1, 2, 2, 1, 3};
      ed->GetScene()->DrawTriangles(v, 4, indx, sizeof(indx)/sizeof(word),
         S_vector(0, 0, 0), sizeof(S_vectorw), I3DDRAW_TRANSFORMED | 0x80);
      if(is_zb)
         drv->SetState(RS_USEZB, true);
   }

//----------------------------

   C_poly_text texts;         //polygon font system
   C_poly_group *text, *text_fps;
   float rect_sy;

//----------------------------

   C_poly_group *MakeText(const char *cp, float x, float y, float size, float &rect_sy){


      float ratio = 1.0f;
      int numl;
      C_poly_group *text = texts.CreateText(cp, x, y, size*ratio, 0, NULL, NULL, NULL, &numl);
      text->SetColor(0xc0ffffff);

      rect_sy = size * (numl + .5f);

      return text;
   }

//----------------------------

   void DestroyText(){
      if(text){
         texts.DeleteText(text);
         text = NULL;
      }
      rect_sy = 0.0f;
   }

//----------------------------

   void DestroyFPS(){
      if(text_fps){
         texts.DeleteText(text_fps);
         text_fps = NULL;
      }
   }

//----------------------------

   static bool GetVisualInfo(PI3D_frame frm, dword &num_verts, dword &num_faces){
      switch(frm->GetType()){
      case FRAME_VISUAL:
         {
            PI3D_visual vis = I3DCAST_VISUAL(frm);
            PI3D_mesh_base mb = vis->GetMesh();
            if(mb){
               num_verts += mb->NumVertices();
               num_faces += mb->NumFaces();
               return true;
            }
         }
      }
      return false;
   }

//----------------------------
public:
   C_edit_Stats():
      stats_index(E_STATS_SCENE),
      show_fps(true),
      text(NULL), text_fps(NULL),
      rect_sy(0.0f),
      curr_count(0)
   {
   }
   ~C_edit_Stats(){
      Close();
   }

   virtual bool Init(){
      int i = texts.Init(ed->GetDriver(), "gamefont.prj", "font.bmp", NULL,
         TEXTMAP_TRANSP, GetHInstance());
      if(!i)
         return false;

      ed->AddShortcut(this, E_STATS_FPS, "%90 &Stats\\&FPS\tF", K_F, 0);
      ed->AddShortcut(this, E_STATS_VIDEO, "&Stats\\&Video\tCtrl+1", K_1, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_SYSMEM, "&Stats\\System &memory\tCtrl+2", K_2, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_COLLISIONS, "&Stats\\&Collisions\tCtrl+3", K_3, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_SCENE, "&Stats\\&Scene\tCtrl+4", K_4, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_SELECTION, "&Stats\\&Selection\tCtrl+5", K_5, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_SOUND, "&Stats\\Sound\tCtrl+6", K_6, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_NET, "&Stats\\Network\tCtrl+7", K_7, SKEY_CTRL);
      //ed->AddShortcut(this, E_STATS_DETAILED, "&Stats\\D&etailed\tCtrl+8", K_8, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_BSP, "&Stats\\&Bsp\tCtrl+8", K_8, SKEY_CTRL);
      ed->AddShortcut(this, E_STATS_VOLTREE, "&Stats\\Volume &tree", 0, 0);
      ed->AddShortcut(this, E_STATS_DISABLE, "&Stats\\&Disable\tCtrl+0", K_0, SKEY_CTRL);

      ed->RegForRender(this, 900);
      return true;
   }

   virtual void Close(){

      ed->UnregForRender(this);
      DestroyText();
      DestroyFPS();
      texts.Close();
   }

   virtual dword C_edit_Stats::Action(int id, void *context){

      switch(id){
      case E_STATS_VIDEO:
      case E_STATS_SYSMEM:
      case E_STATS_COLLISIONS:
      case E_STATS_SCENE:
      case E_STATS_SELECTION:
      case E_STATS_SOUND:
      case E_STATS_NET:
      //case E_STATS_DETAILED:
      case E_STATS_BSP:
      case E_STATS_VOLTREE:
         {
            if(stats_index!=-1) ed->CheckMenu(this, stats_index, false);
            if(stats_index==id){
               stats_index = -1;
               DestroyText();
               break;
            }
            stats_index = id;
            ed->CheckMenu(this, id, true);
            curr_count = 0;
         }
         break;

      case E_STATS_FPS:
         if(show_fps){
            DestroyFPS();
         }
         show_fps = !show_fps;
         ed->CheckMenu(this, id, show_fps);
         break;

      case EDIT_ACTION_WINDOW_RESET:
         DestroyText();
         DestroyFPS();
         curr_count = 0;
         break;

      case E_STATS_SET_NET_TEXT:
         if(stats_index==E_STATS_NET)
            net_stats = (const char*)context;
         break;

      case EDIT_ACTION_AFTER_LOAD:
         curr_count = 0;
         break;
      }
      return 0;
   }

//----------------------------

   virtual void Render(){

      if(!show_fps && stats_index==-1)
         return;

      bool was_wire = ed->GetDriver()->GetState(RS_WIREFRAME);
      if(was_wire)
         ed->GetDriver()->SetState(RS_WIREFRAME, false);

      float y = .005f;
      const float size = .025f;
      if(show_fps){
         DrawRect(.005f, y, .16f, .035f);
         DestroyFPS();
         float rect_sy;
         text_fps = MakeText(C_fstring("fps: %.2f", curr_fps_count), .01f, y+size, size, rect_sy);
         y += .04f;
      }

      C_string str;

      switch(stats_index){
      case E_STATS_VIDEO: 
         if(curr_count<0){
            curr_count = 200;
            I3D_stats_video st;
            ed->GetDriver()->GetStats(I3DSTATS_VIDEO, &st);
            str = C_fstring(
               "Video memory:\n"
               " total: %.1f MB\n"
               " free: %.1f MB\n"
               "Texture memory:\n"
               " total: %.1f MB\n"
               " free: %.1f MB\n"
               "Textures:\n"
               " total: %i\n"
               " in video mem: %i\n"
               " sysmem used: %.1f MB\n"
               " vidmem used: %.1f MB",
               (float)st.vidmem_total/1048576,
               (float)st.vidmem_free/1048576,
               (float)st.txtmem_total/1048576,
               (float)st.txtmem_free/1048576,
               st.txt_total,
               st.txt_vidmem,
               (float)st.txt_sys_mem/1048576,
               (float)st.txt_video_mem/1048576
               );
         }
         break;

      case E_STATS_SYSMEM:
         if(curr_count<0){
            curr_count = 500;

            I3D_stats_memory st;
            ed->GetDriver()->GetStats(I3DSTATS_MEMORY, &st);
            dword mem_used = 0, num_blocks = 0;
            /*
#if defined _MSC_VER
#ifdef _DEBUG
            _CrtMemState state;
            _CrtMemCheckpoint(&state);
            num_blocks = state.lCounts[_NORMAL_BLOCK];
            mem_used = state.lSizes[_NORMAL_BLOCK];
#endif
#endif
            */
            str = C_fstring(
               "System memory stats:\n"
               " used (total): %.1f kb\n"
               " used (engine): %.1f kb\n"
               "  in %i blocks\n"
               " used (game): %.1f kb\n"
               "  in %i blocks\n"
               "Database:\n"
               " total: %i kb\n"
               " used: %i kb",
               //(float)st.available/1024.0f,
               (float)(st.used + st.used_other + mem_used)/1024.0f,
               (float)(st.used + st.used_other)/1024.0f,
               st.num_blocks,
               (float)mem_used/1024.0f,
               num_blocks,
               st.dbase_size / 1024,
               st.dbase_used / 1024
               );
         }
         break;

      case E_STATS_COLLISIONS:
         {
            I3D_stats_volume st;
            ed->GetDriver()->GetStats(I3DSTATS_VOLUME, &st);
            str = C_fstring(
               "Collision stats:\n"
               " ls: %i"
               " lr: %i\n"
               " lb: %i"
               " lf: %i\n"
               " ss: %i"
               " sr: %i\n"
               " sb: %i"
               " sf: %i"
               ,
               st.line_sphere,
               st.line_rectangle,
               st.line_box,
               st.line_face,
               st.sphere_sphere,
               st.sphere_rectangle,
               st.sphere_box,
               st.sphere_face
               );
         }
         break;

      case E_STATS_SCENE:
         {
            I3D_stats_render st;
            ed->GetDriver()->GetStats(I3DSTATS_RENDER, &st);
            I3D_stats_scene st1;
            ed->GetDriver()->GetStats(I3DSTATS_SCENE, &st1);
            str = C_fstring(
               "Scene stats:\n"
               " tri: %i\n"
               " vert: %i\n"
               " vis: %i\n"
               " snd: %i\n"
               " sct: %i\n"
               " occ: %i\n"
               " comp.: %i\n"
               " shd: %i",
               st.triangle,
               st.vert_trans,
               st1.frame_count[FRAME_VISUAL],
               st1.frame_count[FRAME_SOUND],
               st1.frame_count[FRAME_SECTOR],
               st1.frame_count[FRAME_OCCLUDER],
               st1.num_vis_computed,
               st1.dyn_shd_casters
               );
         }
         break;

      case E_STATS_SELECTION:
         {
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)ed->FindPlugin("Selection")->Action(E_SELECTION_GETVECTOR);
            switch(sel_list.size()){
            case 0:
               str += "<no selection>";
               break;
            case 1:
               {
                  PI3D_frame frm = sel_list.front();
                  switch(frm->GetType()){
                  case FRAME_VISUAL:
                     {
                        PI3D_visual vis = I3DCAST_VISUAL(frm);
                        PI3D_mesh_base mb = vis->GetMesh();
                        if(mb){
                           str = C_fstring("%s v: %i, f: %i\n",
                              (const char*)str,
                              mb->NumVertices(),
                              mb->NumFaces());
                        }else{
                           str = "";
                        }
                     }
                     break;

                  case FRAME_MODEL:
                     {
                        PI3D_model mod = I3DCAST_MODEL(frm);
                        struct S_hlp{
                           dword num_verts;
                           dword num_faces;
                           dword num_meshes;
                           bool (*GetVisualInfo)(PI3D_frame, dword&, dword&);
                           static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                              S_hlp &hlp = *(S_hlp*)c;
                              if(hlp.GetVisualInfo(frm, hlp.num_verts, hlp.num_faces))
                                 ++hlp.num_meshes;
                              return I3DENUMRET_OK;
                           }
                        } hlp;
                        hlp.num_verts = 0;
                        hlp.num_faces = 0;
                        hlp.num_meshes = 0;
                        hlp.GetVisualInfo = GetVisualInfo;

                        mod->EnumFrames(S_hlp::cbEnum, (dword)&hlp);
                        str = C_fstring("Meshes: %i\nv: %i, f: %i",
                           hlp.num_meshes, hlp.num_verts, hlp.num_faces);
                     }
                     break;

                  case FRAME_SECTOR:
                     {
                        PI3D_sector sct = I3DCAST_SECTOR(frm);
                        str += C_fstring(
                           "\n"
                           "lights: %i"
                           ,
                           sct->NumLights()
                           );
                     }
                     break;

                  default:
                     str = "";
                  }
               }
               break;
            default:
               {
                  dword num_verts = 0;
                  dword num_faces = 0;
                  dword num_meshes = 0;
                  for(int i=sel_list.size(); i--; ){
                     if(GetVisualInfo(sel_list[i], num_verts, num_faces))
                        ++num_meshes;
                  }
                  str = C_fstring("Meshes: %i\nv: %i, f: %i",
                     num_meshes, num_verts, num_faces);
               }
            }
         }
         break;

      case E_STATS_SOUND:
         {
            I3D_stats_sound st;
            ed->GetDriver()->GetStats(I3DSTATS_SOUND, &st);
            str = C_fstring(
               "Sound stats:\n"
               " all: %i\n"
               " audible: %i\n"
               " in card: %i\n"
               " mem used: %.1f Kb\n"
               "Hardware:\n"
               " hw all: %i free: %i\n"
               " 3D all: %i free: %i\n"
               " mem all: %i\n"
               " mem free: %i",
               st.snds_all,
               st.snds_audible,
               st.snds_in_card,
               (float)st.mem_used / 1024.0f,
               st.buf_hw_all,
               st.buf_hw_free,
               st.buf_3d_all,
               st.buf_3d_free,
               st.hw_mem_all,
               st.hw_mem_free
               );
         }
         break;

      case E_STATS_BSP:
         {
            I3D_stats_bsp st;
            ed->GetDriver()->GetStats(I3DSTATS_BSP, &st);
            str = C_fstring(
               "Bsp stats:\n"
               " num nodes: %i\n"
               " depth: %i\n"
               " num planes: %i\n"
               " num faces: %i\n"
               " faces ref: %i\n"
               " mem used: %.1f Kb",
               st.num_nodes,
               st.depth,
               st.num_planes,
               st.num_faces,
               st.face_references,
               (float)st.mem_used / 1024.0f
               );
         }
         break;
      case E_STATS_VOLTREE:
         {
            I3D_stats_voltree st;
            ed->GetDriver()->GetStats(I3DSTATS_VOLTREE, &st);
            str = C_fstring(
               "Volume tree stats:\n"
               " total volumes: %i\n"
               " root volumes: %i\n"
               " moving static: %i\n"
               //" reserved2: %i\n"
               "",
               st.total_num,
               st.root_num,
               st.moving_static
               //st.reserved2
               );
         }
         break;

      case E_STATS_NET:
         {
            str = net_stats;
         }
         break;

         /*
      case E_STATS_DETAILED:
         if(curr_count<0){
            curr_count = 1000;

            struct S_hlp{
               dword frm_mem[FRAME_LAST];
               dword vis_mem_obj;
               dword vis_mem_prt;
               dword vis_mem_lit;
               dword vis_mem_sm;
               dword vis_mem_rest;
               dword lmap_mem;
               dword total;
               dword num_frames;
               set<PI3D_mesh_base> mb_list;

               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                  S_hlp *hp = (S_hlp*)c;
                  I3D_stats_frame st;
                  frm->GetStats(I3DSTATS_FRAME, &st);
                  if(frm->GetType()!=FRAME_VISUAL){
                     hp->frm_mem[frm->GetType()] += st.class_mem;
                  }else{
                     PI3D_mesh_base mb = I3DCAST_VISUAL(frm)->GetMesh();
                     if(mb){
                        hp->mb_list.insert(mb);
                     }
                     hp->vis_mem_rest += st.class_mem;
                  }
                  hp->total += st.class_mem;
                  ++hp->num_frames;
                  return I3DENUMRET_OK;
               }
            } hlp;
            memset(&hlp, 0, offsetof(S_hlp, mb_list));
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&hlp);

            set<PI3D_material> mat_list;
                              //collect stuff from collected meshes
            I3D_stats_mesh mesh_st;
            memset(&mesh_st, 0, sizeof(mesh_st));

            set<PI3D_texture> txt_list;

                              //collect stuff from collected materials
            I3D_stats_material mat_st;
            memset(&mat_st, 0, sizeof(mat_st));

            set<PI3D_material>::const_iterator mat_it;
            for(mat_it=mat_list.begin(); mat_it!=mat_list.end(); mat_it++){
               PI3D_material mat = *mat_it;
               I3D_stats_material st1;
               mat->GetStats(I3DSTATS_MATERIAL, &st1);
               mat_st.class_mem += st1.class_mem;
               if(mat->GetTexture(MTI_DIFFUSE))
                  txt_list.insert(mat->GetTexture(MTI_DIFFUSE));
            }

                              //collect stuff from collected textures
            I3D_stats_texture txt_st;
            memset(&txt_st, 0, sizeof(txt_st));

            set<PI3D_texture>::const_iterator txt_it;
            for(txt_it=txt_list.begin(); txt_it!=txt_list.end(); txt_it++){
               PI3D_texture txt = *txt_it;
               I3D_stats_texture st1;
               txt->GetStats(I3DSTATS_TEXTURE, &st1);
               txt_st.class_mem += st1.class_mem;
               txt_st.num_texels += st1.num_texels;
               txt_st.mem_size += st1.mem_size;
            }

            str = C_fstring(
               "Detailed 3D stats:\n"
               "Frames: %i\n"
               "%s"           //obj
               "%s"           //prt
               "%s"           //lit
               "%s"           //sm
               "%s"           //vis
               "%s"           //lgt
               "%s"           //snd
               "%s"           //cam
               "%s"           //sec
               "%s"           //dum
               "%s"           //tar
               "%s"           //usr
               "%s"           //mod
               "%s"           //jnt
               "%s"           //vol
               "%s"           //occ
               "  mem used: %.1f kb\n"
               "Meshes: %i\n"
               "  memory: %.1f kb\n"
               "  vertices: %.1f kb\n"
               "  faces: %.1f kb\n"
               "  total: %.1f kb\n"
               "Materials: %i\n"
               "  total: %.1f kb\n"
               "Textures: %i\n"
               "  texels: %i\n"
               "  memory: %.1f kb\n"
               "-----------------\n"
               "Total mem: %.1f kb"
               ,
               hlp.num_frames,
               (!hlp.vis_mem_obj    ? "" : (const char*)C_fstring("  object: %.1f kb\n", (float)hlp.vis_mem_obj/1024.0f)),
               (!hlp.vis_mem_prt    ? "" : (const char*)C_fstring("  particle: %.1f kb\n", (float)hlp.vis_mem_prt/1024.0f)),
               (!hlp.vis_mem_lit    ? "" : (const char*)C_fstring("  lit obj: %.1f kb\n", (float)hlp.vis_mem_lit/1024.0f)),
               (!hlp.vis_mem_sm     ? "" : (const char*)C_fstring("  singlemesh: %.1f kb\n", (float)hlp.vis_mem_sm/1024.0f)),
               (!hlp.vis_mem_rest   ? "" : (const char*)C_fstring("  other vis: %.1f kb\n", (float)hlp.vis_mem_rest/1024.0f)),
               (!hlp.frm_mem[FRAME_LIGHT]    ? "" : (const char*)C_fstring("  light: %.1f kb\n", (float)hlp.frm_mem[FRAME_LIGHT]/1024.0f)),
               (!hlp.frm_mem[FRAME_SOUND]    ? "" : (const char*)C_fstring("  sound: %.1f kb\n", (float)hlp.frm_mem[FRAME_SOUND]/1024.0f)),
               (!hlp.frm_mem[FRAME_CAMERA]   ? "" : (const char*)C_fstring("  camera: %.1f kb\n", (float)hlp.frm_mem[FRAME_CAMERA]/1024.0f)),
               (!hlp.frm_mem[FRAME_SECTOR]   ? "" : (const char*)C_fstring("  sector: %.1f kb\n", (float)hlp.frm_mem[FRAME_SECTOR]/1024.0f)),
               (!hlp.frm_mem[FRAME_DUMMY]    ? "" : (const char*)C_fstring("  dummy: %.1f kb\n", (float)hlp.frm_mem[FRAME_DUMMY]/1024.0f)),
               (!hlp.frm_mem[FRAME_TARGET]   ? "" : (const char*)C_fstring("  target: %.1f kb\n", (float)hlp.frm_mem[FRAME_TARGET]/1024.0f)),
               (!hlp.frm_mem[FRAME_USER]     ? "" : (const char*)C_fstring("  user: %.1f kb\n", (float)hlp.frm_mem[FRAME_USER]/1024.0f)),
               (!hlp.frm_mem[FRAME_MODEL]    ? "" : (const char*)C_fstring("  model: %.1f kb\n", (float)hlp.frm_mem[FRAME_MODEL]/1024.0f)),
               (!hlp.frm_mem[FRAME_JOINT]    ? "" : (const char*)C_fstring("  joint: %.1f kb\n", (float)hlp.frm_mem[FRAME_JOINT]/1024.0f)),
               (!hlp.frm_mem[FRAME_VOLUME]   ? "" : (const char*)C_fstring("  volume: %.1f kb\n", (float)hlp.frm_mem[FRAME_VOLUME]/1024.0f)),
               (!hlp.frm_mem[FRAME_OCCLUDER] ? "" : (const char*)C_fstring("  occluder: %.1f kb\n", (float)hlp.frm_mem[FRAME_OCCLUDER]/1024.0f)),
               (float)hlp.total/1024.0f,
               hlp.mb_list.size(),
               (float)mesh_st.class_mem/1024.0f,
               (float)mesh_st.vertex_mem/1024.0f,
               (float)mesh_st.face_mem/1024.0f,
               (float)(mesh_st.class_mem + mesh_st.vertex_mem + mesh_st.face_mem)/1024.0f,
               mat_list.size(),
               (float)mat_st.class_mem/1024.0f,
               txt_list.size(),
               txt_st.num_texels,
               (float)txt_st.mem_size/1024.0f,
               (float)(hlp.total + mesh_st.class_mem + mesh_st.vertex_mem + mesh_st.face_mem +
                  mat_st.class_mem + txt_st.mem_size) / 1024.0f,
               -1
               );
         }
         break;
         */

      case -1:
         break;

      default:
         assert(0);
      }

      if(stats_index!=-1){
         if(str.Size()){
            DestroyText();
            text = MakeText(str, .01f, y+size, size, rect_sy);
         }
         if(rect_sy)
            DrawRect(.005f, y, .25f, rect_sy);
      }

      bool was_zb = ed->GetDriver()->GetState(RS_USEZB);
      if(was_zb) ed->GetDriver()->SetState(RS_USEZB, false);
      texts.Draw(ed->GetScene());
      if(was_zb) ed->GetDriver()->SetState(RS_USEZB, true);
      if(was_wire) ed->GetDriver()->SetState(RS_WIREFRAME, true);
   }

//----------------------------

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){
      curr_count -= time;
      if(show_fps){
         fps_cnt.Tick(time);
         curr_fps_count = fps_cnt.GetCount();
      }
   }

//----------------------------

   virtual bool LoadState(C_chunk *ck){
      
      int version = 0;
      ck->Read(&version, sizeof(int));
      if(version!=101) return false;

                              //write stats type
      ck->Read(&stats_index, sizeof(int));
      ck->Read(&show_fps, sizeof(byte));
      if(stats_index!=-1){
         ed->CheckMenu(this, stats_index, true);
      }
      ed->CheckMenu(this, E_STATS_FPS, show_fps);
      return true;
   }

//----------------------------

   virtual bool SaveState(C_chunk *ck) const{

      int version = 101;
      ck->Write(&version, sizeof(int));
                              //write stats type
      ck->Write(&stats_index, sizeof(int));
      ck->Write(&show_fps, sizeof(byte));
      return true;
   }
};

//----------------------------

void CreateStats(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Stats;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
