#include "all.h"
#include "common.h"

//----------------------------

enum E_ACTION_SUBEDIT{
   E_SUBEDIT_PICK_DEF = 1000,
   E_SUBEDIT_PICK_SEL,
   E_SUBEDIT_NEW_PT,
   E_SUBEDIT_PICK_SEL_HIDDEN,
   E_SUBEDIT_SET_INDICIES,
   E_SUBEDIT_TOGGLE_INDEX,
   E_SUBEDIT_MOVE_XZ,
   E_SUBEDIT_MOVE_Y,
   E_SUBEDIT_SCALE,
   E_SUBEDIT_DELETE,

   E_SUBEDIT_SET_VERTS,       // - internal (undo)
};

//----------------------------

class C_edit_SubEdit: public C_editor_item{
                              //returns plugin's name
   virtual const char *GetName() const{ return "SubEdit"; }

   C_smart_ptr<C_editor_item> e_medit, e_undo;
   C_smart_ptr<I3D_material> mat_edit[3]; //[0=unselected, 1=selected, 2=pointed]

                              //currently edited frame
   C_smart_ptr<I3D_frame> edit_frm;
   vector<C_smart_ptr<I3D_frame> > hidden_frames;

   set<int> selection;        //sub-selection (indicies)
   bool undo_reset;
   float vertex_render_size;

//----------------------------

   void CreateEditMat(){

      PI3D_driver drv = ed->GetDriver();
      for(int i=0; i<3; i++){
         mat_edit[i] = ed->GetDriver()->CreateMaterial();
         const char *bitmap_name = "OCCL_CIRCLE";
         switch(i){
         case 0:
            mat_edit[i]->SetDiffuse(S_vector(0, 1, 0));
            break;
         case 1:
            mat_edit[i]->SetDiffuse(S_vector(1, 1, 0));
            break;
         case 2:
            mat_edit[i]->SetDiffuse(S_vector(1, 1, 1));
            break;
         }
         I3D_CREATETEXTURE ct;
         memset(&ct, 0, sizeof(ct));
         ct.flags = TEXTMAP_NOMIPMAP;
         HGLOBAL hg = LoadResource(GetHInstance(),
            FindResource(GetHInstance(), bitmap_name, RT_BITMAP));
         if(hg){
            PI3D_texture tp;
            ct.text_name = (const char*)hg;
            ct.flags |= TEXTMAP_DIFFUSE | TEXTMAP_MEM_PTR | TEXTMAP_TRANSP;
            I3D_RESULT ir = drv->CreateTexture(&ct, &tp);
            if(I3D_SUCCESS(ir)){
               mat_edit[i]->SetTexture(tp);
               tp->Release();
            }
         }
         mat_edit[i]->Release();
      }
   }

//----------------------------
// Get picked vertex index, or -1 if out of range.
   int GetPointedVertex(){

      PI3D_scene scene = ed->GetScene();
      S_vector from, dir;
      if(I3D_FAIL(scene->UnmapScreenPoint(ed->GetIGraph()->Mouse_x(), ed->GetIGraph()->Mouse_y(), from, dir)))
         return -1;

      S_plane pl_pick;
      pl_pick.normal = dir;
      pl_pick.d = -(pl_pick.normal.Dot(from));

      //float best_dist = cd.GetHitFrm() ? cd.GetClosestHit() : 1e+16f;
      float best_dist = 1e+16f;
      int best_index = -1;
      assert(edit_frm);
      switch(edit_frm->GetType()){
      case FRAME_OCCLUDER:
         {
            const S_matrix &m = edit_frm->GetMatrix();
            PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);
            dword numv = occ->NumVertices();
            S_vector *verts = occ->LockVertices();
            //float min_angle = vertex_render_size * PI * .1f;
            float min_dist = vertex_render_size * .5f;
            for(int i=numv; i--; ){
               S_vector v = verts[i] * m;
                              //check distance
               float d = v.DistanceToPlane(pl_pick);
               if(d > best_dist)
                  continue;

               float dl = v.DistanceToLine(from, dir);
               if(dl > min_dist)
                  continue;
               /*
                              //check angle
               float a = dir.AngleTo(v-from);
               if(a > min_angle)
                  continue;
                  */
               best_dist = d;
               best_index = i;
            }
            occ->UnlockVertices();
         }
         break;
      }
      return best_index;
   }

//----------------------------

   void HidePointedVisuals(){

      PI3D_scene scene = ed->GetScene();
      S_vector from, dir;
      if(I3D_SUCCESS(scene->UnmapScreenPoint(ed->GetIGraph()->Mouse_x(), ed->GetIGraph()->Mouse_y(), from, dir))){

         while(true){
            I3D_collision_data cd;
            cd.from = from;
            cd.dir = dir;
            cd.flags = I3DGEOCOL_TEXTURE | I3DGEOCOL_RAY | I3DCOL_EXACT;
            ed->GetScene()->TestCollision(cd);
            PI3D_frame frm = cd.GetHitFrm();
            if(!frm)
               break;
            frm->SetOn(false);
            hidden_frames.push_back(frm);
         }
      }
   }

//----------------------------

   void UnhideHiddenVisuals(){

      for(int i=hidden_frames.size(); i--; )
         hidden_frames[i]->SetOn(true);
      hidden_frames.clear();
   }

//----------------------------
// Clear selection and save into UNDO buffer.
   void SaveSelectionUndo(){

      int *mem = new int[selection.size()+1], *mp = mem;
      for(set<int>::const_iterator it=selection.begin(); it!=selection.end(); it++)
         *mp++ = (*it);
      *mp = -1;
      S_edit_Undo_2 eu(GetName(), E_SUBEDIT_SET_INDICIES, mem, (selection.size()+1)*sizeof(int));
      e_undo->Action(E_UNDO_SAVE, &eu);
      delete[] mem;
   }

//----------------------------
// Translate selection into specified direction.
   void TranslateSelection(const S_vector &delta){

      if(!selection.size())
         return;

      assert(edit_frm);

      S_vector loc_dir = delta;

      if(edit_frm->GetParent())
         loc_dir %= edit_frm->GetInvMatrix();

      switch(edit_frm->GetType()){
      case FRAME_OCCLUDER:
         {
            SaveVertexUndo();

            PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);

            dword numv = occ->NumVertices();
            const S_vector *verts = occ->LockVertices();
            vector<S_vector> new_verts(verts, verts+numv);
            vector<S_vector> save_verts = new_verts;

            for(set<int>::const_iterator it=selection.begin(); it!=selection.end(); it++){
               int indx = (*it);
               assert(indx<numv);
               new_verts[indx] += loc_dir;
            }
            occ->UnlockVertices();

            I3D_RESULT ir = occ->Build(new_verts.begin(), new_verts.size());
            if(I3D_FAIL(ir)){
               ed->Message("Failed to scale vertices.");
               occ->Build(save_verts.begin(), save_verts.size());
            }

            set<int> new_sel;
            verts = occ->LockVertices();
            numv = occ->NumVertices();
            for(it=selection.begin(); it!=selection.end(); it++){
               const S_vector &v = new_verts[*it];
               for(int i=numv; i--; ){
                  if(IsMrgZero((verts[i]-v).Magnitude()))
                     break;
               }
               if(i!=-1)
                  new_sel.insert(i);
            }
            selection = new_sel;
            occ->UnlockVertices();
         }
         break;
      }
      ed->SetModified();
   }

//----------------------------
// Scale selection.
   void ScaleSelection(float scl_delta){

      if(selection.size() <= 1)
         return;

      assert(edit_frm);

      float loc_scl = scl_delta;

      if(edit_frm->GetParent())
         loc_scl *= edit_frm->GetInvMatrix()(0).Magnitude();

      switch(edit_frm->GetType()){
      case FRAME_OCCLUDER:
         {
            SaveVertexUndo();

            PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);

            dword numv = occ->NumVertices();
            const S_vector *verts = occ->LockVertices();
            vector<S_vector> new_verts(verts, verts+numv);
            vector<S_vector> save_verts = new_verts;

            set<int>::const_iterator it;
            S_vector sel_center(0, 0, 0);
            for(it=selection.begin(); it!=selection.end(); it++)
               sel_center += verts[*it];
            sel_center /= (float)selection.size();

            for(it=selection.begin(); it!=selection.end(); it++){
               int indx = (*it);
               assert(indx<numv);
               S_vector &v = new_verts[indx];
               S_vector dir = v - sel_center;
               v = sel_center;
               if(dir.Magnitude() > .01f)
                  v += dir * loc_scl;
            }
            occ->UnlockVertices();
            I3D_RESULT ir = occ->Build(new_verts.begin(), new_verts.size());
            if(I3D_FAIL(ir)){
               ed->Message("Failed to scale vertices.");
               occ->Build(save_verts.begin(), save_verts.size());
            }

            set<int> new_sel;
            verts = occ->LockVertices();
            numv = occ->NumVertices();
            for(it=selection.begin(); it!=selection.end(); it++){
               const S_vector &v = new_verts[*it];
               for(int i=numv; i--; ){
                  if(IsMrgZero((verts[i]-v).Magnitude()))
                     break;
               }
               if(i!=-1)
                  new_sel.insert(i);
            }
            selection = new_sel;
            occ->UnlockVertices();
         }
         break;
      }
      ed->SetModified();
   }

//----------------------------

   void DeleteSelection(){

      assert(edit_frm);
      if(!selection.size()){
         ed->Message("No vertex deleted.");
         return;
      }

      int num_del = 0;
      bool ok = false;

      switch(edit_frm->GetType()){
      case FRAME_OCCLUDER:
         {
            PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);
            SaveSelectionUndo();
            SaveVertexUndo();

            dword numv = occ->NumVertices();
            const S_vector *verts = occ->LockVertices();
            vector<S_vector> new_verts;
            new_verts.reserve(numv);

            num_del = selection.size();
            set<int>::const_iterator it;
            for(int i=numv; i--; ){
                              //if index is not in selection, add it
               it = selection.find(i);
               if(it==selection.end())
                  new_verts.push_back(verts[i]);
            }
            occ->UnlockVertices();
            if(new_verts.size() < 4){
               ed->Message("Failed to delete vertex.");
            }else{
               occ->Build(new_verts.begin(), new_verts.size());
               selection.clear();
               ok = true;
            }
         }
         break;
      }
      if(ok){
         ed->SetModified();
         ed->Message(C_fstring("Vertices deleted: %i.", num_del));
      }
   }

//----------------------------

   void SaveVertexUndo(){

      assert(edit_frm);

      const char *top_name = (const char*)e_undo->Action(E_UNDO_GET_TOP_PLUGIN);
      bool undo_on = undo_reset || (!top_name || strcmp(top_name, GetName()) ||
         e_undo->Action(E_UNDO_GET_TOP_ACTION)!=E_SUBEDIT_SET_VERTS);
      if(!undo_on)
         return;

      switch(edit_frm->GetType()){
      case FRAME_OCCLUDER:
         {
            PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);

            dword numv = occ->NumVertices();
            const S_vector *verts = occ->LockVertices();
            vector<dword> undo_info;
            undo_info.push_back(numv);
            for(int i=0; i<numv; i++){
               undo_info.push_back(FloatAsInt(verts[i].x));
               undo_info.push_back(FloatAsInt(verts[i].y));
               undo_info.push_back(FloatAsInt(verts[i].z));
            }
            S_edit_Undo_2 eu(GetName(), E_SUBEDIT_SET_VERTS, undo_info.begin(), undo_info.size()*sizeof(dword));
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;
      }
      undo_reset = false;
   }
public:
   C_edit_SubEdit():
      undo_reset(true),
      vertex_render_size(0.0f)
   {
   }

//----------------------------

   virtual bool Init(){
      e_medit = ed->FindPlugin("MouseEdit");
      e_undo = ed->FindPlugin("Undo");
      if(!e_medit || !e_undo){
         e_medit = NULL;
         e_undo = NULL;
         return false;
      }

      return true;
   }

//----------------------------
                              //uninitialize plugin
   virtual void Close(){

      e_medit = NULL;
      mat_edit[0] = NULL;
      mat_edit[1] = NULL;
   }

//----------------------------
                              //tick - performed in every rendering frame
   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){

      UnhideHiddenVisuals();
      if(edit_frm){
                              //precompute several values
         switch(edit_frm->GetType()){
         case FRAME_OCCLUDER:
            {
               PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);
               dword numv = occ->NumVertices();
               S_vector *verts = occ->LockVertices();

                              //get occluder's bbox
               I3D_bbox bb;
               bb.Invalidate();
               for(int i=numv; i--; ){
                  const S_vector &v = verts[i];
                  bb.min.Minimal(v);
                  bb.max.Maximal(v);
               }
               float diag = (bb.max - bb.min).Magnitude();
               diag *= edit_frm->GetMatrix()(0).Magnitude();
               vertex_render_size = diag * .025f;

               occ->UnlockVertices();
            }
            break;
         }

         if(skeys==(SKEY_SHIFT|SKEY_CTRL|SKEY_ALT)){
            HidePointedVisuals();
         }
      }
   }

//----------------------------
                              //action - called from menu/keypress, or by other plugins
   virtual dword Action(int id, void *context){

      switch(id){

      case EDIT_ACTION_SUBOBJ_EDIT:
         {
            PI3D_frame frm = (PI3D_frame)context;
            bool ok = false;
            switch(frm->GetType()){
            case FRAME_OCCLUDER:
               if(I3DCAST_OCCLUDER(frm)->GetOccluderType()==I3DOCCLUDER_MESH)
                  ok = true;
               break;
            }
            if(!ok)
               break;
            S_MouseEdit_user_mode m;
            memset(&m, 0, sizeof(m));
            m.ei = this;
            for(int i=0; i<S_MouseEdit_user_mode::MODE_LAST; i++){
               S_MouseEdit_user_mode::S_mode &md = m.modes[i];
               switch(i){
               case S_MouseEdit_user_mode::MODE_DEFAULT:
                  md.hcursor = LoadCursor(GetHInstance(), "IDC_OCCLUDER_DEF");
                  md.action_id = E_SUBEDIT_PICK_DEF;
                  break;
               case S_MouseEdit_user_mode::MODE_SELECT:
                  md.hcursor = LoadCursor(GetHInstance(), "IDC_OCCLUDER_SEL");
                  md.action_id = E_SUBEDIT_PICK_SEL;
                  break;
               case S_MouseEdit_user_mode::MODE_LINK:
                  md.hcursor = LoadCursor(GetHInstance(), "IDC_OCCLUDER_NEW");
                  md.action_id = E_SUBEDIT_NEW_PT;
                  break;
               case S_MouseEdit_user_mode::MODE_PICK_SPEC:
                  md.hcursor = LoadCursor(GetHInstance(), "IDC_OCCLUDER_SEL1");
                  md.action_id = E_SUBEDIT_PICK_SEL_HIDDEN;
                  break;
               case S_MouseEdit_user_mode::MODE_MOVE_XZ: md.action_id = E_SUBEDIT_MOVE_XZ; break;
               case S_MouseEdit_user_mode::MODE_MOVE_Y: md.action_id = E_SUBEDIT_MOVE_Y; break;
               case S_MouseEdit_user_mode::MODE_SCALE: md.action_id = E_SUBEDIT_SCALE; break;
               case S_MouseEdit_user_mode::MODE_DELETE: md.action_id = E_SUBEDIT_DELETE; break;
               }
            }
            e_medit->Action(E_MOUSE_SET_USER_MODE, &m);
            assert(!edit_frm);
            edit_frm = frm;
            ed->RegForRender(this);
            Tick(0, 0, 0, 0, 0, 0);
            return true;
         }
         break;

      case EDIT_ACTION_SUBOBJ_CANCEL:
         if(edit_frm){
            edit_frm = NULL;
            ed->UnregForRender(this);
            selection.clear();
         }
         break;

      case E_SUBEDIT_PICK_DEF:
         {
            undo_reset = true;
            SaveSelectionUndo();
            selection.clear();
            int indx = GetPointedVertex();
            if(indx!=-1)
               selection.insert(indx);
         }
         break;

      case E_SUBEDIT_PICK_SEL:
      case E_SUBEDIT_PICK_SEL_HIDDEN:
         {
            undo_reset = true;
            int indx = GetPointedVertex();
            if(indx==-1)
               break;
            Action(E_SUBEDIT_TOGGLE_INDEX, &indx);
         }
         break;

      case E_SUBEDIT_NEW_PT:
         {
            int indx = GetPointedVertex();
            if(indx!=-1){
               ed->Message("Can't create new point over another point.");
               break;
            }
            const S_MouseEdit_picked *mp = (const S_MouseEdit_picked*)context;
            if(!mp->frm_picked){
               ed->Message("Can't create new point in air.");
               break;
            }

            switch(edit_frm->GetType()){
            case FRAME_OCCLUDER:
               {
                  SaveSelectionUndo();
                  selection.clear();
                  SaveVertexUndo();

                  PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);

                  dword numv = occ->NumVertices();
                  const S_vector *verts = occ->LockVertices();
                  vector<S_vector> new_verts(verts, verts+numv);
                  occ->UnlockVertices();

                  S_vector v = mp->pick_from + mp->pick_dir * mp->pick_dist;
                  v *= occ->GetInvMatrix();

                  new_verts.push_back(v);
                  occ->Build(new_verts.begin(), new_verts.size());

                  verts = occ->LockVertices();
                  for(int i=occ->NumVertices(); i--; ){
                     if(IsMrgZero((verts[i]-v).Magnitude()))
                        break;
                  }
                  occ->UnlockVertices();
                  if(i!=-1){
                     selection.insert(i);
                     ed->Message("Vertex created.");
                  }else{
                     ed->Message("New vertex is fully inside of old convex hull.");
                  }
               }
               break;
            }

            undo_reset = true;
         }
         break;

      case E_SUBEDIT_MOVE_XZ:
      case E_SUBEDIT_MOVE_Y:
         {
            const S_vector *delta = (const S_vector*)context;
            TranslateSelection(*delta);
         }
         break;

      case E_SUBEDIT_SCALE:
         {
            float scl = IntAsFloat((int)context);
            ScaleSelection(scl);
         }
         break;

      case E_SUBEDIT_SET_VERTS:
         {
            if(!edit_frm){
               ed->Message("E_SUBEDIT_SET_VERTS: not in subedit mode");
               break;
            }
            SaveVertexUndo();
            switch(edit_frm->GetType()){
            case FRAME_OCCLUDER:
               {
                  PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);
                  int numv = *((dword*&)context)++;
                  const S_vector *verts = (const S_vector*)context;
                  occ->Build(verts, numv);
               }
               break;
            }
         }
         break;

      case E_SUBEDIT_DELETE:
         DeleteSelection();
         break;

      case E_SUBEDIT_TOGGLE_INDEX:
         {
            int indx = *(int*)context;
            S_edit_Undo_2 eu(GetName(), E_SUBEDIT_TOGGLE_INDEX, &indx, sizeof(int));
            e_undo->Action(E_UNDO_SAVE, &eu);
            set<int>::iterator it = selection.find(indx);
            if(it!=selection.end()) selection.erase(it);
            else selection.insert(indx);
         }
         break;

      case E_SUBEDIT_SET_INDICIES:
         {
            SaveSelectionUndo();
            selection.clear();
            int *ip = (int*)context;
            while(*ip!=-1)
               selection.insert(*ip++);
         }
         break;

      case EDIT_ACTION_BEFORE_FREE:
         UnhideHiddenVisuals();
         break;
      }
      return 0;
   }

//----------------------------
                              //called for registered plugins during rendering
   virtual void Render(){

      assert(edit_frm);
      //PI3D_driver drv = ed->GetDriver();
      PI3D_scene scene = ed->GetScene();

      switch(edit_frm->GetType()){
      case FRAME_OCCLUDER:
         {
            PI3D_occluder occ = I3DCAST_OCCLUDER(edit_frm);

            if(!mat_edit[0])
               CreateEditMat();

            S_vector draw_offset = scene->GetActiveCamera()->GetWorldDir() * -.05f;
            S_matrix m = occ->GetMatrix();
            m(3) += draw_offset;

            int pointed = GetPointedVertex();
            dword numv = occ->NumVertices();
            S_vector *verts = occ->LockVertices();
            set<int>::const_iterator it;

                              //render verices
            for(int i=numv; i--; ){
               it = selection.find(i);
               if(it!=selection.end())
                  continue;
               scene->DrawSprite(verts[i] * m, mat_edit[0], 0x80, vertex_render_size);
            }
                              //render selection
            for(it=selection.begin(); it!=selection.end(); it++){
               int indx = (*it);
               scene->DrawSprite(verts[indx] * m, mat_edit[1], 0, vertex_render_size*1.25);
            }
            if(pointed!=-1){
               m(3) += draw_offset*.5f;
               scene->DrawSprite(verts[pointed] * m, mat_edit[2], 0, vertex_render_size*.75f);
            }
            occ->UnlockVertices();
         }
         break;
      }
   }

//----------------------------
                              //store/restore plugin's state
   //virtual bool LoadState(class C_chunk *ck){ return false; }
   //virtual bool SaveState(class C_chunk *ck) const{ return false; }
};

//----------------------------

void CreateSubEdit(LPC_editor ed){

   LPC_editor_item ei = new C_edit_SubEdit;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
