#include "all.h"
#include "common.h"
#include <deque>

//----------------------------

class C_edit_Undo: public C_editor_item{
   virtual const char *GetName() const{ return "Undo"; }
                              //internal undo/redo buffers
   struct S_undo_entry{
      C_string plugin_name;   //NULL = break
      int action_id;
      byte *data;
      dword data_size;
      bool may_collapse;      //true if entry may be merged with another entry
      S_undo_entry():
         data(NULL),
         data_size(0),
         may_collapse(true)
      {}
      ~S_undo_entry(){
         delete[] data;
      }
      S_undo_entry &operator =(const S_undo_entry &ue){
         plugin_name = ue.plugin_name;
         action_id = ue.action_id;
         data = new byte[data_size = ue.data_size];
         memcpy(data, ue.data, data_size);
         return *this;
      }
      bool operator <(const S_undo_entry &ue) const{
         return (this < &ue);
      }
      bool operator !=(const S_undo_entry &ue) const{
         return (this != &ue);
      }
   };
   deque<S_undo_entry> undo_list[2];  //undo=0, redo=1
   bool in_undo;              //true when un-doing
   bool in_redo;              //true when re-doing

                              //check if top entry has same params like this,
                              // delete it if true
   void CheckTop(const S_undo_entry &ue1){

      if(undo_list[in_undo].size()){
         S_undo_entry &ue = undo_list[in_undo].back();
         if(ue.may_collapse &&
            ue.plugin_name==ue1.plugin_name &&
            ue.action_id==ue1.action_id){

            undo_list[in_undo].pop_back();
         }
      }
   }
public:
   C_edit_Undo():
      in_undo(false),
      in_redo(false)
   {}

   virtual bool Init(){
      ed->AddShortcut(this, E_UNDO_UNDO, "%10 &Edit\\%0 &Undo\tCtrl+Z", K_Z, SKEY_CTRL);
      ed->AddShortcut(this, E_UNDO_REDO, "%10 &Edit\\%0 %a Redo\tCtrl+X", K_X, SKEY_CTRL);
      return true;
   }

   virtual dword C_edit_Undo::Action(int id, void *context){

      switch(id){
      case E_UNDO_UNDO:
      case E_UNDO_REDO:
         {
            bool redo = (id==E_UNDO_REDO);
            if(undo_list[redo].size()){
               in_undo = !redo;
               in_redo = redo;
                              //go down untill not empty or break reached
               if(undo_list[redo].size()){
                              //skip break, if any
                  if(!undo_list[redo].back().plugin_name.Size())
                     undo_list[redo].pop_back();

                  while(undo_list[redo].size()){
                              //get undo entry
                     S_undo_entry &ue = undo_list[redo].back();
                              //break - stop un-doing
                     if(!ue.plugin_name.Size()) break;
                              //find specified plugin
                     LPC_editor_item plugin = ed->FindPlugin(ue.plugin_name);
                              //call its action
                     if(plugin) plugin->Action(ue.action_id, ue.data);
                              //pop entry
                     undo_list[redo].pop_back();
                  }
               }
               /*             //not working properly!
               if(!undo_list[redo].size() || !undo_list[redo].back().plugin_name.size())
                  ed->EnableMenu(this, E_UNDO_UNDO+redo, false);
                  */

               in_undo = false;
               in_redo = false;
            }else
               ed->Message(!redo ? "No more undos" : "Nothing to redo");
         }
         break;

      case E_UNDO_SAVE:
         if(context){
            S_edit_Undo_2 *up = (S_edit_Undo_2*)context;
            undo_list[in_undo].push_back(S_undo_entry());
            S_undo_entry &ue = undo_list[in_undo].back();
            ue.plugin_name = up->plugin_name;
            ue.action_id = up->action_id;
            ue.data = new byte[up->data_size];
            ue.data_size = up->data_size;
            memcpy(ue.data, up->data, ue.data_size);

                              //clear redo list
            if(!in_undo && !in_redo){
               //ed->EnableMenu(this, E_UNDO_REDO, false);
               undo_list[1].clear();
            }
         }else
            ed->Message("Undo(E_UNDO_SAVE): expecting parameter", 0, EM_ERROR);
         break;

      case E_UNDO_BREAK:
         {
                              //insert entry with NULL plugin name
            S_undo_entry ue;
            ue.action_id = 0;
            CheckTop(ue);
            undo_list[in_undo].push_back(ue);

            in_undo = !in_undo;
            CheckTop(ue);
            undo_list[in_undo].push_back(ue);
            in_undo = !in_undo;
         }
         break;

      case E_UNDO_CLEAR:
         undo_list[0].clear();
         undo_list[1].clear();
         break;

      case E_UNDO_GET_TOP_PLUGIN:
         if(undo_list[in_undo].size()){
            S_undo_entry &ue = undo_list[in_undo].back();
            if(ue.plugin_name.Size()) return (dword)&ue.plugin_name;
            if(undo_list[in_undo].size()>=2){
               S_undo_entry &ue = undo_list[in_undo][undo_list[in_undo].size()-2];
               return (dword)(const char*)ue.plugin_name;
            }
         }
         return NULL;

      case E_UNDO_GET_TOP_ACTION:
         if(undo_list[in_undo].size()){
            S_undo_entry &ue = undo_list[in_undo].back();
            if(ue.plugin_name.Size()) return ue.action_id;
            if(undo_list[in_undo].size()>=2){
               S_undo_entry &ue = undo_list[in_undo][undo_list[in_undo].size()-2];
               return ue.action_id;
            }
         }
         return -1;

      case E_UNDO_GET_TOP_CONTEXT:
         if(undo_list[in_undo].size()){
            S_undo_entry &ue = undo_list[in_undo].back();
            if(ue.plugin_name.Size()) return (dword)ue.data;
            if(undo_list[in_undo].size()>=2){
               S_undo_entry &ue = undo_list[in_undo][undo_list[in_undo].size()-2];
               return (dword)ue.data;
            }
         }
         return NULL;

      case E_UNDO_POP_TOP:
         if(undo_list[in_undo].size()){
            if(!undo_list[in_undo].back().plugin_name.Size()) undo_list[in_undo].pop_back();
            if(undo_list[in_undo].size()){
               undo_list[in_undo].pop_back();
               return true;
            }
         }
         break;

      case EDIT_ACTION_BEFORE_FREE:
         Action(E_UNDO_CLEAR, 0);
         break;
      }
      return 0;
   }

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){
      Action(E_UNDO_BREAK, 0);
   }
};

//----------------------------

void CreateUndo(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Undo;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
