#include "all.h"
#include "common.h"
#include <advanced.h>         //GetFileSize

//----------------------------

//#define DETECT_LIGHTS_IN_MODELS

//----------------------------

bool WinCreateDirectoryTree(const char *dst);

//----------------------------

static void MakeVertexMapping(const S_vector *vp, dword pitch, int numv, word *v_map, float thresh){

   for(int i=0; i<numv; i++){
      const S_vector &v1 = *(S_vector*)(((byte*)vp) + pitch * i);
      for(int j=0; j<i; j++){
         const S_vector &v2 = *(S_vector*)(((byte*)vp) + pitch * j);
         if((fabs(v1.x-v2.x) + fabs(v1.y-v2.y) + fabs(v1.z-v2.z)) < thresh)
            break;
      }
      v_map[i] = j;
   }
}

//----------------------------

class C_edit_Validity: public C_editor_item{
   virtual const char *GetName() const{ return "Validity"; }

//----------------------------

   bool ValidateGeometry(LPC_editor_item e_log, LPC_editor_item e_slct){

      bool err = false;

                              //collect all unique meshes
      vector<PI3D_visual> viss;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

            PI3D_visual vis = I3DCAST_VISUAL(frm);
            if(vis->GetMesh()){
               vector<PI3D_visual> &v = *(vector<PI3D_visual>*)c;
               v.push_back(vis);
            }
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&viss, ENUMF_VISUAL | ENUMF_BACKDROP);

      LPC_editor_item e_debugline = ed->FindPlugin("DebugLine");

      for(int vi=viss.size(); vi--; ){

         int i;
         int num_dup_faces = 0;
         PI3D_visual vis = viss[vi];
         PI3D_mesh_base mb = vis->GetMesh();

         const S_vector *verts = mb->LockVertices();
         dword numv = mb->NumVertices();
         vector<word> v_map(numv);
         int vstride = mb->GetSizeOfVertex();
         MakeVertexMapping(verts, vstride, numv, v_map.begin(), .001f);

                              //remapped faces
         vector<I3D_triface> faces;

         const I3D_triface *vfcs = mb->LockFaces();
         faces.assign(vfcs, vfcs+mb->NumFaces());
         mb->UnlockFaces();

         for(i=faces.size(); i--; ){
            I3D_triface &fc = faces[i];
            fc[0] = v_map[fc[0]];
            fc[1] = v_map[fc[1]];
            fc[2] = v_map[fc[2]];
         }

         for(i=faces.size(); i--; ){
            I3D_triface f0 = faces[i];
            for(int j=i; j--; ){
               I3D_triface f1 = faces[j];

               bool dup = false;
               if(f0[0]==f1[0] && f0[1]==f1[1] && f0[2]==f1[2]){
                  dup = true;
               }else
               if(f0[0]==f1[1] && f0[1]==f1[2] && f0[2]==f1[0]){
                  dup = true;
               }else
               if(f0[0]==f1[2] && f0[1]==f1[0] && f0[2]==f1[1]){
                  dup = true;
               }
               if(dup){
                  ++num_dup_faces;
                  if(e_debugline){
                     const S_matrix &m = vis->GetMatrix();
                     S_vector vw[3];
                     for(int i=3; i--; )
                        vw[i] = *(S_vector*)(((byte*)verts) + f0[i]*vstride) * m;
                     for(i=3; i--; )
                        e_debugline->Action(E_DEBUGLINE_ADD, &S_add_line(vw[i], vw[(i+1)%3], 0xffff0000));
                  }
               }
            }
         }
         mb->UnlockVertices();

         if(num_dup_faces){
            err = true;
            e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
               C_fstring("Visual '%s' has %i duplicated face(s).", vis->GetName(), num_dup_faces));
            e_slct->Action(E_SELECTION_ADD_FRAME, vis);
         }
      }

      return err;
   }

//----------------------------

   bool ValidateModels(LPC_editor_item e_log, LPC_editor_item e_slct){

      bool err = false;
                              //check model duplication
      vector<PI3D_model> v;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

            vector<PI3D_model> &v = *(vector<PI3D_model>*)c;
            v.push_back(I3DCAST_MODEL(frm));
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&v, ENUMF_MODEL | ENUMF_BACKDROP);

      for(int i=v.size(); i-- ; ){
         PI3D_model model = v[i];
         const S_vector &pos = model->GetWorldPos();
         const S_quat &rot = model->GetRot();
         for(int j=i; j-- ; ){
            PI3D_model model1 = v[j];
            const S_vector &pos1 = model1->GetWorldPos();
            const S_quat &rot1 = model1->GetRot();
            S_vector dir_to = pos - pos1;
            if(dir_to.Dot(dir_to) < .01f && rot==rot1 &&
               !stricmp(model->GetFileName(), model1->GetFileName())){
               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                  C_fstring("Model duplication on [%.3f, %.3f, %.3f] - '%s', duplicated with '%s'",
                  pos.x, pos.y, pos.z,
                  (const char*)model1->GetName(), model->GetName()));
               e_slct->Action(E_SELECTION_ADD_FRAME, model1);
               err = true;
            }
         }
#ifdef DETECT_LIGHTS_IN_MODELS
                              //check for lights in models
         if(model->FindChildFrame("*", ENUMF_LIGHT | ENUMF_WILDMASK)){
            e_slct->Action(E_SELECTION_ADD_FRAME, model);
            e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
               C_fstring("Model %s contains light(s)", model->GetName()));
            err = true;
         }
#endif
      }
      return err;
   }

//----------------------------

   bool ValidateVisuals(LPC_editor_item e_log, LPC_editor_item e_slct){

      bool err = false;
                              //check model duplication
      vector<PI3D_visual> v;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

            vector<PI3D_visual> &v = *(vector<PI3D_visual>*)c;
            v.push_back(I3DCAST_VISUAL(frm));
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&v, ENUMF_VISUAL | ENUMF_BACKDROP);

      for(int i=v.size(); i-- ; ){
         PI3D_visual vis = v[i];
         if(vis->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
            CPI3D_mesh_base mb = vis->GetMesh();
            if(!mb)
               continue;
            C_string str;
            for(int i=mb->NumFGroups(); i--; ){
               CPI3D_face_group fg = mb->GetFGroup(i);
               CPI3D_material mat = fg->GetMaterial();
               if(mat){
                  if(mat->GetAlpha()!=1.0f){
                     str = C_fstring("'%s' (alpha=%.2f)", mat->GetName(), mat->GetAlpha());
                     break;
                  }
                  if(mat->IsCkey()){
                     str = C_fstring("'%s' (color-keyed)", mat->GetName());
                     break;
                  }
                  CPI3D_texture tp = mat->GetTexture(MTI_DIFFUSE);
                  if(tp){
                     const IImage *img = tp->GetSysmemImage();
                     if(img){
                        if(img->GetPixelFormat()->flags&PIXELFORMAT_ALPHA){
                           const char *fn = tp->GetFileName(0);
                           if(!fn)
                              fn = "<unknown>";
                           str = C_fstring("'%s': (texture %s has alpha channel)", mat->GetName(), fn);
                           break;
                        }
                     }
                  }
               }
            }
            if(i!=-1){
               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                  C_fstring("Lightmapped visual '%s' has invalid material %s", vis->GetName(), (const char*)str));
               e_slct->Action(E_SELECTION_ADD_FRAME, vis);
               err = true;
            }
            PI3D_lit_object lobj = I3DCAST_LIT_OBJECT(vis);
            if(!lobj->IsConstructed()){
               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                  C_fstring("Lightmap '%s' not computed", vis->GetName()));
               e_slct->Action(E_SELECTION_ADD_FRAME, vis);
               err = true;
            }
         }
      }
      return err;
   }

//----------------------------

   bool ValidateLights(LPC_editor_item e_log, LPC_editor_item e_slct){

      bool err = false;

      vector<PI3D_light> list;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbAdd(PI3D_frame frm, dword lst){
            vector<PI3D_light> *frm_list = (vector<PI3D_light>*)lst;
            frm_list->push_back(I3DCAST_LIGHT(frm));
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbAdd, (dword)&list, ENUMF_LIGHT);

      for(int i=list.size(); i--; ){
         PI3D_light lp = list[i];
         if(!lp->NumLightSectors()){
            e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
               C_fstring("Light in no sector: '%s'",
                  (const char*)lp->GetName()));
            e_slct->Action(E_SELECTION_ADD_FRAME, lp);
            err = true;
         }
      }
      return err;
   }

//----------------------------

   bool ValidateSounds(LPC_editor_item e_log, LPC_editor_item e_slct){

      bool err = false;

      vector<PI3D_sound> list;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbAdd(PI3D_frame frm, dword lst){
            vector<PI3D_sound> *frm_list = (vector<PI3D_sound>*)lst;
            frm_list->push_back(I3DCAST_SOUND(frm));
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbAdd, (dword)&list, ENUMF_SOUND);

      for(int i=list.size(); i--; ){
         PI3D_sound sp = list[i];
         PISND_source src = sp->GetSoundSource();
         if(src){
                              //check format
            const S_wave_format *fmt = src->GetFormat();
            if(fmt && fmt->num_channels>1){
                              //only ambient types may be stereo
               switch(sp->GetSoundType()){
               case I3DSOUND_NULL:
               case I3DSOUND_AMBIENT:
               case I3DSOUND_POINTAMBIENT:
                  break;
               default:
                  {
                     e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                        C_fstring("Non-ambient stereo sound: '%s'",
                           (const char*)sp->GetName()));
                     e_slct->Action(E_SELECTION_ADD_FRAME, sp);
                     err = true;
                  }
               }
            }
         }
      }
      return err;
   }

//----------------------------

   bool ValidateSectors(LPC_editor_item e_log, LPC_editor_item e_slct){

      bool err = false;

                              //collect all sectors
      vector<PI3D_sector> sectors;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbAdd(PI3D_frame frm, dword lst){
            vector<PI3D_sector> *frm_list = (vector<PI3D_sector>*)lst;
            frm_list->push_back(I3DCAST_SECTOR(frm));
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbAdd, (dword)&sectors, ENUMF_SECTOR);
      S_hlp::cbAdd(ed->GetScene()->GetPrimarySector(), (dword)&sectors);
      S_hlp::cbAdd(ed->GetScene()->GetBackdropSector(), (dword)&sectors);

                              //check lights
      for(int i=sectors.size(); i--; ){
         PI3D_sector sct = sectors[i];
         vector<PI3D_light> fog_lights;
         for(int j=sct->NumLights(); j--; ){
            PI3D_light lp = sct->GetLight(j);
            switch(lp->GetLightType()){
            case I3DLIGHT_FOG:
               fog_lights.push_back(lp);
               break;
            }
         }
         if(fog_lights.size()>1){
            C_string light_names;
            const vector<PI3D_light> &list = fog_lights;
            for(int i=list.size(); i--; ){
               if(light_names.Size())
                  light_names += "; ";
               light_names = C_fstring("%s\"%s\"", (const char*)light_names, list[i]->GetName());
            }
            C_fstring str("Sector has multiple fog lights: '%s' (%s)",
               (const char*)sct->GetName(), (const char*)light_names);

            e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)str);
            e_slct->Action(E_SELECTION_ADD_FRAME, sct);
            err = true;
         }
      }

      LPC_editor_item e_debugline = ed->FindPlugin("DebugLine");
                              //check portals
      const float THRESH = .002f;

      for(i=sectors.size(); i--; ){
         PI3D_sector sct = sectors[i];
         if(sct->IsPrimary())
            continue;

         dword num_pl = sct->NumBoundPlanes();
         const S_plane *pls = sct->GetBoundPlanes();

         const PI3D_portal *prts = sct->GetPortals();
         for(int pi=sct->NumPortals(); pi--; ){
            PI3D_portal prt = *prts++;

                              //check if all points of portal are on edge of sector
            vector<bool> plane_ok(num_pl, true);
            int num_pl_ok = num_pl;

            dword numv = prt->NumVertices();
            const S_vector *verts = prt->GetVertices();
            for(int vi=numv; vi--; ){
               const S_vector &v = verts[vi];
                              //check if there's such plane on which we lie
               for(int pli=num_pl; pli--; ){
                  if(!plane_ok[pli])
                     continue;
                  float d = v.DistanceToPlane(pls[pli]);
                  if((float)fabs(d) > THRESH){
                     plane_ok[pli] = false;
                     --num_pl_ok;
                  }
               }
            }
            if(!num_pl_ok){
               C_fstring str("Sector '%s': sink portal",
                  (const char*)sct->GetName());

               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)str);
               e_slct->Action(E_SELECTION_ADD_FRAME, sct);

               if(e_debugline){
                  for(vi=numv; vi--; )
                     e_debugline->Action(E_DEBUGLINE_ADD, &S_add_line(verts[vi], verts[(vi+1)%numv], 0xffff0000));
               }

               err = true;

            }
         }
      }
      return err;
   }

//----------------------------

   bool ValidateLinks(LPC_editor_item e_log, LPC_editor_item e_slct){

                              //collect models and visuals
      vector<PI3D_frame> frms;
      vector<PI3D_sector> sectors;

      struct S_hlp{
         static I3DENUMRET I3DAPI cbAdd(PI3D_frame frm, dword c){
            vector<PI3D_frame> &v = *(vector<PI3D_frame>*)c;
            v.push_back(frm);
            return I3DENUMRET_OK;
         }
         static I3DENUMRET I3DAPI cbAddS(PI3D_frame frm, dword c){
            vector<PI3D_sector> &v = *(vector<PI3D_sector>*)c;
            v.push_back(I3DCAST_SECTOR(frm));
            return I3DENUMRET_OK;
         }

         static bool CheckSectorLink(PI3D_sector sct, const S_matrix &m, const I3D_bbox &bbox){

            S_vector bb_center = (bbox.min + bbox.max) * .5f;
            S_vector bb_center_world = bb_center * m;
                              //check using thresh
            const float thresh = .05f;
            for(int i=6; i--; ){
               S_vector pt_check = bb_center_world;
               pt_check[i/2] += (i&1) ? thresh : -thresh;
               bool is_inside = sct->CheckPoint(pt_check);
               if(is_inside)
                  break;
            }
            if(i==-1){
               S_vector bbox_full[8];
               bbox.Expand(bbox_full);
               for(i=8; i--; ){
                  S_vector check_point = bbox_full[i] * m;
                  bool is_iniside = sct->CheckPoint(check_point);
                  if(is_iniside)
                     break;
               }
               if(i==-1)
                  return false;
            }
            return true;
         }

         static bool IsOutOfSectors(vector<PI3D_sector> &sectors, const S_matrix &m, const I3D_bbox &bbox){

            S_vector bb_center = (bbox.min + bbox.max) * .5f;
            S_vector bb_center_world = bb_center * m;
                              //check using thresh
            const float thresh = .05f;
            for(int i=6; i--; ){
               S_vector pt_check = bb_center_world;
               pt_check[i/2] += (i&1) ? thresh : -thresh;
               for(int j=sectors.size(); j--; ){
                  if(sectors[j]->CheckPoint(pt_check))
                     break;
               }
               if(j==-1)
                  break;
            }
            if(i!=-1)
               return true;

            S_vector bbox_full[8];
            bbox.Expand(bbox_full);
            for(i=8; i--; ){
               S_vector check_point = bbox_full[i] * m;
               for(int j=sectors.size(); j--; ){
                  if(sectors[j]->CheckPoint(check_point))
                     break;
               }
               if(j==-1)
                  break;
            }
            if(i!=-1)
               return true;
            return false;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbAddS, (dword)&sectors, ENUMF_SECTOR);
      ed->GetScene()->EnumFrames(S_hlp::cbAdd, (dword)&frms, ENUMF_MODEL | ENUMF_VISUAL);

      bool err = false;

      set<PI3D_frame> err_list;

                              //go from front, because parents are stored earlier than children
      for(int i=0; i<frms.size(); i++){

         PI3D_frame frm = frms[i];
         switch(frm->GetType()){
         case FRAME_VISUAL:
         case FRAME_MODEL:
            {
                              //if any if our parent is already in err list, skip
               {
                  for(PI3D_frame prnt = frm; prnt=prnt->GetParent(); ){
                     if(err_list.find(prnt)!=err_list.end())
                        break;
                  }
                  if(prnt)
                     break;
               }

               const I3D_bound_volume &bvol =
                  (frm->GetType()==FRAME_VISUAL) ? I3DCAST_VISUAL(frm)->GetBoundVolume() :
                  I3DCAST_MODEL(frm)->GetHRBoundVolume();
               if(bvol.bbox.IsValid()){
                           //find its sectpr
                  PI3D_frame fsct = frm;
                  while((fsct=fsct->GetParent()) && fsct->GetType()!=FRAME_SECTOR);
                  assert(fsct);
                  if(!fsct)
                     break;
                  PI3D_sector sct = I3DCAST_SECTOR(fsct);

                  const S_matrix &m = frm->GetMatrix();
                  bool link_ok;
                  if(!sct->IsPrimary()){
                     link_ok = S_hlp::CheckSectorLink(sct, m, bvol.bbox);
                  }else{
                     link_ok = S_hlp::IsOutOfSectors(sectors, m, bvol.bbox);
                  }
                  if(!link_ok){
                           //possibly bad link, report
                     C_fstring str("Frame linked to wrong sector: '%s'", frm->GetName());
                     e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)str);
                     e_slct->Action(E_SELECTION_ADD_FRAME, frm);
                     err = true;
                     err_list.insert(frm);
                  }
               }
            }
            break;
         }
      }


      /*
      vector<PI3D_sector> sectors;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbAdd(PI3D_frame frm, dword lst){
            vector<PI3D_sector> *frm_list = (vector<PI3D_sector>*)lst;
            frm_list->push_back(I3DCAST_SECTOR(frm));
            return I3DENUMRET_OK;
         }
         static I3DENUMRET I3DAPI cbAdd1(PI3D_frame frm, dword lst){
            vector<PI3D_frame> *frm_list = (vector<PI3D_frame>*)lst;
            frm_list->push_back(frm);
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbAdd, (dword)&sectors, ENUMF_SECTOR);

      for(int i=sectors.size(); i--; ){
         PI3D_sector sct = sectors[i];
                              //check all frames (visuals and models)
         vector<PI3D_frame> frm_list;
         sct->EnumFrames(S_hlp::cbAdd1, (dword)&frm_list, ENUMF_MODEL);
         for(int j=frm_list.size(); j--; ){
            PI3D_frame frm = frm_list[j];
            I3D_bound_volume bvol;
            frm->GetHRBoundVolume(&bvol);
            if(bvol.bbox.IsValid()){
               S_vector bb_center = bvol.bbox.min + (bvol.bbox.max - bvol.bbox.min) * .5f;
               S_vector bb_center_world = bb_center * frm->GetMatrix();
                              //check if center of geometry is inside of sector
               bool is_iniside = sct->CheckPoint(bb_center_world);
               if(!is_iniside){
                              //possibli bad link, report
                  C_fstring str("Model linked to wrong sector: '%s'", frm->GetName());
                  e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)str);
                  e_slct->Action(E_SELECTION_ADD_FRAME, frm);
                  err = true;
               }
            }
         }
      }
      */
      return err;
   }

//----------------------------

   bool ValidateVolumes(LPC_editor_item e_log, LPC_editor_item e_slct){

      bool err = false;

      vector<PI3D_volume> list;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbAdd(PI3D_frame frm, dword lst){
            vector<PI3D_volume> *frm_list = (vector<PI3D_volume>*)lst;
            frm_list->push_back(I3DCAST_VOLUME(frm));
            return I3DENUMRET_OK;
         }
      };
      ed->GetScene()->EnumFrames(S_hlp::cbAdd, (dword)&list, ENUMF_VOLUME);

      for(int i=list.size(); i--; ){
         PI3D_volume vol = list[i];
         switch(vol->GetVolumeType()){
         case I3DVOLUME_SPHERE:
         case I3DVOLUME_RECTANGLE:
         case I3DVOLUME_BOX:
            break;
         default:
            e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)C_fstring("Unknown volume type: '%s'", vol->GetName()));
            e_slct->Action(E_SELECTION_ADD_FRAME, vol);
            err = true;
            continue;
         }

         const S_vector &pos = vol->GetPos();
         float scale = vol->GetScale();
         const S_vector &nu_scale = vol->GetNUScale();
         const S_quat &rot = vol->GetRot();
         for(int j=i; j-- ; ){
            PI3D_volume vol1 = list[j];
            if(vol->GetVolumeType()!=vol1->GetVolumeType())
               continue;
            if(vol->GetParent()!=vol1->GetParent())
               continue;
            const S_vector &pos1 = vol1->GetPos();
            if(pos==pos1 && scale==vol1->GetScale() && rot==vol1->GetRot() && nu_scale==vol1->GetNUScale()){

               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                  C_fstring("Volume duplication on [%.3f, %.3f, %.3f] - '%s'",
                  pos.x, pos.y, pos.z, vol1->GetName()));

               e_slct->Action(E_SELECTION_ADD_FRAME, vol1);
               err = true;
            }
         }
      }
      return err;
   }

//----------------------------

   static dword CountBits(dword val){
      dword mask = 0x80000000;
      dword num = 0;
      do{
         if(val&mask) ++num;
      }while(mask>>=1);
      return num;
   }

//----------------------------

   char res_copy_path[256];
public:
   C_edit_Validity(){
      res_copy_path[0] = 0;
   }

   virtual bool Init(){
      ed->AddShortcut(this, 1000, "%30 &Debug\\%40 Scene Validity chec&k", 0, 0);
      ed->AddShortcut(this, 1001, "%30 &Debug\\%40 Resource dump", 0, 0);
      return true;
   }
   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){}

//----------------------------

   virtual dword C_edit_Validity::Action(int id, void *context){

      switch(id){
      case 1000:              //scene validity
         {
            LPC_editor_item e_log = ed->FindPlugin("Log");
            LPC_editor_item e_slct = ed->FindPlugin("Selection");
            if(!e_log || !e_slct){
               break;
            }
                              //clear log
            e_log->Action(E_LOG_CLOSE);
                              //clear selection
            e_slct->Action(E_SELECTION_CLEAR);

            LPC_editor_item e_debugline = ed->FindPlugin("DebugLine");
            if(e_debugline)
               e_debugline->Action(E_DEBUGLINE_CLEAR);

            bool err = false;
            do{

               ed->Message("Checking models...", 0, EM_MESSAGE, true);
               err = ValidateModels(e_log, e_slct) || err;
               if(err) break;

               ed->Message("Checking visuals...", 0, EM_MESSAGE, true);
               err = ValidateVisuals(e_log, e_slct) || err;
               if(err) break;

               ed->Message("Checking lights...", 0, EM_MESSAGE, true);
               err = ValidateLights(e_log, e_slct);
               if(err) break;

               ed->Message("Checking sounds...", 0, EM_MESSAGE, true);
               err = ValidateSounds(e_log, e_slct);
               if(err) break;
                              //todo: check sector light ranges to be within sector

               ed->Message("Checking sectors...", 0, EM_MESSAGE, true);
               err = ValidateSectors(e_log, e_slct);
               if(err) break;

               ed->Message("Checking volumes...", 0, EM_MESSAGE, true);
               err = ValidateVolumes(e_log, e_slct);
               if(err) break;

               ed->Message("Checking links...", 0, EM_MESSAGE, true);
               err = ValidateLinks(e_log, e_slct);
               if(err) break;

                              //broadcast validation message
               struct S_hlp{
                  static bool cbEnum(LPC_editor_item ei, void *c){
                     if(ei->Action(EDIT_ACTION_VALIDATE)){
                        *(bool*)c = true;
                        return false;
                     }
                     return true;
                  }
               };
               ed->EnumPlugins(S_hlp::cbEnum, &err);
               if(err) break;

               ed->Message("Checking geometry...", 0, EM_MESSAGE, true);
               err = ValidateGeometry(e_log, e_slct) || err;
               if(err) break;
            }while(false);

            ed->Message(
               !err ? "Scene is OK!" :
               "Validity check failed!",
               0, EM_MESSAGE, true);
         }
         break;

      case 1001:
         {
            LPC_editor_item e_log = ed->FindPlugin("Log");
            LPC_editor_item e_modify = ed->FindPlugin("Modify");
            if(!e_log) break;

            struct S_hlp2{
               static BOOL CALLBACK dlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

                  switch(uMsg){
                  case WM_INITDIALOG:
                     {
                        const char *cp = (const char*)lParam;
                        SetDlgItemText(hwnd, IDC_EDIT, cp);
                        ShowWindow(hwnd, SW_SHOW);
                        SetWindowLong(hwnd, GWL_USERDATA, lParam);
                     }
                     return 1;
                  case WM_COMMAND:
                     switch(LOWORD(wParam)){
                     case IDCANCEL: EndDialog(hwnd, 0); break;
                     case IDOK:
                        SendDlgItemMessage(hwnd, IDC_EDIT, WM_GETTEXT,
                           256, GetWindowLong(hwnd, GWL_USERDATA));
                        EndDialog(hwnd, 1);
                        break;
                     }
                     break;
                  }
                  return 0;
               }
            };

            int i = DialogBoxParam(GetHInstance(), "IDD_RESOURCE_DUMP", (HWND)ed->GetIGraph()->GetHWND(),
               S_hlp2::dlgProc, (LPARAM)res_copy_path);
            if(!i)
               break;

            int num_errs = 0;
            static const char para_break[] = "--------------";

                              //collect all frames
            vector<PI3D_model> uni_mod_list;
            struct S_hlp{
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                  PI3D_model mod = I3DCAST_MODEL(frm);

                  vector<PI3D_model> &uni_mod_list = *(vector<PI3D_model>*)c;
                  for(int i=uni_mod_list.size(); i--; ){
                     if(!strcmp(uni_mod_list[i]->GetFileName(), mod->GetFileName()))
                        break;
                  }
                  if(i==-1)
                     uni_mod_list.push_back(mod);
                  return I3DENUMRET_OK;
               }
               static bool cbLess(PI3D_model m1, PI3D_model m2){
                  return (strcmp(m1->GetFileName(), m2->GetFileName()) < 0);
               }
               static I3DENUMRET I3DAPI cbEnum1(PI3D_frame frm, dword c){
                  return I3DENUMRET_OK;
               }
               static bool cbLessTxt(PI3D_texture t1, PI3D_texture t2){
                  return (strcmp(t1->GetFileName(), t2->GetFileName()) < 0);
               }
               
               static bool cbLessString(const C_string &s1, const C_string &s2){
                  return false;
               }
               static bool cbLessSnd(PI3D_sound s1, PI3D_sound s2){
                  return (strcmp(s1->GetFileName(), s2->GetFileName()) < 0);
               }

               static I3DENUMRET I3DAPI cbEnumSnd(PI3D_frame frm, dword c){

                  PI3D_sound snd = I3DCAST_SOUND(frm);

                  vector<PI3D_sound> &uni_snd_list = *(vector<PI3D_sound>*)c;
                  for(int i=uni_snd_list.size(); i--; ){
                     if(!strcmp(uni_snd_list[i]->GetFileName(), snd->GetFileName()))
                        break;
                  }
                  if(i==-1)
                     uni_snd_list.push_back(snd);
                  return I3DENUMRET_OK;
               }
            };
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&uni_mod_list, ENUMF_MODEL | ENUMF_BACKDROP);
                              //remove non-created models
            for(i=uni_mod_list.size(); i--; ){
               PI3D_model mod = uni_mod_list[i];
               dword flgs = e_modify->Action(E_MODIFY_GET_FLAGS, mod);
               if(!(flgs&E_MODIFY_FLG_CREATE)){
                  uni_mod_list[i] = uni_mod_list.back();
                  uni_mod_list.pop_back();
               }
            }

                              //sort models by name
            sort(uni_mod_list.begin(), uni_mod_list.end(), S_hlp::cbLess);
            e_log->Action(E_LOG_ADD_TEXT, "Loaded model files:");

                              //compute total size / copy models
            dword total_size = 0;
            for(i=0; i<uni_mod_list.size(); i++){
               C_string fname = uni_mod_list[i]->GetFileName();
               fname += ".i3d";
               dword size = GetFileSize(fname);
               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)C_fstring("\"%s\"\tsize: %i", (const char*)fname, size));
               if(size!=-1)
                  total_size += size;

               if(*res_copy_path && fname[0]!='+'){
                              //copy models to required directory
                  bool ok = false;

                  C_fstring dst_name("%s\\%s", res_copy_path, (const char*)fname);
                  if(WinCreateDirectoryTree(dst_name)){
                     dword size = GetFileSize(fname);
                     if(size!=-1){
                        vector<byte> buf(size);
                        if(Load(fname, buf.begin(), size) == size){
                           if(Save(dst_name, buf.begin(), size) == size){
                              ok = true;
                           }
                        }
                     }
                  }
                  if(!ok){
                     e_log->Action(E_LOG_ADD_RICH_TEXT,
                        &pair<const char*, dword>(C_fstring("Error: failed to copy model '%s' to '%s'", (const char*)fname, (const char*)dst_name),
                        0xffff0000));
                     ++num_errs;
                  }
               }
            }
            e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)C_fstring("total size: %i", total_size));
            e_log->Action(E_LOG_ADD_TEXT, (void*)para_break);

                              //collect all objects
            vector<PI3D_frame> frm_list;
            struct S_hlp1{
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                  vector<PI3D_frame> &frm_list = *(vector<PI3D_frame>*)c;
                  frm_list.push_back(frm);
                  return I3DENUMRET_OK;
               }
            };
            ed->GetScene()->EnumFrames(S_hlp1::cbEnum, (dword)&frm_list, ENUMF_ALL | ENUMF_BACKDROP);

                              //collect unique meshes
            set<PI3D_mesh_base> mb_list;
            set<PI3D_material> mat_list;

            for(i=frm_list.size(); i--; ){
               PI3D_frame frm = frm_list[i];
               switch(frm->GetType()){
               case FRAME_VISUAL:
                  {
                     PI3D_visual vis = I3DCAST_VISUAL(frm);
                     PI3D_mesh_base mb = vis->GetMesh();
                     if(mb){
                        mb_list.insert(mb);
                     }else{
                              //specialized visuals
                        switch(vis->GetVisualType()){
                        case I3D_VISUAL_PARTICLE:
                           mat_list.insert(((PI3D_particle)vis)->GetMaterial());
                           break;
                        case I3D_VISUAL_FLARE:
                           break;
                        }
                     }
                  }
                  break;
               }
            }

                              //collect unique materials
            for(set<PI3D_mesh_base>::const_iterator it=mb_list.begin(); it!=mb_list.end(); it++){
               PI3D_mesh_base mb = *it;
               for(int j=mb->NumFGroups(); j--; ){
                  mat_list.insert(mb->GetFGroup(j)->mat);
               }
            }

                              //collect unique textures
            set<C_string> txt_name_list;
            set<PI3D_material>::const_iterator mat_it;
            for(mat_it=mat_list.begin(); mat_it!=mat_list.end(); mat_it++){
               PI3D_material mat = *mat_it;
               for(int mti=0; mti<MTI_LAST; mti++){
                  PI3D_texture tp = mat->GetTexture((I3D_MATERIAL_TEXTURE_INDEX)mti);
                  if(tp){
                     for(i=0; i<3; i++){
                        const char *fname = tp->GetFileName(i);
                        if(fname)
                           txt_name_list.insert(fname);
                     }
                  }
               }
            }

                              //dump textures
            {
               e_log->Action(E_LOG_ADD_TEXT, "Loaded textures:");
               int num_map_dirs = ed->GetDriver()->NumDirs(I3DDIR_MAPS);

               for(set<C_string>::const_iterator it=txt_name_list.begin(); it!=txt_name_list.end(); ++it){
                  const char *fname = *it;
                  e_log->Action(E_LOG_ADD_TEXT, (void*)fname);

                  if(*res_copy_path){
                              //copy map to required directory
                     bool ok = false;
      
                     for(i=0; i<num_map_dirs; i++){

                        const char *dir = ed->GetDriver()->GetDir(I3DDIR_MAPS, i);
                        C_fstring src_name("%s\\%s", dir, fname);
                        C_fstring dst_name("%s\\%s\\%s", res_copy_path, dir, fname);
                        if(WinCreateDirectoryTree(dst_name)){
                           dword size = GetFileSize(src_name);
                           if(size!=-1){
                              vector<byte> buf(size);
                              if(Load(src_name, buf.begin(), size) == size){
                                 if(Save(dst_name, buf.begin(), size) == size){
                                    ok = true;
                                    break;
                                 }
                              }
                           }     
                        }
                     }
                     if(!ok){
                        const char *mat_name = "<unknown>";
                        for(mat_it=mat_list.begin(); mat_it!=mat_list.end(); mat_it++){
                           PI3D_material mat = *mat_it;
                           for(int mti=MTI_LAST; mti--; ){
                              PI3D_texture tp = mat->GetTexture((I3D_MATERIAL_TEXTURE_INDEX)mti);
                              if(tp){
                                 for(i=0; i<3; i++){
                                    const char *fn = tp->GetFileName(i);
                                    if(fn && !strcmp(fname, fn)){
                                       mat_name = mat->GetName();
                                       mti = 0;
                                       break;
                                    }
                                 }
                              }
                           }
                        }
                        e_log->Action(E_LOG_ADD_RICH_TEXT,
                           &pair<const char*, dword>(C_fstring("Error: failed to copy texture '%s' (material '%s')",
                           fname, mat_name),
                           0xffff0000));
                        ++num_errs;
                     }
                  }
               }
               e_log->Action(E_LOG_ADD_TEXT, (void*)para_break);
            }

                              //collect all sounds
            vector<PI3D_sound> uni_snd_list;
            ed->GetScene()->EnumFrames(S_hlp::cbEnumSnd, (dword)&uni_snd_list, ENUMF_SOUND | ENUMF_BACKDROP);

                              //sort sounds by name
            sort(uni_snd_list.begin(), uni_snd_list.end(), S_hlp::cbLessSnd);
            e_log->Action(E_LOG_ADD_TEXT, "Loaded model files:");

                              //compute total size / copy sounds
            total_size = 0;
            for(i=0; i<uni_snd_list.size(); i++){
               C_string fname = uni_snd_list[i]->GetFileName();
               dword size = GetFileSize(fname);
               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)C_fstring("\"%s\"\tsize: %i", (const char*)fname, size));
               if(size!=-1)
                  total_size += size;

               if(*res_copy_path){
                              //copy sounds to required directory
                  bool ok = false;

                  C_fstring dst_name("%s\\%s", res_copy_path, (const char*)fname);
                  if(WinCreateDirectoryTree(dst_name)){
                     dword size = GetFileSize(fname);
                     if(size!=-1){
                        vector<byte> buf(size);
                        if(Load(fname, buf.begin(), size) == size){
                           if(Save(dst_name, buf.begin(), size) == size){
                              ok = true;
                           }
                        }
                     }
                  }
                  if(!ok){
                     e_log->Action(E_LOG_ADD_RICH_TEXT,
                        &pair<const char*, dword>(C_fstring("Error: failed to copy sound '%s' to '%s'", (const char*)fname, (const char*)dst_name),
                        0xffff0000));
                     ++num_errs;
                  }
               }
            }
            e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)C_fstring("total size: %i", total_size));
            e_log->Action(E_LOG_ADD_TEXT, (void*)para_break);

            if(*res_copy_path)
               e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)C_fstring("errors: %i", num_errs));
         }
         break;
      }
      return 0;
   }

//----------------------------

   //virtual void Render(){}

//----------------------------

   virtual bool LoadState(class C_chunk *ck){
                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version!=1) return false;
                              //read other variables
      ck->Read(res_copy_path, sizeof(res_copy_path));
      return true; 
   }

//----------------------------

   virtual bool SaveState(class C_chunk *ck) const{
                              //write version
      byte version = 1;
      ck->Write(&version, sizeof(byte));
                              //write other variables
      ck->Write(res_copy_path, sizeof(res_copy_path));
      return true; 
   }
};

//----------------------------

void CreateValidity(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Validity;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
