#include "all.h"
#include "common.h"



//----------------------------

class C_edit_View: public C_editor_item{
   virtual const char *GetName() const{ return "View"; }
   enum{
      E_HR_BOX, E_BOUND_BOX, E_CLEAR, E_WIRE,
      E_FILTER, E_MIPMAP, E_DITHER, E_FOG,
      E_LIGHTMAP, E_SECTORS, E_PORTALS, E_VOLUMES,
      E_CAMERAS, E_VISUALS, E_LIGHTS, E_SOUNDS,
      E_LM_TXT, E_DUMMYS, E_OCCLUDERS, E_TEXTURES,
      E_DRAWJOINTS, E_USE_OCLUSION, E_ENVMAPPING, E_BUMPMAPPING,
      E_DRAWMIRRORS, E_USESHADOWS, E_DEBUGDRAWSHADOWS, E_DEBUGDRAWSHDREC,
      E_DEBUGDRAWBSP, E_DEBUGDRAWHRDYNAMIC, E_DEBUGDRAWSTATIC,
      E_LAST
   };
   byte state[E_LAST];
   static I3D_RENDERSTATE i3d_map[];

   int lod_quality;

//----------------------------

   const byte *GetDefaultState(){
      static const byte default_state[] = {
                              //hr_bbox, bound_box, clear_bgnd, wire
         false, false, true, false,
                              //filter, mipmap, dither, fog
         true, true, true, true,  
                              //lmapping, sectors, portals, volumes
         true, false, false, false,
                              //cameras, visuals, lights, sounds
         false, true, false, false,  
                              //lm textures, dummys, occluders, textures
         false, false, false, true,
                              //joints, use_occlusion, env_mapping, bumpmapping
         false, true, true, true,
                              //mirrors, use shd, debug shd, debug shd rec
         true, true, false, false,
                              //bsp tree, dynamic tree, static frames, 
         false, false, false
      };
      return default_state;
   }

//----------------------------

   void ApplyState(const byte *state){

      for(int i=0; i<E_LAST; i++){
         ed->CheckMenu(this, 1000+i, state[i]);
         ed->GetDriver()->SetState(i3d_map[i], state[i]);
      }
      ed->GetDriver()->SetState(RS_LOD_INDEX, lod_quality);
      //ed->GetDriver()->SetState(RS_DEBUGMATVISUALIZE, curr_mat_draw);
      ed->GetDriver()->SetState(RS_DEBUGMATVISUALIZE, mat_draw_enabled ? curr_mat_draw : -1);
   }

//----------------------------

                              //material table
   typedef map<int, C_string> t_mat_tab;
   t_mat_tab mat_table;
   int curr_mat_draw;         //material ID from table, -1 uninitialized/invalid
   bool mat_draw_enabled;     //enable/disable drawing

   const char *GetDrawMatName() const{

      if(curr_mat_draw==-1)
         return NULL;
      for(t_mat_tab::const_iterator it=mat_table.begin(); it!=mat_table.end(); it++){
         if((*it).first==curr_mat_draw)
            break;
      }
      if(it==mat_table.end())
         return NULL;
      return (*it).second;
   }

public:
   C_edit_View():
      lod_quality(-1),
      curr_mat_draw(-1),
      mat_draw_enabled(false)
   {}
   virtual bool Init(){
#define DS "&View\\&Debug\\"
      ed->AddShortcut(this, 1000, DS"&Hierarchy boxes\tShift+Ctrl+H", K_H, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1001, DS"&Bounding boxes\tShift+Ctrl+B", K_B, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1003, DS"&Wire-frame\tW", K_W, 0);
      ed->AddShortcut(this, 1018, DS"&Occluders\tU", K_U, 0);
      ed->AddShortcut(this, 1017, DS"&Dummies\tD", K_D, 0);
      ed->AddShortcut(this, 1009, DS"&Sectors\tR", K_R, 0);
      ed->AddShortcut(this, 1010, DS"&Portals\tP", K_P, 0);
      ed->AddShortcut(this, 1011, DS"&Volumes\tShift+Ctrl+Alt+V", K_V, SKEY_CTRL|SKEY_ALT|SKEY_SHIFT);
      ed->AddShortcut(this, 1012, DS"&Cameras\tC", K_C, 0);
      ed->AddShortcut(this, 1014, DS"&Lights\tShift+Ctrl+L", K_L, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1015, DS"&Sounds\tShift+Ctrl+S", K_S, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1016, DS"&LM textures\tShift+Ctrl+Alt+L", K_L, SKEY_CTRL|SKEY_ALT|SKEY_SHIFT);
      ed->AddShortcut(this, 1020, DS"&Joints\tShift+Ctrl+Alt+B", K_B, SKEY_CTRL|SKEY_ALT|SKEY_SHIFT);
      ed->AddShortcut(this, 1026, DS"Dynamic shadow casters\tCtrl+Shift+Alt+D", K_D, SKEY_CTRL|SKEY_SHIFT|SKEY_ALT);
      ed->AddShortcut(this, 1027, DS"Dynamic shadow receivers\tCtrl+Shift+Alt+R", K_R, SKEY_CTRL|SKEY_SHIFT|SKEY_ALT);
      ed->AddShortcut(this, 1028, DS"&Bsp tree\tCtrl+Shift+V", K_V, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1029, DS"D&ynamic collision hiearchy\tCtrl+Shift+Y", K_Y, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1030, DS"St&atic collision frames\tCtrl+Shift+A", K_A, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1100, DS"%i LOD\\Better level\tCtrl+<", K_2C, SKEY_CTRL);
      ed->AddShortcut(this, 1101, DS"LOD\\Worse level\tCtrl+>", K_2E, SKEY_CTRL);
      ed->AddShortcut(this, 1105, DS"LOD\\Default level\tCtrl+Alt+Shift+<", K_2C, SKEY_ALT|SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1102, DS"LOD\\%i Lower scale\tCtrl+Shift+<", K_2C, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1103, DS"LOD\\Increase scale\tCtrl+Shift+>", K_2E, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1104, DS"LOD\\Default scale\tCtrl+Alt+Shift+>", K_2E, SKEY_ALT|SKEY_CTRL|SKEY_SHIFT);

#define MS "&View\\&Debug\\Material\\"
      ed->AddShortcut(this, 1200, MS"Vizualize material\tCtrl+M", K_M, SKEY_CTRL);
      ed->AddShortcut(this, 1201, MS"Disable\tCtrl+Alt+Shift+M", K_M, SKEY_ALT|SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1202, MS"Assign mode\tCtrl+Alt+M", K_M, SKEY_ALT|SKEY_CTRL);

#define QS "&View\\&Quality\\"
      ed->AddShortcut(this, 1002, QS"Cl&ear background\tCtrl+F7", K_F7, SKEY_CTRL);
      ed->AddShortcut(this, 1013, QS"Draw &visuals\tV", K_V, 0);
      ed->AddShortcut(this, 1004, QS"&Filtering\tCtrl+F5", K_F5, SKEY_CTRL);
      ed->AddShortcut(this, 1005, QS"&MipMapping\tCtrl+F6", K_F6, SKEY_CTRL);
      ed->AddShortcut(this, 1006, QS"&Dithering\tCtrl+F8", K_F8, SKEY_CTRL);
      ed->AddShortcut(this, 1007, QS"F&og\tCtrl+F9", K_F9, SKEY_CTRL);
      ed->AddShortcut(this, 1008, QS"&Light-maps\tShift+Ctrl+Alt+T", K_T, SKEY_CTRL|SKEY_ALT|SKEY_SHIFT);
      ed->AddShortcut(this, 1019, QS"&Textures\tCtrl+Shift+T", K_T, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1022, QS"E&nvironment mapping\tCtrl+Shift+N", K_N, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1023, QS"B&ump mapping\tCtrl+Shift+U", K_U, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1024, QS"&Mirrors\tCtrl+Shift+M", K_M, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1025, QS"&Shadows\tCtrl+Shift+D", K_D, SKEY_CTRL|SKEY_SHIFT);
      ed->AddShortcut(this, 1021, QS"&Use occlusion\tShift+Ctrl+Alt+O", K_O, SKEY_CTRL|SKEY_ALT|SKEY_SHIFT);

      ed->AddShortcut(this, 9999, "%30 &Debug\\%61Simulate cr&ash", 0, 0);
                              //init state
      memcpy(state, GetDefaultState(), sizeof(byte)*E_LAST);
      for(int i=0; i<E_LAST; i++) ed->CheckMenu(this, 1000+i, state[i]);
      ApplyState(state);
      return true;
   }
   virtual void Close(){ }
   //virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){}

   virtual dword Action(int id, void *context){

      switch(id){
      case 1000: case 1001: case 1002: case 1003:
      case 1004: case 1005: case 1006: case 1007:
      case 1008: case 1009: case 1010: case 1011:
      case 1012: case 1013: case 1014: case 1015:
      case 1016: case 1017: case 1018: case 1019:
      case 1020: case 1021: case 1022: case 1023:
      case 1024: case 1025: case 1026: case 1027:
      case 1028: case 1029: case 1030:
         {
            int i = id - 1000;
            state[i] = !state[i];
            ed->GetDriver()->SetState(i3d_map[i], state[i]);
            ed->CheckMenu(this, id, state[i]);
            static const char *state_text[E_LAST] = {
               "Hierarchy bounding boxes", "Visual bounding boxes", "Background clearing", "Wire-frame mode",
               "Texture filtering", "Mipmapping", "Dithering", "Fog",
               "Light-mapping", "Draw sectors", "Draw portals", "Draw volumes",
               "Draw cameras", "Draw visuals", "Draw lights", "Draw sounds",
               "Draw light-map cache", "Draw dummys", "Draw occluders", "Textures",
               "Draw joints", "Use occlusion", "Environment mapping", "Bump mapping",
               "Draw mirrors", "Use shadows", "Debug shadows", "Draw shadow receivers",
               "Draw bsp tree", "Draw dynamic tree", "Draw static frames"
            };
            ed->Message(C_fstring("%s %s", state_text[i], state[i] ? "on" : "off"));
         }
         break;

      case 1200:
         {
                              //visualize material (choose from list)
            vector<char> vc;
            int curr_sel = -1;
            int i = 0;
            for(t_mat_tab::const_iterator it=mat_table.begin(); it!=mat_table.end(); it++, i++){
               const C_string &str = (*it).second;
               vc.insert(vc.end(), (const char*)str, (const char*)str+str.Size()+1);
               if((*it).first==curr_mat_draw)
                  curr_sel = i;
            }
            vc.push_back(0);
            int indx = ChooseItemFromList(ed->GetIGraph(), "Choose material to draw",
               vc.begin(), curr_sel);
            if(indx==-1){
               //Action(1201, 0);
               break;
            }
            for(it=mat_table.begin(); indx--; it++);
            assert(it!=mat_table.end());
            curr_mat_draw = (*it).first;
            ed->Message(C_fstring("Drawing material '%s'", (const char*)(*it).second));
            mat_draw_enabled = true;
            ed->GetDriver()->SetState(RS_DEBUGMATVISUALIZE, curr_mat_draw);
         }
         break;

      case 1201:
         {
            //ed->GetDriver()->SetState(RS_DEBUGMATVISUALIZE, curr_mat_draw = -1);
            //ed->Message("Material visualization disebled");
            mat_draw_enabled = !mat_draw_enabled;
            ed->GetDriver()->SetState(RS_DEBUGMATVISUALIZE, mat_draw_enabled ? curr_mat_draw : -1);
            ed->Message(C_fstring("Material visualization %s", mat_draw_enabled ? "enebled" : "disabled"));
         }
         break;

      case 1202:
         {
            if(curr_mat_draw==-1){
                              //select material to vizualize
               Action(1200, 0);
               if(curr_mat_draw==-1)
                  break;
            }
            const char *curr_mat_name = GetDrawMatName();
            if(!curr_mat_name){
                              //invalid material selected, disable mode
               Action(1201, 0);
               break;
            }
                              //enanble drawing if disabled
            if(!mat_draw_enabled){
               mat_draw_enabled = true;
               ed->GetDriver()->SetState(RS_DEBUGMATVISUALIZE, mat_draw_enabled ? curr_mat_draw : -1);
            }
                              //switch to user pick mode
            S_MouseEdit_pick mp = {this, 1203, LoadCursor(GetHInstance(), MAKEINTRESOURCE(IDC_SEED_PUTCURSOR))};
            ed->FindPlugin("MouseEdit")->Action(E_MOUSE_SET_USER_PICK, &mp);

            ed->Message(C_fstring("Assigning material '%s'", curr_mat_name));
         }
         break;

      case 1203:
         {
            if(!mat_draw_enabled)
               break;
            const char *curr_mat_name = GetDrawMatName();
            if(!curr_mat_name)
               break;
            assert(curr_mat_draw != -1);
            const S_MouseEdit_picked *mp = (S_MouseEdit_picked*)context;
                              //do own (normal-collision) picking
            I3D_collision_data cd;
            cd.from = mp->pick_from;
            cd.dir = mp->pick_dir;
            cd.flags = I3DCOL_TEXTURE | I3DCOL_RAY;
            if(ed->GetScene()->TestCollision(cd) && ed->CanModify()){

               PI3D_frame frm = cd.GetHitFrm();
               int save_mat = -1;
               switch(frm->GetType()){
               case FRAME_VISUAL:
                  {
                     PI3D_visual vis = I3DCAST_VISUAL(frm);
                              //check if volume marked 'static'
                     if(!(vis->GetFlags()&I3D_FRMF_STATIC_COLLISION)){
                        ed->Message(C_fstring("Visual '%s' has not static collision!", frm->GetName()));
                        break;
                     }
                     if(vis->GetCollisionMaterial()!=curr_mat_draw){
                        save_mat = vis->GetCollisionMaterial();
                        vis->SetCollisionMaterial(curr_mat_draw);
                        ed->Message(C_fstring("Visual '%s': material '%s' assigned", frm->GetName(), curr_mat_name));
                        ed->SetModified();
                     }
                  }
                  break;
               case FRAME_VOLUME:
                  {
                     PI3D_volume vol = I3DCAST_VOLUME(frm);
                     if(vol->GetCollisionMaterial()!=curr_mat_draw){
                        save_mat = vol->GetCollisionMaterial();
                        vol->SetCollisionMaterial(curr_mat_draw);
                        ed->Message(C_fstring("Volume '%s': material '%s' assigned", frm->GetName(), curr_mat_name));
                        ed->SetModified();
                     }
                  }
                  break;
               }
               if(save_mat!=-1){
                  LPC_editor_item e_undo = ed->FindPlugin("Undo");
                  int len = sizeof(int) + strlen(frm->GetName()) + 1;
                  char *buf = new char[len];
                  *(dword*)buf = save_mat;
                  strcpy(buf+4, frm->GetName());
                  S_edit_Undo_2 eu(GetName(), 1204, buf, len);
                  e_undo->Action(E_UNDO_SAVE, &eu);
                  delete[] buf;
               }
            }
            return true;
         }
         break;

      case 1204:
         {
            int &save_mat = *(int*)context;
            int mat_id = save_mat;
            const char *name = ((const char*)context)+4;
            PI3D_frame frm = ed->GetScene()->FindFrame(name);
            if(frm){
               switch(frm->GetType()){
               case FRAME_VISUAL:
                  {
                     PI3D_visual vis = I3DCAST_VISUAL(frm);
                     save_mat = vis->GetCollisionMaterial();
                     vis->SetCollisionMaterial(mat_id);
                     ed->Message(C_fstring("Visual '%s': material ID restored", name));
                     ed->SetModified();
                  }
                  break;
               case FRAME_VOLUME:
                  {
                     PI3D_volume vol = I3DCAST_VOLUME(frm);
                     save_mat = vol->GetCollisionMaterial();
                     vol->SetCollisionMaterial(mat_id);
                     ed->Message(C_fstring("Volume '%s': material ID restored", name));
                     ed->SetModified();
                  }
                  break;
               }
                              //save undo back, using the same buffer
               LPC_editor_item e_undo = ed->FindPlugin("Undo");
               int len = sizeof(int) + strlen(name) + 1;
               S_edit_Undo_2 eu(GetName(), 1204, context, len);
               e_undo->Action(E_UNDO_SAVE, &eu);
            }
         }
         break;

      case 1100:
      case 1101:
      case 1105:
         {
            switch(id){
            case 1100:
               lod_quality = Max(-1, lod_quality - 1);
               break;
            case 1101:
               ++lod_quality;
               break;
            case 1105:
               lod_quality = -1;
               break;
            default: assert(0);
            }
            ed->GetDriver()->SetState(RS_LOD_INDEX, lod_quality);
            ed->Message(C_fstring("LOD quality: %i", lod_quality));
         }
         break;

      case 1102:
      case 1103:
      case 1104:
         {
            float curr_scale = IntAsFloat(ed->GetDriver()->GetState(RS_LOD_SCALE));
            const float sd = .8f, su = 1.25f;
            switch(id){
            case 1102:
               curr_scale = Max(sd*sd*sd*sd*sd*sd*sd*sd*sd*sd, curr_scale * sd);
               break;
            case 1103:
               curr_scale = Min(su*su*su*su*su*su*su*su*su*su*su, curr_scale * su);
               break;
            case 1104:
               curr_scale = 1.0f;
               break;
            default: assert(0);
            }
            ed->GetDriver()->SetState(RS_LOD_SCALE, FloatAsInt(curr_scale));
            ed->Message(C_fstring("LOD scale: %.2f", curr_scale));
         }
         break;

      case 9999:
         *(char*)NULL = 0;
         break;

      case EDIT_ACTION_SET_ACTIVE:
         //ApplyState(((bool)context) ? state : GetDefaultState());
         break;

      case EDIT_ACTION_FEED_MATERIALS:
         {
            pair<int, const char*> *pairs = (pair<int, const char*>*)context;

            mat_table.clear();

            for(int i=0; pairs[i].first!=-1; i++)
               mat_table[pairs[i].first] = pairs[i].second;

            //FeedMaterials();
         }
         break;
      }
      return 0;
   }

//----------------------------

   //virtual void Render(){}

//----------------------------

#define STATE_VERSION 0xcd

   virtual bool LoadState(C_chunk *ck){

                              //check version
      byte version = -1;
      ck->Read(&version, sizeof(byte));
      if(version != STATE_VERSION)
         return false;
                              //read state
      ck->Read(&state, sizeof(state));
      ck->Read(&lod_quality, sizeof(lod_quality));
      ck->Read(&curr_mat_draw, sizeof(curr_mat_draw));
      //ck->Read(&mat_draw_enabled, sizeof(mat_draw_enabled));
      ApplyState(state);
      return true;
   }

   virtual bool SaveState(C_chunk *ck) const{

                              //write version
      byte version = STATE_VERSION;
      ck->Write(&version, sizeof(byte));
                              //write state
      ck->Write(&state, sizeof(state));
      ck->Write(&lod_quality, sizeof(lod_quality));
      ck->Write(&curr_mat_draw, sizeof(curr_mat_draw));
      //ck->Write(&mat_draw_enabled, sizeof(mat_draw_enabled));
      return true;
   }

};

I3D_RENDERSTATE C_edit_View::i3d_map[E_LAST] = {
   RS_DRAWHRBOUNDBOX, RS_DRAWBOUNDBOX, RS_CLEAR, RS_WIREFRAME,
   RS_LINEARFILTER, RS_MIPMAP, RS_DITHER, RS_FOG,
   RS_USELMAPPING, RS_DRAWSECTORS, RS_DRAWPORTALS, RS_DRAWVOLUMES, 
   RS_DRAWCAMERAS, RS_DRAWVISUALS, RS_DRAWLIGHTS, RS_DRAWSOUNDS,
   RS_DRAWLMTEXTURES, RS_DRAWDUMMYS, RS_DRAWOCCLUDERS, RS_DRAWTEXTURES,
   RS_DRAWJOINTS, RS_USE_OCCLUSION, RS_ENVMAPPING, RS_BUMPMAPPING,
   RS_DRAWMIRRORS, RS_USESHADOWS, RS_DEBUGDRAWSHADOWS, RS_DEBUGDRAWSHDRECS,
   RS_DEBUGDRAWBSP, RS_DEBUGDRAWDYNAMIC, RS_DEBUGDRAWSTATIC
};

//----------------------------

void CreateView(LPC_editor ed){
   LPC_editor_item ei = new C_edit_View;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------