#include <windows.h>
#include <C_string.hpp>
#include <tlhelp32.h>
#include <advanced.h>         //Save
#include <iexcpt.h>
#include <vector>
#include <algorithm>
#include <c_cache.h>
#include <debug.h>
#include <process.h>
#include <cfloat>
#include <imagehlp.h>
#include <win_reg.h>
#include <direct.h>
#include <fstream>
#include "resource.h"

using namespace std;

#pragma comment(lib, "th32")

static const char *dll_name = "iexcpt.dll";   //name of this DLL for resource location

#define MAX_RECURSE 40         //maximal number of function to recurse up the stack

#define USE_WIN_STACKWALK     //use StackWalk from Win32 API

#ifdef _DEBUG
//#define DEBUG_CRASH_TEST
#endif

//----------------------------
                              //network address is in resources
static bool (__stdcall *send_function)(const char*);
static void (__stdcall *callback_function)();

static char *log_dump;
static bool allow_close = false;
//----------------------------

static BOOL CALLBACK dlgCrash(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

   switch(uMsg){
   case WM_INITDIALOG:
      {
         RECT rc;
         GetWindowRect(hwnd, &rc);
         int x = (GetSystemMetrics(SM_CXSCREEN)-(rc.right-rc.left))/2;
         int y = (GetSystemMetrics(SM_CYSCREEN)-(rc.bottom-rc.top))/2;
         SetWindowPos(hwnd, NULL, x, y, 0, 0, SWP_NOACTIVATE | SWP_NOREDRAW | SWP_NOSIZE);

         EnableWindow(GetDlgItem(hwnd, IDCLOSE), allow_close);

         ShowWindow(hwnd, SW_SHOW);

         SetDlgItemText(hwnd, IDC_EDIT, (const char*)lParam);
      }
      break;
   case WM_COMMAND:
      switch(LOWORD(wParam)){
      case IDC_SEND_OVER_NET:
      case IDC_SAVE_TO_FILE:
         {
                              //get comment
            char *buf = new char[2048];
            GetDlgItemText(hwnd, IDC_COMMENT, buf, 2048);
            if(*buf){
               strcat(log_dump, "\r\nComment:\r\n\r\n");
               strcat(log_dump, buf);
            }
            delete[] buf;
         }
                              //flow...
      case IDCLOSE:
      case IDCANCEL:
      case IDC_DEBUG:
         EndDialog(hwnd, LOWORD(wParam));
         break;
      case IDC_ADD_COMMENT:
         {
            RECT rc;
            GetWindowRect(hwnd, &rc);

            SetWindowPos(hwnd, NULL, 0, 0, rc.right-rc.left, (rc.bottom-rc.top)*1.33f,
              SWP_NOACTIVATE | SWP_NOMOVE);
            EnableWindow(GetDlgItem(hwnd, 3), false);
            SetFocus(GetDlgItem(hwnd, IDC_COMMENT));
         }
         break;
      }
      break;
   }

   return 0;
}

//----------------------------

                              //info about single module
struct S_module_info{
   C_string mod_name;         //name of module (exe|dll)
   C_string exe_path;         //full path to module
   byte *base_addr;           //base address
   dword mod_size;            //size
   dword file_time[2];
                              //single function info
   struct S_function{
      C_string name;          //function name
      byte *addr;             //function address (relocated)
      int source_file;        //index into source_files; -1 = unknown
      S_function():
         source_file(-1)
      {}
      S_function(const S_function &f){
         operator =(f);
      }
      S_function &operator = (const S_function &f){
         name = f.name;
         addr = f.addr;
         source_file = f.source_file;
         return *this;
      }
      bool operator < (const S_function &f) const{
         return (addr<f.addr);
      }
   };
   mutable vector<S_function> func_list;  //list of all functions
   mutable bool db_info_searched;         //flag set if this module has searched for info

   S_module_info():
      db_info_searched(false)
   {}

   bool OpenDebugInfo() const{
      int i;
                              //parse map file
      C_cache ics;
      ifstream ifs;
      bool use_ifstream;

      char fname[256];
      strcpy(fname, exe_path);
      for(i=exe_path.Size(); i--; ){
         if(fname[i]=='.')
            break;
      }
      if(i)
         strcpy(&fname[i], ".map");

      char cwd[512];
      _getcwd(cwd, sizeof(cwd));

      i = -1;
      if(strlen(cwd) <= strlen(fname)){
         if(!strnicmp(cwd, fname, strlen(cwd))){
            i = strlen(cwd);
            if(fname[i]=='\\')
               ++i;
         }
      }
      bool open_ok = false;

      if(i!=-1){
         use_ifstream = true;
         ifs.open(&fname[i], ios::in);
         open_ok = (!ifs.fail());
         if(!open_ok){
            use_ifstream = false;
            ics.open(&fname[i], CACHE_READ);
            open_ok = (!ics.fail());
         }
      }
                              //if failed, try full path (not using DTA)
      if(!open_ok){
         use_ifstream = true;
         ifs.open(fname, ios::in);
         open_ok = (!ifs.fail());
      }

      if(!open_ok)
         return false;

                           //file found
      enum E_INFO{
         INFO_MODULE_NAME,
         INFO_TIMESTAMP,
         INFO_LOAD_ADDRESS,
         INFO_ADDRESS,
         INFO_REST,
      } next_info = INFO_MODULE_NAME;
      int reloc_delta = 0;
      bool result = true;
      while(true){
         char line[4096];
         if(use_ifstream){
            ifs.getline(line, sizeof(line));
            if(ifs.eof())
               break;
         }else{
            ics.getline(line, sizeof(line));
            if(ics.eof())
               break;
         }
         for(int i=0; line[i]; i++){
            if(!isspace(line[i]))
               break;
         }
         if(!line[i])
            continue;
         C_string str = &line[i];

         switch(next_info){
         case INFO_MODULE_NAME:
            {
               C_string mname = mod_name;
                              //check if module name matches this string
               for(i=mname.Size(); i--; ){
                  if(mname[i]=='.'){
                     mname[i] = 0;
                     mname = (const char*)mname;
                     break;
                  }
               }
               str.ToLower();
               if(!(mname==str)){
                              //module name mismatch?
                  result = false;
                  goto finish;
               }
               next_info = INFO_TIMESTAMP;
            }
            break;

         case INFO_TIMESTAMP:
            {
               /*
               dword timestamp;
               char dayname[12], mnt[12];
               int year, day, hr, min, sec;
               int num = sscanf(str, "Timestamp is %x (%s %s %i %i:%i:%i %i)", &timestamp, &dayname, &mnt, &day, &hr, &min, &sec, &year);
               if(num!=8){
                  result = false;
                  goto finish;
               }
               static const char *month_names[] = {
                  "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"
               };
               for(int i=12; i--; ){
                  if(!stricmp(month_names[i], mnt))
                     break;
               }
               if(i!=-1){
                              //check date
                  FILETIME ft;
                  SYSTEMTIME st;
                  st.wYear = year;
                  st.wMonth = i+1;
                  st.wDayOfWeek = 0;
                  st.wDay = day;
                  st.wHour = hr;
                  st.wMinute = min;
                  st.wSecond = sec;
                  st.wMilliseconds = 0;
                  SystemTimeToFileTime(&st, &ft);

                  __int64 time_delta = (*(__int64*)&ft) - (*(__int64*)file_time);
                  if(time_delta < __int64(0x4000000) && time_delta>-__int64(0x4000000)){
                     next_info = INFO_LOAD_ADDRESS;
                     break;
                  }
               }
               result = false;
               goto finish;
               */
               next_info = INFO_LOAD_ADDRESS;
            }
            break;

         case INFO_LOAD_ADDRESS:
            {
               static const char search_text[] = "Preferred load address is*";
               if(str.Match(search_text)){
                              //read relocation base
                  const char *cp = &str[(int)sizeof(search_text)-1];
                  if(cp){
                     byte *reloc_base;
                     sscanf(cp, "%x", &reloc_base);
                     reloc_delta = base_addr - reloc_base;
                     next_info = INFO_ADDRESS;
                     break;
                  }
               }
               result = false;
               goto finish;
            }
            break;

         case INFO_ADDRESS:
            if(str.Match("Address*")){
                              //found what we want, skip 1 more line
               next_info = INFO_REST;
            }
            break;

         case INFO_REST:
                              //parse data
            {
                              //scan address and function name
               byte *seg, *offs;
               int num = sscanf(str, "%x:%x", &seg, &offs);
               if(num!=2)
                  break;
               func_list.push_back(S_function());
               const int FIRST_CHAR = 20;
               if(str.Size() > FIRST_CHAR){
                  int i, j = -1;
                  for(i=FIRST_CHAR; str[i]; i++)
                  if(str[i]==' '){
                     str[i] = 0;
                     j = i+1;
                     while(str[j] && isspace(str[j])) ++j;
                     sscanf(&str[j], "%x", &offs);
                     while(str[j] && !isspace(str[j])) ++j;
                     while(str[j] && isspace(str[j])) ++j;
                     while(str[j] && !isspace(str[j])) ++j;
                     while(str[j] && isspace(str[j])) ++j;
                     break;
                  }
                  if(j==-1)
                     func_list.back().name = &str[FIRST_CHAR];
                  else{
                     func_list.back().name = &str[FIRST_CHAR];
                  }
               }
               func_list.back().addr = offs + reloc_delta;
            }
            break;
         }
      }
   finish:
      if(use_ifstream)
         ifs.close();
      else
         ics.close();
      sort(func_list.begin(), func_list.end());
      return result;
   }
                              //check if address is inside of this module address space
   bool ContainAddress(void *addr) const{

      byte *bp = (byte*)addr;
      if(bp>=base_addr && bp<base_addr+mod_size)
         return true;
      return false;
   }

//----------------------------
                              //prepare formatted info for single address
   void *FormatInfo(void *addr, C_string &str, bool *is_main) const{

      if(!db_info_searched){
         db_info_searched = true;
         OpenDebugInfo();
      }

      byte *bp = (byte*)addr;
      *is_main = false;
                              //find which function it may be
      for(int i=func_list.size(); i--; )
      if(bp>=func_list[i].addr){
         const C_string &fnc_name = func_list[i].name;
         char un_name[4096];
         UnDecorateSymbolName(fnc_name, un_name, sizeof(un_name),
            UNDNAME_COMPLETE | UNDNAME_NO_MS_KEYWORDS | 
            //UNDNAME_NO_THISTYPE
            UNDNAME_NO_ACCESS_SPECIFIERS |
            UNDNAME_NO_MEMBER_TYPE
            );

         str = C_fstring("%s, offset 0x%x", un_name,
            bp-func_list[i].addr);
         if(fnc_name.Matchi("_main*")) *is_main = true;
         else
         if(fnc_name.Matchi("?Main*")) *is_main = true;
         else
         if(fnc_name.Matchi("_WinMain@16*"))
            *is_main = true;

         return func_list[i].addr;
      }
      str = "<unknown function>";
      return NULL;
   }
};

//----------------------------
                              //build list of modules
static void BuildModuleInfo(vector<S_module_info> &mod_info){

   HANDLE h = CreateToolhelp32Snapshot(TH32CS_SNAPALL, 0);

   MODULEENTRY32 m32;
   m32.dwSize = sizeof(m32);
   if(Module32First(h, &m32)){
      do{
         mod_info.push_back(S_module_info());
         S_module_info &mi = mod_info.back();
         mi.mod_name = m32.szModule;
         mi.mod_name.ToLower();
         mi.exe_path = m32.szExePath;
         mi.exe_path.ToLower();
         mi.base_addr = (byte*)m32.hModule;
         mi.mod_size = m32.modBaseSize;

         memset(mi.file_time, 0, sizeof(mi.file_time));
         HANDLE h = CreateFile(mi.exe_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
         if(h!=INVALID_HANDLE_VALUE){
            FILETIME ft;
            GetFileTime(h, NULL, NULL, &ft);
            FileTimeToLocalFileTime(&ft, (LPFILETIME)mi.file_time);
            CloseHandle(h);
         }
      } while(Module32Next(h, &m32));
   }
}

//----------------------------
                              //find module in which given address is contained
static const S_module_info *FindModule(void *addr, const vector<S_module_info> &mod_info){

   for(int i=0; i<mod_info.size(); i++){
      if(mod_info[i].ContainAddress(addr))
         return &mod_info[i];
   }
   return NULL;
}

//----------------------------

static const struct{
   dword id;
   const char *name;
} except_code[] = {
   0, "unknown exception",
   EXCEPTION_ACCESS_VIOLATION, "ACCESS_VIOLATION",
   EXCEPTION_IN_PAGE_ERROR, "IN_PAGE_ERROR",
   EXCEPTION_INVALID_HANDLE, "INVALID_HANDLE",
   EXCEPTION_ILLEGAL_INSTRUCTION, "ILLEGAL_INSTRUCTION",
   EXCEPTION_NONCONTINUABLE_EXCEPTION, "NONCONTINUABLE_EXCEPTION",
   EXCEPTION_INVALID_DISPOSITION, "INVALID_DISPOSITION",
   EXCEPTION_ARRAY_BOUNDS_EXCEEDED, "ARRAY_BOUNDS_EXCEEDED",
   EXCEPTION_FLT_DENORMAL_OPERAND, "FLT_DENORMAL_OPERAND",
   EXCEPTION_FLT_DIVIDE_BY_ZERO, "FLOAT_DIVIDE_BY_ZERO",
   EXCEPTION_FLT_INEXACT_RESULT, "FLOAT_INEXACT_RESULT",
   EXCEPTION_FLT_INVALID_OPERATION, "FLOAT_INVALID_OPERATION",
   EXCEPTION_FLT_OVERFLOW, "FLOAT_OVERFLOW",
   EXCEPTION_FLT_STACK_CHECK, "FLOAT_STACK_CHECK",
   EXCEPTION_FLT_UNDERFLOW, "FLOAT_UNDERFLOW",
   EXCEPTION_INT_DIVIDE_BY_ZERO, "INTEGER_DIVIDE_BY_ZERO",
   EXCEPTION_INT_OVERFLOW, "INTEGER_OVERFLOW",
   EXCEPTION_PRIV_INSTRUCTION, "PRIVILEGED_INSTRUCTION",
   EXCEPTION_STACK_OVERFLOW, "STACK_OVERFLOW",
   EXCEPTION_BREAKPOINT, "BREAKPOINT",
   -1
};

//----------------------------

static dword GetExcCodeIndex(dword id){
   dword code_indx = 0;
   do{
      ++code_indx;
      if(except_code[code_indx].id==-1){
         code_indx = 0;       //set unknown
         break;
      }
   } while(except_code[code_indx].id != id);
   return code_indx;
}

//----------------------------

static void BuildCrashInfo(void *ep1, const char *user_message){


   vector<S_module_info> mod_info;
   BuildModuleInfo(mod_info);

   LPEXCEPTION_POINTERS ep = (LPEXCEPTION_POINTERS)ep1;

   CONTEXT *cp = ep->ContextRecord;
#ifndef USE_WIN_STACKWALK
   dword *reg_ebp = (dword*)cp->Ebp;
#endif
   dword *reg_esp = (dword*)cp->Esp;

   C_string str_path;

   const S_module_info *last_module = NULL;
   void *fnc_address = ep->ExceptionRecord->ExceptionAddress;

#ifdef USE_WIN_STACKWALK
                              //structure for stack walking
   STACKFRAME sf;
   memset(&sf, 0, sizeof(sf));

   sf.AddrPC.Offset = cp->Eip;
   sf.AddrPC.Mode = AddrModeFlat;

   sf.AddrFrame.Offset = cp->Ebp;
   sf.AddrFrame.Mode = AddrModeFlat;

   sf.AddrStack.Offset = cp->Esp;
   sf.AddrStack.Mode = AddrModeFlat;
#endif

                              //limit iteration to avoid endless loops
   for(int k=0; k<MAX_RECURSE; k++){
      if(fnc_address){
         const S_module_info *mi = FindModule(fnc_address, mod_info);
         if(!mi){
            if(k) break;
                              //it may be called virtual function on destroyed interface
                              //try function directly on stack
            fnc_address = (void*)reg_esp[0];
            mi = FindModule(fnc_address, mod_info);
            if(!mi)
               break;
         }

         C_string fmt_text;   //formatted info line
         bool is_main;
         mi->FormatInfo(fnc_address, fmt_text, &is_main);

         if(last_module!=mi){
                              //dump module name
            if(last_module){
               char *buf = new char[4096*MAX_RECURSE];
               sprintf(buf, "Module: %s\r\n%s", (const char*)last_module->mod_name, (const char*)str_path);
               str_path = buf;
               delete[] buf;
            }
            last_module = mi;
         }
                              //dump info line
         char *buf = new char[4096*MAX_RECURSE];
         sprintf(buf, "     %s\r\n%s", (const char*)fmt_text, (const char*)str_path);
         str_path = buf;
         delete[] buf;

         if(is_main)
            break;
      }

#ifdef USE_WIN_STACKWALK
      if(!StackWalk(IMAGE_FILE_MACHINE_I386, GetCurrentProcess(), GetCurrentThread(), &sf, cp, NULL,
         SymFunctionTableAccess, SymGetModuleBase, NULL))
         break;
      if(!sf.AddrFrame.Offset)
         break;
      fnc_address = (void*)sf.AddrReturn.Offset;
#else

                              //unwind:
                              // - caller's address at stack+4
                              // - caller's ebp at stack+0
      __asm{
         mov eax, reg_ebp
         mov eax, [eax+4]
         mov fnc_address, eax
      }
      if(!fnc_address){
                              //give it a blind try on ESP
                              // (if stack frame is not used, this may find the ebp and address on the stack)
         for(int i=0; i<10; i++){
            __asm{
               mov edx, i
               shl edx, 2

               mov eax, reg_esp
               mov eax, [eax + edx]
               mov reg_ebp, eax

               mov eax, reg_esp
               mov eax, [eax+edx+4]
               mov fnc_address, eax
            }
            if(reg_ebp && fnc_address &&
               FindModule(fnc_address, mod_info))
               break;
         }
         if(i==10)
            break;
      }else{
         reg_ebp = (dword*)reg_ebp[0];
      }
#endif
   }

   if(last_module){
                              //dump module name
      char *buf = new char[4096*MAX_RECURSE];
      sprintf(buf, "Module: %s\r\n%s", (const char*)last_module->mod_name, (const char*)str_path);
      str_path = buf;
      delete[] buf;
   }
                              //get info about machine and time/date
   char cname[256];
   dword sz = 256;
   GetComputerName(cname, &sz);
   SYSTEMTIME st;
   GetLocalTime(&st);

   C_string code_name;

   if(!user_message){
                              //get exception code name
      dword code_indx = 0;
      code_indx = GetExcCodeIndex(ep->ExceptionRecord->ExceptionCode);
      code_name = C_fstring(
         "Address: %x:%.8x, Exception code: %.8x (%s)\r\n"
         "Registers:\r\n"
         "eax = %.8x, "
         "ebx = %.8x, "
         "ecx = %.8x, "
         "edx = %.8x\r\n"
         "esi = %.8x, "
         "edi = %.8x, "
         "ebp = %.8x, "
         "esp = %.8x"
         ,
         cp->SegCs, cp->Eip, ep->ExceptionRecord->ExceptionCode, except_code[code_indx].name,
         cp->Eax,
         cp->Ebx,
         cp->Ecx,
         cp->Edx,
         cp->Esi,
         cp->Edi,
         cp->Ebp,
         cp->Esp
         );
   }else{
      code_name = user_message;
   }

   log_dump = new char[4096 + 4096*MAX_RECURSE];
   sprintf(log_dump,
      "Machine: %s\r\nDate: %i-%i-%i %i:%.2i\r\n"
      "%s\r\n"
      "\r\n"
      "Execution path:\r\n"
      "%s"
      ,
      cname, st.wDay, st.wMonth, st.wYear, st.wHour, st.wMinute,
      (const char*)code_name,
      (const char*)str_path
      );
}

//----------------------------


static long ExceptHandle(LPEXCEPTION_POINTERS ep, const char *user_message){

   static int nest_count = 0;

   if(nest_count==1){         //this is our fault
#ifndef USE_WIN_STACKWALK
      CONTEXT *cp = ep->ContextRecord;
      if(*(byte*)cp->Eip == 0x8b){
         static int dummy[2];
                              //that's probably mov eax, [mem]
         cp->Eax = (dword)&dummy;
         return EXCEPTION_CONTINUE_EXECUTION;
      }
#endif
      return EXCEPTION_CONTINUE_SEARCH;
   }
   ++nest_count;

   if(callback_function && !user_message)
      (*callback_function)();

   BuildCrashInfo(ep, user_message);
                              //fast log into a file
   {
      FILE *fh = fopen("crash.log", "wb");
      fwrite(log_dump, 1, strlen(log_dump), fh);
      fclose(fh);
   }

   bool ok = false;
   while(!ok){
      HINSTANCE hinst = GetModuleHandle(dll_name);
      assert(hinst);

      int i = DialogBoxParam(hinst, "IDD_CRASH", NULL, (DLGPROC)dlgCrash, (LPARAM)(const char*)log_dump);
      switch(i){
      case IDC_SEND_OVER_NET:
         if(send_function){
            if(!(*send_function)(log_dump))
               MessageBox(NULL, "Failed to send crash report.", "Crash Report", MB_OK);
            else{
               ok = true;
               allow_close = true;
            }
         }
         break;

      case IDC_SAVE_TO_FILE:
         {
            char buf[256];
            buf[0] = 0;
            OPENFILENAME of;
            memset(&of, 0, sizeof(of));
            of.lStructSize = sizeof(of);
            of.lpstrFilter = "Crash report (*.err)\0*.err\0";
            of.lpstrFile = buf;
            of.nMaxFile = sizeof(buf);
            of.lpstrFileTitle = buf;
            of.nMaxFileTitle = sizeof(buf);
            of.lpstrDefExt = "err";
            of.lpstrTitle = "Save Crash Report";
            of.Flags = OFN_OVERWRITEPROMPT | OFN_NOREADONLYRETURN;
            bool b = GetSaveFileName(&of);
            if(!b) break;

            ok = (Save(buf, (const char*)log_dump, strlen(log_dump)) != -1);
            if(!ok)
               MessageBox(NULL, "Save failed", "Crash Report", MB_OK);
            else
               allow_close = true;
         }
         break;

      case IDC_DEBUG:
         delete[] log_dump;
         SetUnhandledExceptionFilter(NULL);
         return EXCEPTION_CONTINUE_SEARCH;
      default:
         ok = true;
      }
   }
   delete[] log_dump;

   nest_count = 0;

   return EXCEPTION_EXECUTE_HANDLER;         //handle the exception
}

//----------------------------

static long __stdcall ExceptHandle(LPEXCEPTION_POINTERS ep){

   return ExceptHandle(ep, NULL);
}

//----------------------------

void __declspec(dllexport) __stdcall InitializeExceptions(dword fpu_cw){

   SetUnhandledExceptionFilter(ExceptHandle);
                              //setup FPU control word
   _control87(fpu_cw, _MCW_EM);

#ifdef DEBUG_CRASH_TEST
   *(char*)NULL = 0;
#endif
#if 0
                              //test it in debugger
   EXCEPTION_POINTERS ep;
   EXCEPTION_RECORD er;
   CONTEXT cr;
   memset(&ep, 0, sizeof(ep));
   memset(&er, 0, sizeof(er));
   memset(&cr, 0, sizeof(cr));
   ep.ExceptionRecord = &er;
   ep.ContextRecord = &cr;
   __asm{
      mov cr.Ebp, ebp
      call l1
   l1:
      pop cr.Eip
      mov cr.Esp, esp
   }
   ep.ExceptionRecord->ExceptionAddress = (void*)cr.Eip;
   ExceptHandle(&ep);
#endif
}

//----------------------------

void __declspec(dllexport) __stdcall SetExceptionSendFunc(bool (__stdcall *send_fnc)(const char*)){

   send_function = send_fnc;
}

//----------------------------

void __declspec(dllexport) __stdcall SetExceptionCallback(void (__stdcall *fnc)()){

   callback_function = fnc;
}

//----------------------------

void __declspec(dllexport) __stdcall UserException(const char *msg){

   EXCEPTION_POINTERS ep;
   EXCEPTION_RECORD er;
   CONTEXT cr;
   memset(&ep, 0, sizeof(ep));
   memset(&er, 0, sizeof(er));
   memset(&cr, 0, sizeof(cr));
   ep.ExceptionRecord = &er;
   ep.ContextRecord = &cr;
   er.ExceptionAddress = InitializeExceptions;
   __asm{
      mov eax, [ebp+0]
      mov cr.Ebp, eax
      mov eax, [ebp+4]
      lea edi, er
      mov [edi].ExceptionAddress, eax
   }
   dword rtn = ExceptHandle(&ep, msg);
   if(rtn == EXCEPTION_CONTINUE_SEARCH){
      __asm int 3
   }
}

//----------------------------
