#define NO_STARTUP
#include "all.h"
#include <conv_rgb.hpp>
#include <sortlist.hpp>
#include <assert.h>
#include <vector>

#define RGB_MASTER 1          //g
#define RGB_SLAVE1 0          //r
#define RGB_SLAVE2 2          //b


                              //get rid of 'delete will...' warning message
#ifdef _MSC_VER
using namespace std;
#pragma warning(disable:4731)
#endif

//----------------------------

struct S_rgbi{
   byte rgb[3];
   byte index;
};

static const word alpha_and_mask[4] = {0, 0xf, 0xff, 0xfff};

//----------------------------
                              //help functions
inline int FindLastBit(dword val){
   dword rtn;
   __asm{
      mov eax, val
      bsr eax, eax
      jnz ok
      mov eax,-1
ok:
      mov rtn, eax
   }
   return rtn;
}

#ifdef _MSC_VER

inline dword Rol(dword dw, byte b){
   __asm{
      push ecx
      mov cl, b
      rol dw, cl
      pop ecx
   }
   return dw;
}
inline dword Ror(dword dw, byte b){
   __asm{
      push ecx
      mov cl, b
      ror dw, cl
      pop ecx
   }
   return dw;
}
#endif

//----------------------------

                              //count pixels in packed data
static void CountPackedPixels(const void *area, dword sy,
   dword *pixnum, dword *codenum){

   *pixnum=0;
   *codenum=0;
   for(int yy=0; yy<sy; yy++){
      const byte *mem=(const byte*)area+((dword*)area)[yy];
      dword pl;
      while(++*codenum, pl=*mem++){
         if(!(pl&0x80)) *pixnum+=pl, mem+=pl;
      }
   }
}

//----------------------------
//----------------------------

class C_rgb_conversion_i: public C_rgb_conversion{

   dword ref;

   bool inited;
   S_pixelformat pf;

                              //convert with dithering (called internally)
   void Dither(const void *src, void *dst, dword pitch_src, dword pitch_dst,
      dword size_x, dword size_y, dword srcbpp, const dword *src_pal) const;

   union{
      word *wfield;
      byte *bfield;
      struct S_rgbi *rgbi;
   } r;
   union{
      word *wfield;
      byte *bfield;
      byte *r2rgbi;
   }g;
   union{
      word *wfield;
      byte *bfield;
   }b;
   union{
      word *wfield;
   }a;
   byte *init_pal_argb;

                              //true-color formats:
   int r_byte_pos, g_byte_pos, b_byte_pos, a_byte_pos;


                              //RGB element position - lowest bit of
                              //byte value, must be masked by ?BitMask
   byte r_pos, g_pos, b_pos, a_pos;
public:
   byte RPos() const{ return r_pos; }
   byte GPos() const{ return g_pos; }
   byte BPos() const{ return b_pos; }
   byte APos() const{ return a_pos; }

   dword RBitMask() const{ return pf.r_mask; }
   dword GBitMask() const{ return pf.g_mask; }
   dword BBitMask() const{ return pf.b_mask; }
   dword ABitMask() const{ return pf.a_mask; }
public:

   C_rgb_conversion_i(): inited(false), ref(1)
   {}
   ~C_rgb_conversion_i(){
      UnInit();
   }

   I2DMETHOD_(dword,AddRef)(){ return ++ref; }
   I2DMETHOD_(dword,Release)(){
      if(--ref) return ref;
      delete this;
      return 0;
   }

   I2DMETHOD_(void,Init)(const S_pixelformat &pf, const dword *pal_rgba = NULL);
   I2DMETHOD_(void,UnInit)();
   I2DMETHOD_(bool,Convert)(const void *src, void *dst, dword sx, dword sy,
      dword src_pitch, dword dst_pitch, byte srcbpp, dword **pal, dword flags,
      dword color_key) const;

   I2DMETHOD_(bool,AlphaBlend)(void *dst, void *src, const dword *pal, dword sx, dword sy,
      dword src_pitch, dword dst_pitch, byte srcbpp, dword flags=0) const;

   I2DMETHOD_(dword,GetPixel)(byte r, byte g, byte b, byte a, byte *pal=NULL) const;
   I2DMETHOD_(void,InverseRGB)(dword val, byte *r, byte *g, byte *b, byte *a,
      const S_pixelformat* = NULL, byte *pal=NULL) const;
   I2DMETHOD_(const S_pixelformat*,GetPixelFormat)() const{
      assert(inited);
      return inited ? &pf : NULL;
   }

   I2DMETHOD_(bool,IsPaletized)() const;

   I2DMETHOD_(byte,Bpp)() const;
};

//----------------------------
//----------------------------

void C_rgb_conversion_i::Init(const S_pixelformat &pf1, const dword *pal_argb1){

   UnInit();
                              //init known values
   init_pal_argb = NULL;
   pf = pf1;

   if(!(pf.flags&PIXELFORMAT_COMPRESS)){
                              //get rotation values
      r_pos = (FindLastBit(RBitMask())-7) & 31;
      g_pos = (FindLastBit(GBitMask())-7) & 31;
      b_pos = (FindLastBit(BBitMask())-7) & 31;
      a_pos = (FindLastBit(ABitMask())-7) & 31;

      int i;
      switch(pf.bytes_per_pixel){
      case 1:
         if(!(pf.flags&PIXELFORMAT_PALETTE) || !pal_argb1)
         {                       //8-bit rgb
                              //initialize byte tables
            r.bfield = new byte[256*3];
            g.bfield = r.bfield + 256*1;
            b.bfield = r.bfield + 256*2;
            byte *bp;
                              //create table of reds
            bp = r.bfield, i=0;
            do{
               *bp = Rol(i, r_pos) & pf.r_mask;
            }while(++bp, ++i<256);
                              //create table of greens
            bp = g.bfield, i=0;
            do{
               *bp = Rol(i, g_pos) & pf.g_mask;
            }while(++bp, ++i<256);
                              //create table of blues
            bp = b.bfield, i=0;
            do{
               *bp = Rol(i, b_pos) & pf.b_mask;
            }while(++bp, ++i<256);
         }else{               //initialized by palette
                              //store pal
            init_pal_argb = new byte[256 * sizeof(dword)];
            memcpy(init_pal_argb, pal_argb1, 256 * sizeof(dword));
                              //build look-up sorted array
                              //based on red entries
            r.rgbi = new S_rgbi[256];
            S_sort_entry se32[256];
            byte *palp;
            for(i=0; i<256; i++){
               byte *palp = &init_pal_argb[i*sizeof(dword)];
               se32[i].sort_value=(palp[RGB_MASTER]<<16) | (palp[RGB_SLAVE1]<<8) | palp[RGB_SLAVE2];
               se32[i].data = (void*)i;
            }
            RadixSort32(se32, 256);
            for(i=0; i<256; i++){
               dword indx=(dword)se32[i].data;
               palp = &init_pal_argb[indx*4];
               *(dword*)&r.rgbi[i].rgb=*(dword*)palp;
               r.rgbi[i].index=indx;
            }
            g.r2rgbi=new byte[256];
            int j=0;
            i=0;
            do{
               while(j<255 && r.rgbi[j].rgb[RGB_MASTER]<i) ++j;
               g.r2rgbi[i]=j;
            }while(++i<256);
         }
         break;
      case 2:
                              //initialize word tables
         r.wfield = new word[256*4];
         g.wfield = r.wfield + 256*1;
         b.wfield = r.wfield + 256*2;
         a.wfield = r.wfield + 256*3;
         word *wp;
                              //create table of reds
         wp = r.wfield, i=0;
         do{
            *wp = (Rol(i, r_pos) & pf.r_mask);
         }while(++wp, ++i<256);
                              //create table of greens
         wp = g.wfield, i=0;
         do{
            *wp = (Rol(i, g_pos) & pf.g_mask);
         }while(++wp, ++i<256);
                              //create table of blues
         wp = b.wfield, i=0;
         do{
            *wp = (Rol(i, b_pos) & pf.b_mask);
         }while(++wp, ++i<256);
                              //create table of alphas
         wp = a.wfield, i=0;
         do{
            *wp = (Rol(i, a_pos) & pf.a_mask);
         }while(++wp, ++i<256);
         break;
      case 3:
      case 4:
         r_byte_pos = r_pos/8;
         g_byte_pos = g_pos/8;
         b_byte_pos = b_pos/8;
         a_byte_pos = a_pos/8;
         break;
      }
   }else{
                              //init compressed format
   }
                              //initialized
   inited = true;
}

//----------------------------

void C_rgb_conversion_i::UnInit(){
                              //initialized?
   if(!inited) return;

   if(!(pf.flags&PIXELFORMAT_COMPRESS)){
      switch(pf.bytes_per_pixel){
      case 1:
         if(!(pf.flags&PIXELFORMAT_PALETTE) || !init_pal_argb){
            delete[] r.bfield;
         }else{
            if(init_pal_argb){
               delete[] init_pal_argb;
               delete[] r.rgbi;
               delete[] g.r2rgbi;
            }
         }
         break;
      case 2:
                                 //free word tables
         delete[] r.wfield;
         break;
      }
   }else{                     //compressed
   }
   inited = false;
}

//----------------------------

byte C_rgb_conversion_i::Bpp() const{

   if(!(pf.flags&PIXELFORMAT_COMPRESS)){
      return pf.bytes_per_pixel;
   }else{
      return 0;
   }
}

//----------------------------

bool C_rgb_conversion_i::IsPaletized() const{
   return (pf.flags&PIXELFORMAT_PALETTE);
}

//----------------------------

static const byte uni_pal[256][4] = {
                              //0
   0x10, 0x10, 0x00, 0,
   0x10, 0x10, 0x40, 0,
   0x10, 0x10, 0x80, 0,
   0x10, 0x10, 0xc0, 0,
                              //4
   0x10, 0x30, 0x20, 0,
   0x10, 0x30, 0x60, 0,
   0x10, 0x30, 0xa0, 0,
   0x10, 0x30, 0xe0, 0,
                              //8
   0x10, 0x50, 0x20, 0,
   0x10, 0x50, 0x60, 0,
   0x10, 0x50, 0xa0, 0,
   0x10, 0x50, 0xe0, 0,
                              //c
   0x10, 0x70, 0x20, 0,
   0x10, 0x70, 0x60, 0,
   0x10, 0x70, 0xa0, 0,
   0x10, 0x70, 0xe0, 0,
                              //10
   0x10, 0x90, 0x20, 0,
   0x10, 0x90, 0x60, 0,
   0x10, 0x90, 0xa0, 0,
   0x10, 0x90, 0xe0, 0,
                              //14
   0x10, 0xb0, 0x20, 0,
   0x10, 0xb0, 0x60, 0,
   0x10, 0xb0, 0xa0, 0,
   0x10, 0xb0, 0xe0, 0,
                              //18
   0x10, 0xd0, 0x20, 0,
   0x10, 0xd0, 0x60, 0,
   0x10, 0xd0, 0xa0, 0,
   0x10, 0xd0, 0xe0, 0,
                              //1c
   0x10, 0xf0, 0x20, 0,
   0x10, 0xf0, 0x60, 0,
   0x10, 0xf0, 0xa0, 0,
   0x10, 0xf0, 0xe0, 0,

                              //20
   0x30, 0x10, 0x20, 0,
   0x30, 0x10, 0x60, 0,
   0x30, 0x10, 0xa0, 0,
   0x30, 0x10, 0xe0, 0,
                              //24
   0x30, 0x30, 0x20, 0,
   0x30, 0x30, 0x60, 0,
   0x30, 0x30, 0xa0, 0,
   0x30, 0x30, 0xe0, 0,
                              //28
   0x30, 0x50, 0x20, 0,
   0x30, 0x50, 0x60, 0,
   0x30, 0x50, 0xa0, 0,
   0x30, 0x50, 0xe0, 0,
                              //2c
   0x30, 0x70, 0x20, 0,
   0x30, 0x70, 0x60, 0,
   0x30, 0x70, 0xa0, 0,
   0x30, 0x70, 0xe0, 0,
                              //30
   0x30, 0x90, 0x20, 0,
   0x30, 0x90, 0x60, 0,
   0x30, 0x90, 0xa0, 0,
   0x30, 0x90, 0xe0, 0,
                              //34
   0x30, 0xb0, 0x20, 0,
   0x30, 0xb0, 0x60, 0,
   0x30, 0xb0, 0xa0, 0,
   0x30, 0xb0, 0xe0, 0,
                              //38
   0x30, 0xd0, 0x20, 0,
   0x30, 0xd0, 0x60, 0,
   0x30, 0xd0, 0xa0, 0,
   0x30, 0xd0, 0xe0, 0,
                              //3c
   0x30, 0xf0, 0x20, 0,
   0x30, 0xf0, 0x60, 0,
   0x30, 0xf0, 0xa0, 0,
   0x30, 0xf0, 0xe0, 0,

                              //0
   0x50, 0x10, 0x20, 0,
   0x50, 0x10, 0x60, 0,
   0x50, 0x10, 0xa0, 0,
   0x50, 0x10, 0xe0, 0,
                              //4
   0x50, 0x30, 0x20, 0,
   0x50, 0x30, 0x60, 0,
   0x50, 0x30, 0xa0, 0,
   0x50, 0x30, 0xe0, 0,
                              //8
   0x50, 0x50, 0x20, 0,
   0x50, 0x50, 0x60, 0,
   0x50, 0x50, 0xa0, 0,
   0x50, 0x50, 0xe0, 0,
                              //c
   0x50, 0x70, 0x20, 0,
   0x50, 0x70, 0x60, 0,
   0x50, 0x70, 0xa0, 0,
   0x50, 0x70, 0xe0, 0,
                              //10
   0x50, 0x90, 0x20, 0,
   0x50, 0x90, 0x60, 0,
   0x50, 0x90, 0xa0, 0,
   0x50, 0x90, 0xe0, 0,
                              //14
   0x50, 0xb0, 0x20, 0,
   0x50, 0xb0, 0x60, 0,
   0x50, 0xb0, 0xa0, 0,
   0x50, 0xb0, 0xe0, 0,
                              //18
   0x50, 0xd0, 0x20, 0,
   0x50, 0xd0, 0x60, 0,
   0x50, 0xd0, 0xa0, 0,
   0x50, 0xd0, 0xe0, 0,
                              //1c
   0x50, 0xf0, 0x20, 0,
   0x50, 0xf0, 0x60, 0,
   0x50, 0xf0, 0xa0, 0,
   0x50, 0xf0, 0xe0, 0,

                              //20
   0x70, 0x10, 0x20, 0,
   0x70, 0x10, 0x60, 0,
   0x70, 0x10, 0xa0, 0,
   0x70, 0x10, 0xe0, 0,
                              //24
   0x70, 0x30, 0x20, 0,
   0x70, 0x30, 0x60, 0,
   0x70, 0x30, 0xa0, 0,
   0x70, 0x30, 0xe0, 0,
                              //28
   0x70, 0x50, 0x20, 0,
   0x70, 0x50, 0x60, 0,
   0x70, 0x50, 0xa0, 0,
   0x70, 0x50, 0xe0, 0,
                              //2c
   0x70, 0x70, 0x20, 0,
   0x70, 0x70, 0x60, 0,
   0x70, 0x70, 0xa0, 0,
   0x70, 0x70, 0xe0, 0,
                              //30
   0x70, 0x90, 0x20, 0,
   0x70, 0x90, 0x60, 0,
   0x70, 0x90, 0xa0, 0,
   0x70, 0x90, 0xe0, 0,
                              //34
   0x70, 0xb0, 0x20, 0,
   0x70, 0xb0, 0x60, 0,
   0x70, 0xb0, 0xa0, 0,
   0x70, 0xb0, 0xe0, 0,
                              //38
   0x70, 0xd0, 0x20, 0,
   0x70, 0xd0, 0x60, 0,
   0x70, 0xd0, 0xa0, 0,
   0x70, 0xd0, 0xe0, 0,
                              //3c
   0x70, 0xf0, 0x20, 0,
   0x70, 0xf0, 0x60, 0,
   0x70, 0xf0, 0xa0, 0,
   0x70, 0xf0, 0xe0, 0,

                              //0
   0x90, 0x10, 0x00, 0,
   0x90, 0x10, 0x40, 0,
   0x90, 0x10, 0x80, 0,
   0x90, 0x10, 0xc0, 0,
                              //4
   0x90, 0x30, 0x20, 0,
   0x90, 0x30, 0x60, 0,
   0x90, 0x30, 0xa0, 0,
   0x90, 0x30, 0xe0, 0,
                              //8
   0x90, 0x50, 0x20, 0,
   0x90, 0x50, 0x60, 0,
   0x90, 0x50, 0xa0, 0,
   0x90, 0x50, 0xe0, 0,
                              //c
   0x90, 0x70, 0x20, 0,
   0x90, 0x70, 0x60, 0,
   0x90, 0x70, 0xa0, 0,
   0x90, 0x70, 0xe0, 0,
                              //10
   0x90, 0x90, 0x20, 0,
   0x90, 0x90, 0x60, 0,
   0x90, 0x90, 0xa0, 0,
   0x90, 0x90, 0xe0, 0,
                              //14
   0x90, 0xb0, 0x20, 0,
   0x90, 0xb0, 0x60, 0,
   0x90, 0xb0, 0xa0, 0,
   0x90, 0xb0, 0xe0, 0,
                              //18
   0x90, 0xd0, 0x20, 0,
   0x90, 0xd0, 0x60, 0,
   0x90, 0xd0, 0xa0, 0,
   0x90, 0xd0, 0xe0, 0,
                              //1c
   0x90, 0xf0, 0x20, 0,
   0x90, 0xf0, 0x60, 0,
   0x90, 0xf0, 0xa0, 0,
   0x90, 0xf0, 0xe0, 0,

                              //20
   0xb0, 0x10, 0x20, 0,
   0xb0, 0x10, 0x60, 0,
   0xb0, 0x10, 0xa0, 0,
   0xb0, 0x10, 0xe0, 0,
                              //24
   0xb0, 0x30, 0x20, 0,
   0xb0, 0x30, 0x60, 0,
   0xb0, 0x30, 0xa0, 0,
   0xb0, 0x30, 0xe0, 0,
                              //28
   0xb0, 0x50, 0x20, 0,
   0xb0, 0x50, 0x60, 0,
   0xb0, 0x50, 0xa0, 0,
   0xb0, 0x50, 0xe0, 0,
                              //2c
   0xb0, 0x70, 0x20, 0,
   0xb0, 0x70, 0x60, 0,
   0xb0, 0x70, 0xa0, 0,
   0xb0, 0x70, 0xe0, 0,
                              //30
   0xb0, 0x90, 0x20, 0,
   0xb0, 0x90, 0x60, 0,
   0xb0, 0x90, 0xa0, 0,
   0xb0, 0x90, 0xe0, 0,
                              //34
   0xb0, 0xb0, 0x20, 0,
   0xb0, 0xb0, 0x60, 0,
   0xb0, 0xb0, 0xa0, 0,
   0xb0, 0xb0, 0xe0, 0,
                              //38
   0xb0, 0xd0, 0x20, 0,
   0xb0, 0xd0, 0x60, 0,
   0xb0, 0xd0, 0xa0, 0,
   0xb0, 0xd0, 0xe0, 0,
                              //3c
   0xb0, 0xf0, 0x20, 0,
   0xb0, 0xf0, 0x60, 0,
   0xb0, 0xf0, 0xa0, 0,
   0xb0, 0xf0, 0xe0, 0,

                              //0
   0xd0, 0x10, 0x20, 0,
   0xd0, 0x10, 0x60, 0,
   0xd0, 0x10, 0xa0, 0,
   0xd0, 0x10, 0xe0, 0,
                              //4
   0xd0, 0x30, 0x20, 0,
   0xd0, 0x30, 0x60, 0,
   0xd0, 0x30, 0xa0, 0,
   0xd0, 0x30, 0xe0, 0,
                              //8
   0xd0, 0x50, 0x20, 0,
   0xd0, 0x50, 0x60, 0,
   0xd0, 0x50, 0xa0, 0,
   0xd0, 0x50, 0xe0, 0,
                              //c
   0xd0, 0x70, 0x20, 0,
   0xd0, 0x70, 0x60, 0,
   0xd0, 0x70, 0xa0, 0,
   0xd0, 0x70, 0xe0, 0,
                              //10
   0xd0, 0x90, 0x20, 0,
   0xd0, 0x90, 0x60, 0,
   0xd0, 0x90, 0xa0, 0,
   0xd0, 0x90, 0xe0, 0,
                              //14
   0xd0, 0xb0, 0x20, 0,
   0xd0, 0xb0, 0x60, 0,
   0xd0, 0xb0, 0xa0, 0,
   0xd0, 0xb0, 0xe0, 0,
                              //18
   0xd0, 0xd0, 0x20, 0,
   0xd0, 0xd0, 0x60, 0,
   0xd0, 0xd0, 0xa0, 0,
   0xd0, 0xd0, 0xe0, 0,
                              //1c
   0xd0, 0xf0, 0x20, 0,
   0xd0, 0xf0, 0x60, 0,
   0xd0, 0xf0, 0xa0, 0,
   0xd0, 0xf0, 0xe0, 0,

                              //20
   0xf0, 0x10, 0x20, 0,
   0xf0, 0x10, 0x60, 0,
   0xf0, 0x10, 0xa0, 0,
   0xf0, 0x10, 0xe0, 0,
                              //24
   0xf0, 0x30, 0x20, 0,
   0xf0, 0x30, 0x60, 0,
   0xf0, 0x30, 0xa0, 0,
   0xf0, 0x30, 0xe0, 0,
                              //28
   0xf0, 0x50, 0x20, 0,
   0xf0, 0x50, 0x60, 0,
   0xf0, 0x50, 0xa0, 0,
   0xf0, 0x50, 0xe0, 0,
                              //2c
   0xf0, 0x70, 0x20, 0,
   0xf0, 0x70, 0x60, 0,
   0xf0, 0x70, 0xa0, 0,
   0xf0, 0x70, 0xe0, 0,
                              //30
   0xf0, 0x90, 0x20, 0,
   0xf0, 0x90, 0x60, 0,
   0xf0, 0x90, 0xa0, 0,
   0xf0, 0x90, 0xe0, 0,
                              //34
   0xf0, 0xb0, 0x20, 0,
   0xf0, 0xb0, 0x60, 0,
   0xf0, 0xb0, 0xa0, 0,
   0xf0, 0xb0, 0xe0, 0,
                              //38
   0xf0, 0xd0, 0x20, 0,
   0xf0, 0xd0, 0x60, 0,
   0xf0, 0xd0, 0xa0, 0,
   0xf0, 0xd0, 0xe0, 0,
                              //3c
   0xf0, 0xf0, 0x20, 0,
   0xf0, 0xf0, 0x60, 0,
   0xf0, 0xf0, 0xa0, 0,
   0xf0, 0xf0, 0xe0, 0,
};

//----------------------------

bool C_rgb_conversion_i::Convert(const void *src, void *dst, dword sx, dword sy,
   dword src_pitch, dword dst_pitch, byte srcbpp, dword **pal_argb, dword flags,
   dword color_key) const{

   if(!inited) return false;
   int i;

   dword code;                //used for compressed data
   dword *p_offss;            //pointer to compressed data offsets
   dword pixel;
   dword pnum, cnum;
   bool packed_counted=false;

   assert(dst);
   if(!dst_pitch)
      dst_pitch = sx * (!(pf.flags&PIXELFORMAT_COMPRESS) ? pf.bytes_per_pixel : 2);

   if(flags&MCONV_FLIP_Y){
                              //swap source (destination might be compressed)
      src = ((byte*)src) + src_pitch*(sy-1);
      src_pitch = -src_pitch;
   }
                              //source
   const byte *bp=(const byte*)src;
   union{                     //destination
      byte *p_bp;
      word *p_wp;
      dword *p_dp;
      void *p_vp;
   };
   p_vp = dst;

   if(flags&MCONV_PACKED){
                              //skip offsets
      p_offss=p_dp;
      p_bp += sy*4, bp += sy*4;
   }

   if(!(pf.flags&PIXELFORMAT_COMPRESS))
   switch(srcbpp){
   case 1:
      switch(pf.bytes_per_pixel){
      case 1:
         if(pf.flags&PIXELFORMAT_PALETTE){       //just copy
            if(flags&MCONV_PACKED){
                              //copy offsets
               memcpy(p_offss, src, sy*4);
                              //copy data
               if(!packed_counted) CountPackedPixels((byte*)src, sy, &pnum, &cnum);
               memcpy(p_bp, bp, pnum+cnum);
            }else{
               do{
                  memcpy(p_bp, bp, sx);
               }while(bp += src_pitch, p_bp += dst_pitch, --sy);
            }
         }else{               //map paletized into 8-bit rgb
            if(flags&MCONV_PACKED) return false;
            if(flags&MCONV_DITHER){
               Dither(src, dst, src_pitch, dst_pitch, sx, sy, 1, *pal_argb);
            }else{
               do{
                  for(int xx=0; xx<sx; xx++){
                     const byte *palp_argb = ((byte*)*pal_argb) + bp[xx] * sizeof(dword);

                     p_bp[xx] = r.bfield[palp_argb[0]] |
                        g.bfield[palp_argb[1]] |
                        b.bfield[palp_argb[2]];
                  }
               }while(bp += src_pitch, p_bp += dst_pitch, --sy);
            }
            break;
         }
         break;
      case 2:                 //... hicolor
         if(flags&MCONV_PACKED){
            do{
                              //store new offset
               *p_offss++ = p_bp-(byte*)dst;
                              //make line
                              //get counter & do until eol
               while(code=(*p_bp++)=(*bp++), code){
                  if(!(code&0x80)){
                              //data length
                     do{
                        const byte *palp_argb = ((byte*)*pal_argb) + (*bp) * sizeof(dword);
                        *p_wp =
                           r.wfield[palp_argb[0]] |
                           g.wfield[palp_argb[1]] |
                           b.wfield[palp_argb[2]];
                              //map super-black pixels to non-black ones
                        if(!*p_wp) ++*p_wp;
                     }while(++p_wp, ++bp, --code);
                  }
               }
            }while(--sy);
         }else{               //raw data
            if(!*pal_argb)
               *pal_argb = (dword*)uni_pal[0];
            {
#if defined _MSC_VER && 1     //slower than C++ !
               word *tb = r.wfield;
               word *p_dst = p_wp;

                              //eax - free
                              //ebx - fetch RGB from palette
                              //ecx - row counter
                              //edx - source color index
                              //esi - source
                              //edi - dest
                              //ebp - rgb converson tables base
                              //esp[4] - palette*
                              //esp[8] - size_x
                              //esp[12] - src_pitch
                              //esp[16] - dst_pitch
                              //esp[20] - sy
                              //esp[24] - sx/2
               if(sx&1){
                              //version working with word-s
                  __asm{
                     push ecx

                     mov esi, src
                     mov edi, p_dst
                     mov edx, pal_argb
                     push [edx]
                     push ebp
                     xor edx, edx
                  ly:
                     mov ecx, sx
                     mov ebp, tb
                  lx:

                     mov dl, [esi]
                     inc esi
                     lea ebx, [edx*4]
                     add ebx, [esp+4]
                     mov ebx, [ebx]
                     mov dl, bl
                     and ebx, 0ffffffh
                     mov ax, [ebp+edx*2]
                     mov dl, bh
                     shr ebx, 16
                     or ax, [ebp+200h+edx*2]
                     or ax, [ebp+400h+ebx*2]
                     mov [edi], ax
                     add edi, 2
                     dec ecx
                     jnz lx

                     mov ebp, [esp]
                     sub esi, sx
                     add esi, src_pitch
                     sub edi, sx
                     sub edi, sx
                     add edi, dst_pitch

                     dec sy
                     jne ly

                     pop ebp
                     add esp, 4

                     pop ecx
                  }
               }else{
                              //version working with dword-s
                  __asm{
                     push ecx

                     mov esi, src
                     mov edi, p_dst
                     mov edx, pal_argb
                     push sx
                     shr dword ptr [esp], 1
                     push sy
                     push dst_pitch
                     push src_pitch
                     push sx
                     push [edx]
                     push ebp
                     xor edx, edx
                     mov ebp, tb
                  ly1:
                     mov ecx, [esp+24]
                  lx1:
                     mov dl, [esi+1]
                     lea ebx, [edx*4]
                     add ebx, [esp+4]
                     mov ebx, [ebx]
                     mov dl, bl
                     and ebx, 0ffffffh
                     mov ax, [ebp+edx*2]
                     mov dl, bh
                     shr ebx, 16
                     or ax, [ebp+200h+edx*2]
                     mov dl, [esi]
                     or ax, [ebp+400h+ebx*2]

                     lea ebx, [edx*4]
                     rol eax, 16
                     add ebx, [esp+4]
                     add esi, 2
                     mov ebx, [ebx]
                     mov dl, bl
                     and ebx, 0ffffffh
                     mov ax, [ebp+edx*2]
                     mov dl, bh
                     shr ebx, 16
                     or ax, [ebp+200h+edx*2]
                     add edi, 4
                     or ax, [ebp+400h+ebx*2]
                     
                     dec ecx
                     mov [edi-4], eax

                     jnz lx1

                     sub esi, [esp+8]
                     add esi, [esp+12]
                     sub edi, [esp+8]
                     sub edi, [esp+8]
                     add edi, [esp+16]

                     dec dword ptr [esp+20]
                     jne ly1

                     pop ebp
                     add esp, 24

                     pop ecx
                  }
               }
#else
               for(; --sy; bp += src_pitch, p_bp += dst_pitch){
                  word *p_wp1 = p_wp;
                  const byte *bp1 = bp;
                  for(int sx1=sx; sx1--; ){
                     byte *palp = (byte*)&(*pal_argb)[*bp1++];
                     *p_wp1++ =
                        r.wfield[palp[0]] |
                        g.wfield[palp[1]] |
                        b.wfield[palp[2]];// | a_bit_mask;
                  }
               }
#endif
            }
         }
         break;
      case 3:                 //... 24-bit true color
         if(flags&MCONV_PACKED){
            do{
                              //store new offset
               *p_offss++ = p_bp-(byte*)dst;
                              //make line
                              //get counter & do until eol
               while(code=(*p_bp++)=(*bp++), code){
                  if(!(code&0x80))
                              //data length
                  do{
                     const byte *palp_argb = ((byte*)*pal_argb) + (*bp) * sizeof(dword);
                              //get pixel
                     pixel =
                        (palp_argb[0]<<r_pos) |
                        (palp_argb[1]<<g_pos) |
                        (palp_argb[2]<<b_pos);
                              //store pixel
                     *p_wp = pixel;
                     p_bp[2] = pixel>>16;
                  }while(p_bp+=3, ++bp, --code);
               }
            }while(--sy);
         }else{            //raw data
            do{
               union{
                  word *p_wp1;
                  byte *p_bp1;
               };
               p_bp1=p_bp;
               for(int xx=0; xx<sx; xx++, p_bp1+=3){
                  const byte *palp_argb = ((byte*)*pal_argb) + bp[xx] * sizeof(dword);
                              //get pixel
                  pixel =
                     (palp_argb[0]<<r_pos) |
                     (palp_argb[1]<<g_pos) |
                     (palp_argb[2]<<b_pos);
                              //store pixel
                  *p_wp1=pixel;
                  p_bp1[2]=pixel>>16;
               }
            }while(bp += src_pitch, p_bp += dst_pitch, --sy);
         }
         break;
      case 4:                 //... 32-bit true color
         if(flags&MCONV_PACKED){
            do{
                              //store new offset
               *p_offss++ = p_bp-(byte*)dst;
                              //make line
                              //get counter & do until eol
               while(code=(*p_bp++)=(*bp++), code){
                  if(!(code&0x80))
                              //data length
                  do{
                     const byte *palp_argb = ((byte*)*pal_argb) + (*bp) * sizeof(dword);
                     *p_dp =
                        (palp_argb[0]<<r_pos) |
                        (palp_argb[1]<<g_pos) |
                        (palp_argb[2]<<b_pos);
                  }while(++p_dp, ++bp, --code);
               }
            }while(--sy);
         }else{               //raw data
            if(pal_argb){
               do{
                  const byte *bp1=bp;
                  dword *p_dp1=p_dp;
                  dword xx=sx;
                  do{
                     const byte *palp_argb = ((byte*)*pal_argb) + (*bp1++) * sizeof(dword);
                     *p_dp1++ =
                        (palp_argb[0]<<r_pos) |
                        (palp_argb[1]<<g_pos) |
                        (palp_argb[2]<<b_pos);
                  }while(--xx);
               }while(bp += src_pitch, p_bp += dst_pitch, --sy);
            }else{
               assert(0);
            }
         }
         break;
      }
      break;

   case 2:                    //16-bit to...
      switch(pf.bytes_per_pixel){
      case 2:
         do{                  //just copy
            memcpy(p_bp, bp, sx*2);
         }while(bp += src_pitch, p_bp += dst_pitch, --sy);
         break;
      }
      break;

   case 3:                    //24-bit to...
   case 4:                    //32-bit (rgba) to...
      switch(pf.bytes_per_pixel){
      case 1:                 //... 8-bit
         if(IsPaletized() && init_pal_argb){
            do{               //map into specified palette
               const byte *bp1=bp;
               byte *p_bp1=p_bp;
               dword xx=sx;
               do{
                              //get source rgb
                  int rgb[3];
                  rgb[0] = bp1[0];
                  rgb[1] = bp1[1];
                  rgb[2] = bp1[2];
                              //get closest r value (index into rgbi table)
                  byte index=g.r2rgbi[rgb[RGB_MASTER]];
                              //now walk up & down the list to find closest rgb val
                  S_rgbi *p_rgbi=&r.rgbi[index], *p_rgbi1=p_rgbi;
                  dword best_delta, delta;
                  best_delta=rgb[RGB_MASTER]-(int)p_rgbi->rgb[RGB_MASTER] +
                     abs(rgb[RGB_SLAVE1]-(int)p_rgbi->rgb[RGB_SLAVE1]) +
                     abs(rgb[RGB_SLAVE2]-(int)p_rgbi->rgb[RGB_SLAVE2]);
                  byte best_index=p_rgbi->index;
                              //traverse both directions
                  byte index1=index;

                  while(index1<255 && index){
                     ++index1;
                     dword delta1=(int)(++p_rgbi1)->rgb[RGB_MASTER] - rgb[RGB_MASTER];
                              //check if worth further traversing
                     if(delta1>=best_delta) break;
                              //get full delta
                     delta1 +=
                        abs(rgb[RGB_SLAVE1]-(int)p_rgbi1->rgb[RGB_SLAVE1])+
                        abs(rgb[RGB_SLAVE2]-(int)p_rgbi1->rgb[RGB_SLAVE2]);
                     if(best_delta>delta1){
                        best_delta=delta1;
                        best_index=p_rgbi1->index;
                     }

                     delta=rgb[RGB_MASTER]-(int)(--p_rgbi)->rgb[RGB_MASTER];
                              //check if worth further traversing
                     if(delta>=best_delta) break;
                     --index;
                              //get full delta
                     delta +=
                        abs(rgb[RGB_SLAVE1]-(int)p_rgbi->rgb[RGB_SLAVE1])+
                        abs(rgb[RGB_SLAVE2]-(int)p_rgbi->rgb[RGB_SLAVE2]);
                     if(best_delta>delta){
                        best_delta=delta;
                        best_index=p_rgbi->index;
                     }
                  }
                              //traverse up
                  while(++index1){
                     ++p_rgbi1;
                     delta=(int)p_rgbi1->rgb[RGB_MASTER] - rgb[RGB_MASTER];
                              //check if worth further traversing
                     if(delta>=best_delta) break;
                              //get full delta
                     delta +=
                        abs(rgb[RGB_SLAVE1]-(int)p_rgbi1->rgb[RGB_SLAVE1])+
                        abs(rgb[RGB_SLAVE2]-(int)p_rgbi1->rgb[RGB_SLAVE2]);
                     if(best_delta>delta){
                        best_delta=delta;
                        best_index=p_rgbi1->index;
                     }
                  }
                              //traverse down
                  while(index--){
                     --p_rgbi;
                     delta=rgb[RGB_MASTER]-(int)p_rgbi->rgb[RGB_MASTER];
                              //check if worth further traversing
                     if(delta>=best_delta) break;
                              //get full delta
                     delta +=
                        abs(rgb[RGB_SLAVE1]-(int)p_rgbi->rgb[RGB_SLAVE1])+
                        abs(rgb[RGB_SLAVE2]-(int)p_rgbi->rgb[RGB_SLAVE2]);
                     if(best_delta>delta){
                        best_delta=delta;
                        best_index=p_rgbi->index;
                     }
                  }
                  *p_bp1++ = best_index;
               }while(bp1+=srcbpp, --xx);
            }while(bp += src_pitch, p_bp += dst_pitch, --sy);
         }else{               //3-3-2 mode
            if(IsPaletized()){
                              //paletized mode, create 3-3-2 palette
               byte *bp1;
               bp1 = (byte*)(*pal_argb = new dword[256]);
               bp1[0] = bp1[1] = bp1[2] = 0;
               for(i=1; i<256; i++){
                  bp1 += sizeof(dword);
                  bp1[0]=(i&0xe0) | 0x10;
                  bp1[1]=((i<<3)&0xe0) | 0x10;
                  bp1[2]=(i<<6) | 0x20;
                  bp1[3] = 0xff;
               }
            }
            if(flags&MCONV_DITHER){
               Dither(src, dst, src_pitch, dst_pitch, sx, sy, srcbpp, NULL);
            }else{
               do{
                  const byte *bp1=bp;
                  for(int xx=0; xx<sx; xx++, bp1+=srcbpp){
                     p_bp[xx] = r.bfield[bp1[2]] |
                        g.bfield[bp1[1]] |
                        b.bfield[bp1[0]];
                  }
               }while(bp += src_pitch, p_bp += dst_pitch, --sy);
            }
         }
         break;

      case 2:                 //... 16-bit
         if(flags&MCONV_DITHER){
            Dither(src, dst, src_pitch, dst_pitch, sx, sy, srcbpp, NULL);
         }else{
            do{
               const byte *bp1 = bp;
               switch(srcbpp){
               case 3:
                  {
                     word alpha = a.wfield[0xff];
                     for(int xx=0; xx<sx; xx++, bp1+=3){
                        p_wp[xx] =
                           r.wfield[bp1[2]] |
                           g.wfield[bp1[1]] |
                           b.wfield[bp1[0]] |
                           alpha;
                     }
                  }
                  break;
               case 4:
                  {
                     for(int xx=0; xx<sx; xx++, bp1+=4){
                        p_wp[xx] =
                           r.wfield[bp1[2]] |
                           g.wfield[bp1[1]] |
                           b.wfield[bp1[0]] |
                           a.wfield[bp1[3]];
                     }
                  }
                  break;
               }
            }while(bp += src_pitch, p_bp += dst_pitch, --sy);
         }
         break;

      case 3:                 //... 24-bit
      case 4:                 //... 32-bit
         {
            bool do_memcpy;
            if(!(flags&MCONV_BGR))
               do_memcpy = ((srcbpp==pf.bytes_per_pixel) && (r_byte_pos==2) && (g_byte_pos==1) && (b_byte_pos==0));
            else
               do_memcpy = ((srcbpp==pf.bytes_per_pixel) && (r_byte_pos==0) && (g_byte_pos==1) && (b_byte_pos==2));

            if(do_memcpy){
               do{               //just copy
                  memcpy(p_bp, bp, sx*srcbpp);
               }while(bp += src_pitch, p_bp += dst_pitch, --sy);
            }else{
               dword r_pos = (flags&MCONV_BGR) ? 0 : 2;
               dword b_pos = 2 - r_pos;
               if(srcbpp==4 && pf.bytes_per_pixel==4){
                  do{         //copy with alpha
                     const byte *bp1 = bp;
                     for(int xx=0; xx<sx; xx++, bp1 += srcbpp, ++p_dp){
                        p_bp[r_byte_pos] = bp1[r_pos];
                        p_bp[g_byte_pos] = bp1[1];
                        p_bp[b_byte_pos] = bp1[b_pos];
                        p_bp[a_byte_pos] = bp1[3];
                     }
                  }while(bp += src_pitch, p_bp += dst_pitch - sx*4, --sy);
               }else{
                  do{         //copy without alpha
                     const byte *bp1 = bp;
                     for(int xx=0; xx<sx; xx++, bp1 += srcbpp, p_bp += pf.bytes_per_pixel){
                        p_bp[r_byte_pos] = bp1[r_pos];
                        p_bp[g_byte_pos] = bp1[1];
                        p_bp[b_byte_pos] = bp1[b_pos];
                     }
                  }while(bp += src_pitch, p_bp += dst_pitch - sx*pf.bytes_per_pixel, --sy);
               }
            }
         }
         break;
      }
      break;
   }else{
      if(srcbpp==2)
         return false;
                              //compressed destination
      struct S_surf_block{
         word color0;
         word color1;
         dword pixel_mask;
      };
      S_surf_block *sb = (S_surf_block*)dst;

                              //compress single 4x4 block
      union S_argb{
         struct{
            byte b, g, r, a;
         };
         dword argb;
      };
      dword num_blocks_x = sx/4;
      dword num_blocks_y = sy/4;

      bool has_alpha = (((byte*)&pf.four_cc)[3] != '1');

      static const dword opaque_pixel_map[4] = {1, 3, 2, 0},
         ckey_pixel_map[3] = {1, 2, 0},
         *pixel_maps[2] = {opaque_pixel_map, ckey_pixel_map};
                              //mask full alpha if not 32-bit 8888 RGBA source
      for(int loop_y = 0; loop_y<num_blocks_y; loop_y++){
         for(int loop_x = 0; loop_x<num_blocks_x; loop_x++){
            S_argb block_4x4[16];   //4x4 input block
            byte ckey_4x4[16];      //true if pixel is transparent
            byte block_color_key = false; //will be true if any pixel in block is color-key
                                    //read-in 4x4 block
            switch(srcbpp){
            case 1:
               {
                  i = 0;
                  byte *bp = ((byte*)src) + loop_x*4;
                  for(int y=0; y<4; y++){
                     for(int x=0; x<4; x++, i++){
                        byte input = bp[x];
                        S_argb &output = block_4x4[i];
                        if(flags&MCONV_COLORKEY){
                           byte is_ck = (input==color_key);
                           ckey_4x4[i] = is_ck;
                           block_color_key |= is_ck;
                           if(is_ck) continue;
                        }
                        //block_4x4[i].argb = *(dword*)((*pal) + input*3);
                        output.argb = (*pal_argb)[input];
                              //palette has swapped R and B components
                        swap(output.r, output.b);
                     }
                     bp += src_pitch;
                  }
               }
               break;
            case 3:
            case 4:
               {
                  i = 0;
                  dword *dwp = ((dword*)src) + loop_x * srcbpp;
                  switch(srcbpp){
                  case 3:
                     {
                        for(int y=0; y<4; y++){
                           dword *dwp1 = dwp;
                           for(int x=0; x<4; x++, i++, dwp1 = (dword*)(((byte*)dwp1) + srcbpp)){
                              union{
                                 dword input;
                                 struct{
                                    word gb;
                                    byte r;
                                    byte a;
                                 };
                              };
                              gb = *(word*)dwp1;
                              r = *(((byte*)dwp1)+2);
                              a = 0;
                              if(flags&MCONV_COLORKEY){
                                 byte is_ck = (input==color_key);
                                 ckey_4x4[i] = is_ck;
                                 block_color_key |= is_ck;
                              }
                              block_4x4[i].argb = input;
                           }
                           dwp = (dword*)(((byte*)dwp) + src_pitch);
                        }
                     }
                     break;

                  case 4:
                     {
                        for(int y=0; y<4; y++){
                           dword *dwp1 = dwp;
                           for(int x=0; x<4; x++, i++, dwp1 = (dword*)(((byte*)dwp1) + srcbpp)){
                              dword input = (*dwp1);
                              if(flags&MCONV_COLORKEY){
                                 byte is_ck = (input==color_key);
                                 ckey_4x4[i] = is_ck;
                                 block_color_key |= is_ck;
                              }
                              block_4x4[i].argb = input;
                           }
                           dwp = (dword*)(((byte*)dwp) + src_pitch);
                        }
                     }
                     break;
                  }
               }
               break;
            }
            if(has_alpha){
                              //fill-in alpha component
               if(srcbpp==4){
                  *(dword*)((byte*)sb+0) = 0;
                  *(dword*)((byte*)sb+4) = 0;
                  for(i=0; i<16; i++){
                     word *dst = ((word*)sb) + i/4;
                     dword xpos = (i&3);
                     (*dst) |= Rol(block_4x4[i].a&0xf0, xpos*4-4);
                  }
               }else{
                  *(dword*)((byte*)sb+0) = 0xffffffff;
                  *(dword*)((byte*)sb+4) = 0xffffffff;
               }
               ++sb;
            }
            S_argb color_min, color_max;
            color_min.argb = 0;
            color_max.argb = 0;
                              //find 2 border colors
            {
               dword sum_min = 0xffffffff, sum_max = 0;
               for(i=0; i<16; i++){
                  if(block_color_key && ckey_4x4[i])
                     continue;

                  const S_argb &bl = block_4x4[i];
                  dword sum = bl.r + bl.g + bl.b;
                  if(sum_min > sum){
                     sum_min = sum;
                     color_min = bl;
                  }
                  if(sum_max < sum){
                     sum_max = sum;
                     color_max = bl;
                  }
               }
            }

                              //encode 2 border colors into compressed block (format RGB565)
            {
               dword c;
               color_min.argb &= 0x00f8fcf8;
               color_max.argb &= 0x00f8fcf8;
               c = color_max.argb;
               sb->color0 = ((c&0xf80000) >> 8) | ((c&0x00fc00) >> 5) | ((c&0x0000f8) >> 3);
               c = color_min.argb;
               sb->color1 = ((c&0xf80000) >> 8) | ((c&0x00fc00) >> 5) | ((c&0x0000f8) >> 3);
               if(!block_color_key){
                              //color0 must be greater than color1
                  if(sb->color0 < sb->color1){
                     swap(sb->color0, sb->color1);
                     swap(color_min.argb, color_max.argb);
                  }
               }else{
                              //color0 must be less than color1
                  if(sb->color0 > sb->color1){
                     swap(sb->color0, sb->color1);
                     swap(color_min.argb, color_max.argb);
                  }
               }
            }
            int delta_minmax =
               (color_max.r - color_min.r) + (color_max.g - color_min.g) + (color_max.b - color_min.b);
                                    //encode all pixels to border colors
            const dword *pixel_map = pixel_maps[block_color_key];

            int mult_ratio = 3 - block_color_key;

            {
               dword pixel_mask;
               int max_clamp = 3 - block_color_key;
               pixel_mask = 0;
               for(i=0; i<16; i++){
                  if(block_color_key && ckey_4x4[i])
                     pixel_mask |= 3 << (i*2);
                  else{
                     if(delta_minmax){
                        const S_argb &bl = block_4x4[i];
                        int delta_min = (bl.r - color_min.r) + (bl.g - color_min.g) + (bl.b - color_min.b);
                                       //get ratio in range 0...3
                        int ratio = (delta_min*mult_ratio + delta_minmax/2) / delta_minmax;
                        ratio = Max(0, Min(max_clamp, ratio));
                                       //store pixel
                        pixel_mask |= pixel_map[ratio] << (i*2);
                     }
                  }
               }
               sb->pixel_mask = pixel_mask;
            }
                                 //move to next 4x4 block
            ++sb;
         }
                              //next block line
         src = ((byte*)src) + src_pitch*4;
      }
   }
   return true;
}

//----------------------------

bool C_rgb_conversion_i::AlphaBlend(void *dst, void *src, const dword *pal_argb,
   dword sx, dword sy, dword src_pitch, dword dst_pitch, byte srcbpp,
   dword flags) const{

   if(srcbpp==2 || (pf.flags&PIXELFORMAT_PALETTE))
      return false;

   union{
      byte *dst_bp;
      word *dst_wp;
      dword *dst_dwp;
      void *dst_vp;
   };
   union{
      byte *src_bp;
      void *src_vp;
   };
   bool inverse = (flags&MCONV_INVALPHA);
   dword mask_8;
   dword not_alpha;

   if(!pal_argb)
      pal_argb = (const dword*)uni_pal[0];

   bool compressed = (pf.flags&PIXELFORMAT_COMPRESS);
   if(!compressed){
      mask_8 = Ror(pf.a_mask, FindLastBit(pf.a_mask)-7);
      not_alpha = ~pf.a_mask;
      if(pf.bytes_per_pixel==3)
         not_alpha |= 0xff000000;
   }else{
      mask_8 = 0xf0;
   }

   dword *error_y = new dword[sx];
   memset(error_y, 0, sx*sizeof(dword));

   for(int y=0; y<sy; y++){
      if(!compressed)
         dst_vp = dst;
      src_vp = src;

      dword error_x = 0;

      for(int x=0; x<sx; x++){

         dword new_alpha;
         switch(srcbpp){
         case 1:
            {
               byte *bp = ((byte*)pal_argb) + (*src_bp) * sizeof(dword);
               new_alpha = (bp[0]*77 + bp[1]*153 + bp[2]*26)/256;
            }
            break;
         default:
            {
               byte *bp = src_bp;
               new_alpha = (bp[2]*77 + bp[1]*153 + bp[0]*26)/256;
            }
         }

         if(inverse)
            new_alpha = new_alpha^255;

         if(flags&MCONV_DITHER){
                              //dither
                              //get source alpha
            dword as = min(255ul, new_alpha + (error_y[x] + error_x)/2);
                              //get destination alpha
            new_alpha = as&mask_8;
                              //distribute error right & down
            error_y[x] = error_x = (as-new_alpha);
         }

         if(!compressed){
            new_alpha = Rol(new_alpha, a_pos) & pf.a_mask;
            switch(pf.bytes_per_pixel){
            case 1:
               *dst_bp &= not_alpha;
               *dst_bp |= new_alpha;
               break;
            case 2:
               *dst_wp &= not_alpha;
               *dst_wp |= new_alpha;
               break;
            case 3:
               *dst_wp &= not_alpha;
               dst_bp[2] &= (not_alpha>>16);
               *dst_wp |= new_alpha;
               dst_bp[2] |= (new_alpha>>16);
               break;
            case 4:
               *dst_dwp &= not_alpha;
               *dst_dwp |= new_alpha;
               break;
            }
         }else{
            word *dst_bl = (word*)(((byte*)dst) + (y/4)*(sx/4)*16 + (x/4)*16 + (y&3)*2);
            dword xpos = (x&3);
            (*dst_bl) &= alpha_and_mask[xpos];
            (*dst_bl) |= Rol(new_alpha&0xf0, xpos*4-4);
         }
         if(!compressed)
            dst_bp += pf.bytes_per_pixel;
         src_bp += srcbpp;
      }
      if(!compressed)
         dst = (void*)(((byte*)dst)+dst_pitch);
      src = (void*)(((byte*)src)+src_pitch);
   }
   delete[] error_y;

   return true;
}

//----------------------------

static byte SingleIndex(const byte *pal, byte r, byte g, byte b){

   const int firstpal = 1, lastpal = 255;

   const byte *bp = pal + firstpal * sizeof(dword);
   byte col;
   int bestval = 256*256, val;
   for(int cnt = firstpal; ++cnt < lastpal; ){
      int delta_r = 256 + ((int)bp[0]-(int)r);
      delta_r *= delta_r;
      int delta_g = 256+((int)bp[1]-(int)g);
      delta_g *= delta_g;
      int delta_b = 256+((int)bp[2]-(int)b);
      delta_b *= delta_b;

      val = delta_r + delta_g + delta_b;
         
      if(bestval>val){
         bestval = val;
         col = cnt;
      }
      bp += 4;
   }
   return col;
}

//----------------------------

dword C_rgb_conversion_i::GetPixel(byte r, byte g, byte b, byte a, byte *pal_argb) const{

   switch(pf.bytes_per_pixel){
   case 1:                    //remap to palette
      if(pf.flags&PIXELFORMAT_PALETTE){
         if(!pal_argb) return 0;
         return SingleIndex(pal_argb, r, g, b);
      }else{
         return
            (Rol(r, r_pos) & pf.r_mask) |
            (Rol(g, g_pos) & pf.g_mask) |
            (Rol(b, b_pos) & pf.b_mask) ;
      }
   case 2:                    //16-bit
      return
         (Rol(r, r_pos) & pf.r_mask) |
         (Rol(g, g_pos) & pf.g_mask) |
         (Rol(b, b_pos) & pf.b_mask) |
         (Rol(a, a_pos) & pf.a_mask);
   case 3:                    //24-bit
      return (r<<r_pos) | (g<<g_pos) | (b<<b_pos);
   case 4:                    //32-bit
      return (r<<r_pos) | (g<<g_pos) | (b<<b_pos) | (a<<a_pos);
   default:
      return 0;
   }
}

//----------------------------

void C_rgb_conversion_i::InverseRGB(dword val, byte *r, byte *g, byte *b, byte *a,
   const S_pixelformat *pf1, byte *pal_argb) const{

   if(!pf1)
      pf1 = &pf;

   switch(pf1->bytes_per_pixel){
   case 1:                             //paletized
      if(pal_argb && val<256){
                                       //val is index into palette
         pal_argb += (val&0xff) * sizeof(dword);
                                       //return palette entry
         *r = pal_argb[0];
         *g = pal_argb[1];
         *b = pal_argb[2];
         *a = pal_argb[3];
      }else{
         *r = 0, *g = 0, *b = 0;
         *a = 0xff;
      }
      break;
   case 2:                    //16-bit RGB
      {
         val &= 0xffff;
                              //get rotation values
         int r_pos = (FindLastBit(pf1->r_mask)-7) & 31;
         int g_pos = (FindLastBit(pf1->g_mask)-7) & 31;
         int b_pos = (FindLastBit(pf1->b_mask)-7) & 31;
         int a_pos = (FindLastBit(pf1->a_mask)-7) & 31;

         *r = Ror(val&pf1->r_mask, r_pos);
         *g = Ror(val&pf1->g_mask, g_pos);
         *b = Ror(val&pf1->b_mask, b_pos);
         *a = Ror(val&pf1->a_mask, a_pos);
         if(!pf1->a_mask) *a = 255;
      }
      break;
   case 3: 
      *r = val;
      *g = val>>8;
      *b = val>>16;
      *a = 0xff;
      break;
   case 4:   
      *r = val;
      *g = val>>8;
      *b = val>>16;
      *a = val>>24;
      break;
   }
}

//----------------------------
                              //convert paletized or RGB (24, 32-bit) data
                              //into 8 or 16-bit RGB formats

                              //note: dither doesn't support alpha channel
void C_rgb_conversion_i::Dither(const void *src, void *dst, dword pitch_src,
   dword pitch_dst, dword size_x, dword size_y, dword srcbpp, const dword *src_pal_argb) const{

                              //shift RGB masks to 8-bit positions
   dword mask_8[3];
   if(IsPaletized()){
      mask_8[0] = 0xe0;
      mask_8[1] = 0xe0;
      mask_8[2] = 0xc0;
   }else{
      mask_8[2] = Ror(pf.r_mask, FindLastBit(pf.r_mask)-7);
      mask_8[1] = Ror(pf.g_mask, FindLastBit(pf.g_mask)-7);
      mask_8[0] = Ror(pf.b_mask, FindLastBit(pf.b_mask)-7);
   }
                              //alloc buffer for y distribution error
   int *error = new int[size_x*3];
                              //clear buffers (line 0 starts with
                              //               no error)
   memset(error, 0, 3*size_x*sizeof(int));
                              //line loop
   int x, y;
   for(y=0; y<size_y; y++){
                              //x distribution error,
                              //start with 0
      int error_x[3];
      error_x[0] = 0;
      error_x[1] = 0;
      error_x[2] = 0;

      int *error_p = error;
      union{
         byte *dst_bp;
         word *dst_wp;
         void *dst_vp;
      };
      const byte *b_src = (const byte*)src;
      dst_vp = dst;
                              //row loop
      switch(pf.bytes_per_pixel){
      case 1:
         for(x=0; x<size_x; x++, b_src+=srcbpp, error_p+=3){
                              //get source RGB triplet
                              //if paletized, use source index to pal
                              //else use source directly
            const byte *bp = (srcbpp==1) ? (((byte*)src_pal_argb)+(*b_src) * sizeof(dword)) : b_src;
                              //get source pixel and
                              //add distributed error
            int rs = min(255, bp[2]+(error_p[2]+error_x[2])/2);
            int gs = min(255, bp[1]+(error_p[1]+error_x[1])/2);
            int bs = min(255, bp[0]+(error_p[0]+error_x[0])/2);
                              //get destination color
            int rd = rs & mask_8[2];
            int gd = gs & mask_8[1];
            int bd = bs & mask_8[0];
                              //distribute error right & down
            error_p[2] = error_x[2] = rs - rd;
            error_p[1] = error_x[1] = gs - gd;
            error_p[0] = error_x[0] = bs - bd;
                              //store destination pixel
            if(!IsPaletized())
               swap(rd, bd);
            *dst_bp++ =
               r.bfield[rd] |
               g.bfield[gd] |
               b.bfield[bd];
         }
         break;
      case 2:
         for(x=0; x<size_x; x++, b_src+=srcbpp, error_p+=3){
                              //get source RGB triplet
                              //if paletized, use source index to pal
                              //else use source directly
            const byte *bp = (srcbpp==1) ? (byte*)(src_pal_argb + *b_src) : b_src;
                              //get source pixel and
                              //add distributed error
            int rs = min(255, bp[2]+(error_p[2]+error_x[2])/2);
            int rd = rs & mask_8[2];
            error_p[2] = error_x[2] = rs - rd;

            int gs = min(255, bp[1]+(error_p[1]+error_x[1])/2);
            int gd = gs & mask_8[1];
            error_p[1] = error_x[1] = gs-gd;

            int bs = min(255, bp[0]+(error_p[0]+error_x[0])/2);
            int bd = bs & mask_8[0];
            error_p[0] = error_x[0] = bs-bd;

                              //store destination pixel
            *dst_wp++ =
               r.wfield[rd] |
               g.wfield[gd] |
               b.wfield[bd];
         }
         break;
      }
                              //add line's pitch
      src = (void*)(((byte*)src) + pitch_src);
      dst = (void*)(((byte*)dst) + pitch_dst);
   }
   delete[] error;
}

//----------------------------

C_rgb_conversion *Create_rgb_conv(){
   return new C_rgb_conversion_i;
}

//----------------------------
 