/*--------------------------------------------------------
   Copyright (c) 1999, 2000 Michal Bacik. 
   All rights reserved.

   File: D3D_err.cpp
   Content: D3D error-reporting functions.
--------------------------------------------------------*/

#include <windows.h>
#include <d3d8.h>
#include <c_string.hpp>

//----------------------------
//----------------------------

void D3D_Fatal(const char *text, dword hr){

   char *msgp;
   switch(hr){

   case D3D_OK: msgp = "No error occurred."; break;
   case D3DERR_CONFLICTINGRENDERSTATE: msgp = "The currently set render states cannot be used together."; break;
   case D3DERR_CONFLICTINGTEXTUREFILTER: msgp = "The current texture filters cannot be used together."; break;
   case D3DERR_CONFLICTINGTEXTUREPALETTE: msgp = "The current textures cannot be used simultaneously. This generally occurs when a multitexture device requires that all palletized textures simultaneously enabled also share the same palette."; break;
   case D3DERR_DEVICELOST: msgp = "The device is lost and cannot be restored at the current time, so rendering is not possible."; break;
   case D3DERR_DEVICENOTRESET: msgp = "The device cannot be reset."; break;
   case D3DERR_DRIVERINTERNALERROR: msgp = "Internal driver error."; break;
   case D3DERR_INVALIDCALL: msgp = "The method call is invalid. For example, a method's parameter may have an invalid value."; break;
   case D3DERR_INVALIDDEVICE: msgp = "The requested device type is not valid."; break;
   case D3DERR_MOREDATA: msgp = "There is more data available than the specified buffer size can hold."; break;
   case D3DERR_NOTAVAILABLE: msgp = "This device does not support the queried technique."; break;
   case D3DERR_NOTFOUND: msgp = "The requested item was not found."; break;
   case D3DERR_OUTOFVIDEOMEMORY: msgp = "Direct3D does not have enough display memory to perform the operation."; break;
   case D3DERR_TOOMANYOPERATIONS: msgp = "The application is requesting more texture-filtering operations than the device supports."; break;
   case D3DERR_UNSUPPORTEDALPHAARG: msgp = "The device does not support a specified texture-blending argument for the alpha channel."; break;
   case D3DERR_UNSUPPORTEDALPHAOPERATION: msgp = "The device does not support a specified texture-blending operation for the alpha channel."; break;
   case D3DERR_UNSUPPORTEDCOLORARG: msgp = "The device does not support a specified texture-blending argument for color values."; break;
   case D3DERR_UNSUPPORTEDCOLOROPERATION: msgp = "The device does not support a specified texture-blending operation for color values."; break;
   case D3DERR_UNSUPPORTEDFACTORVALUE: msgp = "The device does not support the specified texture factor value."; break;
   case D3DERR_UNSUPPORTEDTEXTUREFILTER: msgp = "The device does not support the specified texture filter."; break;
   case D3DERR_WRONGTEXTUREFORMAT: msgp = "The pixel format of the texture surface is not valid."; break;
   case E_FAIL: msgp = "An undetermined error occurred inside the Direct3D subsystem."; break;
   case E_INVALIDARG: msgp = "An invalid parameter was passed to the returning function."; break;
   //case E_INVALIDCALL: msgp = "The method call is invalid. For example, a method's parameter may have an invalid value."; break;
   case E_OUTOFMEMORY: msgp = "Direct3D could not allocate sufficient memory to complete the call."; break;
   //default: DD_Fatal(text, hr); return;
   }

   char buf[16];
   itoa(hr, buf, 16);

   MessageBox(NULL,
      (C_string)text + (char const*)": " + (char const*)msgp +
      (char const*)" 0x" + (char const*)buf,
      "Fatal Error", MB_OK);
}

//----------------------------
