#include "all.h"
#include <mmsystem.h>
#include <winable.h>
#include "igraph_i.h"
#include <win_reg.h>
#include <zmouse.h>
#include <vector>
#include <algorithm>
#include <iostream>
#include <debug.h>
#include <malloc.h>

//----------------------------

                              //disabled - problems after short times in _get_timer
#define USE_QUERYPERFORMANCECOUNTER

//#ifdef NDEBUG
// #define MOUSE_USE_EVENT      //it makes problems!
//#endif

//----------------------------

using namespace std;

C_rgb_conversion *Create_rgb_conv();

//----------------------------

#ifdef LOG_MODE
C_log IG_log;
#endif                        //LOG_MODE

//----------------------------

static void DI_Fatal(const char *text, HRESULT hr){

   char *msgp;

   switch(hr){
   case DI_DOWNLOADSKIPPED: msgp = "DI_DOWNLOADSKIPPED"; break;
   case DI_EFFECTRESTARTED: msgp = "DI_EFFECTRESTARTED"; break;
   case DI_OK: msgp = "DI_OK"; break;
   case DI_POLLEDDEVICE: msgp = "DI_POLLEDDEVICE"; break;
   case DI_TRUNCATED: msgp = "DI_TRUNCATED"; break;
   case DI_TRUNCATEDANDRESTARTED: msgp = "DI_TRUNCATEDANDRESTARTED"; break;
   case DIERR_ACQUIRED: msgp = "DIERR_ACQUIRED"; break;
   case DIERR_ALREADYINITIALIZED: msgp = "DIERR_ALREADYINITIALIZED"; break;
   case DIERR_BADDRIVERVER: msgp = "DIERR_BADDRIVERVER"; break;
   case DIERR_BETADIRECTINPUTVERSION: msgp = "DIERR_BETADIRECTINPUTVERSION"; break;
   case DIERR_DEVICEFULL: msgp = "DIERR_DEVICEFULL"; break;
   case DIERR_DEVICENOTREG: msgp = "DIERR_DEVICENOTREG"; break;
   case DIERR_EFFECTPLAYING: msgp = "DIERR_EFFECTPLAYING"; break;
   case DIERR_HASEFFECTS: msgp = "DIERR_HASEFFECTS"; break;
   case DIERR_GENERIC: msgp = "DIERR_GENERIC"; break;
   case DIERR_HANDLEEXISTS: msgp = "DIERR_HANDLEEXISTS"; break;
   case DIERR_INCOMPLETEEFFECT: msgp = "DIERR_INCOMPLETEEFFECT"; break;
   case DIERR_INPUTLOST: msgp = "DIERR_INPUTLOST"; break;
   case DIERR_INVALIDPARAM: msgp = "DIERR_INVALIDPARAM"; break;
   case DIERR_MOREDATA: msgp = "DIERR_MOREDATA"; break;
   case DIERR_NOAGGREGATION: msgp = "DIERR_NOAGGREGATION"; break;
   case DIERR_NOINTERFACE: msgp = "DIERR_NOINTERFACE"; break;
   case DIERR_NOTACQUIRED: msgp = "DIERR_NOTACQUIRED"; break;
   case DIERR_NOTBUFFERED: msgp = "DIERR_NOTBUFFERED"; break;
   case DIERR_NOTDOWNLOADED: msgp = "DIERR_NOTDOWNLOADED"; break;
   case DIERR_NOTEXCLUSIVEACQUIRED: msgp = "DIERR_NOTEXCLUSIVEACQUIRED"; break;
   case DIERR_NOTFOUND: msgp = "DIERR_NOTFOUND"; break;
   case DIERR_NOTINITIALIZED: msgp = "DIERR_NOTINITIALIZED"; break;
   //case DIERR_OBJECTNOTFOUND: msgp = "DIERR_OBJECTNOTFOUND"; break;
   case DIERR_OLDDIRECTINPUTVERSION: msgp = "DIERR_OLDDIRECTINPUTVERSION"; break;
   //case DIERR_OTHERAPPHASPRIO: msgp = "DIERR_OTHERAPPHASPRIO"; break;
   case DIERR_OUTOFMEMORY: msgp = "DIERR_OUTOFMEMORY"; break;
   //case DIERR_READONLY: msgp = "DIERR_READONLY"; break;
   case DIERR_REPORTFULL: msgp = "DIERR_UNPLUGGED"; break;
   case DIERR_UNPLUGGED: msgp = "DIERR_UNPLUGGED"; break;
   case DIERR_UNSUPPORTED: msgp = "DIERR_UNSUPPORTED"; break;
   case E_HANDLE: msgp = "E_HANDLE"; break;
   case E_PENDING: msgp = "E_PENDING"; break;
   default: msgp="Unknown Error";
   }
   MessageBox(NULL, C_fstring("%s: %s", text, msgp), "DirectInput", MB_OK);
}

#ifndef NDEBUG
#define CHECK_DI_RESULT(text, hr) if(FAILED(hr)) DI_Fatal(text, hr);
#else
#define CHECK_DI_RESULT(text, hr)
#endif

#ifndef NDEBUG
#define CHECK_DD_RESULT(text, hr) if(FAILED(hr)) DD_Fatal(text, hr);
#else
#define CHECK_DD_RESULT(text, hr)
#endif

//----------------------------

static void DumpCaps(IDirect3D8 *lpD3D8, IDirect3DDevice8 *lpDev8,
   IG_INIT::t_log_func *log_func){

   HRESULT hr;

   D3DCAPS8 cp;
   hr = lpDev8->GetDeviceCaps(&cp);
   CHECK_D3D_RESULT("GetDeviceCaps", hr);
   if(FAILED(hr)) return;

   D3DADAPTER_IDENTIFIER8 ai;
   hr = lpD3D8->GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0, &ai);
   CHECK_D3D_RESULT("GetAdapterIdentifier", hr);
   if(FAILED(hr)) return;

   (*log_func)("----------------------");
   (*log_func)("DirectX 8 Device Caps:");

   {
      (*log_func)(C_fstring("Driver      : %s", ai.Driver));
      (*log_func)(C_fstring("Description : %s", ai.Description));
      (*log_func)(C_fstring("Version     : %u.%u.%u.%u", HIWORD(ai.DriverVersion.HighPart),
         LOWORD(ai.DriverVersion.HighPart), HIWORD(ai.DriverVersion.LowPart),
         LOWORD(ai.DriverVersion.LowPart)));
      (*log_func)(C_fstring("Vendor ID   : %u", ai.VendorId));
      (*log_func)(C_fstring("Device ID   : %u", ai.DeviceId));
      (*log_func)(C_fstring("Subsystem ID: %u", ai.SubSysId));
      (*log_func)(C_fstring("Revision    : %u", ai.Revision));
      (*log_func)(C_fstring("dwWHQLLevel : %u", ai.WHQLLevel));
   }
   {
      (*log_func)("DeviceType: ");
      switch(cp.DeviceType){
      case D3DDEVTYPE_HAL: (*log_func)("HAL"); break;
      case D3DDEVTYPE_REF: (*log_func)("REF"); break;
      case D3DDEVTYPE_SW: (*log_func)("SW"); break;
      }
   }
   {
      (*log_func)("Caps:");
      dword dw = cp.Caps;
      if(dw&D3DCAPS_READ_SCANLINE) (*log_func)(" D3DCAPS_READ_SCANLINE");
   }
   {
      (*log_func)("Caps2:");
      dword dw = cp.Caps2;
      if(dw&D3DCAPS2_CANCALIBRATEGAMMA) (*log_func)(" D3DCAPS2_CANCALIBRATEGAMMA");
      if(dw&D3DCAPS2_CANRENDERWINDOWED) (*log_func)(" D3DCAPS2_CANRENDERWINDOWED");
      if(dw&D3DCAPS2_FULLSCREENGAMMA) (*log_func)(" D3DCAPS2_FULLSCREENGAMMA");
      if(dw&D3DCAPS2_NO2DDURING3DSCENE) (*log_func)(" D3DCAPS2_NO2DDURING3DSCENE");
   }
   {
      (*log_func)("PresentationIntervals:");
      dword dw = cp.PresentationIntervals;
      if(dw&D3DPRESENT_INTERVAL_IMMEDIATE) (*log_func)(" D3DPRESENT_INTERVAL_IMMEDIATE");
      if(dw&D3DPRESENT_INTERVAL_ONE) (*log_func)(" D3DPRESENT_INTERVAL_ONE");
      if(dw&D3DPRESENT_INTERVAL_TWO) (*log_func)(" D3DPRESENT_INTERVAL_TWO");
      if(dw&D3DPRESENT_INTERVAL_THREE) (*log_func)(" D3DPRESENT_INTERVAL_THREE");
      if(dw&D3DPRESENT_INTERVAL_FOUR) (*log_func)(" D3DPRESENT_INTERVAL_FOUR");
   }
   {
      (*log_func)("CursorCaps:");
      dword dw = cp.CursorCaps;
      if(dw&D3DCURSORCAPS_COLOR) (*log_func)(" D3DCURSORCAPS_COLOR");
      if(dw&D3DCURSORCAPS_LOWRES) (*log_func)(" D3DCURSORCAPS_LOWRES");
   }
   {
      (*log_func)("DevCaps:");
      dword dw = cp.DevCaps;
      if(dw&D3DDEVCAPS_CANBLTSYSTONONLOCAL) (*log_func)(" D3DDEVCAPS_CANBLTSYSTONONLOCAL");
      if(dw&D3DDEVCAPS_CANRENDERAFTERFLIP) (*log_func)(" D3DDEVCAPS_CANRENDERAFTERFLIP");
      if(dw&D3DDEVCAPS_DRAWPRIMTLVERTEX) (*log_func)(" D3DDEVCAPS_DRAWPRIMTLVERTEX");
      if(dw&D3DDEVCAPS_EXECUTESYSTEMMEMORY) (*log_func)(" D3DDEVCAPS_EXECUTESYSTEMMEMORY");
      if(dw&D3DDEVCAPS_EXECUTEVIDEOMEMORY) (*log_func)(" D3DDEVCAPS_EXECUTEVIDEOMEMORY");
      if(dw&D3DDEVCAPS_HWRASTERIZATION) (*log_func)(" D3DDEVCAPS_HWRASTERIZATION");
      if(dw&D3DDEVCAPS_HWTRANSFORMANDLIGHT) (*log_func)(" D3DDEVCAPS_HWTRANSFORMANDLIGHT");
      if(dw&D3DDEVCAPS_NPATCHES) (*log_func)(" D3DDEVCAPS_NPATCHES");
      if(dw&D3DDEVCAPS_PUREDEVICE) (*log_func)(" D3DDEVCAPS_PUREDEVICE");
      if(dw&D3DDEVCAPS_QUINTICRTPATCHES) (*log_func)(" D3DDEVCAPS_QUINTICRTPATCHES");
      if(dw&D3DDEVCAPS_RTPATCHES) (*log_func)(" D3DDEVCAPS_RTPATCHES");
      if(dw&D3DDEVCAPS_RTPATCHHANDLEZERO) (*log_func)(" D3DDEVCAPS_RTPATCHHANDLEZERO");
      if(dw&D3DDEVCAPS_SEPARATETEXTUREMEMORIES) (*log_func)(" D3DDEVCAPS_SEPARATETEXTUREMEMORIES");
      if(dw&D3DDEVCAPS_TEXTURENONLOCALVIDMEM) (*log_func)(" D3DDEVCAPS_TEXTURENONLOCALVIDMEM");
      if(dw&D3DDEVCAPS_TEXTURESYSTEMMEMORY) (*log_func)(" D3DDEVCAPS_TEXTURESYSTEMMEMORY");
      if(dw&D3DDEVCAPS_TEXTUREVIDEOMEMORY) (*log_func)(" D3DDEVCAPS_TEXTUREVIDEOMEMORY");
      if(dw&D3DDEVCAPS_TLVERTEXSYSTEMMEMORY) (*log_func)(" D3DDEVCAPS_TLVERTEXSYSTEMMEMORY");
      if(dw&D3DDEVCAPS_TLVERTEXVIDEOMEMORY) (*log_func)(" D3DDEVCAPS_TLVERTEXVIDEOMEMORY");
   }
   {
      (*log_func)("PrimitiveMiscCaps:");
      dword dw = cp.PrimitiveMiscCaps;
      if(dw&D3DPMISCCAPS_BLENDOP) (*log_func)(" D3DPMISCCAPS_BLENDOP");
      if(dw&D3DPMISCCAPS_CLIPPLANESCALEDPOINTS) (*log_func)(" D3DPMISCCAPS_CLIPPLANESCALEDPOINTS");
      if(dw&D3DPMISCCAPS_CLIPTLVERTS) (*log_func)(" D3DPMISCCAPS_CLIPTLVERTS");
      if(dw&D3DPMISCCAPS_COLORWRITEENABLE) (*log_func)(" D3DPMISCCAPS_COLORWRITEENABLE");
      if(dw&D3DPMISCCAPS_CULLCCW) (*log_func)(" D3DPMISCCAPS_CULLCCW");
      if(dw&D3DPMISCCAPS_CULLCW) (*log_func)(" D3DPMISCCAPS_CULLCW");
      if(dw&D3DPMISCCAPS_CULLNONE) (*log_func)(" D3DPMISCCAPS_CULLNONE");
      if(dw&D3DPMISCCAPS_LINEPATTERNREP) (*log_func)(" D3DPMISCCAPS_LINEPATTERNREP");
      if(dw&D3DPMISCCAPS_MASKZ) (*log_func)(" D3DPMISCCAPS_MASKZ");
      if(dw&D3DPMISCCAPS_TSSARGTEMP) (*log_func)(" D3DPMISCCAPS_TSSARGTEMP");
   }
   {
      (*log_func)("RasterCaps:");
      dword dw = cp.RasterCaps;
      if(dw&D3DPRASTERCAPS_ANISOTROPY) (*log_func)(" D3DPRASTERCAPS_ANISOTROPY");
      if(dw&D3DPRASTERCAPS_ANTIALIASEDGES) (*log_func)(" D3DPRASTERCAPS_ANTIALIASEDGES");
      if(dw&D3DPRASTERCAPS_COLORPERSPECTIVE) (*log_func)(" D3DPRASTERCAPS_COLORPERSPECTIVE");
      if(dw&D3DPRASTERCAPS_DITHER) (*log_func)(" D3DPRASTERCAPS_DITHER");
      if(dw&D3DPRASTERCAPS_FOGRANGE) (*log_func)(" D3DPRASTERCAPS_FOGRANGE");
      if(dw&D3DPRASTERCAPS_FOGTABLE) (*log_func)(" D3DPRASTERCAPS_FOGTABLE");
      if(dw&D3DPRASTERCAPS_FOGVERTEX) (*log_func)(" D3DPRASTERCAPS_FOGVERTEX");
      if(dw&D3DPRASTERCAPS_MIPMAPLODBIAS) (*log_func)(" D3DPRASTERCAPS_MIPMAPLODBIAS");
      if(dw&D3DPRASTERCAPS_PAT) (*log_func)(" D3DPRASTERCAPS_PAT");
      if(dw&D3DPRASTERCAPS_STRETCHBLTMULTISAMPLE) (*log_func)(" D3DPRASTERCAPS_STRETCHBLTMULTISAMPLE");
      if(dw&D3DPRASTERCAPS_WBUFFER) (*log_func)(" D3DPRASTERCAPS_WBUFFER");
      if(dw&D3DPRASTERCAPS_WFOG) (*log_func)(" D3DPRASTERCAPS_WFOG");
      if(dw&D3DPRASTERCAPS_ZBIAS) (*log_func)(" D3DPRASTERCAPS_ZBIAS");
      if(dw&D3DPRASTERCAPS_ZBUFFERLESSHSR) (*log_func)(" D3DPRASTERCAPS_ZBUFFERLESSHSR");
      if(dw&D3DPRASTERCAPS_ZBUFFERLESSHSR) (*log_func)(" D3DPRASTERCAPS_ZBUFFERLESSHSR");
      if(dw&D3DPRASTERCAPS_ZFOG) (*log_func)(" D3DPRASTERCAPS_ZFOG");
      if(dw&D3DPRASTERCAPS_ZTEST) (*log_func)(" D3DPRASTERCAPS_ZTEST");
   }
   {
      (*log_func)("ZCmpCaps:");
      dword dw = cp.ZCmpCaps;
      if(dw&D3DPCMPCAPS_ALWAYS) (*log_func)(" D3DPCMPCAPS_ALWAYS");
      if(dw&D3DPCMPCAPS_EQUAL) (*log_func)(" D3DPCMPCAPS_EQUAL");
      if(dw&D3DPCMPCAPS_GREATER) (*log_func)(" D3DPCMPCAPS_GREATER");
      if(dw&D3DPCMPCAPS_GREATEREQUAL) (*log_func)(" D3DPCMPCAPS_GREATEREQUAL");
      if(dw&D3DPCMPCAPS_LESS) (*log_func)(" D3DPCMPCAPS_LESS");
      if(dw&D3DPCMPCAPS_LESSEQUAL) (*log_func)(" D3DPCMPCAPS_LESSEQUAL");
      if(dw&D3DPCMPCAPS_NEVER) (*log_func)(" D3DPCMPCAPS_NEVER");
      if(dw&D3DPCMPCAPS_NOTEQUAL) (*log_func)(" D3DPCMPCAPS_NOTEQUAL");
   }
   {
      (*log_func)("SrcBlendCaps:");
      dword dw = cp.SrcBlendCaps;
      if(dw&D3DPBLENDCAPS_BOTHINVSRCALPHA) (*log_func)(" D3DPBLENDCAPS_BOTHINVSRCALPHA");
      if(dw&D3DPBLENDCAPS_BOTHSRCALPHA) (*log_func)(" D3DPBLENDCAPS_BOTHSRCALPHA");
      if(dw&D3DPBLENDCAPS_DESTALPHA) (*log_func)(" D3DPBLENDCAPS_DESTALPHA");
      if(dw&D3DPBLENDCAPS_DESTCOLOR) (*log_func)(" D3DPBLENDCAPS_DESTCOLOR");
      if(dw&D3DPBLENDCAPS_INVDESTALPHA) (*log_func)(" D3DPBLENDCAPS_INVDESTALPHA");
      if(dw&D3DPBLENDCAPS_INVDESTCOLOR) (*log_func)(" D3DPBLENDCAPS_INVDESTCOLOR");
      if(dw&D3DPBLENDCAPS_INVSRCALPHA) (*log_func)(" D3DPBLENDCAPS_INVSRCALPHA");
      if(dw&D3DPBLENDCAPS_INVSRCCOLOR) (*log_func)(" D3DPBLENDCAPS_INVSRCCOLOR");
      if(dw&D3DPBLENDCAPS_ONE) (*log_func)(" D3DPBLENDCAPS_ONE");
      if(dw&D3DPBLENDCAPS_SRCALPHA) (*log_func)(" D3DPBLENDCAPS_SRCALPHA");
      if(dw&D3DPBLENDCAPS_SRCALPHASAT) (*log_func)(" D3DPBLENDCAPS_SRCALPHASAT");
      if(dw&D3DPBLENDCAPS_SRCCOLOR) (*log_func)(" D3DPBLENDCAPS_SRCCOLOR");
      if(dw&D3DPBLENDCAPS_ZERO) (*log_func)(" D3DPBLENDCAPS_ZERO");
   }
   {
      (*log_func)("DestBlendCaps:");
      dword dw = cp.DestBlendCaps;
      if(dw&D3DPBLENDCAPS_BOTHINVSRCALPHA) (*log_func)(" D3DPBLENDCAPS_BOTHINVSRCALPHA");
      if(dw&D3DPBLENDCAPS_BOTHSRCALPHA) (*log_func)(" D3DPBLENDCAPS_BOTHSRCALPHA");
      if(dw&D3DPBLENDCAPS_DESTALPHA) (*log_func)(" D3DPBLENDCAPS_DESTALPHA");
      if(dw&D3DPBLENDCAPS_DESTCOLOR) (*log_func)(" D3DPBLENDCAPS_DESTCOLOR");
      if(dw&D3DPBLENDCAPS_INVDESTALPHA) (*log_func)(" D3DPBLENDCAPS_INVDESTALPHA");
      if(dw&D3DPBLENDCAPS_INVDESTCOLOR) (*log_func)(" D3DPBLENDCAPS_INVDESTCOLOR");
      if(dw&D3DPBLENDCAPS_INVSRCALPHA) (*log_func)(" D3DPBLENDCAPS_INVSRCALPHA");
      if(dw&D3DPBLENDCAPS_INVSRCCOLOR) (*log_func)(" D3DPBLENDCAPS_INVSRCCOLOR");
      if(dw&D3DPBLENDCAPS_ONE) (*log_func)(" D3DPBLENDCAPS_ONE");
      if(dw&D3DPBLENDCAPS_SRCALPHA) (*log_func)(" D3DPBLENDCAPS_SRCALPHA");
      if(dw&D3DPBLENDCAPS_SRCALPHASAT) (*log_func)(" D3DPBLENDCAPS_SRCALPHASAT");
      if(dw&D3DPBLENDCAPS_SRCCOLOR) (*log_func)(" D3DPBLENDCAPS_SRCCOLOR");
      if(dw&D3DPBLENDCAPS_ZERO) (*log_func)(" D3DPBLENDCAPS_ZERO");
   }
   {
      (*log_func)("AlphaCmpCaps:");
      dword dw = cp.AlphaCmpCaps;
      if(dw&D3DPCMPCAPS_ALWAYS) (*log_func)(" D3DPCMPCAPS_ALWAYS");
      if(dw&D3DPCMPCAPS_EQUAL) (*log_func)(" D3DPCMPCAPS_EQUAL");
      if(dw&D3DPCMPCAPS_GREATER) (*log_func)(" D3DPCMPCAPS_GREATER");
      if(dw&D3DPCMPCAPS_GREATEREQUAL) (*log_func)(" D3DPCMPCAPS_GREATEREQUAL");
      if(dw&D3DPCMPCAPS_LESS) (*log_func)(" D3DPCMPCAPS_LESS");
      if(dw&D3DPCMPCAPS_LESSEQUAL) (*log_func)(" D3DPCMPCAPS_LESSEQUAL");
      if(dw&D3DPCMPCAPS_NEVER) (*log_func)(" D3DPCMPCAPS_NEVER");
      if(dw&D3DPCMPCAPS_NOTEQUAL) (*log_func)(" D3DPCMPCAPS_NOTEQUAL");
   }
   {
      (*log_func)("ShadeCaps:");
      dword dw = cp.ShadeCaps;
      if(dw&D3DPSHADECAPS_ALPHAGOURAUDBLEND) (*log_func)(" D3DPSHADECAPS_ALPHAGOURAUDBLEND");
      if(dw&D3DPSHADECAPS_COLORGOURAUDRGB) (*log_func)(" D3DPSHADECAPS_COLORGOURAUDRGB");
      if(dw&D3DPSHADECAPS_FOGGOURAUD) (*log_func)(" D3DPSHADECAPS_FOGGOURAUD");
      if(dw&D3DPSHADECAPS_SPECULARGOURAUDRGB) (*log_func)(" D3DPSHADECAPS_SPECULARGOURAUDRGB");
   }
   {
      (*log_func)("TextureCaps:");
      dword dw = cp.TextureCaps;
      if(dw&D3DPTEXTURECAPS_ALPHA) (*log_func)(" D3DPTEXTURECAPS_ALPHA");
      if(dw&D3DPTEXTURECAPS_ALPHAPALETTE) (*log_func)(" D3DPTEXTURECAPS_ALPHAPALETTE");
      if(dw&D3DPTEXTURECAPS_CUBEMAP) (*log_func)(" D3DPTEXTURECAPS_CUBEMAP");
      if(dw&D3DPTEXTURECAPS_CUBEMAP_POW2) (*log_func)(" D3DPTEXTURECAPS_CUBEMAP_POW2");
      if(dw&D3DPTEXTURECAPS_MIPCUBEMAP) (*log_func)(" D3DPTEXTURECAPS_MIPCUBEMAP");
      if(dw&D3DPTEXTURECAPS_MIPMAP) (*log_func)(" D3DPTEXTURECAPS_MIPMAP");
      if(dw&D3DPTEXTURECAPS_MIPVOLUMEMAP) (*log_func)(" D3DPTEXTURECAPS_MIPVOLUMEMAP");
      if(dw&D3DPTEXTURECAPS_NONPOW2CONDITIONAL) (*log_func)(" D3DPTEXTURECAPS_NONPOW2CONDITIONAL");
      if(dw&D3DPTEXTURECAPS_PERSPECTIVE) (*log_func)(" D3DPTEXTURECAPS_PERSPECTIVE");
      if(dw&D3DPTEXTURECAPS_POW2) (*log_func)(" D3DPTEXTURECAPS_POW2");
      if(dw&D3DPTEXTURECAPS_PROJECTED) (*log_func)(" D3DPTEXTURECAPS_PROJECTED");
      if(dw&D3DPTEXTURECAPS_SQUAREONLY) (*log_func)(" D3DPTEXTURECAPS_SQUAREONLY");
      if(dw&D3DPTEXTURECAPS_TEXREPEATNOTSCALEDBYSIZE) (*log_func)(" D3DPTEXTURECAPS_TEXREPEATNOTSCALEDBYSIZE");
      if(dw&D3DPTEXTURECAPS_VOLUMEMAP) (*log_func)(" D3DPTEXTURECAPS_VOLUMEMAP");
      if(dw&D3DPTEXTURECAPS_VOLUMEMAP_POW2) (*log_func)(" D3DPTEXTURECAPS_VOLUMEMAP_POW2");
   }
   {
      (*log_func)("TextureFilterCaps:");
      dword dw = cp.TextureFilterCaps;
      if(dw&D3DPTFILTERCAPS_MAGFAFLATCUBIC) (*log_func)(" D3DPTFILTERCAPS_MAGFAFLATCUBIC");
      if(dw&D3DPTFILTERCAPS_MAGFANISOTROPIC) (*log_func)(" D3DPTFILTERCAPS_MAGFANISOTROPIC");
      if(dw&D3DPTFILTERCAPS_MAGFGAUSSIANCUBIC) (*log_func)(" D3DPTFILTERCAPS_MAGFGAUSSIANCUBIC");
      if(dw&D3DPTFILTERCAPS_MAGFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MAGFLINEAR");
      if(dw&D3DPTFILTERCAPS_MAGFPOINT) (*log_func)(" D3DPTFILTERCAPS_MAGFPOINT");
      if(dw&D3DPTFILTERCAPS_MINFANISOTROPIC) (*log_func)(" D3DPTFILTERCAPS_MINFANISOTROPIC");
      if(dw&D3DPTFILTERCAPS_MINFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MINFLINEAR");
      if(dw&D3DPTFILTERCAPS_MINFPOINT) (*log_func)(" D3DPTFILTERCAPS_MINFPOINT");
      if(dw&D3DPTFILTERCAPS_MIPFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MIPFLINEAR");
      if(dw&D3DPTFILTERCAPS_MIPFPOINT) (*log_func)(" D3DPTFILTERCAPS_MIPFPOINT");
   }
   {
      (*log_func)("CubeTextureFilterCaps:");
      dword dw = cp.CubeTextureFilterCaps;
      if(dw&D3DPTFILTERCAPS_MAGFAFLATCUBIC) (*log_func)(" D3DPTFILTERCAPS_MAGFAFLATCUBIC");
      if(dw&D3DPTFILTERCAPS_MAGFANISOTROPIC) (*log_func)(" D3DPTFILTERCAPS_MAGFANISOTROPIC");
      if(dw&D3DPTFILTERCAPS_MAGFGAUSSIANCUBIC) (*log_func)(" D3DPTFILTERCAPS_MAGFGAUSSIANCUBIC");
      if(dw&D3DPTFILTERCAPS_MAGFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MAGFLINEAR");
      if(dw&D3DPTFILTERCAPS_MAGFPOINT) (*log_func)(" D3DPTFILTERCAPS_MAGFPOINT");
      if(dw&D3DPTFILTERCAPS_MINFANISOTROPIC) (*log_func)(" D3DPTFILTERCAPS_MINFANISOTROPIC");
      if(dw&D3DPTFILTERCAPS_MINFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MINFLINEAR");
      if(dw&D3DPTFILTERCAPS_MINFPOINT) (*log_func)(" D3DPTFILTERCAPS_MINFPOINT");
      if(dw&D3DPTFILTERCAPS_MIPFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MIPFLINEAR");
      if(dw&D3DPTFILTERCAPS_MIPFPOINT) (*log_func)(" D3DPTFILTERCAPS_MIPFPOINT");
   }
   {
      (*log_func)("VolumeTextureFilterCaps:");
      dword dw = cp.VolumeTextureFilterCaps;
      if(dw&D3DPTFILTERCAPS_MAGFAFLATCUBIC) (*log_func)(" D3DPTFILTERCAPS_MAGFAFLATCUBIC");
      if(dw&D3DPTFILTERCAPS_MAGFANISOTROPIC) (*log_func)(" D3DPTFILTERCAPS_MAGFANISOTROPIC");
      if(dw&D3DPTFILTERCAPS_MAGFGAUSSIANCUBIC) (*log_func)(" D3DPTFILTERCAPS_MAGFGAUSSIANCUBIC");
      if(dw&D3DPTFILTERCAPS_MAGFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MAGFLINEAR");
      if(dw&D3DPTFILTERCAPS_MAGFPOINT) (*log_func)(" D3DPTFILTERCAPS_MAGFPOINT");
      if(dw&D3DPTFILTERCAPS_MINFANISOTROPIC) (*log_func)(" D3DPTFILTERCAPS_MINFANISOTROPIC");
      if(dw&D3DPTFILTERCAPS_MINFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MINFLINEAR");
      if(dw&D3DPTFILTERCAPS_MINFPOINT) (*log_func)(" D3DPTFILTERCAPS_MINFPOINT");
      if(dw&D3DPTFILTERCAPS_MIPFLINEAR) (*log_func)(" D3DPTFILTERCAPS_MIPFLINEAR");
      if(dw&D3DPTFILTERCAPS_MIPFPOINT) (*log_func)(" D3DPTFILTERCAPS_MIPFPOINT");
   }
   {
      (*log_func)("TextureAddressCaps:");
      dword dw = cp.TextureAddressCaps;
      if(dw&D3DPTADDRESSCAPS_BORDER) (*log_func)(" D3DPTADDRESSCAPS_BORDER");
      if(dw&D3DPTADDRESSCAPS_CLAMP) (*log_func)(" D3DPTADDRESSCAPS_CLAMP");
      if(dw&D3DPTADDRESSCAPS_INDEPENDENTUV) (*log_func)(" D3DPTADDRESSCAPS_INDEPENDENTUV");
      if(dw&D3DPTADDRESSCAPS_MIRROR) (*log_func)(" D3DPTADDRESSCAPS_MIRROR");
      if(dw&D3DPTADDRESSCAPS_MIRRORONCE) (*log_func)(" D3DPTADDRESSCAPS_MIRRORONCE");
      if(dw&D3DPTADDRESSCAPS_WRAP) (*log_func)(" D3DPTADDRESSCAPS_WRAP");
   }
   {
      (*log_func)("VolumeTextureAddressCaps:");
      dword dw = cp.VolumeTextureAddressCaps;
      if(dw&D3DPTADDRESSCAPS_BORDER) (*log_func)(" D3DPTADDRESSCAPS_BORDER");
      if(dw&D3DPTADDRESSCAPS_CLAMP) (*log_func)(" D3DPTADDRESSCAPS_CLAMP");
      if(dw&D3DPTADDRESSCAPS_INDEPENDENTUV) (*log_func)(" D3DPTADDRESSCAPS_INDEPENDENTUV");
      if(dw&D3DPTADDRESSCAPS_MIRROR) (*log_func)(" D3DPTADDRESSCAPS_MIRROR");
      if(dw&D3DPTADDRESSCAPS_MIRRORONCE) (*log_func)(" D3DPTADDRESSCAPS_MIRRORONCE");
      if(dw&D3DPTADDRESSCAPS_WRAP) (*log_func)(" D3DPTADDRESSCAPS_WRAP");
   }
   {
      (*log_func)("LineCaps:");
      dword dw = cp.LineCaps;
      if(dw&D3DLINECAPS_ALPHACMP) (*log_func)(" D3DLINECAPS_ALPHACMP");
      if(dw&D3DLINECAPS_BLEND) (*log_func)(" D3DLINECAPS_BLEND");
      if(dw&D3DLINECAPS_FOG) (*log_func)(" D3DLINECAPS_FOG");
      if(dw&D3DLINECAPS_TEXTURE) (*log_func)(" D3DLINECAPS_TEXTURE");
      if(dw&D3DLINECAPS_ZTEST) (*log_func)(" D3DLINECAPS_ZTEST");
   }
   (*log_func)(C_fstring("MaxTextureWidth: %u", cp.MaxTextureWidth));
   (*log_func)(C_fstring("MaxTextureHeight: %u", cp.MaxTextureHeight));
   (*log_func)(C_fstring("MaxVolumeExtent: %u", cp.MaxVolumeExtent));
   (*log_func)(C_fstring("MaxTextureRepeat: %u", cp.MaxTextureRepeat));
   (*log_func)(C_fstring("MaxTextureAspectRatio: %u", cp.MaxTextureAspectRatio));
   (*log_func)(C_fstring("MaxAnisotropy: %u", cp.MaxAnisotropy));
   (*log_func)(C_fstring("MaxVertexW: %f", cp.MaxVertexW));
   (*log_func)(C_fstring("GuardBandLeft: %f", cp.GuardBandLeft));
   (*log_func)(C_fstring("GuardBandTop: %f", cp.GuardBandTop));
   (*log_func)(C_fstring("GuardBandRight: %f", cp.GuardBandRight));
   (*log_func)(C_fstring("GuardBandBottom: %f", cp.GuardBandBottom));
   (*log_func)(C_fstring("ExtentsAdjust: %f", cp.ExtentsAdjust));
   {
      (*log_func)("StencilCaps:");
      dword dw = cp.StencilCaps;
      if(dw&D3DSTENCILCAPS_DECR) (*log_func)(" D3DSTENCILCAPS_DECR");
      if(dw&D3DSTENCILCAPS_DECRSAT) (*log_func)(" D3DSTENCILCAPS_DECRSAT");
      if(dw&D3DSTENCILCAPS_INCR) (*log_func)(" D3DSTENCILCAPS_INCR");
      if(dw&D3DSTENCILCAPS_INCRSAT) (*log_func)(" D3DSTENCILCAPS_INCRSAT");
      if(dw&D3DSTENCILCAPS_INVERT) (*log_func)(" D3DSTENCILCAPS_INVERT");
      if(dw&D3DSTENCILCAPS_KEEP) (*log_func)(" D3DSTENCILCAPS_KEEP");
      if(dw&D3DSTENCILCAPS_REPLACE) (*log_func)(" D3DSTENCILCAPS_REPLACE");
      if(dw&D3DSTENCILCAPS_ZERO) (*log_func)(" D3DSTENCILCAPS_ZERO");
   }
   {
      (*log_func)("FVFCaps:");
      dword dw = cp.FVFCaps;
      if(dw&D3DFVFCAPS_DONOTSTRIPELEMENTS) (*log_func)(" D3DFVFCAPS_DONOTSTRIPELEMENTS");
      if(dw&D3DFVFCAPS_PSIZE) (*log_func)(" D3DFVFCAPS_PSIZE");
      (*log_func)(C_fstring("Texture coord count: %u", dw&D3DFVFCAPS_TEXCOORDCOUNTMASK));
   }
   {
      (*log_func)("TextureOpCaps:");
      dword dw = cp.TextureOpCaps;
      if(dw&D3DTEXOPCAPS_ADD) (*log_func)(" D3DTEXOPCAPS_ADD");
      if(dw&D3DTEXOPCAPS_ADDSIGNED) (*log_func)(" D3DTEXOPCAPS_ADDSIGNED");
      if(dw&D3DTEXOPCAPS_ADDSIGNED2X) (*log_func)(" D3DTEXOPCAPS_ADDSIGNED2X");
      if(dw&D3DTEXOPCAPS_ADDSMOOTH) (*log_func)(" D3DTEXOPCAPS_ADDSMOOTH");
      if(dw&D3DTEXOPCAPS_BLENDCURRENTALPHA) (*log_func)(" D3DTEXOPCAPS_BLENDCURRENTALPHA");
      if(dw&D3DTEXOPCAPS_BLENDDIFFUSEALPHA) (*log_func)(" D3DTEXOPCAPS_BLENDDIFFUSEALPHA");
      if(dw&D3DTEXOPCAPS_BLENDFACTORALPHA) (*log_func)(" D3DTEXOPCAPS_BLENDFACTORALPHA");
      if(dw&D3DTEXOPCAPS_BLENDTEXTUREALPHA) (*log_func)(" D3DTEXOPCAPS_BLENDTEXTUREALPHA");
      if(dw&D3DTEXOPCAPS_BLENDTEXTUREALPHAPM) (*log_func)(" D3DTEXOPCAPS_BLENDTEXTUREALPHAPM");
      if(dw&D3DTEXOPCAPS_BUMPENVMAP) (*log_func)(" D3DTEXOPCAPS_BUMPENVMAP");
      if(dw&D3DTEXOPCAPS_BUMPENVMAPLUMINANCE) (*log_func)(" D3DTEXOPCAPS_BUMPENVMAPLUMINANCE");
      if(dw&D3DTEXOPCAPS_DISABLE) (*log_func)(" D3DTEXOPCAPS_DISABLE");
      if(dw&D3DTEXOPCAPS_DOTPRODUCT3) (*log_func)(" D3DTEXOPCAPS_DOTPRODUCT3");
      if(dw&D3DTEXOPCAPS_LERP) (*log_func)(" D3DTEXOPCAPS_LERP");
      if(dw&D3DTEXOPCAPS_MODULATE) (*log_func)(" D3DTEXOPCAPS_MODULATE");
      if(dw&D3DTEXOPCAPS_MODULATE2X) (*log_func)(" D3DTEXOPCAPS_MODULATE2X");
      if(dw&D3DTEXOPCAPS_MODULATE4X) (*log_func)(" D3DTEXOPCAPS_MODULATE4X");
      if(dw&D3DTEXOPCAPS_MODULATEALPHA_ADDCOLOR) (*log_func)(" D3DTEXOPCAPS_MODULATEALPHA_ADDCOLOR");
      if(dw&D3DTEXOPCAPS_MODULATECOLOR_ADDALPHA) (*log_func)(" D3DTEXOPCAPS_MODULATECOLOR_ADDALPHA");
      if(dw&D3DTEXOPCAPS_MODULATEINVALPHA_ADDCOLOR) (*log_func)(" D3DTEXOPCAPS_MODULATEINVALPHA_ADDCOLOR");
      if(dw&D3DTEXOPCAPS_MODULATEINVCOLOR_ADDALPHA) (*log_func)(" D3DTEXOPCAPS_MODULATEINVCOLOR_ADDALPHA");
      if(dw&D3DTEXOPCAPS_MULTIPLYADD) (*log_func)(" D3DTEXOPCAPS_MULTIPLYADD");
      if(dw&D3DTEXOPCAPS_PREMODULATE) (*log_func)(" D3DTEXOPCAPS_PREMODULATE");
      if(dw&D3DTEXOPCAPS_SELECTARG1) (*log_func)(" D3DTEXOPCAPS_SELECTARG1");
      if(dw&D3DTEXOPCAPS_SELECTARG2) (*log_func)(" D3DTEXOPCAPS_SELECTARG2");
      if(dw&D3DTEXOPCAPS_SUBTRACT) (*log_func)(" D3DTEXOPCAPS_SUBTRACT");
   }
   (*log_func)(C_fstring("MaxTextureBlendStages: %u", cp.MaxTextureBlendStages));
   (*log_func)(C_fstring("MaxSimultaneousTextures: %u", cp.MaxSimultaneousTextures));
   {
      (*log_func)("VertexProcessingCaps:");
      dword dw = cp.VertexProcessingCaps;
      if(dw&D3DVTXPCAPS_DIRECTIONALLIGHTS) (*log_func)(" D3DVTXPCAPS_DIRECTIONALLIGHTS");
      if(dw&D3DVTXPCAPS_LOCALVIEWER) (*log_func)(" D3DVTXPCAPS_LOCALVIEWER");
      if(dw&D3DVTXPCAPS_MATERIALSOURCE7) (*log_func)(" D3DVTXPCAPS_MATERIALSOURCE7");
      if(dw&D3DVTXPCAPS_POSITIONALLIGHTS) (*log_func)(" D3DVTXPCAPS_POSITIONALLIGHTS");
      if(dw&D3DVTXPCAPS_TEXGEN) (*log_func)(" D3DVTXPCAPS_TEXGEN");
      if(dw&D3DVTXPCAPS_TWEENING) (*log_func)(" D3DVTXPCAPS_TWEENING");
      if(dw&D3DVTXPCAPS_NO_VSDT_UBYTE4) (*log_func)(" D3DVTXPCAPS_NO_VSDT_UBYTE4");
   }
   (*log_func)(C_fstring("MaxActiveLights: %u", cp.MaxActiveLights));
   (*log_func)(C_fstring("MaxUserClipPlanes: %u", cp.MaxUserClipPlanes));
   (*log_func)(C_fstring("MaxVertexBlendMatrices: %u", cp.MaxVertexBlendMatrices));
   (*log_func)(C_fstring("MaxVertexBlendMatrixIndex: %u", cp.MaxVertexBlendMatrixIndex));
   (*log_func)(C_fstring("MaxPointSize: %f", cp.MaxPointSize));
   (*log_func)(C_fstring("MaxPrimitiveCount: %u", cp.MaxPrimitiveCount));
   (*log_func)(C_fstring("MaxVertexIndex: %u", cp.MaxVertexIndex));
   (*log_func)(C_fstring("MaxStreams: %u", cp.MaxStreams));
   (*log_func)(C_fstring("MaxStreamStride: %u", cp.MaxStreamStride));
   (*log_func)(C_fstring("VertexShaderVersion: 0x%x", cp.VertexShaderVersion));
   (*log_func)(C_fstring("MaxVertexShaderConst: %u", cp.MaxVertexShaderConst));
   (*log_func)(C_fstring("PixelShaderVersion: 0x%x", cp.PixelShaderVersion));
   (*log_func)(C_fstring("MaxPixelShaderValue: %f", cp.MaxPixelShaderValue));
   {
      static const struct S_format{
         D3DFORMAT fmt;
         const char *name;
      } formats[] = {
         D3DFMT_R8G8B8, "D3DFMT_R8G8B8",
         D3DFMT_A8R8G8B8, "D3DFMT_A8R8G8B8",
         D3DFMT_X8R8G8B8, "D3DFMT_X8R8G8B8",
         D3DFMT_R5G6B5, "D3DFMT_R5G6B5",
         D3DFMT_X1R5G5B5, "D3DFMT_X1R5G5B5",
         D3DFMT_A1R5G5B5, "D3DFMT_A1R5G5B5",
         D3DFMT_A4R4G4B4, "D3DFMT_A4R4G4B4",
         D3DFMT_R3G3B2, "D3DFMT_R3G3B2",
         D3DFMT_A8, "D3DFMT_A8",
         D3DFMT_A8R3G3B2, "D3DFMT_A8R3G3B2",
         D3DFMT_X4R4G4B4, "D3DFMT_X4R4G4B4",
         D3DFMT_A8P8, "D3DFMT_A8P8",
         D3DFMT_P8, "D3DFMT_P8",
         D3DFMT_L8, "D3DFMT_L8",
         D3DFMT_A8L8, "D3DFMT_A8L8",
         D3DFMT_A4L4, "D3DFMT_A4L4",
         D3DFMT_V8U8, "D3DFMT_V8U8",
         D3DFMT_L6V5U5, "D3DFMT_L6V5U5",
         D3DFMT_X8L8V8U8, "D3DFMT_X8L8V8U8",
         D3DFMT_Q8W8V8U8, "D3DFMT_Q8W8V8U8",
         D3DFMT_V16U16, "D3DFMT_V16U16",
         D3DFMT_W11V11U10, "D3DFMT_W11V11U10",
         D3DFMT_UYVY, "D3DFMT_UYVY",
         D3DFMT_YUY2, "D3DFMT_YUY2",
         D3DFMT_DXT1, "D3DFMT_DXT1",
         D3DFMT_DXT2, "D3DFMT_DXT2",
         D3DFMT_DXT3, "D3DFMT_DXT3",
         D3DFMT_DXT4, "D3DFMT_DXT4",
         D3DFMT_DXT5, "D3DFMT_DXT5",
         D3DFMT_D16_LOCKABLE, "D3DFMT_D16_LOCKABLE",
         D3DFMT_D32, "D3DFMT_D32",
         D3DFMT_D15S1, "D3DFMT_D15S1",
         D3DFMT_D24S8, "D3DFMT_D24S8",
         D3DFMT_D16, "D3DFMT_D16",
         D3DFMT_D24X8, "D3DFMT_D24X8",
         D3DFMT_D24X4S4, "D3DFMT_D24X4S4",
         D3DFMT_VERTEXDATA, "D3DFMT_VERTEXDATA",
         D3DFMT_INDEX16, "D3DFMT_INDEX16",
         D3DFMT_INDEX32, "D3DFMT_INDEX32",
         D3DFMT_UNKNOWN, NULL
      };
      D3DDISPLAYMODE dm;
      lpDev8->GetDisplayMode(&dm);
      for(int use=0; use<2; use++){
         (*log_func)(C_fstring("%s formats:", !use ? "Texture" : "Depth buffer"));

         for(int i=0, j=0; formats[i].name; i++){
            hr = lpD3D8->CheckDeviceFormat(D3DADAPTER_DEFAULT, cp.DeviceType,
               dm.Format,
               !use ? 0 : D3DUSAGE_DEPTHSTENCIL,
               !use ? D3DRTYPE_TEXTURE : D3DRTYPE_SURFACE,
               formats[i].fmt);
            if(SUCCEEDED(hr)){
               (*log_func)(C_fstring(" %2i. %s:", j, formats[i].name));
               ++j;
            }
         }
      }
   }
   (*log_func)("----------------------");
}

//----------------------------
                              //keyboard hook
static LPIGraph lpg_key_hooked;

                              //bug - we receive messages after process is done - this flag helps us (but it's not clean)
static bool process_done;

//----------------------------
                              //keyboard hook proc
LRESULT CALLBACK GKeyboardProc(int code, WPARAM wParam, LPARAM lParam){

   if(process_done || !lpg_key_hooked)
      return 0;

   if(code<0){
      return CallNextHookEx((HHOOK)lpg_key_hooked->kb_hook, code, wParam, lParam);
   }

   word wInfo = (word)(lParam>>16);
   char key_code = wParam;

                              //check for ctrl-alt-del
   if(key_code==VK_DELETE && (wInfo&KF_ALTDOWN) && (lpg_key_hooked->shiftkeys&SKEY_CTRL))
      return CallNextHookEx((HHOOK)lpg_key_hooked->kb_hook, code, wParam, lParam);
   /*
                              //Alt+Space
   if(key_code==VK_SPACE && (wInfo&KF_ALTDOWN))
      return CallNextHookEx((HHOOK)lpg_key_hooked->kb_hook, code, wParam, lParam);
      */
                              //shift keys
   if(key_code==VK_SHIFT){
      if(!(lParam&0x80000000)) lpg_key_hooked->shiftkeys |= SKEY_SHIFT;
      else lpg_key_hooked->shiftkeys &= ~SKEY_SHIFT;
   }else
   if(wParam==VK_CONTROL){
      if(!(lParam&0x80000000)) lpg_key_hooked->shiftkeys |= SKEY_CTRL;
      else lpg_key_hooked->shiftkeys &= ~SKEY_CTRL;
   }else
   if((wInfo&KF_ALTDOWN)){
      lpg_key_hooked->shiftkeys |= SKEY_ALT;
   }else{
      lpg_key_hooked->shiftkeys &= ~SKEY_ALT;
   }
   if(!(lParam&0x80000000)){  //pressed
      lpg_key_hooked->key = key_code;
      lpg_key_hooked->keyboard_map[(byte)key_code] = true;
   }else{
      lpg_key_hooked->keyboard_map[(byte)key_code] = false;
   }
                              //update game-keys table
   int i;
   for(i=0; i<lpg_key_hooked->num_key_scodes; ++i)
   if(key_code==lpg_key_hooked->key_scodes[i]){
      if(!(lParam&0x80000000)) lpg_key_hooked->game_keys |= (1<<i);
      else lpg_key_hooked->game_keys &= (~(1<<i));
   }
   if(lpg_key_hooked->shiftkeys&SKEY_ALT){
                              //allow menu access
      if(lpg_key_hooked->create_flags&IG_LOADMENU){
         if(isalnum(key_code))
            return 0;
      }else{
         if(key_code==VK_F4)
            return 0;
      }
                              //allow task-switching
      switch(key_code){
      case VK_TAB:
      case VK_ESCAPE:
         return 0;
      }
   }
   if(lpg_key_hooked->shiftkeys&SKEY_CTRL){
      switch(key_code){
      case VK_ESCAPE:
         return 0;
      }
   }

   switch(key_code){
   case K_NUMLOCK:
   case K_CAPSLOCK:
   case K_SCROLLLOCK:
      return 0;
   }

   return 1;
}

//----------------------------
                              //un-install grab proc
static bool GrabKeys(LPIGraph igraph, bool on){

   if(on){                    //install
      if(!igraph->kb_hook){
         igraph->shiftkeys = 0;
         igraph->game_keys = 0;
         if(lpg_key_hooked)
            return false;
         lpg_key_hooked = igraph;
         igraph->kb_hook = SetWindowsHookEx(WH_KEYBOARD, (HOOKPROC)GKeyboardProc,
            NULL, GetCurrentThreadId());
                              //read currently pressed keys
         {
            byte kb[256];
            GetKeyboardState(kb);
            if(kb[VK_CONTROL]&0x80) igraph->shiftkeys |= SKEY_CTRL;
            if(kb[VK_MENU]&0x80) igraph->shiftkeys |= SKEY_ALT;
            if(kb[VK_SHIFT]&0x80) igraph->shiftkeys |= SKEY_SHIFT;
         }
      }
   }else{                     //uninstall
      if(igraph->kb_hook){
         if(lpg_key_hooked!=igraph)
            return false;
         UnhookWindowsHookEx((HHOOK)igraph->kb_hook);
         igraph->kb_hook = NULL;
         lpg_key_hooked = NULL;
      }
   }
   return true;
}

//----------------------------
//----------------------------
                              //win32 macros
inline void SETRECT(RECT *rc, int l,int t,int r,int b){
   rc->left=l, rc->top=t, rc->right=r, rc->bottom=b;
}

//----------------------------

#pragma pack(push,1)
struct S_config{              //saved settings
   dword posx, posy;
};
#pragma pack(pop)

//----------------------------

BOOL WINAPI DllMain(HINSTANCE hinstDLL, dword fdwReason, void *lpvReserved){

   switch(fdwReason){
   case DLL_THREAD_ATTACH:
      {
      }
      break;

   case DLL_THREAD_DETACH:
      {
      }
      break;

   case DLL_PROCESS_ATTACH:
      {
                              //this func switches to MSVC mem allocation, instead of system allocation
         _set_sbh_threshold(1016);
      }
      break;

   case DLL_PROCESS_DETACH:
      break;
   }
   return true;
}

//----------------------------
                              //graphics window procedure
long FAR PASCAL GWindowProc(HWND hwnd, UINT message,
   WPARAM wParam, LPARAM lParam){

                              //get interface
   LPIGraph igraph = (LPIGraph)GetWindowLong(hwnd, GWL_USERDATA);

   if(!igraph || process_done)
      return DefWindowProc(hwnd, message, wParam, lParam);

   switch(message){
                              //button down
   case WM_LBUTTONDOWN:
      igraph->mouse_butt1 |= 1;
      break;

   case WM_RBUTTONDOWN:
      igraph->mouse_butt1 |= 2;
      break;
   
   case WM_MBUTTONDOWN:
      igraph->mouse_butt1 |= 4;
      break;
                              //button up
   case WM_LBUTTONUP:
      igraph->mouse_butt1 &= -2;
      break;
   
   case WM_RBUTTONUP:
      igraph->mouse_butt1 &= -3;
      break;
   
   case WM_MBUTTONUP:
      igraph->mouse_butt1 &= -5;
      break;

   case WM_LBUTTONDBLCLK:
   case WM_MBUTTONDBLCLK:
   case WM_RBUTTONDBLCLK:
   case WM_NCLBUTTONDBLCLK:
   case WM_NCMBUTTONDBLCLK:
   case WM_NCRBUTTONDBLCLK:
   case WM_MOUSEACTIVATE:
   case WM_NCHITTEST:
      if((igraph->create_flags&IG_FULLSCREEN) && !(igraph->create_flags&IG_NOHIDEMOUSE)){
         return 0;
      }
      break;

   case WM_MOUSEMOVE:
      if(igraph->mouse_y1!=-2){
         igraph->mouse_x1 = (short)LOWORD(lParam);
         igraph->mouse_y1 = (short)HIWORD(lParam);
      }
      if(igraph->create_flags&IG_FULLSCREEN)
         return 0;
      break;

   case WM_MOUSEWHEEL:
      {
         int i = wParam>>16;
         if(i&0x8000)
            i |= 0xffff0000;
         igraph->mouse_z1 = i / 120;
         igraph->mouse_x1 = LOWORD(lParam);
         igraph->mouse_y1 = HIWORD(lParam);
      }
      break;

      /*
   case WM_GETMINMAXINFO:
      if(igraph->create_flags&IG_FULLSCREEN){
         LPMINMAXINFO lpmmi;
         lpmmi=(LPMINMAXINFO)lParam;
         lpmmi->ptMaxSize.x = igraph->scrn_sx;
         lpmmi->ptMaxSize.y = igraph->scrn_sy;
         lpmmi->ptMaxTrackSize.x = igraph->scrn_sx;
         lpmmi->ptMaxTrackSize.y = igraph->scrn_sy;
         return 0;
      }
      break;
      */

   case WM_SETCURSOR:
                              //no cursor in fullscreen mode
                              //or when inactive
      if(igraph->active && (igraph->create_flags&IG_FULLSCREEN)
         && !(igraph->create_flags&IG_NOHIDEMOUSE)){

         SetCursor(NULL);
         return TRUE;
      }
                              //hide cursor in debug mouse mode
      if(igraph->active && igraph->mouse_init && ((igraph->create_flags&IG_DEBUGMOUSE))){
         SetCursor(NULL);
         return TRUE;
      }
      break;

   case WM_ACTIVATE:
      {
         int i = LOWORD(wParam);
         if(igraph->active == (bool)i)
            break;
         igraph->active = i;
         if(igraph->h_mouse_event)
            igraph->event_signalled = true;
         else
            igraph->mouse_butt1 = 0;
         igraph->mouse_butt = 0;
                              //grab keys only if active
         if(igraph->IsKbInit()){
            GrabKeys(igraph, igraph->active);
            if(igraph->active){
               memset(igraph->keyboard_map, 0, sizeof(igraph->keyboard_map));
            }
         }
         igraph->MouseAcquire(igraph->active);
#ifdef ENABLE_PALETIZED_DESKTOP
         if(!igraph->active && igraph->init_ok && igraph->rgb_conv->Bpp()==1){
                                 //losing focus
                                 //save current palette
            igraph->save_pal = new byte[1024];
            igraph->GetPal(igraph->save_pal);
         }
#endif                        //ENABLE_PALETIZED_DESKTOP
         if(igraph->active){
            if(igraph->create_flags&IG_SUSPENDINACTIVE){
                                 //wake up sleeper (send any message)
               PostMessage(hwnd, WM_USER, 0, 0);
            }
         }
         if(igraph->init_ok){
            for(i=igraph->cb_proc_list.size(); i--; )
               (*igraph->cb_proc_list[i].cb_proc)(CM_ACTIVATE, igraph->active, 0, igraph->cb_proc_list[i].context);
         }
      }
      break;

   case WM_COMMAND:
      if(igraph->init_ok){
         for(int i=igraph->cb_proc_list.size(); i--; )
            (*igraph->cb_proc_list[i].cb_proc)(CM_COMMAND, LOWORD(wParam), HIWORD(wParam), igraph->cb_proc_list[i].context);
      }
      break;

   case WM_SETFOCUS:
      if(igraph->IsKbInit())
         GrabKeys(igraph, true);
      if(igraph->mouse_init){
         if(igraph->h_mouse_event)
            igraph->event_signalled = true;
         igraph->mouse_butt = 0;
      }else
         igraph->mouse_butt1 = 0;
      break;

   case WM_KILLFOCUS:
      if(igraph->IsKbInit())
         GrabKeys(igraph, false);
      igraph->key = K_NOKEY;
      igraph->game_keys = 0;
      igraph->shiftkeys = 0;
      break;

   case WM_ENTERMENULOOP:
   case WM_ENTERSIZEMOVE:
                              //allow move through menu
      if(igraph->IsKbInit())
         GrabKeys(igraph, false);
      igraph->MouseAcquire(false);
      igraph->active = false;
      break;

   case WM_EXITMENULOOP:
   case WM_EXITSIZEMOVE:
                              //menu finished
      if(GetFocus())
         igraph->active = true;
      if(igraph->IsKbInit())
         GrabKeys(igraph, igraph->active);
      igraph->MouseAcquire(true);
      if(igraph->h_mouse_event)
         igraph->event_signalled = true;
      igraph->mouse_butt = 0;
      igraph->NullTimer();
      igraph->key = K_NOKEY;
      if(message==WM_EXITSIZEMOVE){
         RECT rc;
         GetClientRect(hwnd, &rc);
                              //consider status window and other helping windows
         struct S_hlp{
            static BOOL CALLBACK cbEnum(HWND hwnd, LPARAM lParam){
               char buf[256];
               if(GetClassName(hwnd, buf, sizeof(buf))){
                  if(!strcmp(buf, "msctls_statusbar32")){
                     RECT &rc = *(RECT*)lParam;
                     RECT rc1;
                     GetWindowRect(hwnd, &rc1);
                     rc.bottom -= (rc1.bottom - rc1.top);
                  }
               }
               return true;
            }
         };
         EnumChildWindows(hwnd, S_hlp::cbEnum, (LPARAM)&rc);

         rc.right = Max(4L, rc.right);
         rc.bottom = Max(4L, rc.bottom);
         igraph->UpdateParams(rc.right, rc.bottom, 0, 0, 0, false);
      }
      break;

#ifdef ENABLE_PALETIZED_DESKTOP
   case WM_QUERYNEWPALETTE:
                              //set our palette again
      if(igraph->init_ok && !(igraph->create_flags&IG_FULLSCREEN)){
                              //use our saved palette
         if(igraph->save_pal!=NULL){
            igraph->SetPal(igraph->save_pal);
            delete[] igraph->save_pal;
            igraph->save_pal=NULL;
         }
         return TRUE;
      }
      return FALSE;
#endif                        //ENABLE_PALETIZED_DESKTOP

   case WM_MOVE:
      {
         igraph->mouse_butt1 = 0;
      }
      break;

   case WM_PAINT:             //repaint our screen
      if(igraph->init_ok){
         igraph->UpdateScreen(0);
      }
      break;

   case WM_ERASEBKGND:
      return true;

   case WM_CLOSE:
      //igraph->key = (char)K_CLOSE;
      igraph->want_close_app = true;
      return 0;

   case WM_DESTROY:           //immediate exit
      if(igraph->init_ok)
         exit(0);
      break;
   default:
      if(message==igraph->wm_mousewheel)
         igraph->mouse_z1 = (int)wParam / WHEEL_DELTA * igraph->mouse_wheel_lines;
   }
   return DefWindowProc(hwnd, message, wParam, lParam);
}

//----------------------------
//----------------------------
                              //creation / destroying
static bool IRegisterClass(HINSTANCE hInstance){

   WNDCLASSEX wc;
   memset(&wc, 0, sizeof(wc));
   wc.cbSize = sizeof(WNDCLASSEX);
   wc.style = 0;
   wc.lpfnWndProc=GWindowProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
                              //try load icon #1 from resources
   HICON hicon;
   hicon = LoadIcon(hInstance, MAKEINTRESOURCE(1));
                              //if it failed, set default app icon
   wc.hIcon = hicon ? hicon : LoadIcon(NULL, IDI_APPLICATION);
   wc.hCursor = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = (HBRUSH)GetStockObject(NULL_BRUSH);
   wc.lpszMenuName = NULL;
   wc.lpszClassName = "IGraph_Window";
   wc.cbClsExtra = 4;         //for graphics interface

   RegisterClassEx(&wc);

   return true;
}

//----------------------------

bool IGraph::InitWindow(int posx, int posy){


                              //add non-client area
   dword wstyle = WS_POPUP | WS_CLIPCHILDREN;
   if(!(create_flags&IG_NOTITLEBAR) && !(create_flags&IG_FULLSCREEN))
      wstyle |= WS_CAPTION;
   if(create_flags&IG_RESIZEABLE)
      wstyle |= WS_THICKFRAME;
   if(!(create_flags&IG_NOMINIMIZE)) wstyle |= WS_MINIMIZEBOX;
   if(!(create_flags&IG_NOSYSTEMMENU)) wstyle |= WS_SYSMENU;
   wstyle |= WS_CLIPCHILDREN;
                              //get size of window
   dword mode_sx;
   dword mode_sy;
   RECT rc;
   hmenu_main = NULL;

   if(!(create_flags&IG_FULLSCREEN)){
      if(create_flags&IG_LOADMENU){
         hmenu_main = LoadMenu((HINSTANCE)hInstance, "MENU_MAIN");
         if(!hmenu_main)
            hmenu_main = CreateMenu();
      }
   }

   if(create_flags&IG_FULLSCREEN){
      mode_sx = scrn_sx;
      mode_sy = scrn_sy;
#ifdef ENABLE_PALETIZED_DESKTOP
      SetPalRange(1, 254);
#endif
   }else{
      SETRECT(&rc, 0, 0, scrn_sx, scrn_sy);
      AdjustWindowRect(&rc, wstyle, (bool)hmenu_main);
      mode_sx = rc.right - rc.left;
      mode_sy = rc.bottom - rc.top;
#ifdef ENABLE_PALETIZED_DESKTOP
      SetPalRange(10, 236);
#endif
   }

   if(!hwnd){
                              //create window
      hwnd = CreateWindowEx(0,
         "IGraph_Window",     //ClassName
         appName,
         wstyle,
         ((create_flags&IG_FULLSCREEN) ? 0 : posx),
         ((create_flags&IG_FULLSCREEN) ? 0 : posy),
         mode_sx, mode_sy,
         NULL,                //hwndParent
         hmenu_main,
         (HINSTANCE)hInstance,
         NULL);               //lpvParam

      if(!hwnd) return false;
                              //store this into window data
      SetWindowLong(hwnd, GWL_USERDATA, (LONG)this);
   }else{
      SetWindowLong(hwnd, GWL_EXSTYLE, 0);
      SetWindowLong(hwnd, GWL_STYLE, wstyle);
      SetWindowPos(hwnd, (create_flags&IG_FULLSCREEN) ? HWND_TOPMOST : HWND_NOTOPMOST,
         ((create_flags&IG_FULLSCREEN) ? 0 : posx),
         ((create_flags&IG_FULLSCREEN) ? 0 : posy),
         mode_sx, mode_sy,
         0);
   }

   if(!(create_flags&IG_HIDDEN)){
                              //show our window
      ShowWindow(hwnd, SW_SHOW);
      UpdateWindow(hwnd);
      SetFocus(hwnd);
   }

   return true;
}

//----------------------------

void IGraph::DestroyWindow(){
                              //don't let it catch destroy messages
   init_ok = false;
   if(hwnd){
      SetWindowLong(hwnd, GWL_USERDATA, 0);
      ::DestroyWindow(hwnd);
      hwnd = NULL;
   }
}

//----------------------------

static const char key_name[] = "Software\\Insanity3D system\\Applications";

//----------------------------

static bool GetKeyName(HINSTANCE hInstance, C_string &rtn){
                              //create regkey name from application name
   char module_name[257];
   if(!GetModuleFileName((HINSTANCE)hInstance, module_name, sizeof(module_name)))
      return false;
   rtn = C_fstring("%s\\%s", key_name, module_name);
   return true;
}

//----------------------------

static bool ReadConfig(HINSTANCE hi, S_config &cfg,
   dword size_x, dword size_y){

   C_string key_name;
   bool b = GetKeyName(hi, key_name);
   assert(b);
   if(!b) return false;
   int rkey = RegkeyOpen(key_name, E_REGKEY_CURRENT_USER);
   if(rkey!=-1){
                              //read position
      int rlen = RegkeyRdata(rkey, "Window position", &cfg, sizeof(S_config));
      RegkeyClose(rkey);
      if(rlen==sizeof(S_config))
         return true;
   }
                              //setup defaults
   cfg.posx = (GetSystemMetrics(SM_CXSCREEN)-size_x)/2;
   cfg.posy = (GetSystemMetrics(SM_CYSCREEN)-size_y)/2;

   return false;
}

//----------------------------

static bool WriteConfig(HINSTANCE hi, const RECT &rc){

   C_string key_name;
   bool b = GetKeyName(hi, key_name);
   assert(b);
   if(!b) return false;
   int rkey = RegkeyCreate(key_name, E_REGKEY_CURRENT_USER);
   assert(rkey!=-1);
   if(rkey==-1) return false;

   S_config cfg;
   cfg.posx = rc.left;
   cfg.posy = rc.top;
                              //write position
   b = RegkeyWdata(rkey, "Window position", (const byte*)&cfg, sizeof(S_config));
   RegkeyClose(rkey);
   assert(b);
   return b;
}

//----------------------------

IGraph::IGraph():
   ig_flags(0),
   want_close_app(false)
{
}

//----------------------------

IGraph::~IGraph(){

   if(!init_ok)
      return;

   MouseClose();
   CloseKeyboard();
   if(!(create_flags&IG_FULLSCREEN)){
                              //write position only if window is in maximized state
      WINDOWPLACEMENT wp;
      wp.length = sizeof(wp);
      if(GetWindowPlacement(hwnd, &wp) && wp.showCmd==SW_SHOWNORMAL){
         RECT rc;
#ifndef NDEBUG
         bool b = 
#endif
            GetWindowRect(hwnd, &rc);
         assert(b);
         WriteConfig((HINSTANCE)hInstance, rc);
      }
   }

   if(lpDI8){
      lpDI8->Release();
      lpDI8 = NULL;
   }
                                       //DD destroy
   if(lpDev8){
#ifdef ENABLE_PALETIZED_DESKTOP
      if(lpDDPal){
         lpDDPal->Release();
         lpDDPal = NULL;
      }
#endif
      lpDev8->Release();
      lpDev8 = NULL;

      lpD3D8->Release();
      lpD3D8 = NULL;
   }

   DestroyWindow();

   UnregisterClass("IGraph_Window", (HINSTANCE)hInstance);
#ifdef ENABLE_PALETIZED_DESKTOP
   delete[] currpal;
#endif

   if(h_d3d8){
      FreeLibrary(h_d3d8);
      h_d3d8 = NULL;
   }

   if(h_dinput){
      FreeLibrary(h_dinput);
      h_dinput = NULL;
   }
   if(!use_qpc){
      timeEndPeriod(1);
   }

#ifdef LOG_MODE
   IG_log.os_log = NULL;
#endif                        //LOG_MODE
}

//----------------------------

static dword MakeBitMask(dword num_bits, dword shift){
   dword ret = 0;
   for(int i=0; i<(int)num_bits; i++)
      ret |= (1 << (shift+i));
   return ret;
}

//----------------------------

static void ConvertD3DpfToI3Dpf(const D3DFORMAT d3d_fmt, S_pixelformat &pf){

   memset(&pf, 0, sizeof(pf));
                        //setup bpp
   switch(d3d_fmt){
   case D3DFMT_P8:
   case D3DFMT_R3G3B2:
      pf.bytes_per_pixel = 1;
      break;
   case D3DFMT_R5G6B5:
   case D3DFMT_X1R5G5B5:
   case D3DFMT_A1R5G5B5:
   case D3DFMT_A4R4G4B4:
   case D3DFMT_A8R3G3B2:
   case D3DFMT_X4R4G4B4:
   case D3DFMT_A8P8:
      pf.bytes_per_pixel = 2;
      break;
   case D3DFMT_R8G8B8:
      pf.bytes_per_pixel = 3;
      break;
   case D3DFMT_A8R8G8B8:
   case D3DFMT_X8R8G8B8:
      pf.bytes_per_pixel = 4;
      break;
   default:
      assert(0);
   }
                        //setup alpha/palette/compression
   switch(d3d_fmt){
   case D3DFMT_DXT1:
   case D3DFMT_DXT2:
   case D3DFMT_DXT3:
   case D3DFMT_DXT4:
   case D3DFMT_DXT5:
      pf.flags = PIXELFORMAT_COMPRESS;
      pf.four_cc = MAKEFOURCC('D', 'X', 'T', '1' + d3d_fmt - D3DFMT_DXT1);
      break;
   case D3DFMT_P8:
   case D3DFMT_A8P8:
      pf.flags = PIXELFORMAT_PALETTE;
      break;
   case D3DFMT_A8R8G8B8:
   case D3DFMT_A1R5G5B5:
   case D3DFMT_A4R4G4B4:
   case D3DFMT_A8R3G3B2:
      pf.flags = PIXELFORMAT_ALPHA;
      break;
   }
                        //setup RGB components
   switch(d3d_fmt){
   case D3DFMT_R8G8B8:
   case D3DFMT_X8R8G8B8:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x00ff0000;
      pf.g_mask = 0x0000ff00;
      pf.b_mask = 0x000000ff;
      break;
   case D3DFMT_A8R8G8B8:
      pf.a_mask = 0xff000000;
      pf.r_mask = 0x00ff0000;
      pf.g_mask = 0x0000ff00;
      pf.b_mask = 0x000000ff;
      break;
   case D3DFMT_R5G6B5:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x0000f800;
      pf.g_mask = 0x000007e0;
      pf.b_mask = 0x0000001f;
      break;
   case D3DFMT_X1R5G5B5:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x00007c00;
      pf.g_mask = 0x000003e0;
      pf.b_mask = 0x0000001f;
      break;
   case D3DFMT_A1R5G5B5:
      pf.a_mask = 0x00008000;
      pf.r_mask = 0x00007c00;
      pf.g_mask = 0x000003e0;
      pf.b_mask = 0x0000001f;
      break;
   case D3DFMT_A4R4G4B4:
      pf.a_mask = 0x0000f000;
      pf.r_mask = 0x00000f00;
      pf.g_mask = 0x000000f0;
      pf.b_mask = 0x0000000f;
      break;
   case D3DFMT_X4R4G4B4:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x00000f00;
      pf.g_mask = 0x000000f0;
      pf.b_mask = 0x0000000f;
      break;
   case D3DFMT_R3G3B2:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x000000e0;
      pf.g_mask = 0x0000001c;
      pf.b_mask = 0x00000003;
      break;
   case D3DFMT_A8R3G3B2:
      pf.a_mask = 0x0000ff00;
      pf.r_mask = 0x000000e0;
      pf.g_mask = 0x0000001c;
      pf.b_mask = 0x00000003;
      break;
   }
}

//----------------------------

HRESULT IGraph::InitRGBConv(){

   HRESULT hr = D3D_OK;

                              //init color conversion tables
   rgb_conv = Create_rgb_conv();
   rgb_conv->Release();
   S_pixelformat pf;
   memset(&pf, 0, sizeof(pf));

   {
                              //get current color model
      D3DDISPLAYMODE dm;
      HRESULT hr = lpDev8->GetDisplayMode(&dm);
      CHECK_D3D_RESULT("GetDisplayMode", hr);
      if(FAILED(hr))
         return hr;
      ConvertD3DpfToI3Dpf(dm.Format, pf);
   }

   if(pf.bytes_per_pixel == 1)
      pf.flags |= PIXELFORMAT_PALETTE;
   if(pf.a_mask) pf.flags |= PIXELFORMAT_ALPHA;
   rgb_conv->Init(pf);

#ifdef ENABLE_PALETIZED_DESKTOP
                              //create palette object
   PALETTEENTRY pe[256];

   if(rgb_conv->IsPaletized()){
                              //get current palette
      HDC hdc;
      hdc = GetDC(NULL);
      GetSystemPaletteEntries(hdc, 0, 256, pe);
      ReleaseDC(NULL, hdc);
      for(i=firstpal; i<firstpal+numpals; i++){
         pe[i].peRed = (i&0xe0);
         pe[i].peGreen = (i&0x1c)<<3;
         pe[i].peBlue = (i&0x03)<<6;
      }
   }else{
      firstpal = 0, numpals = 256;
      memset(pe, 0, sizeof(pe));
   }
   int i;
                              //initialize
   i = firstpal;
   while(i--) pe[i].peFlags = D3DPAL_READONLY;
   i = 256-(firstpal + numpals);
   while(i--) pe[firstpal + numpals + i].peFlags = D3DPAL_READONLY;
   i = numpals;
   while(i--) pe[firstpal + i].peFlags = D3DPAL_FREE | PC_RESERVED;
#endif   //ENABLE_PALETIZED_DESKTOP

   return hr;
}

//----------------------------

static bool IsDepthFormatOk(IDirect3D8 *lpD3D8, int adapter_id, D3DDEVTYPE dev_type,
   D3DFORMAT DepthFormat, D3DFORMAT AdapterFormat, D3DFORMAT BackBufferFormat){

                              //verify that the depth format exists.
   HRESULT hr = lpD3D8->CheckDeviceFormat(adapter_id,
      dev_type, AdapterFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, DepthFormat);

   if(FAILED(hr))
      return false;

                              //verify that the depth format is compatible.
   hr = lpD3D8->CheckDepthStencilMatch(adapter_id, dev_type,
      AdapterFormat, BackBufferFormat, DepthFormat);
   return SUCCEEDED(hr);
}

//----------------------------
                              //formats are selected in order as they appear here
static const D3DFORMAT depth_formats_16bit[] = {
   D3DFMT_D16,
   D3DFMT_D15S1,
   D3DFMT_D16_LOCKABLE,
}, depth_formats_32bit[] = {
   D3DFMT_D24S8,
   D3DFMT_D24X4S4,
   D3DFMT_D32,
   D3DFMT_D24X8,
};

//----------------------------

static D3DFORMAT ChooseDepthFormat(IDirect3D8 *lpD3D8, int adapter_id, D3DDEVTYPE dev_type,
   dword create_bits_per_pixel, D3DFORMAT BackBufferFormat){

   int i;

   switch(create_bits_per_pixel){
   case 8:
   case 16:
   case 24:
      /*
      for(i=0; i<sizeof(depth_formats_16bit)/sizeof(D3DFORMAT); i++){
         if(IsDepthFormatOk(lpD3D8, adapter_id, dev_type, depth_formats_16bit[i], BackBufferFormat, BackBufferFormat))
            return depth_formats_16bit[i];
      }
      */
      for(i=0; i<sizeof(depth_formats_32bit)/sizeof(D3DFORMAT); i++){
         if(IsDepthFormatOk(lpD3D8, adapter_id, dev_type, depth_formats_32bit[i], BackBufferFormat, BackBufferFormat))
            return depth_formats_32bit[i];
      }
      for(i=0; i<sizeof(depth_formats_16bit)/sizeof(D3DFORMAT); i++){
         if(IsDepthFormatOk(lpD3D8, adapter_id, dev_type, depth_formats_16bit[i], BackBufferFormat, BackBufferFormat))
            return depth_formats_16bit[i];
      }
      break;
   case 32:
      for(i=0; i<sizeof(depth_formats_32bit)/sizeof(D3DFORMAT); i++){
         if(IsDepthFormatOk(lpD3D8, adapter_id, dev_type, depth_formats_32bit[i], BackBufferFormat, BackBufferFormat)){
            return depth_formats_32bit[i];
         }
      }
      for(i=0; i<sizeof(depth_formats_16bit)/sizeof(D3DFORMAT); i++){
         if(IsDepthFormatOk(lpD3D8, adapter_id, dev_type, depth_formats_16bit[i], BackBufferFormat, BackBufferFormat)){
            return depth_formats_16bit[i];
         }
      }
      break;
   default: assert(0);
   }
   return D3DFMT_UNKNOWN;
}

//----------------------------

void IGraph::InitPresentStruct(int adapter_id, D3DDEVTYPE dev_type, D3DPRESENT_PARAMETERS &pp) const{

   dword bpp;

   memset(&pp, 0, sizeof(pp));
   pp.BackBufferWidth = scrn_sx;
   pp.BackBufferHeight = scrn_sy;
   pp.BackBufferFormat = D3DFMT_UNKNOWN;
   pp.BackBufferCount = 1;
   if(create_flags&IG_FULLSCREEN){
      bpp = create_bits_per_pixel;

      switch(create_bits_per_pixel){
      case 8:
         pp.BackBufferFormat = D3DFMT_P8;
         break;
      case 16:
         pp.BackBufferFormat = D3DFMT_R5G6B5;
         break;
      case 24:
         pp.BackBufferFormat = D3DFMT_R8G8B8;
         break;
      case 32:
         pp.BackBufferFormat = D3DFMT_X8R8G8B8;
         break;
      }
      if(create_flags&IG_TRIPPLEBUF)
         ++pp.BackBufferCount;
      pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
   }else{
      D3DDISPLAYMODE dm;
      lpD3D8->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &dm);
      pp.BackBufferFormat = dm.Format;

      S_pixelformat pf;
      ConvertD3DpfToI3Dpf(dm.Format, pf);
      bpp = pf.bytes_per_pixel * 8;

      pp.SwapEffect = D3DSWAPEFFECT_COPY;
   }
   pp.MultiSampleType = D3DMULTISAMPLE_NONE;
   pp.hDeviceWindow = hwnd;
   pp.Windowed = !(create_flags&IG_FULLSCREEN);
   pp.EnableAutoDepthStencil = true;

   D3DFORMAT depth_format = ChooseDepthFormat(lpD3D8, adapter_id, dev_type, bpp, pp.BackBufferFormat);
   pp.AutoDepthStencilFormat = depth_format;

   pp.Flags = 0;
   if(create_flags&IG_FULLSCREEN){
      pp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
   }
}

//----------------------------
                              //initialize graphics
bool IGraph::Initialize(void *hinst, dword pos_x, dword pos_y, dword sx1, dword sy1,
   dword flags, int bits_per_pixel,
   int adapter_id, dword d3d_device_type,
   IG_INIT::t_log_func *log_func1){

                              //clean/setup defaults
   init_ok = false;
   h_dinput = NULL;
   log_func = log_func1;

   backbuffer_count = 0;

   lpD3D8 = NULL;
   lpDev8 = NULL;

#ifdef ENABLE_PALETIZED_DESKTOP
   lpDDPal = NULL;
   currpal = NULL;
   save_pal = NULL;
   currpal = new byte[1024];
#endif

   desktop_sx = GetSystemMetrics(SM_CXSCREEN);
   desktop_sy = GetSystemMetrics(SM_CYSCREEN);

   hwnd = NULL;

   active = false;            //inactive
   sleep_loop = false;

   splitmode = SPLIT_UNIFY;

   hInstance = hinst;
                              //keys
   shiftkeys = 0;
   key = K_NOKEY;
   game_keys = 0;
   num_key_scodes = 0;
   kb_init = false;
   kb_hook = NULL;
   memset(keyboard_map, 0, sizeof(keyboard_map));
                              //mouse
   mouse_init = false;
   lpDIMouse8 = NULL;
   h_mouse_event = NULL;
   h_mouse_thread = NULL;
   event_signalled = false;
   acq_count = 0;
   mouse_butt1 = 0;
   mouse_left = mouse_top = 0;
   mouse_right = sx1;
   mouse_bottom = sy1;
   memset(mouse_pos, 0, sizeof(mouse_pos));
   memset(mouse_rel_pos, 0, sizeof(mouse_rel_pos));
   mouse_x1 = 0; mouse_y1 = -1; mouse_z1 = 0;
   SwapMouseButton(swap_mouse_buttons = SwapMouseButton(false));
                              //init wheel
   wm_mousewheel = 0;

                              //check dimensions
   if(!sx1 || !sy1)
      return false;

#ifdef LOG_MODE
   IG_log.os_log = os_log1;
#endif                        //LOG_MODE

   FARPROC fp_d3d8;
   {
      h_d3d8 = LoadLibraryEx("d3d8.dll", NULL, 0);
      if(!h_d3d8)
         return false;
      fp_d3d8 = GetProcAddress(h_d3d8, "Direct3DCreate8");
      if(!fp_d3d8){
         FreeLibrary(h_d3d8);
         h_d3d8 = NULL;
         return false;
      }
   }

                              //support for wheel mouse
   dword msgMSHWheelSupportActive = RegisterWindowMessage(MSH_WHEELSUPPORT);
   HWND hdlMSHWheel = FindWindow(MSH_WHEELMODULE_CLASS, MSH_WHEELMODULE_TITLE);
   if(hdlMSHWheel && msgMSHWheelSupportActive){
      bool fWheelSupport = SendMessage(hdlMSHWheel, msgMSHWheelSupportActive, 0, 0);
      if(fWheelSupport){
         wm_mousewheel = RegisterWindowMessage(MSH_MOUSEWHEEL);
         mouse_wheel_lines = SendMessage(hdlMSHWheel, RegisterWindowMessage(MSH_SCROLL_LINES), 0, 0);
      }
   }

                              //timer
   use_qpc = false;
#ifdef USE_QUERYPERFORMANCECOUNTER
                              //get frequency
   __int64 qpc_freq;
   use_qpc = QueryPerformanceFrequency((LARGE_INTEGER*)&qpc_freq);
   if(use_qpc || qpc_freq){
      qpc_resolution = 1000.0 / (double)qpc_freq;
   }else
#endif
   {
      timeBeginPeriod(1);
   }
   start_timer = 0;
   start_timer = ReadTimer();
   NullTimer();

   hwnd_dlgs.clear();

   HRESULT hr;

   scrn_sx = sx1, scrn_sy = sy1;

   S_config cfg;

   if(!(flags&IG_FULLSCREEN)){
      ReadConfig((HINSTANCE)hInstance, cfg, scrn_sx, scrn_sy);
   }
                              //register class
   IRegisterClass((HINSTANCE)hInstance);

   create_flags = flags;

   create_bits_per_pixel = bits_per_pixel;

   if(!(flags&IG_INITPOSITION)){
      pos_x = cfg.posx;
      pos_y = cfg.posy;
   }

   if(!InitWindow(pos_x, pos_y))
      goto fail;

   assert(hwnd);

   {                          //set mouse position
      POINT pt;
      GetCursorPos(&pt);
      POINT pt1 = {0, 0};
      ClientToScreen(hwnd, &pt1);
      mouse_pos[0] = pt.x - pt1.x;
      mouse_pos[1] = pt.y - pt1.y;
   }

   h_dinput = LoadLibraryEx("dinput8.dll", NULL, 0);
   if(!h_dinput)
      goto fail;

   {
      FARPROC fp = GetProcAddress(h_dinput, "DirectInput8Create");
      if(!fp)
         goto fail;

                              //create object
      typedef HRESULT WINAPI fp_t(HINSTANCE, dword version, REFIID, LPDIRECTINPUT8*, LPUNKNOWN);
      fp_t *fp1;
      fp1 = (fp_t*)fp;
      hr = (*fp1)((HINSTANCE)hInstance, DIRECTINPUT_VERSION, IID_IDirectInput8A, &lpDI8, NULL);
      CHECK_DI_RESULT("DirectInput8Create", hr);
   }

   {
      typedef IDirect3D8 *WINAPI Direct3DCreate8_t(UINT);
      Direct3DCreate8_t *fp_d3d8_a = (Direct3DCreate8_t*)fp_d3d8;

#ifdef LOG_MODE
      IG_log.Enter("Direct3DCreate8");
#endif                        //LOG_MODE

      lpD3D8 = (*fp_d3d8_a)(D3D_SDK_VERSION);

      if(!lpD3D8){
#ifdef LOG_MODE
         IG_log.Fail();
#endif                        //LOG_MODE
         goto fail;
      }
#ifdef LOG_MODE
      IG_log.Ok();
#endif                        //LOG_MODE
      D3DDEVTYPE dev_type = (D3DDEVTYPE)d3d_device_type;
      if(!dev_type) dev_type = D3DDEVTYPE_HAL;

      D3DPRESENT_PARAMETERS pp;
      InitPresentStruct(adapter_id, dev_type, pp);

      D3DCAPS8 d3d_caps;
      lpD3D8->GetDeviceCaps(adapter_id, dev_type, &d3d_caps);

      dword dev_flags = 0;
      if(d3d_caps.DevCaps&D3DDEVCAPS_HWTRANSFORMANDLIGHT){
         if(create_flags&IG_HW_TNL){
            dev_flags |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
            if(d3d_caps.DevCaps&D3DDEVCAPS_PUREDEVICE){
               dev_flags |= D3DCREATE_PUREDEVICE;
            }
         }else{
            dev_flags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
         }
      }else
         dev_flags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

#ifndef NDEBUG
      //dev_flags |= D3DCREATE_FPU_PRESERVE;
#endif
      hr = lpD3D8->CreateDevice(adapter_id,
         dev_type,
         hwnd, dev_flags, &pp, &lpDev8);

      CHECK_D3D_RESULT("CreateDevice", hr);
      if(FAILED(hr))
         goto fail;
retry_caps:
                              //check if it has appropriate shader caps
      D3DCAPS8 cp;
      hr = lpDev8->GetDeviceCaps(&cp);
      CHECK_D3D_RESULT("GetDeviceCaps", hr);
      if((cp.VertexShaderVersion&0xffff) < 0x0101){
         if(dev_flags&D3DCREATE_SOFTWARE_VERTEXPROCESSING)
            goto fail;

                              //try again in software mode
         lpDev8->Release();
         dev_flags &= ~(D3DCREATE_MIXED_VERTEXPROCESSING | D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE);
         dev_flags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

         hr = lpD3D8->CreateDevice(adapter_id,
            dev_type,
            hwnd, dev_flags, &pp, &lpDev8);
         CHECK_D3D_RESULT("CreateDevice", hr);
         if(FAILED(hr))
            goto fail;
         goto retry_caps;
      }
      backbuffer_count = 0;
      num_swap_buffers = pp.BackBufferCount + 1;

      if((flags&IG_VERBOSE) && log_func1){
         DumpCaps(lpD3D8, lpDev8, log_func1);
      }
   }

   InitRGBConv();
                              //setup globals
   SetViewport(0,0, scrn_sx = sx1, scrn_sy = sy1);

                              //success
   init_ok = true;

   return true;

fail:
   if(lpD3D8){
      lpD3D8->Release();
      lpD3D8 = NULL;
   }
   DestroyWindow();
   UnregisterClass("IGraph_Window", (HINSTANCE)hInstance);

   if(h_d3d8){
      FreeLibrary(h_d3d8);
      h_d3d8 = NULL;
   }

   return false;
}

//----------------------------

static const byte guid_hal[] = {
   0xe0, 0x3d, 0xe6, 0x84,
   0xAA, 0x46, 0xCF, 0x11,
   0x81, 0x6F, 0x00, 0x00,
   0xC0, 0x20, 0x15, 0x6E
}, guid_hal_tnl[] = {
   0x78, 0x9e, 0x04, 0xf5,
   0x61, 0x48, 0xd2, 0x11,
   0xa4, 0x07, 0x00, 0xa0,
   0xc9, 0x06, 0x29, 0xa8
};

//----------------------------

LPIGraph __declspec(dllexport) __stdcall IGraphCreate(LPIG_INIT in){

   if(!in)
      return NULL;

   LPIGraph ig = new IGraph;
   if(!ig)
      return false;

   dword save_fpu = _control87(0, 0);

   if(!ig->Initialize(in->hinst, in->pos_x, in->pos_y, in->size_x, in->size_y,
      in->flags, in->bits_per_pixel,
      D3DADAPTER_DEFAULT,
      in->d3d_device_type,
      in->log_func)){
      ig->Release();
      ig = NULL;
   }else{
      ig->ProcessWinMessages(1);
   }
   _control87(save_fpu, -1);
   return ig;
}

//----------------------------

bool IGraph::AddCallback(LPGRAPHCALLBACK cb, void *context){

                              //check for duplicity of same callbacks
   for(int i=cb_proc_list.size(); i--; ){
      if(cb_proc_list[i].cb_proc==cb && cb_proc_list[i].context==context)
         break;
   }
   if(i==-1){
      cb_proc_list.push_back(S_callback());
      cb_proc_list.back().cb_proc = cb;
      cb_proc_list.back().context = context;
      return true;
   }
   THROW("callback function already in internal list");
   return false;
}

//----------------------------

bool IGraph::RemoveCallback(LPGRAPHCALLBACK cb, void *context){

   for(int i=cb_proc_list.size(); i--; ){
      if(cb_proc_list[i].cb_proc==cb && cb_proc_list[i].context==context)
         break;
   }
   if(i!=-1){
      cb_proc_list.erase(&cb_proc_list[i]);
      return true;
   }
   THROW("callback function not in internal list");
   return false;
}

//----------------------------

bool IGraph::SetViewport(dword l, dword t, dword sx, dword sy){

   if(l<0 || t<0 || l+sx>scrn_sx || t+sy>scrn_sy || sx<0 || sy<0){
      THROW("invalid viewport");
      return false;
   }
   vport[0] = l;
   vport[1] = t;
   vport[2] = l + sx;
   vport[3] = t + sy;
   return true;
}

//----------------------------

void IGraph::FlipToGDI(){

   if(create_flags&IG_FULLSCREEN){
      while(backbuffer_count){
         UpdateScreen(0);
      }
   }
}

//----------------------------

bool IGraph::SetBrightness(float r, float g, float b){

   D3DGAMMARAMP gr;
                              //create linear ramps
   int i;
   word *wp = gr.red;
   for(i=0; i<3; i++, wp += 256){
      float f = i==0 ? r : i==1 ? g : b;
      for(int j=0; j<256; j++)
         wp[j] = (word) Min(65535, Max(0, (int)((j<<8) * f)));
   }

   lpDev8->SetGammaRamp(D3DSGR_CALIBRATE, &gr);
   return true;
}

//----------------------------

bool IGraph::UpdateScreen2(dword flags, void *dest_hwnd){

   HRESULT hr;
   if(!dest_hwnd){
      if(create_flags&IG_FULLSCREEN){
         hr = lpDev8->Present(NULL, NULL, NULL, NULL);
         ++backbuffer_count %= num_swap_buffers;
      }else{
         hr = lpDev8->Present((const RECT*)vport, (const RECT*)vport, NULL, NULL);
      }
   }else{
      hr = lpDev8->Present(
                              //source rect
         NULL,
                              //dest rect
         NULL,
                              //dest window
         (HWND)dest_hwnd,
                              //dirty regions
         NULL);
   }

   switch(hr){
   case D3DERR_DEVICELOST:
      {
                              //check if we may reset
         hr = lpDev8->TestCooperativeLevel();
         if(hr == D3DERR_DEVICENOTRESET){
            UpdateParams(0, 0, 0, 0, 0, true);
         }
      }
      break;

   case D3DERR_INVALIDCALL:
                              //ignore (DirectX gives this when minimizing window)
      break;

   default:
      CHECK_D3D_RESULT("Present", hr);
   }
   return SUCCEEDED(hr);
}

//----------------------------

void IGraph::GetDesktopResolution(int &x, int &y) const{

   x = desktop_sx;
   y = desktop_sy;
}

//----------------------------

dword IGraph::NumDisplayModes() const{

   D3DDEVICE_CREATION_PARAMETERS cp;
   HRESULT hr;
   hr = lpDev8->GetCreationParameters(&cp);
   CHECK_D3D_RESULT("GetCreationParameters", hr);
   return lpD3D8->GetAdapterModeCount(cp.AdapterOrdinal);
}

//----------------------------

bool IGraph::GetDisplayMode(int id, int *sx, int *sy, int *bpp) const{

   D3DDEVICE_CREATION_PARAMETERS cp;
   HRESULT hr;
#ifndef NDEBUG
   hr =
#endif
      lpDev8->GetCreationParameters(&cp);
   CHECK_D3D_RESULT("GetCreationParameters", hr);

   D3DDISPLAYMODE dm;
   S_pixelformat pf;

   hr = lpD3D8->EnumAdapterModes(cp.AdapterOrdinal, id, &dm);
   CHECK_D3D_RESULT("EnumAdapterModes", hr);
   if(FAILED(hr))
      return false;
   ConvertD3DpfToI3Dpf(dm.Format, pf);

   if(sx) *sx = dm.Width;
   if(sy) *sy = dm.Height;
   if(bpp) *bpp = pf.bytes_per_pixel;

   return true;
}

//----------------------------

bool IGraph::Rectangle(int l, int t, dword sx1, dword sy1, dword col){

   int r = l + sx1, b = t+sy1;
                                       //clip input coords
   if(l<vport[0]) l = vport[0];
   if(t<vport[1]) t = vport[1];
   if(r>vport[2]) r = vport[2];
   if(b>vport[3]) b = vport[3];
                                       //check in-window
   if(r <= 0)
      return true;
   if(b <= 0)
      return true;

                              //clear to color
   D3DRECT rc = {l, t, r, b};
   lpDev8->Clear(1, &rc, D3DCLEAR_TARGET, col, 0.0f, 0);

   return true;
}

//----------------------------
                              //keyboard
void IGraph::InitKeyboard(){

   if(!IsKbInit()){
      if(active) GrabKeys(this, true);
      kb_init = true;
   }
}

//----------------------------

void IGraph::CloseKeyboard(){

   if(IsKbInit()){
      if(active) GrabKeys(this, false);
      kb_init = false;
      key = K_NOKEY;
      game_keys = 0;
      shiftkeys = 0;
   }
}

//----------------------------

void IGraph::DefineKeys(const char *sc, dword num){

   char *cp = key_scodes;
   num_key_scodes = (byte)num;
   while(num--){
      (*cp) = (*sc);
      ++cp; ++sc;
   }
   game_keys = 0;
}

//----------------------------

char IGraph::ReadKey(bool pwm){

   if(pwm)
      ProcessWinMessages();
   char c = key;
   key = K_NOKEY;
   return c;
}

//----------------------------

static const byte spec_trans[][2]={
   ';', ':',
   '=', '+',
   ',', '<',
   '-', '_',
   '.', '>',
   '/', '?',
   '`', '~',
   '\'','"'
};

static const byte spec_codes[] = {
   0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xde
};

static const byte spec_codes1[] = {
   '1', '2', '3', '4', '5', '6', '7', '8', '9', '0',
   '\\', '[', ']'
};
static const byte spec_trans1[]={
   '!', '@', '#', '$', '%', '^', '&', '*', '(', ')',
   '|', '{', '}'
};

int IGraph::GetAscii(char k1, bool shift) const{

   byte k = k1;
   byte *b;
   if(k>=0xba && k<=0xde){
                                       //special conversion
      b = (byte*)memchr(spec_codes, k, sizeof(spec_codes));
      if(b) return spec_trans[b-spec_codes][shift];
   }
   if(k==0x20 || k<0x10)
      return k;
   if(k<0x30 || k1==K_NOKEY)
      return 0;
   if(k>=0x5b && k<=0xA0)
      return 0;
   k &= 0x7f;
                                       //lower case
   if(k>='A' && k<='Z')
      k += -'A' + 'a';
   if(k>='a' && k<='z')
      return shift ? (k+'A'-'a') : k;
   if(!shift)
      return k;
   b = (byte*)memchr(spec_codes1, k, sizeof(spec_codes1));
   if(b)
      return spec_trans1[b-spec_codes1];
   return k;
}

//----------------------------
//----------------------------

static dword CALLBACK NotifyThreadMouse(void *ig1){

   LPIGraph IGraph = (LPIGraph)ig1;
   HANDLE h_event = IGraph->h_mouse_event;

   while(true){
      WaitForSingleObject(h_event, INFINITE);
      IGraph->event_signalled = true;
   }
}

//----------------------------
                              //mouse
bool IGraph::MouseInit(dword mode){

   if(mouse_init)
      return false;

   bool ret = false;
   if(create_flags&IG_DEBUGMOUSE){
                              //debug mode, don't init DInput object
      mouse_init = true;
                           //acquire
      MouseAcquire(true);

      ret = true;

      SetCursor(NULL);
      goto done;
   }

   HRESULT hr;

                              //create device
   hr = lpDI8->CreateDevice(GUID_SysMouse, &lpDIMouse8, NULL);
   if(FAILED(hr)){
      CHECK_DI_RESULT("CreateDevice", hr);
      goto done;
   }
                              //set data format
   hr = lpDIMouse8->SetDataFormat(&c_dfDIMouse2);
   if(FAILED(hr)){
      CHECK_DI_RESULT("SetDataFormat", hr);
      goto done;
   }

                              //set coop level
   hr = lpDIMouse8->SetCooperativeLevel(hwnd,
      DISCL_FOREGROUND |
      DISCL_EXCLUSIVE);

   if(FAILED(hr)){
      CHECK_DI_RESULT("SetCooperativeLevel", hr);
      goto done;
   }

   DIPROPDWORD dipdw;
   dipdw.diph.dwSize = sizeof(DIPROPDWORD);
   dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
   dipdw.diph.dwObj = 0;
   dipdw.diph.dwHow = DIPH_DEVICE;
   dipdw.dwData = DIPROPAXISMODE_REL;

   hr = lpDIMouse8->SetProperty(DIPROP_AXISMODE, &dipdw.diph);
   if(FAILED(hr)){
      CHECK_DI_RESULT("SetProperty", hr);
      goto done;
   }
                           //get mouse threshold and speed
   SystemParametersInfo(SPI_GETMOUSE, 0, ms_speed, 0);

#ifdef MOUSE_USE_EVENT
                              //setup mouse event
   h_mouse_event = CreateEvent(NULL, false, false, NULL);
   if(h_mouse_event){
      hr = lpDIMouse8->SetEventNotification(h_mouse_event);
      CHECK_DI_RESULT("SetEventNotification", hr);
      if(FAILED(hr)){
         CloseHandle(h_mouse_event); h_mouse_event=NULL;
      }else{
         dword d;
         h_mouse_thread = CreateThread(NULL, 1024, NotifyThreadMouse,
            this, 0, &d);
         if(!h_mouse_thread){
            CloseHandle(h_mouse_event);
            hr = lpDIMouse8->SetEventNotification(h_mouse_event = NULL);
            CHECK_DI_RESULT("SetEventNotification", hr);
         }else
            event_signalled = false;
      }
   }
#endif

   mouse_init = true;
                           //acquire
   MouseAcquire(true);

   ret = true;
done:
   return ret;
}

//----------------------------

void IGraph::MouseClose(){

   HRESULT hr;
   if(mouse_init){
      {
         MouseAcquire(false);
         if(create_flags&IG_DEBUGMOUSE){
         }else{

            if(h_mouse_event){
               TerminateThread(h_mouse_thread, 0);
               CloseHandle(h_mouse_thread); h_mouse_thread=NULL;
               hr = lpDIMouse8->SetEventNotification(NULL);
               CHECK_DI_RESULT("SetEventNotification", hr);
               CloseHandle(h_mouse_event); h_mouse_event=NULL;
            }

            assert(lpDIMouse8);
            lpDIMouse8->Release(); lpDIMouse8 = NULL;
         }

         POINT pt;
         pt.x = mouse_pos[0];
         pt.y = mouse_pos[1];
         ClientToScreen(hwnd, &pt);
         SetCursorPos(pt.x, pt.y);

         mouse_butt1 = 0;
         mouse_butt = 0;
      }
      mouse_init = false;
   }
}

//----------------------------

void IGraph::MouseSetup(int x, int y, int l, int t, int r, int b){

   mouse_left=l, mouse_top=t, mouse_right=r, mouse_bottom=b;
   bool get=0;
   POINT pt;
   if(mouse_pos[0]==-1){
      GetCursorPos(&pt);
      mouse_pos[0] = pt.x;
      get=1;
   }else
      mouse_pos[0] = Max(mouse_left, Min(mouse_right, x));
   if(mouse_pos[1] ==-1){
      if(!get) GetCursorPos(&pt);
      mouse_pos[1] = pt.y;
   }else mouse_pos[1] = Max(mouse_top, Min(mouse_bottom, y));
   mouse_butt = 0;
}

//----------------------------

dword IGraph::MouseUpdate(){

   int i;
   int nm[3];

   if(!h_mouse_event)
      mouse_butt = 0;

   if(!mouse_init){
      mouse_butt = mouse_butt1;
      if(mouse_y1 < 0){
         for(i=0; i<3; i++) mouse_rel_pos[i] = 0;
         mouse_y1 = -1;
         return 0;
      }
                              //get standard cursor
      nm[0] = mouse_x1;
      nm[1] = mouse_y1;
      nm[2] = mouse_pos[2] + mouse_z1; mouse_z1 = 0;
      if(create_flags&IG_FULLSCREEN){
         if(hmenu_main){
            nm[1] += GetSystemMetrics(SM_CYMENU);
         }
      }
      mouse_y1 = -1;
   }else{
      if(!active)
         return false;

      if(create_flags&IG_DEBUGMOUSE){
         POINT pt = { scrn_sx/2, scrn_sy/2};
         ClientToScreen(hwnd, &pt);
         int mx = pt.x;
         int my = pt.y;

         GetCursorPos(&pt);
         SetCursorPos(mx, my);

         mouse_rel_pos[0] = pt.x - mx;
         mouse_rel_pos[1] = pt.y - my;
         mouse_rel_pos[2] = mouse_z1;
         mouse_z1 = 0;

         mouse_butt = mouse_butt1;

         goto accel_ok;
      }

      if(h_mouse_event){
         if(!event_signalled){
            memset(mouse_rel_pos, 0, sizeof(mouse_rel_pos));
            return 0;
         }
         event_signalled = false;
         mouse_butt = 0;
      }

      DIMOUSESTATE2 ms;
      HRESULT hr;
re_read:
      assert(lpDIMouse8);
      hr = lpDIMouse8->GetDeviceState(sizeof(ms), &ms);
      switch(hr){
      case DI_OK: break;
      case DIERR_NOTACQUIRED:
      case DIERR_INPUTLOST:
         hr = lpDIMouse8->Acquire();
         if(hr!=DI_OK)
            return 0;
         goto re_read;
      default:
         CHECK_DI_RESULT("GetDeviceState", hr);
         return false;
      }
      mouse_rel_pos[0] = ms.lX;
      mouse_rel_pos[1] = ms.lY;
      mouse_rel_pos[2] = ms.lZ/120 + mouse_z1; mouse_z1 = 0;

      if(swap_mouse_buttons)
         swap(ms.rgbButtons[0], ms.rgbButtons[1]);

      if(ms.rgbButtons[0]&0x80) mouse_butt |= 1;
      if(ms.rgbButtons[1]&0x80) mouse_butt |= 2;
      if(ms.rgbButtons[2]&0x80) mouse_butt |= 4;
      if(ms.rgbButtons[3]&0x80) mouse_butt |= 8;

                              //acceleration/speed
      if(ms_speed[2] && abs(mouse_rel_pos[0])>ms_speed[0]){
         if(ms_speed[2]==2 && abs(mouse_rel_pos[0])>ms_speed[1])
            mouse_rel_pos[0] *= 4;
         else mouse_rel_pos[0] *= 2;
      }
      if(ms_speed[2] && abs(mouse_rel_pos[1])>ms_speed[0]){
         if(ms_speed[2]==2 && abs(mouse_rel_pos[1])>ms_speed[1])
            mouse_rel_pos[1] *= 4;
         else mouse_rel_pos[1] *= 2;
      }
   accel_ok:
      for(i=0; i<3; i++)
         nm[i] = mouse_pos[i] + mouse_rel_pos[i];
   }
   for(i=0; i<3; i++)
      mouse_rel_pos[i] = nm[i] - mouse_pos[i];

   nm[0] = Max(mouse_left, Min(mouse_right-1, nm[0]));
   nm[1] = Max(mouse_top, Min(mouse_bottom-1, nm[1]));

   for(i=0; i<3; i++) mouse_pos[i] = nm[i];

   return 0;
}

//----------------------------

void IGraph::SetMousePos(int x, int y){

   mouse_pos[0] = Max(mouse_left, Min(mouse_right, x));
   mouse_pos[1] = Max(mouse_top, Min(mouse_bottom, y));

   if(!mouse_init){
      POINT pt;
      pt.x = mouse_pos[0];
      pt.y = mouse_pos[1];
      ClientToScreen(hwnd, &pt);
      SetCursorPos(pt.x, pt.y);
   }
}

//----------------------------

void IGraph::MouseAcquire(bool b){

   if(!mouse_init){
      return;
   }
   if(create_flags&IG_DEBUGMOUSE){
                              //can't change mouse position if not active window
      if(active){
         POINT pt = { scrn_sx/2, scrn_sy/2};
         ClientToScreen(hwnd, &pt);

         SetCursorPos(pt.x, pt.y);
      }
      return;
   }

   assert(lpDIMouse8);

   HRESULT hr;
   if(b){
      if(!acq_count++){
         hr = lpDIMouse8->Acquire();
         //CHECK_DI_RESULT("Mouse::Acquire", hr);
      }
   }else
   if(!--acq_count){
      hr = lpDIMouse8->Unacquire();
      //CHECK_DI_RESULT("Mouse::Unacquire", hr);
   }
}

//----------------------------
//----------------------------
                              //timer
void IGraph::NullTimer(){

   last_timer = ReadTimer();
}

//----------------------------

dword IGraph::GetTimer(dword min, dword max){

   if(!(!min || !max || min<=max)){
      THROW("min must be <= max");
   }

   dword time = ReadTimer() - last_timer;
   int delta = (int)min - (int)time;
   if(delta >= 4)
      Sleep(delta-3);

   while(time < min){
      Sleep(0);
      time = ReadTimer() - last_timer;
   }

   last_timer += time;

   return max ? Min(time, max) : time;
}

//----------------------------

dword IGraph::ReadTimer() const{

   dword cutt_t;
#ifdef USE_QUERYPERFORMANCECOUNTER
   if(use_qpc){
      __int64 t;
      QueryPerformanceCounter((LARGE_INTEGER*)&t);
      cutt_t = FloatToInt(qpc_resolution * (double)t);
   }else
#endif
      cutt_t = timeGetTime();
   return cutt_t - start_timer;
}

//----------------------------

void IGraph::ProcessWinMessages(dword flags){

   int i;

   MSG msg;

   if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)){
      do{
         if(msg.message==WM_QUIT) exit(1);
         if(i=hwnd_dlgs.size()){
            while(i--) if(IsDialogMessage(hwnd_dlgs[i], &msg)) break;
            if(i!=-1) continue;
         }
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE));
   }
   if(!active && (create_flags&IG_SUSPENDINACTIVE) && !(create_flags&IG_HIDDEN) && !(flags&1)){
      if(i=hwnd_dlgs.size()){
         HWND hwnd_fore = GetForegroundWindow();
         while(i--) if(hwnd_dlgs[i]==hwnd_fore) break;
         if(i!=-1) return;
      }
      SetWindowText(hwnd, C_fstring("%s [paused]", (const char*)appName));
      sleep_loop = true;
      while(!active && sleep_loop){
         if(GetMessage(&msg, NULL, 0, 0)){
            i=hwnd_dlgs.size();
            if(i){
               while(i--) if(IsDialogMessage(hwnd_dlgs[i], &msg)) break;
               i=(i!=-1);
            }
            if(!i){
               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         }
         i=hwnd_dlgs.size();
         if(i){
            HWND hwnd_fore=GetForegroundWindow();
            while(i--) if(hwnd_dlgs[i]==hwnd_fore) break;
            if(i!=-1) break;
         }
      }
      sleep_loop = false;
      SetWindowText(hwnd, appName);
   }
}

//----------------------------
                              //info

void IGraph::SetAppName(const char *cp){

   appName = cp;
   SetWindowText(hwnd, cp);
}

//----------------------------

void IGraph::SetMenu(void *menu){

   hmenu_main = (HMENU)menu;
   ::SetMenu(hwnd, hmenu_main);
}

//----------------------------

int IGraph::GetMousePos(int axis, bool rel) const{

   if(axis<0 || axis>=3){
      THROW("IGraph::GetMousePos: invalid axis");
      return 0;
   }
   return !rel ? mouse_pos[axis] : mouse_rel_pos[axis];
}

//----------------------------

bool IGraph::AddDlgHWND(void *hwnd){

   if(hwnd){
      hwnd_dlgs.push_back((HWND)hwnd);
      return true;
   }
   return false;
}

//----------------------------

bool IGraph::DeleteDlgHWND(void *hwnd){

   for(int i = hwnd_dlgs.size(); i--; )
   if(hwnd_dlgs[i]==hwnd){
      hwnd_dlgs[i] = hwnd_dlgs.back();
      hwnd_dlgs.pop_back();
      return true;
   }
   return false;
}

//----------------------------

bool IGraph::CreateRGBConv(C_rgb_conversion **ret) const{

   *ret = Create_rgb_conv();
   return true;
}

//----------------------------

LPC_rgb_conversion IGraph::GetRGBConv1(const S_pixelformat &pf, const dword *pal_rgba){

                              //look if we've this already initialized
   if(!pal_rgba)
   for(int i=rgb_conv_list.size(); i--; ){
      LPC_rgb_conversion pc = rgb_conv_list[i];
      if(!memcmp(pc->GetPixelFormat(), &pf, sizeof(S_pixelformat))){
         pc->AddRef();
         return pc;
      }
   }
                              //create and initialize new convertor
   LPC_rgb_conversion rgb_conv;
   CreateRGBConv(&rgb_conv);
   rgb_conv->Init(pf, pal_rgba);
                              //put into cache
   if(!pal_rgba)
      rgb_conv_list.push_back(rgb_conv);

   return rgb_conv;
}

//----------------------------

LPIImage IGraph::CreateBackbufferCopy(bool lockable){

   IDirect3DSurface8 *bbuf;
   LPIImage img = NULL;
   HRESULT hr = lpDev8->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &bbuf);
   if(SUCCEEDED(hr)){

      CreateImage(&img);
      const S_pixelformat *pf = GetPixelFormat();
      bool b = img->Open(NULL, IMGOPEN_EMPTY, Scrn_sx(), Scrn_sy(), pf);
      if(b){
         if(!lockable){
                              //try to create vidmem surface, if possible
            D3DDISPLAYMODE dm;
            lpD3D8->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &dm);

            IDirect3DSurface8 *bb;
            HRESULT hr = lpDev8->CreateRenderTarget(scrn_sx, scrn_sy, dm.Format, D3DMULTISAMPLE_NONE, false, &bb);
            if(SUCCEEDED(hr)){
               img->lpImg8 = bb;
               bb->Release();
            }
         }
         IDirect3DSurface8 *s1 = img->GetDirecXSurface();
         hr = lpDev8->CopyRects(bbuf, NULL, 0, s1, NULL);
         if(FAILED(hr)){
            img->Release();
            img = NULL;
         }
      }else{
         img->Release();
         img = NULL;
      }
      bbuf->Release();
   }
   return img;
}

//----------------------------

bool IGraph::UpdateParams(dword sx, dword sy, dword bits_per_pixel, dword flg, dword flg_mask, bool force_recreate){

   int save_sx = scrn_sx, save_sy = scrn_sy, save_bpp = create_bits_per_pixel;
                              //mask only supported features
   flg_mask &= (IG_FULLSCREEN | //IG_NOSYSTEMMENU | IG_NOMINIMIZE |
      IG_SUSPENDINACTIVE | IG_LOADMENU 
      | IG_HIDDEN);

   bool recreate = force_recreate;
   if(sx || sy){
      if(!sx || !sy)
         return false;
      if(sx!=scrn_sx || sy!=scrn_sy){
         scrn_sx = sx;
         scrn_sy = sy;
         recreate = true;
         SetViewport(0, 0, scrn_sx, scrn_sy);
      }
   }

   dword flags_change = (create_flags^flg)&flg_mask;
   if(flags_change){
      if(flags_change&IG_FULLSCREEN){
         bool ok = true;
                              //toggle fullscreen
                              //check if switch possible
         if(create_flags&IG_FULLSCREEN){
            D3DCAPS8 d3d_caps;
            lpDev8->GetDeviceCaps(&d3d_caps);
            if(!(d3d_caps.Caps2&D3DCAPS2_CANRENDERWINDOWED))
               ok = false;
         }
         if(!ok){
            flg ^= IG_FULLSCREEN;
            flags_change &= ~IG_FULLSCREEN;
         }else
            recreate = true;
      }
      if((flags_change&IG_SUSPENDINACTIVE) && (create_flags&IG_SUSPENDINACTIVE)){
                              //wake up sleeper 
         sleep_loop = false;
      }

      if(!recreate){
         dword wstyle = GetWindowLong(hwnd, GWL_STYLE);
         create_flags ^= flags_change;
                              //update window flags directly
         bool wl_change = false;

         if(flags_change&IG_NOSYSTEMMENU){
            wstyle &= ~WS_SYSMENU;
            if(!(create_flags&IG_NOSYSTEMMENU)) wstyle |= WS_SYSMENU;
            wl_change = true;
         }
         if(flags_change&IG_NOMINIMIZE){
            wstyle &= ~WS_MINIMIZEBOX;
            if(!(create_flags&IG_NOMINIMIZE)) wstyle |= WS_MINIMIZEBOX;
            wl_change = true;
         }
         if(flags_change&IG_LOADMENU){
            if(hmenu_main){
               ::SetMenu(hwnd, NULL);
               DestroyMenu(hmenu_main);
               hmenu_main = NULL;
            }
            if((create_flags&IG_LOADMENU) && !(create_flags&IG_FULLSCREEN)){
               hmenu_main = LoadMenu((HINSTANCE)hInstance, "MENU_MAIN");
               if(!hmenu_main)
                  hmenu_main = CreateMenu();
               if(hmenu_main)
                  ::SetMenu(hwnd, hmenu_main);
            }
                              //adjust window size
            RECT rc;
            SETRECT(&rc, 0, 0, scrn_sx, scrn_sy);
            AdjustWindowRect(&rc, wstyle, (bool)hmenu_main);

            dword sx = rc.right-rc.left;
            dword sy = rc.bottom-rc.top;
            SetWindowPos(hwnd, 0, 0, 0, sx, sy, SWP_NOACTIVATE | SWP_NOCOPYBITS | SWP_NOMOVE | SWP_NOZORDER);

            {
               int sy = GetSystemMetrics(SM_CYMENU);
               mouse_pos[1] -= (create_flags&IG_LOADMENU) ? sy : +sy;
               mouse_y1 = -2; //ignore 1st WM_MOUSEMOVE message
            }
         }

         if(flags_change&IG_HIDDEN){
                              //show/hide window
            wstyle |= WS_VISIBLE;
            if(create_flags&IG_HIDDEN)
               wstyle &= ~WS_VISIBLE;
            wl_change = true;
         }

         if(wl_change){
            SetWindowLong(hwnd, GWL_STYLE, wstyle);

            if(flags_change&IG_HIDDEN){
               UpdateWindow(hwnd);
               SetFocus(hwnd);
            }
         }
      }
   }
   if(bits_per_pixel){
      if(create_bits_per_pixel!=bits_per_pixel){
         create_bits_per_pixel = (byte)bits_per_pixel;
         if((create_flags^flags_change)&IG_FULLSCREEN)
            recreate = true;
      }
   }

   if(recreate){
      S_config cfg;
      if(!(ig_flags&IGGLAGS_IN_RESET)){
         for(int i=cb_proc_list.size(); i--; )
            (*cb_proc_list[i].cb_proc)(CM_RECREATE, flg, 0, cb_proc_list[i].context);
         init_ok = false;
                              //release window and DirectDraw components
         if(!(create_flags&IG_FULLSCREEN)){
                              //write config
            {
               RECT rc;
#ifndef NDEBUG
               bool b = 
#endif
                  GetWindowRect(hwnd, &rc);
               assert(b);
               WriteConfig((HINSTANCE)hInstance, rc);
            }
         }
         if(!(flags_change&IG_FULLSCREEN) && !(create_flags&IG_FULLSCREEN)){
            ShowWindow(hwnd, SW_HIDE);
         }

#ifdef ENABLE_PALETIZED_DESKTOP
        if(lpDDPal){
            lpDDPal->Release();
            lpDDPal = NULL;
         }
#endif

         create_flags ^= flags_change;
      }

      HRESULT hr;

      D3DDEVICE_CREATION_PARAMETERS cp;
      lpDev8->GetCreationParameters(&cp);

      D3DPRESENT_PARAMETERS pp;

      InitPresentStruct(cp.AdapterOrdinal, cp.DeviceType, pp);
      hr = lpDev8->Reset(&pp);
      if(FAILED(hr)){
         switch(hr){
         case D3DERR_INVALIDCALL:
            {
               abort();
            }
            break;
         case D3DERR_OUTOFVIDEOMEMORY:
            {
               scrn_sx = save_sx;
               scrn_sy = save_sy;
               create_bits_per_pixel = save_bpp;
               SetViewport(0, 0, scrn_sx, scrn_sy);
               InitPresentStruct(cp.AdapterOrdinal, cp.DeviceType, pp);
               hr = lpDev8->Reset(&pp);
            }
            break;
         }
         if(FAILED(hr)){
            ig_flags |= IGGLAGS_IN_RESET;
            return false;
         }
      }
      backbuffer_count = 0;
      num_swap_buffers = pp.BackBufferCount + 1;

      ig_flags &= ~IGGLAGS_IN_RESET;

      if(!(create_flags&IG_FULLSCREEN)){
                              //read config
         ReadConfig((HINSTANCE)hInstance, cfg, scrn_sx, scrn_sy);
      }

      if(!InitWindow(cfg.posx, cfg.posy)){
         D3D_Fatal("InitWindow", hr);
      }

      InitRGBConv();
      ClearViewport();

      init_ok = true;
      MouseSetup(mouse_pos[0], mouse_pos[1], 0, 0, scrn_sx, scrn_sy);
      {
         for(int i=cb_proc_list.size(); i--; )
            (*cb_proc_list[i].cb_proc)(CM_RECREATE, create_flags, 1, cb_proc_list[i].context);
         for(i=cb_proc_list.size(); i--; )
            (*cb_proc_list[i].cb_proc)(CM_RECREATE, create_flags, 2, cb_proc_list[i].context);
      }
   }
   return true;
}

//----------------------------
//----------------------------
