#ifndef __IGRAPHI_H
#define __IGRAPHI_H

#define NO_STARTUP
#define IG_INTERNAL
#include <igraph2.h>

#define USE_DDRAW

//#define ENABLE_PALETIZED_DESKTOP

#include <d3d8.h>
#define DIRECTINPUT_VERSION 0x0800
#include <dinput.h>
#include <conv_rgb.hpp>
#include <C_string.hpp>
#include <vector>
#include <smartptr.h>
#include <c_cache.h>
#include <vector>

using namespace std;

//----------------------------

#ifdef _DEBUG

void D3D_Fatal(const char *text, dword hr);
#define THROW(what) throw(exception(what))
#define CHECK_D3D_RESULT(text, hr) if(FAILED(hr)) D3D_Fatal(text, hr);

#else

#define D3D_Fatal(t, hr)
#define THROW(what)
#define CHECK_D3D_RESULT(text, hr)

#endif

//----------------------------

#define FLOAT_BITMASK(f) (*(int*)&f)

#if defined _MSC_VER && 1

inline int FloatToInt(float f){
   __asm{
      fld f
      fistp f 
   }
   return *(int*)&f;
}
#else
inline int FloatToInt(float f){
   return (int)f;
}
#endif

inline int FloatAsInt(float f){
   return *(dword*)&f;
}

inline float IntAsFloat(int i){
   return (*(float*)&i);
}

//----------------------------

#ifndef NDEBUG

//#define LOG_MODE              //log all hardware calls
#ifdef LOG_MODE

#include <fstream>

class C_log{
   std::ostream *os_log;
   int indent;
   void MakeIndent(){
      for(int i=0; i<indent*2; i++) (*os_log)<<" ";
   }
   friend class IGraph;
public:
   C_log(): os_log(NULL) {}
   void Enter(const char *cp){
      MakeIndent();
      (*os_log)<<cp <<std::endl;
      ++indent;
   };
   void Leave(const char *cp){
      MakeIndent();
      (*os_log)<<cp <<std::endl;
      --indent;
   }
   inline void Ok(){ Leave("OK"); }
   inline void Fail(){ Leave("Fail"); }
   inline void Done(){ Leave("Done"); }
   C_log &operator <<(const char *cp){
      MakeIndent();
      (*os_log)<<cp <<std::endl;
      return *this;
   }
};
extern C_log IG_log;

#endif                        //LOG_MODE

#endif                        //NDEBUG

//----------------------------

class C_unk_base{
public:
   I2DMETHOD_(dword,AddRef)() PURE;
   I2DMETHOD_(dword,Release)() PURE;
};

template<class T>
class C_unk: public C_unk_base{
   unsigned long ref;
protected:
   C_unk(): ref(1) {}

   I2DMETHOD_(dword,AddRef)(){ return ++ref; }
   I2DMETHOD_(dword,Release)(){
      unsigned long new_ref=--ref;
      if(!new_ref) delete (T*)this;
      return new_ref;
   }
};

//----------------------------
                              //located in DLL
extern"C"{
LPIGraph __declspec(dllexport) __stdcall IGraphCreate(LPIG_INIT);
}

//----------------------------
                              //graphics interface class
class IGraph: public C_unk<IGraph>{

#define IGGLAGS_IN_RESET   1  //currently reseting device

   dword ig_flags;
   bool want_close_app;

   HINSTANCE h_dinput;        //dinput.dll

   void *hInstance;
   E_SPLITMODE splitmode;

                              //palettes:
#ifdef ENABLE_PALETIZED_DESKTOP
   byte *currpal;
   LPDIRECTDRAWPALETTE lpDDPal;
   word firstpal, numpals;    //palette range
   byte *save_pal;
#endif
                              //keyboard
   byte shiftkeys;
   char key;                  //currently pressed key
   dword game_keys;           //game keyboard
   char key_scodes[32];
   byte num_key_scodes;
   bool kb_init;              //keyboard functions use
   void *kb_hook;
   bool keyboard_map[256];    //true when key is down
   friend bool GrabKeys(IGraph*, bool on);
   friend LRESULT CALLBACK GKeyboardProc(int code, WPARAM wParam, LPARAM lParam);

                              //mouse
   IDirectInput8 *lpDI8;
   IDirectInputDevice8 *lpDIMouse8;
   IDirectInputDevice8 *lpDIKeyboard;
   HANDLE h_mouse_event, h_mouse_thread;
   bool event_signalled;
   bool mouse_init;
   int mouse_left, mouse_top, mouse_right, mouse_bottom;
   bool swap_mouse_buttons;
   bool init_ok;
   bool mouse_debug_mode;
   int acq_count;
   int ms_speed[3];
   int mouse_x1, mouse_y1;    //help
   int mouse_z1;
   byte mouse_butt1;
   friend void MouseAcquire(IGraph*, bool b);
   int mouse_pos[3];
   int mouse_rel_pos[3];
   byte mouse_butt;
                              //wheel mouse:
   dword wm_mousewheel;       //0 = no wheel present
   dword mouse_wheel_lines;   //# of lines to scroll on wheel

                              //timer
   double qpc_resolution;
   int start_timer, last_timer;
   bool use_qpc;

   static void DoDrawsClip(LPIGraph, dword flags);
   static void DoDrawsNoClip(LPIGraph, dword flags);

   HINSTANCE h_d3d8;          //d3d8.dll
   IDirect3D8 *lpD3D8;
   IDirect3DDevice8 *lpDev8;

                              //keep internal list of convertors,
                              // to avoid duplication of their initialization
   std::vector<C_smart_ptr<C_rgb_conversion> > rgb_conv_list;
public:
   LPC_rgb_conversion GetRGBConv1(const S_pixelformat&, const dword *pal_rgba = NULL);
private:
                              //counter where is GDI window
                              // used for FlipToGDI
                              // this is 100% reliable, but works on most adapters
   int backbuffer_count;      
   int num_swap_buffers;

   HMENU hmenu_main;
                              //window's style - different in windowed
                              //and fullscreen mode
   dword create_flags;

   byte create_bits_per_pixel;
   C_string appName;          //name of window
   IG_INIT::t_log_func *log_func;

                              //registered dialog windows
   std::vector<HWND> hwnd_dlgs;
   struct S_callback{
      LPGRAPHCALLBACK cb_proc;
      void *context;
   };
   std::vector<S_callback> cb_proc_list;

   friend long FAR PASCAL GWindowProc(HWND, UINT, WPARAM, LPARAM);
   friend dword CALLBACK NotifyThreadMouse(void *igraph);

   friend LPIGraph __declspec(dllexport) __stdcall IGraphCreate(LPIG_INIT);
   bool Initialize(void *hinst, dword pos_x, dword pos_y, dword size_x, dword size_y,
      dword flags, int bits_per_pixel,
      int adapter_id, dword d3d_device_type,
      IG_INIT::t_log_func* = NULL);
   HRESULT CreateFrontBack();

   void InitPresentStruct(int adapter_id, D3DDEVTYPE, D3DPRESENT_PARAMETERS&) const;
public:
   IGraph();
   ~IGraph();

   bool InitWindow(int posx, int posy);
   void DestroyWindow();
   HRESULT InitRGBConv();

                              //windows
   HWND hwnd;
   bool active;               //application active
   bool sleep_loop;           //inside of sleeping loop
   dword scrn_sx, scrn_sy;    //size of screen
   dword desktop_sx, desktop_sy; //desktop resolution

                              //viewport
   int vport[4];              //[left, top, right, bottom]

   C_smart_ptr<C_rgb_conversion> rgb_conv;

#ifdef ENABLE_PALETIZED_DESKTOP
   void SetPalRange(int first, int number_of);
   void SavePalRange(int first, int number_of);
   void RestorePalRange();
#endif                        //ENABLE_PALETIZED_DESKTOP
   void MouseAcquire(bool);

public:
   I2DMETHOD_(dword,AddRef)(){ return C_unk<IGraph>::AddRef(); }
   I2DMETHOD_(dword,Release)(){ return C_unk<IGraph>::Release(); }
                              //interfaces:
   I2DMETHOD_(bool,AddCallback)(LPGRAPHCALLBACK, void *context);
   I2DMETHOD_(bool,RemoveCallback)(LPGRAPHCALLBACK, void *context);

   I2DMETHOD_(class C_rgb_conversion*,GetRGBConv)(){ return rgb_conv; }
   I2DMETHOD_(void*,GetHWND)() const{ return hwnd; }
   I2DMETHOD_(void*,GetMenu)() const{ return hmenu_main; }
   I2DMETHOD_(void,SetMenu)(void*);
   I2DMETHOD_(dword,GetFlags)() const{ return create_flags; }
   I2DMETHOD_(bool,UpdateParams)(dword sx, dword sy, dword bpp, dword flg, dword flg_mask, bool force_recreate);

   I2DMETHOD_(bool,SetViewport)(dword left, dword top, dword sizex, dword sizey);
   I2DMETHOD_(bool,ClearViewport)(){ return Rectangle((int)0, (int)0, (dword)scrn_sx, (dword)scrn_sy, (dword)0); }
   I2DMETHOD_(void,FlipToGDI)();

   I2DMETHOD_(bool,SetBrightness)(float r, float g, float b);

   I2DMETHOD_(const char*,GetAppName)() const{ return appName; }
   I2DMETHOD_(void,SetAppName)(const char*);

   I2DMETHOD_(bool,UpdateScreen)(dword flags = 0){
      return UpdateScreen2(flags);
   }
                              //rendering primitves
   I2DMETHOD_(bool,Rectangle)(int l, int t, dword sx, dword sy, dword color);

                              //keyboard
   I2DMETHOD_(void,InitKeyboard)();
   I2DMETHOD_(void,CloseKeyboard)();
   I2DMETHOD_(bool,IsKbInit)() const{ return kb_init; }
                              //remove!!!
   I2DMETHOD_(void,DefineKeys)(const char*, dword num = 32);
   I2DMETHOD_(char,ReadKey)(bool process_messages = false);
   I2DMETHOD_(dword,ReadKeys)(){ return game_keys; }

   I2DMETHOD_(dword,GetShiftKeys)() const{ return shiftkeys; }
   I2DMETHOD_(int,GetAscii)(char k, bool shift = false) const;
   I2DMETHOD_(const bool*,GetKeyboardMap)() const{ return keyboard_map; }
                              //mouse
   I2DMETHOD_(bool,MouseInit)(dword mode = 0); //normal/exclusive
   I2DMETHOD_(void,MouseClose)();
   I2DMETHOD_(void,MouseSetup)(int mouse_x, int mouse_y, int clip_left, int clip_top, int clip_right, int clip_bottom);
   I2DMETHOD_(dword,MouseUpdate)();
   I2DMETHOD_(int,Mouse_x)() const{ return mouse_pos[0]; }
   I2DMETHOD_(int,Mouse_y)() const{ return mouse_pos[1]; }
   I2DMETHOD_(int,Mouse_rx)() const{ return mouse_rel_pos[0]; }
   I2DMETHOD_(int,Mouse_ry)() const{ return mouse_rel_pos[1]; }
   I2DMETHOD_(int,Mouse_rz)() const{ return mouse_rel_pos[2]; }
   I2DMETHOD_(void,SetMousePos)(int x, int y);
   I2DMETHOD_(dword,GetMouseButtons)() const{ return mouse_butt; }
   I2DMETHOD_(int,GetMousePos)(int axis, bool relative) const;
                              //windows
   I2DMETHOD_(void,ProcessWinMessages)(dword flags=0);
                              //timer
   I2DMETHOD_(void,NullTimer)();
   I2DMETHOD_(dword,GetTimer)(dword min, dword max);
   I2DMETHOD_(dword,ReadTimer)() const;

                              //screen
   I2DMETHOD_(dword,Scrn_sx)() const{ return scrn_sx; }
   I2DMETHOD_(dword,Scrn_sy)() const{ return scrn_sy; }
   I2DMETHOD_(const S_pixelformat*,GetPixelFormat)() const{ return rgb_conv->GetPixelFormat(); }
   I2DMETHOD_(bool,CreateImage)(LPIImage*);
   I2DMETHOD_(bool,CreateRGBConv)(class C_rgb_conversion**) const;

   I2DMETHOD_(bool,GetImageInfo)(const char *name, dword &sx, dword &sy, byte &bpp, dword &id);
                              //Dialog windows
   I2DMETHOD_(bool,AddDlgHWND)(void*);
   I2DMETHOD_(bool,DeleteDlgHWND)(void*);

   I2DMETHOD_(struct IDirect3D8*,GetD3DInterface)() const{ return lpD3D8; }
   I2DMETHOD_(struct IDirect3DDevice8*,GetD3DDevice)() const{ return lpDev8; }
   I2DMETHOD_(bool,UpdateScreen2)(dword flags = 0, void *dest_hwnd = NULL);

   I2DMETHOD_(void,GetDesktopResolution)(int &x, int &y) const;
   I2DMETHOD_(dword,NumDisplayModes)() const;
   I2DMETHOD_(bool,GetDisplayMode)(int id, int *sx, int *sy, int *bpp) const;

   I2DMETHOD_(class IImage*,CreateBackbufferCopy)(bool lockable);
   I2DMETHOD_(dword,GetBackbufferCount)() const{ return num_swap_buffers-1; }

   I2DMETHOD_(bool,GetWindowCloseStatus)(){
      bool ret = want_close_app;
      want_close_app = false;
      return ret;
   }
};

typedef IGraph *LPIGraph;

//----------------------------
                              //image interface
class IImage: public C_unk<IImage>{

   C_smart_ptr<IDirect3DTexture8> lpTxt8;
   C_smart_ptr<IDirect3DSurface8> lpImg8;

   dword size_x, size_y;
   dword *pal_argb;           //pointer to 256 ARGB values, or NULL
   vector<byte> ckey_rle_info;//rle-compressed colorkey info
   LPIGraph igraph;
   friend IGraph;

                              //for restoring when surface gets lost - only nontexture surfaces
   C_string open_fname;        
   dword open_flags;
protected:
   S_pixelformat pixel_format;
public:
   IImage(LPIGraph);
   ~IImage();

   I2DMETHOD_(dword,AddRef)(){ return C_unk<IImage>::AddRef(); }
   I2DMETHOD_(dword,Release)(){ return C_unk<IImage>::Release(); }
                              //initialization
   I2DMETHOD_(bool,Open)(const char *name, dword flags = 0, dword sx = 0, dword sy = 0,
      const S_pixelformat* = NULL, dword mipmapminsize = 0);
   I2DMETHOD_(void,Close)();
   I2DMETHOD_(bool,IsInit)() const;
                              //palette stuff
   I2DMETHOD_(bool,SetPal)(dword*);
   I2DMETHOD_(dword*,GetPal)() const{ return (dword*)pal_argb; }
                              //rendering
   I2DMETHOD_(bool,Draw)(int x=0, int y=0, dword flags=0);
                              //access surface
   I2DMETHOD_(bool,Lock)(void **area, dword *pitch, byte *bpp);
   I2DMETHOD_(bool,Unlock)();
                              //special fncs
   I2DMETHOD_(bool,AlphaBlend)(LPIImage, dword flags = 0);
                              //remove!!!
   I2DMETHOD_(bool,MixMipmap)(dword flags);
   I2DMETHOD_(bool,ConvertBumpMap)(LPIImage, dword flags = 0);
                              //info
   I2DMETHOD_(const S_pixelformat*,GetPixelFormat)() const;
   I2DMETHOD_(const void*,GetColorkeyInfo)() const{ return ckey_rle_info.size() ? ckey_rle_info.begin() : NULL; }
   I2DMETHOD_(bool,CopyColorkeyInfo)(IImage *img){
      if(!img)
         return false;
      ckey_rle_info = img->ckey_rle_info;
      return true;
   }
   I2DMETHOD_(dword,SizeX)() const{ return size_x; }
   I2DMETHOD_(dword,SizeY)() const{ return size_y; }

   I2DMETHOD_(LPIGraph,GetIGraph)() const{ return igraph; }

   I2DMETHOD_(struct IDirect3DTexture8*,GetDirecXTexture)() const{ return lpTxt8; }
   I2DMETHOD_(struct IDirect3DSurface8*,GetDirecXSurface)() const{ return lpImg8; }

   I2DMETHOD_(bool,SaveShot)(C_cache*, const char *ext) const;
   I2DMETHOD_(bool,CopyStretched)(LPIImage src_img);
};

//----------------------------

#endif
