#include "all.h"
#include "loader.h"
#include <algorithm>


#define TEMPSIZE 0x8000

#define COMPRESSED            //include de-compression code

#ifdef _MSC_VER
using namespace std;
#endif

//----------------------------

class C_bmp_loader: public C_image_loader{
public:
   C_bmp_loader(){
      RegisterImageLoader(this);
   }

//----------------------------

   virtual bool GetImageInfo(int h, const char *file_name, void *header, dword hdr_size,
      dword *size_x, dword *size_y, S_pixelformat *pf){


      if(hdr_size<4) return false;
      if((*(word*)header) != 0x4d42) return false;

#pragma pack(push,1)
      struct S_bmp_header{
         dword size;
         word res[2];
         dword data_offs;
         struct{
            dword size;
            long sx, sy;
            word planes;
            word bitcount;
            dword compression;
            dword img_size;
            long x_ppm;
            long y_ppm;
            dword num_clr;
            dword clr_important;
         }bi_hdr;
      } bmp_header;
#pragma pack(pop)

      int i;
      dtaSeek(h, 2, DTA_SEEK_SET);
                              //read header
      i = dtaRead(h, &bmp_header, sizeof(bmp_header));
                              //error reading file header
      if(i != sizeof(bmp_header)) return false;
                              //get to data beg
      dtaSeek(h, bmp_header.data_offs, DTA_SEEK_SET);

      dword sx = bmp_header.bi_hdr.sx;
      dword sy = bmp_header.bi_hdr.sy;
      byte bpp = bmp_header.bi_hdr.bitcount/8;

                              //ok, we've identified file and allocated memory
                              //setup return values
      *size_x = sx;
      *size_y = sy;
      memset(pf, 0, sizeof(S_pixelformat));
      pf->bytes_per_pixel = bpp;
      if(bpp==1){
         pf->flags = PIXELFORMAT_PALETTE;
      }else{
         pf->r_mask = 0xff0000;
         pf->g_mask = 0x00ff00;
         pf->b_mask = 0x0000ff;
      }
      return true;
   }

//----------------------------

   virtual bool Load(int h, const char *file_name, void *header, dword hdr_size,
      void **mem1, void **pal1, dword *size_x, dword *size_y, S_pixelformat *pf){

      if(hdr_size<4)
         return false;
      if((*(word*)header) != 0x4d42)
         return false;

#pragma pack(push,1)
      struct S_bmp_header{
         dword size;
         word res[2];
         dword data_offs;
         struct{
            dword size;
            long sx, sy;
            word planes;
            word bitcount;
            dword compression;
            dword img_size;
            long x_ppm;
            long y_ppm;
            dword num_clr;
            dword clr_important;
         }bi_hdr;
      } bmp_header;
#pragma pack(pop)

      int i;
      dtaSeek(h, 2, DTA_SEEK_SET);
                              //read header
      i = dtaRead(h, &bmp_header, sizeof(bmp_header));
                              //error reading file header
      if(i != sizeof(bmp_header))
         return false;
                              //get to data beg
      dtaSeek(h, bmp_header.data_offs, DTA_SEEK_SET);

                              //check if it's supported bit-count
      switch(bmp_header.bi_hdr.bitcount){
      case 8:
      case 24:
         break;
      default:
         return false;
      }
      dword sx = bmp_header.bi_hdr.sx;
      dword sy = bmp_header.bi_hdr.sy;
      byte bpp = bmp_header.bi_hdr.bitcount/8;

                              //alloc screen memory
      byte *mem = (byte*)MemAlloc(sx*sy*bpp);
      if(!mem)
         return false;
                              //setup pointer to palette
      byte *pal_argb = NULL;
      if(bpp == 1){
         pal_argb = (byte*)MemAlloc(256 * sizeof(dword));
         if(!pal_argb){
            MemFree(mem);
            return false;
         }
      }
                              //ok, we've identified file and allocated memory
                              //setup return values
      *mem1 = mem;
      *pal1 = pal_argb;
      *size_x = sx;
      *size_y = sy;
      memset(pf, 0, sizeof(S_pixelformat));
      pf->bytes_per_pixel = bpp;
      if(bpp==1){
         pf->flags = PIXELFORMAT_PALETTE;
      }else{
         pf->r_mask = 0xff0000;
         pf->g_mask = 0x00ff00;
         pf->b_mask = 0x0000ff;
      }
                              //scanline is dword aligned,
                              //compute padding length
      dword padding;
      padding = ((sx*bpp+3) & -4) - sx*bpp;

      switch(bmp_header.bi_hdr.compression){
      case BI_RGB:            //uncompressed data
                              //read bottom-to-top
         mem += sx*sy*bpp;
         while(sy--){
            mem -= sx*bpp;
            dtaRead(h, mem, sx*bpp);
            if(padding)
               dtaSeek(h, padding, DTA_SEEK_CUR);
         };
         break;
      case BI_RLE8:           //compressed data
#ifndef COMPRESSED
         goto fail;
#else
#define READ_CACHE   if(tmpptr==tmptop){\
                        dword readb=dtaRead(h, tmp, TEMPSIZE);\
                        tmptop=(tmpptr=tmp)+readb; }
                              //allocate temp buffer
         byte *tmp = new byte[TEMPSIZE], *tmpptr=tmp, *tmptop=tmp;
                              //temp allocation error
         if(!tmp) goto fail;

         while(sy--){
            mem = ((byte*)*mem1) + sy*sx*bpp;
            while(true){
               union{
                  byte code[2];
                  word wcode;
               } au;
               READ_CACHE
               au.wcode = *(word*)tmpptr;
               tmpptr += 2;
               if(!au.code[0]){
                  switch(au.code[1]){
                  case 0:     //eol
                     goto eol;
                  case 1:     //eof
                     goto eof;
                  case 2:     //delta
                     dtaRead(h, &au.code, 2);
                     break;
                  default:    //data length
                     i = tmptop-tmpptr;
                     if(i<au.code[1]){
                        memcpy(mem, tmpptr, i);
                        mem += i;
                        au.code[1] -= i;
                        tmpptr = tmptop;
                        READ_CACHE
                     }
                     memcpy(mem, tmpptr, au.code[1]);
                     tmpptr += au.code[1];
                     mem += au.code[1];
                              //align to word
                     if(au.code[1]&1) ++tmpptr;
                  }
               }else{         //color fill
                  memset(mem, au.code[1], au.code[0]);
                  mem += au.code[0];
               }
            }
eol:{}
        }
eof:
         delete[] tmp;
         break;
#endif
      }
                              //read palette
      if(bpp==1){
         i = bmp_header.bi_hdr.num_clr;
         if(!i)
            i = 256;
         byte *tmp = new byte[sizeof(dword) * i];
         if(!tmp)
            goto fail;

         dtaSeek(h, sizeof(bmp_header)+2, DTA_SEEK_SET);
         dtaRead(h, tmp, i * sizeof(dword));
                              //fill the rest of palette by zero
         memset(pal_argb + i*sizeof(dword), 0, (256-i)*sizeof(dword));
         while(i--){
            pal_argb[i*4 + 0] = tmp[i*4+2];
            pal_argb[i*4 + 1] = tmp[i*4+1];
            pal_argb[i*4 + 2] = tmp[i*4+0];
            pal_argb[i*4 + 3] = 0xff;
         }
         delete[] tmp;
      }
      dtaClose(h);
      return true;
   fail:
      MemFree(mem);
      if(pal_argb) MemFree(pal_argb);
      return false;
   }
} bmp_loader;

//----------------------------

