#include "all.h"
#include "loader.h"

#define TEMPSIZE 0x8000


//----------------------------

class C_pcx_loader: public C_image_loader{
public:
   C_pcx_loader(){
      RegisterImageLoader(this);
   }

//----------------------------

   virtual bool GetImageInfo(int h, const char *file_name, void *header, dword hdr_size,
      dword *size_x, dword *size_y, S_pixelformat *pf){

      if(hdr_size<4)
         return false;
      if((*(dword*)header) != 0x0801050a)
         return false;

      word sx, sy;
      byte bpp;
                              //get size y
      dtaSeek(h, 0xa, DTA_SEEK_SET);
      dtaRead(h, &sy, sizeof(word));
      ++sy;
                              //get bpp
      dtaSeek(h, 0x41, DTA_SEEK_SET);
      dtaRead(h, &bpp, sizeof(byte));
                              //get size x
      dtaRead(h, &sx, sizeof(word));
                              //get to data beg
      dtaSeek(h, 0x80, DTA_SEEK_SET);

                              //ok, we've identified file and allocated memory
                              //setup return values
      *size_x = sx;
      *size_y = sy;
      memset(pf, 0, sizeof(S_pixelformat));
      pf->bytes_per_pixel = bpp;
      if(bpp==1){
         pf->flags = PIXELFORMAT_PALETTE;
      }else{
         pf->r_mask = 0xff0000;
         pf->g_mask = 0x00ff00;
         pf->b_mask = 0x0000ff;
      }
      return true;
   }

//----------------------------

   virtual bool Load(int h, const char *file_name, void *header, dword hdr_size,
      void **mem1, void **pal1, dword *size_x, dword *size_y, S_pixelformat *pf){

      if(!GetImageInfo(h, file_name, header, hdr_size, size_x, size_y, pf))
         return false;
      dword sx = *size_x;
      dword sy = *size_y;
      dword bpp = pf->bytes_per_pixel;

                              //try to allocate memory
      byte *mem = (byte*)MemAlloc(sx*sy*bpp);
      if(!mem)
         return false;
                              //setup pointer to palette
      byte *pal_argb = NULL;
      if(bpp==1){
         pal_argb = (byte*)MemAlloc(256 * sizeof(dword));
         if(!pal_argb){
            MemFree(mem);
            return false;
         }
      }
                              //allocate temp buffer
      byte *tmp = new byte[TEMPSIZE], *tmpptr=tmp, *tmptop=tmp;
      if(!tmp){
                              //temp allocation error
         MemFree(mem);
         MemFree(pal_argb);
         return false;
      }
      *mem1 = mem;
      *pal1 = pal_argb;


      int rept;
      byte color;
      dword readb;            //cache

#define READ_CACHE   if(tmpptr==tmptop){\
                        readb=dtaRead(h, tmp, TEMPSIZE);\
                        tmptop=(tmpptr=tmp)+readb;\
                     }
                              //do all lines
      while(sy--){
                              //do all planes
         int plane = bpp - 1;
         dword last_rept = 0;
         do{
            byte *dst = &mem[plane];
                                 //do entire row
            dword row = 0;
            while(row < sx){
               READ_CACHE
               if(last_rept || *tmpptr>=0xc0){
                                 //color row
                  if(last_rept) rept = last_rept, last_rept = 0;
                  else{
                     rept = *tmpptr++ - 0xc0;
                     READ_CACHE
                     color = *tmpptr++;
                  }
                  if((row += rept) > sx)
                     rept -= (row - sx), last_rept = (row - sx);
                  if(bpp==1){
                     memset(dst, color, rept);
                     dst += rept;
                  }else
                  do{
                     *dst = color;
                     dst += bpp;
                  }while(--rept);
               }else{
                                 //single color
                  ++row;
                  *dst = *tmpptr++;
                  dst += bpp;
               }
            }
         }while(--plane >= 0);
         mem += sx*bpp;
      }
      delete[] tmp;
                                 //read palette
      if(bpp==1){
         dtaSeek(h, -768, DTA_SEEK_END);
         byte *pal_3 = new byte[768];
         dtaRead(h, pal_3, 768);
         for(int i=0; i<256; i++){
            pal_argb[i*4 + 0] = pal_3[i*3 + 0];
            pal_argb[i*4 + 1] = pal_3[i*3 + 1];
            pal_argb[i*4 + 2] = pal_3[i*3 + 2];
            pal_argb[i*4 + 3] = 0xff;
         }
         delete[] pal_3;
      }
      dtaClose(h);

      return true;
   }
} pcx_loader;

//----------------------------
//----------------------------
