#include "all.h"
#include "loader.h"
extern"C"{
#include "png\png.h"
}

#pragma comment(lib,"png\\png")
#pragma comment(lib,"zlib")

//----------------------------

static void __cdecl ReadFromDTA(png_structp png_ptr, png_bytep buf, png_size_t size){
   dtaRead((int)png_get_io_ptr(png_ptr), buf, size);
}

//----------------------------

class C_png_loader: public C_image_loader{

   bool Load(int h, const char *file_name, void *header, dword hdr_size,
      void **mem1, void **pal1, dword *size_x, dword *size_y, S_pixelformat *pf,
      bool only_header){

      if(hdr_size<4)
         return false;

      if(!png_check_sig((byte*)header, hdr_size))
         return false;

      bool ok = false;
      png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
      if(!png_ptr)
         return false;
      png_infop info_ptr = png_create_info_struct(png_ptr);
      if(info_ptr){
         png_set_bgr(png_ptr);

         dtaSeek(h, 0, DTA_SEEK_SET);
         png_set_read_fn(png_ptr, (void*)h, &ReadFromDTA);

         png_read_info(png_ptr, info_ptr);

         memset(pf, 0, sizeof(S_pixelformat));
         pf->bytes_per_pixel = info_ptr->pixel_depth / 8;
                              //if 8-bit grayscale, consider it as RGB
         *size_x = info_ptr->width;
         *size_y = info_ptr->height;
         if(pf->bytes_per_pixel==1){
            pf->flags = PIXELFORMAT_PALETTE;
         }else{
            pf->r_mask = 0xff0000;
            pf->g_mask = 0x00ff00;
            pf->b_mask = 0x0000ff;
         }
         if(!only_header){
            switch(pf->bytes_per_pixel){
            case 1: case 3:
               {
                              //supported depth, continue reading
                              //alloc screen memory
                  byte *mem = (byte*)MemAlloc((*size_x) * (*size_y) * pf->bytes_per_pixel);
                  if(!mem)
                     break;
                                          //setup pointer to palette
                  byte *pal_argb = NULL;
                  if(pf->bytes_per_pixel == 1){
                     pal_argb = (byte*)MemAlloc(256 * sizeof(dword));
                     if(!pal_argb){
                        MemFree(mem);
                        break;
                     }
                     if(info_ptr->color_type==PNG_COLOR_TYPE_GRAY){
                              //grayscale - create fake palette
                        for(int i=0, mask = 0; i<256; i++, mask += 0x010101){
                           ((dword*)pal_argb)[i] = mask;
                        }
                     }else{
                              //real palette, fill-in indicies
                        const byte *pp = (const byte*)png_ptr->palette;
                        for(int pi=png_ptr->num_palette; pi--; ){
                           pal_argb[pi*4+0] = pp[pi*3+0];
                           pal_argb[pi*4+1] = pp[pi*3+1];
                           pal_argb[pi*4+2] = pp[pi*3+2];
                           pal_argb[pi*4+3] = 0;
                        }
                     }
                  }
                  *mem1 = mem;
                  *pal1 = pal_argb;

                  byte **ptr_buf = new byte*[*size_y];

                  for(int y=*size_y; y--; ){
                     ptr_buf[y] = mem + y * (*size_x) * pf->bytes_per_pixel;
                  }
                  png_read_image(png_ptr, ptr_buf);

                  delete[] ptr_buf;

                  ok = true;
               }
               break;
            }
         }else{
            ok = true;
         }
      }
      png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
      return ok;
   }

public:
   C_png_loader(){
      RegisterImageLoader(this);
   }

//----------------------------

   virtual bool GetImageInfo(int h, const char *file_name, void *header, dword hdr_size,
      dword *size_x, dword *size_y, S_pixelformat *pf){

      return Load(h, file_name, header, hdr_size, NULL, NULL, size_x, size_y, pf, true);
   }

//----------------------------

   virtual bool Load(int h, const char *file_name, void *header, dword hdr_size,
      void **mem1, void **pal1, dword *size_x, dword *size_y, S_pixelformat *pf){

      bool b = Load(h, file_name, header, hdr_size, mem1, pal1, size_x, size_y, pf, false);
      if(b)
         dtaClose(h);
      return b;
   }
} png_loader;

//----------------------------
//----------------------------

