#include <insanity\os.h>
#include <windows.h>
#include <stdio.h>

//----------------------------

bool OsCreateDirectory(const char *cp){

   SECURITY_ATTRIBUTES sa;
   memset(&sa, 0, sizeof(sa));
   sa.nLength = sizeof(sa);
   return CreateDirectory(cp, &sa);
}

//----------------------------

bool OsCreateDirectoryTree(const char *dst){

   if(!*dst)
      return false;
   
   SECURITY_ATTRIBUTES sa;
   memset(&sa, 0, sizeof(sa));
   sa.nLength = sizeof(sa);
   int cpos, npos = 0;
   C_string str = dst;
   while(true){
      cpos = npos;
      while(str[++npos]) if(str[npos]=='\\') break;
      if(cpos){
         str[cpos] = 0;
         //bool st = 
         CreateDirectory(str, &sa);
         /*
         if(!st)
            return false;
            */
         str[cpos] = '\\';
      }
      if(!str[npos])
         break;
   }
   return true;
}

//----------------------------

bool OsDeleteFile(const char *cp){
   return DeleteFile(cp);
}

//----------------------------

bool OsGetComputerName(C_string &str){

   char buf[255];
   dword sz = sizeof(buf);
   if(!GetComputerName(buf, &sz)){
      str = NULL;
      return false;
   }
   str = buf;
   return true;
}

//----------------------------

bool OsMoveFile(const char *src, const char *dst){

   bool ok;
   ok = DeleteFile(dst);
   if(!ok){
                              //try to clear the read-only flag
      SetFileAttributes(dst, GetFileAttributes(dst) & ~FILE_ATTRIBUTE_READONLY);
   }
   ok = MoveFile(src, dst);
   if(!ok){
      OsCreateDirectoryTree(dst);
      DeleteFile(dst);
      ok = MoveFile(src, dst);
   }
   return ok;
}

//----------------------------

void OsDestroyWindow(void *hwnd){
   DestroyWindow((HWND)hwnd);
}

//----------------------------

void OsShowWindow(void *hwnd, bool on){

   if(on){
      ShowWindow((HWND)hwnd, SW_SHOW);
   }else{
      ShowWindow((HWND)hwnd, SW_HIDE);
   }
   UpdateWindow((HWND)hwnd);
}

//----------------------------

void OsCenterWindow(void *hwnd, void *hwnd_other){

   RECT rc;
   GetWindowRect((HWND)hwnd, &rc);
   int sx = GetSystemMetrics(SM_CXSCREEN);
   int sy = GetSystemMetrics(SM_CYSCREEN);
   SetWindowPos((HWND)hwnd, NULL,
      (sx-(rc.right-rc.left))/2,
      (sy-(rc.bottom-rc.top))/2,
      0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

//----------------------------

bool OsIsFileReadOnly(const char *fname){

   dword dw = GetFileAttributes(fname);
   if(dw==0xffffffff) return false;
   return (dw&FILE_ATTRIBUTE_READONLY);
}

//----------------------------

bool OsMakeFileReadOnly(const char *fname, bool rdonly){

   dword dw = GetFileAttributes(fname);
   if(dw==0xffffffff)
      return false;
   bool is_rdonly = (dw&FILE_ATTRIBUTE_READONLY);
   if(is_rdonly==rdonly)
      return false;
   dw ^= FILE_ATTRIBUTE_READONLY;
   return SetFileAttributes(fname, dw);
}

//----------------------------

void OsSleep(int time){
   Sleep(time);
}

//----------------------------

bool OsIsCDROMDrive(const char *path){

   char drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];
   _splitpath(path, drive, dir, fname, ext);
   UINT dt = GetDriveType(drive);
   return (dt==DRIVE_CDROM);
}

//----------------------------

void OsCollectFiles(const char *path1, const char *wildmask, vector<C_string> &ret){

   C_fstring path = path1;
   if(path.Size() && path[(int)path.Size()-1]!='\\')
      path += "\\";
   C_fstring findname("%s%s", (const char*)path, wildmask);

   ret.clear();
   WIN32_FIND_DATA fd;
   HANDLE h = FindFirstFile(findname, &fd);
   if(h!=INVALID_HANDLE_VALUE){
      do{
         ret.push_back(C_fstring("%s%s", (const char*)path, fd.cFileName));
      }while(FindNextFile(h, &fd));
      FindClose(h);
   }
}

//----------------------------

void OsCreateUniqueFileName(const char *path, const char *ext, C_string &str){

                              //make sure name contains valid characters only
   for(int i=str.Size(); i--; ){
      char &c = str[i];
      if(!isalnum(c) && !isspace(c))
         c = '_';
   }
   if(!str.Size())
      str = "name";

                              //make sure such name doesn't exist
   int digit_index = str.Size();
   while(digit_index && isdigit(str[digit_index])) --digit_index;
   int digit_base = atoi(&str[digit_index]);
   while(true){
      C_fstring full_path("%s\\%s.%s", path, (const char*)str, ext);
      FILE *fp = fopen(full_path, "rb");
      if(!fp)
         break;
      fclose(fp);
                              //modify name
      str[digit_index] = 0;
      str = C_fstring("%s%i", (const char*)str, ++digit_base);
   }
}

//----------------------------

void OsGetCurrentTimeDate(S_date_time &dt){

   SYSTEMTIME st;
   GetLocalTime(&st);
   dt.year = st.wYear;
   dt.month = st.wMonth;
   dt.day = st.wDay;
   dt.hour = st.wHour;
   dt.minute = st.wMinute;
   dt.second = st.wSecond;
}

//----------------------------

/*
bool OsGetFileTimeDate(const char *fname, S_date_time &dt){

   dword ftime[2];
   if(!GetFileTime(fname, ftime))
      return false;
   SYSTEMTIME st;
   FileTimeToSystemTime((const FILETIME*)ftime, &st);
   dt.year = st.wYear;
   dt.month = st.wMonth;
   dt.day = st.wDay;
   dt.hour = st.wHour;
   dt.minute = st.wMinute;
   dt.second = st.wSecond;

   return true;
}
*/

//----------------------------

bool OsGetExecutableVersion(const char *fname, int &major, int &minor){

   FILE *fp = fopen(fname, "rb");
   if(!fp)
      return false;
   IMAGE_DOS_HEADER dh;
   fread(&dh, sizeof(dh), 1, fp);

   int seek = dh.e_lfanew + sizeof(dword) + sizeof(IMAGE_FILE_HEADER);
   fseek(fp, seek, SEEK_SET);
   IMAGE_OPTIONAL_HEADER oh;
   fread(&oh, sizeof(oh), 1, fp);
   major = oh.MajorImageVersion;
   minor = oh.MinorImageVersion;

   fclose(fp);
   return true;
}

//----------------------------

void *OsCreateFont(int size){

   HFONT h_font =
      CreateFont(
      -size,                  //height
      0,                      //width
      0,                      //escapement
      0,                      //orientation
      0,                      //weight
      false,                  //italic
      false,                  //underline
      false,                  //stike-out
      OEM_CHARSET,
      OUT_RASTER_PRECIS,
      CLIP_DEFAULT_PRECIS,
      DEFAULT_QUALITY,
      VARIABLE_PITCH, NULL
      );
   return h_font;
}

//----------------------------

void OsDestroyFont(void *h_font){

   DeleteObject((HFONT)h_font);
}

//----------------------------
//----------------------------
