#include <windows.h>
#include <insanity\texts.h>
#include <C_cache.h>
#include <insanity\os.h>
#include <map>

#ifdef _MSC_VER
using namespace std;
#endif

static const char TEXT_MESSAGE[] = "Initializing texts";

//----------------------------
                              //leave implementation private to this source file

                              //akk_texts implementation - using STL map
class C_all_texts_rep{
   struct S_text{
      C_string str;
      bool unicode;

      S_text(){}
      S_text(const S_text &t){ operator =(t); }
      S_text &operator =(const S_text &t){
         str = t.str;
         unicode = t.unicode;
         return *this;
      }
   };
   typedef map<int, S_text> t_text_map;
   t_text_map m;
   friend C_all_texts;
public:

   void AddLines(const char *name, C_cache &is, int delta_index){

      int i;
      char line[4096], *cp;
      int line_num = 0;
      while(!is.eof()){
         ++line_num;
         is.getline(line, 4096);
         if(line[0]==';') continue;
         cp = line;
         int id;
         int num = sscanf(cp, "%i%n", &id, &i);
         if(!num || num==EOF)
            continue;
         id += delta_index;
#ifdef _DEBUG
                                 //check duplication
         t_text_map::const_iterator it;
         it = m.find((const int&)id);
         if(it!=m.end()){
            OsMessageBox(C_fstring("File '%s': text duplication on ID %i", name, id),
               TEXT_MESSAGE, MBOX_OK);
         }
#endif
         cp += i;
         while(*cp && *cp!='"') ++cp;
         if(*cp++){
            i = 0;
            
            bool loop = true;
            while(loop)
            switch(cp[i]){
            case '"':
               {
                              //check to eol to find bracing errors
                  for(int j=i+1; cp[j] && isspace(cp[j]); j++);
                  if(cp[j] && cp[j]!=';'){
                     MessageBox(NULL, C_fstring("File %s: text error on line %i\n(text after last quote)", name, line_num),
                        TEXT_MESSAGE, MB_OK);
                  }
                  cp[i] = 0;
                  loop = false;
               }
               break;
            case '\\':
               cp[i] = '\n';
               break;
            case 0:
               {
                  MessageBox(NULL, C_fstring("File %s: text error on line %i\n(no ending quote)", name, line_num),
                     TEXT_MESSAGE, MB_OK);
                  loop = false;
               }
               break;
            default: 
               ++i;
            }
            
            m[id].str = cp;
            m[id].unicode = false;
         }
      }
   }

//----------------------------

   void AddLinesW(const char *name, C_cache &is, int delta_index){
                              //read whole file
      int num_chars = is.filesize()/sizeof(wchar_t);
      wchar_t *buf = new wchar_t[num_chars];
      is.read((char*)buf, is.filesize());

      int curr_pos = 0;

      int i;
      bool line1_read = false;
      while(curr_pos < num_chars){
                              //find end of current line
         for(int eol = curr_pos; eol<num_chars; eol++){
            if(buf[eol]=='\r' && buf[eol+1]=='\n')
               break;
         }

         if(!line1_read){     //ignore 1st line
            line1_read = true;
            curr_pos = eol + 2;
            continue;
         }
                              //skip empty lines and comments
         if(eol!=curr_pos && buf[curr_pos]!=';'){
                              //scan number
            int id = 0x80000000;
            for(; buf[curr_pos]<256 && isdigit(buf[curr_pos]); curr_pos++)
               id = id*10 + buf[curr_pos]-'0';

            if(id<0){
               curr_pos = eol + 2;
               continue;
            }
            id += delta_index;
#ifdef _DEBUG
                                 //check duplication
            t_text_map::const_iterator it;
            it = m.find((const int&)id);
            if(it!=m.end()){
               OsMessageBox(C_fstring("File %s: text duplication on ID %i", name, id),
                  TEXT_MESSAGE, MBOX_OK);
            }
#endif
                              //move to beginning
            while(curr_pos<eol && buf[curr_pos]!='\"') ++curr_pos;
            ++curr_pos;
                              //find end
                              //find end
            for(i=curr_pos; i<eol && buf[i]!='\"'; ++i){
               wchar_t c = buf[i];
               switch(c){
               case '\\':
                  buf[i] = '\n';
                  break;
               }
            }
                              //make string
            C_string str;
            str.Assign((const char*)&buf[(int)curr_pos], (i-curr_pos)*sizeof(wchar_t) + 2);
            str[(int)str.Size()-2] = 0;
            str[(int)str.Size()-1] = 0;
                              //save
            m[id].str = str;
            m[id].unicode = true;
         }
                              //next line
         curr_pos = eol + 2;
      }
      delete[] buf;
   }

//----------------------------

   bool AddFile(class C_cache &is, const char *name, int delta_index){

                              //read 1st word to detect unicode format
      word w;
      is.read((char*)&w, sizeof(word));
      is.seekp(0);
      if(w!=0xfeff){
         AddLines(name, is, delta_index);
      }else{
         AddLinesW(name, is, delta_index);
      }
      return true;
   }

//----------------------------

   bool AddFile(class C_cache &is, int delta_index){
      return AddFile(is, "unknown", delta_index);
   }

//----------------------------

   bool AddFile(const char *name, int delta_index){
      
      C_cache is;
      is.open(name, CACHE_READ);
      if(!is.good())
         return false;
      return AddFile(is, name, delta_index);
   }
};

//----------------------------

C_all_texts::C_all_texts():
   rep(NULL)
{}

//----------------------------

C_all_texts &C_all_texts::operator =(const C_all_texts &at){

   Close();
   if(at.rep){
      rep = new C_all_texts_rep;
      rep->m = at.rep->m;
   }
   return *this;
}

//----------------------------

bool C_all_texts::AddFile(const char *name, int delta_index){

   if(!rep)
      rep = new C_all_texts_rep;
   return rep->AddFile(name, delta_index);
}

//----------------------------

bool C_all_texts::AddFile(class C_cache &is, int delta_index){

   if(!rep)
      rep = new C_all_texts_rep;
   return rep->AddFile(is, delta_index);
}

//----------------------------

void C_all_texts::Close(){

   delete rep;
   rep = NULL;
}

//----------------------------

const C_string &C_all_texts::operator [](int id) const{

   if(!rep){
      static C_string err_str;
      err_str = C_fstring("error: text #%i not defined", id);
      return err_str;
   }
   C_all_texts_rep::t_text_map::const_iterator it;
   it = rep->m.find((const int&)id);
   if(it==rep->m.end()){
      static C_string err_str;
      err_str = C_fstring("error: text #%i not defined", id);
      return err_str;
   }
   return (*it).second.str;
}

//----------------------------

bool C_all_texts::IsDefined(int id) const{

   C_all_texts_rep::t_text_map::const_iterator it;
   it = rep->m.find((const int&)id);
   return (it!=rep->m.end());
}

//----------------------------

bool C_all_texts::IsWide(int id) const{

   C_all_texts_rep::t_text_map::const_iterator it;
   it = rep->m.find((const int&)id);
   if(it==rep->m.end())
      return false;
   return (*it).second.unicode;
}

//----------------------------

void C_all_texts::EnumTexts(void(*cbEnum)(int, const C_string&, void *context), void *context){

   C_all_texts_rep::t_text_map::const_iterator it;
   for(it=rep->m.begin(); it!=rep->m.end(); it++){
      (*cbEnum)((*it).first, (*it).second.str, context);
   }
}

//----------------------------

void C_all_texts::DelTexts(int base_id, int num_ids){

   C_all_texts_rep::t_text_map::iterator it, it_next;
   for(it=rep->m.begin(); it!=rep->m.end(); ){
      int id = (*it).first;
      if(id >= base_id && id < (base_id + num_ids))
         it = rep->m.erase(it);
      else
         ++it;
   }
}

//----------------------------
//----------------------------
