#include "Max.h"
#include "surf_api.h"
#include "i3dexp.h"
#include "..\common\mtldef.h"
#include "stdmat.h"
#include "list"
#include "algorithm"
#include "decomp.h"

#include"modstack.h"          //nurbs modifiers
#include "Notetrck.h"         //notes

#pragma comment(lib, "edmodel.lib")

using namespace std;


                              //max. number of map channels we do export
                              // this limitation is only because we don't want
                              // redundant channels we don't use
#define MAX_EXPORT_MAP_CHANNELS 2
//----------------------------

// 3DS Key structures

#pragma pack(1)

struct fc_wrt{
   unsigned short a;
   unsigned short b;
   unsigned short c;
   unsigned short flags;
};

#pragma pack()

//----------------------------

#define PRIMARY_NODE (1<<14)

enum E_OBJTYPE{
   OBTYPE_MESH = 0,
   OBTYPE_CAMERA = 1,
   OBTYPE_OMNILIGHT = 2,
   OBTYPE_SPOTLIGHT = 3,
   OBTYPE_DUMMY = 5,
   OBTYPE_CTARGET = 6,
   OBTYPE_LTARGET = 7,

   OBTYPE_FREEDIRECTLIGHT = 9,
   OBTYPE_FREESPOTLIGHT = 10,
   OBTYPE_TARGETDIRECTLIGHT = 11,

   OBTYPE_NURBS = 13,
   OBTYPE_BONE = 20,
};

//----------------------------

static Point3 colorVal;
static float floatVal;

//----------------------------

struct MeshAndNode{
   Mesh *mesh;
   INode *node;
};

//----------------------------

BOOL WINAPI DllMain(HINSTANCE hinstDLL,ULONG fdwReason,LPVOID lpvReserved){

   static int controlsInit = false;

   if(!controlsInit){
      controlsInit = true;
      
      InitCustomControls(hinstDLL);
      InitCommonControls();
   }
   return true;
}

//----------------------------

#pragma pack(1)

struct CameraI3D{
   float x;
   float y;
   float z;
   float tx;
   float ty;
   float tz;
   float bank;
   float focal;
   /*
   WORD flags;
   float nearplane;
   float farplane;
   void *appdata;
   */
};

#pragma pack()

//- Material Export -------------------------------------------------------------

struct MEntry{
   SMtl *sm;
   Mtl *m;
};

static inline bool IsStdMulti(Mtl *m) {
   return (m->ClassID()==Class_ID(MULTI_CLASS_ID, 0)) ? true : false; 
}

//----------------------------

class C_MeshMtlList: public Tab<MEntry> {

   static void IncrName(char *name, char *inm, int n){

      char buf[20];
      strncpy(buf,name,13);
      buf[13] = 0;
      sprintf(inm,"%s%d\0",buf,n);
   }

//----------------------------

   static inline Color_24 C24(Color c){

      Color_24 a;
      a.r = (int)(255.0f*c.r);
      a.g = (int)(255.0f*c.g);
      a.b = (int)(255.0f*c.b);
      return a;
   }

//----------------------------

   static inline int Pcnt(float f) { return (int)(f*100.0f+.5f); }

//----------------------------

   static int MAXMapIndex(int i) {

      switch(i) {
         case Ntex:  return ID_DI;
         case Ntex2: return ID_DI;
         case Nopac: return ID_OP;
         case Nbump: return ID_BU;
         case Nspec: return ID_SP;
         case Nshin: return ID_SH;
         case Nselfi:return ID_SI; 
         case Nrefl: return ID_RL;
         default:    return ID_DI;
      }
   }

//----------------------------

   static void ConvertMaxMtlToSMtl(SMtl *s, Mtl *m){

      Mapping *mp;
      Interval v;
      m->Update(0, v);
      s->amb = C24(m->GetAmbient());
      s->diff = C24(m->GetDiffuse());
      s->spec = C24(m->GetSpecular());
      s->shading = 3;
      s->shininess = Pcnt(m->GetShininess());
      s->shin2pct = Pcnt(m->GetShinStr());
      s->shin3pct = 0;
      s->transparency = Pcnt(m->GetXParency());
      s->wiresize = m->WireSize();
      if(m->ClassID()==Class_ID(DMTL_CLASS_ID,0)){
         StdMat* std = (StdMat *)m;
         switch(std->GetShading()) {
            case SHADE_CONST: s->shading = 1; break;
            case SHADE_PHONG: s->shading = 3; break;
            case SHADE_METAL: s->shading = 4; break;
            }  
         s->xpfall = Pcnt(std->GetOpacFalloff(0));    
         s->selfipct = Pcnt(std->GetSelfIllum(0));    
         if (std->GetTwoSided()) s->flags |= MF_TWOSIDE;
         if (std->GetWire()) s->flags |= MF_WIRE;
         if (!std->GetFalloffOut()) s->flags |= MF_XPFALLIN;
         if (std->GetFaceMap()) s->flags |= MF_FACEMAP;
         if (std->GetSoften()) s->flags |= MF_PHONGSOFT;
         if (std->GetWireUnits()) s->flags |= MF_WIREABS;
         switch (std->GetTransparencyType()) {
            case TRANSP_FILTER:
            case TRANSP_SUBTRACTIVE: break;
            case TRANSP_ADDITIVE: s->flags |= MF_ADDITIVE;
            }

         for (int i=0; i<8; i++){
            if (i==Ntex2) continue;
            int n = MAXMapIndex(i);
            Texmap *tx;
            if ((tx=std->GetSubTexmap(n))==NULL) 
               continue;
            if (i==Nrefl) {
               if (tx->ClassID()==Class_ID(ACUBIC_CLASS_ID,0)) {
                  StdCubic *stdcub = (StdCubic*)tx;
                  s->map[i] = mp = NewMapping(i,0);
                  mp->amt.pct = Pcnt(std->GetTexmapAmt(n,0));        
                  mp->use = std->MapEnabled(n);
                  AutoCubicParams &ac = mp->map.p.ref.acb;
                  ac.flags=AC_ON;
                  if (!stdcub->GetDoNth()) ac.flags |= AC_FIRSTONLY;
                  ac.nth = stdcub->GetNth();
                  ac.size = stdcub->GetSize(0);
                  continue;
                  }
               else 
               if (tx->ClassID()==Class_ID(MIRROR_CLASS_ID,0)) {
                  StdMirror *stdmir = (StdMirror*)tx;
                  s->map[i] = mp = NewMapping(i,0);
                  mp->amt.pct = Pcnt(std->GetTexmapAmt(n,0));        
                  mp->use = std->MapEnabled(n);
                  AutoCubicParams &ac = mp->map.p.ref.acb;
                  if (!stdmir->GetDoNth()) ac.flags |= AC_FIRSTONLY;
                  ac.nth = stdmir->GetNth();
                  ac.flags= AC_MIRROR|AC_ON;
                  continue;
                  }
               }
            // just do bitmap textures for now
            if (tx->ClassID()!=Class_ID(BMTEX_CLASS_ID,0))
               continue;               
            BitmapTex *bmt = (BitmapTex*)tx;
            s->map[i] = mp = NewMapping(i,0);
            mp->amt.pct = Pcnt(std->GetTexmapAmt(n,0));        
            /*                   //MB: don't screw up bumpmap's percentage
            if (i==Nbump) 
               mp->amt.pct = abs(mp->amt.pct)/100;
               */
            mp->use = std->MapEnabled(n);
            TSTR filename;
            SplitPathFile(TSTR(bmt->GetMapName()),NULL,&filename);
            CStr name(filename);
            
            // convert bitmap filename to 8.3
            //EightDotThreeName(name);

            // convert to Uppercase
            name.toUpper();

            strncpy(mp->map.map_name, name.data(), MAX_MAP_NAME+1);

            if(i==Nrefl){

            }else{
               StdUVGen *uv = bmt->GetUVGen();
               MapParams &par = mp->map.p.tex;
               par.type = MAP_TYPE_UV;
               par.uoffset = uv->GetUOffs(0); 
               par.voffset = -uv->GetVOffs(0); 
               par.uscale = uv->GetUScl(0); 
               par.vscale = uv->GetVScl(0); 
               float ang = uv->GetAng(0);
               par.ang_sin = -(float)sin(ang);
               par.ang_cos = (float)cos(ang);
               par.texblur  = uv->GetBlur(0)-1.0f;
               int tile = uv->GetTextureTiling();
               if (tile&(U_MIRROR|V_MIRROR))
                  par.texflags|= TEX_MIRROR;
               if (tile&(U_WRAP|V_WRAP))
                  par.texflags &= ~TEX_NOWRAP;
               else
                  par.texflags |= TEX_NOWRAP;
               TextureOutput* texout = bmt->GetTexout();
               if (texout->GetInvert()) 
                  par.texflags |= TEX_INVERT;
               if (bmt->GetFilterType()==FILTER_SAT)
                  par.texflags |= TEX_SAT;
            }
         }
      }
   }

//----------------------------

   void ReallyAddMtl(Mtl *m){

      MEntry me;
      me.sm = NULL;
      me.m = m;
      SMtl *s = new SMtl;
      memset(s, 0, sizeof(SMtl));
      me.sm = s;
      char buf[20];
      strncpy(buf, me.m->GetName(), 16);
      if(strlen(buf)==0) strcpy(buf, "Matl");
      buf[16] = 0;
      strcpy(s->name, buf);
      int n=0;
      while(FindSName(s->name) >=0)
         IncrName(buf,s->name,++n); 
      buf[16] = 0;
      ConvertMaxMtlToSMtl(s, m);
      Append(1, &me, 20);
   }                       

//----------------------------
public:
   void AddMtl(Mtl *m){

      if(!m)
         return;
      Interval v;
      m->Update(0, v);
      if(IsStdMulti(m)){
         for(int i=0; i<m->NumSubMtls(); i++){
            Mtl *sub = m->GetSubMtl(i);
            if(sub&&FindMtl(sub)<0)
               ReallyAddMtl(sub);
         }
      }else{
         if(FindMtl(m)<0) 
            ReallyAddMtl(m);
      }
   }

//----------------------------

   int FindMtl(Mtl *m){

      for(int i=0; i<Count(); i++) 
         if((*this)[i].m == m)
            return i;
      return -1;
   }

//----------------------------

   int FindSName(char *name) const{

      for(int i=0; i<Count(); i++)
         if(strcmp(name,(*this)[i].sm->name)==0)
            return i;
      return -1;
   }

//----------------------------

   ~C_MeshMtlList() {
      for (int i=0; i<Count(); i++) {
         FreeMatRefs((*this)[i].sm);
         delete (*this)[i].sm;
      }
   }
};

//----------------------------
//----------------------------

class SceneEntry{
public:
	TSTR name;
	INode *node, *tnode;
	Object *obj;
	E_OBJTYPE type;
	int id;

   SceneEntry(INode *n, Object *o, E_OBJTYPE t):
      node(n),
      tnode(n->GetTarget()),
      obj(o),
      type(t),
      id(-1)
   {
   }

//----------------------------
   void SetID(int id1){
      assert(id==-1);
      id = id1;
   }
};

//----------------------------
//----------------------------

class C_SceneEnumProc: public ITreeEnumProc{
   C_MeshMtlList *mtlList;

                              //entries automatically sorted by number of parents!
   list<SceneEntry*> entries;
   bool selection_only;

//----------------------------

   class ObjNameList{

      class ObjName{
      public:
         TSTR name;
         ObjName *next;
         ObjName(TSTR n){
            name = n; next = NULL;
         }
      };

   public:
      ObjName *head;
      ObjName *tail;
      int count;

      ObjNameList(){
         head = tail = NULL; count = 0;
      }

      ~ObjNameList(){
         while(head) {
            ObjName *next = head->next;
            delete head;
            head = next;
         }
         head = tail = NULL;
         count = 0;  
      }

      int Count(){ return count; }

   //----------------------------

      int Contains(TSTR &n){

         ObjName *e = head;
         int index = 0;
         while(e) {
            if(e->name == n)
               return index;
            e = e->next;
            index++;
         }
         return -1;
      }

   //----------------------------

      void Append(TSTR &n){

         ObjName *entry = new ObjName(n);
         if(tail)
            tail->next = entry;
         tail = entry;
         if(!head)
            head = entry;
         count++; 
      }

   //----------------------------

      void MakeUnique(TSTR &n){
                                    //First make it less than MAX_NAME_SIZE chars.
         if(n.Length()>MAX_NAME_SIZE) n.Resize(MAX_NAME_SIZE);
   
         if(Contains(n) < 0){
            Append(n);
            return;
         }
                                    //Make it unique and keep it n chars or less.
         for(int i = 0; i < 100000; ++i){
            char buf[MAX_NAME_SIZE + 2];
            sprintf(buf,"%d",i);
            TSTR num(buf);
            TSTR work = n;
            int totlen = num.Length() + work.Length();
            if(totlen > MAX_NAME_SIZE)
               work.Resize(MAX_NAME_SIZE - (totlen - MAX_NAME_SIZE));
            work = work + num;
            if(Contains(work) < 0) {
               Append(work);
               n = work;
               return;
            }
         }
         // Forget it!
      }
   };

//----------------------------

   void Append(INode *node, Object *obj, E_OBJTYPE type){

      entries.push_back(new SceneEntry(node, obj, type));
   }

//----------------------------

   bool Delete(SceneEntry *target){

      list<SceneEntry*>::iterator it;
      it = find(entries.begin(), entries.end(), target);
      if(it==entries.end())
         return false;
      entries.erase(it);
      return true;
   }

//----------------------------

   virtual int callback(INode *node){

      assert(node);
      if(selection_only){
         if(!node->Selected())
            return TREE_CONTINUE;
      }

      Object *obj = node->EvalWorldState(time).obj;
      Class_ID id = obj->ClassID();

                              //NURBS
      if(id==EDITABLE_SURF_CLASS_ID){
         Append(node, obj, OBTYPE_NURBS);
         mtlList->AddMtl(node->GetMtl());
         return TREE_CONTINUE;
      }

      if(obj->CanConvertToType(triObjectClassID)){
         Append(node, obj, OBTYPE_MESH);
         mtlList->AddMtl(node->GetMtl());
         return TREE_CONTINUE;
      }
      if(node->IsTarget()){
         INode* ln = node->GetLookatNode();
         if(ln){
            Object *lobj = ln->EvalWorldState(time).obj;
            switch(lobj->SuperClassID()){
               case LIGHT_CLASS_ID:  Append(node, obj, OBTYPE_LTARGET); break;
               case CAMERA_CLASS_ID: Append(node, obj, OBTYPE_CTARGET); break;
            }
         }
         return TREE_CONTINUE;
      }

      switch (obj->SuperClassID()){
      case HELPER_CLASS_ID:
         if(id==Class_ID(DUMMY_CLASS_ID,0))
            Append(node, obj, OBTYPE_DUMMY);
         else
         if(id==Class_ID(BONE_CLASS_ID,0))
            Append(node, obj, OBTYPE_BONE);
         break;

      case LIGHT_CLASS_ID:
         {
            if(id==Class_ID(OMNI_LIGHT_CLASS_ID,0))
               Append(node, obj, OBTYPE_OMNILIGHT);
            if(id==Class_ID(SPOT_LIGHT_CLASS_ID,0)) 
               Append(node, obj, OBTYPE_SPOTLIGHT);
            if(id==Class_ID(DIR_LIGHT_CLASS_ID, 0)) {
//             Append(node, obj, OBTYPE_FREEDIRECTLIGHT);
            }
            if(id==Class_ID(FSPOT_LIGHT_CLASS_ID,0)){
//             Append(node, obj, OBTYPE_FREESPOTLIGHT);
            }
            if(id==Class_ID(TDIR_LIGHT_CLASS_ID,0)){
               Append(node, obj, OBTYPE_TARGETDIRECTLIGHT);
            }
         }
         break;

      case CAMERA_CLASS_ID:
         Append(node, obj, OBTYPE_CAMERA);
         break;
      }
      return TREE_CONTINUE;
   }

//----------------------------

   void BuildNames(){

      ObjNameList nameList;
      list<SceneEntry*>::iterator it;
      for(it=entries.begin(); it!=entries.end(); it++){
         (*it)->name = (*it)->node->GetName();
         nameList.MakeUnique((*it)->name);
      }
   }

//----------------------------

public:
   TimeValue time;
   
   C_SceneEnumProc(IScene *scene, TimeValue t, C_MeshMtlList *ml, bool sel):
      time(t),
      mtlList(ml),
      selection_only(sel)
   {
      scene->EnumTree(this);
      BuildNames();
   }

//----------------------------

   ~C_SceneEnumProc(){
      list<SceneEntry*>::iterator it;
      for(it=entries.begin(); it!=entries.end(); it++){
         delete (*it);
      }
   }

   inline int Count() { return entries.size(); }
   inline const list<SceneEntry*> &GetEntries() const{ return entries; }

//----------------------------

   SceneEntry *Find(INode *node){

      list<SceneEntry*>::iterator it;
      for(it=entries.begin(); it!=entries.end(); it++){
         if((*it)->node==node)
            return *it;
      }
      return NULL;
   }
};

//----------------------------

class FindDepNodeEnum: public DependentEnumProc{
public:
   ReferenceTarget *targ;
   INode *depNode;

   FindDepNodeEnum(){
      targ = NULL;
      depNode = NULL;
   }

//----------------------------
// proc should return 1 when it wants enumeration to halt.
   virtual int proc(ReferenceMaker *rmaker){

      if(rmaker->SuperClassID()!=BASENODE_CLASS_ID)
         return 0;
      INode* node = (INode *)rmaker;
      if(node->GetTarget()==targ){
         depNode = node;
         return 1;
      }
      return 0;
   }
};

//----------------------------

// We need to maintain a list of the unique objects in the scene
class ObjectEntry{
public:
	TriObject *tri;
	SceneEntry *entry;

   ObjectEntry(SceneEntry *e):
      entry(e),
      tri(NULL)
   {
   }
};

//----------------------------

class C_ObjectList{
   list<ObjectEntry*> entries;

   void Append(SceneEntry *e){
      entries.push_back(new ObjectEntry(e));
   }

public:
   C_ObjectList(C_SceneEnumProc &scene){
                              //zip thru the object list and record all unique objects
                              // (Some may be used by more than one node)
      const list<SceneEntry*> &entries = scene.GetEntries();
      list<SceneEntry*>::const_iterator it;
      for(it=entries.begin(); it!=entries.end(); it++){

                              //can't multiple instance lights and cameras
                              // make them all unique
         if(((*it)->type!=OBTYPE_MESH)|| !Contains((*it)->obj))
            Append((*it));
      }
   }

//----------------------------

   ~C_ObjectList(){

      list<ObjectEntry*>::iterator it;
      for(it=entries.begin(); it!=entries.end(); it++){
         delete (*it);
      }
   }

//----------------------------

   ObjectEntry *Contains(Object *obj) const{

      list<ObjectEntry*>::const_iterator it;
      for(it=entries.begin(); it!=entries.end(); it++){
         if((*it)->entry->obj==obj)
            return *it;
      }
      return NULL;
   }

//----------------------------

   ObjectEntry *FindLookatNode(INode *node){

      FindDepNodeEnum finder;

      list<ObjectEntry*>::const_iterator it;
      for(it=entries.begin(); it!=entries.end(); it++){
         finder.targ = node;
         (*it)->entry->node->EnumDependents(&finder);
         if(finder.depNode)
            return *it;
      }
      return NULL;
   }

//----------------------------

   inline const list<ObjectEntry*> &GetEntries() const{ return entries; }
};

//----------------------------

static Tab<int> mtlNumMap;
static int mtlnum;
static BOOL multiMtl;


struct S_rgb{
   byte b, g, r;
};

//----------------------------
// Main routine for writing a mesh file.
// Recursive chunk writer -- keeps track of each chunk.
static bool DumpMeshChunk(USHORT tag, void *data, S_dump_context &dc){

   dc.ck->WAscend(tag);

   switch(tag){
   case MMAGIC:
      {
                              //save materials
         for(int ix=0; ix<dc.materials->Count(); ix++){
            savemtl = (*dc.materials)[ix].sm;
            if(DumpMaterialChunk(MAT_ENTRY, NULL, dc)==0)
               return false;
         }

         const list<ObjectEntry*> &entries = dc.objects->GetEntries();
         for(list<ObjectEntry*>::const_iterator it=entries.begin(); it!=entries.end(); it++){
            ObjectEntry *oe = (*it);

            switch(oe->entry->type){
            case OBTYPE_DUMMY:
            case OBTYPE_LTARGET:
            case OBTYPE_CTARGET:
               break;
            case OBTYPE_MESH:
               {
                  Object *obj = oe->entry->obj;
                  oe->tri = (TriObject *)obj->ConvertToType(dc.i->GetTime(), triObjectClassID);
                  Mesh &mesh = oe->tri->mesh;
                  if(mesh.numFaces > 65530)
                     continue;
                  if(mesh.numVerts > 65530)
                     continue;
               }
                              //flow...
            default:
               {
                  if(!DumpMeshChunk(NAMED_OBJECT, oe, dc))
                     return false;
               }
            }
         }
      }
      break;

   case NAMED_OBJECT:
      {
         ObjectEntry *oe = (ObjectEntry *)data;
         //INode *node = oe->entry->node;       
         CStr name = oe->entry->name;
         /*
         if(name.Length()>MAX_NAME_SIZE)
            name.Resize(MAX_NAME_SIZE);
            */
         dc.ck->Write(name.data(), name.Length()+1);

         switch(oe->entry->type){
         case OBTYPE_MESH:
            if(!DumpMeshChunk(N_TRI_OBJECT, oe, dc))
               return false;
            break;

         case OBTYPE_CAMERA:
            if(!DumpMeshChunk(N_CAMERA, oe, dc))
               return false;
            break;

         case OBTYPE_OMNILIGHT:
            if(!DumpMeshChunk(N_DIRECT_LIGHT, oe, dc))
               return false;
            break;

         case OBTYPE_SPOTLIGHT:
            if(!DumpMeshChunk(N_DIRECT_LIGHT, oe, dc))
               return false;
            break;

         case OBTYPE_FREEDIRECTLIGHT:
            return false;

         case OBTYPE_FREESPOTLIGHT:
            return false;

         case OBTYPE_TARGETDIRECTLIGHT:
            if(!DumpMeshChunk(N_DIRECT_LIGHT_TARGET, oe, dc))
               return false;
            break;

         case OBTYPE_NURBS:
            if(!DumpMeshChunk(N_NURBS, oe, dc))
               return false;
            break;

         case OBTYPE_BONE:
            if(!DumpMeshChunk(N_BONE, oe, dc))
               return false;
            break;

         default:
            assert(0);
            break;
         }
//       if(n->flags & NO_HIDDEN){
//          if(dump_mchunk(OBJ_HIDDEN, n)==0)
//             return(0);
//       }
//       if(n->flags & NO_DOESNT_CAST) {
//          if(dump_mchunk(OBJ_DOESNT_CAST, n)==0)
//             return(0);
//       }     
//       if(n->flags & NO_MATTE){
//          if(dump_mchunk(OBJ_MATTE, n)==0)
//             return(0);
//       }     
//       if(n->flags & NO_DONT_RCVSHADOW) {
//          if(dump_mchunk(OBJ_DONT_RCVSHADOW, n)==0)
//             return(0);
//       }     
//       if(n->flags & NO_FAST){
//          if(dump_mchunk(OBJ_FAST, n)==0)
//             return(0);
//       }
//       if(n->flags & NO_FROZEN){
//          if(dump_mchunk(OBJ_FROZEN, n)==0)
//             return(0);
//       }
      }
      break;

   case N_TRI_OBJECT:
      {
         ObjectEntry *oe = (ObjectEntry*)data;
         INode *node = oe->entry->node;
         /*
         Matrix3 ident(TRUE);
         Matrix3 mat = node->GetNodeTM(dc.scene_enum->time);
            
         // RB: For some reason this fixes a problem where
         // the pivot gets screwed up if the object is at 
         // the origin. ???!!!         
         Point3 pos = mat.GetRow(3);
         if(pos.x==0.0f && pos.y==0.0f && pos.z==0.0f){
            pos.x = 0.001f;
            pos.y = 0.001f;
            pos.z = 0.001f;
            mat.SetTrans(pos);
         }
         */
         Object *obj = oe->entry->obj;
         Mesh &mesh = oe->tri->mesh;
         MeshAndNode mt;
         mt.mesh = &mesh;
         mt.node = node;

         // RB 6/11/96: This is kindof a hack, but if the node has
         // WSMs applied then it will already be transformed into
         // world space and so it doesn't need to be transformed again.
         /*
         if(node->GetProperty(PROPID_HAS_WSM)){
            mt.tm = &ident;
         }
         */

         // if( (t->flags & TRI_HAS_PROCDATA) && t->proc_data!=NULL) {
         //     if(dump_mchunk(PROC_DATA, t)==0)
         //        return(0);
         // }
         // if(strlen(t->proc_name)>0) {
         //     if(dump_mchunk(PROC_NAME, t)==0)
         //        return(0);
         // }

         {
            if(mesh.getNumVerts()){
               if(!DumpMeshChunk(POINT_ARRAY, &mesh, dc))
                  return false;
            }
            if(mesh.getNumFaces()){
               if(!DumpMeshChunk(FACE_ARRAY, &mt, dc))
                  return false;
            }
                              //write vertex colors
            if(mesh.getNumVerts() && mesh.mapSupport(0)){
               if(!DumpMeshChunk(MESH_COLOR, &mesh, dc))
                  return false;
            }

                              //export mapping channels
            for(int i=1; i<mesh.getNumMaps(); i++){
               if(!mesh.mapSupport(i))
                  continue;
               struct S_hlp{
                  ObjectEntry *oe;
                  int num;
               } hlp = {oe, i};
               if(!DumpMeshChunk(FACE_MAP_CHANNEL, &hlp, dc))
                  return false;
            }
         }

                              //Delete the working object, if necessary
         if(obj != (Object *)(oe->tri)){
            oe->tri->DeleteThis();
            oe->tri = NULL;
         }
      }
      break;

   case POINT_ARRAY:
      {
         Mesh *mesh = (Mesh*)data;

         word count = mesh->getNumVerts();
         dc.ck->Write(&count, 2);

         for(int i=0; i<count; ++i)
            dc.ck->Write(&mesh->verts[i], sizeof(S_vector));
      }
      break;

   case FACE_MAP_CHANNEL:
      {
         struct S_hlp{
            ObjectEntry *oe;
            int num;
         } *hlp = (S_hlp*)data;
         Mesh &mesh = hlp->oe->tri->mesh;
         const UVVert *uvverts = mesh.mapVerts(hlp->num);
         int num_uv_verts = mesh.getNumMapVerts(hlp->num);
         const TVFace *uvfaces = mesh.mapFaces(hlp->num);
         int num_faces = mesh.numFaces;

         int i;
                              //save map channel index
         dc.ck->Write(&hlp->num, sizeof(int));

                              //save map vertices
         dc.ck->Write(&num_uv_verts, sizeof(word));
         for(i=0; i<num_uv_verts; i++){
            dc.ck->Write(&uvverts[i].x, sizeof(float));
            dc.ck->Write(&uvverts[i].y, sizeof(float));
         }

                              //save map faces
         dc.ck->Write(&num_faces, sizeof(word));
         for(i=0; i<num_faces; i++){
            for(int j=0; j<3; j++){
               dc.ck->Write(&uvfaces[i].t[j], sizeof(word));
            }
         }
      }
      break;

   case MESH_COLOR:
      {
         Mesh *mesh = (Mesh*)data;
         const UVVert *uvverts = mesh->mapVerts(0);
         const TVFace *uvfaces = mesh->mapFaces(0);

                              //re-map UV-style color mapping to per-vertex color
         int num_verts = mesh->getNumVerts();
         int num_faces = mesh->getNumFaces();
         vector<S_vector> v_color(num_verts, S_vector(0, 0, 0));
         vector<bool> v_used(num_verts, false);
         for(int i=num_faces; i--; ){
            const Face &f = mesh->faces[i];
            const TVFace &cf = uvfaces[i];
            for(int j=3; j--; ){
               int vi = f.v[j];
               int ci = cf.t[j];
               const S_vector &color = *(S_vector*)&uvverts[ci];
               if(!v_used[vi]){
                  v_color[vi] = color;
                  v_used[vi] = true;
               }else{
                  v_color[vi] = (v_color[vi] + color) * .5f;
               }
            }
         }
                              //convert S_vector to RGB triplet
         vector<S_rgb> rgb_colors(num_verts);
         for(i=num_verts; i--; ){
            const S_vector &src = v_color[i];
            S_rgb &dst = rgb_colors[i];
            dst.r = Max(0, Min((int)(src.x*255.0f), 255));
            dst.g = Max(0, Min((int)(src.y*255.0f), 255));
            dst.b = Max(0, Min((int)(src.z*255.0f), 255));
         }
         dc.ck->Write(rgb_colors.begin(), num_verts*sizeof(S_rgb));
      }
      break;

   case FACE_ARRAY: 
      {
         MeshAndNode *mt = (MeshAndNode*)data;
         Mesh *m = mt->mesh;
         word count = m->getNumFaces();
         dc.ck->Write(&count, 2);
         bool anySmooth = false;
         for(int ix=0; ix<count; ++ix){
            const Face &f = m->faces[ix];
            if(f.smGroup)
               anySmooth = true;

            fc_wrt Fc_wrt;
            Fc_wrt.a = (unsigned short)f.v[0];
            Fc_wrt.b = (unsigned short)f.v[1];
            Fc_wrt.c = (unsigned short)f.v[2];
            Fc_wrt.flags = 0;
            if(f.flags & EDGE_A) Fc_wrt.flags |= ABLINE;
            if(f.flags & EDGE_B) Fc_wrt.flags |= BCLINE;
            if(f.flags & EDGE_C) Fc_wrt.flags |= CALINE;
            dc.ck->Write(&Fc_wrt, 8);
         }

         Mtl *mtl;
         if(NULL!=(mtl=mt->node->GetMtl())){
            if(IsStdMulti(mtl)){
               multiMtl = true;
                              //make a table that maps sub-mtl number to the proper index
                              // into dc.materials
               mtlNumMap.SetCount(mtl->NumSubMtls());
               for(int i=0; i<mtl->NumSubMtls(); i++){
                  Mtl *sub  = mtl->GetSubMtl(i);
                  if(sub){
                     int mat_i = dc.materials->FindMtl(sub);
                              //check if material already present
                     for(int j=i; j--; ){
                        if(mtlNumMap[j]==mat_i)
                           break;
                     }
                              //material already there? name duplication, forget second
                              // instance
                     if(j!=-1)
                        mat_i = -1;
                     mtlNumMap[i] = mat_i;
                  }else
                     mtlNumMap[i] = -1;
               }
               for(i=0; i<mtl->NumSubMtls(); i++){
                  Mtl *sub  = mtl->GetSubMtl(i);
                  if(sub){
                     mtlnum = mtlNumMap[i];
                     if(mtlnum!=-1){
                        if(!DumpMeshChunk(MSH_MAT_GROUP, m, dc))
                           return false;
                     }
                  }
               }
            }else{
               mtlnum = dc.materials->FindMtl(mtl);
               multiMtl = false;
               if(!DumpMeshChunk(MSH_MAT_GROUP, m, dc))
                  return false;
            } 
         }

                              //save smoothing groups if any
         if(anySmooth){
            if(!DumpMeshChunk(SMOOTH_GROUP, m, dc))
               return false;
         }
      }
      break;

   case MSH_MAT_GROUP:
      {
         Mesh *m = (Mesh*)data;
         SMtl *mtl;
         short temp;
         assert(mtlnum>=0 && mtlnum<dc.materials->Count());
         mtl = (*dc.materials)[mtlnum].sm;
         assert(mtl);
         word count;

         if(multiMtl){
            count = 0;
            for(int ix=0; ix<m->numFaces; ix++){

               int faceIndex = m->getFaceMtlIndex(ix);

               int mtlCnt = mtlNumMap.Count(); // this is to prevent compilier warnings
               faceIndex = (faceIndex % mtlCnt);

               int mtlIndex = mtlNumMap[faceIndex];
               if(mtlIndex==mtlnum)
                  count++;
            }
         }else
            count = m->numFaces;

         dc.ck->Write(mtl->name, strlen(mtl->name) + 1);
         temp = (short)count;
         dc.ck->Write(&temp, 2);

         if(multiMtl){
            for(int ix=0; ix<m->numFaces; ++ix) {

                    int faceIndex = m->getFaceMtlIndex(ix);

                    int mtlCnt = mtlNumMap.Count(); // this is to prevent compilier warnings
                    faceIndex = (faceIndex % mtlCnt);

               int mtlIndex = mtlNumMap[faceIndex];
               if (mtlIndex==mtlnum) {
                  temp = (short)ix;
                  dc.ck->Write(&temp, 2);
                  if(--count==0)
                     break;
                  }
               }
         }else {
            for(int ix=0; ix<m->numFaces; ++ix){
               temp = (short)ix;
               dc.ck->Write(&temp, 2);
            }
         }
      }
      break;

   case SMOOTH_GROUP:
      {
         Mesh *m = (Mesh *)data;

                              //now dump all object faces' smoothing group flags
         int count = m->getNumFaces();
         for(int ix=0; ix<count; ++ix) {
            Face f = m->faces[ix];
            dc.ck->Write(&f.smGroup, 4);
         }
      }
      break;

   case N_BONE:
      {
         ObjectEntry *oe = (ObjectEntry *)data;
         INode *node = oe->entry->node;
         
         Matrix3 mat = node->GetNodeTM(dc.scene_enum->time);
         Point3 pos = mat.GetRow(3);

         dc.ck->Write(&mat, 48);
      }
      break;

   case N_NURBS:
      {
         ObjectEntry *oe = (ObjectEntry *)data;
         INode *node = oe->entry->node;
         Object *obj = oe->entry->obj;
         Matrix3 mat = node->GetNodeTM(dc.scene_enum->time);       
         Point3 pos = mat.GetRow(3);

         Object* xobj = node->GetObjectRef();

         Class_ID c = xobj->ClassID();
         SClass_ID s = xobj->SuperClassID();

         
         IDerivedObject *xobjderived=NULL;
         int nummods=0;

         if(s == GEN_DERIVOB_CLASS_ID){
            xobjderived = (IDerivedObject *)xobj;
            
            
            xobj = xobj->FindBaseObject();


            //bool iii = xobj->IsWorldSpaceObject();

            nummods = xobjderived->NumModifiers();

            //-----------------------------for each modifier
            for(int i=0;i<nummods;i++){
               Modifier *mod = xobjderived->GetModifier(i);
               ModContext *modContext = xobjderived->GetModContext(i);

               //ObjectState *os=new ObjectState(obj);
               ObjectState os(obj);

               INode *nd=NULL;

               Class_ID id = mod->InputType();

               if(id.PartA() == MAPPABLE_OBJ_CLASS_ID){
//                  int a = theSceneEnum->i->GetTime();
                  mod->ModifyObject(0, *modContext, &os, nd);
                  //xobjderived->AddModifier(mod); 
                  //xobjderived->DeleteModifier(0); 
                  }//if

               }//for

            }//derived object

            NURBSSet getSet;
            BOOL okay = GetNURBSSet(obj, dc.i->GetTime(), getSet, FALSE);
            if (okay) {
               NURBSCVSurface *nurbs;
               int numObj = getSet.GetNumObjects();
               if(!numObj){
                  /*
                  char buf[64];
                  sprintf(buf, "Nurbs object unactive. (%s)", node->GetName());
                  AddToMsgList(msgList, buf);
                  */
                  return false;
               }

               nurbs = (NURBSCVSurface *)getSet.GetNURBSObject(0);

       /*        if(dump_mchunk(NURBS_UV_ORDER, nurbs)==0)
                  return(0);
               if(dump_mchunk(NURBS_CVS, nurbs)==0)
                  return(0);
               if(dump_mchunk(NURBS_KNOTS, nurbs)==0)
                  return(0);
               if(dump_mchunk(NURBS_TRANSFORM_MATRIX, nurbs)==0)
                  return(0);
               if(dump_mchunk(NURBS_FLIP_NORMALS, nurbs)==0)
                     return(0);

               Mtl *mat = node->GetMtl();

               if(mat){
                  if(dump_mchunk(NURBS_MAT, mat)==0)
                     return(0);
                  if(dump_mchunk(NURBS_UV_MAP, nurbs)==0)
                     return(0);
                  }
         */
              }//if (okay)
      }
      break;

   case NURBS_UV_MAP:
      {
         NURBSCVSurface *nurbs = (NURBSCVSurface *)data;

         float uoffset;
         float voffset;
         float utile;
         float vtile;

                              //upgrade to v3.1
           // nurbs->GetTileOffset(theSceneEnum->i->GetTime(), utile, vtile, uoffset, voffset);
          
         dc.ck->Write(&uoffset, sizeof(float));
         dc.ck->Write(&voffset, sizeof(float));
         dc.ck->Write(&utile, sizeof(float));
         dc.ck->Write(&vtile, sizeof(float));

         Point2 pt;
         //   bool b = nurbs->GenerateUVs();
          //  nurbs->SetGenerateUVs(true);
          //  b = nurbs->GenerateUVs();

         for(int j = 0;j<4;j++){
            pt = nurbs->GetTextureUVs(dc.i->GetTime(), j);
                //  pt = nurbs->GetTextureUVs(theSceneEnum->i->GetTime(), j, 1);
            dc.ck->Write(&pt.x, sizeof(float));
            dc.ck->Write(&pt.y, sizeof(float));
         }

//Nurbs texture (ALL TO DELETE)

                              //v3.1
          /*     NURBSTextureSurface *texturesurf = &nurbs->mTexSurface[0];

               bool clp    = texturesurf->IsCLP();

               if(clp){
                  int uorder  = texturesurf->GetUOrder();
                  //uorder  = nurbs->GetUOrder();
                  int vorder  = texturesurf->GetVOrder();
                  //uorder  = nurbs->GetUOrder();
                  int uknots  = texturesurf->GetNumUKnots();
                  //uknots=nurbs->GetNumUKnots();

                  int vknots  = texturesurf->GetNumVKnots();
                  //vknots=nurbs->GetNumUKnots();
                  int ucvs    = texturesurf->GetNumUCVs();
                 // ucvs    = nurbs->GetNumUCVs();
                  int vcvs    = texturesurf->GetNumVCVs();
                  //vcvs    = nurbs->GetNumVCVs();

                  int uvalue, vvalue;

                  for(int i=0; i<uknots;i++){
//                     uvalue = texturesurf->GetUKnot(i);
                     }

                  for(j=0; j<vknots;j++){
//                     vvalue = texturesurf->GetVKnot(j);
                     }

                  NURBSControlVertex *cv;

                  for(i = 0; i < ucvs;i++)
                     for(j = 0; j < vcvs;j++){
                       // cv = texturesurf->GetCV(i, j);
                        }//for
                  }//not clp
      */
      }
      break;

   case NURBS_FLIP_NORMALS:
      {
         NURBSCVSurface *nurbs = (NURBSCVSurface *)data;
         bool FlipNormals;
         if(nurbs->FlipNormals())
            FlipNormals = true;
         else
            FlipNormals = false;
         dc.ck->Write(&FlipNormals, sizeof(bool));
      }
      break;

   case NURBS_MAT:
      {
         Mtl *mat = (Mtl *)data;
         char *name = mat->GetName();
         dc.ck->Write(name, strlen(name)+1);
      }
      break;

   case NURBS_TRANSFORM_MATRIX:
      {
         NURBSCVSurface *nurbs = (NURBSCVSurface *)data;
         Matrix3 TransformMatrix = nurbs->GetTransformMatrix(dc.i->GetTime());
         dc.ck->Write(&TransformMatrix, sizeof(Matrix3));
      }
      break;

   case NURBS_KNOTS:
      {
         NURBSCVSurface *nurbs = (NURBSCVSurface *)data;
         int numuknots, numvknots;
         
         numuknots = nurbs->GetNumUKnots();
         numvknots = nurbs->GetNumVKnots();
         dc.ck->Write(&numuknots, sizeof(int));
         dc.ck->Write(&numvknots, sizeof(int));
         
         double knot;
         for(int i=0;i<numuknots;i++){
            knot = nurbs->GetUKnot(i);
            dc.ck->Write(&knot,sizeof(double));
         }
         for(i=0;i<numvknots;i++){
            knot = nurbs->GetVKnot(i);
            dc.ck->Write(&knot,sizeof(double));
         }
      }
      break;

   case NURBS_UV_ORDER:
      {
         NURBSCVSurface *nurbs = (NURBSCVSurface *)data;
         int uorder = nurbs->GetUOrder();
         int vorder = nurbs->GetVOrder();
         
         dc.ck->Write(&uorder, sizeof(int));
         dc.ck->Write(&vorder, sizeof(int));

      }
      break;

   case NURBS_CVS:
      {
         int x, y;
         float u, v, w, weight;
         NURBSControlVertex* cv;

         NURBSCVSurface *nurbs = (NURBSCVSurface *)data;

         nurbs->GetNumCVs(x, y);
         
         dc.ck->Write(&x, sizeof(int));
         dc.ck->Write(&y, sizeof(int));

         for(int i=0;i<x;i++){
            for(int j=0;j<y;j++){
               cv = nurbs->GetCV(i, j);

               cv->GetPosition(dc.i->GetTime(), u, v, w);
               cv->GetWeight(dc.i->GetTime(), weight);

               dc.ck->Write(&u,sizeof(float));
               dc.ck->Write(&v,sizeof(float));
               dc.ck->Write(&w,sizeof(float));
               dc.ck->Write(&weight,sizeof(float));
            }
         }
      }
      break;

   case N_DIRECT_LIGHT_TARGET:
   case N_DIRECT_LIGHT:
      {
         ObjectEntry *oe = (ObjectEntry *)data;
         INode *node = oe->entry->node;
         Object *obj = oe->entry->obj;
         LightObject *lt= (LightObject *)obj;
         LightState ls;
         Interval valid;
         lt->EvalLightState(dc.scene_enum->time, valid, &ls);
         Matrix3 mat = node->GetNodeTM(dc.scene_enum->time);       
         Point3 pos = mat.GetRow(3);
         dc.ck->Write(&pos, sizeof(S_vector));

         dc.ck->WVectorChunk(COLOR_F, *(S_vector*)&ls.color);
         dc.ck->WFloatChunk(DL_OUTER_RANGE, ls.attenEnd);
         dc.ck->WFloatChunk(DL_INNER_RANGE, ls.attenStart);
         dc.ck->WFloatChunk(DL_MULTIPLIER, ls.intens);

         switch(tag){
         case N_DIRECT_LIGHT_TARGET:
            if(!DumpMeshChunk(DL_DIRLIGHT, data, dc))
               return false;
            break;
         case N_DIRECT_LIGHT:
            if(oe->entry->type==OBTYPE_SPOTLIGHT){
               if(!DumpMeshChunk(DL_SPOTLIGHT, data, dc))
               return false;
            }
            break;
         }

         if(!ls.on){
            if(!DumpMeshChunk(DL_OFF, data, dc))
               return false;
         }
         if(ls.useAtten) {
            if(!DumpMeshChunk(DL_ATTENUATE, data, dc))
               return false;
         }
      }    
      break;

   case DL_DIRLIGHT:
      {
         ObjectEntry *oe = (ObjectEntry *)data;
         INode *tnode = oe->entry->tnode;
         Object *obj = oe->entry->obj;

         LightObject *lt= (LightObject *)obj;

         LightState ls;
         Interval valid;
         lt->EvalLightState(dc.scene_enum->time, valid, &ls);

         Matrix3 mat = tnode->GetNodeTM(dc.scene_enum->time);         
         Point3 pos = mat.GetRow(3);
         dc.ck->Write(&pos, sizeof(S_vector));
         dc.ck->Write(&ls.hotsize,sizeof(float));
         dc.ck->Write(&ls.fallsize,sizeof(float));
      }
      break;

   case DL_SPOTLIGHT: 
      {
         ObjectEntry *oe = (ObjectEntry *)data;
         INode *tnode = oe->entry->tnode;
         Object *obj = oe->entry->obj;

         if(obj->ClassID()==Class_ID(DIR_LIGHT_CLASS_ID, 0)){
            break;
         }

         if(obj->ClassID()==Class_ID(FSPOT_LIGHT_CLASS_ID,0)){
            break;
         }
         LightObject *lt = (LightObject*)obj;
         LightState ls;
         Interval valid;
         lt->EvalLightState(dc.scene_enum->time, valid, &ls);
         Point3 pos;
         if(tnode){
            Matrix3 mat = tnode->GetNodeTM(dc.scene_enum->time);
            pos = mat.GetRow(3);
         }else{
            pos.x = 0;
            pos.y = 0;
            pos.z = 0;
         }
         dc.ck->Write(&pos, sizeof(S_vector));
         dc.ck->Write(&ls.hotsize,sizeof(float));
         dc.ck->Write(&ls.fallsize,sizeof(float));
      }
      break;

   case DL_OFF:               //special null chunk
   case DL_ATTENUATE:
      break;

   case N_CAMERA: 
      {
         ObjectEntry *oe = (ObjectEntry *)data;
         INode *node = oe->entry->node;
         CameraObject* camobj = (CameraObject *)oe->entry->obj;
         GenCamera *gc = (GenCamera*)oe->entry->obj;
         CameraState cs;
         Interval valid;
         camobj->EvalCameraState(dc.scene_enum->time, valid, &cs);
         CameraI3D c;
         Matrix3 mat = node->GetNodeTM(dc.scene_enum->time);       
         Point3 pos = mat.GetRow(3);
         c.x = pos.x;
         c.y = pos.y;
         c.z = pos.z;
         if(oe->entry->tnode){
            mat = oe->entry->tnode->GetNodeTM(dc.scene_enum->time);         
            pos = mat.GetRow(3);
         }
         c.tx = pos.x;
         c.ty = pos.y;
         c.tz = pos.z;
         c.bank = 0.0f;
         c.focal = 2400.0f/RadToDeg(cs.fov);
         //c.flags = 0;
         //c.nearplane = gc->GetClipDist(0,0);
         //c.farplane = gc->GetClipDist(0,1);
         //WERR(&c, sizeof(c), fp);
         dc.ck->Write(&c, sizeof(c));

         if(!DumpMeshChunk(CAM_RANGES, gc, dc))
            return false;
      }                
      break;

   case CAM_RANGES:
      {
         GenCamera *gc = (GenCamera *)data;
         float cnear = gc->GetEnvRange(0,0);
         float cfar = gc->GetEnvRange(0,1);
         dc.ck->Write(&cnear, sizeof(float));
         dc.ck->Write(&cfar, sizeof(float));
      }
      break;

   case DUMMY:
      break;

   default:
      assert(0);
      break;
   }

   dc.ck->Descend();
   return true;
}

//----------------------------

static bool WriteStd1KeyHdr(S_dump_context &dc){

   short trackflags = 0;
   dc.ck->Write(&trackflags, sizeof(short));

   long lwork = 0;
                              //Track min
   dc.ck->Write(&lwork, sizeof(dword));

                              //Track max
   dc.ck->Write(&lwork, sizeof(dword));

   long nkeys = 1;
   dc.ck->Write(&nkeys, sizeof(dword));

   long keytime = 0;
   dc.ck->Write(&keytime, sizeof(dword));

   short wflags = 0;    // Ignoring tension, continuity, bias, eases
   dc.ck->Write(&wflags, sizeof(short));
   return true;
}

//----------------------------

static bool IsTCBContol(Control *cont){

   return(
      cont->ClassID()==Class_ID(TCBINTERP_FLOAT_CLASS_ID,0) ||
      cont->ClassID()==Class_ID(TCBINTERP_POSITION_CLASS_ID,0) ||
      cont->ClassID()==Class_ID(TCBINTERP_ROTATION_CLASS_ID,0) ||
      cont->ClassID()==Class_ID(TCBINTERP_POINT3_CLASS_ID,0) ||
      cont->ClassID()==Class_ID(TCBINTERP_SCALE_CLASS_ID,0));
}

//----------------------------

static bool WriteControllerChunk(S_dump_context &dc, Control *cont, int type, float scale = 1.0f){

   ITCBFloatKey fkey;
   ITCBPoint3Key pkey;
   ITCBRotKey rkey;
   ITCBScaleKey skey;
   ITCBKey *k; 
   int num = cont->NumKeys();
   float fval;
   Point3 pval;
   Quat q, qLast = IdentQuat();
   AngAxis rval;
   ScaleValue sval;
   Interval valid;
   TimeValue t;

                              //set up 'k' to point at the right derived class
   switch(type){
   case KEY_FLOAT: k = &fkey; break;
   case KEY_POS:   k = &pkey; break;
   case KEY_ROT:   k = &rkey; break;
   case KEY_SCL:   k = &skey; break;
   case KEY_COLOR: k = &pkey; break;
   default: return false;
   }

                              //get the keyframe interface
   IKeyControl *ikeys = GetKeyControlInterface(cont);
   
                              //gotta have some keys
   if(num==NOT_KEYFRAMEABLE || num==0){
      return false;
   }
   
                              //write track some stuff
   short trackflags = 0;
   dc.ck->Write(&trackflags, sizeof(short));

   long lwork = cont->GetKeyTime(0) / GetTicksPerFrame();
                              //track min
   dc.ck->Write(&lwork, sizeof(dword));

   lwork = cont->GetKeyTime(num-1) / GetTicksPerFrame();
                              //track max
   dc.ck->Write(&lwork, sizeof(dword));

   long nkeys = num;
   dc.ck->Write(&nkeys, sizeof(dword));

   for(int i=0; i<num; i++){
      if(IsTCBContol(cont) && ikeys){
         ikeys->GetKey(i,k);

                              //write key time
         long keytime = k->time / GetTicksPerFrame();
         dc.ck->Write(&keytime, sizeof(dword));
         
                              //write flags
         short wflags = 0;
         if(k->tens   != 0.0f) wflags |= W_TENS;
         if(k->cont   != 0.0f) wflags |= W_CONT;
         if(k->bias   != 0.0f) wflags |= W_BIAS;
         if(k->easeIn != 0.0f) wflags |= W_EASETO;
         if(k->easeOut!= 0.0f) wflags |= W_EASEFROM;       
         dc.ck->Write(&wflags, sizeof(short));
         
                              //write TCB and ease
         if(k->tens   != 0.0f) dc.ck->Write(&k->tens, sizeof(float));
         if(k->cont   != 0.0f) dc.ck->Write(&k->cont, sizeof(float));
         if(k->bias   != 0.0f) dc.ck->Write(&k->bias, sizeof(float));
         if(k->easeIn != 0.0f) dc.ck->Write(&k->easeIn, sizeof(float));
         if(k->easeOut!= 0.0f) dc.ck->Write(&k->easeOut, sizeof(float));
         
                              //write values
         switch(type){
         case KEY_FLOAT: 
            fkey.val *= scale;
            dc.ck->Write(&fkey.val, sizeof(float));
            break;
            
         case KEY_SCL:
            dc.ck->Write(&skey.val.s, sizeof(S_vector));
            break;

         case KEY_COLOR:
         case KEY_POS:
            dc.ck->Write(&pkey.val, sizeof(S_vector));
            break;

         case KEY_ROT:
            dc.ck->Write(&rkey.val.angle, sizeof(float));
            dc.ck->Write(&rkey.val.axis, sizeof(S_vector));
            break;
         }

      }else{
                              //sample the control at each key time
         t = cont->GetKeyTime(i);

                              //write key time
         long keytime = t / GetTicksPerFrame();
         dc.ck->Write(&keytime, sizeof(dword));
         
                              //write flags
         short wflags = 0;       
         dc.ck->Write(&wflags, sizeof(short));
         
                              //write values
         switch(type){
         case KEY_FLOAT: 
            cont->GetValue(t,&fval,valid);
            fval*=scale;
            dc.ck->Write(&fval, sizeof(float));
            break;
         
         case KEY_SCL:
            cont->GetValue(t, &sval,valid);
            dc.ck->Write(&sval.s, sizeof(S_vector));
            break;

         case KEY_COLOR:
         case KEY_POS:
            cont->GetValue(t,&pval,valid);
            dc.ck->Write(&pval, sizeof(S_vector));
            break;

         case KEY_ROT:
            cont->GetValue(t,&q,valid);
            rval = AngAxis(q/qLast);
                 qLast = q;
            dc.ck->Write(&rval.angle, sizeof(float));
            dc.ck->Write(&rval.axis, sizeof(S_vector));
            break;
         }
      }
   }
   return true;
}

//----------------------------

static bool WriteFloatTrack(Control *c, float scale, S_dump_context &dc){

   Interval valid;
   if(!WriteControllerChunk(dc, c, KEY_FLOAT, scale)){
      float v;
      c->GetValue(0,&v,valid);
      v *= scale;
      if(!WriteStd1KeyHdr(dc))
         return false;
      dc.ck->Write(&v, sizeof(float));
   }
   return true;
}

//----------------------------

static bool WriteFloatTrackOrStd1KeyHdr(Control *c, float scale, S_dump_context &dc){

   if(c){
      if(!WriteFloatTrack(c, scale, dc))
         return false;
   }else{
      if(!WriteStd1KeyHdr(dc))
         return false;
      dc.ck->Write(&floatVal, sizeof(float));
   }
   return true;
}

//----------------------------
// Remove the rotational and scale comonents from the max pivot
// to get the pivot offset.
static Point3 GetPivotOffset(INode* node){

                              //get the pivot offset and remove the rotational/scale component
    Matrix3 mat(TRUE);
    Quat qRot = node->GetObjOffsetRot();
    qRot.MakeMatrix(mat);
                              //max scales the object if the pivot is scaled so skip.
                              //m.SetScale( ((SceneEntry *)data)->node->GetObjOffsetScale().s );
    Point3 pivot = -node->GetObjOffsetPos();
    mat = Inverse(mat);
    Point3 pOff = VectorTransform(mat, pivot);
    return pOff;
}

//----------------------------
// Remove the offset effects of the parent from the child.
static bool UndoParentsOffset(INode* node, Point3& pt, Quat& rOff){

   INode* parent;
   Matrix3 mat(TRUE);
   Quat pOff;
   if(node && pt && rOff){
      if(!node->GetParentNode()->IsRootNode()){
         parent = node->GetParentNode();
         pOff   = parent->GetObjOffsetRot();
         if(pOff == IdentQuat())
            return false;
         pOff.MakeMatrix(mat);
         mat   = Inverse(mat);
         pt    = VectorTransform(mat, pt);
         rOff  = rOff/pOff;
         return true;
      }
   }
   return false;
}

//----------------------------
// See if we can write TCB keys instead of generating keys by sampling frames.
static bool IsTCBControl(Control *cont){

   return(
      cont->ClassID()==Class_ID(TCBINTERP_POSITION_CLASS_ID,0) ||
      cont->ClassID()==Class_ID(TCBINTERP_ROTATION_CLASS_ID,0) ||
      cont->ClassID()==Class_ID(TCBINTERP_SCALE_CLASS_ID,0));
}

//----------------------------
// Write a TCB PRS track.
static bool WriteTCBKeysChunk(S_dump_context &dc, INode* node, Control* cont, int type){

   IKeyControl*    ikeys;
   ITCBPoint3Key   pkey;
   ITCBRotKey      rkey;
   ITCBScaleKey    skey;
   ITCBKey*        k;
   Quat qLast = IdentQuat();
   int num = cont->NumKeys();
   Point3 firstpKey;
   
                              //Get the keyframe interface to the TCB Controller
   ikeys = GetKeyControlInterface(cont);

                              //Gotta have some keys
   if(num == NOT_KEYFRAMEABLE || num == 0 || !ikeys){
      return false;
   }

                              //Set up 'k' to point at the right derived class
   switch(type){
   case KEY_POS: k = &pkey; break;
   case KEY_ROT: k = &rkey; break;
   case KEY_SCL: k = &skey; break;
   default: assert(0);
   }

                              //write track some stuff
   short trackflags = 0;
   dc.ck->Write(&trackflags, sizeof(short));

   long lwork = cont->GetKeyTime(0)/GetTicksPerFrame();
   dc.ck->Write(&lwork, sizeof(dword));

   lwork = cont->GetKeyTime(num-1)/GetTicksPerFrame();
   dc.ck->Write(&lwork, sizeof(dword));

   long nkeys = num;
   dc.ck->Write(&nkeys, sizeof(dword));

   for(int i = 0; i < num; i++){

      ikeys->GetKey(i, k);

                              //write key time
      long keytime = k->time/GetTicksPerFrame();
      dc.ck->Write(&keytime, sizeof(dword));

                              //write flags
      short wflags = 0;
      if(k->tens   != 0.0f) wflags |= W_TENS;
      if(k->cont   != 0.0f) wflags |= W_CONT;
      if(k->bias   != 0.0f) wflags |= W_BIAS;
      if(k->easeIn != 0.0f) wflags |= W_EASETO;
      if(k->easeOut!= 0.0f) wflags |= W_EASEFROM;       
      dc.ck->Write(&wflags, sizeof(short));

                              //write TCB and ease
      if(k->tens   != 0.0f) dc.ck->Write(&k->tens, sizeof(float));
      if(k->cont   != 0.0f) dc.ck->Write(&k->cont, sizeof(float));
      if(k->bias   != 0.0f) dc.ck->Write(&k->bias, sizeof(float));
      if(k->easeIn != 0.0f) dc.ck->Write(&k->easeIn, sizeof(float));
      if(k->easeOut!= 0.0f) dc.ck->Write(&k->easeOut, sizeof(float));

                              //write values
      switch(type){
      case KEY_POS:
         {
            if(!node->GetParentNode()->IsRootNode()){
               if(!(node->GetParentNode()->GetObjOffsetRot() == IdentQuat())){
                  Quat q;
                  UndoParentsOffset(node, pkey.val, q);
               }
            }
            dc.ck->Write(&pkey.val, sizeof(S_vector));
         }
         break;

      case KEY_ROT:
         if(i){
            dc.ck->Write(&rkey.val.angle, sizeof(float));
            dc.ck->Write(&rkey.val.axis, sizeof(S_vector));
         }else{
            AngAxis aa;
            Quat qkey = QFromAngAxis(rkey.val.angle, rkey.val.axis);
                           //remove the pivot's rotation
            Quat qOff = qkey / Inverse(node->GetObjOffsetRot());
                           //remove the parents rotation
            if(!node->GetParentNode()->IsRootNode()){
               if(!(node->GetParentNode()->GetObjOffsetRot() == IdentQuat())){
                  Point3 p;
                  UndoParentsOffset(node, p, qOff);
               }
            }
            AngAxisFromQ(qOff, &aa.angle, aa.axis);
            dc.ck->Write(&aa.angle, sizeof(float));
            dc.ck->Write(&aa.axis, sizeof(S_vector));
         }
         break;

      case KEY_SCL:
         dc.ck->Write(&skey.val.s, sizeof(S_vector));
         break;
      }
   }
   return true;
}

//----------------------------

static Matrix3 GetLocalNodeTM(INode* node, TimeValue t){

   Matrix3 tm(TRUE);
   if(node){
      tm = node->GetNodeTM(t);
      if(!node->GetParentNode()->IsRootNode()){
         Matrix3 ip = Inverse(node->GetParentNode()->GetNodeTM(t));
         tm = tm * ip;
      }
   }
   return tm;
}

//----------------------------

static bool WriteSingleKey(S_dump_context &dc, INode* node, int type){
    
   const Matrix3 tm = GetLocalNodeTM(node, 0);
   AffineParts parts;
   decomp_affine(tm, &parts);

                              //write track header
   word trackflags = 0;
   dc.ck->Write(&trackflags, sizeof(word));
   long lwork = 0;
   dc.ck->Write(&lwork, sizeof(dword));
   lwork = 1;
   dc.ck->Write(&lwork, sizeof(dword));
                              //write num of keys
   long nkeys = 0;
   dc.ck->Write(&nkeys, sizeof(dword));
                              //write time
   TimeValue tv(0);
   dc.ck->Write(&tv, sizeof(dword));
                              //write flags
	short wflags = 0;			
   dc.ck->Write(&wflags, sizeof(word));

   switch(type){
   case KEY_POS:
      {
                              //write values
         Point3 pkey = parts.t;
         if(!node->GetParentNode()->IsRootNode()){
            if(!(node->GetParentNode()->GetObjOffsetRot() == IdentQuat())){
               Quat q;
               UndoParentsOffset(node, pkey, q);
            }
         }
         dc.ck->Write(&pkey, sizeof(S_vector));
      }
      break;

   case KEY_SCL:
      {
         Point3 s = ScaleValue(parts.k, parts.u).s;
         if(parts.f < 0.0f)
            s = - s;
         dc.ck->Write(&s, sizeof(S_vector));
      }
      break;

   case KEY_ROT:
      {
         AngAxis aa;
         Quat qkey = parts.q;
                              //remove the pivot's rotation
         Quat qOff = qkey / Inverse(node->GetObjOffsetRot());
                              //remove the parents rotation
         if(!node->GetParentNode()->IsRootNode()){
            if(!(node->GetParentNode()->GetObjOffsetRot() == IdentQuat())){
               Point3 p;
               UndoParentsOffset(node, p, qOff);
            }
         }
         AngAxisFromQ(qOff, &aa.angle, aa.axis);
         dc.ck->Write(&aa, sizeof(S_quat));
      }
      break;
   }
	return true;
}

//----------------------------

// Write the tracks.
static BOOL WriteController(INode* node, S_dump_context &dc){

   Interval valid;
   TimeValue start = dc.i->GetAnimRange().Start();

   int Track_Flags;
   int Force_Keys;

   Control *cpos = node->GetTMController()->GetPositionController();
   Control *crot = node->GetTMController()->GetRotationController();
   Control *cscl = node->GetTMController()->GetScaleController();

//   assert(cpos && crot && cscl);
   Class_ID cid = node->GetTMController()->ClassID();

   if(cid == Class_ID(PRS_CONTROL_CLASS_ID, 0)){
      assert(cpos && crot && cscl);
      Track_Flags = 0;
      Force_Keys = 0;
                              //this is a special case (path controller) modifies both pos and rot
      Class_ID cID = cpos->ClassID();
      if(cID == Class_ID(HYBRIDINTERP_COLOR_CLASS_ID, 0)) 
         Force_Keys |= KEYS_ROT;
   }else{
                              //if it's not a PRS controller (transform controller) try to write PRS
      Track_Flags = KEYS_POS | KEYS_ROT | KEYS_SCL;
      Force_Keys = 0;
   }

                              //write POSITION Track keys
   dc.ck->WAscend(POS_TRACK_TAG);
   if(!cpos || (Force_Keys&KEYS_POS)){
                              //this is a transform controller that doesn't support GetPositionController()
      if(WriteSingleKey(dc, node, KEY_POS))
         Track_Flags &= ~KEYS_POS;
   }else{
      if(NeedsKeys(cpos->NumKeys()) || (Track_Flags&KEYS_POS)){
         if(IsTCBControl(cpos)){
            if(WriteTCBKeysChunk(dc, node, cpos, KEY_POS)){
               Track_Flags &= ~KEYS_POS;
            }else
               goto redo_pos;
         }else{
            if(WriteSingleKey(dc, node, KEY_POS))
               Track_Flags &= ~KEYS_POS;
         }
      }else{
      redo_pos:
         Point3 pkey;
         cpos->GetValue(start, &pkey, valid);
         if(!node->GetParentNode()->IsRootNode()){
            if(!(node->GetParentNode()->GetObjOffsetRot() == IdentQuat())){
               Quat q;
               UndoParentsOffset(node, pkey, q);
            }
         }
            
         if(!WriteStd1KeyHdr(dc))
            return 0;
         dc.ck->Write(&pkey, sizeof(S_vector));
         Track_Flags &= ~KEYS_POS;
      }
   }
   dc.ck->Descend();

                              //write ROTATION Track keys
   dc.ck->WAscend(ROT_TRACK_TAG);
   if(!crot || (Force_Keys&KEYS_ROT)){
                              //this is a transform controller that doesn't support GetRotationController()
      if(WriteSingleKey(dc, node, KEY_ROT))
         Track_Flags &= ~KEYS_ROT;
   }else{
      if(NeedsKeys(crot->NumKeys()) || (Track_Flags & KEYS_ROT)){
         if(IsTCBControl(crot)){
            if(WriteTCBKeysChunk(dc, node, crot, KEY_ROT)){
               Track_Flags &= ~KEYS_ROT;
            }else
               goto redo_rot;
         }else{
            if(WriteSingleKey(dc, node, KEY_ROT))
               Track_Flags &= ~KEYS_ROT;
         }
      }else{
      redo_rot:
         Quat rkey;
         AngAxis aa;
         crot->GetValue(start, &rkey, valid);
                              //remove the pivot's rotation
         Quat qOff = rkey / Inverse(node->GetObjOffsetRot());
                              //remove the parents rotation
         if(!node->GetParentNode()->IsRootNode()){
            if(!(node->GetParentNode()->GetObjOffsetRot() == IdentQuat())){
               Point3 p;
               UndoParentsOffset(node, p, qOff);
            }
         }
         if(!WriteStd1KeyHdr(dc))
            return 0;
         AngAxisFromQ(qOff, &aa.angle, aa.axis);
         dc.ck->Write(&aa.angle, sizeof(float));
         dc.ck->Write(&aa.axis, sizeof(S_vector));
         Track_Flags &= ~KEYS_ROT;
      }
   }
   dc.ck->Descend();

                              //write SCALE Track keys    
   dc.ck->WAscend(SCL_TRACK_TAG);
   if(!cscl || (Force_Keys & KEYS_SCL)){
     // this is a transform controller that doesn't support
    // support GetScaleController()
        if(WriteSingleKey(dc, node, KEY_SCL))
            Track_Flags &= ~KEYS_SCL;
   }else{
      if(NeedsKeys(cscl->NumKeys()) || (Track_Flags & KEYS_SCL)){
         if(IsTCBControl(cscl)){
            if(WriteTCBKeysChunk(dc, node, cscl, KEY_SCL)){
               Track_Flags &= ~KEYS_SCL;
            }else
               goto redo_scl;
         }else{
            if(WriteSingleKey(dc, node, KEY_SCL))
               Track_Flags &= ~KEYS_SCL;
         }
      }else{
      redo_scl:
         ScaleValue skey;
         cscl->GetValue(start, &skey, valid);
         if(!WriteStd1KeyHdr(dc))
            return 0;
         dc.ck->Write(&skey.s, sizeof(S_vector));
         Track_Flags &= ~KEYS_SCL;
      }
   }
   dc.ck->Descend();
    
   return true;
}

//----------------------------

static bool KFWriteChunk(unsigned short tag, void *data, S_dump_context &dc);

//----------------------------

static bool WriteNoteController(INode* node, S_dump_context &dc){

   bool isnote = node->HasNoteTracks();
   if(isnote){
      if(!KFWriteChunk(NOTE_TRACK_TAG, node, dc))
         return false;
   }
   return true;
}

//----------------------------

static bool WriteVisibilityController(INode* node, S_dump_context &dc){

   Control* cvis = node->GetVisController();

   if(cvis){
     if(!KFWriteChunk(HIDE_TRACK_TAG, cvis, dc))
        return false;
   }
   return true;
}

//----------------------------

static bool WriteProperties(INode* node, S_dump_context &dc){

   TSTR str;
   node->GetUserPropBuffer(str);
   
   if(str.Length() > 0){
      dc.ck->WStringChunk(PROPERTIES_TAG, str.data());
      /*
      if(!KFWriteChunk(PROPERTIES_TAG, str.data(), dc))
         return false;
         */
   }
   return true;
}

//----------------------------

static bool KFWriteChunk(unsigned short tag, void *data, S_dump_context &dc){

   dc.ck->WAscend(tag);

   switch(tag){
   case KFDATA: 
      {
                              //keyframe main chunk
         if(!KFWriteChunk(KFHDR, data, dc))
            return(0);
         if(!KFWriteChunk(KFSEG, data, dc))
            return(0); 
         //if(!KFWriteChunk(KFCURTIME, data, dc)) return(0); 
         dc.ck->WIntChunk(KFCURTIME, dc.i->GetTime());

         int nodeid = 0;
         list<SceneEntry*>::const_iterator it;
         const list<SceneEntry*> &entries = dc.scene_enum->GetEntries();
         for(it=entries.begin(); it!=entries.end(); it++){
            SceneEntry *se = (*it);

            switch(se->type){
            case OBTYPE_BONE:
               se->SetID(nodeid++);
               if(!KFWriteChunk(BONE_NODE_TAG, se, dc))
                  return(0);
               break;     

            case OBTYPE_DUMMY:
            case OBTYPE_MESH: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(OBJECT_NODE_TAG, se, dc))
                  return(0);
               break;
            case OBTYPE_OMNILIGHT: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(LIGHT_NODE_TAG, se, dc))
                  return(0);
               break;     
            case OBTYPE_SPOTLIGHT: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(SPOTLIGHT_NODE_TAG, se, dc))
                  return(0);
               break;     

            case OBTYPE_TARGETDIRECTLIGHT: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(DIRECTLIGHT_NODE_TAG, se, dc))
                  return(0);
               break;     

            case OBTYPE_NURBS: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(OBJECT_NODE_TAG, se, dc))
                  return(0);
               break;     

            case OBTYPE_LTARGET: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(L_TARGET_NODE_TAG, se, dc))
                  return(0);
               break;
            case OBTYPE_CTARGET: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(TARGET_NODE_TAG, se, dc))
                  return(0);
               break;
            case OBTYPE_CAMERA: 
               se->SetID(nodeid++);
               if(!KFWriteChunk(CAMERA_NODE_TAG, se, dc))
                  return(0); 
               break;
//             case TARGET_NODE: 
//                {
//                Namedobj *ob = node->object;
//                node->SetID(nodeid++);
//                switch(ob->type) {
//                   case N_CAMERA:
//                      if (!KFWriteChunk(TARGET_NODE_TAG,node))
//                         return(0);
//                      break;
//                   case N_DIRECT_LIGHT:
//                      if (!KFWriteChunk(L_TARGET_NODE_TAG,node))
//                         return(0);
//                      break;
//                   }
//                }
//                break;
//             case OBTYPE_SPOTLIGHT: 
//                node->SetID(nodeid++);
//                if (!KFWriteChunk(LIGHT_NODE_TAG,node))return(0);
//                break;     
//             case SPOTLIGHT_NODE: 
//                node->SetID(nodeid++);
//                if (!KFWriteChunk(SPOTLIGHT_NODE_TAG,node))return(0);
//                break;
            }
         }
      }
      break;

   case KFHDR:
      {
         short version = KFVERSION;
         dc.ck->Write(&version, sizeof(short));
         dc.ck->Write("I3DSCENE", 9);

         long animLength = dc.i->GetAnimRange().End()/GetTicksPerFrame();
         dc.ck->Write(&animLength, sizeof(dword));
      }
      break;

   case KFSEG:
      {
         Interval i = dc.i->GetAnimRange();
         long s = i.Start()/GetTicksPerFrame();
         long e = i.End()/GetTicksPerFrame();
         dc.ck->Write(&s, sizeof(dword));
         dc.ck->Write(&e, sizeof(dword));
      }
      break;

      /*
   case KFCURTIME:
      {
         long t = dc.i->GetTime();
         dc.ck->Write(&t, sizeof(dword));
      }
      break;

   case PIVOT:
      {
                              //make pivot out of a max pivot.
         Point3 pivot = GetPivotOffset(((SceneEntry *)data)->node);
         dc.ck->Write(&pivot[0], sizeof(S_vector));
      }
      break;
      */

   case BOUNDBOX:
      {
         Box3 bb;
         ViewExp *vpt = dc.i->GetViewport(NULL);
         SceneEntry *se = (SceneEntry *)data;
         se->obj->GetLocalBoundBox(dc.i->GetTime(), se->node, vpt, bb);
         dc.ck->Write(&bb.pmin[0], sizeof(S_vector));
         dc.ck->Write(&bb.pmax[0], sizeof(S_vector));
      }
      break;

      /*
   case INSTANCE_NAME:
      {
         CStr *name = (CStr *)data;
         if(name->Length()>MAX_NAME_SIZE)
            name->Resize(MAX_NAME_SIZE);
         dc.ck->Write(name->data(), (name->Length()+1));
      }
      break;

   case NODE_ID: 
      {
         SceneEntry *se = (SceneEntry *)data;
         assert(se->id!=-1);
         dc.ck->Write(&se->id, sizeof(short));
      }
      break;
      */

   case NODE_HDR:
      {
         SceneEntry *se = (SceneEntry*)data;
         bool isInstance = false;
         INode *node = se->node;
         Object *ob = se->obj;
                              //write out the primary name -- It's the name of the first node using the object.
                              //we do this to work with 3DS's system where a primary node uses an object and
                              //shares its name, and all other instances use this name as a primary name with
                              //a dot (.) followed by the instance name.
         if(se->type == OBTYPE_DUMMY){
            dc.ck->Write("$$$DUMMY", 9);
         }else
         if(se->type == OBTYPE_CTARGET || se->type == OBTYPE_LTARGET){
                              //the object name written should be the same as that of the associated
                              // light or camera.
            ObjectEntry *oe = dc.objects->FindLookatNode(se->node);
            assert(oe);
            CStr name(oe->entry->name); 
            if(name.Length()>MAX_NAME_SIZE)
               name.Resize(MAX_NAME_SIZE);
            dc.ck->Write(name.data(), (name.Length()+1));
         }else{
            TSTR name; 
            
            // for cameras, lights, which can't be multiple instanced in 3DS
            // just get name from  scene entry  -- DS 4/6/96
            if(se->type != OBTYPE_MESH){
               name = se->name;
            }else{
                              //search for the object entry 
               ObjectEntry *oe = dc.objects->Contains(ob);
               assert(oe);

               name = oe->entry->name;
                              //if the name is not equal to the master name then
                              // it is an instance.
               if(!(oe->entry->name==se->name))
                  isInstance = true;
            }
            
            if(name.Length()>MAX_NAME_SIZE)
               name.Resize(MAX_NAME_SIZE);
            CStr cname(name);    // Make it char*
            dc.ck->Write(cname.data(), (cname.Length()+1));
         }
         short zero = 0;
         short fl = isInstance ? 0 : PRIMARY_NODE;

         dc.ck->Write(&fl, sizeof(short));

         dc.ck->Write(&zero, sizeof(short));

         INode *parentNode = node->GetParentNode();
         SceneEntry *pse = dc.scene_enum->Find(parentNode);
         if(pse){
            dc.ck->Write(&pse->id, sizeof(short));
            assert(pse->id!=-1);
         }else{
            short none = -1;
            dc.ck->Write(&none, sizeof(short));
         }
      }
      break;

   case OBJECT_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         Object *ob = se->obj;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return 0;
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return 0;
//       if (on->appdata) if (!KFWriteChunk(APP_DATA,on->appdata)) return(0);
         {
                              //write pivot only if it's non-zero
            Point3 pivot = GetPivotOffset(((SceneEntry *)data)->node);
            if(pivot.x!=0.0f || pivot.y!=0.0f || pivot.z!=0.0f){
               //if(!KFWriteChunk(PIVOT, se, dc)) return 0;
               dc.ck->WVectorChunk(PIVOT, *(S_vector*)&pivot);
            }
         }
         TSTR name = se->name;
         // MAX doesn't have the notion of instance names, but 3DS does -- we must use the name
         // of the original object as the node name, and this node's name as the instance name.
         // We don't write an instance name if the instance name is the same as the master object's
         // name.
         ObjectEntry *oe = dc.objects->Contains(ob);
         assert(oe);
         TSTR mname(oe->entry->name);     // Master name
         if(se->type!=OBTYPE_CTARGET&&se->type!=OBTYPE_LTARGET){
            if(se->type==OBTYPE_DUMMY || !(mname == name)){ 
               CStr wname(name);
               //if(!KFWriteChunk(INSTANCE_NAME, &wname, dc)) return false;
               dc.ck->WStringChunk(INSTANCE_NAME, wname);
            }
         }
         WriteController(se->node, dc);

         if(!WriteNoteController(se->node, dc))
            return false;

         if(!WriteVisibilityController(se->node, dc))
            return false;


//       if (on->mtrack.keytab)
//          if (!KFWriteChunk(MORPH_TRACK_TAG, &on->mtrack)) return(0);
//       if (on->htrack.keytab)
//          if (!KFWriteChunk(H_IDE_TRACK_TAG, &on->htrack)) return(0);
         if(se->type==OBTYPE_DUMMY){ //dummy Node
            if(!KFWriteChunk(BOUNDBOX, data, dc)) return false;
         }
//       if (!KFWriteChunk(MORPH_SMOOTH, data)) return(0);

         if(!WriteProperties(se->node, dc))
            return false;
      }
      break;

   case CAMERA_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         GenCamera *ob = (GenCamera *)se->obj;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return false;
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return false;
         WriteController(se->node, dc);

         Control *c;
         
         floatVal = RadToDeg(ob->GetFOV(0));
         c = ob->GetFOVControl();
         if(!KFWriteChunk(FOV_TRACK_TAG, c, dc))
            return(0);

         Control* tmc = se->node->GetTMController();
         c = tmc->GetRollController();
         if(c)
            if(!KFWriteChunk(ROLL_TRACK_TAG, c, dc))
               return(0);

         if(!WriteNoteController(se->node, dc))
            return 0;

         if(!WriteVisibilityController(se->node, dc))
            return 0;
      }
      break;

   case L_TARGET_NODE_TAG:
   case TARGET_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return(0);
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return(0);
         Control *c;
         c = se->node->GetTMController()->GetPositionController();
         if(c)
            if(!KFWriteChunk(POS_TRACK_TAG, c, dc))
               return(0);

         if(!WriteNoteController(se->node, dc))
            return 0;

         if(!WriteVisibilityController(se->node, dc))
            return 0;

         if(!WriteProperties(se->node, dc))
            return 0;
      }
      break;

   case LIGHT_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         GenLight* ob = (GenLight *)se->obj;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return(0);
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return(0);
         Control *c;
         
         colorVal = ob->GetRGBColor(0);
         c = ob->GetColorControl();
         if(!KFWriteChunk(COL_TRACK_TAG, c, dc))
            return(0);
         
         c = se->node->GetTMController()->GetPositionController();
         if(c)
            if(!KFWriteChunk(POS_TRACK_TAG, c, dc))
               return(0);

         if(!WriteNoteController(se->node, dc))
            return 0;

         if(!WriteVisibilityController(se->node, dc))
            return 0;
      }
      break;

   case NURBS_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return(0);
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return(0);

         if(!WriteNoteController(se->node, dc))
            return 0;

         if(!WriteVisibilityController(se->node, dc))
            return 0;
      }
      break;
         
   case BONE_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return(0);
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return(0);
         {
                              //write pivot only if it's non-zero
            Point3 pivot = GetPivotOffset(((SceneEntry *)data)->node);
            if(pivot.x!=0 || pivot.y!=0 || pivot.z!=0){
               //if(!KFWriteChunk(PIVOT, se, dc)) return(0);
               dc.ck->WVectorChunk(PIVOT, *(S_vector*)&pivot);
            }
         }

         WriteController(se->node, dc);

         if(!WriteNoteController(se->node, dc))
            return 0;
         if(!WriteVisibilityController(se->node, dc))
            return 0;
         if(!WriteProperties(se->node, dc))
            return 0;
      }
      break;

   case DIRECTLIGHT_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         GenLight* ob = (GenLight *)se->obj;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return(0);
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return(0);
         Control *c;
         c = se->node->GetTMController()->GetPositionController();
         if(c)
            if(!KFWriteChunk(POS_TRACK_TAG, c, dc))
               return(0);

         colorVal = ob->GetRGBColor(0);
         c = ob->GetColorControl();
         if(!KFWriteChunk(COL_TRACK_TAG, c, dc))
            return(0);
         
         floatVal = ob->GetHotspot(0);
         c = ob->GetHotSpotControl();
         if(!KFWriteChunk(HOT_TRACK_TAG, c, dc))
            return(0);

         floatVal = ob->GetFallsize(0);
         c = ob->GetFalloffControl();
         if(!KFWriteChunk(FALL_TRACK_TAG, c, dc))
            return(0);

         Control* tmc = se->node->GetTMController();
         c = tmc->GetRollController();
         if(!KFWriteChunk(ROLL_TRACK_TAG, c, dc))
            return(0);

         if(!WriteNoteController(se->node, dc))
            return 0;

         if(!WriteVisibilityController(se->node, dc))
            return 0;

      }
      break;

   case SPOTLIGHT_NODE_TAG:
      {
         SceneEntry *se = (SceneEntry *)data;
         GenLight* ob = (GenLight *)se->obj;
         //if(!KFWriteChunk(NODE_ID, se, dc)) return(0);
         dc.ck->WWordChunk(NODE_ID, se->id);
         if(!KFWriteChunk(NODE_HDR, se, dc))
            return(0);
         Control *c;
         c = se->node->GetTMController()->GetPositionController();
         if(c)
            if(!KFWriteChunk(POS_TRACK_TAG, c, dc))
               return(0);

         colorVal = ob->GetRGBColor(0);
         c = ob->GetColorControl();
         if(!KFWriteChunk(COL_TRACK_TAG, c, dc))
            return(0);
         
         floatVal = ob->GetHotspot(0);
         c = ob->GetHotSpotControl();
         if(!KFWriteChunk(HOT_TRACK_TAG, c, dc))
            return(0);

         floatVal = ob->GetFallsize(0);
         c = ob->GetFalloffControl();
         if(!KFWriteChunk(FALL_TRACK_TAG, c, dc))
            return(0);

         Control* tmc = se->node->GetTMController();
         c = tmc->GetRollController();
         if(!KFWriteChunk(ROLL_TRACK_TAG, c, dc))
            return(0);

         if(!WriteNoteController(se->node, dc))
            return 0;

         if(!WriteVisibilityController(se->node, dc))
            return 0;
      }
      break;

   case POS_TRACK_TAG:
      {
         Control *c = (Control*)data;
         if(!WriteControllerChunk(dc, c, KEY_POS)) {
            Point3 pos;
            Interval valid;
            c->GetValue(0, &pos, valid);             
            if(!WriteStd1KeyHdr(dc))
               return 0;
            dc.ck->Write(&pos, sizeof(S_vector));
         }
      }
      break;

   case FOV_TRACK_TAG:
      if(!WriteFloatTrackOrStd1KeyHdr((Control *)data, RAD_TO_DEG, dc))
         return 0;
      break;

   case FALL_TRACK_TAG:
   case HOT_TRACK_TAG: 
      if(!WriteFloatTrackOrStd1KeyHdr((Control *)data, 1.0f, dc))
         return 0;
      break;

   case ROLL_TRACK_TAG:    
      if(!WriteFloatTrackOrStd1KeyHdr((Control *)data, -RAD_TO_DEG, dc))
         return 0;
      break;

   case COL_TRACK_TAG:
      {
         Control *c = (Control*)data;
         if(!c || !WriteControllerChunk(dc, c, KEY_COLOR)){
            Point3 pos;
            Interval valid;
            if(c) c->GetValue(0, &pos, valid);               
            else pos = colorVal;
            if(!WriteStd1KeyHdr(dc))
               return 0;
            dc.ck->Write(&pos, sizeof(S_vector));
         }
      }
      break;

   case HIDE_TRACK_TAG:
      {
         SuspendAnimate();
         AnimateOn();
            
         Control *c = (Control *)data;
         Class_ID id = c->ClassID();

         int numkeys = c->NumKeys();

         if(numkeys == NOT_KEYFRAMEABLE)
            numkeys = 0;
               
         dc.ck->Write(&numkeys, sizeof(int));

         TimeValue t;

         float p;
         Interval valid;

         for(int i =0; i<numkeys;i++){
            t = c->GetKeyTime(i)/GetTicksPerFrame();
            c->GetValue(t, &p, valid);
            dc.ck->Write(&t, sizeof(TimeValue));
         }
         ResumeAnimate();
      }
      break;

   case NOTE_TRACK_TAG:
      {
         INode *n = (INode*)data;
         int ntracks = n->NumNoteTracks();
         DefNoteTrack *track;
         
         dc.ck->Write(&ntracks, sizeof(int));

         for(int i = 0; i<ntracks; i++){
            track = (DefNoteTrack *)n->GetNoteTrack(i);            
            int numkeys = track->NumKeys();
            dc.ck->Write(&numkeys, sizeof(int));

            for(int j = 0; j<numkeys;j++){
               TimeValue t = track->GetKeyTime(j) / GetTicksPerFrame();

               dc.ck->Write(&t, sizeof(TimeValue));

               TSTR note = track->keys[j]->note;
               dc.ck->Write(note.data(), strlen(note.data())+1);
            }
         }
      }
      break;

      /*
   case PROPERTIES_TAG:
      dc.ck->Write(data, strlen((const char*)data)+1);
      break;
      */

   default:
      assert(0);
      break;
   }           

   dc.ck->Descend();
   return true;
}

//----------------------------

static int SaveMain(const TCHAR *filename, ExpInterface *ei, Interface *gi, bool selection_only){

   C_chunk ck;
   if(!ck.WOpen(filename))
      return 0;

   C_MeshMtlList materials;
                              //Make sure there are nodes we're interested in.
                              //Ask the scene to enumerate all its nodes
                              // so we can determine if there are any we can use
   C_SceneEnumProc scene_enum(ei->theScene, gi->GetTime(), &materials, selection_only);

   C_ObjectList objects(scene_enum);

   S_dump_context dc;
   dc.scene_enum = &scene_enum;
   dc.objects = &objects;
   dc.materials = &materials;
   dc.ck = &ck;
   dc.i = gi;

   int error = 0;
   
   ck.WAscend(M3DMAGIC);
   {
      ck.WIntChunk(M3D_VERSION, 200);
      DumpMeshChunk(MMAGIC, NULL, dc);
      KFWriteChunk(KFDATA, NULL, dc);
   }
   dc.ck->Descend();
   /*
   if(!DumpChunk(M3DMAGIC, dc))
      error = 1;
      */
   
   if(error){
      MessageBox(GetActiveWindow(), C_fstring("Failed to write to file: %s.", filename),
         "I3D exporter", MB_OK);
      remove(filename);
      return 0;
   }
   return 1;   
}

//----------------------------

class C_I3DExport: public SceneExport{

public:
   C_I3DExport(){
   }

   virtual ~C_I3DExport(){
   }

//----------------------------

   virtual int ExtCount(){ return 1; }

//----------------------------

   virtual const TCHAR *Ext(int n){
      
      switch(n){
      case 0: return _T("I3D");
      }
      return _T("");
   }

   virtual const TCHAR *LongDesc(){ return "Insanity 3D Resource File"; }
   virtual const TCHAR *ShortDesc(){ return "I3D"; }
   virtual const TCHAR *AuthorName(){ return "Lonely Cat Games"; }
   virtual const TCHAR *CopyrightMessage(){ return "Copyright (c) 2001 Lonely Cat Games"; }
   virtual const TCHAR *OtherMessage1(){ return ""; }
   virtual const TCHAR *OtherMessage2(){ return ""; }
   virtual unsigned int Version(){ return 200; }
   virtual void ShowAbout(HWND hWnd){ }

//----------------------------

   virtual int DoExport(const TCHAR *filename, ExpInterface *ei, Interface *gi, BOOL suppressPrompts, DWORD options){

      int status = SaveMain(filename, ei, gi, (options&SCENE_EXPORT_SELECTED));

      if(status == 0)
         return 1;      //Dialog cancelled
      if(status < 0)
         return 0;      //real error
      return status;
   }

//----------------------------

   virtual BOOL SupportsOptions(int ext, DWORD options){
      return (options==SCENE_EXPORT_SELECTED);
   }
};

//----------------------------

class C_I3DClassDesc: public ClassDesc{

public:
   virtual int IsPublic(){ return 1; }
   virtual void *Create(BOOL loading = false){ return new C_I3DExport; }
   virtual const TCHAR *ClassName(){ return "3D Studio"; }
   virtual SClass_ID SuperClassID(){ return SCENE_EXPORT_CLASS_ID; }
   virtual Class_ID ClassID(){ return Class_ID(0x51b52649, 0x1662350c); }
   virtual const TCHAR *Category(){ return "Scene Export"; }
};

static C_I3DClassDesc I3DDesc;

//----------------------------

__declspec(dllexport) const TCHAR *LibDescription(){ return "I3D file exporter"; }

//----------------------------

__declspec(dllexport) int LibNumberClasses(){ return 1; }

//----------------------------
// Return version so can detect obsolete DLLs
__declspec(dllexport) ULONG LibVersion(){ return VERSION_3DSMAX; }

//----------------------------

__declspec(dllexport) ClassDesc *LibClassDesc(int i){
   switch(i){
      case 0: return &I3DDesc; break;
      default: return 0; break;
   }
}

//----------------------------
//----------------------------
