#include "Max.h"
#include "dummy.h"
#include "i3dexp.h"
#include "..\common\mtldef.h"
#include "gamma.h"


#define WRTERR(ptr, sz) { if(!dc.ck->Write(ptr, (sz))) return 0; }

SMtl *savemtl;

static const ULONG map_chunks[8] = {
   MAT_TEXMAP, MAT_TEX2MAP, MAT_OPACMAP, MAT_BUMPMAP,
   MAT_SPECMAP, MAT_SHINMAP, MAT_SELFIMAP, MAT_REFLMAP
};

static const ULONG mask_chunks[8] = {
   MAT_TEXMASK,MAT_TEX2MASK,MAT_OPACMASK,MAT_BUMPMASK,
   MAT_SPECMASK,MAT_SHINMASK,MAT_SELFIMASK,MAT_REFLMASK
};

static const ULONG map_sxp_chunks[8] = {
   MAT_SXP_TEXT_DATA,MAT_SXP_TEXT2_DATA,MAT_SXP_OPAC_DATA,MAT_SXP_BUMP_DATA,
   MAT_SXP_SPEC_DATA,MAT_SXP_SHIN_DATA,MAT_SXP_SELFI_DATA,0 
};

static const ULONG mask_sxp_chunks[8] = {
   MAT_SXP_TEXT_MASKDATA,MAT_SXP_TEXT2_MASKDATA,MAT_SXP_OPAC_MASKDATA,MAT_SXP_BUMP_MASKDATA,
   MAT_SXP_SPEC_MASKDATA,MAT_SXP_SHIN_MASKDATA,MAT_SXP_SELFI_MASKDATA,MAT_SXP_REFL_MASKDATA  
};

// Main routine for writing a file
// Recursive chunk writer -- keeps track of each chunk

#define DMP_MATCHUNK(id,data) { if(!DumpMaterialChunk((ushort)id, data, dc)) return 0;}

//----------------------------

int dmp_map_params(MapParams *mp, S_dump_context &dc){

   DMP_MATCHUNK(MAT_MAP_TILING,&mp->texflags);
   DMP_MATCHUNK(MAT_MAP_TEXBLUR,&mp->texblur);
   if (mp->uscale!=1.0)
      DMP_MATCHUNK(MAT_MAP_USCALE,&mp->uscale);
   if (mp->vscale!=1.0)
      DMP_MATCHUNK(MAT_MAP_VSCALE,&mp->vscale);
   if (mp->uoffset!=0.0)
      DMP_MATCHUNK(MAT_MAP_UOFFSET,&mp->uoffset);
   if (mp->voffset!=0.0)
      DMP_MATCHUNK(MAT_MAP_VOFFSET,&mp->voffset);
   if (mp->ang_sin!=0.0) {
      DMP_MATCHUNK(MAT_MAP_ANG,mp);
      }
   if (mp->col1.r!=0||mp->col1.g!=0||mp->col1.b!=0) 
      DMP_MATCHUNK(MAT_MAP_COL1,&mp->col1);
   if (mp->col2.r!=255||mp->col2.g!=255||mp->col2.b!=255) 
      DMP_MATCHUNK(MAT_MAP_COL2,&mp->col2);
   if (mp->rcol.r!=255||mp->rcol.g!=0  ||mp->rcol.b!=0  ) 
      DMP_MATCHUNK(MAT_MAP_RCOL,&mp->rcol);
   if (mp->gcol.r!=0  ||mp->gcol.g!=255||mp->gcol.b!=0  ) 
      DMP_MATCHUNK(MAT_MAP_GCOL,&mp->gcol);
   if (mp->bcol.r!=0  ||mp->bcol.g!=0  ||mp->bcol.b!=255) 
      DMP_MATCHUNK(MAT_MAP_BCOL,&mp->bcol);
   return(1);
}

//----------------------------

int dmp_map(int i, S_dump_context &dc){

   Mapping *m = savemtl->map[i];
   /*                         //MB: bump percentage
   if(i==Nbump) {
      int t = m->amt.pct/4;      //for old versions
      DMP_MATCHUNK(INT_PERCENTAGE, &t);
      DMP_MATCHUNK(MAT_MAPNAME, m->map.map_name);
      DMP_MATCHUNK(MAT_BUMP_PERCENT,&m->amt.pct); //new version of bump percent
   }else*/
   {
      DMP_MATCHUNK(INT_PERCENTAGE, &m->amt.pct);
      //DMP_MATCHUNK(MAT_MAPNAME,m->map.name);
      DMP_MATCHUNK(MAT_MAPNAME, m->map.map_name);
   }
   if(m->map.kind==0) 
      return(dmp_map_params(&m->map.p.tex, dc));
   else
      return 1;//(dmp_map_paramsR(&m->map.p.ref));
}

//----------------------------

int dmp_mask(int i, S_dump_context &dc){

   Mapping *m = savemtl->map[i];
   DMP_MATCHUNK(MAT_MAPNAME, m->mask.map_name);
   if (m->mask.kind==0) 
      return(dmp_map_params(&m->mask.p.tex, dc));
   else 
      return(1);
}

//----------------------------

int isSXPMap(MapData *md) {
   return(md->kind==0&&md->p.tex.sxp_data!=NULL);
}

//----------------------------

bool DumpMaterialChunk(ushort tag, void *data, S_dump_context &dc){

   Color_f *cf;
   Color_24 *c24;
   char *string;
   
   dc.ck->WAscend(tag);

   switch(tag){

   case MAT_ENTRY:
      {
         if(!DumpMaterialChunk(MAT_NAME, savemtl->name, dc)) return false;
         if(!DumpMaterialChunk(MAT_AMBIENT, &savemtl->amb, dc)) return false;
         if(!DumpMaterialChunk(MAT_DIFFUSE, &savemtl->diff, dc)) return false;
         if(!DumpMaterialChunk(MAT_SPECULAR, &savemtl->spec, dc))return false;
         if(!DumpMaterialChunk(MAT_SHININESS, NULL, dc))return false;
         if(!DumpMaterialChunk(MAT_SHIN2PCT, NULL, dc)) return false;
         if(!DumpMaterialChunk(MAT_TRANSPARENCY, NULL, dc)) return false;
         if(!DumpMaterialChunk(MAT_XPFALL, NULL, dc)) return false;
         if(!DumpMaterialChunk(MAT_REFBLUR, NULL, dc)) return false;
         if(!DumpMaterialChunk(MAT_SHADING, NULL, dc)) return false;
         if(!DumpMaterialChunk(MAT_SELF_ILPCT, NULL, dc)) return false;
   
         if(savemtl->use&MATUSE_XPFALL){
            if(!DumpMaterialChunk(MAT_USE_XPFALL, NULL, dc))
               return false;
         }
         if(savemtl->use&MATUSE_REFBLUR){
            if(!DumpMaterialChunk(MAT_USE_REFBLUR,NULL, dc))
               return false;
         }
         if(savemtl->flags & MF_TWOSIDE){
            if(!DumpMaterialChunk(MAT_TWO_SIDE,NULL, dc))
               return(0);
         }
         if(savemtl->flags & MF_ADDITIVE){
            if(!DumpMaterialChunk(MAT_ADDITIVE,NULL, dc))
               return(0);
         }
         if(savemtl->flags & MF_WIRE){
            if(!DumpMaterialChunk(MAT_WIRE,NULL, dc))
               return(0);
         }
         if(savemtl->flags & MF_FACEMAP){
            if(!DumpMaterialChunk(MAT_FACEMAP,NULL, dc))
               return(0);
         }

         if(savemtl->flags & MF_XPFALLIN){
            if(!DumpMaterialChunk(MAT_XPFALLIN,NULL, dc))
               return(0);
         }

         if(savemtl->flags & MF_PHONGSOFT){
            if(!DumpMaterialChunk(MAT_PHONGSOFT,NULL, dc))
               return(0);
         }

         if(savemtl->flags & MF_WIREABS){
            if(!DumpMaterialChunk(MAT_WIREABS,NULL, dc))
               return(0);
         }

         if(!DumpMaterialChunk(MAT_WIRESIZE,&savemtl->wiresize, dc))
               return(0);

                              //save out any texture maps, masks, sxp's
         for(int k=0; k<NMAPTYPES; k++){
            Mapping *mp = savemtl->map[k];
            /*
            {
               if(savemtl->flags & MF_NEEDUV){
                  //int a = 0;
               }
            }
            */
            if(mp==NULL)
               continue;
            if(mp->use){
               DMP_MATCHUNK(map_chunks[k], NULL);
               if(k!=Nrefl){
                  if(isSXPMap(&mp->map)) 
                     DMP_MATCHUNK(map_sxp_chunks[k],mp->map.p.tex.sxp_data);
               }
               if(mp->mask.map_name[0]!=0){
                  DMP_MATCHUNK(mask_chunks[k],NULL);
                  if(isSXPMap(&mp->mask)) 
                     DMP_MATCHUNK(mask_sxp_chunks[k],mp->mask.p.tex.sxp_data);
               }
            }
         }
                              //dump auto-cubic chunk
         {
            Mapping *rm = savemtl->map[Nrefl];
            if(rm&&rm->use&&rm->map.p.ref.acb.flags & AC_ON)
               DMP_MATCHUNK(MAT_ACUBIC,&rm->map.p.ref.acb);
         }
         if(savemtl->appdata) 
            if(!DumpMaterialChunk(APP_DATA,savemtl->appdata, dc))
               return(0);
      }           
      break;
   
   case APP_DATA:
      {
         ULONG *plong = (ULONG *)data;
         WRTERR(&plong[1],plong[0]);
      }
      break;

   case MAT_AMBIENT:
   case MAT_DIFFUSE:
   case MAT_SPECULAR:
      if(gammaMgr.enable){
         Color_24 gc;
         c24 = (Color_24 *)data;
         gc.r = gammaMgr.file_in_gamtab[c24->r];
         gc.g = gammaMgr.file_in_gamtab[c24->g];
         gc.b = gammaMgr.file_in_gamtab[c24->b];
         if(!DumpMaterialChunk(COLOR_24,&gc, dc))
            return(0);
         if(!DumpMaterialChunk(LIN_COLOR_24,data, dc))
            return(0);
      }else{
         if(!DumpMaterialChunk(COLOR_24,data, dc))
            return(0);
      }
      break;

   case MAT_SHININESS:
      if(!DumpMaterialChunk(INT_PERCENTAGE,&savemtl->shininess, dc))
         return(0);
      break;

   case MAT_SHIN2PCT:
      if(!DumpMaterialChunk(INT_PERCENTAGE,&savemtl->shin2pct, dc))
         return(0);
      break;

   case MAT_SHIN3PCT:
      if(!DumpMaterialChunk(INT_PERCENTAGE,&savemtl->shin3pct, dc))
         return(0);
      break;

   case MAT_TRANSPARENCY:
      if(!DumpMaterialChunk(INT_PERCENTAGE,&savemtl->transparency, dc))
         return(0);
      break;

   case MAT_XPFALL:
      if(!DumpMaterialChunk(INT_PERCENTAGE,&savemtl->xpfall, dc))
         return(0);
      break;

   case MAT_REFBLUR:
      if(!DumpMaterialChunk(INT_PERCENTAGE,&savemtl->refblur, dc))
         return(0);
      break;

   case MAT_SELF_ILPCT:
      if(!DumpMaterialChunk(INT_PERCENTAGE,&savemtl->selfipct, dc))
         return(0);
      break;

   case MAT_SHADING:
      WRTERR(&savemtl->shading,2);
      break;

   case MAT_TEXMAP:
      if(!dmp_map(Ntex, dc))
         return(0);
      break;

   case MAT_TEX2MAP:
      if (!dmp_map(Ntex2, dc))
         return(0);
      break;

   case MAT_OPACMAP:
      if(!dmp_map(Nopac, dc))
         return(0);
      break;

   case MAT_BUMPMAP:
      if(!dmp_map(Nbump, dc))
         return(0);
      break;

   case MAT_SPECMAP:
      if(!dmp_map(Nspec, dc))
         return(0);
      break;

   case MAT_SHINMAP:
      if (!dmp_map(Nshin, dc))
         return(0);
      break;

   case MAT_SELFIMAP:
      if(!dmp_map(Nselfi, dc))
         return(0);
      break;

   case MAT_REFLMAP:
      if(!dmp_map(Nrefl, dc))
         return(0);
      break;

   case MAT_TEXMASK:  if (!dmp_mask(Ntex, dc)) return(0); break;
   case MAT_TEX2MASK: if (!dmp_mask(Ntex2, dc)) return(0); break;
   case MAT_OPACMASK: if (!dmp_mask(Nopac, dc)) return(0); break;
   case MAT_BUMPMASK: if (!dmp_mask(Nbump, dc)) return(0); break;
   case MAT_SPECMASK: if (!dmp_mask(Nspec, dc)) return(0); break;
   case MAT_SHINMASK: if (!dmp_mask(Nshin, dc)) return(0); break;
   case MAT_SELFIMASK:if (!dmp_mask(Nselfi, dc)) return(0); break;
   case MAT_REFLMASK: if (!dmp_mask(Nrefl, dc)) return(0); break;

   case MAT_MAP_TILING: 
      WRTERR(data,2);
      break;

   case MAT_MAP_TEXBLUR: 
   case MAT_MAP_USCALE: 
   case MAT_MAP_VSCALE: 
   case MAT_MAP_UOFFSET: 
   case MAT_MAP_VOFFSET: 
      WRTERR(data,4);
      break;

   case MAT_MAP_COL1:
   case MAT_MAP_COL2:
   case MAT_MAP_RCOL:
   case MAT_MAP_GCOL:
   case MAT_MAP_BCOL:
      c24=(Color_24 *)data;
      WRTERR(c24,3);
      break;

   case MAT_MAP_ANG:
      {
         MapParams *mp = (MapParams *)data;
         float ang,dang;
         ang = (float)atan2(mp->ang_sin,mp->ang_cos);
         dang = RadToDeg(ang);
         WRTERR(&dang,4);
      }
      break;

   case COLOR_F:
      cf = (Color_f*)data;
      WRTERR(cf,12);
      break;

   case COLOR_24:
      c24=(Color_24 *)data;
      WRTERR(c24,3);
      break;

   case LIN_COLOR_24:
      c24 = (Color_24 *)data;
      WRTERR(c24,3);
      break;

   case MAT_NAME:
   case MAT_MAPNAME:
      string = (char *)data;
      WRTERR(string, strlen(string)+1);
      break;

   //case MAT_BUMP_PERCENT:
   case INT_PERCENTAGE:
      WRTERR(data,2);
      break;

   case MAT_WIRESIZE:
      WRTERR(data,4);
      break;

   case MAT_TWO_SIDE:
   case MAT_SUPERSMP:
   case MAT_ADDITIVE:
   case MAT_WIRE:
   case MAT_FACEMAP:
   case MAT_XPFALLIN:
   case MAT_USE_XPFALL:
   case MAT_USE_REFBLUR:
   case MAT_PHONGSOFT:
   case MAT_WIREABS:
   case DUMMY:
      break;

   case MAT_ACUBIC:
      {
         AutoCubicParams *ac = (AutoCubicParams *)data;
         WRTERR(&ac->shade,1);
         WRTERR(&ac->aalevel,1);
         WRTERR(&ac->flags,2);
         WRTERR(&ac->size,4);
         WRTERR(&ac->nth,4);
      }
      break;

   case MAT_SXP_TEXT_DATA:
   case MAT_SXP_TEXT2_DATA:
   case MAT_SXP_OPAC_DATA:
   case MAT_SXP_BUMP_DATA:
   case MAT_SXP_SPEC_DATA:
   case MAT_SXP_SHIN_DATA:
   case MAT_SXP_SELFI_DATA:

   case MAT_SXP_TEXT_MASKDATA:
   case MAT_SXP_TEXT2_MASKDATA:
   case MAT_SXP_OPAC_MASKDATA:
   case MAT_SXP_BUMP_MASKDATA:
   case MAT_SXP_SPEC_MASKDATA:
   case MAT_SXP_SHIN_MASKDATA:
   case MAT_SXP_SELFI_MASKDATA:
   case MAT_SXP_REFL_MASKDATA:
      {
         ULONG *plong = (ULONG *)data;
         if(plong!=NULL)
            WRTERR(&plong[1],plong[0]);
      }
      break;
   }
   dc.ck->Descend();
   return true;
}

//----------------------------
