#include "Max.h"
#include <stdio.h>
#include <direct.h>
#include <commdlg.h>
#include "imtl.h"
#include "dummy.h"
#include "i3dimp.h"
#include "..\common\mtldef.h"
#include "istdplug.h"
#include "stdmat.h"
#include "gamma.h"
#include "Notetrck.h"

#pragma comment(lib, "edmodel.lib")

#define NextChunk(hdr)  ((Chunk_hdr *)(((char  *)hdr)+hdr->size))
#define SubChunk(hdr)  ((Chunk_hdr *)(((char  *)hdr)+6))

//----------------------------

// 3DS-to-MAX time constant multiplier
#define TIME_CONSTANT GetTicksPerFrame()

// The file stream

// Some stuff we're defining (for now)

struct Bkgrad{
   float midpct;
   Color botcolor;
   Color midcolor;
   Color topcolor;
};

struct Fogdata{
   float nearplane;
   float neardens;
   float farplane;
   float fardens;
   Color color;
};

struct LFogData{
   float zmin,zmax;
   float density;
   short type;
   short fog_bg;
   Color color;
};

struct Distcue{
   float nearplane;
   float neardim;
   float farplane;
   float fardim;
};

#define ENV_DISTCUE 1
#define ENV_FOG 2
#define ENV_LAYFOG 3

#define BG_SOLID 1
#define BG_GRADIENT 2
#define BG_BITMAP 3


struct BGdata{
   int bgType;
   int envType;
   Color bkgd_solid;
   Color amb_light;
   Bkgrad bkgd_gradient;
   Fogdata fog_data;
   Distcue distance_cue;
   int fog_bg,dim_bg;
   char bkgd_map[81];
   LFogData lfog_data;
};


static int merging = 0;
#define OBJ_NAME_LEN 100
static char obname[101];
static short cur_node_id, nodeflags, nodeflags2;
static int nodetag;
//static int skipNode = 0;
static ImpNode *obnode;
static short readVers;
static BGdata BG;

//----------------------------

static Color ColorFrom24(Color_24 c){
   Color a;
   a.r = (float)c.r/255.0f;
   a.g = (float)c.g/255.0f;
   a.b = (float)c.b/255.0f;
   return a;
}

//----------------------------

// 3DS R4 IK joint data
typedef struct{
   int      freeJoints[6];
   float jointMins[6];
   float jointMaxs[6];
   float precedence[6];
   float damping[6];
   int      limited[6];
   int      ease[6];
   int      lastModified;
} JointData3DSR4;

//----------------------------
//----------------------------

class ObjWorker: public C_chunk{

   float msc_wk;
   int nodeLoadNumber;

//----------------------------

   static Object *CreateObjectFromAppData(TriObject *tobj, void *data, DWORD len){

      Chunk_hdr *sub, *hdr = (Chunk_hdr*)data;
      Object *obj;
      char idstring[100];
      int noff, nbytes = (int)len;  

      noff = 0;

      // DS 6/7/96: put in check for hdr->size<0 to catch bad appdata
      // chunks.
      while(noff<nbytes){
         if (hdr->tag != XDATA_ENTRY) goto next; // if it aint XDATA_ENTRY we don't want it
         sub = SubChunk(hdr);
         if (sub->tag!= XDATA_APPNAME) goto next;
      
         // Try to convert it to an object
         GetIDStr((char*)hdr,idstring);
         obj = ObjectFromAppData(tobj, idstring, (void*)sub, hdr->size-6);
         if (obj) return obj;
      
        next:
         if (hdr->size<0) break; 
         noff += hdr->size;
         hdr = NextChunk(hdr);
      }

      return NULL;
   }

//----------------------------
// Match 3DStudio's "Default" material
   static StdMat *New3DSDefaultMtl(){
      StdMat *m = NewDefaultStdMat();
      m->SetName(_T("Default"));
      m->SetAmbient(Color(.7f,.7f,.7f),0);
      m->SetDiffuse(Color(.7f,.7f,.7f),0);
      m->SetSpecular(Color(1.0f,1.0f,1.0f),0);
      m->SetShininess(0.5f,0);
      m->SetShinStr(.707f,0);
      return m;
   }

//----------------------------

   static int MAXMapIndex(int i){
      switch(i) {
      case Ntex:  return ID_DI;
      case Ntex2: return ID_DI;
      case Nopac: return ID_OP;
      case Nbump: return ID_BU;
      case Nspec: return ID_SP;
      case Nshin: return ID_SS;
      case Nselfi:return ID_SI; 
      case Nrefl: return ID_RL;
      default:    return ID_DI;
      }
   }

//----------------------------

   static bool IsSXPName(char *name){

      char fname[30];
      char ext[5];
      _splitpath(name, NULL, NULL, fname, ext );
      return stricmp(ext,".sxp")==0?1:0;
   }

//----------------------------

   static Texmap* MakeTex(MapData& map, SMtl *smtl, BOOL &wasSXP){
      Texmap *txm; 
      wasSXP = FALSE;
      if (map.kind==0){ 
         // Texture Map
   //    if (IsSXPName(map.name)&&(map.p.tex.sxp_data!=NULL)) 
         if (IsSXPName(map.map_name)) {    // DS - 6/11/96
            Tex3D *t3d = GetSXPReaderClass(map.map_name);
            if (t3d) {
               if (map.p.tex.sxp_data) {
                  ULONG *p = (ULONG *)map.p.tex.sxp_data;
                  t3d->ReadSXPData(map.map_name, (void *)(p+1));
                  wasSXP = TRUE;
                  }
               }
            txm = t3d;                    
            }
         else {
            BitmapTex *bmt = NewDefaultBitmapTex();
            bmt->SetMapName(TSTR(map.map_name));
            MapParams &par = map.p.tex;
            bmt->SetAlphaAsMono((par.texflags&TEX_ALPHA_SOURCE)?1:0);
            bmt->SetAlphaSource((par.texflags&TEX_DONT_USE_ALPHA)?ALPHA_NONE:ALPHA_FILE);
            bmt->SetFilterType((par.texflags&TEX_SAT)?FILTER_SAT:FILTER_PYR);
            StdUVGen *uv = bmt->GetUVGen();
            uv->SetUOffs(par.uoffset,0);
            uv->SetVOffs(-par.voffset,0);
            uv->SetUScl(par.uscale,0);
            uv->SetVScl(par.vscale,0);
            uv->SetAng(-((float)atan2(par.ang_sin, par.ang_cos)),0);
            uv->SetBlur(par.texblur+1.0f,0);
            int tile=0;
            if (par.texflags&TEX_MIRROR) tile|= U_MIRROR|V_MIRROR;
            else {
               if (0==(par.texflags&TEX_NOWRAP)) tile|= U_WRAP|V_WRAP;
               }
            uv->SetTextureTiling(tile);
            TextureOutput *txout = bmt->GetTexout();
            txout->SetInvert(par.texflags&TEX_INVERT?1:0);
            txm = bmt;
            }
         if (map.p.tex.texflags&TEX_TINT) {
            // map.p.tex.col1, col2 : stuff into Mix  
            MultiTex* mix = NewDefaultMixTex();
            mix->SetColor(0,ColorFrom24(map.p.tex.col1));
            mix->SetColor(1,ColorFrom24(map.p.tex.col2));
            mix->SetSubTexmap(2, txm);
            txm = mix;        
            }
         else if (map.p.tex.texflags&TEX_RGB_TINT) {
            // map.p.tex.rcol,gcol,bcol : stuf into tint
            MultiTex* mix = NewDefaultTintTex();
            mix->SetColor(0,ColorFrom24(map.p.tex.rcol));
            mix->SetColor(1,ColorFrom24(map.p.tex.gcol));
            mix->SetColor(2,ColorFrom24(map.p.tex.bcol));
            mix->SetSubTexmap(0, txm);
            txm = mix;        
            }
         }
      else {  
         // kind == 1 :  Reflection Map
         BitmapTex *bmt = NewDefaultBitmapTex();
         bmt->SetMapName(TSTR(map.map_name));
         StdUVGen *uv = bmt->GetUVGen();

         // TBD: REFLECTION BLUR SETTING:
         uv->SetBlurOffs((float)smtl->refblur/400.0f+.001f,0);
         bmt->InitSlotType(MAPSLOT_ENVIRON);
         txm = bmt;
         }
      return txm;
   }

//----------------------------

   void ParseIKData(INode *node){

      AppDataChunk *ad = 
         node->GetAppDataChunk(node->ClassID(),node->SuperClassID(),0);
      if(!ad)
         return;
                              //Get the IK data chunk
      Chunk_hdr *hdr = (Chunk_hdr*)
         GetAppDataChunk(ad->data,ad->length,"IK KXPv1 62094j39dlj3i3h42");
      if (!hdr) return;

      // Get the joint data sub chunk
      Chunk_hdr *nhdr = hdr+1;
      DWORD len = hdr->size - nhdr->size - 6;
      void *ikdata = (void*)(((char*)nhdr)+nhdr->size);
      hdr = (Chunk_hdr*)GetAppDataChunk(ikdata,len,"JOINTDATA");
      if (!hdr) return;

      // The first 4 bytes is the version number
      nhdr = hdr+1;
      int *version = (int*)(((char*)nhdr)+nhdr->size);
      if (*version!=3) return; // gotta be version 3

      // Then the joint data
      JointData3DSR4 *jd = (JointData3DSR4*)(version+1);

      // Copy it into our data structures
      InitJointData posData, rotData;
      for (int i=0; i<3; i++) {
         posData.active[i]  = jd->freeJoints[i+3];
         posData.limit[i]   = jd->limited[i+3];
         posData.ease[i]    = jd->ease[i+3];
         posData.min[i]     = jd->jointMins[i+3];
         posData.max[i]     = jd->jointMaxs[i+3];
         posData.damping[i] = jd->damping[i+3];

         rotData.active[i]  = jd->freeJoints[i];
         rotData.limit[i]   = jd->limited[i];
         rotData.ease[i]    = jd->ease[i];
         rotData.min[i]     = jd->jointMins[i];
         rotData.max[i]     = jd->jointMaxs[i];
         rotData.damping[i] = jd->damping[i];
         }

      // Give the data to the TM controller
      node->GetTMController()->InitIKJoints(&posData,&rotData);
   }

//----------------------------

   void Reset(){
      mode = WORKER_IDLE;
      tm.IdentityMatrix();
      gotverts = false;
      gottverts = false;
      gotfaces = false;
      verts = faces = 0;
      id = -1;
      object=NULL;

      nurbs = NULL;
      bone = NULL;
      light = NULL;
      camera=NULL;
      dummy = NULL;
      isDummy = false;
      mesh = NULL;
      shape=NULL;
      spline=NULL;
      splShape = NULL;
      parentNode = thisNode = NULL;
      // NEW INITS
      workNode = NULL;  
      pivot = Point3(0, 0, 0);
      cstShad = 1;
      rcvShad = 1;
      newTV.SetCount(0);
   }

//----------------------------

   dword AddNewTVert(UVVert p){

                              //func shouldn't be called!
      assert(0);
      return -1;
      /*
                              //see if already have it in the new verts.
      int ntv = newTV.Count();
      for(int i=0; i<ntv; i++)
         if(p == newTV[i])
            return i + mesh->numVerts;
                              //otherwise, add it.
      return (DWORD)newTV.Append(1,&p,10) + mesh->numVerts;
      */
   }

//----------------------------

   void FreeObjList(){

      /*
      WkObjList *ptr = objects;
      while(ptr){
         WkObjList *next = ptr->next_obj;
         free(ptr);
         ptr = next;
      }
      objects = NULL;
      */
      obj_map.clear();
   }

//----------------------------

   void FreeNodeList(){

      WkNodeList *ptr = nodes;
      while(ptr){
         WkNodeList *next = ptr->next_node;
         FreeKeyList(&ptr->posList);
         FreeKeyList(&ptr->rotList);
         FreeKeyList(&ptr->scList);
         FreeKeyList(&ptr->colList);
         FreeKeyList(&ptr->hotList);
         FreeKeyList(&ptr->fallList);
         FreeKeyList(&ptr->fovList);
         FreeKeyList(&ptr->rollList);
         free(ptr);
         ptr = next;
      }
   }

//----------------------------

   void *FindObjFromNode(ImpNode *node){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->node == node){
            /*
            WkObjList *optr = objects;
            while(optr){
               if(optr->name == ptr->name)
                  return optr->object;
               optr = optr->next_obj;
            }
            */
            t_obj_map::const_iterator it = obj_map.find(C_string(ptr->name));
            if(it!=obj_map.end())
               return (*it).second.object;
            assert(0);
            return NULL;
         }
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   //bool AddKey(KeyList **list1, Key *key)
   bool AddKey(KeyList **list1, S_key_hdr *key, int key_size){

      if(mode != WORKER_KF){

         return true;
      }
      KeyList *ptr = new KeyList;
      if(!ptr)
         return false;
      KeyList *work = *list1, *last = NULL;
      while(work){
         last = work;
         work = work->next_key;
      }
      if(last)
         last->next_key = ptr;
      else
         *list1 = ptr;
      //ptr->key.key_hdr = *key;
      memcpy(&ptr->key, key, key_size);
      ptr->next_key = NULL;
      return true;
   }

//----------------------------

   void FreeKeyList(KeyList **list1){

      KeyList *ptr = *list1;
      while(ptr){
         KeyList *next = ptr->next_key;
         free(ptr);
         ptr = next;
      }
      *list1 = NULL;
   }

//----------------------------

   static void check_for_wrap(UVVert *tv, int flags){
      float d;
      if(flags&UWRAP){
         float maxu,minu;
         maxu = minu = tv[0].x;
         if (tv[1].x>maxu) maxu = tv[1].x;
         else if (tv[1].x<minu) minu = tv[1].x;
         if (tv[2].x>maxu) maxu = tv[2].x;
         else if (tv[2].x<minu) minu = tv[2].x;
         if((maxu-minu)>0.8f){
            d = (float)ceil(maxu-minu);
            if (tv[0].x<.5f)  tv[0].x += d; 
            if (tv[1].x<.5f)  tv[1].x += d; 
            if (tv[2].x<.5f)  tv[2].x += d; 
         }
      }
      if(flags&VWRAP){
         float maxv,minv;
         maxv = minv = tv[0].y;
         if (tv[1].y>maxv) maxv = tv[1].y;
         else if (tv[1].y<minv) minv = tv[1].y;
         if (tv[2].y>maxv) maxv = tv[2].y;
         else if (tv[2].y<minv) minv = tv[2].y;
         if((maxv-minv)>0.8f){
            d = (float)ceil(maxv-minv);
            if (tv[0].y<.5f)  tv[0].y += d; 
            if (tv[1].y<.5f)  tv[1].y += d; 
            if (tv[2].y<.5f)  tv[2].y += d; 
         }
      }
   }

//----------------------------

public:
   int okay;
   ImpInterface *i;
   Interface *ip;
   TSTR name;
   int mode;
   bool gotverts;
   int verts;
   int tverts;
   bool gottverts;
   bool gotfaces;
   int faces;
   int cstShad;
   int rcvShad;
   BOOL gotM3DMAGIC;
   //Mtl  *sceneMtls[MAX_NUM_OF_MATERIALS];
   //MtlName mtlNames[MAX_NUM_OF_MATERIALS];
   TriObject *object;
   GenLight *light;
   Object  *bone;
   Object   *nurbs;
   NURBSSet nset;  
   int mtlnum;

   Dirlight studioLt;
   GenCamera *camera;
   Camera3DS studioCam;
   SplineShape *splShape;
   BezierShape *shape;
   Spline3D *spline;
   Mesh *mesh;
   ImpNode *thisNode;
   ImpNode *parentNode;
   short id;
   //WkObjList *objects;
   typedef map<C_string, C_work_object> t_obj_map;
   t_obj_map obj_map;
   WkNodeList *nodes;
   WkNodeList *workNode;
   Matrix3 tm;
   Point3 pivot;
   DummyObject *dummy;
   bool isDummy;
   int lightType;
   TSTR nodename;
   // Time stuff:
   bool lengthSet;
   TimeValue length;
   bool segmentSet;
   Interval segment; 
   SHORT trackFlags[NUMTRACKS];
   MtlList loadMtls;
   //vector<Mesh*> mat_meshes;
   vector<vector<Mesh*> > mat_meshes;
   void *appdata;
   DWORD appdataLen;
   Tab<UVVert>newTV;

   float hook_x, hook_y;

//----------------------------

   ObjWorker(ImpInterface *iptr,Interface *ip):
      msc_wk(1.0f),
      nodeLoadNumber(-1),
      appdata(NULL)
   {
      okay=1;
      i=iptr;
      shape=NULL;
      //objects = NULL;
      nodes = NULL;
      dummy=NULL;
      isDummy = false;
      Reset();
      lengthSet = false;
      segmentSet = false;
      appdataLen = 0;
      gotM3DMAGIC = false;
      this->ip = ip;
      /*
      for (int i=0;i<MAX_NUM_OF_MATERIALS;i++){
         sceneMtls[i] = NULL;
         mtlNames[i].s[0] = 0;
      }
      */
      hook_x = hook_y = 0.0f;
   }

//----------------------------

   ~ObjWorker() {

      FinishUp();
      FreeObjList();
      FreeNodeList();
      i->RedrawViews();
   }

//----------------------------

   bool FinishUp(){

      switch(mode){

      case WORKER_IDLE:
                              //nothing to do, return success!
         return true;

      case WORKER_MESH:
         {
                              //must have defined verts!
            if(gotverts){

               // Make this an inverse of the mesh matrix.  This sets up a transform which will
               // be used to transform the mesh from the 3DS Editor-space to the neutral space
               // used by the keyframer.
               Matrix3 itm = Inverse(tm);
               int ix;  
      
               // Transform verts through the inverted mesh transform
               for(ix=0; ix<verts; ++ix){
                  Point3 &p = mesh->getVert(ix);
                  p = p * itm;
                  mesh->setVert(ix,p);
               }

               /* Check for objects that have been flipped: their 3D
                  "parity" will be off */    
               Point3 cp = CrossProd(tm.GetRow(0),tm.GetRow(1));
               if(DotProd(cp,tm.GetRow(2)) < 0){
                  Matrix3 tmFlipX(1);
                  Point3 row0 = tmFlipX.GetRow(0);
                  row0.x = -1.0f;
                  tmFlipX.SetRow(0,row0);
                              //transform verts through the mirror transform
                  for(ix=0; ix<verts; ++ix){
                     Point3 &p = mesh->getVert(ix);
                     p = p * tmFlipX;
                     mesh->setVert(ix,p);
                  }
               }

               mesh->buildNormals();
               mesh->EnableEdgeList(1);            
            }

            if(gottverts){
               int ntv;
               if((ntv=newTV.Count())>0){
                  int oldn = mesh->numTVerts;
                  mesh->setNumTVerts(oldn+ntv, TRUE); 
                  for (int i=0; i<ntv; i++)
                     mesh->tVerts[oldn+i] = newTV[i];
               }
            }

            if(appdata){
                              //see if we can create an object from the appdata
               Object *obj = CreateObjectFromAppData(object,appdata,appdataLen);
               if(obj){
                  obj->AddAppDataChunk(
                     triObjectClassID, 
                     GEOMOBJECT_CLASS_ID, 
                     0, appdataLen, appdata);
                  // first need to save away its mtl num
                  int mnum = -1;
                  if (mesh&&mesh->numFaces)
                     mnum = mesh->faces[0].getMatID();
                  // Then can toss it
                  object->DeleteThis(); // don't need tri object
                  object = NULL;
                  AddObject(obj,OBJ_OTHER,name,NULL,mnum);
                  appdata = NULL;  //DS 3/27/96
                  Reset();
                  return true;
               }
                              //stick app data in the object
               object->AddAppDataChunk(triObjectClassID, GEOMOBJECT_CLASS_ID, 0, appdataLen, appdata);
               appdata = NULL;
               }

            AddObject(object,OBJ_MESH,name,&tm);
            Reset();
         }
         return true;

      case WORKER_KF:
         {
            if(appdata){
                              //stick app data in the node
               thisNode->GetINode()->AddAppDataChunk(thisNode->GetINode()->ClassID(), 
                  thisNode->GetINode()->SuperClassID(), 0, appdataLen, appdata);
               appdata = NULL;
            }

            int type = FindTypeFromNode(thisNode, &mesh);

            if(parentNode) {
               parentNode->GetINode()->AttachChild(thisNode->GetINode());
            }

            switch(type){
            case OBJ_MESH:
            case OBJ_DUMMY:
            case OBJ_TARGET:
            case OBJ_OTHER:
               thisNode->SetPivot(-pivot);
               break;
            }
            Reset();
         }
         return true;

      case WORKER_LIGHT:
         {
                              //must have a light
            if(light){
               //DebugPrint("Light calling AddObject %p (%d) %s\n",light,lightType,name);
               C_work_object &obj =
                  AddObject(light, lightType, name, &tm);            
               //obj = *objects;
               obj.targPos = S_vector(studioLt.tx, studioLt.ty, studioLt.tz);
               obj.srcPos = S_vector(studioLt.x, studioLt.y, studioLt.z);
               Reset();
               return true;
            }
            delete light;
            Reset();
         }
         return true;

      case WORKER_NURBS:
         {
            if(nurbs) {
               int nummat = mtlnum;

               AddObject(nurbs, OBJ_NURBS, name, &tm, nummat);
               Reset();
               return true;
            }
            Reset();
         }
         return true;

      case WORKER_BONE:
         {
            if(bone){
              // thisNode->SetPivot(-pivot);
                              //first input; tm not set
               AddObject(bone, OBJ_BONE, name, &tm);
               //ImpNode *bonenode = FindNode(name);
               Reset();
               return true;
            }
            assert(false);
            Reset();
         }
         return true;

      case WORKER_CAMERA:
                              //must have a camera
         if(camera){
            C_work_object &obj = AddObject(camera, OBJ_CAMERA, name, &tm);
            obj.targPos = S_vector(studioCam.tx, studioCam.ty, studioCam.tz);
            obj.srcPos = S_vector(studioCam.x, studioCam.y, studioCam.z);
            Reset();
            return true;
         }
         delete camera;
         Reset();
         return true;
      }
                              //undefined state
      return false;
   }

//----------------------------

   bool StartMesh(const char *iname){

      if(!FinishUp())
         return false;

      tm.IdentityMatrix();
      name = TSTR(iname);
      object = CreateNewTriObject();
      if(!object)
         return false;

      mesh = &object->GetMesh();
      mode = WORKER_MESH;
      newTV.SetCount(0);
      return true;
   }

//----------------------------

   bool StartNurbs(const char *iname){

      if(!FinishUp())
         return false;

      tm.IdentityMatrix();
      name = TSTR(iname);
      nurbs = NULL;
      mode = WORKER_NURBS; 
      return true;
   }

//----------------------------

   bool CreateNurbs(int type){
      
      assert(mode==WORKER_NURBS);

      SuspendAnimate();
      AnimateOff();

      Matrix3 mat(1);

      nurbs = CreateNURBSObject(NULL, &nset, mat);

      if(!nurbs){
         DebugPrint("Nurbs type %d create failure\n",type);
         return false;
      }

      ResumeAnimate();
      return true;
   }

//----------------------------

   bool StartBone(const char *iname){

      if(!FinishUp())
         return false;
      tm.IdentityMatrix();
      name = TSTR(iname);
      bone = NULL;
      mode = WORKER_BONE;  
      return true;
   }

//----------------------------

   bool CreateBone(){

      assert(mode==WORKER_BONE);

      SuspendAnimate();
      AnimateOff();
                                
      //bone = (Object *)ip->CreateInstance(HELPER_CLASS_ID,Class_ID(BONE_CLASS_ID, 0));
      bone = (Object *)i->Create(HELPER_CLASS_ID,Class_ID(BONE_CLASS_ID, 0));

      //INode *node = bone->node;
      //node->ShowBone(1);

      ResumeAnimate();
      return true;
   }

//----------------------------

   bool StartLight(const char *iname){

      if(!FinishUp())
         return false;
      tm.IdentityMatrix();
      name = TSTR(iname);
      light = NULL;
      mode = WORKER_LIGHT; 
      return true;
   }

//----------------------------

   bool CreateLight(int type){

      assert(mode==WORKER_LIGHT);
      light = i->CreateLightObject(type);
      if(!light){
         DebugPrint("Light type %d create failure\n",type);
         return false;
      }

      lightType = (type == OMNI_LIGHT) ? OBJ_OMNILIGHT : OBJ_SPOTLIGHT;

      SuspendAnimate();
      AnimateOff();

      Point3 col;
      col[0] = studioLt.color[0];
      col[1] = studioLt.color[1];
      col[2] = studioLt.color[2];
      light->SetRGBColor(0,col);
      light->SetUseLight(studioLt.flags&NO_LT_ON);
      light->SetSpotShape(studioLt.flags&NO_LT_RECT ? RECT_LIGHT : CIRCLE_LIGHT);
      light->SetConeDisplay(studioLt.flags&NO_LT_CONE ? 1 : 0, FALSE);
      light->SetUseAtten(studioLt.flags&NO_LT_ATTEN ? 1 : 0);
      light->SetAttenDisplay(0);
      light->SetUseGlobal(studioLt.flags&NO_LT_LOCAL ? 0 : 1);
      light->SetShadow(studioLt.flags&NO_LT_SHAD ? 1 : 0);
      light->SetOvershoot(studioLt.flags&NO_LT_OVER ? 1 : 0);
      light->SetShadowType(studioLt.flags&NO_LT_RAYTR ? 1 : 0);
      light->SetAtten(0, ATTEN_START, studioLt.in_range);
      light->SetAtten(0, ATTEN_END, studioLt.out_range);
      light->SetIntensity(0, studioLt.mult);
      float aspect = 1.0f/studioLt.aspect;   
      light->SetAspect(0, aspect);  
      light->SetMapBias(0, studioLt.lo_bias);
      light->SetMapRange(0, studioLt.shadfilter);
      light->SetMapSize(0, studioLt.shadsize);
      light->SetRayBias(0, studioLt.ray_bias);
      // ? light->SetAbsMapBias(int a);   
      light->SetExclusionList(studioLt.excList);
      light->SetHotspot(0,studioLt.hotsize);
      light->SetFallsize(0,studioLt.fallsize);
      studioLt.excList.SetSize(0);

      ResumeAnimate();
      return true;
   }

//----------------------------

   bool StartCamera(const char *iname){

      if(!FinishUp())
         return false;
      tm.IdentityMatrix();
      name = TSTR(iname);
      camera = NULL;
      mode = WORKER_CAMERA;
      return true;
   }

//----------------------------

   bool CreateCamera(int type){

      assert(mode==WORKER_CAMERA);
      camera = i->CreateCameraObject(type);
      camera->SetEnvRange(0, 0, studioCam.nearplane);
      camera->SetEnvRange(0, 1, studioCam.farplane);
      camera->SetFOV(0, DegToRad(2400.0/studioCam.focal));

      if(!camera){
         DebugPrint("Camera type %d create failure\n",type);
         return false;
      }
      return true;
   }

//----------------------------

   bool StartKF(ImpNode *node){

      if(!FinishUp())
         return false;
      thisNode = node;
      parentNode = NULL;
      mode = WORKER_KF;
      pivot = Point3(0, 0, 0);  // Default pivot
      return true;
   }

//----------------------------

   void SetTm(Matrix3 *transform){ tm = *transform; }

//----------------------------

   bool SetVerts(int count){

      if(gotverts)
         return false;
      if(!mesh->setNumVerts(count))
         return false;
      verts = count;
      gotverts = true;
      return true;
   }

//----------------------------

   bool SetTVerts(int count){

      if(gottverts)
         return false;
      if(!mesh->setNumTVerts(count))
         return false;
      //DebugPrint("Set %d tverts\n",count);
      tverts = count;
      gottverts = true;
      return true;
   }

//----------------------------
   inline int GetVerts() const{ return verts; }

//----------------------------

   bool SetFaces(int count){

      if(gotfaces)
         return false;
      if(!mesh->setNumFaces(count))
         return false;

      // If got texture vertices, set up an equal number of texture faces
      if(gottverts){
         if(!mesh->setNumTVFaces(count))
            return false;
      }
      faces = count;
      gotfaces = true;
      return true;
   }

//----------------------------
   inline int GetFaces() const{ return faces; }

//----------------------------

   bool PutVertex(int index, Verts *v){

      if(!gotverts)
         return false;

      if(index<0 || index>=verts)
         return false;

      mesh->setVert(index, v->x, v->y, v->z);
      return true;
   }

//----------------------------

   bool PutTVertex(int index, UVVert *v){

      if(!gottverts)
         return false;
      if(index<0 || index>=tverts)
         return false;
      mesh->setTVert(index, *v);
      return true;
   }

//----------------------------

   bool PutFace(int index,Faces *f){

      Face *fp = &mesh->faces[index];
      if(!gotfaces)
         return false;
      if(index<0 || index>=faces)
         return false;

                              //if we've got texture vertices, also put out a texture face
      if(gottverts){
         // DS 6/6/96
         // check for wrap.
         SetTVerts(index, f);
   //    TVFace *tf = &mesh->tvFace[index];
   //    tf->setTVerts((DWORD)f->a, (DWORD)f->b, (DWORD)f->c);
      }
      fp->setVerts((int)f->a,(int)f->b,(int)f->c);
      fp->setEdgeVisFlags(f->flags & ABLINE,f->flags & BCLINE,f->flags & CALINE);
      fp->setSmGroup(f->sm_group);
      return true;
   }

//----------------------------

   bool PutSmooth(int index,unsigned long smooth){

      Face *fp = &mesh->faces[index];
      if(!gotfaces)
         return false;
      if(index<0 || index>=faces)
         return false;
      fp->setSmGroup(smooth);
      return true;
   }

//----------------------------

   bool PutFaceMtl(int index, int imtl){

      if(index<mesh->numFaces){
         mesh->faces[index].setMatID(imtl);
         return true;
      } 
      return false;
   }

//----------------------------

   void SetTVerts(int nf, Faces *f){

      UVVert uv[3], uvnew[3]; 
      DWORD ntv[3];
      uvnew[0] = uv[0] = mesh->tVerts[f->a];
      uvnew[1] = uv[1] = mesh->tVerts[f->b];
      uvnew[2] = uv[2] = mesh->tVerts[f->c];
      check_for_wrap(uvnew, f->flags);
      ntv[0] = (uvnew[0]==uv[0]) ? f->a: AddNewTVert(uvnew[0]);
      ntv[1] = (uvnew[1]==uv[1]) ? f->b: AddNewTVert(uvnew[1]);
      ntv[2] = (uvnew[2]==uv[2]) ? f->c: AddNewTVert(uvnew[2]);
      TVFace *tf = &mesh->tvFace[nf];
      tf->setTVerts(ntv[0], ntv[1], ntv[2]);
   }

//----------------------------

   C_work_object &AddObject(Object *object, int type, const TCHAR *name, Matrix3 *tm, int mtlNum = -1){

      pair<t_obj_map::iterator, bool> p =
         obj_map.insert(pair<C_string, C_work_object>(C_string(name), C_work_object()));
      assert(p.second);
      C_work_object &obj = (*p.first).second;

      obj.object = object;
      obj.type = type;
      obj.cstShad = cstShad;
      obj.rcvShad = rcvShad;
      obj.mtln = mtlNum;
      if(tm)
         obj.tm = *tm;
      else
         obj.tm.IdentityMatrix();

      return obj;
   }

//----------------------------

   bool AddNode(ImpNode *node, const TCHAR *name, int type,Mesh *msh, char *owner, int mtlNum = -1){

      WkNodeList *ptr = new WkNodeList;
      if(!ptr)
         return false;
      ptr->node = node;
      ptr->id = -1;
      ptr->type = type;
      ptr->name = name;
      ptr->mesh = msh;
      ptr->mnum = mtlNum;
      ptr->next_node = nodes;
      ptr->parent = NULL;
      ptr->posList = ptr->rotList = ptr->scList = ptr->colList = ptr->hotList = ptr->fallList = ptr->fovList = ptr->rollList = NULL;
      if(type==OBJ_TARGET)
         ptr->owner = TSTR(owner);
      workNode = nodes = ptr;
      thisNode = node;
      return true;
   }

//----------------------------

   bool SetNodeId(ImpNode *node,short id){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->node == node){
            ptr->id = id;
            return true;
         }
         ptr = ptr->next_node;
      }
      this->id = id;
      return false;   
   }

//----------------------------

   bool SetNodesParent(ImpNode *node, ImpNode *parent){

      WkNodeList *ptr = nodes;
      while(ptr) {
         if(node == ptr->node) {
            ptr->parent = parent;
            return true;
         }
         ptr = ptr->next_node;
      }
      return false;
   }

//----------------------------

   C_work_object *FindObject(const C_string &name, int &type, int &cstShad, int &rcvShad, int &mtlNum){

      t_obj_map::iterator it = obj_map.find(name);
      if(it==obj_map.end())
         return NULL;
      C_work_object &obj = (*it).second;
      type = obj.type;
      cstShad = obj.cstShad;
      rcvShad = obj.rcvShad;
      mtlNum = obj.mtln;
      return &obj;
      /*
      TSTR wname(name);
      WkObjList *ptr = objects;
      while(ptr){
         if(_tcscmp(ptr->name, wname)==0){
            type = ptr->type;
            cstShad = ptr->cstShad;
            rcvShad = ptr->rcvShad;
            mtlNum = ptr->mtln;
            return ptr->object;
         }
         ptr = ptr->next_obj;
      }
      return NULL;
      */
   }

//----------------------------

   /*
   bool UseObject(char *name){

      t_obj_map::iterator it = obj_map.find(name);
      if(it==obj_map.end()){
         assert(0);
         return false;
      }
      (*it).second.used = true;
      return true;
   }
   */

//----------------------------

   bool CompleteScene(){

      FinishUp();
                                 //If we need to expand time, ask the user!
      if(lengthSet){
         Interval cur = i->GetAnimRange();
         Interval nrange = Interval(0,length * TIME_CONSTANT);
         if(!(cur==nrange)){
            i->SetAnimRange(nrange);
         }
         /*    
         if((length * TIME_CONSTANT) > cur.End()) {
            if(MessageBox( IDS_TH_EXPANDANIMLENGTH, IDS_RB_3DSIMP, MB_YESNO) == IDYES) {
               cur.SetEnd(length * TIME_CONSTANT);
               i->SetAnimRange(cur);
               }
            }
         */
      }

      WkNodeList *nptr = nodes;
      while(nptr){
         if(nptr->type==OBJ_TARGET && nptr->node){
            WkNodeList *nptr2 = nodes;
            while(nptr2){
               if((nptr2->type==OBJ_CAMERA || nptr2->type==OBJ_SPOTLIGHT) && nptr->owner==nptr2->name && nptr2->node){
                  i->BindToTarget(nptr2->node,nptr->node);
                  goto next_target;       
               }
               nptr2 = nptr2->next_node;
            }
         }
      next_target:
         nptr = nptr->next_node;
      }
      nptr = nodes;

      while(nptr){
         Control *control;

         if(nptr->node){
            ImpNode *theNode = nptr->node;
            INode *theINode = theNode->GetINode();

            switch(nptr->type){
            case OBJ_NURBS:
            case OBJ_MESH:
               if(nptr->type == OBJ_MESH){
                  AssignMtl(theINode, nptr->mesh);
               }else{
                  theINode->SetMtl(loadMtls[FindMatFromNode(theNode)]);
               }
                                 //flow...
            case OBJ_BONE:
                                 //Create interface theINode
               if(nptr->type==OBJ_BONE){
                  theINode->ShowBone(1);
                  theINode->SetWireColor(0x6EFFEB);
               }
                                 //flow...
            case OBJ_DUMMY:
               {
                  control = theINode->GetTMController();
                  MakeControlsTCB(control, nptr->trackFlags);
                  if(control){
                     SuspendAnimate();
                     AnimateOn();
                     if(nptr->posList){
                        Control *posControl = control->GetPositionController();
                        if(posControl)
                           SetControllerKeys(posControl, nptr->posList, KEY_POS);
                     }
                     if(nptr->rotList){
                        Control *rotControl = control->GetRotationController();
                        if(rotControl)
                           SetControllerKeys(rotControl, nptr->rotList, KEY_ROT);
                     }
                     if(nptr->scList){
                        Control *scControl = control->GetScaleController();
                        if(scControl)
                           SetControllerKeys(scControl, nptr->scList, KEY_SCL);
                     }
                     ResumeAnimate();
                  }
               }
               break;

            case OBJ_OMNILIGHT:
               {
                  GenLight *lt = (GenLight *)FindObjFromNode(theNode);
                  if(!lt){
                     assert(0);
                     break;
                  }
                  SuspendAnimate();
                  AnimateOn();
                  control = theINode->GetTMController();
                  MakeControlsTCB(control,nptr->trackFlags);
                  if(control){
                     if(nptr->posList){
                        Control *posControl = control->GetPositionController();
                        if(posControl)
                           SetControllerKeys(posControl, nptr->posList, KEY_POS);
                     }
                  }
                  if(nptr->colList){
                     Control *cont = (Control*)ip->CreateInstance(CTRL_POINT3_CLASS_ID,
                        Class_ID(TCBINTERP_POINT3_CLASS_ID,0));
                     lt->SetColorControl(cont);
                     SetControllerKeys(cont, nptr->colList, KEY_COLOR);
                  }
                  lt->Enable(TRUE);
                  ResumeAnimate();
               }
               break;

            case OBJ_SPOTLIGHT:
               {
                  GenLight *lt = (GenLight *)FindObjFromNode(theNode);
                  float aspect = lt->GetAspect(0);
                  if(!lt){
                     assert(0);
                     break;
                  }
                  SuspendAnimate();
                  AnimateOn();
                  control = theINode->GetTMController();
                  MakeControlsTCB(control,nptr->trackFlags);
                  if(control){
                     if(nptr->posList){
                        Control *posControl = control->GetPositionController();
                        if(posControl) 
                           SetControllerKeys(posControl, nptr->posList, KEY_POS);
                     }
                  }
                  if(nptr->colList){
                     Control *cont = (Control*)ip->CreateInstance(CTRL_POINT3_CLASS_ID,
                        Class_ID(TCBINTERP_POINT3_CLASS_ID,0));
                     lt->SetColorControl(cont);
                     SetControllerKeys(cont, nptr->colList, KEY_COLOR);
                  }
                  if(nptr->hotList){
                     Control *cont = (Control*)ip->CreateInstance(CTRL_FLOAT_CLASS_ID,
                        Class_ID(TCBINTERP_FLOAT_CLASS_ID, 0));
                     lt->SetHotSpotControl(cont);
                     SetControllerKeys(cont, nptr->hotList, KEY_FLOAT, 1.0f, aspect);
                  }
                  if(nptr->fallList){
                     Control *cont = (Control*)ip->CreateInstance(CTRL_FLOAT_CLASS_ID,
                        Class_ID(TCBINTERP_FLOAT_CLASS_ID,0));
                     lt->SetFalloffControl(cont);
                     SetControllerKeys(cont, nptr->fallList, KEY_FLOAT, 1.0f, aspect);
                  }
                  control = theINode->GetTMController();
                  if(control){
                     if(nptr->rollList){
                        Control *rollControl = control->GetRollController();
                        if(rollControl)
                           SetControllerKeys(rollControl, nptr->rollList, KEY_FLOAT, -DEG_TO_RAD);
                     }
                  }
                  lt->Enable(TRUE);
                  ResumeAnimate();
               }
               break;

            case OBJ_CAMERA:
               {
                  GenCamera *cam = (GenCamera *)FindObjFromNode(theNode);
                  if(!cam){
                     assert(0);
                     break;
                  }
                  control = theINode->GetTMController();
                  MakeControlsTCB(control,nptr->trackFlags);
                  if(control){
                     SuspendAnimate();
                     AnimateOn();
                     if(nptr->posList){
                        Control *posControl = control->GetPositionController();
                        if(posControl)
                           SetControllerKeys(posControl, nptr->posList, KEY_POS);
                     }
                     if(nptr->rotList){
                        Control *rotControl = control->GetRotationController();
                        if(rotControl)
                           SetControllerKeys(rotControl, nptr->rotList, KEY_ROT);
                     }
                     if(nptr->scList){
                        Control *scControl = control->GetScaleController();
                        if(scControl)
                           SetControllerKeys(scControl, nptr->scList, KEY_SCL);
                     }
                  }
                  if(nptr->fovList){
                     Control *cont = (Control*)ip->CreateInstance(
                        CTRL_FLOAT_CLASS_ID,
                        Class_ID(TCBINTERP_FLOAT_CLASS_ID,0));
                     cam->SetFOVControl(cont);
                     SetControllerKeys(
                        cont,
                        nptr->fovList,
                        KEY_FLOAT,DEG_TO_RAD);
                  }
                  control = theINode->GetTMController();
                  if(control){
                     if(nptr->rollList) {
                        Control *rollControl = control->GetRollController();
                        if(rollControl) {
                           SetControllerKeys(
                              rollControl,
                              nptr->rollList,
                              KEY_FLOAT,-DEG_TO_RAD);
                           }
                        }
                  }
                  cam->Enable(TRUE);
                  ResumeAnimate();
               }
               break;

            case OBJ_OTHER:
               theINode->SetMtl(GetMaxMtl(nptr->mnum));
                                 //flow...

            case OBJ_TARGET:
               {
                                 //Unload all key info into the Jaguar node
                  control = theINode->GetTMController();
                  MakeControlsTCB(control,nptr->trackFlags);
                  if(control) {
                     SuspendAnimate();
                     AnimateOn();
                     if(nptr->posList){
                        Control *posControl = control->GetPositionController();
                        if(posControl){
                           SetControllerKeys(
                              posControl,
                              nptr->posList,
                              KEY_POS);
                           /*
                           KeyList *k = nptr->posList;
                           while(k) {
                              PosKey *p = &k->key.pos;
                              Point3 trans(p->e[0].p,p->e[1].p,p->e[2].p);
                              posControl->SetValue(k->key.key.time * TIME_CONSTANT, &trans);                   
                              k = (KeyList *)(k->next);
                              }
                           */
                        }
                     }
                     ResumeAnimate();
                  }
               }
               break;

            default:
               assert(0);
               break;
            }

            ParseIKData(theINode);
         }
         nptr = nptr->next_node;
      }

      t_obj_map::const_iterator it;
      for(it=obj_map.begin(); it!=obj_map.end(); it++){
         const C_work_object &wobj = (*it).second;

      //WkObjList *ptr = objects;
      //while(ptr){
         //if(!ptr->used)
         if(!wobj.used){
            ImpNode *node1, *node2;
            //node1 = MakeANode(ptr->name, FALSE, "");
            node1 = MakeANode((*it).first, FALSE, "");

            // DS: 4/30/97
            // This fixes import of files created in 3DS-DOS with the
            // Save Selected command, which don't have keyframe data.
            if(!gotM3DMAGIC) 
               //node1->SetTransform(0, ptr->tm);
               node1->SetTransform(0, wobj.tm);

            //switch(ptr->type)
            switch(wobj.type){
            case OBJ_MESH:
               {
                  INode *inode = node1->GetINode();
                  Mesh *mesh = &(((TriObject *)inode->GetObjectRef())->GetMesh());
                  AssignMtl(inode, mesh);
               }
               break;

            case OBJ_SPOTLIGHT:
            case OBJ_CAMERA:
               {
                  Matrix3 tm(1);
            
                  if(wobj.type==OBJ_SPOTLIGHT){
                     GenLight *lt = (GenLight *)FindObjFromNode(node1);
                     lt->Enable(TRUE);
                  }else{
                     GenCamera *cam = (GenCamera *)FindObjFromNode(node1);
                     cam->Enable(TRUE);
                  }

                                 //create a target
                  //TSTR name = ptr->name + TSTR(_T(".Target"));
                  C_fstring name("%s.Target", (const char*)((*it).first));
                  node2 = MakeANode(name, TRUE, "");                       
                  tm.SetTrans((Point3&)wobj.targPos);
                  node2->SetTransform(0, tm);
                  i->BindToTarget(node1,node2);

                                 //set the position of the light or camera.
                  tm.SetTrans((Point3&)wobj.srcPos);
                  node1->SetTransform(0, tm);
               }
               break;

            case OBJ_OMNILIGHT:
               {
                  GenLight *lt = (GenLight *)FindObjFromNode(node1);
                  lt->Enable(TRUE);
               }
               break;
            }
         }
         //ptr = ptr->next_obj;
      }
      return true;
   }

//----------------------------

   int SetupEnvironment(){

      i->SetAmbient(0, BG.amb_light);   
      i->SetBackGround(0, BG.bkgd_solid);  
      i->SetUseMap(FALSE);

      switch(BG.bgType){
      case BG_SOLID: 
         break;

      case BG_GRADIENT: 
         {
            GradTex *gt = NewDefaultGradTex();
            gt->SetColor(0, BG.bkgd_gradient.topcolor);
            gt->SetColor(1, BG.bkgd_gradient.midcolor);
            gt->SetColor(2, BG.bkgd_gradient.botcolor);
            gt->SetMidPoint(1.0f-BG.bkgd_gradient.midpct);
            gt->GetUVGen()->SetCoordMapping(UVMAP_SCREEN_ENV);
            i->SetEnvironmentMap(gt);
            i->SetUseMap(TRUE);
         }
         break;
      case BG_BITMAP: 
         if (strlen(BG.bkgd_map)>0) {
            BitmapTex *bmt = NewDefaultBitmapTex();
            bmt->SetMapName(TSTR(BG.bkgd_map));
            bmt->GetUVGen()->SetCoordMapping(UVMAP_SCREEN_ENV);
            i->SetEnvironmentMap(bmt);
            i->SetUseMap(TRUE);
            }
         break;
      }  

      switch(BG.envType) {
      case ENV_DISTCUE:
         // BG.distance_cue.nearplane; ??
         // BG.distance_cue.farplane;  ??
         {
         StdFog *fog = NewDefaultStdFog();
         fog->SetType(0);  
         fog->SetColor(Color(0.0f,0.0f,0.0f),0);  
         fog->SetNear(BG.distance_cue.neardim/100.0f,0);  
         fog->SetFar(BG.distance_cue.fardim/100.0f,0);  
         fog->SetFogBackground(BG.dim_bg);
         i->AddAtmosphere(fog);        
         }           
         break;
      case ENV_FOG:{
         StdFog *fog = NewDefaultStdFog();
         fog->SetType(0);  
         fog->SetColor(BG.fog_data.color,0);  
         fog->SetNear(BG.fog_data.neardens/100.0f,0);  
         fog->SetFar(BG.fog_data.fardens/100.0f,0);  
         fog->SetFogBackground(BG.fog_bg);
         i->AddAtmosphere(fog);        
         }
         break;
      case ENV_LAYFOG:{
         StdFog *fog = NewDefaultStdFog();
         fog->SetType(1);  
         int ftype;
         switch(BG.lfog_data.type) {
            case 1:  ftype = FALLOFF_BOTTOM; break;
            case 2:  ftype = FALLOFF_TOP; break;
            default: ftype = FALLOFF_NONE; break;
            }
         fog->SetFalloffType(ftype);  
         fog->SetColor(BG.lfog_data.color,0);  
         fog->SetDensity(BG.lfog_data.density*100.0f,0);  
         fog->SetTop(BG.lfog_data.zmax,0);  
         fog->SetBottom(BG.lfog_data.zmin,0);  
         fog->SetFogBackground(BG.lfog_data.fog_bg);
         i->AddAtmosphere(fog);        
         }
         break;
      }
      i->SetAmbient(0, BG.amb_light);
      return 1;
   }

//----------------------------

   int FindMatFromNode(ImpNode *node){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->node == node){
            /*
            WkObjList *optr = objects;
            while(optr){
               if(optr->name == ptr->name)
                  return ptr->mnum;
               optr = optr->next_obj;
            }
            */
            t_obj_map::const_iterator it = obj_map.find(C_string(name));
            if(it!=obj_map.end())
               return ptr->mnum;

            assert(0);
            return NULL;
         }
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   ImpNode *FindNode(char *name){

      TSTR wname(name);
      WkNodeList *ptr = nodes;
      while(ptr){
         if(_tcscmp(ptr->name,wname)==0)
            return ptr->node;
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   ImpNode *FindNodeFromId(short id){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->id == id)
            return ptr->node;
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   WkNodeList *FindEntry(char *name){

      TSTR wname(name);
      WkNodeList *ptr = nodes;
      while(ptr){
         if(_tcscmp(ptr->name,wname)==0)
            return ptr;
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   WkNodeList *FindEntryFromId(short id){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->id == id)
            return ptr;
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   WkNodeList *FindNodeListEntry(ImpNode *node){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->node == node)
            return ptr;
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   /*
   C_work_object *FindObjListEntry(TSTR &name){

      WkObjList *ptr = objects;
      while(ptr){
         if(ptr->name == name)
            return ptr;
         ptr = ptr->next_obj;
      }
      return NULL;
   }
   */

//----------------------------

   int FindTypeFromNode(ImpNode *node, Mesh **mesh){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->node == node) {
            *mesh =  ptr->mesh;
            return ptr->type;
         }
         ptr = ptr->next_node;
      }
      return -1;
   }

//----------------------------

   TCHAR *NodeName(ImpNode *node){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(node == ptr->node)
            return ptr->name;
         ptr = ptr->next_node;
      }
      return NULL;
   }

//----------------------------

   inline void SetParentNode(ImpNode *node){ parentNode = node; }

   inline void SetPivot(Point3 p){ pivot = p; }

   inline int AddPositionKey(PosKey *key){ return AddKey(&workNode->posList, key, sizeof(PosKey)); }
   inline int AddRotationKey(RotKey *key){ return AddKey(&workNode->rotList, key, sizeof(RotKey)); }
   inline int AddScaleKey(PosKey *key){ return AddKey(&workNode->scList, key, sizeof(PosKey)); }
   inline int AddColorKey(ColorKey *key){ return AddKey(&workNode->colList, key, sizeof(ColorKey)); }
   inline int AddHotKey(ScalarKey *key){ return AddKey(&workNode->hotList, key, sizeof(ScalarKey)); }
   inline int AddFallKey(ScalarKey *key){ return AddKey(&workNode->fallList, key, sizeof(ScalarKey)); }
   inline int AddFOVKey(ScalarKey *key){ return AddKey(&workNode->fovList, key, sizeof(ScalarKey)); }
   inline int AddRollKey(ScalarKey *key){ return AddKey(&workNode->rollList, key, sizeof(ScalarKey)); }

//----------------------------

   /*
   int SetTransform(ImpNode *node, Matrix3 &m){

      WkNodeList *ptr = nodes;
      while(ptr){
         if(ptr->node == node){
            ptr->tm = m;
            return 1;
         }
         ptr = (WkNodeList *)ptr->next;
      }
      return 0;
   }
   */

//----------------------------

   bool ReadyDummy(){

      if(dummy)
         return true;         //already exists!

      dummy = new DummyObject();
      dummy->SetBox(Box3(
         -Point3(0.5f,0.5f,0.5f),
          Point3(0.5f,0.5f,0.5f)));

#if 0
      static int dummyFaceA[] =
         {0,2,0,5,1,6,2,7,3,4,4,6};
      static int dummyFaceB[] =
         {3,1,1,4,2,5,3,6,0,7,5,7};
      static int dummyFaceC[] =
         {2,0,5,0,6,1,7,2,4,3,6,4};

      // Create a triobj (for now)
      TriObject *object = CreateNewTriObject();
      if(!object)
         return false;
      mesh = &object->Mesh();

      if(!mesh->setNumVerts(8)) {
         delete object;
         return false;
      }
      if(!mesh->setNumFaces(12)) {
         delete object;
         return false;
      }

      dummy = object;

      int ix;

      // Stuff faces
      for(ix=0; ix<12; ++ix) {
         Face *fp = &mesh->faces[ix];
         fp->setVerts(dummyFaceA[ix],dummyFaceB[ix],dummyFaceC[ix]);
         fp->setEdgeVisFlags(1,1,0);
         fp->setSmGroup(1 << (ix/2));
      }

      // Stuff verts
      SetDummyBounds(Point3((float)-0.5,(float)-0.5,(float)-0.5),Point3((float)0.5,(float)0.5,(float)0.5));

      mesh->EnableEdgeList(1);
#endif
      return true;
   }

//----------------------------

   ImpNode *MakeDummy(const TCHAR *name){

      if(!ReadyDummy())
         return NULL;
      ImpNode *node = i->CreateNode();
      if(!node || !dummy){
         return NULL;
      }
      node->Reference(dummy);
      tm.IdentityMatrix();    // Reset initial matrix to identity
      node->SetTransform(0, tm);
      i->AddNodeToScene(node);
      node->SetName(name);
      AddNode(node,name,OBJ_DUMMY,mesh,"");
      return node;
   }

//----------------------------

   ImpNode *MakeANode(const C_string &name, BOOL target, char *owner){

      int type, cstShad, rcvShad, mtlNum;
      Object *obj;
                              //it MUST have an object unless it's a target!
      if(!target){
         C_work_object *wobj;
         wobj = FindObject(name, type, cstShad, rcvShad, mtlNum);
         if(!wobj){
            return NULL;
         }
         //UseObject((TCHAR*)name);
         wobj->used = true;
         obj = wobj->object;
      }else{
         type = OBJ_TARGET;
         obj = i->CreateTargetObject();
      }
      if(type==OBJ_MESH)
                                 //v3.1
   //    mesh = &(((TriObject *)obj)->Mesh());
         mesh = &(((TriObject *)obj)->GetMesh());
      else
         mesh = NULL;
      ImpNode *node = i->CreateNode();
      if(!node)
         return NULL;
      node->Reference((ObjectHandle)obj);
      tm.IdentityMatrix();    // Reset initial matrix to identity
      node->SetTransform(0, tm);
      node->GetINode()->SetCastShadows(cstShad);
      node->GetINode()->SetRcvShadows(rcvShad);
      i->AddNodeToScene(node);
      node->SetName(name);
      AddNode(node, name, type, mesh, owner, mtlNum);
      return node;
   }

//----------------------------

   inline void SetDummy(bool x){ isDummy = x; }

//----------------------------

   inline bool IsDummy() const{ return isDummy; }

//----------------------------

   bool SetDummyBounds(Point3& dMin,Point3& dMax){

      if(!isDummy)
         return false;

      dummy->SetBox(Box3(dMin,dMax));  

#if 0
         float dummyVertX[] =
         {(float)dMin.x,(float)dMax.x,(float)dMax.x,(float)dMin.x,(float)dMin.x,(float)dMax.x,(float)dMax.x,(float)dMin.x};
      float dummyVertY[] =
         {(float)dMin.y,(float)dMin.y,(float)dMax.y,(float)dMax.y,(float)dMin.y,(float)dMin.y,(float)dMax.y,(float)dMax.y};
      float dummyVertZ[] =
         {(float)dMin.z,(float)dMin.z,(float)dMin.z,(float)dMin.z,(float)dMax.z,(float)dMax.z,(float)dMax.z,(float)dMax.z};

      int ix;

      Mesh *mesh = &dummy->Mesh();

      // Stuff verts
      for(ix=0; ix<8; ++ix)
         mesh->setVert(ix,Point3(dummyVertX[ix],dummyVertY[ix],dummyVertZ[ix]));

      mesh->buildNormals();
#endif
      return true;
   }

//----------------------------
   inline void SetNodeName(const TCHAR *name){ nodename = name; }

//----------------------------

   void SetInstanceName(ImpNode *node, const TCHAR *iname){

      TSTR instancename(iname);
      if(!nodename.Length() && instancename.Length())
         node->SetName(iname);
      else
      if(!instancename.Length() && nodename.Length())
         node->SetName(nodename);
      else{

      // TSTR fullname = nodename + TSTR(_T(".")) + iname;
      // node->SetName(fullname);

         node->SetName(iname);
      }
   }

//----------------------------

   inline void SetAnimLength(TimeValue l){ length = l; lengthSet = true; }

//----------------------------
   inline void SetSegment(Interval seg){ segment = seg; segmentSet = true; }

//----------------------------

   void SetControllerKeys(Control *cont, KeyList *keys, int type, float f=1.0f,float aspect=-1.0f){

      KeyList *ptr = keys;
      int ct=0;
      ITCBFloatKey fkey;
      ITCBPoint3Key pkey;
      ITCBRotKey rkey;
      ITCBScaleKey skey;
      ITCBKey *k;

                              //Set up 'k' to point at the right derived class
      switch (type) {
      case KEY_FLOAT: k = &fkey; break;
      case KEY_POS: k = &pkey; break;
      case KEY_ROT: k = &rkey; break;
      case KEY_SCL: k = &skey; break;
      case KEY_COLOR: k = &pkey; break;
      default: return;
      }

      // Count how many keys we got.
      while(ptr){
         ct++;
         ptr = ptr->next_key;
      }

      // Special case where there is only one key at frame 0.
      // Just set the controller's value
      if(ct==1 && keys->key.key_hdr.time==0){
         switch(type){
         case KEY_FLOAT:
            {
               float flt;
               if(aspect>0){
                  float tn = (float)tan(DegToRad(keys->key.sc.e[0].p/2.0f));
                  flt = (float)RadToDeg(atan(tn/sqrt(aspect)))*2.0f;
               }else{
                  flt = keys->key.sc.e[0].p * f;
               }
               cont->SetValue(0,&flt);
            }
            break;

         case KEY_COLOR:
            {
               Point3 p(
                  keys->key.col.c[0],
                  keys->key.col.c[1],
                  keys->key.col.c[2]);
               cont->SetValue(0,&p);
            }
            break;

         case KEY_POS:
            {
               Point3 p(
                  keys->key.pos.e[0].p,
                  keys->key.pos.e[1].p,
                  keys->key.pos.e[2].p);
               cont->SetValue(0,&p);
            }
            break;

         case KEY_ROT:
            {
               Point3 axis(
                  keys->key.rot.axis[0],
                  keys->key.rot.axis[1],
                  keys->key.rot.axis[2]);
               Quat q = QFromAngAxis(keys->key.rot.angle, axis);
               cont->SetValue(0,&q);
            }
            break;

         case KEY_SCL:
            {
               ScaleValue s(
                  Point3(
                  keys->key.pos.e[0].p,
                  keys->key.pos.e[1].p,
                  keys->key.pos.e[2].p));
               cont->SetValue(0,&s);
            }
            break;
         }
         return;
      }

                              //Make sure we can get the interface and have some keys
      IKeyControl *ikeys = GetKeyControlInterface(cont);
      if (!ct || !ikeys) return;
   
                              //Allocate the keys in the table
      ikeys->SetNumKeys(ct);

      ptr = keys;
      ct  = 0;
      while (ptr) {
         // Set the common values
         k->time    = ptr->key.key_hdr.time * GetTicksPerFrame();
         k->tens    = ptr->key.key_hdr.tens;
         k->cont    = ptr->key.key_hdr.cont;
         k->bias    = ptr->key.key_hdr.bias;
         k->easeIn  = ptr->key.key_hdr.easeTo;
         k->easeOut = ptr->key.key_hdr.easeFrom;
      
         // Set the key type specific values
         switch (type) {
            case KEY_FLOAT:
               if (aspect>0) {
                  float tn = (float)tan(DegToRad(ptr->key.sc.e[0].p/2.0f));
                  fkey.val = (float)RadToDeg(atan(tn/sqrt(aspect)))*2.0f;
               } else {
                  fkey.val = ptr->key.sc.e[0].p * f;
                  }
               break;

            case KEY_POS:  
               pkey.val[0] = ptr->key.pos.e[0].p; 
               pkey.val[1] = ptr->key.pos.e[1].p; 
               pkey.val[2] = ptr->key.pos.e[2].p; 
               break;

            case KEY_ROT:  
               rkey.val.angle   = ptr->key.rot.angle; 
               rkey.val.axis[0] = ptr->key.rot.axis[0];
               rkey.val.axis[1] = ptr->key.rot.axis[1];
               rkey.val.axis[2] = ptr->key.rot.axis[2];
               break;

            case KEY_SCL:  
               skey.val.s[0] = ptr->key.pos.e[0].p; 
               skey.val.s[1] = ptr->key.pos.e[1].p; 
               skey.val.s[2] = ptr->key.pos.e[2].p;
               break;

            case KEY_COLOR:   
               pkey.val[0] = ptr->key.col.c[0]; 
               pkey.val[1] = ptr->key.col.c[1]; 
               pkey.val[2] = ptr->key.col.c[2]; 
               break;
            }
      
                              //set the key in the table
         ikeys->SetKey(ct++, k);     
         ptr = ptr->next_key;
      }
   
      // This will ensure that the keys are sorted by time at the
      // track is invalidated so the tangents will be recomputed.
      ikeys->SortKeys();
   }

//----------------------------

   void MakeControlsTCB(Control *tmCont,SHORT *tflags){

      Control *c;
      DWORD flags=INHERIT_ALL;

      // Setup inheritance flags.
      if(tflags[POS_TRACK_INDEX] & NO_LNK_X) flags &= ~INHERIT_POS_X;
      if(tflags[POS_TRACK_INDEX] & NO_LNK_Y) flags &= ~INHERIT_POS_Y;
      if(tflags[POS_TRACK_INDEX] & NO_LNK_Z) flags &= ~INHERIT_POS_Z;

      if(tflags[ROT_TRACK_INDEX] & NO_LNK_X) flags &= ~INHERIT_ROT_X;
      if(tflags[ROT_TRACK_INDEX] & NO_LNK_Y) flags &= ~INHERIT_ROT_Y;
      if(tflags[ROT_TRACK_INDEX] & NO_LNK_Z) flags &= ~INHERIT_ROT_Z;

      if(tflags[SCL_TRACK_INDEX] & NO_LNK_X) flags &= ~INHERIT_SCL_X;
      if(tflags[SCL_TRACK_INDEX] & NO_LNK_Y) flags &= ~INHERIT_SCL_Y;
      if(tflags[SCL_TRACK_INDEX] & NO_LNK_Z) flags &= ~INHERIT_SCL_Z;

      tmCont->SetInheritanceFlags(flags,FALSE);

      c = tmCont->GetPositionController();
      if(c && c->ClassID()!=Class_ID(TCBINTERP_POSITION_CLASS_ID,0)){
         Control *tcb = (Control*)ip->CreateInstance(
            CTRL_POSITION_CLASS_ID,
            Class_ID(TCBINTERP_POSITION_CLASS_ID,0));
         if(!tmCont->SetPositionController(tcb)){
            tcb->DeleteThis();
         }
      }

      c = tmCont->GetRotationController();
      if(c && c->ClassID()!=Class_ID(TCBINTERP_ROTATION_CLASS_ID,0)){
         Control *tcb = (Control*)ip->CreateInstance(
            CTRL_ROTATION_CLASS_ID,
            Class_ID(TCBINTERP_ROTATION_CLASS_ID,0));
         if (!tmCont->SetRotationController(tcb)) {
            tcb->DeleteThis();
         }
      }

      c = tmCont->GetRollController();
      if(c && c->ClassID()!=Class_ID(TCBINTERP_FLOAT_CLASS_ID,0)){
         Control *tcb = (Control*)ip->CreateInstance(
            CTRL_FLOAT_CLASS_ID,
            Class_ID(TCBINTERP_FLOAT_CLASS_ID,0));
         if (!tmCont->SetRollController(tcb)) {
            tcb->DeleteThis();
         }
      }

      c = tmCont->GetScaleController();
      if(c && c->ClassID()!=Class_ID(TCBINTERP_SCALE_CLASS_ID,0)){
         Control *tcb = (Control*)ip->CreateInstance(
            CTRL_SCALE_CLASS_ID,
            Class_ID(TCBINTERP_SCALE_CLASS_ID,0));
         if (!tmCont->SetScaleController(tcb)) {
            tcb->DeleteThis();
         }
      }
   }

//----------------------------

   Mtl *GetMaxMtl(int i){

      if(i<0)
         return NULL;
      if(i>=loadMtls.Count())
         return NULL;
      return loadMtls[i];
   }

//----------------------------

   void AssignMtl(INode *theINode, Mesh *mesh){

      short used[MAX_NUM_OF_MATERIALS],
         cused[MAX_NUM_OF_MATERIALS],
         remap[MAX_NUM_OF_MATERIALS];
      int i;

                              //first check if another instance of this mesh has already
                              // had a material assigned to it
      //for(i=0; i<loadMtlMesh.Count(); i++)
      for(i=mat_meshes.size(); i--; ){
         //if(loadMtlMesh[i]==mesh)
         if(find(mat_meshes[i].begin(), mat_meshes[i].end(), mesh) != mat_meshes[i].end()){
            theINode->SetMtl(loadMtls[i]);
            return;
         }
      }
      for(i=0; i<MAX_NUM_OF_MATERIALS; i++)
         used[i] = 0;
                              //see if a multi-mtl is required
      int nmtl, numMtls = 0;
      for(i =0; i<mesh->numFaces; i++){
         nmtl = mesh->faces[i].getMatID();
         assert(nmtl<MAX_NUM_OF_MATERIALS);
         if(!used[nmtl]){
            used[nmtl] = 1;
            remap[nmtl] = numMtls;
            cused[numMtls++] = nmtl;
         }
      }
      if(numMtls>1){
                              //need a Multi-mtl
                              // scrunch the numbers down to be local to this multi-mtl
         for(i =0; i<mesh->numFaces; i++){
            Face &f = mesh->faces[i];
            int id = f.getMatID();
            f.setMatID(remap[id]);
         }
                              //create a new multi with numMtls, and set them
                              // to GetMaxMtl(cused[i]), i==0..numMtls-1
         MultiMtl *newmat = NewDefaultMultiMtl();
         newmat->SetNumSubMtls(numMtls);
         for(i=0; i<numMtls; i++) 
            newmat->SetSubMtl(i,GetMaxMtl(cused[i]));
         theINode->SetMtl(newmat);
         loadMtls.AddMtl(newmat);         
         //loadMtlMesh.Append(1,&mesh,10);
         //mat_meshes.push_back(mesh);
         mat_meshes.push_back(vector<Mesh*>());
         mat_meshes.back().push_back(mesh);
      }else{
         if(mesh->getNumFaces()){
            nmtl = mesh->faces[0].getMatID();
            for(i =0; i<mesh->numFaces; i++) 
               mesh->faces[i].setMatID(0);
            theINode->SetMtl(GetMaxMtl(nmtl));
            //loadMtlMesh[nmtl] = mesh;
            //mat_meshes[nmtl] = mesh;
            mat_meshes[nmtl].push_back(mesh);
         }
      }
   }

//----------------------------
   
   int GetMatNum(char *name){
      TSTR s(name);
      for (int i=0; i<loadMtls.Count(); i++) {
         if (_tcscmp(s,loadMtls[i]->GetName())==0)
            return i;
         }
      assert(0);
      return 0;
   }

//----------------------------
// Convert mesh mtl to max standard matl.
   void AddMeshMtl(SMtl *smtl){

      StdMat *m;
      //Mesh *mesh = NULL;
      if(smtl==NULL){
         m = New3DSDefaultMtl();
         loadMtls.AddMtl(m);        
         //loadMtlMesh.Append(1,&mesh,10);
         //mat_meshes.push_back(NULL);
         mat_meshes.push_back(vector<Mesh*>());
         return;
      }
      m = NewDefaultStdMat();
      m->SetName(TSTR(smtl->name));
      int shade;
      switch(smtl->shading){
      case REND_FLAT:  shade = SHADE_CONST; break;
      case REND_METAL: shade = SHADE_METAL; break;
      default:     shade = SHADE_PHONG; break;
      }
      m->SetShading(shade);
      m->SetAmbient(ColorFrom24(smtl->amb),0);
      m->SetDiffuse(ColorFrom24(smtl->diff),0);
      m->SetFilter(ColorFrom24(smtl->diff),0);
      m->SetSpecular(ColorFrom24(smtl->spec),0);
      m->SetShininess((float)smtl->shininess/100.0f,0);
      m->SetShinStr((float)smtl->shin2pct/100.0f,0);
      m->SetOpacity(1.0f-(float)smtl->transparency/100.0f,0);
      m->SetOpacFalloff((float)smtl->xpfall/100.0f, 0);     
      m->SetFalloffOut(smtl->flags&MF_XPFALLIN?0:1);  
      m->SetSelfIllum((float)smtl->selfipct/100.0f,0);
      m->SetWireSize(smtl->wiresize,0);
      m->SetFaceMap(smtl->flags&MF_FACEMAP?1:0);
      m->SetSoften(smtl->flags&MF_PHONGSOFT?1:0);
      m->SetWire(smtl->flags&MF_WIRE?1:0);
      m->SetTwoSided(smtl->flags&MF_TWOSIDE?1:0);
      m->SetTransparencyType(smtl->flags&MF_ADDITIVE ? TRANSP_ADDITIVE : TRANSP_FILTER);
      m->SetWireUnits(smtl->flags&MF_WIREABS?1:0);


      if(smtl->map){
         Texmap *txm;
         float amt,amt0;
         BOOL gotTex=0;
         BOOL dum;
         for(int i=0; i<8; i++){
            if(smtl->map[i]==NULL) 
               continue;
            Mapping &mp = *(smtl->map[i]);
            int n = MAXMapIndex(i);
            if(i==Nrefl){
               amt = (float)mp.amt.pct/100.0f;
               RMapParams &par = mp.map.p.ref;
               if(par.acb.flags&AC_ON){ 
               // Mirror or Auto-cubic
               if (par.acb.flags&AC_MIRROR) {
                  StdMirror *mir = NewDefaultStdMirror();
                  txm = (Texmap *)mir;
                  mir->SetDoNth(par.acb.flags&AC_FIRSTONLY?0:1);
                  mir->SetNth(par.acb.nth);
                  }
               else {
                  StdCubic *cub = NewDefaultStdCubic();
                  txm = (Texmap *)cub;
                  cub->SetSize(par.acb.size,0);
                  cub->SetDoNth(par.acb.flags&AC_FIRSTONLY?0:1);
                  cub->SetNth(par.acb.nth);
                  }
               }else{   
               // Environment map
               txm = MakeTex(mp.map,smtl,dum);
               }


            if (strlen(mp.mask.map_name)>0) {
               // make a Mask texmap.
               Texmap *masktex = (Texmap *)CreateInstance(TEXMAP_CLASS_ID, Class_ID(MASK_CLASS_ID,0));
               masktex->SetSubTexmap(1,MakeTex(mp.mask,smtl,dum));
               masktex->SetSubTexmap(0,txm);
               txm = masktex;
               }

            m->SetSubTexmap(n,txm);
            amt = (float)mp.amt.pct/100.0f;
            m->SetTexmapAmt(n, amt, 0);
            }else{
            
            // non-reflection maps
            amt = (float)mp.amt.pct/100.0f;

            // DS: 4/30/97 correct for new interpretation of the
            // amount sliders.
            /*
            switch(i) {
               case Nopac:  
                  if (amt<1.0f) 
                     m->SetOpacity(0.0f,0); 
                  break;
               case Nselfi: 
                  if (amt<1.0f) 
                     m->SetSelfIllum(0.0f,0); 
                  break;
               case Nshin:  
                  // Shininess mapping in 3DS was really shininess strength mapping
                  amt*= (float)smtl->shin2pct/100.0f;
                  m->SetShinStr(0.0f,0);
                  break;
               }
               */

            BOOL wasSXP;
            txm = MakeTex(mp.map,smtl,wasSXP);
            if (n==ID_BU&&!wasSXP) amt *= 10.0f;
            m->SetTexmapAmt(n, amt, 0);
            if (strlen(mp.mask.map_name)>0) {
               // make a Mask texmap.
               Texmap *masktex = (Texmap *)CreateInstance(TEXMAP_CLASS_ID, Class_ID(MASK_CLASS_ID,0));
               masktex->SetSubTexmap(1,MakeTex(mp.mask,smtl,dum));
               masktex->SetSubTexmap(0,txm);
               txm = masktex;
               }
            if (i==Ntex2) {
               if (gotTex) {
                  // Make a Composite texmap
                  MultiTex *comp = NewDefaultCompositeTex();
                  comp->SetNumSubTexmaps(2);
                  Texmap *tm0 = m->GetSubTexmap(ID_DI);
                  comp->SetSubTexmap(0,tm0);
                  comp->SetSubTexmap(1,txm);
                  m->SetSubTexmap(ID_DI,comp);                 
                  if (tm0) 
                     tm0->SetOutputLevel(0,amt0);
                  if (txm)
                     txm->SetOutputLevel(0,amt);
                  m->SetTexmapAmt(ID_DI,1.0f,0);
                  }
               else {
                  m->SetSubTexmap(ID_DI,txm);                  
                  m->SetTexmapAmt(n,amt,0);
                  }
               } 
            else 
               m->SetSubTexmap(n,txm);
            if (i==Ntex&&txm) {
               gotTex=1;
               amt0 = amt;
               }
            }
            m->EnableMap(n,mp.use);
         }
      }
      loadMtls.AddMtl(m);        
      //loadMtlMesh.Append(1,&mesh,10);
      //mat_meshes.push_back(NULL);
      mat_meshes.push_back(vector<Mesh*>());
   }

//----------------------------

   bool LoadAppData(){

      appdata = malloc(Size());
      if(!appdata)
         return false;
      appdataLen = Size();
      C_chunk::Read(appdata, appdataLen);
      return true;
   }

//----------------------------

   bool GetChunk(word tag, void *data);
};

//==========================================================================
// Setup Environment
//==========================================================================

class StudioImport: public SceneImport{
public:
   StudioImport(){
   }

//----------------------------

   virtual ~StudioImport(){
   }

//----------------------------

   virtual int ExtCount(){ return 1; }

//----------------------------

   virtual const TCHAR *Ext(int n){
      switch(n) {
      case 0:
         return _T("I3D");
      }
      return _T("");
   }

   virtual const TCHAR *LongDesc(){ return "Insanity import"; }
   virtual const TCHAR *ShortDesc(){ return "I3D"; }
   virtual const TCHAR *AuthorName(){ return "Lonely Cat Games"; }
   virtual const TCHAR *CopyrightMessage(){ return "Copyright (c) 2002 Lonely Cat Games"; }
   virtual const TCHAR *OtherMessage1(){ return _T(""); }
   virtual const TCHAR *OtherMessage2(){ return _T(""); }
   virtual unsigned int Version(){ return 200; }
   virtual void ShowAbout(HWND hWnd){ }

   virtual int DoImport(const TCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts=FALSE);
};

//----------------------------

typedef unsigned short ushort;

typedef struct{
   ushort id;
   ulong length;
   char data[100];
} ChunkHdr;

//----------------------------
                              //v3.1

BOOL WINAPI DllMain(HINSTANCE hinstDLL,ULONG fdwReason,LPVOID lpvReserved){

   static bool controlsInit = false;
   if(!controlsInit){
      controlsInit = true;
      
      InitCustomControls(hinstDLL);
      InitCommonControls();
   }
   switch(fdwReason){
   case DLL_PROCESS_ATTACH:
      break;
   case DLL_THREAD_ATTACH:
      break;
   case DLL_THREAD_DETACH:
      break;
   case DLL_PROCESS_DETACH:
      break;
   }
   return(TRUE);
}

//----------------------------
// Get a null-terminated string from the file.
bool read_string(C_chunk *ck, char *string, int maxsize){

   while(maxsize--){
      if(!ck->Read(string, 1))
         return false;
      if(*(string++)==0)
         return true;
   }
   return false;
}

//----------------------------

bool tchar_read_string(C_chunk *ck, TSTR &tstr){

#define JUSTEASYINFOABOUTMAXLENGTHOFSTRING 4096

   int maxsize = JUSTEASYINFOABOUTMAXLENGTHOFSTRING;
   char string;
   char buf[2];

   while(maxsize--){
      if(!ck->Read(&string, 1))
         return false;
      sprintf(buf, "%c", string);
      tstr += TSTR(buf);
      
      if(string == 0)
         return true;
   }
   return false;
}

//----------------------------

struct S_rgb{
   byte b, g, r;
};

//----------------------------
#undef RDLONG
#define RDLONG(p) C_chunk::Read(p, sizeof(long))
#undef RDFLOAT
#define RDFLOAT(p) C_chunk::Read(p, sizeof(float));
#undef RD3FLOAT
#define RD3FLOAT(p) C_chunk::Read(p, sizeof(S_vector));

bool ObjWorker::GetChunk(word tag, void *data){

   switch(tag){
   case KFDATA:
      while(C_chunk::Size()){
         if(!RAscend(tag))
            return false;
         switch(tag){
         case OBJECT_NODE_TAG:
         case CAMERA_NODE_TAG: 
         case TARGET_NODE_TAG:   
         case L_TARGET_NODE_TAG:    
         case LIGHT_NODE_TAG:
         case SPOTLIGHT_NODE_TAG:
         case DIRECTLIGHT_NODE_TAG:
         case BONE_NODE_TAG:

            nodeLoadNumber++;
                              //flow...
         case KFHDR:
         case KFSEG:
            if(!GetChunk(tag, NULL))
               return false;
            break;
         default:
            Descend();
         }
      }
      break;

   case KFHDR:
      {
         char dumstr[10];
         LONG length;
         C_chunk::Read(&readVers, sizeof(short));
         if(!read_string(this, dumstr, 9))
            return false;
         C_chunk::Read(&length, sizeof(dword));
         SetAnimLength((TimeValue)length);
      }
      break;

   case KFSEG:
      {
         long segStart, segEnd;
         C_chunk::Read(&segStart, sizeof(dword));
         C_chunk::Read(&segEnd, sizeof(dword));
         SetSegment(Interval((TimeValue)segStart, (TimeValue)segEnd));
      }
      break;

   case CAMERA_NODE_TAG:
   case OBJECT_NODE_TAG:
   case LIGHT_NODE_TAG:
   case BONE_NODE_TAG:
   case DIRECTLIGHT_NODE_TAG:
   case SPOTLIGHT_NODE_TAG:
      {
         FinishUp();
         cur_node_id = -32000;
         while(C_chunk::Size()){
            if(!C_chunk::RAscend(tag))
               return false;
            switch(tag){
            case NODE_HDR:
            case NODE_ID:
            case APP_DATA:
            case BOUNDBOX:
            case PIVOT:
            case INSTANCE_NAME:
            case POS_TRACK_TAG: 
            case ROT_TRACK_TAG:
            case SCL_TRACK_TAG:
            case ROLL_TRACK_TAG: 
            case FOV_TRACK_TAG: 
            case HIDE_TRACK_TAG:
            case NOTE_TRACK_TAG:
            case PROPERTIES_TAG:
            case COL_TRACK_TAG: 
            case HOT_TRACK_TAG: 
            case FALL_TRACK_TAG: 
               if(!GetChunk(tag, NULL))
                  return false;
               break;
            default:
               Descend();
            }
         }
      }
      break;

   case L_TARGET_NODE_TAG: 
   case TARGET_NODE_TAG: 
      {
         FinishUp();
         cur_node_id = -32000;   
         while(C_chunk::Size()){
            if(!C_chunk::RAscend(tag))
               return false;
            switch(tag){
            case NODE_ID:
            case NODE_HDR:
               if(!GetChunk(tag, ".target"))
                  return false;
               break;
            case APP_DATA:
            case POS_TRACK_TAG: 
            case HIDE_TRACK_TAG:
            case NOTE_TRACK_TAG:
               if(!GetChunk(tag, NULL))
                  return false;
               break;
            default:
               Descend();
            }
         }
      }
      break;

   case NODE_ID:
      {
         C_chunk::Read(&cur_node_id, sizeof(short));
      }
      break;

   case NODE_HDR:
      {
         USHORT npar;
         ImpNode *pnode;
                              //if we have a suffix, this is a target object
         char *suffix = (char*)data;
                  
         if(!read_string(this, obname, OBJ_NAME_LEN))
            return false;
         obname[OBJ_NAME_LEN] = 0;
         char prefix[OBJ_NAME_LEN];
         if(suffix){
            strcpy(prefix, obname);           
            strcat(obname, suffix);
         }else
            prefix[0] = 0;

         //TSTR Wname(obname);
         C_string Wname = obname;

         SetDummy(0);

         // Check for dummies -- If it is a dummy, we create a node now because
         // none was created for it in the mesh section
         if(strcmp(obname,"DUMMY")==0 || strcmp(obname,"$$$DUMMY")==0){
            if(!(obnode = MakeDummy(_T(""))))
               return false;
            SetDummy(1);
            SetNodeName(_T(""));
         }else{
            //TSTR wname(obname);
            if(!(obnode = MakeANode(obname, suffix ? true : false, prefix)))
               return false;
            SetNodeName(Wname);
         }

         StartKF(obnode);   // Start creating a KF node
                     
         if(cur_node_id!=-32000)
            SetNodeId(obnode,cur_node_id);
         else 
            SetNodeId(obnode, nodeLoadNumber);
         C_chunk::Read(&nodeflags, sizeof(short));

         if(readVers<4) nodeflags &= 7;
                              //read second flag word
         C_chunk::Read(&nodeflags2, sizeof(short));
                              //get parent node if any
         C_chunk::Read(&npar, sizeof(short));
         
         pnode = NULL;
         if(npar != 0xffff){
            WkNodeList *ptr = FindEntryFromId(npar);
            if(!ptr){
            }else{
               pnode = ptr->node;
               SetNodesParent(obnode,pnode);
            }
         }
         SetParentNode(pnode);
      }
      break;

   case PIVOT:
      {
         float pivot[3];
         C_chunk::Read(&pivot, sizeof(S_vector));
         if(msc_wk!=1.0f){
            pivot[0]*=msc_wk;
            pivot[1]*=msc_wk;
            pivot[2]*=msc_wk;
         }
         SetPivot(Point3(pivot[0], pivot[1], pivot[2]));
      }
      break;

   case BOUNDBOX:
      {
         Point3 min,max;
         C_chunk::Read(&min, sizeof(S_vector));
         C_chunk::Read(&max, sizeof(S_vector));
         SetDummyBounds(min,max);
      }
      break;

   case INSTANCE_NAME:
      {
         char iname[MAX_NAME_SIZE];
         if(!read_string(this, iname, MAX_NAME_SIZE))
            return false;

         iname[MAX_NAME_SIZE - 1] = 0;
         SetInstanceName(obnode, TSTR(iname));
      }
      break;

   case PROPERTIES_TAG:
      {
         TSTR buffer;
         if(!tchar_read_string(this, buffer))
            return false;
         thisNode->GetINode()->SetUserPropBuffer(buffer);
      }
      break;

   case NOTE_TRACK_TAG:
      {
         int ntracks=0;
         DefNoteTrack *ntra;

         C_chunk::Read(&ntracks, sizeof(int));

         for(int i =0; i<ntracks;i++){
            ntra = (DefNoteTrack *)NewDefaultNoteTrack();

            int nkeys;
            C_chunk::Read(&nkeys, sizeof(int));

            for(int j = 0; j<nkeys;j++){
               TimeValue t;
               C_chunk::Read(&t, sizeof(TimeValue));
               TSTR buffer;

               if(!tchar_read_string(this, buffer)) 
                  return false;

               ntra->AddNewKey(t*GetTicksPerFrame(), 0);
               ntra->keys[j]->note = buffer;
            }
                                                   
            thisNode->GetINode()->AddNoteTrack(ntra);
         }
      }
      break;

#define BOOL_CONTROL_CLASS_ID Class_ID(0x984b8d27,0x938f3e43)
   case HIDE_TRACK_TAG:
      {
         SuspendAnimate();
         AnimateOn();

         Control *cont = (Control *)CreateInstance(CTRL_FLOAT_CLASS_ID, BOOL_CONTROL_CLASS_ID);

         int numkeys;
         C_chunk::Read(&numkeys, sizeof(int));
         TimeValue t;
         for(int i =0; i<numkeys;i++){
            C_chunk::Read(&t, sizeof(TimeValue));
            cont->AddNewKey(t*GetTicksPerFrame(), 0);
         }
         thisNode->GetINode()->SetVisController(cont);
         ResumeAnimate();
      }
      break;
      
   case POS_TRACK_TAG:
   case ROT_TRACK_TAG:
   case SCL_TRACK_TAG:
   case FOV_TRACK_TAG:
   case ROLL_TRACK_TAG:
   case COL_TRACK_TAG:
   case HOT_TRACK_TAG:
   case FALL_TRACK_TAG:
      {
         LONG nkeys;
         SHORT rflags;
         //Key key;
         S_key_hdr key_hdr;
         short trflags;
         long trtmin,trtmax;
         C_chunk::Read(&trflags, sizeof(short));
         RDLONG(&trtmin);
         RDLONG(&trtmax);
         RDLONG(&nkeys);
         for(int ix=0; ix<nkeys; ix++){
            memset(&key_hdr, 0, sizeof(key_hdr));
            RDLONG(&key_hdr.time);
            C_chunk::Read(&rflags, sizeof(short));
            if(rflags&W_TENS)
               RDFLOAT(&key_hdr.tens);
            if(rflags&W_CONT)
               RDFLOAT(&key_hdr.cont);
            if(rflags&W_BIAS)
               RDFLOAT(&key_hdr.bias);
            if(rflags&W_EASETO)
               RDFLOAT(&key_hdr.easeTo);
            if(rflags&W_EASEFROM)
               RDFLOAT(&key_hdr.easeFrom);

            switch(tag){
            case POS_TRACK_TAG:  
               {
                  PosKey pkey;
                  memset(&pkey, 0, sizeof(PosKey));
                  //memcpy(&pkey, &key, sizeof(PosKey));
                  memcpy(&pkey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&pkey.e[0].p);
                  RDFLOAT(&pkey.e[1].p);
                  RDFLOAT(&pkey.e[2].p);

                  if(msc_wk!=1.0f){
                     pkey.e[0].p *= msc_wk;
                     pkey.e[1].p *= msc_wk;
                     pkey.e[2].p *= msc_wk;
                  }
                  //trackFlags[POS_TRACK_INDEX] = trflags;
                  workNode->trackFlags[POS_TRACK_INDEX] = trflags;

                  if(!AddPositionKey(&pkey))
                     return false;
               }
               break;

            case SCL_TRACK_TAG:
               {
                  PosKey skey;
                  memset(&skey, 0, sizeof(PosKey));
                  //memcpy(&skey,&key,sizeof(PosKey));
                  memcpy(&skey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&skey.e[0].p);
                  RDFLOAT(&skey.e[1].p);
                  RDFLOAT(&skey.e[2].p);
                  //trackFlags[SCL_TRACK_INDEX] = trflags;
                  workNode->trackFlags[SCL_TRACK_INDEX] = trflags;                

                  if(!AddScaleKey(&skey))
                     return false;
               }
               break;

            case COL_TRACK_TAG:
               {
                  ColorKey ckey;
                  memset(&ckey, 0, sizeof(ColorKey));
                  memcpy(&ckey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&ckey.c[0]);
                  RDFLOAT(&ckey.c[1]);
                  RDFLOAT(&ckey.c[2]);
                  //trackFlags[COL_TRACK_INDEX] = trflags;
                  workNode->trackFlags[COL_TRACK_INDEX] = trflags;                

                  if(!AddColorKey(&ckey))
                     return false;
               }
               break;

            case ROT_TRACK_TAG:
               {
                  RotKey rkey;
                  memset(&rkey, 0, sizeof(RotKey));
                  memcpy(&rkey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&rkey.angle);
                  assert(rkey.angle>=0);

                  RD3FLOAT(rkey.axis);
                  //trackFlags[ROT_TRACK_INDEX] = trflags;
                  workNode->trackFlags[ROT_TRACK_INDEX] = trflags;
                  
                  if(!AddRotationKey(&rkey))
                     return false;
               }
               break;

            case HOT_TRACK_TAG:
               {
                  ScalarKey sckey;
                  memset(&sckey, 0, sizeof(ScalarKey));
                  memcpy(&sckey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&sckey.e[0].p);
                  //trackFlags[HOT_TRACK_INDEX] = trflags;
                  workNode->trackFlags[HOT_TRACK_INDEX] = trflags;
                  
                  if(!AddHotKey(&sckey))
                     return false;
               }
               break;

            case FALL_TRACK_TAG:
               {
                  ScalarKey sckey;
                  memset(&sckey,0,sizeof(ScalarKey));
                  memcpy(&sckey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&sckey.e[0].p);
                  //trackFlags[FALL_TRACK_INDEX] = trflags;
                  workNode->trackFlags[FALL_TRACK_INDEX] = trflags;
                  
                  if(!AddFallKey(&sckey))
                     return false;
               }
               break;

            case FOV_TRACK_TAG:
               {
                  ScalarKey sckey;
                  memset(&sckey, 0, sizeof(ScalarKey));
                  memcpy(&sckey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&sckey.e[0].p);
                  //trackFlags[FOV_TRACK_INDEX] = trflags;
                  workNode->trackFlags[FOV_TRACK_INDEX] = trflags;
                  
                  if(!AddFOVKey(&sckey))
                     return false;
               }
               break;

            case ROLL_TRACK_TAG:
               {
                  ScalarKey sckey;
                  memset(&sckey, 0, sizeof(ScalarKey));
                  memcpy(&sckey, &key_hdr, sizeof(S_key_hdr));
                  RDFLOAT(&sckey.e[0].p);
                  //trackFlags[ROLL_TRACK_INDEX] = trflags;
                  workNode->trackFlags[ROLL_TRACK_INDEX] = trflags;
                  
                  if(!AddRollKey(&sckey))
                     return false;
               }
               break;
            }
         }
      }
      break;

   case M3DMAGIC:
      gotM3DMAGIC = TRUE;
                              //flow...
   case MMAGIC:
      while(C_chunk::Size()){
         if(!C_chunk::RAscend(tag))
            return false;
         switch(tag){
         case MMAGIC:
         case NAMED_OBJECT:
         case KFDATA:
            if(!GetChunk(tag, NULL))
               return false;
            break;

         case MASTER_SCALE:
            if(!GetChunk(tag, NULL))
               return false;
            break;

         case MAT_ENTRY:
            loadmtl = &inmtl;
                              //zro out data structure first
            init_mtl_struct(loadmtl);
            if(get_mtlchunk(this, tag, NULL)==0)
               return false;
            if(merging==0||(merging==1))
               AddMeshMtl(loadmtl);
            FreeMatRefs(loadmtl);
            break;     
         default:
            Descend();
         }
      }
      break;

   case LIN_COLOR_F:
   case COLOR_F:
      {
         Color_f *cf = (Color_f *)data;
         C_chunk::Read(&cf, 12);
      }
      break;

   case COLOR_24:
      {
         Color_24 c24;
         Color_f *cf=(Color_f *)data;
         C_chunk::Read(&c24, 3);

         cf->r=(float)c24.r/255.0f;
         cf->g=(float)c24.g/255.0f;
         cf->b=(float)c24.b/255.0f;
      }
      break;

   case MASTER_SCALE:
      {
         C_chunk::Read(&msc_wk, 4);
         int type;
         float scale;
         GetMasterUnitInfo(&type, &scale);
         float msc_factor = (float)GetMasterScale(UNITS_INCHES);
         msc_wk = (float)((double)msc_wk / msc_factor);     //Turn into a mult factor
      }
      break;

   //case DL_RANGE:
   case DL_OUTER_RANGE:
   case DL_INNER_RANGE:
      {
         float *fptr = (float *)data;
         C_chunk::Read(fptr, sizeof(float));
         *fptr *= msc_wk;
      }
      break;

   case DL_MULTIPLIER:
      {
         Dirlight *d=(Dirlight *)data;
         C_chunk::Read(&d->mult, sizeof(float));
      }
      break;

   case NAMED_OBJECT:
      {
         if(!read_string(this, obname, OBJ_NAME_LEN+1))
            return false;
         while(C_chunk::Size()){
            if(!C_chunk::RAscend(tag))
               return false;
            switch(tag){
            case N_TRI_OBJECT:
               if(!StartMesh(obname))
                  return false;
               if(!GetChunk(tag, this))
                  return false;
               break;
               
            case N_BONE:
               {
                  if(!StartBone(obname))
                     return false;
                  if(GetChunk(tag, NULL)==0)
                     return false;
                  if(!CreateBone())
                     return false;
               }
               break;

            case N_NURBS:
               {
                  if(!StartNurbs(obname))
                     return false;
                  if(GetChunk(tag, NULL)==0)
                     return false;
                  if(!CreateNurbs(0))
                     return false;
               }
               break;

            case N_DIRECT_LIGHT_TARGET:
               {
                  Dirlight &d = studioLt;
                  d.x=d.y=d.z=d.tx=d.ty=d.tz=0.0f;
                  d.flags=NO_LT_ON;
                  d.color.r=d.color.g=d.color.b=1.0f;
                  d.hotsize=d.fallsize=360.0f;
                  d.lo_bias=3.0f;
                  d.shadsize=256;   
                  d.in_range=10.0f;
                  d.out_range=100.0f;
                  d.shadfilter=5.0f;   
                  strcpy(d.imgfile,"");
                  d.ray_bias = 0.2f;
                  d.bank=0.0f;
                  d.aspect=1.0f;
                  d.mult=1.0f;
                  d.appdata = NULL;
                  if(!StartLight(obname))
                     return false;
                  if(GetChunk(tag, &d)==0)
                      return false;
                  int type = TDIR_LIGHT ;
                  if(!CreateLight(type))
                     return false;
               }
               break;

            //case N_D_L_OLD:
            case N_DIRECT_LIGHT:
               {
                  Dirlight &d = studioLt;
                                 //init the direct light to an omni
                  d.x = d.y = d.z = d.tx = d.ty = d.tz = 0.0f;
                  d.flags = NO_LT_ON;
                  d.color.r = d.color.g = d.color.b = 1.0f;
                  d.hotsize = d.fallsize=360.0f;
                  d.lo_bias = 3.0f;   // TO DO: USE GLOBAL?
   //                d.exclude=NULL;
                  d.shadsize = 256;   // TO DO: USE GLOBAL?
                  d.in_range = 10.0f;
                  d.out_range = 100.0f;
                  d.shadfilter = 5.0f;   // TO DO: USE GLOBAL?
                  strcpy(d.imgfile, "");
                  d.ray_bias = 0.2f;   // TO DO: USE GLOBAL?
                  d.bank = 0.0f;
                  d.aspect = 1.0f;
                  d.mult = 1.0f;
                  d.appdata = NULL;
                  if(!StartLight(obname))
                     return false;
                  if(GetChunk(tag, &d)==0)
                     return false;
                  int type = d.hotsize==360.0f ? OMNI_LIGHT : TSPOT_LIGHT;
                  if(!CreateLight(type))
                     return false;
               }
               break;

            case N_CAMERA:
               {
                  Camera3DS &c = studioCam;
                  // Init the camera struct
                  c.x = c.y = c.z = c.tx = c.ty = c.tz = c.bank = 0.0f;
                  c.focal = 50.0f;
                  c.flags = 0;
                  c.nearplane = 10.0f;
                  c.farplane = 1000.0f;
                  c.appdata = NULL;
                  if(!StartCamera(obname))
                     return false;
                  if(GetChunk(tag, &c)==0)
                     return false;
                  if(!CreateCamera(TARGETED_CAMERA))
                     return false;
               }
               break;

            default:
               Descend();
            }
         }
      }
      break;
            
   case APP_DATA:       
      LoadAppData();
      break;

// TO DO: Actually load & store appdata
//       if (!load_app_data(stream, data, chunk.size))
//          return(0);
//       chunk.size = 0;
//       goto skiprest;
      

   case N_TRI_OBJECT:
      while(C_chunk::Size()){
         if(!C_chunk::RAscend(tag))
            return false;
         switch(tag){
         case POINT_ARRAY:
         case FACE_ARRAY:
         case MESH_MATRIX:
         case TEX_VERTS:
         case FACE_MAP_CHANNEL:
            if(!GetChunk(tag, data))
               return false;
            break;

         case APP_DATA:
            if(GetChunk(tag, data)==0)
               return false;
            break;

         case MESH_COLOR:
            if(GetChunk(tag, data)==0)
               return false;
            break;

         default:
            Descend();
         }
      }
      break;

   case POINT_ARRAY:
      {
         word wk_count;
         C_chunk::Read(&wk_count, sizeof(word));
         if(!SetVerts((int)wk_count))
            return false;
         Verts v;
         v.flags = 0;
         for(int ix=0; ix<wk_count; ++ix){
            C_chunk::Read(&v, sizeof(S_vector));
            if(!PutVertex(ix, &v))
               return false;
         }
      }
      break;

   case TEX_VERTS:
      {
         word wk_count;
         C_chunk::Read(&wk_count, sizeof(word));
         if(!SetTVerts((int)wk_count))
            return false;
         UVVert tv;
         tv.z = 0.0f;
         for(int ix=0; ix<wk_count; ++ix){
            C_chunk::Read(&tv, sizeof(float)*2);

            // filter out bogus texture coord values: old versions of 
            // 3ds could have 0xfefefefe as the value.
            // This should catch those and other bad values 
            // derived from them.  --DS 3/16/96
            if (tv.x>1.0e10) tv.x = 0.0f;
            if (tv.x<-1.0e10) tv.x = 0.0f;
            if (tv.y>1.0e10) tv.y = 0.0f;
            if (tv.y<-1.0e10) tv.y = 0.0f;
            if(!PutTVertex(ix,&tv))
               return false;
         }
      }
      break;

   case FACE_MAP_CHANNEL:
      {
         int i;
         int channel_i;
                              //read map channel index
         C_chunk::Read(&channel_i, sizeof(int));

         dword num_uv_verts = 0;
                              //read map vertices
         C_chunk::Read(&num_uv_verts, sizeof(word));
         UVVert *uvverts = new UVVert[num_uv_verts];
         for(i=0; i<num_uv_verts; i++){
            C_chunk::Read(&uvverts[i].x, sizeof(float));
            C_chunk::Read(&uvverts[i].y, sizeof(float));
            uvverts[i].z = 0.0f;
         }

         dword num_faces = 0;
                              //read map faces
         C_chunk::Read(&num_faces, sizeof(word));
         TVFace *uvfaces = new TVFace[num_faces];

         for(i=0; i<num_faces; i++){
            for(int j=0; j<3; j++){
               uvfaces[i].t[j] = 0;
               C_chunk::Read(&uvfaces[i].t[j], sizeof(word));
            }
         }
         if(channel_i>=2 && mesh->getNumMaps() < (channel_i+1))
            mesh->setNumMaps(channel_i+1, true);
         mesh->setMapSupport(channel_i, true);
         mesh->setNumMapVerts(channel_i, num_uv_verts, false);
         for(i=0; i<num_uv_verts; i++)
            mesh->setMapVert(channel_i, i, uvverts[i]);

         mesh->setNumMapFaces(channel_i, num_faces, false);
         for(i=0; i<num_faces; i++)
            mesh->mapFaces(channel_i)[i] = uvfaces[i];

         delete[] uvverts;
         delete[] uvfaces;
      }
      break;

   case POINT_FLAG_ARRAY:
      break;

   case FACE_ARRAY:
      {
         word wk_count;
         C_chunk::Read(&wk_count, sizeof(word));
         if(!SetFaces((int)wk_count))
            return false;

         struct fc_wrt{
            unsigned short a;
            unsigned short b;
            unsigned short c;
            unsigned short flags;
         } Fc_wrt;
         Faces f;

         for(int ix=0; ix<wk_count; ++ix){
            C_chunk::Read(&Fc_wrt, 8);

            f.a=Fc_wrt.a;
            f.b=Fc_wrt.b;
            f.c=Fc_wrt.c;
            f.material = 0;
            f.sm_group = 0;
            f.flags=Fc_wrt.flags;
            if(!PutFace(ix,&f))
               return false;
         }

         while(C_chunk::Size()){
            if(!C_chunk::RAscend(tag))
               return false;
            switch(tag){
            case MSH_MAT_GROUP:
            case SMOOTH_GROUP:
               if(!GetChunk(tag, NULL))
                  return false;
               break;
            default:
               Descend();
            }
         }
      }
      break;

   case MESH_MATRIX:
      {
         Matrix3 tm(1);
         C_chunk::Read(tm.GetAddr(), sizeof(float) * 12);
         C_chunk::Read(tm.GetAddr(), sizeof(float) * 12);
         tm.SetNotIdent();
         if(msc_wk!=1.0f){
            tm.SetTrans(tm.GetTrans()*msc_wk);
         }
         SetTm(&tm);
      }
      break;

   case MESH_COLOR:
      {
         int num_verts = mesh->getNumVerts();
         int num_faces = mesh->getNumFaces();

         UVVert *uvverts = new UVVert[num_verts];
         for(int i=0; i<num_verts; i++){
            S_rgb rgb;
            C_chunk::Read(&rgb, sizeof(S_rgb));
            S_vector &color = *(S_vector*)&uvverts[i];
            color.x = (float)rgb.r / 255.0f;
            color.y = (float)rgb.g / 255.0f;
            color.z = (float)rgb.b / 255.0f;
         }
         TVFace *c_faces = new TVFace[num_faces];
         for(i=num_faces; i--; ){
            for(int j=3; j--; ){
               c_faces[i].t[j] = mesh->faces[i].v[j];
            }
         }

         mesh->setNumMapVerts(0, num_verts, false);
         for(i=0; i<num_verts; i++)
            mesh->setMapVert(0, i, uvverts[i]);

         mesh->setNumMapFaces(0, num_faces, false);
         for(i=0; i<num_faces; i++)
            mesh->mapFaces(0)[i] = c_faces[i];

         delete[] uvverts;
         delete[] c_faces;
      }
      break;

   case MSH_MAT_GROUP:
      {
         char mtlname[256];
         int mtlnum;
         unsigned short wkface;
         if(!read_string(this, mtlname, sizeof(mtlname)))
            return false;
         mtlnum = GetMatNum(mtlname);
         
         word wk_count;
         C_chunk::Read(&wk_count, 2);
         for(int ix=0; ix<wk_count; ++ix){
            C_chunk::Read(&wkface,2);
            if(!PutFaceMtl(wkface, mtlnum))
               return false;
         }
      }
      break;

   case SMOOTH_GROUP:
      {
         int faces = GetFaces();
         unsigned long smgroup;
         for(int ix=0; ix<faces; ++ix){
            C_chunk::Read(&smgroup,sizeof(unsigned long));
            if(!PutSmooth(ix,smgroup))
               return false;
         }
      }
      break;

   case N_BONE:
      {
         Matrix3 m;
         C_chunk::Read(&m, 48);

         SetTm(&m);
      }
      break;

   case N_NURBS:
      {
         NURBSCVSurface *cc = new NURBSCVSurface();
         while(C_chunk::Size()){
            if(!RAscend(tag))
               return false;
            switch(tag){
            case NURBS_UV_ORDER:
            case NURBS_CVS:
            case NURBS_KNOTS:   
            case NURBS_TRANSFORM_MATRIX:
            case NURBS_MAT:
            case NURBS_FLIP_NORMALS:
            case NURBS_UV_MAP:
               if(GetChunk(tag, cc)==0)
                  return false;
               break;
            default :
               Descend();
            }
         }
         if(nset.GetNumObjects() == 1)
            nset.RemoveObject(0);
         nset.AppendObject(cc);
      }
      break;

   case NURBS_UV_MAP:
      {
         NURBSCVSurface *cc = (NURBSCVSurface *)data;
         float uoffset, voffset, utile, vtile;
         C_chunk::Read(&uoffset, sizeof(float));
         C_chunk::Read(&voffset, sizeof(float));
         C_chunk::Read(&utile, sizeof(float));
         C_chunk::Read(&vtile, sizeof(float));

         Point2 pt;
         for(int j = 0;j<4;j++){
            C_chunk::Read(&pt.x, sizeof(float));
            C_chunk::Read(&pt.y, sizeof(float));
            cc->SetTextureUVs(0, j, pt);
         }
      }
      break;

   case NURBS_MAT:
      {
         NURBSCVSurface *cc = (NURBSCVSurface *)data;

         int mtlnum;

         TSTR buffer;

         if(!tchar_read_string(this, buffer)) 
            return false;

         mtlnum = GetMatNum(buffer.data());

         cc->MatID(mtlnum);
         mtlnum = mtlnum;
      }
      break;

   case NURBS_FLIP_NORMALS:
      {
         NURBSCVSurface *cc = (NURBSCVSurface *)data;
         bool FlipNormals;
         C_chunk::Read(&FlipNormals, sizeof(bool));
         cc->FlipNormals(FlipNormals);
      }
      break;

   case NURBS_TRANSFORM_MATRIX:
      {
         NURBSCVSurface *cc = (NURBSCVSurface *)data;
         Matrix3 TransformMatrix;
         C_chunk::Read(&TransformMatrix, sizeof(Matrix3));
         SetXFormPacket mat(TransformMatrix);
         cc->SetTransformMatrix(0, mat);
      }
      break;

   case NURBS_KNOTS:
      {
         NURBSCVSurface *cc = (NURBSCVSurface *)data;
         int numuknots, numvknots;
         C_chunk::Read(&numuknots, sizeof(int));
         C_chunk::Read(&numvknots, sizeof(int));
         cc->SetNumUKnots(numuknots);
         cc->SetNumVKnots(numvknots);
         double knot;
         for(int i=0;i<numuknots;i++){
            C_chunk::Read(&knot,sizeof(double));
            cc->SetUKnot(i, knot);            
         }
         for(i=0;i<numvknots;i++){
            C_chunk::Read(&knot,sizeof(double));
            cc->SetVKnot(i, knot);            
         }
      }
      break;

   case NURBS_UV_ORDER:
      {
         int uorder, vorder;
         NURBSCVSurface *cc = (NURBSCVSurface *)data;
         C_chunk::Read(&uorder, sizeof(int));
         C_chunk::Read(&vorder, sizeof(int));
         cc->SetUOrder(uorder);
         cc->SetVOrder(vorder);
      }
      break;

   case NURBS_CVS:
      {
         float x, y, z, w;
         int ix, iy;
         NURBSCVSurface *cc = (NURBSCVSurface *)data;
         C_chunk::Read(&ix, sizeof(int));
         C_chunk::Read(&iy, sizeof(int));
         cc->SetNumCVs(ix, iy);
         cc->SetName(_T("CV Surface"));
         NURBSControlVertex cv;

         for(int i=0;i<ix;i++){
            for(int j=0;j<iy;j++){
               C_chunk::Read(&x,sizeof(float));
               C_chunk::Read(&y,sizeof(float));
               C_chunk::Read(&z,sizeof(float));
               C_chunk::Read(&w,sizeof(float));
               cv.SetPosition(0, Point3(x, y, z));
               cv.SetWeight(0, w);
               cc->SetCV(i, j, cv);               
            }
         }
      }
      break;

   case N_DIRECT_LIGHT_TARGET:
      {
         Dirlight *d = (Dirlight *)data;
         C_chunk::Read(&d->x,sizeof(float));
         C_chunk::Read(&d->y,sizeof(float));
         C_chunk::Read(&d->z,sizeof(float));
         if(msc_wk != 1.0f){
            d->x *= msc_wk;
            d->y *= msc_wk;
            d->z *= msc_wk;
         }
         while(C_chunk::Size()){
            if(!RAscend(tag))
               return false;
            switch(tag){
            case COLOR_F:
            case COLOR_24:
               if(GetChunk(tag, &d->color)==0)
                  return false;
               break;
            case DL_INNER_RANGE:
               if(GetChunk(tag, &d->in_range)==0)
                  return false;
               break;
            case DL_OUTER_RANGE:
               if(GetChunk(tag, &d->out_range)==0)
                  return false;
               break;
            //case DL_EXCLUDE:
            case DL_SPOTLIGHT:
            //case DL_DIRLIGHT:
            case DL_OFF:
            case DL_ATTENUATE:
            case DL_MULTIPLIER:
               if(GetChunk(tag, d)==0)
                  return false;
               break;
            case APP_DATA:
               if(GetChunk(tag, &d->appdata)==0)
                  return false;
               break;
            default:
               Descend();
            }
         }
      }
      break;

   case N_DIRECT_LIGHT:
      {
         Dirlight *d = (Dirlight *)data;
         C_chunk::Read(&d->x,sizeof(float));
         C_chunk::Read(&d->y,sizeof(float));
         C_chunk::Read(&d->z,sizeof(float));
         if(msc_wk != 1.0f){
            d->x *= msc_wk;
            d->y *= msc_wk;
            d->z *= msc_wk;
         }
         while(C_chunk::Size()){
            if(!RAscend(tag))
               return false;
            switch(tag){
            case COLOR_F:
            case COLOR_24:
               if(GetChunk(tag, &d->color)==0)
                  return false;
               break;
            case DL_INNER_RANGE:
               if(GetChunk(tag, &d->in_range)==0)
                  return false;
               break;
            case DL_OUTER_RANGE:
               if(GetChunk(tag, &d->out_range)==0)
                  return false;
               break;
            //case DL_EXCLUDE:
            case DL_SPOTLIGHT:
            case DL_OFF:
            case DL_ATTENUATE:
            case DL_MULTIPLIER:
               if(GetChunk(tag, d)==0)
                  return false;
               break;
            case APP_DATA:
               if(GetChunk(tag, &d->appdata)==0)
                  return false;
               break;
            default:
               Descend();
            }
         }
      }
      break;

   case DL_DIRLIGHT:
   case DL_SPOTLIGHT:
      {
         Dirlight *d = (Dirlight *)data;
         C_chunk::Read(&d->tx, sizeof(float));
         C_chunk::Read(&d->ty, sizeof(float));
         C_chunk::Read(&d->tz, sizeof(float));
         if(msc_wk != 1.0f){
            d->tx *= msc_wk;
            d->ty *= msc_wk;
            d->tz *= msc_wk;
         }
         C_chunk::Read(&d->hotsize, sizeof(float));
         C_chunk::Read(&d->fallsize, sizeof(float));

                              //enforce keep .5 degree gap for antialiasing
         if(d->hotsize>d->fallsize-.5&&d->fallsize>=.5f) 
            d->hotsize = d->fallsize-.5f;
      }
      break;

   case DL_OFF:
      {
         Dirlight *d = (Dirlight *)data;
         d->flags &= NO_LT_OFF;
      }
      break;

   case DL_ATTENUATE:
      {
         Dirlight *d=(Dirlight *)data;
         d->flags |= NO_LT_ATTEN;
      }
      break;

   case N_CAMERA: 
      {
         Camera3DS *c = (Camera3DS *)data;
         C_chunk::Read(c, 32);
         if(msc_wk != 1.0f){
            c->x *= msc_wk;
            c->y *= msc_wk;
            c->z *= msc_wk;
            c->tx *= msc_wk;
            c->ty *= msc_wk;
            c->tz *= msc_wk;
         }
         while(C_chunk::Size()){
            if(!RAscend(tag))
               return false;
            switch(tag){
            case APP_DATA:
               if(GetChunk(tag, &c->appdata)==0)
                  return false;
               break;
            case CAM_RANGES:
               if(GetChunk(tag, c)==0)
                  return false;
               break;
            default:
               Descend();
            }
         }
      }
      break;

   case CAM_RANGES:
      {
         Camera3DS *c = (Camera3DS *)data;
         C_chunk::Read(&c->nearplane,sizeof(float));
         C_chunk::Read(&c->farplane,sizeof(float));
         c->nearplane *= msc_wk;
         c->farplane *= msc_wk;
      }
      break;
   }
   Descend();
   return true;
}

//----------------------------

int StudioImport::DoImport(const TCHAR *filename, ImpInterface *i, Interface *gi, BOOL suppressPrompts){

   ObjWorker ow(i, gi);
   if(!ow.ROpen(filename))
      return IMPEXP_FAIL;

   if(!i->NewScene())
      return IMPEXP_CANCEL;

   SetCursor(LoadCursor(NULL, IDC_WAIT));

   memset(&BG, 0, sizeof(BG));
   ow.AddMeshMtl(NULL); // Create "Default" mtl

   word tag;
   if(!ow.RAscend(tag))
      return IMPEXP_FAIL;
   if(tag != MMAGIC && tag != M3DMAGIC)
      return IMPEXP_FAIL;

   if(!ow.GetChunk(tag, NULL)){
      //ow.Abandon();      // Abandon any pending object construction
      return IMPEXP_FAIL;
   }

                              //Make sure any objects without nodes are used!
   if(!ow.CompleteScene())
      return IMPEXP_FAIL;

   SetCursor(LoadCursor(NULL, IDC_ARROW));

   if(!ow.SetupEnvironment())
      return IMPEXP_FAIL;

   return IMPEXP_SUCCESS;
}

//----------------------------
//----------------------------

class StudioClassDesc:public ClassDesc{
public:
   virtual int IsPublic(){ return 1; }
   void *Create(BOOL loading = FALSE){ return new StudioImport; }
   const TCHAR *ClassName(){ return "I3D Import"; }
   SClass_ID SuperClassID(){ return SCENE_IMPORT_CLASS_ID; }
   Class_ID ClassID(){ return Class_ID(0x694e4941, 0x70491ec0); }
   const TCHAR *Category(){ return "Import"; }
};

static StudioClassDesc StudioDesc;

//----------------------------

__declspec( dllexport ) int LibNumberClasses(){ return 1; }

//----------------------------

__declspec( dllexport ) ClassDesc *LibClassDesc(int i){

   switch(i) {
   case 0: return &StudioDesc; break;
   }
   return NULL;
}

// Return version so can detect obsolete DLLs
__declspec( dllexport ) ULONG LibVersion() { return VERSION_3DSMAX; }

//----------------------------
