    
/**********************************************************************
 *<
   FILE: 3dsimp.h

   DESCRIPTION:  .3DS file import module header file

   CREATED BY: Tom Hudson

   HISTORY: created 26 December 1994

 *>   Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#pragma pack(1)

#include "surf_api.h"         //NURBs
#include "..\common\common.h"

/* 3DS Shape point structure */

struct shppt{
   float x; /* Control point */
   float y;
   float z;
   float inx;  /* Incoming vector */
   float iny;
   float inz;
   float outx; /* Outgoing vector */
   float outy;
   float outz;
   unsigned short flags;
};

typedef struct shppt Shppt;

#pragma pack()

// Handy file class

/*
class WorkFile{
private:
   FILE *stream;
public:
               WorkFile(const TCHAR *filename,const TCHAR *mode) { stream = _tfopen(filename,mode); };
               ~WorkFile() { if(stream) fclose(stream); stream = NULL; };
   FILE *         Stream() { return stream; };
};
*/

// 3DS face edge vis flags
#define ABLINE (1<<2)
#define BCLINE (1<<1)
#define CALINE 1

#define VWRAP (1<<11)      /* Texture coord V wraps on this face */
#define UWRAP (1<<3)    /* Texture coord U wraps on this face */

// Node list structure

#define OBJ_MESH 0
#define OBJ_OMNILIGHT 1
#define OBJ_SPOTLIGHT 2
#define OBJ_CAMERA 3
#define OBJ_DUMMY 4
#define OBJ_TARGET 5
#define OBJ_OTHER  6 // generated from app data
#define OBJ_NURBS  7
#define OBJ_BONE   8

// key types
#define KEY_FLOAT 0
#define KEY_POS      1
#define KEY_ROT      2
#define KEY_SCL      3
#define KEY_COLOR 4

#define NUMTRACKS 8

#define POS_TRACK_INDEX 0
#define ROT_TRACK_INDEX 1
#define SCL_TRACK_INDEX 2
#define FOV_TRACK_INDEX 3
#define ROLL_TRACK_INDEX 4
#define COL_TRACK_INDEX 5
#define HOT_TRACK_INDEX 6
#define FALL_TRACK_INDEX 7

/*
struct S_key_hdr_tmp: public S_key_hdr{
   PosElem e[8];
};
*/

struct KeyList{
   union{
      S_key_hdr key_hdr;
      PosKey pos;
      RotKey rot;
      ColorKey col;
      ScalarKey sc;
      HideKey hide;
   } key;
   KeyList *next_key;
};

// A list of 3DS objects with their names and types
class C_work_object{
   //C_work_object(const C_work_object&);
   C_work_object &operator =(const C_work_object&);
public:
   Object *object;
   S_vector srcPos;
   S_vector targPos;
   int type;
   int used;
   int cstShad;
   int rcvShad;
   int mtln;
   Matrix3 tm;

   C_work_object():
      used(false)
   {}
};

// A list of the nodes and their various keys
struct WkNodeList{
   ImpNode *node;
   short id;
   int type;
   int mnum;
   TSTR name;
   TSTR owner;
   Mesh *mesh;
   ImpNode *parent;
   Matrix3 tm;
   KeyList *posList;
   KeyList *rotList;
   KeyList *scList;
   KeyList *colList;
   KeyList *hotList;
   KeyList *fallList;
   KeyList *fovList;
   KeyList *rollList;
   SHORT trackFlags[NUMTRACKS];
   WkNodeList *next_node;
};

/* Camera flag bit meanings */

#define NO_CAM_CONE  0x0001
#define NO_CAM_TEMP_APPDATA 0x0002 /* Free appdata after rendering complete  */  

#define NO_CAM_CONE_OFF (~NO_CAM_CONE)

/* Light flag bit meanings */

#define NO_LT_ON  0x0001
#define NO_LT_SHAD   0x0002
#define NO_LT_LOCAL  0x0004
#define NO_LT_CONE   0x0008
#define NO_LT_RECT   0x0010
#define NO_LT_PROJ   0x0020
#define NO_LT_OVER   0x0040
#define NO_LT_ATTEN  0x0080
#define NO_LT_RAYTR  0x0100
#define NO_LT_TEMP_APPDATA 0x0200 /* Free appdata after rendering complete  */   

#define NO_LT_OFF (~NO_LT_ON)
#define NO_LT_SHAD_OFF  (~NO_LT_SHAD)
#define NO_LT_LOCAL_OFF (~NO_LT_LOCAL)
#define NO_LT_CONE_OFF  (~NO_LT_CONE)
#define NO_LT_RECT_OFF  (~NO_LT_RECT)
#define NO_LT_PROJ_OFF  (~NO_LT_PROJ)
#define NO_LT_OVER_OFF  (~NO_LT_OVER)
#define NO_LT_ATTEN_OFF (~NO_LT_ATTEN)

/*--------- Track flags bits------------ */

/*-- This bit causes the spline to be cyclic */
#define ANIM_CYCLIC 1
/*-- This bit causes a track to continue "modulo" its duration */
#define ANIM_LOOP  (1<<1)
/*-- This bit is used by anim.c, but clients need not worry about it*/
#define ANIM_NEGWRAP (1<<2)

#define X_LOCKED (1<<3)
#define Y_LOCKED (1<<4)
#define Z_LOCKED (1<<5)
#define ALL_LOCKED (X_LOCKED|Y_LOCKED|Z_LOCKED)
#define TRACK_ATKEY (1<<6)

/* these flags specify which coords are NOT inherited from parent */ 
#define LNKSHFT 7
#define NO_LNK_X (1<<LNKSHFT)
#define NO_LNK_Y (1<<(LNKSHFT+1))
#define NO_LNK_Z (1<<(LNKSHFT+2))
#define LASTAXIS_SHFT 10
#define LASTAXIS_MASK (3<<LASTAXIS_SHFT)

// A worker object for dealing with creating the objects.
// Useful in the chunk-oriented 3DS file format

#define MAX_NUM_OF_MATERIALS 4096

// Worker types

#define WORKER_IDLE     0
#define WORKER_MESH     1
#define WORKER_KF    2
#define WORKER_SHAPE 3
#define WORKER_LIGHT 4
#define WORKER_CAMERA   5
#define WORKER_NURBS   6
#define WORKER_BONE   7

struct SMtl;

/*
struct MtlName{
   char s[20];
};
*/


int skip_chunk(FILE *stream);
int get_next_chunk(FILE *stream,Chunk_hdr *hdr);
int SkipRead(FILE *stream,long bytes);
bool read_string(class C_chunk*, char *string, int maxsize);
int load_app_data(FILE *stream,void **pdata, int size);

#define RDERR(ptr,count) { if(!fread(ptr,count,1,stream)) return 0; }
#define RD3FLOAT(p) RDERR(p,3*sizeof(FLOAT))
#define RDFLOAT(p)   RDERR(p,sizeof(FLOAT))
#define RDLONG(p) RDERR(p,sizeof(LONG))
#define RDSHORT(p) RDERR(p,sizeof(SHORT))
#define DUMNUM 0x7fff
