#include "all.h"
#include "tables.h"



//----------------------------

C_inventory::C_inventory(C_game_menu &gm):
   game_menu(gm),
   active_item(0)
{}

//----------------------------

C_inventory::~C_inventory(){
}

//----------------------------

bool C_inventory::HasAmmo(int itm_id){

   switch(tab_inventory->ItemI(TAB_I_INV_ID, itm_id)){
   case INV_WEAPON:
   case INV_COLT:
   case INV_BAZOOKA:
   case INV_FLAREGUN:
      return true;
   default: return false;
   }
}

//----------------------------

float C_inventory::GetItemWeight(int itm, int amnt){

   bool has_ammo = HasAmmo(itm);
   bool mult = tab_inventory->ItemB(TAB_B_INV_MULTIPLE, itm);
   int pieces = has_ammo ? 1 : !mult ? 1 : amnt;
   int ammo = has_ammo ? amnt : 0;
   return pieces*tab_inventory->ItemF(TAB_F_INV_WEIGTH, itm) + ammo*tab_inventory->ItemF(TAB_F_INV_AMMO_WEIGTH, itm);
}

//----------------------------

void C_inventory::DeleteAllItems(){

   if(IsVisible())
      ReleaseModels();
   if(items.size() > 1)
      items.erase(items.begin()+1, items.end());
}

//----------------------------

void C_inventory::LoadModels(PI3D_scene is, C_poly_text &texts){

   if(IsVisible())
      return;
   if(!items.size())
      return;

   for(int i=0; i<items.size(); i++){
      PI3D_model model = driver->CreateModel();
      const char *name = tab_inventory->ItemS(TAB_S16_INV_MODEL_I, items[i]->itm, 16);
      model_cache.Open(model, name, is, 0, ErrReport, editor);
      model->LinkTo(is->GetPrimarySector());
      //model->StopAnimation(0);
      model->SetAnimation(0, NULL);
      items[i]->model = model;
      model->Release();

                              //create number (if needed)
      items[i]->num = NULL;
      if(tab_inventory->ItemB(TAB_B_INV_MULTIPLE, items[i]->itm)){
         LPC_text tx = texts.CreateText(C_fstring("%i", items[i]->amount), -1, 0.0f, 0.0f, .0225f);
         items[i]->num = tx;
         tx->Release();
      }
   }
   inv_scene = is;
   MakeOrder();
}

//----------------------------

void C_inventory::ReleaseModels(){

   if(!IsVisible())
      return;
   for(int i=0; i<items.size(); i++){
      items[i]->model = NULL;
      if(items[i]->intp){
         inv_scene->DeleteInterpolator(items[i]->intp);
         items[i]->intp = NULL;
      }
      items[i]->num = NULL;
   }
   inv_scene = NULL;
}

//----------------------------
#define INV_MODEL_DIST  .9f

static void ModelSetPos(PI3D_model model, LPC_text num, int pos1, bool place_models){

   if(place_models){
      S_vector pos(0, -pos1*INV_MODEL_DIST, 0);
      model->SetPos(pos);
      model->SetOn(true);
      float f = 1.0f - abs(pos1)*.4f;
      model->SetScale(f);
      model->Update();
   }
   if(num){
      num->SetPos(.72f, .26f + pos1*.11f);
      num->SetOn(true);
   }
}

//----------------------------

void C_inventory::MakeOrder(bool place_models){

   assert(IsVisible());
   if(!items.size())
      return;
   for(int i=0; i<items.size(); i++){
      if(place_models)
         items[i]->model->SetOn(false);
      if(items[i]->num)
         items[i]->num->SetOn(false);
   }
   ModelSetPos(items[active_item]->model, items[active_item]->num, 0, place_models);

   switch(items.size()){
   case 1:
      break;
   case 2:
      ModelSetPos(items[!active_item]->model, items[!active_item]->num, active_item ? -1 : 1, place_models);
      break;
   default:
      {
         i = active_item;
         if(++i==items.size())
            i = 0;
         ModelSetPos(items[i]->model, items[i]->num, 1, place_models);
         if(!(i=active_item))
            i = items.size(); --i;
         ModelSetPos(items[i]->model, items[i]->num, -1, place_models);
      }
   }
}

//----------------------------

void C_inventory::Scroll(bool left){

   assert(IsVisible());
   if(!IsVisible())
      return;

   switch(items.size()){
   case 0:
   case 1:
      return;
   case 2:
      if(left && active_item==0)
         return;
      if(!left && active_item==1)
         return;
      break;
   }
   
   PI3D_scene inv_scene = game_menu.inv_scene;

                              //delete previous interpolator
   for(int i=0; i<items.size(); i++){
      PI3D_interpolator intp = items[i]->intp;
      if(intp){
         items[i]->model->SetOn(false);
         inv_scene->DeleteInterpolator(intp);
         items[i]->intp = NULL;
      }
   }

   int beg = -2, end = 2;
   switch(items.size()){
   case 2:
      beg = -active_item;
      end = !active_item;
      break;
   }

   for(i=beg; i<=end; i++){
      if(i==-2 && !left)
         continue;
      if(i==2 && left)
         continue;
      bool itm3_repeat = false;
      if(items.size()==3){
         if(abs(i)==2)
            continue;
         itm3_repeat = ((i==-1 && !left) || (i==1 && left));
      }

      int indx = active_item + i;
      if(indx<0)
         indx += items.size();
      if(indx>=items.size())
         indx -= items.size();

      S_item &itm = *items[indx];
                              //place model
      PI3D_model model = itm.model;
      model->SetOn(true);
      model->SetPos(S_vector(0, -i*INV_MODEL_DIST, 0));

      PI3D_keyframe_anim anim = (PI3D_keyframe_anim)driver->CreateAnimation(I3DANIM_KEYFRAME);

#define INV_SCROLL_TIME 120

      int time_beg = (left ? (1-i) : (1+i))*INV_SCROLL_TIME/3;
      int time_end = time_beg + INV_SCROLL_TIME;

      vector<I3D_anim_pos_tcb> pos_keys;
      I3D_anim_pos_tcb pos_key;
      memset(&pos_key, 0, sizeof(pos_key));
      pos_key.easy_from = 1.0f;
      pos_key.easy_to = 1.0f;

      vector<I3D_anim_power> scl_keys;
      I3D_anim_power scl_key;
      scl_key.Clear();

      {
                              //setup position keys
         pos_key.v = model->GetPos();
         pos_key.time = time_beg;
         pos_keys.push_back(pos_key);

         pos_key.v = model->GetPos() + S_vector(0, left ? -INV_MODEL_DIST : INV_MODEL_DIST, 0);
         pos_key.time = time_end;
         pos_keys.push_back(pos_key);
      }

      {                       //setup scale keys
                              //setup keys
         if(abs(i) == 1){
            scl_key.easy_from = 0.0f;
            scl_key.easy_to = 1.0f;
         }else{
            scl_key.easy_from = 1.0f;
            scl_key.easy_to = 0.0f;
         }
         float f;
         f = 1.0f - abs(i)*.4f;

         scl_key.power = f;
         scl_key.time = time_beg;
         scl_keys.push_back(scl_key);
         model->SetScale(f);

         int ni = i + (left ? 1 : -1);
         f = 1.0f - abs(ni)*.4f;
         scl_key.power = f;
         scl_key.time = time_end;
         scl_keys.push_back(scl_key);
      }
      if(itm3_repeat){
         time_beg = time_end;
         time_end = time_beg + INV_SCROLL_TIME;
                              //item re-appear from other side, setup more keys
         int i1 = i * -2;
         {
                              //setup position keys
            S_vector pos(0, -i1*INV_MODEL_DIST, 0);
            pos_key.v = pos;
            pos_key.time = time_beg;
            pos_keys.push_back(pos_key);

            pos_key.v = pos + S_vector(0, left ? -INV_MODEL_DIST : INV_MODEL_DIST, 0);
            pos_key.time = time_end;
            pos_keys.push_back(pos_key);
         }

         {                       //setup scale keys
                                 //setup keys
            scl_key.easy_from = 1.0f;
            scl_key.easy_to = 0.0f;
            float f;
            f = 1.0f - abs(i1)*.4f;

            scl_key.power = f;
            scl_key.time = time_beg;
            scl_keys.push_back(scl_key);

            int ni = i1 + (left ? 1 : -1);
            f = 1.0f - abs(ni) * .4f;
            scl_key.power = f;
            scl_key.time = time_end;
            scl_keys.push_back(scl_key);
         }
      }
      anim->SetPositionKeys(&pos_keys.front(), pos_keys.size());
      anim->SetPowerKeys(&scl_keys.front(), scl_keys.size());

      anim->SetEndTime(time_end);

                              //make interpolator
      PI3D_interpolator intp = driver->CreateInterpolator();
      intp->SetFrame(model);
      intp->SetAnimation(0, anim);
      anim->Release();
      inv_scene->AddInterpolator(intp);
      itm.intp = intp;
      intp->Release();

      model->Update();
   }

   if(!left){
      if(++active_item == items.size())
         active_item = 0;
   }else{
      if(!active_item)
         active_item = items.size();
      --active_item;
   }
   MakeOrder(false);
}

//----------------------------

void C_inventory::SetItemAmount(int i, dword a){

   assert(i<items.size());
   if(i<items.size())
      items[i]->amount = a;
}

//----------------------------

void C_inventory::SetBullets(int i, dword nb){

   assert(i<items.size());
   if(i<items.size())
      items[i]->bullets_in_stack = nb;
}

//----------------------------

void C_inventory::SetActiveItem(int i){

   assert(i<items.size());
   if(i<items.size())
      active_item = i;
}
//----------------------------

int C_inventory::AddItem(int itm, dword num){
                              //check type
   bool b = tab_inventory->ItemB(TAB_B_INV_MULTIPLE, itm);
   if(b){
                              //try to find the same
      int i=NumItems();
      while(i--)
      if(items[i]->itm == itm){
         items[i]->amount += num;
         return i;
      }
   }
   C_inventory::S_item *it = new C_inventory::S_item(itm, num);
   items.push_back(it);
   it->Release();
   return items.size()-1;
}

//----------------------------

bool C_inventory::Reload(int i, bool set){

                              //check if re-loading reasonable - weapons and colt only
   int itm = GetItem(i);
   int id = tab_inventory->ItemI(TAB_I_INV_ID, itm);
   switch(id){
   case INV_WEAPON:
   case INV_COLT:
      switch(tab_inventory->ItemE(TAB_E_INV_FIREMODE, itm)){
      case 1:
      case 2:
         dword batch_size = tab_inventory->ItemI(TAB_I_INV_STACKSIZE, itm);
         dword nb = GetBullets(i);
         dword num_ammo = GetItemAmount(i);
         //if(num_ammo && nb!=batch_size && num_ammo>=batch_size){
         if(num_ammo && nb!=batch_size){
            if(set){
               SetBullets(i, Min(num_ammo, batch_size));
            }
            return true;
         }
         break;
      }
      break;
   }
   return false;
}

//----------------------------

void C_inventory::DeleteItem(int i){

   assert(i<items.size());

   if(i>=items.size())
      return;

   if(IsVisible()){
      //items[i]->model = NULL;
      if(items[i]->intp)
         inv_scene->DeleteInterpolator(items[i]->intp);
   }

   items.erase(items.begin()+i);

   if(active_item >= i)
      --active_item;
}

//----------------------------
//----------------------------
