#include "all.h"
#include "net.h"


//----------------------------
                              //network support:
                              // - original:
                              //    - play animation
                              //    - sync anim position with net players
                              //    - restart anim
                              // - network_actor:
                              //    - play anim non-looped
                              //    - update sync
                              //    - restart anim when told
                              // - network messages:
                              //    - bits:1 = command
                              //       - 0 = set anim position
                              //       - 1 = restart
class C_kill_anim: public C_actor{
   S_vector last_pos;
   bool last_pos_ok;
   bool init_ok;
   int check_count;
   int net_out_count;
   C_smart_ptr<I3D_animation_base> anim;
   C_smart_ptr<I3D_interpolator> intp;

//----------------------------

   static dword I3DAPI cbAnim(PI3D_frame frm, FRM_CALLBACK_MESSAGE msg, dword prm1, dword prm2){

      LPC_actor act = (LPC_actor)frm->GetData();
      assert(act);
      C_kill_anim *ka = (C_kill_anim*)act;
      switch(msg){

      case CBM_ANIM_FINISH:
         ka->last_pos_ok = false;
         ka->check_count = -1000;

         if(net && !ka->network_actor){
                              //must create NM_GAME chunk because this callback is called out of update loop
                              //01.02.02: commneted out: cbAnim IS inside update loop(is called during scene tick,
                              // which is called by C_mission tick and it is inside game_mission tick).
            //net_msg->Ascend(NM_GAME);
                              //restart anim
            const bool g = true;
            net_msg->AddID(ka->net_id, g);
            net_msg->AddBits(1, 1, g);
            //net_msg->Descend();
         }
         break;
      }
      return 0;
   }

//----------------------------

   void InitFirst(){

      assert(!init_ok);

      frame->SetCallback(cbAnim, FRMCALLBACK_ANIMFINISH);
                              //make killing volume
      struct S_hlp{
         static I3DENUMRET I3DAPI cbVol(PI3D_frame frm, dword c){
            I3DCAST_VOLUME(frm)->SetOwner((PI3D_frame)c);
            return I3DENUMRET_OK;
         }
      };
      frame->EnumFrames(S_hlp::cbVol, (dword)frame, ENUMF_VOLUME);
                              //find frame's interpolator
      const PI3D_interpolator *intps = mission.GetScene()->GetInterpolators();
      for(int i = mission.GetScene()->NumInterpolators(); i--; )
      if(intps[i]->GetFrame()==frame){
         intp = intps[i];
         //if(network_actor){
                              //unloop animation
            anim = intp->GetAnimation(0);
            intp->SetAnimation(0, anim);
         //}
         break;
      }
      init_ok = true;
   }
public:
   C_kill_anim(C_game_mission &m):
      C_actor(m, ACTOR_KILL_ANIM),
      last_pos_ok(false),
      init_ok(false),
      net_out_count(0),
      intp(NULL),
      check_count(0)
   {
      if(net)
         network_actor = net->IsHost() ? 0 : net->GetHostPID();
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      if(!init_ok){
         InitFirst();
      }else{
         if(check_count==-1000){
            intp->SetAnimation(0, anim);
            mission.GetScene()->AddInterpolator(intp);
         }

         if((check_count += tc.time) >= 200){
            if(!last_pos_ok){
               last_pos = frame->GetWorldPos();
               last_pos_ok = true;
               return 0;
            }
            check_count=0;
            const S_vector &curr_pos = frame->GetWorldPos();

                              //check collisions on the movement path
            struct S_hlp{
               PI3D_frame owner;
               S_vector delta;
               C_game_mission *mission;
               LPC_actor this_actor;

               static bool I3DAPI ColRespKill(I3D_collision_data &cd){
                              //just kill anyone in the way
                  LPC_actor act = NULL;
                  PI3D_frame frm = cd.GetHitFrm();
                  while(frm=frm->GetParent()){
                     act = (LPC_actor)frm->GetData();
                     if(act)
                        break;
                  }
                  if(act)
                     act->Crash((LPC_actor)cd.cresp_context, .4f);
                  return false;
               }

               static I3DENUMRET I3DAPI cbVol(PI3D_frame frm, dword c){
                  S_hlp *hp = (S_hlp*)c;

                  I3D_collision_data cd;
                  cd.from = frm->GetWorldPos()-hp->delta;
                  cd.dir = hp->delta;
                  cd.radius = frm->GetWorldScale();
                  cd.flags = I3DCOL_SPHERE;
                  cd.frm_ignored = hp->owner;
                  cd.cresp_face = ColRespKill;
                  cd.cresp_context = hp->this_actor;
                  hp->mission->GetScene()->TestCollision(cd);
                  return I3DENUMRET_OK;
               }
            } hlp;
            hlp.owner = frame;
            hlp.delta = curr_pos-last_pos;
            hlp.mission = &mission;
            hlp.this_actor = this;
            frame->EnumFrames(S_hlp::cbVol, (dword)&hlp, ENUMF_VOLUME);
            last_pos = curr_pos;
         }
         if(net && !network_actor && intp && (net_out_count+=tc.time) >= 500){
            net_out_count = 0;
                              //send new interpolator's time
            const bool g = false;
            net_msg->AddID(net_id, g);
            net_msg->AddBits(0, 1, g);
            net_msg->AddWord(intp->GetCurrTime(0), g);
         }
      }
      return 0;
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

                           //write data
      dword curr_t = intp ? intp->GetCurrTime(0) : 0;
      ck->Write(&curr_t, sizeof(curr_t));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

                              //read data
      dword curr_t;
      ck->Read(&curr_t, sizeof(curr_t));
      last_pos_ok = false;

      InitFirst();
      if(intp)
         intp->SetCurrTime(0, curr_t);
      return true;
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      switch(in_msg->GetBits(1)){
      case 0:                 //position
         {
            word w = in_msg->GetWord();
            if(intp)
               intp->SetCurrTime(0, w + in_msg->GetLatency());
         }
         break;
      case 1:                 //reset
         last_pos_ok = false;
         check_count = -1000;
         break;
      }
   }
};

//----------------------------

LPC_actor CreateKillAnim(C_game_mission &gm){

   return new C_kill_anim(gm);
}

//----------------------------

