#include "all.h"
#include "mainmenu.h"
#include <insanity\controller.h>
#include "net.h"
#include "winstuff.h"
#include "tables.h"

//----------------------------

//#define ALLOW_MODEM           //OK to offer modem


#ifdef _DEBUG

//#define DEBUG_TEST_NET        //start in network menu when beginning
//#define DEBUG_CONFIG_NET      //start in config menu
#define DEBUG_NET_ALLOW_HOST_SINGLE //allow hosting also if no joining players are connected

#endif

//----------------------------

C_string MakeGameText(const C_string &str);
C_string AddGameText(const C_string&, const C_string&);
extern LPC_controller controller;

//----------------------------

#define COLOR_ACTIVE 0xffffc080
#define COLOR_POINTED 0xc0f4e0c0
#define COLOR_NORMAL 0xc0c0c0c0

//----------------------------
                              //menu sound indicies
enum E_MENU_SOUND{
   MS_MOVE,                   //navigate through menu
   MS_MOVE_FAIL,              //navigation failed
   MS_FORWARD,                //move forward
   MS_FORWARD_FAIL,           //moving forward failed
   MS_BACK,                   //move backward
   MS_BACK_FAIL,              //move backward failed
   MS_POINT_AT,               //mouse pointed over active item
   MS_CLICK_ITEM,             //click on selectable item
   MS_EDIT_OK,                //editing in edit box successfully finished
   MS_MODIFY_OK,              //adding/modifying succeeded
   MS_MODIFY_FAIL,            //adding/modifying failed
   MS_APPEAR,                 //menu, or layout appears

   MS_LAST
};

//----------------------------

                              //input slots used in menu's controller
enum E_INPUT_SLOT{
   IS_LEFT,                   //move left
   IS_UP,                     //move up
   IS_RIGHT,                  //move right
   IS_DOWN,                   //move down
   IS_ENTER,                  //enter, move forward
   IS_CLICK,                  //activate, press
   IS_BACK,                   //navigate back, erase last item
   IS_ESCAPE,                 //navigate back, escape action
   IS_HOME,                   //move to beginning
   IS_END,                    //move to end
   IS_ADD,                    //add element, increase
   IS_SUBTRACT,               //sub element, decrease

   IS_LAST
};

                              //default delay for controller key hits used in menus
const int KEYBOARD_DELAY_TIME = 160;

const float MCT_DEFAUL_TEXT_SIZE = .04f;
const float TEXT_LINE_DISTANCE_RATIO = 1.0f;

//----------------------------

class C_menu_background{
protected:

   enum E_TEXTURE_INDEX{      //used textures - indicies
      TI_BASE,
                              //tiled textures:
      TI_HORIZONTAL,
      TI_VERTICAL,
      TI_CENTER,

      TI_LAST
   };

   C_smart_ptr<C_sprite_image> spr_images[TI_LAST];

   C_smart_ptr<ISND_source> snd_sources[MS_LAST];
   C_smart_ptr<ISND_sound> play_sound;    //single sound for playing back any source
   E_MENU_SOUND last_played_sound;

   enum E_RECTANGLE{          //rectangles forming background - indicies
      RC_LEFT_TOP,
      RC_TOP,
      RC_RIGHT_TOP,
      RC_LEFT,
      RC_CENTER,
      RC_RIGHT,
      RC_LEFT_BOTTOM,
      RC_BOTTOM,
      RC_RIGHT_BOTTOM,

      RC_LAST
   };

   struct S_vertex: public S_vectorw{
      I3D_text_coor uv;
   };

                              //20 vertices forming menu
                              // they come from left-top corner, and continue
                              // to the right and bottom
   struct S_rectangle{
      S_vertex v[4];
      C_smart_ptr<I3D_visual_dynamic> dyn_vis;

      S_rectangle(){
         for(int i=4; i--; ){
            v[i].Zero();
            v[i].w = 1.0f;
         }
      }
   
      void SetupUV(const C_sprite_image::S_spr_rect &rc){
         v[0].uv = I3D_text_coor(rc.l, rc.t);
         v[3].uv = I3D_text_coor(rc.l + rc.sx, rc.t + rc.sy);
         v[1].uv.u = v[3].uv.u; v[1].uv.v = v[0].uv.v;
         v[2].uv.u = v[0].uv.u; v[2].uv.v = v[3].uv.v;
      }
      void SetupTile(float uc, float vc){
         v[3].uv.u *= uc;
         v[3].uv.v *= vc;
         v[1].uv.u = v[3].uv.u;
         v[2].uv.v = v[3].uv.v;
      }

      inline void Render(){ dyn_vis->Render(); }

      void InitVisual(PI3D_texture txt){
         dyn_vis = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);
         dyn_vis->Release();

         I3D_triface fc[2];
         fc[0][0] = 2;
         fc[0][1] = 0;
         fc[0][2] = 1;
         fc[1][0] = 2;
         fc[1][1] = 1;
         fc[1][2] = 3;

         I3D_face_group fg;
         PI3D_material mat = driver->CreateMaterial();
         fg.base_index = 0;
         fg.num_faces = 2;
         fg.SetMaterial(mat);
         mat->SetTexture(txt);
         mat->Release();

         dyn_vis->Build(v, 4, I3DVC_XYZRHW | /*I3DVC_DIFFUSE | I3DVC_SPECULAR | */(1<<I3DVC_TEXCOUNT_SHIFT), fc, 2, &fg, 1);
      }
   };

   S_rectangle rectangles[9];

                              //texel size at which this menu background was created
                              // see InitRectangle for details
   float texel_size;

protected:
   C_smart_ptr<I3D_driver> drv;

                              //rectange covering this menu
                              // in normalized coords X<0 - 1>, Y<0 - .7f>
   struct S_rect{
      float l, t, r, b;
   } menu_rect, menu_client_rect;
                              // in screen coords, non-clipped, thus may be outside
                              // visible area
   I3D_rectangle menu_screen_rect;
                              // menu client area, in screen coords
   I3D_rectangle menu_client_screen_rect;

   inline float GetTexelSize() const{ return texel_size; }

//----------------------------
// Initialize base rectangles - compute screen and UV coordinates.
   void InitRectangle(float left, float top, float size_x, float size_y, float texel_size1){

      texel_size = texel_size1;

      int scrn_sx = igraph->Scrn_sx();
      //int scrn_sy = igraph->Scrn_sy();

      menu_rect.l = left;
      menu_rect.t = top;
      menu_rect.r = left + size_x;
      menu_rect.b = top + size_y;

      menu_screen_rect.l = menu_rect.l * scrn_sx;
      menu_screen_rect.t = menu_rect.t * scrn_sx;
      menu_screen_rect.r = menu_rect.r * scrn_sx;
      menu_screen_rect.b = menu_rect.b * scrn_sx;
   
      float screen_aspect_ratio = ::screen_aspect_ratio * 1.33333333f;
                                 //compute 2 corners in screen coords
      float sl = left * scrn_sx;
      float st = top * scrn_sx * screen_aspect_ratio;
      float sr = (left + size_x) * scrn_sx;
      float sb = (top + size_y) * scrn_sx * screen_aspect_ratio;

      S_rectangle *r = rectangles;
                              
                                 //init corners
      {
         LPC_sprite_image img = spr_images[TI_BASE];
                                 //init left top corner
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[0];
            float rc_sx = rc.isx * texel_size * scrn_sx;
            float rc_sy = rc.isy * texel_size * scrn_sx * screen_aspect_ratio;

            S_vertex *v = r[0].v;
            v[0].x = sl; v[0].y = st;
            v[3].x = sl + rc_sx; v[3].y = st + rc_sy;

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[0].SetupUV(rc);
            r[0].InitVisual(img->GetTexture());
         }
                                 //init right bottom corner
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[3];
            float rc_sx = rc.isx * texel_size * scrn_sx;
            float rc_sy = rc.isy * texel_size * scrn_sx * screen_aspect_ratio;

            S_vertex *v = r[3].v;
            v[0].x = sr - rc_sx; v[0].y = sb - rc_sy;
            v[3].x = sr; v[3].y = sb;

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[3].SetupUV(rc);
            r[3].InitVisual(img->GetTexture());
         }
                                 //init right top corner
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[1];
            float rc_sx = rc.isx * texel_size * scrn_sx;

            S_vertex *v = r[1].v;
            v[0].x = sr - rc_sx; v[0].y = r[0].v[0].y;
            v[3].x = sr; v[3].y = r[0].v[3].y;

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[1].SetupUV(rc);
            r[1].InitVisual(img->GetTexture());
         }
                                 //init left bottom corner
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[2];
            float rc_sx = rc.isx * texel_size * scrn_sx;

            S_vertex *v = r[2].v;
            v[0].x = sl; v[0].y = r[3].v[0].y;
            v[3].x = sl + rc_sx; v[3].y = r[3].v[3].y;

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[2].SetupUV(rc);
            r[2].InitVisual(img->GetTexture());
         }
      }
                                 //init vertical strips
      {
         LPC_sprite_image img = spr_images[TI_VERTICAL];
                                 //init left side
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[0];
            float rc_sx = rc.isx * texel_size * scrn_sx;
            float rc_sy = rc.isy * texel_size * scrn_sx * screen_aspect_ratio;

            S_vertex *v = r[4].v;
            v[0] = r[0].v[2];
            v[3].x = v[0].x + rc_sx; v[3].y = r[2].v[0].y;

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[4].SetupUV(rc);
            r[4].SetupTile(1.0f, (v[2].y - v[0].y) / rc_sy);
            r[4].InitVisual(img->GetTexture());
         }
                                 //init right side
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[1];
            float rc_sx = rc.isx * texel_size * scrn_sx;
            float rc_sy = rc.isy * texel_size * scrn_sx * screen_aspect_ratio;

            S_vertex *v = r[5].v;
            v[3] = r[3].v[1];
            v[0].x = v[3].x - rc_sx; v[0].y = r[1].v[3].y;

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[5].SetupUV(rc);
            r[5].SetupTile(1.0f, (v[2].y - v[0].y) / rc_sy);
            r[5].InitVisual(img->GetTexture());
         }
      }
                                 //init horizontal strips
      {
         LPC_sprite_image img = spr_images[TI_HORIZONTAL];
                                 //init top side
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[0];
            float rc_sx = rc.isx * texel_size * scrn_sx;

            S_vertex *v = r[6].v;
            v[0] = r[0].v[1];
            v[3] = r[1].v[2];

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[6].SetupUV(rc);
            r[6].SetupTile((v[1].x - v[0].x) / rc_sx, 1.0f);
            r[6].InitVisual(img->GetTexture());
         }
                                 //init bottom side
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[1];
            float rc_sx = rc.isx * texel_size * scrn_sx;

            S_vertex *v = r[7].v;
            v[0] = r[2].v[1];
            v[3] = r[3].v[2];

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[7].SetupUV(rc);
            r[7].SetupTile((v[1].x - v[0].x) / rc_sx, 1.0f);
            r[7].InitVisual(img->GetTexture());
         }
      }
                                 //init center strips
      {
         LPC_sprite_image img = spr_images[TI_CENTER];
         {
            const C_sprite_image::S_spr_rect &rc = img->GetRects()[0];
            float rc_sx = rc.isx * texel_size * scrn_sx;
            float rc_sy = rc.isy * texel_size * scrn_sx * screen_aspect_ratio;

            S_vertex *v = r[8].v;
            v[0] = r[4].v[1];
            v[3] = r[5].v[2];

            v[1].x = v[3].x; v[1].y = v[0].y;
            v[2].x = v[0].x; v[2].y = v[3].y;

            r[8].SetupUV(rc);
            r[8].SetupTile((v[1].x - v[0].x) / rc_sx, (v[2].y - v[0].y) / rc_sy);
            r[8].InitVisual(img->GetTexture());
         }
      }

                                 //init common parts of vertices
      for(int i=0; i<9; i++){
         for(int j=0; j<4; j++){
            r[i].v[j].z = 1.0f;
            r[i].v[j].w = 1.0f;
         }
      }

      menu_client_screen_rect.l = r[4].v[1].x;
      menu_client_screen_rect.t = r[4].v[1].y;
      menu_client_screen_rect.r = r[5].v[2].x;
      menu_client_screen_rect.b = r[5].v[2].y;

      menu_client_rect.l = menu_client_screen_rect.l / (float)scrn_sx;
      menu_client_rect.t = menu_client_screen_rect.t / ((float)scrn_sx * screen_aspect_ratio);
      menu_client_rect.r = menu_client_screen_rect.r / (float)scrn_sx;
      menu_client_rect.b = menu_client_screen_rect.b / ((float)scrn_sx * screen_aspect_ratio);
   }

//----------------------------
// Play sound, specified by index.
   bool PlaySound(E_MENU_SOUND indx){

      if(indx >= MS_LAST)
         return false;
      if(isound){
         if(!snd_sources[indx])
            return false;
         if(last_played_sound!=indx){
            ISND_RESULT ir;
            ir = play_sound->PlaySource(snd_sources[indx]);
            assert(ISND_SUCCESS(ir));
         }else{
            if(play_sound->IsPlaying()){
               play_sound->SetCurrTime(0);
               if(!play_sound->IsPlaying())
                  play_sound->Play();
            }else{
               play_sound->Play();
            }
         }
         last_played_sound = indx;
      }
      return true;
   }

public:
   C_menu_background():
      drv(driver)
   {
      int i;

      {
                                 //load all sprite images
         static const char *set_names[TI_LAST] = {
            "base", "horizontal", "vertical", "center"
         };
         for(i=0; i<TI_LAST; i++){
            C_sprite_image *si = new C_sprite_image;
            bool b = si->Init(drv, C_fstring("tables\\menu\\%s.txt", set_names[i]),
               C_fstring("menu\\%s.png", set_names[i]),
               C_fstring("menu\\%s_o.png", set_names[i]),
               TEXTMAP_TRUECOLOR | TEXTMAP_MIPMAP,
               true,
               -.5f);
            assert(b);
            if(b)
               spr_images[i] = si;
            si->Release();
         }
                                 //load all sounds
         if(isound){
            //play_sound = NULL;
            PISND_sound snd;
            ISND_RESULT ir = isound->CreateSound(&snd);
            play_sound = snd;
            snd->Release();
            if(ISND_SUCCESS(ir)){
               static const char *snd_names[] = {
                  "move", "move_fail",
                  "forward", "forward_fail",
                  "backward", "backward_fail",
                  "point_at", "click",
                  "edit_end",
                  "modify", "modify_fail",
                  "appear",
               };
               dword open_flags = ISND_NOVOLUME | ISND_NOPANNING;
               for(i=0; i<MS_LAST; i++){
                  PISND_source src;
                  ISND_RESULT ir = isound->CreateSource(&src);
                  if(ISND_SUCCESS(ir)){
                     ir = src->Open(C_fstring("sounds\\Common\\menu\\%s.ogg", snd_names[i]), open_flags);
                     if(ISND_FAIL(ir)){
                        ir = src->Open(C_fstring("sounds\\Common\\menu\\%s.wav", snd_names[i]), open_flags);
                        if(ISND_FAIL(ir)){
                           ErrReport(C_fstring("Cannot open sound: 'Common\\menu\\%s.wav'", snd_names[i]), editor);
                        }
                     }
                     if(ISND_SUCCESS(ir))
                        snd_sources[i] = src;
                     src->Release();
                  }
               }
            }
         }
         last_played_sound = MS_LAST;
      }
   }

//----------------------------

   void Render(){
      for(int i=9; i--; )
         rectangles[i].Render();
   }
};

//----------------------------
//----------------------------
                              //base class of all menu controls
class C_menu_control_base: public C_unknown{
   E_MENU_CONTROL_TYPE type;
   int hint_text_id;          //ID of hint text, 0 = no hint
protected:
   C_menu_base *menu;
   S_vector2 pos;
   float extent[4];           //extent of the item, in normalized screen coords [l|t|r|b]
   bool active;               //set to true when control is active, one item at a time
   bool pointed;              //set to true when control is pointed-at, one item at a time
   bool clicked;              //set to true when control is pointed-at, one item at a time
   dword user_data;

   C_menu_control_base(const C_menu_control_base&);
   C_menu_control_base &operator =(const C_menu_control_base&);

//----------------------------
// Render thin rectangle around control's extent.
   void RenderExtentRectangle(const float *extent = NULL) const{

      if(!extent)
         extent = C_menu_control_base::extent;
                           //convert extent to screen coords
      int sx = igraph->Scrn_sx();

      const S_vector2 *c = (const S_vector2*)extent;
      struct S_vertex: public S_vectorw{
         dword diffuse;
         //dword specular;      //padding to 32 bytes
         dword uv[2];
      } vv[8];
      float line_width = (int)Max(1.0f, sx * .003f);
      float aspect = screen_aspect_ratio * 1.33333333f;
      for(int i=0; i<4; i++){
         S_vertex &v = vv[i];
         v.z = 1.0f;
         v.w = 1.0f;
         v.x = c[i&1].x * sx;
         v.y = c[i/2].y * sx * aspect;
         v.diffuse = 0xffff0000;
         v.uv[0] = 0; v.uv[1] = 0;

         S_vertex &v1 = vv[4+i];
         v1 = v;
         v1.x += line_width * (!(i&1) ? 1.0f : -1.0f);
         v1.y += line_width * (!(i/2) ? 1.0f : -1.0f) * aspect;
         v1.diffuse = 0xffff0000;
         v1.uv[0] = 0; v1.uv[1] = 0;
      }
      static const word indx[] = {
         0, 1, 4,  4, 1, 5,
         2, 0, 6,  6, 0, 4,
         2, 6, 7,  2, 7, 3,
         7, 5, 1,  7, 1, 3,
      };

      PI3D_visual_dynamic dyn_vis = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);

      I3D_face_group fg;
      PI3D_material mat = driver->CreateMaterial();
      fg.base_index = 0;
      fg.num_faces = 8;
      fg.SetMaterial(mat);
      mat->Release();

      dyn_vis->Build(vv, 8, I3DVC_XYZRHW | I3DVC_DIFFUSE | /*I3DVC_SPECULAR | */(1<<I3DVC_TEXCOUNT_SHIFT), (I3D_triface*)indx, 8, &fg, 1);

      dyn_vis->Render();
      dyn_vis->Release();
   }
public:
   enum E_UPDATE_ACTION{
      UA_NULL,
      UA_MOVE_LEFT,           //move one step to direction
      UA_MOVE_UP,
      UA_MOVE_RIGHT,
      UA_MOVE_DOWN,
      UA_MOVE_HOME,
      UA_MOVE_END,
      UA_MOVE_RELATIVE,       //move depending on input relative direction
   };
                              //where do we go, when clicked
                              // special value int(1) means returning one level back
   int control_id;
   union{
      dword context_data;
      const void *link_to;
      bool (*callback)(E_MENU_CALLBACK_MESSAGE, class C_menu_base*, class C_menu_control_base*,
         void *context);
   };
   dword control_flags;       //flags spcifying this control, from template

   C_menu_control_base(E_MENU_CONTROL_TYPE t1, C_menu_base *m, const S_vector2 &pos1, int hint_id):
      menu(m),
      user_data(0),
      type(t1),
      hint_text_id(hint_id),
      context_data(0),
      control_id(0),
      control_flags(0),
      active(false),
      pointed(false),
      clicked(false),
      pos(pos1)
   {
      //memset(extent, 0, sizeof(extent));
                              //invalidate extent
      extent[0] = 1.0f;
      extent[1] = .75f;
      extent[2] = 0.0f;
      extent[3] = 0.0f;
   }

   virtual E_MENU_CONTROL_TYPE GetType() const{ return type; }
   virtual int GetID() const{ return control_id; }
   virtual dword GetFlags() const{ return control_flags; }

   inline const S_vector2 &GetPos() const{ return pos; }
   inline const float *GetExtent() const{ return extent; }
   inline int GetHintTextID() const{ return hint_text_id; }

   virtual bool MayBeActivated() const{ return true; }

//----------------------------
// Render this control.
   virtual void Render() const{}

//----------------------------
// Activate/deactivate control.
   virtual void SetActive(bool b){
      active = b;
   }

//----------------------------
// Make control pointed-at (by mouse).
   virtual void SetPointed(bool b){
      pointed = b;
   }

//----------------------------
// Make control pointed-at (by mouse).
   virtual void SetClicked(bool b){
      clicked = b;
   }

//----------------------------
// Update control. Only active control is updated by this method.
// If this method returns true, the input data are 'ate', which means
// that they're not used for navigating through menu.
// Parameter 'play_sound' is used to play particular sound when return back to menu.
   virtual bool Tick(const S_tick_context&, LPC_controller, E_UPDATE_ACTION ua,
      E_MENU_SOUND &play_sound, void *context){
      return false;
   }

   virtual void BuildText(const char *cp, bool wide){}

   inline void SetUserData(dword dw){ user_data = dw; }
   inline dword GetUserData() const{ return user_data; }
};

typedef C_menu_control_base *LPC_menu_control_base;

//----------------------------
//----------------------------

class C_menu_control_text: public C_menu_control_base{
   C_smart_ptr<C_text> text[2];  //text and its shadow
   char activation_hot_key;

   dword color;
   dword justify;

protected:
   void UpdateExtent(){
      assert(text[0]);
      text[0]->GetExtent(extent[0], extent[1], extent[2], extent[3]);
   }

   C_smart_ptr<C_poly_text> texts;

   S_vector2 text_pos;
   float size;

   inline float GetShadowOffset() const{ return .003f; }

//----------------------------
// Get color of text, depending on current control's state.
   dword GetCurrentColor() const{
      //return active ? 0xffffffff : pointed ? 0xc0f4f4f4 : color;
      //return active ? 0xffffc080 : pointed ? 0xc0f4e0c0 : color;
      return active ? COLOR_ACTIVE : pointed ? COLOR_POINTED : COLOR_NORMAL;
   }
public:
   C_menu_control_text(C_menu_base *m, const S_vector2 &pos, float sz, dword just1, C_poly_text &texts1,
      int text_id, int hint_id, E_MENU_CONTROL_TYPE type = MC_TEXT):
      C_menu_control_base(type, m, pos, hint_id),
      activation_hot_key(K_NOKEY),
      justify(just1 | TEXT_JUSTIFY_CONVERT_EOL),
      size(sz),
      texts(&texts1)
   {
      color = 0xa0f0f0f0;
      text_pos = pos;
      text_pos.y += size;

      if(text_id && all_txt.IsDefined(text_id)){
         const char *cp = all_txt[text_id];
         bool wide = all_txt.IsWide(text_id);
         BuildText(cp, wide);
                              //detect activation hot-key
         activation_hot_key = !wide ? tolower(cp[0]) : 0;
      }
   }

   virtual bool MayBeActivated() const{ return false; }

//----------------------------

   virtual void SetActive(bool b){
      if(active==b) return;
      C_menu_control_base::SetActive(b);

      if(text[0]){
         text[0]->SetColor(GetCurrentColor());
         float x = text_pos.x, y = text_pos.y;
         float sx = x + GetShadowOffset(), sy = y + GetShadowOffset();
         if(active){
            x -= .001f; y -= .001f;
            sx += .002f; sy += .002f;
         }
         text[0]->SetPos(x, y);
         if(text[1])
            text[1]->SetPos(sx, sy);
      }
   }

//----------------------------

   virtual void SetPointed(bool b){

      if(pointed==b) return;
      C_menu_control_base::SetPointed(b);
      if(text[0])
         text[0]->SetColor(GetCurrentColor());
   }

//----------------------------

   virtual void SetClicked(bool b){
      if(clicked==b) return;
      C_menu_control_base::SetClicked(b);

      float soffs = GetShadowOffset();
      float x = text_pos.x, y = text_pos.y;
      float sx = x + soffs, sy = sy = y + soffs;
      if(clicked){
         x += .001f; y += .001f;
      }else
      if(active){
         x -= .001f; y -= .001f;
         sx += .002f; sy += .002f;
      }
      text[0]->SetPos(x, y);
      if(text[1])
         text[1]->SetPos(sx, sy);
   }

//----------------------------

   inline char GetActivationHotKey() const{ return activation_hot_key; }

   void BuildText(const char *cp, bool wide = false){

      text[0] = texts->CreateText(cp, color, text_pos.x, text_pos.y, size, justify, wide);
      text[0]->Release();
      if(!wide){
         float so = GetShadowOffset();
         text[1] = texts->CreateText(cp, 0x60000000, text_pos.x + so, text_pos.y + so,
            size, justify, wide, NULL, NULL, NULL, NULL, -1);
         text[1]->Release();
      }else{
         text[1] = NULL;
      }
      UpdateExtent();
   }
};

//----------------------------
//----------------------------

class C_menu_control_text_button: public C_menu_control_text{
public:
   C_menu_control_text_button(C_menu_base *m, const S_vector2 &pos, float sz, dword justify, C_poly_text &texts,
      int text_id, int hint_id):
      C_menu_control_text(m, pos, sz, justify, texts, text_id, hint_id, MC_TEXT_BUTTON)
   {}

   virtual bool MayBeActivated() const{ return true; }
};

//----------------------------
//----------------------------

class C_menu_control_editbox: public C_menu_control_text{
protected:
   C_string edit_text;        //text which is displayed
   C_string edited_text;      //copy of text which is being edited
   bool in_edit;              //when true, we're operating on 'edited_text'
   bool cursor_on;
   C_smart_ptr<C_text> edit[2];  //edited text and its shadow
   float text_size;
private:

                              //max number of characters in edited text
                              // (not including terminating '\0')
   int max_edit_len;

   int cursor_flash_countdown;
   enum{
      CURSOR_FLASH_COUNT = 200
   };

protected:
   inline float GetEditOffset() const{ return .005f; }

   float edit_extent[4];      //extent of edit text, in normalized screen coords [l|t|r|b]

   void UpdateEditExtent(){
      assert(edit[0]);
      edit[0]->GetExtent(edit_extent[0], edit_extent[1], edit_extent[2], edit_extent[3]);
                              //expand control's extent by our text (to the right only)
      if(edit_extent[2])
         extent[2] = edit_extent[2];

      const float c = .002f;
      edit_extent[0] -= c;
      edit_extent[1] -= c;
      edit_extent[2] += c * 5.0f;   //account for possible cursor
      edit_extent[3] += c;
   }

//----------------------------
// Build graphical text from edited text. The text may contain cursor at the end,
// if 'cursor_on' is set to true.
   virtual void BuildText(){

      const C_string &display_text = !in_edit ? edit_text : edited_text;
      C_string et = !cursor_on ? display_text : C_fstring("%s|", (const char*)display_text);
      float scl_x = .8f;
      edit[0] = texts->CreateText(et, GetCurrentColor(), text_pos.x + GetEditOffset(), text_pos.y, text_size, TEXT_JUSTIFY_LEFT,
         false, NULL, NULL, NULL, NULL, 0, 0.0f, scl_x, 1.0f);
      edit[1] = texts->CreateText(et, 0x60000000, text_pos.x + GetEditOffset() + GetShadowOffset(), text_pos.y + GetShadowOffset(), text_size, TEXT_JUSTIFY_LEFT,
         false, NULL, NULL, NULL, NULL, -1, 0.0f, scl_x, 1.0f);
      edit[0]->Release();
      edit[1]->Release();
   }

//----------------------------
// Reset flashing of cursor, so that next time it is displayed for the full time.
   void ResetCursor(){
      cursor_flash_countdown = 0;
      cursor_on = false;
   }
public:
   C_menu_control_editbox(C_menu_base *m, const S_vector2 &pos, float size, C_poly_text &texts1,
      int text_id, int hint_id, E_MENU_CONTROL_TYPE ct = MC_EDITBOX):

      C_menu_control_text(m, pos, size, TEXT_JUSTIFY_RIGHT, texts1, text_id, hint_id, ct),
      text_size(size),
      max_edit_len(20),
      in_edit(false),
      cursor_flash_countdown(0),
      cursor_on(false)
   {
      BuildText();

      UpdateEditExtent();
   }

   virtual bool MayBeActivated() const{ return true; }

//----------------------------

   virtual void SetActive(bool b){
      if(active==b) return;
      C_menu_control_text::SetActive(b);
      if(!active && in_edit){
         ToggleEditMode();
      }
      edit[0]->SetColor(GetCurrentColor());
   }

//----------------------------

   virtual void SetPointed(bool b){
      if(pointed==b) return;
      C_menu_control_text::SetPointed(b);
      edit[0]->SetColor(GetCurrentColor());
   }

//----------------------------

   virtual void Render() const{
      if(in_edit)
         RenderExtentRectangle(edit_extent);
      C_menu_control_text::Render();
   }

//----------------------------
// Enable/disable edit mode.
   void ToggleEditMode(){

      if(control_flags&MCT_CALLBACK){
         assert(callback);
         if(!callback(in_edit ? MCM_CTRL_EDIT_DONE : MCM_CTRL_EDIT_BEGIN, menu, this, 0))
            return;
      }
      in_edit = !in_edit;
      if(in_edit){
         edited_text = edit_text;
                              //cause to update extents
         cursor_on = false;
         BuildText();
         UpdateEditExtent();

         cursor_on = true;
         BuildText();
      }else{
         edited_text = NULL;
                              //rebuild original text, also cause to update extents
         cursor_on = false;
         BuildText();
         UpdateEditExtent();
      }
   }

//----------------------------

   virtual bool Tick(const S_tick_context &tc, LPC_controller ctrl, E_UPDATE_ACTION ua,
      E_MENU_SOUND &play_sound, void *context){

      assert(active);
      if(in_edit){
         if(ctrl->Get(IS_ENTER)){
            ctrl->IgnoreValue(IS_ENTER);
            play_sound = MS_EDIT_OK;
                              //end of editing, accept changes, and back to passive mode
            edit_text = edited_text;
            ToggleEditMode();
            return true;
         }else
         if(ctrl->Get(IS_ESCAPE)){
            ctrl->IgnoreValue(IS_ESCAPE);
            play_sound = MS_BACK;
                              //end of editing, cancel changes, and back to passive mode
            ToggleEditMode();
            return true;
         }else
         if(ctrl->Get(IS_BACK)){
            ctrl->IgnoreValue(IS_BACK, KEYBOARD_DELAY_TIME);

            if(!edited_text.Size()){
               play_sound = MS_MODIFY_FAIL;
            }else{
               play_sound = MS_MODIFY_OK;
                              //strip last character
               edited_text[(int)edited_text.Size() - 1] = 0;
               edited_text = (const char*)edited_text;
                              //cause to update extents
               cursor_on = false;
               BuildText();
               UpdateEditExtent();

               ResetCursor();
            }
         }else{
            char c = igraph->GetAscii(tc.key, tc.modify_keys&SKEY_SHIFT);
            if(c && (isalnum(c) || isspace(c) || ispunct(c))){
                              //check if we're still allowed to add characters
               if(edited_text.Size() >= max_edit_len){
                  play_sound = MS_MODIFY_FAIL;
               }else{
                  edited_text = C_fstring("%s%c", (const char*)edited_text, c);
                              //cause to update extents
                  cursor_on = false;
                  BuildText();
                  UpdateEditExtent();

                  ResetCursor();
                  play_sound = MS_MODIFY_OK;
               }
            }
         }
                              //flash cursor
         if((cursor_flash_countdown -= tc.time) <= 0){
            cursor_flash_countdown = CURSOR_FLASH_COUNT;
            cursor_on = !cursor_on;
            BuildText();
         }
         return true;
      }
      return false;
   }

//----------------------------
// Get edited text.
   const C_string &GetEditText() const{
      return edit_text;
   }

//----------------------------
// Set maximal editable size.
   inline void SetMaxEditLen(int l){ max_edit_len = l; }

//----------------------------
// Setup initial edited text.
   void SetEditText(const char *cp){

      assert(!in_edit);
      edit_text = cp;
      BuildText();
   }
};

//----------------------------

class C_menu_control_enumbox: public C_menu_control_text{
   struct S_enum_entry{
      C_string text;
      bool wide;
      dword user_data;
      S_enum_entry(){}
      S_enum_entry(const C_string &s1, dword dw, bool w): text(s1), user_data(dw), wide(w){}
      S_enum_entry(const S_enum_entry &ee){ operator =(ee); }
      S_enum_entry &operator =(const S_enum_entry &ee){
         text = ee.text;
         user_data = ee.user_data;
         wide = ee.wide;
         return *this;
      }
   };
   typedef map<int, S_enum_entry> t_enum_map;
   t_enum_map enums;

   float text_size;

   C_smart_ptr<C_text> text[2];  //text and its shadow
   C_string curr_text;
   bool curr_wide;
//----------------------------

   void BuildText(){

      const float offset = .005f;
      const float scl_x = .8f;
      text[0] = texts->CreateText(curr_text, GetCurrentColor(), text_pos.x + offset, text_pos.y, text_size, TEXT_JUSTIFY_LEFT,
         curr_wide, NULL, NULL, NULL, NULL, 0, 0.0f, scl_x, 1.0f);
      text[0]->Release();

      text[1] = NULL;
      if(!curr_wide){
         text[1] = texts->CreateText(curr_text, 0x60000000, text_pos.x + offset + GetShadowOffset(), text_pos.y + GetShadowOffset(), text_size, TEXT_JUSTIFY_LEFT,
            curr_wide, NULL, NULL, NULL, NULL, -1, 0.0f, scl_x, 1.0f);
         text[1]->Release();
      }
      C_menu_control_text::UpdateExtent();
      float ext[4];
      text[0]->GetExtent(ext[0], ext[1], ext[2], ext[3]);
      extent[2] = Max(extent[2], ext[2]);
   }

//----------------------------

   int curr_enum;             //-1 = no selection
public:
   C_menu_control_enumbox(C_menu_base *m, const S_vector2 &pos, float size, C_poly_text &texts1,
      int text_id, int hint_id):
      C_menu_control_text(m, pos, size, TEXT_JUSTIFY_RIGHT, texts1, text_id, hint_id, MC_ENUMBOX),
      text_size(size),
      curr_enum(-1)
   {
   }

   virtual bool MayBeActivated() const{ return true; }

//----------------------------

   virtual void SetActive(bool b){

      if(active==b) return;
      C_menu_control_text::SetActive(b);
      if(text[0])
         text[0]->SetColor(GetCurrentColor());
   }

//----------------------------

   virtual void SetPointed(bool b){

      if(pointed==b) return;
      C_menu_control_text::SetPointed(b);
      if(text[0])
         text[0]->SetColor(GetCurrentColor());
   }

//----------------------------
// Return number of possible enumerations.
   inline int GetNumEnums() const{ return enums.size(); }

//----------------------------
// Get currently selected enumeration.
   inline int GetCurrEnum() const{ return curr_enum; }

//----------------------------
// Set current enumeration.
   void SetCurrEnum(int id){

      if(id==curr_enum)
         return;
      curr_enum = id;
      t_enum_map::const_iterator it = enums.find(id);
      if(it==enums.end()){
         curr_text = NULL;
      }else{
         curr_text = (*it).second.text;
         curr_wide = (*it).second.wide;
      }
      BuildText();
   }

//----------------------------
// Add enumeration on specified position.
   void SetupEnum(int id, const C_string &str, dword user_data = 0, bool wide = false){

      enums[id] = S_enum_entry(str, user_data, wide);
      if(curr_enum==id){
         curr_text = str;
         curr_wide = wide;
         BuildText();
      }
   }

//----------------------------
// Clear enumeration on specified position.
   void ClearEnum(int id){

      t_enum_map::iterator it = enums.find(id);
      assert(it != enums.end());
      enums.erase(it);
      if(curr_enum == id)
         SetCurrEnum(-1);
   }

//----------------------------
// Get data associated with specified enum.
   dword GetEnumData(int id) const{

      t_enum_map::const_iterator it = enums.find(id);
      if(it==enums.end())
         return -1;
      return (*it).second.user_data;
   }
};

//----------------------------
//----------------------------

class C_menu_control_controller_box: public C_menu_control_editbox{
   int slot;
   C_smart_ptr<C_controller> cfg_controller;

   bool in_scan_mode;         //set to true when we're scanning
   int flash_countdown;
   bool flash_phase;

//----------------------------
// Build graphical text from edited text. The text may contain cursor at the end,
// if 'cursor_on' is set to true.
   void BuildText(bool wide){

      const C_string &display_text = !in_edit ? edit_text : edited_text;
      C_string et = !cursor_on ? display_text : C_fstring("%s|", (const char*)display_text);
      float scl_x = .8f;
      edit[0] = texts->CreateText(et, GetCurrentColor(), text_pos.x + GetEditOffset(), text_pos.y, text_size, TEXT_JUSTIFY_LEFT,
         wide, NULL, NULL, NULL, NULL, 0, 0.0f, scl_x, 1.0f);
      edit[0]->Release();

      edit[1] = NULL;
      if(!wide){
         edit[1] = texts->CreateText(et, 0x60000000, text_pos.x + GetEditOffset() + GetShadowOffset(), text_pos.y + GetShadowOffset(), text_size, TEXT_JUSTIFY_LEFT,
            wide, NULL, NULL, NULL, NULL, -1, 0.0f, scl_x, 1.0f);
         edit[1]->Release();
      }
   }

public:
   C_menu_control_controller_box(C_menu_base *m, const S_vector2 &pos, float size, C_poly_text &texts1,
      int text_id, int hint_id, LPC_controller c1, int s1):
      C_menu_control_editbox(m, pos, size, texts1, text_id, hint_id, MC_CONTROLLER_BOX),
      cfg_controller(c1),
      slot(s1),
      in_scan_mode(false)
   {
      BuildControlText();
   }

//----------------------------
// Build text using the enumerated element.
   void BuildControlText(){
                              //get names of assigned device(s)
      C_controller::E_VALUE_MODE vm;
      C_controller::E_INPUT_DEVICE dev[2];
      dword data[2];
      bool ok = false;
      for(int i=0; i<2; i++){
         if(cfg_controller->GetSlotSetting(slot, i, vm, dev[i], data[i]))
            ok = true;
      }

      bool wide = all_txt.IsWide(10256);
      if(ok){
         edit_text = NULL;

         for(i=0; i<2; i++){
                              //add separator
            if(edit_text.Size() && dev[i]!=C_controller::ID_NULL){
               if(!wide)
                  edit_text += ", ";
               else
                  edit_text = AddGameText(edit_text, MakeGameText(", "));
            }
            switch(dev[i]){
            case C_controller::ID_NULL:
               break;
            case C_controller::ID_KEYBOARD:
               {
                  int tid = 10000 + data[i];
                  if(all_txt.IsDefined(tid)){
                     if(!wide)
                        edit_text += all_txt[tid];
                     else
                        edit_text = AddGameText(edit_text, all_txt[tid]);
                  }else{
                     if(!wide)
                        edit_text += all_txt[10256];
                     else
                        edit_text = AddGameText(edit_text, all_txt[10256]);
                  }
               }
               break;

            case C_controller::ID_MOUSE_BUTTON:
               {
                  if(!wide)
                     edit_text += C_fstring("%s %i", (const char*)all_txt[10257], data[i]);
                  else{
                     edit_text = AddGameText(edit_text, all_txt[10257]);
                     edit_text = AddGameText(edit_text, MakeGameText(C_fstring(" %i", data[i])));
                  }
               }
               break;

            case C_controller::ID_MOUSE_AXIS:
               {
                  int axe = data[i] & 0x7fffffff;
                  /*
                  if(axe==2){
                              //we know this is wheel, write it directly
                     edit_text += all_txt[10258 + ((data[i]&0x80000000) != 0)];
                  }else{
                     edit_text += C_fstring(all_txt[6282], axe);
                  }
                  */
                  if(!wide){
                     edit_text += C_fstring("%s %c%c", (const char*)all_txt[10258], 'X'+axe, !(data[i]&0x80000000) ? '+' : '-');
                  }else{
                     edit_text = AddGameText(edit_text, all_txt[10258]);
                     edit_text = AddGameText(edit_text, MakeGameText(C_fstring(" %c%c", 'X'+axe, !(data[i]&0x80000000) ? '+' : '-')));
                  }
               }
               break;

            case C_controller::ID_JOY_AXIS:
               {
                  int axe = data[i] & 0x7fffffff;
                  static const char axis_name[] = {'X', 'Y', 'Z', 'U', 'V', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', ' '};
                  axe = Min(axe, (int)sizeof(axis_name)-1);
                  if(!wide){
                     edit_text += C_fstring("%s %c%c", (const char*)all_txt[10260], axis_name[axe], !(data[i]&0x80000000) ? '+' : '-');
                  }else{
                     edit_text = AddGameText(edit_text, all_txt[10260]);
                     edit_text = AddGameText(edit_text, MakeGameText(C_fstring(" %c%c", axis_name[axe], !(data[i]&0x80000000) ? '+' : '-')));
                  }
               }
               break;

            case C_controller::ID_JOY_BUTTON:
               {
                  if(!wide)
                     edit_text += C_fstring("%s %i", (const char*)all_txt[10259], data[i]);
                  else{
                     edit_text = AddGameText(edit_text, all_txt[10259]);
                     edit_text = AddGameText(edit_text, MakeGameText(C_fstring(" %i", data[i])));
                  }
               }
               break;

            case C_controller::ID_JOY_POV:
               {
                  int pov_i = data[i] & 0xffff;
                  int a = data[i]>>16;
                  if(!wide)
                     edit_text += C_fstring("%s%i (%i)", (const char*)all_txt[10261], pov_i, a);
                  else{
                     edit_text = AddGameText(edit_text, all_txt[10261]);
                     edit_text = AddGameText(edit_text, MakeGameText(C_fstring("%i (%i)", pov_i, a)));
                  }
               }
               break;

            default:
               assert(0);
            }
         }
      }else{
         edit_text = "error - unknown slot";
      }
      BuildText(wide);
      UpdateEditExtent();
   }

//----------------------------
// Enable/disable scan mode.
   void ToggleScanMode(){

      in_scan_mode = !in_scan_mode;
      if(in_scan_mode){
                              //display message, that we're waiting for a key
         edit_text = all_txt[10262];
         BuildText(all_txt.IsWide(10262));
         UpdateEditExtent();
         flash_countdown = 0;
         flash_phase = false;
      }else{
                              //display configured data
         BuildControlText();
      }
   }

   virtual void SetActive(bool b){
      if(active==b) return;
      C_menu_control_editbox::SetActive(b);
      if(!active && in_scan_mode){
         ToggleScanMode();
      }
   }

   virtual void Render() const{
      if(in_scan_mode){
         if(flash_phase)
            RenderExtentRectangle(edit_extent);
      }
      C_menu_control_editbox::Render();
   }

   virtual bool Tick(const S_tick_context &tc, LPC_controller ctrl, E_UPDATE_ACTION ua,
      E_MENU_SOUND &play_sound, void *context);
};

//----------------------------
//----------------------------

class C_menu_control_checkbox: public C_menu_control_text{
   C_smart_ptr<C_poly_sprite> spr[2];  //[off | on]
   S_vector2 spr_pos;
   bool checked;

//----------------------------
// Set colors of button, depending on current control state.
   void SetupColors(){
      //dword color = active ? 0xffffffff : pointed ? 0xe0e0e0e0 : 0xc0c0c0c0;
      dword color = active ? COLOR_ACTIVE : pointed ? COLOR_POINTED : COLOR_NORMAL;

      spr[0]->SetColor(color);
      spr[1]->SetColor(color);
   }
public:
   C_menu_control_checkbox(C_menu_base *m, const S_vector2 &pos, float size, dword justify, C_poly_text &texts,
      int text_id, LPC_sprite_image spr_img, int spri_off, int spri_on, int hint_id):
      C_menu_control_text(m, pos, size, justify, texts, text_id, hint_id, MC_CHECKBOX),
      checked(false),
      spr_pos(pos)
   {
                              //init sprites
      spr[0] = new C_poly_sprite(spr_pos.x, spr_pos.y, size * 1.2f, spr_img, spri_off);
      spr[0]->Release();
      spr[1] = new C_poly_sprite(spr_pos.x, spr_pos.y, size * 1.2f, spr_img, spri_on);
      spr[1]->Release();
                              //position check mark accordingly to the text
      spr_pos.y += size * .6f;
      if((justify&TEXT_JUSTIFY_MASK) == TEXT_JUSTIFY_LEFT){
         const S_vector2 &size1 = spr[0]->GetSize();
         spr_pos.x -= size1.x + size * .3f;
      }else{
         spr_pos.x += size * .3f;
      }
      spr[0]->SetPos(spr_pos.x, spr_pos.y);
      spr[1]->SetPos(spr_pos.x, spr_pos.y);

                              //expand our extent
      {
         float e[4];
         spr[0]->GetExtent(e[0], e[1], e[2], e[3]);
         extent[0] = Min(extent[0], e[0]);
         extent[1] = Min(extent[1], e[1]);
         extent[2] = Max(extent[2], e[2]);
         extent[3] = Max(extent[3], e[3]);
      }
      SetupColors();
   }

   virtual bool MayBeActivated() const{ return true; }

   virtual void SetActive(bool b){
      if(active==b) return;
      C_menu_control_text::SetActive(b);
      SetupColors();
   }

   virtual void SetPointed(bool b){
      if(pointed==b) return;
      C_menu_control_text::SetPointed(b);
      SetupColors();
   }

   virtual void Render() const{
      C_menu_control_text::Render();
      spr[checked ^ clicked]->Render();
   }

   virtual void SetClicked(bool b){
      if(clicked==b) return;
      C_menu_control_text::SetClicked(b);
      S_vector2 pos = spr_pos;
      if(clicked){
         pos.x += .001f; pos.y += .001f;
      }
      spr[0]->SetPos(pos.x, pos.y);
      spr[1]->SetPos(pos.x, pos.y);
   }

//----------------------------
// Check if control is currently checked.
   virtual bool IsChecked() const{ return checked; }

//----------------------------
// Check/uncheck the control.
   virtual void SetCheck(bool b){ checked = b; }
};

//----------------------------
//----------------------------

class C_menu_control_slider: public C_menu_control_text{
   C_smart_ptr<C_poly_sprite> spr_base, spr_arrow;
   S_vector2 spr_pos;
   int num_positions;
   float slide_pos;
   bool first_click;

   void UpdateArrowPos(){
      S_vector2 v = spr_base->GetSize();
      v.x = spr_pos.x + v.x * (.05f + .9f * (slide_pos / (float)(num_positions - 1)));
      v.y = spr_pos.y;
      if(clicked){
         v.x += .001f;
         v.y += .001f;
      }
      spr_arrow->SetPos(v.x, v.y);
   }

//----------------------------
// Set colors of button, depending on current control state.
   void SetupColors(){
      //dword color = active ? 0xffffffff : pointed ? 0xe0e0e0e0 : 0xc0c0c0c0;
      dword color = active ? COLOR_ACTIVE : pointed ? COLOR_POINTED : COLOR_NORMAL;

      spr_base->SetColor(color);
      spr_arrow->SetColor(color);
   }
public:
   C_menu_control_slider(C_menu_base *m, const S_vector2 &pos, float size, dword justify, C_poly_text &texts,
      int text_id, LPC_sprite_image spr_img, int spri_base, int spri_arrow,
      int num_pos, int hint_id):
      C_menu_control_text(m, pos, size, justify, texts, text_id, hint_id, MC_SLIDER),
      num_positions(num_pos),
      slide_pos(0),
      first_click(false),
      spr_pos(pos)
   {
                              //init sprites
      spr_base = new C_poly_sprite(spr_pos.x, spr_pos.y, size * .7f, spr_img, spri_base);
      spr_base->Release();
      spr_arrow = new C_poly_sprite(spr_pos.x, spr_pos.y, size * 1.2f, spr_img, spri_arrow);
      spr_arrow->Release();
                              //position check mark accordingly to the text
      spr_pos.y += size * .6f;
      if((justify&TEXT_JUSTIFY_MASK) == TEXT_JUSTIFY_LEFT){
         const S_vector2 &spr_size = spr_base->GetSize();
         spr_pos.x -= spr_size.x + size * .6f;
      }else{
         spr_pos.x += size * .6f;
      }
      spr_base->SetPos(spr_pos.x, spr_pos.y);
      SetupColors();
      UpdateArrowPos();

                              //expand our extent
      {
         float e[4];
         spr_base->GetExtent(e[0], e[1], e[2], e[3]);
         extent[0] = Min(extent[0], e[0]);
         extent[1] = Min(extent[1], e[1]);
         extent[2] = Max(extent[2], e[2]);
         extent[3] = Max(extent[3], e[3]);
      }
   }

   virtual bool MayBeActivated() const{ return true; }

   virtual void Render() const{
      C_menu_control_text::Render();
      spr_base->Render();
      spr_arrow->Render();
   }

   virtual void SetActive(bool b){
      if(active==b) return;
      C_menu_control_text::SetActive(b);
      SetupColors();
      if(!active)
         first_click = false;
   }

   virtual void SetPointed(bool b){
      if(pointed==b) return;
      C_menu_control_text::SetPointed(b);
      SetupColors();
   }

   virtual void SetClicked(bool b){
      if(clicked==b) return;
                              //override text's method
      C_menu_control_base::SetClicked(b);
      UpdateArrowPos();
   }

   virtual bool Tick(const S_tick_context &tc, LPC_controller, E_UPDATE_ACTION ua,
      E_MENU_SOUND &play_sound, void *context){

      if(tc.mouse_buttons&1){
         if(clicked && !first_click){
            first_click = true;
                              //clicked on control,
                              // compute where should the slider move to
            const S_vector2 &v = spr_base->GetSize();
            float x = spr_pos.x + .05f * v.x;
            float sz = v.x * .9f;
            float scrn_sx = (float)igraph->Scrn_sx();

            float mx = igraph->Mouse_x() / scrn_sx;

            float rel_x;
            if(mx < x)
               rel_x = 0.0f;
            else
            if(mx > (x+sz))
               rel_x = 1.0f;
            else{
               rel_x = (mx-x) / sz;
            }
            slide_pos = rel_x * (float)(num_positions-1);
         }else
         if(!clicked)
            first_click = false;
      }else{
         first_click = false;
      }

      switch(ua){
      case UA_MOVE_LEFT: slide_pos = Max(0.0f, slide_pos - 1.0f); break;
      case UA_MOVE_RIGHT: slide_pos = Min(num_positions - 1.0f, slide_pos + 1.0f); break;
      case UA_MOVE_HOME: slide_pos = 0; break;
      case UA_MOVE_END: slide_pos = num_positions - 1; break;
      case UA_MOVE_RELATIVE:
                              //move by mouse only if clicked
         if(clicked){
            float rx = tc.mouse_rel[0];
                              //compute how much is single unit on the screen
            const S_vector2 &v = spr_base->GetSize();
            float sz = v.x * .9f;
            sz *= (float)igraph->Scrn_sx();
            float rel_change = rx / sz;
            float real_change = rel_change * (float)num_positions;

            slide_pos = Max(0.0f, Min((float)num_positions - 1.0f, slide_pos + real_change));
         }
         break;
      }
      UpdateArrowPos();

      if(control_flags&MCT_CALLBACK){
         (*callback)(MCM_CTRL_SLIDED, menu, this, context);
      }

      return false;
   }

//----------------------------
// Get current slider position.
   inline int GetSlidePos() const{ return slide_pos; }

//----------------------------
// Set current slider position.
   inline void SetSlidePos(int i){
      slide_pos = Max(0, Min(num_positions - 1, i));
      UpdateArrowPos();
   }

//----------------------------
// Get number of slider positions.
   inline int GetNumPositions() const{ return num_positions; }
};

//----------------------------
//----------------------------

class C_menu_control_button: public C_menu_control_base{
   C_smart_ptr<C_poly_sprite> spr[2];  //[released | pressed]

//----------------------------
// Set colors of button, depending on current control state.
   void SetupColors(){
      dword color = active ? COLOR_ACTIVE : pointed ? COLOR_POINTED : COLOR_NORMAL;

      spr[0]->SetColor(color);
      spr[1]->SetColor(color);
   }
public:
   C_menu_control_button(C_menu_base *m, const S_vector2 &pos, float size,
      LPC_sprite_image spr_img, int spri_rel, int spri_pres, int hint_id):
      C_menu_control_base(MC_BUTTON, m, pos, hint_id)
   {
                              //init sprites
      spr[0] = new C_poly_sprite(pos.x, pos.y, size, spr_img, spri_rel);
      spr[0]->Release();

      spr[1] = new C_poly_sprite(pos.x, pos.y, size, spr_img, spri_pres);
      spr[1]->Release();

      spr[0]->GetExtent(extent[0], extent[1], extent[2], extent[3]);

      SetupColors();
   }

   virtual void SetActive(bool b){
      if(active==b) return;
      C_menu_control_base::SetActive(b);
      SetupColors();
   }

   virtual void SetPointed(bool b){
      if(pointed==b) return;
      C_menu_control_base::SetPointed(b);
      SetupColors();
   }

   virtual void Render() const{
      spr[clicked]->Render();
   }
};

//----------------------------
//----------------------------

class C_menu_base: public C_class_to_be_ticked_hd, public C_menu_background{

   C_smart_ptr<C_text> menu_name[2];   //text and its shadow
   C_smart_ptr<C_text> hint_text;      //hint text, shown in hint bar

   int active_control_index;  //currently active control
   int pointed_control_index; //currently pointed-at control
   int clicked_control_index; //currently clicked control

   C_smart_ptr<C_poly_sprite> cursor;

   enum E_CURSOR_MOVEMENT{
      MOVE_UNKNOWN,
      MOVE_LEFT,
      MOVE_UP,
      MOVE_RIGHT,
      MOVE_DOWN,
      MOVE_HOME,
      MOVE_END,
   };

protected:
   C_smart_ptr<I3D_scene> scene; //dummy scene for rendering texts and sprites
   C_poly_text texts;
   C_smart_ptr<C_controller> menu_controller;  //for scanning keys
                              //tick class on which which we operate on
   C_smart_ptr<C_class_to_be_ticked_hd> saved_tick_class;

                              //all controls of this menu
   vector<C_smart_ptr<C_menu_control_base> > controls;
                              //history of menu layouts we've navigated through
                              // current layout is always at the top
   struct S_history_element{
      S_menu_template menu_templ;
      const S_menu_layout_template *layout_templ;
      int active_control_index;
   };
   vector<S_history_element> history;

private:

//----------------------------
// Create cursor sprite.
   void CreateCursor(){

      assert(!cursor);
      float sz = .05f;
      cursor = new C_poly_sprite(0, 0, sz, spr_images[TI_BASE], 4);

      cursor->SetScreenPos(igraph->Mouse_x(), (float)igraph->Mouse_y() / (screen_aspect_ratio * 1.33333333f));

      cursor->Release();
   }

//----------------------------
// Clear entire contents of menu - controls, menu name, etc.
   void ClearContents(){

      controls.clear();
      for(int i=0; i<2; i++)
         menu_name[i] = NULL;
      hint_text = NULL;
      active_control_index = -1;

                              //clear our history, if we don't want to keep one
      if(history.size()){
         if(history.back().layout_templ->flags&MLT_NO_HISTORY){
            history.pop_back();
         }
      }
   }

//----------------------------
// Delete old hint message, and create a new one. If 'text_id' is zero,
// no hint message is only destroyed.
   void SetupHintMessage(int text_id){

                              //clear hint text
      hint_text = NULL;
      if(text_id && all_txt.IsDefined(text_id)){
         const char *cp = all_txt[text_id];
         //dword color = 0xffc0c040;
         dword color = 0xffc0c0c0;
         float sz = .015f + 12.0f * GetTexelSize();
         float x = menu_rect.l + 40.0f * GetTexelSize();
         float y = menu_rect.b - 14.0f * GetTexelSize();
         const float it_angle = 0.0f;
         hint_text = texts.CreateText(cp, color, x, y, sz, TEXT_JUSTIFY_LEFT, false, NULL, NULL, NULL, NULL, 0,
            it_angle, .7f, 1.0f);
         hint_text->Release();
      }
   }

//----------------------------
// Set pointed control. This also updates hint message.
// The input parameter may be -1 to set no pointed control.
   void SetPointedControl(int indx, bool play_sound){

      assert(indx==-1 || indx < controls.size());
      if(pointed_control_index == indx)
         return;

      if(pointed_control_index!=-1 && pointed_control_index != active_control_index)
         controls[pointed_control_index]->SetPointed(false);

      pointed_control_index = indx;

      int hint_id = 0;
      if(pointed_control_index!=-1){
         LPC_menu_control_base ctrl = controls[pointed_control_index];
         if(pointed_control_index != active_control_index)
            ctrl->SetPointed(true);
         hint_id = ctrl->GetHintTextID();
      }else
      if(active_control_index!=-1){
         LPC_menu_control_base ctrl = controls[active_control_index];
         hint_id = ctrl->GetHintTextID();
      }
      SetupHintMessage(hint_id);

      if(pointed_control_index!=-1 && play_sound)
         PlaySound(MS_POINT_AT);
   }

//----------------------------
// Move cursor to desired direction. This may cause de-activating old control
// and activating new control.
   bool ChangeActiveItem(E_INPUT_SLOT is){

                              //find control which is closest to desired direction,
                              // and may be activated
      S_vector2 curr_pos;
      int ignore_item = active_control_index;

      switch(is){
      case IS_HOME:
         curr_pos.x = menu_client_rect.l;
         curr_pos.y = menu_client_rect.t;
         is = IS_RIGHT;
         ignore_item = -1;
         break;
      case IS_END:
         curr_pos.x = menu_client_rect.r;
         curr_pos.y = menu_client_rect.b;
         is = IS_LEFT;
         ignore_item = -1;
         break;
      default:
         if(active_control_index!=-1){
            curr_pos = controls[active_control_index]->GetPos();
         }else{
            curr_pos.x = menu_client_rect.l;
            curr_pos.y = menu_client_rect.t;
         }
      }
      float best_score;
      int best_index = -1;

      for(int i=controls.size(); i--; ){
         if(i==ignore_item)
            continue;
         LPC_menu_control_base ctrl = controls[i];
         if(ctrl->MayBeActivated()){
                              //check if it's in desired direction
            const S_vector2 &pos = ctrl->GetPos();
            const S_vector2 &dir = pos - curr_pos;
            bool dir_ok = false;
            float score;
            const float C = 3.0f;

            switch(is){
            case IS_LEFT: 
               dir_ok = (dir.x < 0.0f);
               if(dir_ok) score = -dir.x + fabs(dir.y) * C;
               break;
            case IS_RIGHT:
               dir_ok = (dir.x > 0.0f);
               if(dir_ok) score = dir.x + fabs(dir.y) * C;
               break;
            case IS_UP:
               dir_ok = (dir.y < 0.0f);
               if(dir_ok) score = -dir.y + fabs(dir.x) * C;
               break;
            case IS_DOWN:
               dir_ok = (dir.y > 0.0f);
               if(dir_ok) score = dir.y + fabs(dir.x) * C;
               break;
            default: assert(0);
            }
            if(dir_ok){
                           //control in good direction...
               if(best_index==-1 || best_score > score){
                  best_score = score;
                  best_index = i;
               }
            }
         }
      }
      if(best_index==-1 || best_index==active_control_index)
         return false;

      SetActiveControl(best_index);
      return true;
   }

//----------------------------
// Get menu control, which is currently clicked.
   C_menu_control_base *GetClickedControl(const S_tick_context &tc, bool only_mouse){

      bool ok = false;
      if(!only_mouse && menu_controller->Get(IS_ENTER)){
         if(clicked_control_index == -1){
            if(clicked_control_index != active_control_index){
               pointed_control_index =  active_control_index;
               clicked_control_index = active_control_index;
               if(clicked_control_index!=-1){
                  controls[clicked_control_index]->SetClicked(true);
                  PlaySound(MS_CLICK_ITEM);
               }
            }
         }
      }else
      if(menu_controller->Get(IS_CLICK)){
         if(clicked_control_index == -1 && pointed_control_index!=-1){
            if(active_control_index != pointed_control_index){
                              //make this one active
               SetActiveControl(pointed_control_index);
            }
            if(clicked_control_index != active_control_index){
               clicked_control_index = active_control_index;
               if(clicked_control_index!=-1){
                  controls[clicked_control_index]->SetClicked(true);
                  PlaySound(MS_CLICK_ITEM);
               }
            }
         }
         if(clicked_control_index!=-1)
            controls[clicked_control_index]->SetClicked(clicked_control_index == pointed_control_index);
      }else{
         if(clicked_control_index != -1){
            if(clicked_control_index == pointed_control_index){
               ok = true;
            }
            controls[clicked_control_index]->SetClicked(false);
            clicked_control_index = -1;
         }
      }

      if(!ok)
         return NULL;
      if(active_control_index==-1)
         return NULL;

      return controls[active_control_index];
   }

//----------------------------
// Update cursor movement, also process status bar and checking of control
// under the cursor.
   void UpdateMouseMovement(int mrx, int mry, bool force = false){

      if(cursor && (mrx || mry || force)){
         float scrn_sx = igraph->Scrn_sx();

         float mx = igraph->Mouse_x();
         float my = igraph->Mouse_y();
         float ar = screen_aspect_ratio * 1.33333333f;

         float pos_x = (float)mx / scrn_sx;
         float pos_y = (float)my / scrn_sx / ar;

         cursor->SetScreenPos(mx, my / ar);

                              //determine control, which we point at
         for(int i=controls.size(); i--; ){
            LPC_menu_control_base ctrl = controls[i];
            if(!ctrl->MayBeActivated())
               continue;
            const float *ctrl_extent = ctrl->GetExtent();
            if(pos_x >= ctrl_extent[0] &&
               pos_y >= ctrl_extent[1] &&
               pos_x < ctrl_extent[2] &&
               pos_y < ctrl_extent[3]){
                              //this is the control we point at
               break;
            }
         }
         SetPointedControl(i, !force);
      }
   }

//----------------------------
protected:
   void *context;

//----------------------------
// Create new control item from template. This func may return new (suggested)
// position and scale for next elements.
   C_menu_control_base *CreateAndAddControl(const S_menu_control_template &templ,
      S_vector2 &pos, float &scale){

      LPC_menu_control_base ct = NULL;

      const S_rect &mrc = (templ.flags&MCT_POS_NO_CLIENT) ? menu_rect : menu_client_rect;

                              //init position
      if(templ.flags&MCT_POS_X_CENTER){
         pos.x = mrc.l + (mrc.r - mrc.l) * .5f + templ.pos_x;
      }else
      if(templ.flags&MCT_POS_X_LEFT){
         pos.x = mrc.l + templ.pos_x;
      }else
      if(templ.flags&MCT_POS_X_RIGHT){
         pos.x = mrc.r + templ.pos_x;
      }else
         pos.x += templ.pos_x;

      if(templ.flags&MCT_POS_Y_CENTER){
         pos.y = mrc.t + (mrc.b - mrc.t) * .5f + templ.pos_y;
      }else
      if(templ.flags&MCT_POS_Y_TOP){
         pos.y = mrc.t + templ.pos_y;
      }else
      if(templ.flags&MCT_POS_Y_BOTTOM){
         pos.y = mrc.b + templ.pos_y;
      }else
         pos.y += templ.pos_y;

      switch(templ.flags&(MCT_SIZE_1|MCT_SIZE_2|MCT_SIZE_3|MCT_SIZE_4|MCT_SIZE_5)){
      case MCT_SIZE_1: scale *= .64f; break;
      case MCT_SIZE_2: scale *= .8f; break;
      case MCT_SIZE_3: scale *= 1.0f; break;
      case MCT_SIZE_4: scale *= 1.25f; break;
      case MCT_SIZE_5: scale *= 1.5625f; break;
      default: assert(0);
      }

      switch(templ.ctrl){
      case MC_TEXT_BUTTON:
         {
            dword justify =
               (templ.flags&MCT_JUSTIFY_LEFT) ? TEXT_JUSTIFY_LEFT :
               (templ.flags&MCT_JUSTIFY_RIGHT) ? TEXT_JUSTIFY_RIGHT :
               TEXT_JUSTIFY_CENTER;
            float sz = MCT_DEFAUL_TEXT_SIZE * scale;
            ct = new C_menu_control_text_button(this, pos, sz, justify, texts,
               templ.text_id, templ.text_id_hint);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      case MC_EDITBOX:
         {
            float sz = MCT_DEFAUL_TEXT_SIZE * scale;
            ct = new C_menu_control_editbox(this, pos, sz, texts,
               templ.text_id, templ.text_id_hint);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      case MC_ENUMBOX:
         {
            float sz = MCT_DEFAUL_TEXT_SIZE * scale;
            ct = new C_menu_control_enumbox(this, pos, sz, texts,
               templ.text_id, templ.text_id_hint);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      case MC_TEXT:
         {
            dword justify =
               (templ.flags&MCT_JUSTIFY_LEFT) ? TEXT_JUSTIFY_LEFT :
               (templ.flags&MCT_JUSTIFY_RIGHT) ? TEXT_JUSTIFY_RIGHT :
               TEXT_JUSTIFY_CENTER;
            float sz = MCT_DEFAUL_TEXT_SIZE * scale;
            ct = new C_menu_control_text(this, pos, sz, justify, texts, templ.text_id,
               templ.text_id_hint);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      case MC_CONTROLLER_BOX:
         {
            float sz = MCT_DEFAUL_TEXT_SIZE * scale;
            ct = new C_menu_control_controller_box(this, pos, sz, texts, templ.text_id,
               templ.text_id_hint, controller, templ.controller_slot);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      case MC_BUTTON:
         {
            float sz = 32.0f * GetTexelSize() * scale;;
            ct = new C_menu_control_button(this, pos, sz, spr_images[TI_BASE],
               templ.sprite_index_released, templ.sprite_index_pressed, templ.text_id_hint);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      case MC_CHECKBOX:
         {
                              //text may be either on left or on right side
            dword justify =
               (templ.flags&MCT_JUSTIFY_LEFT) ? TEXT_JUSTIFY_LEFT :
               TEXT_JUSTIFY_RIGHT;
            float sz = MCT_DEFAUL_TEXT_SIZE * scale;
            ct = new C_menu_control_checkbox(this, pos, sz, justify, texts,
               templ.text_id, spr_images[TI_BASE], 6, 7,
               templ.text_id_hint);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      case MC_SLIDER:
         {
                              //text may be either on left or on right side
            dword justify =
               (templ.flags&MCT_JUSTIFY_LEFT) ? TEXT_JUSTIFY_LEFT :
               TEXT_JUSTIFY_RIGHT;
            float sz = MCT_DEFAUL_TEXT_SIZE * scale;
            ct = new C_menu_control_slider(this, pos, sz, justify, texts,
               templ.text_id, spr_images[TI_BASE], 8, 9, 101,
               templ.text_id_hint);
            pos.y += sz * TEXT_LINE_DISTANCE_RATIO;
         }
         break;

      default:
         assert(0);
      }

      if(ct){
         ct->context_data = templ.context_data;
         ct->control_id = templ.control_id;
         ct->control_flags = templ.flags;
         controls.push_back(ct);
         ct->Release();
      }
      return ct;
   }

//----------------------------
// Process shortcuts - allow direct response to input.
// By default, this causes moving to control which name begins by pressed key.
   virtual void ProcessShortcuts(const S_tick_context &tc){

      if(isalnum(tc.key)){
         char hot_key = tolower(tc.key);
         for(int i=(active_control_index + 1) % controls.size(); i!=active_control_index; ++i %= controls.size()){

            LPC_menu_control_base ctrl = controls[i];
            if(!ctrl->MayBeActivated())
               continue;
            switch(ctrl->GetType()){
            case MC_TEXT_BUTTON:
            case MC_CHECKBOX:
            case MC_SLIDER:
            //case MC_LISTBOX:
            case MC_CONTROLLER_BOX:
            case MC_EDITBOX:
            case MC_ENUMBOX:
               {
                  C_menu_control_text *ct = (C_menu_control_text*)ctrl;
                  if(ct->GetActivationHotKey() == hot_key){
                     ctrl = NULL;
                     break;
                  }
               }
               break;
            }
            if(!ctrl)
               break;
         }
         if(i!=active_control_index){
            SetActiveControl(i);
            PlaySound(MS_MOVE);
         }else
            PlaySound(MS_MOVE_FAIL);
      }
   }

public:
   C_menu_base(C_class_to_be_ticked_hd *tc1 = NULL, void *c1 = NULL):
      context(c1),
      saved_tick_class(tc1),
      active_control_index(-1)
   {
      scene = driver->CreateScene();
      scene->Release();

                              //init menu font
      bool b;
      b = texts.Init();
      assert(b);

                              //init input controller
      {
         menu_controller = CreateController();
         menu_controller->Init(igraph, NULL, IS_LAST);
         static const struct S_input_slot{
            C_controller::E_INPUT_DEVICE input_device;
            dword data;
         } input_slots[][2] = {
            {C_controller::ID_KEYBOARD, K_CURSORLEFT},
            {C_controller::ID_KEYBOARD, K_CURSORUP},
            {C_controller::ID_KEYBOARD, K_CURSORRIGHT},
            {C_controller::ID_KEYBOARD, K_CURSORDOWN},
            {C_controller::ID_KEYBOARD, K_ENTER},
            {C_controller::ID_MOUSE_BUTTON, 0},
            {C_controller::ID_KEYBOARD, K_BACKSPACE},
            {C_controller::ID_KEYBOARD, K_ESC, C_controller::ID_MOUSE_BUTTON, 1},
            {C_controller::ID_KEYBOARD, K_HOME},
            {C_controller::ID_KEYBOARD, K_END},
            {C_controller::ID_KEYBOARD, K_GREYPLUS, C_controller::ID_KEYBOARD, K_PAGEUP},
            {C_controller::ID_KEYBOARD, K_GREYMINUS, C_controller::ID_KEYBOARD, K_PAGEDOWN},
         };
         for(int i=0; i<IS_LAST; i++){
            for(int j=0; j<2; j++){
               menu_controller->SetupSlotInput(i, j, C_controller::VM_BOOL, input_slots[i][j].input_device, input_slots[i][j].data);
            }
         }
         menu_controller->IgnoreAllPressed();
         menu_controller->Release();
      }
   }

   ~C_menu_base(){
      ClearContents();
   }

//----------------------------

   inline void SetContext(void *c){ context = c; }
   inline void *GetContext() const{ return context; }

//----------------------------
// Initialize menu from template.
   bool InitMenu(const S_menu_template &templ){

      InitLayout(templ, templ.begin_layout);

      if(cursor)
         cursor->SetScreenPos(igraph->Mouse_x(), (float)igraph->Mouse_y() / (screen_aspect_ratio * 1.33333333f));

      if(templ.flags&MT_PLAY_SOUND)
         PlaySound(MS_APPEAR);

      return true;
   }

//----------------------------
// Initialize elements of menu.
   void InitLayout(const S_menu_template &menu_templ, const S_menu_layout_template *templ = NULL,
      int active_control_index1 = -1){

      if(!templ)
         templ = menu_templ.begin_layout;

      InitRectangle(menu_templ.left, menu_templ.top, menu_templ.size_x, menu_templ.size_y,
         1.0f / menu_templ.texel_size);


      ClearContents();
                              //init menu name
      if(templ->menu_name_id){
         int txt_id = templ->menu_name_id;
         const char *cp = all_txt[txt_id];
         bool wide = all_txt.IsWide(txt_id);
         //dword color = 0xffe0e060;
         dword color = 0xffffffff;
         float sz = .005f + 40.0f * GetTexelSize();
         //float sz = 40.0f * GetTexelSize();
         float x = menu_client_rect.l + (menu_client_rect.r - menu_client_rect.l) * .5f;
         float y = menu_rect.t + (menu_client_rect.t - menu_rect.t) * .5f;
         const float shadow_offset = .004f;
         const float it_angle = PI * .1f;
         menu_name[0] = texts.CreateText(cp, color, x, y, sz, TEXT_JUSTIFY_CENTER,
            wide, NULL, NULL, NULL, NULL, 0, it_angle);
         menu_name[0]->Release();
         if(!wide){
            menu_name[1] = texts.CreateText(cp, 0xa0000000, x + shadow_offset, y + shadow_offset, sz, TEXT_JUSTIFY_CENTER,
               wide, NULL, NULL, NULL, NULL, -1, it_angle);
            menu_name[1]->Release();
         }
      }

      S_vector2 pos;
      pos.x = menu_client_rect.l + (menu_client_rect.r - menu_client_rect.l) * .5f;
      pos.y = menu_client_rect.t;
      float scale = 1.0f + templ->global_scale_delta;

      for(const S_menu_control_template *ct = templ->lines; ct->ctrl != MC_NULL; ++ct){
         C_menu_control_base *ctrl = CreateAndAddControl(*ct, pos, scale);

         if(ct->flags&MCT_CALLBACK){
                              //do creation callback
            assert(ct->callback);
            if(!(*ct->callback)(MCM_CTRL_CREATED, this, ctrl, context)){
               controls.pop_back();
               ctrl = NULL;
            }
         }
         if(ctrl && (ct->flags&MCT_DEFAULT_ITEM)){
            assert(controls.back()->MayBeActivated());
            if(active_control_index1==-1)
               active_control_index1 = controls.size() - 1;
         }
      }

      if(templ->flags&MLT_CLEAR_HISTORY)
         history.clear();
      
      if(templ->flags&MLT_NO_MOUSE_CURSOR){
         cursor = NULL;
      }else
      if(!cursor)
         CreateCursor();

      bool may_go_back = (history.size() && history.back().layout_templ != templ);
      if(!may_go_back && !(templ->flags&MLT_CLEAR_HISTORY))
         may_go_back = (saved_tick_class!=NULL);

      if(templ->flags&MLT_NO_BACK_BUTTON)
         may_go_back = false;
      else
      if(templ->flags&MLT_FORCE_BACK_BUTTON)
         may_go_back = true;
                              //add back button (if we have some history)
      if(may_go_back){
         float scale = 1.06f;
         const S_menu_control_template ct_back = {
            MC_BUTTON, 0, 0, 0, MCT_POS_Y_TOP | MCT_POS_X_LEFT | MCT_POS_NO_CLIENT,
               MT_MAKE_SPR_INDEX(16, 17), 1, 0
         };
         CreateAndAddControl(ct_back, pos, scale);
      }

                              //save history info
      {
                              //don't save the same menu twice
         if(!history.size() || history.back().layout_templ != templ){
            S_history_element he;
            he.menu_templ = menu_templ;
            he.layout_templ = templ;
            he.active_control_index = -1;
            history.push_back(he);
         }
      }

      pointed_control_index = -1;
      clicked_control_index = -1;

      int ai;
      if(active_control_index1==-1){
                              //make 1st possible control active
         for(int i=0; i<controls.size(); i++){
            if(controls[i]->MayBeActivated())
               break;
         }
         ai = (i!=controls.size()) ? i : -1;
      }else{
         ai = active_control_index1;
                              //check if valid control is activated
         if(ai >= controls.size() || !controls[ai]->MayBeActivated()){
            assert(0);
            ai = -1;
         }
      }
      SetActiveControl(ai);
                              //re-compute pointed-at control
      UpdateMouseMovement(0, 0, true);
   }

//----------------------------
// Reset the contents of current layout - rebuild.
   void ResetLayout(){
      if(history.size()){
         InitLayout(history.back().menu_templ, history.back().layout_templ, active_control_index);
      }
   }

//----------------------------
// Return one step back in history - re-init menu.
   virtual bool ReturnBackInHistory(){

      assert(history.size());
      if(history.size()==1){
         if(!(history.back().layout_templ->flags&MLT_CLEAR_HISTORY)){
            if(saved_tick_class){
               tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(saved_tick_class);
            }
            return true;
         }
         return false;
      }
                        //try to navigate one menu back
      if(history.size() < 2){
         return false;
      }
      S_history_element he = history[history.size() - 2];

      history.pop_back();
      history.pop_back();

      InitLayout(he.menu_templ, he.layout_templ, he.active_control_index);
      return true;
   }

//----------------------------
// Process click of control. If inherited class returns false, the control is not clicked (and processed further).
   virtual bool OnClick(C_menu_control_base *ctrl){

      return true;
   }

//----------------------------
// Set active control. This also updates hint message.
// The input parameter may be -1 to set no active control.
   void SetActiveControl(int indx){

      assert(indx==-1 || indx < controls.size());
      if(active_control_index == indx)
         return;
      if(active_control_index!=-1){
         controls[active_control_index]->SetActive(false);
         if(pointed_control_index != active_control_index)
            controls[active_control_index]->SetPointed(false);
      }

      int hint_id = 0;

      active_control_index = indx;

      if(active_control_index!=-1){
         LPC_menu_control_base ctrl = controls[active_control_index];
         ctrl->SetActive(true);
         hint_id = ctrl->GetHintTextID();
      }
      SetupHintMessage(hint_id);
   }

//----------------------------

   virtual void Render(){

      driver->BeginScene();

                              //render base first, then menu above this
      if(saved_tick_class)
         saved_tick_class->Render();

      bool use_zb = driver->GetState(RS_USEZB);
      if(use_zb) driver->SetState(RS_USEZB, false);

                              //render background
      C_menu_background::Render();

      if(hint_text)
         hint_text->Render();

                              //render all controls
      for(int i=controls.size(); i--; )
         controls[i]->Render();
      texts.Render();

      if(cursor)
         cursor->Render();

      if(use_zb) driver->SetState(RS_USEZB, true);

      driver->EndScene();
   }

//----------------------------

   virtual void ResetScreen(){

      if(saved_tick_class)
         saved_tick_class->ResetScreen();
                              //reset menu
      ResetLayout();
      if(cursor){
         cursor = NULL;
         CreateCursor();
      }
   }

//----------------------------

   virtual void Tick(const S_tick_context &tc){

      AddRef();               //for case we'll end released

      if(saved_tick_class){
                              //tick saved without input
         S_tick_context tc1 = tc;
         tc1.ClearInput();
         if(!net && saved_tick_class->GetID()=='GMIS')
            tc1.time = 0;
         saved_tick_class->Tick(tc1);
      }

      menu_controller->Update(tc);
                              //update cursor movement
      UpdateMouseMovement(tc.mouse_rel[0], tc.mouse_rel[1]);

      bool input_ate = false;

      C_menu_control_base::E_UPDATE_ACTION ua = C_menu_control_base::UA_NULL;
      if(clicked_control_index==-1){

         E_INPUT_SLOT is;

         if(active_control_index != -1){
            C_menu_control_base::E_UPDATE_ACTION ua1;
            if((ua1 = C_menu_control_base::UA_MOVE_RIGHT, menu_controller->Get(is = IS_ADD)) ||
               (ua1 = C_menu_control_base::UA_MOVE_LEFT, menu_controller->Get(is = IS_SUBTRACT))){

               menu_controller->IgnoreValue(is, KEYBOARD_DELAY_TIME);
               ua = ua1;
            }
            E_MENU_SOUND play_sound = MS_LAST;
            input_ate = controls[active_control_index]->Tick(tc, menu_controller, ua,
               play_sound, context);
            if(play_sound!=MS_LAST) PlaySound(play_sound);
         }

         if(!input_ate){
                              //navigate through menu
            if(menu_controller->Get(is = IS_LEFT) ||
               menu_controller->Get(is = IS_RIGHT) ||
               menu_controller->Get(is = IS_UP) ||
               menu_controller->Get(is = IS_DOWN) ||
               menu_controller->Get(is = IS_HOME) ||
               menu_controller->Get(is = IS_END)){

               menu_controller->IgnoreValue(is, KEYBOARD_DELAY_TIME);
               bool st = ChangeActiveItem(is);
               PlaySound(st ? MS_MOVE : MS_MOVE_FAIL);
            }

            if(menu_controller->Get(is = IS_BACK) || menu_controller->Get(is = IS_ESCAPE)){
               if(!(history.back().layout_templ->flags&MLT_NO_BACK_BUTTON)){
                  if(ReturnBackInHistory()){
                     menu_controller->IgnoreValue(is);
                     PlaySound(MS_BACK);
                  }
               }
            }
         }
      }else{
                              //redirect keys into control, if interested
         E_INPUT_SLOT is;
         C_menu_control_base::E_UPDATE_ACTION ua1;
         if((ua1 = C_menu_control_base::UA_MOVE_LEFT, menu_controller->Get(is = IS_LEFT)) ||
            (ua1 = C_menu_control_base::UA_MOVE_RIGHT, menu_controller->Get(is = IS_RIGHT)) ||
            (ua1 = C_menu_control_base::UA_MOVE_UP, menu_controller->Get(is = IS_UP)) ||
            (ua1 = C_menu_control_base::UA_MOVE_DOWN, menu_controller->Get(is = IS_DOWN)) ||
            (ua1 = C_menu_control_base::UA_MOVE_HOME, menu_controller->Get(is = IS_HOME)) ||
            (ua1 = C_menu_control_base::UA_MOVE_END, menu_controller->Get(is = IS_END))){

            menu_controller->IgnoreValue(is, KEYBOARD_DELAY_TIME);
            ua = ua1;
         }else
         if(tc.mouse_rel[0] || tc.mouse_rel[1]){
            ua = C_menu_control_base::UA_MOVE_RELATIVE;
         }

         E_MENU_SOUND play_sound = MS_LAST;
         input_ate = controls[active_control_index]->Tick(tc, menu_controller, ua, play_sound, context);
         if(play_sound!=MS_LAST)
            PlaySound(play_sound);
      }

      if(!input_ate)
         ProcessShortcuts(tc);

      C_menu_control_base *ctrl_clicked = GetClickedControl(tc, input_ate);
      if(ctrl_clicked){

                              //set to true if result clicking successful (play sound)
         bool success = false;

         switch(ctrl_clicked->GetType()){

         case MC_TEXT_BUTTON:
         case MC_BUTTON:
            {
               LPC_menu_control_base ctrlb = (LPC_menu_control_base)ctrl_clicked;
               success = OnClick(ctrl_clicked);

               if(success && ctrlb->context_data){
                                 //check if we're returning back
                  bool return_back = (ctrlb->context_data == 1);
                  if(return_back){
                     success = ReturnBackInHistory();
                     break;
                  }
                                 //save currently active item index
                  if(history.size())
                     history.back().active_control_index = active_control_index;

                  success = true;
                              //re-initialize menu using new template
                  if(ctrlb->GetFlags()&MCT_MENU_LINK){
                     const S_menu_template &templ = *(const S_menu_template*)ctrlb->link_to;
                                 //re-size menu
                     InitLayout(templ, templ.begin_layout);
                  }else
                  if(ctrlb->GetFlags()&MCT_CALLBACK){
                                 //do callback
                     success = (*ctrlb->callback)(MCM_CTRL_CLICKED, this, ctrl_clicked, context);
                  }else{
                     assert(history.size());
                     InitLayout(history.back().menu_templ, (S_menu_layout_template*)ctrlb->link_to);
                  }
               }
            }
            break;

         case MC_CHECKBOX:
            {
               C_menu_control_checkbox *cc = (C_menu_control_checkbox*)ctrl_clicked;
               cc->SetCheck(!cc->IsChecked());
               if(cc->context_data && (cc->GetFlags()&MCT_CALLBACK)){
                              //do callback
                  success = (*cc->callback)(MCM_CTRL_CLICKED, this, ctrl_clicked, context);
               }
            }
            break;

         case MC_SLIDER:
                              //no action for clicking
            break;

         case MC_EDITBOX:
            {
               C_menu_control_editbox *ce = (C_menu_control_editbox*)ctrl_clicked;
               ce->ToggleEditMode();
            }
            break;

         case MC_CONTROLLER_BOX:
            {
               C_menu_control_controller_box *cc = (C_menu_control_controller_box*)ctrl_clicked;
               cc->ToggleScanMode();
            }
            break;

         case MC_ENUMBOX:
            {
               C_menu_control_enumbox *ce = (C_menu_control_enumbox*)ctrl_clicked;
               int nume = ce->GetNumEnums();
               if(nume)
                  ce->SetCurrEnum((ce->GetCurrEnum() + 1) % nume);
               if(ce->context_data && (ce->GetFlags()&MCT_CALLBACK)){
                              //do callback
                  success = (*ce->callback)(MCM_CTRL_CLICKED, this, ctrl_clicked, context);
               }
            }
            break;

         default:
            assert(0);
         }
         PlaySound(success ? MS_FORWARD : MS_FORWARD_FAIL);
      }

      Release();
   }

//----------------------------

   virtual bool ActivateControl(int ctrl_id){

      for(int i=0; i<controls.size(); i++){
         C_menu_control_base *ctrl = controls[i];
         if(ctrl->MayBeActivated() && ctrl->GetID() == ctrl_id){
            SetActiveControl(i);
            return true;
         }
      }
      return false;
   }

//----------------------------

   virtual dword GetID() const{ return 'MMNU'; }

//----------------------------

   virtual void Notify(PI3D_frame frm, dword notify_key, dword anim_time){
      if(saved_tick_class)
         saved_tick_class->Notify(frm, notify_key, anim_time);
   }

//----------------------------
// Get control with specified (non-zero) ID.
   C_menu_control_base *GetControl(int id) const{

      if(id)
      for(int i=controls.size(); i--; ){
         if(controls[i]->GetID()==id)
            return controls[i];
      }
      return NULL;
   }

//----------------------------
// Get currently active control (if any).
   inline C_menu_control_base *GetActiveControl() const{
      assert(active_control_index < controls.size());
      return (active_control_index==-1) ? NULL : controls[active_control_index];
   }
};

//----------------------------
//----------------------------

bool C_menu_control_controller_box::Tick(const S_tick_context &tc, LPC_controller ctrl, E_UPDATE_ACTION ua,
   E_MENU_SOUND &play_sound, void *context){

   assert(active);
   if(in_scan_mode){
      switch(tc.key){
      case K_ESC:
         {
            int slot, input_index;
            if(ctrl->FindSlot(C_controller::ID_KEYBOARD, K_ESC, slot, input_index))
               ctrl->IgnoreValue(slot);
         
            play_sound = MS_BACK;
                              //end of editing, cancel changes, and back to passive mode
            ToggleScanMode();
            return true;
         }
         break;
                           //ignore F? keys
      case K_F1: case K_F2: case K_F3: case K_F4:
      case K_F5: case K_F6: case K_F7: case K_F8:
      case K_F9: case K_F10: case K_F11: case K_F12:
         return true;
      }
                           //perform scanning
      C_controller::E_INPUT_DEVICE id;
      dword data;
      if(cfg_controller->ScanInput(id, data)){
                           //check if it is supported one
         bool ok = true;
         switch(id){
         case C_controller::ID_MOUSE_AXIS:
                           //ignore main mouse axes, wheel is ok
            if((data&0x7fffffff) != 2)
               ok = false;
            break;
         }
         if(ok){
                           //add / replace in the slot
            C_controller::E_VALUE_MODE curr_vm;
            C_controller::E_INPUT_DEVICE curr_id[2];
            dword curr_data[2];
            bool ok;
            for(int i=0; i<2; i++)
               ok = cfg_controller->GetSlotSetting(slot, i, curr_vm, curr_id[i], curr_data[i]);

                           //remove this item from all other slots
            while(true){
               int slot, input_index;
               bool ok = cfg_controller->FindSlot(id, data, slot, input_index);
               if(!ok)
                  break;
               cfg_controller->SetupSlotInput(slot, input_index, C_controller::VM_NULL, C_controller::ID_NULL, 0);
            }


            bool set_ok = false;
            if(curr_id[0]!=C_controller::ID_NULL && curr_id[1]!=C_controller::ID_NULL){
                           //controller previously full, clear contents
               set_ok = cfg_controller->SetupSlotInput(slot, 0, C_controller::VM_NULL, id, data);
               cfg_controller->SetupSlotInput(slot, 1, C_controller::VM_NULL, C_controller::ID_NULL, 0);
            }else
            if(curr_id[0]==id && curr_data[0]==data){
                           //this value already present, ignore
               set_ok = cfg_controller->SetupSlotInput(slot, 0, curr_vm, id, data);
            }else
            if(curr_id[1]==id && curr_data[1]==data){
                           //this value already present, ignore
               set_ok = cfg_controller->SetupSlotInput(slot, 1, curr_vm, id, data);
            }else{
                           //add this input to unused index
               set_ok = cfg_controller->SetupSlotInput(slot, (curr_id[0]==C_controller::VM_NULL) ? 0 : 1,
                  C_controller::VM_NULL, id, data);
            }
            {
                           //ignore this value in menu's controller
               int slot, input_index;
               if(ctrl->FindSlot(id, data, slot, input_index))
                  ctrl->IgnoreValue(slot);
            }
            play_sound = set_ok ? MS_EDIT_OK : MS_MODIFY_FAIL;
            ToggleScanMode();

            menu->ResetLayout();
            return true;
         }
      }

                           //make flashing
      if((flash_countdown -= tc.time) <= 0){
         flash_phase = !flash_phase;
         flash_countdown = 150;
      }
      return true;
   }
   return false;
}

//----------------------------
//----------------------------
                              //complete implementation of controller configuration

enum E_CONTROLLER_CONFIG_CTRL_ID{
   CTRLID_INVERT_MOUSE = 1,
   CTRLID_ALWAYS_RUN = 2,
   CTRLID_MOUSE_SENS = 3,
};

//----------------------------
// Controller config - reset to defaults.
static bool mc_ControllerConfigDefaults(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
   C_menu_control_base *ctrl, void *context){

   switch(msg){
   case MCM_CTRL_CLICKED:
      SetupControllerDefaults();
      menu->ResetLayout();
      break;
   }
   return true;
}

//----------------------------
// Controller config - process checkboxes.
static bool mc_ControllerConfigCheckbox(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
   C_menu_control_base *ctrl, void *context){

   switch(msg){
   case MCM_CTRL_CLICKED:
      if(ctrl->GetType() == MC_CHECKBOX){
         C_menu_control_checkbox *cc = (C_menu_control_checkbox*)ctrl;
         C_controller::E_CONFIG cfg_item;
         switch(ctrl->GetID()){
         case CTRLID_INVERT_MOUSE: cfg_item = C_controller::CFG_INVERT_MOUSE_Y; break;
         case CTRLID_ALWAYS_RUN: cfg_item = C_controller::CFG_ALWAYS_RUN; break;
         default: assert(0);
         }
         controller->SetConfigValue(cfg_item, cc->IsChecked());
      }
      break;

   case MCM_CTRL_CREATED:
      if(ctrl->GetType() == MC_CHECKBOX){
         C_menu_control_checkbox *cc = (C_menu_control_checkbox*)ctrl;
         C_controller::E_CONFIG cfg_item;
         switch(ctrl->GetID()){
         case CTRLID_INVERT_MOUSE: cfg_item = C_controller::CFG_INVERT_MOUSE_Y; break;
         case CTRLID_ALWAYS_RUN: cfg_item = C_controller::CFG_ALWAYS_RUN; break;
         default: assert(0);
         }
         cc->SetCheck(controller->GetConfigValue(cfg_item));
      }else
      if(ctrl->GetType() == MC_SLIDER){
         C_menu_control_slider *cs = (C_menu_control_slider*)ctrl;
         switch(ctrl->GetID()){
         case CTRLID_MOUSE_SENS:
            cs->SetSlidePos(controller->GetConfigValue(C_controller::CFG_MOUSE_SENSITIVITY)/2);
            break;
         default: assert(0);
         }
      }
      break;

   case MCM_CTRL_SLIDED:
      if(ctrl->GetType() == MC_SLIDER){
         C_menu_control_slider *cs = (C_menu_control_slider*)ctrl;
         switch(ctrl->GetID()){
         case CTRLID_MOUSE_SENS:
            controller->SetConfigValue(C_controller::CFG_MOUSE_SENSITIVITY, cs->GetSlidePos() * 2);
            break;
         default: assert(0);
         }
      }
      break;
   }
   return true;
}

//----------------------------

extern const S_menu_layout_template mlt_advanced, mlt_weapons, mlt_camera;
static const S_menu_layout_template mlt_basic = {
   {240},
   {
      {MC_CONTROLLER_BOX, 0, 0, .01f, MCT_SIZE_1 | MCT_JUSTIFY_RIGHT, 241, GKEY_FORWARD, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 242, GKEY_BACKWARD, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 243, GKEY_LEFT, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 244, GKEY_RIGHT, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 245, GKEY_JUMP, 0},

      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 247, GKEY_MOVE_LEFT, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 248, GKEY_MOVE_RIGHT, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 249, GKEY_STRAIFE, 0},
      
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 250, GKEY_RUN, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 251, GKEY_UP, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 252, GKEY_DOWN, 0},

                        //additional description for this layout
      //{MC_TEXT, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_SIZE_4 | MCT_POS_Y_TOP | MCT_POS_X_CENTER, 1, NULL},
                        //defaults/prev/next buttons
      {MC_TEXT_BUTTON, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_CALLBACK | MCT_POS_Y_BOTTOM | MCT_POS_X_CENTER | MCT_JUSTIFY_CENTER, 20212, (dword)&mc_ControllerConfigDefaults, 0},
      {MC_TEXT_BUTTON, 0, .04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_LEFT | MCT_JUSTIFY_LEFT, 20210, (dword)&mlt_camera, 0},
      {MC_TEXT_BUTTON, 0, -.04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_RIGHT | MCT_JUSTIFY_RIGHT, 20211, (dword)&mlt_advanced, 0},

      {MC_NULL}               //terminator
   }
}, mlt_advanced = {
   {240, MLT_NO_HISTORY},
   {
      {MC_CONTROLLER_BOX, 0, 0, .01f, MCT_SIZE_1 | MCT_JUSTIFY_RIGHT, 246, GKEY_FIRE, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 254, GKEY_USE, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 255, GKEY_INV_SCROLL_LEFT, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 256, GKEY_INV_SCROLL_RIGHT, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 257, GKEY_ITEM_USE, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 253, GKEY_MAP, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 269, GKEY_RELOAD, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 270, GKEY_CENTER, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 275, GKEY_ZOOM, 0},

                        //additional description for this layout
      //{MC_TEXT, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_SIZE_4 | MCT_POS_Y_TOP | MCT_POS_X_CENTER, 2013, NULL},
                        //defaults/prev/next buttons
      {MC_TEXT_BUTTON, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_CALLBACK | MCT_POS_Y_BOTTOM | MCT_POS_X_CENTER | MCT_JUSTIFY_CENTER, 20212, (dword)&mc_ControllerConfigDefaults, 0},
      {MC_TEXT_BUTTON, 0, .04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_LEFT | MCT_JUSTIFY_LEFT, 20210, (dword)&mlt_basic, 0},
      {MC_TEXT_BUTTON, 0, -.04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_RIGHT | MCT_JUSTIFY_RIGHT, 20211, (dword)&mlt_weapons, 0},

      {MC_NULL}               //terminator
   }
}, mlt_weapons = {
   {240, MLT_NO_HISTORY},
   {
      {MC_CONTROLLER_BOX, 0, 0, .01f, MCT_SIZE_1 | MCT_JUSTIFY_RIGHT, 258, GKEY_PLR_SWITCH, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 259, GKEY_PLR_SWITCH_1, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 260, GKEY_PLR_SWITCH_2, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 261, GKEY_PLR_SWITCH_3, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 262, GKEY_PLR_SWITCH_4, 0},

      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 263, GKEY_CMD_FOLLOW, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 264, GKEY_CMD_STOP, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 265, GKEY_CMD_MOVE, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 266, GKEY_CMD_HOLD_FIRE, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 267, GKEY_CMD_HEY, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 268, GKEY_CMD_PRG_CLR, 0},

                        //additional description for this layout
      //{MC_TEXT, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_SIZE_4 | MCT_POS_Y_TOP | MCT_POS_X_CENTER, 2016, NULL},
                        //defaults/prev/next buttons
      {MC_TEXT_BUTTON, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_CALLBACK | MCT_POS_Y_BOTTOM | MCT_POS_X_CENTER | MCT_JUSTIFY_CENTER, 20212, (dword)&mc_ControllerConfigDefaults, 0},
      {MC_TEXT_BUTTON, 0, .04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_LEFT | MCT_JUSTIFY_LEFT, 20210, (dword)&mlt_advanced, 0},
      {MC_TEXT_BUTTON, 0, -.04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_RIGHT | MCT_JUSTIFY_RIGHT, 20211, (dword)&mlt_camera, 0},

      {MC_NULL}               //terminator
   }
}, mlt_camera = {
   {240, MLT_NO_HISTORY},
   {
      {MC_CONTROLLER_BOX, 0, 0, .01f, MCT_SIZE_1 | MCT_JUSTIFY_RIGHT, 271, GKEY_CAM_ZOOM_IN, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 272, GKEY_CAM_ZOOM_OUT, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 273, GKEY_CAM_VIEW_UP, 0},
      {MC_CONTROLLER_BOX, 0, 0, 0, MCT_JUSTIFY_RIGHT, 274, GKEY_CAM_VIEW_DOWN, 0},

      {MC_CHECKBOX, CTRLID_ALWAYS_RUN, 0, .02f, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20213, (dword)&mc_ControllerConfigCheckbox, 0},
      {MC_SLIDER, CTRLID_MOUSE_SENS, 0, 0, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20214, (dword)&mc_ControllerConfigCheckbox, 0},
      {MC_CHECKBOX, CTRLID_INVERT_MOUSE, 0, 0, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20215, (dword)&mc_ControllerConfigCheckbox, 0},


                        //additional description for this layout
      //{MC_TEXT, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_SIZE_3 | MCT_POS_Y_TOP | MCT_POS_X_CENTER, 2017, NULL},
                        //defaults/prev/next buttons
      {MC_TEXT_BUTTON, 0, 0, -MCT_DEFAUL_TEXT_SIZE, MCT_CALLBACK | MCT_POS_Y_BOTTOM | MCT_POS_X_CENTER | MCT_JUSTIFY_CENTER, 20212, (dword)&mc_ControllerConfigDefaults, 0},
      {MC_TEXT_BUTTON, 0, .04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_LEFT | MCT_JUSTIFY_LEFT, 20210, (dword)&mlt_weapons, 0},
      {MC_TEXT_BUTTON, 0, -.04f, -MCT_DEFAUL_TEXT_SIZE, MCT_POS_Y_BOTTOM | MCT_POS_X_RIGHT | MCT_JUSTIFY_RIGHT, 20211, (dword)&mlt_basic, 0},

      {MC_NULL}               //terminator
   }
};

extern const S_menu_template controller_menu_templ;
const S_menu_template controller_menu_templ = {
   {.5f - .35f, .1f, .7f, .52f, 800.0f},
   &mlt_basic
};

//----------------------------
//----------------------------
                              //complete implementation of video configuration
enum E_VIDEO_CONFIG_CTRL_ID{
   VIDCFGID_BRIGHTNESS = 1,   //global brightness
   VIDCFGID_FULLSCREEN,
   VIDCFGID_RESOLUTION,
   VIDCFGID_COLORDEPTH,
   VIDCFGID_TXT_COMPRESS,
   VIDCFGID_SHADOWS,
};

//----------------------------
// Sound config - reset to defaults.
static bool mc_VideoConfigDefaults(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
   C_menu_control_base *ctrl, void *context){

   switch(msg){
   case MCM_CTRL_CLICKED:
      game_configuration.ResetVideoDefaults();
      game_configuration.WriteSysConfig();
      ((C_menu_base*)menu)->ResetLayout();
      break;
   }
   return true;
}

//----------------------------

static bool mc_VideoConfigKeepChanges(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
   C_menu_control_base *ctrl,  void *context){

   switch(msg){
   case MCM_CTRL_CLICKED:
      {
                              //ok, changes accepted, save it and return
         game_configuration.mode_x = igraph->Scrn_sx();
         game_configuration.mode_y = igraph->Scrn_sy();
         game_configuration.bits_per_pixel = igraph->GetPixelFormat()->bytes_per_pixel * 8;
         game_configuration.fullscreen = igraph->GetFlags()&IG_FULLSCREEN;

         C_menu_base *menub = (C_menu_base*)menu;
         return menub->ReturnBackInHistory();
      }
      break;
   }
   return true;
}

//----------------------------

static bool mc_VideoConfigDiscardChanges(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
   C_menu_control_base *ctrl, void *context){

   switch(msg){
   case MCM_CTRL_CLICKED:
      {
                              //discard changes, put original settings back
         game_configuration.ApplyVideoConfig();
         C_menu_base *menub = (C_menu_base*)menu;
         return menub->ReturnBackInHistory();
      }
      break;
   }
   return true;
}

//----------------------------

static const S_menu_layout_template mlt_video_apply = {
   {20507, MLT_NO_BACK_BUTTON},
   {
      {MC_TEXT,        0,     0,  .03f, MCT_SIZE_3 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 20500, NULL},
      {MC_TEXT_BUTTON, 0, -.08f, -.09f, MCT_CALLBACK | MCT_SIZE_3 | MCT_POS_X_CENTER | MCT_POS_Y_BOTTOM, 349, (dword)&mc_VideoConfigKeepChanges},
      {MC_TEXT_BUTTON, 0, +.08f, -.09f, MCT_CALLBACK | MCT_DEFAULT_ITEM | MCT_POS_X_CENTER | MCT_POS_Y_BOTTOM, 348, (dword)&mc_VideoConfigDiscardChanges},

      {MC_NULL}               //terminator
   }
};
extern const S_menu_template video_menu_templ;

//----------------------------

static bool mc_VideoConfigApply(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
   C_menu_control_base *ctrl, void *context){

   switch(msg){
   case MCM_CTRL_CLICKED:
      {
                              //for windowed modes, don't allow the window size
                              // be greater than screen size
         if(!game_configuration.fullscreen){
            int dsx, dsy;
            igraph->GetDesktopResolution(dsx, dsy);
            if(game_configuration.mode_x > dsx ||
               game_configuration.mode_y > dsy)
            return false;
         }

         int mode_x = igraph->Scrn_sx(), mode_y = igraph->Scrn_sy();
         int curr_bpp = igraph->GetPixelFormat()->bytes_per_pixel;
         bool curr_fs = (igraph->GetFlags()&IG_FULLSCREEN);
         bool resolution_changed = (game_configuration.mode_x!=mode_x ||
            game_configuration.mode_y!=mode_y ||
            (game_configuration.bits_per_pixel/8)!=curr_bpp);
         bool confirm_changes = ((game_configuration.fullscreen && resolution_changed) ||
            curr_fs != game_configuration.fullscreen);
         game_configuration.ApplyVideoConfig();
         if(confirm_changes){
                              //put old settings back, and keep changes if user accepts that
            game_configuration.mode_x = mode_x;
            game_configuration.mode_y = mode_y;
            game_configuration.bits_per_pixel = curr_bpp * 8;
            game_configuration.fullscreen = curr_fs;

            C_menu_base *menub = (C_menu_base*)menu;
            menub->InitLayout(video_menu_templ, &mlt_video_apply);
         }
         if(resolution_changed)
            menu->ResetScreen();
      }
      break;
   }
   return true;
}

//----------------------------

static bool mc_VideoConfigCallback(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
   C_menu_control_base *ctrl, void *context){

   switch(ctrl->GetID()){
   case VIDCFGID_BRIGHTNESS:
      {
         assert(ctrl->GetType()==MC_SLIDER);
         C_menu_control_slider *cs = (C_menu_control_slider*)(C_menu_control_base*)ctrl;
         switch(msg){
         case MCM_CTRL_CREATED:
            cs->SetSlidePos(game_configuration.brightness / 2);
            break;
         case MCM_CTRL_SLIDED:
            {
                              //limit min brightness, avoid absolute black
               game_configuration.brightness = Max(20, cs->GetSlidePos() * 2);
               float f = game_configuration.brightness * .01f;
               game_configuration.WriteSysConfig();
               igraph->SetBrightness(f, f, f);
            }
            break;
         }
      }
      break;

   case VIDCFGID_FULLSCREEN:
   case VIDCFGID_TXT_COMPRESS:
   case VIDCFGID_SHADOWS:
      {
         assert(ctrl->GetType()==MC_CHECKBOX);
         C_menu_control_checkbox *cc = (C_menu_control_checkbox*)(C_menu_control_base*)ctrl;
         switch(msg){
         case MCM_CTRL_CREATED:
            {
               bool b;
               switch(ctrl->GetID()){
               case VIDCFGID_FULLSCREEN: b = game_configuration.fullscreen; break;
               case VIDCFGID_TXT_COMPRESS: b = game_configuration.tx_compress; break;
               case VIDCFGID_SHADOWS: b = game_configuration.render_shadows; break;
               default: assert(0);
               }
               cc->SetCheck(b);
            }
            break;
         case MCM_CTRL_CLICKED:
            {
               bool b = cc->IsChecked();;
               switch(ctrl->GetID()){
               case VIDCFGID_FULLSCREEN: game_configuration.fullscreen = b; break;
               case VIDCFGID_TXT_COMPRESS:
                  game_configuration.tx_compress = b;
                  if(!game_configuration.safe_mode)
                     driver->SetState(RS_TEXTURECOMPRESS, b);
                  break;
               case VIDCFGID_SHADOWS:
                  game_configuration.render_shadows = b;
                  if(!game_configuration.safe_mode)
                     driver->SetState(RS_USESHADOWS, b);
                  break;
               default: assert(0);
               }
               game_configuration.WriteSysConfig();
            }
            break;
         }
      }
      break;

   case VIDCFGID_RESOLUTION:
   case VIDCFGID_COLORDEPTH:
      {
         assert(ctrl->GetType()==MC_ENUMBOX);
         C_menu_control_enumbox *ce = (C_menu_control_enumbox*)(C_menu_control_base*)ctrl;
         switch(msg){
         case MCM_CTRL_CREATED:
            switch(ctrl->GetID()){
            case VIDCFGID_COLORDEPTH:
               {
                              //enum all display modes
                  vector<dword> all_widths;
                  dword num_modes = igraph->NumDisplayModes();
                  for(int i=0; i<num_modes; i++){
                     int sx, sy, bpp;
                     bool b = igraph->GetDisplayMode(i, &sx, &sy, &bpp);
                     assert(b);
                     if(b){
                        for(int j=all_widths.size(); j--; ){
                           if(all_widths[j]==bpp)
                              break;
                        }
                        if(j==-1)
                           all_widths.push_back(bpp);
                     }
                  }
                  sort(all_widths.begin(), all_widths.end());
                  dword curr_bpp = igraph->GetPixelFormat()->bytes_per_pixel;
                  for(i=0; i<all_widths.size(); i++){
                     dword bpp = all_widths[i];
                     ce->SetupEnum(i, C_fstring("%i-bit", bpp*8), bpp*8);
                     if(bpp == curr_bpp)
                        ce->SetCurrEnum(i);
                  }
               }
               break;

            case VIDCFGID_RESOLUTION:
               {
                  vector<dword> all_modes;
                              //enum all display modes
                  dword num_modes = igraph->NumDisplayModes();
                  for(int i=0; i<num_modes; i++){
                     int sx, sy, bpp;
                     bool b = igraph->GetDisplayMode(i, &sx, &sy, &bpp);
                     assert(b);
                     if(b){
                        if(sx<512 || sy<384)
                           continue;
                        dword mode_code = sx | (sy << 16);
                        for(int j=all_modes.size(); j--; ){
                           if(all_modes[j] == mode_code)
                              break;
                        }
                        if(j==-1)
                           all_modes.push_back(mode_code);
                     }
                  }
                              //sort modes
                  struct S_hlp{
                     static bool cbModeLess(dword m1, dword m2){
                        return ( ((m1&0xffff) * (m1>>16)) < ((m2&0xffff) * (m2>>16)));
                     }
                  };
                  sort(all_modes.begin(), all_modes.end(), S_hlp::cbModeLess);
                  for(i=0; i<all_modes.size(); i++){
                     dword sx = all_modes[i] & 0xffff;
                     dword sy = all_modes[i] >> 16;
                     ce->SetupEnum(i, C_fstring("%i x %i", sx, sy), all_modes[i]);
                     if(sx==igraph->Scrn_sx() && sy==igraph->Scrn_sy())
                        ce->SetCurrEnum(i);
                  }
               }
               break;
            }
            break;
         case MCM_CTRL_CLICKED:
            {
               dword user_data = ce->GetEnumData(ce->GetCurrEnum());
               switch(ctrl->GetID()){
               case VIDCFGID_COLORDEPTH:
                  game_configuration.bits_per_pixel = user_data;
                  break;
               case VIDCFGID_RESOLUTION:
                  game_configuration.mode_x = user_data&0xffff;
                  game_configuration.mode_y = user_data>>16;
                  break;
               }
            }
            break;
         }
      }
      break;
   }
   return true;
}

//----------------------------

static const S_menu_layout_template mlt_video = {
   {20507},
   {
      {MC_SLIDER,    VIDCFGID_BRIGHTNESS,    0, .02f, MCT_CALLBACK | MCT_SIZE_2 | MCT_JUSTIFY_RIGHT, 20501, (dword)&mc_VideoConfigCallback},
      {MC_CHECKBOX,  VIDCFGID_FULLSCREEN,    0,    0, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20502, (dword)&mc_VideoConfigCallback},
      {MC_ENUMBOX,  VIDCFGID_RESOLUTION,    0,    0, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20503, (dword)&mc_VideoConfigCallback},
      {MC_ENUMBOX,  VIDCFGID_COLORDEPTH,    0,    0, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20504, (dword)&mc_VideoConfigCallback},
      //{MC_CHECKBOX,  VIDCFGID_TXT_COMPRESS,  0,    0, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20505, (dword)&mc_VideoConfigCallback},
      {MC_CHECKBOX,  VIDCFGID_SHADOWS,  0,    0, MCT_CALLBACK | MCT_JUSTIFY_RIGHT, 20508, (dword)&mc_VideoConfigCallback},

                              //apply changes button
      {MC_TEXT_BUTTON, 0, 0, -MCT_DEFAUL_TEXT_SIZE*2.0f, MCT_CALLBACK | MCT_POS_Y_BOTTOM | MCT_POS_X_CENTER | MCT_JUSTIFY_CENTER, 20506, (dword)&mc_VideoConfigApply},

                              //defaults button
      {MC_TEXT_BUTTON, 0, 0, 0, MCT_CALLBACK | MCT_JUSTIFY_CENTER, 20212, (dword)&mc_VideoConfigDefaults},

      {MC_NULL}               //terminator
   }
};

extern const S_menu_template video_menu_templ;
const S_menu_template video_menu_templ = {
   {.5f - .35f, .1f, .7f, .52f, 800.0f},
   &mlt_video
};

//----------------------------
//----------------------------

C_class_to_be_ticked_hd *CreateMenu(const S_menu_template &templ, C_class_to_be_ticked_hd *tc,
   void *context){

   C_menu_base *mp = new C_menu_base(tc, context);
   mp->InitMenu(templ);
   return mp;
}

//----------------------------
//----------------------------

                              //Main menu class - not present in editor compilation
#ifndef EDITOR

#define NUM_ENUM_SESSIONS 8

enum E_MENU_ID_MAINMENU{
   IDMM_NULL,
                              //main menu:
   IDMM_SINGLE_PLAYER,
   IDMM_NET_HOST,
   IDMM_NET_JOIN,
   IDMM_CREDITS,
   IDMM_QUIT_LINK,
   IDMM_QUIT,
                              //network:
   IDMM_SESSION_0,
   IDMM_SESSION_LAST = IDMM_SESSION_0 + NUM_ENUM_SESSIONS - 1,

   IDMM_NETPLR_0,
   IDMM_NETPLR_LAST = IDMM_NETPLR_0 + 7,
   IDMM_NET_BEGIN,
   IDMM_NET_PLR_NAME,
                              //net config
   IDMM_NET_CONNECTION_TYPE,                                 
   IDMM_NET_TCPIP_ADDRESS,
   IDMM_NET_TCPIP_PORT,
   IDMM_NET_PHONE_NUMBER,
   IDMM_NET_VOICECHAT,
   IDMM_NET_VOICE_COMPRESSION,
   IDMM_NET_VOICE_TEST,

                              //config
   IDMM_SOUND_VOLUME_MUSIC,
   IDMM_SOUND_VOLUME_EFFECTS,
   IDMM_AIM_LINE,
   IDMM_AIM_TYPE,
                              //in-game
   IDMM_RESUME_GAME,
   IDMM_TO_MAIN_MENU,
   IDMM_RESTART_MISSION,
   IDMM_SAVE_GAME,
   IDMM_LOAD_GAME,
                              //Add-on
   IDMM_ADDON_PACK,
   IDMM_ADDON_SINGLE,
   IDMM_ADDON_NET_HOST,
};

extern const S_menu_template controller_menu_templ, sounds_menu_templ, video_menu_templ;


//----------------------------
                              //Credits
static const S_menu_layout_template layout_credits = {
   {203},
   {
      {MC_NULL}
   }
};
static const S_menu_template templ_credits = {
   { .1f, .05f, .8f, .65f, 1200.0f }, &layout_credits
};

//----------------------------
extern const S_menu_template templ_net_game;
extern const S_menu_layout_template layout_net_join, layout_opt_game, layout_net_wait_players;

//----------------------------

extern const S_menu_layout_template layout_opt_sound, layout_add_on;

static const S_menu_template templ_addon_net_host = {
   { .2f, .15f, .6f, .45f, 1200.0f }, &layout_net_wait_players
};

static const S_menu_layout_template
layout_quit = {
   {0},
   {
      {MC_TEXT, 0, 0, MCT_DEFAUL_TEXT_SIZE*.3f, MCT_SIZE_3 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 215},
      {MC_TEXT_BUTTON, IDMM_QUIT, -.06f, -MCT_DEFAUL_TEXT_SIZE*2.4f, MCT_DEFAULT_ITEM | MCT_SIZE_3 | MCT_POS_X_CENTER | MCT_POS_Y_BOTTOM, 216, NULL, 0},
      {MC_TEXT_BUTTON, 0, .06f, -MCT_DEFAUL_TEXT_SIZE*2.4f, MCT_POS_X_CENTER | MCT_POS_Y_BOTTOM, 217, 1, 0},
      {MC_NULL}               //terminator
   }
}, layout_options = {
   {202},
   {
      {MC_TEXT_BUTTON, 0, 0, .02f, MCT_MENU_LINK, 220, (dword)&controller_menu_templ},
      {MC_TEXT_BUTTON, 0, 0, 0, 0, 20220, (dword)&layout_opt_game, 0},
      {MC_TEXT_BUTTON, 0, 0, 0, MCT_MENU_LINK, 224, (dword)&video_menu_templ},
      {MC_TEXT_BUTTON, 0, 0, 0, 0, 222, (dword)&layout_opt_sound},
      {MC_NULL}
   }
}, layout_main_menu = {
   {0, MLT_CLEAR_HISTORY},
   {
      {MC_TEXT_BUTTON, IDMM_SINGLE_PLAYER,0, 0, 0, 200, NULL, 0},
      {MC_TEXT_BUTTON, 0, 0, 0, MCT_MENU_LINK, 201, (dword)&templ_net_game},
      {MC_TEXT_BUTTON, 0, 0, 0, 0, 20221, (dword)&layout_add_on},
      {MC_TEXT_BUTTON, 0, 0, 0, 0, 202, (dword)&layout_options},
      {MC_TEXT_BUTTON, IDMM_CREDITS,      0, 0, MCT_MENU_LINK, 203, (dword)&templ_credits},
      {MC_TEXT_BUTTON, IDMM_QUIT_LINK,    0, 0, 0, 204, (dword)&layout_quit},
      {MC_NULL}               //terminator
   }
};

static const S_menu_template templ_main_menu = {
   {
      .55f, .02f, .43f, .36f, 1200.0f
   },
   &layout_main_menu
};

//----------------------------
//----------------------------

static C_mission *CreateBackgroundMission(const char *name = "Logo"){

                           //create dummy background scene to run
   C_mission *back_scene = new C_mission;
   E_MISSION_IO mio = back_scene->Open(name, OPEN_USE_SCENE_CAMERA | 
      //OPEN_NO_SHOW_PROGRESS |
      OPEN_NO_EDITOR);
   if(mio!=MIO_OK){
      ErrReport("Failed to initialize background mission for menu", editor);
      back_scene->Release();
      back_scene = NULL;
   }
   return back_scene;
}

//----------------------------

class C_tick_class_main_menu: public C_menu_base{

                              //mission prepared to be played in multiplayer mode
   C_smart_ptr<C_mission> net_mission;

   int sess_enum_count;
   vector<C_smart_ptr<INET_session_address> > enum_sessions;
   vector<dword> enum_players;
   vector<C_string> addon_names;
   int net_host_addon_pack;   //-1 if not in add-on mode

   enum{
      STATE_NULL,
      STATE_MENU,             //waiting for user input from the menu
      STATE_SCROLLTEXT,       //displaying scrolltext
      STATE_NET_HOST_WAIT,    //host waiting for players to join
      STATE_NET_SESS_ENUM,    //enumerating sessions
      STATE_NET_SESS_JOIN,    //joining session
      STATE_NET_SYNC_WAIT,    //waiting for synchro start (clients)
   } state;

   C_smart_ptr<C_scrolltext> scroll_text;

//----------------------------

   static bool cbSessionLess(C_smart_ptr<INET_session_address> &s1, C_smart_ptr<INET_session_address> &s2){
      return (strcmp(s1->GetName(), s2->GetName()) < 0);
   }

//----------------------------
public:
   static bool cbCtrl(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
      C_menu_control_base *ctrl, void *context){

      int id = ctrl->GetID();

      switch(msg){
      case MCM_CTRL_CREATED:
         {
            C_tick_class_main_menu *mmnu = (C_tick_class_main_menu*)menu;

            switch(id){
            case IDMM_SOUND_VOLUME_MUSIC:
            case IDMM_SOUND_VOLUME_EFFECTS:
               {
                  assert(ctrl->GetType() == MC_SLIDER);
                  C_menu_control_slider *cs = (C_menu_control_slider*)ctrl;
                  int pc;
                  switch(id){
                  case IDMM_SOUND_VOLUME_MUSIC: pc = game_configuration.volume_music; break;
                  case IDMM_SOUND_VOLUME_EFFECTS: pc = game_configuration.volume_effects; break;
                  default: assert(0);
                  }
                  cs->SetSlidePos((cs->GetNumPositions() - 1) * pc / 100);
               }
               break;

            case IDMM_AIM_LINE:
               {
                  assert(ctrl->GetType() == MC_CHECKBOX);
                  C_menu_control_checkbox *cb = (C_menu_control_checkbox*)ctrl;
                  cb->SetCheck(game_configuration.draw_aim_line);
               }
               break;

            case IDMM_AIM_TYPE:
               {
                  assert(ctrl->GetType() == MC_ENUMBOX);
                  C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
                  for(int i=0; i<3; i++)
                     eb->SetupEnum(i, all_txt[20602+i], 0, all_txt.IsWide(20602+i));
                  eb->SetCurrEnum(game_configuration.aim_mode);
               }
               break;

            case IDMM_NET_BEGIN:
                              //available only if hosting, and if there're players in the list
               switch(mmnu->state){
               case STATE_NET_HOST_WAIT:
                  {
#ifndef DEBUG_NET_ALLOW_HOST_SINGLE
                     return (mmnu->enum_players.size()!=0);
#endif
                  }
                  break;
               default:
                  return false;
               }
               break;

            case IDMM_NET_PLR_NAME:
               {
                  assert(ctrl->GetType() == MC_EDITBOX);
                  C_menu_control_editbox *eb = (C_menu_control_editbox*)ctrl;
                  eb->SetEditText(game_configuration.net_plr_name);
               }
               break;

            case IDMM_NET_CONNECTION_TYPE:
               {
                  assert(ctrl->GetType() == MC_ENUMBOX);
                  C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
                  static struct{
                     IN_CONNECTION con;
                     const char *desc;
                  } data[] = {
                     INCON_TCPIP, "TCP/IP (Internet, LAN)",
                     INCON_IPX, "IPX",
#ifdef ALLOW_MODEM
                     INCON_MODEM, "Modem",
#endif
                     INCON_NULL
                  };
                  for(int i=0; data[i].con; i++){
                     eb->SetupEnum(i, data[i].desc, data[i].con);
                     if(game_configuration.net_ctype==data[i].con)
                        eb->SetCurrEnum(i);
                  }
               }
               break;

            case IDMM_NET_VOICE_COMPRESSION:
               {
                  assert(ctrl->GetType() == MC_ENUMBOX);
                  C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
                  static const char *compr_names[] = {
                     "1.2 Kb/sec",
                     "3.2 Kb/sec",
                     "6.4 Kb/sec",
                     "8 Kb/sec",
                     "13 Kb/sec",
                     "32 Kb/sec",
                     "64 Kb/sec"
                  };
                  for(int i=0; i<S_game_configuration::VCC_LAST; i++){
                     eb->SetupEnum(i, compr_names[i]);
                  }
                  eb->SetCurrEnum(game_configuration.voice_chat_compression);
               }
               break;

            case IDMM_NET_TCPIP_ADDRESS:
            case IDMM_NET_TCPIP_PORT:
            case IDMM_NET_PHONE_NUMBER:
               {
                  assert(ctrl->GetType() == MC_EDITBOX);
                  C_menu_control_editbox *eb = (C_menu_control_editbox*)ctrl;
                  const char *txt;
                  switch(id){
                  case IDMM_NET_TCPIP_ADDRESS:
                     txt = game_configuration.net_tcpip_addr;
                     eb->SetMaxEditLen(28);
                     break;
                  case IDMM_NET_TCPIP_PORT:
                     txt = game_configuration.net_tcpip_port;
                     eb->SetMaxEditLen(5);
                     break;
                  case IDMM_NET_PHONE_NUMBER:
                     txt = game_configuration.net_modem_num;
                     break;
                  default: assert(0);
                  }
                  eb->SetEditText(txt);
               }
               break;

            case IDMM_NET_VOICECHAT:
               {
                  assert(ctrl->GetType() == MC_CHECKBOX);
                  C_menu_control_checkbox *cb = (C_menu_control_checkbox*)ctrl;
                  cb->SetCheck(game_configuration.enable_voice_chat);
               }
               break;

            case IDMM_ADDON_PACK:
               {
                  assert(ctrl->GetType() == MC_ENUMBOX);
                  C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
                              //search add-ons
                  mmnu->addon_names.clear();
                  vector<C_string> dta_files;
                  OsCollectFiles("Data", "*.add", dta_files);
                  sort(dta_files.begin(), dta_files.end());
                  for(int i=0; i<dta_files.size(); i++){
                     const C_string &str = dta_files[i];
                     LPC_dta_read dta = dtaCreate(str);
                     if(dta){
                              //keep pure name (no base dir, no extension)
                        assert(str.Size() > 5+4);
                        C_string name = str;
                        name[(int)name.Size()-4] = 0;
                        name = &name[5];
                        name.ToLower();
                        name[0] = toupper(name[0]);
                        mmnu->addon_names.push_back(str);
                        dta->Release();

                        eb->SetupEnum(i, name);
                        if(game_profile.addon_name.Match(str))
                           eb->SetCurrEnum(i);
                     }
                  }
                              //check if local add-on is being available
                  {
                     const char *mname = tab_campaigns->ItemS(TAB_S12_CM_MISS_DIR, NUM_MISSIONS, 12);
                     C_fstring fname("Missions\\%s\\Game\\scene.bin", mname);
                     int h = dtaOpen(fname);
                     if(h!=-1){
                        dtaClose(h);
                        i = mmnu->addon_names.size();
                        mmnu->addon_names.push_back(".");
                        eb->SetupEnum(i, "<local test>");
                        if(game_profile.addon_name.Match("."))
                           eb->SetCurrEnum(i);
                     }
                  }
               }
               break;

            case IDMM_LOAD_GAME:
               {
                  if(net)
                     return false;
                              //check if we have currently saved a game
                  if(!game_profile.mission_states[game_profile.MissionIndex()].save_game.IsOpen())
                     return false;
               }
               break;

            case IDMM_SAVE_GAME:
               if(net)
                  return false;
               break;

            default:
               if(id>=IDMM_SESSION_0 && id<=IDMM_SESSION_LAST){
                              //session text buttons
                  assert(ctrl->GetType() == MC_TEXT_BUTTON);
                  int indx = id - IDMM_SESSION_0;
                              //check if button may be mapped to a session
                  if(indx >= mmnu->enum_sessions.size())
                     return false;
                  C_menu_control_text_button *tb = (C_menu_control_text_button*)ctrl;
                  PINET_session_address addr = mmnu->enum_sessions[indx];
                  tb->BuildText(C_fstring("%s (%i)", addr->GetName(), addr->GetNumPlayers()));
                  tb->SetUserData(addr->GetNumPlayers());
               }else
               if(id>=IDMM_NETPLR_0 && id<=IDMM_NETPLR_LAST){
                              //connecting players
                  assert(ctrl->GetType() == MC_TEXT);
                  int indx = id - IDMM_NETPLR_0;
                              //check if there's a player for this text
                  if(indx >= mmnu->enum_players.size())
                     return false;
                              //get name of the player
                  char buf[256];
                  if(!net->GetPlayerName(mmnu->enum_players[indx], buf, sizeof(buf)))
                     return false;

                  C_menu_control_text *tc = (C_menu_control_text*)ctrl;
                  tc->BuildText(buf);
               }
            }
         }
         break;

      case MCM_CTRL_CLICKED:
         {
            C_tick_class_main_menu *mmnu = (C_tick_class_main_menu*)menu;
            switch(id){
            case IDMM_NET_BEGIN:
               {
                              //user decided to begin
#ifndef DEBUG_NET_ALLOW_HOST_SINGLE
                  if(!mmnu->enum_players.size())
                     return false;
#endif
                  mmnu->NetBegin();
               }
               break;

            case IDMM_AIM_LINE:
               {
                  assert(ctrl->GetType() == MC_CHECKBOX);
                  C_menu_control_checkbox *cb = (C_menu_control_checkbox*)ctrl;
                  game_configuration.draw_aim_line = cb->IsChecked();
                  game_configuration.WriteSysConfig();
               }
               break;

            case IDMM_AIM_TYPE:
               {
                  assert(ctrl->GetType() == MC_ENUMBOX);
                  C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
                  game_configuration.aim_mode = eb->GetCurrEnum();
                  game_configuration.WriteSysConfig();
                  if(mmnu->saved_tick_class && mmnu->saved_tick_class->GetID()=='GMIS'){
                     C_game_mission *gmis = (C_game_mission*)((C_class_to_be_ticked_hd*)mmnu->saved_tick_class);
                     gmis->BroadcastMessage(CB_SETFOCUS, 2);
                  }
               }
               break;

            case IDMM_NET_VOICE_COMPRESSION:
               {
                  assert(ctrl->GetType() == MC_ENUMBOX);
                  C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
                  game_configuration.voice_chat_compression = (S_game_configuration::E_VOICE_COMPRESSION)eb->GetCurrEnum();
                  game_configuration.WriteSysConfig();
               }
               break;

            case IDMM_NET_CONNECTION_TYPE:
               {
                  assert(ctrl->GetType() == MC_ENUMBOX);
                  C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
                  game_configuration.net_ctype = (IN_CONNECTION)eb->GetEnumData(eb->GetCurrEnum());
                  game_configuration.WriteSysConfig();
               }
               break;

            case IDMM_NET_VOICECHAT:
               {
                  assert(ctrl->GetType() == MC_CHECKBOX);
                  C_menu_control_checkbox *cb = (C_menu_control_checkbox*)ctrl;
                  game_configuration.enable_voice_chat = cb->IsChecked();
                  game_configuration.WriteSysConfig();
               }
               break;

            default:
               if(id>=IDMM_SESSION_0 && id<=IDMM_SESSION_LAST){
                              //session text buttons
                  assert(ctrl->GetType() == MC_TEXT_BUTTON);
                  int indx = id - IDMM_SESSION_0;
                              //join selected session
                  if(!mmnu->JoinSession(mmnu->enum_sessions[indx]))
                     return false;
               }
            }
         }
         break;

      case MCM_CTRL_SLIDED:
         {
            C_menu_control_slider *cs = (C_menu_control_slider*)ctrl;
            int pos = cs->GetSlidePos();

            switch(id){
            case IDMM_SOUND_VOLUME_MUSIC:
               game_configuration.volume_music = pos;
               game_configuration.ApplySoundConfig();
               break;
            case IDMM_SOUND_VOLUME_EFFECTS:
               game_configuration.volume_effects = pos;
               game_configuration.ApplySoundConfig();
               break;
            default: assert(0);
            }
            game_configuration.WriteSysConfig();
         }
         break;

      case MCM_CTRL_EDIT_DONE:
         {
            assert(ctrl->GetType() == MC_EDITBOX);
            C_menu_control_editbox *eb = (C_menu_control_editbox*)ctrl;

            switch(id){
            case IDMM_NET_PLR_NAME:
               game_configuration.net_plr_name = eb->GetEditText();
               game_configuration.WriteSysConfig();
               break;

            case IDMM_NET_TCPIP_ADDRESS:
            case IDMM_NET_TCPIP_PORT:
            case IDMM_NET_PHONE_NUMBER:
               switch(id){
               case IDMM_NET_TCPIP_ADDRESS: game_configuration.net_tcpip_addr = eb->GetEditText(); break;
               case IDMM_NET_TCPIP_PORT: game_configuration.net_tcpip_port = eb->GetEditText(); break;
               case IDMM_NET_PHONE_NUMBER: game_configuration.net_modem_num = eb->GetEditText(); break;
               default: assert(0);
               }
               game_configuration.WriteSysConfig();
               break;
            }
         }
         break;
      }
      return true;
   }

//----------------------------

   bool HostBegin(){
                              //connect to network
      if(NetConnect()){
                              //host session now
         if(net->HostSession(INHOST_INIT_VOICE_CHAT)){
                              //hosting ok, go to menu where we wait for players
            InitVoiceChat();
            state = STATE_NET_HOST_WAIT;
            return true;
         }
         NetQuitConnecting(20004);
      }
      NetQuitConnecting(20003);
      return false;
   }

//----------------------------

   bool JoinBegin(){
                        //connect to network
      if(NetConnect()){
         state = STATE_NET_SESS_ENUM;
         return true;
      }
      NetQuitConnecting(20003);
      return false;
   }

//----------------------------

   bool JoinSession(PINET_session_address addr){

      if(net->JoinSession(addr)){
         state = STATE_NET_SESS_JOIN;
         enum_sessions.clear();
         InitLayout(templ_net_game, &layout_net_wait_players);

         InitVoiceChat();
         return true;
      }
      NetQuitConnecting(20006, true);
      return false;
   }

//----------------------------

   bool InitVoiceChat(){

      assert(net);
      if(game_configuration.enable_voice_chat){
         return net->InitVoice(igraph->GetHWND(), isound ? isound->GetDS() : NULL);
      }
      return false;
   }

//----------------------------

   void NetBegin(){
                  //close session, noone is allowed to join later
      net->SetOpen(false);

#ifdef _DEBUG
      if(game_configuration.net_host &&
         game_configuration.net_cmdline_mission.Size()){

         const C_string &mission_name = game_configuration.net_cmdline_mission;
         dword mcrc;
         bool b = net_mission->GetMissionCRC(mission_name, mcrc);
         if(b){
                              //send info to other party, so that they may start loading
                              //send network code with mission name
            net_msg->Ascend(NM_BEGIN_MISSION);
            for(int i=0; i<mission_name.Size()+1; i++){
               net_msg->AddByte(mission_name[i], true);
            }
            net_msg->AddDword(mcrc, true);
            net_msg->Descend();
            NetSend();

            C_game_mission *mission = CreateGameMission();
            E_MISSION_IO mio = mission->Open(mission_name, 0);
            if(mio==MIO_OK){
               mission->GameBegin(true, true);

               tick_class = mission;
            }
            mission->Release();
         }
      }else
#endif
      {
                              //proceed to menu browser
         NET_LOG("User wants Begin, sending game profile, begin code and loading campaign menu.");

         game_profile.curr_camp = 0;
         if(net_host_addon_pack!=-1){
            assert(net_host_addon_pack<addon_names.size());
            const C_string &addon_name = addon_names[net_host_addon_pack];
            if(game_profile.OpenAddOn(addon_name)){
               game_profile.curr_camp = 9;
               game_profile.curr_mission = 0;
            }else{
               return;
            }
         }

         game_profile.NetSend();
         NetSend();

                              //move on, create menu
         tick_class = CreateMenu((net_host_addon_pack==-1) ? GAME_MENU_CAMPAIGN_SELECT : GAME_MENU_MISSION_SELECT);
         if(tick_class)
            tick_class->Release();
      }
   }

//----------------------------

   virtual void OnSessionLost(){

      net_mission = NULL;
      enum_sessions.clear();
      enum_players.clear();
      state = STATE_MENU;

      C_menu_base::OnSessionLost();
   }

private:

//----------------------------

   virtual bool OnClick(C_menu_control_base *ctrl){

      int id = ctrl->GetID();

      switch(id){

      case IDMM_CREDITS:
         {
            state = STATE_SCROLLTEXT;
            scroll_text = CreateScrolltext(texts, NULL, NULL, false, .09f);
            scroll_text->Release();
         }
         break;

      case IDMM_SINGLE_PLAYER:
         RenderAllBackbuffers();
         game_profile.curr_camp = 0;
         tick_class = CreateMenu(GAME_MENU_CAMPAIGN_SELECT);
         if(tick_class)
            tick_class->Release();
         break;

      case IDMM_ADDON_SINGLE:
         {
            C_menu_control_base *ctrl = GetControl(IDMM_ADDON_PACK);
            assert(ctrl && ctrl->GetType() == MC_ENUMBOX);
            C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
            int enum_id = eb->GetCurrEnum();
            if(enum_id==-1)
               break;
            assert(enum_id<addon_names.size());
            const C_string &addon_name = addon_names[enum_id];

            if(game_profile.OpenAddOn(addon_name)){
               game_profile.curr_camp = 9;
               game_profile.curr_mission = 0;
               game_profile.Save();
               RenderAllBackbuffers();
               tick_class = CreateMenu(GAME_MENU_MISSION_SELECT);
               if(tick_class)
                  tick_class->Release();
            }
         }
         break;

      case IDMM_ADDON_NET_HOST:
         {
            C_menu_control_base *ctrl = GetControl(IDMM_ADDON_PACK);
            assert(ctrl && ctrl->GetType() == MC_ENUMBOX);
            C_menu_control_enumbox *eb = (C_menu_control_enumbox*)ctrl;
            net_host_addon_pack = eb->GetCurrEnum();
            if(net_host_addon_pack==-1)
               return false;
            if(!HostBegin())
               return false;
         }
         break;

      case IDMM_NET_HOST:
         if(!HostBegin())
            return false;
         net_host_addon_pack = -1;
         break;

      case IDMM_NET_JOIN:
         if(!JoinBegin())
            return false;
         break;

      case IDMM_TO_MAIN_MENU:
      case IDMM_RESTART_MISSION:
         {
            assert(saved_tick_class && saved_tick_class->GetID()=='GMIS');
            C_game_mission *mission = (C_game_mission*)(C_class_to_be_ticked_hd*)saved_tick_class;
            tick_class = saved_tick_class;
            mission->SetGameState(id==IDMM_TO_MAIN_MENU ? MS_EXIT : MS_DONE_FAIL, true);
         }
         break;

      case IDMM_QUIT:
         tick_class = NULL;
         break;

      case IDMM_RESUME_GAME:
         {
            assert(saved_tick_class);
                              //Note: must pass it through temp variable,
                              // because ours reference is destroyed during assignment
            tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(saved_tick_class);
         }
         break;

      case IDMM_LOAD_GAME:
         {
            RenderAllBackbuffers();
            assert(!net);
            assert(saved_tick_class && saved_tick_class->GetID()=='GMIS');
            C_game_mission *mission = (C_game_mission*)(C_class_to_be_ticked_hd*)saved_tick_class;
            bool ok = mission->LoadGame();
            if(!ok){
               tick_class = CreateMainMenu();
               if(tick_class)
                  tick_class->Release();
            }else{
               tick_class = mission;
            }
         }
         break;

      case IDMM_SAVE_GAME:
         {
            assert(!net);
            assert(saved_tick_class && saved_tick_class->GetID()=='GMIS');
            C_game_mission *mission = (C_game_mission*)(C_class_to_be_ticked_hd*)saved_tick_class;
            mission->SaveGame();
            tick_class = C_smart_ptr<C_class_to_be_ticked_hd>(saved_tick_class);
         }
         break;

      case IDMM_NET_VOICE_TEST:
         {
            INET_INIT ni;
            memset(&ni, 0, sizeof(ni));
            PI_net net;
#ifdef _DEBUG
            net = InetCreate(&ni, true);
#else
            net = InetCreate(&ni, false);
#endif
            if(net){
               void *hwnd = igraph->GetHWND();
               if(igraph->GetFlags()&IG_FULLSCREEN){
                  OsShowWindow(igraph->GetHWND(), false);
                  hwnd = NULL;
               }
               net->TestVoice(hwnd, true);
               if(igraph->GetFlags()&IG_FULLSCREEN){
                  OsShowWindow(igraph->GetHWND(), true);
                  igraph->ProcessWinMessages();
               }
               net->Release();
            }
         }
         break;
      }
      return true;
   }

//----------------------------

   virtual bool ReturnBackInHistory(){

      bool ok = C_menu_base::ReturnBackInHistory();
      if(ok){
         switch(state){
         case STATE_SCROLLTEXT:
            scroll_text = NULL;
            state = STATE_MENU;
            break;

         case STATE_NET_SYNC_WAIT:
         case STATE_NET_SESS_JOIN:
            NetQuitConnecting(20002, true);
            break;

         case STATE_NET_HOST_WAIT:
         case STATE_NET_SESS_ENUM:
            NetQuitConnecting(20002);
            break;
         }
      }
      return ok;
   }

//----------------------------

   void NetQuitConnecting(int id_reason = 0, bool return_back = false){

      if(id_reason)
         console_main.OutText(all_txt[id_reason], -1, all_txt.IsWide(id_reason));

      net_mission = NULL;
      enum_sessions.clear();
      enum_players.clear();
      NetCloseAll();

      state = STATE_MENU;
      if(return_back)
         ReturnBackInHistory();
   }

//----------------------------
public:
   C_tick_class_main_menu(C_class_to_be_ticked_hd *save_tc, E_MENU_ID_MAINMENU start_menu_item = IDMM_NULL):
      C_menu_base(save_tc),
      sess_enum_count(0),
      net_host_addon_pack(-1),
      state(STATE_MENU)
   {
   }

//----------------------------

   virtual void Tick(const S_tick_context &tc){

      AddRef();
      C_menu_base::Tick(tc);

      switch(state){

      case STATE_MENU:
         if(menu_controller->Get(IS_ESCAPE)){
                              //go to Quit or Resume control
            if(ActivateControl(IDMM_QUIT_LINK) ||
               ActivateControl(IDMM_RESUME_GAME)){

               menu_controller->IgnoreValue(IS_ESCAPE);
               PlaySound(MS_MOVE);
            }
         }
         break;

      case STATE_SCROLLTEXT:
         {
                              //display scrolltext until it's not done
            if(scroll_text->IsEnd()){
               ReturnBackInHistory();
               break;
            }
            scroll_text->Tick(tc);
         }
         break;

      case STATE_NET_SESS_ENUM:
         {
            if((sess_enum_count-=tc.time) <= 0){
               sess_enum_count = 200;
                              //try to enum sessions
               dword num_sess = 0;
               C_smart_ptr<INET_session_address> *addys = NULL;
               if(!net->EnumSessions(NULL, &num_sess)){
                  do{
                     delete[] addys;
                     addys = new C_smart_ptr<INET_session_address>[num_sess];
                  } while(!net->EnumSessions((PINET_session_address*)addys, &num_sess) && num_sess);
                              //sort addresses
                  sort(addys, addys+num_sess, cbSessionLess);
               }
                              //check if they differs from what we already have
               bool match = false;
               if(enum_sessions.size()==num_sess){
                  match = true;
                  for(int i=num_sess; i--; ){
                     C_menu_control_base *ctrl = GetControl(IDMM_SESSION_0+i);
                     assert(ctrl);
                     if(strcmp(enum_sessions[i]->GetName(), addys[i]->GetName()) ||
                        ctrl->GetUserData() != addys[i]->GetNumPlayers()){
                        match = false;
                        break;
                     }
                  }
               }
               if(!match){
                              //re-init layout, keep the same selection (if possible)
                  bool was_empty = (enum_sessions.size()==0);
                  PINET_session_address curr_addy = NULL;
                  C_menu_control_base *ctrl = GetActiveControl();
                  if(ctrl && ctrl->GetType()==MC_TEXT_BUTTON){
                     int indx = ctrl->GetID();
                     if(indx>=IDMM_SESSION_0 && indx<=IDMM_SESSION_LAST){
                        indx -= IDMM_SESSION_0;
                        assert(indx<enum_sessions.size());
                        curr_addy = enum_sessions[indx];
                     }
                  }
                  enum_sessions.clear();
                  enum_sessions.insert(enum_sessions.end(), addys, addys+num_sess);

                  InitLayout(templ_net_game, &layout_net_join);

                  if(curr_addy){
                              //try to find index of address
                     for(int i=enum_sessions.size(); i--; ){
                        if(enum_sessions[i]==curr_addy){
                           ActivateControl(IDMM_SESSION_0+i);
                           break;
                        }
                     }
                              //if session is gone, activate no control
                     if(i==-1)
                        SetActiveControl(-1);
                  }else{
                  if(!was_empty)
                              //activate no control
                     ActivateControl(0);
                  }
               }
               delete[] addys;
                              //auto-join 1st session
               if(game_configuration.net_join && enum_sessions.size())
                  JoinSession(enum_sessions.front());
            }
         }
         break;

      case STATE_NET_SESS_JOIN:
         {
                              //wait until not joined
            E_NET_STATUS nst = net->GetStatus();
            switch(nst){
            case INET_ST_CONNECTING:
               break;
            case INET_ST_CONNECTED:
               {
                  state = STATE_NET_SYNC_WAIT;
                  NET_LOG("Waiting for sync starting...");
               }
               break;
            default:
               ReturnBackInHistory();
            }
         }
         break;

      case STATE_NET_HOST_WAIT:
      case STATE_NET_SYNC_WAIT:
         {
                              //check status (if still connected)
            E_NET_STATUS nst = net->GetStatus();
            if(nst!=INET_ST_CONNECTED){
               NetQuitConnecting(20007, true);
               ReturnBackInHistory();
               break;
            }
                              //periodically enum players and feed list
            dword num_plrs = 0;
            dword *plrs = NULL;
            if(!net->GetPlayers(NULL, &num_plrs)){
               do{
                  delete[] plrs;
                  plrs = new dword[num_plrs];
               } while(!net->GetPlayers(plrs, &num_plrs));
                              //sort players
               sort(plrs, plrs+num_plrs);
            }
                              //check if they differs from what we already have
            bool match = false;
            if(enum_players.size()==num_plrs){
               for(int i=num_plrs; i--; ){
                  if(enum_players[i] != plrs[i])
                     break;
               }
               match = (i==-1);
            }
            if(!match){
                              //re-init layout
               bool was_empty = (enum_players.size()==0);
               C_menu_control_base *ctrl = GetActiveControl();
               int curr_ctrl = -1;
               if(ctrl && ctrl->GetType()==MC_TEXT_BUTTON){
                  curr_ctrl = ctrl->GetID();
               }
               enum_players.clear();
               enum_players.insert(enum_players.end(), plrs, plrs+num_plrs);

               InitLayout(templ_net_game, &layout_net_wait_players);

               if(curr_ctrl!=-1){
                  ActivateControl(curr_ctrl);
               }else
               if(!was_empty){
                  SetActiveControl(-1);
               }
            }
            delete[] plrs;
#ifdef _DEBUG
                              //auto-begin game when anybody connects
            if(state==STATE_NET_HOST_WAIT && game_configuration.net_host &&
               enum_players.size()==game_configuration.net_debug_numplayers){
               NetBegin();
               Release();
               return;
            }
#endif
         }
         break;
      default:
         assert(0);
      }

      /*
#ifndef BETAVERSION
      if(tick_class != this && tick_class){
                        //since we process this in Tick, update the loaded mission
         tick_class->Tick(tc);
         //if(music) music->Close();
      }
#endif
      */
      Release();
   }

//----------------------------

   virtual void NetIn(enum E_NETWORK_MESSAGE msg_id, class C_net_in_msg *in_msg){

      switch(msg_id){
      case NM_GAME_PROFILE:
                              //host's game profile arrived to us,
                              // this is also signal to move on to menu
         game_profile.NetImport(in_msg);

                              //move on, create menu
         if(!game_profile.IsAddOn()){
            RenderAllBackbuffers();
            tick_class = CreateMenu(GAME_MENU_CAMPAIGN_SELECT);
            if(tick_class)
               tick_class->Release();
         }else{
            int id_err = 0;
            dword save_crc = game_profile.addon_crc;
            if(game_profile.OpenAddOn(game_profile.addon_name)){
               if(save_crc==game_profile.addon_crc){
                  tick_class = CreateMenu(GAME_MENU_MISSION_SELECT);
                  if(tick_class)
                     tick_class->Release();
               }else{
                  id_err = 20404;
               }
            }else{
               id_err = 20403;
            }
            if(id_err){
               game_profile.ResetAddOn();
               NetQuitConnecting(id_err, true);
            }
         }
         return;

      case NM_BEGIN_MISSION:
         {
            in_msg->Ascend();
            NET_LOG("received mission info from host.");
                              //get mission name
            char mission_name[256];
            in_msg->GetString(mission_name, sizeof(mission_name));
                              //get mission crc
            dword mcrc = in_msg->GetDword();
            NET_LOG(C_fstring("mission name: '%s', crc: 0x%x", mission_name, mcrc));
            //E_MISSION_MODE game_mode = (E_MISSION_MODE)in_msg->GetWord();
            //net_mission->SetLevelTime(in_msg->GetDword() - igraph->ReadTimer());
            in_msg->Descend();

                              //check if we have got such mission
            dword mcrc1;
            bool b = net_mission->GetMissionCRC(mission_name, mcrc1);
            if(b && mcrc == mcrc1){
               NET_LOG("Loading scene...");

               C_game_mission *mission = CreateGameMission();
               if(mission){
                  E_MISSION_IO mio = mission->Open(mission_name, 0);
                  NET_LOG(C_fstring("Scene load %s", (mio==MIO_OK) ? "OK" : "failed"));
                  if(mio==MIO_OK){
                     mission->GameBegin(true, true);
                     tick_class = mission;
                  }else{
                     NET_LOG("Failed to load mission.");
                     NetQuitConnecting(20008, true);
                     ReturnBackInHistory();
                  }
                  mission->Release();
               }else{
                  NetQuitConnecting(20008, true);
                  ReturnBackInHistory();
               }
            }else{
                              //mission not found, or crc mismatch
               NET_LOG("Mission mismatch (not found, or CRC not same)...");
               NetQuitConnecting(20008, true);
               ReturnBackInHistory();
            }
         }
         break;

      case NM_GAME:
                              //route to game, if saved tick class is game
         if(saved_tick_class && saved_tick_class->GetID()=='GMIS'){
            saved_tick_class->NetIn(msg_id, in_msg);
            break;
         }
                              //flow...
      default:
         //assert(0);
         in_msg->SkipChunk();
      }
   }
};

//----------------------------

                              //network game
static const S_menu_layout_template layout_net_wait_players = {
   {315},
   {
      {MC_TEXT, IDMM_NETPLR_0+0, 0, .04f, MCT_CALLBACK | MCT_SIZE_2, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT, IDMM_NETPLR_0+1, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT, IDMM_NETPLR_0+2, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT, IDMM_NETPLR_0+3, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT, IDMM_NETPLR_0+4, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT, IDMM_NETPLR_0+5, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT, IDMM_NETPLR_0+6, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT, IDMM_NETPLR_0+7, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_NET_BEGIN, 0, -.07f, MCT_SIZE_4 | MCT_POS_Y_BOTTOM | MCT_CALLBACK, 316, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_NULL}
   }
}, layout_net_join = {
   {312},
   {
      {MC_TEXT, 0, 0, -.01f, MCT_SIZE_2, 20001, NULL},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+0, 0, .01f, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+1, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+2, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+3, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+4, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+5, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+6, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SESSION_0+7, 0, 0, MCT_CALLBACK, 0, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_NULL}
   }
}, layout_net_config = {
   {285},
   {
                              //connection type
      {MC_ENUMBOX, IDMM_NET_CONNECTION_TYPE, 0, 0, MCT_CALLBACK, 286, (dword)C_tick_class_main_menu::cbCtrl},

      {MC_TEXT, 0, .01f, .01f, MCT_POS_X_LEFT | MCT_JUSTIFY_LEFT | MCT_SIZE_2, 290, NULL},
      {MC_EDITBOX, IDMM_NET_TCPIP_ADDRESS, 0, 0, MCT_CALLBACK | MCT_POS_X_CENTER | MCT_JUSTIFY_CENTER | MCT_SIZE_4, 291, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_EDITBOX, IDMM_NET_TCPIP_PORT, 0, 0, MCT_CALLBACK, 292, (dword)C_tick_class_main_menu::cbCtrl},
#ifdef ALLOW_MODEM
      {MC_TEXT, 0, .01f, .01f, MCT_POS_X_LEFT | MCT_JUSTIFY_LEFT | MCT_SIZE_2, 295, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_EDITBOX, IDMM_NET_PHONE_NUMBER, .03f, 0, MCT_CALLBACK | MCT_POS_X_CENTER | MCT_JUSTIFY_CENTER | MCT_SIZE_4, 297, (dword)C_tick_class_main_menu::cbCtrl},
#endif//ALLOW_MODEM
      {MC_CHECKBOX, IDMM_NET_VOICECHAT, .01f, .01f, MCT_CALLBACK, 20100, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_ENUMBOX, IDMM_NET_VOICE_COMPRESSION, 0, 0, MCT_CALLBACK, 20101, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_NET_VOICE_TEST, -.01f, 0, 0, 20102, 0},
      {MC_NULL}
   }
}, layout_net_game = {
   {201},
   {
      {MC_TEXT, 0, 0, -.01f, MCT_SIZE_2, 310, NULL},
      {MC_TEXT_BUTTON, IDMM_NET_HOST, 0, .02f, MCT_SIZE_4, 311, (dword)&layout_net_wait_players},
      {MC_TEXT_BUTTON, IDMM_NET_JOIN, 0, 0, 0, 312, (dword)&layout_net_join},
      {MC_EDITBOX, IDMM_NET_PLR_NAME, 0, -.12f, MCT_SIZE_2 | MCT_CALLBACK | MCT_POS_Y_BOTTOM, 288, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, 0, 0, -.07f, MCT_POS_Y_BOTTOM, 285, (dword)&layout_net_config},
      {MC_NULL}
   }
}, layout_add_on = {
   {20221},
   {
      {MC_ENUMBOX, IDMM_ADDON_PACK, 0, 0, MCT_CALLBACK, 20222, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_ADDON_SINGLE, 0, 0, MCT_DEFAULT_ITEM, 200, NULL},
      {MC_TEXT_BUTTON, IDMM_ADDON_NET_HOST, 0, 0, MCT_MENU_LINK, 311, (dword)&templ_addon_net_host},
      {MC_NULL}
   }
};

static const S_menu_template templ_net_game = {
   { .18f, .15f, .64f, .45f, 1200.0f }, &layout_net_game
};

//----------------------------

static const S_menu_layout_template
layout_opt_game = {
   {20220},
   {
      {MC_CHECKBOX, IDMM_AIM_LINE, .02f, 0, MCT_CALLBACK | MCT_SIZE_1, 20600, (dword)&C_tick_class_main_menu::cbCtrl},
      {MC_ENUMBOX, IDMM_AIM_TYPE, 0, 0, MCT_CALLBACK, 20601, (dword)&C_tick_class_main_menu::cbCtrl},

      {MC_NULL}
   }
}, layout_opt_sound = {
   {235},
   {
      {MC_TEXT, 0, 0, 0, MCT_SIZE_2, 236},
      {MC_SLIDER, IDMM_SOUND_VOLUME_MUSIC, -.15f, 0, MCT_SIZE_4 | MCT_CALLBACK, 0, (dword)&C_tick_class_main_menu::cbCtrl, 0},

      {MC_TEXT, 0, .15f, .02f, MCT_SIZE_2, 237},
      {MC_SLIDER, IDMM_SOUND_VOLUME_EFFECTS, -.15f, 0, MCT_SIZE_4 | MCT_CALLBACK, 0, (dword)&C_tick_class_main_menu::cbCtrl, 0},
      {MC_NULL}
   }
};

//----------------------------
                              //in-game layout and menu
static const S_menu_layout_template templ_ingame_to_menu = {
   {810, 0, -.1f},
   {
      {MC_TEXT, 0, 0, .02f, MCT_SIZE_2 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 215},
      {MC_TEXT_BUTTON, IDMM_TO_MAIN_MENU, -.06f, .09f, MCT_DEFAULT_ITEM | MCT_SIZE_4 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 216},
      {MC_TEXT_BUTTON, 0, +.06f, .09f, MCT_POS_X_CENTER | MCT_POS_Y_TOP, 217, 1},
      {MC_NULL}               //terminator
   }
}, templ_ingame_restart = {
   {810, 0, -.1f},
   {
      {MC_TEXT, 0, 0, .02f, MCT_SIZE_2 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 20231},
      {MC_TEXT_BUTTON, IDMM_RESTART_MISSION, -.06f, .09f, MCT_DEFAULT_ITEM | MCT_SIZE_4 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 216},
      {MC_TEXT_BUTTON, 0, +.06f, .09f, MCT_POS_X_CENTER | MCT_POS_Y_TOP, 217, 1},
      {MC_NULL}               //terminator
   }
}, templ_ingame_layout = {
   {810, MLT_CLEAR_HISTORY, -.1f},
   {
      {MC_TEXT_BUTTON, IDMM_RESUME_GAME, 0, 0, 0, 811, NULL},
      {MC_TEXT_BUTTON, IDMM_LOAD_GAME, 0, 0, MCT_CALLBACK, 812, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, IDMM_SAVE_GAME, 0, 0, MCT_CALLBACK, 813, (dword)C_tick_class_main_menu::cbCtrl},
      {MC_TEXT_BUTTON, 0, 0, 0, 0, 202, (dword)&layout_options},
      {MC_TEXT_BUTTON, 0, 0, 0, 0, 20230, (dword)&templ_ingame_restart},
      {MC_TEXT_BUTTON, 0, 0, 0, 0, 814, (dword)&templ_ingame_to_menu},

      {MC_NULL}               //terminator
   }
};

static const S_menu_template templ_ingame_menu = {
   {
      .5f - .225f, .375f - .17f, .45f, .34f, 1200.0f,
      MT_PLAY_SOUND
   },
   &templ_ingame_layout
};

//----------------------------
//----------------------------

extern const S_menu_template templ_log_on, templ_logon_quit;
extern const S_menu_layout_template layout_new_profile, layout_del_profile, layout_logon_quit;

class C_menu_log_on: public C_menu_base{
   vector<S_logon_account> accounts;
   int deleted_account;

//----------------------------

   virtual bool ReturnBackInHistory(){

      if(history.size()!=1)
         return C_menu_base::ReturnBackInHistory();
      InitLayout(templ_logon_quit, &layout_logon_quit);
      return true;
   }

//----------------------------

   virtual void ProcessShortcuts(const S_tick_context &tc){

      if(tc.key==K_DEL){
         C_menu_control_base *ctrl = GetActiveControl();
         if(ctrl && ctrl->GetType()==MC_TEXT_BUTTON){
            int id = ctrl->GetID();
            if(id>=0 && id<8){
                              //give warning before deletion
               deleted_account = id;
               InitLayout(templ_log_on, &layout_del_profile);
            }
         }
         return;
      }
      C_menu_base::ProcessShortcuts(tc);
   }

public:

   static bool cbCtrl(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu,
      C_menu_control_base *ctrl, void *context){

      switch(msg){
      case MCM_CTRL_CREATED:
         switch(ctrl->GetType()){
         case MC_TEXT_BUTTON:
            {
               int id = ctrl->GetID();
               vector<S_logon_account> &accounts = *(vector<S_logon_account>*)context;

               switch(id){
               case 1010:
               case 1011:
               case 1012:
               case 1002:
               case 2000:
                  return true;
               case 1000:
                              //allow creation button only if there's space
                  return (accounts.size() < 8);
               }
               if(id >= accounts.size())
                  return false;
                           //set name for this control
               const C_string &user_name = accounts[id].user_name;
               int diff_tid = GT_MENU_PROFILE_DIFFICULTY+accounts[id].difficulty;
               const C_string &diff = all_txt[diff_tid];

               C_string str;

               bool wide = all_txt.IsWide(diff_tid);
               if(!wide){
                  str = C_fstring("\"%s\"   %s", (const char*)user_name, (const char*)diff);
               }else{
                  str = MakeGameText(C_fstring("\"%s\"   ", (const char*)user_name));
                  str = AddGameText(str, diff);
               }
               ctrl->BuildText(str, wide);
            }
            break;

         case MC_EDITBOX:
            {
                              //automatically switch to edit mode
               C_menu_control_editbox *eb = (C_menu_control_editbox*)ctrl;
               eb->SetMaxEditLen(16);
               eb->ToggleEditMode();
            }
            break;
         }
         break;

      case MCM_CTRL_CLICKED:
         switch(ctrl->GetType()){
         case MC_TEXT_BUTTON:
            {
                              //game profile selected, move on!
               int id = ctrl->GetID();
               switch(id){
               case 1010:
               case 1011:
               case 1012:
                  {
                              //new profile created
                     C_menu_control_base *ctrl_eb = ((C_menu_base*)menu)->GetControl(1001);
                     assert(ctrl_eb && ctrl_eb->GetType()==MC_EDITBOX);
                     C_menu_control_editbox *eb = (C_menu_control_editbox*)ctrl_eb;
                     const C_string &prof_name = eb->GetEditText();
                              //don't allow empty names
                     if(!prof_name.Size())
                        return false;
                     S_game_profile gp;
                     gp.name = prof_name;
                     gp.file_name = prof_name;
                     gp.difficulty = (S_game_profile::E_DIFFICULTY)(S_game_profile::DIF_EASY + id-1010);
                     OsCreateUniqueFileName("Savegame", "bin", gp.file_name);
                     gp.file_name = C_fstring("Savegame\\%s.bin", (const char*)gp.file_name);
                     gp.Save();
                              //re-create menu now
                     C_menu_log_on *m_logon = (C_menu_log_on*)menu;
                     SearchAccounts(m_logon->accounts);
                     m_logon->InitLayout(templ_log_on, templ_log_on.begin_layout);
                  }
                  break;

               case 1000:
                  {
                     C_menu_log_on *m_logon = (C_menu_log_on*)menu;
                     m_logon->InitLayout(templ_log_on, &layout_new_profile);
                  }
                  break;

               case 1002:
                  {
                              //delete profile confirmed
                     C_menu_log_on *m_logon = (C_menu_log_on*)menu;
                     assert(m_logon->deleted_account!=-1);

                     const char *fname = m_logon->accounts[m_logon->deleted_account].file_name;
                     m_logon->deleted_account = -1;
                     bool b = OsDeleteFile(fname);
                     if(!b)
                        return false;
                              //re-create menu now
                     SearchAccounts(m_logon->accounts);
                     m_logon->InitLayout(templ_log_on, templ_log_on.begin_layout);
                  }
                  break;

               case 2000:
                  tick_class = NULL;
                  break;

               default:
                  {
                     vector<S_logon_account> &accounts = *(vector<S_logon_account>*)context;
                     assert(id < accounts.size());
                     bool b = game_profile.Open(accounts[id].file_name);
                     assert(b);
                     if(!b)
                        return false;
                              //save now, so that we have set latest access date
                     game_profile.Save();
                     RenderAllBackbuffers();
                     tick_class = CreateMainMenu();
                     if(tick_class)
                        tick_class->Release();
                  }
               }
            }
            break;
         }
         break;
      }

      return true;
   }

//----------------------------

   C_menu_log_on():
      deleted_account(-1)
   {
      SearchAccounts(accounts);
      context = &accounts;

      InitMenu(templ_log_on);
      saved_tick_class = CreateBackgroundMission("Log_on");
      saved_tick_class->Release();
   }
};

//----------------------------

                              //logon menu
static const S_menu_layout_template layout_new_profile = {
   {225},
   {
      {MC_EDITBOX, 1001, -.05f, .02f, MCT_CALLBACK | MCT_SIZE_2, 226, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT, 0, 0, .1f, MCT_POS_X_CENTER, 210},
      {MC_TEXT_BUTTON, 1010, 0, .02f, MCT_CALLBACK, 211, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 1011, 0, 0, MCT_CALLBACK, 212, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 1012, 0, 0, MCT_CALLBACK, 213, (dword)C_menu_log_on::cbCtrl},
      {MC_NULL}
   }
}, layout_del_profile = {
   {225},
   {
      {MC_TEXT, 0, 0, 0, 0, 120},
      {MC_TEXT_BUTTON, 1002, -.06f, MCT_DEFAUL_TEXT_SIZE*1.4f, MCT_CALLBACK | MCT_SIZE_3 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 121, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 0, .06f, MCT_DEFAUL_TEXT_SIZE*1.4f, MCT_POS_X_CENTER | MCT_POS_Y_TOP, 122, 1, 0},
      {MC_NULL}
   }
}, layout_logon_quit = {
   {0},
   {
      {MC_TEXT, 0, 0, MCT_DEFAUL_TEXT_SIZE*.3f, MCT_SIZE_3 | MCT_POS_X_CENTER | MCT_POS_Y_TOP, 215},
      {MC_TEXT_BUTTON, 2000, -.06f, -MCT_DEFAUL_TEXT_SIZE*2.4f, MCT_CALLBACK | MCT_DEFAULT_ITEM | MCT_SIZE_3 | MCT_POS_X_CENTER | MCT_POS_Y_BOTTOM, 216, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 0, .06f, -MCT_DEFAUL_TEXT_SIZE*2.4f, MCT_POS_X_CENTER | MCT_POS_Y_BOTTOM, 217, 1, 0},
      {MC_NULL}               //terminator
   }
}, layout_log_on = {
   {225, MLT_CLEAR_HISTORY | MLT_FORCE_BACK_BUTTON},
   {
      //{MC_TEXT, 0, 0, -.02f, MCT_SIZE_1 | MCT_POS_Y_TOP, 225},
      {MC_TEXT_BUTTON, 0, 0, .03f, MCT_CALLBACK | MCT_SIZE_2, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 1, 0, 0, MCT_CALLBACK, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 2, 0, 0, MCT_CALLBACK, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 3, 0, 0, MCT_CALLBACK, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 4, 0, 0, MCT_CALLBACK, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 5, 0, 0, MCT_CALLBACK, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 6, 0, 0, MCT_CALLBACK, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 7, 0, 0, MCT_CALLBACK, 0, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT_BUTTON, 1000, 0, 0, MCT_CALLBACK, 231, (dword)C_menu_log_on::cbCtrl},
      {MC_TEXT, 0, 0, -.03f, MCT_SIZE_2 | MCT_POS_Y_BOTTOM, 232},
      {MC_NULL}
   }
};

static const S_menu_template templ_log_on = {
   //{.25f, .1f, .5f, .46f, 1200.0f},
   {.2f, .1f, .6f, .46f, 1200.0f},
   &layout_log_on
}, templ_logon_quit = {
   {
      .3f, .2f, .4f, .25f, 1800.0f
   },
   &layout_logon_quit
};


//----------------------------
#pragma warning(disable:4189)

C_class_to_be_ticked_hd *CreateMainMenu(C_class_to_be_ticked_hd *save_tc){

   bool ingame = (save_tc);
   if(!ingame)
      game_profile.CloseAddOn();

   C_smart_ptr<C_mission> back_scene;
   if(!save_tc){
      const char *back_misison = "logo";
      if(game_configuration.net_host || game_configuration.net_join)
         back_misison = "log_on";

      back_scene = CreateBackgroundMission(back_misison);
      if(!back_scene)
         return NULL;
      if(!(game_configuration.net_host || game_configuration.net_join))
         game_music.Play("logo", true);
      save_tc = back_scene;
      back_scene->Release();
   }
   C_tick_class_main_menu *mp = new C_tick_class_main_menu(save_tc);
#ifdef DEBUG_TEST_NET
   mp->InitMenu(templ_net_game);
#elif defined DEBUG_CONFIG_NET
   mp->InitMenu(controller_menu_templ);
#else
   mp->InitMenu(ingame ? templ_ingame_menu : templ_main_menu);
   if(!ingame){
                              //auto host|join
      if(game_configuration.net_host){
         if(mp->HostBegin())
            mp->InitLayout(templ_net_game, &layout_net_wait_players);
      }else
      if(game_configuration.net_join){
         if(mp->JoinBegin())
            mp->InitLayout(templ_net_game, &layout_net_join);
      }
   }
#endif
   return mp;
}

//----------------------------

C_class_to_be_ticked_hd *CreateLogonMenu(){

   C_menu_log_on *mp = new C_menu_log_on;
   return mp;
}

//----------------------------
//----------------------------
extern const S_menu_template templ_sync_menu;

class C_sync_menu_imp: public C_sync_menu{
   C_menu_base menu_base;

                              //keep sorted net players
   vector<S_net_player> net_players;

   bool cbCtrl1(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu, C_menu_control_base *ctrl){

      int id = ctrl->GetID();
      if(id>=200 && id<250){
         assert(ctrl->GetType()==MC_CHECKBOX);
         C_menu_control_checkbox *cb = (C_menu_control_checkbox*)ctrl;
         int i = id-200;
         if(i>=net_players.size())
            return false;
         cb->BuildText(net_players[i].name, false);
      }
      return true;
   }
public:
   C_sync_menu_imp(int title_id, dword initiator_pid, bool check_our_plr){
      menu_base.SetContext(this);

      assert(net);
      int our_pid = net->GetPID();
      MakeSortedNetList(net_players, true);
      menu_base.InitMenu(templ_sync_menu);

      if(title_id){
         C_menu_control_base *ctrl = menu_base.GetControl(100);
         if(ctrl){
            assert(ctrl->GetType()==MC_TEXT);
            C_menu_control_text *ct = (C_menu_control_text*)ctrl;
            ct->BuildText(all_txt[title_id], all_txt.IsWide(title_id));
         }
      }
                              //mark our name and check it now
      for(int i=net_players.size(); i--; ){
         if(net_players[i].pid==our_pid){
            menu_base.ActivateControl(200+i);
            if(check_our_plr){
               C_menu_control_base *ctrl = menu_base.GetControl(200+i);
               ((C_menu_control_checkbox*)ctrl)->SetCheck(true);
            }
            break;
         }
      }
      //menu_base.SetActiveControl(-1);
   }

   virtual void Tick(const S_tick_context &tc){ menu_base.Tick(tc); }
   virtual void Render(){ menu_base.Render(); }
   virtual dword GetID() const{ return 'SMNU'; }

//----------------------------
   static bool cbCtrl(E_MENU_CALLBACK_MESSAGE msg, C_menu_base *menu, C_menu_control_base *ctrl, void *c){
      return ((C_sync_menu_imp*)c)->cbCtrl1(msg, menu, ctrl);
   }

//----------------------------

   virtual void AddCheckMarks(dword *pids_checked, int num_checked){

      for(int i=net_players.size(); i--; ){
         C_menu_control_base *ctrl = menu_base.GetControl(200+i);
         assert(ctrl && ctrl->GetType()==MC_CHECKBOX);
         C_menu_control_checkbox *cb = (C_menu_control_checkbox*)ctrl;

         bool is_checked = (find(pids_checked, pids_checked+num_checked, net_players[i].pid) != pids_checked+num_checked);
         if(is_checked)
            cb->SetCheck(true);
      }
   }

//----------------------------

   virtual void AddAllCheckMarks(){

      for(int i=net_players.size(); i--; ){
         C_menu_control_base *ctrl = menu_base.GetControl(200+i);
         assert(ctrl && ctrl->GetType()==MC_CHECKBOX);
         C_menu_control_checkbox *cb = (C_menu_control_checkbox*)ctrl;
         cb->SetCheck(true);
      }
   }
};

//----------------------------

                              //Net sync menu (display only)
static const S_menu_layout_template templ_sync = {
   {0, MLT_NO_HISTORY | MLT_NO_MOUSE_CURSOR | MLT_NO_BACK_BUTTON, -.5f},
   {
      {MC_TEXT, 100, 0, 0, 0, 0, NULL},
      {MC_TEXT, 101, .005f, 0, MCT_POS_X_LEFT | MCT_JUSTIFY_LEFT, 20301, NULL},

      {MC_CHECKBOX, 200, .035f, .008f, MCT_JUSTIFY_LEFT | MCT_CALLBACK | MCT_POS_X_LEFT, 0, (dword)C_sync_menu_imp::cbCtrl},
      {MC_CHECKBOX, 201, 0, 0, MCT_JUSTIFY_LEFT | MCT_CALLBACK, 0, (dword)C_sync_menu_imp::cbCtrl},
      {MC_CHECKBOX, 202, 0, 0, MCT_JUSTIFY_LEFT | MCT_CALLBACK, 0, (dword)C_sync_menu_imp::cbCtrl},
      {MC_CHECKBOX, 203, 0, 0, MCT_JUSTIFY_LEFT | MCT_CALLBACK, 0, (dword)C_sync_menu_imp::cbCtrl},

      {MC_NULL}               //terminator
   }
};

static const S_menu_template templ_sync_menu = {
   {
      .99f - .25f, .74f - .17f, .25f, .17f, 5000.0f, 0
   },
   &templ_sync
};

//----------------------------

C_sync_menu *CreateSyncMenu(int title_id, dword initiator_pid){

   return new C_sync_menu_imp(title_id, initiator_pid, true);
}

//----------------------------
//----------------------------

class C_net_leave_menu: public C_menu_base{
public:
   C_net_leave_menu(C_class_to_be_ticked_hd *save_tc):
      C_menu_base(save_tc)
   {
      InitLayout(templ_logon_quit);
   }

   virtual void NetIn(enum E_NETWORK_MESSAGE msg_id, class C_net_in_msg *in_msg){

      if(saved_tick_class){
         saved_tick_class->NetIn(msg_id, in_msg);
      }
   }

   virtual void Tick(const S_tick_context &tc){

      AddRef();
      C_menu_base::Tick(tc);

      if(menu_controller->Get(IS_ESCAPE)){
         tick_class = saved_tick_class;
      }
      Release();
   }

   virtual bool OnClick(C_menu_control_base *ctrl){

      switch(ctrl->GetID()){
      case 2000:
         NetCloseAll();
         tick_class = CreateMainMenu();
         if(tick_class)
            tick_class->Release();
         return false;
      }
      return true;
   }
};

//----------------------------

C_class_to_be_ticked_hd *CreateNetLeaveMenu(C_class_to_be_ticked_hd *save_tc){
   return new C_net_leave_menu(save_tc);
}

//----------------------------

#endif//!EDITOR

//----------------------------
//----------------------------
