#ifndef __MENU_H_
#define __MENU_H_

//----------------------------
                              //menu controls type identifier
enum E_MENU_CONTROL_TYPE{
   MC_NULL,
   MC_TEXT,                   //static non-selectable text
   MC_TEXT_BUTTON,            //text button
   MC_BUTTON,                 //sprite button
   MC_CHECKBOX,               //checkbox with description text
   MC_SLIDER,                 //horizontal slider with description text
   //MC_LISTBOX,
   MC_EDITBOX,                //editable box with description text
   MC_CONTROLLER_BOX,         //box displaying controller input
   MC_ENUMBOX,               //box displaying one of enumerated values
};

//----------------------------
                              //flags for menu layout template
enum{
   MLT_CLEAR_HISTORY = 0x0001,//clear all navigation history (no return back from this layout)
   MLT_NO_HISTORY = 0x0002,   //don't save navigation history for this layout
   MLT_NO_MOUSE_CURSOR = 0x0004, //no mouse cursor available in this layout
   MLT_NO_BACK_BUTTON = 0x0008,  //don't create back button, nor scan back keys
   MLT_FORCE_BACK_BUTTON = 0x0010,  //force create back button
};

//----------------------------
                              //flags for menu-controls templates
enum{
                              //X position is specified relative to menu's client area ...
   MCT_POS_X_CENTER = 0x0001, // ... center
   MCT_POS_X_LEFT = 0x0002,   // ... left border
   MCT_POS_X_RIGHT = 0x0004,  // ... right border
                              //Y position is specified relative to menu's client area ...
   MCT_POS_Y_TOP = 0x0008,    // ... top border
   MCT_POS_Y_CENTER = 0x0010, // ... center
   MCT_POS_Y_BOTTOM = 0x0020, // ... bottom border

   MCT_POS_NO_CLIENT = 0x0040,//the position of control is relative to menu rectangle, not to client-area rectangle
   MCT_DEFAULT_ITEM = 0x0080, //make this item the default one
   MCT_MENU_LINK = 0x0100,    //this control's link points to menu, rather than to layout ('menu_link_to')
   MCT_CALLBACK = 0x0800,     //when this control is clicked/updated, etc, callback is called ('callback')
                              //text justify is set to...
   MCT_JUSTIFY_CENTER = 0x0000,  // ... center (default)
   MCT_JUSTIFY_LEFT = 0x0200,    // ... left
   MCT_JUSTIFY_RIGHT = 0x0400,   // ... right

                              //size of element
   MCT_SIZE_1 = 0x10000,      //very small (64%)
   MCT_SIZE_2 = 0x20000,      //small (80%)
   MCT_SIZE_3 = 0x00000,      //default size
   MCT_SIZE_4 = 0x30000,      //big (125%)
   MCT_SIZE_5 = 0x40000,      //very bix (156%)

};

//----------------------------
                              //flags for menu template
enum{
   MT_PLAY_SOUND = 0x0001,    //play sound when menu appears
};

//----------------------------
                              //menu control callback message
enum E_MENU_CALLBACK_MESSAGE{
   MCM_NULL,
   MCM_CTRL_CREATED,          //control is just created
   MCM_CTRL_CLICKED,          //control is clicked/(un)checked
   MCM_CTRL_SLIDED,           //slider position changed
   MCM_CTRL_EDIT_BEGIN,       //editing in edit box begins
   MCM_CTRL_EDIT_DONE,        //editing in edit box is done
};

//----------------------------
                              //menu control template

                              //make compressed sprite indicies for 2-state sprite controls
#define MT_MAKE_SPR_INDEX(released, pressed) (released | (pressed << 16))

struct S_menu_control_template{
   E_MENU_CONTROL_TYPE ctrl;       //control identifier
   int control_id;            //unique identifier of control
   float pos_x, pos_y;        //position in normalized coordinates
   dword flags;               //combination of MTC_??? flags

   union{                     //type-dependent info
      union{
         int text_id;         //text index (MC_TEXT_BUTTON, MC_CHECKBOX, MC_SLIDER, MC_EDITBOX, MC_CONTROLLER_BOX)

                              //sprite indicies (MC_BUTTON)
         int sprite_index;
         struct{              
            word sprite_index_released, sprite_index_pressed;
         };
      };
   };
   union{
      dword context_data;     //control-dependent data
      //int slider_num_pos;
                              //where do we go, when clicked (MC_TEXT_BUTTON, MC_BUTTON)
      struct S_menu_layout_template *layout_link_to;
                              //the same as above, but if MCT_MENU_LINK flag specified
      struct S_menu_template *menu_link_to;
      bool (*callback)(E_MENU_CALLBACK_MESSAGE, class C_menu_base*, class C_menu_control_base*,
         void *context);
      int controller_slot;  //controller slow which we modify (E_CONTROLLER_SLOT)
   };
   int text_id_hint;          //hint for this control, 0 for no hint
};

//----------------------------
                              //menu layout template - single screen in a menu
struct S_menu_layout_template{
   struct{
      int menu_name_id;          //text index
      dword flags;               //combination of MT_??? flags
      float global_scale_delta;  //all controls scaled by (1.0f + global_scale_delta)
   };
#pragma warning(disable: 4200)//using non-standard extension
   S_menu_control_template lines[];
#pragma warning(default: 4200)
};

//----------------------------
                              //menu template
struct S_menu_template{
   struct{
      float left, top, size_x, size_y;
      float texel_size;          //size of original texel in menu sprite texture
      dword flags;
   };
   const S_menu_layout_template *begin_layout;  //fist layout page to display
};

//----------------------------
// Create menu from template. If 'save_tc' is non-NULL, it is saved in the menu class,
// and updated (ticked, rendered, etc) duruing the menu's lifetime. When 'Back' button
// is pressed, the 'save_tc' is made active 'tick_class'.
C_class_to_be_ticked_hd *CreateMenu(const S_menu_template&, C_class_to_be_ticked_hd *save_tc = NULL,
   void *context = NULL);

//----------------------------
// Create main menu. The initial layout is selected dependent on 'save_tc'.
// If it is !NULL, the game may be resumed.
C_class_to_be_ticked_hd *CreateMainMenu(C_class_to_be_ticked_hd *save_tc = NULL);

//----------------------------
// Create logon menu. The menu accepts player's name and logs on (creates main menu), or quits.
C_class_to_be_ticked_hd *CreateLogonMenu();

//----------------------------
//----------------------------
                              //small informational menu for network synchonizations
class C_sync_menu: public C_class_to_be_ticked_hd{
public:
//----------------------------
// Set checkmarks next to network players.
   virtual void AddCheckMarks(dword *pids_checked, int num_checked) = 0;
   virtual void AddAllCheckMarks() = 0;
};

C_sync_menu *CreateSyncMenu(int title_id, dword initiator_pid = 0);

//----------------------------
// Create net-leave menu for joining player, which enables him to leave host-controlled menu.
C_class_to_be_ticked_hd *CreateNetLeaveMenu(C_class_to_be_ticked_hd *save_tc);

//----------------------------
//----------------------------

#endif