#include "all.h"
#include "tables.h"
#include "actors.h"
#include <insanity\controller.h>
#include "mainmenu.h"
#include "net.h"

                              //map
#define PRG_HINT_SWITCH_COUNT  200
#define MAP_NUM_CTRL_ICONS 6
#define MAP_BUTTON_INACTIVE   0x80ffffff
#define MAP_ICON_INACTIVE     0xff808080
#define MAP_BUTTON_ACTIVE     0xffffffff

#define MAP_ZOOM_MIN 48.0f

#define MAP_PRG_SIZE .04275f
#define MAP_PRG_X(i) (MAP_PRG_SIZE*.7f)
#define MAP_PRG_Y(i) ((i)*MAP_PRG_SIZE + .0975f)

#define MAP_CAM_Y -1.5f

//----------------------------
//----------------------------

static int GetKeyWord(char* &line, const char *kwords){

   int i = 0;
   while(kwords[0]){
      int linelen=strlen(kwords);
      if(!strnicmp(line, kwords, linelen)){
         line += linelen;
         return i;
      }
      ++i;
      kwords += strlen(kwords) + 1;
   }
   return -1;
}

//----------------------------

static bool ReadInt(char* &line, int &i1){

   int i;
   if(sscanf(line, "%i%n", &i1, &i) == 1){
      line += i;
      return true;
   }
   return false;
}

//----------------------------

float Angle2D(const S_vector2 &v1, const S_vector2 &v2);
float Angle2D(const S_vector2 &v1);
bool I3DAPI WatchColResp2(I3D_collision_data&);

//----------------------------
//----------------------------

static const float map_img_pos[][3] = {  //x, y, size
   .30f, .037f, .045f,       //level
   .36f, .037f, .045f,
   .12f, .037f, .045f,       //rotate
   .18f, .037f, .045f,       //zoom
   .24f, .037f, .045f,       //pan
   .045f,.039f, .06f,        //ok
};

//----------------------------

static const E_PRG_ITEM prg_icon_id[] = {
   PRG_MOVE, PRG_FOLLOW, PRG_USE, PRG_USE_INV,
   PRG_ATTACK, PRG_GUARD, PRG_WAIT_SIGNAL, PRG_STAY_UP,
   PRG_STAY_MID, PRG_STAY_DOWN,
};

//----------------------------
                              //programming keys - flags
#define PRGF_DEST       1     //destination required
#define PRGF_ACTIVE     2     //active item required (actor)
#define PRGF_INVENTORY  4     //inventory selection

static const dword prg_flags[PRG_LAST] = {
                              //move, follow, use, use inv
   PRGF_DEST, PRGF_DEST|PRGF_ACTIVE, PRGF_DEST|PRGF_ACTIVE, PRGF_INVENTORY,
                              //attack, guard, watch, stand
   PRGF_DEST|PRGF_ACTIVE, PRGF_DEST, PRGF_DEST, 0,
                              //bend, lay, ?, ?
   0, 0, 0, 0,
                              //?, ?, ?, ?
   0, 0, 0, 0,
                              //?, ?, ?, ?
   0, 0, 0, 0,
                              //?, ?, ?, WAIT_SIGNAL
   0, 0, 0, 0,
};

//----------------------------

class C_map_actor: public C_unknown{
   C_smart_ptr<I3D_model> model;
   C_smart_ptr<I3D_model> hlight;
   bool active;
   int model_id;

//----------------------------

   void SetHLight(PI3D_scene scene){

      if(!model)
         return;
                              //create actor's highlight
      if(!hlight){
         hlight = driver->CreateModel();
         hlight->Release();
      }
      C_fstring name("map_mrk%i", model_id);
      model_cache.Open(hlight, name, scene, 0, ErrReport, editor);

                              //make visuals of the marker non-colliding,
                              // so that other model's don't accidentally link to it
      struct S_hlp{
         static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
            frm->SetFlags(I3D_FRMF_NOCOLLISION, I3D_FRMF_NOCOLLISION);
            return I3DENUMRET_OK;
         };
      };
      hlight->EnumFrames(S_hlp::cbEnum, 0, ENUMF_VISUAL);

      hlight->SetPos(S_vector(0, .6f, 0));
      hlight->Update();
      hlight->LinkTo(model);
   }

//----------------------------
public:
   PI3D_frame floor_dummy;
   LPC_actor actor;

   C_map_actor():
      actor(NULL),
      model_id(0),
      floor_dummy(NULL),
      active(false)
   {}

//----------------------------

   void SetModel(PI3D_model m, int id, PI3D_scene scene){

      model = m;
      model_id = id;
      if(model_id!=-1)
         SetHLight(scene);
   }

//----------------------------

   inline PI3D_model GetModel() const{ return model; }

//----------------------------

   void SetActive(bool b, PI3D_scene scene){

      active = b;
      SetHLight(scene);
      if(active)
         hlight->SetAnimation(0, NULL, I3DANIMOP_LOOP);
      else
         hlight->StopAnimation(0);
   }

//----------------------------

   void Tick(int time){

      if(hlight)
         hlight->Tick(time);
   }
};

typedef C_map_actor *LPC_map_actor;

//----------------------------
//----------------------------
                              //map hint ids - due to update of translated localication texts
static const short hint_id_map[] = {
   0, 1, 2, 3,
   4, 5, 11, 7,
   8, 9, 0, 0,
   0, 0, 0, 0,
   0, 0, 0, 0,
   0, 0, 0, 10,
};

//----------------------------

class C_map_manager_imp: public C_map_manager{
                              //resources
   C_game_mission *mission;
   C_poly_text texts;

   enum E_MAP_BUTTON{
      MBUT_LEVEL_UP, MBUT_LEVEL_DOWN,
      MBUT_ROTATE,
      MBUT_ZOOM,
      MBUT_PAN,
      MBUT_EXIT,
      //MBUT_CANCEL,
   };

   enum E_MAP_SOUND{
      MAP_SOUND_ICON,
      MAP_SOUND_FLOOR,
      MAP_SOUND_HINT,
      MAP_SOUND_PRG_ADD,

      MAP_SOUND_LAST
   };

   C_smart_ptr<I3D_interpolator> cam_intp;
   C_smart_ptr<ISND_sound> snd[MAP_SOUND_LAST]; //loaded sounds

                              //mouse caching
   byte last_mouse_butt;
   int last_mx, last_my;

                              //visual effects
   int flash_ctrl_button_id;
   int flash_prg_icon_id;     //prg icon (from stock) under cursor (-1 = no)
   int prg_item_hint_count;
                              //hint text & its fade
   C_smart_ptr<C_hint_base> hint;
   C_smart_ptr<C_hint_base> flt_msg;
   C_fade<S_fade_color> sel_icon_fade;
   C_fade<S_fade_color> path_fade;
   dword color_path_fade;

                              //icons
   int curr_arrow_shape;
   C_smart_ptr<C_poly_sprite> spr_arrow;
   C_smart_ptr<C_poly_sprite> ctrl_icons[MAP_NUM_CTRL_ICONS];  //controlling icons - fixed
   vector<C_smart_ptr<C_poly_sprite> > prg_icons;  //programming icons

                              //floors
   int curr_floor;            //current
   vector<vector<int> > floor_def;  //floor definitions - parts to be 'on'

   C_smart_ptr<C_net_sync> net_sync;
   C_smart_ptr<C_sync_menu> sync_menu;

                              //operation mode
   enum{
      MODE_NORMAL,
      MODE_AXIS_POINT, MODE_AXIS_SLIDE,
      MODE_ICON_POINT, MODE_ICON_SLIDE,
      MODE_DEST_SELECT, MODE_INVENTORY,
   } mode;

                              //world resolution (2D)
   I3D_bbox world_box;
   float map_zoom_max;

   int mouse_x, mouse_y;

                              //camera movement
   S_vector look_at;
   float look_angle, look_dist;
   float pan_speed;

                              //programming
   vector<C_smart_ptr<C_map_actor> > actors;
   vector<C_smart_ptr<I3D_model> > map_models;
   int prg_icon_dest;         //prg icon for which destination is taken
   int selected_prg_key, sel_actor; //selected prg key on player's time axis

//----------------------------
// Play selected sound.
   void PlaySound(E_MAP_SOUND id){
      assert(id<MAP_SOUND_LAST);
      if(snd[id] && !snd[id]->IsPlaying())
         snd[id]->Play();
   }

//----------------------------

   void ShowHint(const C_string &text, bool wide, bool sound=true){

      float sx = igraph->Scrn_sx();
      float x = (float)mouse_x / sx, y = (float)mouse_y/sx;
      hint = CreateHint(text, wide, &texts, x, y, .025f);
      hint->Release();
      if(sound) PlaySound(MAP_SOUND_HINT);
   }

//----------------------------

   void ShowFloor(int floor, bool on){

      PI3D_frame frm;
      int i;
      for(i=0; i<floor_def[floor].size(); i++){
         int pi=floor_def[floor][i];
         frm = scene->FindFrame(C_fstring("area %.2i", pi));
         if(!frm){
#ifdef _DEBUG
            OUTTEXT(C_fstring("cannot find map area part #%i", pi));
#endif
         }else
            frm->SetOn(on);
      }
                              //show/hide map actors
      for(i=0; i<actors.size(); i++){
         if(actors[i]->floor_dummy)
            actors[i]->GetModel()->SetOn(actors[i]->floor_dummy->IsOn());
      }
   }

//----------------------------

   void SwitchFloor(bool up){

      int new_floor;
      if(up){
         if((new_floor = curr_floor+1) >= floor_def.size()) return;
      }else{
         if((new_floor = curr_floor-1) < 0) return;
      }
                              //switch off current parts
      ShowFloor(curr_floor, false);
                              //switch on new parts
      ShowFloor(new_floor, true);

      curr_floor = new_floor;

      PlaySound(MAP_SOUND_FLOOR);
   }

//----------------------------
// Set cursor shape.
   void SetCursor(int i, bool on = true){

      if(curr_arrow_shape!=i){
         spr_arrow = new C_poly_sprite(0, 0, .03f, &icons_img, curr_arrow_shape = i);
         spr_arrow->Release();
         spr_arrow->SetScreenPos(mouse_x, mouse_y);
      }
      spr_arrow->SetOn(on);
   }

//----------------------------
// Create floating message.
   void FltMsg(int txt_id){

      float sx = igraph->Scrn_sx();
      float x = (float)mouse_x / sx, y = (float)mouse_y / sx;
      flt_msg = CreateHint(all_txt[txt_id], all_txt.IsWide(txt_id), &texts, x, y, .025f, -1, true);
      flt_msg->Release();
   }

//----------------------------

   void AddPlayerKey(int prg_id){

      int i;
                              //add to player's program
      LPC_map_actor ma = FindActiveActor();
      assert(ma);
      if(!ma) return;
                              //get insert position
      if(selected_prg_key!=-1) i = selected_prg_key;
      else i = ma->actor->cbProc(CB_GET_PRG_SIZE);
                              //add key at position
      i = ma->actor->AddProgram(i, prg_icon_id[prg_id], S_prg_add(), true);
      if(i){
         ma->actor->cbProc(CB_PRG_PREPROCESS, (selected_prg_key==-1) ? 2 : 0);
         ma->actor->cbProc(CB_PRG_SHOW, false);
         PlaySound(MAP_SOUND_PRG_ADD);
      }
   }

//----------------------------

   void LinkModel(PI3D_model model, LPC_map_actor ma = NULL){

      const I3D_bbox &bb = model->GetHRBoundVolume().bbox;
      S_vector center = bb.min + (bb.max-bb.min)*.5f;
      S_vector from = center*model->GetMatrix();

      PI3D_frame frm = NULL;
      I3D_collision_data cd(from, S_vector(0, -10, 0),
         I3DCOL_INVISIBLE |
         I3DCOL_EXACT);
      if(scene->TestCollision(cd))
         frm = cd.GetHitFrm();

      if(frm){
         if(ma){
                              //find floor dummy
            while(frm){
               frm = frm->GetParent();
               if(!frm) break;
               if(frm->GetType()==FRAME_DUMMY)
               if(C_string(frm->GetName()).Match("area *")){
                  ma->floor_dummy = frm;
                  break;
               }
            }
         }else
            model->LinkTo(frm, I3DLINK_UPDMATRIX);
      }else{
         if(ma)
            ma->floor_dummy = NULL;
         else
            model->LinkTo(scene->GetPrimarySector());
      }
   }

//----------------------------
// Find actor by game-menu id.
   LPC_map_actor FindActor(int gm_id){

      for(int i=0; i<actors.size(); i++){
         if(actors[i]->actor->GetMenuID()==gm_id)
            return actors[i];
      }
      return NULL;
   }

//----------------------------

   LPC_map_actor FindActiveActor(){

      for(int i=0; i<actors.size(); i++){
         if(actors[i]->actor->IsActive())
            return actors[i];
      }
      return NULL;
   }

//----------------------------

   void SetActive(LPC_map_actor ma, bool cam_anim = true){

      if(selected_prg_key!=-1){
         mission->game_menu.SetPrgKeyColor(FindActor(sel_actor)->actor->GetMenuID(), selected_prg_key, -1);
         selected_prg_key = -1;
      }
      S_vector new_look_at = *ma->actor->GetPos();
      new_look_at.y = 0.0f;
      S_vector new_dir;

      S_vector dir1(*ma->actor->GetDir());
      float new_look_angle = Angle2D(S_vector2(-dir1.x, dir1.z));
      new_dir.x = sin(new_look_angle);
      new_dir.z = cos(new_look_angle);
      new_dir.y = MAP_CAM_Y;
      new_dir.Normalize();
      new_dir *= look_dist;

      look_angle = new_look_angle;

      S_vector new_pos = new_look_at - new_dir;

      look_at = new_look_at;

      if(cam_anim){
                              //update camera to look at him

#define MAP_CAM_ANIM_TIME 600

         PI3D_keyframe_anim anim = (PI3D_keyframe_anim)driver->CreateAnimation(I3DANIM_KEYFRAME);
         I3D_anim_pos_tcb keys[2];
         keys[0].Clear();
         keys[1].Clear();
         keys[1].time=MAP_CAM_ANIM_TIME;
         keys[0].easy_from = 1.0f;
         keys[1].easy_to = 1.0f;
         {
            keys[0].v = scene->GetActiveCamera()->GetPos();
            keys[1].v = new_pos;
            anim->SetPositionKeys(keys, 2);
         }
         {
            I3D_anim_look_at_tcb keys[2];
            keys[0].Clear();
            keys[0].up_vector = S_vector(0, 1, 0);
            keys[1].Clear();
            keys[1].up_vector = S_vector(0, 1, 0);
            keys[0].easy_from = 1.0f;
            keys[1].easy_to = 1.0f;
            keys[1].time=MAP_CAM_ANIM_TIME;
            keys[0].v = scene->GetActiveCamera()->GetDir();
            keys[1].v = new_dir;
            anim->SetLookAtKeys(keys, 2);
         }
         anim->SetEndTime(MAP_CAM_ANIM_TIME);

         if(!cam_intp){
            cam_intp = driver->CreateInterpolator();
            cam_intp->Release();
         }
         cam_intp->SetFrame(scene->GetActiveCamera());
         cam_intp->SetAnimation(0, anim);

         anim->Release();
      }else{
         PositionCamera();
      }
      mission->game_menu.SetCompassAngle(-new_look_angle);

                              //determine floor to show
      {
         vector<int> off_parts;
         S_vector to = *ma->actor->GetPos();
         S_vector from = to;
         to.y += 1.9f;         //height of human
         from.y += 1000.0f;
                              //get any layer, which is above out soldier
         while(true){
            PI3D_frame frm = NULL;

            I3D_collision_data cd(from, to-from, I3DCOL_INVISIBLE | I3DCOL_EXACT);
            if(scene->TestCollision(cd)){
               frm = cd.GetHitFrm();
               from = cd.GetHitPosition();
            }

                              //get model/dummy parent
            if(!frm)
               break;
            bool find = true;
            while(find){
               frm = frm->GetParent();
               if(!frm)
                  break;
               switch(frm->GetType()){
               case FRAME_MODEL:
                  frm = NULL;
                  find = false;
                  break;

               case FRAME_DUMMY:
                  if(!strncmp(frm->GetName(), "area ", 5))
                     find = false;
                  break;
               }
            }
            if(!frm)
               break;
            int i;
            if(sscanf(&frm->GetName()[5], "%i", &i) == 1){
               for(int j=off_parts.size(); j--; ) if(off_parts[j]==i) break;
               if(j==-1) off_parts.push_back(i);
            }
            from.y -= .2f;
            if(from.y <= to.y)
               break;
         }
                              //find what's under feets
         int on_part = -1;
         {
            S_vector from = *ma->actor->GetPos();
            from.y += .1f;

            do{
               S_vector dir(0, -.5f, 0);
               PI3D_frame frm = NULL;

               I3D_collision_data cd(from, S_vector(0, -.5, 0), I3DCOL_INVISIBLE | I3DCOL_EXACT);
               if(scene->TestCollision(cd)){
                  frm = cd.GetHitFrm();
                  from = cd.GetHitPosition();
               }

               if(!frm)
                  break;
               bool find = true;
               bool out = false;
               while(find){
                  frm = frm->GetParent();
                  if(!frm){
                     out = true;
                     break;
                  }
                  switch(frm->GetType()){
                  case FRAME_MODEL:
                     frm = NULL;
                     find = false;
                     break;

                  case FRAME_DUMMY:
                     if(!strncmp(frm->GetName(), "area ", 5))
                        find = false;
                     break;
                  }
               }
               if(frm){
                  int i;
                  if(sscanf(&frm->GetName()[5], "%i", &i) == 1){
                     for(int j=off_parts.size(); j--; ) if(off_parts[j]==i){
                        off_parts[j] = off_parts.back(); off_parts.pop_back();
                        break;
                     }
                     on_part = i;
                  }
               }
               from.y -= .001f;
            }while(on_part==-1);
         }

                              //try to unhide what's necessary
         int best_score = -1000, new_floor = -1;
         for(int i=floor_def.size(); i--; ){
            int score = 0;
            for(int j=0; j<off_parts.size(); j++){
               for(int k=floor_def[i].size(); k--; )
                  if(floor_def[i][k]==off_parts[j])
                     break;
               score += k==-1 ? 1 : -1;
            }
            if(on_part!=-1){
               for(int k=floor_def[i].size(); k--; )
                  if(floor_def[i][k]==on_part){
                     ++score;
                     break;
                  }
            }
            if(score>0 && best_score<score){
               best_score = score;
               new_floor = i;
            }
         }
                              //switch to new floor
         if(new_floor!=-1 && new_floor!=curr_floor){
            ShowFloor(curr_floor, false);
            ShowFloor(new_floor, true);
            curr_floor = new_floor;
         }
      }
   }

//----------------------------

   void SetActive(int plr_id, bool sw = true){

      if(plr_id<0)
         return;

      LPC_map_actor act = FindActor(plr_id);
      if(!act)
         return;
      if(!sw || (!act->actor->IsActive() && act->actor->IsAlive())){
         SetActive(act);
         if(sw){
                              //switch active
            for(int i=0; i<actors.size(); i++){
               if(actors[i]->actor->IsActive()){
                  actors[i]->actor->SetActive(false);
                  break;
               }
            }
            act->actor->SetActive(true);
         }
      }
   }

//----------------------------

   void SelectPrgKey(int plr_id, int key_id){

      if(selected_prg_key!=-1)
         mission->game_menu.SetPrgKeyColor(FindActor(sel_actor)->actor->GetMenuID(), selected_prg_key, -1);
      selected_prg_key=key_id;
      sel_actor=plr_id;
   }

//----------------------------

   void PrgIconPressed(int id){

      LPC_map_actor ma = FindActiveActor();
      if(!ma)
         return;
      LPC_actor a = ma->actor;
                              //can't program dead
      if(!a->IsAlive())
         return;
                              //cannot program net player
      if(a->cbProc(CB_IS_NET_PLAYER))
         return;

      int i = prg_icon_id[id];

      if(prg_icon_dest != -1)
         prg_icons[prg_icon_dest]->SetColor(MAP_ICON_INACTIVE);

      if(prg_flags[i]&PRGF_DEST){
                              //put info message
         switch(prg_icon_id[id]){
         case PRG_ATTACK:
         case PRG_USE:
         case PRG_FOLLOW:
            FltMsg(GT_MAP_ACTIVE_REQ);
            break;
         case PRG_MOVE:
         default:
            FltMsg(GT_MAP_DEST_REQ);
            break;
         }
         mode = MODE_DEST_SELECT;
         SetCursor(2);
         prg_icon_dest = id;
      }else
      if(prg_flags[i]&PRGF_INVENTORY){
                              //set inventory mode
         FltMsg(GT_MAP_INVENTORY);
         mode = MODE_INVENTORY;
         spr_arrow->SetOn(false);
         prg_icon_dest = id;
                              //show inventory
         LPC_map_actor ma = FindActiveActor();
         ma->actor->cbProc(CB_INV_GRAB, 1, true);
      }else{
                              //put key
         AddPlayerKey(id);
         prg_icon_dest = -1;
      }

      if(prg_icon_dest!=-1 && prg_icon_dest!=flash_prg_icon_id)
         prg_icons[prg_icon_dest]->SetColor(MAP_BUTTON_ACTIVE);
   }

//----------------------------

   S_vector GetLookDir(){

      S_vector dir;
      dir.x = sin(look_angle);
      dir.z = cos(look_angle);
      dir.y = MAP_CAM_Y;
      dir.Normalize();
      dir *= look_dist;
      return dir;
   }

//----------------------------

   void PositionCamera(){

      S_vector dir = GetLookDir();
      PI3D_camera cam = scene->GetActiveCamera();
      cam->SetPos(look_at - dir);
      cam->SetDir(dir);
      cam->Update();
      cam->SetRange(look_dist/3, cam_range);
   }

//----------------------------
// Cancel curent (special) mode, revert to normal. Set normal cursor, hide messages, etc.
   void CancelMode(){

      if(mode!=MODE_NORMAL){
         if(prg_icon_dest != flash_prg_icon_id)
            prg_icons[prg_icon_dest]->SetColor(MAP_ICON_INACTIVE);
         prg_icon_dest = -1;
         FltMsg(GT_MAP_CANCELED);
         SetCursor(0);
         mode = MODE_NORMAL;
      }
   }

//----------------------------
   
   void ExitMapMode(const S_tick_context &tc){

      if(net){
         net_sync->BeginEvent();
         return;
      }
      mission->ExitMapMode();
      tc.p_ctrl->IgnoreAllPressed();
   }

//----------------------------

public:
   C_map_manager_imp(C_game_mission *gm):
      mission(gm),
      selected_prg_key(-1),
      curr_arrow_shape(-1),
      prg_icon_dest(-1),
      sel_actor(-1),
      curr_floor(0)
   {
                              //our net_sync creates a sub-chunk of NM_GAME, because we're always ticked
                              // from inside of NM_GAME
      net_sync = CreateNetSync(C_game_mission::NM_MAP);
      net_sync->Release();
   }

   ~C_map_manager_imp(){
      Close();
   }

//----------------------------

   bool Init();

//----------------------------

   virtual void Activate();

//----------------------------

   void Suspend(){

      hint = NULL;
      flt_msg = NULL;
      if(selected_prg_key!=-1){
         mission->game_menu.SetPrgKeyColor(FindActor(sel_actor)->actor->GetMenuID(), selected_prg_key, -1);
         selected_prg_key = -1;
      }
      actors.clear();
      map_models.clear();

      SetCursor(5, false);
      cam_intp = NULL;
   }

//----------------------------

   void Close(){

      C_mission::Close();

      hint = NULL;
      flt_msg = NULL;

      for(int i=0; i<MAP_NUM_CTRL_ICONS; i++)
         ctrl_icons[i] = NULL;

      prg_icons.clear();

      for(i=0; i<MAP_SOUND_LAST; i++)
         snd[i] = NULL;

      cam_intp = NULL;
      floor_def.clear();
   }

//----------------------------

   void ProcessInput(const S_tick_context &tc){


                              //update mouse position
      mouse_x = Max(0, Min((int)igraph->Scrn_sx(), mouse_x + tc.mouse_rel[0]));
      mouse_y = Max(0, Min((int)igraph->Scrn_sy(), mouse_y + tc.mouse_rel[1]));

      int i;

      if(mode==MODE_INVENTORY){
         mouse_x = last_mx;
         mouse_y = last_my;

         bool off = false;
         enum{
            A_NONE, A_SCRLUP, A_SCRLDN, A_SELECT
         }action = A_NONE;

         switch(tc.key){
         case K_ESC:
            FltMsg(GT_MAP_CANCELED);
            off = true;
            break;
         case K_CURSORUP:
            action = A_SCRLUP;
            break;
         case K_CURSORDOWN:
            action = A_SCRLDN;
            break;
         }
         if(tc.p_ctrl->Get(GKEY_INV_SCROLL_LEFT)){
            tc.p_ctrl->IgnoreValue(GKEY_INV_SCROLL_LEFT, 250);
            action = A_SCRLUP;
         }else
         if(tc.p_ctrl->Get(GKEY_INV_SCROLL_RIGHT)){
            tc.p_ctrl->IgnoreValue(GKEY_INV_SCROLL_RIGHT, 250);
            action = A_SCRLDN;
         }else
         if(tc.p_ctrl->Get(GKEY_ITEM_USE)){
            tc.p_ctrl->IgnoreValue(GKEY_ITEM_USE);
            action = A_SELECT;
         }

         switch(action){
         case A_SCRLUP:
         case A_SCRLDN:
                              //scroll in inventory through grab messages - sending over net is done
            FindActiveActor()->actor->cbProc(CB_INV_GRAB, 2, (action==A_SCRLUP));
            break;
         case A_SELECT:
            {
               LPC_actor act = FindActiveActor()->actor;
                              //get insert position
               i = (selected_prg_key!=-1) ? selected_prg_key : act->cbProc(CB_GET_PRG_SIZE);
               int itm = act->cbProc(CB_INV_GRAB, 6, NULL);
               if(itm!=-1){
                                       //add key at position
                  S_prg_add pa;
                  pa.d[0] = itm;
                  act->AddProgram(i, prg_icon_id[prg_icon_dest], pa, true);

                  PlaySound(MAP_SOUND_PRG_ADD);
                  act->cbProc(CB_PRG_PREPROCESS, (selected_prg_key==-1) ? 2 : 0);
                  act->cbProc(CB_PRG_SHOW, false);
               }else
                  FltMsg(GT_MAP_NOT_FOR_USE);
               off = true;
            }
            break;
         }
         if(off){
            spr_arrow->SetOn(true);
            FindActiveActor()->actor->cbProc(CB_INV_GRAB, 1, false);
            mode = MODE_NORMAL;
            if(prg_icon_dest != flash_prg_icon_id)
               prg_icons[prg_icon_dest]->SetColor(MAP_ICON_INACTIVE);
            prg_icon_dest = -1;
         }
         return;
      }

                              //camera update requests
      S_vector2 pan_request(0, 0);
      float rotate_request = 0.0f;
      int zoom_request = 0;
      bool want_level_up = false;
      bool want_level_dn = false;

      if(tc.key!=K_NOKEY){
                              //check shortcut
         char c = tolower(igraph->GetAscii(tc.key));
         for(i=0; i<prg_icons.size(); i++){
            int txt_id = GT_PRG_HINT_BASE + hint_id_map[prg_icon_id[i]];
            const C_string &str = all_txt[txt_id];
            const char *cp = find(&str[0], &str[(int)str.Size()], '&');
            if(cp!=&str[(int)str.Size()]){
               char shortcut;
               if(!all_txt.IsWide(txt_id))
                  shortcut = tolower(cp[1]);
               else
                  shortcut = !cp[3] ? tolower(cp[2]) : 0;
               if(c==shortcut){
                  PrgIconPressed(i);
                  break;
               }
            }
         }
      }

      switch(tc.mouse_buttons){
      case 1:
         if(mode==MODE_DEST_SELECT){
            if(!(last_mouse_butt&1)){
               SetCursor(0);

               mode = MODE_NORMAL;

               LPC_map_actor ma = FindActiveActor();
               assert(ma);

               if(!ma->actor->cbProc(CB_IS_NET_PLAYER))
               do{
                                 //evaluate destination
                  S_vector from, dir;
                  scene->UnmapScreenPoint(mouse_x, mouse_y, from, dir);
                  S_vector dest, norm;
                  PI3D_frame frm(NULL);

                  I3D_collision_data cd(from, dir, I3DCOL_EXACT | I3DCOL_RAY);
                  if(scene->TestCollision(cd)){
                     frm = cd.GetHitFrm();
                     dest = cd.GetHitPosition();
                     norm = cd.GetHitNormal();
                  }
                  if(!frm)
                     break;
                                 //check what's hit
                  if(dest.x<=world_box.min.x || dest.x>=world_box.max.x ||
                     dest.z<=world_box.min.z || dest.z>=world_box.max.z){

                     FltMsg(GT_MAP_DEST_OUT);
                     break;
                  }
                  while(frm && frm->GetType()!=FRAME_MODEL) frm=frm->GetParent();
                  LPC_actor active=NULL;
                  if(frm){       //find whom belongs this
                     for(i=0; i<actors.size(); i++)
                     if(actors[i]->GetModel() == frm){
                        active = actors[i]->actor;
                        break;
                     }
                  }
                                 //add to player's program
                                 //get insert position
                  i = (selected_prg_key!=-1) ? selected_prg_key :
                     ma->actor->cbProc(CB_GET_PRG_SIZE);

                  S_prg_add pa;
                                          //add key at position
                  bool b = true;
                  {
                                 //setup key data
                     switch(prg_icon_id[prg_icon_dest]){
                     case PRG_MOVE:
                        pa.d[0] = (dword)&dest;
                        pa.d[1] = true;
                        break;
                     case PRG_FOLLOW:
                        if(!active){
                           FltMsg(GT_MAP_NO_ACTIVE);
                           b = false;
                        }else{
                           pa.d[0] = (dword)active;
                        }
                        break;

                     case PRG_USE:
                        if(!active){
                           FltMsg(GT_MAP_NO_ACTIVE);
                           b = false;
                        }else
                        if(!active->NumUsePositions(ma->actor)){
                           FltMsg(GT_MAP_NOT_FOR_USE);
                           b = false;
                        }else{
                           pa.d[0] = (dword)active;
                        }
                        break;

                     case PRG_ATTACK:
                        if(!active){
                           FltMsg(GT_MAP_NO_ACTIVE);
                           b  =false;
                        }else    //can't attack non-enemy
                        if(active->IsHuman() && !active->IsEnemy(ma->actor)){
                           FltMsg(GT_MAP_CANNOT_DO_THIS);
                           b = false;
                        }else{
                           pa.d[0] = (dword)active;
                           pa.d[2] = AM_MUST_SUCCEED;
                        }
                        break;

                     case PRG_GUARD:
                        if(active){
                           pa.d[3] = (dword)&active->GetFrame()->GetWorldPos();
                        }else{
                           pa.d[3] = (dword)&dest;
                        }
                        break;

                     case PRG_WATCH:
                        if(active) dest = active->GetFrame()->GetWorldPos();
                        dest.y += 1.0f;
                        pa.d[0] = (dword)&dest;
                        break;

                     default:
                        assert(0);
                     }
                     if(b){
                        PlaySound(MAP_SOUND_PRG_ADD);
                        ma->actor->AddProgram(i, prg_icon_id[prg_icon_dest], pa, true);
                        ma->actor->cbProc(CB_PRG_PREPROCESS, (selected_prg_key==-1) ? 2 : 0);
                        ma->actor->cbProc(CB_PRG_SHOW, false, 1);
                        //ma->actor->AddProgram(i, prg_icons[prg_icon_dest].itm, pa, true);
                     }
                  }
                  if(prg_icon_dest != flash_prg_icon_id)
                     prg_icons[prg_icon_dest]->SetColor(MAP_ICON_INACTIVE);
                  prg_icon_dest = -1;
               }while(false);
            }
            break;
         }
         if(mode!=MODE_NORMAL){
                              //slide time axis

                              //no Y move
            mouse_y = last_my;
            if(!tc.mouse_rel[0])
               break;
                              //slide by amount
            i = mission->game_menu.GetTimeOffset();
            float f0, f1;
            mission->game_menu.IconX(i, f0);
            mission->game_menu.IconX(i+10000, f1);
            i = tc.mouse_rel[0]*10000 / ((f1-f0)*igraph->Scrn_sx());

            switch(mode){
            case MODE_AXIS_POINT:
               mode = MODE_AXIS_SLIDE;
            case MODE_AXIS_SLIDE:
               mission->game_menu.ScrollTimeAxis(i);
               break;
            case MODE_ICON_POINT:
               mode = MODE_ICON_SLIDE;
            case MODE_ICON_SLIDE:
               {
                  LPC_actor a = FindActor(sel_actor)->actor;
                  if(!a->cbProc(CB_IS_NET_PLAYER))
                     a->cbProc(CB_SET_PRG_TIME, selected_prg_key, i);
               }
               break;
            }
            break;
         }

         switch(flash_ctrl_button_id){
         case MBUT_ROTATE:
            rotate_request = (float)tc.mouse_rel[0] * .002f;
            mouse_x = last_mx, mouse_y = last_my;
            break;

         case MBUT_ZOOM:
            zoom_request = tc.mouse_rel[1] - tc.mouse_rel[0];
            mouse_x = last_mx, mouse_y = last_my;
            break;

         case MBUT_PAN:
            pan_request.x = tc.mouse_rel[0];
            pan_request.y = tc.mouse_rel[1];
            mouse_x = last_mx, mouse_y = last_my;
            break;

         case MBUT_LEVEL_UP:
            want_level_up = true;
            break;

         case MBUT_LEVEL_DOWN:
            want_level_dn = true;
            break;

         case MBUT_EXIT:
            ExitMapMode(tc);
            return;

         default:
            if(!(last_mouse_butt&1)){
                              //ask game-menu for info on player's menus
               int plr_id, sub_id;
               float scrn_sx = igraph->Scrn_sx();
               bool b = mission->game_menu.HitCheck((float)mouse_x/scrn_sx, (float)mouse_y/scrn_sx, plr_id, sub_id);
               if(b){
                  SelectPrgKey(-1, -1);
                  switch(sub_id){
                  case -1:    //player's face
                     SetActive(plr_id);
                     break;
                  case -2:    //time axis
                     break;
                  default:
                     SelectPrgKey(plr_id, sub_id);
                     SetActive(plr_id);
                     SetCursor(3);
                     mode = MODE_ICON_SLIDE;
                  }
                  break;
               }
               if(flash_prg_icon_id != -1){

                  PrgIconPressed(flash_prg_icon_id);
                  break;
               }
                              //check map hit
               S_vector mpos, mdir;
               scene->UnmapScreenPoint(mouse_x, mouse_y, mpos, mdir);

                              //find clicked model
               I3D_collision_data cd(mpos, mdir, I3DCOL_EXACT | I3DCOL_RAY);
               if(scene->TestCollision(cd)){
                  PI3D_frame frm = cd.GetHitFrm();

                  while(frm && frm->GetParent()){
                     if((frm=frm->GetParent())->GetType()==FRAME_MODEL){
                                    //find out who's this model
                        for(int i=0; i<actors.size(); i++){
                           if(actors[i]->GetModel() == frm){
                              SetActive(actors[i]->actor->GetMenuID());
                              SetActive(actors[i]);
                              break;
                           }
                        }
                        break;
                     }
                  }
               }
            }
         }
         break;

      case 2:
         {
            pan_request.x = tc.mouse_rel[0];
            pan_request.y = tc.mouse_rel[1];
            mouse_x = last_mx, mouse_y = last_my;
         }
         break;

      case 3:
         if(mode==MODE_NORMAL){
            zoom_request = tc.mouse_rel[1];
            rotate_request = (float)tc.mouse_rel[0] * .002f;
            mouse_x = last_mx, mouse_y = last_my;
         }
         break;
      }

      if(tc.p_ctrl){
         if(tc.p_ctrl->Get(GKEY_MAP)){   
            tc.p_ctrl->IgnoreValue(GKEY_MAP);
            ExitMapMode(tc);
         }else
         if(tc.p_ctrl->Get(GKEY_MAP_FLOOR_UP)){
            tc.p_ctrl->IgnoreValue(GKEY_MAP_FLOOR_UP);
            want_level_up = true;
         }else
         if(tc.p_ctrl->Get(GKEY_MAP_FLOOR_DOWN)){
            tc.p_ctrl->IgnoreValue(GKEY_MAP_FLOOR_DOWN);
            want_level_dn = true;
         }else
         if((i=GKEY_PLR_SWITCH,tc.p_ctrl->Get(i)) || (i=GKEY_PLR_SWITCH_1,tc.p_ctrl->Get(i)) ||
            (i=GKEY_PLR_SWITCH_2,tc.p_ctrl->Get(i)) || (i=GKEY_PLR_SWITCH_3,tc.p_ctrl->Get(i)) ||
            (i=GKEY_PLR_SWITCH_4,tc.p_ctrl->Get(i))){

            tc.p_ctrl->IgnoreValue(i);

            LPC_map_actor old_a = FindActiveActor();
            if(old_a)
               old_a->SetActive(false, scene);

            i = mission->PlayerSwitch(igraph->GetShiftKeys()&SKEY_SHIFT,
               (i==GKEY_PLR_SWITCH) ? -1 : i-GKEY_PLR_SWITCH_1);
            SetActive(i, false);
            LPC_map_actor new_a = FindActiveActor();
            if(new_a)
               new_a->SetActive(true, scene);
         }
      }

      switch(tc.key){
      case K_ESC:
         CancelMode();
         break;
      case K_DEL:
      case K_BACKSPACE:
         {
            LPC_map_actor ma = FindActiveActor();
            if(!ma->actor->cbProc(CB_IS_NET_PLAYER)){
               if((tc.key==K_BACKSPACE) || selected_prg_key==-1){
                  i = ma->actor->cbProc(CB_GET_PRG_SIZE);
                                 //delete last key
                  if(i)
                     ma->actor->DelProgram(i-1, true);
               }else
               if(selected_prg_key!=-1){
                                 //delete selected key
                  ma->actor->DelProgram(selected_prg_key, true);
                  selected_prg_key = -1;
                  ma->actor->cbProc(CB_PRG_PREPROCESS, 0);
                  ma->actor->cbProc(CB_PRG_SHOW, false);
               }
            }
         }
         break;
      }

      if(tc.mouse_buttons!=1){
         switch(mode){
         case MODE_AXIS_SLIDE:
         case MODE_ICON_SLIDE:
            int plr_id, sub_id;
            float scrn_sx = igraph->Scrn_sx();
            bool b = mission->game_menu.HitCheck((float)mouse_x/scrn_sx, (float)mouse_y/scrn_sx, plr_id, sub_id);
            if(b && sub_id==-2){
               mode = MODE_AXIS_POINT;
            }else{
               mode = MODE_NORMAL;
               SetCursor(0);
            }
            break;
         }
      }

      if(want_level_up){
         if(!(last_mouse_butt&1))
            SwitchFloor(true);
      }else
      if(want_level_dn){
         if(!(last_mouse_butt&1))
            SwitchFloor(false);
      }

      if(!cam_intp){
                              //update camera position/direction
         float &dir_angle = look_angle;

         bool upd = false, upd_dir = false;

                              //rotation
         if(rotate_request!=0.0f){
            dir_angle += rotate_request;
            upd = true;
            upd_dir = true;
            if(flash_ctrl_button_id!=MBUT_ROTATE)
               ctrl_icons[MBUT_ROTATE]->SetColor(MAP_BUTTON_ACTIVE);
         }else{
            if(flash_ctrl_button_id!=MBUT_ROTATE)
               ctrl_icons[MBUT_ROTATE]->SetColor(MAP_BUTTON_INACTIVE);
         }

                              //zoom
         if(zoom_request){
            look_dist = Min(map_zoom_max, Max(MAP_ZOOM_MIN,
               look_dist + (float)zoom_request * .16f * pan_speed));
            upd = true;
            if(flash_ctrl_button_id!=MBUT_ZOOM)
               ctrl_icons[MBUT_ZOOM]->SetColor(MAP_BUTTON_ACTIVE);
         }else{
            if(flash_ctrl_button_id!=MBUT_ZOOM)
               ctrl_icons[MBUT_ZOOM]->SetColor(MAP_BUTTON_INACTIVE);
         }
                                 //panning
         if(pan_request.x || pan_request.y){

            float fx = (float)pan_request.x * .08f;
            float fz = (float)pan_request.y * .08f;
            fx *= pan_speed;
            fz *= pan_speed;

            float c = cos(dir_angle), s = sin(dir_angle);
            look_at.x += -fx*c + fz*s;
            look_at.z +=  fx*s + fz*c;
            look_at.x = Max(world_box.min.x+.3f, Min(world_box.max.x-.3f, look_at.x));
            look_at.z = Max(world_box.min.z+.3f, Min(world_box.max.z-.3f, look_at.z));

            upd = true;
            if(flash_ctrl_button_id!=MBUT_PAN)
               ctrl_icons[MBUT_PAN]->SetColor(MAP_BUTTON_ACTIVE);
         }else{
            if(flash_ctrl_button_id!=MBUT_PAN)
               ctrl_icons[MBUT_PAN]->SetColor(MAP_BUTTON_INACTIVE);
         }
         if(upd){
            PositionCamera();
            if(upd_dir)
               mission->game_menu.SetCompassAngle(-dir_angle);
         }
      }

      spr_arrow->SetScreenPos(mouse_x, mouse_y / (screen_aspect_ratio * 1.33333333f));
      last_mouse_butt = tc.mouse_buttons;

      if(last_mx!=mouse_x || last_my!=mouse_y)
      do{
                              //check icons
         float scrn_sx = igraph->Scrn_sx();
         float sx = (float)mouse_x / scrn_sx;
         float sy = (float)mouse_y / scrn_sx;

         for(i=MAP_NUM_CTRL_ICONS; i--; ){
            float size = .85f * map_img_pos[i][2] / 2.0f;
            if(map_img_pos[i][0]-size < sx && map_img_pos[i][0]+size > sx &&
               map_img_pos[i][1]-size < sy && map_img_pos[i][1]+size > sy)
               break;
         }
                              //update pointed control button
         if(flash_ctrl_button_id != i){
                              //unlit old
            if(flash_ctrl_button_id != -1)
               ctrl_icons[flash_ctrl_button_id]->SetColor(MAP_BUTTON_INACTIVE);
            flash_ctrl_button_id = i;
                              //lit new
            if(i != -1){
               ctrl_icons[flash_ctrl_button_id]->SetColor(MAP_BUTTON_ACTIVE);
               PlaySound(MAP_SOUND_ICON);
            }
         }
         if(flash_ctrl_button_id!=-1)
            break;

                              //check programming icons
         const float SIZE = .85f * MAP_PRG_SIZE / 2.0f;
         for(i=prg_icons.size(); i--; ){
            if(sx>MAP_PRG_X(i)-SIZE && sx<MAP_PRG_X(i)+SIZE &&
               sy>MAP_PRG_Y(i)-SIZE && sy<MAP_PRG_Y(i)+SIZE)
               break;
         }
                              //update pointed programming icon
         if(i != flash_prg_icon_id){
                              //unlit old
            if(flash_prg_icon_id!=-1 && flash_prg_icon_id!=prg_icon_dest)
               prg_icons[flash_prg_icon_id]->SetColor(MAP_ICON_INACTIVE);
            flash_prg_icon_id = i;
                              //lit new
            if(i != -1){
               prg_item_hint_count = PRG_HINT_SWITCH_COUNT;
               prg_icons[flash_prg_icon_id]->SetColor(MAP_BUTTON_ACTIVE);
            }else
            if(hint)
               hint->FadeOff();
            break;
         }

         int plr_id, sub_id;
         bool b = mission->game_menu.HitCheck((float)mouse_x/scrn_sx, (float)mouse_y/scrn_sx, plr_id, sub_id);
         if(b){
            switch(sub_id){
            case -2:
               if(mode==MODE_NORMAL){
                  mode = MODE_AXIS_POINT;
                  SetCursor(3);
               }
               break;
            case -1:
               break;
            default:
               switch(mode){
               case MODE_NORMAL:
               case MODE_AXIS_POINT:
               case MODE_ICON_POINT:
                  if(selected_prg_key==sub_id && FindActor(plr_id)==FindActiveActor()){
                     /*
                     switch(mode)
                     case MODE_NORMAL:
                     case MODE_AXIS_POINT:
                     */
                        mode = MODE_ICON_POINT;
                        SetCursor(3);
                        /*
                        break;
                     */
                  }else{
                     mode = MODE_NORMAL;
                     SetCursor(0);
                  }
                  break;
               }
            }
         }else
         switch(mode){
         case MODE_AXIS_POINT:
         case MODE_ICON_POINT:
            mode = MODE_NORMAL;
            SetCursor(0);
            break;
         }
      }while(false);

      last_mx = mouse_x, last_my = mouse_y;
   }

//----------------------------

   virtual void Tick(const S_tick_context &tc){

      if(net){
         switch(net_sync->GetState()){
         case C_net_sync::SS_RESOLVED:
            {
               if(sync_menu){
                  sync_menu->AddAllCheckMarks();
                  RenderScreen();
                  sync_menu = NULL;
               }

               net_sync->Reset();
               mission->ExitMapMode();
               if(tc.p_ctrl)
                  tc.p_ctrl->IgnoreAllPressed();
               return;
            }
            break;

         case C_net_sync::SS_PROGRESS:
            {
#ifndef EDITOR
                              //waiting for sync event
               if(!sync_menu){
                  sync_menu = CreateSyncMenu(20304);
                  sync_menu->Release();
               }
               vector<dword> pids;
               net_sync->GetAnsweredPIDs(pids);
               sync_menu->AddCheckMarks(pids.begin(), pids.size());
#endif
            }
            return;
         }
      }

      ProcessInput(tc);                              

      C_mission::Tick(tc);

      for(int i=actors.size(); i--; )
         actors[i]->Tick(tc.time);

                              //update hint
      if(hint && !hint->Tick(tc.time))
         hint = NULL;

                              //update float message
      if(flt_msg && !flt_msg->Tick(tc.time))
         flt_msg = NULL;

                              //icon hints
      if(prg_item_hint_count && (prg_item_hint_count-=tc.time)<=0){
         prg_item_hint_count = 0;
         if(flash_prg_icon_id != -1){
            int txt_id = hint_id_map[prg_icon_id[flash_prg_icon_id]];
            C_string str(all_txt[GT_PRG_HINT_BASE + txt_id]);
            bool wide = all_txt.IsWide(txt_id);
            if(!wide){
               for(int i=str.Size(); i--; )
               if(str[i]=='&'){
                  strcpy(&str[i], &str[i+1]);
                  break;
               }
            }else{
               wchar_t *wp = (wchar_t*)&str[0];
               for(i=0; i<str.Size()/2; i++){
                  if(wp[i]=='&')
                     break;
               }
               for(; i<str.Size()/2; i++)
                  wp[i] = wp[i+1];
               str.Assign((const char*)wp, str.Size()-2);
            }
            ShowHint(str, wide, false);
         }
      }

                              //flash selected prg key
      if(selected_prg_key!=-1){
         LPC_map_actor ma = FindActor(sel_actor);
         if(sel_icon_fade.IsEnd()){
            static const dword colors[]={
               0xffa0a0, 0xa0ffa0, 0xa0a0ff,
            };
            static int col_indx=0;
            if(++col_indx==sizeof(colors)/sizeof(dword)) col_indx=0;
            sel_icon_fade.Setup(240,
               mission->game_menu.GetPrgKeyColor(ma->actor->GetMenuID(), selected_prg_key),
               colors[col_indx]|0xff000000);
               //Random(0xffffff)|0xff000000);
         }
         mission->game_menu.SetPrgKeyColor(ma->actor->GetMenuID(), selected_prg_key,
            sel_icon_fade.Tick(tc.time));
      }

      {
                              //flash path color
         if(path_fade.IsEnd()){
            static const dword colors[] = {
               0xff2020, 0xff8828,
            };
            static int col_indx = 0;
            int old_indx = col_indx;
            if(++col_indx==sizeof(colors)/sizeof(dword))
               col_indx = 0;
            path_fade.Setup(500,
               colors[old_indx]|0xff000000,
               colors[col_indx]|0xff000000);
         }
         color_path_fade = path_fade.Tick(tc.time);
      }

                              //interpolate camera
      if(cam_intp && cam_intp->Tick(tc.time)==I3D_DONE)
         cam_intp = NULL;
   }

//----------------------------

   virtual void Render(){

      driver->BeginScene();
      C_mission::Render();

                              //get move path
      LPC_map_actor ma=FindActiveActor();
      if(!ma) return;
      LPC_actor act=ma->actor;
      int prg_size=act->cbProc(CB_GET_PRG_SIZE);
      int i;

      vector<S_vector> path;
      vector<word> indx;
//----------------------------
      const S_matrix &m = scene->GetActiveCamera()->GetMatrix();
      for(i=0; i<prg_size; i++){
         E_PRG_ITEM itm=(E_PRG_ITEM)act->cbProc(CB_GET_PRG, i);
         switch(itm){
         case PRG_MOVE:
            {
               vector<S_vector> mp;
                              //get proposed move path from actor
               act->cbProc(CB_GET_PRG_DATA, i, (dword)&mp);
               if(mp.size()){
                  int sz = mp.size();
                  for(int j=0; j<sz-1; j++){
                     const S_vector &s0 = mp[j];
                     const S_vector &s1 = mp[j+1];
                     const S_vector line_dir(s1 - s0);
                     S_vector right = (m(3)-s0).Cross(line_dir);
                     right.Normalize();
                     right *= .2f;
                     
                     S_vector v0 = s0 - right;
                     S_vector v1 = s0 + right * 2.0f;
                     S_vector v2 = s1 - right;
                     S_vector v3 = s1 + right * 2.0f;

                              //blend with previous segment
                     if(path.size()){
                        S_vector &lv0 = *(path.end()-2);
                        //S_vector &lv1 = *(path.end()-1);
                        int ii = path.size();
                        float pol = lv0.PositionOnLine(s0, line_dir);
                        if(pol < 0.0f){
                           //v1 = (lv1 + v1) * .5f; lv1 = v1;
                           indx.push_back(ii-2);
                           indx.push_back(ii-0);
                           indx.push_back(ii-1);
                        }else{
                           //v0 = (lv0 + v0) * .5f; lv0 = v0;
                           indx.push_back(ii+0);
                           indx.push_back(ii+1);
                           indx.push_back(ii-1);
                        }
                     }

                     int ii = path.size();
                     indx.push_back(ii+0);
                     indx.push_back(ii+2);
                     indx.push_back(ii+1);
                     indx.push_back(ii+1);
                     indx.push_back(ii+2);
                     indx.push_back(ii+3);
                     path.push_back(v0);
                     path.push_back(v1);
                     path.push_back(v2);
                     path.push_back(v3);
                  }
               }
            }
            break;
         }
      }
      if(indx.size()){
         scene->SetRenderMatrix(GetIdentityMatrix());
         driver->SetTexture(NULL);

         dword &col = color_path_fade;
         S_vector color(((col>>16)&255)/255.0f, ((col>>8)&255)/255.0f, (col&255)/255.0f);
         //byte alpha = 160;
         byte alpha = 0;
         scene->DrawTriangles(path.begin(), path.size(), indx.begin(), indx.size(), color, sizeof(S_vector), alpha);
      }
//----------------------------
      /*
      for(i=0; i<prg_size; i++){
         E_PRG_ITEM itm=(E_PRG_ITEM)act->cbProc(CB_GET_PRG, i);
         switch(itm){
         case PRG_MOVE:
            {
                              //get proposed move path from actor
               int old_nv = path.size();
               act->cbProc(CB_GET_PRG_DATA, i, (dword)&path);
               if(path.size()){
                              //create indicies
                  for(int j=old_nv; j<path.size()-1; j++){
                     indx.push_back(j);
                     indx.push_back(j+1);
                  }
               }
            }
            break;
         }
      }
      if(indx.size()){
         scene->SetRenderMatrix(GetIdentityMatrix());

         dword &col = color_path_fade;
         S_vector color(((col>>16)&255)/255.0f, ((col>>8)&255)/255.0f, (col&255)/255.0f);
         scene->DrawLines(&path[0], path.size(), &indx[0], indx.size(), color);
      }
      */

      mission->game_menu.Render();

      bool use_zb = driver->GetState(RS_USEZB);
      if(use_zb) driver->SetState(RS_USEZB, false);

      for(i=0; i<MAP_NUM_CTRL_ICONS; i++)
         ctrl_icons[i]->Render();
      for(i=prg_icons.size(); i--; )
         prg_icons[i]->Render();

      texts.Render();
      //mission->texts.Render();

      spr_arrow->Render();

      if(use_zb) driver->SetState(RS_USEZB, true);

      if(sync_menu)
         sync_menu->Render();

      driver->EndScene();
   }

//----------------------------

   virtual void NetIn(E_NETWORK_MESSAGE msg_id, PC_net_in_msg in_msg){

      switch(msg_id){
      case C_game_mission::NM_MAP:
         net_sync->NetIn(in_msg, true);
         break;

      default:
         assert(0);
         in_msg->SkipChunk();
      }
   }

//----------------------------

   virtual void OnPlayerDisconnect(dword pid, const char *name){

      net_sync->OnPlayerDisconnect(pid);
      C_mission::OnPlayerDisconnect(pid, name);
   }

//----------------------------

   static I3DENUMRET I3DAPI cbZeroData(PI3D_frame frm, dword c){
      frm->SetData(0);
      return I3DENUMRET_OK;
   }
};

//----------------------------

bool C_map_manager_imp::Init(){

   int i;

   C_string m_name = mission->GetMissionName();
   if(m_name.Size()<5 || stricmp(&m_name[(int)m_name.Size()-5], "\\Game"))
      return false;
   m_name[(int)m_name.Size()-5] = 0;

   E_MISSION_IO mio = Open(C_fstring("%s\\Map", (const char*)m_name), OPEN_USE_SCENE_CAMERA | OPEN_NO_OPEN_FAIL_MSG);
   if(mio!=MIO_OK)
      return false;
                           //init camera
   PI3D_camera map_cam = driver->CreateCamera();
   map_cam->SetFOV(65.0f*PI/180.0f);
   map_cam->SetName("map_camera");
   scene->AddFrame(map_cam);
   map_cam->SetRange(CAM_NEAR_RANGE, cam_range);
   scene->SetActiveCamera(map_cam);
   map_cam->Release();

   mouse_x = igraph->Scrn_sx()/2;
   mouse_y = igraph->Scrn_sy()/2;


   curr_floor = 0;
                           //read map init file
   C_cache is;
   C_fstring ini_name("missions\\%s\\map\\map.ini", (const char*)m_name);
   is.open(ini_name, CACHE_READ);

   if(is.good()){
      char line[256];
      char *cp;
      while(!is.eof()){
         is.getline(line, 256);
         if(line[0]==';' || !line[0])
            continue;
         cp = line;
         switch(GetKeyWord(cp,
            "icons\0"
            "start_floor\0"
            "floor\0"
            )){
         case 0:
                              //available icons - ignored
            break;

         case 1:              //default floor
            ReadInt(cp, curr_floor);
            break;

         case 2:              //floor definition
            {
               vector<int> vi;
               int pi;
               while(ReadInt(cp, pi))
                  vi.push_back(pi);
               floor_def.push_back(vi);
            }
            break;
         }
      }
      is.close();
   }//else
      //ErrReport(C_fstring("can't load '%s'", (const char*)ini_name), editor);

   curr_floor = Max(0, Min((int)floor_def.size()-1, (int)curr_floor));

                              //find all floor parts
   for(i=0; ; i++){
      PI3D_frame frm = scene->FindFrame(C_fstring("area %.2i", i));
      if(!frm)
         break;
      frm->SetOn(false);
   }
                           //switch on current floor
   if(curr_floor < floor_def.size())
      ShowFloor(curr_floor, true);

                           //init controll icons
   for(i=0; i<MAP_NUM_CTRL_ICONS; i++){
      C_poly_sprite *spr = new C_poly_sprite(map_img_pos[i][0], map_img_pos[i][1], map_img_pos[i][2], &icons_img, 8+i);
      spr->SetColor(MAP_BUTTON_INACTIVE);
      ctrl_icons[i] = spr;
      spr->Release();
   }

                           //init programming icons
   for(i=0; i<sizeof(prg_icon_id)/sizeof(E_PRG_ITEM); i++){
      C_poly_sprite *spr = new C_poly_sprite(MAP_PRG_X(i), MAP_PRG_Y(i), MAP_PRG_SIZE, &icons_img, 16 + prg_icon_id[i]);
      spr->SetColor(MAP_ICON_INACTIVE);
      prg_icons.push_back(spr);
      spr->Release();
   }
   last_mx = igraph->Mouse_x();
   last_my = igraph->Mouse_y();
   flash_ctrl_button_id = -1;

                              //load sounds
   if(isound){
      for(i=0; i<MAP_SOUND_LAST; i++){
         PISND_source src;
         isound->CreateSource(&src);
         PISND_sound sp;
         isound->CreateSound(&sp);
         snd[i] = sp;
         sp->Release();
         {
            ISND_RESULT ir;
            ir = src->Open(C_fstring("Sounds\\Common\\map_fx%.2i.ogg", i), ISND_NOPANNING);
            if(ISND_FAIL(ir))
               ir = src->Open(C_fstring("Sounds\\Common\\map_fx%.2i.wav", i), ISND_NOPANNING);
            if(ISND_FAIL(ir)){
               ErrReport(C_fstring("Can't open sound: 'map_fx%.2i'", i), editor);
            }
            snd[i]->Open(src);
            snd[i]->SetVolume(.5f);
         }
         src->Release();
      }
   }
   texts.Init();
                           //get bounding space
   PI3D_dummy mr = I3DCAST_DUMMY(scene->FindFrame("map 00", ENUMF_DUMMY));
   if(mr){
      world_box = *mr->GetBBox();
      const S_matrix *mat;
      mat = &mr->GetMatrix();
      world_box.min = world_box.min * *mat;
      world_box.max = world_box.max * *mat;
   }else{
                              //consider entire geometry as bounding space
      I3D_bound_volume bvol;
      scene->GetPrimarySector()->ComputeHRBoundVolume(&bvol);
      world_box = bvol.bbox;
   }
   map_zoom_max = Max(world_box.max.x-world_box.min.x, world_box.max.z-world_box.min.z);
   map_zoom_max = Max(MAP_ZOOM_MIN, map_zoom_max);
   pan_speed = (world_box.max-world_box.min).Magnitude() / 80.0f;

   SetCursor(0);

                              //init values
   look_at.Zero();
   look_angle = 0.0f;
   look_dist = MAP_ZOOM_MIN + (map_zoom_max - MAP_ZOOM_MIN) * .5f;
   PositionCamera();

   flash_prg_icon_id = -1;
   last_mouse_butt = 0;
   prg_item_hint_count = 0;

   return true;
}

//----------------------------

void C_map_manager_imp::Activate(){

   int i;
   prg_icon_dest = -1;
   selected_prg_key = -1;
   sel_actor = -1;
   mode = MODE_NORMAL;

   const S_vector &dir = mission->GetScene()->GetActiveCamera()->GetDir();
   float dir_angle = Angle2D(S_vector2(dir.x, dir.z), S_vector2(0, 0));
   mission->game_menu.SetCompassAngle(-dir_angle);

                              //enum actors, put them into the list
   struct S_help{
      static bool cbEnum(LPC_actor act, dword a){
         if(act->IsOnMap()){
                              //only actors with models may be on map
            PI3D_frame frm = act->GetFrame();
            if(frm->GetType()==FRAME_MODEL){

               vector<C_smart_ptr<C_map_actor> > *actors = (vector<C_smart_ptr<C_map_actor> >*)a;
               LPC_map_actor ma = new C_map_actor;
               ma->actor = act;
               actors->push_back(ma);
               ma->Release();
            }
         }
         return true;
      };
   };
   mission->EnumActors(S_help::cbEnum, (dword)&actors);

                              //remove unseen enemies
   for(i=actors.size(); i--; ){
      LPC_actor a = actors[i]->actor;
      bool on = true;
      switch(a->GetType()){
      case ACTOR_ENEMY:
         {
            on = false;
            if(a->GetFrame()->IsOn())
            for(int j=actors.size(); !on && j--; ){
               LPC_actor a1 = actors[j]->actor;
               if(a1->GetType()==ACTOR_PLAYER && a1->IsAlive()){
                              //check if he's seeing him
                  const S_vector &from = *a1->GetHeadPos();
                  const S_vector &to = *a->GetBrestPos();
                  S_vector dir = to-from;
                                 //check view angle, can't see behind
                  float angle = dir.AngleTo(- *a1->GetHeadDir());
                  if(angle<PI*.5f){
   #if 0
                     DebugLine(from, from+dir, 0);
   #endif
                     I3D_collision_data cd(from, dir, 0, a1->GetFrame(), WatchColResp2);
                     if(mission->GetScene()->TestCollision(cd)){
                        PI3D_frame vol = cd.GetHitFrm();
                        const S_vector &ret_pos = cd.GetHitPosition();

                        LPC_actor act = (LPC_actor)vol->GetData();
                        if(act==a){
                           float max_dist = mission->tab_property->ItemF(TAB_F_AI_WATCH_MAX_DIST);
                           if(!max_dist) max_dist = tab_game_cfg->ItemF(TAB_F_AI_WATCH_MAX_DIST);
                           if((ret_pos-from).Magnitude() < max_dist*.9f){
                              on = true;
                           }
                        }
                     }else{
                                 //no collision = visible
                        on = true;
                     }
                  }
               }
            }
         }
         break;
      }
      if(on){
         switch(a->GetType()){
         case ACTOR_PLAYER:
            break;
         default:
            {
                              //check if actor is inside of map rectangle
               const S_vector &pos = a->GetFrame()->GetWorldPos();
               on = (pos.x > world_box.min.x &&
                  pos.z > world_box.min.z &&
                  pos.x < world_box.max.x &&
                  pos.z < world_box.max.z);
            }
         }
      }
      if(!on){
         actors[i] = actors.back(); actors.pop_back();
      }      
   }

   LPC_map_actor ma_active = NULL;

   for(i=0; i<actors.size(); i++){
      LPC_map_actor ma = actors[i];
      PI3D_frame frm = ma->actor->GetFrame();
      assert(frm->GetType()==FRAME_MODEL);
      PI3D_model mp = I3DCAST_MODEL(frm);
      PI3D_model model = driver->CreateModel();
      model->Duplicate(mp);
      model->EnumFrames(cbZeroData, 0);
      cbZeroData(model, 0);

      model->SetRot(mp->GetWorldRot());
      model->SetScale(mp->GetWorldScale());
      model->SetPos(mp->GetWorldPos());
      model->SetAnimation(0, NULL);
      model->Update();

      LinkModel(model, ma);
      ma->SetModel(model, -1, scene);
      model->Release();
   }

   for(i=0; i<actors.size(); i++){
      LPC_map_actor ma = actors[i];
      int type = -1;
      switch(ma->actor->GetType()){
      case ACTOR_PLAYER:
         type = 0;
         break;
      case ACTOR_ENEMY:
         {
            //int eg = ma->actor->cbProc(CB_GETENEMYGROUP);
            int eg = ma->actor->GetEnemyGroup();
            type = eg==0 ? 1 : 2;
         }
         break;
      }
      if(type!=-1)
         if(!ma->actor->IsAlive())
            type += 3;
      ma->SetModel(ma->GetModel(), type, scene);
      ma->GetModel()->LinkTo(scene->GetPrimarySector());

      if(ma->actor->IsActive()){
         ma_active = ma;
         ma->SetActive(true, scene);
      }
   }

   if(ma_active){
      SetActive(ma_active, false);
   }

                              //add models to map
                              //get all models first
   struct S_hlp{
      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
         vector<PI3D_model> &lst = *(vector<PI3D_model>*)c;
         lst.push_back(I3DCAST_MODEL(frm));
         return I3DENUMRET_OK;
      }
   };
   vector<PI3D_model> lst;
   mission->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&lst, ENUMF_MODEL);
                              //now add models with special flags set
   for(i=lst.size(); i--; ){
      PI3D_model mp = lst[i];
      dword flgs = mp->GetUserFlags();
      if(flgs&1){
         PI3D_model mod = driver->CreateModel();
         mod->Duplicate(mp);

         mod->EnumFrames(cbZeroData, 0);
         cbZeroData(mod, 0);

         mod->SetPos(mp->GetWorldPos());
         mod->SetRot(mp->GetWorldRot());
         mod->SetScale(mp->GetWorldScale());
         mod->SetAnimation(0, NULL);
         mod->Update();
         LinkModel(mod);
         map_models.push_back(mod);
         mod->Release();
      }
   }

   PositionCamera();
   SetCursor(0, true);

   if(floor_def.size()){
      ShowFloor(curr_floor, false);
      ShowFloor(curr_floor, true);
   }
}

//----------------------------
//----------------------------

C_map_manager *CreateMap(C_game_mission &gm){

   RenderAllBackbuffers();
   C_map_manager_imp *map = new C_map_manager_imp(&gm);
   if(!map->Init()){
      map->Release();
      map = NULL;
   }
   return map;
}

//----------------------------
//----------------------------
