#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"


//----------------------------
                              //mine over network:
                              // - sent over network:
                              //    - byte id
                              //    - word net_id
                              //    - S_vector pos
                              // - sent over network (not at creation time):
                              //    - explosion synchronization
#define MINE_RANGE 10.0f
#define MINE_POWER 2000

class C_mine: public C_actor{
   LPC_actor who_layed_it;    //used to show killer
   dword inv_id;
   C_smart_ptr<C_net_sync> net_sync;
   bool runtime_created;
   int init_delay;            //delay after installing

                              //network_actor: no
   void Explode(){
                           //create explosion
      S_explosion_init ei;
      ei.pos = frame->GetWorldPos();
      ei.id = tab_inventory->ItemI(TAB_I_INV_MINE_EXPL, inv_id);
      ei.actor_reason = who_layed_it;
      mission.CreateActor(ACTOR_EXPLOSION, (dword)&ei);
   }

   void SetupModel(){
                              //install collision callback
      PI3D_volume vol = I3DCAST_VOLUME(model->FindChildFrame(NULL, ENUMF_VOLUME));
      if(vol){
                              //set special type of collisions
         vol->SetData((dword)this);
         vol->SetCollisionMaterial(40);
         vol->SetScene(mission.GetScene());
         vol->SetOn(true);
      }
   }
public:
   float explode_range;
   int destruct_power;


   C_mine(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_MINE),
      explode_range(MINE_RANGE), destruct_power(MINE_POWER),
      init_delay(0),
      inv_id(9),
      runtime_created(init_data!=NULL),
      who_layed_it(NULL)
   {
                              //allow init data to be NULL, used during LoadGame or for in-scene mines
      if(!init_data)
         return;

      const S_mine_init &mi = *(S_mine_init*)init_data;

      inv_id = mi.id;
      init_delay = 1000;

      frame = model = driver->CreateModel();

      const char *name = tab_inventory->ItemS(TAB_S16_INV_MODEL_S, inv_id, 16);
      model_cache.Open(model, name, mission.GetScene(), 0, ErrReport, editor);
      //S_model_load s = {model, &mission};
      //model->EnumFrames(cbModelLoad, (dword)&s, ENUMF_DUMMY | ENUMF_USEMASK, "~0*");
      model->EnumFrames(cbAddVolumeToScene, (dword)mission.GetScene(), ENUMF_VOLUME);

      network_actor = mi.netw_actor;

      /*
                              //dig into earth
      S_vector r_pos, r_norm;
      I3D_collision_data cd(mi.pos, S_vector(0, -1, 0), 0);
      if(mission.GetScene()->TestCollision(cd)){
         r_pos = cd.GetHitPosition();
         r_norm = cd.GetHitNormal();
      }else
         r_pos = mi.pos;
      frame->SetPos(r_pos);
      frame->Update();
      mission.GetScene()->SetFrameSector(frame);
      */
      mission.LinkToFloor(frame, mi.pos, frame, true);

      SetupModel();

      if(net && !network_actor){
                              //assign ID
         mission.CreateActorNetID(this);

         bool g = true;
         net_msg->AddWord(C_game_mission::NM_GAME_CREATE_MINE, g);
         net_msg->AddByte(mi.id, g);
         net_msg->AddWord(net_id, g);
         net_msg->AddVector(mi.pos, g);
      }
   }

   ~C_mine(){
      PI3D_volume vol = I3DCAST_VOLUME(model->FindChildFrame(NULL, ENUMF_VOLUME));
      if(vol)
         vol->SetData(NULL);
   }

//----------------------------

   virtual void SetNetID(word id){
      
      net_sync = CreateNetSync(id);
      net_sync->Release();
      C_actor::SetNetID(id);
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      ck->Write(&runtime_created, sizeof(runtime_created));
      if(runtime_created)
         SG_SaveModelInfo(ck);
                           //write data
      ck->Write(&inv_id, sizeof(inv_id));
      ck->Write(&explode_range, sizeof(explode_range));
      ck->Write(&destruct_power, sizeof(destruct_power));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      ck->Read(&runtime_created, sizeof(runtime_created));
      if(runtime_created)
         SG_LoadModelInfo(ck);
                           //read data
      ck->Read(&inv_id, sizeof(inv_id));
      ck->Read(&explode_range, sizeof(explode_range));
      ck->Read(&destruct_power, sizeof(destruct_power));
      SetupModel();

      return true;
   }

//----------------------------

#define MINE_EXPLODE_TIME 300

   dword C_mine::cbProc(dword msg, dword prm1, dword prm2){

      switch(msg){

      case CB_COLLISION:
         {
            if(init_delay)
               break;
                              //anti-tank mines may only be triggered by vehicles
            if(inv_id==8){
               switch(((LPC_actor)prm1)->GetType()){
               case ACTOR_AUTOMOBIL:
               case ACTOR_AUTO_CANNON:
                  break;
               default:
                  return 0;
               }
            }
                              //we've hit explosion mechanism,
                              // so explode with full power
            return Explode((LPC_actor)prm1, 10000);
         }
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual bool MayExplode() const{ return true; }

//----------------------------

   virtual bool Explode(C_actor *caller, int power){

      if(power<4000)
         return false;
      if(!net){
                           //explode directly
         Explode();
         mission.DestroyActor(this);
      }else
      if(!net_sync->GetState()){
         net_sync->BeginEvent();
      }
      return true;
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){
      net_sync->NetIn(in_msg);
   }

//----------------------------

   virtual void NetPlrDisconnect(dword pid){

      net_sync->OnPlayerDisconnect(pid);
      C_actor::NetPlrDisconnect(pid);
   }

//----------------------------

   virtual dword Tick(const S_tick_context &tc){

      if(init_delay){
         if((init_delay -= tc.time) < 0)
            init_delay = 0;
      }
      if(net){
         dword winner_pid;
         if(net_sync->GetState(&winner_pid)==C_net_sync::SS_RESOLVED){
            if(winner_pid==net->GetPID()){
               Explode();
            }
            mission.DestroyActor(this);
         }
      }
      return 0;
   }

//----------------------------

   virtual bool IsOnMap() const{ return true; }

//----------------------------

   virtual void TableUpdate(int index, bool not_load_prg){

      if(model)
         SetupModel();
   }

};

//----------------------------

LPC_actor CreateMine(C_game_mission &gm, void *init_data){

   return new C_mine(gm, init_data);
}

//----------------------------
