#include "all.h"
#include "loader.h"
#include <windows.h>
#include <commctrl.h>
#include "resource.h"
#include "winstuff.h"
#include "tables.h"
#include "actors.h"
#include <malloc.h>
#include <map>
#include <tabler2.h>
#include <insanity\sourcecontrol.h>

//#define SAFE_FRAME 10         //draw safe frame with specified size

//----------------------------


#ifdef EDITOR

extern const char *actor_names[];

//----------------------------

static bool GetFileName(const char *title, const char *filter,
   const char *def_ext, char *buf1, dword bufsize, const char *init_dir){

   OPENFILENAME on;
   memset(&on, 0, sizeof(on));
   on.lStructSize = sizeof(on);
   on.hwndOwner = (HWND)igraph->GetHWND();
   on.lpstrFilter = filter;
   on.nFilterIndex = 1;
   on.lpstrFile = buf1;
   buf1[0] = 0;
   on.nMaxFile = bufsize;
   on.lpstrInitialDir = init_dir;
   on.lpstrTitle = title;
   on.Flags = OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
   on.lpstrDefExt = def_ext;
   return (GetOpenFileName(&on)!=0);
}

//----------------------------

static const float time_scales[] = {0.0f, .05f, .1f, .2f, .5f, 1.0f, 2.0f, 5.0f, 10.0f};

static const int simulate_fps_table[] = {0, 5, 10, 15, 20, 50};

//----------------------------

static int GetSimFpsID(int sim_fps){

   int fps_table_size = sizeof(simulate_fps_table) / sizeof(int);
   for(int i = fps_table_size; i--; )
      if(simulate_fps_table[i] == sim_fps)
         return i;
   return 0;
}

//----------------------------
                              //editor plugin: Mission
class C_edit_Mission: public C_editor_item{
   virtual const char *GetName() const{ return "Mission"; }

   C_string mission_name;
   C_string last_mission_name;

   C_smart_ptr<C_game_mission> mission;
   C_smart_ptr<C_editor_item> e_slct;
   
   bool make_log;
   bool ask_to_save;
   bool debug_cam_acq;
   bool enabled;              //plugin enabled (failed if lock-ed)
   int time_scale_id;
   int simulate_fps;          //(0 for disable)
   float sim_fps_last_timer;
   bool keep_checked_out;

                              //flags for merging:
   dword merge_load;
   dword merge_erase;

//----------------------------

   static void AddDirs(HWND hwnd_tc, const C_string &root_dir, int base_dir_size,
      const C_string &curr_dir, HTREEITEM &curr_i, HTREEITEM tc_parent = TVI_ROOT){

      WIN32_FIND_DATA fd;
      HANDLE h = FindFirstFile(C_fstring("%s*.*", (const char*)root_dir), &fd);
      if(h!=INVALID_HANDLE_VALUE)
      do{
         if(fd.cFileName[0]=='.')
            continue;
         if(!(fd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY))
            continue;
                              //make it nice - to lower case, except first letters
         for(int i=strlen(fd.cFileName); i--; )
            fd.cFileName[i] = (!i ? toupper : tolower)(fd.cFileName[i]);

         bool has_mission = false;
         int hh = dtaOpen(C_fstring("%s%s\\scene.i3d", (const char*)root_dir, fd.cFileName));
         if(hh!=-1){
            dtaClose(hh);
            has_mission = true;
         }
         C_fstring str("%s%s", (const char*)&root_dir[base_dir_size], fd.cFileName);
         bool recurse = (str.Size()<curr_dir.Size() && !strnicmp(str, curr_dir, str.Size()));
                           //add this dir
         TVINSERTSTRUCT itm;
         memset(&itm, 0, sizeof(itm));
         itm.hParent = tc_parent;
         itm.hInsertAfter = TVI_SORT;
         itm.item.mask = TVIF_TEXT;
         itm.item.pszText = fd.cFileName;
         itm.item.state = 0;
         itm.item.stateMask = 0xff;
         if(has_mission){
            itm.item.mask |= TVIF_STATE;
            itm.item.state |= TVIS_BOLD;
         }
         bool active = str.Matchi(curr_dir);
         if(recurse){
            itm.item.mask |= TVIF_STATE;
            itm.item.state |= TVIS_EXPANDED;
         }
         HTREEITEM hi = (HTREEITEM)SendMessage(hwnd_tc, TVM_INSERTITEM, 0, (LPARAM)&itm);
         if(active){
            SendMessage(hwnd_tc, TVM_SELECTITEM, TVGN_CARET, (LPARAM)hi);
            curr_i = hi;
         }
                              // //if our item may be inside of this dir, go to sub-dir
         AddDirs(hwnd_tc, C_fstring("%s%s\\", (const char*)root_dir, fd.cFileName),
            base_dir_size, curr_dir, curr_i, hi);
      }while(FindNextFile(h, &fd));
      FindClose(h);
   }

//----------------------------

   static bool GetFullDir(HWND hwnd, C_string &str, bool &is_valid){

      is_valid = true;
      str = NULL;
      HTREEITEM hi = (HTREEITEM)SendDlgItemMessage(hwnd, IDC_LIST1, TVM_GETNEXTITEM, TVGN_CARET, 0);
      if(hi){
         do{
            TVITEM tvi;
            char buf[MAX_PATH];
            tvi.mask = TVIF_HANDLE | TVIF_TEXT | TVIF_STATE;
            tvi.hItem = hi;
            tvi.stateMask = 0xff;
            tvi.pszText = buf;
            tvi.cchTextMax = sizeof(buf);
            if(SendDlgItemMessage(hwnd, IDC_LIST1, TVM_GETITEM, 0, (LPARAM)&tvi)){
               if(!str.Size() && !(tvi.state&TVIS_BOLD))
                  is_valid = false;
               str = C_fstring("%s%s%s", buf, str.Size() ? "\\" : "", (const char*)str);
            }else
               return false;
         }while(hi = (HTREEITEM)SendDlgItemMessage(hwnd, IDC_LIST1, TVM_GETNEXTITEM, TVGN_PARENT, (LPARAM)hi));
         return true;
      }
      return false;
   }

//----------------------------
// Process mission selection. The returned value is:
// 0 - some error occured
// 1 - mission name saved
// 2 - mission name the same
   static int Clicked(HWND hwnd){

      C_string dir;
      bool is_valid;
      if(!GetFullDir(hwnd, dir, is_valid) || !is_valid)
         return false;

      C_edit_Mission *em = (C_edit_Mission*)GetWindowLong(hwnd, GWL_USERDATA);
      assert(em);
      if(!em->mission_name.Matchi(dir)){
         em->mission_name = dir;
         return 1;
      }
      return 2;
   }

//----------------------------

   static BOOL CALLBACK dlgSelect(HWND hwnd, UINT uiMsg, WPARAM wParam, LPARAM lParam){

      switch(uiMsg){
      case WM_INITDIALOG:
         {
            SetWindowLong(hwnd, GWL_USERDATA, lParam);
            C_edit_Mission *em = (C_edit_Mission*)lParam;
            WinInitDialog(hwnd);

            SendDlgItemMessage(hwnd, IDC_STATIC_CURR_DIR, WM_SETTEXT, 0, (LPARAM)(const char*)em->mission_name);
            if(em->last_mission_name.Size())
               SendDlgItemMessage(hwnd, IDC_LAST_MISSION, WM_SETTEXT, 0, (LPARAM)(const char*)em->last_mission_name);
            else
               EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_OPEN_LAST), false);

            HTREEITEM curr_i = NULL;
            C_fstring root("missions\\");
            AddDirs(GetDlgItem(hwnd, IDC_LIST1),
               root, root.Size(),
               em->mission_name, curr_i);
            if(curr_i)
               SendDlgItemMessage(hwnd, IDC_LIST1, TVM_ENSUREVISIBLE, 0, (LPARAM)curr_i);

            ShowWindow(hwnd, SW_SHOW);
         }
         return 1;

      case WM_COMMAND:
         switch(LOWORD(wParam)){
         case IDCANCEL:
            EndDialog(hwnd, 0);
            break;
         case IDOK:
            {
               switch(Clicked(hwnd)){
               case 0:
                              //expand/collapse
                  {
                     HTREEITEM hi = (HTREEITEM)SendDlgItemMessage(hwnd, IDC_LIST1, TVM_GETNEXTITEM, TVGN_CARET, 0);
                     if(hi){
                        SendDlgItemMessage(hwnd, IDC_LIST1, TVM_EXPAND, TVE_TOGGLE, (LPARAM)hi);
                     }
                  }
                  break;
               case 1: EndDialog(hwnd, 1); break;
               case 2: EndDialog(hwnd, 2); break;
               }
            }
            break;
         case IDC_BUTTON_OPEN_LAST:
            {
               C_edit_Mission *em = (C_edit_Mission*)GetWindowLong(hwnd, GWL_USERDATA);
               assert(em);
               em->mission_name = em->last_mission_name;
               EndDialog(hwnd, 1);
            }
            break;
         }
         break;

      case WM_NOTIFY:
         if(wParam==IDC_LIST1){
            LPNMHDR hdr = (LPNMHDR)lParam;
            switch(hdr->code){
            case NM_DBLCLK:
               {
                  switch(Clicked(hwnd)){
                  case 1: 
                     EndDialog(hwnd, 1);
                     break;
                  case 2:
                     EndDialog(hwnd, 2);
                     break;
                  }
               }
               break;
            case TVN_SELCHANGED:
               {
                  C_string dir;
                  bool is_valid;
                  GetFullDir(hwnd, dir, is_valid);
                  SendDlgItemMessage(hwnd, IDC_STATIC_NEW_DIR, WM_SETTEXT, 0, (LPARAM)(const char*)dir);
               }
               break;
            }
         }
         break;
      }
      return 0;
   }

//----------------------------

   void GetDlgButtonState(HWND hwnd){

      dword ml = 0, me = 0;

      if(IsDlgButtonChecked(hwnd, IDC_MRG_SETTINGS)) ml |= OPEN_MERGE_SETTINGS;

      if(IsDlgButtonChecked(hwnd, IDC_MRG_LIGHT_L)) ml |= OPEN_MERGE_LIGHTS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_SOUND_L)) ml |= OPEN_MERGE_SOUNDS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_MODEL_L)) ml |= OPEN_MERGE_MODELS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_VOLUME_L)) ml |= OPEN_MERGE_VOLUMES;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_DUMMY_L)) ml |= OPEN_MERGE_DUMMYS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_CAMERA_L)) ml |= OPEN_MERGE_CAMERAS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_TARGET_L)) ml |= OPEN_MERGE_TARGETS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_OCCLUDER_L)) ml |= OPEN_MERGE_OCCLUDERS;

      if(IsDlgButtonChecked(hwnd, IDC_MRG_LMAP)) ml |= OPEN_MERGE_LIGHTMAPS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_SCRIPT_L)) ml |= OPEN_MERGE_SCRIPTS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_CHECKPOINTS)) ml |= OPEN_MERGE_CHECKPOINTS;

      merge_load = ml;

      if(IsDlgButtonChecked(hwnd, IDC_MRG_LIGHT_E)) me |= OPEN_MERGE_LIGHTS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_SOUND_E)) me |= OPEN_MERGE_SOUNDS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_MODEL_E)) me |= OPEN_MERGE_MODELS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_VOLUME_E)) me |= OPEN_MERGE_VOLUMES;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_DUMMY_E)) me |= OPEN_MERGE_DUMMYS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_CAMERA_E)) me |= OPEN_MERGE_CAMERAS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_TARGET_E)) me |= OPEN_MERGE_TARGETS;
      if(IsDlgButtonChecked(hwnd, IDC_MRG_OCCLUDER_E)) me |= OPEN_MERGE_OCCLUDERS;

      if(IsDlgButtonChecked(hwnd, IDC_MRG_SCRIPT_E)) me |= OPEN_MERGE_SCRIPTS;

      merge_erase = me;
   }


//----------------------------

   int tab_pos_size[4];

   int last_create_actor_index;


   struct S_name_texture{
                              //name of actor type stored in texture - for editing purposes
      C_smart_ptr<I3D_texture> tp;
                              //size of initialized part of texture
      int tp_size[2];
      S_name_texture(){}
      S_name_texture(const S_name_texture &nt){ operator =(nt); }
      S_name_texture &operator =(const S_name_texture &nt){
         tp = nt.tp;
         memcpy(tp_size, nt.tp_size, sizeof(tp_size));
         return *this;
      }
   };
   map<E_ACTOR_TYPE, S_name_texture> actor_name_textures;
   void *h_font;
   bool draw_actor_names;
   int sb_rt_msg_pos;         //position in statusbar for run-time error messages

                              //property template definition
   enum{
      TAB_CV_SCN_BGND = 0,          //bgnd color
      
      TAB_I_SCN_CAM_FOV = 2,        //camera FOV
      TAB_F_SCN_CAM_RANGE = 3,      //camera far range
      TAB_B_SCN_CAM_ORTHO = 4,      //orthogonal camera view
      TAB_F_SCN_CAM_ORTHO_SCL = 5,  //orthogonal camera scale
      TAB_F_SCN_BDROP_RANGE_N = 6,  //backdrop near range
      TAB_F_SCN_BDROP_RANGE_F = 7,  //backdrop far range
   };

   static LPC_table_template GetPropTemplate(){
      static C_table_template templ_scene_property = {
         { 0xbd765780, 0x3c52, 0x11d2, { 0x9b, 0x9d, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
         "Scene properties",
         NULL,
         {
            {TE_BRANCH, 0, "Environment", 2},
               {TE_COLOR_VECTOR, TAB_CV_SCN_BGND,     "Background color"},
               {TE_BRANCH, 0, "Backdrop range", 2, (dword)"n: %[0], f: %[1]"},
                  {TE_FLOAT, TAB_F_SCN_BDROP_RANGE_N, "near", 0, 0, 1},
                  {TE_FLOAT, TAB_F_SCN_BDROP_RANGE_F, "far", 0, 0, 100},
            {TE_BRANCH, 0, "Camera", 4, (int)"FOV=%[0], range=%[1]"},
               {TE_INT, TAB_I_SCN_CAM_FOV, "FOV", 10, 135},
               {TE_FLOAT, TAB_F_SCN_CAM_RANGE, "Range"},
               {TE_BOOL, TAB_B_SCN_CAM_ORTHO, "Orthogonal"},
               {TE_FLOAT, TAB_F_SCN_CAM_ORTHO_SCL, "Orthogonal scale", 0, 0, 1},
            {TE_NULL}
         }
      };
      return &templ_scene_property;
   }

   static void cbProp(LPC_table tab, dword msg, dword cb_user, dword prm2, dword prm3){

      switch(msg){
      case TCM_CLOSE:
         {
            C_edit_Mission *em = (C_edit_Mission*)cb_user;
            //em->ed->GetIGraph()->DeleteDlgHWND(em->hwnd_prop);
            //em->hwnd_prop = NULL;
            memcpy(em->tab_pos_size, (LPRECT)prm2, sizeof(int)*4);
         }
         break;
      case TCM_MODIFY:
         {
            C_edit_Mission *em = (C_edit_Mission*)cb_user;
            PI3D_scene scene = em->ed->GetScene();
            switch(prm2){
            case TAB_CV_SCN_BGND:
               scene->SetBgndColor(*(S_vector*)tab->Item(TAB_CV_SCN_BGND));
               em->mission->GetScene()->Render(0, &em->mission);
               em->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
               em->ed->SetModified();
               break;
            case TAB_I_SCN_CAM_FOV:
            case TAB_F_SCN_CAM_RANGE:
            case TAB_B_SCN_CAM_ORTHO:
            case TAB_F_SCN_CAM_ORTHO_SCL:
               {
                  PI3D_camera cam = scene->GetActiveCamera();
                  if(cam){
                     cam->SetFOV(tab->ItemI(TAB_I_SCN_CAM_FOV)*PI/180.0f);
                     cam->SetRange(CAM_NEAR_RANGE, tab->ItemF(TAB_F_SCN_CAM_RANGE));
                     cam->SetOrthogonal(tab->ItemB(TAB_B_SCN_CAM_ORTHO));
                     cam->SetOrthoScale(tab->ItemF(TAB_F_SCN_CAM_ORTHO_SCL));
                     em->ed->SetModified();
                  }
                  scene->Render(0, &em->mission);
                  em->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
               }
               break;
            case TAB_F_SCN_BDROP_RANGE_N:
            case TAB_F_SCN_BDROP_RANGE_F:
               {
                  scene->SetBackdropRange(tab->ItemF(TAB_F_SCN_BDROP_RANGE_N), tab->ItemF(TAB_F_SCN_BDROP_RANGE_F));
                  em->ed->SetModified();
                  scene->Render(0, &em->mission);
                  em->ed->GetIGraph()->UpdateScreen(IGUPDATE_ENTIRE_AREA);
               }
               break;
            }
            em->ed->SetModified();
         }
         break;
      }
   }
   
   C_string FormatLoadTime(int save_time){
      return C_fstring("Scene loaded in %.2f seconds",
         (float)(ed->GetIGraph()->ReadTimer()-save_time)/1000.0f);
   }

   void WriteStatusMessage(E_MISSION_IO mio,
      const char *fail_msg, const char *ok_msg, bool allow_msgbox = true){

      C_string str;
      switch(mio){
      case MIO_OK:
         str = ok_msg;
         //ed->GetIGraph()->SetAppName(C_fstring("%s - %s", (const char*)app_name, (const char*)mission_name));
         break;
      case MIO_CORRUPT:
         str = C_fstring("%s: file corrupted", (fail_msg));
         break;
      case MIO_NOFILE:
         str = C_fstring("%s: file not found", (fail_msg));
         break;
      case MIO_ACCESSFAIL:
         str = C_fstring("%s: can't write to file", (fail_msg));
         break;
      case MIO_CANCELLED:
         str = C_fstring("%s: cancelled", (fail_msg));
         break;
      default: assert(0);
      }
      if(mio!=MIO_OK && allow_msgbox)
         OsMessageBox(str, "Mission open/write", MBOX_OK);
      ed->Message(str);
   }

//----------------------------

   void DrawActorName(LPC_actor act){
      
      E_ACTOR_TYPE type = act->GetType();
      map<E_ACTOR_TYPE, S_name_texture>::iterator it = actor_name_textures.find(type);
      if(it==actor_name_textures.end()){
                              //init texture
         if(!h_font) h_font = OsCreateFont(18);
         actor_name_textures[type] = S_name_texture();
         it = actor_name_textures.find(type);

         PI3D_texture tp = CreateStringTexture(actor_names[type],
            (*it).second.tp_size, ed->GetDriver(), h_font);
         (*it).second.tp = tp;
         tp->Release();
      }
      const S_name_texture &nt = (*it).second;

                              //find (or create) texture associated with actor
      PI3D_texture tp = nt.tp;
      assert(tp);
      DrawStringTexture(act->GetFrame(), 0.0f, ed->GetScene(), tp, nt.tp_size,
         S_vector(1, .5f, 1),
         20,
         30,
         .6f);
   }

//----------------------------
#pragma pack(push,1)
   struct S_actor_undo_info{
      E_ACTOR_TYPE type;
      char frm_name[1];
   };
#pragma pack(pop)
   
   bool DestroyActor(PI3D_frame frm){

      LPC_actor act = (LPC_actor)frm->GetData();
      if(act){

         LPC_editor_item e_undo = ed->FindPlugin("Undo");
         {           //save undo info
            int len = sizeof(S_actor_undo_info) + strlen(frm->GetName());
            S_actor_undo_info *ui = (S_actor_undo_info*)alloca(len);
            ui->type = act->GetType();
            strcpy(ui->frm_name, frm->GetName());
            e_undo->Action(E_UNDO_SAVE, 
               &S_edit_Undo_2(GetName(), E_MISSION_SET_ACTOR, ui, len));
         }
         mission->DestroyActor(act);
         return true;
      }
      return false;
   }

//----------------------------

   LPC_actor CreateActor(PI3D_frame frm, E_ACTOR_TYPE at){

      LPC_actor act = (LPC_actor)frm->GetData();
                     //save undo info
      LPC_editor_item e_undo = ed->FindPlugin("Undo");
      if(e_undo){
         int len = sizeof(S_actor_undo_info) + strlen(frm->GetName());
         S_actor_undo_info *ui = (S_actor_undo_info*)(new byte[len]);
         ui->type = (act) ? act->GetType() : ACTOR_NULL;
         strcpy(ui->frm_name, frm->GetName());

         e_undo->Action(E_UNDO_SAVE, 
            &S_edit_Undo_2(GetName(), E_MISSION_SET_ACTOR, ui, len));
         delete[] (byte*)ui;
      }
      if(act)
         mission->DestroyActor(act);

      act = mission->CreateActor(at);
      act->SetFrame(frm);
      act->TableUpdate(-1, true);
      return act;
   }

//----------------------------

   static bool CheckIfSoundExists(const char *name){

      C_string snd_name = name;
                              //check wrong slash
      for(int i=snd_name.Size(); i--; ){
         if(snd_name[i]=='/')
            return false;
      }

      
      if(snd_name.Size()>=4 && snd_name[(int)snd_name.Size()-4]=='.'){
         snd_name[(int)snd_name.Size()-4] = 0;
      }
      C_fstring fn("Sounds\\Common\\%s", (const char*)snd_name);
      int h = dtaOpen(C_fstring("%s.ogg", (const char*)fn));
      if(h==-1)
         h = dtaOpen(C_fstring("%s.wav", (const char*)fn));
      if(h==-1)
         return false;

      dtaClose(h);
      return true;
   }

//----------------------------

public:
   C_edit_Mission(C_game_mission *gm):
      mission(gm),
      debug_cam_acq(false),
      draw_actor_names(true),
      make_log(false),
      time_scale_id(5),
      simulate_fps(0),
      sim_fps_last_timer(.0f),
      keep_checked_out(false),
      enabled(true),
      h_font(NULL),
      merge_load(0),
      last_create_actor_index(-1),
      merge_erase(0),
      ask_to_save(true)
   {
      memset(tab_pos_size, 0, sizeof(tab_pos_size));
      mission_name = "!test";
   }

   ~C_edit_Mission(){
      if(h_font) OsDestroyFont(h_font);
   }

   virtual bool Init(){
      e_slct = ed->FindPlugin("Selection");
      if(!e_slct)
         return false;

      ed->AddShortcut(this, E_MISSION_OPEN, "%0 &File\\%0 &Open mission\tO", K_O, 0);
      ed->AddShortcut(this, E_MISSION_SAVE, "&File\\%0 &Save mission\tS", K_S, 0);
      ed->AddShortcut(this, E_MISSION_RELOAD, "&File\\%0 &Reload mission\tL", K_L, 0);
      ed->AddShortcut(this, E_MISSION_SAVE_ASK_TOGGLE, "&File\\%0 As&k to save", 0, 0);
      ed->AddShortcut(this, E_MISSION_PROPERTIES, "&File\\%30 %i Scene &properties", 0, 0);

      ed->AddShortcut(this, E_MISSION_LOG_TOGGLE, "%30 &Debug\\%80 %i Scene open &log", 0, 0);

      ed->AddShortcut(this, E_MISSION_GCAM_DIST_P, "%30 &Debug\\%70 Game camera\\Distance mode\\Previous\t-", K_2D, 0);
      ed->AddShortcut(this, E_MISSION_GCAM_DIST_N, "%30 &Debug\\%70 Game camera\\Distance mode\\Next\t=", K_3D, 0);
      ed->AddShortcut(this, E_MISSION_GCAM_ANGLE_P, "%30 &Debug\\%70 Game camera\\Angle\\Up\t|", K_5C, 0);
      ed->AddShortcut(this, E_MISSION_GCAM_ANGLE_N, "%30 &Debug\\%70 Game camera\\Angle\\Down\tBackspace", K_BACKSPACE, 0);
      ed->AddShortcut(this, E_MISSION_GCAM_ACQUIRE, "%30 &Debug\\%70 Game camera\\Acquire\tF8", K_F8, 0);

      ed->AddShortcut(this, E_MISSION_TIME_0,   "%30 &Debug\\%80 &Time scale\\Stop\tNum0", K_NUM0, 0);
      ed->AddShortcut(this, E_MISSION_TIME_5,   "%30 &Debug\\%80 &Time scale\\5 %\tNum1", K_NUM1, 0);
      ed->AddShortcut(this, E_MISSION_TIME_10,  "%30 &Debug\\%80 &Time scale\\10 %\tNum2", K_NUM2, 0);
      ed->AddShortcut(this, E_MISSION_TIME_20,  "%30 &Debug\\%80 &Time scale\\20 %\tNum3", K_NUM3, 0);
      ed->AddShortcut(this, E_MISSION_TIME_50,  "%30 &Debug\\%80 &Time scale\\50 %\tNum4", K_NUM4, 0);
      ed->AddShortcut(this, E_MISSION_TIME_100, "%30 &Debug\\%80 &Time scale\\100 %\tNum5", K_NUM5, 0);
      ed->AddShortcut(this, E_MISSION_TIME_200, "%30 &Debug\\%80 &Time scale\\2 x\tNum6", K_NUM6, 0);
      ed->AddShortcut(this, E_MISSION_TIME_500, "%30 &Debug\\%80 &Time scale\\5 x %\tNum7", K_NUM7, 0);
      ed->AddShortcut(this, E_MISSION_TIME_1000, "%30 &Debug\\%80 &Time scale\\10 x %\tNum8", K_NUM8, 0);

      ed->AddShortcut(this, E_MISSION_SIMFPS_OFF,   "%30 &Debug\\%85 &Simulate fps\\off", 0, 0);
      ed->AddShortcut(this, E_MISSION_SIMFPS_5,     "%30 &Debug\\%85 &Simulate fps\\5", 0, 0);
      ed->AddShortcut(this, E_MISSION_SIMFPS_10,   "%30 &Debug\\%85 &Simulate fps\\10", 0, 0);
      ed->AddShortcut(this, E_MISSION_SIMFPS_15,   "%30 &Debug\\%85 &Simulate fps\\15", 0, 0);
      ed->AddShortcut(this, E_MISSION_SIMFPS_20,   "%30 &Debug\\%85 &Simulate fps\\20", 0, 0);
      ed->AddShortcut(this, E_MISSION_SIMFPS_50,   "%30 &Debug\\%85 &Simulate fps\\50", 0, 0);

      ed->AddShortcut(this, E_MISSION_ACTOR_CREATE, "%60 &Game\\%0 &Create actor\tA", K_A, 0);
      ed->AddShortcut(this, E_MISSION_ACTOR_DESTROY, "%60 &Game\\%a &Destroy actor", 0, 0);
      ed->AddShortcut(this, E_MISSION_ACTOR_DRAW_NAMES, "%60 &Game\\Draw &names\tCtrl+Shift+Alt+A", K_A, SKEY_CTRL|SKEY_ALT|SKEY_SHIFT);

      ed->AddShortcut(this, E_MISSION_MODE_EDIT, "%60 &Game\\%i &Edit mode\tF5", K_F5, 0);
      ed->AddShortcut(this, E_MISSION_MODE_GAME, "%60 &Game\\&Game mode\tF6", K_F6, 0);

      ed->AddShortcut(this, E_MISSION_ADD_SEL_TO_MAP, "%60 &Game\\&Models on map\\&Mark selection", 0, 0);
      ed->AddShortcut(this, E_MISSION_DEL_SEL_FROM_MAP, "%60 &Game\\&Models on map\\&Unmark selection", 0, 0);
      ed->AddShortcut(this, E_MISSION_SEL_MAP_MARKED, "%60 &Game\\&Models on map\\&Select marked", 0, 0);
      ed->AddShortcut(this, E_MISSION_FLASH_MAP_MARKED, "%60 &Game\\&Models on map\\&Flash marked\tCtrl+Alt+F", K_F, SKEY_CTRL|SKEY_ALT);

      ed->AddShortcut(this, E_MISSION_VERSION, "%100 &Help\\&Version", 0, 0);

      C_string vss_exe, vss_database;
      if(SCGetPath(vss_exe, vss_database)){
         ed->AddShortcut(this, E_MISSION_GET, "&File\\%10 &Version control\\&Get latest\tShift+G", K_G, SKEY_SHIFT);
         ed->AddShortcut(this, E_MISSION_CHECK_IN, "&File\\%10 &Version control\\Check &in mission\tCtrl+I", K_I, SKEY_CTRL);
         ed->AddShortcut(this, E_MISSION_CHECK_OUT, "&File\\%10 &Version control\\Check &out mission\tCtrl+O", K_O, SKEY_CTRL);
      }

      ed->CheckMenu(this, E_MISSION_SAVE_ASK_TOGGLE, ask_to_save);
      ed->CheckMenu(this, E_MISSION_ACTOR_DRAW_NAMES, draw_actor_names);

      if(debug_cam_acq)
         Action(E_MISSION_GCAM_ACQUIRE, 0);
      ed->CheckMenu(this, E_MISSION_TIME_100, true);

      ed->RegForRender(this);

      //app_name = ed->GetIGraph()->GetAppName();
      sb_rt_msg_pos = ed->CreateStatusIndex(300);

      return true;
   }

   virtual void Close(){ 
   }

   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){

      if(simulate_fps){
                              //slow down so that we get desired FPS
         float sleep_time = (1000.0f/(float)simulate_fps);
                              //sleep for the time,
                              // minus time really elapsed
         sleep_time -= (ed->GetIGraph()->ReadTimer() - sim_fps_last_timer);
         sleep_time -= .3f;
         if(sleep_time>0){
            Sleep(sleep_time);
         }
         sim_fps_last_timer = ed->GetIGraph()->ReadTimer() + fmod(sleep_time, 1.0f);
      }
   }

//----------------------------

   virtual dword Action(int id, void *context){

      switch(id){
      case E_MISSION_OPEN:
         {
            ed->GetScene()->AttenuateSounds();
            ed->GetIGraph()->FlipToGDI();
            C_string save_name = mission_name;
            int i = DialogBoxParam(GetModuleHandle(NULL), "IDD_OPEN_MISSION",
               (HWND)ed->GetIGraph()->GetHWND(), dlgSelect, (LPARAM)this);
            if(i==1){
               C_string new_name = mission_name;
               mission_name = save_name;
               if(!Action(E_MISSION_SAVE, (void*)true))
                  break;
               mission_name = new_name;
               int save_time = ed->GetIGraph()->ReadTimer();

               mission->Close();

               E_MISSION_IO mio = mission->Open(mission_name, (make_log ? OPEN_LOG : 0));
               WriteStatusMessage(mio, "Load failed", FormatLoadTime(save_time), false);
               last_mission_name = save_name;
            }
         }
         break;

      case E_MISSION_RELOAD:
         {
            if(!context){
               if(!Action(E_MISSION_SAVE, (void*)true))
                  break;
            }
            int save_time = ed->GetIGraph()->ReadTimer();
            mission->Close();

            E_MISSION_IO mio = mission->Open(mission_name, (make_log ? OPEN_LOG : 0));
            WriteStatusMessage(mio, "Load failed", FormatLoadTime(save_time), false);
#if SAFE_FRAME
            igraph->Rectangle(0, 0, igraph->Scrn_sx(), igraph->Scrn_sy(), 0);
            mission->GetScene()->SetViewport(5, 5, igraph->Scrn_sx()-SAFE_FRAME, igraph->Scrn_sy()-SAFE_FRAME);
            //mission->GetScene()->SetViewport(SAFE_FRAME, SAFE_FRAME, igraph->Scrn_sx()-SAFE_FRAME, igraph->Scrn_sy()-SAFE_FRAME);
#endif   //SAFE_FRAME
         }
         break;

      case E_MISSION_SAVE:
         if(!enabled) return true;
         if(!context || ask_to_save)
         if(ed->IsModified() || !context){
            bool ok = false;
            if(context){
               ed->GetIGraph()->FlipToGDI();
               switch(OsMessageBox("Scene modified. Save changes?", "", MBOX_YESNOCANCEL)){
               case MBOX_RETURN_YES:
                  ok = true;
                  break;
               case MBOX_RETURN_CANCEL:
                  return false;
               }
            }else ok = true;
            if(ok){
               if(!Action(E_MISSION_CHECK_BSP, (void*)true))
                  return false;
               E_MISSION_IO mio = mission->Save(mission_name);
               WriteStatusMessage(mio, "Save failed", "Scene saved");
            }
         }
         if(ed->IsModified())
            ed->SetModified(false);
         return true;

      case E_MISSION_CHECK_BSP:
         if(!ed->GetScene()->IsBspBuild()){
                              //check only for game files, not maps or briefings or others
            static const int GAME_DIR_SIZE = 4;
            static const char *game_dir ="game";
            int name_pos = mission_name.Size();
            if(name_pos < GAME_DIR_SIZE)
               return true;
            name_pos -= GAME_DIR_SIZE;
            C_string name_type = mission_name;
            name_type.ToLower();
            for(int jj1 = GAME_DIR_SIZE; jj1--; ){
               int pos1 = name_pos + jj1;
               if(name_type[pos1] != game_dir[jj1])
                  return true;
            }

            int id = OsMessageBox("BSP tree is not built, do you want to do it now?",
               "Checking bsp.", MBOX_YESNOCANCEL);
            if(id==MBOX_RETURN_CANCEL)
               return false;
            if(id==MBOX_RETURN_YES){
               LPC_editor_item ei = ed->FindPlugin("BSP-tools");
               if(!ei){
                  ed->Message("Failed to locate BPS build plugin");
                  return false;
               }
               if(!ei->Action(E_BSP_CREATE))
                  return false;
            }
         }
         return true;

      case E_MISSION_BROWSE:
         {
            pair<char*,HWND> &p = *(pair<char*,HWND>*)context;
            C_string save_name = mission_name;
            C_string save_name1 = last_mission_name;
            mission_name = p.first;
            last_mission_name = NULL;
            int i = DialogBoxParam(GetModuleHandle(NULL), "IDD_OPEN_MISSION",
               p.second, dlgSelect, (LPARAM)this);
            if(i)
               strcpy(p.first, mission_name);
            mission_name = save_name;
            last_mission_name = save_name1;
            return (bool)i;
         }
         break;

      case E_MISSION_ENABLE:
         {
            enabled = (bool)context;
            ed->EnableMenu(this, E_MISSION_OPEN, enabled);
            ed->EnableMenu(this, E_MISSION_SAVE, enabled);
            ed->EnableMenu(this, E_MISSION_RELOAD, enabled);
            ed->EnableMenu(this, E_MISSION_MERGE, enabled);
         }
         break;

      case E_MISSION_SET_MODEL:
         {
            Action(E_MISSION_ENABLE, (void*)false);

            I3D_LOAD_CB_PROC cbLoadScene;

            ed->GetScene()->Open((const char*)context, I3DLOAD_PROGRESS | I3DLOAD_ERRORS, cbLoadScene);
            {                 //finish progress indicator
               float f = 1.0f;
               cbLoadScene((I3D_LOADMESSAGE)10000, *(dword*)&f, 2, NULL);
            }
                              //setup lighting
            ed->FindPlugin("Light")->Action(E_LIGHT_MAKE_HELP);
            ed->GetScene()->SetBgndColor(S_vector(.3f, .4f, .55f));
                              //setup camera to see entire model

            PI3D_camera cam = (PI3D_camera)ed->FindPlugin("MouseEdit")->Action(E_MOUSE_GET_EDIT_CAMERA);

            I3D_bound_volume bvol;
            ed->GetScene()->GetPrimarySector()->ComputeHRBoundVolume(&bvol);
            const I3D_bbox &bbox = bvol.bbox;
            float cam_dist;
            S_vector cam_look_at;
            float diag_size = 0.0f;
            if(bbox.IsValid()){
               S_vector diag = bbox.max - bbox.min;
               cam_look_at = bbox.min + diag*.5f;
               diag_size = diag.Magnitude();
               cam_dist = diag_size * .6f;
               cam_dist = Max(cam_dist, 2.0f);
            }else{
               cam_look_at.Zero();
               cam_dist = 0.0f;
            }
            S_vector pos = cam_look_at + S_vector(.8f, .4f, -.8f) * cam_dist;
            float nr, fr;
            cam->GetRange(nr, fr);
            fr = Max(200.0f, cam_dist + diag_size * .6f);
            cam->SetRange(CAM_NEAR_RANGE, fr);
            cam->SetFOV(PI*.4f);
            cam->SetPos(pos);
            cam->SetDir(cam_look_at - pos);
            cam->Update();
         }
         break;

      case E_MISSION_SET_MISSION:
         mission_name = (const char*)context;
         break;

      case E_MISSION_SAVE_ASK_TOGGLE:
         ask_to_save = !ask_to_save;
         ed->CheckMenu(this, E_MISSION_SAVE_ASK_TOGGLE, ask_to_save);
         ed->Message(C_fstring("Ask to save scene option %s", ask_to_save ? "on" : "off"));
         break;

      case E_MISSION_GET:
         {
            C_fstring filename("missions\\%s\\scene.bin", (const char*)mission_name);
            ed->Message(C_fstring("Getting '%s'...", (const char*)filename), 0, EM_MESSAGE, true);
            OsMakeFileReadOnly(filename, true);
            bool ok = SCGetFile("HDT", filename);
            if(!ok){
               ed->Message(C_fstring("Failed to get: '%s'", (const char*)filename));
               break;
            }
                        //file checked out, must reload now (if different)
            dword mission_crc;
            if(mission->GetMissionCRC(mission_name, mission_crc)){
               if(mission_crc!=mission->GetCRC()){
                  int i = OsMessageBox("Mission must be reloaded. Click OK to reload now.", "Insanity editor", MBOX_OKCANCEL);
                  if(i==MBOX_RETURN_YES){
                                 //reload without attempting to save
                     Action(E_MISSION_RELOAD, (void*)true);
                     return 2;
                  }
               }
            }
            ed->Message(C_fstring("Mission get: '%s'.", (const char*)filename));
         }
         break;

      case E_MISSION_CHECK_OUT:
         {
            C_fstring filename("missions\\%s\\scene.bin", (const char*)mission_name);
            bool is_rdonly = OsIsFileReadOnly(filename);
                              //ask user what to do
            struct S_hlp{
               const char *filename;
               bool can_make_wr;
               static BOOL CALLBACK dlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

                  switch(uMsg){
                  case WM_INITDIALOG:
                     {
                        S_hlp *hp = (S_hlp*)lParam;
                        WinInitDialog(hwnd);
                        SetDlgItemText(hwnd, IDC_EDIT, (const char*)hp->filename);
                        EnableWindow(GetDlgItem(hwnd, IDC_WRITABLE), hp->can_make_wr);
                        ShowWindow(hwnd, SW_SHOW);
                     }
                     return 1;
                  case WM_COMMAND:
                     if(HIWORD(wParam)==BN_CLICKED)
                        EndDialog(hwnd, LOWORD(wParam));
                     break;
                  }
                  return 0;
               }
            } hlp;
            hlp.filename = filename;
            hlp.can_make_wr = is_rdonly;
            ed->GetIGraph()->FlipToGDI();
            int i = DialogBoxParam(GetModuleHandle(NULL), "IDD_CHECKOUT", (HWND)ed->GetIGraph()->GetHWND(),
               S_hlp::dlgProc, (LPARAM)&hlp);

            switch(i){
            case IDOK:
               {
                              //try to check out the file
                  ed->Message(C_fstring("Checking out '%s'...", (const char*)filename), 0, EM_MESSAGE, true);
                  if(!is_rdonly)
                     OsMakeFileReadOnly(filename, true);
                  bool ok = SCCheckOutFile("HDT", filename);
                  if(!ok){
                     ed->Message(C_fstring("Failed to check out: '%s'", (const char*)filename));
                     break;
                  }
                              //file checked out, must reload now (if different)
                  dword mission_crc;
                  if(mission->GetMissionCRC(mission_name, mission_crc)){
                     if(mission_crc!=mission->GetCRC()){
                        int i = OsMessageBox("Mission must be reloaded. Click OK to reload now.", "Insanity editor", MBOX_OKCANCEL);
                        if(i==MBOX_RETURN_YES){
                                       //reload without attempting to save
                           Action(E_MISSION_RELOAD, (void*)true);
                           return 2;
                        }
                     }
                  }
               }
               ed->Message(C_fstring("Mission checked out: '%s'.", (const char*)filename));
               return 1;

            case IDC_WRITABLE:
               if(OsMakeFileReadOnly(filename, false))
                  return 1;
               break;

            default:
               ed->Message("Check out canceled.");
            }
         }
         break;

      case E_MISSION_CHECK_IN:
         {
            bool tmp = ask_to_save;
            ask_to_save = true;
            if(!Action(E_MISSION_SAVE, (void*)true)){
               ask_to_save = tmp;
               break;
            }
            ask_to_save = tmp;
            C_fstring filename("missions\\%s\\scene.bin", (const char*)mission_name);
            ed->GetIGraph()->FlipToGDI();

                              //ask user what to do
            struct S_hlp{
               const char *filename;
               bool keep;
               static BOOL CALLBACK dlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

                  switch(uMsg){
                  case WM_INITDIALOG:
                     {
                        S_hlp *hp = (S_hlp*)lParam;
                        SetWindowLong(hwnd, GWL_USERDATA, lParam);
                        WinInitDialog(hwnd);
                        SetDlgItemText(hwnd, IDC_EDIT, hp->filename);
                        CheckDlgButton(hwnd, IDC_CHECKIN_KEEP, hp->keep);
                        ShowWindow(hwnd, SW_SHOW);
                     }
                     return 1;
                  case WM_COMMAND:
                     if(HIWORD(wParam)==BN_CLICKED){
                        switch(LOWORD(wParam)){
                        case IDOK:
                        case IDCANCEL:
                           {
                              S_hlp *hp = (S_hlp*)GetWindowLong(hwnd, GWL_USERDATA);
                              hp->keep = IsDlgButtonChecked(hwnd, IDC_CHECKIN_KEEP);
                              EndDialog(hwnd, LOWORD(wParam));
                           }
                           break;
                        }
                     }
                     break;
                  }
                  return 0;
               }
            } hlp;
            hlp.filename = filename;
            hlp.keep = keep_checked_out;
            int i = DialogBoxParam(GetModuleHandle(NULL), "IDD_CHECKIN", (HWND)ed->GetIGraph()->GetHWND(),
               S_hlp::dlgProc, (LPARAM)&hlp);
            keep_checked_out = hlp.keep;
            if(i==IDOK){
                              //try to check in the file
               ed->Message(C_fstring("Checking in '%s'...", (const char*)filename), 0, EM_MESSAGE, true);
               bool ok = SCCheckInFile("HDT", filename, keep_checked_out);
               if(!ok){
                  ed->Message(C_fstring("Failed to check in: '%s'", (const char*)filename));
               }else{
                  ed->Message(C_fstring("Mission checked in: '%s'.", (const char*)filename));
               }
            }else{
               ed->Message("Check in canceled.");
            }
         }
         break;

      case E_MISSION_LOG_TOGGLE:
         make_log = !make_log;
         ed->CheckMenu(this, E_MISSION_LOG_TOGGLE, make_log);
         break;

      case E_MISSION_PROPERTIES:
         {
            LPC_table tab_prop = CreateTable();
            tab_prop->Open((dword)GetPropTemplate(), TABOPEN_TEMPLATE);

            {
               *(S_vector*)tab_prop->Item(TAB_CV_SCN_BGND) = ed->GetScene()->GetBgndColor();
               PI3D_camera cam = ed->GetScene()->GetActiveCamera();
               if(cam){
                  tab_prop->ItemI(TAB_I_SCN_CAM_FOV) = cam->GetFOV() * 180.0f/PI;
                  float rn;
                  cam->GetRange(rn, tab_prop->ItemF(TAB_F_SCN_CAM_RANGE));
                  tab_prop->ItemB(TAB_B_SCN_CAM_ORTHO) = cam->GetOrthogonal();
                  tab_prop->ItemF(TAB_F_SCN_CAM_ORTHO_SCL) = cam->GetOrthoScale();
               }
               ed->GetScene()->GetBackdropRange(tab_prop->ItemF(TAB_F_SCN_BDROP_RANGE_N),
                  tab_prop->ItemF(TAB_F_SCN_BDROP_RANGE_F));
               //hwnd_prop = (HWND)
               tab_prop->Edit(GetPropTemplate(), ed->GetIGraph()->GetHWND(), cbProp,
                  (dword)this, 
                  /*TABEDIT_MODELESS | */TABEDIT_IMPORT | TABEDIT_EXPORT | TABEDIT_INFO,
                  NULL, tab_pos_size);
               //if(hwnd_prop) ed->GetIGraph()->AddDlgHWND(hwnd_prop);
            }
            tab_prop->Release();
         }
         break;

      case E_MISSION_ACTOR_DRAW_NAMES:
         draw_actor_names = !draw_actor_names;
         ed->CheckMenu(this, E_MISSION_ACTOR_DRAW_NAMES, draw_actor_names);
         //if(draw_actor_names) ed->RegForRender(this);
         //else ed->UnregForRender(this);
         ed->Message(C_fstring("Draw actor names %s", draw_actor_names ? "on" : "off"));
         break;

      case E_MISSION_MODE_EDIT:
         if(!ed->IsActive()){
            //NetCloseAll();
            mission->GameEnd();

            if(ed->CanModify(false))
               ed->SetModified(true);
            ed->SetActive(true);
            ed->Message("Edit mode");
         }
         break;

      case E_MISSION_MODE_GAME:
         if(ed->IsActive()){
                              //save changes now
            if(!Action(E_MISSION_SAVE, (void*)true))
               break;
                              //switch to view mode, so that mouse is hidden
            ed->FindPlugin("MouseEdit")->Action(E_MOUSE_SET_MODE, (void*)1);
                              //check if BSP tree is built, if not, ask user to do it now
            if(!Action(E_MISSION_CHECK_BSP, (void*)true))
               break;

            mission->GameBegin(false);
            ed->SetActive(false);
            ed->Message("Game mode");
         }
         break;

      case E_MISSION_VERSION:
         {
            OsMessageBox("Compile date: "__DATE__", "__TIME__, "Application version", MBOX_OK);
         }
         break;

      case E_MISSION_RT_ERR_MESSAGE:
         ed->Message((const char*)context, sb_rt_msg_pos, EM_ERROR);
         break;

      case E_MISSION_TIME_0:
      case E_MISSION_TIME_5:
      case E_MISSION_TIME_10:
      case E_MISSION_TIME_20:
      case E_MISSION_TIME_50:
      case E_MISSION_TIME_100:
      case E_MISSION_TIME_200:
      case E_MISSION_TIME_500:
      case E_MISSION_TIME_1000:
         {
            ed->CheckMenu(this, time_scale_id, false);
            time_scale_id = id - E_MISSION_TIME_0;
            ed->CheckMenu(this, time_scale_id, true);
            ed->Message(C_fstring("Time scale set to %i%%", (int)(time_scales[time_scale_id]*100.0f)));
         }
         break;

      case E_MISSION_SIMFPS_OFF:
      case E_MISSION_SIMFPS_5:
      case E_MISSION_SIMFPS_10:
      case E_MISSION_SIMFPS_15:
      case E_MISSION_SIMFPS_20:
      case E_MISSION_SIMFPS_50:
         {
            ed->CheckMenu(this, E_MISSION_SIMFPS_OFF + GetSimFpsID(simulate_fps), false);
            simulate_fps = simulate_fps_table[id - E_MISSION_SIMFPS_OFF];
            ed->CheckMenu(this, id, true);
            sim_fps_last_timer = ed->GetIGraph()->ReadTimer();
            ed->Message(simulate_fps ?
               C_fstring("Simulate fps %i", simulate_fps) : "Simulate fps off");
         }
         break;

      case E_MISSION_TIME_GET:
         return *(dword*)&time_scales[time_scale_id];

      case E_MISSION_ACTOR_CREATE:
         {
            if(!ed->CanModify())
               break;
            PI3D_frame frm = (PI3D_frame)e_slct->Action(E_SELECTION_GET_SINGLE);
            if(frm){
               switch(frm->GetType()){
               case FRAME_VOLUME:
               case FRAME_CHECKPOINT:
               case FRAME_CAMERA:
                  frm = NULL;
                  break;
               }
            }
            if(!frm){
               ed->Message("Single selection required");
               break;
            }
            static const E_ACTOR_TYPE types[] = {
               ACTOR_PLAYER,
               ACTOR_ENEMY,
               ACTOR_DOOR,
               ACTOR_RESOURCE,
               ACTOR_AUTOMOBIL,
               ACTOR_RANDOM_GEN,
               ACTOR_TEXTURE_SHIFT,
               ACTOR_CANNON,
               ACTOR_VERSION,
               ACTOR_AUTO_CANNON,
               ACTOR_DETECTOR,
               ACTOR_LADDER,
               ACTOR_KILL_ANIM,
               ACTOR_MINE,
               ACTOR_NULL
            };
                              //let user select actor to create
            char buf[2048];
            int len = 0;
            for(int i=0; types[i]; i++){
               int len1 = strlen(actor_names[types[i]]) + 1;
               memcpy(&buf[len], actor_names[types[i]], len1);
               len += len1;
            }
            buf[len] = 0;
            i = WinSelectItem("Create actor:", buf, last_create_actor_index);
            if(i==-1)
               break;
            last_create_actor_index = i;

            CreateActor(frm, types[i]);

            ed->Message("Actor created");
            ed->SetModified();
         }
         break;

      case E_MISSION_ACTOR_DESTROY:
         {
            if(!ed->CanModify())
               break;
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            if(!sel_list.size()){
               ed->Message("Actor destruction: selection required");
               break;
            }

            for(int i=sel_list.size(), count = 0; i--; ){
               PI3D_frame frm = sel_list[i];
               switch(frm->GetType()){
               case FRAME_VOLUME:
               case FRAME_CHECKPOINT:
               case FRAME_CAMERA:
                  frm = NULL;
                  break;
               }
               if(frm && DestroyActor(frm))
                  ++count;
            }
            ed->Message(C_fstring("Actors destroyed: %i", count));
            if(count)
               ed->SetModified();
         }
         break;

      case E_MISSION_SET_ACTOR:
         {
            S_actor_undo_info *ui = (S_actor_undo_info*)context;
            PI3D_frame frm = ed->GetScene()->FindFrame(ui->frm_name, ENUMF_ALL | ENUMF_BACKDROP);
            if(frm){
               LPC_actor act = (LPC_actor)frm->GetData();
               if(ui->type == ACTOR_NULL){
                  if(act){
                     ui->type = act->GetType();
                              //destroy actor
                     mission->DestroyActor(act);
                  }
               }else{
                  E_ACTOR_TYPE save_type = (act) ? act->GetType() : ACTOR_NULL;
                  LPC_actor act1 = mission->CreateActor(ui->type);
                  act1->SetFrame(frm);
                  ui->type = save_type;
               }
                              //save undo info
               int len = sizeof(S_actor_undo_info) + strlen(frm->GetName());
               LPC_editor_item e_undo = ed->FindPlugin("Undo");
               if(e_undo)
               e_undo->Action(E_UNDO_SAVE, 
                  &S_edit_Undo_2(GetName(), E_MISSION_SET_ACTOR, ui, len));
            }else
               ed->Message(C_fstring("C_edit_Mission(E_MISSION_SET_ACTOR): can't find frame %s", ui->frm_name));
         }
         break;

      case EDIT_ACTION_FRAME_DELETE:
         {
            PI3D_frame frm = (PI3D_frame)context;
                              //don't destroy actors on volumes
            if(frm->GetType()!=FRAME_VOLUME)
               DestroyActor(frm);
         }
         break;

      case EDIT_ACTION_FRAME_DUPLICATE:
         {
            pair<PI3D_frame, PI3D_frame> *p = (pair<PI3D_frame, PI3D_frame>*)context;
            PI3D_frame frm_old = p->first, frm_new = p->second;
            LPC_actor actor_old = (LPC_actor)frm_old->GetData();
            if(actor_old){
               LPC_actor actor_new = CreateActor(frm_new, actor_old->GetType());
               actor_new->SetFrame(frm_new);

               for(int i=0; i<3; i++){
                  LPC_table tab_old = actor_old->GetTable(i);
                  if(tab_old){
                              //copy table (if any)
                     LPC_table tab_new = actor_new->GetTable(i);
                     assert(tab_new);
                     tab_new->Open((dword)tab_old, TABOPEN_UPDATE | TABOPEN_DUPLICATE);
                  }
               }
            }
         }
         break;

         /*
      case EDIT_ACTION_MISSION_SAVE_0:
         {
                              //save mission table
            C_chunk *ck = (C_chunk*)context;
            ck->WAscend(CT_MISSION_TABLE);
            mission->GetConfigTab()->Save((dword)ck->GetHandle(), TABOPEN_FILEHANDLE);
            ck->Descend();
         }
         break;

      case EDIT_ACTION_MISSION_SAVE_2:
         {
                              //save actors
            C_chunk *ck = (C_chunk*)context;
            const vector<C_smart_ptr<C_actor> > &actors = mission->GetActorsVector();

            for(int i=0; i<actors.size(); i++){
               LPC_actor actor = actors[i];
               E_ACTOR_TYPE at = actor->GetActorType();
                              //check if it may be written onto disk
               if(actor_type_info[at].allow_edit_creation){
                  ck->WAscend(CT_ACTOR);
                  ck->WStringChunk(CT_NAME, actor->GetName());
                  ck->Write((const char*)&at, sizeof(word));
                              //let actor write its data
                  //actor->MissionSave(ck);

                              //write actor's table
                  LPC_table tab = actor->GetTable();
                  if(tab){
                  //S_frame_info *fi = GetFrameInfo(actor->GetFrame());
                  //if(fi->tab && actor->GetTableTemplate() && (!fi->script || !fi->script->GetTableTemplate()))
                     ck->WAscend(CT_ACTOR_TABLE);
                     tab->Save((dword)ck->GetHandle(), TABOPEN_FILEHANDLE);
                     ck->Descend();
                  }
                  ck->Descend();
               }
            }
         }
         break;
         */

      case E_MISSION_ADD_SEL_TO_MAP:
      case E_MISSION_DEL_SEL_FROM_MAP:
         {
            if(!ed->CanModify())
               break;
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);
            int num_changed = 0;
            for(int i=sel_list.size(); i--; ){
               PI3D_frame frm = sel_list[i];
               if(frm->GetType()==FRAME_MODEL){
                  dword flgs = frm->GetUserFlags();
                  if(id==E_MISSION_ADD_SEL_TO_MAP){
                     if(!(flgs&1)){
                        flgs |= 1;
                        frm->SetUserFlags(flgs);
                        ++num_changed;
                        e_slct->Action(E_SELECTION_FLASH_FRAME, frm);
                     }
                  }else{
                     if(flgs&1){
                        flgs &= ~1;
                        frm->SetUserFlags(flgs);
                        ++num_changed;
                        e_slct->Action(E_SELECTION_FLASH_FRAME, frm);
                     }
                  }
               }
            }
            if(num_changed)
               ed->SetModified();
            ed->Message(C_fstring("Models %s: %i",
               (id==E_MISSION_ADD_SEL_TO_MAP) ? "marked" : "unmarked", num_changed));
         }
         break;

      case E_MISSION_SEL_MAP_MARKED:
         e_slct->Action(E_SELECTION_CLEAR);
                              //flow...
      case E_MISSION_FLASH_MAP_MARKED:
         {
            struct S_hlp{
               C_editor_item *e_slct;
               bool select;
               int num;
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                  dword flgs = frm->GetUserFlags();
                  if(flgs&1){
                     S_hlp *hp = (S_hlp*)c;
                     hp->e_slct->Action(hp->select ? E_SELECTION_ADD_FRAME : E_SELECTION_FLASH_FRAME, frm);
                     ++hp->num;
                  }
                  return I3DENUMRET_OK;
               }
            } hlp = {e_slct, (id==E_MISSION_SEL_MAP_MARKED), 0};
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&hlp, ENUMF_MODEL);
            ed->Message(C_fstring("Models marked: %i", hlp.num));
         }
         break;
      case E_MISSION_GCAM_ACQUIRE:
         {
            debug_cam_acq = !debug_cam_acq;
            mission->game_cam.Acquire(debug_cam_acq);
            ed->CheckMenu(this, E_MISSION_GCAM_ACQUIRE, debug_cam_acq);
            ed->Message(C_fstring("Camera acquire %s", debug_cam_acq ? "on":"off"));
         }
         break;
      case E_MISSION_GCAM_IS_ACQUIRED:
         return debug_cam_acq;
         break;

      case EDIT_ACTION_AFTER_LOAD:
         igraph->GetTimer(1, 1);
         break;

      case EDIT_ACTION_BEFORE_FREE:
         {
            ed->Message("", 4);
                              //always flush out model cache in editor
            model_cache.Clear();
            sound_cache.Clear();
                              //no old sins into new mission
            LPC_editor_item ei = ed->FindPlugin("Log");
            if(ei)
               ei->Action(E_LOG_CLOSE);
         }
         break;

      case EDIT_ACTION_WINDOW_RESET:
                              //destroy all current name textures
         actor_name_textures.clear();
         break;

      case EDIT_ACTION_QUERY_MODIFY:
         {
            bool quiet = (bool)context;
                              //check if current mission is readonly
            C_fstring filename("missions\\%s\\scene.bin", (const char*)mission_name);
            if(OsIsFileReadOnly(filename)){
               if(!quiet){
                  int i = Action(E_MISSION_CHECK_OUT, 0);
                  switch(i){
                  case 0: return 1;
                  case 1: return 0;
                  case 2: return 1;
                  default: assert(0);
                  }
               }
               return 1;
            }
         }
         return 0;

      case EDIT_ACTION_VALIDATE:
         {
            LPC_editor_item e_log = ed->FindPlugin("Log");
            if(!e_log)
               break;
            ed->Message("Checking actors...", 0, EM_MESSAGE, true);
            const vector<C_smart_ptr<C_actor> > &actors = mission->GetActors();
            bool err = false;
            for(int i=actors.size(); i--; ){
               LPC_actor a = actors[i];
               CPI3D_frame frm = a->GetFrame();
               switch(a->GetType()){
               case ACTOR_DOOR:
                  {
                              //check door's sector in opened position
                     CPI3D_frame prnt = frm;
                     LPC_table tab = a->GetTable(0);

                     while((prnt=prnt->GetParent()) && (prnt->GetType()!=FRAME_SECTOR));
                     assert(prnt);
                     if(!prnt)
                        break;
                     I3D_bound_volume bvol;
                     frm->ComputeHRBoundVolume(&bvol);
                     if(bvol.bbox.IsValid()){
                        int oa = tab->ItemI(TAB_I_DOR_ANGLE_OPEN);
                        float a = (float)oa * PI / 180.0f;
                        S_quat q;
                        q.Make(S_vector(0, 1, 0), a);
                        S_matrix m(q);
                        m(3) = frm->GetPos();
                        m *= frm->GetParent()->GetMatrix();
                              //get bbox' center
                        S_vector pos = (bvol.bbox.min + bvol.bbox.max) * .5f;
                        pos *= frm->GetScale();
                              //apply opened angle, and put to world coords
                        pos *= m;
                              //check if sector matches to parent sector
                        CPI3D_sector sct = ed->GetScene()->GetSector(pos);
                        if(sct!=prnt){
                           e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                           e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                              C_fstring("Door's opened position is in wrong sector: '%s'", frm->GetName()));
                           err = true;
                        }
                     }
                              //check open sound
                     const char *name = tab->ItemS(TAB_S12_DOR_SOUND);
                     if(*name){
                        if(!CheckIfSoundExists(name)){
                           e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                           e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                              C_fstring("Door '%s': can't play sound '%s'", frm->GetName(), name));
                           err = true;
                        }
                     }
                  }
                  break;

               case ACTOR_AUTOMOBIL:
               case ACTOR_AUTO_CANNON:
                  {
                              //check sounds
                     LPC_table tab = a->GetTable(1);
                     static int indxs[3] = {TAB_S12_AUTO_SOUND, TAB_S12_AUTO_SOUND_ON, TAB_S12_AUTO_SOUND_OFF};
                     for(int i=0; i<3; i++){
                        const char *name = tab->ItemS(indxs[i]);
                        if(*name){
                           if(!CheckIfSoundExists(name)){
                              e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                              e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                                 C_fstring("Automobil '%s': can't play sound '%s'", frm->GetName(), name));
                              err = true;
                           }
                        }
                     }
                              //check collision body (airplanes can be ignored)
                     if(frm && frm->GetType()==FRAME_MODEL){
                        PI3D_frame col_parent = frm->FindChildFrame("col");
                        if(col_parent){
                           if(!col_parent->FindChildFrame("*", ENUMF_VOLUME | ENUMF_WILDMASK)){
                              e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                              e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                                 C_fstring("Automobil '%s': object 'col' has no volume child (automobil has no physic body)", frm->GetName()));
                              err = true;
                              //ReportActorError("object 'col' has no volume child (automobil has no physic body)");
                           }
                        }else{
                           e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                           e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                              C_fstring("Automobil '%s': cannot find dummy object 'col'", frm->GetName()));
                           err = true;
                           //ReportActorError("cannot find dummy object 'col'");
                        }
                     }
                  }
                  break;

               case ACTOR_DETECTOR:
                  {
                     LPC_table tab = a->GetTable(0);
                     for(int i=0; i<tab->ArrayLen(TAB_E_DET_ACTION); i++){
                        switch(tab->ItemE(TAB_E_DET_ACTION, i)){
                        case 1:
                           if(isound){
                              const char *name = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
                              if(!CheckIfSoundExists(name)){
                                 e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                                 e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                                    C_fstring("Detector '%s': can't play sound '%s'", frm->GetName(), name));
                                 err = true;
                              }
                           }
                           break;
                        case 2:
                        case 8:
                        case 11:
                           {
                              const char *cp = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
                              if(*cp){
                                 LPC_actor a = mission->FindActor(cp);
                                 if(!a){
                                    e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                                    e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                                       C_fstring("Detector '%s': can't find actor '%s'", frm->GetName(), cp));
                                    err = true;
                                 }
                              }
                           }
                           break;

                        case 9:
                           {
                              const char *cp = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
                              if(*cp){
                                 const vector<C_smart_ptr<C_actor> > &actors = mission->GetActors();
                                 for(int i=actors.size(); i--; ){
                                    LPC_actor a = actors[i];
                                    if(C_string(a->GetName()).Match(cp))
                                       break;
                                 }
                                 if(i==-1){
                                    e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                                    e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                                       C_fstring("Detector '%s': can't find actor '%s'", frm->GetName(), cp));
                                    err = true;
                                 }
                              }
                           }
                           break;

                        case 6:
                           {
                              const char *cp1 = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
                              LPC_checkpoint cp = mission->FindCP(cp1);
                              if(!cp){
                                 e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                                 e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                                    C_fstring("Detector '%s': can't find checkpoint '%s'", frm->GetName(), cp1));
                                 err = true;
                              }
                           }
                           break;
                        case 7:
                           {
                              C_string str = tab->ItemS(TAB_S32_DET_ACTION_DATA, i, 32);
                              if(str.Size() >= 3 && str[(int)str.Size()-1]=='@' && str[(int)str.Size()-1]=='@'){
                                 str[(int)str.Size()-3] = 0;
                              }
                              struct S_hlp{
                                 static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                                    *(bool*)c = true;
                                    return I3DENUMRET_OK;
                                 }
                              };
                              bool any = false;
                              mission->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&any, ENUMF_ALL | ENUMF_USEMASK, str);
                              if(!any){
                                 e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                                 e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                                    C_fstring("Detector '%s': can't find frame '%s'", frm->GetName(), (const char*)str));
                                 err = true;
                              }
                           }
                           break;
                        }
                     }
                  }
                  break;

               case ACTOR_LADDER:
                  {
                              //validate bottom ones only
                     bool is_up = a->cbProc(CB_USE_LADDER, 1);
                     if(!is_up)
                        break;
                     LPC_actor up_part = NULL;
                     float best_dist;
                     S_vector pos1_2d(a->GetPos()->x, 0.0f, a->GetPos()->z);
                     for(int j=actors.size(); j--; ){
                        LPC_actor a1 = actors[j];
                        if(a1->GetType()==ACTOR_LADDER && !a1->cbProc(CB_USE_LADDER, 1)){
                              //check distance
                           S_vector pos2_2d(a1->GetPos()->x, 0.0f, a1->GetPos()->z);
                           float d = (pos1_2d - pos2_2d).Magnitude();
                           if(!up_part || (best_dist>d)){
                              best_dist = d;
                              up_part = a1;
                           }
                        }
                     }
                     C_string str;
                     if(!up_part){
                        str = "can't find upper counterpart";
                     }else{
                        PI3D_frame frm_up = up_part->GetFrame();
                              //check their world directions
                        float a = frm->GetWorldDir().AngleTo(frm_up->GetWorldDir());
                        if(a > PI*.4f){
                           str = C_fstring("invalid angle to upper part '%s'", frm_up->GetName());
                        }else{
                                 //check if bottom is behind top
                           S_vector loc_pos = frm->GetWorldPos() * frm_up->GetInvMatrix();
                           loc_pos.y = 0.0f;
                           loc_pos.Normalize();
                           if(loc_pos.z > -.75f){
                              str = C_fstring("invalid position against upper part '%s'", frm_up->GetName());
                           }
                        }
                        if(str.Size())
                           e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm_up);
                     }
                     if(str.Size()){
                        e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                        e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                           C_fstring("Ladder '%s': %s", frm->GetName(), (const char*)str));
                        err = true;
                     }
                  }
                  break;
               }
                              //check actor's static collisions
               bool can_have_st_vols = false;
               switch(a->GetType()){
               case ACTOR_VERSION:
               case ACTOR_KILL_ANIM:
                  can_have_st_vols = true;
                  break;
               case ACTOR_AUTO_CANNON:
               case ACTOR_AUTOMOBIL:
                  {
                     LPC_table tab = a->GetTable(1);
                     can_have_st_vols = (tab->ItemF(TAB_F_AUTO_FUEL)==0);
                  }
                  break;
               }
               if(!can_have_st_vols){
                  PI3D_frame frm_sc = NULL;
                  struct S_hlp{
                     static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                        if(frm->GetFlags()&I3D_FRMF_STATIC_COLLISION){
                           *(PI3D_frame*)c = frm;
                           return I3DENUMRET_CANCEL;
                        }
                        return I3DENUMRET_OK;
                     }
                  };
                  frm->EnumFrames(S_hlp::cbEnum, (dword)&frm_sc, ENUMF_VISUAL | ENUMF_VOLUME);
                  S_hlp::cbEnum((PI3D_frame)frm, (dword)&frm_sc);
                  if(frm_sc){
                     e_slct->Action(E_SELECTION_ADD_FRAME, (void*)frm);
                     e_log->Action(E_LOG_ADD_TEXT, (void*)(const char*)
                        C_fstring("Actor '%s' has static collision (%s)", frm->GetName(), frm_sc->GetName()));
                     err = true;
                  }
               }
            }
            return err;
         }
         break;

      case E_MISSION_GET_MISSION:
         return (dword)(C_game_mission*)mission;
      }
      return 0;
   }

//----------------------------

   virtual void Render(){

      if(ed->IsActive()){
         if(draw_actor_names){
            const vector<C_smart_ptr<C_actor> > &actors = mission->GetActors();
            for(int i=actors.size(); i--; ){  
               DrawActorName(actors[i]);
            }
         }       
      }
   }

//----------------------------
#define VERSION 5

   virtual bool LoadState(class C_chunk *ck){

                              //check version
      word version = VERSION-1;
      ck->Read(&version, sizeof(word));
      if(version!=VERSION)
         return false;

      ed->CheckMenu(this, E_MISSION_TIME_0 + time_scale_id, false);

      ck->ReadString(CT_NAME, mission_name);
      ck->Read(&make_log, sizeof(byte));
      ck->Read(&tab_pos_size, sizeof(tab_pos_size));

      ck->Read(&ask_to_save, sizeof(byte));
      ck->Read(&debug_cam_acq, sizeof(byte));
      ck->Read(&merge_load, sizeof(dword));
      ck->Read(&merge_erase, sizeof(dword));
      ck->Read(&last_create_actor_index, sizeof(last_create_actor_index));
      ck->Read(&draw_actor_names, sizeof(draw_actor_names));
      ck->Read(&time_scale_id, sizeof(time_scale_id));
      ck->Read(&simulate_fps, sizeof(simulate_fps));
      ck->Read(&keep_checked_out, sizeof(keep_checked_out));

      //if(draw_actor_names) ed->RegForRender(this);
      //else ed->UnregForRender(this);
      ed->CheckMenu(this, E_MISSION_ACTOR_DRAW_NAMES, draw_actor_names);

      ed->CheckMenu(this, E_MISSION_LOG_TOGGLE, make_log);
      ed->CheckMenu(this, E_MISSION_SAVE_ASK_TOGGLE, ask_to_save);
      ed->CheckMenu(this, E_MISSION_TIME_0 + time_scale_id, true);

      ed->CheckMenu(this, E_MISSION_SIMFPS_OFF + GetSimFpsID(simulate_fps), true);
      sim_fps_last_timer = ed->GetIGraph()->ReadTimer();

      ck->ReadString(CT_NAME, last_mission_name);

      return true;
   }

   virtual bool SaveState(class C_chunk *ck) const{

                              //write version
      word version = VERSION;
      ck->Write(&version, sizeof(word));

      ck->WStringChunk(CT_NAME, mission_name);
      ck->Write(&make_log, sizeof(byte));
      ck->Write(&tab_pos_size, sizeof(tab_pos_size));

      ck->Write(&ask_to_save, sizeof(byte));
      ck->Write(&debug_cam_acq, sizeof(byte));
      ck->Write(&merge_load, sizeof(dword));
      ck->Write(&merge_erase, sizeof(dword));
      ck->Write(&last_create_actor_index, sizeof(last_create_actor_index));
      ck->Write(&draw_actor_names, sizeof(draw_actor_names));
      ck->Write(&time_scale_id, sizeof(time_scale_id));
      ck->Write(&simulate_fps, sizeof(simulate_fps));
      ck->Write(&keep_checked_out, sizeof(keep_checked_out));
      ck->WStringChunk(CT_NAME, last_mission_name);
      return true;
   }
};

//----------------------------

void InitMissionPlugin(C_game_mission *mission, LPC_editor editor){
   C_edit_Mission *em = new C_edit_Mission(mission); editor->InstallPlugin(em); em->Release();
}

#endif                        //EDITOR

//----------------------------
