#include "all.h"
#include "loader.h"
#include "tables.h"
#include "net.h"


#ifdef EDITOR
//# define CHECK_VISUAL_MATERIAL
#endif

//----------------------------
//----------------------------

C_mission::C_mission():
   tab_property(NULL),
   scene(NULL),
   open_ok(false)
{
   scene = driver->CreateScene();
   scene->SetViewport(0, 0, igraph->Scrn_sx(), igraph->Scrn_sy(), false);

   igraph->AddCallback(GraphCallback_thunk, this);
}

//----------------------------

C_mission::~C_mission(){

   igraph->RemoveCallback(GraphCallback_thunk, this);
   Close();
   //UnInit();
   scene->Release();
}

//----------------------------

void C_mission::ResetScreen(){
   scene->SetViewport(0, 0, igraph->Scrn_sx(), igraph->Scrn_sy(), false);
}

//----------------------------

dword C_mission::GraphCallback(dword msg, dword par1, dword par2){

   switch(msg){
   case CM_RECREATE:
      {
                              //re-size viewport when window resized
         switch(par2){
         case 2:
            int sx = igraph->Scrn_sx();
            int sy = igraph->Scrn_sy();
            scene->SetViewport(0, 0, sx, sy, !(igraph->GetFlags()&IG_FULLSCREEN));
            screen_aspect_ratio = (float)sy / (float)sx;
            break;
         }
      }
      break;
   }
   return 0;
}

//----------------------------

dword C_mission::GraphCallback_thunk(dword msg, dword par1, dword par2, void *context){

   C_mission *mission = (C_mission*)context;
   assert(mission);
   return mission->GraphCallback(msg, par1, par2);
}

//----------------------------

void C_mission::Close(){

   if(!open_ok)
      return;

#ifdef EDITOR
   if(editor)
      editor->BroadcastAction(EDIT_ACTION_BEFORE_FREE, 0);
#endif

                              //remove data of all ours frames
   struct S_hlp{
      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
         switch(frm->GetType()){
         case FRAME_CAMERA: case FRAME_USER: case FRAME_VOLUME:
            break;
         default:
            frm->SetData(NULL);
         }
         return I3DENUMRET_OK;
      }
   };
   scene->EnumFrames(S_hlp::cbEnum, (dword)this);

   scene->Close();

   if(tab_property){
      tab_property->Release();
      tab_property = NULL;
   }

   open_ok = false;
}

//----------------------------

void C_mission::Render(){

   //driver->BeginScene();
   scene->Render(I3DRENDER_UPDATELISTENER, this);

   //driver->EndScene();
}

//----------------------------

static PI3D_volume SetFrameVolume(PI3D_frame frm, I3D_VOLUMETYPE vt, bool model_child, PI3D_scene scene){

   PI3D_volume vol = driver->CreateVolume();

   vol->LinkTo(frm);
                              //name volume after the original frame
   if(model_child){
      vol->SetName(frm->GetOrigName());
   }else{
      C_fstring name("volume %s", frm->GetOrigName());
                              //check if this name already exists
      if(scene->FindFrame(name, ENUMF_ALL | ENUMF_BACKDROP)){
         for(int i=0; ;++i){
            C_fstring n1("%s%.2i", (const char*)name, i);
            if(!scene->FindFrame(n1, ENUMF_ALL | ENUMF_BACKDROP)){
               name = n1;
               break;
            }
         };
      }
      vol->SetName(name);
   }
                              //extract volume info from parent object
   vol->SetVolumeType(vt);
   vol->SetScene(scene);

   vol->Update();
   if(model_child){
                              //setup owner (if any)
      while(frm && frm->GetType()!=FRAME_MODEL)
         frm = frm->GetParent();
      assert(frm);
      vol->SetOwner(frm);
   }

   return vol;
}

//----------------------------

/*
static PI3D_volume MakeVols(PI3D_frame frm, PI3D_model model, PI3D_scene scene){

   PI3D_dummy dum = I3DCAST_DUMMY(frm);
   const I3D_bbox *bbox = dum->GetBBox();
   float scl = frm->GetScale();

   PI3D_volume vol;

   const char *str = frm->GetOrigName();
   switch(str[3]){
   case '0':                  //sphere
      {
         vol = driver->CreateVolume();
         vol->SetVolumeType(I3DVOLUME_SPHERE);

         float f = (bbox->max.x-bbox->min.x) * scl / 2;
         vol->SetScale(f);
         frm->SetScale(1.0f);
         frm->Update();
      }
      break;
   case '1':                  //box
      {
         vol = driver->CreateVolume();
         vol->SetVolumeType(I3DVOLUME_BOX);
         vol->SetNUScale((bbox->max-bbox->min) * scl * .5f);
         frm->SetScale(1.0f);
         frm->Update();
      }
      break;
   default:
      return NULL;
   }
   vol->SetName(frm->GetOrigName());
   vol->LinkTo(frm->GetParent());
   vol->SetPos(frm->GetPos());
   vol->SetRot(frm->GetRot());
   vol->SetScale(frm->GetScale() * vol->GetScale());
   vol->Update();

   while(frm && frm->GetType()!=FRAME_MODEL)
      frm = frm->GetParent();
   assert(frm);
   vol->SetOwner(frm);
   vol->SetScene(scene);

   model->DeleteFrame(frm);
   model->AddFrame(vol);
   vol->Release();

   return vol;
}
*/
//----------------------------

/*
I3DENUMRET I3DAPI cbModelLoad(PI3D_frame frm, dword c){

   S_model_load *s = (S_model_load*)s1;
   MakeVols(frm, s->model, s->scene);

   return I3DENUMRET_OK;
}
*/

I3DENUMRET I3DAPI cbAddVolumeToScene(PI3D_frame frm, dword c){
   I3DCAST_VOLUME(frm)->SetScene((PI3D_scene)c);
   return I3DENUMRET_OK;
}

//----------------------------

I3DENUMRET I3DAPI cbSetAllCallback(PI3D_frame frm, dword mask);

//----------------------------

void C_mission::Tick(const S_tick_context &tc){

   scene->Tick(tc.time);
   int i;

   level_time += tc.time;

                              //process scene time
#ifdef EDITOR
   if(!editor->IsActive())
#endif
   {
      while(next_time_value != -1){
         if(level_time >= tab_property->ItemI(TAB_I_SCN_TIME_VAL, next_time_value)){
            used_time_value[next_time_value] = true;
                                 //do action
            char *frm_name = tab_property->ItemS(TAB_S24_SCN_TIME_OBJ, next_time_value, 24);
            if(*frm_name){
               PI3D_frame frm = scene->FindFrame(frm_name, ENUMF_ALL);

                              //compatibility with old tables - strip '.@@'
               if(!frm){
                  int slen = strlen(frm_name);
                  if(slen>=3 && frm_name[slen-1]=='@' && frm_name[slen-2]=='@' && frm_name[slen-3]=='.'){
                     frm_name[slen-3] = 0;
                     frm = scene->FindFrame(frm_name, ENUMF_ALL);
                  }
               }
               if(frm){
                  bool b = tab_property->ItemB(TAB_B_SCN_TIME_STATUS, next_time_value);
                  frm->SetOn(b);
               }else
                  ErrReport(C_fstring("Time course: can't find frame %s", frm_name), editor);
            }
                                 //broadcast signal
            int sig_num = tab_property->ItemI(TAB_I_SCN_TIME_SIGNUM, next_time_value);
            if(sig_num){
               S_signal_data sd;
               memset(&sd, 0, sizeof(sd));
               sd.type = SIGNAL_TIME;
               sd.sig_id = sig_num;
               BroadcastMessage(CB_SIGNAL, (dword)&sd);
            }
                                 //get next key
            int j = 1e+8;
            next_time_value = -1;
            for(i=tab_property->ArrayLen(TAB_E_SCN_TIME_UNIT); i--; ){
               if(tab_property->ItemE(TAB_E_SCN_TIME_UNIT, i) == 1 && !used_time_value[i]){
                  int &ir = tab_property->ItemI(TAB_I_SCN_TIME_VAL, i);
                  if(ir<j){
                     j = ir;
                     next_time_value = i;
                  }
               }
            }
         }else
            break;
      }
   }
}

//----------------------------

void C_mission::Notify(PI3D_frame, dword notify_key, dword anim_time){

#ifdef EDITOR
   if(editor->IsActive())
      return;
#endif

   int i;
                              //find if key exists
   for(i = tab_property->ArrayLen(TAB_E_SCN_TIME_UNIT); i--; ){
      if(tab_property->ItemE(TAB_E_SCN_TIME_UNIT, i) == 2 &&
         tab_property->ItemI(TAB_I_SCN_TIME_VAL, i) == notify_key){

                              //make action
         const char *frm_name = tab_property->ItemS(TAB_S24_SCN_TIME_OBJ, i, 24);
         PI3D_frame frm = scene->FindFrame(frm_name, ENUMF_ALL);
                              //compatibility with old tables - strip '.@@'
         if(!frm){
            int slen = strlen(frm_name);
            if(slen>=3 && frm_name[slen-1]=='@' && frm_name[slen-2]=='@' && frm_name[slen-3]=='.'){
               C_string fname1 = frm_name;
               fname1[slen-3] = 0;
               frm = scene->FindFrame(fname1, ENUMF_ALL);
            }
         }
         if(frm){
            bool b = tab_property->ItemB(TAB_B_SCN_TIME_STATUS, i);
            frm->SetOn(b);
         }else
            ErrReport(C_fstring("C_mission::Notify: can't find frame %s", frm_name), editor);
      }
   }
}

//----------------------------

void TableToLight(LPC_table tp, PI3D_light lp);
void TableToSound(LPC_table tp, PI3D_sound sp, PI3D_scene);
void TableToParticle(LPC_table tp, PI3D_particle sp);

//----------------------------

                           //help class assuring that frame's name will appear
                           // before first error message
struct S_err{

#ifdef EDITOR
   LPC_editor_item e_log;
   bool first;
   const char *msg;
   PI3D_frame frm;
   S_err(LPC_editor_item ei, const char *cp, PI3D_frame f1 = NULL):
      e_log(ei), msg(cp), frm(f1), first(ei!=NULL)
   {}
   static void OpenErrFnc(const char *cp, void *context){
      S_err *ep = (S_err*)context;
      if(ep->first){
         ep->first = false;
         C_string str;
         if(ep->frm){      //flash badly, so that artists notice they've got REAL problem
            LPC_editor_item e_slct = editor->FindPlugin("Selection");
            if(e_slct) e_slct->Action(E_SELECTION_BLINK_FRAME, ep->frm);
            str = C_fstring("*** %s load errors (frame %s) ***", ep->msg, ep->frm->GetName());
         }else{
            str = C_fstring("*** %s load errors ***", ep->msg);
         }
         ep->e_log->Action(E_LOG_ADD_RICH_TEXT, &pair<const char*, dword>(str, 0xff0000));
      }
      if(ep->e_log){
         //ep->e_log->Action(E_LOG_ADD_TEXT, (void*)cp);
         ep->e_log->Action(E_LOG_ADD_RICH_TEXT, &pair<const char*, dword>(cp, 0x800000));
      }
   }
#else //EDITOR
   static void OpenErrFncRel(const char *cp, void *context){
      ErrReport(cp, NULL);
   }
#endif   //!EDITOR
};

//----------------------------

static I3DENUMRET I3DAPI AddFramesToMap(PI3D_frame frm, dword c){
   map<C_string, PI3D_frame> &frame_map = *(map<C_string, PI3D_frame>*)c;
   frame_map[frm->GetName()] = frm;
   return I3DENUMRET_OK;
}

static I3DENUMRET I3DAPI AddFramesToMap_OrigName(PI3D_frame frm, dword c){
   map<C_string, PI3D_frame> &frame_map = *(map<C_string, PI3D_frame>*)c;
   frame_map[frm->GetOrigName()] = frm;
   return I3DENUMRET_OK;
}

//----------------------------

static PI3D_visual CastVisual(PI3D_visual orig, dword vtype, PI3D_scene scene){

                              //cast visual to other type
   PI3D_visual new_vis = driver->CreateVisual(vtype);
   if(!new_vis)
      return NULL;

   I3D_RESULT ir = new_vis->Duplicate(orig);
   if(I3D_FAIL(ir)){
      new_vis->Release();
      return NULL;
   }
                              //re-link
   new_vis->LinkTo(orig->GetParent());
   while(orig->NumChildren()) orig->GetChild(0)->LinkTo(new_vis);
   new_vis->Update();
                           //close/destroy old
   orig->LinkTo(NULL);

   for(PI3D_frame model_root=new_vis; (model_root=model_root->GetParent()) && model_root->GetType()!=FRAME_MODEL; );
   if(model_root){
      I3DCAST_MODEL(model_root)->DeleteFrame(orig);
      I3DCAST_MODEL(model_root)->AddFrame(new_vis);
   }else{
      scene->DeleteFrame(orig);
      scene->AddFrame(new_vis);
   }
   new_vis->Release();
   return new_vis;
}

//----------------------------

struct S_scene_load{
   PI3D_scene scene;
   class C_mission *mission;
   dword mask;
};

static I3DENUMRET I3DAPI cbAfterLoad(PI3D_frame frm, dword s1){

   S_scene_load *s = (S_scene_load*)s1;
#ifdef EDITOR
   if(frm->GetType()==FRAME_DUMMY){
                              //make sure no obsolete volumes are in i3d scenes
      const char *name = frm->GetName();
      if(name[0]=='~' && name[1]=='0' && name[2]=='2')
         ErrReport(C_fstring("In-scene volume: %s", frm->GetName()), editor);
   }
#endif
   frm->SetCallback(cbFrame, s->mask);
   return I3DENUMRET_OK;
}

//----------------------------

/*#ifdef CHECK_VISUAL_MATERIAL

static int debug_visusals_converted_cnt;

static I3DENUMRET I3DAPI cbEnumStaticVisual(PI3D_frame frm, dword c){

   if(!(frm->GetFlags()&I3D_FRMF_STATIC_COLLISION))
      return I3DENUMRET_OK;

   assert(c);
   C_mission::S_load_context &lc = *(C_mission::S_load_context*)c;

                              //visulas without modify flag ie moutly object didn't save mat_id.
                              //to fix it we check all visuals, get material id from volume 
                              //fullface which belong to it and assign to it.
                              //if visual has no fullface, we could mark all which has no modify
                              //flag - they are sure bad.
                              //finally every other visual with  default mat can be bad because
                              //it could be changed to lit-objet after dafault mat was saved.
   switch(frm->GetType()){
   case FRAME_VISUAL:
      {
                              //visual material
         PI3D_visual vis = I3DCAST_VISUAL(frm);
         int mat_id = vis->GetCollisionMaterial();

//----------------------------
#if 0
         if(mat_id == 12){
            if(editor->CanModify())
               editor->SetModified();
            mat_id = 29;
            vis->SetCollisionMaterial(mat_id);
            dword warn_clr = 0xff0000ff;
            const char *mat_name = mat_table->ItemS(TAB_S16_MAT_NAME, mat_id, 16);
            lc.Log(C_fstring("set %i(%s) on '%s', convert from fullface",
               vis->GetCollisionMaterial(), mat_name, frm->GetName()), warn_clr ? frm : NULL, warn_clr);
            debug_visusals_converted_cnt++;
         }
         return I3DENUMRET_OK;
#endif
//----------------------------
         if(mat_id)
            break;
                              //fulface volume
         for(int i = frm->NumChildren(); i--; ){
            PI3D_frame frm_child = frm->GetChild(i);
            assert(frm_child);
            if(frm_child->GetType() != FRAME_VOLUME)
               continue;
            PI3D_volume volf = I3DCAST_VOLUME(frm_child);
            if(volf->GetVolumeType() != I3DVOLUME_FACE)
               continue;
                              //copy material
            int volf_mat = volf->GetCollisionMaterial();
            if(volf_mat){
               vis->SetCollisionMaterial(volf_mat);

               assert(lc.e_modify);
               pair<PI3D_frame, dword> p(frm, E_MODIFY_FLG_VISUAL);
               lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &p);
               assert(editor);
               if(editor->CanModify())
                  editor->SetModified();
               int warn_clr = 0;
//----------------------------//warn on material which may cause problem if not assigned properly
               if(mat_table->ItemB(TAB_B_MAT_STAIRS, volf_mat) ||
                  mat_table->ItemB(TAB_B_MAT_BORDER, volf_mat) ||
                  //mat_table->ItemB(TAB_B_MAT_AI_SEE_THROUGH, volf_mat) ||
                  //mat_table->ItemB(TAB_B_MAT_CAM_SEE_THROUGH, volf_mat) ||
                  mat_table->ItemB(TAB_B_MAT_WATER, volf_mat) ||
                  mat_table->ItemE(TAB_E_MAT_KILL, volf_mat)){
#if 0  //konvert water surface to water solid(datadisk)
                  if(volf_mat == 12){
                     volf_mat = 29; 
                     warn_clr = 0xff0000f0;
                  }else
#endif
                  warn_clr = 0xffff00ff;
               }
//----------------------------
               const char *mat_name = mat_table->ItemS(TAB_S16_MAT_NAME, volf_mat, 16);
               lc.Log(C_fstring("set %i(%s) on '%s', convert from fullface",
                  vis->GetCollisionMaterial(), mat_name, frm->GetName()), warn_clr ? frm : NULL, warn_clr);
               debug_visusals_converted_cnt++;
            }
            break;
         }

                              //if no fullface found, check for modify flag
         if(i == -1){
            assert(lc.e_modify);
            dword flags = lc.e_modify->Action(E_MODIFY_GET_FLAGS, frm);
            if(!(flags&E_MODIFY_FLG_VISUAL)){
                                 //this volume didn't saving mat_id, report it to user
               lc.WarnLog(C_fstring("usaved material on %s, probably bad.", frm->GetName()), frm);
            }else{
               lc.WarnLog(C_fstring("default material on %s, can be bad.", frm->GetName()), frm);
            }
         }
      }
      break;
   default:
      assert(0);
   }
   return I3DENUMRET_OK;
}
#endif //CHECK_VISUAL_MATERIAL
*/

//----------------------------

E_MISSION_IO C_mission::Open(const char *dir, dword open_flags){

   //bool scn_prop_read = false;

//----------------------------

   S_load_context lc;
   lc.load_cb_proc = cbLoadScene;
   lc.load_cb_context = NULL;
                              
   lc.open_flags = open_flags;
                              //debuggin support - log errors
                              // (debug compilation only)
   lc.open_dir = dir;
#ifdef EDITOR
   lc.e_log = editor ? editor->FindPlugin("Log") : NULL;
   struct S_hlp{
      static bool I3DAPI SceneErrFnc(I3D_LOADMESSAGE msg, dword prm1, dword prm2, void *context){

         C_mission::S_load_context &lc = *(C_mission::S_load_context*)context;
         switch(msg){
         case CBM_ERROR:
         case CBM_LOG:
            if(context)
               //lc.e_log->Action(E_LOG_ADD_TEXT, (void*)prm1);
               lc.e_log->Action(E_LOG_ADD_RICH_TEXT, &pair<const char*, dword>((const char*)prm1,
                  msg==CBM_ERROR ? 0xff0000 : 0x505050));
            break;
         default:
            return (*lc.load_cb_proc)(msg, prm1, prm2, lc.load_cb_context);
         }
         return false;
      }
   };
#ifdef EDITOR
   //if(!(open_flags&OPEN_NO_EDITOR))
   {
      lc.e_modify = editor ? editor->FindPlugin("Modify") : NULL;
      lc.e_undo = editor ? editor->FindPlugin("Undo") : NULL;
   }
#endif

# define OPEN_ERR_LOG(name, frm) S_err::OpenErrFnc, (void*)&S_err(lc.e_log, name, frm)
# define SCENE_LOAD_LOG S_hlp::SceneErrFnc, &lc

#else //EDITOR
# define OPEN_ERR_LOG(name, frm) S_err::OpenErrFncRel, NULL
# define SCENE_LOAD_LOG lc.load_cb_proc, lc.load_cb_context
#endif   //!EDITOR

   lc.merge_mode = (open_flags&OPEN_MERGE_MODE);

   bool load_cancelled = false;
   if(!lc.merge_mode){
                              //re-load scene
      Close();
      open_ok = true;

                              //init table
      //if(!ai_open)
      {
         if(!tab_property){
            tab_property = CreateTable();
            tab_property->Open((dword)&tab_scene_property, TABOPEN_TEMPLATE);
         }
      }

      //if(open_flags&OPEN_LOAD_SCENE)
      {
         dword i3d_open_flags = I3DLOAD_ERRORS | I3DLOAD_ANIM_COMPATIBILITY;
         if(!(open_flags&OPEN_NO_SHOW_PROGRESS))
            i3d_open_flags |= I3DLOAD_PROGRESS;
         if(open_flags&OPEN_LOG) i3d_open_flags |= I3DLOAD_LOG;
         I3D_RESULT ir;
         ir = scene->Open(C_fstring("missions\\%s\\scene.i3d", dir),
            i3d_open_flags, SCENE_LOAD_LOG);

                              //report errors (if any)
         if(I3D_FAIL(ir) && !(open_flags&OPEN_NO_OPEN_FAIL_MSG)){
            const char *err;
            switch(ir){
            case I3DERR_NOFILE: 
               err = "no such file";
               break;
            case I3DERR_OUTOFMEM:
               err = "out of memory";
               break;
            case I3DERR_FILECORRUPTED:
               err = "file corrupted";
               break;
            case I3DERR_CANCELED:
               err = "loading cancelled";
               load_cancelled = true;
               break;
            default:
               err = "uknown error";
            }
            ErrReport(C_fstring("Scene open error: %s.", err), editor);
            open_ok = false;
            //if(ir==I3DERR_CANCELED) return MIO_CANCELLED;
         }

         S_scene_load s = {scene, this, FRMCALLBACK_ANIMNOTE };
         scene->EnumFrames(cbAfterLoad, (dword)&s, ENUMF_VISUAL | ENUMF_MODEL | ENUMF_SECTOR | ENUMF_DUMMY);
      }
   }else{
      if(!open_ok)
         return MIO_ACCESSFAIL;
      if(!(open_flags&OPEN_NO_SHOW_PROGRESS)){
                              //init progress indicator
         float f = 0.0f;
         (*lc.load_cb_proc)((I3D_LOADMESSAGE)10000, *(dword*)&f, 0, lc.load_cb_context);
      }
   }

   if(open_ok){

                              //keep all in-scene frames in a map,
                              // which allows much faster name look-up
                              // than I3D_scene::FindFrame
   scene->EnumFrames(AddFramesToMap, (dword)&lc.frame_map, ENUMF_ALL | ENUMF_BACKDROP);
   lc.frame_map[scene->GetPrimarySector()->GetName()] = scene->GetPrimarySector();
   lc.frame_map[scene->GetBackdropSector()->GetName()] = scene->GetBackdropSector();

   C_string str;
   S_vector v;

   if(!lc.merge_mode){
                              //create binary filename
      str = C_fstring("missions\\%s\\scene.bin", dir);
   }else{
                              //in merge mode, the filename is full
      str = dir;
   }

   if(!lc.ck.ROpen(str)){
      if(!(open_flags&OPEN_NO_SHOW_PROGRESS)){
         float f = 1.0f;
         (*lc.load_cb_proc)((I3D_LOADMESSAGE)10000, *(dword*)&f, 2, lc.load_cb_context);
      }
      return MIO_NOFILE;
   }
   }//open_ok

   lc.cam = scene->GetActiveCamera();
   if(!lc.cam){
      if(open_ok && (open_flags&OPEN_USE_SCENE_CAMERA)){
         lc.cam = I3DCAST_CAMERA(scene->FindFrame(NULL, ENUMF_CAMERA));
         if(lc.cam)
            scene->SetActiveCamera(lc.cam);
      }
#ifdef EDITOR
      if(!lc.cam)
      if(!(open_flags&OPEN_NO_EDITOR)){
         if(editor){
            LPC_editor_item ei = editor->FindPlugin("MouseEdit");
            if(ei){
               lc.cam = (PI3D_camera)ei->Action(E_MOUSE_GET_EDIT_CAMERA);
               scene->SetActiveCamera(lc.cam);
            }
         }
      }
#endif   //EDITOR
      /*
      if(!lc.cam){
         lc.cam = driver->CreateCamera();
         scene->SetActiveCamera(lc.cam);
         scene->AddFrame(lc.cam);
         lc.cam->Release();
      }
      */
   }

   if(open_ok){

   lc.is_ortho = false;
   lc.ortho_scale = 1.0f;
   //lc.dump_chunks = true;

   if(!lc.merge_mode){
      cam_range = 100.0f;
      cam_fov = PI*.5f;
      /*
      if(lc.cam){
                              //setup defaults
         lc.cam->SetOrthogonal(lc.is_ortho);
         lc.cam->SetOrthoScale(lc.ortho_scale);
         lc.cam->SetRange(CAM_NEAR_RANGE, cam_range);
         lc.cam->SetFOV(cam_fov);
      }
      */
   }

   CK_TYPE t;
   if(!lc.ck.RAscend(t)){
      open_ok = false;
   }else
   if(t==CT_BASECHUNK){
      if(!lc.merge_mode)
         GetMissionCRC(dir, mission_crc);

      while(lc.ck.Size()){
         if(!lc.ck.RAscend(t)){
            open_ok = false;
            break;
         }
         if(!(open_flags&OPEN_NO_SHOW_PROGRESS)){
            if(lc.ShowProgress()){
               load_cancelled = true;
               open_ok = false;
               break;
            }
         }
         if(lc.dump_chunks)
            lc.WarnLog(C_fstring("Chunk 0x%x, size %i", t, lc.ck.Size()), NULL);

                              //let virtual func to load this chunk
         if(!LoadChunk(t, lc)){
            open_ok = false;
            load_cancelled = lc.cancelled;
            break;
         }
         lc.ck.Descend();
      }
                              //load BSP tree as final
      if(open_ok && lc.bsp_file_pos!=-1){
         C_cache *ch = lc.ck.GetHandle();
         ch->seekp(lc.bsp_file_pos);

         bool bsp_ok = scene->LoadBsp(lc.ck.GetHandle(), lc.load_cb_proc, lc.load_cb_context
#ifdef EDITOR            
            ,FBSPLOAD_CHECKSUM | FBSPLOAD_CONSISTENCE
#endif
            );
         if(!bsp_ok)
            lc.ErrLog("Failed to load bsp information.", NULL);
      }
   }
   lc.ck.Close();

   if(open_ok){

#ifdef CHECK_VISUAL_MATERIAL
      assert(scene);
      debug_visusals_converted_cnt = 0;
      scene->EnumFrames(cbEnumStaticVisual, (dword)&lc, ENUMF_VISUAL);
      lc.Log(C_fstring("converted %i fullfaces.",debug_visusals_converted_cnt),0,0);
#endif

#ifdef EDITOR
      if(!(open_flags&OPEN_NO_EDITOR)){
         if(!lc.merge_mode && editor){
            editor->BroadcastAction(EDIT_ACTION_AFTER_LOAD, 0);
         }
      }
#else
      //mission.game_cam.SetCamera(scene->GetActiveCamera());
#endif

      if(!lc.merge_mode)
         mission_name = dir;

      PI3D_camera cam = scene->GetActiveCamera();
      if(cam){
         cam->SetOrthogonal(lc.is_ortho);
         cam->SetOrthoScale(lc.ortho_scale);
         cam->SetRange(CAM_NEAR_RANGE, cam_range);
         cam->SetFOV(cam_fov);
      }
      {
         used_time_value.clear();
         used_time_value.reserve(tab_property->ArrayLen(TAB_E_SCN_TIME_UNIT));
                                    //get 1st key in time course
         {
            int j = 1e8;
            next_time_value = -1;
            for(int i=tab_property->ArrayLen(TAB_E_SCN_TIME_UNIT); i--; ){
               used_time_value.push_back(false);
               if(((byte*)tab_property->Item(TAB_E_SCN_TIME_UNIT))[i] == 1){
                  int &ir=((int*)tab_property->Item(TAB_I_SCN_TIME_VAL))[i];
                  if(ir<j){
                     j=ir;
                     next_time_value=i;
                  }
               }
            }
         }
      }
      level_time = 0;
   }

   }//open_ok

   if(!(open_flags&OPEN_NO_SHOW_PROGRESS)){
      float f = 1.0f;
      (*lc.load_cb_proc)((I3D_LOADMESSAGE)10000, *(dword*)&f, 2, lc.load_cb_context);
   }
   igraph->GetTimer(1, 1);

   return open_ok ? MIO_OK : load_cancelled ? MIO_CANCELLED : MIO_NOFILE;
}

//----------------------------
//#define CONV_WARNINGS         //enable conversion warnin messages

bool C_mission::LoadChunk(CK_TYPE ck_t, S_load_context &lc){

   switch(ck_t){

//----------------------------//compatible importer:
#if defined EDITOR
   case CT_TABLE:
      {
         lc.WarnLog("Converting mission property table", NULL);

         assert(tab_property);
         //bool ok = tab_property->Open((dword)lc.ck.GetHandle(), TABOPEN_FILEHANDLE | TABOPEN_UPDATE);
         bool ok = tab_property->Open((dword)lc.ck.GetHandle(), TABOPEN_FILEHANDLE);
         assert(ok);
         if(ok){
            PI3D_light l_b_a = NULL;
            PI3D_light l_b_f = NULL;
            {
                              //backdrop
               char *cp = tab_property->ItemS(TAB_S_SCN_BACKDROPNAME);
               if(*cp){
#if defined CONV_WARNINGS || 0
                  lc.WarnLog(C_fstring(" Importing backdrop: '%s' - converting to model", cp), NULL);
#endif
                              //try to import backdrop
                  C_smart_ptr<I3D_model> mp = driver->CreateModel(); mp->Release();
                  C_string str = cp;
                  for(int i=str.Size(); i--; ){
                     if(str[i]=='.'){
                        str[i] = 0;
                        break;
                     }
                  }
                  C_fstring mname("Backdrops\\%s\\%s", (const char*)lc.open_dir, (const char*)str);
                  bool b = I3D_SUCCESS(model_cache.Open(mp, mname, scene, 0, ErrReport, editor));
                  if(!b){
                     lc.ErrLog(C_fstring("  - can't load model: %s", (const char*)mname), NULL);
                  }else{
                              //setup backdrop range
                     const I3D_bound_volume &bvol = mp->GetHRBoundVolume();
                     float r = (bvol.bbox.max - bvol.bbox.min).Magnitude() * .5f;
                     scene->SetBackdropRange(1.0f, r);
                              //todo: setup all its lights to shine on it

                     dword flags = E_MODIFY_FLG_CREATE | E_MODIFY_FLG_POSITION |
                        E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_LINK;
                              //save modification
                     if(lc.e_modify){
                        pair<PI3D_frame, dword> modify_pair(mp, flags);
                        lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);
                     }
                     mp->SetName("<backdrop>");
                     scene->AddFrame(mp);

                     l_b_a = I3DCAST_LIGHT(mp->FindChildFrame("^*", ENUMF_WILDMASK | ENUMF_LIGHT));
                     if(l_b_a){
                        l_b_a->SetLightType(I3DLIGHT_AMBIENT);
                        if(lc.e_modify)
                           lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(l_b_a, E_MODIFY_FLG_LIGHT));
                     }
                     l_b_f = I3DCAST_LIGHT(mp->FindChildFrame("=*", ENUMF_WILDMASK | ENUMF_LIGHT));
                     if(l_b_f){
                        l_b_f->SetLightType(I3DLIGHT_FOG);
                        if(lc.e_modify)
                           lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(l_b_f, E_MODIFY_FLG_LIGHT));
                     }
                              //add into map
                     mp->EnumFrames(AddFramesToMap_OrigName, (dword)&lc.frame_map);
                     lc.frame_map[mp->GetName()] = mp;
                     mp->LinkTo(scene->GetBackdropSector());
                              //setup all its lights to backdrop sectors
                     struct S_hlp{
                        S_load_context *lc;
                        PI3D_sector bsect;
                        static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                           S_hlp *hp = (S_hlp*)c;
                           hp->bsect->AddLight(I3DCAST_LIGHT(frm));
                           if(hp->lc->e_modify){
                              pair<PI3D_frame, dword> modify_pair(frm, E_MODIFY_FLG_LIGHT);
                              hp->lc->e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);
                           }
                           return I3DENUMRET_OK;
                        }
                     } hlp = {&lc, scene->GetBackdropSector()};
                     mp->EnumFrames(S_hlp::cbEnum, (dword)&hlp, ENUMF_LIGHT);
                  }
               }
            }
                              //global lightness
            float f = tab_property->ItemF(TAB_F_SCN_CGLOBAL);
            //f *= tab_game_cfg->ItemF(TAB_F_GAME_GAMMA);
            if(!(!f || f==1.0f)){
               lc.ErrLog(C_fstring("  Unsupported global lightness: %.2f", f), NULL);
            }
            {
                              //background color
               S_vector c = tab_property->ItemV(TAB_CV_SCN_BGND);
               scene->SetBgndColor(c);
#if defined CONV_WARNINGS || 0
               lc.WarnLog(C_fstring("  Converting background color from table"), NULL);
#endif
            }
            PI3D_light l_p_a = I3DCAST_LIGHT(scene->GetPrimarySector()->FindChildFrame("^*", ENUMF_WILDMASK | ENUMF_LIGHT));
            if(l_p_a){
               l_p_a->SetLightType(I3DLIGHT_AMBIENT);
               if(lc.e_modify)
                  lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(l_p_a, E_MODIFY_FLG_LIGHT));
            }
            PI3D_light l_p_f = I3DCAST_LIGHT(scene->GetPrimarySector()->FindChildFrame("=*", ENUMF_WILDMASK | ENUMF_LIGHT));
            if(l_p_f){
               l_p_f->SetLightType(I3DLIGHT_FOG);
               if(lc.e_modify)
                  lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &pair<PI3D_frame, dword>(l_p_f, E_MODIFY_FLG_LIGHT));
            }

            for(int i=0; i<2; i++){
                              //default ambient/fog light
               S_vector l = tab_property->ItemV(!i ? TAB_CV_SCN_AMBIENT : TAB_CV_SCN_FOG);
               bool valid = !i ? !l.IsNull() : tab_property->ItemB(TAB_B_SCN_FOG_ON);
               if(valid){
                  const char *type_name = !i ? "ambient" : "fog";
#if defined CONV_WARNINGS || 0
                  lc.WarnLog(C_fstring("  Converting default %s to light '<default %s>'",
                     type_name, type_name), NULL);
#endif
                  PI3D_light lp = driver->CreateLight();
                  lp->SetLightType(!i ? I3DLIGHT_AMBIENT : I3DLIGHT_FOG);
                  lp->SetColor(l);
                  lp->SetName(C_fstring("<default %s>", type_name));
                  if(i){
                     lp->SetRange(tab_property->ItemF(TAB_F_SCN_FOG_NEAR), 0.0f);
                  }
                              //setup to all sectors
                  struct S_hlp{
                     static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){
                        I3DCAST_SECTOR(frm)->AddLight((PI3D_light)c);
                        return I3DENUMRET_OK;
                     }
                  };
                  bool add_to_primary = (!i ? (l_p_a==NULL) : (l_p_f==NULL));
                  if(add_to_primary){
                     scene->EnumFrames(S_hlp::cbEnum, (dword)lp, ENUMF_SECTOR);
                     S_hlp::cbEnum(scene->GetPrimarySector(), (dword)lp);
                  }
                  if(!i){
                     bool add_to_backdrop = (l_b_f==NULL);
                     if(add_to_backdrop)
                        S_hlp::cbEnum(scene->GetBackdropSector(), (dword)lp);
                  }
                  if(lp->NumLightSectors()){
                     lp->LinkTo(scene->GetPrimarySector());
                     scene->AddFrame(lp);
                                          //add into map
                     lc.frame_map[lp->GetName()] = lp;
                              //save modification
                     if(lc.e_modify){
                        pair<PI3D_frame, dword> modify_pair(lp,
                           E_MODIFY_FLG_CREATE | E_MODIFY_FLG_LIGHT | E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_LINK);
                        lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);
                     }
                  }
                  lp->Release();
               }
            }
            {
                              //camera
               int fov = tab_property->ItemI(TAB_I_SCN_CAM_FOV);
               if(fov){
                  cam_fov = (float)fov*PI/180.0f;
                  if(lc.cam)
                     lc.cam->SetFOV(cam_fov);

#if defined CONV_WARNINGS || 0
                  lc.WarnLog(C_fstring("  Using camera FOV: %i", fov), NULL);
#endif
               }
               float r = tab_property->ItemF(TAB_F_SCN_CAM_RANGE);
               if(r){
                  cam_range = r;
                  if(lc.cam)
                     lc.cam->SetRange(CAM_NEAR_RANGE, cam_range);
#if defined CONV_WARNINGS || 0
                  lc.WarnLog(C_fstring("  Using camera range: %.1f", r), NULL);
#endif
               }
            }
         }
         if(editor){
            if(editor->CanModify())
               editor->SetModified(true);
         }
      }
      break;

   case CT_MODELS:
      {
         lc.WarnLog("Importing models...", NULL);
         int count = 0;

         while(lc.ck.Size()){
            CK_TYPE ck_t;
            if(!lc.ck.RAscend(ck_t))
               return false;
            lc.ShowProgress();
            switch(ck_t){
            case CT_MODEL:    //read name, filename, pos, rot
               C_smart_ptr<I3D_model> mp = driver->CreateModel(); mp->Release();
               dword flags = E_MODIFY_FLG_CREATE;
               C_string fname;

               while(lc.ck.Size()){
                  if(!lc.ck.RAscend(ck_t))
                     return false;
                  switch(ck_t){
                  case CT_MODELFILENAME:
                     {
                        fname = lc.ck.RStringChunk();
                        bool b = I3D_SUCCESS(model_cache.Open(mp, fname, scene, 0, ErrReport, editor));
                        if(b){
                                          //init volumes from dummies
                           //S_model_load s = {mp, scene};
                           //mp->EnumFrames(cbModelLoad, (dword)&s, ENUMF_DUMMY | ENUMF_USEMASK, "~0*");
                           mp->EnumFrames(cbSetAllCallback, FRMCALLBACK_ANIMNOTE, ENUMF_VISUAL | ENUMF_DUMMY | ENUMF_JOINT, NULL);
                           mp->EnumFrames(cbAddVolumeToScene, (dword)scene, ENUMF_VOLUME);
                        }else{
                           lc.ErrLog(C_fstring("can't load model: %s", (const char*)fname), NULL);
                        }
                     }
                     break;

                  case CT_MODELNAME:
                     {
                        C_string str = lc.ck.RStringChunk();
                        mp->SetName(str);
                     }
                     break;

                  case CT_POSITION:
                     mp->SetPos(lc.ck.RVectorChunk());
                     flags |= E_MODIFY_FLG_POSITION;
                     break;

                  case CT_QUATERNION:
                     mp->SetRot(lc.ck.RQuaternionChunk());
                     flags |= E_MODIFY_FLG_ROTATION;
                     break;

                  case CT_SCALE:
                     {
                        S_vector s = lc.ck.RVectorChunk();
                        float avg_scale = (s.x + s.y + s.z) / 3.0f;
                        avg_scale = Min(avg_scale, I3D_MAX_FRAME_SCALE);
#if defined CONV_WARNINGS || 1
                        if(fabs(s.x - s.y) > .001f || fabs(s.x - s.z) > .001f){
                           lc.ErrLog(C_fstring("Non-uniform scale on model: '%s': %.2f, %.2f, %.2f - setting average: %.2f", mp->GetName(), s.x, s.y, s.z, avg_scale), mp);
                        }
#endif
                        mp->SetScale(avg_scale);
                     }
                     flags |= E_MODIFY_FLG_SCALE;
                     break;

                  case CT_MODEL_FLAGS:
                     {
                                 //read flags
                        word flg;
                        lc.ck.Read(&flg, sizeof(word));
                        assert(flg==1);
                        mp->SetUserFlags(flg&1);
                        lc.ck.Descend();
#if defined CONV_WARNINGS || 0
                        lc.WarnLog(C_fstring("Model set on map: %s", mp->GetName()), mp);
#endif
                     }
                     break;
                  default:
                     lc.ck.Descend();
                  }
               }
#if defined CONV_WARNINGS || 0
               lc.WarnLog(C_fstring("Importing model: %s (file: %s)", mp->GetName(), mp->GetFileName()), NULL);
#endif
               C_smart_ptr<I3D_frame> frm = mp;
                              //check for special models - sounds and lights
               if(fname.Matchi("sound*")){
                              //change to sound
                  PI3D_sound snd = scene->CreateSound();
                  snd->SetSoundType(I3DSOUND_POINT);
                  snd->SetPos(mp->GetPos());
                  snd->SetRot(mp->GetRot());
                  snd->SetScale(mp->GetScale());
                  snd->SetName(mp->GetName());
                  frm = snd;
                  snd->Release();
#if defined CONV_WARNINGS || 0
                  lc.WarnLog(C_fstring("Converting model to sound: %s", frm->GetName()), NULL);
#endif
               }else
               if(fname.Matchi("l_*")){
                              //change to light
                  PI3D_light orig = I3DCAST_LIGHT(mp->FindChildFrame(NULL, ENUMF_LIGHT));
                  if(orig){
                     PI3D_light lp = driver->CreateLight();
                     lp->Duplicate(orig);
                     lp->SetPos(mp->GetPos());
                     lp->SetRot(mp->GetRot());
                     lp->SetScale(mp->GetScale());
                     lp->SetName(mp->GetName());
                     switch(orig->GetOrigName()[0]){
                     case '|': lp->SetLightType(I3DLIGHT_DIRECTIONAL); break;
                     case '^': lp->SetLightType(I3DLIGHT_AMBIENT); break;
                     case '=': lp->SetLightType(I3DLIGHT_FOG); break;
                     }
                     frm = lp;
                     lp->Release();
#if defined CONV_WARNINGS || 0
                     lc.WarnLog(C_fstring("Converting model to light: %s", frm->GetName()), NULL);
#endif
                  }
               }else{
                  //S_model_load s = {mp, scene};
                  //mp->EnumFrames(cbModelLoad, (dword)&s, ENUMF_DUMMY | ENUMF_USEMASK, "~0*");
                  mp->EnumFrames(cbSetAllCallback, FRMCALLBACK_ANIMNOTE, ENUMF_VISUAL | ENUMF_DUMMY | ENUMF_JOINT, NULL);
                  mp->EnumFrames(cbAddVolumeToScene, (dword)scene, ENUMF_VOLUME);
                  mp->SetAnimation(0, NULL, I3DANIMOP_LOOP);
               }
               frm->Update();

               ++count;
                                 //save modification
               if(lc.e_modify){
                  pair<PI3D_frame, dword> modify_pair(frm, flags);
                  lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);
               }
               scene->AddFrame(frm);
                                 //add into map
               frm->EnumFrames(AddFramesToMap, (dword)&lc.frame_map);
               lc.frame_map[frm->GetName()] = frm;
               scene->SetFrameSector(frm);
            }
            lc.ck.Descend();
         }
         lc.WarnLog(C_fstring(" %i models imported", count), NULL);
      }
      break;

   case CT_VOLUME_BOXES:
      {
         lc.WarnLog("Importing volumes...", NULL);
         int count = 0;

         while(lc.ck.Size()){
            CK_TYPE ck_t;
            if(!lc.ck.RAscend(ck_t))
               return false;
            lc.ShowProgress();
            switch(ck_t){
            case CT_VOLUME:
               {
                  dword flags = 0;

                  word wtype;
                  lc.ck.Read(&wtype, 2);
                  static const I3D_VOLUMETYPE vtypes[] = {I3DVOLUME_SPHERE, I3DVOLUME_BOX,
                     (I3D_VOLUMETYPE)0xff,
                     I3DVOLUME_RECTANGLE};
                  bool created = (wtype!=0);
                  I3D_VOLUMETYPE type;
                  if(created)
                     type = vtypes[wtype-128];
                  C_string str;

                  PI3D_volume vol = NULL;
                  word vd_flags = 0;
                  word vd_mat_id = 0;
                  dword vd_data_d;
                  while(lc.ck.Size()){
                     if(!lc.ck.RAscend(ck_t))
                        return false;
                     switch(ck_t){
                     case CT_NAME:
                        {
                           str = lc.ck.RStringChunk();
                              //skip fullface, obsolote
                           if(type == 0xff){
                              lc.WarnLog(C_fstring("Skipping fullface '%s'.", str), NULL);
                              break;
                           }
                           if(created){
                              map<C_string, PI3D_frame>::const_iterator it_fmap = lc.frame_map.find(str);
                              PI3D_frame owner = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;

                              if(owner){
                                 vol = SetFrameVolume(owner, type, false, scene);
                                 scene->AddFrame(vol);
                                 vol->Release();
                                 //vd = GetVolData(vol);
                                 //vd->flags |= VF_USER_CREATED;
                              }
                              flags |= E_MODIFY_FLG_LINK | E_MODIFY_FLG_CREATE |
                                 E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION;
                           }else{
                              /*
                              for(int i=volumes.size(); i--; ){
                                 if(!strcmp(volumes[i]->GetName(), str))
                                    break;
                              }
                              if(i!=-1){
                                 vol = volumes[i];
                              }
                              */
                              map<C_string, PI3D_frame>::const_iterator it_fmap = lc.frame_map.find(str);
                              PI3D_frame frm = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
                              if(frm && frm->GetType()==FRAME_VOLUME)
                                 vol = I3DCAST_VOLUME(frm);
                           }
                        }
                        break;

                     case CT_FLOAT:
                        {
                           float f = lc.ck.RFloatChunk();
                           if(vol){
                              if(f > I3D_MAX_FRAME_SCALE){
                                 lc.ErrLog(C_fstring("Invalid scale on '%s'", vol->GetName()), vol);
                              }else
                                 vol->SetScale(f);
                           }
                        }
                        break;

                     case CT_POSITION:
                        if(vol) vol->SetPos(lc.ck.RVectorChunk());
                        else
                           lc.ck.Descend();
                        flags |= E_MODIFY_FLG_POSITION;
                        break;

                     case CT_QUATERNION:
                        if(vol) vol->SetRot(lc.ck.RQuaternionChunk());
                        else
                           lc.ck.Descend();
                        flags |= E_MODIFY_FLG_ROTATION;
                        break;

                     case CT_SCALE:
                        if(vol){
                           vol->SetNUScale(lc.ck.RVectorChunk());
                        }else
                           lc.ck.Descend();
                        flags |= E_MODIFY_FLG_NU_SCALE;
                        break;

                     case CT_VOL_FLAGS:
                        lc.ck.Read(&vd_flags, 2);
                        vd_flags &= 0xff0f;
                        /*
                        if(vd){
                           word w;
                           lc.ck.Read(&w, 2);
                           vd->SetFlags(w);
                        }
                        */
                        lc.ck.Descend();
                        break;

                     case CT_VOL_MAT_ID:
                        lc.ck.Read(&vd_mat_id, 2);
                        /*
                        if(vd){
                           word mat_id;
                           lc.ck.Read(&mat_id, 2);
                           vd->mat_id = mat_id;
                        }
                        */
                        lc.ck.Descend();
                        break;

                     case CT_VOL_USER_DATA:
                        lc.ck.Read(&vd_data_d, 4);
                        /*
                        if(vd){
                           lc.ck.Read(&vd->data_d, 4);
                        }
                        */
                        lc.ck.Descend();
                        break;

                     default:
                        lc.ck.Descend();
                     }
                  }
                  if(vol){
                     //assert(vd_mat_id < MAX_MATERIALS);
                     if(vd_mat_id<0 || vd_mat_id>=MAX_MATERIALS || !*mat_table->ItemS(TAB_S16_MAT_NAME, vd_mat_id, 16)){
                        lc.ErrLog(C_fstring("Volume '%s' has invalid material: %i - setting default", vol->GetName(), vd_mat_id), vol);
                        vd_mat_id = 0;
                     }

                              //find compatible material in table
                     bool stairs = (vd_flags&4);
                     bool border = (vd_flags&1);
                     bool ai_see_through = (vd_flags&0x1000);
                     bool cam_see_through = (vd_flags&0x2000);
                     bool water = (vd_flags&2);
                     int killmode = (vd_flags&0xf00)>>8;
                     static const char *mode_name[] = {"", "kill_spasm", "kill_water", "kill_minefield", "kill_throwout", "?!?!"};

                     if(stairs || border || ai_see_through || water || killmode){
                        const char *mat_name = mat_table->ItemS(TAB_S16_MAT_NAME, vd_mat_id, 16);
                              //find compatible material containing our flags
                        for(int i=0; i<MAX_MATERIALS; i++){
                           if(mat_table->ItemB(TAB_B_MAT_STAIRS, i) == stairs &&
                              mat_table->ItemB(TAB_B_MAT_BORDER, i) == border &&
                              mat_table->ItemB(TAB_B_MAT_AI_SEE_THROUGH, i) == ai_see_through &&
                              mat_table->ItemB(TAB_B_MAT_CAM_SEE_THROUGH, i) == cam_see_through &&
                              mat_table->ItemB(TAB_B_MAT_WATER, i) == water &&
                              mat_table->ItemE(TAB_E_MAT_KILL, i) == killmode){

                              break;
                           }
                        }
                        if(i==MAX_MATERIALS){
                           lc.ErrLog(C_fstring("Volume: '%s' - can't convert material: %s, "
                              "%s%s%s%s%s%s",
                              vol->GetName(), mat_name,
                              !stairs ? "" : "stairs ",
                              !border ? "" : "border ",
                              !ai_see_through ? "" : "ai_see_through ",
                              !cam_see_through ? "" : "cam_see_through ",
                              !water ? "" : "water ",
                              mode_name[Min(killmode, 5)]), vol);
                        }else{
                           if(i!=vd_mat_id && vd_mat_id){
                                    //check for common case - metal1 to AI_transp
                              if(vd_mat_id==2 && i==32)
                                 i = 37;

                              const char *new_name = mat_table->ItemS(TAB_S16_MAT_NAME, i, 16);
                              lc.WarnLog(C_fstring("Volume: '%s' - converting material '%s' to '%s'",
                                 vol->GetName(), mat_name, new_name), vol);
                           }
                           vd_mat_id = i;
                        }
                     }
                     vol->SetCollisionMaterial(vd_mat_id);
                     if(vd_mat_id)
                        flags |= E_MODIFY_FLG_VOLUME;

                     if(created){
                        vol->Update();
#if defined CONV_WARNINGS || 0
                        lc.WarnLog(C_fstring("Importing volume: %s", (const char*)str), NULL);
#endif
                                       //save modification
                        scene->AddFrame(vol);
                     }else{
#if defined CONV_WARNINGS || 0
                        lc.WarnLog(C_fstring("Importing volume settings: %s", (const char*)str), NULL);
#endif
                     }
                              //save modification
                     if(lc.e_modify && flags){
                        pair<PI3D_frame, dword> modify_pair(vol, flags);
                        lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);
                     }
                     ++count;
                                 //add into map
                     lc.frame_map[vol->GetName()] = vol;
                  }else{
                     lc.ErrLog(C_fstring("Failed to import volume: %s", (const char*)str), NULL);
                  }
               }
               break;
            }
            lc.ck.Descend();
         }
         lc.WarnLog(C_fstring(" %i volumes imported", count), NULL);
      }
      break;

   case CT_SELECTION:
      lc.WarnLog("Skipping selection chunk", NULL);
      break;

#endif//EDITOR
//----------------------------
   case CT_TABLE_EX:
      {
         assert(tab_property);
         bool ok;
         ok = tab_property->Open((dword)lc.ck.GetHandle(), TABOPEN_FILEHANDLE | TABOPEN_UPDATE);
         assert(ok);
      }
      break;

   case CT_SCENE_BGND_COLOR:
      if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_SETTINGS))
         break;
      {
         S_vector v;
         lc.ck.Read(&v, sizeof(v));
         scene->SetBgndColor(v);
      }
      break;

   case CT_CAMERA_FOV:
   case CT_CAMERA_RANGE:
   case CT_CAMERA_ORTHO_SCALE:
      if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_SETTINGS))
         break;
      {
         float f;
         lc.ck.Read(&f, sizeof(float));
         switch(ck_t){
         case CT_CAMERA_FOV:
            cam_fov = f;
            if(lc.cam)
               lc.cam->SetFOV(f);
            break;
         case CT_CAMERA_RANGE:
            cam_range = f;
            if(lc.cam)
               lc.cam->SetRange(CAM_NEAR_RANGE, f);
            break;
         case CT_CAMERA_ORTHO_SCALE:
            lc.ortho_scale = f;
            if(lc.cam)
               lc.cam->SetOrthoScale(f);
            break;
         }
      }
      break;

   case CT_CAMERA_ORTHOGONAL:
      if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_SETTINGS))
         break;
      lc.is_ortho = true;
      if(lc.cam)
         lc.cam->SetOrthogonal(true);
      break;

   case CT_SCENE_BACKDROP:
      if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_SETTINGS))
         break;
      {
         float r[2];
         lc.ck.Read((char*)r, sizeof(r));
         scene->SetBackdropRange(r[0], r[1]);
      }
      break;

      /*
   case CT_POLAR_MODE:
      if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_SETTINGS))
         break;
      if(!lc.merge_mode)
         SetPolarMode(true);
      break;
      */

   case CT_MODIFICATIONS:
      if(lc.merge_mode && !(lc.open_flags&(OPEN_MERGE_SOUNDS|OPEN_MERGE_LIGHTS|OPEN_MERGE_MODELS|OPEN_MERGE_VOLUMES|
         OPEN_MERGE_DUMMYS|OPEN_MERGE_CAMERAS|OPEN_MERGE_TARGETS|OPEN_MERGE_OCCLUDERS|OPEN_MERGE_LIGHTMAPS)))
         break;
      {
#ifdef EDITOR
         pair<PI3D_frame, dword> modify_pair;
#endif                        //EDITOR

         while(lc.ck.Size()){
            if(!lc.ck.RAscend(ck_t))
               return false;
            if(!(lc.open_flags&OPEN_NO_SHOW_PROGRESS)){
               if(lc.ShowProgress()){
                  lc.cancelled = true;
                  return false;
               }
            }

            switch(ck_t){
            case CT_MODIFICATION:
               I3D_FRAME_TYPE frame_type = FRAME_NULL;   //when FRAME_NULL, find in scene
               dword frame_subtype;
               PI3D_frame frm = NULL;
#ifdef EDITOR
               dword flags = 0;  //modification flags
#endif
               S_vector pos, world_pos, nu_scale, color, pivot;
               bool world_pos_ok = false;
               float scale;
               S_quat rot;
               PI3D_frame link_frm = NULL;
               bool pos_ok = false, rot_ok = false, scl_ok = false,
                  nu_scl_ok = false, hide = false, created = false;
               bool snd_or_light = false;

               C_string str;
               map<C_string, PI3D_frame>::const_iterator it_fmap;

               while(lc.ck.Size()){
                  if(!lc.ck.RAscend(ck_t))
                     return false;
                  switch(ck_t){
                  case CT_FRAME_TYPE:
                     frame_type = (I3D_FRAME_TYPE)lc.ck.RIntChunk();
                     break;

                  case CT_FRAME_SUBTYPE:
                     frame_subtype = lc.ck.RIntChunk();
                     break;

                  case CT_NAME:
                     str = lc.ck.RStringChunk();
                     if(frame_type==FRAME_NULL){
                                          //in merge mode, we only operate on created frames, or lightmaps
                        if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_LIGHTMAPS))
                           break;
                        it_fmap = lc.frame_map.find(str);
                        frm = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
#if defined _DEBUG
                        if(!frm){
                              //try find dot in the name
                           for(int i=str.Size(); i--; ) if(str[i]=='.') break;
                           if(i!=-1){
                                       //for lights and sounds, strip frame name
                              C_string str1 = str;
                              str1[i] = 0;
                              it_fmap = lc.frame_map.find(str1);
                              frm = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
                              if(frm){
                                          //lights and sounds only!
                                 if(frm->GetType()==FRAME_LIGHT || frm->GetType()==FRAME_SOUND)
                                    snd_or_light = true;
                                 else
                                    frm = NULL;
                              }
                           }
                        }
#endif
                     }
                     if(frame_type!=FRAME_NULL){
                        {
                           switch(frame_type){

                           case FRAME_LIGHT:
                              if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_LIGHTS))
                                 break;
                              frm = driver->CreateLight();
                              break;

                           case FRAME_SOUND:
                              if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_SOUNDS))
                                 break;
                              frm = scene->CreateSound();
                              break;

                           case FRAME_MODEL:
                              if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_MODELS)) break;
                              frm = driver->CreateModel();
                              break;

                           case FRAME_VOLUME: 
                              if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_VOLUMES)) break;
                              {
                                 PI3D_volume vol = driver->CreateVolume();
                                 vol->SetScene(scene);
                                 frm = vol;
                              }
                              break;

                           case FRAME_DUMMY:
                              if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_DUMMYS)) break;
                              frm = driver->CreateDummy();
                              break;

                           case FRAME_USER: frm = driver->CreateUserFrame(); break;

                           case FRAME_CAMERA:
                              {
                                 if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_CAMERAS))
                                    break;
                                 PI3D_camera cam;
                                 frm = cam = driver->CreateCamera();
                                 cam->SetRange(CAM_NEAR_RANGE, cam_range);
                                 cam->SetFOV(cam_fov);
                                 cam->SetOrthogonal(lc.is_ortho);
                                 cam->SetOrthoScale(lc.ortho_scale);

                                             //set first created camera to scene, if not set yet
                                 if(!scene->GetActiveCamera())
                                    scene->SetActiveCamera(cam);
                              }
                              break;

                           case FRAME_VISUAL:
                              if(lc.merge_mode) break;
                              frm = driver->CreateVisual(frame_subtype);
                              break;

                           case FRAME_TARGET:
                              if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_TARGETS)) break;
                              frm = driver->CreateTarget();
                              break;

                           case FRAME_OCCLUDER:
                              if(lc.merge_mode && !(lc.open_flags&OPEN_MERGE_OCCLUDERS)) break;
                              frm = driver->CreateFrame(FRAME_OCCLUDER);
                              break;

                           default: 
                              lc.ErrLog(C_fstring("Cannot create frame %s: requested type = %i", (const char*)str, frame_type), NULL);
                           }
#ifdef EDITOR
                           if(frm){
                                             //check frame duplication
                              if(lc.frame_map.find(str)!=lc.frame_map.end()){
                                 lc.ErrLog(C_fstring("Cannot create frame %s: already exist", (const char*)str), NULL);
                                 frm->Release();
                                 frm = NULL;
                              }
                           }
#endif                                          //EDITOR

                           if(frm){
#ifdef EDITOR
                              if(lc.merge_mode && lc.e_undo){
                                             //save undo info (merge mode only)
                                 S_edit_Undo_2 eu("Create", E_CREATE_DELETE_NAMED, (void*)(const char*)str, str.Size()+1);
                                 lc.e_undo->Action(E_UNDO_SAVE, &eu);
                              }
                              flags |= E_MODIFY_FLG_CREATE;
#endif   //EDITOR
                              created = true;
                              frm->SetName(str);
                              scene->AddFrame(frm);
                              frm->Release();
                                          //add into map
                              lc.frame_map[str] = frm;
                           }
                        }
                     }else{
                        if(!frm){
                           lc.ErrLog(C_fstring("Cannot initialize frame %s: not found", (const char*)str), NULL);
                        }
                     }
                     break;

                  case CT_FRAME_FLAGS:
                     {
                        dword frm_flags = lc.ck.RIntChunk();
                        if(frm)
                           frm->SetFlags(frm_flags);
                     }
                     break;

                  case CT_MODIFY_TARGET:
                     if(frm && frm->GetType()==FRAME_TARGET){
                        str = lc.ck.RStringChunk();
                        it_fmap = lc.frame_map.find(str);
                        PI3D_frame tgt = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
                        if(tgt){
                           I3DCAST_TARGET(frm)->SetFrame(tgt);
                        }else
                           lc.ErrLog(C_fstring("Target '%s': can't find frame '%s'", frm->GetName(), (const char*)str), frm);
                     }else
                        lc.ck.Descend();
                     break;

                  case CT_POSITION:
                     pos = lc.ck.RVectorChunk();
                     pos_ok = true;
#ifdef EDITOR
                     flags |= E_MODIFY_FLG_POSITION;
#endif
                     break;

                  case CT_WORLD_POSITION:
                     world_pos = lc.ck.RVectorChunk();
                     world_pos_ok = true;
#ifdef EDITOR
                     flags |= E_MODIFY_FLG_POSITION;
#endif
                     break;

                  case CT_QUATERNION:
                     rot = lc.ck.RQuaternionChunk();
                     rot_ok = true;
#ifdef EDITOR
                     flags |= E_MODIFY_FLG_ROTATION;
#endif
                     break;

                  case CT_USCALE:
                     scale = lc.ck.RFloatChunk();
                     if(scale > I3D_MAX_FRAME_SCALE)
                        lc.ErrLog(C_fstring("Invalid scale on '%s': %.3f - ignoring ", frm ? frm->GetName() : "?", scale), frm);
                     else
                        scl_ok = true;
#ifdef EDITOR
                     flags |= E_MODIFY_FLG_SCALE;
#endif
                     break;

                  case CT_NUSCALE:
                     nu_scale = lc.ck.RVectorChunk();
                     nu_scl_ok = true;
#ifdef EDITOR
                     flags |= E_MODIFY_FLG_NU_SCALE;
#endif
                     break;

                  case CT_MODIFY_LIGHT:
                     if(frm && frm->GetType()==FRAME_LIGHT){
                        PI3D_light lp = I3DCAST_LIGHT(frm);
                        while(lp->NumLightSectors())
                           lp->GetLightSectors()[0]->DeleteLight(lp);
                        while(lc.ck.Size()){
                           if(!lc.ck.RAscend(ck_t))
                              return false;
                           switch(ck_t){
                           case CT_LIGHT_TYPE:
                              {
                                 I3D_LIGHTTYPE lt = (I3D_LIGHTTYPE)lc.ck.RIntChunk();
                                 if(lt==I3DLIGHT_SPOT){
                                    lc.ErrLog(C_fstring("Light '%s': spotlight is obsolete", lp->GetName()), lp);
                                 }
                                 lp->SetLightType(lt);
                              }
                              break;
                           case CT_COLOR: lp->SetColor(lc.ck.RVectorChunk()); break;
                           case CT_LIGHT_POWER: lp->SetPower(lc.ck.RFloatChunk()); break;
                           case CT_LIGHT_CONE: 
                              {
                                 float c[2];
                                 lc.ck.Read(c, sizeof(c));
                                 lc.ck.Descend();
                                 lp->SetCone(c[0], c[1]);
                              }
                              break;
                           case CT_LIGHT_RANGE:
                              {
                                 float r[2];
                                 lc.ck.Read(r, sizeof(r));
                                 lc.ck.Descend();
                                 lp->SetRange(r[0], r[1]);
                              }
                              break;

                           case CT_LIGHT_MODE:
                              lp->SetMode(lc.ck.RIntChunk());
                              break;

                           case CT_LIGHT_SPECULAR_COLOR: lp->SetSpecularColor(lc.ck.RVectorChunk()); break;
                           case CT_LIGHT_SPECULAR_POWER: lp->SetSpecularPower(lc.ck.RFloatChunk()); break;
                           //case CT_LIGHT_SPECULAR_ANGLE: lp->SetSpecularAngle(lc.ck.RFloatChunk()); break;

                           case CT_LIGHT_SECTOR:
                              {
                                 str = lc.ck.RStringChunk();
                                 PI3D_sector sct = NULL;
                                 it_fmap = lc.frame_map.find(str);
                                 PI3D_frame frm1 = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
                                 if(frm1 && frm1->GetType()==FRAME_SECTOR) sct = I3DCAST_SECTOR(frm1);
                                 if(sct) sct->AddLight(lp);
                                 else{
                                    lc.ErrLog(C_fstring("Cannot set light '%s' to sector '%s': not found", lp->GetName(), (const char*)str), lp);
                                 }
                              }
                              break;

                           default:
                              lc.ck.Descend();
                           }
                        }
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_LIGHT;
#endif
                     }
                     lc.ck.Descend();
                     break;

                  case CT_MODIFY_SOUND:
                     if(frm && frm->GetType()==FRAME_SOUND){
                        PI3D_sound sp = I3DCAST_SOUND(frm);
                        sp->SetLoop(false);
                        sp->SetStreaming(false);
                        sp->SetOn(false);
                        bool on = false;
                        while(lc.ck.Size()){
                           if(!lc.ck.RAscend(ck_t))
                              return false;
                           switch(ck_t){
                           case CT_NAME:
                              str = lc.ck.RStringChunk();
                              if(isound){
                                 sound_cache.Open(sp, str, scene, 0, OPEN_ERR_LOG(str, sp));
                              }else{
                                 sound_cache.Open(sp, str, scene, 0, NULL, NULL);
                              }
                              break;

                           case CT_SOUND_TYPE: sp->SetSoundType((I3D_SOUNDTYPE)lc.ck.RIntChunk()); break;
                           case CT_SOUND_VOLUME: sp->SetVolume(lc.ck.RFloatChunk()); break;
                           case CT_SOUND_OUTVOL: sp->SetOutVol(lc.ck.RFloatChunk()); break;
                           case CT_SOUND_CONE: 
                              {
                                 float c[2];
                                 lc.ck.Read(c, sizeof(c));
                                 lc.ck.Descend();
                                 sp->SetCone(c[0], c[1]);
                              }
                              break;

                           case CT_SOUND_RANGE:
                              {
                                 S_vector r;
                                 lc.ck.Read(&r, sizeof(r));
                                 lc.ck.Descend();
                                 sp->SetRange(r[0], r[1], r[2]);
                              }
                              break;

                           case CT_SOUND_LOOP: 
                              sp->SetLoop(true); 
                              lc.ck.Descend();
                              break;

                           case CT_SOUND_STREAMING: 
                              sp->SetStreaming(true); 
                              lc.ck.Descend();
                              break;

                           case CT_SOUND_ENABLE: 
                              on = true;
                              lc.ck.Descend();
                              break;

                           case CT_SOUND_SECTOR:
                              {
                                 str = lc.ck.RStringChunk();
                                 PI3D_sector sct = NULL;
                                 it_fmap = lc.frame_map.find(str);
                                 PI3D_frame frm1 = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
                                 if(frm1 && frm1->GetType()==FRAME_SECTOR)
                                    sct = I3DCAST_SECTOR(frm1);
                                 if(sct) sct->AddSound(sp);
                                 else{
                                    lc.ErrLog(C_fstring("Cannot set sound to sector '%s': not found", (const char*)str), sp);
                                 }
                              }
                              break;

                           default:
                              lc.ck.Descend();
                           }
                        }
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_SOUND;
#endif
                        if(on)
                           sp->SetOn(true);
                     } 
                     lc.ck.Descend();
                     break;

                  case CT_MODIFY_VOLUME:
                     if(frm && frm->GetType()==FRAME_VOLUME){
                        PI3D_volume vp = I3DCAST_VOLUME(frm);
                        while(lc.ck.Size()){
                           if(!lc.ck.RAscend(ck_t))
                              return false;
                           switch(ck_t){
                           case CT_VOLUME_TYPE: vp->SetVolumeType((I3D_VOLUMETYPE)lc.ck.RIntChunk()); break;
                           case CT_VOLUME_MATERIAL:
                              {
                                 int mid = lc.ck.RIntChunk();
                                 if(mid<0 || mid>=MAX_MATERIALS || !*mat_table->ItemS(TAB_S16_MAT_NAME, mid, 16)){
                                    lc.ErrLog(C_fstring("Volume '%s' has invalid material: %i - setting default", vp->GetName(), mid), vp);
                                    mid = 0;
                                 }
                                 vp->SetCollisionMaterial(mid);
                              }
                              break;
                           case CT_NUSCALE: vp->SetNUScale(lc.ck.RVectorChunk()); break;
                           default:
                              lc.ck.Descend();
                           }
                        }
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_VOLUME;
#endif
                     } 
                     lc.ck.Descend();
                     break;

                  case CT_MODIFY_OCCLUDER:
                     if(frm && frm->GetType()==FRAME_OCCLUDER){
                        PI3D_occluder occ = I3DCAST_OCCLUDER(frm);
                        I3D_OCCLUDERTYPE ot = I3DOCCLUDER_SPHERE;

                        while(lc.ck.Size()){
                           if(!lc.ck.RAscend(ck_t))
                              return false;
                           switch(ck_t){
                           case CT_OCCLUDER_VERTICES:
                              {
                                 dword numv = 0;
                                 lc.ck.Read(&numv, sizeof(numv));
                                 S_vector *vp = new S_vector[numv];
                                 lc.ck.Read(vp, numv*sizeof(S_vector));
                                 occ->Build(vp, numv);
                                 delete[] vp;
                              }
                              ot = I3DOCCLUDER_MESH;
                              break;
                           }
                           lc.ck.Descend();
                        }
                        occ->SetOccluderType(ot);
                     } 
                     lc.ck.Descend();
                     break;

                  case CT_MODIFY_SECTOR:
                     if(frm && frm->GetType()==FRAME_SECTOR){
                        PI3D_sector sct = I3DCAST_SECTOR(frm);
                        while(lc.ck.Size()){
                           if(!lc.ck.RAscend(ck_t))
                              return false;
                           switch(ck_t){
                              /*
                           case CT_SECTOR_SOUND_DATA:
                              {
                                 word indx = 0;
                                 lc.ck.Read(&indx, sizeof(word));
                                 dword data;
                                 lc.ck.Read(&data, sizeof(dword));
                                 sct->SetSoundData((I3D_SECTOR_SOUND_DATA_INDEX)indx, data);
                              }
                              break;
                              */
                           case CT_SECTOR_ENVIRONMENT:
                              sct->SetEnvironmentID(lc.ck.RWordChunk());
                              break;
                           default:
                              lc.ck.Descend();
                           }
                        }
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_SECTOR;
#endif
                     } 
                     lc.ck.Descend();
                     break;

                  case CT_LINK:
                     if(frm){
                        if(!lc.ck.RAscend(ck_t))
                           return false;
                        if(ck_t==CT_NAME){
                           str = lc.ck.RStringChunk();
                           it_fmap = lc.frame_map.find(str);
                           link_frm = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;
                           if(link_frm){
#ifdef EDITOR
                              flags |= E_MODIFY_FLG_LINK;
#endif
                           }else{
                              lc.ErrLog(C_fstring("frame %s: failed to link to %s", frm->GetName(), (const char*)str), NULL);
                              if(world_pos_ok)
                                 pos = world_pos;
                           }
                        }
                     }
                     lc.ck.Descend();
                     break;

                  case CT_HIDDEN:
                     hide = true;
#ifdef EDITOR
                     flags |= E_MODIFY_FLG_HIDE;
#endif
                     lc.ck.Descend();
                     break;

                  case CT_MODELFILENAME:
                     if(frm && frm->GetType()==FRAME_MODEL){
                        str = lc.ck.RStringChunk();
                        PI3D_model mod = I3DCAST_MODEL(frm);

                              //strip extension (old-style format)
                        if(str.Size()>=4 && str[(int)str.Size()-4]=='.'){
                           lc.WarnLog(C_fstring("Name '%s': stripping extension", frm->GetName(), (const char*)str), NULL);
                           str[(int)str.Size()-4] = 0;
                        }
                        if(I3D_SUCCESS(model_cache.Open(mod, str, scene, 0, OPEN_ERR_LOG(str, mod)))){
                           mod->SetAnimation(0, NULL, I3DANIMOP_LOOP);
                                          //init volumes from dummies
                           //S_model_load s = {mod, scene};
                           //mod->EnumFrames(cbModelLoad, (dword)&s, ENUMF_DUMMY | ENUMF_USEMASK, "~0*");
                           mod->EnumFrames(cbSetAllCallback, FRMCALLBACK_ANIMNOTE, ENUMF_VISUAL | ENUMF_DUMMY | ENUMF_JOINT, NULL);
                           mod->EnumFrames(cbAddVolumeToScene, (dword)scene, ENUMF_VOLUME);

                                          //add children into map
                           mod->EnumFrames(AddFramesToMap, (dword)&lc.frame_map);
                        }else{
                           lc.ErrLog(C_fstring("Frame '%s': can't load model: %s", frm->GetName(), (const char*)str), NULL);
                        }
                     }else
                        lc.ck.Descend();
                     break;

                  case CT_MODEL_DEL_FRAME:
                     if(frm && frm->GetType()==FRAME_MODEL){
                        str = lc.ck.RStringChunk();
                        PI3D_model mod = I3DCAST_MODEL(frm);
                                       //find specified frame in model
                        PI3D_frame frm_del = mod->FindChildFrame(str);
                        if(frm_del){
                                       //remove also from frame-map
                           map<C_string, PI3D_frame>::iterator it = lc.frame_map.find(frm_del->GetName());
                           assert(it != lc.frame_map.end());
                           lc.frame_map.erase(it);

                           mod->DeleteFrame(frm_del);
#ifdef EDITOR
                           if(lc.e_modify){
                              lc.e_modify->Action(E_MODIFY_ADD_MODEL_DEL_CHILD, 
                                 &pair<PI3D_frame, const char*>(mod, str));
                           }
#endif
                        }else{
                           lc.ErrLog(C_fstring("can't delete frame %s in model %s",
                           (const char*)str, mod->GetName()), NULL);
                        }
                     }else
                        lc.ck.Descend();
                     break;

                  case CT_MODIFY_DUMMY:
                     {
                        I3D_bbox bb;
                        lc.ck.Read(&bb, sizeof(I3D_bbox));
                        lc.ck.Descend();
                        if(frm && frm->GetType()==FRAME_DUMMY)
                           I3DCAST_DUMMY(frm)->SetBBox(bb);
                     }
                     break;

                  case CT_MODIFY_VISUAL:
                     {
                        dword vtype = lc.ck.RIntChunk();
                        if(frm && frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()!=vtype){
                                             //cast visual to other type
                           PI3D_visual new_vis = CastVisual(I3DCAST_VISUAL(frm), vtype, scene);
                           if(new_vis){
                              lc.frame_map[new_vis->GetName()] = new_vis;
                              frm = new_vis;
#ifdef EDITOR
                              flags |= E_MODIFY_FLG_VISUAL;
#endif
                           }else{
                              lc.ErrLog(C_fstring("failed to cast visual %s to type 0x%x", frm->GetName(), vtype), frm);
                           }
                        }
                     }
                     break;

                  case CT_MODIFY_LIT_OBJECT:
                     if(frm && frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_LIT_OBJECT){
                        PI3D_lit_object lmap = (PI3D_lit_object)frm;

                        /*
#ifdef EDITOR
                        if(lc.merge_mode && lc.e_undo)
                           SaveLMUndo(lmap, lc.e_undo);
#endif                                       //EDITOR
                           */

                        I3D_RESULT ir = lmap->Load((dword)lc.ck.GetHandle());
                        if(I3D_FAIL(ir)){
                           lc.ErrLog(C_fstring("can't load light-map: %s", (const char*)str), lmap);
                        }
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_VISUAL;
#endif
                     }
                     lc.ck.Descend();
                     break;

                  case CT_MODIFY_VISUAL_PROPERTY:
                     if(frm && frm->GetType()==FRAME_VISUAL){
                        PI3D_visual vis = I3DCAST_VISUAL(frm);

                        I3D_PROPERTYTYPE ptype = I3DPROP_NULL;
                        lc.ck.Read(&ptype, sizeof(byte));

                        word prop_index;
                        lc.ck.Read(&prop_index, sizeof(prop_index));

                        byte tmp_data[16];
                        dword value;
                        switch(ptype){
                        case I3DPROP_BOOL:
                           value = 0;
                           lc.ck.Read(&value, sizeof(byte));
                           vis->SetProperty(prop_index, value);
                           break;
                        case I3DPROP_INT:
                        case I3DPROP_FLOAT:
                        case I3DPROP_ENUM:
                           lc.ck.Read(&value, sizeof(value));
                           vis->SetProperty(prop_index, value);
                           break;
                        case I3DPROP_VECTOR3:
                        case I3DPROP_COLOR:
                           {
                              lc.ck.Read(tmp_data, 12);
                              vis->SetProperty((dword)tmp_data, value);
                           }
                           break;
                        default:
                           assert(0);
                        }
                     }
                     lc.ck.Descend();
                     break;

//----------------------------//compatible importer:
#if defined _DEBUG

                  case CT_LSECTOR:
                     if(frm && frm->GetType()==FRAME_LIGHT){
                        str = lc.ck.RStringChunk();
                        PI3D_sector sct = NULL;

                        it_fmap = lc.frame_map.find(str);
                        PI3D_frame fsct = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;

                        if(fsct && fsct->GetType()==FRAME_SECTOR) sct = I3DCAST_SECTOR(fsct);
                        if(!sct){
                                       //try new sector name format "sector %s"
                           str = C_fstring("sector %s", ((const char*)str)+1);
                           it_fmap = lc.frame_map.find(str);
                           fsct = (it_fmap!=lc.frame_map.end()) ? (*it_fmap).second : NULL;

                           if(fsct && fsct->GetType()==FRAME_SECTOR) sct = I3DCAST_SECTOR(fsct);
                        }
#if defined CONV_WARNINGS || 0
                        lc.WarnLog(C_fstring("Light '%s' - conversion - adding to sector '%s'", frm->GetName(), (const char*)str), NULL);
#endif
                        if(sct){
                           sct->AddLight(I3DCAST_LIGHT(frm));
                        }else{
                           lc.ErrLog(C_fstring("Cannot set light '%s' to sector '%s': not found", frm->GetName(), (const char*)str), frm);
                        }
                     }else
                        lc.ck.Descend();
                     break;

                  case CT_LIGHT:
                     if(frm && frm->GetType()==FRAME_LIGHT){
#if defined CONV_WARNINGS || 0
                        lc.WarnLog(C_fstring("Light '%s' - importing table.", frm->GetName()), NULL);
#endif
                        PI3D_light lp = I3DCAST_LIGHT(frm);
                        LPC_table table = CreateTable();
                        table->Open((dword)lc.ck.GetHandle(), TABOPEN_FILEHANDLE);
                        TableToLight(table, lp);
                        table->Release();
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_LIGHT;
#endif
                     }
                     lc.ck.Descend();
                     break;

                  case CT_LIGHTCOLOR:
                     if(frm && frm->GetType()==FRAME_LIGHT){
                        I3DCAST_LIGHT(frm)->SetColor(lc.ck.RVectorChunk());
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_LIGHT;
#endif
                     }else
                        lc.ck.Descend();
                     break;

                  case CT_LIGHTPOWER:
                     if(frm && frm->GetType()==FRAME_LIGHT){
                        I3DCAST_LIGHT(frm)->SetPower(lc.ck.RFloatChunk());
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_LIGHT;
#endif
                     }else
                        lc.ck.Descend();
                     break;

                  case CT_SOUND:
                     if(frm && frm->GetType()==FRAME_SOUND){
#if defined CONV_WARNINGS || 0
                        lc.WarnLog(C_fstring("Sound '%s' - importing table.", frm->GetName()), NULL);
#endif
                        PI3D_sound sp = I3DCAST_SOUND(frm);
                        LPC_table table = CreateTable();
                        //table->Open((dword)&tab_sound_params, TABOPEN_TEMPLATE);
                        table->Open((dword)lc.ck.GetHandle(), TABOPEN_FILEHANDLE);
                        TableToSound(table, sp, scene);
                              //import sound sectors
                        for(int i=0; i<table->ArrayLen(TAB_S12_SOUND_SCTNAME); i++){
                           const char *cp = table->ItemS(TAB_S12_SOUND_SCTNAME, i, 12);
                           if(*cp){
                              PI3D_frame fsct = scene->FindFrame(cp, ENUMF_SECTOR | ENUMF_WILDMASK | ENUMF_BACKDROP);
                              if(!fsct){
                                       //todo: try modified sector name
                                       //try new sector name format "sector %s"
                                 str = C_fstring("sector %s", cp+1);
                                 fsct = scene->FindFrame(str, ENUMF_SECTOR | ENUMF_WILDMASK | ENUMF_BACKDROP);
                              }
                              if(fsct){
                                 I3DCAST_SECTOR(fsct)->AddSound(sp);
                              }else{
                                 lc.ErrLog(C_fstring("Can't find sound sector: '%s'.", cp), NULL);
                              }
                           }
                        }

                        table->Release();
                     }
                     lc.ck.Descend();
                     break;

                  case CT_PARTICLE:
                     if(frm && frm->GetType()==FRAME_VISUAL){
#if defined CONV_WARNINGS || 0
                        lc.WarnLog(C_fstring("Frame '%s' - converting to particle.", frm->GetName()), NULL);
#endif
                        PI3D_visual vis = I3DCAST_VISUAL(frm);
                        if(vis->GetVisualType()!=I3D_VISUAL_PARTICLE){
                                       //cast visual to particle
                           PI3D_visual new_vis = CastVisual(vis, I3D_VISUAL_PARTICLE, scene);
                           assert(new_vis);
                           lc.frame_map[new_vis->GetName()] = new_vis;
                           frm = new_vis;
                        }
#ifdef EDITOR
                        flags |= E_MODIFY_FLG_VISUAL;
#endif

                        PI3D_particle prt = (PI3D_particle)frm;
                        LPC_table table = CreateTable();
                        table->Open((dword)&tab_particle_params, TABOPEN_TEMPLATE);
                        table->Open((dword)lc.ck.GetHandle(), TABOPEN_FILEHANDLE | TABOPEN_UPDATE);
                        TableToParticle(table, prt);
                        table->Release();
                     }
                     lc.ck.Descend();
                     break;

#endif//EDITOR
//----------------------------
                  case CT_VISUAL_MATERIAL:
                     if(frm && frm->GetType()==FRAME_VISUAL){
                        {
                           PI3D_visual vis = I3DCAST_VISUAL(frm);
                           int mid = lc.ck.RIntChunk();
                           if(mid<0 || mid>=MAX_MATERIALS || !*mat_table->ItemS(TAB_S16_MAT_NAME, mid, 16)){
                              lc.ErrLog(C_fstring("Visual '%s' has invalid material: %i - setting default", vis->GetName(), mid), vis);
                              mid = 0;
                           }
#ifdef EDITOR
                           flags |= E_MODIFY_FLG_VISUAL;
#endif
                           vis->SetCollisionMaterial(mid);
                        }
                     }else
                        lc.ck.Descend();
                     break;

                  default:
                     lc.ErrLog(C_fstring("skipping chunk: 0x%.4x, size: %i", ck_t, lc.ck.Size()), NULL);
                     lc.ck.Descend();
                  }
               }

               if(frm){
#ifdef EDITOR                 //save modification (must do it before actually applying changes)
                  if(lc.e_modify){
                     modify_pair.first = frm;
                     modify_pair.second = flags;
                     lc.e_modify->Action(E_MODIFY_ADD_FLAGS, &modify_pair);
                  }
#endif//EDITOR

                  if(!snd_or_light){
                     if(link_frm){
                              //compatibility - if only link specified, update matrix
                        if(!((pos_ok && rot_ok) || (frm->GetType()==FRAME_MODEL))){
#ifdef EDITOR
                           flags |= E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE;
#endif
                           frm->LinkTo(link_frm, I3DLINK_UPDMATRIX);
                        }else
                           frm->LinkTo(link_frm);
                     }
                     if(pos_ok){
                        if(I3D_FAIL(frm->SetPos(pos))){
                           lc.ErrLog(C_fstring("Failed to set position on frame '%s'", frm->GetName()), frm);
#ifdef EDITOR
                           lc.e_modify->Action(E_MODIFY_DEL_FLAGS, &pair<PI3D_frame, dword>(frm, E_MODIFY_FLG_POSITION));
#endif
                        }
                     }
                     if(rot_ok){
                        if(I3D_FAIL(frm->SetRot(rot))){
                           lc.ErrLog(C_fstring("Failed to set rotation on frame '%s'", frm->GetName()), frm);
#ifdef EDITOR
                           lc.e_modify->Action(E_MODIFY_DEL_FLAGS, &pair<PI3D_frame, dword>(frm, E_MODIFY_FLG_ROTATION));
#endif
                        }
                     }
                     if(scl_ok){
                        if(I3D_FAIL(frm->SetScale(scale))){
                           lc.ErrLog(C_fstring("Failed to set scale on frame '%s'", frm->GetName()), frm);
#ifdef EDITOR
                           lc.e_modify->Action(E_MODIFY_DEL_FLAGS, &pair<PI3D_frame, dword>(frm, E_MODIFY_FLG_SCALE));
#endif
                        }
                     }
                     if(nu_scl_ok){
                        switch(frm->GetType()){
                        case FRAME_VOLUME: I3DCAST_VOLUME(frm)->SetNUScale(nu_scale); break;
                        default:
                           {
#ifdef EDITOR
                              const S_vector &s = nu_scale;
                              float avg_scale = (s.x + s.y + s.z) / 3.0f;
#if defined CONV_WARNINGS || 1
                              if(fabs(s.x - s.y) > .001f || fabs(s.x - s.z) > .001f){
                                 lc.ErrLog(C_fstring("Non-uniform scale on frame '%s': %.2f, %.2f, %.2f - setting average: %.2f", frm->GetName(), s.x, s.y, s.z, avg_scale), frm);
                              }
#endif
                              flags |= E_MODIFY_FLG_SCALE;
                              if(scl_ok)
                                 avg_scale *= scale;
                              frm->SetScale(avg_scale);
#endif//EDITOR
                           }
                        }
                     }
                     if(hide) frm->SetOn(false);
                  }
                  frm->Update();
                           //created un-linked frames are automatically placed into sector
                  if(created && !link_frm){
                     //scene->SetFrameSector(frm);
                     SetFrameSector(frm);
                  }

                           //validity check
                  switch(frm->GetType()){
                  case FRAME_VOLUME:
                     {
                        /*PI3D_volume vol = I3DCAST_VOLUME(frm);
                        switch(vol->GetVolumeType()){
                        case I3DVOLUME_FACE:
                           PI3D_frame prnt = vol->GetParent();
                           bool ok = false;
                           if(prnt->GetType() == FRAME_VISUAL)
                           switch(I3DCAST_VISUAL(prnt)->GetVisualType()){
                           case I3D_VISUAL_OBJECT:
                           case I3D_VISUAL_LIT_OBJECT:
                              ok = true;
                              break;
                           }
                           if(!ok){
                              if(created){
                                 lc.ErrLog(C_fstring("fullface volume '%s': not linked to geometry frame, removing from scene.", frm->GetName()), NULL);
#ifdef EDITOR                                          //remove modification
                                 if(lc.e_modify){
                                    lc.e_modify->Action(E_MODIFY_DEL_FRAME, frm);
                                 }
                                 scene->DeleteFrame(frm);
                                 frm = NULL;
#endif   //EDITOR
                              }else{
                                 lc.ErrLog(C_fstring("fullface volume '%s': not linked to geometry frame.", frm->GetName()), NULL);
                              }
                           }
                           break;
                        }*/
                     }
                     break;
                  }
               }
               break;
            }
            lc.ck.Descend();
         }
      }
      break;

   case CT_BSP_FILE:
                              //save position in file, postpone loading
      lc.bsp_file_pos = lc.ck.GetHandle()->tellg();
      break;

#ifdef EDITOR
   case CT_EDITOR_PLUGIN:
      if(lc.merge_mode || !editor)
         break;
      if(!(lc.open_flags&OPEN_NO_EDITOR)){
         C_string str;
         if(lc.ck.ReadString(CT_NAME, str)){
            LPC_editor_item ei = editor->FindPlugin(str);
            if(ei){
               ei->Action(EDIT_ACTION_MISSION_LOAD, &lc.ck);
            }else{
               lc.ErrLog(C_fstring("Cannot find plugin: %s", (const char*)str), NULL);
            }
         }
      }
      break;
#endif   //EDITOR

   case CT_COPYRIGHT:
   case CT_VERSION:
   case CT_CHECKPOINTS:
   case CT_CAMERA:
                              //silently skip these
      break;

   default:
      lc.ErrLog(C_fstring("skipping chunk: 0x%.4x, size: %i", ck_t, lc.ck.Size()), NULL);
   }
   return true;
}

//----------------------------

bool C_mission::GetMissionCRC(const C_string &name, dword &out_crc){

   C_fstring scene_name("missions\\%s\\scene.i3d", (const char*)name);
   C_fstring bin_name("missions\\%s\\scene.bin", (const char*)name);

   out_crc = 0;
                              //try to open both, get size and date
   int h;
   bool b;
   dword ftime[2];

                              //get crc of scene file
   h = dtaOpen(scene_name);
   if(h == -1)
      return false;

   dtaSeek(h, 0, DTA_SEEK_END);
   out_crc ^= dtaSeek(h, 0, DTA_SEEK_CUR);

   b = dtaGetTime(h, ftime);
   dtaClose(h);
   if(!b)
      return false;
   out_crc ^= ftime[0] ^ ftime[1];

                                 //get crc of bin file
   h = dtaOpen(bin_name);
   if(h == -1)
      return false;

   dtaSeek(h, 0, DTA_SEEK_END);
   out_crc ^= dtaSeek(h, 0, DTA_SEEK_CUR);

   b = dtaGetTime(h, ftime);
   dtaClose(h);
   if(!b)
      return false;
   out_crc ^= ftime[0] ^ ftime[1];

   return true;
}

//----------------------------

void C_mission::SetFrameSector(PI3D_frame frm){

   S_vector pos;
   switch(frm->GetType()){
   case FRAME_MODEL:
      {
         const I3D_bound_volume &bvol = I3DCAST_MODEL(frm)->GetHRBoundVolume();
         pos = ((bvol.bbox.min + bvol.bbox.max) * .5f) * frm->GetMatrix();
      }
      break;
   case FRAME_VISUAL:
      {
         const I3D_bound_volume &bvol = I3DCAST_VISUAL(frm)->GetBoundVolume();
         pos = ((bvol.bbox.min + bvol.bbox.max) * .5f) * frm->GetMatrix();
      }
      break;
   default:
      {
         I3D_bound_volume bvol;
         frm->ComputeHRBoundVolume(&bvol);
         if(bvol.bbox.IsValid())
            pos = ((bvol.bbox.min + bvol.bbox.max) * .5f) * frm->GetMatrix();
         else
            pos = frm->GetMatrix()(3);
      }
   }
   scene->SetFrameSectorPos(frm, pos);
}

//----------------------------

void C_mission::LinkToFloor(PI3D_frame frame, const S_vector &pos, PI3D_frame frm_ignore, bool change_pos){
                           //link to floor
   S_vector from = pos;
   PI3D_frame vol;
   S_vector r_pos, r_norm;
   while(true){
      struct S_hlp{
         static bool I3DAPI ColRespWaterIgnore(I3D_collision_data &cd){
            int mat_id = 0;
            PI3D_frame frm = cd.GetHitFrm();
            switch(frm->GetType()){
            case FRAME_VOLUME:
               mat_id = I3DCAST_VOLUME(frm)->GetCollisionMaterial();
               break;
            case FRAME_VISUAL:
               mat_id = I3DCAST_VISUAL(frm)->GetCollisionMaterial();
               break;
            case FRAME_JOINT:
               break;
            }
            if(mat_table->ItemB(TAB_B_MAT_BORDER, mat_id) ||
               mat_table->ItemB(TAB_B_MAT_WATER, mat_id))
               return false;
                           //fly through humans volumes(react only on joints)
            if(mat_id == HUMAN_SPHERE_MAT)
               return false;

            return true;
         }
      };

      vol = NULL;
      I3D_collision_data cd(from, S_vector(0, -1, 0), 0, frm_ignore, S_hlp::ColRespWaterIgnore);
      if(scene->TestCollision(cd)){
         vol = cd.GetHitFrm();
         r_pos = cd.GetHitPosition();
         r_norm = cd.GetHitNormal();
      }
      if(vol){
                           //find actor
         LPC_actor act = (LPC_actor)vol->GetData();
         if(act){
            bool brk = false;
            switch(act->GetType()){
            case ACTOR_AUTOMOBIL:
            case ACTOR_AUTO_CANNON:
            case ACTOR_VERSION:
            case ACTOR_CANNON:
               brk = true;
               break;
            }
            if(!brk){
               from = r_pos;
               from.y -= .001f;
               continue;
            }
         }
         break;
      }else break;
   }
   if(vol && vol->GetParent()){
      if(change_pos){
         frame->SetPos(r_pos);
         frame->Update();
      }
      frame->LinkTo(vol->GetParent(), I3DLINK_UPDMATRIX);
   }else{
      if(change_pos){
         frame->SetPos(pos);
         frame->Update();
      }
      scene->SetFrameSector(frame);
   }
}

//----------------------------
//----------------------------

