#include "all.h"
#include "winstuff.h"
#include <dplay.h>
#include "net.h"


//----------------------------

                              //GUID for network game
//static const GUID GAME_GUID = 
//{ 0xcfd51f35, 0xc3db, 0x4626, { 0xb6, 0x10, 0x95, 0x46, 0xcd, 0xa3, 0x6e, 0x55 } };
static const WIN_GUID NET_APP_GUID = {
   0x35, 0x1f, 0xd5, 0xcf, 0xdb, 0xc3, 0x26, 0x46, 0xb6, 0x10, 0x95, 0x46, 0xcd, 0xa3, 0x6e, 0x55
};

static C_log_mt_safe net_log;

//----------------------------

static void __stdcall NetLogFunc(const char *cp){

   if(net)
      net_log.WriteString(cp);
}

//----------------------------

C_string MakeGameText(const C_string&);

//----------------------------

bool NetConnect(){

   assert(!net);
   assert(igraph);

   igraph->UpdateParams(0, 0, 0, 0, IG_SUSPENDINACTIVE);


   INET_INIT ni;
   memset(&ni, 0, sizeof(ni));
   ni.flags = 0;
   ni.app_guid = (const void*)&NET_APP_GUID;
   ni.session_name = game_configuration.net_sess_name;
   ni.player_name = game_configuration.net_plr_name;
   ni.connection = game_configuration.net_ctype;
   ni.max_players = NET_MAX_PLAYERS;
   ni.ip_host_name = game_configuration.net_tcpip_addr;
   ni.port = game_configuration.net_tcpip_port;
   ni.modem_phone_number = game_configuration.net_modem_num;
   ni.drop_timeout = 15000;

                              //override by command-line
   if(game_configuration.net_cmdline_sess_name.Size())
      ni.session_name = game_configuration.net_cmdline_sess_name;
   if(game_configuration.net_cmdline_plr_name.Size())
      ni.player_name = game_configuration.net_cmdline_plr_name;
   if(game_configuration.net_cmdline_tcpip_addr.Size())
      ni.ip_host_name = game_configuration.net_cmdline_tcpip_addr;

                              //if no session name provided, use player's name
   if(!*ni.session_name)
      ni.session_name = ni.player_name;

   switch(game_configuration.voice_chat_compression){
   case S_game_configuration::VCC_1_2: ni.voice_compression = INET_INIT::VS_1_2; break;
   case S_game_configuration::VCC_3_2: ni.voice_compression = INET_INIT::VS_3_2; break;
   case S_game_configuration::VCC_6_4: ni.voice_compression = INET_INIT::VS_6_4; break;
   case S_game_configuration::VCC_8: ni.voice_compression = INET_INIT::VS_8; break;
   case S_game_configuration::VCC_13: ni.voice_compression = INET_INIT::VS_13; break;
   case S_game_configuration::VCC_32: ni.voice_compression = INET_INIT::VS_32; break;
   case S_game_configuration::VCC_64: ni.voice_compression = INET_INIT::VS_64; break;
   default: assert(0);
   }

//#ifndef NDEBUG
   if(game_configuration.net_debug_log){
      net_log.log_name = C_fstring("%s.log", (const char*)ni.player_name);
      net_log.f_log = fopen(net_log.log_name, "wt");
      ni.log_func = NetLogFunc;
   }
//#endif

#ifndef NDEBUG
   ni.drop_timeout = 300000;
   net = InetCreate(&ni, game_configuration.net_debug_log);
#else
   net = InetCreate(&ni, false);
#endif
   if(!net)
      goto fail1;

   net_msg = net->CreateOutMsg();   
   return true;

fail1:
   igraph->UpdateParams(0, 0, 0, IG_SUSPENDINACTIVE, IG_SUSPENDINACTIVE);
   return false;
}

//----------------------------

void NetCloseAll(bool allow_suspend){

   bool was_join = false;
   if(net && net_msg){
                              //flush messages
      net->Send(net_msg);
   }
                              //destroy interfaces
   if(net_msg){
      net_msg->Release();
      net_msg = NULL;
   }
   if(net){
      was_join = (!net->IsHost());
      net->Close();
      net->Release();
      net = NULL;
   }
#ifdef _DEBUG
   if(net_log.f_log){
      fclose(net_log.f_log);
      net_log.f_log = NULL;
      net_log.log_name = NULL;
   }
#endif
   if(allow_suspend && igraph){
                              //let inactive application get paused
      igraph->UpdateParams(0, 0, 0, IG_SUSPENDINACTIVE, IG_SUSPENDINACTIVE);
   }

#ifdef FINAL
                              //if joiner, re-open last profile, so that we play on what we loaded
   if(was_join){
      vector<S_logon_account> accounts;
      SearchAccounts(accounts);
      if(accounts.size())
         game_profile.Open(accounts.front().file_name);
   }
#endif
}

//----------------------------

#ifdef EDITOR
void NetStatUpdate(){

   if(!editor)
      return;
   LPC_editor_item ei = editor->FindPlugin("Stats");
   if(!ei)
      return;

   if(!net){
      ei->Action(E_STATS_SET_NET_TEXT,(void*)(const char*)C_fstring("Uninitialized"));
      return;
   }

   char name[256];
   dword nlen = sizeof(name);
   net->GetPlayerName(net->GetPID(), name, nlen);
   dword in, out_g, out_ng;
   net->GetStats(&in, &out_g, &out_ng);

   C_fstring str(
      "Name: %s\n"
      "%s\n"
      "In: %i\n"
      "Out (g) : %i\n"
      "Out (ng): %i",
      name,
      net->IsHost() ? "host" : "join",
      in,
      out_g,
      out_ng
      );
   ei->Action(E_STATS_SET_NET_TEXT, (void*)(const char*)str);
}
#endif   //EDITOR

//----------------------------
// Send message in outgoing queue, if specified time from last send elapsed,
// or message is long enough.
//----------------------------
#define NET_SEND_MIN_TIME_DELTA 50        //in milliseconds
#define NET_SEND_MIN_MSG_SIZE (64*8)      //in bits

bool NetSend(bool immediate){

   static dword last_send_time = 0;
   static dword last_send_time_sys = 0;

   assert(net_msg && net);

   if(immediate){
      net->Send(net_msg);
   }else{
      dword time = igraph->ReadTimer();
      dword msg_size = net_msg->GetSize(true) + net_msg->GetSize(false);
      if((msg_size > NET_SEND_MIN_MSG_SIZE) || ((time-last_send_time) > NET_SEND_MIN_TIME_DELTA)){
         net->Send(net_msg);
         last_send_time = time;
      }
   }
   return true;
}

//----------------------------
// Player message processing callback.
static bool INETAPI PlrMessagesCallback(PI_net inet, PC_net_in_msg msg, dword context){

   dword sender_pid = msg->GetSenderPID();
   if(!sender_pid){
                              //system message
      //tick_class->ProcessSystemNetMessage(inet, msg, context);
      switch(msg->GetID()){
      case INET_SYSMSG_PLAYER_DESTROYED:
         {
            dword pid = msg->GetDword();
            char buf[256];
            msg->GetString(buf, sizeof(buf));
            OUTTEXT(C_fstring("Player disconnected: %s", buf));
            tick_class->OnPlayerDisconnect(pid, buf);
         }
         break;

      case INET_SYSMSG_SESSION_OVER:
         {
            console_main.OutText(all_txt[20007], -1, all_txt.IsWide(20007));
            tick_class->OnSessionLost();
            NetCloseAll();
            return false;
         }
         break;
      }
      return true;
   }

                              //process all chunks in the message
   E_NETWORK_MESSAGE msg_id;
   while(msg_id = (E_NETWORK_MESSAGE)msg->GetID()){
                              //process known messages, dispatch unknown
      switch(msg_id){
      case NM_TEXT_CHAT:
         {
            msg->Ascend();
                              //find who's sending
            char plr_name[256];
            plr_name[0] = 0;
            inet->GetPlayerName(sender_pid, plr_name, sizeof(plr_name));
            OUTTEXT(C_fstring("%s:  %s", plr_name, (const char*)msg->GetBuf()));
            msg->Descend();
         }
         break;

      default:
         tick_class->NetIn(msg_id, msg);
      }
   }
   return true;
}

//----------------------------
// Receive new messages, send out
void NetProcessMessages(){

   assert(net);
   net->ProcessMessages(PlrMessagesCallback, 0, INPROCESS_SKIPOLDER);
   if(net)
      NetSend(false);
}

//----------------------------

void MakeSortedNetList(vector<S_net_player> &v, bool including_ours){

   assert(net);
                              //get names of all players
   dword num_plrs = 0;
   net->GetPlayers(NULL, &num_plrs);
   vector<dword> pids(num_plrs);
   net->GetPlayers(pids.begin(), &num_plrs);
   if(including_ours)
      pids.push_back(net->GetPID());
   for(int i=num_plrs+1; i--; ){
      char buf[256];
      if(net->GetPlayerName(pids[i], buf, sizeof(buf)))
         v.push_back(S_net_player(buf, pids[i]));
   }
   sort(v.begin(), v.end());

}

//----------------------------
//----------------------------

