#include "all.h"
#include "net.h"
#include <insanity\controller.h>

#define CURSOR_BLINK_TIME 150

//----------------------------
                              //C_chat - network chatting class

void C_net_chat::ShowText(const char *cp, bool rc){

   HideText();
   console_main.texts.Init();
   text = console_main.texts.CreateText(cp, -1, .075f, .6f, .045f);
   text->Release();
                              //automatically reset cursor
   if(rc){
      chat_cursor_phase = false;
      chat_cursor_count = 0;
   }
}

//----------------------------

void C_net_chat::Open(){

   if(!net) return;

   chat_buffer[chat_buf_pos = 0] = 0;
   chat_cursor_count = 0;
   chat_cursor_phase = true;
   mode = MODE_ENTER;
}

//----------------------------

void C_net_chat::Close(){

   HideText();
   mode = MODE_NULL;
}

//----------------------------

void C_net_chat::Tick(const S_tick_context &tc, bool auto_open){

   if(!net)
      return;

   char ascii_key = igraph->GetAscii(tc.key, igraph->GetShiftKeys()&SKEY_SHIFT);
   if(ascii_key=='`'){
                              //special key, used in text formatting
                              // cannot be displayed
      ascii_key = 0;
   }

   switch(mode){
   case MODE_NULL:
      if(auto_open && isprint(ascii_key))
         Open();
      else
         break;
                              //flow...
   case MODE_ENTER:
      switch(tc.key){
      case K_ENTER:
         {
            const bool g = true;
                              //send chat message
            net_msg->Ascend(chat_msg_id);
                              //message data (including terminating NULL character)
            for(int i=0; i<chat_buf_pos+1; i++)
               net_msg->AddByte(chat_buffer[i], g);

            net_msg->Descend();
                              //output text
            char name[256];
            name[0] = 0;
            net->GetPlayerName(net->GetPID(), name, sizeof(name));
            C_fstring str("%s:  %s", name, chat_buffer);
            OUTTEXT(str);
         }
                              //flow...
      case K_ESC:             //quit chat mode
         {
            HideText();
            mode = MODE_KEY_RELEASE;
                              //ignore pressed key
            int slot, ii;
            if(tc.p_ctrl->FindSlot(C_controller::ID_KEYBOARD, tc.key, slot, ii)){
               tc.p_ctrl->IgnoreValue(slot, ii);
            }
         }
         return;

      case K_BACKSPACE:
      case K_DEL:
         --chat_buf_pos;
         if(chat_buf_pos>0){
            chat_buffer[chat_buf_pos] = 0;
            ShowText(C_fstring("%s_", (const char*)chat_buffer), true);
         }else{
            HideText();
            mode = MODE_KEY_RELEASE;
            return;
         }
         break;
      }
                              //add char
      if(isprint(ascii_key)){
         if(chat_buf_pos<CHAT_BUFFER_SIZE-1){
            chat_buffer[chat_buf_pos++] = ascii_key;
            chat_buffer[chat_buf_pos] = 0;
            ShowText(C_fstring("%s_", (const char*)chat_buffer), true);
         }
      }else{
                              //blink chat cursor
         if((chat_cursor_count -= tc.time) <= 0){
            chat_cursor_count = CURSOR_BLINK_TIME;
            chat_cursor_phase = !chat_cursor_phase;
            ShowText(C_fstring("%s%s", chat_buffer, !chat_cursor_phase ? "" : "_"));
         }
      }
      break;

   case MODE_KEY_RELEASE:
      if(tc.key==K_NOKEY)
         mode = MODE_NULL;
      break;
   }
}

//----------------------------

void C_net_chat::NetIn(class C_net_in_msg *msg){

   msg->Ascend();
                     //find who's sending
   char plr_name[256];
   plr_name[0] = 0;
   dword sender_pid = msg->GetSenderPID();
   net->GetPlayerName(sender_pid, plr_name, sizeof(plr_name));
   OUTTEXT(C_fstring("%s:  %s", plr_name, (const char*)msg->GetBuf()));
   msg->Descend();

}

//----------------------------
//----------------------------
