#include "all.h"
#include "net.h"


//----------------------------

class C_net_sync_imp: public C_net_sync{
                              //possible codes (2 significant bits sent)
   enum E_NET_CODE{
      CODE_REQUEST_EVENT_0 = 0,  //request event (when event_id = 0, to save bits)
      CODE_REQUEST_EVENT = 1, //request event (sent another 16 bits with event_id)
      CODE_ANSWER = 2,        //answer to event request
      CODE_ACK_RESET = 3,     //acknowledge reset
   };

   const word net_id;         //message chunk id
   const dword hdr_bits, num_hdr_bits;  //message header

   
   E_SYNC_STATE state;        //current state

   bool lock;                 //lock - disabling event request
   dword winner_pid;          //PID of winner, available in SS_RESOLVED state
   word winner_event_id;      //event ID of winner, available in SS_RESOLVED state

   vector<dword> pid_to_answer;  //IDs of players not answered yet to recent query
   vector<dword> save_answer_pids;//IDs of players which sent answer code earlier than we actually knew about an event

   vector<dword> pid_to_reset;   //IDs of players not answered yet to reset request
   vector<dword> save_reset_pids;//IDs of players which sent reset code earlier than we actually wanted to Reset

                              //another cached event
   bool event_in_cache;

   word cached_event_id;      //cached sync event id

//----------------------------
// Add data to message. Data are determined by code (2 lower bits are significant),
// and optional event ID.
   void SendCode(E_NET_CODE code, word event_id = 0){

      const bool g = true;
                              //insert chunk ID or header
      if(net_id)
         net_msg->Ascend(net_id);
      if(num_hdr_bits)
         net_msg->AddBits(hdr_bits, num_hdr_bits, g);
                              //insert code
      assert(code < 4);
      net_msg->AddBits(code, 2, g);
                              //insert event ID (if any)
      if(event_id)
         net_msg->AddWord(event_id, g);
                              //close chunk
      if(net_id)
         net_msg->Descend();
   }

//----------------------------
// Add all net players into a vector.
   void AddPlayers(vector<dword> &v){

      vector<dword> va;
      //while(true){
         dword num_pl = 0;
         net->GetPlayers(NULL, &num_pl);
         va.assign(num_pl);
         bool ok = net->GetPlayers(va.begin(), &num_pl);
         //if(ok) break;
         if(!ok)
            return;
      //}
      v.insert(v.end(), va.begin(), va.end());
   }

//----------------------------

   void AddAnswer(dword pid){

      for(int i=pid_to_answer.size(); i--; ){
         if(pid_to_answer[i]==pid){
            pid_to_answer[i] = pid_to_answer.back(); pid_to_answer.pop_back();
            break;
         }
      }
      if(i==-1){
         save_answer_pids.push_back(pid);
         NET_LOG(C_fstring("C_net_sync: early reset received, saving (0x%x)", pid));
      }else{
         if(!pid_to_answer.size()){
#ifdef _DEBUG
            char plr_name[256], *pp;
            if(winner_pid!=net->GetPID()){
               plr_name[0] = 0;
               net->GetPlayerName(winner_pid, plr_name, sizeof(plr_name));
               pp = plr_name;
            }else{
               pp = "<this>";
            }
            NET_LOG(C_fstring("C_net_sync: all answered, event winner: %s, event_id = %i", pp, winner_event_id));
#endif
            state = SS_RESOLVED;
         }else{
            NET_LOG(C_fstring("C_net_sync: new answer_count = %i", pid_to_answer.size()));
         }
      }
   }

//----------------------------

   void AddReset(dword pid){

      for(int i=pid_to_reset.size(); i--; ){
         if(pid_to_reset[i]==pid){
            pid_to_reset[i] = pid_to_reset.back(); pid_to_reset.pop_back();
            break;
         }
      }
      if(i==-1){
         save_reset_pids.push_back(pid);
         NET_LOG(C_fstring("C_net_sync: early reset received, saving (0x%x)", pid));
      }else{
         if(!pid_to_reset.size()){
                           //all participants have sent reset, we're really reset now
            state = SS_NULL;
            NET_LOG("C_net_sync: all reset");
            if(event_in_cache)
               BeginCachedEvent();
         }else{
            NET_LOG(C_fstring("C_net_sync: new reset_count = %i", pid_to_reset.size()));
         }
      }
   }

//----------------------------

   void BeginCachedEvent(){

      assert(event_in_cache);
      assert(state==SS_NULL);
      NET_LOG(C_fstring("C_net_sync: beginning cached event (%i)", cached_event_id));
      event_in_cache = false;
      BeginEvent(cached_event_id);
   }

//----------------------------

   void PutToProgressState(){

      assert(state==SS_NULL);

      NET_LOG("C_net_sync: putting to progress state");
      SendCode(CODE_ANSWER);

      AddPlayers(pid_to_answer);

                              //erase all early answers
      for(int i=save_answer_pids.size(); i--; ){
         dword pid = save_answer_pids[i];
         for(int j=pid_to_answer.size(); j--; ){
            if(pid_to_answer[j]==pid){
               pid_to_answer[j] = pid_to_answer.back(); pid_to_answer.pop_back();
               break;
            }
         }
         //assert(j!=-1);
      }
      save_answer_pids.clear();
      NET_LOG(C_fstring("C_net_sync:: answer_count = %i", pid_to_answer.size()));

      state = SS_PROGRESS;

      if(!pid_to_answer.size()){
         assert(0);
         state = SS_RESOLVED;
      }
   }

//----------------------------

public:
   C_net_sync_imp(word nid, dword hdr_bits1, byte num_hdr_bits1):
      net_id(nid),
      hdr_bits(hdr_bits1),
      num_hdr_bits(num_hdr_bits1),
      lock(false),
      event_in_cache(false),
      winner_pid(0),
      state(SS_NULL)
   {}

//----------------------------

   virtual E_SYNC_STATE GetState(dword *wid, dword *eid) const{

      switch(state){
      case SS_RESOLVED:
      case SS_PROGRESS:
         if(wid)
            *wid = winner_pid;
         if(eid)
            *eid = winner_event_id;
         break;
      }
      return state;
   }

//----------------------------

   virtual bool BeginEvent(word event_id){

      if(lock){
         NET_LOG("C_net_sync: cannot begin event - lock");
         return false;
      }
      if(state!=SS_NULL){
         NET_LOG(C_fstring("C_net_sync: cannot begin event - state = %i", state));
         return false;
      }
      if(pid_to_reset.size()){
         if(event_in_cache){
            NET_LOG(C_fstring("C_net_sync: cannot begin event - reset_count = %i & event_in_cache", pid_to_reset.size()));
            return false;
         }
         NET_LOG(C_fstring("C_net_sync: cannot begin event - reset_count = %i, putting into cache", pid_to_reset.size()));
         assert(state==SS_NULL);
         event_in_cache = true;
         cached_event_id = event_id;
         return true;
      }
      if(save_answer_pids.size()){
         NET_LOG(C_fstring("C_net_sync: cannot begin event - num_saved_answers = %i", save_answer_pids.size()));
         return false;
      }

      NET_LOG("C_net_sync: putting to sync mode");
      SendCode(!event_id ? CODE_REQUEST_EVENT_0 : CODE_REQUEST_EVENT, event_id);
                              //by default we're the winner
      winner_pid = net->GetPID();
      winner_event_id = event_id;
      state = SS_PROGRESS;

      assert(!pid_to_answer.size());
      AddPlayers(pid_to_answer);
      if(!pid_to_answer.size())
         state = SS_RESOLVED;
      return true;
   }

//----------------------------

   virtual void Reset(){

      assert(state==SS_RESOLVED);
      winner_pid = 0;
      winner_event_id = 0;
      assert(!pid_to_answer.size());
      state = SS_NULL;

      SendCode(CODE_ACK_RESET);

      AddPlayers(pid_to_reset);
                              //erase all early resets
      for(int i=save_reset_pids.size(); i--; ){
         dword pid = save_reset_pids[i];
         for(int j=pid_to_reset.size(); j--; ){
            if(pid_to_reset[j]==pid){
               pid_to_reset[j] = pid_to_reset.back(); pid_to_reset.pop_back();
               break;
            }
         }
         //assert(j!=-1);
      }  
      save_reset_pids.clear();

      NET_LOG(C_fstring("C_net_sync::Reset, reset_count = %i", pid_to_reset.size()));
      if(!pid_to_reset.size()){
         NET_LOG("C_net_sync: all reset");
         if(event_in_cache)
            BeginCachedEvent();
      }
   }

//----------------------------

   virtual void Lock(bool b){

      lock = b;
      assert(!event_in_cache);
      /*
      if(!lock && event_in_cache)
         SendCachedEvent();
      */
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg, bool skip_chunk_header){

                              //if we use chunks, ascend and descend at the end
      if(skip_chunk_header)
         in_msg->Ascend();
      dword sender_pid = in_msg->GetSenderPID();
      word event_id = 0;

      do{
         switch(in_msg->GetBits(2)){
         case CODE_REQUEST_EVENT:
            event_id = in_msg->GetWord();
         case CODE_REQUEST_EVENT_0:
                              //event request from someone
            switch(state){
            case SS_NULL:           
                              //answering to asker
               PutToProgressState();
                              //flow...
            case SS_PROGRESS:
                              //collecting answers
               {
                  if(!winner_pid || winner_pid<sender_pid){
                     winner_pid = sender_pid;
                     winner_event_id = event_id;
                  }
                  AddAnswer(sender_pid);
               }
               break;

            default:
               assert(0);
            }
            event_id = 0;
            break;

         case CODE_ANSWER:
                              //passive answer from participant
            AddAnswer(sender_pid);
            break;

         case CODE_ACK_RESET:
                              //reset code from participant
            AddReset(sender_pid);
            break;

         default:
            assert(0);
         }
                              //loop if we use chunks, because more codes may be stored in the same chunk
      }while(skip_chunk_header && in_msg->GetChunkSize());

                              //descend the chunk
      if(skip_chunk_header)
         in_msg->Descend();
   }

//----------------------------

   virtual bool GetAnsweredPIDs(vector<dword> &pids){

      if(state!=SS_PROGRESS)
         return false;
                              //get all lpayers
      AddPlayers(pids);
                              //remove non-answered
      for(int i=pid_to_answer.size(); i--; ){
         for(int j=pids.size(); j--; ){
            if(pids[j]==pid_to_answer[i]){
               pids[j] = pids.back(); pids.pop_back();
               break;
            }
         }
      }
      return true;
   }

//----------------------------

   virtual void OnPlayerDisconnect(dword pid){

      NET_LOG(C_fstring("C_net_sync: disconnecting player 0x%x", pid));
                              //forget this player
      switch(state){
      case SS_PROGRESS:
         {
            assert(pid_to_answer.size());
            for(int i=pid_to_answer.size(); i--; ){
               if(pid_to_answer[i]==pid){
                  pid_to_answer[i] = pid_to_answer.back(); pid_to_answer.pop_back();
               }
            }
            if(!pid_to_answer.size())
               state = SS_RESOLVED;
         }
         break;
      }
      if(pid_to_reset.size()){
         for(int i=pid_to_reset.size(); i--; ){
            if(pid_to_reset[i]==pid){
               pid_to_reset[i] = pid_to_reset.back(); pid_to_reset.pop_back();
            }
         }
         if(!pid_to_reset.size()){
                              //all participants have sent reset, we're really reset now
            state = SS_NULL;
            if(event_in_cache)
               BeginCachedEvent();
         }
      }
      if(pid_to_answer.size()){
         for(int i=pid_to_answer.size(); i--; ){
            if(pid_to_answer[i]==pid){
               pid_to_answer[i] = pid_to_answer.back(); pid_to_answer.pop_back();
            }
         }
         if(!pid_to_answer.size()){
                              //all participants have answered, we're really resolved now
            state = SS_RESOLVED;
         }
      }
   }
};

//----------------------------

C_net_sync *CreateNetSync(word net_id, dword hdr_bits, byte num_hdr_bits){

   return new C_net_sync_imp(net_id, hdr_bits, num_hdr_bits);
}

//----------------------------

