#include "all.h"
#include "net.h"
#include "tables.h"


//----------------------------

void TableToParticle(LPC_table tp, PI3D_particle sp);

//----------------------------
                              //particles over network:
                              // - sent over network:
                              //    - S_vector position
                              //    - S_vector direction

class C_particle: public C_actor{
   PI3D_particle prt;
   PI3D_frame sector;
public:
   C_particle(C_game_mission &m):
      C_actor(m, ACTOR_PARTICLE),
      sector(NULL),
      prt(NULL)
   {
   }
   ~C_particle(){
      struct S_hlp{
         static I3DENUMRET I3DAPI cbLight(PI3D_frame frm, dword sct){
            ((PI3D_sector)sct)->DeleteLight(I3DCAST_LIGHT(frm));
            return I3DENUMRET_OK;
         }
      };
      model->EnumFrames(S_hlp::cbLight, (dword)sector, ENUMF_LIGHT, NULL);
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      C_actor::SetFrame(frm);
      assert(frm->GetType()==FRAME_MODEL);
      PI3D_model m = I3DCAST_MODEL(frm);
      PI3D_animation_set as = m->GetAnimationSet();
      if(as)
         m->SetAnimation(0, as);
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      if(model) model->Tick(tc.time);
      int i = 0;
      if(prt){
                                 //check if any elements alive
         i = prt->GetProperty(PRT_NUMELEMENTS);
      }
      if(!i && model){
                                 //determine if there's some KF interpolator
         const PI3D_interpolator *intps = model->GetInterpolators();
         for(i=model->NumInterpolators(); i--; ){
            PI3D_interpolator intp = intps[i];
            if(intp->GetAnimation(0))
               break;
         }
         i = (i!=-1);
      }
      if(!i)
         mission.DestroyActor(this);
      return 0;
   }

//----------------------------

   dword cbProc(dword msg, dword prm1, dword prm2){

      switch(msg){

      case CB_PAR_SETPARAMS:
         if(model){
            int i;
                                 //open table
            if(prm1){
               LPC_table table=CreateTable();
//#ifdef EDITOR
               table->Open((dword)&tab_particle_params, TABOPEN_TEMPLATE);
//#endif
               C_fstring tab_name("tables\\%s", (char*)prm1);
               i = table->Open((dword)(const char*)tab_name, TABOPEN_FILENAME | TABOPEN_UPDATE);
               if(!i){
                  ReportActorError(C_fstring("Cannot open table: %s", (const char*)tab_name));
               }else{
                  prt = (PI3D_particle)model->FindChildFrame("*", ENUMF_VISUAL | ENUMF_USEMASK);
                  if(prt && I3DCAST_VISUAL(prt)->GetVisualType()!=I3D_VISUAL_PARTICLE)
                     prt = NULL;
                  if(prt){
                     //prt->Reset();
                     TableToParticle(table, prt);
                  }
               }
               table->Release();
            }
                                 //setup lights
            struct S_hlp{
               static I3DENUMRET I3DAPI cbLight(PI3D_frame frm, dword sct){
                  PI3D_light lp = I3DCAST_LIGHT(frm);
                  ((PI3D_sector)sct)->AddLight(lp);
                              //modify light to be also dynamic for lightmaps
                  lp->SetMode(lp->GetMode() | I3DLIGHTMODE_DYNAMIC_LM);
                  return I3DENUMRET_OK;
               }
            };
            sector = model->GetParent();
            while(sector && sector->GetType()!=FRAME_SECTOR) sector=sector->GetParent();
            model->EnumFrames(S_hlp::cbLight, (dword)sector, ENUMF_LIGHT, NULL);
            break;
         }
         break;
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }


//----------------------------
   virtual bool MaySaveGame() const{ return false; }
};

//----------------------------

LPC_actor CreateParticle(C_game_mission &gm){

   return new C_particle(gm);
}

//----------------------------
