#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"



//----------------------------

C_table_template templ_resc_table = {
   // {A1068C80-5937-11d2-9B9E-004F4905FF1A}
   { 0xa1068c80, 0x5937, 0x11d2, { 0x9b, 0x9e, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Resource setup",
   NULL,
   {
      {TE_INT, TAB_I_RES_INVTYPE,  "Inventory type"},
      {TE_INT, TAB_I_RES_AMOUNT,       "Amount", 0, 0, 1},
      {TE_STRING, TAB_S8_RES_RND_CP,   "Random position CP", 8},
      {TE_NULL}
   }
};

//----------------------------
                              //resource over network:
                              // - sent over network:
                              //    - word net_id
                              //    - S_vector pos
                              //    - byte itm_id
                              //    - word amount
                              // - sent over network (not at creation time):
                              //    - pick synchronization

class C_resource: public C_actor{
   LPC_table tab;
   C_smart_ptr<C_net_sync> net_sync;

   LPC_actor net_sync_picker; //used in network game
   bool created_run_time;
public:
   C_resource(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_RESOURCE),
      net_sync_picker(NULL),
      created_run_time(true),
      tab(CreateTable())
   {
      tab->Open((dword)&templ_resc_table, TABOPEN_TEMPLATE);

      if(init_data){
         const S_resource_init &ri = *(S_resource_init*)init_data;

         model = driver->CreateModel();
         frame = model;
         const char *m_name = tab_inventory->ItemS(TAB_S16_INV_MODEL_RES, ri.inv_id, 16);
         if(!*m_name)
            m_name = tab_inventory->ItemS(TAB_S16_INV_MODEL_S, ri.inv_id, 16);
         model_cache.Open(model, m_name, mission.GetScene(), 0, ErrReport, editor);
         /*
         {
                              //dig into earth
            S_vector r_pos, r_norm;
            I3D_collision_data cd(ri.pos, S_vector(0, -1, 0), 0);
            bool b = mission.GetScene()->TestCollision(cd);
            if(b){
               r_pos = cd.GetHitPosition();
               r_norm = cd.GetHitNormal();
            }else
               r_pos = ri.pos;
            model->SetPos(r_pos + S_vector(0, .05f, 0));
            model->Update();
            mission.GetScene()->SetFrameSector(model);
         }
         */
         mission.LinkToFloor(frame, ri.pos, NULL, true);

         tab->ItemI(TAB_I_RES_INVTYPE) = ri.inv_id;
         tab->ItemI(TAB_I_RES_AMOUNT) = ri.amount;

         network_actor = ri.netw_actor;
                                 //send over network
         if(net && !network_actor){
                              //assign ID
            mission.CreateActorNetID(this);

            bool g = true;
            net_msg->AddWord(C_game_mission::NM_GAME_CREATE_RESOURCE, g);
            net_msg->AddWord(net_id, g);
            net_msg->AddVector(ri.pos, g);
            net_msg->AddByte(ri.inv_id, g);
            net_msg->AddWord(ri.amount, g);
         }
      }
   }

   ~C_resource(){
      tab->Release();
   }

//----------------------------

   virtual void SetNetID(word id){

      net_sync = CreateNetSync(id);
      net_sync->Release();
      C_actor::SetNetID(id);
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      if(created_run_time){
         SG_SaveModelInfo(ck);
                           //write item and amount
         ck->Write(&tab->ItemI(TAB_I_RES_INVTYPE), sizeof(byte));
         ck->Write(&tab->ItemI(TAB_I_RES_AMOUNT), sizeof(word));
      }else
      if(*tab->ItemS(TAB_S8_RES_RND_CP)){
         SG_SaveModelInfo(ck);
      }
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      if(ck->Size()){
         SG_LoadModelInfo(ck);
         if(!*tab->ItemS(TAB_S8_RES_RND_CP)){
                           //read item and amount
            byte b;
            word w;
            ck->Read(&b, sizeof(byte));
            tab->ItemI(TAB_I_RES_INVTYPE) = b;
            ck->Read(&w, sizeof(word));
            tab->ItemI(TAB_I_RES_AMOUNT) = w;
         }
      }

      return true;
   }

//----------------------------

   dword C_resource::cbProc(dword msg, dword prm1, dword prm2){

      switch(msg){

      case CB_USE_RESOURCE:
         switch(prm1){
         case 0:                 //pick request
            if(NumUsePositions()){
               LPC_actor picker = (LPC_actor)prm2;
               if(!net){
                                 //pick directly
                  return picker->cbProc(CB_USE_RESOURCE, 2, (dword)this);
               }else{
                  if(net_sync->BeginEvent())
                     net_sync_picker = picker;
                  return false;
               }
            }
            break;
         case 1:                 //get ID
            return tab->ItemI(TAB_I_RES_INVTYPE);
         case 3:                 //set ID
            tab->ItemI(TAB_I_RES_INVTYPE) = prm2;
            break;
         case 4:                 //set amount
            tab->ItemI(TAB_I_RES_AMOUNT) = prm2;
            break;
         }
         break;
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      net_sync->NetIn(in_msg);
   }

//----------------------------

   virtual void NetPlrDisconnect(dword pid){

      net_sync->OnPlayerDisconnect(pid);
      C_actor::NetPlrDisconnect(pid);
   }

//----------------------------

   virtual int NumUsePositions(C_actor *asker = NULL) const{
      if(init_ok){
         if(net && net_sync->GetState())
            return 0;
         return 1;
      }
      return 0;
   }

//----------------------------

   virtual int Use(LPC_actor caller, dword prm, bool force_use_off){

      if(init_ok){
         if(!net || !net_sync->GetState())
            return caller->cbProc(CB_ITEM_PICK);
      }
      return C_actor::Use(caller, prm, force_use_off);
   }

//----------------------------

   virtual dword Tick(const S_tick_context &tc){

      if(net){
         dword winner_pid;
         if(net_sync->GetState(&winner_pid) == C_net_sync::SS_RESOLVED){
            if(winner_pid==net->GetPID()){
               assert(net_sync_picker);
               net_sync_picker->cbProc(CB_USE_RESOURCE, 2, (dword)this);
            }else{
               mission.DestroyActor(this);
            }
         }
      }
      return 0;
   }

//----------------------------

   virtual bool IsOnMap() const{
      if(!init_ok)
         return false;
      if(net_sync && net_sync->GetState())
         return false;
      return true;
   }

//----------------------------

   virtual LPC_table GetTable(int index) const{
      if(index==0)
         return tab;
      return C_actor::GetTable(index);
   }

//----------------------------

   virtual class C_table_template *GetTemplate(int) const{ return &templ_resc_table; }

   virtual void MissionSave(C_chunk *ck) const{ SaveTable(ck, tab); }
   virtual bool MissionLoad(C_chunk *ck, int table_id){
      created_run_time = false;
      return LoadTable(ck, tab);
   }

//----------------------------

   virtual void TableUpdate(int index, bool not_load_prg){

                              //apply random checkpoint in single-player only
      if(!net){
         const char *cp1 = tab->ItemS(TAB_S8_RES_RND_CP);
         if(*cp1){
            LPC_checkpoint cp = mission.FindCP(cp1);
            if(cp){
                              //make a few hyper-links
               for(int i=S_int_random(10)+5; i--; )
                  cp = cp->GetCPoint(S_int_random(cp->NumConnections()));

               model->SetPos(cp->GetPos());
               model->Update();
               mission.SetFrameSector(model);
            }else{
               ReportActorError(C_fstring("can't find checkpoint %s", cp1));
            }
         }
      }
   }
};

//----------------------------

LPC_actor CreateResource(C_game_mission &gm, void *data){

   return new C_resource(gm, data);
}

//----------------------------
