#include "all.h"
#include <math\ease.hpp>


//----------------------------

C_string MakeGameText(const C_string &str);

//----------------------------

static S_scrolltext_style credits_styles[] = {
// size     color         italic    align                   pause
   {.0375f,   0xffffffff,    PI*.1f,  TEXT_JUSTIFY_CENTER,    800},          //italic, big, centered
   {.026f,  0xc0fff0a0,    0.0f,    TEXT_JUSTIFY_LEFT,      1000},           //small, left-aligned
   {.045f,   0xffffffc0,    PI*.05f, TEXT_JUSTIFY_CENTER,    2000},          //italic, big, centered
};

static void CreateCreditsText(C_string &txt){

                              //build credits text
   const int TEXT_BASE = 11000;
   for(int i=0; ; i++){
      if(!all_txt.IsDefined(TEXT_BASE+i))
         continue;
      const C_string &str = all_txt[TEXT_BASE+i];
      //if(game_configuration.language!=5){
         txt += str;
         txt += "\n";
         /*
      }else{
         int sum = the_end_text.Size() + str.Size();
         char *cp = new char[sum];
         if(the_end_text.Size()) memcpy(cp, &the_end_text[0], the_end_text.Size());
         memcpy(cp+the_end_text.Size(), &str[0], str.Size());
         the_end_text.Assign(cp, sum);
         delete[] cp;
      }
      */
      if(!str[0])
         break;
   }
}

#define SCROLL_SPEED     .045f
#define TEXT_FADE_Y      .0375f


//----------------------------

class C_scrolltext_imp: public C_scrolltext{
   struct S_text_line{
      LPC_text text;             //for other scrolls
      dword color;
      float alpha;
      byte phase;
      float base_y;
      float pos[2];
      bool operator < (const S_text_line &tl) const{ return this < &tl; }
      bool operator == (const S_text_line &tl) const{ return this == &tl; }
   };
   C_poly_text &texts;

   bool use_localized;        //true to use localized text output
   vector<S_text_line> text_lines;
   int counter;
   const S_scrolltext_style *style;
   C_string text_save;
   const char *next_text;
   C_ease_interpolator fade_ease;
   float clip_y;

//----------------------------

   void ExtractInfo(int &style, int &pause, int &size){

      style = 0;
      pause = 0;
      size = 0;
      if(use_localized && game_configuration.language==5){
         wchar_t *wp = (wchar_t*)next_text;
         if(wp[0]!='~') return;
         int i = 0;
         swscanf(&wp[++size], (wchar_t*)"%\0i\0%\0n\0\0", &style, &i);
         size += i;
         if(wp[size]!='~') return;    //this breaks our rules
         if(wp[++size]=='~'){
            ++size;
            return;
         }
         i = 0;
         swscanf(&wp[size], (wchar_t*)"%\0i\0%\0n\0\0", &pause, &i);
         size += i;
         if(wp[size]!='~') return;    //this breaks our rules
         ++size;
         size *= 2;
      }else{
         if(next_text[size]!='~') return;
         int i = 0;
         sscanf(&next_text[++size], "%i%n", &style, &i);
         size += i;
         if(next_text[size]!='~') return;    //this breaks our rules
         if(next_text[++size]=='~'){
            ++size;
            return;
         }
         i = 0;
         sscanf(&next_text[size], "%i%n", &pause, &i);
         size += i;
         if(next_text[size]!='~') return;    //this breaks our rules
         ++size;
      }
   }

//----------------------------

   void SetCount(){

      int st, pos;
      ExtractInfo(st, counter, pos);
      if(!counter)
         counter = style[st].pause;
   }
public:
   C_scrolltext_imp(C_poly_text &t, S_scrolltext_style *cs, const C_string &ct, bool use_loc, float cy):
      texts(t),
      clip_y(cy),
      text_save(ct),
      use_localized(use_loc),
      style(cs)
   {
      next_text = &text_save[0];
      SetCount();
   }

   ~C_scrolltext_imp(){

      for(int i=text_lines.size(); i--; )
         text_lines[i].text->Release();
   }

   virtual void Tick(const S_tick_context&);

   virtual dword GetID() const{ return 'SCTX'; }
   virtual void Render(){}
   
   virtual bool IsEnd() const{
      return (!*next_text && !text_lines.size());
   }
};

//----------------------------

void C_scrolltext_imp::Tick(const S_tick_context &tc){

   if(*next_text)
   while((counter -= tc.time) <= 0){
      int st, pause, pos;
      ExtractInfo(st, pause, pos);
      next_text += pos;

      float size = style[st].size;
      float base_y = clip_y + size * .5f;

      byte align = style[st].align;
      char *buf;
      int len;
      bool wide = (use_localized && game_configuration.language==5);

      if(wide){
         wchar_t *wp = (wchar_t*)next_text;
         wchar_t *tmp = wcschr(wp, '\n');
         len = tmp ? (tmp-wp) : wcslen(wp);
         buf = new char[len*2+2];
         wcsncpy((wchar_t*)buf, wp, len);
         *((wchar_t*)&buf[len*2]) = 0;
         next_text += len*2+2;
      }else{
         char *tmp = strchr(next_text, '\n');
         len = tmp ? (tmp-next_text) : strlen(next_text);
         buf = new char[len+1];
         strncpy(buf, next_text, len);
         buf[len]=0;
         next_text += len+1;
      }
      float x = (align ? .5f : .3f);
      float y = .75f - base_y;
                              //align decimal point to fit with previous line on sub-pixel
      if(text_lines.size()){
         float last_y = text_lines.back().pos[1] * igraph->Scrn_sy();
         y *= igraph->Scrn_sy();
         y = y - fmod(y, 1.0f) + fmod(last_y, 1.0f);
         y /= igraph->Scrn_sy();
      }
      LPC_text txt;
      dword color = style[st].color;

      txt = texts.CreateText(buf, color, x, y, size, align,
         wide, NULL, NULL, NULL, NULL, 0, style[st].italic);
      delete[] buf;

      text_lines.push_back(S_text_line());
      text_lines.back().text = txt;
      text_lines.back().color = color&0xffffff;
      text_lines.back().alpha = (float)(color>>24)/255.0f;
      text_lines.back().phase = 0;
      text_lines.back().base_y = base_y;
      text_lines.back().pos[0] = x;
      text_lines.back().pos[1] = y;

      SetCount();
   }
                           //scroll all texts
   float y_scroll = (float)tc.time * SCROLL_SPEED * .001f;
   for(int i=0; i<text_lines.size(); i++){
      S_text_line &line = text_lines[i];
      float &x = line.pos[0];
      float &y = line.pos[1];
      y -= y_scroll;
      if(y <= line.base_y){
         line.text->Release();
         text_lines.erase(text_lines.begin());
         --i;
      }else{
         line.text->SetPos(x, y);

         float f;
         switch(line.phase){
         case 0:           //fade-in
            if(y < (.75f-line.base_y - TEXT_FADE_Y)) ++line.phase;
            else
               f = (.75f-line.base_y-y) / TEXT_FADE_Y;
            break;
         case 1:           //full
            if(y > (line.base_y + TEXT_FADE_Y)) break;
            ++line.phase;
                           //flow...
         case 2:           //fade-out
            f = (y-line.base_y) / TEXT_FADE_Y;
            break;
         }
         if(line.phase != 1){
            float alpha = line.alpha*fade_ease[f];
            line.text->SetColor(line.color | (dword(alpha*255.0f)<<24));
         }
      }
   }
}

//----------------------------

C_scrolltext *CreateScrolltext(C_poly_text &texts, struct S_scrolltext_style *cs, const C_string *ct, bool wide, float clip_y){

   C_string tmp;
   if(!cs) cs = credits_styles;
   if(!ct){
      CreateCreditsText(tmp);
      ct = &tmp;
   }
   return new C_scrolltext_imp(texts, cs, *ct, wide, clip_y);
}

//----------------------------
