#include "all.h"
#include "winstuff.h"
#include "Tables.h"

//#define USE_EAX

#ifdef USE_EAX
#include "eax.h"
#endif

//----------------------------

LPC_table tab_sys_cfg;
extern C_table_template templ_sys_cfg;

//----------------------------

void ApplySysSetting(LPC_table tab){

   for(int i=0; i<8; i++){
      driver->SetState((I3D_RENDERSTATE)(RS_DEBUG_INT0+i), tab->ItemI(TAB_I_DEBUG, i));
      driver->SetState((I3D_RENDERSTATE)(RS_DEBUG_FLOAT0+i), tab->ItemI(TAB_F_DEBUG, i));
   }
}

//----------------------------
#ifdef USE_EAX

#define MAX_SOUND_ENVIRONMENTS 32

static const C_table_template templ_sound_envs = {
   // {67B1CF20-11A1-11d4-BCB2-80CA6EC10000}
   { 0x67b1cf20, 0x11a1, 0x11d4, { 0xbc, 0xb2, 0x80, 0xca, 0x6e, 0xc1, 0x0, 0x0 } },
   "Sector sound properties",
   NULL,
   {
      {TE_ARRAY, I3D_SENV_S_NAME, "Environments", MAX_SOUND_ENVIRONMENTS},
         {TE_BRANCH, I3D_SENV_S_NAME, NULL, 6, (dword)"%[0]"},
            {TE_STRING, I3D_SENV_S_NAME, "Name", 24},
            {TE_BRANCH, 0, "Room level", 2, (dword)"LF: %[0] HF: %[1]"},
               {TE_FLOAT, I3D_SENV_F_ROOM_LEVEL_LF, "Low frequency", 0, 1, 0},
               {TE_FLOAT, I3D_SENV_F_ROOM_LEVEL_HF, "High frequency", 0, 1, 0},

            {TE_BRANCH, 0, "Decay", 2, (dword)"Time: %[0] ms HF: %[1]"},
               {TE_INT, I3D_SENV_I_REVERB_DECAY_TIME, "Time (msec)", EAXLISTENER_MINDECAYTIME*1000, EAXLISTENER_MAXDECAYTIME*1000, EAXLISTENER_DEFAULTDECAYTIME*1000},
               {TE_FLOAT, I3D_SENV_F_DECAY_HF_RATIO, "HF ratio", 0, 1, 0},

            {TE_BRANCH, 0, "Reflections", 2, (dword)"Ratio: %[0] Delay: %[1] ms"},
               {TE_FLOAT, I3D_SENV_F_REFLECTIONS_RATIO, "Ratio", 0, 1, 0},
               {TE_INT, I3D_SENV_I_REFLECTIONS_DELAY, "Delay (msec)", EAXLISTENER_MINREFLECTIONSDELAY*1000, EAXLISTENER_MAXREFLECTIONSDELAY*1000, EAXLISTENER_DEFAULTREFLECTIONSDELAY*1000},

            {TE_BRANCH, 0, "Reverbation", 2, (dword)"Level: %[0] Delay: %[1] ms"},
               {TE_FLOAT, I3D_SENV_F_REVERB_RATIO, "Level", 0, 1, 0},
               {TE_INT, I3D_SENV_I_REVERB_DELAY, "Delay (msec)", EAXLISTENER_MINREVERBDELAY*1000, EAXLISTENER_MAXREVERBDELAY*1000, EAXLISTENER_DEFAULTREVERBDELAY*1000},
      
            {TE_FLOAT, I3D_SENV_F_ENV_DIFFUSION, "Env. diffusion", 0, 1},
            /*
            {TE_INT, I3D_SENV_I_SWITCH_FADE, "Fade", 0, 20000, 1000},
            {TE_BRANCH, 0, "Emulation (non-EAX)", 1, (dword)"Volume %[0]"},
               {TE_FLOAT, I3D_SENV_F_EMUL_VOLUME, "Volume", 0, 1, 1},
               */
      {TE_NULL}
   }
};
#endif//USE_EAX

//----------------------------

const C_table_template tab_particle_params = {
   // {C504D1A0-3D49-11d2-9B9D-004F4905FF1A}
   { 0xc504d1a0, 0x3d49, 0x11d2, { 0x9b, 0x9d, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },

   "Particle setup",
   NULL,
   {
      {TE_BOOL, TAB_B_PRT_ON, "Enabled"},
      {TE_BRANCH, 0, "Out time", 2, (int)"Time=%[0], thresh=%[1]"},
         {TE_INT, PRT_OUT_TIME, "Time"},
         {TE_INT, PRT_OUTT_THRESH, "Thresh"},
      {TE_BRANCH, 0, "Life time", 2, (int)"Time=%[0], thresh=%[1]"},
         {TE_INT, PRT_LIFE_LEN, "Time"},
         {TE_INT, PRT_LIFEL_THRESH, "Thresh"},
      {TE_BRANCH, 0, "Scale", 2, (int)"min=%[0], max=%[1]"},
         {TE_FLOAT, PRT_SCALE_GROW_MIN, "Min range", -2, 2},
         {TE_FLOAT, PRT_SCALE_GROW_MAX, "Max range", -2, 2},
      {TE_FLOAT, PRT_INIT_POS_THRESH, "Init position thresh"},
      {TE_BRANCH, 0, "Affect", 3, (int)"Time=%[0], thresh=%[1], dir=%[2]"},
         {TE_INT, PRT_AFFECT_TIME, "Time"},
         {TE_INT, PRT_AFFECTT_THRESH, "Thresh"},
         {TE_FLOAT, PRT_AFFECT_DIR_THRESH, "Dir thresh"},
      {TE_BRANCH, 0, "Opacity", 2, (int)"In=%[0], Out=%[1]"},
         {TE_FLOAT, PRT_OPACITY_IN, "In", 0, 1},
         {TE_FLOAT, PRT_OPACITY_OUT, "Out", 0, 1},
      {TE_BRANCH, 0, "Base direction", 5, (int)"[%[1], %[2], %[3]], thresh=%[4]"},
         {TE_BOOL, PRT_BASE_DIR_WORLD, "World coordinates"},
         {TE_FLOAT, PRT_BASE_DIR_X, "X"},
         {TE_FLOAT, PRT_BASE_DIR_Y, "Y"},
         {TE_FLOAT, PRT_BASE_DIR_Z, "Z"},
         {TE_FLOAT, PRT_INIT_DIR_THRESH, "thresh"},
      {TE_BRANCH, 0, "Dest direction", 4, (int)"[%[1], %[2], %[3]]"},
         {TE_BOOL, PRT_DEST_DIR_WORLD, "World coordinates"},
         {TE_FLOAT, PRT_DEST_DIR_X, "X"},
         {TE_FLOAT, PRT_DEST_DIR_Y, "Y"},
         {TE_FLOAT, PRT_DEST_DIR_Z, "Z"},
      {TE_ENUM, PRT_DRAW_MODE, "Draw mode", (int)"OPAQUE\0BLEND\0ADD\0"},
      {TE_INT, PRT_SETCOUNT, "Count"},

      {TE_NULL}
   }
};

//----------------------------

const C_table_template templ_game_cfg = {
   // {2F2B5AE0-401F-11d2-9B9D-004F4905FF1A}
   { 0x2f2b5ae1, 0x401f, 0x11d2, { 0x9b, 0x9d, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Game configuration",
   NULL,
   {
      //{TE_STRING, TAB_S32_GAME_MISSIONNAME,  "Mission name", 32},

      {TE_BRANCH, 0,  "AI", 11},
         {TE_ENUM, TAB_E_AI_ENABLED,  "Enabled", (dword)ENUM_AI_NOYES},
         {TE_ENUM, TAB_E_AI_LISTEN,  "Listen", (dword)ENUM_AI_NOYES},
         {TE_ENUM, TAB_E_AI_HUNT,  "Hunt enemy...", (dword)ENUM_AI_NOYES},
         {TE_FLOAT, TAB_F_AI_HUNT_MAX_DIST,  "... up to max distance"},
         {TE_ENUM, TAB_E_AI_SEARCH_DEAD,  "Search dead", (dword)ENUM_AI_NOYES},
         {TE_ENUM, TAB_E_AI_SEARCH_HEARD,  "Search heard", (dword)ENUM_AI_NOYES},
         {TE_FLOAT, TAB_F_AI_WATCH_MAX_DIST,  "Watch distance"},
         {TE_INT, TAB_I_AI_SHOOT_QUALITY,  "Shoot quality %"},
         {TE_FLOAT, TAB_F_AI_MIN_SHOOT_DIST,  "Min. shoot distance"},
         {TE_INT, TAB_I_AI_REACTION,  "Reaction %"},
         {TE_INT, TAB_I_AI_ENDURANCE,  "Endurance %"},

      {TE_BRANCH, 0,  "Debug", 8},
         {TE_ARRAY, TAB_I_DEBUG, "Debug int", 8},
            {TE_INT, TAB_I_DEBUG, NULL, 0, 255},
         {TE_ARRAY, TAB_F_DEBUG, "Debug float", 8},
            {TE_FLOAT, TAB_F_DEBUG, NULL, 0, 1},

         {TE_BOOL, TAB_B_GAME_DRAW_PATH, "Draw AI path"},
         {TE_BOOL, TAB_B_GAME_DEBUG_SHOW_LISTEN, "Draw listening"},
         {TE_BOOL, TAB_B_GAME_DEBUG_SHOW_WATCH, "Draw watching"},
         {TE_BOOL, TAB_B_GAME_DEBUG_EXPL_LINES, "Draw explosion wave"},
         {TE_BOOL, TAB_B_GAME_DEBUG_SHOW_SHOTS, "Draw shots"},
         {TE_BOOL, TAB_B_GAME_DEBUG_DETECTOR, "Detector messages"},
         //{TE_BOOL, TAB_B_GAME_DEBUG_NET, "Network"},

         /*
      {TE_BRANCH, 0,  "Pre-load models", 2},
         {TE_BOOL, TAB_B_GAME_PRELOAD_MODELS,   "Use pre-loading"},
         {TE_ARRAY, TAB_S12_GAME_PRELOAD_MODEL,  "List", 16},
            {TE_STRING, TAB_S12_GAME_PRELOAD_MODEL,  0, 12},
            */

      {TE_ARRAY, TAB_I_GAME_CAMP_ITEMS_ID,  "Campaign items", NUM_CAMPAIGNS+1},
         {TE_ARRAY, TAB_I_GAME_CAMP_ITEMS_ID,  "Items", 32},
            {TE_INT, TAB_I_GAME_CAMP_ITEMS_ID,  "Item ID"},
      {TE_ARRAY, TAB_I64_GAME_MIS_ITEM_AMOUNT,  "Mission auto-setup items", 40},
         {TE_ARRAY, TAB_I64_GAME_MIS_ITEM_AMOUNT,  "Items", 64},
            {TE_INT, TAB_I64_GAME_MIS_ITEM_AMOUNT,  "Amount"},

      {TE_BRANCH, 0,  "Tuning *** top secret area ***", 19},
         {TE_INT, TAB_I_GAME_SHOOT_SPEED, "Gun-shoot speed (m/s)", 0, 0, 170},
         {TE_FLOAT, TAB_F_GAME_GUNHOLE_DIST, "Gun-hole distance"},
         {TE_INT, TAB_F_GAME_BARR_CLIMB_COUNT, "Barrier climb count"},
         {TE_BRANCH, 0,  "Sniper mode (zooming)", 3},
            {TE_FLOAT, TAB_F_GAME_MIN_AIM_ZOOM, "Min aim zoom"},
            {TE_FLOAT, TAB_F_GAME_SNIPER_ERR_FACTOR, "Err factor"},
            {TE_FLOAT, TAB_F_GAME_SNIPER_MAGNET_FACTOR, "Magnet factor"},
         {TE_FLOAT, TAB_F_GAME_AIM_MODEL_DIST, "Aim model distance"},
         {TE_FLOAT, TAB_F_GAME_DIE_FALL_DIST, "Killing fall distance"},
         {TE_INT, TAB_I_GAME_HUM_CRASH_RESIST, "Human crash resist %", 0, 100, 30},
         //{TE_FLOAT, TAB_F_GAME_CAMERA_SHAKE_RANGE, "Camera shake range", 0, 100, 10},
         {TE_FLOAT, TAB_F_GAME_WALL_Y_THRESH, "Collision floor/wall thresh", 0, 1, 0},
         {TE_FLOAT, TAB_F_GAME_GRENADE_JUMP, "Grenade jumping ratio", 0, 1, 0},
         {TE_INT, TAB_I_GAME_EXPL_DELAY, "Explosion delay"},
         //{TE_INT, TAB_I_GAME_DETECTOR_COUNT, "Detector counter"},
         {TE_BRANCH, 0,  "Barrier climb", 2},
            {TE_FLOAT, TAB_F_GAME_BARRIER_LOW, "Low"},
            {TE_FLOAT, TAB_F_GAME_BARRIER_HI, "High"},
         {TE_BRANCH, 0,  "Shooting (angles)", 4},
            {TE_FLOAT, TAB_F_GAME_SHOOT_ERR_MIN, "Shoot err min"},
            {TE_FLOAT, TAB_F_GAME_SHOOT_ERR_MAX, "Shoot err max"},
            {TE_FLOAT, TAB_F_GAME_SHOOT_MAG_MIN, "Shoot magnet min"},
            {TE_FLOAT, TAB_F_GAME_SHOOT_MAG_MAX, "Shoot magnet max"},
         {TE_BRANCH, 0,  "Reaction (times)", 2},
            {TE_INT, TAB_F_GAME_REACTION_MIN, "Reaction time min"},
            {TE_INT, TAB_F_GAME_REACTION_MAX, "Reaction time max"},
         {TE_BRANCH, 0,  "Automobil", 1},
            {TE_INT, TAB_I_GAME_AUTO_START_DELAY, "Start delay (ms)"},
         {TE_BRANCH, 0,  "Mouse turning", 2},
            {TE_FLOAT, TAB_F_GAME_MOUSE_2_TURN, "Mouse to angle"},
            {TE_FLOAT, TAB_F_GAME_MOUSE_TURN_CACHE, "Mouse fade speed"},
         {TE_BRANCH, 0,  "Sound ranges", 14},
            {TE_FLOAT, TAB_F_GAME_SNDR_GUN_MIN, "Gun min"},
            {TE_FLOAT, TAB_F_GAME_SNDR_GUN_MAX, "Gun max"},
            {TE_FLOAT, TAB_F_GAME_SNDR_GUNWHEEZ_MIN, "Gun wheez min"},
            {TE_FLOAT, TAB_F_GAME_SNDR_GUNWHEEZ_MAX, "Gun wheez max"},
            {TE_FLOAT, TAB_F_GAME_SNDR_EXPL_MIN, "Explosion min"},
            {TE_FLOAT, TAB_F_GAME_SNDR_EXPL_MAX, "Explosion max"},
            {TE_FLOAT, TAB_F_GAME_SNDR_HUM_MIN, "Voices min"},
            {TE_FLOAT, TAB_F_GAME_SNDR_HUM_MAX, "Voices max"},
            {TE_FLOAT, TAB_F_GAME_SNDR_ROCKET_MIN, "Rocket min"},
            {TE_FLOAT, TAB_F_GAME_SNDR_ROCKET_MAX, "Rocket max"},
            {TE_FLOAT, TAB_F_GAME_SNDR_CANNON_MIN, "Cannon min"},
            {TE_FLOAT, TAB_F_GAME_SNDR_CANNON_MAX, "Cannon max"},
            {TE_FLOAT, TAB_F_GAME_SNDR_AUTO_MIN, "Automobil max"},
            {TE_FLOAT, TAB_F_GAME_SNDR_AUTO_MAX, "Automobil max"},
         {TE_BRANCH, 0,  "Watching", 6},
            {TE_ARRAY, TAB_F_GAME_WATCH_STAY_FACTOR, "Stay factor", 4},
               {TE_FLOAT, TAB_F_GAME_WATCH_STAY_FACTOR},
            {TE_FLOAT, TAB_F_GAME_WATCH_LIGHT_MULT, "Lightness multiplier"},
            {TE_INT, TAB_I_GAME_WATCH_ENUM_COUNT, "Enum count"},
            {TE_INT, TAB_I_GAME_WATCH_CHECK_COUNT, "Check count"},
            {TE_BRANCH, 0,  "Vigilance", 6},
               {TE_FLOAT, TAB_F_GAME_WATCH_CARE_WALK, "Walk"},
               {TE_FLOAT, TAB_F_GAME_WATCH_CARE_RUN, "Run"},
               {TE_FLOAT, TAB_F_GAME_WATCH_CARE_WATCH, "Watch"},
               {TE_FLOAT, TAB_F_GAME_WATCH_CARE_GUARD, "Guard"},
               {TE_FLOAT, TAB_F_GAME_WATCH_CARE_FOLLOW, "Follow"},
               {TE_FLOAT, TAB_F_GAME_WATCH_CARE_ATTACK, "Attack"},
            {TE_BRANCH, 0,  "Uniform confusing factor", 3},
               {TE_FLOAT, TAB_F_GAME_WATCH_UNIF_GERMAN, "German uniform"},
               {TE_FLOAT, TAB_F_GAME_WATCH_UNIF_CIVIL, "Civil suit"},
               {TE_FLOAT, TAB_F_GAME_WATCH_UNIF_CIVIL_W, "Civil with weapon"},
         {TE_BRANCH, 0,  "AI", 4},
            {TE_INT, TAB_I_GAME_AI_ATTACK_KILL_COUNT, "Attack - subj. killed count"},
            {TE_INT, TAB_I_GAME_AI_ATTACK_LOST_COUNT, "Attack - subj. lost count"},
            {TE_INT, TAB_I_GAME_AI_MOVE_HUNT_COUNT, "Move - hunt dist over"},
            {TE_FLOAT, TAB_F_GAME_AI_FOLLOW_CIRCLE_RAD, "Follow - formation radius"},
         {TE_BRANCH, 0,  "Aim line", 2},
            {TE_ARRAY, TAB_F_GAME_AIMLINE_COLOR,  "Color", 3},
               {TE_FLOAT, TAB_F_GAME_AIMLINE_COLOR},
            {TE_INT, TAB_I_GAME_AIMLINE_ALPHA, "Alpha"},
      {TE_NULL}
   }
};

//----------------------------

const C_table_template tab_scene_property = {
   // {BD765780-3C52-11d2-9B9D-004F4905FF1A}
   { 0xbd765780, 0x3c52, 0x11d2, { 0x9b, 0x9d, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Scene properties",
   NULL,
   {
      {TE_BRANCH, 0,  "AI", 11},
         {TE_ENUM, TAB_E_AI_ENABLED,  "Enabled", (dword)ENUM_AI_NOYES},
         {TE_ENUM, TAB_E_AI_LISTEN,  "Listen", (dword)ENUM_AI_NOYES},
         {TE_ENUM, TAB_E_AI_HUNT,  "Hunt enemy...", (dword)ENUM_AI_NOYES},
         {TE_FLOAT, TAB_F_AI_HUNT_MAX_DIST,  "... up to max distance"},
         {TE_ENUM, TAB_E_AI_SEARCH_DEAD,  "Search dead", (dword)ENUM_AI_NOYES},
         {TE_ENUM, TAB_E_AI_SEARCH_HEARD,  "Search heard", (dword)ENUM_AI_NOYES},
         {TE_FLOAT, TAB_F_AI_WATCH_MAX_DIST,  "Watch distance"},
         {TE_INT, TAB_I_AI_SHOOT_QUALITY,  "Shoot quality %"},
         {TE_FLOAT, TAB_F_AI_MIN_SHOOT_DIST,  "Min. shoot distance"},
         {TE_INT, TAB_I_AI_REACTION,  "Reaction %"},
         {TE_INT, TAB_I_AI_ENDURANCE,  "Endurance %"},

         /*
      {TE_BRANCH, 0, "Environment", 7},
         {TE_COLOR_VECTOR, TAB_CV_SCN_BGND,     "Background color"},
         {TE_COLOR_VECTOR, TAB_CV_SCN_AMBIENT,  "Ambient color"},
         {TE_FLOAT, TAB_F_SCN_CGLOBAL,  "Global lighting", 0, 1},
         {TE_BOOL, TAB_B_SCN_FOG_ON, "Fog"},
         {TE_COLOR_VECTOR, TAB_CV_SCN_FOG,      "Fog color"},
         {TE_FLOAT, TAB_F_SCN_FOG_NEAR, "Fog near range", 0, 1},
         {TE_BRANCH, 0, "Backdrop", 3, (int)"Name=\"%[0]\""},
            {TE_STRING, TAB_S_SCN_BACKDROPNAME, "Backdrop", 16},
            {TE_COLOR_VECTOR, TAB_CV_SCN_BACK_AMB,  "Ambient color"},
            {TE_BOOL, TAB_B_SCN_BGND_FORCE,  "Always draw"},
      {TE_BRANCH, 0, "Camera", 2, (int)"FOV=%[0], range=%[1]"},
         {TE_INT, TAB_I_SCN_CAM_FOV, "FOV", 10, 135},
         {TE_FLOAT, TAB_F_SCN_CAM_RANGE, "Range"},
         */
      {TE_ARRAY, TAB_E_SCN_TIME_UNIT, "Time configuration", 64, 64},
         {TE_BRANCH, TAB_E_SCN_TIME_UNIT, "", 5, (dword)"%[0] %[1] - \"%[2]\" %[3] S=%[4]"},
            {TE_ENUM, TAB_E_SCN_TIME_UNIT, "Unit", (dword)"???\0Time\0Notify\0"},
            {TE_INT, TAB_I_SCN_TIME_VAL, "Value"},
            {TE_STRING, TAB_S24_SCN_TIME_OBJ, "Frame name", 24},
            {TE_BOOL, TAB_B_SCN_TIME_STATUS, "Status"},
            {TE_INT, TAB_I_SCN_TIME_SIGNUM, "Signal broadcast #"},
      {TE_BRANCH, TAB_B_SCN_DISABLE_SHDWS, "Optimization", 3},
         {TE_BOOL, TAB_B_SCN_DISABLE_SHDWS, "Disable enemy shadow"},
         {TE_BOOL, TAB_B_SCN_OPTIM_DEAD_BODY, "Optimize dead bodies"},
         {TE_BOOL, TAB_B_SCN_OPTIM_GUNFLR, "Optimize gun flares"},
      {TE_NULL}
   }
};

//----------------------------

static char pose_enum[]="No\0Stay\0Bench\0B + S\0Lay\0L + S\0L + B\0L + B + S\0";
                              //template
const C_table_template templ_inv_table = {
   // {61B396E0-48AC-11d2-9B9E-004F4905FF1A}
   { 0x61b396e0, 0x48ac, 0x11d2, { 0x9b, 0x9e, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Inventory setup",
   NULL,
   {
      {TE_ARRAY, TAB_S32_INV_NAME,  "Items", 64},
         {TE_BRANCH, 0,  "Item", 16, (dword)"%[0] (%[6])"},
            {TE_STRING, TAB_S32_INV_NAME,  "Name", 32},
            {TE_INT, TAB_I_INV_PRIORITY, "Choose priority", 0, 100, 100},
            {TE_STRING, TAB_S16_INV_MODEL_S, "Model (scene)", 16},
            {TE_STRING, TAB_S16_INV_MODEL_I, "Model (inventory)", 16},
            {TE_STRING, TAB_S16_INV_MODEL_RES, "Model (resource)", 16},
            {TE_STRING, TAB_S16_INV_MODEL_AIM, "Model (aim)", 16},
            {TE_INT, TAB_I_INV_ID, "ID"},
            {TE_INT, TAB_I_INV_AUTO_GROUP, "Auto-setup group"},
            {TE_FLOAT, TAB_F_INV_WEIGTH, "Weigth", 0, 25, 2},
            {TE_STRING, TAB_S16_INV_USESOUND, "Use sound", 16},
            {TE_BRANCH, 0,  "Poses", 3, (dword)"%[0], %[1], %[2]"},
               {TE_ENUM, TAB_E_INV_POSEMODE,  "Pose mode", (dword)"N\0" "G\0" "B\0"},
               {TE_ENUM, TAB_E_INV_USEPOSE,  "Use pose(s)", (dword)pose_enum},
               {TE_ENUM, TAB_E_INV_CARRYPOSE,  "Carry pose(s)", (dword)pose_enum},
            {TE_BRANCH, 0,  "Weapons", 12, (dword)"P=%[0], D=%[1]"},
               {TE_INT, TAB_I_INV_WPN_POWER, "Power", 0, 2000, 500},
               {TE_INT, TAB_I_INV_WPN_DELAY, "Shoot delay", 0, 1000},
               {TE_STRING, TAB_S16_INV_SMKNAME, "Smoke model", 16},
               {TE_STRING, TAB_S16_INV_SMKTABNAME, "Smoke table", 16},
               {TE_BOOL, TAB_B_INV_ZOOM, "Zooming"},
               {TE_ENUM, TAB_E_INV_FIREMODE, "Fire mode", (dword)"Auto\0Single\0Batch\0"},
               {TE_INT, TAB_I_INV_STACKSIZE, "Batch size"},
               {TE_INT, TAB_I_INV_NUM_BULLETS, "Bullets out", 0, 5, 1},
               {TE_FLOAT, TAB_F_INV_AMMO_WEIGTH, "Ammo weigth", 0, 1, 0},
               {TE_INT, TAB_I_INV_DEFAULT_AMMO, "Base ammo (menu)", 0, 200, 40},
               {TE_INT, TAB_I_INV_MENU_AMMO_TRANSFER, "Add/remove ammo (menu)", 0, 0, 10},
               {TE_FLOAT, TAB_F_INV_WPN_SHOT_Z, "Gun shot Z origin"},
            {TE_BRANCH, 0,  "Grenade", 2, (dword)"Type=%[0], D=%[1]"},
               {TE_ENUM, TAB_E_INV_GRN_TYPE, "Type", (dword)"Time\0Explosion\0"},
               {TE_FLOAT, TAB_F_INV_GRN_RANGE, "Range", 10},
            {TE_BRANCH, 0,  "Mine", 1, (dword)"Expl=%[0]"},
               {TE_INT, TAB_I_INV_MINE_EXPL, "Explosion"},
            {TE_BOOL, TAB_B_INV_MULTIPLE, "Multiple items"},
            {TE_BOOL, TAB_B_INV_TO_NEXT_MISS, "Pass to next mission", 0, 0, true},
      {TE_NULL}
   }
};

//----------------------------

const C_table_template templ_anim_table = {
   // {F6727780-4CC9-11d2-9B9E-004F4905FF1A}
   { 0xf6727780, 0x4cc9, 0x11d2, { 0x9b, 0x9e, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Animation table",
   NULL,
   {
      {TE_ARRAY, TAB_S16_ANI_FILENAME,  "Animations", 0x120},
         {TE_BRANCH, 0,  "%x", 6, (dword)"%[0] [%[1]]"},
            {TE_STRING, TAB_S16_ANI_FILENAME, "Filename", 16},
            {TE_BRANCH, 0,  "Speed", 3, (dword)"X=%[0], Y=%[1], Z=%[2]"},
               {TE_FLOAT, TAB_F_ANI_SPEEDX,  "X"},
               {TE_FLOAT, TAB_F_ANI_SPEEDY,  "Y"},
               {TE_FLOAT, TAB_F_ANI_SPEEDZ,  "Z"},
            {TE_ENUM, TAB_E_ANI_TYPE,  "Type", (dword)"Normal\0Simple\0Full\0"},
            //{TE_FLOAT, TAB_F_ANI_SEEN_FACTOR,  "Seen factor", 1, 5, 1},
            {TE_INT, TAB_I_ANI_BLENDTIME,  "Blend time %", 0, 100, 50},
            {TE_BOOL, TAB_B_ANI_TURNENABLE, "Turning", 0, 0, true},
            {TE_BOOL, TAB_B_ANI_MOUSE_TURN_X, "Mouse X turn"},
      {TE_NULL}
   }
};

//----------------------------

const C_table_template templ_resset = {
   // {C6A5C820-B481-11d2-87BD-004F4905FF1A}
   { 0xc6a5c820, 0xb481, 0x11d2, { 0x87, 0xbd, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Object resistance sets",
   NULL,
   {
      {TE_ARRAY, TAB_S12_RSET_SET_NAME, "Sets", 20},
         {TE_BRANCH, 0, "", 4, (dword)"%[0] (%[1])"},
            {TE_STRING, TAB_S12_RSET_SET_NAME, "Name", 12},
            {TE_INT, TAB_I_RSET_VALUE, "Resistance", 0, 5000, 1000},
            {TE_INT, TAB_I_RSET_THRESH, "Thresh", 0, 5000, 500},
            //{TE_STRING, TAB_S12_RSET_EXPL_MODEL, "Explosion model", 12},
            {TE_INT, TAB_I_RSET_EXPLOSION, "Explosion ID (-1 = no)", 0, 5, -1},
      {TE_NULL}
   }
};

//----------------------------

const C_table_template templ_campaigns = {
   // {7F5BA260-6D1C-11d3-B3AD-0000B49B8E8B}
   { 0x7f5ba260, 0x6d1c, 0x11d3, { 0xb3, 0xad, 0x0, 0x0, 0xb4, 0x9b, 0x8e, 0x8b } },
   "Campaign & mission setup",
   NULL,
   {
      //{TE_INT, TAB_I_CM_NUM_CAMPAIGNS, "Number of campaigns"},
      {TE_INT, TAB_I_CM_NUM_SOLDIERS, "Number of soldiers", 0, 0, 40},
      {TE_ARRAY, TAB_S12_CM_IMAGE_NAME, "Campaigns", NUM_CAMPAIGNS+1},
         {TE_BRANCH, 0, "Campaign", 3},
            {TE_INT, TAB_I_CM_NUM_MISSIONS, "Number of missions"},
            {TE_STRING, TAB_S12_CM_IMAGE_NAME, "Image name (no ext.)", 12},
            {TE_STRING, TAB_S12_CM_MISS_IMG_BASE, "Mission image base", 12},
      {TE_ARRAY, TAB_S12_CM_MISS_DIR, "Mission names", NUM_MISSIONS+MAX_NUM_ADDON_MISS},
         {TE_STRING, TAB_S12_CM_MISS_DIR, "Directory", 12},

      {TE_NULL}
   }
};

//----------------------------

const C_table_template templ_mat_table = {
   // {7AC35CC0-5B84-11d2-9B9E-004F4905FF1A}
   { 0x7ac35cc0, 0x5b84, 0x11d2, { 0x9b, 0x9e, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Materials setup",
   NULL,
   {
      {TE_ARRAY, TAB_S16_MAT_NAME,  "Materials", MAX_MATERIALS},
         {TE_BRANCH, TAB_S16_MAT_NAME, NULL, 8, (dword)"%[0]"},
            {TE_STRING, TAB_S16_MAT_NAME, "Friendly name", 16},
            {TE_STRING, TAB_S12_SND_GUNSHOOT, "Gunshot sound", 12},
            {TE_BOOL, TAB_B_MAT_STAIRS, "Stairs"},
            {TE_BOOL, TAB_B_MAT_BORDER, "Border"},
            {TE_BOOL, TAB_B_MAT_AI_SEE_THROUGH, "AI see through"},
            {TE_BOOL, TAB_B_MAT_CAM_SEE_THROUGH, "Camera see through"},
            {TE_BOOL, TAB_B_MAT_WATER, "Water surface"},
            {TE_ENUM, TAB_E_MAT_KILL, "Kill mode", (dword)"no\0die in spasm\0water flow\0mine field\0throw to pieces\0"},

      {TE_ARRAY, TAB_S12_SND_SNDNAME,  "Material sounds", MAX_MATERIALS},
         {TE_ARRAY, TAB_S12_SND_SNDNAME,  "Action sounds", MAX_ACTIONS},
            {TE_BRANCH, 0, "Sound", 2, (dword)"%[0] (%[1]%)"},
               {TE_STRING, TAB_S12_SND_SNDNAME, "Name", 12},
               {TE_INT, TAB_I_SND_VOLUME, "Volume %", 0, 100, 100},

      {TE_ARRAY, TAB_S12_EXPL_SOUND,  "Explosions", 32},
         {TE_BRANCH, TAB_S12_EXPL_SOUND, NULL, 3},
            {TE_STRING, TAB_S12_EXPL_SOUND, "Name", 12},
            {TE_INT, TAB_I_EXPL_RANGE, "Range", 0, 50, 15},
            {TE_INT, TAB_I_EXPL_POWER, "Power", 0, 4000, 3000},

      {TE_NULL}
   }
};

//----------------------------
                              //tables open/close
static const struct{
   LPC_table *tab;
   const C_table_template *templ;
   const char *tab_name;
} tab_info[] = {
   {&mat_table, &templ_mat_table, "mat_sound"},
#ifdef USE_EAX
   {&tab_sound_envs, &templ_sound_envs, "sound_envs"},
#endif
   {&tab_game_cfg, &templ_game_cfg, "game_cfg"},
   {&tab_inv[0], &templ_inv_table, "inventry_0"},
   {&tab_inv[1], &templ_inv_table, "inventry_1"},
   {&tab_human_anim, &templ_anim_table, "ani_humn"},
   {&tab_resistance, &templ_resset, "resist"},
   {&tab_campaigns, &templ_campaigns, "camp_stp"},
   {NULL}
};

#ifdef EDITOR
//----------------------------
// Update material table into editor.
void ApplyMaterialTable(){

   LPC_table tab = mat_table;
   if(editor){
                              //build material pairs
      typedef pair<int, const char*> t_mp;
      vector<t_mp> mat_pairs;
      for(int i=0; i<tab->ArrayLen(TAB_S16_MAT_NAME); i++){
         const char *mat_name = tab->ItemS(TAB_S16_MAT_NAME, i, 16);
         if(*mat_name)
            mat_pairs.push_back(t_mp(i, mat_name));
      }
      mat_pairs.push_back(t_mp(-1, NULL));
      editor->BroadcastAction(EDIT_ACTION_FEED_MATERIALS, mat_pairs.begin());
   }
}

//----------------------------

void ApplySoundEnvTable(){

#ifdef USE_EAX
   LPC_table tab = tab_sound_envs;
   for(int i=tab->ArrayLen(I3D_SENV_S_NAME); i--; ){
      const char *name = tab->ItemS(I3D_SENV_S_NAME, i, 24);
      if(*name){
         driver->SetSoundEnvData(i, I3D_SENV_S_NAME, (dword)name);
         static const I3D_SOUND_ENV_DATA_INDEX indxs[] = {
            I3D_SENV_F_ROOM_LEVEL_LF, I3D_SENV_F_ROOM_LEVEL_HF, I3D_SENV_I_REVERB_DECAY_TIME,
            I3D_SENV_F_DECAY_HF_RATIO, I3D_SENV_F_REFLECTIONS_RATIO, I3D_SENV_I_REFLECTIONS_DELAY,
            I3D_SENV_F_REVERB_RATIO, I3D_SENV_I_REVERB_DELAY, I3D_SENV_F_ENV_DIFFUSION,
            //I3D_SENV_F_EMUL_VOLUME,
         };
         for(int j=sizeof(indxs)/sizeof(I3D_SOUND_ENV_DATA_INDEX); j--; ){
            driver->SetSoundEnvData(i, indxs[j], tab->ItemI(indxs[j], i));
         }
      }
   }
   if(editor){
                              //build table pairs
      typedef pair<int, const char*> t_mp;
      vector<t_mp> pairs;
      for(int i=0; i<tab->ArrayLen(TAB_S16_MAT_NAME); i++){
         const char *name = tab->ItemS(I3D_SENV_S_NAME, i, 24);
         if(*name)
            pairs.push_back(t_mp(i, name));
      }
      pairs.push_back(t_mp(-1, NULL));
      editor->BroadcastAction(EDIT_ACTION_FEED_ENVIRONMENTS, pairs.begin());
   }
#endif//USE_EAX
}

//----------------------------

                              //editor plugin: SysConfig
class C_edit_SysConfig: public C_editor_item{
   virtual const char *GetName() const{ return "SysConfig"; }

   C_smart_ptr<C_game_mission> mission;

   void *hwnd;
   int pos_size[4];

   enum E_table_type {
      TAB_NO,
      TAB_MATERIALS = 1000,
      TAB_GAME_CFG,
      TAB_INVENTORY,
      TAB_HUMAN_ANIM,
      TAB_RESISTANCE,
      TAB_CAMPAIGNS,
      TAB_SCENE,
      TAB_ACTOR,
      TAB_ENVS,
   } curr_mode;
   C_smart_ptr<C_actor> edited_actor;

   static void cbTab(LPC_table tab, dword msg, dword cb_user, dword prm2, dword prm3){

      C_edit_SysConfig *ec = (C_edit_SysConfig*)cb_user;
      switch(msg) {

      case TCM_CLOSE:
         {
            igraph->DeleteDlgHWND(ec->hwnd);
            ec->hwnd = NULL;
            int *pos_size = (int*)prm2;
            memcpy(ec->pos_size, pos_size, sizeof(ec->pos_size));
            ec->curr_mode = TAB_NO;
            ec->edited_actor = NULL;
         }
         break;

      case TCM_MODIFY:
         switch(ec->curr_mode){
         case TAB_MATERIALS:
            ApplyMaterialTable();
            break;
         case TAB_ENVS:
            ApplySoundEnvTable();
            break;
         case TAB_SCENE:
            ec->ed->SetModified();
            break;
         case TAB_ACTOR:
            {
               ec->ed->SetModified();
               ec->edited_actor->TableUpdate(prm2, false);
            }
            break;
         case TAB_GAME_CFG:
            ApplySysSetting(tab);
            break;
         }
         break;

      case TCM_FOCUS:
         switch(ec->curr_mode){
         case TAB_ACTOR:
            ec->edited_actor->cbProc(CB_TABLE_MESSAGE, msg, prm2);
            break;
         }
         break;
      }
   }

public:

   C_edit_SysConfig(C_game_mission *m):
      mission(m),
      hwnd(NULL),
      curr_mode(TAB_NO)
   {
      pos_size[0] = 0;
      pos_size[1] = 0;
      pos_size[2] = 80;
      pos_size[3] = 400;
   }

   virtual bool Init(){
#define ET "%0 &File\\%31 Edit table\\"
      ed->AddShortcut(this, TAB_MATERIALS, ET"&Materials\tCtrl+A", K_A, SKEY_CTRL);
      ed->AddShortcut(this, TAB_GAME_CFG, ET"&Game config\tCtrl+Shift+Alt+G", K_G, SKEY_CTRL|SKEY_SHIFT|SKEY_ALT);
      ed->AddShortcut(this, TAB_INVENTORY, ET"&Inventory\tCtrl+Shift+Alt+I", K_I, SKEY_CTRL|SKEY_SHIFT|SKEY_ALT);
      ed->AddShortcut(this, TAB_HUMAN_ANIM, ET"&Human anim\tCtrl+Shift+Alt+H", K_H, SKEY_CTRL|SKEY_SHIFT|SKEY_ALT);
      ed->AddShortcut(this, TAB_RESISTANCE, ET"&Resistance\tCtrl+Shift+Alt+E", K_E, SKEY_CTRL|SKEY_SHIFT|SKEY_ALT);
      ed->AddShortcut(this, TAB_CAMPAIGNS, ET"&Campaigns", 0, 0);
      ed->AddShortcut(this, TAB_SCENE, ET"&Current Scene-Actor\tCtrl+Enter", K_ENTER, SKEY_CTRL);
      ed->AddShortcut(this, TAB_ENVS, ET"Sound environments\tCtrl+Shift+E", K_E, SKEY_CTRL|SKEY_SHIFT);
      return true;
   }

   virtual void Close(){
      if(hwnd) OsDestroyWindow(hwnd);
   }

   virtual dword Action(int id, void *context) {

      switch(id) {
      case TAB_SCENE:
         if(!ed->CanModify())
            break;
                              //flow...
      case TAB_MATERIALS:
      case TAB_GAME_CFG:
      case TAB_INVENTORY:
      case TAB_HUMAN_ANIM:
      case TAB_RESISTANCE:
      case TAB_CAMPAIGNS:
      case TAB_ENVS:
         {
            if(!hwnd){
               LPC_table tab = NULL;
               const C_table_template *templ;
               switch(id){
               case TAB_MATERIALS: tab = mat_table; templ = &templ_mat_table; break;
               case TAB_GAME_CFG: tab = tab_game_cfg; templ = &templ_game_cfg; break;
               case TAB_INVENTORY: tab = tab_inventory; templ = &templ_inv_table; break;
               case TAB_HUMAN_ANIM: tab = tab_human_anim; templ = &templ_anim_table; break;
               case TAB_RESISTANCE: tab = tab_resistance; templ = &templ_resset; break;
               case TAB_CAMPAIGNS: tab = tab_campaigns; templ = &templ_campaigns; break;
               case TAB_ENVS:
#ifdef USE_EAX
                  tab = tab_sound_envs; templ = &templ_sound_envs;
#endif
                  break;
               case TAB_SCENE:
                  {
                     PI3D_frame frm = (PI3D_frame)editor->FindPlugin("Selection")->Action(E_SELECTION_GET_SINGLE);
                     if(!frm){
                        tab = mission->tab_property; templ = &tab_scene_property;
                     }else{
                        LPC_actor act = (LPC_actor)frm->GetData();
                        if(act){
                           int indx = act->GetTableIndex();
                           if(indx!=-1){
                              tab = act->GetTable(indx);
                              if(tab){
                                 templ = act->GetTemplate(indx);
                                 id = TAB_ACTOR;
                                 edited_actor = act;
                              }
                           }
                        }else{
                           ed->Message("Selected frame is not actor.");
                        }
                     }
                  }
                  break;
               default: assert(0);
               }
               if(!tab) break;

               curr_mode = (E_table_type)id;
               const dword flags = TABEDIT_CENTER | TABEDIT_MANUALUPDATE | TABEDIT_MODELESS | TABEDIT_IMPORT | TABEDIT_EXPORT | TABEDIT_INFO;
               hwnd = tab->Edit(templ,
                  igraph->GetHWND(), cbTab, (dword)this,
                  flags,
                  NULL, pos_size);
               if(hwnd) igraph->AddDlgHWND(hwnd);
            }
         }
         break;
      }
      return 0;
   }

   bool LoadState(class C_chunk *ck){ 
      ck->Read(pos_size, sizeof(pos_size));
      return true; 
   }

   bool SaveState(class C_chunk *ck) const{ 
      ck->Write(pos_size, sizeof(pos_size));
      return true; 
   }

};

//----------------------------
// Install sysconfig plugin.
void InitSysConfigPlugin(C_game_mission *m1, LPC_editor editor){
   C_edit_SysConfig *em = new C_edit_SysConfig(m1); editor->InstallPlugin(em); em->Release();
}

#endif//EDITOR

//----------------------------

void TablesCloseAll(){

   for(int i=0; tab_info[i].tab; i++){
      LPC_table tab = *tab_info[i].tab;
      if(!tab) continue;
#ifdef EDITOR
      if(tab_info[i].tab_name && tab->IsModified() || tab==tab_game_cfg){
         C_fstring tab_name("Tables\\%s.tab", tab_info[i].tab_name);
         if(tab->IsModified()){
            while(true){
               bool ok = tab->Save((dword)(const char*)tab_name, TABOPEN_FILENAME | TABOPEN_UPDATE);
               if(ok)
                  break;
                                 //failed to save table, issue warning
               if(OsMessageBox(tab_name, "Failed to save table:", MBOX_RETRYCANCEL) == MBOX_RETURN_CANCEL)
                  break;
            }
         }
      }
#endif
      tab->Release();
      *tab_info[i].tab = NULL;
   }
}

//----------------------------

bool TablesOpenAll(){

   for(int i=0; tab_info[i].tab; i++){
      LPC_table tab = CreateTable();
      *tab_info[i].tab = tab;
      tab->Open((dword)tab_info[i].templ, TABOPEN_TEMPLATE);
      if(tab_info[i].tab_name){
         C_fstring tab_name("Tables\\%s.tab", tab_info[i].tab_name);

         bool b = tab->Open((dword)(const char*)tab_name, TABOPEN_FILENAME | TABOPEN_UPDATE);
         if(!b){
            TablesCloseAll();
            return false;
         }
      }
   }
   return true;
}

//----------------------------
//----------------------------
