#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"


//----------------------------

#define TIME_BOMB_COUNT 15000 //default value
#define TIME_BOMB_MIN_COUNT 2000
#define TIME_BOMB_MAX_COUNT 90000
#define TIME_BOMB_COUNT_STEP 1000

//----------------------------
                              //time_bomb over network:
                              // - primary bomb:
                              //    - countdown
                              //    - create explosion
                              // - network bomb:
                              // - sent over network:
                              //    - S_vector pos
                              // - sent over network (run-time):
                              //    -bits:2 = code
                              //       - 0 = activate
                              //       - 1 = set count
                              //          - word new_count (in seconds)
                              //       - 2 = cancel use
class C_time_bomb: public C_actor{
   int counter;
   C_smart_ptr<C_icon_info> ii;
   C_smart_ptr<I3D_sound> tick_sound;
   bool activated;

//----------------------------

   enum E_NETWORK_MESSAGE{
      NM_ACTIVATE = 0,
      NM_SET_TIME = 1,
         //word
      NM_CANCEL_USE = 2,
   };
#define NM_NUM_SEND_BITS 2

//----------------------------

   void ShowTime(){
      int tt = (counter/1000) + 1;
      ii->SetText(C_fstring("%i:%.2i", tt/60, tt%60));
   }

//----------------------------

   bool DecTime(int t){

      int old_c = counter;
      counter -= t;
      if(counter<=0)
         return true;
      if((counter/1000) != (old_c/1000))
         ShowTime();
      return false;
   }

//----------------------------

   void Activate(){

      activated = true;
                              //send over network
      if(net && !network_actor){
         bool g = true;
         net_msg->Ascend(net_id);
         net_msg->AddBits(NM_ACTIVATE, NM_NUM_SEND_BITS, g);
         net_msg->Descend();
      }
   }

//----------------------------

   void SetCounter(int new_count, bool snd = true, bool net_g = false){

      if(counter!=new_count){
         counter = new_count;
         ShowTime();
         if(snd)
            mission.PlaySoundGlobal("tb_setup");
                              //send over network
         if(net && !network_actor){
            net_msg->Ascend(net_id);
            net_msg->AddBits(NM_SET_TIME, NM_NUM_SEND_BITS, net_g);
            net_msg->AddWord(counter/1000, net_g);
            net_msg->Descend();
         }
      }else
         if(snd)
            mission.PlaySoundGlobal("tb_err");
   }

//----------------------------

   void CancelUse(bool net_send = true){

      if(net_send){
                              //create resource actor
         S_resource_init ri;
         ri.pos = model->GetWorldPos();
         ri.inv_id = 10;
         ri.amount = 1;
         mission.CreateActor(ACTOR_RESOURCE, (dword)&ri);
                              //send cancel message over net
         if(net){
            bool g = true;
            net_msg->Ascend(net_id);
            net_msg->AddBits(NM_CANCEL_USE, NM_NUM_SEND_BITS, g);
            net_msg->Descend();
         }
      }
      mission.DestroyActor(this);
   }

//----------------------------

   void Init(){
      mission.AddIconInfo(ii);
      ShowTime();
                              //tick clock
      tick_sound = mission.CreateSound("tbomb");
      if(tick_sound){
         tick_sound->LinkTo(frame);
         tick_sound->SetRange(2.0f, 4.0f);
         tick_sound->SetLoop(true);
         tick_sound->SetOn(true);
         tick_sound->Release();
      }
   }
public:
   C_time_bomb(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_TIME_BOMB),
      activated(false),
      counter(TIME_BOMB_COUNT)
   {
      ii = new C_icon_info(mission.texts, 49);
      ii->Release();

                              //allow init data to be NULL, used during LoadGame
      if(!init_data)
         return;

      const S_time_bomb_init &ti = *(S_time_bomb_init*)init_data;

      frame = model = driver->CreateModel();
                              //determine unique name for model (due to SaveGame)
      C_string mname;
      for(int i=0; ; i++){
         mname = C_fstring("TimeBomb #%i", i);
         if(!mission.GetScene()->FindFrame(mname, ENUMF_MODEL))
            break;
      }
      frame->SetName(mname);

      const char *model_name = ((char*)tab_inventory->Item(TAB_S16_INV_MODEL_S)) + 16*10;
      model_cache.Open(model, model_name, mission.GetScene(), 0, ErrReport, editor);

      mission.LinkToFloor(frame, ti.pos, ti.who_put_it ? ti.who_put_it->GetFrame() : NULL, true);

      Init();

      network_actor = ti.netw_actor;

      if(net && !network_actor){
                              //assign ID
         mission.CreateActorNetID(this);

         bool g = true;
         net_msg->AddWord(C_game_mission::NM_GAME_CREATE_TIME_BOMB, g);
         net_msg->AddWord(net_id, g);
         net_msg->AddVector(ti.pos, g);
      }
   }

//----------------------------

   ~C_time_bomb(){
      mission.DelIconInfo(ii);
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      SG_SaveModelInfo(ck);
                           //write data
      ck->Write(&counter, sizeof(counter));
      ck->Write(&activated, sizeof(activated));
      ck->WStringChunk(CT_NAME, frame->GetName());
      C_actor::SaveGame(ck);
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      SG_LoadModelInfo(ck);
                              //read data
      ck->Read(&counter, sizeof(counter));
      ck->Read(&activated, sizeof(activated));
      C_string str;
      ck->ReadString(CT_NAME, str);
      frame->SetName(str);

      Init();
      return true;
   }

//----------------------------

   dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_USE_TIME_BOMB:
         switch(prm1){
         case 1:              //inc time
         case 2:              //dec time
            {
               int new_count =
                  (prm1==1) ?
                     Min(TIME_BOMB_MAX_COUNT-1, counter + TIME_BOMB_COUNT_STEP) :
                     Max(TIME_BOMB_MIN_COUNT-1, counter - TIME_BOMB_COUNT_STEP);
               SetCounter(new_count);
            }
            break;
         case 3:              //activate
            Activate();
            break;
         case 4:              //setup time directly
            SetCounter(prm2, false, true);
            break;
         case 5:
            CancelUse();
            return true;
         }
         break;
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual int Use(LPC_actor user, dword prm, bool force_use_off){
      if(!prm)
         Activate();
      return 0;
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      if(!activated)
         return 0;
      if(DecTime(tc.time)){
                              //create explosion
         if(!network_actor){
            S_explosion_init ei;
            ei.pos = frame->GetWorldPos();
            ei.id = EXPLOSION_TIMEBOMB;

                              //check if we're about to explode in water
            I3D_collision_data cd(ei.pos, S_vector(0, 2, 0), 0, frame, NULL);
            if(mission.GetScene()->TestCollision(cd)){
               PI3D_frame frm = cd.GetHitFrm();
               int mat_id = 0;
               switch(frm->GetType()){
               case FRAME_VOLUME: mat_id = I3DCAST_VOLUME(frm)->GetCollisionMaterial(); break;
               case FRAME_VISUAL: mat_id = I3DCAST_VISUAL(frm)->GetCollisionMaterial(); break;
               }
               if(mat_table->ItemB(TAB_B_MAT_WATER, mat_id))
                  ei.id = EXPLOSION_MINE;
            }
            mission.CreateActor(ACTOR_EXPLOSION, (dword)&ei);
         }
         mission.DestroyActor(this);
         return 1;
      }
      return 0;
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      in_msg->Ascend();
      while(in_msg->GetChunkSize()){
         byte code = in_msg->GetBits(NM_NUM_SEND_BITS);
         switch(code){
         case NM_ACTIVATE:
            Activate();
            break;
         case NM_SET_TIME:
            SetCounter(in_msg->GetWord()*1000, false);
            break;
         case NM_CANCEL_USE:
            CancelUse(false);
            break;
         default:
            assert(0);
         }
      }
      in_msg->Descend();
   }
};

//----------------------------

LPC_actor CreateTimeBomb(C_game_mission &gm, void *data){

   return new C_time_bomb(gm, data);
}

//----------------------------
