#include "all.h"
#include "net.h"
#include "actors.h"
#include "tables.h"
#include "winstuff.h"


//----------------------------

#ifdef _DEBUG
# define AUTO_DEBUG(n)
//# define AUTO_DEBUG(n) DEBUG(n)
# define AUTO_DEBUG_LOG(n)
//# define AUTO_DEBUG_LOG(n) OUTTEXT(n)

//# define AUTO_DRAW_AXIS
//#define AUTO_DRAW_TURN_TEST
//#define DRAW_AUTO_DIR_TEST
#else
# define AUTO_DEBUG(n)
# define AUTO_DEBUG_LOG(n)
#endif
                              //allow raise wheels every collision test and put them down at theend.
#define RIDE_SMALL_BARRIER
                              //size of such raise
#define RIDE_STEP_SIZE .32f
//#define RIDE_STEP_SIZE .12f
//#define RIDE_STEP_SIZE .04f
//#define RIDE_STEP_SIZE .0f
                              //how fast player die in max demage collision (msec)
#define PLR_HIT_FATAL_TIME 70
                              //speed at which auto stop if hit wall
#define HIT_STOP_SPEED .001f
                              //speed at which hit make minimum demage
#define HIT_HURT_SPEED_MIN .12f
                              //speed at which hit make maximum demage
#define HIT_HURT_SPEED_MAX 8.0f
                              //how much can wheels compensate terrain and how much can move in y
#define MAX_WHEEL_SPRING .2f
                              //how fast slow down to stop when hiting the wall (msec)
#define SPEED_LOSS_MAX .35f
                              //minimum percentage speed loss when hit wall
#define SPEED_LOSS_MIN .05f
                              //speed to crash/destroy power conversion:
#define DESTROY_SPEED 50.0f     //crash speed for default values (in km/h)
#define EXPLODE_SPEED 80.0f  //explode speed for default vaslues (in km/h)
#define DEFAULT_DESTROY_POWER 150.0f //taken from devalues in vehicle table (magic constant)
#define DEFAULT_EXPLODE_POWER 300.0f //taken from devalues in vehicle table (magic constant)

                              //how often emit sound when vehicle ride
#define VEHICLE_EMIT_SOUND_TIME 1200

#define ENABLE_AUTO_FALL
#define AUTO_CRASH_FALL_SPEED 20.0f
//#define AUTO_FLOOR_TEST_TIME 1000 //interval for test floor under car if not mooving

                              //debug:
//#define DISABLE_CAR_BODY_COL
//#define DISABLE_WHEEL_SPRING

//----------------------------

   struct S_wall_help{
      float wall_slide;
      PI3D_frame wall_vol;
      int time;
   };

//----------------------------
                              //versions over network:
                              // - send version change
                              //    - byte new_version
//----------------------------

enum{
   TAB_I_VER_DEFAULT,
   TAB_I_VER_RESISTANCE_SET,
   TAB_B_VER_LOOP_ANIM,
};

C_table_template templ_version = {
   // {7C5137C0-B470-11d2-87BD-004F4905FF1A}
   { 0x7c5137c0, 0xb470, 0x11d2, { 0x87, 0xbd, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Version properties",
   NULL,
   {
      {TE_INT, TAB_I_VER_DEFAULT, "Default version"},
      {TE_INT, TAB_I_VER_RESISTANCE_SET, "Resistance set"},
      {TE_BOOL, TAB_B_VER_LOOP_ANIM, "Looped animation"},
      {TE_NULL}
   }
};

//----------------------------

class C_version: public C_actor{

   LPC_table tab;
   vector<PI3D_frame> version_list;
   PI3D_frame link_prnt;
   int v_resistance;
protected:
   int curr_version;

//----------------------------

   static I3DENUMRET I3DAPI cbShow(PI3D_frame frm, dword on){
      frm->SetOn(on);
      return I3DENUMRET_OK;
   }
   static I3DENUMRET I3DAPI cbVol(PI3D_frame frm, dword c){
      I3DCAST_VOLUME(frm)->SetScene((PI3D_scene)c);
      return I3DENUMRET_OK;
   }
//----------------------------
   
   void SetVersion(int new_v, bool net_send, bool play_anim){

      if(!version_list.size())
         return;
      new_v %= version_list.size();
      if(curr_version==new_v)
         return;

      version_list[curr_version]->EnumFrames(cbShow, false, ENUMF_LIGHT | ENUMF_SOUND | ENUMF_VISUAL, NULL);
      version_list[curr_version]->EnumFrames(cbVol, NULL, ENUMF_VOLUME, NULL);
      version_list[new_v]->EnumFrames(cbShow, true, ENUMF_LIGHT | ENUMF_SOUND | ENUMF_VISUAL, NULL);
      version_list[new_v]->EnumFrames(cbVol, (dword)mission.GetScene(), ENUMF_VOLUME, NULL);

      version_list[curr_version]->SetOn(false);
      version_list[new_v]->SetOn(true);
      if(frame->GetType()==FRAME_VISUAL){
                              //version directly on object, migrate actor
         version_list[curr_version]->SetData(NULL);
         version_list[new_v]->SetData((dword)this);
      }

      {                          //bug override: if tanks are versions, hide cannon
         PI3D_frame frm = frame->FindChildFrame("joint_y");
         if(frm) frm->SetOn(false);
      }

      curr_version = new_v;

      if(net && net_send){
         const bool g = true;
         net_msg->Ascend(net_id);
                                 //routing
         switch(GetType()){
         case ACTOR_AUTOMOBIL: net_msg->AddBits(0, 1, g); break;
         case ACTOR_AUTO_CANNON: net_msg->AddBits(0, 2, g); break;
         }
         net_msg->AddByte(curr_version, g);
         net_msg->Descend();
      }
                              //reset animation (some versions are done through animated objects)
      if(model){
         model->SetAnimation(0, NULL);
         if(!play_anim)
            model->Tick(10000);
      }
   }

//----------------------------

   void AddResist(){
      v_resistance += tab_resistance->ItemI(TAB_I_RSET_VALUE, tab->ItemI(TAB_I_VER_RESISTANCE_SET));
   }

//----------------------------

   virtual bool HitExplode(C_actor *caller, int power, S_CB_hit *hit, bool make_expl){

      if(v_resistance && curr_version<version_list.size()-1){
         int res_set = tab->ItemI(TAB_I_VER_RESISTANCE_SET);
         const int thresh = tab_resistance->ItemI(TAB_I_RSET_THRESH, res_set);
         if(power < thresh)
            return false;
         if((v_resistance -= power) <= 0){
            SetVersion(curr_version+1, true, true);
            if(curr_version==version_list.size()-1){
               v_resistance = 0;
            }else
               AddResist();
                              //create explosion
            int expl = tab_resistance->ItemI(TAB_I_RSET_EXPLOSION, res_set);
            if(expl!=-1 && expl<6 && make_expl){
               S_explosion_init ei;
               if(hit){
                           //push it something inside
                  ei.pos = hit->hit_pos - hit->hit_norm*.5f;
                  ei.actor_reason = hit->shooter;
               }else{
                     GetCenterPos(ei.pos);
               }
               ei.id = expl;
               ei.creator = this;
               mission.CreateActor(ACTOR_EXPLOSION, (dword)&ei);
            }
         }
      }
      return false;
   }

//----------------------------
public:
   C_version(C_game_mission &m, E_ACTOR_TYPE t = ACTOR_VERSION):
      C_actor(m, t),
      link_prnt(NULL),
      curr_version(0),
      v_resistance(0),
      tab(CreateTable())
   {
      tab->Open((dword)&templ_version, TABOPEN_TEMPLATE);
   }

   ~C_version(){
      if(tab) tab->Release();
      if(frame->GetType()==FRAME_VISUAL)
         SetVersion(0, false, false);
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      C_actor::SetFrame(frm);
      {
         if(frm->GetType()==FRAME_MODEL){
            PI3D_model mp = I3DCAST_MODEL(frm);
            mp->SetAnimation(0, NULL);
         }

         PI3D_frame f = frm->FindChildFrame("~01*", ENUMF_VISUAL | ENUMF_MODEL | ENUMF_DUMMY | ENUMF_WILDMASK);
         if(f){
            f = f->GetParent();
            link_prnt = f->GetParent();
            const S_vector &base_pos = f->GetPos();
            const S_quat &base_rot = f->GetRot();
            const float base_scl = f->GetScale();
            PI3D_frame prev_f;
            do{
                              //hide all versions except first
               if(version_list.size()){
                  f->LinkTo(link_prnt);
                  f->SetOn(false);
                  f->SetPos(base_pos);
                  f->SetRot(base_rot);
                  f->SetScale(base_scl);
                  f->Update();

                  f->EnumFrames(cbShow, false, ENUMF_LIGHT | ENUMF_SOUND | ENUMF_VISUAL, NULL);
                  f->EnumFrames(cbVol, NULL, ENUMF_VOLUME, NULL);

                  /*
#ifdef EDITOR
                              //check if not bad linked
                  if(!prm2){
                     if(prev_f->FindChildFrame("~01*", ENUMF_VISUAL | ENUMF_MODEL | ENUMF_DUMMY | ENUMF_WILDMASK)){
                        ReportActorError("bad linked version");
                     }
                  }
#endif
                  */
               }
               prev_f = f;

               version_list.push_back(f);
            }while(f=f->FindChildFrame("~01*", ENUMF_VISUAL | ENUMF_MODEL | ENUMF_DUMMY | ENUMF_WILDMASK));
         }
      }
      /*
      if(!prm2){
         if(version_list.size()<=1){
            v_resistance = 0;
            ReportActorError("no multiple versions");
         }
      }
      */
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

                           //write data
      ck->Write(&v_resistance, sizeof(v_resistance));
      ck->Write(&curr_version, sizeof(curr_version));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

                              //read data
      ck->Read(&v_resistance, sizeof(v_resistance));
      dword new_v;
      ck->Read(&new_v, sizeof(new_v));
      SetVersion(new_v, false, false);

      return true;
   }

//----------------------------

   dword C_version::cbProc(dword msg, dword prm1, dword prm2){

      switch(msg){

      case CB_HIT:
         HitExplode(NULL, prm2, (S_CB_hit*)prm1, true);
         break;

      case CB_SIGNAL:
         {
            S_signal_data *sd = (S_signal_data*)prm1;
            if(sd->type==SIGNAL_DETECTOR){
               switch(sd->sub_type){
               case 0:        //switch version
                  if(curr_version<version_list.size()-1){
                     SetVersion(curr_version+1, true, true);
                     return true;
                  }
                  break;
               }
            }
         }
         break;
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual bool MayExplode() const{
      return (curr_version != version_list.size()-1);
   }

   virtual bool Explode(C_actor *caller, int power){
      return HitExplode(caller, power, NULL, true);
   }

//----------------------------

   virtual void NetCodeIn(byte msg_code, PC_net_in_msg in_msg){

      SetVersion(msg_code, false, true);
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

                              //skip chunk size
      in_msg->GetWord();
      NetCodeIn(in_msg->GetByte(), in_msg);
   }

//----------------------------

   virtual bool IsDestroyed() const{ return (curr_version!=0); }

//----------------------------

   virtual LPC_table GetTable(int index) const{
      if(index==0)
         return tab;
      return C_actor::GetTable(index);
   }

//----------------------------

   virtual class C_table_template *GetTemplate(int) const{ return &templ_version; }

   virtual void TableUpdate(int index, bool not_load_prg){

      switch(index){
      case -1:
         {
            if(version_list.size()<=1)
               v_resistance = 0;
            else
               v_resistance = tab_resistance->ItemI(TAB_I_RSET_VALUE, tab->ItemI(TAB_I_VER_RESISTANCE_SET));
         }
      case TAB_I_VER_DEFAULT:
         SetVersion(tab->ItemI(TAB_I_VER_DEFAULT), false, false);

      case TAB_B_VER_LOOP_ANIM:
                              //setup loop animation for special cases(animated dogs)
         if(model && tab->ItemB(TAB_B_VER_LOOP_ANIM) && !IsDestroyed())
            model->SetAnimation(0, NULL, I3DANIMOP_LOOP);
         break;
      }
   }

   virtual void MissionSave(C_chunk *ck) const{ SaveTable(ck, tab); }
   virtual bool MissionLoad(C_chunk *ck, int table_id){ return LoadTable(ck, tab); }
};

//----------------------------


LPC_actor CreateVersion(C_game_mission &gm){

   return new C_version(gm);
}

//----------------------------
//----------------------------
                              //automobil over network:
                              // - actor routing:
                              // - bits:1:
                              //    - 0 = version
                              //    - 1 = automobil
                              // - bits:4 command
                              //    - 0 = seat sync
                              //       - bits:3 seat_index
                              //       - message for C_net_sync
                              //    - 1 = seat's user
                              //       - bits:3 seat_index
                              //       - word user_net_id
                              //    - 2 = motor off/on
                              //       - bits:1 motor state
                              //    - 3 = wheel turn angle
                              //       - float angle
                              //    - 4 = fuel
                              //       - float fuel
                              //    - 5 = speed
                              //       - word speed
                              //    - 6 = pos
                              //       - S_vector pos
                              //    - 7 = rot
                              //       - S_quat rot (S_vector, float)
                              //    - 8 = pos + rot
                              //    - 9 = version change
                              //       - bits:1 exploded
                              //          - byte version (if exploded)
                              //    - 10 = reservation
                              //       - bits:3 seat_index
                              //       - word reserver_net_id
                              //       - word time
                              //    - 11 = start airplane animation


#define AUTO_WHEEL_DIST 2.3f  //meters

//----------------------------
#define MAX_AUTO_SEATS 8

C_table_template templ_automobil = {
// {C338ECC0-C99A-11d2-87BD-004F4905FF1A}
   { 0xc338ecc0, 0xc99a, 0x11d2, { 0x87, 0xbd, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Automobil properties",
   NULL,
   {
      //{TE_INT, TAB_I_AUTO_MAX_SPEED, "Maximal speed", 0, 0, 60},
      {TE_INT, TAB_I_AUTO_NUM_TRANSL, "Number of transmissions", 0, 0, 4},
      {TE_FLOAT, TAB_F_AUTO_FUEL, "Fuel", 0, 0, 5},
      {TE_FLOAT, TAB_F_AUTO_CONSUM, "Consumption", 0, 5, 1},
      {TE_BRANCH, 0, "Sound", 3},
         {TE_STRING, TAB_S12_AUTO_SOUND, "Engine sound", 12},
         {TE_STRING, TAB_S12_AUTO_SOUND_ON, "Engine start", 12},
         {TE_STRING, TAB_S12_AUTO_SOUND_OFF, "Engine stop", 12},
      {TE_BRANCH, 0, "Advanced", 12},
         {TE_INT, TAB_I_AUTO_MAX_WHEEL_TURN, "Max wheel turn angle", 0, 90, 45},
         {TE_BRANCH, 0, "Texture shift", 3},
            {TE_ARRAY, TAB_S16_AUTO_TXT_SHIFT, "Objects", 4},
               {TE_STRING, TAB_S16_AUTO_TXT_SHIFT, "Name", 16},
            {TE_INT, TAB_I_AUTO_TXS_ANGLE, "Angle", 0, 359},
            {TE_FLOAT, TAB_F_AUTO_TXS_SPEED, "Speed for 1 m/s", 0, 0, 1},
         {TE_FLOAT, TAB_F_AUTO_ACCEL, "Acceleration constant", 0, 10, 1},
         {TE_FLOAT, TAB_F_AUTO_BRAKE, "Brake constant", 0, 10, 6},
         {TE_INT, TAB_I_AUTO_MAX_MOTOR_TURN, "Max motor turn", 0, 0, 5000},
         {TE_INT, TAB_I_AUTO_TRANS_UP, "Trans up at", 0, 0, 3500},
         {TE_INT, TAB_I_AUTO_TRANS_DOWN, "Trans down at", 0, 0, 1700},
         {TE_ARRAY, TAB_B_WHEEL_ON, "Wheels", 4},
            {TE_BRANCH, 0, "Wheel", 1},
               {TE_BOOL, TAB_B_WHEEL_ON, "Collision", 0, 0, 1},
         {TE_ARRAY, TAB_E_AUTO_SEAT_STAY, "Seats", MAX_AUTO_SEATS},
            {TE_BRANCH, 0, "Seat", 3},
               {TE_ENUM, TAB_E_AUTO_SEAT_STAY, "Use mode", (dword)"Sit\0Stay\0Bike\0Bench\0"},
               {TE_BOOL, TAB_B_AUTO_SEAT_WEAPON, "Use weapons", 0, 0, 1},
               {TE_BOOL, TAB_B_AUTO_SEAT_CANNON, "Cannon position"},
         {TE_ENUM, TAB_E_AUTO_TYPE, "Type", (dword)"Auto\0Boat\0"},
         {TE_FLOAT, TAB_F_AUTO_TURN_RATIO, "Speed ratio", 0, 2, 1},
         {TE_BRANCH, 0, "Resistance", 2},
            {TE_INT, TAB_I_AUTO_EXPL_THRESH, "Speed loss for explode", 0, 500, 300},
            {TE_INT, TAB_I_AUTO_DESTROY_THRESH, "Speed loss for destroy", 0, 500, 150},
      {TE_NULL}
   }
};

//----------------------------

static const S_vector axis_y(0.0, 1.0, 0.0);
                           //speed per AUTO_TRANSLATE_UP t/sec
static const float wanted_speed[]={
   -3.5,
   0,
   2, 8, 16, 32, 50, 72
};

class C_automobil: public C_version{

   enum{
      MODE_STAY,
      MODE_START,
      MODE_GO,
      MODE_STOP,
      MODE_DESTROYED,
   } mode;
   int mode_count;

                              //wheels
   struct S_wheel{
      PI3D_frame wheel;      //wheel object
      PI3D_frame joint;
      S_quat joint_base_rot;
      float volume_radius;    //for wheel collision testing
      float fall_speed;
      float world_y;
      float base_loc_y;

      void SetWorldY(){
         if(joint)
            world_y = joint->GetWorldPos().y;
      }
   } wheels[4];
   bool auto_falling;
   float fall_speed;
   int floor_test_cnt;
   PI3D_frame volant;
   S_quat volant_base_rot;

   int num_of_seats;
   struct S_seat{
      bool driver;
      PI3D_frame seat;
      PI3D_frame entry;
      LPC_actor user;
                              //reservation
      int reserve_count;
      LPC_actor reserver;

      C_smart_ptr<C_net_sync> net_sync;
      LPC_actor net_sync_user;

      bool IsFree() const{
         if(seat){
            if(!user){
               if(!net || !net_sync->GetState())
                  return true;
            }
         }
         return false;
      }
   } seats[MAX_AUTO_SEATS];
   float fuel;                //in litres
   float speed;               //m/sec
   PI3D_sound snd_engine;
   PI3D_frame engine_dummy;
   float last_engine_scale;   //for scaling sound pitch
   PI3D_frame col_parent;
   int transl;                //-1 (reverse) - 4
   int eng_turn;
   float wheel_turn;          //-1.0 to 1.0
   bool proposed_forward;
   float axes_dist;
   float back_axle_len;
   int airplane_time;         //time of airplane animation, or -1 if not animated
   int emit_snd_cnt;
   float airplane_speed;
   S_vector last_airplane_pos;

                              //network
   float net_wheel_turn;
   float net_fuel;
   float net_speed;
   S_vector net_last_sent_pos;//for delta
   int net_last_sent_time;
   int net_pos_count;

   vector<C_smart_ptr<I3D_visual> > tx_shift;

   C_smart_ptr<C_icon_info> fuel_icon;
protected:
   LPC_table tab;

   byte net_route_bits;       //how many bits we use for routing (1 for C_automobil, 2 for C_auto_cannon)

   S_seat *GetFreeSeat(int indx = 0, bool all_indx_based = false){
      return (S_seat*)GetFreeSeatConst(indx, all_indx_based);
   }
   const S_seat *GetFreeSeatConst(int indx = 0, bool all_indx_based = false) const{

      assert(indx>=0 && indx<num_of_seats);
      if(indx >= num_of_seats)
         return NULL;
      if(all_indx_based)
         return seats[indx].IsFree() ? &seats[indx] : NULL;
      for(int i=0; i<MAX_AUTO_SEATS; i++){
         if(seats[i].IsFree() && !indx--) 
            return &seats[i];
      }
      return NULL;
   }

   S_seat *GetUserSeat(LPC_actor u){
      for(int i=MAX_AUTO_SEATS; i--; )
         if(seats[i].user==u) return &seats[i];
      return NULL;
   }

//----------------------------

   int NumFreeSeats() const{
      for(int i=MAX_AUTO_SEATS, n=0; i--; )
         if(seats[i].IsFree()) ++n;
      return n;
   }

//----------------------------

   void SetEngTurn(int et){
      eng_turn = et;
      if(snd_engine){
         snd_engine->SetFrequency(.5f + ((float)eng_turn / 1600.0f)*.5f);
         snd_engine->SetVolume(.5f + ((float)eng_turn / 800.0f)*.5f);
      }
   }

//----------------------------

   void ParticleOn(bool b, PI3D_frame mp = NULL){
      if(!mp) mp = model;
      if(!mp) return;
      struct S_hlp{
         static I3DENUMRET I3DAPI cbPrt(PI3D_frame frm, dword c){
            if(I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_PARTICLE){
               frm->SetOn(c);
            }
            return I3DENUMRET_OK;
         }
      };
      mp->EnumFrames(S_hlp::cbPrt, b, ENUMF_VISUAL);
   }

//----------------------------

   void MotorOn(bool b, bool net_send = true, dword net_route_code = 0, byte net_route_bits = 0){

      ParticleOn(b);

      switch(mode){
      case MODE_GO:
         if(b) return;
         break;
      case MODE_START:
      case MODE_STOP:
         return;
      case MODE_STAY:
         if(!b) return;
         break;
      case MODE_DESTROYED:
         return;
      }
      const char *cp;
      if(b){
         mode = MODE_START;
         cp = tab->ItemS(TAB_S12_AUTO_SOUND_ON);
         if(!*cp) cp = "motor_on";
      }else{
         mode = MODE_STOP;
         cp = tab->ItemS(TAB_S12_AUTO_SOUND_OFF);
         if(!*cp) cp = "motor_of";
         if(snd_engine)
            snd_engine->SetOn(false);
         if(fuel_icon){
            mission.DelIconInfo(fuel_icon);
            fuel_icon = NULL;
         }
         if(speed<.5f){
            speed = 0.0f;
            eng_turn = 0.0f;
         }
      }
      C_string fname = cp;
      for(int i=fname.Size(); i--; ){
         if(fname[i]=='.'){
            fname[i] = 0;
            break;
         }
      }
      mission.PlaySound(fname, engine_dummy,
         tab_game_cfg->ItemF(TAB_F_GAME_SNDR_AUTO_MIN),
         tab_game_cfg->ItemF(TAB_F_GAME_SNDR_AUTO_MAX));
      mode_count = tab_game_cfg->ItemI(TAB_I_GAME_AUTO_START_DELAY);

      if(net && net_send){
         bool g = true;
         net_msg->Ascend(net_id);
         net_msg->AddBits(net_route_code, net_route_bits, g);
         net_msg->AddBits(2, 4, g);          //code
         net_msg->AddBits(b, 1, g);          //motor state
         net_msg->Descend();
                              //update pos/rot
         if(!b)
            NetSendPos(g, false, false, net_route_code, net_route_bits);
      }
   }

//----------------------------

   void SetWheelTurn(float f, bool net_send = true, dword net_route_code = 0, byte net_route_bits = 0){

      if(wheel_turn>0.0){
         if(f>0.0) wheel_turn = Min(1.0f, wheel_turn+f);
         else
            wheel_turn = Max(0.0f, wheel_turn + f*2.0f);
      }else
      if(wheel_turn<0.0){
         if(f<0.0) wheel_turn = Max(-1.0f, wheel_turn+f);
         else
            wheel_turn = Min(0.0f, wheel_turn + f*2.0f);
      }else
         wheel_turn = Min(1.0f, Max(-1.0f, wheel_turn+f));

      S_quat q;
      static const S_vector axis(0, 1, 0);
      for(int i=0; i<2; i++)
      if(wheels[i].joint){
         float max_wt = tab->ItemI(TAB_I_AUTO_MAX_WHEEL_TURN)*GetPI()/180.0f;
         q.Make(axis, wheel_turn * max_wt);
         wheels[i].joint->SetRot(q * wheels[i].joint_base_rot);
         wheels[i].joint->Update();
      }
      if(volant){
         volant->SetRot(q * volant_base_rot);
         volant->Update();
      }
      if(net){
         if(net_send){
            float delta = fabs(net_wheel_turn - wheel_turn);
            if(delta>PI*.05f){
               int g = INET_MSG_NGP;
               net_msg->Ascend(net_id);
               net_msg->AddBits(net_route_code, net_route_bits, g);
               net_msg->AddBits(3, 4, g);
               net_msg->AddFloat(wheel_turn, g);
               net_msg->Descend();
               net_wheel_turn = wheel_turn;
            }
         }else
            net_wheel_turn = wheel_turn;
      }
   }

//----------------------------

   void SetFuel(float f, bool net_send = true, byte net_route_bits = 0){

      fuel = f;
      if(net){
         if(net_send){
            float delta = fabs(net_fuel - fuel);
            if(delta>.1f){
               bool g = false;
               net_msg->Ascend(net_id);
               net_msg->AddBits(1, net_route_bits, g);
               net_msg->AddBits(4, 4, g);
               net_msg->AddFloat(fuel, g);
               net_msg->Descend();

               net_fuel = fuel;
            }
         }else
            net_fuel = fuel;
      }
      ShowFuel();
   }

//----------------------------

   void SetSpeed(float s, bool net_send = true, dword net_route_code = 0, byte net_route_bits = 0){
      speed = s;
      if(net){
         if(net_send){
            float delta = fabs(net_speed - speed);
            if(delta>1.0f || (!speed && net_speed)){
               bool g = false;
               net_msg->Ascend(net_id);
               net_msg->AddBits(net_route_code, net_route_bits, g);
               net_msg->AddBits(5, 4, g);
               net_msg->AddWord((short)speed, g);
               net_msg->Descend();

               net_speed = speed;
            }
         }else
            net_speed = speed;
      }
   }

//----------------------------

   void NetSendPos(int time, bool guaranteed = false, bool force = false, dword net_route_code = 0, byte net_route_bits = 0){

      if(net){
         const S_vector &pos = frame->GetPos();
         if(!force){
            int curr_timer = igraph->ReadTimer();
                              //get delta
            int time_delta = curr_timer - net_last_sent_time;
            if(time_delta<100){
               float delta = (net_last_sent_pos-pos).Magnitude();
                  if(delta<.1f)
                     return;
            }
            net_last_sent_pos = pos;
            net_last_sent_time = curr_timer;
         }
         E_INET_MESSAGE_TYPE mt = INET_MSG_NGP;
         if(guaranteed)
            mt = INET_MSG_G;
         net_msg->Ascend(net_id);
         net_msg->AddBits(net_route_code, net_route_bits, mt);
         net_msg->AddBits(8, 4, mt);
         net_msg->AddVector(pos, mt);
         S_quat rot = frame->GetRot();
         net_msg->AddVector(rot.v, mt);
         net_msg->AddFloat(rot.s, mt);

         net_msg->Descend();
      }
   }

//----------------------------

   void InitFuelIcon(){

      if(seats[0].user && seats[0].user->GetType()==ACTOR_PLAYER){
         fuel_icon = new C_icon_info(mission.texts, IIMG_PETROL);
         fuel_icon->Release();
         mission.AddIconInfo(fuel_icon);
      }
   }

//----------------------------

   void ShowFuel() const{
      if(fuel_icon)
         fuel_icon->SetText(C_fstring("%.1f", fuel));
   }

//----------------------------

   void GasDown(int time){
      SetEngTurn(Max(0.0f, eng_turn - time*.001f * 400.0f));
   }

//----------------------------

   void AdjustEngTurn(float tsec){
      if(transl && (speed || eng_turn)){
                              //map turning back to speed
         //int ideal_eng_turn = (speed/(wanted_speed[transl+1]*tab->ItemF(TAB_F_AUTO_TURN_RATIO))) * tab->ItemI(TAB_I_AUTO_TRANS_UP);

         float max_speed = wanted_speed[transl+1]*tab->ItemF(TAB_F_AUTO_TURN_RATIO);
         AUTO_DEBUG(max_speed*3.6f);
         int ideal_eng_turn = tab->ItemI(TAB_I_AUTO_TRANS_UP) * (speed / max_speed);
         ideal_eng_turn = Min(ideal_eng_turn, tab->ItemI(TAB_I_AUTO_MAX_MOTOR_TURN));
         AUTO_DEBUG(C_fstring("ideal ot: %i",ideal_eng_turn));
         if(eng_turn < ideal_eng_turn){
            if(fuel){
               eng_turn = Min(ideal_eng_turn, (int)(eng_turn + tsec*9000.0f));
            }
         }else
         if(eng_turn > ideal_eng_turn){
            eng_turn = Max(ideal_eng_turn, (int)(eng_turn - tsec*9000.0f));
         }
         SetEngTurn(eng_turn);
      }
   }

//----------------------------

   void Translate(int delta){
      transl += delta;
                              //send message to driver
      if(seats[0].user && tab->ItemE(TAB_E_AUTO_SEAT_STAY, 0)!=2) 
         seats[0].user->cbProc(CB_USE_AUTO, 9);
   }

//----------------------------

   void OpenSound(){

      if(snd_engine){
         snd_engine->Release();
         snd_engine = NULL;
      }
      if(isound){
         snd_engine = mission.GetScene()->CreateSound();
         if(snd_engine){
            const char *cp = tab->ItemS(TAB_S12_AUTO_SOUND);
            ISND_RESULT ir;
            if(!*cp)
               cp = "auto_eng";
            snd_engine->SetSoundType(I3DSOUND_POINT);
                              //remove extension
            C_string str = cp;
            for(int i=str.Size(); i--; ){
               if(str[i]=='.'){
                  str[i] = 0;
                  break;
               }
            }
            ir = sound_cache.Open(snd_engine, str, mission.GetScene(), I3DLOAD_SOUND_FREQUENCY);

            if(ISND_FAIL(ir))
               ReportActorError(C_fstring("can't open sound %s", cp));
            snd_engine->SetRange(2.0f, 40.0f);
            snd_engine->SetLoop(true);
            snd_engine->LinkTo(engine_dummy);
            snd_engine->SetOn(false);
         }
      }
   }

//----------------------------

#define TURN_SPEED 3000
   void SlowDown(int time){

      int i;
      if(!curr_version){
         GasDown(time);
         return;
      }
      if(!transl){
         if(fuel){
            if(eng_turn > 1000){
               i = Max(1000.0f, eng_turn - time*.001f * TURN_SPEED);
               SetEngTurn(i);
            }
         }
      }else
      if(!fuel)
         GasDown(time);
   }

//----------------------------

   void ControlEnginePitch(){

      if(engine_dummy && last_engine_scale!=engine_dummy->GetScale()){
         last_engine_scale = engine_dummy->GetScale();
         if(engine_dummy->NumChildren()){
            PI3D_frame sndf = engine_dummy->GetChild(0);
            if(sndf->GetType()==FRAME_SOUND){
               PI3D_sound snd = I3DCAST_SOUND(sndf);
               snd->SetFrequency(last_engine_scale);
               snd->SetVolume(last_engine_scale);
            }
         }
      }
   }

//----------------------------

   void Destroy(bool exploded, bool net_send, dword net_route_code, byte net_route_bits){

      MotorOn(false, net_send, net_route_code, net_route_bits);
      if(exploded){
         for(int i=0; i<4; i++){
            if(wheels[i].wheel)
               wheels[i].wheel->SetOn(false);
         }
      }
      mode = MODE_DESTROYED;

      if(net && net_send){
         bool g = true;
         net_msg->Ascend(net_id);
         net_msg->AddBits(net_route_code, net_route_bits, g);
         net_msg->AddBits(9, 4, g);
         net_msg->AddBits(exploded, 1, g);
         //if(exploded) net_msg->AddByte(curr_version, g);
         net_msg->Descend();
      }
      AUTO_DEBUG_LOG("Automobil destroyed");
   }

//----------------------------

   void AirplaneStart(bool net_send = true){

                              //sometimes is airplane start send by wrong detectors, so
                              //ignore it for non plane vehicles(airplanes has model NULL)
      if(model)
         return;
                              //start airplane animation
      PI3D_frame mp = frame;
      if(mp->GetType()!=FRAME_MODEL) mp = mp->GetParent();
      if(mp && mp->GetType()==FRAME_MODEL){
         PI3D_model model = I3DCAST_MODEL(mp);
         PI3D_animation_set as = model->GetAnimationSet();
         if(as){
            model->SetAnimation(0, NULL);
            if(snd_engine){
               snd_engine->SetRange(100.0f, 400.0f);
               snd_engine->SetOn(true);
            }
            airplane_time = 0;

            if(net && net_send){
               const bool g = true;
               net_msg->Ascend(net_id);
               net_msg->AddBits(1, net_route_bits, g);
               net_msg->AddBits(11, 4, g);
               net_msg->Descend();
            }
         }
      }
   }

//----------------------------

   struct S_mobile_context{
      bool mobil_boat;
      bool boat_water_coll;
      dword mobil_kill_mask;
      bool coll_wall;        //true if wheel is in collision with wall
      float coll_wall_slide; //if wall collision, this value is in range 0.0 to 1.0 representing collision with wall
      //PI3D_frame wall_vol;
      C_smart_ptr<I3D_frame> wall_vol;
      S_vector auto_dir;
      bool get_wall_info;     //some test are only for get auto to groung, they not return wall collision
      bool coll_floor;
      int col_mat_id;         //currently used in water test.
      LPC_actor this_actor;

      S_mobile_context() :
         mobil_boat(false),
         boat_water_coll(false),
         mobil_kill_mask(0),
         coll_wall(false),     //true if wheel is in collision with wall
         coll_wall_slide(1.0f), //if wall collision, this value is in range 0.0 to 1.0 representing collision with wall
         get_wall_info(true),
         coll_floor(false),
         wall_vol(NULL),
         col_mat_id(0)
      {}

   };

   static bool I3DAPI MobilColResp(I3D_collision_data &cd){

      assert(cd.cresp_context);
      S_mobile_context &cs = *(S_mobile_context *)(cd.cresp_context);
      //AUTO_DEBUG(C_fstring("col: %s (%s)", (const char*)vol->GetName(), v1 ? v1->GetName() : "?"));
      AUTO_DEBUG(C_fstring("col: %s", (const char*)cd.GetHitFrm()->GetName()));

      int mat_id(0);
      switch(cd.GetHitFrm()->GetType()){
      case FRAME_VOLUME:
         mat_id = I3DCAST_VOLUME(cd.GetHitFrm())->GetCollisionMaterial(); break;
      case FRAME_VISUAL:
         mat_id = I3DCAST_VISUAL(cd.GetHitFrm())->GetCollisionMaterial(); break;
      default: assert(0);
      }
      if(mat_id==40){   //callback
         LPC_actor act = (LPC_actor)cd.GetHitFrm()->GetData();
         assert(act);
         if(act)
            act->cbProc(CB_COLLISION, (dword)cs.this_actor);
      }

#if 0 //comment out: (bushes gets inside vehicle)
                              //ride throught bushes
      if(mat_id == BUSH_RIDE_FREE_MAT)
         return false;
#endif

      cs.col_mat_id = mat_id;
                              //store kill mode, also in case when material is water and is not responded.
      dword kill_mask = mat_table->ItemE(TAB_E_MAT_KILL, mat_id);
      if(kill_mask)
         cs.mobil_kill_mask = kill_mask;

      if(mat_table->ItemB(TAB_B_MAT_WATER, mat_id)){
         if(!cs.mobil_boat)
            return false;
         cs.boat_water_coll = true;
      }
                              //to detect wall check real face normal
      const S_vector limit_normal = cd.GetFaceNormal();

                              //ignore cails - mainly for humans laying on cars
      static const float CEIL_ANGLE = -.7f;
      //if(limit_normal.y < CEIL_ANGLE)      
      if((limit_normal.y < CEIL_ANGLE) && (mat_id == HUMAN_SPHERE_MAT))
      {
         //OUTTEXT(C_fstring("%s ignored, %.2f", (const char*)cd.GetHitFrm()->GetName(), limit_normal.y));
         return false;
      }

      if(fabs(limit_normal.y) >= .8f){
         driver->CollisionResponse(cd);
         if(limit_normal.y >= 0.0f){
            cs.coll_floor = cd.GetHitFrm();
         }
         return true;
      }else{
         if(cs.get_wall_info)
         {
                              //to detect slide get collision normal(modified on edge)
            const S_vector &limit_normal = cd.GetHitNormal();
            //DebugLine(cd.from, cd.from + limit_normal, 0, 0xffff0000);
            cs.coll_wall = true;
            cs.wall_vol = cd.GetHitFrm();
                                 //get angle against auto_dir
            float angle = cs.auto_dir.AngleTo(limit_normal);
            if(angle >= GetPI()) angle = GetPI2()-angle;
            float slide = 1.0f;
            if(angle>=(PI*.5f))
               slide -= ((angle-PI*.5f)/(PI*.5f));
            assert(slide>=0 && slide<=1);
            cs.coll_wall_slide = slide;
         }
         driver->CollisionResponse(cd);
         return true;
      }
   }

//----------------------------

   void HitExplode(LPC_actor caller, int power, S_CB_hit *hit, bool make_expl, dword net_route_code, byte net_route_bits){

      C_version::HitExplode(caller, power, hit, make_expl);
      if(curr_version){
         Destroy(true, true, net_route_code, net_route_bits);
      }
   }

//----------------------------
//<comment>
   void ProcessNetSeats(byte net_route_bits){

      if(net){
         for(int i=num_of_seats; i--; ){
            S_seat &seat = seats[i];
            dword winner_pid;
            if(seat.net_sync->GetState(&winner_pid) == C_net_sync::SS_RESOLVED){
               if(winner_pid==net->GetPID()){
                  //seat.net_sync->Reset();
                  assert(seat.net_sync_user);
                  seat.user = seat.net_sync_user;
                  bool b = seat.user->cbProc(CB_USE_AUTO, 0, (dword)this);
                  if(!b)
                     seat.user = NULL;
                  seat.net_sync_user = NULL;
                              //say to all who's user
                  bool g = true;
                  net_msg->Ascend(net_id);
                              //route code (for C_automobil)
                  net_msg->AddBits(1, net_route_bits, g);
                              //code (set user seat)
                  net_msg->AddBits(1, 4, g);
                  net_msg->AddBits(i, 3, g);  //seat_index
                  net_msg->AddWord(seat.user ? seat.user->GetNetID() : 0, g);
                  net_msg->Descend();
               }
               seat.net_sync->Reset();
            }
         }
      }
   }

//----------------------------

   static bool I3DAPI MobilTurnColResp(I3D_collision_data &cd){

      int mat_id(0);
      switch(cd.GetHitFrm()->GetType()){
      case FRAME_VOLUME:
         mat_id = I3DCAST_VOLUME(cd.GetHitFrm())->GetCollisionMaterial(); break;
      case FRAME_VISUAL:
         mat_id = I3DCAST_VISUAL(cd.GetHitFrm())->GetCollisionMaterial(); break;
      default: assert(0);
      }
      if(mat_table->ItemB(TAB_B_MAT_WATER, mat_id) || (mat_id == HUMAN_SPHERE_MAT)
#if 0 //comment out: (bushes gets inside vehicle)
         || (mat_id == BUSH_RIDE_FREE_MAT)
#endif
         )
         return false;
      else
         return true;
   }

//----------------------------

   struct S_turn_hlp{

      S_vector center;
      bool turn_is_free;
      S_matrix m_rot;
      S_vector move_dir;
      PI3D_scene scene;
      PI3D_model model;
   };

//----------------------------

   static I3DENUMRET I3DAPI cbCheckTurnCol(PI3D_frame frm, dword hlp){

      const S_vector axis_y(.0f, 1.0f, .0f);
      PI3D_volume vol = (PI3D_volume)frm;

      if(vol->GetVolumeType() == I3DVOLUME_SPHERE){

         S_turn_hlp &th = *(S_turn_hlp*)hlp;

                                 //colision sphere values
         const S_matrix &mv = frm->GetMatrix();
         const S_vector &pos = mv(3);
         float radius = mv(0).Magnitude();
         radius *= .75f; //put some tolerance
         S_vector delta = pos - th.center;
         //DebugLine(pos, th.center, 0, 0xff0000ff);
         S_vector rotated = delta.RotateByMatrix(th.m_rot);
         S_vector rot_pos = th.center + rotated;
         //DebugLine(pos, rot_pos, 0, 0xffffffff);
         S_vector test_dir = th.move_dir.Cross(axis_y) * radius;
         //DebugLine(rot_pos, rot_pos + test_dir, 0, 0xff00ffff);
         float cosh = test_dir.Dot(rot_pos - pos);
         if(cosh < MRG_ZERO)
            test_dir.Invert();

         S_vector to = rot_pos + test_dir;
#ifdef AUTO_DRAW_TURN_TEST
         DebugLine(pos, to, 0, 0xff00ffff);
#endif

         I3D_collision_data cd;
         cd.from = pos;
         cd.dir = to - pos;
         cd.flags = 0;
         assert(th.model);
         cd.frm_ignored = th.model;
         cd.cresp_face = MobilTurnColResp;
         assert(th.scene);
         if(th.scene->TestCollision(cd)){
            th.turn_is_free = false;
            return I3DENUMRET_CANCEL;
         }
      }
      return I3DENUMRET_SKIPCHILDREN;
   }                                                           

//----------------------------

   bool CheckFreeTurn(S_quat &q){

      S_turn_hlp th;
      th.center = model->GetWorldPos();
      th.turn_is_free = true;
      th.m_rot = S_matrix(q);
      th.move_dir = model->GetDir();
      th.scene = mission.GetScene();
      th.model = model;

      //model->EnumFrames(cbCheckTurnCol, (dword)&th, ENUMF_VOLUME, NULL);
      if(col_parent)
         col_parent->EnumFrames(cbCheckTurnCol, (dword)&th, ENUMF_VOLUME, NULL);
      return th.turn_is_free;
   }

//----------------------------

   void SetMobilDir(S_vector &dir, S_vector &up){

      const S_matrix &m = model->GetFrameMatrix();
      const S_vector save_dir = m(2);
      const S_vector save_up = m(1);
      model->SetDir1(dir, up);
      model->Update();

      S_vector wheel_pos[4];
      for(int i = 0; i<4; i++){
         assert(wheels[i].joint);
         const S_matrix &m = wheels[i].joint->GetMatrix();
         wheel_pos[i] = m(3);
         wheel_pos[i].y = wheels[i].world_y;
      }

      S_vector front_dir = wheel_pos[1] - wheel_pos[0];
      S_vector back_dir = wheel_pos[3] - wheel_pos[2];
      S_vector front_center = wheel_pos[0] + front_dir *.5f;
      S_vector back_center = wheel_pos[2] + back_dir *.5f;

#ifdef DRAW_AUTO_DIR_TEST
      DebugLine(wheel_pos[0], wheel_pos[1], 0, 0xffaaaaaa);
      DebugLine(wheel_pos[2], wheel_pos[3], 0, 0xffdddddd);
      DebugLine(front_center, back_center, 0, 0xffaaaaff);
#endif
                              //test axis
      bool dir_collision(false);
      {
         I3D_collision_data cd_front(wheel_pos[0], front_dir, I3DCOL_FORCE2SIDE, model, MobilTurnColResp);
         dir_collision |= mission.GetScene()->TestCollision(cd_front);
      }
      if(!dir_collision){
         I3D_collision_data cd_back(wheel_pos[2], back_dir, I3DCOL_FORCE2SIDE, model, MobilTurnColResp);
         dir_collision |= mission.GetScene()->TestCollision(cd_back);
      }
      if(!dir_collision){
         I3D_collision_data cd_center(front_center, back_center - front_center, I3DCOL_FORCE2SIDE, model, MobilTurnColResp);
         dir_collision |= mission.GetScene()->TestCollision(cd_center);
      }
      if(dir_collision){
                              //undo set dir
         model->SetDir1(save_dir, save_up);
         model->Update();
      }
   }

//----------------------------

   void MoveMobil(float tsec, const S_vector &auto_dir, const S_vector *w_from, bool net_mode, byte net_route_bits){

      int i;
                           //move mobil
      S_mobile_context cs;
      const S_vector &curr_pos = model->GetPos();
      const S_vector old_pos = curr_pos;
      cs.auto_dir = auto_dir;
      cs.this_actor = this;
      S_vector new_pos = curr_pos + cs.auto_dir;

      cs.mobil_boat = (tab->ItemE(TAB_E_AUTO_TYPE)==1);
                           //test collisions of car body
      cs.coll_wall = false;

                           //set ideal position
      model->SetPos(new_pos); 
      model->Update();

                           //check all wheels independently
      S_vector wheel_to[4];
      static const float  fall_base[2] = {.85f, .65f};
      static const float fall_accel[2] = {13.0, 18.0};
      //AUTO_DEBUG(C_fstring("tsec: %.4f", tsec));

      float new_speed = speed;
      bool any_wheel_floor(false);

      cs.get_wall_info = false;
      for(i=0; i<4; i++){
         cs.boat_water_coll = false;
         cs.mobil_kill_mask = 0;

         float &fall_spd = wheels[i].fall_speed;
         float fall_y = Min(tsec * fall_spd, .3f);
         //float fall_y = tsec * fall_spd;

         I3D_collision_data cd;
         cd.from = w_from[i];

         assert(wheels[i].joint);
         wheel_to[i] = wheels[i].joint->GetMatrix()(3) + S_vector(0.0, -fall_y, 0.0);
         cd.dir = wheel_to[i] - w_from[i];

         cd.radius = wheels[i].volume_radius;
         cd.flags = I3DCOL_SPHERE;
         cd.frm_ignored = model;
         cd.cresp_face = MobilColResp;
         cd.cresp_context = &cs;
         bool wheel_collide(false);
#ifdef RIDE_SMALL_BARRIER
         {
            static const S_vector up_help(.0f, RIDE_STEP_SIZE, .0f); //size of y step which help ride over terrain obstacles and also prevent wheels from been pushed into ground
            cd.from += up_help;
            //DebugLine(cd.from, cd.from + cd.dir, 0, 0xffffafff);
            wheel_collide = mission.GetScene()->TestCollision(cd);
            S_vector part_to = cd.GetDestinationPos();
            cd.Reset();
            cd.from = part_to;
            cd.dir = -up_help;
            cd.radius = wheels[i].volume_radius;
            cd.flags = I3DCOL_SPHERE;
            cd.frm_ignored = model;
            cd.cresp_face = MobilColResp;
            cd.cresp_context = &cs;
            //DebugLine(cd.from, cd.from + cd.dir, 0, 0xffff0f0f);
            wheel_collide |= mission.GetScene()->TestCollision(cd);
         }
#else
         wheel_collide = mission.GetScene()->TestCollision(cd);
#endif
         //DebugLine(cd.from, cd.from + cd.dir, 0, 0xffff00ff);
         //DebugLine(cd.from, cd.GetDestinationPos(), 0, 0xff0fffff);
         any_wheel_floor |= cs.coll_floor;
         if(wheel_collide)
         {
            wheel_to[i] = cd.GetDestinationPos();
            //DebugLine(cd.from, cd.GetDestinationPos(), 0, 0xff00ff00);

            if(fall_spd>0.0f) 
               fall_spd = fall_base[i/2]-fall_spd*.05f;
            else 
               fall_spd = 0.0f;

            if(cs.mobil_boat){
               if(!cs.boat_water_coll)
                  new_speed = Max(0.0f, new_speed - tsec*10.0f);
            }else{
               switch(cs.mobil_kill_mask){
               case 2:        //water
                  new_speed = Max(0.0f, new_speed - tsec*0.5f);
                  fuel=0.0f;
                  if(cs.col_mat_id == AUTO_KILL_WATER_MAT)
                     DrownVehicleUsers();
                  Destroy(false, true, 1, net_route_bits);                              
                  break;
               case 1:        //spasm
               case 3:        //minefield
                  {
                     S_explosion_init ei;
                     ei.pos = cd.from;
                     ei.id = EXPLOSION_MINE;
                     mission.CreateActor(ACTOR_EXPLOSION, (dword)&ei);
                  }
                  break;
               }
            }
         }
         else{
            wheel_to[i] = cd.from + cd.dir;
            fall_spd += tsec * fall_accel[i/2];
         }
         wheels[i].world_y = wheel_to[i].y;
      }

      AUTO_DEBUG(C_fstring("any_wheel_floor %i",any_wheel_floor));
      if(!any_wheel_floor){
         #define AIR_FRICTION .05f
                              //scale speed in the air
         new_speed -= (new_speed * (AIR_FRICTION*tsec));
      }

                           //find center of our back axle (= mobil pos)
      S_vector front_dir = wheel_to[1]-wheel_to[0];
      S_vector back_dir = wheel_to[3]-wheel_to[2];

      float u_x, u_z;      //ranges: <-.5f, +.5f>
      u_x = .0f;
      u_z = .3f;
      S_vector front_center = wheel_to[0] + front_dir * (.5f+u_x);
      S_vector back_center = wheel_to[2] + back_dir * (.5f+u_x);

      S_vector dir = front_center - back_center;
      S_vector car_center = back_center + dir*(.5f+u_z);

      S_vector car_pos = car_center;
      {                    //adjust x
         float bdir_len = back_dir.Magnitude();
         if(!IsAbsMrgZero(bdir_len))
            car_pos -= back_dir/bdir_len*back_axle_len * u_x;
      }
      {                    //adjust z
         float dir_len = dir.Magnitude();
         if(!IsAbsMrgZero(dir_len))
            car_pos -= dir/dir_len*axes_dist * (.5f+u_z);
      }
      model->SetPos(car_pos);
                           //get average of both axles
      S_vector v = back_dir + front_dir;
      v.Normalize();

      S_vector up = dir.Cross(v);
      SetMobilDir(dir, up);

#ifndef DISABLE_CAR_BODY_COL

      cs.get_wall_info = true;
                              //Test car body
      if(col_parent)
      {
         //float test_len = (car_pos - old_pos).Magnitude();
         //DEBUG(C_fstring("%.8f",test_len));
         S_vector col_delta = col_parent->GetWorldPos()-car_pos;
         I3D_collision_data cd;
         cd.from = old_pos + col_delta;
         cd.dir = car_pos - old_pos;
         //DebugLine(cd.from, cd.from + cd.dir, 0, 0xff00ffff);
         cd.frm_root = col_parent ? col_parent : model;
         cd.flags = I3DCOL_SPHERE_GROUP;
         cd.frm_ignored = model;
         cd.cresp_face = MobilColResp;
         cd.cresp_context = &cs;
         if(mission.GetScene()->TestCollision(cd)){
            new_pos = cd.GetDestinationPos();
            model->SetPos(new_pos-col_delta);
         }
      }
#endif

      auto_falling = !(any_wheel_floor || cs.coll_floor);
      AUTO_DEBUG(C_fstring("auto_falling %i",auto_falling));

      model->Update();

#ifdef ENABLE_AUTO_FALL
      //DEBUG(C_fstring("fall_speed:%.2f", fall_speed));
      if(auto_falling){
         static const float FALL_ACCEL(40.0f);
         float fall_add = tsec*FALL_ACCEL*.4;
         //fall_speed += tsec*FALL_ACCEL*.4;
                              //Test car body fall
         {
            I3D_collision_data cd;
            cd.from = model->GetWorldPos();
            cd.dir = S_vector(.0f, -(fall_speed+fall_add)*tsec, .0f);
            //DebugLine(cd.from, cd.from + cd.dir, 0, 0xff00ffff);
            cd.frm_root = model;
            cd.flags = I3DCOL_SPHERE_GROUP;
            cd.frm_ignored = model;
            cd.cresp_face = MobilColResp;
            cd.cresp_context = &cs;
            mission.GetScene()->TestCollision(cd);
            model->SetPos(cd.GetDestinationPos());
            model->Update();
                              //recompute fall speed by falled distance
            float falled_dist = (cd.from - cd.GetDestinationPos()).Magnitude();
            float ratio = IsAbsMrgZero(cd.dir.y) ? .0f : fabs(falled_dist / cd.dir.y);
            //DEBUG(C_fstring("ratio:%.2f", ratio));
            fall_speed += (fall_add*ratio);
         }
      }else{
         if(fall_speed > AUTO_CRASH_FALL_SPEED)
            HitExplode(this, 4000, NULL, true, 1, net_route_bits);
         fall_speed = .0f;
      }
#endif

      mission.GetScene()->SetFrameSector(model);

#ifndef DISABLE_WHEEL_SPRING
      PI3D_frame f_inv = NULL;
      S_matrix m_inv;
      for(i=0; i<4; i++){
         PI3D_frame wheel = wheels[i].joint;
         if(f_inv!=wheel->GetParent()){
            f_inv=wheel->GetParent();
            m_inv = f_inv->GetMatrix();
            m_inv.Inverse();
         }
         const float &y = wheels[i].world_y;

         S_vector pos = wheel->GetPos();
         float new_y =
            wheel_to[i].x*m_inv(0, 1) +
            y*m_inv(1, 1) +
            wheel_to[i].z*m_inv(2, 1) +
            m_inv(3, 1);
         float delta = new_y-wheels[i].base_loc_y;
         delta = Min(MAX_WHEEL_SPRING, Max(-MAX_WHEEL_SPRING, delta));
         pos.y = wheels[i].base_loc_y + delta;

         wheel->SetPos(pos);
         wheel->Update();

         wheels[i].world_y = wheel->GetWorldPos().y;
      }
#endif //DISABLE_WHEEL_SPRING

#ifdef AUTO_DRAW_AXIS
                           //draw proposed axles
      DebugLine(wheel_to[0], wheel_to[1], 1, 0xffff0000);
      DebugLine(wheel_to[2], wheel_to[3], 1, 0xffff0000);
                           //draw exles centers connection
      DebugLine(back_center, front_center, 1, 0xff00ff00);
                           //draw up vector
      DebugLine(car_center, car_center+up, 1, 0xff0000ff);
      DebugLine(car_pos, car_pos+up, 1, 0xff00ff80);
#endif

      if(cs.coll_wall){

         S_wall_help wh;
         wh.wall_slide = cs.coll_wall_slide;
         wh.wall_vol = cs.wall_vol;
         wh.time = tsec*1000;
         WallHit(wh, new_speed, net_route_bits);
      }else{
         S_normal dir_n(dir);
         float y_ratio = dir_n.y * ((fabs(dir.y) > .65f) ? 1.8f : 1.4f);
                              //adjust speed by slope
         if((dir.y > 0.0f) && (speed > 0.0f)){
            //new_speed *= Max(.0f, (1.0f - tsec*dir_n.y*.7f));
            new_speed *= Max(.0f, (1.0f - tsec*y_ratio*.7f));
         }else
         if((dir.y < 0.0f) && (speed < 0.0f)){
            //new_speed *= Max(.0f, (1.0f + tsec*dir_n.y*.7f));
            new_speed *= Max(.0f, (1.0f + tsec*y_ratio*.7f));
         }
      }
      if(speed!=new_speed)
         SetSpeed(new_speed, !net_mode, 1, net_route_bits);
   }

//----------------------------

   inline void WallHit(const S_wall_help &wh, float &new_speed, dword net_route_bits){

      if(fabs(speed) < HIT_STOP_SPEED){
         new_speed =.0f;
         return;
      }

      float rel_speed_loss = Min(1.0f, SPEED_LOSS_MIN + (SPEED_LOSS_MAX)*(1.0f-wh.wall_slide));
      float speed_loss = fabs(new_speed) * rel_speed_loss;
                        //find actor whom belongs the wall volume (if any)
      LPC_actor act = NULL;
      PI3D_frame frm = wh.wall_vol;
      while(frm=frm->GetParent()){
         act = (LPC_actor)frm->GetData();
         if(act)
            break;
      }
                              //hit actor
      float resist = 1.0f;
      if(act){
                              //compute hit power (range <0,1>)
         //float power = Min(1.0f, ((float)fabs(speed)-HIT_STOP_SPEED)/(HIT_DEADLY_SPEED-HIT_STOP_SPEED));
         float power = ((float)fabs(speed)-HIT_HURT_SPEED_MIN)/(HIT_HURT_SPEED_MAX);
                           //hurt player soflty - he get hit every frame
         if((act->GetType() == ACTOR_PLAYER) || (act->GetType() == ACTOR_ENEMY))
            power *= ((float)wh.time/PLR_HIT_FATAL_TIME);

         //power = Max(.005f, Min(1.0f, power));
         power = Max(.0f, Min(1.0f, power));
                              //return value of Crash is how much actor rezist to hit - heaviest for more slowdown,or .0f for deault
#ifdef _DEBUG
         //OUTTEXT(power);
#endif
         if(power){
            resist = act->Crash(this, power);
            if(!resist) resist = 1.0f; //default
            speed_loss *= resist;
         }else{
            speed_loss = .0f;
         }
         if((float)fabs(speed) < (1.2f*HIT_HURT_SPEED_MIN)){
            speed_loss += (HIT_HURT_SPEED_MIN * .75f);
         }
      }

      AUTO_DEBUG(C_fstring("Wallhit: speed_loss %.2f, rel_speed_loss %.2f, slide %.2f, speed %.2f, new_speed %.2f",
         speed_loss, rel_speed_loss, wh.wall_slide, speed, new_speed));

      if(mode != MODE_DESTROYED){
         ApplyHit(new_speed, wh, resist, net_route_bits);
      }

      if(new_speed>0.0f)
         new_speed = Max(.0f, new_speed - speed_loss);
      else
         new_speed = Min(new_speed + speed_loss, .0f);
   }

//----------------------------

   inline void ApplyHit(float curr_speed, const S_wall_help &wh, float resist, dword net_route_bits){

      int wall_mat(0);
      assert(wh.wall_vol);
      switch(wh.wall_vol->GetType()){
      case FRAME_VOLUME:
         wall_mat = I3DCAST_VOLUME(wh.wall_vol)->GetCollisionMaterial(); break;
      case FRAME_VISUAL:
         wall_mat = I3DCAST_VISUAL(wh.wall_vol)->GetCollisionMaterial(); break;
      }

      if(!mat_table->ItemB(TAB_B_MAT_BORDER, wall_mat)){
                              //convert to km/h and scale by wallslide 
         curr_speed *= 3.6 * (1.0f-wh.wall_slide);
                              //scale by actor resist
         curr_speed *= resist;
                              //compute crash power
         float explode_ratio = tab->ItemI(TAB_I_AUTO_EXPL_THRESH) / DEFAULT_EXPLODE_POWER;
         float destroy_ratio = tab->ItemI(TAB_I_AUTO_DESTROY_THRESH) / DEFAULT_DESTROY_POWER;
         float destroy_speed = DESTROY_SPEED * destroy_ratio;
         float explode_speed = EXPLODE_SPEED * explode_ratio;

         AUTO_DEBUG_LOG(C_fstring("Crash: speed %.2f, destroy spd %.2f (ratio %.2f), explode spd %.2f (ratio %.2f)",
            curr_speed, destroy_speed, destroy_ratio, explode_speed, explode_ratio));
         if(curr_speed > destroy_speed){  //destroy or explode
            if(curr_speed > explode_speed){ //explode!
               HitExplode(this, 4000, NULL, true, 1, net_route_bits);
            }else{
               Destroy(false, true, 1, net_route_bits);                              
            }
         }
         /*if(crash_power > tab->ItemI(TAB_I_AUTO_DESTROY_THRESH)){      //destroy or crash
            if(crash_power > tab->ItemI(TAB_I_AUTO_EXPL_THRESH)){      //crash! explode!
               //HitExplode(this, 2000, NULL, true, net_route_code, net_route_bits);
               HitExplode(this, 4000, NULL, true, net_route_code, net_route_bits);
            }else{          //only destroy
               Destroy(false, true, net_route_code, net_route_bits);
            }
         }*/
      }
   }

//----------------------------
public:
   C_automobil(C_game_mission &m, E_ACTOR_TYPE at = ACTOR_AUTOMOBIL):
      C_version(m, at),
      mode(MODE_STAY),
      mode_count(0),
      num_of_seats(0),
      fuel(0.0f),
      net_fuel(0.0f),
      speed(0.0f),
      airplane_time(-1),
      net_speed(0.0f),
      eng_turn(0),
      snd_engine(NULL),
      col_parent(NULL),
      wheel_turn(0.0f),
      net_pos_count(0),
      net_wheel_turn(0.0f),
      transl(0),
      proposed_forward(false),
      volant(NULL),
      auto_falling(false),
      fall_speed(.0f),
      floor_test_cnt(0),
      emit_snd_cnt(0),
      airplane_speed(.0f),
      net_route_bits(1),
      tab(CreateTable())
   {
      memset(seats, 0, sizeof(seats));
      memset(wheels, 0, sizeof(wheels));

      tab->Open((dword)&templ_automobil, TABOPEN_TEMPLATE);
#ifdef EDITOR
      tab->ItemI(TAB_I_VER_RESISTANCE_SET) = 4;  //default resistance set
#endif
   }

   ~C_automobil(){
      if(fuel_icon)
         mission.DelIconInfo(fuel_icon);

      if(snd_engine){
         snd_engine->SetOn(false);
         snd_engine->Release();
      }
      if(tab) tab->Release();
   }

//----------------------------

   virtual void SetNetID(word w){

      for(int i=num_of_seats; i--; ){
                              //encode following into the sync class:
                              // bits:net_route_bits route_code (=1)
                              // bits:4 C_vehicle code (=0 -> seat sync)
                              // bits:3 seat_index
         seats[i].net_sync = CreateNetSync(w, 1 | (0 | (i<<4) << net_route_bits), 4+3+net_route_bits);
         seats[i].net_sync->Release();
      }
      C_actor::SetNetID(w);
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      C_version::SetFrame(frm);
                              //try to stop animation - for case of animated plane
      if(frm->GetType()!=FRAME_MODEL){
         PI3D_frame mp1 = frm->GetParent();
         if(mp1 && mp1->GetType()==FRAME_MODEL){
            I3DCAST_MODEL(mp1)->Tick(0); //airplanes can have some parts hidden by animation
            I3DCAST_MODEL(mp1)->StopAnimation(0);
         }
         last_airplane_pos = frm->GetWorldPos();
      }
                        //find wheels
      bool warn = false;
      for(int i=0; i<4; i++){
         PI3D_frame w;
         w = frm->FindChildFrame(C_fstring("wheel_%.2i", i));
         if(!w) warn=true;
         else{
            //w = w->FindChildFrame(NULL, ENUMF_VOLUME);
         }
         wheels[i].wheel = w;

         PI3D_frame &j = wheels[i].joint;
         j = frm->FindChildFrame(C_fstring("joint_%.2i", i), ENUMF_DUMMY);
         if(j){
            wheels[i].joint_base_rot = j->GetRot();
            wheels[i].SetWorldY();
            wheels[i].base_loc_y = j->GetPos().y;
            //j->LinkTo(mp, I3DLINK_UPDMATRIX);

            PI3D_frame vol = j->FindChildFrame(NULL, ENUMF_VOLUME);
            if(vol){
               wheels[i].volume_radius = vol->GetMatrix()(0).Magnitude();
            }else{
               ReportActorError(C_fstring("can't find wheel volume %s", (const char*)frm->GetName()));
            }
         }else warn=true;
      }
      if(warn){
         //OUTTEXT(C_fstring("can't find all wheels/joints in %s", mp->GetName()));
      }else{
         S_vector v0, v1;
         v0 = wheels[0].joint->GetPos();
         v1 = wheels[1].joint->GetPos();
         S_vector mid0 = v0 + (v1-v0)*.5;
         v0 = wheels[2].joint->GetPos();
         v1 = wheels[3].joint->GetPos();
         back_axle_len = (v1-v0).Magnitude();
         S_vector mid1 = v0 + (v1-v0)*.5;
         axes_dist = (mid1 - mid0).Magnitude();
      }
      volant = frm->FindChildFrame("volant");
      if(volant){
         volant_base_rot = volant->GetRot();
      }
                        //find all seats
      for(num_of_seats=0; num_of_seats<MAX_AUTO_SEATS; num_of_seats++){
         PI3D_frame frm_seat = frm->FindChildFrame(C_fstring("seat_%.2i", num_of_seats), ENUMF_DUMMY);
         if(!frm_seat)
            break;
         S_seat &seat = seats[num_of_seats];
         seat.seat = frm_seat;
         seat.driver = (num_of_seats==0);
         PI3D_frame entry = frm_seat->GetParent();
         if(entry->GetType()==FRAME_DUMMY){
            seat.entry = entry;
                        //unlink, so that it's not linked to version
            entry->LinkTo(frm, I3DLINK_UPDMATRIX);
         }else{
            ReportActorError(C_fstring("can't find entry for seat %s", (const char*)frm_seat->GetName()));
         }
         //seats[num_of_seats].type = !num_of_seats ? 0 : 1;
      }
      //if(!num_of_seats) OUTTEXT(C_fstring("no seats in %s", mp->GetName()));
      engine_dummy = frm->FindChildFrame("engine", ENUMF_DUMMY);
      if(engine_dummy){
         engine_dummy->LinkTo(frm, I3DLINK_UPDMATRIX);
         //engine->SetData((dword)this);
         last_engine_scale = engine_dummy->GetScale();
      }else{
         //OUTTEXT(C_fstring("can't find engine in %s", mp->GetName()));
         last_engine_scale = 1.0f;
      }

                              //set all VBs to be seen through
                              // (not planes)
      if(model){                 
         struct S_hlp{
            static I3DENUMRET I3DAPI cbEnum(PI3D_frame vol, dword c){
               assert(mat_table->ItemE(TAB_B_MAT_CAM_SEE_THROUGH, 35)==1);
               I3DCAST_VOLUME(vol)->SetCollisionMaterial(35);
               return I3DENUMRET_OK;
            }
         };
         frm->EnumFrames(S_hlp::cbEnum, 0, ENUMF_VOLUME);
         col_parent = frm->FindChildFrame("col");
         if(!col_parent){
                              //car should have col
            if(frm && frm->GetType()==FRAME_MODEL){
               ReportActorError("cannot find dummy object 'col'");
            }
         }
      }
      ParticleOn(false, frm);

      //prm2 = 1;           //disable complains from C_version

      net_last_sent_pos = frm->GetPos();
      net_last_sent_time = igraph->ReadTimer();
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      SG_SaveModelInfo(ck, false);
                              //let version save first
      C_version::SaveGame(ck);
                           //write data
      ck->Write(&mode, sizeof(mode));
      ck->Write(&mode_count, sizeof(mode_count));
      ck->Write(&fuel, sizeof(fuel));
      ck->Write(&speed, sizeof(speed));
      ck->Write(&transl, sizeof(transl));
      ck->Write(&eng_turn, sizeof(eng_turn));
      ck->Write(&wheel_turn, sizeof(wheel_turn));
      ck->Write(&proposed_forward, sizeof(proposed_forward));
                           //write seats
      for(int i=0; i<num_of_seats; i++){
         const S_seat *sp = seats+i;
         ck->WStringChunk(CT_NAME, sp->user ? sp->user->GetName() : "");
      }
                           //write wheels
      for(i=0; i<4; i++){
         const S_wheel &w = wheels[i];
         ck->Write(&w.fall_speed, sizeof(float));
         ck->Write(&w.world_y, sizeof(float));
      }
                           //write airplane anim
      ck->Write(&airplane_time, sizeof(airplane_time));
                              //write fall speed
      ck->Write(&fall_speed, sizeof(fall_speed));
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      fall_speed = .0f;
      floor_test_cnt = 0;
      SG_LoadModelInfo(ck, true);
                              //let version load first
      if(!C_version::LoadGame(ck))
         return false;
                           //read data
      ck->Read(&mode, sizeof(mode));
      ck->Read(&mode_count, sizeof(mode_count));
      ck->Read(&fuel, sizeof(fuel));
      ck->Read(&speed, sizeof(speed));
      ck->Read(&transl, sizeof(transl));
      ck->Read(&eng_turn, sizeof(eng_turn));
      ck->Read(&wheel_turn, sizeof(wheel_turn));
      ck->Read(&proposed_forward, sizeof(proposed_forward));

      SetEngTurn(eng_turn);

      switch(mode){
      case MODE_GO:
      case MODE_START:
         if(snd_engine) snd_engine->SetOn(true);
         break;
      }

                           //read seats
      for(int i=0; i<num_of_seats; i++){
         S_seat *sp = seats+i;
         C_string str;
         ck->ReadString(CT_NAME, str);
         sp->user = !str.Size() ? NULL : mission.FindActor(str);
      }
                           //read wheels
      for(i=0; i<4; i++){
         S_wheel &w = wheels[i];
         ck->Read(&w.fall_speed, sizeof(float));
         ck->Read(&w.world_y, sizeof(float));
      }
                           //read airplane anim
      ck->Read(&airplane_time, sizeof(airplane_time));
                              //read fall_speed
      ck->Read(&fall_speed, sizeof(fall_speed));

      if(airplane_time!=-1){
         dword save_time = airplane_time;
         AirplaneStart(false);
         airplane_time = save_time;
         PI3D_frame mp = frame;
         if(mp->GetType()!=FRAME_MODEL) mp = mp->GetParent();
         if(mp && mp->GetType()==FRAME_MODEL){
            PI3D_model model = I3DCAST_MODEL(mp);
            model->Tick(airplane_time);
         }
      }
      if(mode==MODE_GO){
         InitFuelIcon();
         ShowFuel();
      }
      airplane_speed = .0f;
      mission.SetFrameSector(frame);
      return true;
   }

//----------------------------

   virtual float Crash(C_actor *who, float power){

      if(power>.5f)
         return Explode(NULL, 0);
      return C_actor::Crash(who, power);
   }

//----------------------------

   virtual const S_vector GetUsePos(int indx, bool abs_index_mode = false) const{

      const S_seat *sp = GetFreeSeatConst(indx, abs_index_mode);
      if(!sp && abs_index_mode){
         sp = &seats[indx];
      }
      if(sp && sp->seat){
         return (sp->entry ? sp->entry : sp->seat)->GetWorldPos();
      }
      return S_vector(0, 0, 0);
   }

//----------------------------

   virtual dword CheckUseStay(dword suggest_stay_mode, int indx) const{
      return SM_STAY;
   }

//----------------------------

   virtual bool CheckUseFreeHands(int indx, bool abs_index_mode) const{

      if(!abs_index_mode){
         for(int i=0; i<MAX_AUTO_SEATS; i++){
            if(seats[i].IsFree() && !indx--)
               break;
         }
         indx = (i < MAX_AUTO_SEATS) ? i : indx;
      }

                           //driver must always be off weapon
      if(!indx && !seats[0].user)
         return true;
      if(!seats[indx].user && !tab->ItemB(TAB_B_AUTO_SEAT_WEAPON, indx))
         return true;
      if(!seats[indx].user && tab->ItemB(TAB_B_AUTO_SEAT_CANNON, indx))
         return true;
      return false;
   }

//----------------------------

   dword C_automobil::cbProc(dword msg, dword prm1, dword prm2){

      switch(msg){

      case CB_SIGNAL:
         {
            S_signal_data *sd = (S_signal_data*)prm1;
            if(sd->type==SIGNAL_DETECTOR && sd->sub_type==999){
               AirplaneStart();
            }
         }
         break;

      case CB_HIT:
         if(!curr_version){
            HitExplode(NULL, prm2, (S_CB_hit*)prm1, true, 1, net_route_bits);
            return 0;
         }
         break;

      case CB_RESERVE_USE:
         return ReserveSeat((LPC_actor)prm1, prm2, true);

      case CB_USE_AUTO:
         return UseAuto(prm1, prm2, 1);

      }
      return C_version::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual bool Explode(C_actor *caller, int power){

      if(!curr_version){
         HitExplode(caller, power, NULL, true, 1, net_route_bits);
         if(curr_version){
                              //explode all users
            for(int i=num_of_seats; i--; ){
               if(seats[i].user)
                  seats[i].user->Explode(caller, power);
            }
         }
         return (curr_version!=0);
      }
      return false;
   }

//----------------------------

   void DrownVehicleUsers(){
                              //test every alive user if has head under water 
      for(int i=num_of_seats; i--; ){
         if(seats[i].user && seats[i].user->IsAlive()){
            seats[i].user->cbProc(CB_DIE_DROWN);
         }
      }
   }

//----------------------------

   dword Tick1(int time, byte net_route_bits);

//----------------------------

   dword Tick(const S_tick_context &tc){

      return Tick1(tc.time, 1);
   }

//----------------------------

   dword UseAuto(dword prm1, dword prm2, byte net_route_bits){

      int i;

      switch(prm1){
      case 1:              //accelerate
         if(!curr_version && (airplane_time==-1) && model){
            if(proposed_forward){
               if(mode==MODE_DESTROYED){
                  SlowDown(prm2);
               }else
               if(mode!=MODE_GO){
                  MotorOn(true, true, 1, net_route_bits);
               }else
               if(fuel){
                  i = Min(tab->ItemI(TAB_I_AUTO_MAX_MOTOR_TURN), (int)(eng_turn + prm2*.001f * TURN_SPEED));
                  eng_turn=i;
                  if(eng_turn >= tab->ItemI(TAB_I_AUTO_TRANS_UP))
                     if(transl < tab->ItemI(TAB_I_AUTO_NUM_TRANSL))
                        Translate(1);
               }else
                  GasDown(prm2);
            }else{
               SetSpeed(Min(0.0f, speed + tab->ItemF(TAB_F_AUTO_BRAKE)*prm2*.001f), true, 1, net_route_bits);
               if(eng_turn < tab->ItemI(TAB_I_AUTO_TRANS_DOWN) && transl)
                  Translate(1);
               if(!speed){
                  transl=0;
                  proposed_forward = true;
               }
            }
         }
         break;

      case 2:              //brake
         if(!curr_version && (airplane_time==-1) && model){

            if(proposed_forward){
               SetSpeed(Max(0.0f, speed - tab->ItemF(TAB_F_AUTO_BRAKE)*prm2*.001f), true, 1, net_route_bits);
               if(eng_turn < tab->ItemI(TAB_I_AUTO_TRANS_DOWN) && transl)
                  Translate(-1);
               if(!speed)
                  proposed_forward=false;
            }else{
               if(mode==MODE_DESTROYED){
                  SlowDown(prm2);
               }else
               if(mode!=MODE_GO){
                  MotorOn(true, true, 1, net_route_bits);
               }else
               if(fuel){
                  i = Min(tab->ItemI(TAB_I_AUTO_MAX_MOTOR_TURN), (int)(eng_turn + prm2*.001f * TURN_SPEED));
                  eng_turn=i;
                  if(transl>0)
                     transl=0; 
                  if(eng_turn >= tab->ItemI(TAB_I_AUTO_TRANS_UP))
                  if(!transl)
                     Translate(-1);
               }else
                  GasDown(prm2);
            }
         }
         break;

      case 4:                 //turn wheels
         {
            float f = IntAsFloat(prm2);
            SetWheelTurn(f*.6f, true, 1, net_route_bits);
         }
         break;

      case 5:                 //return seat type
         {
            S_seat *sp = GetUserSeat((LPC_actor)prm2);
            if(sp)
               return tab->ItemE(TAB_E_AUTO_SEAT_STAY, sp-seats);
         }
         return -1;

      case 6:                 //get seat frame
         {
            S_seat *sp = GetUserSeat((LPC_actor)prm2);
            if(sp)
               return (dword)sp->seat;
         }
         return NULL;

      case 7:                 //get seat entry pos
         {
            S_seat *sp = GetUserSeat((LPC_actor)prm2);
            if(sp)
               return (dword)&(sp->entry ? sp->entry : sp->seat)->GetWorldPos();
         }
         return NULL;

      case 8:                 //get speed
         return (int)(speed*3.6f);

      case 10:                //get driver
         return (dword)seats[0].user;

      case 11:                //get cannon status
         {
            if(type!=ACTOR_AUTO_CANNON)
               return false;
            S_seat *sp = GetUserSeat((LPC_actor)prm2);
            if(sp)
               return tab->ItemE(TAB_B_AUTO_SEAT_CANNON, sp-seats);
         }
         return false;

      case 12:                //get wheels angle (local)
         return FloatAsInt(wheel_turn * tab->ItemI(TAB_I_AUTO_MAX_WHEEL_TURN)*GetPI()/180.0f);

      case 13:             //stop immediately
         MotorOn(false, true, 1, net_route_bits);
         SetSpeed(0, true, 1, net_route_bits);
         break;

      case 14:                //check if weapons allowed
         {
            S_seat *sp = GetUserSeat((LPC_actor)prm2);
            if(sp)
               return tab->ItemE(TAB_B_AUTO_SEAT_WEAPON, sp-seats);
         }
         break;

      case 15:                //check if actor is in
         {
            S_seat *sp = GetUserSeat((LPC_actor)prm2);
            return bool(sp);
         }
         break;

      case 16:
         return -1;
      }
      return 0;
   }

//----------------------------
                              //net_actor: yes

   int ReserveSeat(LPC_actor r, int res_time, bool net_send){
                              //find non-reserved seat
      if(!r)
         return -1;
      if(!res_time)
         res_time = 4000;
      int i;
      for(i=0; i<num_of_seats; i++)
      if(!seats[i].user){
                              //already reserved?
         if(seats[i].reserver==r) return i;
         if(!seats[i].reserver){
            seats[i].reserver = r;
            seats[i].reserve_count = res_time;
                           //send reservation over net
            if(net && net_send){
               const bool g = true;
               net_msg->Ascend(net_id);
               net_msg->AddBits(1, net_route_bits, g);
               net_msg->AddBits(10, 4, g);      //code
               net_msg->AddBits(i, 3, g);       //seat index
               net_msg->AddWord(r->GetNetID(), g);   //reserver
               net_msg->AddWord(res_time, g);   //time
               net_msg->Descend();
            }
            return i;
         }
      }
      return -1;              //no free seat, sorry
   }

//----------------------------

   virtual void NetCodeIn(byte msg_code, PC_net_in_msg in_msg){

      switch(msg_code){
      case 0:              //seat sync
         {
            int seat_indx = in_msg->GetBits(3);
            NET_LOG(C_fstring("Message for seat %i", seat_indx));
            S_seat *sp = &seats[seat_indx];
            sp->net_sync->NetIn(in_msg, false);
         }
         break;

      case 1:              //seat user
         {
            int seat_indx = in_msg->GetBits(3);
            S_seat *sp = &seats[seat_indx];

            LPC_actor user1 = mission.NetFindActor(in_msg->GetWord());
            NET_LOG(C_fstring("C_automobil: seat %i - network user: %s", seat_indx, user1 ? user1->GetName() : "<NULL>"));
            if(user1){
               sp->user = user1;
               sp->user->cbProc(CB_USE_AUTO, 0, (dword)this);
               //sp->net_sync->Reset();
            }else{
               if(sp->user){
                  sp->user->cbProc(CB_USE_AUTO, 0, NULL);
                  sp->user = NULL;
               }
            }
         }
         break;

      case 2:              //motor on
         MotorOn(in_msg->GetBits(1), false);
         break;

      case 3:              //wheel turn
         SetWheelTurn(in_msg->GetFloat(), false);
         break;

      case 4:              //fuel
         SetFuel(in_msg->GetFloat(), false);
         break;

      case 5:              //speed
         SetSpeed((short)in_msg->GetWord(), false);
         break;

      case 8:              //pos + rot
         {
            model->SetPos(in_msg->GetVector());
            S_quat q;
            q.v = in_msg->GetVector();
            q.s = in_msg->GetFloat();
            model->SetRot(q);
            model->Update();
            for(int i=0; i<4; i++) wheels[i].SetWorldY();
            mission.GetScene()->SetFrameSector(model);
         }
         break;

      case 9:
         {
            bool exploded = in_msg->GetBits(1);
            /*
            if(exploded)
               SetVersion(in_msg->GetByte());
               */
            Destroy(exploded, false, 0, 0);
         }
         break;

      case 10:                //seat reservation
         {
            int seat_indx = in_msg->GetBits(3);
            seats[seat_indx].reserver = mission.NetFindActor(in_msg->GetWord());
            seats[seat_indx].reserve_count = in_msg->GetWord();
         }
         break;

      case 11:
         AirplaneStart(false);
         break;

      default:
         assert(0);
      }
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      in_msg->Ascend();
      while(in_msg->GetChunkSize()){
         NET_LOG("reading routing code");
         byte route = in_msg->GetBits(1);
         switch(route){
         case 0:              //version
            NET_LOG("message for C_version");
            C_version::NetCodeIn(in_msg->GetByte(), in_msg);
            break;
         case 1:              //automobil
            NET_LOG("message for C_automobil");
            NetCodeIn(in_msg->GetBits(4), in_msg);
            break;
         }
      }
      in_msg->Descend();
   }

//----------------------------

   virtual void NetPlrDisconnect(dword pid){

      for(int i=num_of_seats; i--; ){
         seats[i].net_sync->OnPlayerDisconnect(pid);
         dword winner_pid;
         if(seats[i].net_sync->GetState(&winner_pid) == C_net_sync::SS_RESOLVED){
            if(winner_pid==net->GetPID())
               seats[i].net_sync->Reset();
         }
      }
      C_actor::NetPlrDisconnect(pid);
   }

//----------------------------

   virtual bool IsDestroyed() const{

      bool destr = (C_version::IsDestroyed() || (mode == MODE_DESTROYED));
      return destr;
   }

//----------------------------

   virtual bool IsBoat() const{ return tab ? (tab->ItemE(TAB_E_AUTO_TYPE)==1) : false; }

//----------------------------

   virtual int NumUsePositions(C_actor *asker = NULL) const{

                              //can't use destroyed vehicle
      if(IsDestroyed())
         return 0;
                              //check speed
      if((fabs(speed) > .4f) || (airplane_speed > .4f))
         return 0;

      return NumFreeSeats();
   }

//----------------------------

   virtual int Use(LPC_actor user, dword prm2, bool force_use_off){

      if(prm2){
                              //get use index
         int use_indx = (prm2>>16);
                              //check if we've got free seat
         S_seat *sp = GetFreeSeat(use_indx, prm2&0x8000);
         if(!sp)
            return 0;

         if(!net){
                              //use directly
            sp->user = user;
            user->cbProc(CB_USE_AUTO, 0, (dword)this);
            return 1;
         }else{
            if(sp->net_sync->GetState())
               return 0;
                              //start sync
            sp->net_sync->BeginEvent();         //seat sync
            sp->net_sync_user = user;
            return 2;         //in progress
         }
      }else{
                              //check speed
         if((fabs(speed) > .4f) || (airplane_speed > .4f)){
            if(!force_use_off)
               return 0;
         }
         S_seat *sp = GetUserSeat(user);
         if(sp){
            if(!user->cbProc(CB_USE_AUTO, 0, NULL))
               return 0;
            sp->user = NULL;
            if(sp->driver)
               MotorOn(false, true, 1, net_route_bits);

            if(net){
               const bool g = true;
               net_msg->Ascend(net_id);
                              //tell to all
                              //routing
               net_msg->AddBits(1, net_route_bits, g);
               net_msg->AddBits(1, 4, g);          //code
               net_msg->AddBits(sp-seats, 3, g);   //seat index
               net_msg->AddWord(0, g);             //user_id (=NULL)
               net_msg->Descend();
            }
            return 1;
         }
      }
      return 0;
   }

//----------------------------

   void ComputeAirplaneSpeed(int time){

      float tick_delta = (last_airplane_pos - frame->GetWorldPos()).Magnitude();
      airplane_speed = (time) ? tick_delta / (time*.001f) : .0f;
      last_airplane_pos = frame->GetWorldPos();
      //DEBUG(C_fstring("airplane speed: %.2f", airplane_speed));
   }


//----------------------------

   virtual bool IsOnMap() const{ return true; }

//----------------------------

   virtual int GetTableIndex() const{ return WinSelectItem("Edit:", "Version\0Automobil\0", 1); }

//----------------------------

   virtual LPC_table GetTable(int index) const{
      return (index==1) ? tab : C_version::GetTable(index);
   }
   virtual class C_table_template *GetTemplate(int index) const{
      if(index==1)
         return &templ_automobil;
      return C_version::GetTemplate(index);
   }

//----------------------------

   virtual void TableUpdate(int index, bool not_load_prg){

      OpenSound();
      fuel = tab->ItemF(TAB_F_AUTO_FUEL);
      net_fuel = fuel;
      for(int i=0; i<4; i++){
         if(!tab->ItemB(TAB_B_WHEEL_ON, i) && wheels[i].wheel){
            PI3D_volume vol = (PI3D_volume)wheels[i].wheel->FindChildFrame(NULL, ENUMF_VOLUME);
            if(vol)
               vol->SetOn(false);
         }
      }
                        //prepate texture-shifting
      tx_shift.clear();
      for(i=0; i<tab->ArrayLen(TAB_S16_AUTO_TXT_SHIFT); i++){
         const char *ts = tab->ItemS(TAB_S16_AUTO_TXT_SHIFT, i, 16);
         if(*ts){
            struct S_hlp{
               static I3DENUMRET I3DAPI cbO(PI3D_frame frm, dword c){
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  vector<C_smart_ptr<I3D_visual> > *vp = (vector<C_smart_ptr<I3D_visual> >*)c;
                  vp->push_back(vis);
                  return I3DENUMRET_OK;
               }
            };
            model->EnumFrames(S_hlp::cbO, (dword)&tx_shift, ENUMF_VISUAL | ENUMF_USEMASK, ts);
                              //validate
            for(int i=tx_shift.size(); i--; ){
               if(tx_shift[i]->GetVisualType()!=I3D_VISUAL_UV_SHIFT){
                  ReportActorError(C_fstring("Visual '%s' is not UV-shift object", tx_shift[i]->GetName()));
                  tx_shift[i] = tx_shift.back(); tx_shift.pop_back();
               }
            }
         }
      }
      C_version::TableUpdate(index, not_load_prg);
                              //some tables went corrupted during port, write warning
      if(tab->ItemI(TAB_I_AUTO_NUM_TRANSL)<0)
         ReportActorError("table corrupted, update!");
   }

//----------------------------

   virtual void MissionSave(C_chunk *ck) const{

      C_version::MissionSave(ck);
      SaveTableEx(ck, 1);
   }

//----------------------------

   virtual bool MissionLoad(C_chunk *ck, int table_id){
      switch(table_id){
      case 1:
         return LoadTable(ck, tab);
      default:
         return C_version::MissionLoad(ck, table_id);
      }
   }
};

//----------------------------

dword C_automobil::Tick1(int time, byte net_route_bits){

   int i;

#ifdef _DEBUG
   //DEBUG(speed);
#endif

   AUTO_DEBUG(C_fstring("T: %i, ot: %i, speed: %i, forward: %i, fuel: %.2f",
      transl, eng_turn, (int)(speed*3.6), proposed_forward, fuel));

                           //process reservations
   for(i=0; i<MAX_AUTO_SEATS; i++){
      if(seats[i].reserver && (seats[i].reserve_count -= time) <= 0){
         seats[i].reserver = NULL;
      }
   }
                              //count airplane's time
   if(airplane_time!=-1){
      airplane_time += time;
      ComputeAirplaneSpeed(time);
   }
                              //break drastically when destroyed
   if(curr_version){
      if(speed)               //(zero speed to let use get out from vehicle)
         speed = .0f;
      return 0;
   }

   ProcessNetSeats(net_route_bits);

                              //for airplanes, control sound by scale
   ControlEnginePitch();
                           //shift texture(s)
   float tx_spd = speed * tab->ItemF(TAB_F_AUTO_TXS_SPEED);
   for(i=tx_shift.size(); i--; ){
      tx_shift[i]->SetProperty(0, FloatAsInt(tx_spd));
   }

   switch(mode){
   case MODE_START:
      if((mode_count -= time) <= 0){
         if(fuel){
            if(snd_engine) snd_engine->SetOn(true);
            SetEngTurn(1000);
            proposed_forward = true;
            InitFuelIcon();
            ShowFuel();
            mode = MODE_GO;
         }else
            mode = MODE_STAY;
      }
      break;
   case MODE_GO:
      emit_snd_cnt = Max(0, emit_snd_cnt - time);
      if(!emit_snd_cnt){
         emit_snd_cnt = VEHICLE_EMIT_SOUND_TIME;
         S_listen_source ls;
         ls.type = LISTEN_VEHICLE;
         ls.pos = model->GetPos();
         ls.power = Min(.5f + ((float)eng_turn / 800.0f), 2.0f) * 5.0f;
                           //don't send over network - shadow shots will emit sound
         mission.EmitListen(ls, 35.0f, false, seats[0].user);         
      }
      break;
   case MODE_STOP:
      if((mode_count -= time) <= 0){
         mode=MODE_STAY;
      }
      break;
   }
                              //check floor under vehicle for start falling again if anything change
   /*if(model && !IsBoat() && ((floor_test_cnt += time) > AUTO_FLOOR_TEST_TIME)){
      floor_test_cnt = 0;
      I3D_collision_data cd(model->GetWorldPos(), -axis_y, 0, model);
      //DebugLine(cd.from, cd.from+cd.dir, 0, 0xffffffff);
      if(!mission.GetScene()->TestCollision(cd))
         auto_falling = true;
   }*/

   if(!eng_turn && !speed && !auto_falling){
      if(transl){             //lets put there neutral while standing
         Translate(-transl);
         assert(!transl);
      }
      return 1;
   }

   bool net_mode = false;

   if(net){
                              //check if driver is network actor
      if(seats[0].user){
         if(seats[0].user->cbProc(CB_IS_NETWORK_ACTOR)){
            net_mode = true;
         }
      }else
      if(network_actor)
         net_mode = true;
   }

   float tsec = time*.001f;
                              
   if(!net_mode){
                              //auto-return wheel_turn to center
      if(wheel_turn){
         wheel_turn *= Max(.0f, (float)(1.0f-.1f*tsec*fabs(speed)));
         SetWheelTurn(0.0f, true, 1, net_route_bits);
      }

      SlowDown(time);

      SetFuel(Max(0.0f, fuel - eng_turn*tsec*.000005f * tab->ItemF(TAB_F_AUTO_CONSUM)), true, net_route_bits);
      if(!fuel){
         MotorOn(false, true, 1, net_route_bits);
      }

      if(transl){
                              //move
         float max_speed = wanted_speed[transl+1]*tab->ItemF(TAB_F_AUTO_TURN_RATIO);
         float wspd = max_speed * eng_turn / tab->ItemI(TAB_I_AUTO_TRANS_UP);
         //AUTO_DEBUG(wspd*3.6f);
         float add = tab->ItemF(TAB_F_AUTO_ACCEL) * tsec;
         if(speed < wspd){
            SetSpeed(Min(wspd, speed+add), true, 1, net_route_bits);
         }else
         if(speed > wspd){
            SetSpeed(Max(wspd, speed-add), true, 1, net_route_bits);
         }
      }
   }

   if(speed || auto_falling){
                           //turn wheels
      static const S_vector axis(1, 0, 0);
      S_quat q;
      q.Make(axis, -speed*tsec*2.0f / AUTO_WHEEL_DIST);
      for(i=0; i<4; i++)
      if(wheels[i].wheel){
         wheels[i].wheel->SetRot(wheels[i].wheel->GetRot() * q);
         wheels[i].wheel->Update();
      }
      S_vector save_auto_dir = model->GetDir() * speed * tsec;
      S_vector save_w_from[4];
      for(int ii = 4; ii--;){
         assert(wheels[ii].joint);
         const S_matrix &m = wheels[ii].joint->GetMatrix();
         save_w_from[ii] = m(3);
         save_w_from[ii].y = wheels[ii].world_y;
      }
                           //turn mobil
      {
         float max_wt = 1.0f - speed*.01f;
         max_wt *= tab->ItemI(TAB_I_AUTO_MAX_WHEEL_TURN)/180.0f;
         q.Make(axis_y, speed * tsec * wheel_turn * max_wt);
         S_quat new_rot = model->GetRot() * q;
         if(CheckFreeTurn(q))
            model->SetRot(new_rot);
      }

      if(net_mode)
         return 1;

      MoveMobil(tsec, save_auto_dir, save_w_from, net_mode, net_route_bits);

      if(net && !net_mode)
         NetSendPos(time, false, false, 1, net_route_bits);
                              //detroy auto when twist on back(stop move and disable use)
      {
         const S_matrix &m = model->GetFrameMatrix();
         const S_vector &current_up = m(1);
         if(current_up.y < -.01f)
            Destroy(false, true, 1, net_route_bits);
      }

   }

   AUTO_DEBUG(C_fstring("transl %i, speed %.2f, eng_turn %i", transl, speed*3.6f, eng_turn));
   AdjustEngTurn(tsec);
   return 0;
}

//----------------------------

LPC_actor CreateAutomobil(C_game_mission &gm){

   return new C_automobil(gm);
}

//----------------------------
//----------------------------
                              //cannon over network:
                              // - not used cannon:
                              //    - sync before using (through C_net_sync)
                              // - used cannon:
                              //    - send sub-actor ID bits:2 = 0
                              //       - bits:3 command:
                              //          - 0, 1 = rotation angle for axis X/Y
                              //             - float angle
                              //          - 3 = fire
                              //             - word num_ammo
                              //          - 3 = reflector state
                              //             bits:1 state (off/on)
                              //          - 5 = net sync mark
                              //          - 6 = proposed user
                              //             - word user_net_id (0 = stop use)

//----------------------------

enum{
   TAB_E_CAN_TYPE,
   TAB_I_CAN_ROT_Y_MIN,
   TAB_I_CAN_ROT_Y_MAX,
   TAB_I_CAN_ROT_X_MIN,
   TAB_I_CAN_ROT_X_MAX,
   TAB_I_CAN_ROT_SPEED_X,
   TAB_I_CAN_ROT_SPEED_Y,
   TAB_I_CAN_AMMO_AMOUNT,
   TAB_I_CAN_FIRE_POWER,
   TAB_I_CAN_FIRE_DELAY,
   TAB_S16_CAN_USESOUND,
   TAB_I_CAN_FIRE_RNDSHAKE,
   TAB_B_CAN_REFL_ON,
   TAB_B_CAN_USER_INVISIBLE,
   TAB_E_CAN_FIRE_TYPE,
   TAB_E_CAN_BULLETS_OUT,
};

C_table_template templ_cannon = {
   // {9DF774C0-A296-11d2-9B9E-004F4905FF1A}
   { 0x9df774c0, 0xa296, 0x11d2, { 0x9b, 0x9e, 0x0, 0x4f, 0x49, 0x5, 0xff, 0x1a } },
   "Cannon properties",
   NULL,
   {
      {TE_ENUM, TAB_E_CAN_TYPE, "Type", (dword)"Cannon\0Reflector\0"},
      {TE_BRANCH, 0, "Rotation", 2},
         {TE_BRANCH, 0, "Ranges (degrees)", 4},
            {TE_INT, TAB_I_CAN_ROT_X_MIN, "Left", -90, 0, -45},
            {TE_INT, TAB_I_CAN_ROT_X_MAX, "Right", 0, 90, 45},
            {TE_INT, TAB_I_CAN_ROT_Y_MIN, "Up", -90, 0, -22},
            {TE_INT, TAB_I_CAN_ROT_Y_MAX, "Down", 0, 90, 22},
         {TE_BRANCH, 0, "Max speed (degrees/sec)", 2},
            {TE_INT, TAB_I_CAN_ROT_SPEED_Y, "Horizontal", 0, 180, 120},
            {TE_INT, TAB_I_CAN_ROT_SPEED_X, "Vertical", 0, 180, 120},
      {TE_BRANCH, 0, "Firing", 7},
         {TE_INT, TAB_I_CAN_AMMO_AMOUNT, "Ammo", 0, 0, 100},
         {TE_INT, TAB_E_CAN_BULLETS_OUT, "Bullets out", 0, 0, 3},
         {TE_INT, TAB_I_CAN_FIRE_POWER, "Power", 0, 2000, 800},
         {TE_INT, TAB_I_CAN_FIRE_DELAY, "Fire delay", 0, 1000, 100},
         {TE_STRING, TAB_S16_CAN_USESOUND, "Use sound", 16},
         {TE_INT, TAB_I_CAN_FIRE_RNDSHAKE, "Fire shaking (angle)", 0, 20, 10},
         {TE_ENUM, TAB_E_CAN_FIRE_TYPE, "Fire type", (dword)"Shot\0Rocket\0"},
      {TE_BRANCH, 0, "Reflector", 1},
         {TE_BOOL, TAB_B_CAN_REFL_ON, "Default state", 0, 0, true},
      {TE_BOOL, TAB_B_CAN_USER_INVISIBLE, "Invisilbe user"},
      {TE_NULL}
   }
};

//----------------------------

class C_cannon: public C_actor{

   PI3D_frame joint[2], head;
   LPC_actor user;            //if not NULL, the cannon is being used by human
   LPC_actor net_sync_user;   //proposed user
   float joint_angle[2];
   int fire_count, ammo_count;
   bool ref_state;
   PI3D_frame use_pos_dummy;
   C_smart_ptr<I3D_sound> snd_fire;
   PI3D_frame sound_source;
   enum{
      TYPE_GUN, TYPE_REFLECTOR
   };
   LPC_table tab;

   C_smart_ptr<C_net_sync> net_sync;

   float net_angle[2];        //reference network angle

//----------------------------

   void ReflectorOn(bool b, bool snd = true, bool net_send = true, dword net_route_code = 0, byte net_route_bits = 0){

      if(ref_state!=b){
         ref_state = b;

         PI3D_frame frm;
         frm = frame->FindChildFrame("*", ENUMF_LIGHT | ENUMF_USEMASK);
         if(frm) frm->SetOn(ref_state);
         frm = frame->FindChildFrame("%|shine", ENUMF_VISUAL | ENUMF_USEMASK);
         if(frm) frm->SetOn(ref_state);

         if(snd)
            mission.PlaySound("switch", frame, 2.0, 4.0);

         if(net && net_send){
            bool g = true;
            net_msg->Ascend(net_id);
            net_msg->AddBits(net_route_code, net_route_bits, g);
            net_msg->AddBits(3, 3, g);
            net_msg->AddBits(ref_state, 1, g);
            net_msg->Descend();
         }
      }
   }

//----------------------------

   void ApplyAngle(int i, float angle){
      PI3D_frame frm(joint[i]);
      if(frm){
         S_quat q;
         q.Make(!i ? S_vector(1, 0, 0) : S_vector(0, -1, 0), angle);
         frm->SetRot(q);
         frm->Update();
      }
   }

//----------------------------

   void GetSound(){
                              //play sound
      const char *use_snd = tab->ItemS(TAB_S16_CAN_USESOUND);
      if(!*use_snd) use_snd = "cannon_0";
      if(!snd_fire){
                              //remove extension from filename
         C_string str = use_snd;
         for(int i=str.Size(); i--; ){
            if(str[i]=='.'){
               str[i] = 0;
               break;
            }
         }
         use_snd = str;
         snd_fire = mission.CreateSound(use_snd);
         if(snd_fire){
            snd_fire->LinkTo(head);
            snd_fire->SetRange(tab_game_cfg->ItemF(TAB_F_GAME_SNDR_CANNON_MIN),
               tab_game_cfg->ItemF(TAB_F_GAME_SNDR_CANNON_MAX));
            snd_fire->Release();
         }
      }
   }

//----------------------------

   void Fire(bool play_snd, LPC_actor user, bool net_send, dword net_route_code = 0, byte net_route_bits = 0){

      if(net_send){
         if(play_snd) GetSound();

         if(tab->ItemE(TAB_E_CAN_FIRE_TYPE)==0){   //shot type
                        //create gunshoot actor
            S_vector pos_dir[2] = {
               head->GetWorldPos(),
               head->GetWorldDir()
            };
            assert(user);
            user->cbProc(CB_USE_CANNON, 12, (dword)pos_dir);

            int num_bullets = Min(tab->ItemI(TAB_E_CAN_BULLETS_OUT), ammo_count);
            ammo_count -= num_bullets;

            S_gun_shoot_init si;
            si.pos = pos_dir[0];
            si.dir = pos_dir[1];
            si.shoot_item = -1;
            si.shooter = user;
            si.power = tab->ItemI(TAB_I_CAN_FIRE_POWER);
            si.weapon = head;
            si.gun_snd = snd_fire;
            si.cannon_actor = this;

            for(int i=0; i<num_bullets; i++){
               si.delay = i*fire_count;
               mission.CreateActor(ACTOR_GUN_SHOOT, (dword)&si);
            }
            fire_count *= num_bullets;
            //mission.CreateActor(ACTOR_GUN_SHOOT, (dword)&si);
            if(user && net_send && user->GetType()==ACTOR_PLAYER) mission.stat_ammo_spent += num_bullets;
         }else{   //rocket type
            S_rocket_init ri;
            ri.pos = head->GetWorldPos();
            ri.dir = head->GetWorldDir();
            ri.id = 34;
            ri.shooter = user;
            mission.CreateActor(ACTOR_ROCKET, (dword)&ri);
            //GunMakeAnim();

            --ammo_count;
         }
         {               //random shaking after fire
            float f[3];
            f[2] = .1f;
            float shake = tab->ItemI(TAB_I_CAN_FIRE_RNDSHAKE)*GetPI()/180.0f;
            //f[0] = shake - Random(shake*2.0f);
            f[0] = shake - S_float_random() * shake * 2.0f;
            //f[1] = shake - Random(shake*2.0f);
            f[1] = shake - S_float_random() * shake * 2.0f;
            //cbProc(CB_USE_CANNON, 1, (dword)f);
            UseCannon(1, (dword)f, net_route_code, net_route_bits);
         }
         if(net){
            const bool g = false;
            net_msg->Ascend(net_id);
            if(net_route_bits)
               net_msg->AddBits(net_route_code, net_route_bits, g);
            net_msg->AddBits(2, 3, g);
            net_msg->AddWord(ammo_count, g);
            net_msg->Descend();
         }
      }
      S_listen_source ls = S_listen_source(LISTEN_SHOOT, sound_source->GetWorldPos(), 30);
      mission.EmitListen(ls, tab_game_cfg->ItemF(TAB_F_GAME_SNDR_CANNON_MAX), false, user);

      GunMakeAnim();
   }

//----------------------------

   bool BeginUse(LPC_actor act){
      assert(!user);
      if(act->cbProc(CB_USE_CANNON, 0, (dword)this)){
         user = act;
         //if(tab->ItemB(TAB_B_CAN_USER_INVISIBLE)) user->cbProc(CB_SETVISIBLE, false);
         act->cbProc(CB_USE_CANNON, 10, 1);
         if(net)
            for(int i=0; i<2; i++) net_angle[i] = joint_angle[i];
         return true;
      }
      return false;
   }

//----------------------------

   void GunMakeAnim(){

      PI3D_frame mp = frame;
      if(!mp) mp = head;
      while(mp && mp->GetType()!=FRAME_MODEL) mp=mp->GetParent();
      if(mp){
         PI3D_model model = I3DCAST_MODEL(mp);
                              //check if some animation already running
         if(model->NumInterpolators()) return;
                              //run gun anim
         PI3D_animation_set anim;
         anim = mission.GetAnim("gunshoot.i3d");
         if(anim){
            model->StopAnimation(0);
            model->SetAnimation(0, anim);
         }
      }
      if(tab->ItemE(TAB_E_CAN_FIRE_TYPE)==0)
         mission.MakeParticle("sh_flash", NULL, head->GetWorldPos(), NULL, NULL);
      else{
         mission.MakeParticle("tank_smk", "tank_smk.tab", head->GetWorldPos(), NULL, NULL);
      }
   }

public:
public:
   C_cannon(C_game_mission &m):
      C_actor(m, ACTOR_CANNON),
      fire_count(0),
      ref_state(true),
      head(NULL),
      net_sync_user(NULL),
      sound_source(NULL),
      use_pos_dummy(NULL),
      user(NULL)
   {
      joint[0] = joint[1] = NULL;
      joint_angle[0] = joint_angle[1] = 0.0;
      memset(net_angle, 0, sizeof(net_angle));

      tab = CreateTable();
      tab->Open((dword)&templ_cannon, TABOPEN_TEMPLATE);
   }

//----------------------------

   ~C_cannon(){
      tab->Release();
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      C_actor::SetFrame(frm);

      for(int i=0; i<2; i++){
         joint[i] = frm->FindChildFrame(C_fstring("joint_%c", 'x'+i), ENUMF_ALL);
         if(!joint[i]){
            ReportActorError(C_fstring("can't find frame 'joint_%c' in model", 'x'+i));
         }
      }
      head = frm->FindChildFrame("head", ENUMF_ALL);
      if(!head){
         ReportActorError("can't find frame 'head' in model");
      }
      sound_source = frm;
   }

//----------------------------

   virtual void SetNetID(word id){

      net_sync = CreateNetSync(id, 5, 3);
      net_sync->Release();
      C_actor::SetNetID(id);
   }

//----------------------------

   void HideCannon(){

      joint[0]->SetOn(false);
      joint[1]->SetOn(false);
   }

//----------------------------

   virtual void SetVisible(bool on, bool vol_only, bool net_send){

      if(!model){
         HideCannon();
         return;
      }
      C_actor::SetVisible(on, vol_only, net_send);
   }

//----------------------------

   virtual bool IsOnMap() const{ return true; }

//---------------------------

   virtual void SaveGame(C_chunk *ck) const{

                           //write data
      ck->Write(joint_angle, sizeof(joint_angle));
      ck->Write(&fire_count, sizeof(fire_count));
      ck->Write(&ammo_count, sizeof(ammo_count));
      ck->Write(&ref_state, sizeof(ref_state));
      ck->WStringChunk(CT_NAME, user ? user->GetName() : "");
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

                              //read data
      ck->Read(joint_angle, sizeof(joint_angle));
      ck->Read(&fire_count, sizeof(fire_count));
      ck->Read(&ammo_count, sizeof(ammo_count));
      bool ref;
      ck->Read(&ref, sizeof(ref_state));
      C_string str;
      ck->ReadString(CT_NAME, str);
      user = !str.Size() ? NULL : mission.FindActor(str);
                              //update
      for(int i=0; i<2; i++)
         ApplyAngle(i, joint_angle[i]);
      ReflectorOn(ref, false, false);

      return true;
   }

//----------------------------

   virtual const S_vector GetUsePos(int indx, bool abs_index_mode = false) const{

      return frame->GetWorldPos() - frame->GetWorldDir();
   }
   virtual bool GetUseDir(S_vector &ret, int indx, bool abs_index_mode = false) const{
      ret = frame->GetWorldDir();
      return true;
   }
   virtual float GetUseDirThresh() const{ return PI*.25f; }
   virtual bool CheckUseFreeHands(int indx, bool abs_index_mode) const{ return true; }
   virtual dword CheckUseStay(dword suggest_stay_mode, int indx) const{
      if(suggest_stay_mode==SM_LAY)
         return SM_BENCH;
      return suggest_stay_mode;
   }

//----------------------------

   dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_GET_GUN_SOUND:
         GetSound();
         return (dword)(PI3D_sound)snd_fire;

      case CB_USE_CANNON:
         return UseCannon(prm1, prm2);
      }
      return C_actor::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual LPC_table GetTable(int index) const{
      if(index==0)
         return tab;
      return C_actor::GetTable(index);
   }

//----------------------------

   virtual class C_table_template *GetTemplate(int) const{ return &templ_cannon; }
   virtual void TableUpdate(int index, bool not_load_prg){
      ammo_count = tab->ItemI(TAB_I_CAN_AMMO_AMOUNT);
      ReflectorOn(tab->ItemB(TAB_B_CAN_REFL_ON), false, false);
   }
   virtual void MissionSave(C_chunk *ck) const{ SaveTable(ck, tab); }
   virtual bool MissionLoad(C_chunk *ck, int table_id){ return LoadTable(ck, tab); }

//----------------------------

   virtual int Use(LPC_actor act, dword prm, bool force_use_off){

      if(prm){
         if(!NumUsePositions())
            return 0;
         if(!net){
                              //use directly
            BeginUse(act);
         }else{
            net_sync_user = act;
            net_sync->BeginEvent();
            return 2;         //in progress
         }
      }else{                  //switch off shooting mode
         if(user!=act)
            return 0;
         //if(tab->ItemB(TAB_B_CAN_USER_INVISIBLE)) user->cbProc(CB_SETVISIBLE, true);
         user->cbProc(CB_USE_CANNON, 0, NULL);
         if(net){
            bool g = true;
            net_msg->Ascend(net_id);
            net_msg->AddBits(6, 3, g);
            net_msg->AddWord(0, g);
            net_msg->Descend();
         }
         user = NULL;
      }
      return 1;
   }

//----------------------------

   virtual int NumUsePositions(C_actor *asker = NULL) const{

      if(!ammo_count || user)
         return 0;
      if(net && net_sync->GetState())
         return 0;
      return 1;
   }

//----------------------------

   dword UseCannon(dword prm1, dword prm2, dword net_route_code = 0, byte net_route_bits = 0){

      int i;

      switch(prm1){
      case 1:                 //rotate joint
         {
            for(i=0; i<2; i++){
               //PI3D_frame frm(joint[i]);
               //if(!frm) break;
               float *fptr = (float*)prm2;
               if(!fptr[i]) continue;
               const float &tsec = fptr[2];
               float &angle = joint_angle[i];
               float add = fptr[i]*tsec;
                                 //limit speed
               float max_speed = tab->ItemI(TAB_I_CAN_ROT_SPEED_X+i);
               max_speed = max_speed*GetPI()/180.0f*tsec;
               if(fabs(add)>max_speed) add = (add>0.0) ? max_speed : -max_speed;
               float range_min = (float)(tab->ItemI(TAB_I_CAN_ROT_Y_MIN+2*i))*GetPI()/180.0f;
               float range_max = (float)(tab->ItemI(TAB_I_CAN_ROT_Y_MAX+2*i))*GetPI()/180.0f;
               angle = Max(range_min, Min(range_max, angle + add));

               if(net){
                  float delta = fabs(net_angle[i]-angle);
                  if(delta>=PI*.01f){
                     const bool g = false;
                     net_msg->Ascend(net_id);
                           //time to send it out
                     net_angle[i] = angle;
                     if(net_route_bits)
                        net_msg->AddBits(net_route_code, net_route_bits, g);
                     net_msg->AddBits(i, 3, g);
                     net_msg->AddFloat(angle, g);
                     net_msg->Descend();
                  }
               }
               ApplyAngle(i, angle);
               //if(prm1==2) user->cbProc(CB_USE_CANNON, 10);
            }
         }
         return true;

      case 3:                 //use gun
         {
            byte type = tab->ItemE(TAB_E_CAN_TYPE);
            switch(type){
            case TYPE_REFLECTOR:
               ReflectorOn(!ref_state, true, true, net_route_code, net_route_bits);
               break;
            case TYPE_GUN:

               if(ammo_count){
                  if(fire_count) break;
                  fire_count = tab->ItemI(TAB_I_CAN_FIRE_DELAY);
                  Fire(true, (LPC_actor)prm2, true, net_route_code, net_route_bits);
                  return true;
               }
               break;
            }
         }
         break;

      case 4:                 //get dir
         {
            if(!head) break;
            S_vector *vp=(S_vector*)prm2;
            *vp = head->GetWorldDir();
         }
         return true;

      case 5:                 //get pos
         {
            if(!head) break;
            S_vector *vp=(S_vector*)prm2;
            *vp = head->GetWorldPos();
         }
         return true;

      case 6:              //get ammo
         return ammo_count;

      case 8:              //get type
         return tab->ItemE(TAB_E_CAN_TYPE);
      case 9:              //get reflector state
         return ref_state;
      case 11:             //get use_pos
         {
            if(!use_pos_dummy) return false;
            return (dword)&use_pos_dummy->GetWorldPos();
         }
         return true;
      }
      return 0;
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      if(net){
         dword winner_pid;
         if(net_sync->GetState(&winner_pid)==C_net_sync::SS_RESOLVED){
            if(winner_pid==net->GetPID()){
               assert(net_sync_user);
               BeginUse(net_sync_user);
                           //say to all who's user
               net_msg->Ascend(net_id);
               net_msg->AddBits(6, 3, true);
               net_msg->AddWord(user ? user->GetNetID() : 0, true);
               net_msg->Descend();

               net_sync_user = NULL;
            }else{         //lock until we know who's user
               net_sync->Lock(true);
            }
            net_sync->Reset();
         }
      }

      if(fire_count && (fire_count -= tc.time) < 0)
         fire_count = 0;
      return 0;
   }

//----------------------------

   virtual void NetCodeIn(byte code, class C_net_in_msg *in_msg){

      switch(code){
      case 0:              //new angle
      case 1:              //new angle
         {
            float f = in_msg->GetFloat();
            joint_angle[code] = net_angle[code] = f;
            ApplyAngle(code, f);
            //NET_LOG("C_cannon: net - new angle message received");
         }
         break;
      case 2:              //fire
         Fire(true, NULL, false);
         ammo_count = in_msg->GetWord();
         break;
      case 3:              //reflector state
      case 4:
         ReflectorOn(in_msg->GetBits(1), true, false);
         break;
      case 5:              //sync
         net_sync->NetIn(in_msg, false);
         break;
      case 6:
         {
            word id = in_msg->GetWord();
            LPC_actor user1 = id ? mission.NetFindActor(id) : NULL;
            NET_LOG(C_fstring("C_cannon: network user: %s", user1 ? user1->GetName() : "<NULL>"));
            if(user1){
               BeginUse(user1);
            }else{
               if(user){
                  user->cbProc(CB_USE_CANNON, 0, NULL);
                  user = NULL;
               }
            }
            net_sync->Lock(false);
         }
         break;
      default:
         assert(0);
      }
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      in_msg->Ascend();
      while(in_msg->GetChunkSize())
         NetCodeIn(in_msg->GetBits(3), in_msg);
      in_msg->Descend();
   }

//----------------------------

   virtual void NetPlrDisconnect(dword pid){

      net_sync->OnPlayerDisconnect(pid);
      C_actor::NetPlrDisconnect(pid);
   }
};

//----------------------------

LPC_actor CreateCannon(C_game_mission &gm){

   return new C_cannon(gm);
}

//----------------------------
//----------------------------
                              //C_auto_cannon

                              //auto+cannon over net:
                              // - route incomming messages:
                              //    - bits:2 = sub_class:
                              //       - 0 = version
                              //       - 1 = auto
                              //       - 2 = cannon
                              //       - 3 = auto_cannon
                              //          - bits:1 = sub_code:
                              //             - 0 = destroy cannon

class C_auto_cannon: public C_automobil{
   C_cannon cannon;

   void DestroyCannon(){
      cannon.HideCannon();
   }

   virtual bool HitExplode(C_actor *caller, int power, S_CB_hit *hit, bool make_expl, dword net_route_code, byte net_route_bits){

      C_automobil::HitExplode(caller, power, hit, make_expl, net_route_code, net_route_bits);
      if(curr_version){
         DestroyCannon();
         if(net){
            const bool g = true;
            net_msg->Ascend(net_id);
            net_msg->AddBits(3, 2, g);
            //net_msg->AddBits(0, 1, g);
            net_msg->Descend();
         }
      }
      return 0;
   }
public:
   C_auto_cannon(C_game_mission &m):
      C_automobil(m, ACTOR_AUTO_CANNON),
      cannon(m)
   {
      net_route_bits = 2;
   }

   virtual void SetNetID(word w){
                              //cannon has the same id
      cannon.SetNetID(w);
      C_automobil::SetNetID(w);
   }

//----------------------------

   virtual void SetFrame(PI3D_frame frm){

      cannon.SetFrame(frm);
      C_automobil::SetFrame(frm);
   }

//----------------------------

   virtual int GetTableIndex() const{ return WinSelectItem("Edit:", "VERSION\0AUTOMOBIL\0CANNON\0", 1); }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      C_automobil::SaveGame(ck);
      cannon.SaveGame(ck);
   }

//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      if(!C_automobil::LoadGame(ck))
         return false;
      return cannon.LoadGame(ck);
   }

//----------------------------

   dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_GET_GUN_SOUND:
         cannon.cbProc(msg, prm1, prm2);
         break;

      case CB_HIT:
         if(!curr_version){
            return HitExplode(NULL, prm2, (S_CB_hit*)prm1, true, 1, net_route_bits);
         }
         break;

      case CB_USE_CANNON:
                              //let cannon do the job
         return cannon.UseCannon(prm1, prm2, 2, 2);

      case CB_USE_AUTO:
         switch(prm1){
         case 16:
            return cannon.UseCannon(6, 0, 2, 2);
         default:
            return C_automobil::UseAuto(prm1, prm2, 2);
         }
         break;
      }
      return C_automobil::cbProc(msg, prm1, prm2);
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      cannon.Tick(tc);
      return C_automobil::Tick1(tc.time, 2);
   }

//----------------------------

   virtual LPC_table GetTable(int index) const{

      if(index==2) return cannon.GetTable(0);
      return C_automobil::GetTable(index);
   }
   virtual class C_table_template *GetTemplate(int index) const{
      if(index==2) return cannon.GetTemplate(0);
      return C_automobil::GetTemplate(index);
   }

   virtual void MissionSave(C_chunk *ck) const{

      C_version::MissionSave(ck);
      SaveTableEx(ck, 1);
      SaveTableEx(ck, 2);
   }
   virtual bool MissionLoad(C_chunk *ck, int table_id){
      switch(table_id){
      case 2:
         return cannon.MissionLoad(ck, table_id);
      default:
         return C_automobil::MissionLoad(ck, table_id);
      }
   }

//----------------------------

   virtual void TableUpdate(int index, bool not_load_prg){
      
      cannon.TableUpdate(index, not_load_prg);
      C_automobil::TableUpdate(index, not_load_prg);
   }

//----------------------------

   virtual void NetIn(PC_net_in_msg in_msg){

      in_msg->Ascend();
      while(in_msg->GetChunkSize()){
         byte route = in_msg->GetBits(2);
         switch(route){
         case 0:              //version
            NET_LOG("message for C_version");
            C_version::NetCodeIn(in_msg->GetByte(), in_msg);
            break;

         case 1:              //automobil
            NET_LOG("message for C_automobil");
            C_automobil::NetCodeIn(in_msg->GetBits(4), in_msg);
            break;

         case 2:
            NET_LOG("message for C_cannon");
            cannon.NetCodeIn(in_msg->GetBits(3), in_msg);
            break;

         case 3:
            DestroyCannon();
            /*
            switch(in_msg->GetBits(1)){
            case 0:
               DestroyCannon();
               break;
            default:
               assert(0);
            }
            */
            break;
         }
      }
      in_msg->Descend();
   }

//----------------------------

   virtual void NetPlrDisconnect(dword pid){

      cannon.NetPlrDisconnect(pid);
      C_automobil::NetPlrDisconnect(pid);
   }
};

//----------------------------

LPC_actor CreateAutoCannon(C_game_mission &gm){

   C_auto_cannon *ac = new C_auto_cannon(gm);
   return ac;
}

//----------------------------
//----------------------------
