/*--------------------------------------------------------
   Wide-char 2D text output.
   Written by Michal Bacik 17.8.1999
   Updated 12.6.2000
--------------------------------------------------------*/
#include "all.h"
#define MBCS
#include <windows.h>
#include "i3d_font.h"

//#define DEBUG_MODE            //define to draw in special mode - no colorkeying


//----------------------------

class C_wide_text_rep{
public:
                              //texture keeping the text
   C_smart_ptr<I3D_texture> tp;
   float size[2];             //in normalized screen coords
   float use_area[2];         //ratio of used texture area
                              //screen position (reference point)
   float pos_x, pos_y;
                              //screen position (upper left corner)
   float real_x, real_y;
   dword color;

   C_smart_ptr<I3D_visual_dynamic> dyn_vis;
};

//----------------------------
                              //simple cache for last created font
class C_font_resource{
public:
   struct S_font_info{
      int res_x, res_y;
      HFONT hfont;
   } fnt_j, fnt_e;
   
   C_font_resource(){
      memset(&fnt_j, 0, sizeof(fnt_j));
      memset(&fnt_e, 0, sizeof(fnt_e));
   }

   ~C_font_resource(){
      if(fnt_j.hfont) DeleteObject(fnt_j.hfont);
      if(fnt_e.hfont) DeleteObject(fnt_e.hfont);
   }
} font_res;

//----------------------------
                              //MS Gothic
static const char font_name[] = "MS Gothic";
//static const char font_name[] = "Times New Roman";  //this is not working for 100%
//static const char font_name[] = "Arial";

static HFONT japCreateFontJ(int charx, int chary){

                              //try to reuse cache
   if(font_res.fnt_j.hfont){
      if(font_res.fnt_j.res_x==charx && font_res.fnt_j.res_y==chary)
         return font_res.fnt_j.hfont;
      bool b;
      b = DeleteObject(font_res.fnt_j.hfont);
      assert(b);
   }

   font_res.fnt_j.hfont = CreateFont(
      chary,                  //logical height of font
      charx/2,
      0,                      //angle of escapement
      0,                      //base-line orientation angle
      FW_NORMAL,              //font weight
      //FW_THIN,              //font weight
      FALSE,                  //italic attribute flag
      FALSE,                  //underline attribute flag
      FALSE,                  //strikeout attribute flag
      SHIFTJIS_CHARSET,       //character set identifier
      OUT_DEFAULT_PRECIS,     //output precision
      CLIP_DEFAULT_PRECIS,    //clipping precision
      DEFAULT_QUALITY,        //output quality
      FIXED_PITCH | FF_DONTCARE, //pitch and family
      (const char*)font_name);//pointer to typeface name string);

   font_res.fnt_j.res_x = charx;
   font_res.fnt_j.res_y = chary;
   return font_res.fnt_j.hfont;
}

//----------------------------

static HFONT japCreateFontE(int charx, int chary){

                              //try to reuse cache
   if(font_res.fnt_e.hfont){
      if(font_res.fnt_e.res_x==charx && font_res.fnt_e.res_y==chary)
         return font_res.fnt_e.hfont;
      bool b;
      b = DeleteObject(font_res.fnt_e.hfont);
      assert(b);
   }
   font_res.fnt_e.hfont = CreateFont(
      chary,                  //logical height of font
      charx/2,
      0,                      //angle of escapement
      0,                      //base-line orientation angle
      //FW_NORMAL,              //font weight
      //FW_MEDIUM,              //font weight
      FW_THIN,              //font weight
      FALSE,                  //italic attribute flag
      FALSE,                  //underline attribute flag
      FALSE,                  //strikeout attribute flag
      ANSI_CHARSET,       //character set identifier
      OUT_DEFAULT_PRECIS,     //output precision
      CLIP_DEFAULT_PRECIS,    //clipping precision
      DEFAULT_QUALITY,        //output quality
      FIXED_PITCH | FF_DONTCARE, //pitch and family
      //DEFAULT_PITCH | FF_DONTCARE, //pitch and family
      (const char*)font_name);//pointer to typeface name string);

   font_res.fnt_e.res_x = charx;
   font_res.fnt_e.res_y = chary;
   return font_res.fnt_e.hfont;
}

//----------------------------

struct S_wide_line{
   wchar_t *wp;
   float x;
   int len;
   float width;
   int base_indx;
};

//----------------------------

C_wide_text::C_wide_text(const wchar_t *wcp, dword color, float x, float y, float size,
   dword just, PI3D_driver drv, float fmt_rect[4],
   float *xr1, float *yr1, int *num_lines){

   LPIGraph igraph = drv->GetGraphInterface();

   int scrn_sx = igraph->Scrn_sx();
   //int scrn_sy = igraph->Scrn_sy();

   size *= .8f;
   float scrn_width = 1.0f;

   /*
   {
      const float ratio_x = screen_ratio * .75f;
      x *= ratio_x;
      if(fmt_rect){
         fmt_rect[0] *= ratio_x;
         fmt_rect[2] *= ratio_x;
      }
      scrn_width *= ratio_x;
   }
   */
   const float ratio_x = 1.0f;

   int charx = scrn_sx * size * ratio_x;
   int chary = scrn_sx * size;
   int num_chars = wcslen(wcp);

   bool scr_clip = (just&TEXT_JUSTIFY_CLIP_TO_SCREEN);
   bool eol_char = (just&TEXT_JUSTIFY_CONVERT_EOL);
   just &= 0xffff;

   wchar_t *wcp1 = NULL;

   if(eol_char){
      int len = wcslen(wcp);
      wcp1 = new wchar_t[len + 1];

      wcscpy(wcp1, wcp);
      for(int i=0; i<len; i++)
         if(wcp1[i]=='\\') wcp1[i] = '\n';

      wcp = wcp1;
   }

   vector<S_wide_line> lines;

   float xr = x;
   float yr = y;

   float *char_width = NULL;

   HFONT hFontJap = NULL, hFontEng = NULL;
   
   if(num_lines) *num_lines = 0;

   HWND hwnd = (HWND)igraph->GetHWND();

   if(num_chars)
   {

      int i;
      float x_scale = size;
      const wchar_t *wp = wcp;

      {
         HDC hdc = NULL;
         HGDIOBJ hgi_save = NULL;
         char_width = new float[num_chars+1];
         bool is_tt = true;
         for(i=0; i<num_chars; i++){
            wchar_t c = wp[i];
            if(c>=256) char_width[i] = x_scale;
            else{
               if(!hFontEng)
                  hFontEng = japCreateFontE(charx, chary);

               if(!hdc){
                  hdc = GetDC(hwnd);
                  hgi_save = SelectObject(hdc, hFontEng);
               }
               int w;
               bool st = false;
               if(is_tt){
                  ABC abc;
                  is_tt = GetCharABCWidths(hdc, wp[i], wp[i], &abc);
                  if(is_tt){
                     w = abc.abcA + abc.abcB + abc.abcC;
                     st = true;
                  }else{
                     st = GetCharWidth32(hdc, wp[i], wp[i], &w);
                  }
               }else
                  st = GetCharWidth32(hdc, wp[i], wp[i], &w);
               if(!st){
                  w = charx;
               }
               char_width[i] = (float)w / (float)scrn_sx;
            }
         }
         if(hgi_save)
            SelectObject(hdc, hgi_save);
         ReleaseDC(hwnd, hdc);
      }

      float x_clip_delta = 0.0f;

      int base_indx = 0;
      while(true){
         if(num_lines) ++(*num_lines);
                              //get width of line
         const wchar_t *eol = wcschr(wp, '\n');
         bool end = bool(!eol);
         if(end) eol = wp + wcslen(wp);
         int line_len = eol - wp;

         float line_width = 0;
         for(i=0; i<line_len; i++)
            line_width += char_width[base_indx+i];

         float x_base = x;
                              //format text to fit in rectanlge (optional)
         if(fmt_rect){
            x_base = Max(x_base, fmt_rect[0]);
            if(line_width > (fmt_rect[2]-fmt_rect[0])){
                              //remove some words from end
               int last_beg = line_len;
               while(true){
                  i = last_beg - 1;
                  line_width -= char_width[base_indx+i];
                  if(line_width <= (fmt_rect[2]-fmt_rect[0])){
                              //check if we may break the line here
                     if(wp[i-1]>=256 || wp[i]>=256 || wp[i]==' ')
                        break;
                  }
                  last_beg = i;
               };
               assert(i>0);
               line_len = i;
            }
         }

         switch(just){
         case TEXT_JUSTIFY_CENTER:
            x_base -= line_width/2;
            xr -= line_width/2;
            break;
         case TEXT_JUSTIFY_RIGHT:
            x_base -= line_width;
            xr -= line_width;
            break;
         }
         if(scr_clip){
            float new_x_base = x_base;
                              //clip onto screen
            if((x_base+line_width) > scrn_width-.02f) new_x_base = scrn_width - .02f - line_width;
            new_x_base = Max(0.0f, new_x_base);
            if(fabs(x_clip_delta)<fabs(new_x_base-x_base))
               x_clip_delta = new_x_base - x_base;
            x_base = new_x_base;
         }
                              //copy to line
         lines.push_back(S_wide_line());
         lines.back().wp = new wchar_t[line_len+1];
         wcsncpy(lines.back().wp, wp, line_len);
         lines.back().wp[line_len] = 0;
         lines.back().len = line_len;
         lines.back().x = x_base;
         lines.back().width = line_width;
         lines.back().base_indx = base_indx;

         wp += line_len;
         if(!(*wp)) break;
         base_indx += line_len;
                           //skip eol
         if(*wp=='\n'){
            ++base_indx;
            ++wp;
         }

         y += size;
      }
      xr += x_clip_delta;
   }
   delete[] wcp1;

   C_wide_text_rep *rep1 = new C_wide_text_rep;

   if(num_chars)
   {
      PI3D_texture tp = NULL;

      float min_x = 1000;
      float max_x = -1000;
                              //determine boundaries
      for(int i=0; i<lines.size(); i++){
         min_x = Min(min_x, lines[i].x);
         max_x = Max(max_x, lines[i].x + lines[i].width);
      }

      float rel_size_x = (max_x - min_x);
      float rel_size_y = lines.size() * size;

      int sx = rel_size_x * scrn_sx;
      int sy = rel_size_y * scrn_sx;

      int used_sx = sx;
      int used_sy = sy;

      if(sx && sy){
         int size1[2];
         size1[0] = sx;
         size1[1] = sy;
         for(i=0; i<16; i++) if((1<<i) >= sx){ sx = (1<<i); break; }
         for(i=0; i<16; i++) if((1<<i) >= sy){ sy = (1<<i); break; }

         HDC hdc = CreateCompatibleDC(NULL);
         assert(hdc);

         void *bitmap_bits = NULL;

         BITMAPINFO bmi;
         memset(&bmi, 0, sizeof(bmi));
         bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
         bmi.bmiHeader.biWidth = sx;
         bmi.bmiHeader.biHeight = sy;
         bmi.bmiHeader.biPlanes = 1;
         bmi.bmiHeader.biBitCount = 24;
         bmi.bmiHeader.biCompression = BI_RGB;
         bmi.bmiHeader.biSizeImage = 0;

         HBITMAP h_bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, &bitmap_bits, NULL, 0);
         assert(h_bm);

         i = SetMapMode(hdc, MM_TEXT);
         assert(i);
         HGDIOBJ hgi_save = SelectObject(hdc, h_bm);

         i = SetTextAlign(hdc, TA_TOP | TA_LEFT);
         assert(i!=GDI_ERROR);

         {
            dword fill_color = 0;
   #ifdef DEBUG_MODE
            fill_color = 0x00101010;
   #endif
            HBRUSH hb = CreateSolidBrush(fill_color);
            assert(hb);
            RECT rc = {0, 0, size1[0], size1[1]};
            i = FillRect(hdc, &rc, hb);
            assert(i);
            i = DeleteObject(hb);
            assert(i);
         }

   #ifndef DEBUG_MODE
         i = SetBkMode(hdc, TRANSPARENT);
         assert(i);
   #else
         i = SetBkMode(hdc, OPAQUE);
         assert(i);
         i = SetBkColor(hdc, 0x00000080);
         assert(i!=CLR_INVALID);
   #endif

         SetTextCharacterExtra(hdc, 0);
         int shadow_dist = Max(1, scrn_sx*3 / 1024);

         HFONT last_font = NULL;

         float py = 0;
         for(i=0; i<lines.size(); i++){
            float px = lines[i].x-min_x;
            for(int j=0; j<lines[i].len; j++){
               wchar_t c = lines[i].wp[j];

               HFONT hf;
               if(c<256){
                  if(!hFontEng)
                     hFontEng = japCreateFontE(charx, chary);
                  hf = hFontEng;
               }else{
                  if(!hFontJap) 
                     hFontJap = japCreateFontJ(charx, chary);
                  hf = hFontJap;
               }
               if(last_font!=hf)
                  SelectObject(hdc, last_font = hf);
                                 //draw shadow
               SetTextColor(hdc, 0x080808);
               TextOutW(hdc, (int)(px*scrn_sx) + shadow_dist, (int)(py*scrn_sx) + shadow_dist, &c, 1);

                                 //draw text
               SetTextColor(hdc, 0xffffff);
   #ifdef DEBUG_MODE
               SetBkMode(hdc, TRANSPARENT);
   #endif
               TextOutW(hdc, px * scrn_sx, py * scrn_sx, &c, 1);
   #ifdef DEBUG_MODE
               SetBkMode(hdc, OPAQUE);
   #endif
               px += char_width[lines[i].base_indx+j];
            }
            py += size;
         }
                                 //create texture from given bitmap
         {
            S_pixelformat pf;
            dword find_flags = FINDTF_ALPHA1;
   //#ifndef DEBUG_MODE
            find_flags |= FINDTF_COMPRESSED;
   //#endif
            I3D_RESULT ir = drv->FindTextureFormat(pf, find_flags);
            if(I3D_SUCCESS(ir)){
                                 //create texture of desired resolution and pixelformat
               I3D_CREATETEXTURE ct;
               memset(&ct, 0, sizeof(ct));
               ct.flags = TEXTMAP_TRANSP | TEXTMAP_NOMIPMAP | TEXTMAP_USEPIXELFORMAT |
                  TEXTMAP_DIFFUSE | TEXTMAP_MEM_PTR;
               ct.pf = &pf;
               bmi.bmiHeader.biCompression = (long)bitmap_bits;
               ct.text_mem = &bmi;
               ct.size_x = sx;
               ct.size_y = sy;

               ir = drv->CreateTexture(&ct, &tp);
               assert(I3D_SUCCESS(ir));
            }
         }
         SelectObject(hdc, hgi_save);

         i = DeleteObject(h_bm);
         assert(i);
         i = DeleteDC(hdc);
         assert(i);
      }
      rep1->real_x = min_x;
      rep1->real_y = yr - size*.81f;

      rep1->use_area[0] = !IsAbsMrgZero(sx) ? (float)used_sx / (float)sx : .0f;
      rep1->use_area[1] = !IsAbsMrgZero(sy) ? (float)used_sy / (float)sy : .0f;
//      rep1->use_area[0] = (float)used_sx / (float)sx;
//      rep1->use_area[1] = (float)used_sy / (float)sy;
      //rep1->used_size[0] = used_sx;
      rep1->size[0] = rel_size_x;
      //rep1->used_size[1] = used_sy;
      rep1->size[1] = rel_size_y;

      rep1->tp = tp;
      if(tp)
         tp->Release();
   }else{
      rep1->real_x = 0.0f;
      rep1->real_y = 0.0f;
      memset(rep1->size, 0, sizeof(rep1->size));
      memset(rep1->use_area, 0, sizeof(rep1->use_area));
   }
   rep1->pos_x = x;
   rep1->pos_y = y;

   delete[] char_width;
   for(int i=lines.size(); i--; )
      delete lines[i].wp;
   lines.clear();

   if(xr1) *xr1 = xr;
   if(yr1) *yr1 = yr;

   rep = rep1;

   SetColor(color);
}

//----------------------------

C_wide_text::~C_wide_text(){

   C_wide_text_rep *rep1 = (C_wide_text_rep*)rep;
   delete rep1;
}

//----------------------------

void C_wide_text::Render() const{

   C_wide_text_rep *rep1 = (C_wide_text_rep*)rep;

   assert(rep1);
   if(!rep1->tp)
      return;
    
   if(!rep1->dyn_vis){
      rep1->dyn_vis = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);
      rep1->dyn_vis->Release();

      I3D_face_group fg;
      PI3D_material mat = driver->CreateMaterial();
      fg.base_index = 0;
      fg.num_faces = 2;
      mat->SetTexture(rep1->tp);
      fg.SetMaterial(mat);
      mat->Release();

      I3D_triface fc[2];
      fc[0][0] = 2;
      fc[0][1] = 0;
      fc[0][2] = 1;
      fc[1][0] = 2;
      fc[1][1] = 1;
      fc[1][2] = 3;

      struct S_vertex{
         S_vectorw v;
         dword diffuse;
         //dword specular;      //padding to 32 bytes
         I3D_text_coor tex;
      } v[4];

      float sx = igraph->Scrn_sx();
      float sy = igraph->Scrn_sy();

      float size_x = floor(rep1->size[0] * sx) - .1f;
      float size_y = floor(rep1->size[1] * sx) - .1f;
                                 //copy verts, also perform simple clipping
      for(int i=0; i<4; i++){
         v[i].tex.u = 0.0f;
         v[i].tex.v = 0.0f;
         v[i].v.x = rep1->real_x * sx;
         v[i].v.y = rep1->real_y * sx;
         v[i].v.z = 1.0f;
         v[i].v.w = 1.0f;
         v[i].diffuse = rep1->color;
         if(i&1){
            //v[i].v.x += rep1->used_size[0];
            v[i].v.x += size_x;
            v[i].tex.u = rep1->use_area[0];
         }else{
            if(v[i].v.x >= sx) return;
         }
         if(i&2){
            //v[i].v.y += rep1->used_size[1];
            v[i].v.y += size_y;
            v[i].tex.v = rep1->use_area[1];
         }else{
            if(v[i].v.y >= sy) return;
         }
      }

      rep1->dyn_vis->Build(v, 4, I3DVC_XYZRHW | I3DVC_DIFFUSE | /*I3DVC_SPECULAR | */(1<<I3DVC_TEXCOUNT_SHIFT),
         fc, 2, &fg, 1);
   }

   bool filter = driver->GetState(RS_LINEARFILTER);
   if(filter) driver->SetState(RS_LINEARFILTER, false);

   rep1->dyn_vis->Render();

   if(filter) driver->SetState(RS_LINEARFILTER, true);
}
//----------------------------

void C_wide_text::SetPos(float x, float y){

   C_wide_text_rep *rep1 = (C_wide_text_rep*)rep;
   rep1->real_x = x + (rep1->real_x - rep1->pos_x);
   rep1->pos_x = x;

   rep1->real_y = y + (rep1->real_y - rep1->pos_y);
   rep1->pos_y = y;

   rep1->dyn_vis = NULL;
}

//----------------------------

void C_wide_text::GetPos(float &x, float &y) const{

   C_wide_text_rep *rep1 = (C_wide_text_rep*)rep;
   x = rep1->pos_x;
   y = rep1->pos_y;
}

//----------------------------

void C_wide_text::SetColor(dword color){

   C_wide_text_rep *rep1 = (C_wide_text_rep*)rep;

   /*
   rep1->color.x = (float)((color>>16) & 255) / 255.0f;
   rep1->color.y = (float)((color>>8) & 255) / 255.0f;
   rep1->color.z = (float)((color>>0) & 255) / 255.0f;
   rep1->alpha = (float)((color>>24) & 255) / 255.0f;
   */
   rep1->color = color;
   rep1->dyn_vis = NULL;
}

//----------------------------

void C_wide_text::GetExtent(float &l, float &t, float &r, float &b){

   C_wide_text_rep *rep1 = (C_wide_text_rep*)rep;
   l = rep1->real_x;
   t = rep1->real_y;
   r = l + rep1->size[0];
   b = t + rep1->size[1];
}

//----------------------------
