#include "all.h"
#include <windows.h>
#include <advanced.h>         //Load/Save
#include <shellapi.h>
#include <win_reg.h>
#include "winstuff.h"
#include "resource.h"

//----------------------------

                              //where crash reports are being sent
#define CRASH_REPORT_PATH "\\\\Mike\\!In\\Crashes\\"

//----------------------------

C_log_mt_safe::C_log_mt_safe(){

   critical_section = new CRITICAL_SECTION;
   InitializeCriticalSection((LPCRITICAL_SECTION)critical_section);
}

//----------------------------

C_log_mt_safe::~C_log_mt_safe(){

   DeleteCriticalSection((LPCRITICAL_SECTION)critical_section);
   delete (LPCRITICAL_SECTION)critical_section;
}

//----------------------------

void C_log_mt_safe::WriteString(const char *cp){

   EnterCriticalSection((LPCRITICAL_SECTION)critical_section);
   //queue.insert(queue.end(), cp, cp + strlen(cp));
                              //write to log
   fwrite(cp, sizeof(char), strlen(cp), f_log);
                              //flush log
   fclose(f_log);
   f_log = fopen(log_name, "at");

   LeaveCriticalSection((LPCRITICAL_SECTION)critical_section);
}

//----------------------------

int WinSelectItem(const char *title, const char *item_list, int curr_sel){

   struct S_hlp{
      const char *title;
      const char *item_list;
      int curr_sel;
      static BOOL CALLBACK dlgListSelect(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

         switch(uMsg){
         case WM_INITDIALOG:
            {
               S_hlp *hp = (S_hlp*)lParam;
               WinInitDialog(hwnd);
               SetWindowText(hwnd, hp->title);
               const char *cp = hp->item_list;
               while(*cp){
                  SendDlgItemMessage(hwnd, IDC_ITEM_SELECT, LB_ADDSTRING, 0, (LPARAM)cp);
                  cp += strlen(cp) + 1;
               }
               SendDlgItemMessage(hwnd, IDC_ITEM_SELECT, LB_SETCURSEL, hp->curr_sel, 0);
               ShowWindow(hwnd, SW_SHOW);
            }
            return 1;
         case WM_COMMAND:
            switch(LOWORD(wParam)){
            case IDOK:
               {
                  EndDialog(hwnd, SendDlgItemMessage(hwnd, IDC_ITEM_SELECT, LB_GETCURSEL, 0, 0));
               }
               break;
            case IDCANCEL: EndDialog(hwnd, -1); break;
            }
            break;
         }
         return 0;
      }
   } hlp;
   hlp.title = title;
   hlp.item_list = item_list;
   hlp.curr_sel = curr_sel;

   return DialogBoxParam(GetModuleHandle(NULL), "IDD_LISTSELECT", (HWND)igraph->GetHWND(),
      (DLGPROC)S_hlp::dlgListSelect, (LPARAM)&hlp);
}

//----------------------------

bool WinGetName(const char *title, C_string &str){

   struct S_hlp{
      const char *title;
      C_string *str;
      static BOOL CALLBACK dlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

         switch(uMsg){
         case WM_INITDIALOG:
            {
               S_hlp *hp = (S_hlp*)lParam;
               SetWindowLong(hwnd, GWL_USERDATA, lParam);
               WinInitDialog(hwnd);
               SetWindowText(hwnd, hp->title);
               SetDlgItemText(hwnd, IDC_EDIT, *hp->str);
               ShowWindow(hwnd, SW_SHOW);
            }
            return 1;
         case WM_COMMAND:
            switch(LOWORD(wParam)){
            case IDOK:
               {
                  S_hlp *hp = (S_hlp*)GetWindowLong(hwnd, GWL_USERDATA);
                  char buf[512];
                  GetDlgItemText(hwnd, IDC_EDIT, buf, sizeof(buf));
                  *hp->str = buf;
                  EndDialog(hwnd, 1);
               }
               break;
            case IDCANCEL: EndDialog(hwnd, 0); break;
            }
            break;
         }
         return 0;
      }
   } hlp;
   hlp.title = title;
   hlp.str = &str;

   return DialogBoxParam(GetModuleHandle(NULL), "IDD_GETNAME", (HWND)igraph->GetHWND(),
      (DLGPROC)S_hlp::dlgProc, (LPARAM)&hlp);
}

//----------------------------

void WinInitDialog(void *hwnd){
                              //center dialog
   RECT rc;
   GetWindowRect((HWND)hwnd, &rc);
   int x = rc.left + (igraph->Scrn_sx() - (rc.right-rc.left)) / 2;
   int y = Max(0, rc.top  + ((int)igraph->Scrn_sy() - (rc.bottom-rc.top)) / 2);
   SetWindowPos((HWND)hwnd, NULL, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

//----------------------------

bool OsGetFileTimeDate(const char *fname, S_date_time &dt){

   dword ftime[2];
   if(!GetFileTime(fname, ftime))
      return false;
   SYSTEMTIME st;
   FileTimeToSystemTime((const FILETIME*)ftime, &st);
   dt.year = st.wYear;
   dt.month = st.wMonth;
   dt.day = st.wDay;
   dt.hour = st.wHour;
   dt.minute = st.wMinute;
   dt.second = st.wSecond;

   return true;
}

//----------------------------

bool __stdcall WinSendCrashReport(const char *log_dump){

#if defined _DEBUG && 1
                                 //get name to be saved
   C_string comp_name;
   OsGetComputerName(comp_name);
   SYSTEMTIME st;
   GetLocalTime(&st);

   C_fstring name_base("%s%s %i-%.2i-%.2i.err",
      CRASH_REPORT_PATH,
      (const char*)comp_name, st.wYear, st.wMonth, st.wDay);
   C_string name;
   for(int i=0; i<1000; i++){
      name = C_fstring("%s - %.3i", (const char*)name_base, i);
      if(Load(name, NULL, 0)==-1){
         if(Save(name, (const char*)log_dump, strlen(log_dump)) != -1){
            return true;
         }
      }
   }
   return false;
#else
   {
      static const char header[] =
         "mailto:hddsupport@lonelycatgames.com?subject=Bug Report&body="
         "Problem description:\f"
         "(please put few words about when this crash happened)\f\f\f"
         ;
      char *buf = new char[strlen(log_dump) + sizeof(header)];
      strcpy(buf, header);
      strcat(buf, log_dump);
                              //convert '\n' into '\f'
      for(int i=strlen(buf); --i >= sizeof(header)-1; ){
         if(buf[i]=='\n')
            buf[i] = '\f';
      }
      HINSTANCE hi=
         ShellExecute(NULL, NULL, buf, NULL, NULL, SW_SHOWDEFAULT);
      delete[] buf;
      if(((dword)hi) >= 32)
         return true;
                              //some error occured
      MessageBox(NULL,
         "Cannot execute mail client.\nPlease save report to file and send manually.\nThanks for making this product better.",
         "Sending e-mail", MB_OK);
   }
   return false;
#endif
}

//----------------------------

bool IsSpecialKeyDown(){
   return ((GetAsyncKeyState(VK_CONTROL)&0x8000) && (GetAsyncKeyState(VK_MENU)&0x8000));
}

//----------------------------

#ifdef EDITOR

PI3D_texture CreateStringTexture(const char *cp, int size1[2], PI3D_driver drv, void *winHFONT){

   HFONT h_font = (HFONT)winHFONT;
   if(!h_font)
      return NULL;

   HWND hwnd = (HWND)drv->GetGraphInterface()->GetHWND();
   SIZE size;
   {
      HDC hdc = GetDC(hwnd);
      SelectObject(hdc, h_font);
      GetTextExtentPoint32(hdc, cp, strlen(cp), &size);
      ReleaseDC(hwnd, hdc);
   }
   size1[0] = size.cx;
   size1[1] = size.cy;
   int i;
   for(i=0; i<9; i++) if((1<<i) >= size.cx){ size.cx = (1<<i); break; }
   for(i=0; i<9; i++) if((1<<i) >= size.cy){ size.cy = (1<<i); break; }

   PI3D_texture tp = NULL;
   {
      void *bitmap_bits = NULL;

      HDC hdc = CreateCompatibleDC(NULL);

      BITMAPINFO bmi;
      memset(&bmi, 0, sizeof(bmi));
      bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
      bmi.bmiHeader.biWidth = size.cx;
      bmi.bmiHeader.biHeight = size.cy;
      bmi.bmiHeader.biPlanes = 1;
      bmi.bmiHeader.biBitCount = 24;
      bmi.bmiHeader.biCompression = BI_RGB;
      bmi.bmiHeader.biSizeImage = 0;

      HBITMAP h_bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, &bitmap_bits, NULL, 0);
      SetMapMode(hdc, MM_TEXT);
      SelectObject(hdc, h_bm);

                              //paint into the bitmap
      SetTextAlign(hdc, TA_TOP | TA_LEFT);
      SelectObject(hdc, h_font);
      {
         HBRUSH hb = CreateSolidBrush(0x00101010);
         RECT rc = {0, 0, size1[0], size1[1]};
         FillRect(hdc, &rc, hb);
         DeleteObject(hb);
      }
      SetBkMode(hdc, TRANSPARENT);
      //SetBkColor(hdc, (COLORREF)0);
      //SetTextColor(hdc, 0x00101010);
      //TextOut(hdc, 1, 1, cp, strlen(cp));
      SetTextColor(hdc, 0xffffff);
      TextOut(hdc, 0, 0, cp, strlen(cp));

                              //convert to the texture
      {
         S_pixelformat pf;
         I3D_RESULT ir = drv->FindTextureFormat(pf, FINDTF_ALPHA1);
         if(I3D_SUCCESS(ir)){
                              //create texture of desired resolution and pixelformat
            I3D_CREATETEXTURE ct;
            memset(&ct, 0, sizeof(ct));
            ct.flags = TEXTMAP_TRANSP | TEXTMAP_NOMIPMAP | TEXTMAP_USEPIXELFORMAT |
               TEXTMAP_DIFFUSE | TEXTMAP_MEM_PTR;
            ct.pf = &pf;
            bmi.bmiHeader.biCompression = (long)bitmap_bits;
            ct.text_mem = &bmi;
            ct.size_x = size.cx;
            ct.size_y = size.cy;

            ir = drv->CreateTexture(&ct, &tp);
            /*
            if(I3D_SUCCESS(ir)){
               LPIImage img = tp->GetImage(false);
               img->CKeyToAlpha(img->GetColorKey());
            }
            */
         }
      }

      DeleteObject(h_bm);
      DeleteDC(hdc);
   }
   return tp;
}

//----------------------------

void DrawStringTexture(PI3D_frame frm, float y_screen_displace, PI3D_scene scene, PI3D_texture tp, const int size[2],
   const S_vector &color, float min_dist, float max_dist, float opacity, float y_world_displace,
   const S_vector *pos1){

                              //check if visual was rendered
                              // but only if visuals are drawn
   if(driver->GetState(RS_DRAWVISUALS)){
      PI3D_frame vis_parent = frm;
      while(vis_parent && vis_parent->GetType()!=FRAME_VISUAL) vis_parent = vis_parent->GetParent();
      if(vis_parent){
         dword last_frm_render_time = I3DCAST_VISUAL(vis_parent)->GetLastRenderTime();
         dword last_scn_render_time = scene->GetLastRenderTime();
                              //don't wast time on invisible visuals
         if(last_frm_render_time != last_scn_render_time)
            return;
      }
   }

   assert(tp && size);

   scene->SetRenderMatrix(GetIdentityMatrix());
   S_vectorw scrn_pos;
   S_vector pos;
   if(frm){
      const S_matrix &mat = frm->GetMatrix();
      pos = mat(3);
                              //adjust position, if necessary
      const I3D_bbox *bbox = NULL;
      switch(frm->GetType()){
      case FRAME_VISUAL:
         bbox = &I3DCAST_VISUAL(frm)->GetBoundVolume().bbox;
         break;
      case FRAME_MODEL:
         bbox = &I3DCAST_MODEL(frm)->GetHRBoundVolume().bbox;
         break;
      case FRAME_DUMMY:
         bbox = I3DCAST_DUMMY(frm)->GetBBox();
         break;
      }
      if(bbox && bbox->IsValid())
         pos += mat(1) * bbox->max.y;
      pos += mat(1) * y_world_displace;
   }else{
      if(!pos1)
         return;
      pos = *pos1;
      pos.y += y_world_displace;
   }

                              //blend at least by this amount, so that texture alpha is used
   byte alpha = Max(1, (int)(255.0f - 255.0f * opacity));
                              //check dist to camera
   float dist = (pos - scene->GetActiveCamera()->GetWorldPos()).Magnitude();
   if(dist>=max_dist) return;
   if(dist>=min_dist){
      alpha += (opacity * 255.0f * (dist-min_dist) / (max_dist-min_dist));
   }
   scene->TransformPoints(&pos, &scrn_pos, 1);

                              //check z
   if(scrn_pos.z < CAM_NEAR_RANGE)
      return;

   scrn_pos.y += y_screen_displace;

   struct S_vertex{
      S_vectorw v;
      I3D_text_coor tex;
   } v[4];
   int sx = driver->GetGraphInterface()->Scrn_sx();
   int sy = driver->GetGraphInterface()->Scrn_sy();

   float scale = scrn_pos.w * 2.0f;
   float fsize[2] = {size[0] * scale, size[1] * scale};
                              //copy verts, also perform simple clipping
   for(int i=0; i<4; i++){
      v[i].tex.u = 0.0f;
      v[i].tex.v = 0.0f;
      v[i].v = scrn_pos;
      v[i].v.x -= fsize[0]/2;
      v[i].v.y -= fsize[1];
      if(i&1){
         v[i].v.x += fsize[0];
         if(v[i].v.x <= 0) return;
         v[i].tex.u = (float)size[0] / (float)tp->SizeX();
      }else{
         if(v[i].v.x >= sx) return;
      }
      if(i&2){
         v[i].v.y += fsize[1];
         if(v[i].v.y <= 0) return;
         v[i].tex.v = (float)size[1] / (float)tp->SizeY();
      }else{
         if(v[i].v.y >= sy) return;
      }
   }
   driver->SetTexture(tp);
   static const word indx[] = {0, 1, 2, 2, 1, 3};
   scene->DrawTriangles(&v[0].v, 4, indx, sizeof(indx)/sizeof(word), color,
      sizeof(S_vertex), I3DDRAW_TRANSFORMED | alpha,
      &v[0].tex);
}

//----------------------------

#endif

//----------------------------

