#include "all.h"
#include "tables.h"
#include "net.h"
#include "mainmenu.h"


#define MENU_HINT_SIZE .0285f
#define DOUBLE_CLICK_TIME 250 //in ms

#define ENABLE_NET_SYNC_MENU  //display network synchronization progress (menu)

//----------------------------

C_string MakeGameText(const C_string &str);
C_string AddGameText(const C_string &str1, const C_string &str2);
bool ScaleObject(PI3D_frame frm, PI3D_mesh_level ml_src, const S_vector &scl);

//----------------------------

inline int F2I(float f){ return *(dword*)&f; }

bool LoadSoldierTable(LPC_table tab, int man_indx);

//----------------------------
//----------------------------

float CountManCarriage(LPC_table tab);

//----------------------------
//----------------------------

                              //callback messages
enum{
   CB_POINTED_RES_CHANGE,     //(S_resource *curr_res, S_resource *prev_res) - NULL = no resource
   CB_ENUM_RESOURCES,         //(0, dword context)
   CB_ENUM_CALL,              //(S_resource*); ret = (bool)stop enumeration
   CB_CLICK,                  //(S_resource*, int phase <1=press, 0=release> );
   CB_KEY,                    //(char key)
   CB_SELECT_IMAGE,           //(S_resource*, bool sound); ret: bool ok
   CB_SELECT_MAN,             //(S_resource*, bool sound); ret: bool ok
   CB_UNSELECT,               //(bool export_net); ret: bool ok
   CB_ADD_SELECTED,           //ret: bool ok
   CB_ADD_SELECTED1,          //(bool add_ammo | (bool snd<<16), int num_ammo); ret: bool ok
   CB_ADD_SELECTED_MAN,       //ret: bool ok
   CB_REMOVE_SELECTED,        //ret: bool ok
   CB_REMOVE_SELECTED1,       //(bool add_ammo | (bool snd<<16), int num_ammo); ret: bool ok
   CB_REMOVE_SELECTED_MAN,    //ret: bool ok
   CB_SCROLL,                 //(bool up, bool snd); ret: bool ok
   CB_SCROLL_GRID,            //(bool up, bool snd); ret: bool ok
   CB_REMOVE_ALL,             //ret: bool ok
};

enum{                         //button sounds
   GM_SND_BUTT_DOWN,
   GM_SND_BUTT_UP,
   GM_SND_HINT,
   GM_SND_HLIGHT,
   GM_SND_SELECT,
   GM_SND_CANNOT,
   GM_SND_SCROLL,
   GM_SND_ADD,
   GM_SND_REMOVE,
};

enum E_MENU_BUTTON{
   ID_BUTTON_OK,
   ID_BUTTON_BACK,
   ID_BUTTON_CAMP_ANIM,
   ID_BUTTON_MISSION_STATUS,
   ID_BUTTON_MAN_ADD,
   ID_BUTTON_MAN_REMOVE,
   ID_BUTTON_AUTO_SELECT,
   ID_BUTTON_UP,
   ID_BUTTON_DOWN,
   ID_BUTTON_UP1,
   ID_BUTTON_DOWN1,
   ID_BUTTON_ITEM_ADD,
   ID_BUTTON_ITEM_REMOVE,
   ID_BUTTON_AMMO_ADD,
   ID_BUTTON_AMMO_REMOVE,
   ID_BUTTON_SOLDIER_INFO,
   ID_BUTTON_MISSION_LOAD,
};

                              //general resource flags
#define RFLG_MAN_40     1
#define RFLG_MAN_8      2
#define RFLG_MAN_4      4
#define RFLG_MAN_PHOTO  8
#define RFLG_SELECTION_IMAGE  0x10
#define RFLG_MAIN_PIC         0x20
#define RFLG_CAMPAIGN_PHOTO   0x40
#define RFLG_ITEM             0x80
#define RFLG_ITEM_GRID        0x100

#define RFLG_HINT       0x10000
#define RFLG_BUTTON     0x20000
#define RFLG_HIGHLIGHT  0x40000
#define RFLG_INVISIBLE  0x80000
#define RFLG_ARROW_UP   0x100000
#define RFLG_ARROW_DN   0x200000

//----------------------------
#define HINT_SIZE .04f
                              //general menu implementation
class C_menu_base1: public C_mission{
protected:
   struct S_resource_def{
      const char *name;
      int id;
      dword flags;
   };

   struct S_resource{
      PI3D_frame frm;
      int id;
      dword flags;
   };
private:

   E_GAME_MENU menu_id;
   int last_mx, last_my;
   bool cursor_dirty;
   S_resource *last_pointed_res;
   S_resource *curr_clicked_button;
   S_resource *last_pointed_button;
   dword last_button_state;
   dword last_click_time;
                              //sound management
   struct S_sound_entry{
      C_smart_ptr<ISND_sound> snd;
      int id;
      float volume;
      S_sound_entry():
         volume(1.0f)
      {}
   };
   vector<S_sound_entry> sound_list;
protected:

   C_poly_text texts;

   C_smart_ptr<C_poly_sprite> spr_arrow;
   int curr_arrow_shape;

                              //internal network messages
   enum E_NETWORK_MESSAGE_GM{

      NM_MENU_MOUSE_POS = 0x1000,
         //float x, y

      NM_GM_SYNC = 0x1010,
         //sync class data

      NM_CBPROC = 0x1020,
         //byte cb_msg, dword prm1, dword prm2

      NM_CBPROC_R1 = 0x1030,
         //byte cb_msg, byte res_indx, bits:1 prm2

      NM_MENU_CHAT = 0x1040,
         //C_net_chat data
   };

   enum E_NET_SYNC_EVENT{

      NS_NULL,
      NS_BACK_BUTTON,
      NS_PROCEED,
      NS_LOAD_CAMPANIM,
      NS_SOLDIER_INFO,
   };

//----------------------------

   void InitSounds(){

      if(!isound)
         return;

      static const S_resource_def ss_list[] = {
         {"butt_dn", GM_SND_BUTT_DOWN},
         {"hint",    GM_SND_HINT, 30},
         {"hlight",  GM_SND_HLIGHT},
         {"select",  GM_SND_SELECT},
         {"cannot",  GM_SND_CANNOT},
         {"scroll",  GM_SND_SCROLL},
         {"add",     GM_SND_ADD},
         {"remove",  GM_SND_REMOVE},
         {NULL}
      };
      const S_resource_def *s_list = ss_list;

      while(s_list->name){
         PISND_source src;
         isound->CreateSource(&src);
         if(!src)
            return;
         ISND_RESULT ir = ISNDERR_NOFILE;
         for(int ii=0; ii<sound_cache.NumDirs(); ii++){
            const char *snd_dir = sound_cache.GetDir(ii);
            C_fstring fname("%sgamemenu\\%s", snd_dir, s_list->name);
            ir = src->Open(C_fstring("%s.ogg", (const char*)fname), ISND_NOPANNING);
            if(ir==ISNDERR_NOFILE)
               ir = src->Open(C_fstring("%s.wav", (const char*)fname), ISND_NOPANNING);
            if(ISND_SUCCESS(ir))
               break;
         }
         if(ISND_FAIL(ir))
            ErrReport(C_fstring("unable to open sound %s", s_list->name), editor);
         sound_list.push_back(S_sound_entry());
         S_sound_entry &se = sound_list.back();
         PISND_sound snd;
         isound->CreateSound(&snd);
         se.snd = snd;
         snd->Release();

         se.snd->Open(src);
         src->Release();
         se.id = s_list->id;
         if(s_list->flags)
            se.snd->SetVolume((float)s_list->flags/100.0f);
         ++s_list;
      }
   }

//----------------------------

   void PlaySound(int id){
      if(!isound) return;
      for(int i=sound_list.size(); i--;)
      if(sound_list[i].id==id){
         sound_list[i].snd->Play();
         break;
      }
      if(i==-1)
         ErrReport(C_fstring("Can't play sound (id %i)", id), editor);
   }

//----------------------------
                              //internal resource list
   vector<S_resource> res_list;

//----------------------------

   void InitResources(const S_resource_def *r_list){

      for(; r_list->name; ++r_list){
         PI3D_frame frm = scene->FindFrame(r_list->name);
         if(!frm){
            //ErrReport(C_fstring("can't find resource: %s", r_list->name), editor);
         }else{
            res_list.push_back(S_resource());
            S_resource &res = res_list.back();
            res.frm = frm;
            res.id = r_list->id;
            res.flags = r_list->flags;
                              //initialize special resources
            if(res.flags&RFLG_BUTTON){
               if(frm->NumChildren()==1){
                  PI3D_frame chld(frm->GetChild(0));
                  chld->SetOn(false);
                  chld->SetPos(frm->GetPos());
                  chld->SetRot(frm->GetRot());
                  chld->SetScale(frm->GetScale());
                  chld->Update();
               }
            }
            if(res.flags&RFLG_INVISIBLE){
               frm->SetOn(false);
            }
         }
      }
   }

//----------------------------

   void Map3DTo2D(const S_vector &in, float &x, float &y) const{

      S_vectorw v;
      scene->TransformPoints(&in, &v, 1);
      x = v.x / igraph->Scrn_sx();
      y = v.y / igraph->Scrn_sx();
   }

//----------------------------

   void SetHLight(const S_resource *res, bool on){
      if(res){
         int id = res->id;
         if(res->flags&RFLG_BUTTON) id += 100;
         const S_resource *hgl = FindResource(id, RFLG_HIGHLIGHT);
         //assert(hgl);
         if(hgl)
            hgl->frm->SetOn(on);
      }
   }

   void ShowButton(int id, bool on){
      const S_resource *but = FindResource(id, RFLG_BUTTON);
      if(but){
         but->frm->SetOn(on);
         if(!on){
            SetHLight(but, false);
            if(last_pointed_res==but){
               cbProc(CB_POINTED_RES_CHANGE, NULL, (dword)last_pointed_res);
               if(last_pointed_res->flags&RFLG_HINT)
                  HintManage(NULL);
               last_pointed_res = NULL;
            }
         }
      }else
         ErrReport(C_fstring("Can't find button id %i", id), editor);
   }

                              //hint management
   C_smart_ptr<C_hint_base> hint;
   C_smart_ptr<C_text> menu_desc;

private:
   S_resource *last_hint_res;
   int hint_on_count;

   void HintManage(S_resource *res, int time = 0, float mx = 0, float my = 0){

      if(!res){
         if(hint)
            hint->FadeOff();
         last_hint_res = NULL;
      }else
      if(last_hint_res!=res){
         last_hint_res = res;
         hint_on_count = 500;
      }else
      if(hint_on_count && (hint_on_count -= time) <= 0){
         hint_on_count = 0;
                              //time to show the hint
         dword hint_color = -1;
         bool wide;
         const C_string &str = GetHintMessage(res, hint_color, wide);
         if(str.Size()){
            float sx = igraph->Scrn_sx();
            float x = (float)mx / sx, y = (float)my/sx - .02f;
            hint = CreateHint(str, wide, &texts, x, y, MENU_HINT_SIZE, hint_color);
            hint->Release();
            //PlaySound(GM_SND_HINT);
         }
      }
   }

public:
//----------------------------
// Initialize class, return false if failed.
   virtual bool C_menu_base1::Init(E_GAME_MENU id, dword = 0){

      menu_id = id;

      net_sync = CreateNetSync(NM_GAME_MENU+menu_id, NM_GM_SYNC, 16);
      net_sync->Release();

      E_MISSION_IO mio = Open(C_fstring("menu\\%i", id), OPEN_USE_SCENE_CAMERA);
      if(mio!=MIO_OK)
         return false;

      bool b;
      b = texts.Init();
      if(!b)
         return false;

      spr_arrow = new C_poly_sprite(0, 0, .03f, &icons_img, curr_arrow_shape = 0);
      spr_arrow->Release();
      spr_arrow->SetScreenPos(igraph->Mouse_x(), igraph->Mouse_y());

      PI3D_camera cam = scene->GetActiveCamera();
      if(!cam)
         return false;

      InitSounds();

      {
         float x = .885f;
         float y = .075f;
         const float size = .0337f;
         PI3D_frame frm = scene->FindFrame("description");
         if(frm)
            Map3DTo2D(frm->GetWorldPos(), x, y);
         y /= (screen_aspect_ratio * 1.33333333f);

         int txt_id = GT_MENU_DESCRIPTION + menu_id - 1;
         menu_desc = texts.CreateText(all_txt[txt_id],
            0xffffffff, x, y, size,
            TEXT_JUSTIFY_CENTER | TEXT_JUSTIFY_CONVERT_EOL | TEXT_JUSTIFY_CLIP_TO_SCREEN,
            all_txt.IsWide(txt_id),
            NULL, NULL, NULL, NULL, 0, PI*.05f);
         menu_desc->Release();
      }
      return true;
   }

//----------------------------

                              //network messaging, in_msg=received message, out_msg=feedbac message
   virtual void NetIn(E_NETWORK_MESSAGE msg_id, PC_net_in_msg in_msg);

//----------------------------

   virtual void Render(){

      driver->BeginScene();

      C_mission::Render();

      bool use_zb = driver->GetState(RS_USEZB);
      if(use_zb) driver->SetState(RS_USEZB, false);

      texts.Render();

      spr_arrow->Render();
      if(sync_menu)
         sync_menu->Render();

      if(use_zb) driver->SetState(RS_USEZB, true);

      driver->EndScene();
   }
                 
//----------------------------

   E_GAME_MENU GetType(){ return menu_id; }

//----------------------------

   C_menu_base1():
      net_chat(NM_MENU_CHAT),
      hint_on_count(0), last_hint_res(NULL),
      last_mx(igraph->Scrn_sx()/2), last_my(igraph->Scrn_sy()/2),
      cursor_dirty(true),
      last_pointed_res(NULL),
      curr_clicked_button(NULL),
      last_pointed_button(NULL),
      last_button_state(0),
      //state(GMS_NULL),
      last_click_time(-1000),
      menu_id((E_GAME_MENU)0)
   {
   }

//----------------------------
protected:

                              //network support
   C_smart_ptr<C_net_sync> net_sync;
   C_smart_ptr<C_sync_menu> sync_menu;

                              //network chatting
   C_net_chat net_chat;
public:
   int ResToInt(const S_resource *res) const{
      return (res - res_list.begin());
   }

   virtual dword cbProc(dword msg, dword prm1 = 0, dword prm2 = 0);

//----------------------------

   virtual void Tick(const S_tick_context &tc);

//----------------------------

   const S_resource *FindResource(int id, dword flags) const{
      for(int i=0; i<res_list.size(); i++)
      if(res_list[i].id==id && (res_list[i].flags&flags)==flags){
         return &res_list[i];
      }
      return NULL;
   }

//----------------------------

   void SetResourceFrame(const S_resource *res, PI3D_frame frm){
      ((S_resource*)res)->frm = frm;
   }

//----------------------------

   static PI3D_texture ChangeTexture(PI3D_frame frm, const char *txt_name, const char *dir = NULL){

      assert(frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_OBJECT);
                              //put image onto rectangle
      C_fstring str("%s\\%s.png", dir ? dir : "gamemenu", txt_name);
      I3D_CREATETEXTURE ct;
      ct.flags = TEXTMAP_DIFFUSE | TEXTMAP_NOMIPMAP;
      ct.text_name = str;
      PI3D_texture tp = NULL;
      if(I3D_FAIL(driver->CreateTexture(&ct, &tp))){
         ct.text_name = "unknown.png";
         driver->CreateTexture(&ct, &tp);
      }
      if(tp){
         PI3D_object obj = I3DCAST_OBJECT(frm);
         PI3D_mesh_level ml = obj->GetMeshObject()->GetLOD(0);
         PI3D_face_group fg = ml->GetFGroup(0);
         PI3D_material mp1 = fg->GetMaterial();
         PI3D_material mp = driver->CreateMaterial();
         mp->Duplicate(mp1);
         fg->SetMaterial(mp);
         mp->SetTexture(tp);
         mp->Release();
         tp->Release();
      }
      return tp;
   }

private:
   S_resource *ManageButton(S_resource *new_id){
      S_resource *ret = NULL;
      for(int i=0; i<2; i++){
         if(curr_clicked_button){
                              //hide previouse button
            PI3D_frame frm = curr_clicked_button->frm;
            if(frm->NumChildren()==1){
               PI3D_frame chld(frm->GetChild(0));
               chld->LinkTo(frm->GetParent());
               chld->SetOn(true);
               frm->LinkTo(chld);
               frm->SetOn(false);
               curr_clicked_button->frm = chld;
            }
         }
         if(!i){
            ret = curr_clicked_button;
            curr_clicked_button = new_id;
         }
      }
      return ret;
   }

//----------------------------

   virtual void GetTextExtents(float &l, float &t, float &sx, float &sy){

      l = 0.0f;
      t = 0.0f;
      sx = 1.0f;
      sy = .75f;
   }

//----------------------------
protected:

   bool GetItemRect(PI3D_frame frm, float *fr){

      if(!(frm->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_OBJECT))
         return false;

      PI3D_object obj = I3DCAST_OBJECT(frm);
      const S_matrix &m = obj->GetMatrix();
      const float BORDER = .0015f;
      //const float &k = igraph->Scrn_sy();
      const float &k = igraph->Scrn_sx();

      const I3D_bbox *bbox;
      S_vector bb[2];

      bbox = &obj->GetBoundVolume().bbox;
      bb[0] = (*bbox)[0]*m;
      bb[1] = (*bbox)[1]*m;

      S_vectorw bb4[2];
      scene->TransformPoints(bb, bb4, 2);
      bb[0] = bb4[0];
      bb[1] = bb4[1];

      fr[0] = Min(bb[0].x, bb[1].x);
      fr[2] = Max(bb[0].x, bb[1].x);
      fr[0] /= k, fr[2] /= k;
      fr[0] += BORDER, fr[2] -= BORDER;
      //fr[2] -= fr[0];

      fr[1] = Min(bb[0].y, bb[1].y);
      fr[3] = Max(bb[0].y, bb[1].y);
      fr[1] /= k, fr[3] /= k;
      fr[1] += BORDER, fr[3] -= BORDER;
      //fr[3] -= fr[1];

      fr[0] += .01f;
      fr[2] -= .02f;

      /*
      {
         float f = screen_aspect_ratio * 1.33333333f;
         fr[1] /= f;
         fr[3] /= f;
      }
      */
      return true;
   }

//----------------------------
protected:

   bool HasAmmo(int itm_id) const{
      switch(tab_inventory->ItemI(TAB_I_INV_ID, itm_id)){
      case INV_WEAPON:
      case INV_COLT:
      case INV_BAZOOKA:
      //case INV_PANZER:
      case INV_FLAREGUN:
         return true;
      default: return false;
      }
   }
private:
   S_resource *FindRes(PI3D_frame frm){
      if(!frm) return NULL;
      while(true){
         for(int i=res_list.size(); i--; )
         if(res_list[i].frm==frm) return &res_list[i];
                           //find model parent
         PI3D_frame prnt=frm->GetParent();
         while(prnt && prnt->GetType()!=FRAME_MODEL) prnt=prnt->GetParent();
         if(!prnt) return NULL;
         frm=prnt;
      }
   }

//----------------------------

   void MouseUpdate(int mx, int my, int time = 0, bool force = false){

      if(force || last_mx!=mx || last_my!=my){
         spr_arrow->SetScreenPos(mx, my / (screen_aspect_ratio * 1.33333333f));

                                 //check what's under cursor
         S_vector pos, dir;
         scene->UnmapScreenPoint(mx, my, pos, dir);
         PI3D_frame frm = NULL;
         {
            I3D_collision_data cd(pos, dir, I3DCOL_TEXTURE | I3DCOL_RAY | I3DCOL_EXACT);
            if(scene->TestCollision(cd))
               frm = cd.GetHitFrm();
         }

         S_resource *res = FindRes(frm);

         {
            I3D_collision_data cd(pos, dir, I3DCOL_TEXTURE | I3DCOL_RAY | I3DCOL_INVISIBLE | I3DCOL_EXACT);
            if(scene->TestCollision(cd))
               frm = cd.GetHitFrm();
            else
               frm = NULL;
         }

         S_resource *res_i = FindRes(frm);
         if(res_i && !(res_i->flags&RFLG_INVISIBLE)) res_i=NULL;
         if(res_i != res && res_i && (!res || !(res->flags&RFLG_BUTTON))) res = res_i;

         if(last_pointed_res != res){
            cbProc(CB_POINTED_RES_CHANGE, (dword)res, (dword)last_pointed_res);
            last_pointed_res = res;
            int ai = 0;
            if(last_pointed_res){
               if(last_pointed_res->flags&RFLG_ARROW_UP) ai = 5;
               if(last_pointed_res->flags&RFLG_ARROW_DN) ai = 6;
               if(last_pointed_res->flags&RFLG_HINT)
                  HintManage(last_pointed_res, 0, mx, my);
            }else
               HintManage(NULL);
            if(ai!=curr_arrow_shape){
               S_vector2 spos = spr_arrow->GetScreenPos();
               spr_arrow = new C_poly_sprite(0, 0, .03f, &icons_img, curr_arrow_shape = ai);
               spr_arrow->SetScreenPos(spos.x, spos.y);
               spr_arrow->Release();
            }
         }
         last_mx = mx, last_my = my;
      }
   }

//----------------------------

   virtual dword GetID() const{ return 'MENU'; }

//----------------------------
protected:

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){
      case ID_BUTTON_OK:
         if(net)
            BeginSyncEvent(NS_PROCEED);
         else
            Proceed();
         break;

      case ID_BUTTON_BACK:
         if(net)
            BeginSyncEvent(NS_BACK_BUTTON);
         else
            ReturnBack();
         break;
      }
   }

//----------------------------

   virtual void OnDoubleClick(S_resource *res){
   }

//----------------------------

   virtual C_string GetHintMessage(S_resource *res, dword &ret_color, bool &wide){

      if(res->flags&RFLG_BUTTON){
         int tid = GT_GAME_MENU_BUTTONS+res->id;
         wide = all_txt.IsWide(tid);
         return all_txt[tid];
      }
      return NULL;
   }

//----------------------------
// Proceed forward (OK button).
   virtual void Proceed(){
      assert(0);
   }

//----------------------------
// Proceed back (BACK button).
   virtual void ReturnBack(){
      assert(0);
   }

//----------------------------

   bool BeginSyncEvent(E_NET_SYNC_EVENT event, dword data = 0){

      assert(net && net_sync);
      assert(net->IsHost());
      assert(data<256);
      bool ok = net_sync->BeginEvent(event | (data<<8));
      if(ok) NetSend();
      return ok;
   }

//----------------------------

   virtual void OnNetSyncEvent(E_NET_SYNC_EVENT event, dword data){

      switch(event){
      case NS_PROCEED:
         Proceed();
         break;

      case NS_BACK_BUTTON:
         ReturnBack();
         break;

      default:
         assert(0);
      }
   }

//----------------------------

   virtual void OnPlayerDisconnect(dword pid, const char *name){

      net_sync->OnPlayerDisconnect(pid);
      C_mission::OnPlayerDisconnect(pid, name);
   }

//----------------------------

   virtual void OnSessionLost(){

                              //go to Main Menu
      game_configuration.net_host = false;
      game_configuration.net_join = false;
      RenderAllBackbuffers();
      tick_class = CreateMainMenu();
      if(tick_class)
         tick_class->Release();
   }
};

//----------------------------

dword C_menu_base1::cbProc(dword msg, dword prm1, dword prm2){

   int i;
   switch(msg){

   case CB_SELECT_IMAGE:
   case CB_SELECT_MAN:
      if(net && net->IsHost()){
         const S_resource *res = (const S_resource*)prm1;
                        //propagate to others
         const bool g = true;
         net_msg->Ascend(NM_GAME_MENU+menu_id);
         net_msg->AddID(NM_CBPROC_R1, g);
         net_msg->AddByte(msg, g);
         net_msg->AddByte(ResToInt(res), g);
         net_msg->AddBits(prm2, 1, g);
         net_msg->Descend();
      }
      return true;
                           //controlling messages - send over network
   case CB_UNSELECT:
   case CB_ADD_SELECTED:
   case CB_REMOVE_SELECTED:
   case CB_ADD_SELECTED1:
   case CB_REMOVE_SELECTED1:
   case CB_ADD_SELECTED_MAN:
   case CB_REMOVE_SELECTED_MAN:
   case CB_SCROLL:
   case CB_SCROLL_GRID:
   case CB_REMOVE_ALL:
      if(net && net->IsHost()){
         const bool g = true;
         net_msg->Ascend(NM_GAME_MENU+menu_id);
         net_msg->AddID(NM_CBPROC, g);
         net_msg->AddByte(msg, g);
         net_msg->AddDword(prm1, g);
         net_msg->AddDword(prm2, g);
         net_msg->Descend();
      }
      return true;

   case CB_ENUM_RESOURCES:
      for(i=res_list.size(); i--; )
         cbProc(CB_ENUM_CALL, (dword)&res_list[i], prm2);
      break;

   case CB_POINTED_RES_CHANGE:
      {
         S_resource *res = (S_resource*)prm1;
         if(last_pointed_button){
            const S_resource *hgl = FindResource(last_pointed_button->id+100, RFLG_HIGHLIGHT);
            if(hgl) hgl->frm->SetOn(false);
         }
         if(!res || (res->flags&RFLG_BUTTON)){
            last_pointed_button = res;
            if(last_pointed_button){
               const S_resource *hgl = FindResource(last_pointed_button->id+100, RFLG_HIGHLIGHT);
               if(hgl) hgl->frm->SetOn(true);
            }
         }
      }
   }
   return 0;
}

//----------------------------

void C_menu_base1::Tick(const S_tick_context &tc){

   AddRef();

   C_mission::Tick(tc);
   if(hint && !hint->Tick(tc.time))
      hint = NULL;
   dword mouse_butt = tc.mouse_buttons;
   int mx = igraph->Mouse_x();
   int my = igraph->Mouse_y();

   if(!net || net->IsHost()){
      if(net){
         if(last_mx!=mx || last_my!=my){
                              //propagate mouse update
            const bool g = false;
            net_msg->Ascend(NM_GAME_MENU+menu_id);
            net_msg->AddID(NM_MENU_MOUSE_POS, g);
            net_msg->AddFloat((float)mx/igraph->Scrn_sy(), g);
            net_msg->AddFloat((float)my/igraph->Scrn_sy(), g);
            net_msg->Descend();
         }
      }
      MouseUpdate(mx, my, tc.time, cursor_dirty);
      cursor_dirty = false;
   }

   if(last_pointed_res){
      if(last_pointed_res->flags&RFLG_HINT){
         S_vector2 pos = spr_arrow->GetScreenPos();
         HintManage(last_pointed_res, tc.time, pos.x, pos.y);
      }
   }
                           //blink net chat
   net_chat.Tick(tc, (tc.key!=K_NOKEY));


   if(net && net_sync->GetState()){
      dword net_event;
      switch(net_sync->GetState(NULL, &net_event)){
      case C_net_sync::SS_RESOLVED:
         {
            if(sync_menu){
               sync_menu->AddAllCheckMarks();
               RenderScreen();
               sync_menu = NULL;
            }

            NET_LOG(C_fstring("Game menu: received event %i (data = %i), reseting sync and processing event", (net_event&0xff), net_event>>8));
            net_sync->Reset();
            NetSend();
            OnNetSyncEvent((E_NET_SYNC_EVENT)(net_event&0xff), net_event>>8);
         }
         break;
      case C_net_sync::SS_PROGRESS:
         {
#ifdef ENABLE_NET_SYNC_MENU
                              //waiting for sync event
            if(!sync_menu){
               sync_menu = CreateSyncMenu(0);
               sync_menu->Release();
            }
            vector<dword> pids;
            net_sync->GetAnsweredPIDs(pids);
            sync_menu->AddCheckMarks(pids.begin(), pids.size());
#endif//ENABLE_NET_SYNC_MENU
         }
         break;
      }
   }else
   if(!(net && !net->IsHost())){
                              //singleplayer or host
      if(last_button_state!=mouse_butt){
         if((last_button_state&1) != (mouse_butt&1)){
            bool pressed = (mouse_butt&1);

                              //check if resource under cursor
            if(last_pointed_res){
               bool done = false;
               if(pressed){
                  dword curr_t = igraph->ReadTimer();
                  if(curr_t-last_click_time < DOUBLE_CLICK_TIME){
                     OnDoubleClick(last_pointed_res);
                     done = true;
                  }
                  last_click_time = curr_t;
               }
               if(!done){
                  cbProc(CB_CLICK, (dword)last_pointed_res, pressed);
                  done = true;
               }
               if(done){
                  MouseUpdate(mx, my, tc.time, true);
               }
            }
            if(pressed){
               if(last_pointed_res && (last_pointed_res->flags&RFLG_BUTTON)){
                  ManageButton(last_pointed_res);
                  PlaySound(GM_SND_BUTT_DOWN);
               }
            }else{
               S_resource *but = ManageButton(NULL);
               if(but && last_pointed_res==but){
                  //PlaySound(GM_SND_BUTT_DOWN);
                  OnButtonPressed(but);
               }
            }
         }
         last_button_state = mouse_butt;
      }else{
         if(!net_chat.IsChatMode())
         switch(tc.key){
         case K_ENTER:
            {
               const S_resource *res = FindResource(ID_BUTTON_OK, RFLG_BUTTON);
               if(res && res->frm->IsOn()){
                  if(net)
                     BeginSyncEvent(NS_PROCEED);
                  else
                     Proceed();
               }
            }
            break;
         case K_BACKSPACE:
            {
               const S_resource *res = FindResource(ID_BUTTON_BACK, RFLG_BUTTON);
               if(res && res->frm->IsOn()){
                  if(net)
                     BeginSyncEvent(NS_BACK_BUTTON);
                  else
                     ReturnBack();
               }
            }
            break;
         }
      }
   }else{
                              //allow leaving menus
      if(!net_chat.IsChatMode())
      switch(tc.key){
      case K_ESC:
         if(tick_class==this){
            tick_class = CreateNetLeaveMenu(this);
            if(tick_class)
               tick_class->Release();
         }
         break;
      }
   }
   Release();
}

//----------------------------

void C_menu_base1::NetIn(E_NETWORK_MESSAGE msg_id, PC_net_in_msg in_msg){

   AddRef();

   switch(msg_id){
   case NM_GAME_MENU+1: case NM_GAME_MENU+2: case NM_GAME_MENU+3: case NM_GAME_MENU+4:
   case NM_GAME_MENU+5: case NM_GAME_MENU+6: case NM_GAME_MENU+7: case NM_GAME_MENU+8:
      {
         int for_menu_id = msg_id-NM_GAME_MENU;
         if(for_menu_id!=menu_id){
            NET_LOG("Message for different menu, skipping");
            in_msg->SkipChunk();
            break;
         }

         in_msg->Ascend();

         while(net && in_msg->GetChunkSize()){
            switch(in_msg->GetID()){

            case NM_MENU_MOUSE_POS:
               {
                  float mx = in_msg->GetFloat();
                  float my = in_msg->GetFloat();
                  MouseUpdate(mx*igraph->Scrn_sy(), my*igraph->Scrn_sy());
               }
               break;

            case NM_GM_SYNC:
               {
                  net_sync->NetIn(in_msg, false);
                  dword net_event;
                  switch(net_sync->GetState(NULL, &net_event)){
                  case C_net_sync::SS_RESOLVED:
                     {
                        if(sync_menu){
                           sync_menu->AddAllCheckMarks();
                           RenderScreen();
                           sync_menu = NULL;
                        }
                        NET_LOG(C_fstring("Game menu: received event %i (data = %i), reseting sync and processing event", (net_event&0xff), net_event>>8));
                        net_sync->Reset();
                        NetSend();
                        OnNetSyncEvent((E_NET_SYNC_EVENT)(net_event&0xff), net_event>>8);
                     }
                     break;
                  case C_net_sync::SS_PROGRESS:
                     {
#ifdef ENABLE_NET_SYNC_MENU
                                          //waiting for sync event
                        if(!sync_menu){
                           sync_menu = CreateSyncMenu(0);
                           sync_menu->Release();
                        }
                        vector<dword> pids;
                        net_sync->GetAnsweredPIDs(pids);
                        sync_menu->AddCheckMarks(pids.begin(), pids.size());
#endif//ENABLE_NET_SYNC_MENU
                     }
                     break;
                  }
               }
               break;

            case NM_CBPROC:
               {
                  byte cb_msg = in_msg->GetByte();
                  dword prm1 = in_msg->GetDword();
                  dword prm2 = in_msg->GetDword();
                  cbProc(cb_msg, prm1, prm2);
               }
               break;
            case NM_CBPROC_R1:
               {
                  byte cb_msg = in_msg->GetByte();
                  S_resource *res = &res_list[in_msg->GetByte()];
                  dword prm2 = in_msg->GetBits(1);
                  cbProc(cb_msg, (dword)res, prm2);
               }
               break;

            default:
               assert(0);
            }
         }
         in_msg->Descend();
      }
      break;

   case NM_GAME:
      NET_LOG("Game menu: silently skipping NM_GAME message");
      in_msg->SkipChunk();
      break;

   case NM_MENU_CHAT:
      net_chat.NetIn(in_msg);
      break;

   case NM_BRIEFING:
      NET_LOG("Game menu: silently skipping NM_BRIEFING message");
      in_msg->SkipChunk();
      break;

   default:
      assert(0);
      in_msg->SkipChunk();
   }
   Release();
}

//----------------------------
//----------------------------
                              //find first inventory item matching requested group,
                              // if nothing is found, return 0
                              //Groups:
                              // 1 = automatic guns
                              // 2 = sniper
                              // 3 = heavy guns
                              // 4 = colt
                              // 5 = grenades

static int asFindItem(int group, const vector<S_item> &avl_items){

   int i;
   for(i=0; i<avl_items.size(); i++){
      if(tab_inventory->ItemI(TAB_I_INV_AUTO_GROUP, avl_items[i].itm) == group)
         return avl_items[i].itm;
   }
   return 0;
}

static int asFindItem(int group, const vector<int> &avl_items){

   int i;
   for(i=0; i<avl_items.size(); i++){
      if(tab_inventory->ItemI(TAB_I_INV_AUTO_GROUP, avl_items[i]) == group)
         return avl_items[i];
   }
   return 0;
}

//----------------------------
//----------------------------

                              //concrete menu - select 4 of 8 and assign ammo
                              //GAME_MENU_4_OF_8
class C_menu_4_of_8: public C_menu_base1{

   S_mission_state *mstate;
                              //items:
   const S_resource *pointed_image;
   const S_resource *selected_image;
                              //men:
   const S_resource *pointed_man;
   const S_resource *selected_man;

   int max_carriage;

                              //item lists
   vector<S_item> *avl_items;
   vector<S_item> *man_items;

                              //scroll selection among available items
   int curr_itm_scroll;

   PI3D_model inv_models[5];
   C_smart_ptr<C_text> man_ammo_txt[5];
//----------------------------

   int FindUpper(int i) const{
      if(i==-1) return -1;
      if(curr_itm_scroll!=-1)
      for(int j=i; j=(++j==man_items->size()) ? 0 : j, j!=i; )
         return j;
      return -1;
   }

//----------------------------

   int FindLower(int i) const{
      if(i==-1) return -1;
      if(curr_itm_scroll!=-1)
      for(int j=i; j=(j--==0) ? man_items->size()-1 : j, j!=i; )
         return j;
      return -1;
   }

//----------------------------

   void ShowScrollItems(bool hide_all = false){

      int lower(curr_itm_scroll), upper(curr_itm_scroll);
      for(int i1=0; i1<5; i1++){

         static const imap[] = {2, 1, 3, 0, 4};
         int i(imap[i1]);
                              //hide previous
         if(inv_models[i]){
            inv_models[i]->LinkTo(NULL);
            inv_models[i]->Release();
            inv_models[i]=NULL;
         }
         man_ammo_txt[i] = NULL;

         if(hide_all) continue;
         assert(man_items);

         if(i1>=man_items->size()) continue;

         int ii = curr_itm_scroll;
         if(ii!=-1){
            switch(i){
            case 0:
               lower = ii = FindLower(ii);
               if(ii==upper) ii=-1;
            case 1:
               lower = ii = FindLower(ii);
               if(ii==upper) ii=-1;
               break;
            case 4:
               upper = ii = FindUpper(ii); if(ii==lower) ii=-1;
            case 3:
               upper = ii = FindUpper(ii); if(ii==lower) ii=-1;
               break;
            }
         }
         float scrn_sx = igraph->Scrn_sx();
         if(ii!=-1 && !hide_all){
            const S_resource *res = FindResource(i+10, RFLG_ITEM);
            if(res){
               S_item &itm = (*man_items)[ii];
               PI3D_model mod = driver->CreateModel();
               inv_models[i] = mod;

               const char *mname = tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16);
               model_cache.Open(mod, mname, scene, 0, ErrReport, editor);

#ifdef _DEBUG
               mod->SetName(tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16));
#endif
               mod->SetAnimation(0, NULL);
               mod->LinkTo(res->frm);

                              //show text info
               S_vector in(res->frm->GetWorldPos()), out;
               S_vectorw v4;
               scene->TransformPoints(&in, &v4, 1);
               out = v4;

               float x1 = out.x / scrn_sx, y1 = out.y / scrn_sx;
               //x1 -= .1f;
               y1 += .0225f;
               C_fstring str("%i", itm.pieces);
               if(HasAmmo(itm.itm)) str += C_fstring("/%i", itm.ammo);

               y1 /= (screen_aspect_ratio * 1.33333333f);
               man_ammo_txt[i] = texts.CreateText(str, -1, x1, y1, .0225f, TEXT_JUSTIFY_CENTER);
               man_ammo_txt[i]->Release();
            }
         }
      }
   }

#define ITM_GRID_X 3
#define ITM_GRID_Y 5
   PI3D_frame itm_grid[3][ITM_GRID_Y][ITM_GRID_X];   //[dummy/hlight/item][][]
   C_smart_ptr<C_text> ammo_txt[ITM_GRID_Y][ITM_GRID_X];
   int grid_scroll_line;

//----------------------------

   void ShowGridItems(bool hide_all = false){

      if(hide_all && !avl_items)
         return;
      assert(avl_items);
      int i, x, y;
                              //hide previous
      for(y=0; y<ITM_GRID_Y; y++)
      for(x=0; x<ITM_GRID_X; x++){
         if(itm_grid[2][y][x]){
            itm_grid[2][y][x]->LinkTo(NULL);
            itm_grid[2][y][x]->Release();
            itm_grid[2][y][x] = NULL;
         }
         ammo_txt[y][x] = NULL;
      }
      if(hide_all)
         return;
                              //show new
      float scrn_sx = igraph->Scrn_sx();
      int num = Min(ITM_GRID_X*ITM_GRID_Y, (int)avl_items->size()-grid_scroll_line*ITM_GRID_X);
      for(i=0; i<num; i++){
         x = i%ITM_GRID_X;
         y = i/ITM_GRID_X;

         S_item &itm = (*avl_items)[i+grid_scroll_line*ITM_GRID_X];
         PI3D_model mod = driver->CreateModel();
         itm_grid[2][y][x] = mod;

         const char *mname = tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16);
         model_cache.Open(mod, mname, scene, 0, ErrReport, editor);

#ifdef _DEBUG
         mod->SetName(tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16));
#endif
         mod->SetAnimation(0, NULL);
         mod->LinkTo(itm_grid[0][y][x]);
                              //show text info
         S_vector in(itm_grid[0][y][x]->GetWorldPos()), out;
         S_vectorw v4;
         scene->TransformPoints(&in, &v4, 1);
         out = v4;
         float x1 = out.x / scrn_sx, y1 = out.y / scrn_sx;
         y1 += .0225f;
         C_fstring str("%i", itm.pieces);
         if(HasAmmo(itm.itm)) str += C_fstring("/%i", itm.ammo);
         y1 /= (screen_aspect_ratio * 1.33333333f);

         ammo_txt[y][x] = texts.CreateText(str, -1, x1, y1, .0225f, TEXT_JUSTIFY_CENTER);
         ammo_txt[y][x]->Release();
      }
   }

   void ShowTeam8(int indx){
      const S_resource *res = FindResource(indx, RFLG_MAN_8);
      int mi = mstate->team_8[indx];
      ChangeTexture(res->frm, mi==-1 ? "face_no" : C_fstring("man%.2i", mi+1), "faces");
   }

   void ShowTeam4(int indx){
      const S_resource *res = FindResource(indx+20, RFLG_MAN_4);
      int mi = mstate->team_4[indx];
      ChangeTexture(res->frm, mi==-1 ? "face_no" : C_fstring("man%.2i", mi+1), "faces");
   }

   int GetItemID(const S_resource *res) const{

      if(!res) return -1;
      if(res->flags&RFLG_MAN_8) return res->id;
      if(res->flags&RFLG_MAN_4) return res->id-20;

      if(res->flags&RFLG_ITEM_GRID){
         return res->id - 1000 + grid_scroll_line*ITM_GRID_X;
      }
      assert(res->flags&RFLG_ITEM);
      int i(-1);
      switch(res->id){
      case 10: i = FindLower(FindLower(curr_itm_scroll)); break;
      case 11: i = FindLower(curr_itm_scroll); break;
      case 12: i = curr_itm_scroll; break;
      case 13: i = FindUpper(curr_itm_scroll); break;
      case 14: i = FindUpper(FindUpper(curr_itm_scroll)); break;
      }
      return i;
   }

   int GetItemIDID(const S_resource *res) const{
      int id = GetItemID(res);
      if(res->flags&RFLG_ITEM) return (*man_items)[id].itm;
      else return (*avl_items)[id].itm;
   }

   int GetManPos(const S_resource *res) const{
      if(!res) return -1;
      if(res->flags&RFLG_MAN_4) return res->id-20;
      return res->id;
   }

   int GetManID(const S_resource *res) const{
      if(!res) return -1;
      if(res->flags&RFLG_MAN_4) return mstate->team_4[res->id-20];
      return mstate->team_8[res->id];
   }

   C_smart_ptr<C_text> desc, man_desc;
   C_smart_ptr<C_text> bar_num;
   C_smart_ptr<I3D_mesh_level> ml_bar;
protected:

public:
   C_menu_4_of_8():
      mode(MODE_NORMAL),
      selected_image(NULL),
      pointed_image(NULL),
      selected_man(NULL),
      pointed_man(NULL),
      avl_items(NULL),
      man_items(NULL),
      curr_itm_scroll(0),
      grid_scroll_line(0)
   {
      mstate = &game_profile.mission_states[game_profile.MissionIndex()];
      memset(inv_models, 0, sizeof(inv_models));
      memset(itm_grid, 0, sizeof(itm_grid));
   }

   ~C_menu_4_of_8(){
      ShowScrollItems(true);
      ShowGridItems(true);
      HideInfo();
      HideManInfo();
      HideWeight();

      for(int y=0; y<ITM_GRID_Y; y++)
      for(int x=0; x<ITM_GRID_X; x++)
      for(int i=0; i<2; i++)
      if(itm_grid[i][y][x]) itm_grid[i][y][x]->Release();
   }

//----------------------------
                              //context = flags (1 = image should be primarily selected)
   virtual bool Init(E_GAME_MENU id, dword context){

      if(!C_menu_base1::Init(id))
         return false;
      {
         PI3D_frame fm = scene->FindFrame("item05");
         const S_vector &mn = fm->GetPos(),
            &mx = scene->FindFrame("item06")->GetPos();
         const float scl = fm->GetScale();
         float dx = (mx.x-mn.x) / (ITM_GRID_X-1);
         float dy = (mx.y-mn.y) / (ITM_GRID_Y-1);
                              //create grid of dummys and highlights
         for(int y=0; y<ITM_GRID_Y; y++)
         for(int x=0; x<ITM_GRID_X; x++){
            for(int i=0; i<2; i++){
               PI3D_model mod = driver->CreateModel();
               if(i){
                  model_cache.Open(mod, "inv_frm", scene, 0, ErrReport, editor);
#ifdef _DEBUG
                  mod->SetName("inv_frm");
#endif
               }
               PI3D_frame frm = mod;
               itm_grid[i][y][x] = frm;
               frm->SetPos(S_vector(mn.x + dx*x, mn.y + dy*y, !i ? mn.z : mn.z+.5f));
               frm->SetScale(scl);
               frm->Update();
               frm->LinkTo(scene->GetPrimarySector());
               res_list.push_back(S_resource());
               S_resource &res = res_list.back();
               res.frm = frm;
               res.id = y*ITM_GRID_X + x;
               //if(i)
                  res.id += 1000;
               res.flags = i ? RFLG_HIGHLIGHT : (RFLG_ITEM_GRID | RFLG_HINT);
            }
         }
         PI3D_frame bar = scene->FindFrame("LEVEL");
         if(bar && bar->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(bar)->GetVisualType()==I3D_VISUAL_OBJECT){
            ml_bar = I3DCAST_OBJECT(bar)->GetMeshObject()->GetLOD(0);
         }
      }

      static const S_resource_def r_list[] = {
         {"OK", ID_BUTTON_OK, RFLG_BUTTON | RFLG_HINT},
         {"BACK", ID_BUTTON_BACK, RFLG_BUTTON | RFLG_HINT},
         {"AUTO", ID_BUTTON_AUTO_SELECT, RFLG_BUTTON | RFLG_HINT},
         {"INFO", ID_BUTTON_SOLDIER_INFO, RFLG_BUTTON | RFLG_HINT},
         {"BUT00", ID_BUTTON_ITEM_ADD, RFLG_BUTTON | RFLG_HINT},
         {"BUT01", ID_BUTTON_ITEM_REMOVE, RFLG_BUTTON | RFLG_HINT},
         {"BUT02", ID_BUTTON_AMMO_ADD, RFLG_BUTTON | RFLG_HINT},
         {"BUT03", ID_BUTTON_AMMO_REMOVE, RFLG_BUTTON | RFLG_HINT},
         {"BUT04", ID_BUTTON_MAN_ADD, RFLG_BUTTON | RFLG_HINT},
         {"BUT05", ID_BUTTON_MAN_REMOVE, RFLG_BUTTON | RFLG_HINT},
         {"DOWN00", ID_BUTTON_UP, RFLG_BUTTON | RFLG_INVISIBLE | RFLG_ARROW_UP},
         {"UP00", ID_BUTTON_DOWN, RFLG_BUTTON | RFLG_INVISIBLE | RFLG_ARROW_DN},
         {"UP01", ID_BUTTON_UP1, RFLG_BUTTON | RFLG_ARROW_UP | RFLG_INVISIBLE},
         {"DOWN01", ID_BUTTON_DOWN1, RFLG_BUTTON | RFLG_INVISIBLE | RFLG_ARROW_DN},

         {"slct_ok", ID_BUTTON_OK+100, RFLG_HIGHLIGHT},
         {"slct_back", ID_BUTTON_BACK+100, RFLG_HIGHLIGHT},
         {"slct_auto", ID_BUTTON_AUTO_SELECT+100, RFLG_HIGHLIGHT},
         {"slct_info", ID_BUTTON_SOLDIER_INFO+100, RFLG_HIGHLIGHT},
         {"slct_butt00", ID_BUTTON_ITEM_ADD+100, RFLG_HIGHLIGHT},
         {"slct_butt01", ID_BUTTON_ITEM_REMOVE+100, RFLG_HIGHLIGHT},
         {"slct_butt02", ID_BUTTON_AMMO_ADD+100, RFLG_HIGHLIGHT},
         {"slct_butt03", ID_BUTTON_AMMO_REMOVE+100, RFLG_HIGHLIGHT},
         {"slct_butt04", ID_BUTTON_MAN_ADD+100, RFLG_HIGHLIGHT},
         {"slct_butt05", ID_BUTTON_MAN_REMOVE+100, RFLG_HIGHLIGHT},
         //{"slct_up01", ID_BUTTON_UP+100, RFLG_HIGHLIGHT},
         //{"slct_down01", ID_BUTTON_DOWN+100, RFLG_HIGHLIGHT},
         //{"slct_up00", ID_BUTTON_UP1+100, RFLG_HIGHLIGHT},
         //{"slct_down00", ID_BUTTON_DOWN1+100, RFLG_HIGHLIGHT},
                              //8 men
         {"pic1", 0, RFLG_MAN_8 | RFLG_HINT},
         {"pic2", 1, RFLG_MAN_8 | RFLG_HINT},
         {"pic3", 2, RFLG_MAN_8 | RFLG_HINT},
         {"pic4", 3, RFLG_MAN_8 | RFLG_HINT},
         {"pic5", 4, RFLG_MAN_8 | RFLG_HINT},
         {"pic6", 5, RFLG_MAN_8 | RFLG_HINT},
         {"pic7", 6, RFLG_MAN_8 | RFLG_HINT},
         {"pic8", 7, RFLG_MAN_8 | RFLG_HINT},
         {"slct_00", 0, RFLG_HIGHLIGHT},
         {"slct_01", 1, RFLG_HIGHLIGHT},
         {"slct_02", 2, RFLG_HIGHLIGHT},
         {"slct_03", 3, RFLG_HIGHLIGHT},
         {"slct_04", 4, RFLG_HIGHLIGHT},
         {"slct_05", 5, RFLG_HIGHLIGHT},
         {"slct_06", 6, RFLG_HIGHLIGHT},
         {"slct_07", 7, RFLG_HIGHLIGHT},
                              //4 men
         {"MAN1", 20, RFLG_MAN_4 | RFLG_HINT},
         {"MAN2", 21, RFLG_MAN_4 | RFLG_HINT},
         {"MAN3", 22, RFLG_MAN_4 | RFLG_HINT},
         {"MAN4", 23, RFLG_MAN_4 | RFLG_HINT},
         {"slct_10", 20, RFLG_HIGHLIGHT},
         {"slct_11", 21, RFLG_HIGHLIGHT},
         {"slct_12", 22, RFLG_HIGHLIGHT},
         {"slct_13", 23, RFLG_HIGHLIGHT},

         {"item00", 10, RFLG_ITEM | RFLG_HINT},
         {"item01", 11, RFLG_ITEM | RFLG_HINT},
         {"item02", 12, RFLG_ITEM | RFLG_HINT},
         {"item03", 13, RFLG_ITEM | RFLG_HINT},
         {"item04", 14, RFLG_ITEM | RFLG_HINT},
         {"inv_frm00", 10, RFLG_HIGHLIGHT},
         {"inv_frm01", 11, RFLG_HIGHLIGHT},
         {"inv_frm02", 12, RFLG_HIGHLIGHT},
         {"inv_frm03", 13, RFLG_HIGHLIGHT},
         {"inv_frm04", 14, RFLG_HIGHLIGHT},

         {NULL}
      };
      InitResources(r_list);

      cbProc(CB_ENUM_RESOURCES, 0, 0);

      int i;
      for(i=0; i<8; i++)
         ShowTeam8(i);
      for(i=0; i<4; i++)
         ShowTeam4(i);

      avl_items = &mstate->item_list;
      //if(!avl_items->size()) grid_scroll_line = -1;
      curr_itm_scroll = -1;

      ShowGridItems();
      ShowWeigth();

      ManageOk();
      return true;
   }

//----------------------------

   void MakeAutoList(){

      int i, j;

      for(i=0; i<4; i++){
         as_want_list[i].clear();
      }

      int miss_indx = game_profile.MissionIndex();
                              //spread mission items among men
      int man_indx = -1;
      for(i=0; i<64; i++){
         int amount = tab_game_cfg->ItemI(TAB_I64_GAME_MIS_ITEM_AMOUNT, miss_indx*64 + i);
         if(amount){
                              //find next available man
            while(true){
               ++man_indx %= 4;
               if(mstate->team_4[man_indx]!=-1) break;
            }

            bool has_ammo = C_inventory::HasAmmo(i);
            int pcs = has_ammo ? 1 : amount;
            int ammo = has_ammo ? amount : 0;
            if(has_ammo && !ammo) ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, i);

            vector<S_item> &want_list = as_want_list[man_indx];
                              //find among existing
            for(j=want_list.size(); j--; )
            if(want_list[j].itm==i) break;
            if(j==-1)
               want_list.push_back(S_item(i, ammo, pcs));
            else{
               want_list[j].pieces += pcs;
               want_list[j].ammo += ammo;
            }
         }
      }

                              //now assign rest of items

                              //binoculars: 1st man in mission
      for(i=0; i<4; i++) if(mstate->team_4[i]!=-1) break;
      as_want_list[i].push_back(S_item(1, 0, 1));

                              //camera: 2nd man in miss.
      for(++i; i<4; i++) if(mstate->team_4[i]!=-1) break;
      if(i<4) as_want_list[i].push_back(S_item(11, 0, 1));

      int wpn_pattern_indx = 0;
      static const byte wpn_pattern[] = {
                              //assign: 2x autom. gun, sniper, heavy gun, colt
         4, 1, 1, 2, 3
      };
      vector<S_item> avl_list = *avl_items;
                              //now assign weapons to each man
      int k;
      for(i=0; i<4; i++){
         if(mstate->team_4[i]==-1) continue;
                              //give each man a knife
         as_want_list[i].push_back(S_item(29, 0, 1));
                              //and 1 grenade
         for(k = avl_list.size(); k--; ){
            if(tab_inventory->ItemI(TAB_I_INV_AUTO_GROUP, avl_list[k].itm) == 5) break;
         }
         if(k!=-1){
            as_want_list[i].push_back(S_item(avl_list[k].itm, 0, 1));
            if(!--avl_list[k].pieces){
               avl_list[k] = avl_list.back();
               avl_list.pop_back();
            }
         }

                              //assign weapon by pattern

                              //find requested wpn
         for(j=wpn_pattern_indx; ++j%=sizeof(wpn_pattern), j!=wpn_pattern_indx; ){
            for(k = avl_list.size(); k--; )
            if(avl_list[k].pieces && avl_list[k].ammo)
            {
               if(tab_inventory->ItemI(TAB_I_INV_AUTO_GROUP, avl_list[k].itm) == wpn_pattern[j]) break;
            }
            if(k!=-1) break;
         }
         if(j!=wpn_pattern_indx){
                              //requested weapon found
            wpn_pattern_indx = j;
            int inv_id = avl_list[k].itm;
            int ammo = Min(tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, inv_id), (int)avl_list[k].ammo);
                              //remove from avl list
            avl_list[k].ammo -= ammo;
            avl_list[k].pieces--;
            if(!avl_list[k].pieces && !avl_list[k].ammo){
               avl_list[k] = avl_list.back();
               avl_list.pop_back();
            }

            as_want_list[i].push_back(S_item(inv_id, ammo, 1));
         }
      }
   }

//----------------------------

   enum{
      MODE_NORMAL,
      MODE_AUTO_SELECT,
      MODE_END,
      MODE_PAUSE,
      MODE_CHECK,
      MODE_MAN_ADD,
      MODE_POSITION,
      MODE_FILL,
   } mode, paused_mode;
                              //auto-select values
   vector<S_item> as_want_list[4];
   int as_curr_indx;
   int as_man_indx;
   bool as_fast;
   int as_pause_count;

//----------------------------

   void SetPause(int p = 100){
      if(as_fast) return;
      paused_mode=mode;
      mode=MODE_PAUSE;
      as_pause_count = p;
   }

//----------------------------

   virtual void Tick(const S_tick_context &tc){

      int i;

      if(mode==MODE_NORMAL){
         C_menu_base1::Tick(tc);
         return;
      }

      if(tc.key==K_ESC)
         as_fast = true;

      do{
      switch(mode){
      case MODE_AUTO_SELECT:
         spr_arrow->SetOn(false);
         as_fast=false;
         mode=MODE_CHECK;
         break;

      case MODE_END:
                              //end of fun, back to user
         spr_arrow->SetOn(true);
         mode=MODE_NORMAL;
         break;

      case MODE_CHECK:
         {
            cbProc(CB_UNSELECT, true);
                              //check if we've some empty space for man
            i = GetEmptySlot();
            if(i!=-1){
                              //check if we've man in stock
               int numf = 0;
               for(i=8; i--; ) if(mstate->team_8[i]!=-1) ++numf;
               if(numf){
                              //select random man
                  numf = S_int_random(numf);
                  for(i=8; i--; ) if(mstate->team_8[i]!=-1 && !numf--) break;
                  cbProc(CB_SELECT_MAN, (dword)FindResource(i, RFLG_MAN_8), !as_fast);
                  mode=MODE_MAN_ADD;
                  SetPause(50);
                  break;
               }
            }
                              //remove all items of all men
            for(i=0; i<4; i++)
            if(mstate->team_4[i]!=-1){
               cbProc(CB_SELECT_MAN, (dword)FindResource(i+20, RFLG_MAN_4), false);
               cbProc(CB_REMOVE_ALL);
            }
            MakeAutoList();
            mode = MODE_POSITION;
            as_man_indx = -1;
         }
         break;

      case MODE_MAN_ADD:
         {
            cbProc(CB_ADD_SELECTED_MAN);
            mode=MODE_CHECK;
            SetPause(50);
         }
         break;

      case MODE_POSITION:
         {
            while(++as_man_indx<4)
            if(mstate->team_4[as_man_indx]!=-1){
               cbProc(CB_SELECT_MAN, (dword)FindResource(as_man_indx+20, RFLG_MAN_4), false);
               DisplayManInfo(selected_man);
               as_curr_indx=0;
               cbProc(CB_UNSELECT, true);
               mode=MODE_FILL;
               SetPause(50);
               break;
            }
            if(as_man_indx==4)
               mode=MODE_END;
         }
         break;

      case MODE_FILL:
         {
                              //check end
            if(as_curr_indx >= as_want_list[as_man_indx].size()){
               mode = MODE_POSITION;
               break;
            }

            int id = as_want_list[as_man_indx][as_curr_indx].itm;
                              //check if we need to process this
            for(i=man_items->size(); i--; ) if((*man_items)[i].itm==id) break;
            if(i!=-1){
               if((*man_items)[i].pieces==as_want_list[as_man_indx][as_curr_indx].pieces &&
                  (*man_items)[i].ammo==as_want_list[as_man_indx][as_curr_indx].ammo){
                              //item ok, proceed
                  ++as_curr_indx;
                  cbProc(CB_UNSELECT, true);
                  break;
               }
            }
                              //find among available
            for(i=avl_items->size(); i--; ) if((*avl_items)[i].itm==id) break;
            if(i==-1){        //this shouldn't happen in final
               ++as_curr_indx;
               break;
            }
                              //scroll to desired item
            if(MakeVisible(i)){
               SetPause(30);
               break;
            }

            if(!selected_image){
               cbProc(CB_SELECT_IMAGE, (dword)FindResource(i+1000-grid_scroll_line*ITM_GRID_X, RFLG_ITEM_GRID), !as_fast);
               SetPause(30);
               break;
            }
                              //now add pieces/ammo
            for(i=man_items->size(); i--; ) if((*man_items)[i].itm==id) break;
                              //check existence | pieces
            if(i==-1 ||
               (*man_items)[i].pieces < as_want_list[as_man_indx][as_curr_indx].pieces){

               i = cbProc(CB_ADD_SELECTED1, (!as_fast ? 0x10000 : 0));
               if(!i){        //overload - should be computed properly in final
                  cbProc(CB_UNSELECT, true);
                  ++as_curr_indx;
               }
               SetPause(20);
               break;
            }
            if(HasAmmo(id)){
                              //check ammo
               int delta_ammo = as_want_list[as_man_indx][as_curr_indx].ammo - (*man_items)[i].ammo;
               if(delta_ammo > 0){
                  i = cbProc(CB_ADD_SELECTED1, (!as_fast ? 0x10000 : 0) | 1, Min(10, delta_ammo));
                  if(!i){        //overload - should be computed properly in final
                     cbProc(CB_UNSELECT, true);
                     ++as_curr_indx;
                  }
                  SetPause(10);
                  break;
               }
            }
            ++as_curr_indx;
         }
         break;

      case MODE_PAUSE:
         as_pause_count -= tc.time;
         if(as_pause_count<=0) mode=paused_mode;
         break;
      }
      }while(as_fast && mode!=MODE_NORMAL);

      C_mission::Tick(tc);
      if(hint && !hint->Tick(tc.time))
         hint = NULL;
   }

//----------------------------

   bool MakeVisible(int indx){
      if(indx < grid_scroll_line*ITM_GRID_X){
         cbProc(CB_SCROLL_GRID, true, true);
         return true;
      }
      int maxx = (grid_scroll_line+ITM_GRID_Y)*ITM_GRID_X;
      if(indx >= maxx){
         cbProc(CB_SCROLL_GRID, false, true);
         return true;
      }
      return false;
   }

//----------------------------

   void HideInfo(){
      desc = NULL;
   }

//----------------------------

   void DisplayInfo(const S_resource *res){

      HideInfo();

                              //feed texts
      float x = .5f, y = .5f;
      PI3D_frame frm = scene->FindFrame("TEXT00");
      if(frm) Map3DTo2D(frm->GetWorldPos(), x, y);

      int tid = GetItemIDID(res);
      if(game_profile.IsDatadisk())
         tid += 1e6;
      C_string str;
      str = all_txt[GT_GAME_MENU_ITEM_NAME+tid];
      bool wide = all_txt.IsWide(GT_GAME_MENU_ITEM_NAME+tid);
      str = AddGameText(str, MakeGameText("\n"));

      str = AddGameText(str, all_txt[GT_GAME_MENU_ITEM_DESC+tid]);
      float fmt_rect[4];
      GetItemRect(frm, fmt_rect);
      y /= (screen_aspect_ratio * 1.33333333f);

      float sz = .021f;
      desc = texts.CreateText(str, -1, x, y, sz, TEXT_JUSTIFY_CONVERT_EOL, wide, fmt_rect);
      desc->Release();
   }

//----------------------------

   void HideManInfo(){
      man_desc = NULL;
   }

//----------------------------

   void DisplayManInfo(const S_resource *res){

      HideManInfo();

      int id = GetManID(res);
      if(id==-1) return;
                              //feed texts
      float x(.5), y(.5);
      PI3D_frame frm = scene->FindFrame("TEXT01");
      if(frm) Map3DTo2D(frm->GetWorldPos(), x, y);
      C_string str;
      str = all_txt[GT_GAME_MENU_SOLDIER_TITL+id];
      bool wide = all_txt.IsWide(GT_GAME_MENU_SOLDIER_TITL+id);
      //str += "\n";
      str = AddGameText(str, MakeGameText("\n"));
      //str += all_txt[GT_GAME_MENU_SOLDIER_NAME+id];
      str = AddGameText(str, all_txt[GT_GAME_MENU_SOLDIER_NAME+id]);
      float fmt_rect[4];
      GetItemRect(frm, fmt_rect);
      y /= (screen_aspect_ratio * 1.33333333f);
      float sz = .021f;
      man_desc = texts.CreateText(str, -1, x, y, sz, TEXT_JUSTIFY_CONVERT_EOL, wide, fmt_rect);
      man_desc->Release();
   }

//----------------------------

   int MaxGridScroll() const{
      return Max(0, (int)(avl_items->size()-1)/ITM_GRID_X);
   }

//----------------------------

   int GetEmptySlot() const{
      for(int i=0; i<4; i++)
         if(mstate->team_4[i]==-1) return i;
      return -1;
   }

//----------------------------

   int GetEmptySlot8() const{
      for(int i=0; i<8; i++)
         if(mstate->team_8[i]==-1) return i;
      return -1;
   }

//----------------------------
                              //show OK button when appropriate
   void ManageOk(){

      for(int i=4; i--; ){
         if(mstate->team_4[i]!=-1)
            break;
      }
      ShowButton(ID_BUTTON_OK, bool(i!=-1));
   }

//----------------------------

   float CountWeight() const{
      assert(man_items);
      float w(0);
      for(int i=0; i<man_items->size(); i++){
         const S_item &itm = (*man_items)[i];
         w += tab_inventory->ItemF(TAB_F_INV_WEIGTH, itm.itm) * itm.pieces;
         if(HasAmmo(itm.itm))
            w += tab_inventory->ItemF(TAB_F_INV_AMMO_WEIGTH, itm.itm) * itm.ammo;
      }
      return w;
   }

//----------------------------

   void ShowWeigth(){

      float f, w;

      PI3D_frame frm = scene->FindFrame("LEVEL");
      if(frm){
         if(!man_items){
            frm->SetOn(false);
            return;
         }
         w = CountWeight();
         f = w / max_carriage;
         frm->SetOn(true);
         S_vector scl(Min(1.0f, f), 1.0f, 1.0f);
         ScaleObject(frm, ml_bar, scl);
      }
      HideWeight();
      {
         float x, y;
         PI3D_frame frm = scene->FindFrame("bar");
         if(frm) Map3DTo2D(frm->GetWorldPos(), x, y);
         y /= (screen_aspect_ratio * 1.33333333f);

         bar_num = texts.CreateText(C_fstring("%i%%", Min(100, (int)(f*100.0f))), -1, x, y+.006f, .0225f, TEXT_JUSTIFY_CENTER,
            false,
            NULL, NULL, NULL, NULL, 0,
            PI*.1f);
         bar_num->Release();
      }
   }

//----------------------------

   void HideWeight(){
      bar_num = NULL;
   }

//----------------------------

public:
   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){
      case CB_REMOVE_ALL:
         {
            assert(man_items);
                                    //return all ammo back
            for(int i=0; i<man_items->size(); i++){
                                    //find in avl list
               for(int j=avl_items->size(); j--; ) if((*avl_items)[j].itm == (*man_items)[i].itm) break;
               if(j==-1) avl_items->push_back((*man_items)[i]);
               else{
                  (*avl_items)[j].ammo += (*man_items)[i].ammo;
                  (*avl_items)[j].pieces += (*man_items)[i].pieces;
               }
            }
            man_items->clear();
            ShowGridItems();
            if(prm1) return true;
         }
         break;

      case CB_SCROLL:
         {
            C_menu_base1::cbProc(msg, prm1, prm2);
            bool up = prm1;

            int i;
            if(!up) i=FindUpper(curr_itm_scroll);
            else i=FindLower(curr_itm_scroll);
            if(i!=-1){
               if(selected_image && (selected_image->flags&RFLG_ITEM))
                  cbProc(CB_UNSELECT);
               curr_itm_scroll=i;
               ShowScrollItems();
               PlaySound(GM_SND_SCROLL);
            }
            return true;
         }
         break;

      case CB_SCROLL_GRID:
         {
            bool up = prm1;
            if(up){
               if(!grid_scroll_line) return false;
               --grid_scroll_line;
            }else{
               if(grid_scroll_line==MaxGridScroll()) return false;
               ++grid_scroll_line;
            }
            ShowGridItems();
         }
         break;

      case CB_SELECT_IMAGE:
         {
            const S_resource *res = (const S_resource*)prm1;
            bool snd = prm2;
                                    //check if selection possible
            if(selected_image!=res){
               ShowButton(ID_BUTTON_ITEM_ADD, false);
               ShowButton(ID_BUTTON_AMMO_ADD, false);
               ShowButton(ID_BUTTON_ITEM_REMOVE, false);
               ShowButton(ID_BUTTON_AMMO_REMOVE, false);

               int id = GetItemIDID(res);

               if(res->flags&RFLG_ITEM){
                  ShowButton(ID_BUTTON_ITEM_REMOVE, true);
                  if(HasAmmo(id))
                     ShowButton(ID_BUTTON_AMMO_REMOVE, true);
               }
               if(res->flags&RFLG_ITEM_GRID){
                  ShowButton(ID_BUTTON_ITEM_ADD, true);
                  if(HasAmmo(id))
                     ShowButton(ID_BUTTON_AMMO_ADD, true);
               }
               if(selected_image) SetHLight(selected_image, false);
               selected_image=res;
               if(selected_image) SetHLight(selected_image, true);
               if(snd) PlaySound(GM_SND_SELECT);
            }
         }
         break;

      case CB_SELECT_MAN:

         {
            const S_resource *res = (const S_resource*)prm1;
            bool snd = prm2;
                                    //check if selection possible
            if(selected_man!=res){
               int id = GetManID(res);
               if(id==-1){
                  if(snd) PlaySound(GM_SND_CANNOT);
                  return false;
               }
               int pos = GetManPos(res);

               ShowButton(ID_BUTTON_MAN_ADD, false);
               ShowButton(ID_BUTTON_MAN_REMOVE, false);

               if(res->flags&RFLG_MAN_4){
                  ShowButton(ID_BUTTON_MAN_REMOVE, true);
               }
               if(res->flags&RFLG_MAN_8){
                  ShowButton(ID_BUTTON_MAN_ADD, true);
               }
               if(selected_man) SetHLight(selected_man, false);
               selected_man = res;
               if(selected_man) SetHLight(selected_man, true);
               if(snd) PlaySound(GM_SND_SELECT);

               if(selected_man->flags&RFLG_MAN_4){
                                    //get man's carriage
                  LPC_table tab(CreateTable());
                  LoadSoldierTable(tab, id);
                  max_carriage = CountManCarriage(tab);
                  tab->Release();
                                    //show man's inv
                  man_items = &mstate->team_4[pos].item_list;
                  curr_itm_scroll = !man_items->size() ? -1 : 0;
                  ShowScrollItems();
               }else{
                  man_items=NULL;
                  curr_itm_scroll = -1;
                  ShowScrollItems(true);
               }
               C_menu_base1::cbProc(msg, prm1, prm2);

               cbProc(CB_UNSELECT);

               ShowWeigth();

               ShowButton(ID_BUTTON_SOLDIER_INFO, true);
            }
            return true;
         }
         break;

      case CB_UNSELECT:
         if(selected_image){
            HideInfo();
            if(pointed_image!=selected_image)
               SetHLight(selected_image, false);
            selected_image=NULL;
            pointed_image=NULL;
            ShowButton(ID_BUTTON_ITEM_ADD, false);
            ShowButton(ID_BUTTON_AMMO_ADD, false);
            ShowButton(ID_BUTTON_ITEM_REMOVE, false);
            ShowButton(ID_BUTTON_AMMO_REMOVE, false);
            return true;
         }
         //if(!prm1) return true;
         break;

      case CB_ADD_SELECTED_MAN:
         {
            assert(selected_man);
            int i=GetEmptySlot();
            if(i==-1) return false;
            int id(GetManID(selected_man));
            int t8_pos = selected_man->id;
            mstate->team_4[i] = (short)id;
            ShowTeam4(i);
                                    //change selection to team4
            SetHLight(selected_man, false);

            C_menu_base1::cbProc(msg, prm1, prm2);
            cbProc(CB_UNSELECT);
            cbProc(CB_SELECT_MAN, (dword)FindResource(i+20, RFLG_MAN_4), false);
                                    //adjust team8
            mstate->team_8[t8_pos] = -1;
            ShowTeam8(t8_pos);
            ManageOk();
            return true;
         }
         break;

      case CB_REMOVE_SELECTED_MAN:
         {
            C_menu_base1::cbProc(msg, prm1, prm2);
            assert(selected_man);
            assert(man_items);
            cbProc(CB_REMOVE_ALL, true);
            int i;

            int id = GetManID(selected_man);
            int t4_pos = selected_man->id-20;
            i = GetEmptySlot8();
            assert(i!=-1);
            mstate->team_8[i] = id;
            ShowTeam8(i);
                                    //change selection to team8
            SetHLight(selected_man, false);
            cbProc(CB_UNSELECT);
            cbProc(CB_SELECT_MAN, (dword)FindResource(i, RFLG_MAN_8), true);
                                    //adjust team4
            mstate->team_4[t4_pos] = -1;
            ShowTeam4(t4_pos);
            ManageOk();
            return true;
         }
         break;

      case CB_ADD_SELECTED1:
         {
            bool add_ammo = (prm1&1);
            bool snd = (prm1&0x10000);
            int num_ammo = prm2;
            //if(add_ammo && !num_ammo) num_ammo=10;

            assert(selected_image);
            assert(man_items && avl_items);
            int id(GetItemIDID(selected_image));
            int i;

            float w = CountWeight();
                                    //check if we have such item already
            for(i=man_items->size(); i--; ) if((*man_items)[i].itm==id) break;
                              //cannot add ammo if there's not an instance of the item
            if(i==-1 && add_ammo){
               if(snd) PlaySound(GM_SND_CANNOT);
               return false;
            }

            bool mult = tab_inventory->ItemB(TAB_B_INV_MULTIPLE, id);
            bool new_ref;
                              //always new instance if not multiple item
            if(!mult) new_ref=true;
            else new_ref = (i==-1);

            int id1 = GetItemID(selected_image);
            if(add_ammo){
               if(!HasAmmo(id)) return false;
               if(!num_ammo) num_ammo = tab_inventory->ItemI(TAB_I_INV_MENU_AMMO_TRANSFER, id);
               num_ammo = Min(num_ammo, (int)(*avl_items)[id1].ammo);
               if(!num_ammo) return false;
            }else{
               if(mult && HasAmmo(id) && !new_ref){
                  return false;
               }
               if(!(*avl_items)[id1].pieces) return false;
            }

                                    //check weight
            float item_w;
            if(!add_ammo) item_w = tab_inventory->ItemF(TAB_F_INV_WEIGTH, id);
            else item_w = tab_inventory->ItemF(TAB_F_INV_AMMO_WEIGTH, id) * num_ammo;
            if(w+item_w > max_carriage){
               if(snd) PlaySound(GM_SND_CANNOT);
               return false;
            }

            if(new_ref){
               //if(!num_ammo) num_ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, id);
                                    //add item
               curr_itm_scroll=Max(0, curr_itm_scroll);
               man_items->insert(&(*man_items)[curr_itm_scroll], S_item(id, 0, 1));

               if(HasAmmo(id))
                  i = cbProc(CB_ADD_SELECTED1, 0x20001, tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, id));
            }else{
               if(add_ammo){
                  (*man_items)[i].ammo += num_ammo;
               }else{
                  ++(*man_items)[i].pieces;
               }
            }
                                    //adjust available list
            if(!add_ammo){
               (*avl_items)[id1].pieces--;
            }else{
               (*avl_items)[id1].ammo -= num_ammo;
            }
            if(!(prm1&0x20000)) C_menu_base1::cbProc(msg, prm1, prm2);
            if(!(*avl_items)[id1].pieces && (!HasAmmo(id) || !(*avl_items)[id1].ammo)){
               cbProc(CB_UNSELECT);
               avl_items->erase(&(*avl_items)[id1]);

               grid_scroll_line = Min(grid_scroll_line, Max(0, ((int)avl_items->size()-1)/ITM_GRID_X));
            }

            //if(prm1&0x20000) return true;

            ShowGridItems();
            ShowScrollItems();

            ShowWeigth();
            if(snd) PlaySound(GM_SND_ADD);
            return true;
         }
         break;

      case CB_REMOVE_SELECTED1:
         {
            C_menu_base1::cbProc(msg, prm1, prm2);

            bool rem_ammo = (prm1&1), snd = (prm1&0x10000);
            int ammo_count = prm2;
            //if(!ammo_count) ammo_count=10;

            assert(selected_image);
            assert(man_items && avl_items);
            int id = GetItemID(selected_image);
            int idid = GetItemIDID(selected_image);
            S_item &itm = (*man_items)[id];
            int i;
            for(i=avl_items->size(); i--; ) if((*avl_items)[i].itm==itm.itm) break;
            if(i==-1){
               avl_items->push_back(S_item(itm.itm, 0, 0));
               i = avl_items->size()-1;
            }

            if(rem_ammo){
               if(!ammo_count) ammo_count = tab_inventory->ItemI(TAB_I_INV_MENU_AMMO_TRANSFER, idid);
               ammo_count = Min(ammo_count, (int)itm.ammo);
               itm.ammo -= ammo_count;
               (*avl_items)[i].ammo += ammo_count;
            }else{
               --itm.pieces;
               (*avl_items)[i].pieces++;
               if(!itm.pieces){
                                    //return back all ammo
                  (*avl_items)[i].ammo += itm.ammo;

                  cbProc(CB_UNSELECT);
                  man_items->erase(&itm);
                  curr_itm_scroll = Min(curr_itm_scroll, (int)man_items->size()-1);
               }
            }
            ShowGridItems();
            ShowScrollItems();
            ShowWeigth();
            if(snd) PlaySound(GM_SND_REMOVE);

            return true;
         }
         break;

      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            PI3D_frame frm(res.frm);
            dword flg(res.flags);
            if(flg&RFLG_HIGHLIGHT)
               res.frm->SetOn(false);
            if(flg&RFLG_BUTTON){
               switch(res.id){
               case ID_BUTTON_MAN_ADD:
               case ID_BUTTON_MAN_REMOVE:
               case ID_BUTTON_ITEM_ADD:
               case ID_BUTTON_ITEM_REMOVE:
               case ID_BUTTON_AMMO_ADD:
               case ID_BUTTON_AMMO_REMOVE:
               case ID_BUTTON_SOLDIER_INFO:
                  frm->SetOn(false);
               }
            }
         }
         break;

      case CB_POINTED_RES_CHANGE:
         {
            S_resource *res = (S_resource*)prm1;
            S_resource *old = (S_resource*)prm2;

            if(old && old!=selected_man){
               if(old->flags&(RFLG_BUTTON|RFLG_MAN_4|RFLG_MAN_8|RFLG_ITEM|RFLG_ITEM_GRID))
                  SetHLight(old, false);
               if(old->flags&(RFLG_ITEM|RFLG_ITEM_GRID)) HideInfo();
               if(old->flags&(RFLG_MAN_4|RFLG_MAN_8)) HideManInfo();
            }
            if(res){
               bool ok(true);
               if(res->flags&(RFLG_MAN_4|RFLG_MAN_8)){
                  ok = (GetManID(res)!=-1);
                  if(ok){
                     if(res!=selected_man) PlaySound(GM_SND_HLIGHT);
                  }
                  pointed_man = res;
                  if(!pointed_man && selected_man) DisplayManInfo(selected_man);
                  else DisplayManInfo(pointed_man);
               }
               if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
                  if(res!=selected_image) PlaySound(GM_SND_HLIGHT);
                  pointed_image = res;
                  if(!pointed_image && selected_image) DisplayInfo(selected_image);
                  else DisplayInfo(pointed_image);
               }
               //if(ok)         //don't hilight empty slot
               if(res->flags&(RFLG_BUTTON|RFLG_MAN_4|RFLG_MAN_8|RFLG_ITEM|RFLG_ITEM_GRID))
                  SetHLight(res, true);
            }else{
               pointed_image=NULL;
               pointed_man=NULL;
               if(selected_image){
                  SetHLight(selected_image, true);
                  DisplayInfo(selected_image);
               }
               if(selected_man){
                  SetHLight(selected_man, true);
                  DisplayManInfo(selected_man);
               }
            }
         }
         break;

      case CB_CLICK:
         {
            S_resource *res = (S_resource*)prm1;
            if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
               if(prm2){
                              //must be selected man in team4
                  if(selected_man && (selected_man->flags&RFLG_MAN_4)){
                     cbProc(CB_SELECT_IMAGE, (dword)res, true);
                  }else
                     PlaySound(GM_SND_CANNOT);
               }
            }
            if(res->flags&(RFLG_MAN_4|RFLG_MAN_8)){
               if(prm2){
                  cbProc(CB_SELECT_MAN, (dword)res, true);
                  if(res->flags&RFLG_MAN_8)
                     cbProc(CB_UNSELECT);
               }
            }
         }
         break;
      }
      return C_menu_base1::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void Proceed(){

      C_game_mission *mission = CreateGameMission();
      if(!mission)
         return;
      game_music.Stop();
      RenderAllBackbuffers();
      C_fstring mname("%s\\game", game_profile.GetCurrMissionName());
      E_MISSION_IO mio = mission->Open(mname, 0);
      if(mio==MIO_OK){
         mission->GameBegin();
         tick_class = mission;
         S_tick_context tc;
         memset(&tc, 0, sizeof(tc));
         tc.time = 1;
         mission->Tick(tc);

                              //try to play before-game animation
         C_class_to_be_ticked_hd *canim = CreateCampAnim(C_fstring("Ingame_%.2i", game_profile.MissionIndex()), mission, false);
         if(canim){
            tick_class = canim;
            canim->Release();
         }
         mission->DisplayObjectives();
      }else{
         ErrReport(C_fstring("Failed to start mission: %s", (const char*)mname), editor);
      }
      mission->Release();
   }

//----------------------------

   virtual void ReturnBack(){

      RenderAllBackbuffers();
      tick_class = CreateMenu(GAME_MENU_MISSION_SELECT);
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   void ShowSoldierInfo(int indx){

      if(net) NetSend();
      RenderAllBackbuffers();
      S_menu_soldier_info si = {indx, this };
      tick_class = CreateMenu(GAME_MENU_SOLDIER_INFO, (dword)&si);
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void OnNetSyncEvent(E_NET_SYNC_EVENT event, dword data){

      switch(event){
      case NS_SOLDIER_INFO:
         ShowSoldierInfo(data);
         break;
      default:
         C_menu_base1::OnNetSyncEvent(event, data);
      }
   }

//----------------------------

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){
      case ID_BUTTON_DOWN:
      case ID_BUTTON_UP:
         {
            int i = -1;
            if(selected_image && (selected_image->flags&RFLG_ITEM))
               i = selected_image->id;
            cbProc(CB_SCROLL, (res->id==ID_BUTTON_UP), true);
            if(i!=-1){
               cbProc(CB_SELECT_IMAGE, (dword)FindResource(i, RFLG_ITEM), false);
               DisplayInfo(selected_image);
               SetHLight(selected_image, true);
            }
         }
         break;

      case ID_BUTTON_DOWN1:
      case ID_BUTTON_UP1:
         {
            cbProc(CB_UNSELECT);
            int i = cbProc(CB_SCROLL_GRID, (res->id==ID_BUTTON_UP1), true);
            if(!i) PlaySound(GM_SND_CANNOT);
         }
         break;

      case ID_BUTTON_MAN_ADD:
         {
            int i = cbProc(CB_ADD_SELECTED_MAN);
            if(!i) PlaySound(GM_SND_CANNOT);
         }
         break;

      case ID_BUTTON_MAN_REMOVE:
         cbProc(CB_REMOVE_SELECTED_MAN);
         break;

      case ID_BUTTON_ITEM_ADD:
         cbProc(CB_ADD_SELECTED1, 0x10000);
         break;
      case ID_BUTTON_AMMO_ADD:
         cbProc(CB_ADD_SELECTED1, 0x10001);
         break;
      case ID_BUTTON_ITEM_REMOVE:
         cbProc(CB_REMOVE_SELECTED1, 0x10000);
         break;
      case ID_BUTTON_AMMO_REMOVE:
         cbProc(CB_REMOVE_SELECTED1, 0x10001);
         break;
      case ID_BUTTON_AUTO_SELECT:
                        //setup auto-selection mode
         mode=MODE_AUTO_SELECT;
         break;

      case ID_BUTTON_SOLDIER_INFO:
         if(selected_man){
            int indx = GetManID(selected_man);
            if(!net){
               ShowSoldierInfo(indx);
            }else
               BeginSyncEvent(NS_SOLDIER_INFO, indx);
         }
         break;
      }
      C_menu_base1::OnButtonPressed(res);
   }

//----------------------------

   virtual void OnDoubleClick(S_resource *res){

      if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
         if(selected_image==res){
            if(res->flags&RFLG_ITEM_GRID){
               int i = cbProc(CB_ADD_SELECTED1, 0x10000);
               if(!i)
                  i = cbProc(CB_ADD_SELECTED1, 0x10001);
            }else
               cbProc(CB_REMOVE_SELECTED1, 0x10000);
         }
      }
      if(res->flags&(RFLG_MAN_4|RFLG_MAN_8)){
         if(selected_man==res){
            bool b;
            if(res->flags&RFLG_MAN_4) b = cbProc(CB_REMOVE_SELECTED_MAN);
            else b = cbProc(CB_ADD_SELECTED_MAN);
            if(!b)
               PlaySound(GM_SND_CANNOT);
         }
      }
   }

//----------------------------

   C_string GetHintMessage(S_resource *res, dword &ret_color, bool &wide){

      if(res->flags&RFLG_MAN_8){
         int man_id = mstate->team_8[res->id];
         if(man_id!=-1){
            int tid = GT_GAME_MENU_SOLDIER_NAME+man_id;
            wide = all_txt.IsWide(tid);
            return all_txt[tid];
         }
      }
      if(res->flags&RFLG_MAN_4){
         int man_id = mstate->team_4[res->id-20];
         if(man_id!=-1){
                              //get name of network player
            if(net){
                        //determine which man it is (will be in game)
               int indx = res->id - 20;
               int j=0;
               for(int i=0; i<indx; i++){
                  if(mstate->team_4[i]!=-1)
                     ++j;
               }
               vector<S_net_player> net_players;
               MakeSortedNetList(net_players, true);
               if(j<net_players.size()){
                  ret_color = 0xffff0000;   //setup color
                  wide = false;
                  return net_players[j].name;
               }
            }
            int tid = GT_GAME_MENU_SOLDIER_NAME+man_id;
            wide = all_txt.IsWide(tid);
            return all_txt[tid];
         }
      }
      if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
         int id = GetItemID(res);
         if(id!=-1){
            int tid;
            if(res->flags&RFLG_ITEM)
               tid = (*man_items)[id].itm;
            else
               tid = (*avl_items)[id].itm;
            if(game_profile.IsDatadisk())
               tid += 1e6;
            tid += GT_GAME_MENU_ITEM_NAME;
            wide = all_txt.IsWide(tid);
            return all_txt[tid];
         }
      }
      return C_menu_base1::GetHintMessage(res, ret_color, wide);
   }
};

//----------------------------
//----------------------------
                              //concrete menu - soldier info
class C_menu_soldier_info: public C_menu_base1{
protected:
   S_mission_state *mstate;
   const S_resource *pointed_image;
   C_smart_ptr<C_text> desc;

                              //on-menu team_8
   short team_8[8];

   C_smart_ptr<C_class_to_be_ticked_hd> save_tc;

#define NUM_BARS 5
   C_smart_ptr<C_text> bar_title_a[NUM_BARS];
   C_smart_ptr<C_text> bar_title_w[NUM_BARS];
   C_smart_ptr<I3D_mesh_level> ml_bar[NUM_BARS];
public:
   C_menu_soldier_info():
      pointed_image(NULL),
      mstate(NULL)
   {
      mstate = &game_profile.mission_states[game_profile.MissionIndex()];
   }

   ~C_menu_soldier_info(){
      HideInfo();
   }
protected:
                              //context: LOWORD = man ID, HIWORD = calling menu ID
   virtual bool Init(E_GAME_MENU id, dword context){

      if(!C_menu_base1::Init(id))
         return false;

      static const S_resource_def r_list[] = {
         {"OK", ID_BUTTON_OK, RFLG_BUTTON | RFLG_HINT},
         {"BACK", ID_BUTTON_BACK, RFLG_BUTTON | RFLG_HINT},
         {"slct_ok", ID_BUTTON_OK+100, RFLG_HIGHLIGHT},
         {"slct_back", ID_BUTTON_BACK+100, RFLG_HIGHLIGHT},
                              //8 men
         {"pic1", 0, RFLG_MAN_8},
         {"pic2", 1, RFLG_MAN_8},
         {"pic3", 2, RFLG_MAN_8},
         {"pic4", 3, RFLG_MAN_8},
         {"pic5", 4, RFLG_MAN_8},
         {"pic6", 5, RFLG_MAN_8},
         {"pic7", 6, RFLG_MAN_8},
         {"pic8", 7, RFLG_MAN_8},
         {"slct_00", 0, RFLG_HIGHLIGHT},
         {"slct_01", 1, RFLG_HIGHLIGHT},
         {"slct_02", 2, RFLG_HIGHLIGHT},
         {"slct_03", 3, RFLG_HIGHLIGHT},
         {"slct_04", 4, RFLG_HIGHLIGHT},
         {"slct_05", 5, RFLG_HIGHLIGHT},
         {"slct_06", 6, RFLG_HIGHLIGHT},
         {"slct_07", 7, RFLG_HIGHLIGHT},
                              //photo
         {"MAN", 0, RFLG_MAN_PHOTO},
         {NULL}
      };
      InitResources(r_list);
      cbProc(CB_ENUM_RESOURCES, 0, 0);

      int i, j;
      for(i=0; i<8; i++)
         team_8[i] = -1;
      for(i=j=0; i<4; i++)
         if(mstate->team_4[i] != (short)-1)
            team_8[j++] = (short)mstate->team_4[i];
      for(i=0; i<8; i++)
         if(mstate->team_8[i]!=-1)
            team_8[j++] = mstate->team_8[i];
      for(i=0; i<8; i++)
         ShowTeam8(i);
                              //init bar texts
      for(i=0; i<NUM_BARS; i++){
                              //find background
         C_fstring str("TEXT%.2i", i+1);
         PI3D_frame frm = scene->FindFrame(str);
         float x = .5f, y = .5f;
         if(frm){
            Map3DTo2D(frm->GetWorldPos(), x, y);
         }else
            OUTTEXT(C_fstring("can't find frame: %s", (const char*)str));

         y /= (screen_aspect_ratio * 1.33333333f);

         bar_title_w[i] = texts.CreateText(all_txt[GT_GAME_MENU_SOLDIER_BAR+i], -1, x, y, .022f,
            0, all_txt.IsWide(GT_GAME_MENU_SOLDIER_BAR+i));
         bar_title_w[i]->Release();

         PI3D_frame bar = scene->FindFrame(C_fstring("LEVEL%.2i", i));
         if(bar && bar->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(bar)->GetVisualType()==I3D_VISUAL_OBJECT){
            ml_bar[i] = I3DCAST_OBJECT(bar)->GetMeshObject()->GetLOD(0);
         }
      }
      HideInfo();

      if(GetType()==GAME_MENU_SOLDIER_INFO){
         const S_menu_soldier_info *si = (S_menu_soldier_info*)context;
         assert(si);
         DisplayInfo(si->soldier_id);
         save_tc = si->save_tc;
      }
      return true;
   }

//----------------------------

   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      int i;

      switch(msg){
      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            dword flg = res.flags;
            if(flg&RFLG_HIGHLIGHT)
               res.frm->SetOn(false);
         }
         break;

      case CB_POINTED_RES_CHANGE:
         {
            S_resource *res = (S_resource*)prm1;
            S_resource *old = (S_resource*)prm2;

            if(old){
               if(old->flags&(RFLG_BUTTON|RFLG_MAN_8)) SetHLight(old, false);
            }
            if(res){
               if(res->flags&(RFLG_BUTTON|RFLG_MAN_8)) SetHLight(res, true);
            }
            pointed_image = res;
         }
         break;
      case CB_CLICK:
         if(prm2==1)          //if pressed...
         {
            S_resource *res = (S_resource*)prm1;
            if(res->flags&RFLG_MAN_8)
               cbProc(CB_SELECT_IMAGE, (dword)res, true);
         }
         break;

      case CB_SELECT_IMAGE:
         {
            S_resource *res = (S_resource*)prm1;
            i = team_8[res->id];
            if(i!=-1){
               DisplayInfo(i);
               if(prm2) PlaySound(GM_SND_SELECT);
            }
         }
         break;
      }

      return C_menu_base1::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void ReturnBack(){

      tick_class = save_tc;
   }

//----------------------------

   void ShowTeam8(int indx){
      const S_resource *res = FindResource(indx, RFLG_MAN_8);
      int mi = team_8[indx];
      ChangeTexture(res->frm, mi==-1 ? "face_no" : C_fstring("man%.2i", mi+1), "faces");
   }

//----------------------------

   void NetIn(E_NETWORK_MESSAGE msg_id, PC_net_in_msg in_msg){

      switch(msg_id){
      case NM_GAME_MENU+1: case NM_GAME_MENU+2: case NM_GAME_MENU+3: case NM_GAME_MENU+4:
      case NM_GAME_MENU+5: case NM_GAME_MENU+6: case NM_GAME_MENU+7: case NM_GAME_MENU+8:
         {
            int for_menu_id = msg_id - NM_GAME_MENU;
            if(for_menu_id!=GetType()){
               if(save_tc){
                  save_tc->NetIn(msg_id, in_msg);
               }else{
                  in_msg->SkipChunk();
                  NET_LOG("Message for different menu, skipping");
               }
               return;
            }
         }
         break;
      }
      C_menu_base1::NetIn(msg_id, in_msg);
   }


//----------------------------

   void HideInfo(){

      int i;
      const S_resource *main = FindResource(0, RFLG_MAN_PHOTO);
      main->frm->SetOn(false);
      desc = NULL;
                              //hide bars
      for(i=0; i<NUM_BARS; i++){
                              //get bar frame
         PI3D_frame frm = scene->FindFrame(C_fstring("LEVEL%.2i", i));
         if(frm)
            frm->SetOn(false);

         bar_title_a[i] = NULL;
      }
   }

//----------------------------

   void DisplayInfo(int man_indx){

      HideInfo();

      const S_resource *main = FindResource(0, RFLG_MAN_PHOTO);
      main->frm->SetOn(true);
                              //make main photo
      ChangeTexture(main->frm, C_fstring("man%.2i", man_indx+1), "faces");
                              //feed texts
      float x = .5f, y = .5f;
      PI3D_frame frm = scene->FindFrame("TEXT00");
      if(frm) Map3DTo2D(frm->GetWorldPos(), x, y);

      bool wide = (game_configuration.language==5);      
      C_string str;
      str = all_txt[GT_GAME_MENU_SOLDIER_NAME+man_indx];
      str = AddGameText(str, MakeGameText("\n"));
      str = AddGameText(str, all_txt[GT_GAME_MENU_SOLDIER_TITL+man_indx]);
      str = AddGameText(str, MakeGameText("\n\n"));
      str = AddGameText(str, all_txt[GT_GAME_MENU_SOLDIER_DESC+man_indx]);
      float fmt_rect[4];
      GetItemRect(frm, fmt_rect);
      y /= (screen_aspect_ratio * 1.33333333f);

      desc = texts.CreateText(str, -1, x, y, .0225f, TEXT_JUSTIFY_CONVERT_EOL,
         wide, fmt_rect);
      desc->Release();
                              //adjust bars
      LPC_table tab(CreateTable());
      LoadSoldierTable(tab, man_indx);
      int i;
      for(i=0; i<NUM_BARS; i++){
                              //get bar frame
         PI3D_frame frm = scene->FindFrame(C_fstring("LEVEL%.2i", i));
         if(!frm) continue;
         int value = tab->ItemI(TAB_I_HUM_BAR1_SHOOTING+i);

         float f = (float)value/100.0f;
         S_vector scl(Min(1.0f, f), 1.0f, 1.0f);
         ScaleObject(frm, ml_bar[i], scl);
         frm->SetOn(true);

         {
            PI3D_frame frm = scene->FindFrame(C_fstring("back%.2i", i+6));
            if(frm) Map3DTo2D(frm->GetWorldPos(), x, y);
            y /= (screen_aspect_ratio * 1.33333333f);

            bar_title_a[i] = texts.CreateText(C_fstring("%i%%", value), -1, x, y+.006f, .0225f, TEXT_JUSTIFY_CENTER,
               false,
               NULL, NULL, NULL, NULL, 0,
               PI*.1f);
            bar_title_a[i]->Release();
         }
      }
      tab->Release();
   }
public:
};

//----------------------------
//----------------------------
                              //concrete menu - select 8 men out of maximum
class C_menu_8_of_40: public C_menu_soldier_info{

   const S_resource *selected_image;

                              //scroll selection among maximum
   int curr_scroll_40;        //index 0-<maximum> into available men in mission state, if man is !avl, he's skipped; -1 = no available

//----------------------------

   int FindUpper(int i) const{

      if(curr_scroll_40!=-1)
      for(int j=i; j=(++j==tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS)) ? 0 : j, j!=i; )
         if(mstate->base_soldier_avl[j]) return j;
      return -1;
   }

//----------------------------

   int FindLower(int i) const{
      if(curr_scroll_40!=-1)
      for(int j=i; j=(j--==0) ? tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS)-1 : j, j!=i; )
         if(mstate->base_soldier_avl[j]) return j;
      return -1;
   }

//----------------------------

   void ShowScrollMen(){

      int lower;
      for(int i=0; i<3; i++){
         int ii=curr_scroll_40;
         if(ii!=-1){
            switch(i){
            case 0: lower=ii=FindLower(ii); break;
            case 2: ii=FindUpper(ii); if(ii==lower) ii=-1; break;
            }
         }
         const S_resource *res = FindResource(i+10, RFLG_MAN_40);
         if(ii==-1 || !mstate->base_soldier_avl[ii]){
            res->frm->SetOn(false);
            FindResource(i+10, RFLG_HIGHLIGHT)->frm->SetOn(false);
         }else{
            res->frm->SetOn(true);
                              //set face
            ChangeTexture(res->frm, C_fstring("man%.2i", ii+1), "faces");
         }
      }
   }

//----------------------------

   int GetManID(const S_resource *res) const{

      if(!res) return -1;
      switch(res->id){
      case 10: return FindLower(curr_scroll_40); break;
      case 11: return curr_scroll_40; break;
      case 12: return FindUpper(curr_scroll_40); break;
      default:                //man from team 8
         //return mstate->team_8[res->id];
         return team_8[res->id];
      }
   }

//----------------------------

   int GetEmptySlot() const{

      for(int i=0; i<8; i++)
         if(team_8[i]==-1)
            return i;
      return -1;
   }

//----------------------------

   int GetEmptySlot1() const{

      for(int i=0; i<8; i++)
         if(mstate->team_8[i]==-1)
            return i;
      return -1;
   }

//----------------------------

   C_smart_ptr<I3D_texture> tx_cache[MAX_BASE_SOLDIERS];

   enum{
      MODE_NORMAL,
      MODE_AUTO_SELECT,
      MODE_CHECK,
      MODE_SCROLL,
      MODE_END,
      MODE_PAUSE,
      MODE_SCROLLED,
   } mode, paused_mode;
                              //auto-select values
   int as_slot;
   int as_avl_indx;
   int as_pause_count;
   bool as_fast;

//----------------------------

   void SetPause(int p = 100){

      if(as_fast)
         return;
      paused_mode = mode;
      mode = MODE_PAUSE;
      as_pause_count = p;
   }

//----------------------------

                              //show OK button when appropriate
   void ManageOk(){

      for(int i=8; i--; )
         if(team_8[i]!=-1)
            break;
      ShowButton(ID_BUTTON_OK, bool(i!=-1));
   }

//----------------------------

   void Unselect(){
      if(selected_image){
         HideInfo();
         if(pointed_image!=selected_image)
            SetHLight(selected_image, false);
         selected_image=NULL;
         ShowButton(ID_BUTTON_MAN_ADD, false);
         ShowButton(ID_BUTTON_MAN_REMOVE, false);
      }
   }

public:
   C_menu_8_of_40():
      mode(MODE_NORMAL),
      selected_image(NULL)
   {
                              //point man40 to 1st available (if any)
      for(curr_scroll_40=0; curr_scroll_40<tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS); curr_scroll_40++)
         if(mstate->base_soldier_avl[curr_scroll_40]) break;
      if(curr_scroll_40==tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS)) curr_scroll_40=-1;
   }

//----------------------------
                              //context = flags (1 = image should be primarily selected)
   virtual bool Init(E_GAME_MENU id, dword context){

      if(!C_menu_soldier_info::Init(id, context))
         return false;

      for(int i=0; i<4; i++)
         game_profile.mission_states[game_profile.MissionIndex()].Move4To8(i);

      static const S_resource_def r_list[] = {
         {"AUTO", ID_BUTTON_AUTO_SELECT, RFLG_BUTTON | RFLG_HINT},
         {"BUT00", ID_BUTTON_MAN_ADD, RFLG_BUTTON | RFLG_HINT},
         {"BUT01", ID_BUTTON_MAN_REMOVE, RFLG_BUTTON | RFLG_HINT},
         {"UP00", ID_BUTTON_UP, RFLG_BUTTON | RFLG_ARROW_UP},
         {"DOWN00", ID_BUTTON_DOWN, RFLG_BUTTON | RFLG_ARROW_DN},

         {"slct_auto", ID_BUTTON_AUTO_SELECT+100, RFLG_HIGHLIGHT},
         {"slct_butt00", ID_BUTTON_MAN_ADD+100, RFLG_HIGHLIGHT},
         {"slct_butt01", ID_BUTTON_MAN_REMOVE+100, RFLG_HIGHLIGHT},
         {"slct_up00", ID_BUTTON_UP+100, RFLG_HIGHLIGHT | RFLG_ARROW_UP},
         {"slct_down00", ID_BUTTON_DOWN+100, RFLG_HIGHLIGHT | RFLG_ARROW_DN},
                              //available men men (3 shown)
         {"MAN1", 10, RFLG_MAN_40},
         {"MAN2", 11, RFLG_MAN_40},
         {"MAN3", 12, RFLG_MAN_40},
         {"slct_10", 10, RFLG_HIGHLIGHT},
         {"slct_11", 11, RFLG_HIGHLIGHT},
         {"slct_12", 12, RFLG_HIGHLIGHT},
         {NULL}
      };
      InitResources(r_list);

      cbProc(CB_ENUM_RESOURCES, 0, 0);
      ShowScrollMen();
                              //pre-load all face's textures
      for(i=0; i<tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS); i++){
         C_fstring str("faces\\man%.2i.png", i+1);
         I3D_CREATETEXTURE ct;
         ct.flags = TEXTMAP_DIFFUSE;
         ct.text_name = str;
         PI3D_texture tp;
         if(I3D_SUCCESS(driver->CreateTexture(&ct, &tp))){
            tx_cache[i] = tp;
            tp->Release();
         }
      }

      HideInfo();
      ManageOk();
      return true;
   }

//----------------------------

   virtual void Tick(const S_tick_context &tc){

      int i;

      if(mode==MODE_NORMAL){
         C_menu_base1::Tick(tc);
         return;
      }

      if(tc.key==K_ESC)
         as_fast = true;

      do{
         switch(mode){
         case MODE_AUTO_SELECT:
            spr_arrow->SetOn(false);
            as_fast = false;
            mode = MODE_CHECK;
            break;

         case MODE_END:

                                 //end of fun, back to user
            spr_arrow->SetOn(true);
            mode = MODE_NORMAL;
            break;

         case MODE_CHECK:
            {
                                 //find empty slot to put into
               as_slot = GetEmptySlot();
               if(as_slot!=-1){
                                 //choose random available man
                  //as_avl_indx = FindUpper(Random(tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS)));
                  as_avl_indx = FindUpper(S_int_random(tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS)));
                  if(as_avl_indx!=-1){
                     Unselect();
                     mode=MODE_SCROLL;
                     SetHLight(FindResource(11, RFLG_MAN_40), true);
                     break;
                  }
               }
               mode = MODE_END;
            }
            break;

         case MODE_SCROLL:
            {
                                 //check if we're at desired position
               if(curr_scroll_40!=as_avl_indx){
                                 //determine shorter direction
                  i = as_avl_indx-curr_scroll_40;
                  if(i>tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS)/2)
                     i -= tab_campaigns->ItemI(TAB_I_CM_NUM_SOLDIERS);
                  bool up = (i<0);

                  cbProc(CB_SCROLL, up, !as_fast);
                  SetPause(20);
                  break;
               }
                                 //select scrolled image
               cbProc(CB_SELECT_IMAGE, (dword)FindResource(11, RFLG_MAN_40), !as_fast);
               DisplayInfo(GetManID(selected_image));
               mode = MODE_SCROLLED;
               SetPause(120);
            }
            break;

         case MODE_SCROLLED:
            {
                                 //add selected image
               cbProc(CB_ADD_SELECTED);
               mode = MODE_CHECK;
               SetPause(120);
            }
            break;

         case MODE_PAUSE:
            as_pause_count -= tc.time;
            if(as_pause_count<=0){
               mode=paused_mode;
            }
            break;
         }
      }while(as_fast && mode!=MODE_NORMAL);

      C_mission::Tick(tc);
      if(hint && !hint->Tick(tc.time))
         hint = NULL;
   }

//----------------------------

   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){
      case CB_SELECT_IMAGE:
         {
            const S_resource *res = (const S_resource*)prm1;
            bool snd = prm2;
                              //check if selection possible
            if(selected_image!=res){
               if(res->flags&RFLG_MAN_8){
                                 //check if something in slot
                  if(team_8[res->id]==-1){
                     PlaySound(GM_SND_CANNOT);
                     return false;
                  }
                  ShowButton(ID_BUTTON_MAN_REMOVE, true);
               }
               ShowButton(ID_BUTTON_MAN_ADD, false);
               if(res->flags&RFLG_MAN_40){
                  ShowButton(ID_BUTTON_MAN_REMOVE, false);
                  if(GetEmptySlot()!=-1){
                                 //unhide ADD button
                     ShowButton(ID_BUTTON_MAN_ADD, true);
                  }
               }
               if(selected_image) SetHLight(selected_image, false);
               selected_image=res;
               if(snd)
                  PlaySound(GM_SND_SELECT);
            }
         }
         break;

      case CB_ADD_SELECTED:
         {
            assert(selected_image);
            int i=GetEmptySlot();
            if(i==-1) return false;
            int id(GetManID(selected_image));
            team_8[i] = id;
            mstate->team_8[GetEmptySlot1()] = id;
            ShowTeam8(i);
                                    //change selection to team8
            Unselect();
            selected_image=FindResource(i, RFLG_MAN_8);
            SetHLight(selected_image, true);
            DisplayInfo(GetManID(selected_image));
                                    //adjust team40
            mstate->base_soldier_avl[id]=false;
            if(id==curr_scroll_40) curr_scroll_40=FindUpper(curr_scroll_40);
            ShowScrollMen();
            ManageOk();
                                    //change buttons to REMOVE
            ShowButton(ID_BUTTON_MAN_REMOVE, true);
            ShowButton(ID_BUTTON_MAN_ADD, false);
         }
         break;

      case CB_REMOVE_SELECTED:
         {

            assert(selected_image);
            int id = GetManID(selected_image);
            mstate->base_soldier_avl[id]=true;
            if(curr_scroll_40==-1) curr_scroll_40=id;

            int i = selected_image->id;
            team_8[i] = -1;
            ShowTeam8(i);

            for(i=4; i--; ) if(mstate->team_4[i]==id) break;
            if(i!=-1) mstate->Move4To8(i);
            for(i=8; i--; ) if(mstate->team_8[i]==id) break;
            assert(i!=-1);
            mstate->team_8[i] = -1;

            Unselect();
            ShowScrollMen();
            ManageOk();
         }
         break;

      case CB_SCROLL:
         {
            bool up = prm1;
            bool snd = prm2;
            int i;
            if(!up) i=FindUpper(curr_scroll_40);
            else i=FindLower(curr_scroll_40);
            if(i!=-1){
               if(selected_image && (selected_image->flags&RFLG_MAN_40))
                  Unselect();
               curr_scroll_40=i;
               ShowScrollMen();
               if(snd) PlaySound(GM_SND_SCROLL);
            }
         }
         break;

      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            PI3D_frame frm(res.frm);
            dword flg(res.flags);
            if(flg&RFLG_MAN_PHOTO)
               frm->SetOn(false);
            if(flg&RFLG_HIGHLIGHT)
               res.frm->SetOn(false);
            if(flg&RFLG_BUTTON){
               switch(res.id){
               case ID_BUTTON_OK:
               case ID_BUTTON_MAN_ADD:
               case ID_BUTTON_MAN_REMOVE:
                  frm->SetOn(false);
               }
            }
         }
         break;

      case CB_POINTED_RES_CHANGE:
         {
            S_resource *res = (S_resource*)prm1;
            S_resource *old = (S_resource*)prm2;

            if(old && old!=selected_image){
               if(old->flags&(RFLG_BUTTON|RFLG_MAN_40|RFLG_MAN_8))
                  SetHLight(old, false);
               if(old->flags&(RFLG_MAN_40|RFLG_MAN_8))
                  HideInfo();
            }
            if(res){
               bool ok(true);
               if(res->flags&(RFLG_MAN_40|RFLG_MAN_8)){
                  if(res->flags&RFLG_MAN_8){
                              //check if something in slot
                     if(team_8[res->id]==-1)
                        ok=false;
                  }
                  if(ok){
                     DisplayInfo(GetManID(res));
                     if(res!=selected_image){
                        PlaySound(GM_SND_HLIGHT);
                     }
                  }
               }
               //if(ok)         //don't hilight empty slot
               if(res->flags&(RFLG_BUTTON|RFLG_MAN_40|RFLG_MAN_8))
                  SetHLight(res, true);
            }

            pointed_image = res;

            if(!pointed_image && selected_image)
               DisplayInfo(GetManID(selected_image));
         }
         break;

      case CB_CLICK:
         {
            S_resource *res = (S_resource*)prm1;
            if(res->flags&(RFLG_MAN_40|RFLG_MAN_8)){
               if(prm2)
                  cbProc(CB_SELECT_IMAGE, (dword)res, true);
            }
         }
         break;
      }
      return C_menu_base1::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void Proceed(){

      RenderAllBackbuffers();
      tick_class = CreateMenu(GAME_MENU_CAMPAIGN_LOAD);
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void ReturnBack(){

      RenderAllBackbuffers();
      tick_class = CreateMenu(GAME_MENU_MISSION_SELECT);
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){
      case ID_BUTTON_DOWN:
      case ID_BUTTON_UP:
         {
            int i = -1;
            if(selected_image && (selected_image->flags&RFLG_MAN_40))
               i = selected_image->id;
            cbProc(CB_SCROLL, (res->id==ID_BUTTON_UP), true);
            if(i!=-1){
               cbProc(CB_SELECT_IMAGE, (dword)FindResource(i, RFLG_MAN_40), false);
               DisplayInfo(GetManID(selected_image));
               SetHLight(selected_image, true);
            }
         }
         break;

      case ID_BUTTON_MAN_ADD:
         cbProc(CB_ADD_SELECTED);
         break;

      case ID_BUTTON_MAN_REMOVE:
         cbProc(CB_REMOVE_SELECTED);
         break;

      case ID_BUTTON_AUTO_SELECT:
                        //setup auto-selection mode
         mode = MODE_AUTO_SELECT;
         break;
      }
      C_menu_base1::OnButtonPressed(res);
   }

//----------------------------

   virtual void OnDoubleClick(S_resource *res){

      if(res->flags&(RFLG_MAN_40|RFLG_MAN_8))
      if(selected_image==res){
         if(res->flags&RFLG_MAN_8){
                        //remove man
            cbProc(CB_REMOVE_SELECTED);
         }else{
            bool b = cbProc(CB_ADD_SELECTED);
            if(!b)
               PlaySound(GM_SND_CANNOT);
         }
      }
   }
};

//----------------------------
//----------------------------
                              //concrete menu - select ammo into campaign
class C_menu_campaign_load: public C_menu_base1{

   S_mission_state *mstate;
                              //on-menu team_8
   short team_8[8];
                              //on-menu item_list
   vector<S_item> item_list;

   const S_resource *pointed_image;
   const S_resource *selected_image;

   int max_carriage;


   vector<int> avl_items;
                              //scroll selection among available items
   int curr_avl_scroll;

   PI3D_model inv_models[5];
//----------------------------

   int FindUpper(int i) const{
      if(i==-1) return -1;
      if(curr_avl_scroll!=-1)
      for(int j=i; j=(++j==avl_items.size()) ? 0 : j, j!=i; )
         return j;
      return -1;
   }

//----------------------------

   int FindLower(int i) const{
      if(i==-1) return -1;
      if(curr_avl_scroll!=-1)
      for(int j=i; j=(j--==0) ? avl_items.size()-1 : j, j!=i; )
         return j;
      return -1;
   }

//----------------------------

   void ShowScrollItems(bool hide_all = false){

      int lower(curr_avl_scroll), upper(curr_avl_scroll);
      for(int i1=0; i1<5; i1++){

         static const imap[] = {2, 1, 3, 0, 4};
         int i(imap[i1]);
                              //hide previous
         if(inv_models[i]){
            inv_models[i]->LinkTo(NULL);
            inv_models[i]->Release();
            inv_models[i]=NULL;
         }
         if(i1>=avl_items.size()) continue;
         if(hide_all) continue;

         int ii = curr_avl_scroll;
         if(ii!=-1){
            switch(i){
            case 0:
               lower = ii = FindLower(ii);
               if(ii==upper) ii=-1;
            case 1:
               lower = ii = FindLower(ii);
               if(ii==upper) ii=-1;
               break;
            case 4:
               upper = ii = FindUpper(ii); if(ii==lower) ii=-1;
            case 3:
               upper = ii = FindUpper(ii); if(ii==lower) ii=-1;
               break;
            }
         }
         if(ii!=-1 && !hide_all){
            const S_resource *res = FindResource(i+10, RFLG_ITEM);
            //S_item &itm = avl_items[ii];
            int itm = avl_items[ii];
            PI3D_model mod = driver->CreateModel();
            inv_models[i] = mod;

            const char *mname = tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm, 16);
            model_cache.Open(mod, mname, scene, 0, ErrReport, editor);

#ifdef _DEBUG
            mod->SetName(mname);
#endif
            mod->SetAnimation(0, NULL);
            mod->LinkTo(res->frm);
         }
      }
   }

//----------------------------

//#define ITM_GRID_X 4
#define ITM_GRID_Y 5
   PI3D_frame itm_grid[3][ITM_GRID_Y][ITM_GRID_X];   //[dummy/hlight/item][][]
   C_smart_ptr<C_text> ammo_txt[ITM_GRID_Y][ITM_GRID_X];
   int grid_scroll_line;

//----------------------------

   void ShowGridItems(bool hide_all = false){

      int i, x, y;
                              //hide previous
      for(y=0; y<ITM_GRID_Y; y++)
      for(x=0; x<ITM_GRID_X; x++){
         if(itm_grid[2][y][x]){
            itm_grid[2][y][x]->LinkTo(NULL);
            itm_grid[2][y][x]->Release();
            itm_grid[2][y][x] = NULL;
         }
         ammo_txt[y][x] = NULL;
      }
      if(hide_all)
         return;
                              //show new
      int num = Min(ITM_GRID_X*ITM_GRID_Y, (int)item_list.size()-grid_scroll_line*ITM_GRID_X);
      float scrn_sx = igraph->Scrn_sx();
      for(i=0; i<num; i++){
         x = i%ITM_GRID_X;
         y = i/ITM_GRID_X;

         S_item &itm = item_list[i+grid_scroll_line*ITM_GRID_X];
         PI3D_model mod = driver->CreateModel();
         itm_grid[2][y][x] = mod;

         const char *mname = tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16);
         model_cache.Open(mod, mname, scene, 0, ErrReport, editor);

#ifdef _DEBUG
         mod->SetName(tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16));
#endif
         mod->SetAnimation(0, NULL);
         mod->LinkTo(itm_grid[0][y][x]);

         S_vector in(itm_grid[0][y][x]->GetWorldPos()), out;
         S_vectorw v4;
         scene->TransformPoints(&in, &v4, 1);
         out = v4;
         float x1 = out.x / scrn_sx, y1 = out.y / scrn_sx;
         //x1 -= .1f;
         y1 += .0225f;
         C_fstring str("%i", itm.pieces);
         if(HasAmmo(itm.itm)) str += C_fstring("/%i", itm.ammo);
         y1 /= (screen_aspect_ratio * 1.33333333f);

         ammo_txt[y][x] = texts.CreateText(str, -1, x1, y1, .0225f, TEXT_JUSTIFY_CENTER);
         ammo_txt[y][x]->Release();
      }
   }

//----------------------------

   void ShowTeam8(int indx){

      const S_resource *res = FindResource(indx, RFLG_MAN_8);
      int mi = team_8[indx];
      ChangeTexture(res->frm, mi==-1 ? "face_no" : C_fstring("man%.2i", mi+1), "faces");
   }

//----------------------------

   int GetItemID(const S_resource *res) const{

      int i(-1);
      if(res)
      switch(res->id){
      case 10: i = FindLower(FindLower(curr_avl_scroll)); break;
      case 11: i = FindLower(curr_avl_scroll); break;
      case 12: i = curr_avl_scroll; break;
      case 13: i = FindUpper(curr_avl_scroll); break;
      case 14: i = FindUpper(FindUpper(curr_avl_scroll)); break;
      default:
         return res->id - 1000 + grid_scroll_line*ITM_GRID_X;
      }
      return i;
   }

//----------------------------

   int GetItemIDID(const S_resource *res) const{

      int id = GetItemID(res);
      if(res->flags&RFLG_ITEM)
         return avl_items[id];
      return item_list[id].itm;
   }

//----------------------------

   void asAddMissionItems(int miss_indx, vector<S_item> &want_list) const{

      int i, j;
      for(i=0; i<64; i++){
         int amount = tab_game_cfg->ItemI(TAB_I64_GAME_MIS_ITEM_AMOUNT, miss_indx*64 + i);
         if(amount){
            bool has_ammo = C_inventory::HasAmmo(i);
            int pcs = has_ammo ? 1 : amount;
            int ammo = has_ammo ? amount : 0;
            if(has_ammo && !ammo) ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, i);
                                 //find among existing
            for(j=want_list.size(); j--; )
            if(want_list[j].itm==i) break;
            if(j==-1)
               want_list.push_back(S_item(i, ammo, pcs));
            else{
               want_list[j].pieces += pcs;
               want_list[j].ammo += ammo;
            }
         }
      }
   }

//----------------------------

   C_smart_ptr<C_text> desc;
   C_smart_ptr<C_text> bar_num;
   C_smart_ptr<I3D_mesh_level> ml_bar;

protected:

public:
   C_menu_campaign_load():
      mode(MODE_NORMAL),
      selected_image(NULL),
      pointed_image(NULL),
      max_carriage(0),
      curr_avl_scroll(0),
      grid_scroll_line(0)
   {
      mstate = &game_profile.mission_states[game_profile.MissionIndex()];
      memset(inv_models, 0, sizeof(inv_models));
      memset(itm_grid, 0, sizeof(itm_grid));

      int i, j;
      for(i=0; i<8; i++) team_8[i] = -1;
      for(i=j=0; i<4; i++) if(mstate->team_4[i]!=-1) team_8[j++] = (short)mstate->team_4[i];
      for(i=0; i<8; i++) if(mstate->team_8[i]!=-1) team_8[j++] = mstate->team_8[i];

      for(i=0; i<8; i++){
         short plr_id = team_8[i];
         if(plr_id!=-1){
            LPC_table tab(CreateTable());
            LoadSoldierTable(tab, plr_id);
            max_carriage += CountManCarriage(tab);
            tab->Release();
         }
      }
                              //collect all items from men
      for(i=0; i<4; i++){
         if(mstate->team_4[i]==-1) continue;
         for(j=0; j<mstate->team_4[i].item_list.size(); j++)
            item_list.push_back(mstate->team_4[i].item_list[j]);
      }
      for(j=0; j<mstate->item_list.size(); j++)
         item_list.push_back(mstate->item_list[j]);
   }

   ~C_menu_campaign_load(){
      ShowScrollItems(true);
      ShowGridItems(true);
      HideInfo();
      HideWeight();

      for(int y=0; y<ITM_GRID_Y; y++)
      for(int x=0; x<ITM_GRID_X; x++)
      for(int i=0; i<2; i++)
      if(itm_grid[i][y][x]) itm_grid[i][y][x]->Release();
   }

//----------------------------
                              //context = flags (1 = image should be primarily selected)
   virtual bool Init(E_GAME_MENU id, dword context){

      if(!C_menu_base1::Init(id))
         return false;

                              //remove all men from team4
      for(int i=0; i<4; i++)
         game_profile.mission_states[game_profile.MissionIndex()].Move4To8(i);
      {
         PI3D_frame fm = scene->FindFrame("item05");

         const S_vector &mn = fm->GetPos(),
            &mx = scene->FindFrame("item06")->GetPos();
         const float scl = fm->GetScale();
         float dx = (mx.x-mn.x) / (ITM_GRID_X-1);
         float dy = (mx.y-mn.y) / (ITM_GRID_Y-1);
                              //create grid of dummys and highlights
         for(int y=0; y<ITM_GRID_Y; y++)
         for(int x=0; x<ITM_GRID_X; x++){
            for(int i=0; i<2; i++){
               PI3D_model mod = driver->CreateModel();
               if(i){
                  model_cache.Open(mod, "inv_frm", scene, 0, ErrReport, editor);
#ifdef _DEBUG
                  mod->SetName("inv_frm");
#endif
               }
               PI3D_frame frm = mod;
               itm_grid[i][y][x] = frm;
               frm->SetPos(S_vector(mn.x + dx*x, mn.y + dy*y, !i ? mn.z : mn.z+.5f));
               frm->SetScale(scl);
               frm->Update();
               frm->LinkTo(scene->GetPrimarySector());
               res_list.push_back(S_resource());
               S_resource &res = res_list.back();
               res.frm = frm;
               res.id = y*ITM_GRID_X + x;
               //if(i)
                  res.id += 1000;
               res.flags = i ? RFLG_HIGHLIGHT : (RFLG_ITEM_GRID | RFLG_HINT);
            }
         }
      }

      static const S_resource_def r_list[] = {
         {"OK", ID_BUTTON_OK, RFLG_BUTTON | RFLG_HINT},
         {"BACK", ID_BUTTON_BACK, RFLG_BUTTON | RFLG_HINT},
         {"AUTO", ID_BUTTON_AUTO_SELECT, RFLG_BUTTON | RFLG_HINT},
         {"BUT00", ID_BUTTON_ITEM_ADD, RFLG_BUTTON | RFLG_HINT},
         {"BUT01", ID_BUTTON_ITEM_REMOVE, RFLG_BUTTON | RFLG_HINT},
         {"BUT02", ID_BUTTON_AMMO_ADD, RFLG_BUTTON | RFLG_HINT},
         {"BUT03", ID_BUTTON_AMMO_REMOVE, RFLG_BUTTON | RFLG_HINT},
         {"UP01", ID_BUTTON_UP, RFLG_BUTTON | RFLG_ARROW_UP},
         {"DOWN01", ID_BUTTON_DOWN, RFLG_BUTTON | RFLG_ARROW_DN},
         {"UP00", ID_BUTTON_UP1, RFLG_BUTTON | RFLG_INVISIBLE | RFLG_ARROW_UP},
         {"DOWN00", ID_BUTTON_DOWN1, RFLG_BUTTON | RFLG_INVISIBLE | RFLG_ARROW_DN},

         {"slct_ok", ID_BUTTON_OK+100, RFLG_HIGHLIGHT},
         {"slct_back", ID_BUTTON_BACK+100, RFLG_HIGHLIGHT},
         {"slct_auto", ID_BUTTON_AUTO_SELECT+100, RFLG_HIGHLIGHT},
         {"slct_butt00", ID_BUTTON_ITEM_ADD+100, RFLG_HIGHLIGHT},
         {"slct_butt01", ID_BUTTON_ITEM_REMOVE+100, RFLG_HIGHLIGHT},
         {"slct_butt02", ID_BUTTON_AMMO_ADD+100, RFLG_HIGHLIGHT},
         {"slct_butt03", ID_BUTTON_AMMO_REMOVE+100, RFLG_HIGHLIGHT},
         {"slct_up01", ID_BUTTON_UP+100, RFLG_HIGHLIGHT | RFLG_ARROW_UP},
         {"slct_down01", ID_BUTTON_DOWN+100, RFLG_HIGHLIGHT | RFLG_ARROW_DN},
         //{"slct_up00", ID_BUTTON_UP1+100, RFLG_HIGHLIGHT},
         //{"slct_down00", ID_BUTTON_DOWN1+100, RFLG_HIGHLIGHT},
                              //8 men
         {"pic1", 0, RFLG_MAN_8 | RFLG_HINT},
         {"pic2", 1, RFLG_MAN_8 | RFLG_HINT},
         {"pic3", 2, RFLG_MAN_8 | RFLG_HINT},
         {"pic4", 3, RFLG_MAN_8 | RFLG_HINT},
         {"pic5", 4, RFLG_MAN_8 | RFLG_HINT},
         {"pic6", 5, RFLG_MAN_8 | RFLG_HINT},
         {"pic7", 6, RFLG_MAN_8 | RFLG_HINT},
         {"pic8", 7, RFLG_MAN_8 | RFLG_HINT},
         {"slct_00", 0, RFLG_HIGHLIGHT},
         {"slct_01", 1, RFLG_HIGHLIGHT},
         {"slct_02", 2, RFLG_HIGHLIGHT},
         {"slct_03", 3, RFLG_HIGHLIGHT},
         {"slct_04", 4, RFLG_HIGHLIGHT},
         {"slct_05", 5, RFLG_HIGHLIGHT},
         {"slct_06", 6, RFLG_HIGHLIGHT},
         {"slct_07", 7, RFLG_HIGHLIGHT},

         {"item00", 10, RFLG_ITEM | RFLG_HINT},
         {"item01", 11, RFLG_ITEM | RFLG_HINT},
         {"item02", 12, RFLG_ITEM | RFLG_HINT},
         {"item03", 13, RFLG_ITEM | RFLG_HINT},
         {"item04", 14, RFLG_ITEM | RFLG_HINT},
         {"inv_frm00", 10, RFLG_HIGHLIGHT},
         {"inv_frm01", 11, RFLG_HIGHLIGHT},
         {"inv_frm02", 12, RFLG_HIGHLIGHT},
         {"inv_frm03", 13, RFLG_HIGHLIGHT},
         {"inv_frm04", 14, RFLG_HIGHLIGHT},

         {NULL}
      };
      InitResources(r_list);
      cbProc(CB_ENUM_RESOURCES, 0, 0);

      PI3D_frame bar = scene->FindFrame("LEVEL");
      if(bar && bar->GetType()==FRAME_VISUAL && I3DCAST_VISUAL(bar)->GetVisualType()==I3D_VISUAL_OBJECT){
         ml_bar = I3DCAST_OBJECT(bar)->GetMeshObject()->GetLOD(0);
      }


      for(i=0; i<8; i++)
         ShowTeam8(i);

                              //fill-in campaign items
      for(i=0; i<32; i++){
         int id = tab_game_cfg->ItemI(TAB_I_GAME_CAMP_ITEMS_ID, i+game_profile.curr_camp*32);
         if(id)
            avl_items.push_back(id);
      }

      if(!avl_items.size())
         curr_avl_scroll = -1;
      ShowScrollItems();
      ShowGridItems();
      ShowWeigth();
      return true;
   }

//----------------------------

   void MakeAutoList(){

      int i, j;

      as_want_list.clear();
      int num_miss = game_profile.NumMissionsInCamp(game_profile.curr_camp);
                              //add 'must' items of all missions
      for(i=0; i<num_miss; i++){
         asAddMissionItems(game_profile.MissionIndex()+i , as_want_list);
      }

      int num_men = 0;
      for(i=0; i<8; i++) if(team_8[i]!=-1) ++num_men;

                              //add selected items

                              //binoculars: Min(2, num_miss)
                              //camera:      ''
      as_want_list.push_back(S_item(1, 0, Min(2, num_miss)));
      as_want_list.push_back(S_item(11, 0, Min(2, num_miss)));

                              //knives: num_men
      as_want_list.push_back(S_item(29, 0, num_men));

                              //grenades: num_miss * num_men
      i = asFindItem(5, avl_items);
      if(i) as_want_list.push_back(S_item(i, 0, num_miss*num_men));

                              //autom. guns: num_men/2
      i = asFindItem(1, avl_items);
      if(i){
         j = (num_men+1)/2;
         if(j){
            int ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, i) * j * num_miss;
            as_want_list.push_back(S_item(i, ammo, j));
         }
      }

                              //snipers: num_men/4
      i = asFindItem(2, avl_items);
      if(i){
         j = (num_men+2)/4;
         if(j){
            int ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, i) * j * num_miss;
            as_want_list.push_back(S_item(i, ammo, j));
         }
      }
                              //heavy guns: Min(2, num_men/4)
      i = asFindItem(3, avl_items);
      if(i){
         j = Min(2, (num_men+2)/4);
         if(j){
            int ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, i) * j * num_miss;
            as_want_list.push_back(S_item(i, ammo, j));
         }
      }
                              //colts: num_men/2, ammo = base_ammo * num_colts * num_miss
      i = asFindItem(4, avl_items);
      if(i){
         j = (num_men+1)/2;
         if(j){
            int ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, i) * j * num_miss;
            as_want_list.push_back(S_item(i, ammo, j));
         }
      }
   }

//----------------------------

   enum{
      MODE_NORMAL,
      MODE_AUTO_SELECT,
      MODE_END,
      MODE_PAUSE,
      MODE_PROCESS_EXIST,
      MODE_FILL,
   } mode, paused_mode;
                              //auto-select values
   vector<S_item> as_want_list;
   int as_curr_indx;
   bool as_fast;
   int as_pause_count;

   void SetPause(int p = 100){
      if(as_fast) return;
      paused_mode=mode;
      mode=MODE_PAUSE;
      as_pause_count=p;
   }

   virtual void Tick(const S_tick_context &tc){

      int i;

      if(mode==MODE_NORMAL){
         C_menu_base1::Tick(tc);
         return;
      }

      if(tc.key==K_ESC)
         as_fast = true;

      do{
      switch(mode){
      case MODE_AUTO_SELECT:
         spr_arrow->SetOn(false);
         as_fast=false;
         as_want_list.clear();
         if(!avl_items.size()){
            mode=MODE_END;
            break;
         }

         MakeAutoList();

         Unselect();
         mode=MODE_PROCESS_EXIST;
         as_curr_indx = 0;
         break;

      case MODE_END:

                              //end of fun, back to user
         spr_arrow->SetOn(true);
         mode=MODE_NORMAL;
         break;

      case MODE_PROCESS_EXIST:
         //if(as_curr_indx >= mstate->item_list.size())
         if(as_curr_indx >= item_list.size()){
            as_curr_indx = 0;
            mode = MODE_FILL;
            cbProc(CB_SELECT_IMAGE, (dword)FindResource(12, RFLG_ITEM), !as_fast);
            break;
         }
         if(!selected_image){
            i = MakeVisible(as_curr_indx);
            if(i){
               SetPause(30);
               break;
            }
            cbProc(CB_SELECT_IMAGE, (dword)FindResource(1000+as_curr_indx - grid_scroll_line*ITM_GRID_X, RFLG_ITEM_GRID), !as_fast);
            SetPause(30);
            break;
         }
                              //check if matching
         //for(i=as_want_list.size(); i--; ) if(as_want_list[i].itm == mstate->item_list[as_curr_indx].itm) break;
         for(i=as_want_list.size(); i--; ) if(as_want_list[i].itm == item_list[as_curr_indx].itm) break;
         if(i==-1){
                              //drop this item
            cbProc(CB_REMOVE_SELECTED1, (!as_fast ? 0x10000 : 0));
            SetPause(50);
            break;
         }
         {
                              //get delta pieces
            //int pcs_delta = mstate->item_list[as_curr_indx].pieces - as_want_list[i].pieces;
            int pcs_delta = item_list[as_curr_indx].pieces - as_want_list[i].pieces;
            if(pcs_delta > 0){
               cbProc(CB_REMOVE_SELECTED1, (!as_fast ? 0x10000 : 0));
               SetPause(50);
               break;
            }
            //if(HasAmmo(mstate->item_list[as_curr_indx].itm))
            if(HasAmmo(item_list[as_curr_indx].itm)){
                              //get delta ammo
               //int ammo_delta = mstate->item_list[as_curr_indx].ammo - as_want_list[i].ammo;
               int ammo_delta = item_list[as_curr_indx].ammo - as_want_list[i].ammo;
               if(ammo_delta > 0){
                  cbProc(CB_REMOVE_SELECTED1, (!as_fast ? 0x10000 : 0) | 1, Min(10, ammo_delta));
                  SetPause(50);
                  break;
               }
            }
         }
         Unselect();
         ++as_curr_indx;
         break;

      case MODE_FILL:
                              //check end
         if(as_curr_indx >= as_want_list.size()){
            mode = MODE_END;
            break;
         }
         {
            int id = as_want_list[as_curr_indx].itm;
                              //check if we need to process this
            //for(i=mstate->item_list.size(); i--; ) if(mstate->item_list[i].itm==id) break;
            for(i=item_list.size(); i--; ) if(item_list[i].itm==id) break;
            if(i!=-1){
               //if(mstate->item_list[i].pieces==as_want_list[as_curr_indx].pieces &&
               if(item_list[i].pieces==as_want_list[as_curr_indx].pieces &&
                  //mstate->item_list[i].ammo==as_want_list[as_curr_indx].ammo)
                  item_list[i].ammo==as_want_list[as_curr_indx].ammo){
                              //item ok, proceed
                  ++as_curr_indx;
                  break;
               }
            }
                              //find among available
            for(i=avl_items.size(); i--; )
               if(avl_items[i]==id)
                  break;
            if(i==-1){        //this shouldn't happen in final
               ++as_curr_indx;
               break;
            }
                              //scroll to desired item
            if(i!=curr_avl_scroll){
                              //determine shorter direction
               i = i-curr_avl_scroll;
               if(i>avl_items.size()/2) i -= avl_items.size();
               bool up = (i<0);
               cbProc(CB_SCROLL, up, !as_fast);
               cbProc(CB_SELECT_IMAGE, (dword)FindResource(12, RFLG_ITEM), !as_fast);
               DisplayInfo(selected_image);
               SetPause(20);
               break;
            }
                              //now add pieces/ammo
            //for(i=mstate->item_list.size(); i--; ) if(mstate->item_list[i].itm==id) break;
            for(i=item_list.size(); i--; ) if(item_list[i].itm==id) break;
                              //check existence | pieces
            if(i==-1 ||
               //mstate->item_list[i].pieces < as_want_list[as_curr_indx].pieces)
               item_list[i].pieces < as_want_list[as_curr_indx].pieces){

               i = cbProc(CB_ADD_SELECTED1, (!as_fast ? 0x10000 : 0), as_want_list[as_curr_indx].ammo);
               if(!i)         //overload - should be computed properly in final
                  ++as_curr_indx;
               SetPause(50);
               break;
            }
            if(HasAmmo(id)){
                              //check ammo
               //int delta_ammo = as_want_list[as_curr_indx].ammo - mstate->item_list[i].ammo;
               int delta_ammo = as_want_list[as_curr_indx].ammo - item_list[i].ammo;
               if(delta_ammo > 0){
                  i = cbProc(CB_ADD_SELECTED1, (!as_fast ? 0x10000 : 0) | 1, Min(10, delta_ammo));
                  if(!i)         //overload - should be computed properly in final
                     ++as_curr_indx;
                  SetPause(50);
                  break;
               }
            }
            ++as_curr_indx;
         }
         break;

      case MODE_PAUSE:
         as_pause_count -= tc.time;
         if(as_pause_count<=0) mode = paused_mode;
         break;
      }
      }while(as_fast && mode!=MODE_NORMAL);

      C_mission::Tick(tc);
      if(hint && !hint->Tick(tc.time))
         hint = NULL;
   }

//----------------------------

   bool MakeVisible(int indx){
      if(indx < grid_scroll_line*ITM_GRID_X){
         cbProc(CB_SCROLL_GRID, true);
         return true;
      }
      int maxx = (grid_scroll_line+ITM_GRID_Y)*ITM_GRID_X;
      if(indx >= maxx){
         cbProc(CB_SCROLL_GRID, false);
         return true;
      }
      return false;
   }

//----------------------------

   void HideInfo(){
      desc = NULL;
   }

//----------------------------

   void DisplayInfo(const S_resource *res){

      HideInfo();

                              //feed texts
      float x(.5), y(.5);
      PI3D_frame frm = scene->FindFrame("TEXT00");
      if(frm) Map3DTo2D(frm->GetWorldPos(), x, y);

      int tid = GetItemIDID(res);
      if(game_profile.IsDatadisk())
         tid += 1e6;
      C_string str;
      str = all_txt[GT_GAME_MENU_ITEM_NAME + tid];
      bool wide = all_txt.IsWide(GT_GAME_MENU_ITEM_NAME + tid);
      str = AddGameText(str, MakeGameText("\n"));

      str = AddGameText(str, all_txt[GT_GAME_MENU_ITEM_DESC + tid]);
      float fmt_rect[4];
      GetItemRect(frm, fmt_rect);
      y /= (screen_aspect_ratio * 1.33333333f);
      desc = texts.CreateText(str, -1, x, y, .0225f, TEXT_JUSTIFY_CONVERT_EOL, wide, fmt_rect);
      desc->Release();
   }

//----------------------------

   void Unselect(){
      if(selected_image){
         HideInfo();
         if(pointed_image!=selected_image)
            SetHLight(selected_image, false);
         selected_image = NULL;
         pointed_image = NULL;
         ShowButton(ID_BUTTON_ITEM_ADD, false);
         ShowButton(ID_BUTTON_AMMO_ADD, false);
         ShowButton(ID_BUTTON_ITEM_REMOVE, false);
         ShowButton(ID_BUTTON_AMMO_REMOVE, false);
      }
   }

//----------------------------

   int MaxGridScroll() const{
      //return Max(0, (int)(mstate->item_list.size()-1)/ITM_GRID_X);
      return Max(0, (int)(item_list.size()-1)/ITM_GRID_X);
   }

//----------------------------

   float CountWeight() const{
      float w(0);
      for(int i=0; i<item_list.size(); i++){
         const S_item &itm = item_list[i];
         w += tab_inventory->ItemF(TAB_F_INV_WEIGTH, itm.itm) * itm.pieces;
         if(HasAmmo(itm.itm))
            w += tab_inventory->ItemF(TAB_F_INV_AMMO_WEIGTH, itm.itm) * itm.ammo;
      }
      return w;
   }

//----------------------------

   void ShowWeigth(){

      float w = CountWeight();
      float f = 0.0f;
      if(max_carriage)
         f = w / (float)max_carriage;

      PI3D_frame frm = scene->FindFrame("LEVEL");
      if(frm){
         S_vector scl(Min(1.0f, f), 1.0f, 1.0f);
         ScaleObject(frm, ml_bar, scl);
      }
      HideWeight();
      {
         float x, y;
         PI3D_frame frm = scene->FindFrame("bar");
         if(frm) Map3DTo2D(frm->GetWorldPos(), x, y);
         y /= (screen_aspect_ratio * 1.33333333f);
         bar_num = texts.CreateText(C_fstring("%i%%", Min(100, (int)(f*100.0f))), -1, x, y+.006f, .0225f, TEXT_JUSTIFY_CENTER);
         bar_num->Release();
      }
   }

//----------------------------

   void HideWeight(){

      bar_num = NULL;
   }

//----------------------------

                              //find particular item at some man's inventory
   bool FindItem(const S_item &itm, int &t4_i, int &itm_i) const{
      for(t4_i=4; t4_i--; ){
         for(itm_i=mstate->team_4[t4_i].item_list.size(); itm_i--; )
         if(mstate->team_4[t4_i].item_list[itm_i]==itm) return true;
      }
      return false;
   }

public:
   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_SELECT_IMAGE:
         {
            const S_resource *res = (const S_resource *)prm1;
            bool snd = prm2;
                                    //check if selection possible
            if(selected_image!=res){
               ShowButton(ID_BUTTON_ITEM_ADD, false);
               ShowButton(ID_BUTTON_AMMO_ADD, false);
               ShowButton(ID_BUTTON_ITEM_REMOVE, false);
               ShowButton(ID_BUTTON_AMMO_REMOVE, false);

               int id = GetItemIDID(res);

               if(res->flags&RFLG_ITEM){
                  ShowButton(ID_BUTTON_ITEM_ADD, true);
                  if(HasAmmo(id))
                     ShowButton(ID_BUTTON_AMMO_ADD, true);
               }
               if(res->flags&RFLG_ITEM_GRID){
                  ShowButton(ID_BUTTON_ITEM_REMOVE, true);
                  if(HasAmmo(id))
                     ShowButton(ID_BUTTON_AMMO_REMOVE, true);
               }
               if(selected_image) SetHLight(selected_image, false);
               selected_image=res;
               if(selected_image) SetHLight(selected_image, true);
               if(snd)
                  PlaySound(GM_SND_SELECT);
            }
         }
         break;

      case CB_ADD_SELECTED1:
         {
            bool add_ammo = (prm1&1), snd = (prm1&0x10000);
            int num_ammo = prm2;

            assert(selected_image);
            int id(GetItemIDID(selected_image));
            int grid_id;
            int i;

            float w = CountWeight();
                                    //check if we have such item already
            //for(i=mstate->item_list.size(); i--; ) if(mstate->item_list[i].itm==id) break;
            for(i=item_list.size(); i--; ) if(item_list[i].itm==id) break;
            bool skip1(false);
            if(i==-1){
               if(add_ammo){
                  if(snd) PlaySound(GM_SND_CANNOT);
                  return false;
               }
                                    //check weight
               float item_w = tab_inventory->ItemF(TAB_F_INV_WEIGTH, id);
               if(w+item_w > max_carriage){
                  if(snd) PlaySound(GM_SND_CANNOT);
                  return false;
               }
                                    //add item
               item_list.push_back(S_item(id, 0, 1));
               mstate->item_list.push_back(S_item(id, 0, 1));
                                    //add minimal munition
               //i = cbProc(CB_ADD_SELECTED1, 0x20001, tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, id));
               if(!num_ammo) num_ammo = tab_inventory->ItemI(TAB_I_INV_DEFAULT_AMMO, id);
               i = cbProc(CB_ADD_SELECTED1, 0x20001, num_ammo);
               if(i){
                  skip1 = true;
                  if(snd) PlaySound(GM_SND_ADD);
               }
               //else grid_id = mstate->item_list.size()-1;
               else grid_id = item_list.size()-1;
            }else{
               if(add_ammo){
                  if(!num_ammo) num_ammo = tab_inventory->ItemI(TAB_I_INV_MENU_AMMO_TRANSFER, id);
                                    //check weight
                  float item_w = tab_inventory->ItemF(TAB_F_INV_AMMO_WEIGTH, id) * num_ammo;
                  if(w+item_w > max_carriage){
                     if(snd) PlaySound(GM_SND_CANNOT);
                     return false;
                  }
                              //find this item in real heap|at man
                  for(int j=mstate->item_list.size(); j--; ) if(mstate->item_list[j]==item_list[i]) break;
                  if(j!=-1) mstate->item_list[j].ammo += num_ammo;
                  else{       //scan all soldiers in team4
                     int t4_i, itm_i;
                     bool b;
                     b = FindItem(item_list[i], t4_i, itm_i);
                     assert(b);
                     mstate->team_4[t4_i].item_list[itm_i].ammo += num_ammo;
                  }

                  //mstate->item_list[i].ammo += num_ammo;
                  item_list[i].ammo += num_ammo;
               }else{
                                    //check weight
                  float item_w = tab_inventory->ItemF(TAB_F_INV_WEIGTH, id);
                  if(w+item_w > max_carriage){
                     if(snd) PlaySound(GM_SND_CANNOT);
                     return false;
                  }
                              //find this item in real heap|at man
                  for(int j=mstate->item_list.size(); j--; ) if(mstate->item_list[j]==item_list[i]) break;
                  if(j!=-1) ++mstate->item_list[j].pieces;
                  else{       //scan all soldiers in team4
                     int t4_i, itm_i;
                     bool b;
                     b = FindItem(item_list[i], t4_i, itm_i);
                     assert(b);
                     ++mstate->team_4[t4_i].item_list[itm_i].pieces;
                  }

                  //++mstate->item_list[i].pieces;
                  ++item_list[i].pieces;
               }
               grid_id = i;
            }
            if(!skip1){
               ShowGridItems();
               ShowWeigth();
               if(snd) PlaySound(GM_SND_ADD);

               while(MakeVisible(grid_id));
            }
            if(prm1&0x20000) return true;
         }
         break;

      case CB_REMOVE_SELECTED1:
         {
            bool rem_ammo = (prm1&1), snd = (prm1&0x10000);
            int ammo_count = prm2;
            //if(!ammo_count) ammo_count=10;

            assert(selected_image);
            int id = GetItemID(selected_image);
            int idid(GetItemIDID(selected_image));

            //S_item &itm = mstate->item_list[id];
            S_item &itm = item_list[id];
            if(rem_ammo){
               if(!ammo_count) ammo_count = tab_inventory->ItemI(TAB_I_INV_MENU_AMMO_TRANSFER, idid);
                              //find this item in real heap|at man
               for(int j=mstate->item_list.size(); j--; ) if(mstate->item_list[j]==itm) break;
               if(j!=-1) mstate->item_list[j].ammo = Max(0, (int)mstate->item_list[j].ammo-ammo_count);
               else{       //scan all soldiers in team4
                  int t4_i, itm_i;
                  bool b;
                  b = FindItem(itm, t4_i, itm_i);
                  assert(b);
                  mstate->team_4[t4_i].item_list[itm_i].ammo = Max(0, (int)mstate->team_4[t4_i].item_list[itm_i].ammo-ammo_count);
               }

               itm.ammo = Max(0, (int)itm.ammo-ammo_count);
            }else{
               if(itm.pieces==1){
                              //find this item in real heap|at man
                  for(int j=mstate->item_list.size(); j--; ) if(mstate->item_list[j]==itm) break;
                  if(j!=-1) mstate->item_list.erase(&mstate->item_list[j]);
                  else{       //scan all soldiers in team4
                     int t4_i, itm_i;
                     bool b;
                     b = FindItem(itm, t4_i, itm_i);
                     assert(b);
                     mstate->team_4[t4_i].item_list.erase(&mstate->team_4[t4_i].item_list[itm_i]);
                  }

                  //mstate->item_list.erase(&itm);
                  item_list.erase(&itm);
                  Unselect();

                  grid_scroll_line = Min(grid_scroll_line, MaxGridScroll());
               }else{
                              //find this item in real heap|at man
                  for(int j=mstate->item_list.size(); j--; ) if(mstate->item_list[j]==itm) break;
                  if(j!=-1) --mstate->item_list[j].pieces;
                  else{       //scan all soldiers in team4
                     int t4_i, itm_i;
                     bool b;
                     b = FindItem(itm, t4_i, itm_i);
                     assert(b);
                     --mstate->team_4[t4_i].item_list[itm_i].pieces;
                  }
                  --itm.pieces;
               }
            }
            ShowGridItems();
            ShowWeigth();
            if(snd) PlaySound(GM_SND_REMOVE);
         }
         break;

      case CB_SCROLL:
         {
            bool up = prm1, snd = prm2;

            int i;
            if(!up) i=FindUpper(curr_avl_scroll);
            else i=FindLower(curr_avl_scroll);
            if(i!=-1){
               if(selected_image && (selected_image->flags&RFLG_ITEM))
                  Unselect();
               curr_avl_scroll=i;
               ShowScrollItems();
               if(snd) PlaySound(GM_SND_SCROLL);
            }
         }
         break;

      case CB_SCROLL_GRID:
         {
            bool up = prm1;
            if(up){
               if(!grid_scroll_line) return false;
               --grid_scroll_line;
            }else{
               if(grid_scroll_line==MaxGridScroll()) return false;
               ++grid_scroll_line;
            }
            ShowGridItems();
         }
         break;

      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            PI3D_frame frm(res.frm);
            dword flg(res.flags);
            if(flg&RFLG_HIGHLIGHT)
               res.frm->SetOn(false);
            if(flg&RFLG_BUTTON){
               switch(res.id){
               case ID_BUTTON_ITEM_ADD:
               case ID_BUTTON_ITEM_REMOVE:
               case ID_BUTTON_AMMO_ADD:
               case ID_BUTTON_AMMO_REMOVE:
                  frm->SetOn(false);
               }
            }
         }
         break;

      case CB_POINTED_RES_CHANGE:
         {
            S_resource *res = (S_resource*)prm1;
            S_resource *old = (S_resource*)prm2;

            if(old && old!=selected_image){
               if(old->flags&(RFLG_BUTTON|RFLG_MAN_8|RFLG_ITEM|RFLG_ITEM_GRID))
                  SetHLight(old, false);
               if(old->flags&(RFLG_ITEM|RFLG_ITEM_GRID)) HideInfo();
            }
            if(res){
               if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
                  DisplayInfo(res);
                  if(res!=selected_image) PlaySound(GM_SND_HLIGHT);
               }
               //if(ok)         //don't hilight empty slot
               if(res->flags&(RFLG_BUTTON|RFLG_MAN_8|RFLG_ITEM|RFLG_ITEM_GRID))
                  SetHLight(res, true);
            }

            pointed_image = res;

            if(!pointed_image && selected_image)
               DisplayInfo(selected_image);
         }
         break;

      case CB_CLICK:
         {
            S_resource *res = (S_resource*)prm1;
            if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
               if(prm2)
                  cbProc(CB_SELECT_IMAGE, (dword)res);
            }
         }
         break;
      }
      return C_menu_base1::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void Proceed(){

      RenderAllBackbuffers();
      const char *mname = tab_campaigns->ItemS(TAB_S12_CM_MISS_DIR, game_profile.MissionIndex(), 12);
      tick_class = CreateBriefing(mname);
      if(tick_class){
         game_music.FadeVolume(.3f);
      }else{
         tick_class = CreateMenu(GAME_MENU_4_OF_8);
      }
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void ReturnBack(){

      RenderAllBackbuffers();
      tick_class = CreateMenu(GAME_MENU_8_OF_40);
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){
      case ID_BUTTON_DOWN:
      case ID_BUTTON_UP:
         {
            int i = -1;
            if(selected_image && (selected_image->flags&RFLG_ITEM))
               i = selected_image->id;
            cbProc(CB_SCROLL, (res->id==ID_BUTTON_UP));
            if(i!=-1){
               cbProc(CB_SELECT_IMAGE, (dword)FindResource(i, RFLG_ITEM), false);
               DisplayInfo(selected_image);
               SetHLight(selected_image, true);
            }
         }
         break;

      case ID_BUTTON_DOWN1:
      case ID_BUTTON_UP1:
         {
            Unselect();
            int i = cbProc(CB_SCROLL_GRID, res->id==ID_BUTTON_UP1);
            if(!i) PlaySound(GM_SND_CANNOT);
         }
         break;

      case ID_BUTTON_ITEM_ADD:
         cbProc(CB_ADD_SELECTED1, 0x10000);
         break;

      case ID_BUTTON_AMMO_ADD:
         cbProc(CB_ADD_SELECTED1, 0x10001);
         break;

      case ID_BUTTON_ITEM_REMOVE:
         cbProc(CB_REMOVE_SELECTED1, 0x10000);
         break;

      case ID_BUTTON_AMMO_REMOVE:
         cbProc(CB_REMOVE_SELECTED1, 0x10001);
         break;

      case ID_BUTTON_AUTO_SELECT:
                        //setup auto-selection mode
         mode=MODE_AUTO_SELECT;
         break;
      }
      C_menu_base1::OnButtonPressed(res);
   }

//----------------------------

   virtual void OnDoubleClick(S_resource *res){

      if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
         if(selected_image==res){
            if(res->flags&RFLG_ITEM_GRID){
               cbProc(CB_REMOVE_SELECTED1, 0x10000);
            }else{
               cbProc(CB_ADD_SELECTED1, 0x10000);
            }
         }
      }
   }

//----------------------------

   C_string GetHintMessage(S_resource *res, dword &ret_color, bool &wide){

      if(res->flags&RFLG_MAN_8){
         int man_id = team_8[res->id];
         if(man_id!=-1){
            int tid = GT_GAME_MENU_SOLDIER_NAME+man_id;
            wide = all_txt.IsWide(tid);
            return all_txt[tid];
         }
      }
      if(res->flags&(RFLG_ITEM|RFLG_ITEM_GRID)){
         int id = GetItemID(res);
         if(id!=-1){
            int tid;
            if(res->flags&RFLG_ITEM)
               tid = avl_items[id];
            else
               tid = item_list[id].itm;
            if(game_profile.IsDatadisk())
               tid += 1e6;
            tid += GT_GAME_MENU_ITEM_NAME;
            wide = all_txt.IsWide(tid);
            return all_txt[tid];
         }
      }
      return C_menu_base1::GetHintMessage(res, ret_color, wide);
   }
};

//----------------------------
//----------------------------
                              //messages for menu5 and menu6
enum{
   CB_GET_IMAGE_NAME = 1000,  //(C_string *ret, int indx)
};

class C_menu_select: public C_menu_base1{

   int pointed_image_id;

   void DisplayInfo(const S_resource *curr, const S_resource *old){

      const S_resource *res[2] = {old, curr};
                              //switch highlights
      for(int i=0; i<2; i++)
      if(res[i] && (res[i]->flags&RFLG_SELECTION_IMAGE)){
         const S_resource *hgl = FindResource(res[i]->id, RFLG_HIGHLIGHT);
         assert(hgl);
         hgl->frm->SetOn(i);
         const S_resource *main = FindResource(0, RFLG_MAIN_PIC);
         if(i){
            if(main){
               C_string str;
               cbProc(CB_GET_IMAGE_NAME, (dword)&str, res[i]->id);

               bool is_grayed = (res[i]->id>=max_possible_slct);
               ChangeTexture(main->frm, !is_grayed ? str : C_fstring("%s_bw", (const char*)str));
               main->frm->SetOn(true);
            }
                              //set info text
            if(curr->id<max_possible_slct)
               ShowText(text_base+res[i]->id);
         }else{
            if(!res[1] || !(res[1]->flags&RFLG_SELECTION_IMAGE)){
               main->frm->SetOn(false);
            }
            menu_text = NULL;
         }
      }
   }

//----------------------------
   C_smart_ptr<C_text> menu_text;

   void ShowText(int id){

      float l, t, sx, sy;
      GetTextExtents(l, t, sx, sy);
      float ext[4] = {l, t, l+sx, t+sy};
      float size = .0225f;

      menu_text = texts.CreateText(all_txt[id], 0xffffffff, l + sx*.5f, t, size,
         TEXT_JUSTIFY_CONVERT_EOL | TEXT_JUSTIFY_CENTER,
         all_txt.IsWide(id),
         ext, NULL, NULL, NULL,
         0, 0.0f);
      menu_text->Release();
   }

protected:
   int selected_image_id;
   int max_possible_slct;
   int text_base;

   C_menu_select():
      text_base(0),
      selected_image_id(-1),
      pointed_image_id(-1)
   {
      max_possible_slct = game_profile.max_camp + 1;
   }

//----------------------------
public:
   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_SELECT_IMAGE:
         {
            const S_resource *res = (const S_resource*)prm1;
            if(selected_image_id==-1){
                        //unhide OK and other button
               FindResource(ID_BUTTON_OK, RFLG_BUTTON)->frm->SetOn(true);
               if(GetType()==GAME_MENU_CAMPAIGN_SELECT){
                  //dword b1 = GetType()==GAME_MENU_CAMPAIGN_SELECT ? ID_BUTTON_CAMP_ANIM : ID_BUTTON_MISSION_STATUS;
                  FindResource(ID_BUTTON_CAMP_ANIM, RFLG_BUTTON)->frm->SetOn(true);
               }
            }
            const S_resource *old = selected_image_id!=-1 ? FindResource(selected_image_id, RFLG_SELECTION_IMAGE) :
               pointed_image_id!=-1 ? FindResource(pointed_image_id, RFLG_SELECTION_IMAGE) : NULL;
            DisplayInfo(res, old);
            selected_image_id = res->id;
            pointed_image_id = -1;
            if(prm2) PlaySound(GM_SND_SELECT);
         }
         break;

      case CB_GET_IMAGE_NAME:
         {
            C_string *sp = (C_string*)prm1;
            *sp = tab_campaigns->ItemS(TAB_S12_CM_IMAGE_NAME, prm2, 12);
         }
         return 0;

      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            PI3D_frame frm(res.frm);
            dword flg(res.flags);
            if(frm->IsOn())
            if(flg&RFLG_SELECTION_IMAGE){
               C_string str;
               cbProc(CB_GET_IMAGE_NAME, (dword)&str, res.id);
               int max_sel;
               if(GetType()==GAME_MENU_CAMPAIGN_SELECT)
                  max_sel = NUM_CAMPAIGNS;
               else
               if(!game_profile.IsAddOn())
                  max_sel = game_profile.NumMissionsInCamp();
               else{
                  max_sel = game_profile.num_addon_missions;
               }
               if(res.id >= max_sel){
                  frm->SetOn(false);
               }else{
                  bool is_grayed = (res.id>=max_possible_slct);
                  ChangeTexture(frm, !is_grayed ? str : C_fstring("%s_bw", (const char*)str));
               }
            }
            if(flg&RFLG_MAIN_PIC)
               frm->SetOn(false);
            if(flg&RFLG_HIGHLIGHT)
               res.frm->SetOn(false);
            if(flg&RFLG_BUTTON){
               switch(res.id){
               case ID_BUTTON_OK:
               case ID_BUTTON_CAMP_ANIM:
               case ID_BUTTON_MISSION_STATUS:
               case ID_BUTTON_MISSION_LOAD:
                  frm->SetOn(false);
               }
            }
         }
         break;

      case CB_POINTED_RES_CHANGE:
         {
            S_resource *res = (S_resource*)prm1;
            if(selected_image_id==-1){
               DisplayInfo(res, (S_resource*)prm2);
               if(res){
                  if(res->flags&RFLG_SELECTION_IMAGE)
                     PlaySound(GM_SND_HLIGHT);
                  pointed_image_id = res->id;
               }else pointed_image_id = -1;
            }
         }
         break;

      case CB_CLICK:
         {
            S_resource *res = (S_resource*)prm1;
            if(res->flags&RFLG_SELECTION_IMAGE)
            if(prm2){
                                 //check if selection possible
               if(res->id<max_possible_slct){
                  if(selected_image_id!=res->id){
                     cbProc(CB_SELECT_IMAGE, (dword)res, true);
                  }
               }else{
                  PlaySound(GM_SND_CANNOT);
               }
            }
         }
         break;
      }
      return C_menu_base1::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void GetTextExtents(float &l, float &t, float &sx, float &sy){

      l = .2f;
      t = .5025f;
      sx = .76f;
      sy = .0f;
   }
};

//----------------------------
//----------------------------
                              //concrete menu - select campaign
class C_menu_campaign_select: public C_menu_select{

   C_string GetHintMessage(S_resource *res, dword &ret_color, bool &wide){

      if(res->flags&RFLG_SELECTION_IMAGE){
         if(res->id<max_possible_slct){
            int tid = GT_GAME_MENU_CAMP_NAME + res->id;
            wide = all_txt.IsWide(tid);
            return all_txt[tid];
         }
      }
      return C_menu_base1::GetHintMessage(res, ret_color, wide);
   }

public:
   virtual bool Init(E_GAME_MENU id, dword context){

      //if(!C_menu_select::Init(id, context & ~1)) return false;
      if(!C_menu_base1::Init(id))
         return false;

      static const S_resource_def r_list[] = {
         {"PIC1", 0, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC2", 1, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC3", 2, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC4", 3, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC5", 4, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC6", 5, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC7", 6, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC8", 7, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC9", 8, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"slct_00", 0, RFLG_HIGHLIGHT},
         {"slct_01", 1, RFLG_HIGHLIGHT},
         {"slct_02", 2, RFLG_HIGHLIGHT},
         {"slct_03", 3, RFLG_HIGHLIGHT},
         {"slct_04", 4, RFLG_HIGHLIGHT},
         {"slct_05", 5, RFLG_HIGHLIGHT},
         {"slct_06", 6, RFLG_HIGHLIGHT},
         {"slct_07", 7, RFLG_HIGHLIGHT},
         {"slct_08", 8, RFLG_HIGHLIGHT},
         {"OK", ID_BUTTON_OK, RFLG_BUTTON | RFLG_HINT},
         {"BACK", ID_BUTTON_BACK, RFLG_BUTTON | RFLG_HINT},
         {"slct_ok", ID_BUTTON_OK+100, RFLG_HIGHLIGHT},
         {"slct_back", ID_BUTTON_BACK+100, RFLG_HIGHLIGHT},
         {"PIC", 0, RFLG_MAIN_PIC},

         {"BUT00", ID_BUTTON_CAMP_ANIM, RFLG_BUTTON | RFLG_HINT},
         {"slct_butt00", ID_BUTTON_CAMP_ANIM+100, RFLG_HIGHLIGHT},
         {NULL}
      };
      InitResources(r_list);
      cbProc(CB_ENUM_RESOURCES, 0, 0);
      text_base = GT_GAME_MENU_CAMP_TEXT;

      if(context&1)
         cbProc(CB_SELECT_IMAGE, (dword)FindResource(game_profile.curr_camp, RFLG_SELECTION_IMAGE));

      return true;
   }

//----------------------------

   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){
      case CB_SELECT_IMAGE:
         {
            dword ret = C_menu_select::cbProc(msg, prm1, prm2);
            game_profile.curr_camp = selected_image_id;
            return ret;
         }
         break;
      }
      return C_menu_select::cbProc(msg, prm1, prm2);
   }

//----------------------------

   void RunCampAnim(int camp_indx){

      RenderAllBackbuffers();
      if(net) NetSend();
      game_music.Stop();
      tick_class = CreateCampAnim(C_fstring("campaign_%i", camp_indx),
         C_smart_ptr<C_class_to_be_ticked_hd>(this));
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   void Proceed(){

      RenderAllBackbuffers();
      tick_class = CreateMenu(GAME_MENU_MISSION_SELECT);
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void ReturnBack(){

      NetCloseAll();
      game_configuration.net_host = false;
      game_configuration.net_join = false;
      RenderAllBackbuffers();
      tick_class = CreateMainMenu();
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void OnNetSyncEvent(E_NET_SYNC_EVENT event, dword data){

      switch(event){
      case NS_LOAD_CAMPANIM:
         RunCampAnim(data);
         break;
      default:
         C_menu_select::OnNetSyncEvent(event, data);
      }
   }

//----------------------------

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){
      case ID_BUTTON_CAMP_ANIM:
         if(net){
            BeginSyncEvent(NS_LOAD_CAMPANIM, selected_image_id);
         }else{
            RunCampAnim(selected_image_id);
         }
         break;
      }
      C_menu_base1::OnButtonPressed(res);
   }

//----------------------------

   virtual void OnDoubleClick(S_resource *res){

      if((res->flags&RFLG_SELECTION_IMAGE) && (selected_image_id==res->id)){
         if(net)
            BeginSyncEvent(NS_PROCEED);
         else
            Proceed();
      }
   }
};

//----------------------------
//----------------------------
                              //select mission
class C_menu_mission_select: public C_menu_select{
   int curr_camp;

   C_string GetImageName1(int i) const{
      return C_fstring("%s%.1i", tab_campaigns->ItemS(TAB_S12_CM_MISS_IMG_BASE, curr_camp, 12), i+1);
   }

//----------------------------

public:
   C_menu_mission_select()
   {
      curr_camp = game_profile.curr_camp;
      if(!game_profile.IsAddOn()){
         max_possible_slct = (curr_camp==game_profile.max_camp) ?
            game_profile.max_mission + 1 :
            game_profile.NumMissionsInCamp(curr_camp);
      }else{
         max_possible_slct = game_profile.max_addon_mission+1;
      }
   }

//----------------------------

   virtual bool Init(E_GAME_MENU id, dword context){

      if(!C_menu_select::Init(id, context & ~1))
         return false;
      static const S_resource_def r_list[] = {
         {"PIC1", 0, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC2", 1, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC3", 2, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC4", 3, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC5", 4, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"PIC6", 5, RFLG_SELECTION_IMAGE | RFLG_HINT},
         {"slct_00", 0, RFLG_HIGHLIGHT},
         {"slct_01", 1, RFLG_HIGHLIGHT},
         {"slct_02", 2, RFLG_HIGHLIGHT},
         {"slct_03", 3, RFLG_HIGHLIGHT},
         {"slct_04", 4, RFLG_HIGHLIGHT},
         {"slct_05", 5, RFLG_HIGHLIGHT},
         {"OK", ID_BUTTON_OK, RFLG_BUTTON | RFLG_HINT},
         {"BACK", ID_BUTTON_BACK, RFLG_BUTTON | RFLG_HINT},
         {"slct_ok", ID_BUTTON_OK+100, RFLG_HIGHLIGHT},
         {"slct_back", ID_BUTTON_BACK+100, RFLG_HIGHLIGHT},
         {"PIC", 0, RFLG_MAIN_PIC},

         {"BUT00", ID_BUTTON_MISSION_STATUS, RFLG_BUTTON | RFLG_HINT},
         {"slct_butt00", ID_BUTTON_MISSION_STATUS+100, RFLG_HIGHLIGHT},
         {"BUT02", ID_BUTTON_MISSION_LOAD, RFLG_BUTTON | RFLG_HINT},
         {"slct_load", ID_BUTTON_MISSION_LOAD+100, RFLG_HIGHLIGHT},
         {NULL}
      };
      InitResources(r_list);
      cbProc(CB_ENUM_RESOURCES, 0, 0);

      const S_resource *cpht = FindResource(0, RFLG_CAMPAIGN_PHOTO);
      if(cpht){
         C_string str;
         C_menu_select::cbProc(CB_GET_IMAGE_NAME, (dword)&str, curr_camp);
         ChangeTexture(cpht->frm, str);
      }
      text_base = GT_GAME_MENU_MISS_TEXT + game_profile.MissionIndex(-1, 0);
      if(game_profile.IsDatadisk())
         text_base += 1e6 - game_profile.MissionIndex(6, 0);
      else
      if(game_profile.IsAddOn())
         text_base = 100010;

      if(context&1)
         cbProc(CB_SELECT_IMAGE, (dword)FindResource(game_profile.curr_mission, RFLG_SELECTION_IMAGE));

      return true;
   }

//----------------------------

   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_GET_IMAGE_NAME:
         {
            C_string *sp = (C_string*)prm1;
            *sp = GetImageName1(prm2);
         }
         return 0;

      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            PI3D_frame frm(res.frm);
            dword flg(res.flags);
            if(flg&RFLG_SELECTION_IMAGE){
               if(res.id>=game_profile.NumMissionsInCamp(curr_camp)){
                  frm->SetOn(false);
               }
            }
         }
         break;

      case CB_SELECT_IMAGE:
         {
            dword ret = C_menu_select::cbProc(msg, prm1, prm2);
            game_profile.curr_mission = selected_image_id;
            if(ret){
                              //check if there's any saved game available
               S_mission_state *mstate = &game_profile.mission_states[game_profile.MissionIndex()];
               bool gs = (!net && mstate->save_game.IsOpen());
               FindResource(ID_BUTTON_MISSION_LOAD, RFLG_BUTTON)->frm->SetOn(gs);
            }
            return ret;
         }
         break;

      case CB_CLICK:
         {
            S_resource *res = (S_resource*)prm1;
            if(res->flags&RFLG_SELECTION_IMAGE)
            if(prm2){
                                 //check if selection possible
               if(res->id<max_possible_slct){
                  if(selected_image_id!=res->id){
                     cbProc(CB_SELECT_IMAGE, (dword)res, true);
                  }
               }else{
                  PlaySound(GM_SND_CANNOT);
               }
            }
         }
         break;
      }
      return C_menu_select::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void Proceed(){

      RenderAllBackbuffers();
      if(!selected_image_id){
                              //1st mission in campaign - go to 8-of-40 selection
         tick_class = CreateMenu(GAME_MENU_8_OF_40);
      }else{
         const char *mname = tab_campaigns->ItemS(TAB_S12_CM_MISS_DIR, game_profile.MissionIndex(), 12);
         tick_class = CreateBriefing(mname);
         if(tick_class)
            game_music.FadeVolume(.3f);
         else{
            tick_class = CreateMenu(GAME_MENU_4_OF_8);
         }
      }
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void ReturnBack(){

      RenderAllBackbuffers();
      if(!game_profile.IsAddOn()){
         tick_class = CreateMenu(GAME_MENU_CAMPAIGN_SELECT);
      }else{
                              //close Add-on
         NetCloseAll();
         tick_class = CreateMainMenu();
      }
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){

      case ID_BUTTON_MISSION_LOAD:
         {
            if(net) NetSend();
            C_game_mission *mission = CreateGameMission();
            if(mission){
               RenderAllBackbuffers();
               if(mission->LoadGame()){
                  mission->DisplayObjectives();
                  game_music.Stop();
                  tick_class = mission;
               }
               mission->Release();
            }
         }
         break;
      }
      C_menu_select::OnButtonPressed(res);
   }

//----------------------------

   virtual void OnDoubleClick(S_resource *res){

      if((res->flags&RFLG_SELECTION_IMAGE) && (selected_image_id==res->id)){
         if(net)
            BeginSyncEvent(NS_PROCEED);
         else
            Proceed();
      }
   }

//----------------------------

   C_string GetHintMessage(S_resource *res, dword &ret_color, bool &wide){

      if(res->flags&RFLG_CAMPAIGN_PHOTO){
                              //get hint for shown campaign
         int tid = GT_GAME_MENU_CAMP_NAME+curr_camp;
         wide = all_txt.IsWide(tid);
         return all_txt[tid];
      }
      if(res->flags&RFLG_SELECTION_IMAGE){
         if(res->id<max_possible_slct){

            int tid = game_profile.MissionIndex(-1, res->id);
            if(!game_profile.IsAddOn()){
               if(game_profile.IsDatadisk())
                  tid += 1e6 - game_profile.MissionIndex(6, 0);
               tid += GT_MISSION_NAMES;
            }else
               tid += 100000 - NUM_MISSIONS;
            wide = all_txt.IsWide(tid);
            return all_txt[tid];
         }
      }
      return C_menu_select::GetHintMessage(res, ret_color, wide);
   }
};

//----------------------------
//----------------------------
/*
                              //concrete menu - mission info
class C_menu7: public C_menu_base1{

   S_mission_state *mstate;
                              //on-menu item_list
   vector<S_item> item_list;

                              //on-menu team_8
   short team_8[8];

   const S_resource *pointed_image;
   const S_resource *selected_image;

#undef ITM_GRID_X
#define ITM_GRID_X 4
#define ITM_GRID_Y 5
   PI3D_frame itm_grid[3][ITM_GRID_Y][ITM_GRID_X];   //[dummy/hlight/item][][]
   C_text *ammo_txt[ITM_GRID_Y][ITM_GRID_X];
   int grid_scroll_line;

   void ShowGridItems(bool hide_all = false){

      int i, x, y;
                              //hide previous
      for(y=0; y<ITM_GRID_Y; y++)
      for(x=0; x<ITM_GRID_X; x++){
         if(itm_grid[2][y][x]){
            itm_grid[2][y][x]->LinkTo(NULL);
            itm_grid[2][y][x]->Release();
            itm_grid[2][y][x] = NULL;
         }
         if(ammo_txt[y][x]){
            texts.DeleteText(ammo_txt[y][x]);
            ammo_txt[y][x]=NULL;
         }
      }
      if(hide_all) return;
                              //show new
      int num = Min(ITM_GRID_X*ITM_GRID_Y, (int)item_list.size()-grid_scroll_line*ITM_GRID_X);
      for(i=0; i<num; i++){
         x = i%ITM_GRID_X;
         y = i/ITM_GRID_X;

         S_item &itm = item_list[i+grid_scroll_line*ITM_GRID_X];
         PI3D_model mod = driver->CreateModel();
         itm_grid[2][y][x] = mod;

         const char *mname = tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16);
         model_cache.Open(mod, mname, driver, 0, ErrReport, editor);

#ifdef _DEBUG
         mod->SetName(tab_inventory->ItemS(TAB_S16_INV_MODEL_I, itm.itm, 16));
#endif
         mod->SetAnimation(0, NULL);
         mod->LinkTo(itm_grid[0][y][x]);

         S_vector in(itm_grid[0][y][x]->GetWorldPos()), out;
         S_vectorw v4;
         scene->TransformPoints(&in, &v4, 1);
         out = v4.xyz;
         float x1(out.x/igraph->Scrn_sy()), y1(out.y/igraph->Scrn_sy());
         //x1 -= .1f;
         y1 += .03f;
         C_fstring str("%i", itm.pieces);
         if(HasAmmo(itm.itm)) str += C_fstring("/%i", itm.ammo);
         y1 /= (screen_aspect_ratio * 1.33333333f);
         ammo_txt[y][x] = texts_old.CreateText(str, x1, y1, .03f, 1);
      }
   }

   void ShowTeam8(int indx){
      const S_resource *res = FindResource(indx, RFLG_MAN_8);
      int mi = team_8[indx];
      ChangeTexture(res->frm, mi==-1 ? "face_no" : C_fstring("man%.2i", mi+1), "faces");
   }

   int GetItemID(const S_resource *res) const{
      if(res)
         return res->id - 1000 + grid_scroll_line*ITM_GRID_X;
      return -1;
   }

   int GetItemIDID(const S_resource *res) const{
      int id = GetItemID(res);
      return item_list[id].itm;
   }

   C_string GetImageName1() const{
      return C_fstring("%s%.1i", tab_campaigns->ItemS(TAB_S12_CM_MISS_IMG_BASE, game_profile.curr_camp, 12), game_profile.curr_mission+1);
   }
protected:

public:
   C_menu7():
      selected_image(NULL),
      pointed_image(NULL),
      //desc(NULL),
      grid_scroll_line(0)
   {
      mstate = &game_profile.mission_states[game_profile.MissionIndex()];
      memset(itm_grid, 0, sizeof(itm_grid));
      memset(ammo_txt, 0, sizeof(ammo_txt));
      int i, j;

                              //collect all items from men
      for(i=0; i<4; i++){
         if(mstate->team_4[i]==-1) continue;
         for(int k=0; k<mstate->team_4[i].item_list.size(); k++){
            //item_list.push_back(mstate->team_4[i].item_list[j]);
            for(j=item_list.size(); j--; ) if(item_list[j].itm==mstate->team_4[i].item_list[k].itm) break;
            if(j==-1) item_list.push_back(mstate->team_4[i].item_list[k]);
            else{
               item_list[j].ammo += mstate->team_4[i].item_list[k].ammo;
               item_list[j].pieces += mstate->team_4[i].item_list[k].pieces;
            }
         }
      }
      for(i=0; i<mstate->item_list.size(); i++){
         for(j=item_list.size(); j--; ) if(item_list[j].itm==mstate->item_list[i].itm) break;
         if(j==-1) item_list.push_back(mstate->item_list[i]);
         else{
            item_list[j].ammo += mstate->item_list[i].ammo;
            item_list[j].pieces += mstate->item_list[i].pieces;
         }
      }
   }
   ~C_menu7(){
      ShowGridItems(true);

      for(int y=0; y<ITM_GRID_Y; y++)
      for(int x=0; x<ITM_GRID_X; x++)
      for(int i=0; i<2; i++)
      if(itm_grid[i][y][x]) itm_grid[i][y][x]->Release();
   }
                              //context = flags (1 = image should be primarily selected)
   virtual bool Init(E_GAME_MENU id, dword context){

      if(!C_menu_base1::Init(id))
         return false;

      {
         PI3D_frame fm = scene->FindFrame("item05");

         const S_vector &mn = fm->GetPos(),
            &mx = scene->FindFrame("item06")->GetPos();
         const float scl = fm->GetScale();
         float dx = (mx.x-mn.x) / (ITM_GRID_X-1);
         float dy = (mx.y-mn.y) / (ITM_GRID_Y-1);
                              //create grid of dummys and highlights
         for(int y=0; y<ITM_GRID_Y; y++)
         for(int x=0; x<ITM_GRID_X; x++){
            for(int i=0; i<2; i++){
               PI3D_model mod = driver->CreateModel();
               if(i){
                  model_cache.Open(mod, "inv_frm", driver, 0, ErrReport, editor);
#ifdef _DEBUG
                  mod->SetName("inv_frm");
#endif
               }
               PI3D_frame frm = mod;
               itm_grid[i][y][x] = frm;
               frm->SetPos(S_vector(mn.x + dx*x, mn.y + dy*y, !i ? mn.z : mn.z+.5f));
               frm->SetScale(scl);
               frm->Update();
               frm->LinkTo(scene->GetPrimarySector());
               res_list.push_back(S_resource());
               S_resource &res = res_list.back();
               res.frm = frm;
               res.id = y*ITM_GRID_X + x;
               //if(i)
                  res.id += 1000;
               res.flags = i ? RFLG_HIGHLIGHT : (RFLG_ITEM_GRID | RFLG_HINT);
            }
         }
      }

      static const S_resource_def r_list[] = {
         //{"OK", ID_BUTTON_OK, RFLG_BUTTON | RFLG_HINT},
         {"BACK", ID_BUTTON_BACK, RFLG_BUTTON | RFLG_HINT},
         {"INFO", ID_BUTTON_SOLDIER_INFO, RFLG_BUTTON | RFLG_HINT},
         {"UP01", ID_BUTTON_UP, RFLG_BUTTON | RFLG_INVISIBLE | RFLG_ARROW_UP},
         {"DOWN01", ID_BUTTON_DOWN, RFLG_BUTTON | RFLG_INVISIBLE | RFLG_ARROW_DN},

         {"slct_ok", ID_BUTTON_BACK+100, RFLG_HIGHLIGHT},
         {"slct_info", ID_BUTTON_SOLDIER_INFO+100, RFLG_HIGHLIGHT},
         //{"slct_up01", ID_BUTTON_UP+100, RFLG_HIGHLIGHT},
         //{"slct_down01", ID_BUTTON_DOWN+100, RFLG_HIGHLIGHT},
                              //8 men
         {"pic1", 0, RFLG_MAN_8 | RFLG_HINT},
         {"pic2", 1, RFLG_MAN_8 | RFLG_HINT},
         {"pic3", 2, RFLG_MAN_8 | RFLG_HINT},
         {"pic4", 3, RFLG_MAN_8 | RFLG_HINT},
         {"pic5", 4, RFLG_MAN_8 | RFLG_HINT},
         {"pic6", 5, RFLG_MAN_8 | RFLG_HINT},
         {"pic7", 6, RFLG_MAN_8 | RFLG_HINT},
         {"pic8", 7, RFLG_MAN_8 | RFLG_HINT},
         {"slct_00", 0, RFLG_HIGHLIGHT},
         {"slct_01", 1, RFLG_HIGHLIGHT},
         {"slct_02", 2, RFLG_HIGHLIGHT},
         {"slct_03", 3, RFLG_HIGHLIGHT},
         {"slct_04", 4, RFLG_HIGHLIGHT},
         {"slct_05", 5, RFLG_HIGHLIGHT},
         {"slct_06", 6, RFLG_HIGHLIGHT},
         {"slct_07", 7, RFLG_HIGHLIGHT},

         //{"mission.photo", 0, RFLG_CAMPAIGN_PHOTO | RFLG_HINT},

         {NULL}
      };
      InitResources(r_list);

      cbProc(CB_ENUM_RESOURCES, 0, 0);

      const S_resource *mpht = FindResource(0, RFLG_CAMPAIGN_PHOTO);
      if(mpht){
         C_string str;
         ChangeTexture(mpht->frm, GetImageName1());
      }

      int i, j;
      for(i=0; i<8; i++) team_8[i] = -1;
      for(i=j=0; i<4; i++) if(mstate->team_4[i]!=-1) team_8[j++] = (short)mstate->team_4[i];
      for(i=0; i<8; i++) if(mstate->team_8[i]!=-1) team_8[j++] = mstate->team_8[i];
      for(i=0; i<8; i++) ShowTeam8(i);

      ShowGridItems();
      return true;
   }

   int MaxGridScroll() const{
      return Max(0, (int)(item_list.size()-1)/ITM_GRID_X);
   }

public:
   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){
      int i;

      switch(msg){
      case CB_SCROLL_GRID:
         {
            bool up = prm1;
            if(up){
               if(!grid_scroll_line) return false;
               --grid_scroll_line;
            }else{
               if(grid_scroll_line==MaxGridScroll()) return false;
               ++grid_scroll_line;
            }
            ShowGridItems();
         }
         break;

      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            dword flg(res.flags);
            if(flg&RFLG_HIGHLIGHT)
               res.frm->SetOn(false);
         }
         break;

      case CB_POINTED_RES_CHANGE:
         {
            S_resource *res = (S_resource*)prm1;
            S_resource *old = (S_resource*)prm2;

            if(old && old!=selected_image){
               if(old->flags&(RFLG_BUTTON|RFLG_MAN_8|RFLG_ITEM_GRID))
                  SetHLight(old, false);
               //if(old->flags&(RFLG_ITEM|RFLG_ITEM_GRID)) HideInfo();
            }
            if(res){
               //if(ok)         //don't hilight empty slot
               if(res->flags&(RFLG_BUTTON|RFLG_MAN_8|RFLG_ITEM_GRID))
                  SetHLight(res, true);
            }
            pointed_image = res;
            //if(!pointed_image && selected_image) DisplayInfo(selected_image);
         }
         break;

      case CB_CLICK:
         {
            S_resource *res = (S_resource*)prm1;
            if(res->flags&RFLG_MAN_8){
               if(prm2 && res==pointed_image){
                  i = GetManID(res);
                  if(i!=-1){
                     SetMenuState(GMS_SOLDIER_INFO);
                  }else
                     PlaySound(GM_SND_CANNOT);
               }
            }
         }
         break;
      }
      return C_menu_base1::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){
      case ID_BUTTON_DOWN:
      case ID_BUTTON_UP:
         cbProc(CB_SCROLL_GRID, (res->id==ID_BUTTON_UP));
         //if(!i)       //uninitialized!?!
            PlaySound(GM_SND_CANNOT);
         break;
      case ID_BUTTON_SOLDIER_INFO:
         SetMenuState(GMS_SOLDIER_INFO);
         break;
      }
      C_menu_base1::OnButtonPressed(res);
   }

//----------------------------

   C_string GetHintMessage(S_resource *res, dword &ret_color){

      if(res->flags&RFLG_MAN_8){
         int man_id = team_8[res->id];
         if(man_id!=-1)
            return all_txt[GT_GAME_MENU_SOLDIER_NAME+man_id];
      }
      if(res->flags&RFLG_ITEM_GRID){
         int id = GetItemID(res);
         if(id!=-1){
            id = item_list[id].itm;
            return all_txt[GT_GAME_MENU_ITEM_NAME+id];
         }
      }
      if(res->flags&RFLG_CAMPAIGN_PHOTO){
                        //get hint for shown mission
         return all_txt[GT_MISSION_NAMES+game_profile.MissionIndex()];
      }
      return C_menu_base1::GetHintMessage(res, ret_color);
   }

//----------------------------

   int GetManID(const S_resource *res) const{
      if(!res) return -1;
      return team_8[res->id];
   }
};
*/

//----------------------------
//----------------------------

                              //concrete menu - debriefing
class C_menu_debriefing: public C_menu_base1{
   S_mission_state *mstate;
   short team_8[8];
   const S_resource *pointed_image;
   C_smart_ptr<C_text> desc[3];          //[title|description|info]

   S_menu_debriefing_info::E_status status;

public:
   C_menu_debriefing():
      pointed_image(NULL),
      mstate(NULL)
   {
      mstate = &game_profile.mission_states[game_profile.MissionIndex()];
   }

protected:

//----------------------------
                              //context: LOWORD = man ID, HIWORD = calling menu ID
   virtual bool Init(E_GAME_MENU id, dword context){

      if(!C_menu_base1::Init(id))
         return false;
      static const S_resource_def r_list[] = {
         {"OK", ID_BUTTON_OK, RFLG_BUTTON | RFLG_HINT},
         {"slct_ok", ID_BUTTON_OK+100, RFLG_HIGHLIGHT},
         {"OK02", ID_BUTTON_MISSION_LOAD, RFLG_BUTTON | RFLG_HINT},
         {"slct_load", ID_BUTTON_MISSION_LOAD+100, RFLG_HIGHLIGHT},
                              //8 men
         {"pic1", 0, RFLG_MAN_8},
         {"pic2", 1, RFLG_MAN_8},
         {"pic3", 2, RFLG_MAN_8},
         {"pic4", 3, RFLG_MAN_8},
         {"pic5", 4, RFLG_MAN_8},
         {"pic6", 5, RFLG_MAN_8},
         {"pic7", 6, RFLG_MAN_8},
         {"pic8", 7, RFLG_MAN_8},
         {"slct_00", 0, RFLG_HIGHLIGHT},
         {"slct_01", 1, RFLG_HIGHLIGHT},
         {"slct_02", 2, RFLG_HIGHLIGHT},
         {"slct_03", 3, RFLG_HIGHLIGHT},
         {"slct_04", 4, RFLG_HIGHLIGHT},
         {"slct_05", 5, RFLG_HIGHLIGHT},
         {"slct_06", 6, RFLG_HIGHLIGHT},
         {"slct_07", 7, RFLG_HIGHLIGHT},
         {NULL}
      };
      const S_menu_debriefing_info *di = (S_menu_debriefing_info*)context;
      status = di->status;

      InitResources(r_list);
      cbProc(CB_ENUM_RESOURCES, 0, 0);

      int i, j;
      for(i=0; i<8; i++)
         team_8[i] = -1;
      for(i=j=0; i<4; i++)
         if(mstate->team_4[i]!=-1)
            team_8[j++] = (short)mstate->team_4[i];
      for(i=0; i<8; i++)
         if(mstate->team_8[i]!=-1)
            team_8[j++] = mstate->team_8[i];
      for(i=0; i<8; i++)
         ShowTeam8(i);

      DisplayInfo(di);

      const char *music_name = NULL;
      switch(status){
      case S_menu_debriefing_info::DEBRIEF_SUCCESS_MISSION:
         music_name = "success_mission";
         break;
      case S_menu_debriefing_info::DEBRIEF_SUCCESS_CAMPAIGN:
      case S_menu_debriefing_info::DEBRIEF_SUCCESS_GAME_DONE:
         music_name = "success_campaign";
         break;
      case S_menu_debriefing_info::DEBRIEF_FAIL:
         music_name = "fail_mission";
         break;
      case S_menu_debriefing_info::DEBRIEF_SUCCESS_ADDON_DONE:
         music_name = "addon_done";
         break;
      default:
         assert(0);
      }
      if(music_name)
         game_music.Play(music_name, true);

      return true;
   }

//----------------------------

   virtual dword cbProc(dword msg, dword prm1=0, dword prm2=0){

      switch(msg){

      case CB_ENUM_CALL:
         {
            const S_resource &res = *(S_resource*)prm1;
            dword flg(res.flags);
            if(flg&RFLG_HIGHLIGHT)
               res.frm->SetOn(false);
            if((flg&RFLG_BUTTON) && res.id==ID_BUTTON_MISSION_LOAD){
               if(status == S_menu_debriefing_info::DEBRIEF_FAIL){
                              //check if there's any saved game available
                  S_mission_state *mstate = &game_profile.mission_states[game_profile.MissionIndex()];
                  bool gs = (!net && mstate->save_game.IsOpen());
                  res.frm->SetOn(gs);
               }else
                  res.frm->SetOn(false);
            }
         }
         break;

      case CB_POINTED_RES_CHANGE:
         {
            S_resource *res = (S_resource*)prm1;
            S_resource *old = (S_resource*)prm2;

            if(old){
               if(old->flags&(RFLG_BUTTON|RFLG_MAN_8)) SetHLight(old, false);
            }
            if(res){
               if(res->flags&(RFLG_BUTTON|RFLG_MAN_8)) SetHLight(res, true);
            }
            pointed_image = res;
         }
         break;
      }
      return C_menu_base1::cbProc(msg, prm1, prm2);
   }

//----------------------------

   virtual void Proceed(){

      RenderAllBackbuffers();
      switch(status){
      case S_menu_debriefing_info::DEBRIEF_SUCCESS_MISSION:
         tick_class = CreateMenu(GAME_MENU_MISSION_SELECT, 1);
         break;

      case S_menu_debriefing_info::DEBRIEF_FAIL:
         tick_class = CreateMenu(GAME_MENU_4_OF_8);
         break;

      case S_menu_debriefing_info::DEBRIEF_SUCCESS_CAMPAIGN:
         tick_class = CreateMenu(GAME_MENU_CAMPAIGN_SELECT, 1);
         break;

      case S_menu_debriefing_info::DEBRIEF_SUCCESS_GAME_DONE:
         NetCloseAll();
         tick_class = CreateTheEnd();
         break;

      case S_menu_debriefing_info::DEBRIEF_SUCCESS_ADDON_DONE:
         NetCloseAll();
         tick_class = CreateMainMenu();
         break;

      default:
         assert(0);
      }
      if(tick_class)
         tick_class->Release();
   }

//----------------------------

   virtual void OnButtonPressed(S_resource *res){

      switch(res->id){
      case ID_BUTTON_MISSION_LOAD:
         {
            if(net) NetSend();
            C_game_mission *mission = CreateGameMission();
            if(mission){
               RenderAllBackbuffers();
               if(mission->LoadGame()){
                  mission->DisplayObjectives();
                  game_music.Stop();
                  tick_class = mission;
               }
               mission->Release();
            }
         }
         break;
      }
      C_menu_base1::OnButtonPressed(res);
   }

//----------------------------

   void ShowTeam8(int indx){

      const S_resource *res = FindResource(indx, RFLG_MAN_8);
      int mi = team_8[indx];
      ChangeTexture(res->frm, mi==-1 ? "face_no" : C_fstring("man%.2i", mi+1), "faces");
   }

//----------------------------

   void HideInfo1(){

      for(int i=0; i<3; i++)
         desc[i] = NULL;
   }

//----------------------------

   void DisplayInfo(const S_menu_debriefing_info *di){

      const float x = .52f;
      const float y = .225f;
                              //feed texts
      int txt_id = status;
      if(txt_id==S_menu_debriefing_info::DEBRIEF_SUCCESS_GAME_DONE)
         txt_id = S_menu_debriefing_info::DEBRIEF_SUCCESS_CAMPAIGN;
      desc[0] = texts.CreateText(all_txt[GT_DEBRIEF_TITLE+txt_id], -1, x, y, .0375f, TEXT_JUSTIFY_CENTER,
         all_txt.IsWide(GT_DEBRIEF_TITLE+txt_id));
      desc[0]->Release();

      bool wide = (game_configuration.language==5);
      C_string str[2];
      int i;
      for(i=0; i<6; i++){
         str[0] = AddGameText(str[0], all_txt[GT_DEBRIEF_STATS+i]);
         switch(i){
         case 0:              //ammo spent
            str[1] = AddGameText(str[1], MakeGameText(C_fstring("%i\n", di->ammo_spent)));
            break;
         case 1:              //ammo hit
            str[1] = AddGameText(str[1], MakeGameText(C_fstring("%i\n\n", di->num_hits)));
            str[0] = AddGameText(str[0], MakeGameText("\n"));
            break;
         case 2:              //enemies killed
            str[1] = AddGameText(str[1], MakeGameText(C_fstring("%i\n", di->enemy_killed)));
            break;
         case 3:              //civilists killed
            str[1] = AddGameText(str[1], MakeGameText(C_fstring("%i\n", di->civil_killed)));
            break;
         case 4:              //allies killed
            str[1] = AddGameText(str[1], MakeGameText(C_fstring("%i\n\n", di->allies_killed.size())));
            str[0] = AddGameText(str[0], MakeGameText("\n"));
            break;
         case 5:              //missed in action
            for(int j=0; j<di->allies_killed.size(); j++){
               str[1] = AddGameText(str[1], all_txt[GT_GAME_MENU_SOLDIER_NAME + di->allies_killed[j]]);
               str[1] = AddGameText(str[1], MakeGameText("\n"));
            }
            break;
         }
         str[0] = AddGameText(str[0], MakeGameText("\n"));
      }
      const float size = .0245f;
      desc[1] = texts.CreateText(str[0], -1, x, y+.0525f, size, TEXT_JUSTIFY_RIGHT, wide);
      desc[1]->Release();
      desc[2] = texts.CreateText(str[1], -1, x+.01f, y+.0525f, size, TEXT_JUSTIFY_LEFT, wide);
      desc[2]->Release();
   }
};

//----------------------------
//----------------------------
                              //briefing creation

C_class_to_be_ticked_hd *CreateMenu(E_GAME_MENU id, dword context){

   if(!CDPresenceCheck())
      return NULL;

   C_menu_base1 *mp = NULL;
   switch(id){
   case GAME_MENU_4_OF_8: mp = new C_menu_4_of_8; break;
   case GAME_MENU_8_OF_40: mp = new C_menu_8_of_40; break;
   case GAME_MENU_SOLDIER_INFO: mp = new C_menu_soldier_info; break;
   case GAME_MENU_CAMPAIGN_LOAD: mp = new C_menu_campaign_load; break;
   case GAME_MENU_CAMPAIGN_SELECT: mp = new C_menu_campaign_select; break;
   case GAME_MENU_MISSION_SELECT: mp = new C_menu_mission_select; break;
   //case GAME_MENU_MISSION_STATUS: mp = new C_menu7; break;
   case GAME_MENU_DEBRIEF: mp = new C_menu_debriefing; break;
   default:
      ErrReport(C_fstring("CreateMenu - unknown menu id: %i", id), editor);
   }
   if(mp){
                              //try to initialize created menu
      if(!mp->Init(id, context)){
         delete mp;
         mp = NULL;
      }
   }
   return mp;
}

//----------------------------

